automaton ComputeActionThread_ba =
( ?  
    event Dispatch, Start, Resume;
    boolean inObstacleDetected; 
    integer inObstacleDistance; 
    integer inPreferredSpeed; 
    real inActualSpeed; 
    boolean inConsoleOn;
  ! event outFullBreak; 
    real outSpeedUp; 
    real outSpeedDown;
    event Complete;
    event Error;
    )
  (| inObstacleDetected ^= inObstacleDistance ^= inPreferredSpeed ^= inActualSpeed ^= inConsoleOn   %  inpust are synchronous except the added ones %
   | obstacle_distance ^= previous_obstacle_distance ^= actual_speed ^= previous_actual_speed ^= obstacle_speed ^= previous_obstacle_speed ^= preferred_acceleration ^= preferred_speed ^= outSpeedDown
   | outSpeedDown := outSpeedDown_loc  
   | Automaton_SetClock(% vClkAuto,% Dispatch ^+  Start ^+  Resume )
   | (| Complete :=  t18 ^+ t19 | Error ^= ^0 |)   %  Added Control Signals %
    | (| Automaton_Initial_State(S_Waiting) |)
    | (| (| t0 :: Automaton_Transition(S_Waiting, S_Started, Dispatch )
          | on t0 :: (| actual_speed ::= inActualSpeed |)
          |)
       | (| t1 :: Automaton_Transition(S_Started, S_Detected, [:inObstacleDetected and (actual_speed/=0.0)])
          | on t1 :: (| obstacle_distance ::= inObstacleDistance |)
                       |)
       | (| t2 :: Automaton_Transition(S_Detected, S_Emergency, [:obstacle_distance<UNSAFE_RANGE]) %  NO actions % |)
       | (| t3 :: Automaton_Transition(S_Emergency, S_SaveValues, [: true])
          | outFullBreak ^= t3
          |)
       | (| t4 :: Automaton_Transition(S_Detected, S_NoEmergency, [:obstacle_distance>=UNSAFE_RANGE])
          | on t4 :: (| obstacle_speed ::= (actual_speed-previous_actual_speed)+(real(obstacle_distance-previous_obstacle_distance)/real(PERIOD)) |)
          |)
       | (| t5 :: Automaton_Transition(S_NoEmergency, S_SaveValues, [:obstacle_speed<actual_speed])
          | on t5 :: (| relative_speed := obstacle_speed-actual_speed 
                         | vehicle_acceleration := (actual_speed-previous_actual_speed)/real(PERIOD)
                         | obstacle_acceleration := (obstacle_speed-previous_obstacle_speed)/real(PERIOD)
                         | relative_acceleration := obstacle_acceleration-vehicle_acceleration
                         | xbreak := relative_acceleration-((relative_speed*relative_speed)/real(obstacle_distance ))
                         | outSpeedDown_loc ::= xbreak
                         |) where 
                            %  locals %
                              real relative_speed; 
                              real vehicle_acceleration; 
                              real obstacle_acceleration init 0.0; 
                              real relative_acceleration; 
                              real xbreak; 
                            end
                      |)
       | (| t6 :: Automaton_Transition(S_NoEmergency, S_SaveValues, [:(obstacle_speed>=actual_speed) and (not inConsoleOn)]) %  NO actions % |)
       | (| t7 :: Automaton_Transition(S_NoEmergency, S_ObstacleAndConsole, [:(obstacle_speed>=actual_speed) and inConsoleOn])
          |  on t7 :: (| preferred_speed ::= real(inPreferredSpeed) |)
          |)
       | (| t8 :: Automaton_Transition(S_ObstacleAndConsole, S_SpeedUp, [:obstacle_speed<preferred_speed])
          | on t8 :: (| preferred_acceleration ::= (obstacle_speed-actual_speed)/real(PERIOD) |)
                      |)
       | (| t9 :: Automaton_Transition(S_ObstacleAndConsole, S_Console, [:obstacle_speed>=preferred_speed]) %  No actions % |)
       | (| t10 :: Automaton_Transition(S_Started, S_Console, [:inConsoleOn and ((not inObstacleDetected) or (actual_speed=0.0))])
          | on t10 :: (| preferred_speed ::= real(inPreferredSpeed) |)
          |)
       | (| t11 :: Automaton_Transition(S_Console, S_SpeedUp, [:actual_speed<preferred_speed])
          | on t11 :: (| preferred_acceleration ::= (preferred_speed-actual_speed)/real(PERIOD) |)
                           |)
       | (| t12 :: Automaton_Transition(S_Console, S_SlowDown, [:actual_speed>preferred_speed])
          | on t12 :: (| preferred_acceleration ::= (actual_speed-preferred_speed)/real(PERIOD) |)
          |)
       | t13 :: Automaton_Transition(S_Console, S_SaveValues, [:actual_speed=preferred_speed])
       
       | (| t14 :: Automaton_Transition(S_SpeedUp, S_SaveValues, [:MAX_ACCELERATION<=preferred_acceleration])
          | outSpeedUp := MAX_ACCELERATION when ^t14
          |)
       | (| t15 :: Automaton_Transition(S_SpeedUp, S_SaveValues, [:MAX_ACCELERATION>preferred_acceleration])
          | on t15 :: (| outSpeedDown_loc ::= preferred_acceleration |)
          |)
       | (| t16 :: Automaton_Transition(S_SlowDown, S_SaveValues, [:MAX_DECELERATION<=preferred_acceleration])
          | on t16 :: (| outSpeedDown_loc ::= MAX_DECELERATION |)
          |)
       | (| t17 :: Automaton_Transition(S_SlowDown, S_SaveValues, [:MAX_DECELERATION>preferred_acceleration])
          | on t17 :: (| outSpeedDown_loc ::= preferred_acceleration  |)
          |)
       | (| t18 :: Automaton_Transition(S_SaveValues, S_Waiting, [:true])
          | on t18:: (| previous_actual_speed ::= actual_speed  
                         | previous_obstacle_speed ::= obstacle_speed
                         | previous_obstacle_distance ::= obstacle_distance
                         |)
          |)
       | t19 :: Automaton_Transition(S_Started, S_Waiting, [:(not inConsoleOn) and (not inObstacleDetected)])
       |)
    | %  Default value dor the shared signals %
      (| obstacle_distance ::= defaultvalue obstacle_distance $1 
       | previous_obstacle_distance ::=defaultvalue previous_obstacle_distance $1
       | actual_speed ::=defaultvalue actual_speed $1
       | previous_actual_speed ::= defaultvalue previous_actual_speed $1
       | obstacle_speed ::= defaultvalue obstacle_speed $1
       | previous_obstacle_speed ::=defaultvalue  previous_obstacle_speed $1
       | preferred_acceleration ::=defaultvalue  preferred_acceleration $1
       | preferred_speed ::=defaultvalue preferred_speed$1
       | outSpeedDown_loc ::=defaultvalue outSpeedDown_loc $1
       |)
    |)
  where 
    %  Transitions %
    label t0, t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17, t18, t19; 
    % STATES %
    label S_Waiting, S_Started, S_Detected, S_Emergency, S_NoEmergency, S_SaveValues, S_ObstacleAndConsole, S_Console, S_SlowDown, S_SpeedUp; 
%    event vClkAuto; %

   constant integer UNSAFE_RANGE = 10; 
   constant integer PERIOD = 100; 
   constant real MAX_ACCELERATION = 5.0; 
   constant real MAX_DECELERATION = 2.0; 
   
%   Shared %
   shared integer obstacle_distance; 
   shared integer previous_obstacle_distance init 0; 
   shared real actual_speed; 
   shared real previous_actual_speed init 0.0; 
   shared real obstacle_speed; 
   shared real previous_obstacle_speed init 2.0; 
   shared real preferred_acceleration; 
   shared real preferred_speed; 
   shared real outSpeedDown_loc;
   end; 
