/*
  Authors: Loïc Besnard, François Fabre, Thierry Gautier.
  Copyright (C) by INRIA-CNRS - developed by Espresso project - EPL License - 2010.

  Contact: polychronycontact@listes.irisa.fr

  This file is part of Polychrony SSME Platform (website: http://www.irisa.fr/espresso/polychrony).

  Polychrony SSME Platform is free software; you can redistribute it and/or modify it under
  the terms of the EPL as published by the Eclipse Foundation.

  Polychrony SSME Platform are distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the Eclipse Public License - v 1.0 for more details.

  You should have received a copy of the Eclipse Public License - v 1.0
  along with Polychrony SSME Platform; see the file COPYING. If not, consult the url
  http://www.eclipse.org/org/documents/epl-v10.php, or write to the Eclipse Foundation, Inc.
  102 Centrepointe Drive. Ottawa, Ontario,Canada, K2G 6B1.
*/

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    newMachine
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_newMachine(JNIEnv * env, jobject obj)
{
  setJNIEnv(env);
  vpKServices = newMachine(&OM);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    initMachine
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_initMachine(JNIEnv * env, jobject obj)
{
  setJNIEnv(env);
  vpKServices->pK_initialize(eclipsePlugin);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    resetMachine
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_resetMachine(JNIEnv * env, jobject obj)
{
  setJNIEnv(env);
  vpKServices->reset();
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    displayWarnings
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_displayWarnings(JNIEnv * env, jobject obj, jboolean enable)
{
  setJNIEnv(env);
  // 2 is the value of the Warning mode
  vpKServices->setMessengerOption(2, enable);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_finalize(JNIEnv * env, jobject obj)
{
  setJNIEnv(env);
  vpKServices->finalyze();
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    tree2IGraph
 * Signature: (IZIZLjava/lang/String;Z)Lfr/irisa/espresso/sme/polychrony/GraphValue;
 */
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_tree2DCG(JNIEnv * env, jobject obj, jlong tree, jboolean retTreeError, jlong paramTree, jboolean retParamTreeError, jstring dirName, jboolean genLIS, jboolean bannounce)
{
  setJNIEnv(env);
  ptDCG iGraph = NULL;
  ptDCG_basic gb = NULL;

  if (bannounce) announce_POLYCHRONY();
  char* directory = jstring2CString(dirName, env);
  ntrees mpp = (ntrees) tree;
  if (noper(mpp) == oplmodu) mpp = nchild(1, mpp);
  jint ret = (jint) vpKServices->SIGNALTree2DCG(mpp, (boolean) retTreeError, (ntrees) paramTree, (boolean) retParamTreeError, directory, (boolean) genLIS, (&iGraph));
  return mkReturnGraphValue ( env, ret, (jlong) new DCG_basic(iGraph));
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    signal2SME
 * Signature: (JLjava/lang/String;)I
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_signal2SME__JLjava_lang_String_2(JNIEnv * env, jobject obj, jlong tree, jstring dirName)
{
  int res;
  setJNIEnv(env);
  char* directory = jstring2CString(dirName, env);
  res = vpKServices->unparse_SME((ntrees) tree, directory);
  return (jboolean)(res == 0);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    signal2SME
 * Signature: (Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_signal2SME__Ljava_lang_String_2Ljava_lang_String_2(JNIEnv * env, jobject obj, jstring srcFile, jstring dirName)
{
  int res;	
  setJNIEnv(env);
  char* file = jstring2CString(srcFile, env);
  char* directory = jstring2CString(dirName, env);
  res = vpKServices->unparse_parse2SME(file, directory);
  return (jboolean)(res == 0);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    signal2SSME
 * Signature: (JLjava/lang/String;)I
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_signal2SSME__JLjava_lang_String_2(JNIEnv * env, jobject obj, jlong tree, jstring dirName)
{
  int res;
  ptrepstr enteringOutputDir;
  setJNIEnv(env);
  char* directory = jstring2CString(dirName, env);
  enteringOutputDir = sys_get_userOutputDir();
  sys_set_userOutputDir(string2repstr(directory));
  vpKServices->signalTree2SSME((ntrees) tree);
  sys_restore_userOutputDir(enteringOutputDir);
  return (jboolean)(pk_true);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    signal2SSSME
 * Signature: (Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_signal2SSME__Ljava_lang_String_2Ljava_lang_String_2(JNIEnv * env, jobject obj, jstring srcFile, jstring dirName)
{
  int res;
  setJNIEnv(env);
  char* file = jstring2CString(srcFile, env);
  char* directory = jstring2CString(dirName, env);
  res = vpKServices->signal2SSME(file, directory);
  return (jboolean)(res == 0);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2C
 * Signature: (IZZZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_genCodeC(JNIEnv * env, jobject obj, jlong graphNode, jboolean mainFile, jboolean bodyFile, jboolean ioFile)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->genCodeC( (ptDCG_sch) graphNode, 1);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2CPP
 * Signature: (IZZZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_genCodeCPP(JNIEnv * env, jobject obj, jlong graphNode, jboolean mainFile, jboolean bodyFile, jboolean ioFile)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->genCodeCPP( (ptDCG_sch) graphNode, 1);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2CPPSeparate
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_genCodeCPPSeparate(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->genCodeCPP( (ptDCG_sch) graphNode, 3);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2CSeparate
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_genCodeCSeparate(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->genCodeC( (ptDCG_sch) graphNode, 3);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2DotCode
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_IGraph2DotCode(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return pk_true;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2Java
 * Signature: (IZZZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_genCodeJava(JNIEnv * env, jobject obj, jlong graphNode, jboolean mainFile, jboolean bodyFile, jboolean ioFile)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->genCodeJava( (ptDCG_sch) graphNode, 1);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2JavaSeparate
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_genCodeJavaSeparate(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->genCodeJava( (ptDCG_sch) graphNode, 3);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2threadsJava
 * Signature: (IZZZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_genCodeJavaThreads(JNIEnv * env, jobject obj, jlong graphNode, jboolean mainFile, jboolean bodyFile, jboolean ioFile)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->genCodeJava( (ptDCG_sch) graphNode, 2); 
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2Kernel
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_expressionExpand(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->expressionExpand((ptDCG_basic) graphNode);
}


/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2Signal
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_unparse(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->unparse((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2SigTree
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_produce(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jlong) vpKServices->produce_tree((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2SigTreeFlat
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_produceFlat(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jlong) vpKServices->produceFlat((ptDCG_basic) graphNode, pk_false);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2SynDEx
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_SynDExUnparse(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->export_SynDEx((ptDCG_flat) graphNode, pk_true);
}


/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraph2Z3z
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_SigaliPrintExtraction(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->SigaliPrintExtraction((ptDCG_basic) graphNode, false);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphAbstraction2SigTree
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_produceInterfaceAbstraction(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jlong) vpKServices->produceInterfaceAbstraction((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphAbstractionGraph
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_buildAbstraction(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  ptDCG_poly * gResult = 0;
  if(vpKServices->buildAbstraction((ptDCG_poly) graphNode, gResult)) return (jlong) gResult;
  return (jlong) ERROR_NODE;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphAggregatingMultiplexing
 * Signature: (IZ)V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_memoryOptimizations(JNIEnv * env, jobject obj, jlong graphNode, jboolean delayMultiplexing)
{
  setJNIEnv(env);
  vpKServices->memoryOptimizations((ptDCG_poly)graphNode, delayMultiplexing);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphBooleans2Events
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_transformBooleanClocks2Events(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->transformBooleanClocks2Events((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphBooleansOthersPartitionning
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_separateBooleansOthers(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  vpKServices->separateBooleansOthers((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphClockEnvBuilding
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_setIntrinsicClocks(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->setIntrinsicClocks((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphCodeGenerationReinforcing
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_dependenceReinforcing(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->dependenceReinforcing((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphConstraintsAsExceptions
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_addExceptionsForConstraints(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->addExceptionsForConstraints((ptDCG_bool) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphCycleDetection
 * Signature: (IZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_hasStaticCycle(JNIEnv * env, jobject obj, jlong graphNode, jboolean debug)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->isStaticCycleFree((ptDCG_poly) graphNode, pk_true, debug);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphDefaultSpliting
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_splitDefaultDefinitions(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->splitDefaultDefinitions((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphDelaysMemoryOptimisation
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_setDelayMemoryOverloading(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->setDelayMemoryOverloading((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphDownStreamRetiming
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_moveIn2OutDelays(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->moveIn2OutDelays((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    n_IGraphExclusiveClocks
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_computeClockExclusions(JNIEnv * env, jobject obj, jlong graphNode, jint ecLevel)
{
  setJNIEnv(env);
  vpKServices->computeClockExclusions((ptDCG_poly)graphNode, (integer) ecLevel);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphExpand
 * Signature: (IZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_expandModelReference(JNIEnv * env, jobject obj, jlong graphNode, jboolean asSubGraph)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->expandModelReference((ptDCG_basic) graphNode, asSubGraph);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphFlatten
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_breakCPG(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->breakCPG((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphGenCodeVerifs
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_checkExplicitTyping(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->checkExplicitTyping((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphHierarchyStatistics
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_printCHStatistics(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->printCHStatistics((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphInputsTrain
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_phylumClustering(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->phylumClustering((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphInterfaceCalculus
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_buildClockProcessInterface(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->buildClockProcessInterface((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphIsWithConstraints
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_hasUnsolvedClockConstraints(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->hasUnsolvedClockConstraints((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphMakeTreeDelays
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_delayTreeBuild(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->delayTreeBuild((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphPartialDefinitionsMerging
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_groupPartialDefinitions(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->groupPartialDefinitions((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphProfiling
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_applyTabularMorphism(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->applyTabularMorphism((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphProjection
 * Signature: (IZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_buildIOClockSubHierarchy(JNIEnv * env, jobject obj, jlong graphNode, jboolean clockOfDependences)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->buildIOClockSubHierarchy((ptDCG_poly) graphNode, clockOfDependences);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphPruning
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_nonProductiveSignalElimination(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->nonProductiveSignalElimination((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphRenamesEliminating
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_removeSignalEqualities(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->removeSignalEqualities((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSeparatedCompiling
 * Signature: (IZ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_separatedCompilingPoly(JNIEnv * env, jobject obj, jlong graphNode, jboolean genCode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->InterfaceAbstraction((ptDCG_poly) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSequentializing
 * Signature: (IZZZ)Z
 */ 
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_Serializing(JNIEnv * env, jobject obj, jlong graphNode, jboolean btotalOrder, jboolean balwaysCode, jboolean bcheck, jboolean bdelayMultiplexing)
{
  boolean b;
  int ret;
  ptDCG_sch gSeq = NULL;

  setJNIEnv(env);
  if (vpKServices->Serializing((ptDCG_bool) graphNode,  btotalOrder, balwaysCode, bcheck, bdelayMultiplexing, &gSeq)) ret=0; else ret=1;
  pK_messenger->announce_SCHED();
  return mkReturnGraphValue(env, ret, (jlong) gSeq);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSetDependences
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_setClockDependences(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->setClockDependences((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSIGNALAbstractionPrinting
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_unparseAbstraction(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  vpKServices->unparseAbstraction((ptDCG_poly) graphNode);
  return (jboolean)pk_true;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSplitingStructure
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_splitDataStructure(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->splitDataStructure((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphStatesOthersPartitionning
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_separateStatesOthers(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  vpKServices->separateStatesOthers( (ptDCG_poly)graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphStateVarUtilityClock
 * Signature: (I)Z
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_computeProductiveClockStateVar(JNIEnv * env, jobject obj, jlong graphNode)
{
  /*  setJNIEnv(env);
      vpKServices->computeProductiveClockStateVar((ptDCG_poly) graphNode);
  */
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSubGraph2SigTree
 * Signature: (II)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_produceSubgraph(JNIEnv * env, jobject obj, jlong graphNode, jlong subGraph)
{
  setJNIEnv(env);
  return (jlong) vpKServices->produceSubgraph((ptDCG_basic) graphNode, (ptINode) subGraph);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSubstitute
 * Signature: (II)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_substituteExpr2Var(JNIEnv * env, jobject obj, jlong graphNode, jint maxRef)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->substituteExpr2Var((ptDCG_basic) graphNode, maxRef);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphTransitiveClosure
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_computeIOTransitiveClosure(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->computeIOTransitiveClosure((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphUnifyingNodes
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_removeSignalClones(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->removeSignalClones((ptDCG_basic) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphUtilityClock
 * Signature: (I)Z
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_computeProductiveClock(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  vpKServices->computeProductiveClock((ptDCG_poly) graphNode);
}


/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    Signal2Lustre
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_LustreUnparse(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->export_Lustre((ptDCG_bool) graphNode);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    parse
 * Signature: (Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_parse(JNIEnv * env, jobject obj, jstring src, jstring param)
{
  setJNIEnv(env);
  char* paramfile= jstring2CString(param, env);
  char *srcfile = jstring2CString(src, env);

  int n = vpKServices->SIGNALFiles2DCG(srcfile, paramfile, (char*) ".", pk_false, &vgraph);
  pK_messenger->gmessload(mkfgmess);
  if(n == 0) return (jlong) vgraph;
  return (jlong) ERROR_NODE;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    printSignalLIS
 * Signature: (JJ)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_printSignalLIS
(JNIEnv * env, jobject obj, jlong tree, jlong params)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->unparse_SignalLis((ntrees)tree,(ntrees)params, pk_false); /* standard announce */
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    setEnv
 * Signature: (Ljava/lang/String;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_setEnv(JNIEnv * env, jobject obj, jstring key, jstring value)
{
  setJNIEnv(env);
  char *cKey = jstring2CString(key, env);
  char *cValue = jstring2CString(value, env);
  /* Modified DEC 2015 LBesnard. 
     allows to use the SignalToolBox by adding the address under Eclipse, and the polychrony GUI as external editor for a gpk file.
     // old code : if(vpKServices != NULL)  { vpKServices->setEnv(cKey, cValue); } 
     */
#if (xKHOST != xKwinNT)
  int err = setenv(cKey, cValue, true);  // setenv is not standard.
#else
  ptrepstr r1,r2;
  r1 = string2repstr (cKey);
  strconcchastr('=', string2repstr (cValue), &r2); // =xxxx 
  strconc2str(r1, r2, pk_false, &r1);
  _putenv(repstr2string (r1));
#endif
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    getEnv
 * Signature: (Ljava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_getEnv(JNIEnv * env, jobject obj, jstring key)
{
  setJNIEnv(env);
  char *cKey = jstring2CString(key, env);
  char *cValue = NULL;
  
  if(vpKServices != NULL) cValue = vpKServices->getEnv(cKey);
  if(cValue == NULL) return (jstring) NULL;
  return (jstring) env->NewStringUTF(cValue);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphBooleanizing
 * Signature: (I)Z
 */
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_DCGPoly2DCGBool(JNIEnv * env, jobject obj, jlong graphNode)
{
  ptDCG_bool gBool=NULL;
  jint ret;

  setJNIEnv(env);
  // if (! vpKServices->DCGPoly2DCGBool((ptDCG_poly) graphNode, &gBool, OM->req_traceALL())) ret = 1; else ret=0;
  if (DCGPoly2DCGBool((ptDCG_poly) graphNode, &gBool, OM->req_traceALL())) ret=0; else ret=1;
  return mkReturnGraphValue(env, ret, (jlong) gBool);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphSTSFlattening
 * Signature: (I)Z
 */
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_DCGBool2DCGFlat(JNIEnv * env, jobject obj, jlong graphNode)
{
  // ptDCG_bool gBool=NULL;
  ptDCG_flat gf=NULL;
  jint ret;

  setJNIEnv(env);
  ret=0;
  //  if (!DCGPoly2DCGBool((ptDCG_poly) graphNode, &gBool, OM->req_traceALL()))  //    ret=1;   // else
  {
    if (! vpKServices->DCGBool2DCGFlat((ptDCG_bool) graphNode, &gf)) ret=1;
    pK_messenger->announce_FLAT();
  }
  return mkReturnGraphValue(env, ret, (jlong) gf);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    IGraphClockCalculus
 * Signature: (I)I
 */
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_DCGBasic2DCGPoly(JNIEnv * env, jobject obj, jlong graphNode)
{
  ptDCG_poly gp;
  jint ret=0;

  setJNIEnv(env);
  if (!vpKServices->DCGBasic2DCGPoly((ptDCG_basic) graphNode, 5, &gp)) ret =1;
  pK_messenger->announce_POLY();
  return mkReturnGraphValue(env, ret, (jlong) gp);

}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_PKServices
 * Method:    DCG2dotCode
 * Signature: (J)Z
 */
JNIEXPORT jboolean JNICALL Java_fr_inria_tea_ssme_polychrony_PKServices_DotUnparse(JNIEnv * env, jobject obj, jlong graphNode)
{
  setJNIEnv(env);
  return (jboolean) vpKServices->export_Dot((ptDCG_basic) graphNode);
}
