/*
   Authors: Loïc Besnard, François Fabre, Thierry Gautier.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - EPL License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony SSME Platform (website: http://www.irisa.fr/espresso/polychrony).

   Polychrony SSME Platform is free software; you can redistribute it and/or modify it under
   the terms of the EPL as published by the Eclipse Foundation.

   Polychrony SSME Platform are distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the Eclipse Public License - v 1.0 for more details.

   You should have received a copy of the Eclipse Public License - v 1.0
   along with Polychrony SSME Platform; see the file COPYING. If not, consult the url
   http://www.eclipse.org/org/documents/epl-v10.php, or write to the Eclipse Foundation, Inc.
   102 Centrepointe Drive. Ottawa, Ontario,Canada, K2G 6B1.
*/

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    init
 * Signature: ()V
 */

#include <cstdlib>

JNIEXPORT void JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_init(JNIEnv * env, jobject obj)
{
  setJNIEnv(env);
  parserFileName = sysGetnameTempfile();
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseParamfile
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseParamfile(JNIEnv * env, jobject obj, jstring filename)
{
	setJNIEnv(env);
	ntrees result;
	if(SIGNALParseParamfile( jstring2CString(filename, env), &result)) return (jlong) result;
	pK_messenger->parsing_logFilePrint(nilstr);// it prints the errors to the Eclipse Console.
	return (jlong) ERROR_NODE;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseSrcfile
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseSrcfile(JNIEnv * env, jobject obj, jstring filename)
{
  ntrees result;
  boolean b;
  setJNIEnv(env);
  char * vfile;

  announce_POLYCHRONY();
  pK_messenger->announce_PARSING();
  vfile= jstring2CString(filename, env);
  if(SIGNALParseSrcfile( vfile, &result, &b)) return (jlong) result;
  pK_messenger->parsing_logFileSave(vfile);
  pK_messenger->parsing_logFilePrint(nilstr);// it prints the errors to the Eclipse Console.
  return (jlong) ERROR_NODE;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseString
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseString(JNIEnv * env, jobject obj, jstring expression)
{
  return parse(env, obj, expression, ph_cst_cha);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseIntegerConstant
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseIntegerConstant(JNIEnv * env, jobject obj, jstring expression)
{
  return parse(env, obj, expression, ph_e_cst);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseDoubleReal
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseDoubleReal(JNIEnv * env, jobject obj, jstring expression)
{
  return parse(env, obj, expression, ph_e_cst);
}


/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseChar
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseReal(JNIEnv * env, jobject obj, jstring expression)
{
  return parse(env, obj, expression, ph_e_cst);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseChar
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseChar(JNIEnv * env, jobject obj, jstring expression)
{
  return parse(env, obj, expression, ph_e_cst);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseChar
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseIdent(JNIEnv * env, jobject obj, jstring expression)
{
  return parse(env, obj, expression, ph_ident);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalParser
 * Method:    parseChar
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalParser_parseOperator(JNIEnv * env, jobject obj, jstring expression)
{
  return parse(env, obj, expression,  ph_ident_op);
}
