/*
   Authors: Loïc Besnard, François Fabre, Thierry Gautier.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - EPL License - 2010.

   Contact: polychronycontact@listes.irisa.fr

   This file is part of Polychrony SSME Platform (website: http://www.irisa.fr/espresso/polychrony).

   Polychrony SSME Platform is free software; you can redistribute it and/or modify it under
   the terms of the EPL as published by the Eclipse Foundation.

   Polychrony SSME Platform are distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the Eclipse Public License - v 1.0 for more details.

   You should have received a copy of the Eclipse Public License - v 1.0
   along with Polychrony SSME Platform; see the file COPYING. If not, consult the url
   http://www.eclipse.org/org/documents/epl-v10.php, or write to the Eclipse Foundation, Inc.
   102 Centrepointe Drive. Ottawa, Ontario,Canada, K2G 6B1.
*/

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getBody
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getBody(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STreeModel_getBody((ntrees) node);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getCallToIn
 * Signature: (II)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getCallToIn(JNIEnv * env, jobject obj, jlong name, jlong node)
{
	ntrees *aCall = NULL;
	if(STreeModel_getCallToIn((ptident)name, (ntrees) node, aCall) )
		return (jlong) aCall;
	else
		return (jlong) ERROR_NODE;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    n_getDeclaration
 * Signature: (II)Lfr/irisa/espresso/ssme/polychrony/DeclarationValue;
 */
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_n_1getDeclaration(JNIEnv * env, jobject obj, jlong node, jint kind)
{
    int iDecl = 0;
	jlong ret = (jlong) STree_getDeclaration((ntrees) node, kind, &iDecl);
		
	jclass cls = env->FindClass("fr/irisa/espresso/ssme/polychrony/DeclarationValue");
	if (cls == NULL)
         return NULL; /* exception thrown */
	
	/* Get the method ID for the DeclarationValue constructor */
	jmethodID cid = env->GetMethodID(cls, "<init>", "(IL)V");
	if (cid == NULL)
	  return NULL; /* exception thrown */
	
	/* Construct a DeclarationValue object */
	jobject result = env->NewObject(cls, cid, ret, (jlong) iDecl);
	
	/* Free local references */
	env->DeleteLocalRef(cls);
	return result;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getDeclarationWithoutType
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getDeclarationWithoutType(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong) STree_getDeclarationWithoutType ((ntrees) node);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getDescription
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getDescription(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getDescription((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getExternalValue
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getExternalValue(JNIEnv * env, jobject obj, jlong node)
{
	ptrepstr str = STree_getExternalValue((ntrees) node);
	return str2jstring(str, env);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getIdentStr
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getIdentStr(JNIEnv * env, jobject obj, jlong node)
{
	ptrepstr str = STree_getIdentStr((ntrees) node);
	return str2jstring(str, env);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getInitExpr
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getInitExpr(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getInitExpr((ntrees) node);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getInputs
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getInputs(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getInputs((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getInputsOutputs
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getInputsOutputs(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getInputsOutputs((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getInterface
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getInterface(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getInterface((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getIthModel
 * Signature: (II)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getIthModel(JNIEnv * env, jobject obj, jlong node, jint rank)
{
	return (jlong)STree_getIthModel((ntrees) node, (int) rank);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getLocalDeclarations
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getLocalDeclarations(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong) STreeModel_getLocalDeclarations((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getModelFather
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getModelFather(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getModelFather((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getModelInWhich
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getModelInWhich(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getModelInWhich((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getModelName
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getModelName(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STreeModelModule_getName((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getModelNameStr
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getModelNameStr(JNIEnv * env, jobject obj, jlong node)
{
	ptrepstr str = STreeModelModule_getNameStr((ntrees) node);
	return str2jstring(str, env);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getNameCalledModel
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getNameCalledModel(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getNameCalledModel((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getNameOfDeclaration
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getNameOfDeclaration(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getNameOfDeclaration((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getNbInputs
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getNbInputs(JNIEnv * env, jobject obj, jlong node)
{
  return (jint)STree_getNbInputs((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getNbOutputs
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getNbOutputs(JNIEnv * env, jobject obj, jlong node)
{
	return (jint)STree_getNbOutputs((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getNbParameters
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getNbParameters(JNIEnv * env, jobject obj, jlong node)
{
	return (jint)STree_getNbParameters((ntrees) node);	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getNilTree
 * Signature: ()I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getNilTree(JNIEnv * env, jobject obj)
{
	return (jlong)STree_getNilTree();	
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getOperatorArity
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getOperatorArity(JNIEnv * env, jobject obj, jint codeOp)
{
	return (jint)STree_getOperatorArity((noperators) codeOp);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getOutputs
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getOutputs(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getOutputs((ntrees) node);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getParameters
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getParameters(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getParameters((ntrees) node);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getPragmas
 * Signature: (L)L
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getPragmas(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getPragmas((ntrees) node);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getPragmaValues
 * Signature: (I)Lfr/irisa/espresso/ssme/polychrony/PragmaValues;
 */
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getPragmaValues(JNIEnv * env, jobject obj, jlong pragmaNode)
{
	ptrepstr * name = NULL;
	ntrees * objList = NULL;
	ntrees * enonce = NULL;
		
	STree_getPragmaValues((ntrees) pragmaNode, name, objList, enonce);
	
	jclass cls = env->FindClass("fr/irisa/espresso/ssme/polychrony/PragmaValues");
	if (cls == NULL)
         return NULL; /* exception thrown */
	
	/* Get the method ID for the PragmaValues constructor */
    jmethodID cid = env->GetMethodID(cls, "<init>", "(Ljava/lang/String;JJ)V");
    if (cid == NULL)
		return NULL; /* exception thrown */
 
	/* Construct a PragmaValues object */
	jobject result = env->NewObject(cls, cid, str2jstring((*name), env), (jlong) objList, (jlong) enonce);
 
	/* Free local references */
	env->DeleteLocalRef(cls);
	return result;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getRootTree
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getRootTree(JNIEnv * env, jobject obj, jlong node)
{
	ntrees *res = NULL;
	if(STree_getRootTree((ntrees) node, res))
		return (jlong) *res;
	else
		return (jlong) ERROR_NODE;
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getSpec
 * Signature: (I)I
 */
JNIEXPORT jlong JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getSpec(JNIEnv * env, jobject obj, jlong node)
{
	return (jlong)STree_getSpec((ntrees) node);
}

/*
 * Class:     fr_irisa_espresso_ssme_polychrony_SignalTreeExploring
 * Method:    getType
 * Signature: (L)Lfr/irisa/espresso/ssme/polychrony/TypeValue;
 */
JNIEXPORT jobject JNICALL Java_fr_inria_tea_ssme_polychrony_SignalTreeExploring_getType(JNIEnv * env, jobject obj, jlong node)
{
	ntrees * treeType = NULL;
		
	jlong ret = (jlong) STreeDecl_getType((ntrees) node, treeType);  
	
	jclass cls = env->FindClass("fr/irisa/espresso/ssme/polychrony/TypeValue");
	if (cls == NULL)
         return NULL; /* exception thrown */
	
	/* Get the method ID for the TypeValue constructor */
    jmethodID cid = env->GetMethodID(cls, "<init>", "(IL)V");
    if (cid == NULL)
		return NULL; /* exception thrown */
 
	/* Construct a TypeValue object */
	jobject result = env->NewObject(cls, cid, ret, (jlong) treeType);
 
	/* Free local references */
	env->DeleteLocalRef(cls);
	return result;
}

 
