/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_

// #include <iostream> 

/**
   Closure calculus for a node (node) in the current mode (the successors are not informed).
*/
void
ClosureCalculus::nodeClosure (ptINode node)
{
  switch (NM->get_oper (node))
    {
    case opdefpart:
    case opdefsig:
      equClosure (node);
      break;

    case opmoti:
      arrayClosure (node);
      break;

    case opret:
      delayClosure (node);
      break;

    case opappe:
      procCallClosure (node);
      break;

    case opsubgraph:
    case opldefpart:
    case opbund:		// Symbolise struct/bundle
    case opnodecall:		// Rien a faire : traitement reportés sur les sorties du "nodeCall"
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}

/**
    If the clock (clk) has not been explored, it is marked and 
       - its successors are visited (recursively)
       - then, the closure calculus is applied for the clock (clockClosure() method)  
*/
_VIRTUAL_ void
ClosureCalculus::visitClock (ptinsth clk)
{
  if (clk == CKM->getRepresentativeFormula (clk))
    {
      clockClosure (clk);
      //      std::cout << " VisitClock = " << CKM->getNum(clk) << std::endl;
      if (! CKM->satisfiesMarked (clk, vmarkexplored))
	{
	  if (! CKM->satisfiesMarked(clk, vmarkpushed))
	    {
	      CKM->addMark (clk, vmarkpushed);
	      visitReferences(clk);
	    }
	  CKM->removeMark (clk, vmarkpushed);
	}
    }
}

/**
   If a node (node) has not been explored, it is marked and 
   - its successors are visited (recursively)
   - then, the closure calculus is applied for the node (nodeClosure() method).
*/
_VIRTUAL_ void
ClosureCalculus::visitNode (ptINode node)
{
  ptINode imoti;

  nodeClosure (node);
  //  std::cout << " VisitNode = " << NM->get_num(node) << std::endl;
  if (! NM->isMarked (node, vmarkexplored))
    {
      if (!NM->isMarked (node, vmarkpushed))
	{
	  NM->addMark (node, vmarkpushed);
	  visitReferences(node);

	  if (NM->is_partialDefinition (node))
	    {
	      imoti = EM->getNode (NM->get_leftOperand (node));
	      visitNode (imoti);
	    }
	  else
	    imoti = node;
	  while (NM->get_inComposedNode (imoti, &imoti))
	    visitNode (imoti);
  	}
      NM->removeMark (node, vmarkpushed);
    }
}


/**
   A node (node) is marked explored if all its predecessors are marked as explored.
*/
_VIRTUAL_ void ClosureCalculus::markExplored(ptINode node)
{
  ptldependence lpred;
  ptdependence aDep;
  ptinsth hdep;
  ptINode n1, n2;

  boolean allvisited = pk_true;
  lpred = NM->get_predecessors(node);
  while (allvisited && lpred->get_next(&aDep, &lpred))
    {
      hdep = aDep->getDependence(&n1, &n2); // n2 is node.
      if (! NM->isMarked (n1, vmarkexplored)) allvisited = pk_false;
    }
  if (allvisited) NM->addMark (node, vmarkexplored);
}


