/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   \copydoc CircuitCalculus::CircuitCalculus
*/
ClosureCalculus::ClosureCalculus (ptDCG g):CircuitCalculus (g)
{
  lexpclos = NILLINSTANCES;
  lhclos = NILLINSTH;
}



/**
  This function computes the closure for the signals (lsig) of a graph (parameter of the constructor).
  The following rules are applied (closure calculus definition)
   - for two dependences n2 -> n1 when H and n1 -> n when K then n2 -> n when H^*K.
   - for two dependences m -> n when H and m -> n when K, then m -> n when K^+ K.

  The result of the calculus is available as an attribute associated with each signal of the parameter list (lsig).
  This attribute is a global one (see CompilerAttributes.h file).
  The list lsig is splitted into two lists:  the list of the events (levents) and the others (lexpclos).\n  
  \b lexpclos and \b levents are attributes members of the class.
*/
boolean
ClosureCalculus::transitiveClosure (ptlinstances lsig)
{
  ptlinstances levents;
  boolean bres;

  //  CKM->setSimplifyMode(pk_true);
  closureInitialize (lsig);
  PARC->initialize ();
  lexpclos = NILLINSTANCES;
  levents = NILLINSTANCES;
  EM->splitEvents (lsig, &lexpclos, &levents);
  lhclos = EM->getClockOfSignals (levents);
  
  graphTravellingInitialClock ();
  bres= pk_false;
  if (PARC->initTravelForClock (AGRAPH->getRootClock ()))
    {
      graphTravelling (NILINODE);
      if (existLockedNodes ())
	{
	  addInputsInReached(lsig);
	  visitReached ();
	}
      importTransitiveClosure (lsig);
      // importTransitiveClosure (levents); // added Oct 2019 TEST
      bres= pk_true;
    }
  PARC->finalize ();
  //  CKM->setSimplifyMode(pk_false);
  return bres;
}
