/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   Generating of the test for a clock (h).
*/
  _VIRTUAL_ void
cGenerator::genClockCondition (ptinsth h, boolean reqIf)
{
  if (reqIf) TextForm->printkwif ();
  TextForm->printLP ();
  genClockExpression (h, opnullary);
  if (sameLine)
    {
      TextForm->printRP ();
      TextForm->printSpace ();
    }
  else
    TextForm->printRPlineindent ();
  okForElse = pk_true;
}

/**
   Generating of N closing blocks where N is defined by the relative depth of a clock (h) to another one (vhemb).
   @param h  (in) clock that must be in the sub-tree with vhemb as root.
   @param vhemb  (in) a clock.
*/
_VIRTUAL_ void
cGenerator::genCloseConditional (ptinsth h, ptinsth vhemb)
{
  while (h != vhemb)
    {
      TextForm->printblockIfClosing(); 
      h = CKM->getFather (h);
    }
  setCurrentContextClockGenerator (vhemb);
}

/** 
    Generating  N opening blocks, where N is defined by the relative depth of h parameter to vhemb parameter.
    Opening a block induces the code generating of the test of the clock (genClockCondition()).
    @param h, vhemb  (in) 2 clocks ; h must be in the sub-tree with vhemb as root.
*/
_VIRTUAL_ void
cGenerator::genOpenConditional (ptinsth h, ptinsth vhemb)
{
  if (h != vhemb)
    {
      genOpenConditional (CKM->getFather (h), vhemb);
      genClockCondition (h, pk_true);
      TextForm->printkwthen ();
      TextForm->OpenBlock();
      setCurrentContextClockGenerator (h);
    }

}

/** 
    Generating  the code of the initializing of clocks to "false" ("bottom value").
    @param aNode (in) a node; for each output signals of the node defining a boolean clock, a initiliazing is generated.
*/
_VIRTUAL_ void
cGenerator::set_booleanClocks_undefined (ptINode aNode)
{
  ptlinstances lsig;
  ptexpcal aSig;
  ptinsth h;

  lsig = NM->getInodelOutputFlows (aNode);
  while (ptlinstances_get_next (&lsig, &aSig))
    if (EM->getExtractionLink (aSig, &h))
      if (CKM->getRepresentativeFormula (h) == h)
	if (!CKM->satisfiesMarked (h, alreadyVisisted))
	  {
	    set_booleanClocks_undefined (h);
	    CKM->addMark (h, alreadyVisisted);	// to avoid  code duplication.
	  }
}


/** 
    Generating  the code of the initializing clocks to "false" ("bottom value") 
    For each signal of a list (lsig) that defines a boolean clock, 
    an initialisation to "false" (bottom) is generated 
    - if it is a useful clock  and
    - and if it is not in the list of the clocks that must not be initialized (lClocksNotToBeInitiliazed attribute member).
    - and if it is not synchronous to clock (when a boolean signal, used as a clock, is local to the sub-tree with ^x as root, it is not useful to initialize it to
    bottom).
*/
_VIRTUAL_ void
cGenerator::setUndefinedValueForBooleanList (ptinsth clock, ptlinstances lsig)
{
  ptexpcal aSig;
  ptINode node;
  ptinsth h, hsig;

  while (ptlinstances_get_next (&lsig, &aSig))
    {
      node = EM->getNode (aSig);
      if (mustBeGenerated (node))
	if (EM->getExtractionLink (aSig, &h))
	  if (!ptlinstances_isin_exp (lClocksNotToBeInitiliazed, aSig))
	    {
	      aSig = EM->getAggregation (aSig);
	      hsig = EM->getClock (aSig);
	      if ((hsig != clock) && !ptlinstances_isin_exp (lClocksNotToBeInitiliazed, aSig))
		genBottomValue (aSig);
	    }
    }
}

/** 
    Generating the code of the initializing clocks to "false" ("bottom value") from the local signals of a clock (clock).
    It calls setUndefinedValueForBooleanList() method.
*/
_VIRTUAL_ void
cGenerator::set_booleanClocks_undefined (ptinsth clock)
{
  setUndefinedValueForBooleanList (clock, ptinterface_get_localSignals (CKM->getInterfaceSubHierarchy (clock)));
}


/** 
    This method sets the context clock (currentClock attribute member) to the clock h.
*/
_VIRTUAL_ void
cGenerator::initContextGenerator (ptinsth h)
{
  setCurrentContextClockGenerator (h);
  okForElse = pk_false;
}

_VIRTUAL_ void
cGenerator::setCurrentContextClockGenerator (ptinsth h)
{
  currentClock = h;
}

/**
   @return "true" is a clock (h) is the "esle" part of the currentClock, "false" otherwise.
   vhemb is assumed to be the branching clock of currentClock and h.
*/
boolean
cGenerator::testElseBlock (ptinsth currentClock, ptinsth h, ptinsth vhemb)
{
  ptinsth vhaux;
  ptexpcal vcondCurrent, vcond, vdef;

  return pk_false;

  if (CKM->eqClock (h, currentClock))
    return pk_false;
  if (!okForElse)
    return pk_false;
  vhaux = CKM->getFather (h);
  if (vhaux != CKM->getFather (currentClock))
    return pk_false;
  if (vhaux != vhemb)
    return pk_false;
  if (CKM->getComplExtraction (h, &vhaux))
    if (CKM->eqClock (vhaux, currentClock))
      return pk_true;
  if (!CKM->satisfiesSub (currentClock))
    return pk_false;
  if (!CKM->satisfiesSub (h))
    return pk_false;
  vcondCurrent = CKM->getConditionSub (currentClock);
  vcondCurrent = getExpressionToGenerate (vcondCurrent);
  vcond = CKM->getConditionSub (h);
  vcond = getExpressionToGenerate (vcond);
  if (EM->getSyncExprDefSig (vcond, &vdef) && (EM->getOperator (vdef) == opnon))
    {
      vdef = EM->getUnaryArg (vdef);
      vdef = getExpressionToGenerate (vdef);
      return (vdef == vcondCurrent);
    }
  if (EM->getSyncExprDefSig (vcondCurrent, &vdef) && (EM->getOperator (vdef) == opnon))
    {
      vdef = EM->getUnaryArg (vdef);
      vdef = getExpressionToGenerate (vdef);
      return (vdef == vcond);
    }
  return pk_false;
}

/** 
    Generating a set of block closings/block openings to have a clock (h) as the current one.
    The "currentClock" member is set to this clock.
    It calls genCloseConditional(), genOpenConditional().
*/
_VIRTUAL_ void
cGenerator::genBlockOpeningOf (ptinsth h)
{
  ptinsth vhemb;
  boolean b;

  vhemb = CKM->getBranching (h, currentClock);
  if (vhemb == h)
    genCloseConditional (currentClock, vhemb);
  else if (vhemb == currentClock)
    genOpenConditional (h, vhemb);
  else
    {
      b = testElseBlock (currentClock, h, vhemb);
      genCloseConditional (currentClock, vhemb);
      // "if (c) {statements} else {statements}" instead of "if (c) {statements}; if (!c) {statements}"
      if (b)
	{
	  TextForm->printkwelse ();
	  TextForm->printOpenBlockIndent ();
	  okForElse = pk_false;
	}
      else
	genOpenConditional (h, vhemb);
    }
  setCurrentContextClockGenerator (h);
}


/**
   Generating the definition of a clock (h).
   @param h  (in) clock which expression definition must be generated,
   @param opparent (in) operator giving the context to put expression in brackets.
*/
_VIRTUAL_ void
cGenerator::genClockExpression (ptinsth h, noperators opparent)
{
  ptexpcal cond;
  if (CKM->satisfiesNullClock (h))	// no occurrence, but kept
    {
      TextForm->printLP ();
      TextForm->printkwlogicons (pk_false);
      TextForm->printRP ();
    }
  else
    {
      cond = CKM->getConditionSub (h);
      if (CKM->getOperator (h) == nosub)
	TextForm->printkwnot ();
      genexpression (cond, opparent);
    }
}
