/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_ 

_VIRTUAL_ boolean cGeneratorDist::getInputNodes (ptlINode * ln)
{
  ptlINode l;
  ptINode n;

  (*ln) = NILLINODE;
  if (NM->getInputs (&l))
    while (ptlINode_get_elem (&l, &n))
      if (ptlinstances_isin_exp (Inputs_env, NM->get_io_arg (n)))
	ptlINode_insert_elem (ln, n);
  return ((*ln) != NILLINODE);
}

_VIRTUAL_
  boolean
cGeneratorDist::getOutputNodes (ptlINode * ln)
{
  ptlINode l;
  ptINode n;
  ptexpcal aSig;

  (*ln) = NILLINODE;
  if (NM->getOutputs (&l))
    while (ptlINode_get_elem (&l, &n))
      {
	aSig = NM->get_io_arg (n);
	if (ptlinstances_isin_exp (Outputs_env, aSig))
	  ptlINode_insert_elem (ln, n);
      }
  return ((*ln) != NILLINODE);
}


_VIRTUAL_ boolean
cGeneratorDist::getOutputNodesToGenerate (ptlINode * ln)
{
  ptlINode l;
  ptINode n;
  ptexpcal aSig;

  (*ln) = NILLINODE;
  if (NM->getOutputs (&l))
    while (ptlINode_get_elem (&l, &n))
      {
	aSig = NM->get_io_arg (n);
	if (ptlinstances_isin_exp (Outputs_env, aSig))
	  if (!ptlinstances_isin_exp (Outputs_comm, aSig))
	    ptlINode_insert_elem (ln, n);
      }
  return ((*ln) != NILLINODE);
}


_VIRTUAL_ boolean
cGeneratorDist::getCommInputNodes (ptlINode * ln)
{
  ptlINode l;
  ptINode n;

  (*ln) = NILLINODE;
  if (NM->getInputs (&l))
    while (ptlINode_get_elem (&l, &n))
      if (ptlinstances_isin_exp (Inputs_comm, NM->get_io_arg (n)))
	ptlINode_insert_elem (ln, n);
  return ((*ln) != NILLINODE);
}

_VIRTUAL_ boolean
cGeneratorDist::getCommOutputNodes (ptlINode * ln)
{
  ptlINode l;
  ptINode n;

  (*ln) = NILLINODE;
  if (NM->getOutputs (&l))
    while (ptlINode_get_elem (&l, &n))
      if (ptlinstances_isin_exp (Outputs_comm, NM->get_io_arg (n)))
	ptlINode_insert_elem (ln, n);
  return ((*ln) != NILLINODE);
}

/**
   Generation of the declarations of the signals of a list of signals (lsig).
*/
_VIRTUAL_ void
cGeneratorDist::genCommDclSignals (ptlinstances lsig)
{
  ptexpcal sig;
  TypeType t;

  cdebdecl ();
  while (ptlinstances_get_next (&lsig, &sig))
    {
      t = EM->getType (sig);
      gentype (t);
      TextForm->printidentCommModule (EM->getName (sig));
      genDclDimensions (t);
    }
  cfindecl ();
}

_VIRTUAL_ void
cGeneratorDist::genDclCommTasks ()
{
  TextForm->printNumberOfIOTasks (ptlinstances_get_size (Inputs_comm) + ptlinstances_get_size (Outputs_comm));
}


/**
   Generation of the declarations of the input/output.
   It could be FIFOS (but it is not yet implemented).

   note : not used when not simpleComm.
*/
_VIRTUAL_ void
cGeneratorDist::genCommDeclarations ()
{
  genComment (comm_dclent);
  genCommDclSignals (Inputs_comm);
  genComment (comm_dclsort);
  genCommDclSignals (Outputs_comm);
  TextForm->printLine ();
  // Declaration of the semaphores for IO operations.
  genDclSemaphoresSignals (Inputs_comm);
  genDclSemaphoresSignals (Outputs_comm);
  genDclCommTasks ();
  TextForm->printLine ();
}

/**
   Generation of the declarations of the semaphores used for communications.
   It applies genDclSemaphoresForSignal() for each element of a list of signals (lsig).
*/
_VIRTUAL_ void
cGeneratorDist::genDclSemaphoresSignals (ptlinstances lsig)
{
  ptexpcal sig;

  // A adpater pour des fifos.
  while (ptlinstances_get_next (&lsig, &sig))
    TextForm->printDclSemaphoreComm (EM->getName (sig));
}

_VIRTUAL_ void
cGeneratorDist::genInitSemaphores (ptlinstances lsig, boolean boutput)
{
  ptexpcal sig;
  while (ptlinstances_get_next (&lsig, &sig))
    TextForm->printInitSemaphore (EM->getName (sig), boutput);
}

_VIRTUAL_ void
cGeneratorDist::genCreateCommTasks (ptlinstances lsig, int *n, boolean boutput)
{
  ptexpcal sig;

  while (ptlinstances_get_next (&lsig, &sig))
    {
      TextForm->printCreateIOTask (*n, EM->getName (sig), boutput);
      (*n)++;
    }
}


_VIRTUAL_ boolean
cGeneratorDist::getCommInfos (ptpragma aPragma, ptexpcal * id, ntrees * pValues)
{
  ptlinstances ls;
  ls = (ptlinstances) getPragmaObjects (aPragma);
  if (!ptlinstances_get_next (&ls, id))
    return pk_false;
  *pValues = getPragmaStatement (aPragma);
  return pk_true;
}


/**
   note : Called when not simpleComm.
*/
_VIRTUAL_ void
cGeneratorDist::genCodeCommTasks ()
{
  ptpragma aPragma;
  ptrepstr nPragma;
  ptexpcal id;
  ptlpragmas lprag;
  ntrees pragmaValues;


  lprag = AGRAPH->getPragmas ();
  while (ptlpragmas_get_next (&lprag, &aPragma))
    {
      nPragma = getPragmaName (aPragma);
      if (streq (nPragma, vtrInput))  // receiving
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  TextForm->printReceiveTaskDef (EM->getName (id), EM->getType (id), pragmaValues);
	}
      else if (streq (nPragma, vtrOutput)) // sending
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  TextForm->printSendTaskDef (EM->getName (id), EM->getType (id), pragmaValues, ptlinstances_isin_exp (Outputs, id));
	}
    }
}

_VIRTUAL_ boolean
cGeneratorDist::genDeclCommunicationCounter ()
{
  ptpragma aPragma;
  ptrepstr nPragma;
  ptexpcal id;
  ptlpragmas lprag;
  ntrees pragmaValues;
  boolean b;

  lprag = AGRAPH->getPragmas ();
  b = pk_false;
  while (ptlpragmas_get_next (&lprag, &aPragma))
    {
      nPragma = getPragmaName (aPragma);
      if (streq (nPragma, vtrOutput))
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  TextForm->printDeclCommunicationCounter (EM->getName (id));
	  b = pk_true;
	}
    }
  TextForm->printLine ();
  return pk_true;
}

/**
   It generates the call to the printer of the Communication counters ( _printCommCounters_()) defined in the 
   io module.   
*/
_VIRTUAL_ void
cGeneratorDist::genPrintInfosStep ()
{
  TextForm->printCallToCommunicationCounter ();
  TextForm->printIncrStepCounter ();
}

/**
   Generates the declaraion of the counter of steps.
*/
_VIRTUAL_ void
cGeneratorDist::genDeclStepCounter()
{
  TextForm->DeclStepCounter ();
}

_VIRTUAL_ void
cGeneratorDist::genEntryPrintCommunicationCounter ()
{
  ptpragma aPragma;
  ptrepstr nPragma;
  ptexpcal id;
  ptlpragmas lprag;
  ntrees pragmaValues;
  boolean bfirst;

  TextForm->printCommunicationCounterBegin ();
  
  lprag = AGRAPH->getPragmas ();
  bfirst = pk_true;
  while (ptlpragmas_get_next (&lprag, &aPragma))
    {
      nPragma = getPragmaName (aPragma);
      if (streq (nPragma, vtrOutput))
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  TextForm->genCommunicationCounter (EM->getName (id));
	  bfirst = pk_false;
	}
    }
  if (! bfirst )
    {
      lprag = AGRAPH->getPragmas ();
      TextForm->printNumberCommunicationsBegin ();
      
      bfirst = pk_true;
      while (ptlpragmas_get_next (&lprag, &aPragma))
	{
	  nPragma = getPragmaName (aPragma);
	  if (streq (nPragma, vtrOutput))
	    {
	      getCommInfos (aPragma, &id, &pragmaValues);
	      TextForm->genCommunicationNamePlus (EM->getName (id), !bfirst);
	      bfirst = pk_false;
	    }
	}
      TextForm->printNumberCommunicationsEnd ();
    }
  TextForm->printEndIfDirectiveCommunicationCounter ();
  TextForm->printTailCommunicationCounter ();
}


_VIRTUAL_ void
cGeneratorDist::genCommNodesSimple ()
{
  ptpragma aPragma;
  ptrepstr nPragma;
  ptexpcal id;
  ptlpragmas lprag;
  ntrees pragmaValues;
  boolean bcomms, b;

  bcomms = genDeclCommunicationCounter ();	// To have the number of communications.
  lprag = AGRAPH->getPragmas ();
  while (ptlpragmas_get_next (&lprag, &aPragma))
    {
      nPragma = getPragmaName (aPragma);
      if (streq (nPragma, vtrInput))
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  genReadHeaderExterncall (id, pk_true);
	  TextForm->printReceiveTaskDefSimple (EM->getName (id), EM->getType (id), pragmaValues);
	}
      else if (streq (nPragma, vtrOutput))
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  genWriteHeaderExterncall (id, pk_true);
	  // A signal can be send to other processors, and can be a "real" output.
	  b = ptlinstances_isin_exp (Outputs_env, id);
	  TextForm->printSendTaskDefSimple (EM->getName (id), EM->getType (id), pragmaValues, b);
	  if (b)
	    {
	      genDefWritingSignal (id);	// an output of the application (environment).
	      TextForm->printCloseBlockIndent (0);
	    }
	}
    }
  if (bcomms)
    genEntryPrintCommunicationCounter ();

}


_VIRTUAL_ void
cGeneratorDist::genMainTask ()
{
  TextForm->printTasksMainEntry (SemSynchroEndClusters, pk_true);
}

/** 
   Generation of the entry point foo_OpenComm() of the file foo_comm.c.
   It generates the call to the initialisation of the support of the execution (such that LAM/MPI,...).
*/
_VIRTUAL_ void
cGeneratorDist::genCommInit ()
{
  int n;

  TextForm->printdclbegcommfunchead ();
  TextForm->printCommInit ();	// initialising of the support of the execution.
  if (!simpleComm)
    {
      genInitSemaphores (Inputs_comm, pk_false);
      genInitSemaphores (Outputs_comm, pk_true);
      n = 0;
      genCreateCommTasks (Inputs_comm, &n, pk_false);
      genCreateCommTasks (Outputs_comm, &n, pk_true);
    }

  TextForm->printdclfuncend ();
}

/**
   Generation of  the entry point Foo_closeComm() of the file foo_comm.c, where Foo denotes the name
   of the current graph. 
   It generates the call to the finalizing of the support of execution (such that LAM/MPI,...).
*/
_VIRTUAL_ void
cGeneratorDist::genCommFinalize ()
{
  TextForm->printdclendcommfunchead ();
  TextForm->printCommFinalize ();
  TextForm->printdclfuncend ();
}

/**
  It generates the body of the reading function of a signal.
  @param aSig (in) an input signal. 
*/
_VIRTUAL_ void
cGeneratorDist::genCommDefReadingSignal (ptexpcal aSig)
{
  boolean bref;

  bref = EM->getNumberOfDimensions (aSig) == 0 && (EM->getType (aSig) != StringType);
  TextForm->printBodyReceive (EM->getName (aSig), bref);
}

/**
  Generation of  the body of the writing function of a signal.
  @param aSig (in) an output signal. 
*/
_VIRTUAL_ void
cGeneratorDist::genCommDefWritingSignal (ptexpcal aSig)
{
  boolean bref;

  bref = EM->getNumberOfDimensions (aSig) == 0 && (EM->getType (aSig) != StringType);
  TextForm->printBodySend (EM->getName (aSig), bref);
}

/** Generation of the "reading function" entry point of a signal. 
   @param aSig (in) a input signal.
*/
_VIRTUAL_ void
cGeneratorDist::genCommDeclAndDefReadingSignal (ptexpcal aSig)
{
  genReadHeaderExterncall (aSig, pk_true);
  genCommDefReadingSignal (aSig);
}

/** Generation of  the "writing function" entry point of a signal. 
   @param aSig (in) a output signal.
*/
_VIRTUAL_ void
cGeneratorDist::genCommDeclAndDefWritingSignal (ptexpcal aSig)
{
  genWriteHeaderExterncall (aSig, pk_true);
  genCommDefWritingSignal (aSig);
}

/** 
   Generation of  reading/writing operations for input/output signals of the program.

   note : not used when not simpleComm.
 */
_VIRTUAL_ void
cGeneratorDist::genCommNodes ()
{
  ptlINode ln;
  ptINode n;

  if (getCommInputNodes (&ln))
    while (ptlINode_get_elem (&ln, &n))
      genCommDeclAndDefReadingSignal (NM->get_io_arg (n));

  if (getCommOutputNodes (&ln))
    while (ptlINode_get_elem (&ln, &n))
      genCommDeclAndDefWritingSignal (NM->get_io_arg (n));
}

/**
   It closes the file (foo_comm.c), previously opened in genCommModuleInit().
   _VIRTUAL_ void
   cGeneratorDist::genCommModuleFinalize()
   {
   TextForm->closefile();  
   }
*/


/**
   It generates the declaration of some specific includes for the distributed code.
   \code
   #include "pK_semaphores.h"
   #include "pK_threads.h"
   #include "runConfig.h"
   #include "pK_distrib.h"
   \endcode
   and if derivedTypesUsed is pk_true, it generates the include of the file in which the
   definition of the derived types (i.e structures) used for the communication will be 
   defined by the user.
   For example, for MPI support the following code will be generated
   \code
   #include "MPI_derived_Types.c"
   \endcode
*/
_VIRTUAL_ void
cGeneratorDist::genCommModuleInit ()
{
  TextForm->printIncludeForDistribution (simpleComm, derivedTypesUsed);
}

void
cGeneratorDist::setDerivedType (ptexpcal id)
{
  TypeType t;

  if (!derivedTypesUsed)
    {
      t = EM->getType (id);
      if (TextForm->getIndexTypeCode (t) < 0)
	derivedTypesUsed = pk_true;
    }
}

_VIRTUAL_ void
cGeneratorDist::selectingIO ()
{
  ptpragma aPragma;
  ptrepstr nPragma;
  ptexpcal id;
  ptlpragmas lprag;
  ntrees pragmaValues;

  Inputs_env = NILLINSTANCES;
  Inputs_comm = NILLINSTANCES;
  Outputs_env = NILLINSTANCES;
  Outputs_comm = NILLINSTANCES;
  derivedTypesUsed = pk_false;
  lprag = AGRAPH->getPragmas ();
  while (ptlpragmas_get_next (&lprag, &aPragma))
    {
      nPragma = getPragmaName (aPragma);
      if (streq (nPragma, vtrEnv))
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  if (ptlinstances_isin_exp (Inputs, id))
	    Inputs_env = ptlinstances_insert (Inputs_env, id);
	  else
	    Outputs_env = ptlinstances_insert (Outputs_env, id);
	}
      else if (streq (nPragma, vtrInput))
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  setDerivedType (id);
	  Inputs_comm = ptlinstances_insert (Inputs_comm, id);
	}
      else if (streq (nPragma, vtrOutput))
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  setDerivedType (id);
	  Outputs_comm = ptlinstances_insert (Outputs_comm, id);
	}
    }

}


_VIRTUAL_ void
cGeneratorDist::genDistIni ()
{
  vtrEnv = getPredefinedPragma (PRAGMA_COMM_ENV);
  vtrInput = getPredefinedPragma (PRAGMA_COMM_RECEIVE);
  vtrOutput = getPredefinedPragma (PRAGMA_COMM_SEND);
  selectingIO ();
}

_VIRTUAL_ void
cGeneratorDist::initialize (ptDCG_sch vGraph, boolean genMode, ptrepstr vprefix)
{

  cGeneratorTasks::initialize (vGraph, genMode, vprefix);
  genDistIni ();
}

/**
   \copydoc cGeneratorDist::geniomodule().
   Specialisation of cGeneratorDist::geniomodule()
   The communications are also generated in the Foo_io.c componnent, where Foo is the name of the graph.
*/
_VIRTUAL_ void
cGeneratorDist::geniomodule ()
{
  ptlINode ins, outs, outs1;

  getInputNodes (&ins);
  getOutputNodes (&outs);

  geniomoduleInit ();		// std io
  genCommModuleInit ();		// comm declarations
  genIOFileDeclarations (ins, outs);
  genIOOpening (ins, outs);
  genIOClosing (ins, outs);

  getOutputNodesToGenerate (&outs1);
  genIOReadingWriting (ins, outs1);
  // Communications.
  if (simpleComm)		// no fifos
    {
      genCommInit ();
      genCommFinalize ();
      genCommNodesSimple ();
    }
  else
    {
      genCommDeclarations ();
      genCodeCommTasks ();
      genCommInit ();
      genCommFinalize ();
      genCommNodes ();
    }
  geniomoduleFinalize ();
}


/**
   Declaration of some particular functions used for debugging, polyendochrony,...
*/
_VIRTUAL_ void
cGeneratorDist::genSpecifics()
{
  TextForm->printHeaderCommunicationCounter ();
}



_VIRTUAL_ void
cGeneratorDist::genRendezVousNode(ptident nom, ptexpcal lparams, ptexpcal linputs, ptexpcal loutputs)
{
  ptexpcal vSend;

  TextForm->printBeginComment ();
  genExternalCall (nom, linputs, loutputs, pk_true, NILPRAGMA, NILTREE);
  TextForm->printEndComment ();
  TextForm->printLine ();
  if (isSendId(nom))
    {
      // A RDV variable is defined by the local step counter.
      vSend = EM->getNaryArgs (&loutputs);
      gensig (vSend, pk_false);
      TextForm->printassign ();
      TextForm->printStepCounter();
      TextForm-> printSemiColonLine ();
    }
  else 
    if (isReceiveId(nom))
      {
	// Nothing to do: see the "rendezVous" node.
      }
    else
     if (isRendezVousId(nom)) 
       {
	 vSend = EM->getNaryArgs (&linputs);
	 genInstants(vSend, genMax(linputs));
       } 
}

/**
   Generates (nbInstants-vSend) instants for the resynchronisation of this part of program
   with the other ones afer a rendez-vous; where vSend is the 'step counter' of this part.
   For that the boolean clocks defined at the rrot are sent with 'false' value.
 */
_VIRTUAL_ void cGeneratorDist::genInstants(ptexpcal vSend, ptident nbInstants)
{

  /* generate:
     if (nbInstants-vcpt) 
     {
       for (int i=1; i<= (nbInstants - vSend); i++)
       {
         Send the boolean clocks for a bottom.
       };
       update the local 'step counter'
     }
*/

  TextForm->printkwif();
  TextForm->printPO();
  TextForm->printident(nbInstants);
  TextForm->printkwGreaterthan ();
  gensig(vSend, pk_false);
  TextForm->printPF();
  TextForm-> printLine (); 
  TextForm-> printblockopening ();
  
  ptlinstances lClocksToSend = genBooleanClocksToSend();
  if (lClocksToSend != NILLINSTANCES)
    {
      ptident id = string2ident("i");

      TextForm->printForLP ();
      TextForm->printtype(IntegerType, pk_true);
      TextForm->printident(id);
      TextForm->printassign ();
      TextForm->printint(1);
      TextForm->printSemiColon ();
      TextForm->printident(string2ident("i"));
      TextForm->printkwLessthanEqual ();
      TextForm->printPO();
      TextForm->printident(nbInstants);
      TextForm->printkwminus ();
      gensig(vSend, pk_false);
      TextForm->printPF();
      TextForm->printSemiColon ();
      TextForm->printident(string2ident("i"));
      TextForm->printkwincr ();
      TextForm->printPF();
      TextForm-> printLine (); 
      TextForm-> printblockopening ();
      genInstant(lClocksToSend);
      TextForm-> printBlockClosingBackIndent ();
    }
  // Updating the 'step counter'
  TextForm->printStepCounter();
  TextForm->printassign ();
  TextForm->printident(nbInstants);
  TextForm->printSemiColonLine ();
  TextForm-> printBlockClosingBackIndent ();
}

/**
   Generates the definition of the max of the elements of linputs.
   @return the generate identifier.
 */
_VIRTUAL_  ptident cGeneratorDist::genMax(ptexpcal linputs)
{
  ptexpcal  vReceive;

  // Declaration 
  ptident idMax = string2ident("__polyEndo_RDV_");
  TextForm->printtype(IntegerType, pk_true);
  TextForm->printident(idMax);
  TextForm-> printSemiColonLine ();

  // init __nbInstants__ by the first element of linputs
  EM->getNaryArgs (&linputs, &vReceive);
  TextForm->printident(idMax);
  TextForm->printassign ();
  gensig (vReceive, pk_false);
  TextForm-> printSemiColonLine ();
  // max computation
  while (EM->getNaryArgs (&linputs, &vReceive))
    {
      TextForm->printkwif();
      TextForm->printPO();
      gensig (vReceive, pk_false);
      TextForm->printkwGreaterthan ();
      TextForm->printident(idMax);
      TextForm->printPF();
      TextForm->printident(idMax);
      TextForm->printassign ();
      gensig (vReceive, pk_false);
      TextForm-> printSemiColon ();
    }
  TextForm-> printLine ();
  return idMax;
}

_VIRTUAL_  ptlinstances cGeneratorDist::genBooleanClocksToSend()
{
  ptpragma aPragma;
  ptrepstr nPragma;
  ptexpcal id;
  ptlpragmas lprag;
  ntrees pragmaValues;
  ptinsth masterClock;
  
  ptlinstances lclocks= NILLINSTANCES;
  
  masterClock = CKM->getTickClock();
  lprag = AGRAPH->getPragmas ();
  while (ptlpragmas_get_next (&lprag, &aPragma))
    {
      nPragma = getPragmaName (aPragma);
      if (streq (nPragma, vtrOutput)) // Sending {id} '9' dest1 '10' dest2 ...
	{
	  getCommInfos (aPragma, &id, &pragmaValues);
	  if (EM->IsBoolean(id))
	    if (EM->getClock(id) == masterClock )
	      lclocks = ptlinstances_insert(lclocks, id);
	}
    }
  return lclocks;
}


_VIRTUAL_  void cGeneratorDist::genInstant(ptlinstances lclocks)
{
  ptexpcal vsig;

   while (ptlinstances_get_next (&lclocks, &vsig))
     {
       TextForm->printWriteFunctionLP (getIOCompleteIdentifier (vsig), progName);
       TextForm->printkwlogicons (pk_false);
       TextForm->printRPSemiColonLine ();
     }
}
