/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PROTECTED_
/**
   @return "true" if the type(t) of an expression(e) is a structure one, "false" otherwise.
  
   For a structure, the code is expanded for each of the field.
*/
  boolean cGenerator::isStructureExpr (ptexpcal e, TypeType * t)
{
  *t = EM->getType (e);
  if (!TypeTestNamedClass (*t))
    return pk_false;
  *t = NamedTypeGetTerm (*t);
  return TypeTestStructureClass (*t);
}

/**
   @return the expression to generate associated with exp.
   The returned expression can be different from exp when some compiler attributes are associated 
   with the expression (OPTIMIZED_ATTRIBUTE, OPTIMIZED_MEMORY_ATTRIBUTE). It also looks at 
   the clone (getCloneClassHead()) and the aggregation getAggregation()).
*/
ptexpcal cGenerator::getExpressionToGenerate (ptexpcal exp)
{
  ptexpcal exp1;
  if (exp == NILEXPCAL)
    return exp;
  exp = EM->getCloneClassHead (exp);
  // while (EM->getAttribute(exp, OPTIMIZED_ATTRIBUTE, (attributeValue *) &exp1)) exp = exp1;
  if (EM->getAttribute (exp, OPTIMIZED_ATTRIBUTE, (attributeValue *) & exp1))
    return getExpressionToGenerate (exp1);
  if (EM->getAttribute (exp, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *) & exp1))
    return getExpressionToGenerate (exp1);
  exp1 = EM->getAggregation (exp);
  if (exp1 != exp)
    return getExpressionToGenerate (exp1);
  return exp;
}

/**
   @param exp (in) an access to an element of an array  (ie x[i]....[k])
   @param l (out) the list of the indeices of exp (ie [i]....[k])
   @return getExpressionToGenerate() of the array (ie x).
*/
_VIRTUAL_
  ptexpcal
cGenerator::mkSignalAndListElem (ptexpcal exp, ptexpcal * l)
{
  ptexpcal exp1, exp2;
  boolean b;

  b = EM->setModeTypeChecking (pk_false);
  EM->getBinaryArgs (exp, &exp1, &exp2);
  if (EM->getOperator (exp1) != opelem)
    {
      *l = exp2;
      exp = exp1;
    }
  else
    {
      *l = NILEXPCAL;
      while (EM->getOperator (exp) == opelem)
	{
	  EM->getBinaryArgs (exp, &exp1, &exp2);
	  *l = EM->mkNaryExp (opleflot, exp2, *l, NILBACKANNOTATION);
	  exp = exp1;
	}
    }
  b = EM->setModeTypeChecking (b);
  return getExpressionToGenerate (exp);
}

/**
   @return "true" if ope is a text operator (=, <, <=, >, >=, !=, ==, <<=), "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::isTestOperator (noperators ope)
{
  switch ((int) ope)
    {
    case opegal:
    case opdiff:
    case opinfe:
    case opiegal:
    case opsupe:
    case opsegal:
    case opegeg:
    case opiieg:
      return pk_true;
    }
  return pk_false;
}

/**
   @return "true" if the signal exp is defined at the initial clock and if exp is not an array, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::isScalarTypeAndConstant (ptexpcal exp, ptexpcal * eDef)
{
  if (EM->getOperatorClass (exp) != variabl)
    return pk_false;
  if (EM->IsIndex (exp))
    return pk_false;
  if (!TypeTestPredef (EM->getUnamedType (exp)))
    return pk_false;
  if (!isNodeAtInitialClock (EM->getNode (exp), eDef))
    return pk_false;
  return pk_true;
}


/**
   It generates the name of (exp) assumed to be an identifier.
*/
_VIRTUAL_ void
cGenerator::printidentifier (ptexpcal exp)
{
  ptident id;

  if (EM->getOperatorClass (exp) == variabl)
    id = EM->getName (exp);
  else
    id = string2ident ("BUG_BUG");
  TextForm->printident (id);
}

/**
   It generates the code associated with an operator (ope) on strings
   ( pK_String followed by EQ | NEQ | LT | GT | LTE | GTE for respectively =, !=, <, >, <=, >= operator.)
*/
_VIRTUAL_ void
cGenerator::genCodeStringOperator (noperators ope)
{
  TextForm->printkwPolychrony (pk_true);
  TextForm->printstr (string2repstr ("String"));
  switch ((int) ope)
    {
    case opegal:
      TextForm->printstr (string2repstr ("EQ"));
      break;
    case opdiff:
      TextForm->printstr (string2repstr ("NEQ"));
      break;
    case opinfe:
      TextForm->printstr (string2repstr ("LT"));
      break;
    case opsupe:
      TextForm->printstr (string2repstr ("GT"));
      break;
    case opiegal:
      TextForm->printstr (string2repstr ("LTE"));
      break;
    case opsegal:
      TextForm->printstr (string2repstr ("GTE"));
      break;
    default:
      TextForm->printkwnotimpl ();
    }
}

/**
   @return "true" if h = K - hleft, "false" otherwise
   @param h (in) a clock.
   @param hleft (in) a clock.
   @param exp (out) when  h = K - hleft, exp is the boolean expression of K.
*/
_VIRTUAL_ boolean
cGenerator::genOptimizeClockAssign (ptinsth h, ptinsth hleft, ptexpcal * exp)
{
  ptexpcal condLeft, cond, expd;
  ptINode aNode;

  if (CKM->satisfiesUndefClock (hleft))
    return pk_false;
  condLeft = CKM->getConditionSub (hleft);
  cond = CKM->getConditionSub (h);
  if (!EM->IsSignal (cond))
    return pk_false;
  aNode = EM->getNode (cond);
  if (!NM->get_rightOperand (aNode, &expd))
    return pk_false;
  expd = EM->getBasicExpr (expd);
  // si expd = not(condLeft) && exp then if (exp) is generated
  if (!is_andNotCondition (expd, condLeft, exp))
    return pk_false;
  return EM->IsSignal (*exp);
}


/**
   @return "true" if the operator t is in set  {modulo, unary -, *, /}, "false" otherwise.

   It is used to parenthesize expressions.
*/
_VIRTUAL_ boolean
cGenerator::isinset2 (noperators t)
{
  return ((t == opmod) || (t == opnega) || (t == opmult) || (t == opdivi));
}

/**
  @return  the value "True" if the operator is in set  {modulo, unary -, *, /}U{ binary-, +}, "false" otherwise.
  
  It is used to parenthesize expressions.
*/
_VIRTUAL_ boolean
cGenerator::isinset3 (noperators t)
{
  return (isinset2 (t) || (t == opmoin) || (t == opplus));
}

/**
  @return  "true" if the operator is in set  {or, not}, "false" otherwise.

  It is used to parenthesize expressions.
*/
_VIRTUAL_ boolean
cGenerator::isinset4 (noperators t)
{
  return ((boolean) ((t == opou) || (t == opnon)));

}

/**
  @return  "true" if the operator is in set  {and, not}, "false" otherwise.

  It is used to parenthesize expressions.
*/
_VIRTUAL_ boolean
cGenerator::isinset5 (noperators t)
{
  return ((boolean) ((t == opet) || (t == opnon)));
}

/** 
    @return "true" if an operator (opparent) have the priority greater than an anoter operator (ope), "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::insertBraketForSynchronousExp (noperators ope, noperators opparent)
{
  boolean parenthese;
  if (LL_is_listOp (ope))
    return pk_false;
  switch ((int) ope)
    {
    case opvvarin:
    case oppuis:
    case opcplx:
    case opcham:
    case opappe:
      parenthese = pk_false;
      break;
    case opmult:
    case opdivi:
      parenthese = isinset2 (opparent);
      break;
    case opplus:
    case opmoin:
      parenthese = isinset3 (opparent);
      break;

    case opinfe:
    case opsupe:
    case opiegal:
    case opsegal:
    case opegal:
    case opdiff:
    case opleft_tt:
    case opet:
    case opou:
    case opouex:
    case opegeg:
    case opiieg:
      parenthese = (boolean) (opparent != opnullary);
      break;

    case opmod:
      parenthese = (boolean) (opparent != opnullary);
      break;

    default:
      parenthese = pk_true;
      pK_messenger->errorMethodFile ("insertBraketForSynchronousExp", "cGeneratorExpr_i.c");
    }
  return parenthese;
}


/**
   Specific to the management of the nodes of delay trees.
   @return  "true" if the implementation of the dynamic expressions induces index compression for the element nm (a node of a delay tree), 
   "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::iscompressindex (ptmemnode mn)
{
  if (mn == NILMEMNODE)
    return pk_false;
  return (ptmemnode_get_nb_indexes (mn) != 0);
}

/**
   @return "true" if a signal (exp) is optimized with it delayed signal (BELONG_MEMORY is set in 
   optimisation functionalty to specify that x and zx:=x$1 can be merge in the same memory), "false" otherwise. 
*/
_VIRTUAL_ boolean
cGenerator::satifies_memory_optimized (ptexpcal exp)
{
  exp = getExpressionToGenerate (exp);
  if (!EM->IsSignal (exp))
    return pk_false;
  return NM->isMarked (EM->getNode (exp), BELONG_MEMORY);
}

/**
   @return "true" if exp = not(cond) && OtherExpr, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::is_andNotCondition (ptexpcal exp, ptexpcal cond, ptexpcal * OtherExp)
{
  ptexpcal exp1, exp2;

  if (EM->getOperator (exp) != opet)
    return pk_false;
  EM->getBinaryArgs (exp, &exp1, &exp2);
  if (EM->getOperator (exp1) == opnon)
    {
      exp1 = EM->getUnaryArg (exp1);
      if (exp1 == cond)
	if (!satifies_memory_optimized (exp2))
	  {
	    *OtherExp = exp2;
	    return pk_true;
	  }
    }
  if (EM->getOperator (exp2) == opnon)
    {
      exp2 = EM->getUnaryArg (exp2);
      if (exp2 == cond)
	if (!satifies_memory_optimized (exp1))
	  {
	    *OtherExp = exp1;
	    return pk_true;
	  }
    }
  return pk_false;
}


/** @return  the number of dimensions of pt. */
_VIRTUAL_ int
cGenerator::getexpcalnbdim (ptexpcal pt)
{
  ptexpcal lindices;
  int ndim;
  boolean b;
  noperators t;

  t = EM->getOperator (pt);
  b = ((t == opelevec) || (t == opensvec) || (t == oplelevec));
  if (!b)
    return EM->getNumberOfDimensions (pt);	// signal

  if (EM->getOperator (pt) == oplelevec)
    pt = EM->getNaryArgs (&pt);
  if (EM->getOperator (pt) == opensvec)
    EM->getBinaryArgs (pt, &lindices, &pt);
  if (EM->getOperator (pt) == opelevec)
    EM->getBinaryArgs (pt, &lindices, &pt);
  ndim = 0;
  while (lindices != NILEXPCAL)
    {
      ndim = ndim + 1;
      pt = EM->getNaryArgs (&lindices);
    }
  return ndim;
}


/**
   @return "true" if exp (or its associated signal) is a node of a delay tree (mn), "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::get_tree_delay_indi (ptexpcal exp, ptmemnode * mn)
{
  ptexpcal e;
  e = EM->getUnindexedSignal (exp);
  e = EM->getAggregation (e);
  *mn = ptexpcal_get_node_tree_delay (e);
  return ((*mn) != NILMEMNODE);
}

/**
  @return "true" if exp (or its associated signal) is a node of a delay tree (mn), "false" otherwise.
  exp (out) if modified, it is the associated signal with exp.
*/
_VIRTUAL_ boolean
cGenerator::getindielemdim (ptexpcal * exp, ptmemnode * mn)
{
  ptexpcal e;

  e = EM->getUnindexedSignal (*exp);
  e = EM->getAggregation (e);
  *mn = ptexpcal_get_node_tree_delay (e);
  if ((*mn) != NILMEMNODE)
    {
      *exp = e;
      return pk_true;
    }
  return pk_false;
}

/**
   Generation of the subscripting of exp by dim (i.e. exp[dim]) or exp [i1,...,in]
   where the indices are given by embedded loops (parc).
*/
_VIRTUAL_ void
cGenerator::genelemdim (ptexpcal exp, ptexpcal dim, ptexpcal parc)
{
  ptmemnode mn;
  ptident id;
  boolean done, bgenprefix;
  noperators vop;

  vop = EM->getOperator (exp);
  if (vop == opsdec)
    exp = EM->getUnaryArg (exp);

  if (EM->IsParameter (exp))
    bgenprefix = pk_false;
  else if (EM->IsConstant (exp))
    if (EM->IsExternalConstant (exp))
      bgenprefix = pk_false;
    else
      bgenprefix = !TypeTestPredef (EM->getUnamedType (exp));
  else
    bgenprefix = pk_true;


  done = pk_false;
  if (!EM->getAttribute (exp, AttributeNameGencode, (attributeValue *) & id))	//  --id non exploité
    if (getindielemdim (&exp, &mn))
      {
	if (bgenprefix)
	  genPrefix ();
	genelemdelay (mn, dim, parc);
	done = pk_true;
      }
  if (!done)
    {
      if (vop != opcham)
	genPrefix ();
      exp = getExpressionToGenerate (exp);
      genelemsimple (exp, dim, parc);
    }
}

/** 
    Generation of the code for an identifier (exp) as an operand of an operator (opparent).
    The identifier can be a signal, an index (printIndex()),...
    The generated code can be optimized (genNodeOptimize(ptexpcal, noperators), genValue(), isScalarTypeAndConstant()) otherwise it calls gensig() method.
*/
_VIRTUAL_ void
cGenerator::genIdentifier (ptexpcal exp, noperators opparent)
{
  ptexpcal exp2;
  boolean b;

  if (!printIndex (exp))
    {
      exp = EM->getAggregation (exp);
      if (opparent == opsdec)	// exp[]
	gensig (exp, pk_false);
      else if (!genNodeOptimize (exp, opparent))
	if (!genValue (exp))
	  { 
	    if (isScalarTypeAndConstant (exp, &exp2))
	      {
		b = (EM->getUnamedType (exp2) != StringType);
		if (b)
		  TextForm->printLP ();
		genexpression (exp2, opparent);
		if (b)
		  TextForm->printRP ();
	      }
	    else
	      /* sinon écriture du signal */
	      gensig (exp, pk_false);
	  }
    }
}


/**
  @return  "true" if the signal cond is an optimized one, "false" otherwise.
  When it is optimized, its definition expression is generated as operand of the opparent operator (genNodeOptimize(ptINode, noperators).
*/
_VIRTUAL_ boolean
cGenerator::genNodeOptimize (ptexpcal cond, noperators opparent)
{
  if (!EM->IsSignal (cond))
    return pk_false;
  return genNodeOptimize (EM->getNode (cond), opparent);
}

/**
   @return "true" if a node (aNode) is marked by substituteGenCodeMark, "false" otherwise.
   When it is marked, its definition expression is generated as operand of the opparent operator.
*/
boolean
cGenerator::genNodeOptimize (ptINode aNode, noperators opparent)
{
  ptexpcal expd;
  if (!NM->isMarked (aNode, substituteGenCodeMark))
    return pk_false;
  NM->get_rightOperand (aNode, &expd);
  expd = EM->getBasicExpr (expd);
  genexpression (expd, opparent);
  return pk_true;
}


/**
  Specific to the management of the nodes of delay trees.  
  It generates the code of the index incrementing assigned to the signal x in the delay tree.
  When the signal is defined by partial definitions, index is updated for each definition (they are exclusive).
*/
_VIRTUAL_ void
cGenerator::genIncrIndexDelayTreeNode (ptexpcal x)
{
  genIncrIndexDelayTreeNode (x, pk_false);
}

/**
  Specific to the management of the nodes of delay trees.  
  It generates the code of the index incrementing assigned to the signal x in the delay tree.
  When the signal is defined by partial definitions, index is updated for each definition (they are exclusive).
*/
_VIRTUAL_ void
cGenerator::genIncrIndexDelayTreeNode (ptexpcal x, boolean b)
{
  ptmemnode mn, elem;
  ptlmemnode lmn;
  ptexpcal e;

  if (getIndexStackSize () > 0)
    return;			// delay managment must be done outside of array of processes.
  x = EM->getCloneClassHead (x);
  if (!EM->getSignalAssignedTo (x, &e))
    return;
  e = EM->getCloneClassHead (e);

  if (!ptexpcal_satisfies_node_tree_delay (e, &mn))
    {
      e = EM->getDeclaredSignal (x);
      if (EM->getOperatorClass (e) != variabl)
	return;
      e = EM->getCloneClassHead (e);
      if (!ptexpcal_satisfies_node_tree_delay (e, &mn))
	return;
    }
  x = e;

  if (b || (!b && (ptmemnode_get_delay_exp (mn) == NILEXPCAL)))
    {
      if (ptlinstances_isin_exp (lIncrDelay, x))
	return;
      lIncrDelay = ptlinstances_insert (lIncrDelay, x);

      ptinsth h = EM->getClock (x);
      genBlockOpeningOf (h);

      genIncrement (mn);
      lmn = ptmemnode_get_equiv (mn);
      while (ptlmemnode_getNaryArgs (&lmn, &elem))
	genIncrement (elem);
    }
}


/** 
   Generating the code expg = hexp where hexp denotes an expression (E when H) sub-term of a default expression.
   The clock hleft is used to optimized the code.
   @return true if the clock H  is not the null clock (always, I think).
*/
_VIRTUAL_ boolean
cGenerator::genExpWhenClockDefault (ptexpcal expg, ptexpcal hexp, ptinsth hleft, boolean bfirst)
{
  ptinsth h, hAt;
  ptexpcal expd, exp;
  boolean b;

  h = EM->getContributionClock (hexp);
  if (CKM->satisfiesNullClock (h)) return pk_false; // ununseful but security..;
  hleft = CKM->getRepresentativeFormula (hleft);
  if (CKM->satisfiesUndefClock (hleft))
    {
      hAt = hleft;
      b = pk_true;
    }
  else
    b = CKM->getBooleanClockAt (hleft, CKM->getBranching (hleft, currentClock), &hAt);

  expd = EM->getWhenOrCellArg (hexp);
  if (b && genOptimizeClockAssign (h, hAt, &exp))
    genConditionalAssign (exp, expg, expd, bfirst);
  else
    genConditionalAssign (h, expg, expd, bfirst);
  return pk_true;
}


/** 
   Generation of the code of the expression EXP defined by "exp1 default exp2"
*/
_VIRTUAL_ void
cGenerator::genDefaultExpression (ptexpcal exp)
{
  if (EM->getOperator (exp) == opmele)
    {
      genDefaultExpression (EM->getDefaultArgstAt (exp, 1));
      genWhenExpression (EM->getDefaultArgstAt (exp, 2));
    }
  else
    genWhenExpression (exp);
}

/** 
    This method writes the access to a signal which is an node NM of a delay tree.
    The form of such a code is X[i1, ...,in ], where X is the identifier associates with the node of the delay tree mn.
    The index expressions are built using the parameters lind, ldim, lparc in the current loop context given by nbind (number of opened loops).
*/
_VIRTUAL_ void
cGenerator::genelemdelay (ptmemnode mn, ptexpcal ldim, ptexpcal lparc)
{
  int n, nbic, nbDim;
  /* nbic : number of compacted indexes in MN. */

  nbic = ptmemnode_get_nb_indexes (mn);
  if (nbic == 0)
    {
      ecrsimpleindexdelay (mn);	/* writing tX[iX] */
      n = EM->getSize (ldim) + EM->getSize (lparc);
      nbDim = EM->getNumberOfDimensions (ptmemnode_get_signal (mn));
      genListOfIndexes (ldim, lparc, n, nbDim);
    }
  else
    gennnelemdelay (mn, ldim, lparc, nbic, pk_false);
}

/** 
   Generation of the assignment of a index for a node of a delay tree to a specific value.
   @param mn (in) a node of a delay tree.
   @param i (in) value assigned to the index.
*/
_VIRTUAL_ void
cGenerator::genAssignDelay (ptmemnode mn, int i)
{
  ecrnominddelay (mn);
  TextForm->printassign ();
  TextForm->printint (i);
  TextForm->printSemiColonLine ();
}

/**
  Generation of the increment (modulo the size of the area) of the index assigned to a node of a delay tree.
  @param mn (in) a node of a delay tree.
*/
_VIRTUAL_ void
cGenerator::genIncrement (ptmemnode mn)
{
  ecrnominddelay (mn);
  TextForm->printassign ();
  ecrnominddelay (mn);
  TextForm->printplusOne ();
  TextForm->printSemiColon ();
  TextForm->printkwif ();
  TextForm->printLP ();
  ecrnominddelay (mn);
  TextForm->printkwequal ();
  TextForm->printint (ptmemnode_get_size (mn));
  TextForm->printRP ();
  TextForm->printSpace ();
  TextForm->printkwthen ();
  genAssignDelay (mn, 0);
  TextForm->printkwEndIf ();
}

/** 
   Generation of the array identifier assigned to a signal of a delay tree.
   @param mn (in) a node of a delay tree. The code is \em Tident where \em ident is the name of 
   the signal that is the root of the delay tree.
 */
_VIRTUAL_ void
cGenerator::ecrnomrootinddelay (ptmemnode mn)
{
  TextForm->printidfromchaintid (pk_false, ptmemnode_get_root_num (mn), EM->getName (ptmemnode_get_signal_root (mn)));
}


/**
   Generation of the identifier of the index assigned to a node of a delay tree.
   @param mn (in) a node of a delay tree. The code is \b Iident where \b ident is the name of the signal assigned with mn.
 */
_VIRTUAL_ void
cGenerator::ecrnominddelay (ptmemnode mn)
{
  genPrefix ();
  TextForm->printidfromchaintid  (pk_true, ptmemnode_get_index_num (mn), ptmemnode_get_signal_ident (mn));
}

/** 
    Generation of the access to a signal, node of a delay tree.
    @param mn (in) a node of a delay tree. The code is ARRAYNAME[INDEX].
    \see{ecrnomrootinddelay, ecrnominddelay}
*/
_VIRTUAL_ void
cGenerator::ecrsimpleindexdelay (ptmemnode mn)
{
  ecrnomrootinddelay (mn);	/* tx */
  TextForm->printLeftSquareBracket ();
  ecrnominddelay (mn);
  TextForm->printRightSquareBracket ();
}

/**
  Generation of an index using decreasing priority.
  @param opparent (in) gives the context to put expression in brackets.
  @param ldim (inout) list of dimensions, the first index is generated if lind is empty;
  @param lparc (inout) list of travel, the first index is generated if lind and ldim are empty.
  @param n (inout) the value of n is increasing and generated if the previous list are empty.
  For the lists, the returned value is for the selected one, the list without the first index.
*/
_VIRTUAL_ void
cGenerator::genOneIndexOf (noperators opparent, ptexpcal * ldim, ptexpcal * lparc, int *n)
{
  ptexpcal expindice;

  if ((*ldim) != NILEXPCAL)
    {
      expindice = EM->getNaryArgs (ldim);
      genexpression (expindice, opparent);
    }
  else if ((*lparc) != NILEXPCAL)
    {
      expindice = EM->getBinaryArgstAt (EM->getNaryArgs (lparc), 1);
      genexpression (expindice, opparent);
    }
  else
    {
      *n = (*n) + 1;
      expindice = topIndex ();
      if (expindice == NILEXPCAL)
	TextForm->printindex (*n);	// ERROR (the index In is not declared)
      else
	genexpression (expindice, opparent);
    }
}

/**
   Generation of the access to a signal taking account to indexes.
   @param exp (in) a signal
   @param ldim  (inout) list of dimensions
   @param lparc (inout) list of Signal travels
   \see{genListOfIndexes} 
*/
_VIRTUAL_ void
cGenerator::genIdentsignal (ptexpcal exp, ptexpcal ldim, ptexpcal lparc)
{
  int n, nbDim;
  noperators op;
  ptexpcal exp1, exp2;

  nbDim = getNbDimFlattenedType (EM->getType (exp)); /*EM->getNumberOfDimensions(exp) */ ;

  n = EM->getSize (ldim) + EM->getSize (lparc);
  if (!genAssignedCode (exp, opnullary))
    {
      op = EM->getOperator (exp);
      switch ((int) op)
	{
	case opident:
	  printidentifier (exp);
	  break;
	case opcham:
	  EM->getBinaryArgs (exp, &exp1, &exp2);
	  genexpression (exp1, opcham);
	  genOperator (opcham);
	  genConstant (exp2);	// name of the field
	  break;
	default:
	  TextForm->printkwnotimpl ();
	  setSomeProblem ();
	}
    }
  genListOfIndexes (ldim, lparc, n, nbDim);
  // genDimAdjust(exp);
}

/**
  Generation of the access to a signal taking account to indexes and exceptions.
  @param exp (in) a signal.
  @param ldim  (in) list of dimensions, 
  @param lparc (in) list of travellings.
*/
_VIRTUAL_ void
cGenerator::genelemsimple (ptexpcal exp, ptexpcal ldim, ptexpcal lparc)
{
  ptexpcal indice, indexExpr, vexception, lb;
  ptINode indexNode;
  noperators op;
  boolean bexcept;

  op = EM->getOperator (exp);
  if (op == opident)
    {
      vexception = EM->getException (exp);
      bexcept = (vexception != NILEXPCAL);
    }
  else
    bexcept = pk_false;
  if (bexcept)
    {				/* In an array of process. */
      indice = topIndex ();
      indexNode = EM->getNode (indice);
      indexExpr = NM->get_index_def (indexNode);
      TextForm->printLP ();
      printIndex (indice);
      TextForm->printkwequal ();
      lb = EM->getTernaryArgstAt (indexExpr, 1);
      genexpression (lb, opnullary);

      TextForm->printSpace ();
      TextForm->printQuestionMark ();
      TextForm->printSpace ();
      genIdentifier (vexception, opnullary);
      TextForm->printSpace ();
      TextForm->printColon ();
      TextForm->printSpace ();
      genIdentsignal (exp, ldim, lparc);
      TextForm->printRP ();
    }
  else
    genIdentsignal (exp, ldim, lparc);
}

/** 
   Generation of compacted index for defining access to a node of a delay tree.
   @param nbic (in) number of indexes (of ldim, lparc, [n+1, nbic]) to be sumed.
   @param ldim  (inout) list of dimensions, 
   @param lparc (inout) list of Signal travels, 
   @param n (inout) gives the upper bound of the interval [n+1, nbic].
*/
_VIRTUAL_ void
cGenerator::genCompactedIndexes (int nbic, ptexpcal * ldim, ptexpcal * lparc, int *n)
{
  while (nbic > 0)
    {
      TextForm->printkwplus ();
      genOneIndexOf (opplus, ldim, lparc, n);
      nbic--;
    }
}

/** 
   Generation of the access to a node of a delay tree with compacted index.
   This method writes the access to an node NM of a delay tree with compacted indexes.
   The form of such a code is  <em> ident_delay_tree[index_expr]  </em>
   with  <em> index_expr = compacted_indexes modulo memorySize  </em>.
    <em> compacted_indexes  </em> is the sum of nbic indexes taken in the input parameters lind, ldim, lparc, nind. 
    <em> memorySize </em> is the size of the array used to manage the delay tree in which exp is a node.

   @param mn (in) a node of a delay tree,
   @param nbic (in) number of indexes of ldim,
   @param ldim (in) list of dimensions, 
   @param lparc (in) list of Signal travels, 
   @param decr (in) strange argument (?).
*/
_VIRTUAL_ void
cGenerator::gennnelemdelay (ptmemnode mn, ptexpcal ldim, ptexpcal lparc, int nbic, boolean decr)
{
  int nbDim, m, n, memtaille, i;

  ecrnomrootinddelay (mn);
  TextForm->printLeftSquareBracket ();
  TextForm->printLP ();
  ecrnominddelay (mn);
  m = EM->getSize (ldim) + EM->getSize (lparc);
  n = 0;
  genCompactedIndexes (nbic, &ldim, &lparc, &n);
  memtaille = ptmemnode_get_size (mn);
  i = memtaille - nbic;
  if (decr)
    i = i - 1;
  else
    i = i + 1;
  if (i != 0)
    TextForm->printplus (i);
  TextForm->printRP ();
  TextForm->printmodulo (memtaille);
  TextForm->printRightSquareBracket ();
  nbDim = EM->getNumberOfDimensions (ptmemnode_get_signal(mn));
  genListOfIndexes (ldim, lparc, nbic, nbDim);
}

/** 
   Generation of the access or the reference to a signal (exp).
   @param exp (in) a signal.
   @param cref (in) must be "true" for the generation of a reference to the signal, "false" for an access.
*/
_VIRTUAL_ void
cGenerator::gensig (ptexpcal exp, boolean cref)
{
  ptexpcal exp1, exp2;
  noperators op;

  if (!printIndex (exp))
    {
      exp = getExpressionToGenerate (exp);
      if (!genAssignedCode (exp, opnullary))
	{
	  if (cref)
	    TextForm->printkwref ();
	  op = EM->getOperator (exp);
	  if (op == opelem)
	    {
	      exp1 = mkSignalAndListElem (exp, &exp2);
	      genelemdim (exp1, exp2, NILEXPCAL);
	    }
	  else
	    genelemdim (exp, NILEXPCAL, NILEXPCAL);
	}
    }
}

/** 
    It generates the code of an operator (op).
    (see opersignal.h file for the set of SIGNAL operators).
*/
_VIRTUAL_ void
cGenerator::genOperator (noperators op)
{
  switch ((int) op)
    {
    case opegeg:		/* == */
      TextForm->printkwegeg ();
      return;
    case opiieg:		/* <<= */
      TextForm->printkwiieg ();
      return;
    case opvvarin:		/*  <state_variable> */
      return;
    }

  if ((LL_is_binaryOp (op)) || (LL_is_listOp (op)) || (LL_is_ternaryOp (op)))
    if (op != opcham)
      TextForm->printSpace ();

  switch ((int) op)
    {
    case opmoin:
    case opnega:
      TextForm->printkwminus ();
      break;
    case opnon:
      TextForm->printkwnot ();
      break;
    case opdiff:
      TextForm->printkwdiff ();
      break;
    case opdivi:
      TextForm->printkwdivi ();
      break;
    case opegal:
      TextForm->printkwequal ();
      break;
    case opet:
      TextForm->printkwand ();
      break;
    case opiegal:
      TextForm->printkwLessthanEqual ();
      break;
    case opinfe:
      TextForm->printkwLessthan ();
      break;
    case opmult:
      TextForm->printstar ();
      break;
    case opou:
      TextForm->printkwor ();
      break;
    case opouex:
      TextForm->printkworexclusif ();
      break;

    case opplus:
      TextForm->printkwplus ();
      break;
    case opsegal:
      TextForm->printkwGreaterthanEqual ();
      break;
    case opsupe:
      TextForm->printkwGreaterthan ();
      break;

    case opmod:
      TextForm->printkwmod ();
      break;
    case opcham:		// access to a field
      TextForm->printDot ();
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }

  if ((LL_is_binaryOp (op)) || (LL_is_listOp (op)) || (LL_is_ternaryOp (op)))
    if (op != opcham)
      TextForm->printSpace ();
}

/** 
   Generation of a valued expression.
   @return "true" if an expression (exp) is a valued expression which is generated, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::genValue (ptexpcal exp)
{
  TypeType TT;
  long int val;

  if (EM->IsIndex (exp))
    return pk_false;
  if (!EM->IsAssigned (exp))
    return pk_false;
  TT = EM->getUnamedType (exp);
  if (!TypeTestPredef (TT))
    return pk_false;
  val = EM->getValue (exp);
  if (TypeTestLogical (TT))
    {
      TextForm->printkwlogicons (val != 0);
      return pk_true;
    }
  if (TypeTestIntNumber (TT))
    {
      TextForm->printint (val);
      return pk_true;
    }
  // default
  return pk_false;
}


/**
   @return "true" if an expression is an int valued expression (val), "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::getValue (ptexpcal exp, long int *val)
{
  TypeType TT;

  if (EM->IsIndex (exp))
    return pk_false;
  if (!EM->IsAssigned (exp))
    return pk_false;
  //  TT = EM->getType(exp);
  TT = EM->getUnamedType (exp);
  if (!TypeTestPredef (TT))
    return pk_false;
  if (TypeTestLogical (TT))
    return pk_false;
  *val = EM->getValue (exp);
  return pk_true;
}


/** 
    Generation of a constant expression (exp) (i.e. get_operatorClass(exp) == constant).
   
*/
_VIRTUAL_ void
cGenerator::genConstant (ptexpcal exp)
{
  noperators ope;
  ptrepstr strval;

  ope = EM->getOperator (exp);
  strval = EM->getConstant (exp);
  switch ((int) ope)
    {
    case optrue:
      TextForm->printkwlogicons (pk_true);
      break;
    case opfalse:
      TextForm->printkwlogicons (pk_false);
      break;
    case opcstdp:
      strval = strdoubletoCdouble (strval);
    default:
      TextForm->printstr (strval);
    }
}

/** 
   Generation of the code of a left_tt expression (exp) operand of an expression with opparent as operator.
   @return "true" if the expression exp is a left_tt one, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::genexpressionLeft_tt (ptexpcal exp, noperators opparent)
{
  boolean parenthese;
  ptexpcal exp1, exp2;
  ptinsth vhexp1;

  if (EM->getOperator (exp) != opleft_tt)
    return pk_false;
  EM->getBinaryArgs (exp, &exp1, &exp2);
  if (EM->getSynchronousExprHorl (exp1, &vhexp1))
    {
      parenthese = (boolean) (opparent != opnullary);
      if (parenthese)
	TextForm->printLP ();

      TextForm->printLP ();
      genClockExpression (vhexp1, opleft_tt);
      TextForm->printSpace ();
      TextForm->printQuestionMark ();
      TextForm->printSpace ();
      genexpression (exp1, opleft_tt);
      TextForm->printSpace ();
      TextForm->printColon ();
      TextForm->printSpace ();
      TextForm->printkwlogicons (pk_false);
      TextForm->printRP ();
      if (parenthese)
	TextForm->printRP ();
      return pk_true;
    }
  pK_messenger->Caseerror (currentLine, currentFile);
  return pk_false;			// default
}

/** Generation of the code of a list of string constant expressions (exp).
    @return "true" if the expression (exp) is a list of string constant expressions, "false" otherwise.
    It calls genConstant().
*/
_VIRTUAL_ boolean
cGenerator::genString (ptexpcal exp)
{
  ptexpcal exp1, exp2;
  if (EM->getOperator (exp) != opchai)
    return pk_false;
  EM->getBinaryArgs (exp, &exp1, &exp2);
  genConstant (exp1);
  if (exp2 != NILEXPCAL)
    genString (exp2);
  return pk_true;
}


/**
   It generates the test <= (v-1), where v is the expression aDim or the value associated with it.
*/
_VIRTUAL_ void
cGenerator::genInfUpperBound (ptexpcal aDim)
{
  long int v;

  TextForm->printkwLessthanEqual ();
  if (getValue (aDim, &v))
    TextForm->printint (v - 1);
  else
    {
      genexpression (aDim, opnullary);
      TextForm->printint (-1);
    }
}

/** 
    Generation of the test of the "out of bounds" for the expression exp 
    that must be a select expression ( i.i T[e1]...[en])
    \see{genexpression}
*/
_VIRTUAL_ boolean
cGenerator::genTestException (ptexpcal exp)
{
  TypeType t;
  ptexpcal exp1, exp2, vdim, aDim, velem;
  boolean first;


  if (EM->getOperator (exp) != opelem)
    return pk_false;
  EM->getBinaryArgs (exp, &exp1, &exp2);
  t = EM->getUnamedType (exp1);
  if (!TypeTestArrayClass (t))
    return pk_false;
  vdim = ArrayMatrixTypeGetDimLst (t);
  first = pk_true;
  while (EM->getNaryArgs (&exp2, &velem))
    {
      EM->getNaryArgs (&vdim, &aDim);
      if (first)
	first = pk_false;
      else
	TextForm->printkwand ();
      TextForm->printLP ();
      genexpression (velem, opnullary);
      TextForm->printkwGreaterthanEqual ();
      TextForm->printint (0);
      TextForm->printSpace ();
      TextForm->printkwand ();
      TextForm->printSpace ();
      genexpression (velem, opnullary);
      genInfUpperBound (aDim);
      TextForm->printRP ();
    }
  return pk_true;
}

/** 
    Generation of the code for the exception operator Exp1 \\ Exp2.
    \see{genexpression}
*/
_VIRTUAL_ void
cGenerator::genException (ptexpcal exp)
{
  ptexpcal exp1, exp2;

  EM->getBinaryArgs (exp, &exp1, &exp2);
  TextForm->printLP ();
  genTestException (exp1);
  TextForm->printSpace ();
  TextForm->printQuestionMark ();
  TextForm->printSpace ();
  genexpression (exp1, opnullary);
  TextForm->printColon ();
  genexpression (exp2, opnullary);
  TextForm->printRP ();
}


/** 
    It generates (exp1, exp2).
    \see{genexpression}.
*/
_VIRTUAL_ void
cGenerator::genExpressionsAsListOfParameters (ptexpcal exp1, ptexpcal exp2)
{
  TextForm->printLP ();
  genexpression (exp1, opnullary);
  TextForm->printComma ();
  genexpression (exp2, opnullary);
  TextForm->printRP ();
}

/** Generation of the code of a STRING binary expression (exp1 ope exp2)
    It calls genCodeStringOperator() followed by genExpressionsAsListOfParameters().
    \code
    Example:  for x = y expression with x, y some strings,
    it generates
    pK_StringEQ( genexpression(x), genexpression(y)))
    \endcode
*/
_VIRTUAL_ void
cGenerator::genStringBinaryExpression (noperators ope, ptexpcal exp1, ptexpcal exp2)
{
  genCodeStringOperator (ope);
  genExpressionsAsListOfParameters (exp1, exp2);
}

/** 
    Generation of the code of a STRING binary expression (exp)
    @return "true" when some code is generated, "false" otherwise.
    It calls:
    - genCmpOperator() for == or <<= operators
    - genStringBinaryExpression(noperators, ptexpcal, ptexpcal) otherwise.    
*/
_VIRTUAL_ boolean
cGenerator::genStringBinaryExpression (ptexpcal exp)
{
  noperators ope;
  ptexpcal exp1, exp2;
  TypeType t;

  ope = EM->getOperator (exp);
  if (!isTestOperator (ope))
    return pk_false;
  EM->getBinaryArgs (exp, &exp1, &exp2);
  //  t = EM->getType(exp1);
  t = EM->getUnamedType (exp1);	
  if (!TypeTestPredef (t))
    if (TypeTestArrayClass (t))
      t = ArrayMatrixTypeGetTerm (t);	// to be completed
  if (t != StringType)
    return pk_false;
  if ((ope == opegeg) || (ope == opiieg))	// == and <<= operators
    {
      genCmpOperator (ope, exp1, exp2);
      return pk_true;
    }
  genStringBinaryExpression (ope, exp1, exp2);
  return pk_true;
}

/** 
    Generation of the code of a complex or double complex (given by TT type) binary expression (exp1 ope op2). 
    It is generated as an external call. A library that defines the complex operations is available in the 
    Polychrony distribution (lib directory).
    \code
    For example:
    for <em> x + y </em>  where \em x and \em y are complex expressions,
    it generates <em> pK_complexPlus(x, y).</em>.
    \endcode
*/
_VIRTUAL_ void
cGenerator::genComplexBinaryExpression (noperators ope, ptexpcal exp1, ptexpcal exp2, TypeType TT)
{
  TextForm->printtype (TT, pk_false);
  switch ((int) ope)
    {
    case opcplx:
      TextForm->printstr (string2repstr ("New"));
      break;

    case opplus:
      TextForm->printstr (string2repstr ("Plus"));
      break;

    case opmult:
      TextForm->printstr (string2repstr ("Mult"));
      break;

    case opmoin:
      TextForm->printstr (string2repstr ("Minus"));
      break;
    case opdivi:
      TextForm->printstr (string2repstr ("Div"));
      break;
    case opegal:
      TextForm->printstr (string2repstr ("Eq"));
      break;
    case opdiff:
      TextForm->printstr (string2repstr ("Neq"));
      break;
    case oppuis:
      TextForm->printstr (string2repstr ("Pow"));
      break;
    }
  TextForm->printLP ();
  genexpression (exp1, opnullary);
  TextForm->printComma ();
  genexpression (exp2, opnullary);
  TextForm->printRP ();
}

/**
   Generation of the code of a complex or double complex binary expression 
   @return "true" if exp is complex expression, "false" otherwise.
   It calls genComplexBinaryExpression(noperators, ptexpcal,ptexpcal,TypeType).
*/
_VIRTUAL_ boolean
cGenerator::genComplexBinaryExpression (ptexpcal exp)
{
  noperators ope;
  ptexpcal exp1, exp2;
  TypeType t;

  t = EM->getUnamedType (exp);
  if (!TypeTestComplexNumber (t))
    return pk_false;
  if (TypeTestArrayClass (t))
    t = ArrayMatrixTypeGetTerm (t);	// to be completed
  EM->getBinaryArgs (exp, &exp1, &exp2);
  ope = EM->getOperator (exp);
  genComplexBinaryExpression (ope, exp1, exp2, t);
  return pk_true;
}

/** 
    Generation of the code of the expression exp1 codop exp2 with t as type for codop a equal to (==) or smaller/equal to (<<=) operator.
*/
_VIRTUAL_ void
cGenerator::genCmpOperator (noperators codop, ptexpcal exp1, ptexpcal exp2)
{
  TypeType t;

  t = EM->getType (exp1);
  if (isDefinedCmp (t, codop))
    TextForm->printident (getIdCmpTable (t, codop));
  else
    {
      TextForm->printkwnotimpl ();
      setSomeProblem ();
    }
  genExpressionsAsListOfParameters (exp1, exp2);
}

/**
   It generates the code for the access to a value (exp) of a enumerated type (t).
*/
_VIRTUAL_ void
cGenerator::genAccessEnumValue (TypeType t, ptexpcal exp)
{
  genexpression (exp, opnullary);
}

/** Generation of the code of a binary expression.
    It calls
    - the following methods for particular operators:
       - genfieldAccess() for the generation of the access to a field of a signal (x.y)
       - genString() for some string type expressions
       - genexpressionLeft_tt() for the left_tt expression
       - genCmpOperator() for == and <<= operator
       - genAccessEnumValue() for # operator (access to a value of a enumerated type)
       - genException() for the \\ operator (exception)
       - mkSignalAndListElem() followed by genelemdim() for the access to an element of an array (x[i])
       - genListExpr() for a list of expressions
    - and for the other ones:
       - genComplexBinaryExpression() for complex expressions
       - genStringBinaryExpression() for string expressions
*/
_VIRTUAL_ void
cGenerator::genBinExpression (ptexpcal exp, noperators opparent)
{
  noperators ope;
  ptexpcal exp1, exp2;
  boolean bracket;

  if (!genfieldAccess (exp, opparent))	// acces to a field (opcham operator).
    {
      ope = EM->getOperator (exp);
      if (ope == opinit)
	{
	  // opinit referenced only in counter from/after
	  genexpression (EM->getBinaryArgstAt (exp, 1), opparent);
	}
      else if (!genString (exp))
	if (!genexpressionLeft_tt (exp, opparent))
	  {
	    EM->getBinaryArgs (exp, &exp1, &exp2);
	    switch ((int) ope)
	      {
	      case opegeg:	// ==
	      case opiieg:	// <<=
		genCmpOperator (ope, exp1, exp2);
		break;
	      case opelem:
		exp1 = mkSignalAndListElem (exp, &exp2);
		genelemdim (exp1, exp2, NILEXPCAL);
		break;
	      case opvenum:	// type_enumere # valeur_du type Ex : couleur#bleu
		genAccessEnumValue (EM->getType (exp), exp2);
		break;
	      case oppuis:
		TextForm->printPow ();
		TextForm->printLP ();
		genexpression (exp1, opnullary);
		TextForm->printComma ();
		genexpression (exp2, opnullary);
		TextForm->printRP ();
		break;
	      case oprecu:	// exception  X1 \\ Vexception
		genException (exp);
		break;

	      case opleflot:
		TextForm->printLBracket ();
		genListExpr (exp, string2repstr (","));
		TextForm->printRBracket ();
		break;

	      default:
		if (!genComplexBinaryExpression (exp))
		  if (!genStringBinaryExpression (exp))
		    {
		      bracket = insertBraketForSynchronousExp (ope, opparent);
		      if (bracket)
			TextForm->printLP ();
		      genexpression (exp1, ope);
		      genOperator (ope);
		      genexpression (exp2, ope);
		      if (bracket)
			TextForm->printRP ();
		    }
	      }
	  }
    }
}


/**
   It generates a list of expression (exp) with sepa as separator.
*/
_VIRTUAL_ void
cGenerator::genListExpr (ptexpcal exp, ptrepstr sepa)
{
  if (exp != NILEXPCAL)
    while (pk_true)
      {
	genexpression (EM->getNaryArgs (&exp), opnullary);
	if (exp == NILEXPCAL)
	  return;
	TextForm->printstr (sepa);
      }
}

/** 
    @return "true" if the type of the operand expression (exp) is complex or double complex, "false" otherwise.
    Generation of the code of a complex or double complex unary expression (exp). 
    It exists only one operator(-).
    It is generated as an external call. A library that defines the complex operations is available in the 
    Polychrony distribution (lib directory).
    \code
    For example:
    for <em> -x </em>  where \em x is a complex expressions,
    it generates <em> pK_complexUMinus(x).</em>.
    \endcode
*/
_VIRTUAL_ boolean
cGenerator::genComplexUnaryExpression (ptexpcal exp)
{
  TypeType t;
  noperators ope;
  ptexpcal exp1;

  //  t = EM->getType(exp);
  t = EM->getUnamedType (exp);
  if (!TypeTestComplexNumber (t))
    return pk_false;
  if (TypeTestArrayClass (t))
    t = ArrayMatrixTypeGetTerm (t);	// to be completed
  if ((t != ComplexType) && (t != DComplexType))
    return pk_false;
  ope = EM->getOperator (exp);
  switch ((int) ope)
    {
    case opnega:
      TextForm->printtype (t, pk_false);
      TextForm->printstr (string2repstr ("UMinus"));
      break;
      // other 
    }
  TextForm->printLP ();
  exp1 = EM->getUnaryArg (exp);
  genexpression (exp1, opnullary);
  TextForm->printRP ();
  return pk_true;
}

/**
   @return "true" if the operand expression (exp) is an access to a field of a signal (ie x.y), "false" otherwise.
   The code is generated when it returns "true".
*/
boolean cGenerator::genfieldAccess (ptexpcal exp, noperators opparent)
{
  ptexpcal
    exp1;
  ptmemnode
    mn;
  boolean
    done;

  if (EM->getOperator (exp) != opcham)
    return pk_false;
  done = pk_false;
  if (EM->getSignalAssignedTo (exp, &exp1))
    {
      if (genNodeOptimize (EM->getNode (exp1), opparent))
	return pk_true;
      if (get_tree_delay_indi (exp1, &mn))	//  if (!getindielemdim(&exp1, &mn)) return pk_false;
	{
	  genelemdelay (mn, NILEXPCAL, NILEXPCAL);
	  done = pk_true;
	}
    }
  if (!done)
    genelemsimple (exp, NILEXPCAL, NILEXPCAL);
  return pk_true;
}

/**
   It generates the code of the operand expression (exp) when it is annotated with specific code
   given by the AttributeNameGencode internal attribute (See stateVariablesAttribute()).

   @return "true" if some code is generated, "false" otherwise.
*/
boolean cGenerator::genAssignedCode (ptexpcal exp, noperators opparent)
{
  ptident id;

  if (!EM->getAttribute (exp, AttributeNameGencode, (attributeValue *) & id)) return pk_false;
  printPrefixedIdent(id);
  genIndexes (); // contextual indexes
  return pk_true;
}


/**
   It generates a function call (fc) as a child of an expression with opparent as operator.
*/
_VIRTUAL_ void
cGenerator::genFunctionCall (ptexpcal fc, noperators opparent)
{
  ptexpcal inputs, params;
  ptident id;
  ptpragma valPragma;
  ntrees statement;

  EM->getFunctionCall (fc, &id, &params, &inputs);
  if (getPragmaStrKeyword (EM->getPragmas (fc), PragmaGencodeLanguage, &valPragma))
    {
      statement = getPragmaStatement (valPragma);
      genPragmaCode (id, statement, params, inputs, fc);
    }
  else
    {
      TextForm->printident (id);
      TextForm->printLP ();
      genListExpr (inputs, string2repstr (","));
      TextForm->printRP ();
    }
}


/**
   It generates an identifier(exp) taking into account a prefix identifier.
   See: genPrefix(), printidentifier().
*/
_VIRTUAL_ void
cGenerator::genPrefixedIdentifier (ptexpcal exp)
{
  genPrefix ();
  printidentifier (exp);
}

/**
   @return "true" if the operand expression is an index, "false" otherwise.
   It generates an identifier (exp) when it is an index.
   See genPrefixedIdentifier().
*/
_VIRTUAL_ boolean
cGenerator::printIndex (ptexpcal exp)
{
  if (!EM->IsIndex (exp))
    return pk_false;
  genPrefixedIdentifier (exp);
  return pk_true;
}


/**
   It generates an unary expression (exp) as a child of an expression with opparent as operator.
*/
_VIRTUAL_ void
cGenerator::genUnaryExpression (ptexpcal exp, noperators opparent)
{
  ptexpcal exp1;
  noperators ope;

  ope = EM->getOperator (exp);
  if (ope == opindirection)
    genexpression (EM->getUnaryArg (exp), opparent);	// urgence
  else if (!genComplexUnaryExpression (exp))
    {
      exp1 = EM->getUnaryArg (exp);
      genOperator (ope);
      genexpression (exp1, ope);
    }
}

/**
   It generates a temporal-binary expression (exp) as a child of an expression with opparent as operator.
   It is reduced to the default expression.
*/
_VIRTUAL_ void
cGenerator::genBinTpExpression (ptexpcal exp, noperators opparent)
{
  // default expression
  sameLine = pk_true;
  TextForm->printLP ();
  genDefaultExpression (EM->getDefaultArgstAt (exp, 1));
  // The last one: no condition.
  genexpression (EM->getWhenOrCellArg (EM->getDefaultArgstAt (exp, 2)), opparent);
  TextForm->printRP ();
  sameLine = pk_false;
}



/** 
    Generation of the code of an expression  as a child of an expression with opparent as operator.
    It calls the following methods for particular cases:
    - genTravelIndex()
    - genAssignedCode()
    - genValue()
    otherwise, it calls the followings depending on the class of the expression:
    - genIdentifier() for identifiers
    - genConstant() for constant
    - genUnaryExpression() for unary expression (synchronous)
    - genBinExpression() for binary expression (synchronous)
    - genBinTpExpression() for binary expression  (default expression)
    - genexpression() for ternary expression and (E when C) expressions
    - genFunctionCall() for function call expression
*/
_VIRTUAL_ void
cGenerator::genexpression (ptexpcal exp, noperators opparent)
{
  boolean done;

  if (exp != NILEXPCAL)
    {
      exp = getExpressionToGenerate (exp);
      if (!genTravelIndex (exp))
	{
	  if (EM->getOperator (exp) == opsdec)
	    {
	      opparent = opsdec;
	      exp = EM->getUnaryArg (exp);	// []
	      done = pk_false;
	    }
	  else if (genAssignedCode (exp, opparent))
	    done = pk_true;
	  else
	    done = genValue (exp);

	  if (!done)
	    switch ((int) (EM->getOperatorClass (exp)))
	      {
	      case variabl:
		genIdentifier (exp, opparent);
		break;
	      case constant:
		genConstant (exp);
		break;
	      case oper1:
		genUnaryExpression (exp, opparent);
		break;
	      case oper2:
		genBinExpression (exp, opparent);
		break;
	      case oper2tp:
		genBinTpExpression (exp, opparent);
		break;
	      case oper3:
		genexpression (EM->getTernaryArgstAt (exp, 1), opparent);
		break;

	      case opereh:
		// Attention, l'horloge de contribution n'est pas prise en compte.
		genexpression (EM->getWhenOrCellArg (exp), opparent);
		break;
	      case call:
		genFunctionCall (exp, opparent);
		break;
	      case opertyp:
		TextForm->printtype (EM->getExpType (exp), pk_false);
		break;

	      default:
		pK_messenger->Caseerror (currentLine, currentFile);
	      }
	}
    }
}

/** 
   Specific to the management of the nodes of delay tree(misc).\n
   Written for masking the management of an circular area when a node of a tree delay is an argument of a external call.
   It generates a assignment such that <em> left_part := exp </em>  where 
   <em> left_part </em> is the expression assigned to exp in the delay tree.
*/
_VIRTUAL_ void
cGenerator::gencopie (ptexpcal exp)
{
  ptmemnode mn;
  int n;

  mn = ptexpcal_get_node_tree_delay (exp);
  n = genOpenIteration (exp);

  genelemdelay (mn, NILEXPCAL, NILEXPCAL);
  TextForm->printassign ();
  genelemsimple (exp, NILEXPCAL, NILEXPCAL);

  printptvlinebi (n);
  genCloseIteration (n);
}


/**
   Generation of the code for a "when expression" (exp), sub-term of a default expression.
*/
_VIRTUAL_ void
cGenerator::genWhenExpression (ptexpcal exp, ptinsth h)
{
  genClockExpression (h, opnullary);
  TextForm->printQuestionMark ();
  TextForm->printSpace ();
  genexpression (exp, opnullary);
  TextForm->printSpace ();
  TextForm->printColon ();
  TextForm->printSpace ();
}
/**
   Generation of the code for a "when expression" (exp), sub-term of a default expression.
*/
_VIRTUAL_ void
cGenerator::genWhenExpression (ptexpcal exp)
{
  ptinsth h;
  ptexpcal exp1;

  h = EM->getContributionClock (exp);
  exp1 = EM->getWhenOrCellArg (exp);
  genWhenExpression (exp1, h);
}




/** 
    Generation of the assignment of the "false" value (bottom value) to a signal (aSig).
*/
_VIRTUAL_ void
cGenerator::genBottomValue (ptexpcal aSig)
{
  // TextForm-> printBeginComment(); TextForm->printint(EM->getNum(aSig)); TextForm-> printEndComment();
  gensig (aSig, pk_false);
  TextForm->printassign ();
  TextForm->printlogiline (pk_false);
}


/**
  Generation of the access, or the reference if islect, to a signal (exp) if it is a node of a delay tree.
  @return true if the parameter exp is a node of a delay tree, "false" otherwise.

  The output value (exp) is the equivalent signal using equivalences.
*/
_VIRTUAL_ boolean
cGenerator::genindielemdim (ptexpcal * exp, boolean islect)
{
  ptmemnode mn;

  if (!getindielemdim (exp, &mn))
    return pk_false;
  if (iscompressindex (mn))
    return pk_false;
  if (islect)
    if (EM->getNumberOfDimensions (*exp) == 0)
      TextForm->printkwref ();
  ecrsimpleindexdelay (mn);
  return pk_true;
}

/**
   Specific to the management of the nodes of delay trees.  
   It generates the incrementing (genIncrIndexDelayTreeNode()) for all the signals of a list (exp).
*/
_VIRTUAL_ void
cGenerator::genIncrIndexDelayTreeNodes (ptexpcal exp)
{
  ptexpcal e;
  if (EM->isNILEXPCAL (exp))
    return;
  if (EM->satisfiesList (exp))
    while (EM->getNaryArgs (&exp, &e))
      genIncrIndexDelayTreeNode (e);
  else
    genIncrIndexDelayTreeNode (exp);
}



/**
   MISC : WITHOUT EFFECT NOW.\n
   Specific to the management of the nodes of delay trees.
   @return  "true" if the signal exp is a node of a delay tree also used as parameter for an external call, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::copyForExternal (ptexpcal exp)
{
  ptmemnode mn;
  exp = EM->getCloneClassHead (exp);
  if (!ptexpcal_satisfies_node_tree_delay (exp, &mn))
    return pk_false;
  if (ptmemnode_get_use_in_external_call (mn) == nouse)
    return pk_false;
  return (ptmemnode_get_window_value (mn) > 1);
}

/**
   For each element of a list of signals (lexp) that satisfies copyForExternal(),
   it generates a copy ( gencopie()).
*/
_VIRTUAL_ void
cGenerator::copyForExternalList (ptexpcal lexp)
{
  ptexpcal exp;
  while (lexp != NILEXPCAL)
    {
      exp = EM->getNaryArgs (&lexp);
      if (copyForExternal (exp))
	gencopie (exp);
    }
}


/**
   It generates in reverse ordrer a list (lexp) composed of fields of structure or access to array elements (list of indexes).
   Used for extension of the operators.
*/
_VIRTUAL_ void
cGenerator::genFieldsList (ptexpcal lexp)
{
  ptexpcal exp;

  if (EM->getNaryArgs (&lexp, &exp))
    {
      genFieldsList (lexp);
      if (EM->getOperator (exp) == oplesign)
	genAccessElementArray (exp);
      else			// constant : a name of a field of a structure.
	{
	  TextForm->printDot ();
	  genConstant (exp);
	}
    }
}
