/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** @return "true" if a external call must be declared, false otherwise.*/
  _VIRTUAL_ boolean cGenerator::mustBeDeclared (ptINode aNode)
{
  ptident
    id;
  if (!NM->isMarked (aNode, NOTGREYBOXBLACKBOXREFERENCE))
    return pk_false;
  if (!NM->is_intrinsicfunc (aNode))
    return pk_true;
  if (existPredefinedPragma (NM->getPragmas (aNode), PRAGMA_CONVERSION))
    return pk_false;
  if (!getName (aNode, &id))
    return pk_true;
  if (isCartesianProductId (id))
    return pk_false;
  if (isSampleId (id))
    return pk_false;
 if (isExceptionId (id))
    return pk_false;
  return pk_true;
}

/**
   @return "true" if the node call node is annotated by the "grey box" pragma, "false" otherwise.
   @param node (in) a node call node.
   @param valPragma (out) the associated grey box pragma.
*/
_VIRTUAL_
  boolean
cGenerator::isAnnotatedGreyBox (ptINode node, ptpragma * valPragma)
{
  // Doesn't reference textform object
  if (!getPragmaStrKeyword (NM->getPragmas (node), repstrGreyBox, valPragma))
    return pk_false;
  return pk_true;
}

/**
   @param node (in) a node call node.
   @param valPragma (out) the associated "black box" pragma.
   @return "true" if the Graph of the node call node is annotated by the "black box" pragma, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::isAnnotatedBlackBoxGraph (ptINode node, ptpragma * valPragma)
{
  // Doesn't reference textform object
  ptlpragmas prags;
  ptDCG g;

  if (!NM->isNodeCall (node))
    return pk_false;
  if (getPragmaStrKeyword (NM->getPragmas (node), repstrBlackBox, valPragma))
    return pk_true;
  // ?????
  g = NM->get_nodeCall_graph (node);
  prags = g->getPragmas ();
  if (!getPragmaStrKeyword (prags, repstrBlackBox, valPragma))
    return pk_false;
  return pk_true;
}


/**
   @return "true" if it exists some external constants, "false" otherwise.
*/
_VIRTUAL_ boolean
cGenerator::TestExternalConstants ()
{
  ptlinstances lconst;
  ptexpcal vconst;
  boolean b;

  b = pk_false;
  lconst = Parameters;
  if (lconst != NILLINSTANCES)
    {
      while (ptlinstances_get_next (&lconst, &vconst))
	if (EM->getOperatorClass (vconst) == variabl)	// Filtrage des decl de types
	  return pk_true;
    }

  lconst = Constants;
  while (ptlinstances_get_next (&lconst, &vconst))
    if (EM->IsExternalConstant (vconst))
      if (mustBeGenerated (vconst)) return pk_true;
  return pk_false;
}

/** @return "true" if it exists some external types, "false" otherwise. */
_VIRTUAL_ boolean
cGenerator::TestExternalTypes ()
{
  ptlinstances ltypes;
  TypeType TT;
  ptexpcal elem;

  ltypes = Types;
  while (ptlinstances_get_next (&ltypes, &elem))
    {
      TT = EM->getExpType (elem);
      if (TypeTestExternClass (TT) || TypeTestMetaClass (TT))
	return pk_true;
    }
  return pk_false;
}


boolean
cGenerator::getName (ptINode node, ptident * id)
{
  if (NM->get_nameCall (node, id))
    return pk_true;
  if (!NM->isNodeCall (node))
    return pk_false;
  ptDCG g = NM->get_nodeCall_graph (node);
  *id = g->getName ();
  return pk_true;
}

ptlINode
cGenerator::UnicityDeclProc (ptlINode l)
{
  ptlINode ln, laux;
  ptINode node, n;
  boolean found;
  ptident id1, id2;

  ln = NILLINODE;
  while (ptlINode_get_elem (&l, &node))
    if (getName (node, &id1))
      {
	laux = ln;
	found = pk_false;
	while (!found && ptlINode_get_elem (&laux, &n))
	  if (getName (n, &id2))
	    found = eq_idents (id1, id2);
	if (!found)
	  ln = ptlINode_append_head (node, ln);
      }
  return ln;
}

/** 
   Generation of  declarations of external functions/procedures of the program.
   It generates a declaration for the node calls annotated by the NOTGREYBOXBLACKBOXREFERENCE attribute. This
   attribute is set in mkListOfBlackGreyBoxesDeclarations() method.
*/
_VIRTUAL_ void
cGenerator::genExternalProcedures ()
{
  ptident nom;
  ptexpcal linputs, loutputs, par;
  ptINode node;
  ptlINode lfonctions;

  lfonctions = UnicityDeclProc (ExternalsProcs);
  while (ptlINode_get_elem (&lfonctions, &node))
    {
      if (mustBeDeclared (node))
	{
	  NM->get_procCall_arg (node, &nom, &par, &linputs, &loutputs);
	  if (genExternalFunction (nom, linputs, loutputs, pk_true, pk_false))
	    {
	      TextForm->printRPSemiColonLine ();
	      externalObjects = pk_true;
	    }
	}
    }
}

ptlinstances
cGenerator::UnicityDeclFunc (ptlinstances l)
{
  ptlinstances ln, laux;
  ptexpcal node, n, x, y;
  boolean found;
  ptident id1, id2;

  ln = NILLINSTANCES;
  while (ptlinstances_get_next (&l, &node))
    if (EM->getFunctionCall (node, &id1, &x, &y))
      {
	laux = ln;
	found = pk_false;
	while (!found && ptlinstances_get_next (&laux, &n))
	  if (EM->getFunctionCall (n, &id2, &x, &y))
	    found = eq_idents (id1, id2);
	if (!found)
	  ln = ptlinstances_insert (ln, node);
      }
  return ln;
}

/** 
   Generation of  declarations of external functions/procedures of the program.
*/
_VIRTUAL_ void
cGenerator::genExternalFunctions ()
{
  ptident id;
  ptexpcal fc, params, inputs;
  ptlinstances lfonctions;
  boolean mode;

  lfonctions = UnicityDeclFunc (ExternalsFuncs);
  mode = functionDeclaration;
  functionDeclaration = pk_true;

  while (ptlinstances_get_next (&lfonctions, &fc))
    {
      if (!EM->isMarked (fc, INTRINSIC))
	{
	  EM->getFunctionCall (fc, &id, &params, &inputs);
	  TextForm->printdclreftypeextfunc (id, EM->getType (fc));
	  genDclFormalParameters (inputs, pk_false, pk_false, pk_false, 0);
	  TextForm->printRPSemiColonLine ();
	  externalObjects = pk_true;
	}
    }
  functionDeclaration = mode;
  TextForm->printLine ();
}

/**
   Generation of the headers of reading operations.
*/
_VIRTUAL_ void
cGenerator::genInputReadFunctions ()
{
  ptlINode ln;
  ptINode n;

  if (!NM->getInputs (&ln))
    return;
  while (ptlINode_get_elem (&ln, &n))
    genReadHeaderExterncall (NM->get_io_arg (n), pk_false);
  TextForm->printLine ();
}


/**  Generation of  headers of writing operations.
*/
_VIRTUAL_ void
cGenerator::genOutputWriteFunctions ()
{
  ptlINode ln;
  ptINode n;

  if (!NM->getOutputs (&ln))
    return;
  while (ptlINode_get_elem (&ln, &n))
    genWriteHeaderExterncall (NM->get_io_arg (n), pk_false);
  TextForm->printLine ();
}


/** 
    Generation of  declarations of external constants referenced in the program. 
 */
_VIRTUAL_ boolean
cGenerator::genExternalConstants ()
{
  ptlinstances lconst;
  ptexpcal vconst;
  boolean b;

  b = pk_false;
  lconst = Parameters;
  if (lconst != NILLINSTANCES)
    {
      genComment (comm_paramNotDefined);
      cdebdecl ();
      while (ptlinstances_get_next (&lconst, &vconst))
	if (EM->getOperatorClass (vconst) == variabl)	// Filtrage des decl de types
	  {
	    cdebdecl ();
	    genSignalDeclaration (vconst);
	    TextForm->printassign ();
	    TextForm->printkwnotimplpopf ();
	    cfindecl ();
	    b = pk_true;
	  }

    }
  lconst = Constants;
  while (ptlinstances_get_next (&lconst, &vconst))
    if (EM->IsExternalConstant (vconst))
      if (mustBeGenerated (vconst))
	{
	  TextForm->printkwdefineExternal (EM->getName (vconst));
	  b = pk_true;
	}

  if (b)
    TextForm->printLine ();
  return b;
}

/**
   It generates the declaration of an (assumed) external type (TT).
*/
_VIRTUAL_ void
cGenerator::genExternalType (TypeType TT)
{
  genTypeDeclaration (TT);
}

/** 
    Generation of  declarations of external types referenced in the program.
 */
_VIRTUAL_ boolean
cGenerator::genExternalTypes ()
{
  ptlinstances ltypes;
  TypeType TT;
  ptexpcal elem;
  boolean b;

  b = pk_false;
  ltypes = Types;
  while (ptlinstances_get_next (&ltypes, &elem))
    {
      TT = EM->getExpType (elem);
      if (TypeTestExternClass (TT) || TypeTestMetaClass (TT))
	{
	  genExternalType (TT);
	  b = pk_true;
	}
    }
  return b;
}

/** Generation of  the declarations of external objects referenced in the program.
    It generates a declaration for the node calls annotated by the NOTGREYBOXBLACKBOXREFERENCE attribute. This
    attribute is set in mkListOfBlackGreyBoxesDeclarations() method.
*/
_VIRTUAL_ void
cGenerator::genExternalNodeDeclarations ()
{
  ptexpcal OrderedParams, OrderedInputs, OrderedOutputs;
  ptlINode lNodes;
  ptINode aNode;
  ptDCG g;
  ptident nom;
  boolean asFunc;

  if (NM->getNodeCalls (&lNodes))
    {
      lNodes = UnicityDeclProc (lNodes);
      while (ptlINode_get_elem (&lNodes, &aNode))
	if (mustBeDeclared (aNode))
	  {
	    NM->get_nodeCall_arg (aNode, &g, &OrderedParams, &OrderedInputs, &OrderedOutputs);
	    nom = g->getName ();
	    asFunc = /* !existPredefinedPragma (g->getPragmas (), PRAGMA_SEPARATE) &&*/ !existPredefinedPragma (g->getPragmas (), PRAGMA_UNEXPANDED);
	    if (genExternalFunction (nom, OrderedInputs, OrderedOutputs, asFunc, pk_false))
	      {
		TextForm->printRPSemiColonLine ();
		externalObjects = pk_true;
	      }
	  }
    }
}

/** 
   Generation of declarations of external objects (constants, types, function, nodes, i/o operations) 
   Note that some of externals can be undefined; a message is send to the user in this case.
*/
_VIRTUAL_ boolean
cGenerator::genExternalDeclarations ()
{
  boolean b;

  b = (TestExternalConstants () || TestExternalTypes ());
  if (b)
    TextForm->printRefExternalInclude (progName);
  return b;

}

/**
   It generates (printincludeBodyheaderFileSeparate()) 
   the import of the header file for each element of the internal list lBlackGreyBoxesDeclarations.
   This list is assumed previously filled by the mkListOfBlackGreyBoxesDeclarations() method.

   For a name <em> foo </em> of lBlackGreyBoxesDeclarations, it generates (C language) the following line:\n
   <em> \#include foo_separate.h </em> \n
   This file <em> foo_separate.h </em> is automaticaly generated when a model (<em>foo </em>) is compiled in separated mode.
*/
_VIRTUAL_ void
cGenerator::genGreyBlackBoxesExternalDeclarations ()
{
  ptrepstr statementValue;
  ptlrepstr lstr;

  lstr = lBlackGreyBoxesDeclarations;
  while (ptlMark_get_elemNext (&lstr, &statementValue))
    TextForm->printincludeBodyheaderFileSeparate (repstr2ident (statementValue));
  TextForm->printLine ();
}

/**
   added to llow the specific declaration of a devired generator( distribution, tasks,...)
*/
_VIRTUAL_ void
cGenerator::genSpecifics()
{

}


/**
   It generates the declaration of external calls in foo_externalsProc.h (foo is the name of the current graph).
   It calls:
   - printdclexternfileProc()  opening file
   - genExternalFunctions()  functions
   - genExternalProcedures() procedures
   - genExternalNodeDeclarations() node calls
   - genGreyBlackBoxesExternalDeclarations() black/grey boxes references (separated compiling of models)
   - genInputReadFunctions() readings
   - genOutputWriteFunctions() writings
   -  closefile() closing file
*/
_VIRTUAL_ void
cGenerator::genExternalProcs ()
{
  TextForm->printdclexternfileProc (progName);
  genUserImports ();
  TextForm->printdclexternfileProcBegin (progName);
  genExternalFunctions ();
  genExternalProcedures ();
  genExternalNodeDeclarations ();
  genGreyBlackBoxesExternalDeclarations ();
  genInputReadFunctions ();
  genOutputWriteFunctions ();
  genSpecifics();
  TextForm->printdclexternfileProcEnd (progName);
}

/**
   @return "true" if it exists external constants or external types in the graph, "false" otherwise.

   It generates the external constants (genExternalConstants()) and the external types (genExternalTypes())
   of the graph. It is done when TestExternalConstants() or TestExternalTypes(), in the file 
   <em> foo_externalsUNDEF_LIS.h  </em> where foo is the name of the current graph).

   Such a file will contain lines such that:\n
   typedef UNDEF() pointer;
   
   The user must define these externals in <em>  foo_externalsUNDEF.h </em> for compiling the generated code.
*/
_VIRTUAL_ boolean
cGenerator::genExternalConstantsTypes ()
{
  boolean b;
  ptident gn;
  ptrepstr f1;

  b = (TestExternalConstants () || TestExternalTypes ());
  if (b)
    {				// Some external objects
      gn = progName;
      f1 = TextForm->printdclexternfileUser (gn);
      genExternalConstants ();
      genExternalTypes ();
      sys_filePathInUserOutputDir (f1, &f1);
      pK_messenger->messagestr2 (infoExterns1, TextForm->getExternalsFileNameUndefForUser (gn), f1);
      TextForm->closefile ();
    }
  return b;
}

/**
   For each node (n) of lNodes:
   - if the node is a "black box" reference or a "grey box" reference (Signal separated compiling), a pragma is 
   associated with it. This pragma gives the name of the external call to use. This name is added to the attribute member
   (lBlackGreyBoxesDeclarations). This list is explored  by the genGreyBlackBoxesExternalDeclarations() method to generate the
   declaration of such objects.
   - otherwise, n is marked by NOTGREYBOXBLACKBOXREFERENCE (internal attribute). This attribute is referenced in 
   genExternalProcedures() method.
*/
void
cGenerator::mkListOfBlackGreyBoxesDeclarations (ptlINode lNodes)
{
  ptINode node;
  ptpragma valPragma;
  ptrepstr externalName;

  while (ptlINode_get_elem (&lNodes, &node))
    if (NM->is_intrinsicfunc (node))
      NM->addMark (node, NOTGREYBOXBLACKBOXREFERENCE);
    else if (isAnnotatedGreyBox (node, &valPragma) || isAnnotatedBlackBoxGraph (node, &valPragma))
      {
	externalName = getPragmaTerm (getPragmaStatement (valPragma), 1);
	ptlMark_insert_elem (&lBlackGreyBoxesDeclarations, externalName);
      }
    else
      NM->addMark (node, NOTGREYBOXBLACKBOXREFERENCE);
}

/**
   It calls mkListOfBlackGreyBoxesDeclarations(ptlINode) for
   - the node calls of the program,
   - the procedure call nodes of the program.

   At the end, the list lBlackGreyBoxesDeclarations contains all the names of the external references to "black boxes" or 
   "grey boxes" models. This list will be explored  by the genGreyBlackBoxesExternalDeclarations() method to generate the
   declaration of such objects.
   
*/
_VIRTUAL_ void
cGenerator::mkListOfBlackGreyBoxesDeclarations ()
{
  ptlINode lNodes;

  lBlackGreyBoxesDeclarations = NILLREPSTR;
  if (NM->getNodeCalls (&lNodes))
    mkListOfBlackGreyBoxesDeclarations (lNodes);
  mkListOfBlackGreyBoxesDeclarations (ExternalsProcs);
}

/** 
   Entry point of the generation of the external object declarations(constants, types, ...) 
   in the file foo_externals.h
   (where foo is the name of the current graph).
   It calls:
   - mkListOfBlackGreyBoxesDeclarations()
   - genExternalProcs()
   - genExternalConstantsTypes()
   - printdclexternfile()
   - printRefExternalInclude()
   - printRefExternalIncludeProc()
   - closefile()  closing file
*/
_VIRTUAL_ void
cGenerator::genExternalDeclarationsFile ()
{
  boolean b;

  mkListOfBlackGreyBoxesDeclarations ();
  genExternalProcs ();
  b = genExternalConstantsTypes ();
  TextForm->printdclexternfile (progName);
  if (b)
    TextForm->printRefExternalInclude (progName);
  TextForm->printRefExternalIncludeProc (progName);
  TextForm->closefile ();
}
