/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @return "true" if the n-th operand of e exists (elem), "false" otherwise.
*/
  _VIRTUAL_ boolean cGenerator::getOperandAtIndex (ptexpcal e, int n, ptexpcal * elem)
{
  int
    k;
  ptexpcal
    aux;

  k = 1;
  while (k <= n)
    {
      if (!EM->getNaryArgs (&e, elem))
	return pk_false;
      k++;
    }

  *elem = getExpressionToGenerate (*elem);
  if (EM->isMarked (*elem, OPTIMIZED_INDEX))
    if (EM->getAttribute (*elem, IteratorAttribute, (attributeValue *) & aux))
      *elem = aux;

  return pk_true;
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It generates the current parametrized code pointed by a scanner (S) when it is &t (btype) or &b (not btype).
   - &t stands for "type of", it must be followed by &iK, &oK, &pK directive
   - &b stands for "base type of", it must be followed by &iK, &oK, &pK directive
   
   It generates, for &t and not for &b, when the type of the k-th element is 
   an array type the string "Array" before the code of the base type.
*/
_VIRTUAL_
  boolean
cGenerator::genTypeParametrizedCode (ptStrScanner S, ptexpcal params, ptexpcal inputs, ptexpcal outputs, boolean btype)
{
  ScanPragmaToken tk;
  int iOp;
  TypeType t, unamedType;
  ptexpcal e, elem;

  tk = S->getToken ();
  switch (tk)
    {
    case tk_param:
      e = params;
      break;
    case tk_input:
      e = inputs;
      break;
    case tk_output:
      e = outputs;
      break;
    default:
      return pk_false;
    }
  if (!S->getInt (&iOp)) return pk_false;
  if (EM->getOperatorClass (e) == call)
    elem = e; 
  else
    if (!getOperandAtIndex (e, iOp, &elem)) return pk_false;

  t = EM->getType (elem);
  unamedType = EM->getUnamedType(elem);
  if ((t != unamedType) && TypeTestPredef(unamedType)) t = unamedType;

  if (btype && TypeTestArrayClass (t)) TextForm->printstr (string2repstr ("Array"));
  return TextForm->printCodetype (t);
}


/**
   @return "true" if the current code pointed by S is an int, "false" otherwise.
   When the current code pointed by S is an int (n), it generates the n-th expression of exp.
*/
_VIRTUAL_ boolean
cGenerator::genExprParametrizedCode (ptStrScanner S, ptexpcal exp)
{
  ptexpcal elem;
  int n;
  if (!S->getInt (&n))
    return pk_false;
  if (!getOperandAtIndex (exp, n, &elem))
    return pk_false;
  genexpression (elem, opnullary);
  return pk_true;
}


/**
   @return "true" if the operation succeeds, "false" otherwise.
   It generates the parametrized code pointed by a scanner (S). It replaces
   - &n by name,
   and  (genExprParametrizedCode())
   - &pk by the k-th element of params 
   - &ik by the k-th element of intputs
   - &ok by the k-th element of outputs
   
   It also allows &t and &b (see genTypeParametrizedCode()).

*/
_VIRTUAL_ boolean
cGenerator::PrintParameter (ptStrScanner S, ptident name, ptexpcal params, ptexpcal inputs, ptexpcal outputs)
{
  ScanPragmaToken tk;
  boolean ok;


  S->setMode (pk_false);
  tk = S->getToken ();
  switch (tk)
    {
    case tk_end:
      return pk_false;

    case tk_ident:
      TextForm->printstr (S->getIdent ());
      ok = pk_true;
      break;

    case tk_name:		/* &n */
      TextForm->printident (name);
      ok = pk_true;
      break;

    case tk_param:		/* &p */
      ok = genExprParametrizedCode (S, params);
      break;
    case tk_input:		/* &i */
      ok = genExprParametrizedCode (S, inputs);
      break;
    case tk_output:		/* &o */
      ok = genExprParametrizedCode (S, outputs);
      break;

    case tk_type:		/* &t */
    case tk_baseType:		/* &b */
      ok = genTypeParametrizedCode (S, params, inputs, outputs, tk == tk_type);
      break;

    default:			// case tk_error: case tk_homo: case tk_homoClock: case tk_clock: case tk_dim:
      // case tk_cardwait: case tk_wait: case tk_done: case tk_list:
      ok = pk_false;
      break;
    }
  if (!ok)
    {
      TextForm->printkwnotimplpopf ();
      pK_messenger->errorIdent (errOnPragma, name);
    }
  return pk_true;
}


/**
   It generates the parametrized code (vtprag) in which
   - &n is replaced by name,
   - &pk is replaced by the k-th element of params
   - &ik is replaced by the k-th element of intputs
   - &ok is replaced by the k-th element of outputs
   
   To achieve this, it uses the services of StrScanner(), when a parameter (it begins with &) is detected,
   it calls PrintParameter().
*/
_VIRTUAL_ void
cGenerator::genParametrizedCode (ntrees vtprag, ptident name, ptexpcal params, ptexpcal inputs, ptexpcal outputs)
{
  boolean bcontinue;
  ptrepstr vstr;
  ptStrScanner S;		// Scanner object for parametrized code.
  boolean b;

  if (!STree_isUndefinedTree (vtprag))
    {
      S = new StrScanner ();
      S->StrScannerPragma (vtprag, pk_false);
      if (outputs != NILEXPCAL)
	if (EM->getOperatorClass (outputs) == call)
	  if (!S->Goto ('='))
	    TextForm->printkwnotimplpopf ();

      b = setnlmode (pk_false);	/* all the text in the same line, a string cannot be on several lines */
      do
	{
	  if (S->getStrUntilSeparator ('&', &vstr))
	    TextForm->printstr (vstr);
	  bcontinue = PrintParameter (S, name, params, inputs, outputs);
	}
      while (bcontinue);
      setnlmode (b);

      delete S;
    }
}

/**
   It generates a comment (numMess) parametrized by a string (id).
   numMess is an index of the current loaded table of messages.
*/
_VIRTUAL_ void
cGenerator::genParametrizedComment (int numMess, ptrepstr id)
{
  boolean bcontinue;
  ScanPragmaToken tk;
  int n;
  ptrepstr vstr;
  ptStrScanner S;		// Scanner object for parametrized code.

  S = new StrScanner ();
  S->StrScannerStr (pK_messenger->recupmessStr (numMess), pk_false);
  TextForm->printBeginComment ();
  bcontinue = pk_true;
  do
    {
      if (S->getStrUntilSeparator ('&', &vstr))
	TextForm->printstr (vstr);
      S->setMode (pk_true);
      tk = S->getToken ();
      if (tk == tk_end)
	bcontinue = pk_false;
      else
	/* &i<integer> */
      if (tk == tk_input)
	if (S->getInt (&n))
	  {
	    TextForm->printstr (id);
	  }
    }
  while (bcontinue);
  TextForm->printEndComment ();
  TextForm->printLine ();
  delete S;
}
