/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/**
   It generates the specific code for the GeenSys simulator.
*/
void
cGenerator::genSimulatorCode ()
{
  instanciationPathTableType::iterator it;
  boolean first;
  pttypeInstanciationPath anInstanciationPath;
  int n, i;
  ptexpcal id;
  ptinstanciationPathTableType instanciationPathTable;
  ptObjectCommonInfos x;
  objectType objtype;
  TypeType TT;

  if (TM->getinstanciationPathTable (&instanciationPathTable))
    {
      TextForm->printTracefilebegin (progName);
      n = (int) instanciationPathTable->size ();
      TextForm->printBeginDeclPathTable (n, maxDepthCall);
      first = pk_true;
      for (it = instanciationPathTable->begin (); it != instanciationPathTable->end (); it++)
	{
	  anInstanciationPath = (*it).first;
	  i = 0;
	  if (!first)
	    TextForm->printComma ();
	  TextForm->printLBracket ();
	  while (i < maxDepthCall)
	    {
	      TextForm->printint (anInstanciationPath[i]);
	      i++;
	      if (i < maxDepthCall)
		TextForm->printComma ();
	    }
	  first = pk_false;
	  TextForm->printRBracket ();
	  TextForm->printLine ();
	}
      TextForm->printEndDeclPathTable ();
      TextForm->printLine ();
      // Here generating the getPath() function.
      TextForm->printGetPathFunction (n, maxDepthCall);
      // Defining the  getValueByIndices() funtion.
      /*
         (for each i, element of the table);
         case i
         pM->buffer[0] = "value of the i-th identifier."
         break;
       */
      TextForm->printBeginGetValue ();
      i = 0;
      for (it = instanciationPathTable->begin (); it != instanciationPathTable->end (); it++)
	{
	  x = (*it).second;	// identifier only currently.
	  objtype = x->getwhoami ();
	  if (objtype == exprObject)
	    {
	      id = (ptexpcal) x;
	      EM->getIndirection (id, &id);	// rewritings !!!
	      if (EM->getOperatorClass (id) == variabl)
		if (mustBeGenerated (id))
		  {
		    TextForm->printCase (i);
		    TextForm->printBufferValueEqual (0); // A generaliser pour les tableaux.
		    TextForm->printcast (DRealType);
		    // genIdentifier(id, opnullary);
		    TT = EM->getUnamedType (id);
		    if (TT == StringType)
		      TextForm->printint (0);
		    else
		      {
			if (TypeTestArrayClass (TT))
			  TextForm->printcast (IntegerType);
			genexpression (id, opnullary);
		      }
		    TextForm->printSemiColonLine ();
		    TextForm->printBreakSemiColonLine ();
		  }
	    }
	  TextForm->printLine ();
	  i++;
	}

      TextForm->printEndGetValue ();
      TextForm->printLine ();
      TextForm->printBeginSetValue ();
      i = 0;
      for (it = instanciationPathTable->begin (); it != instanciationPathTable->end (); it++)
	{
	  x = (*it).second;	// identifier only currently.
	  objtype = x->getwhoami ();
	  if (objtype == exprObject)
	    {
	      id = (ptexpcal) x;
	      EM->getIndirection (id, &id);	// rewritings !!!
	      if (EM->getOperatorClass (id) == variabl)
		if (EM->IsInputFlow (id))
		  {
		    TextForm->printCase (i);
		    genIdentifier (id, opnullary);
		    TextForm->printassign ();
		    TT = EM->getUnamedType (id);	// assumed to be a predefined type. 
		    TextForm->printcast (TT);
		    TextForm->printBufferValue (0);	// A generaliser pour les tableaux.
		    TextForm->printSemiColonLine ();
		    TextForm->printBreakSemiColonLine ();
		    TextForm->printLine ();
		  }
	    }
	  i++;
	}

      TextForm->printEndSetValue ();
      TextForm->closefile ();
    }

}
