/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/*
  @return the index of the MPI Type associated with the predefined Signal type (t), -1 if it is not a prdefined type.
  WARNING: It is assumed that the boolean/event types are represented in C by a short int. If it is not, the execution will be erroneous.
*/
  int
inteprintc::getIndexTypeCode (TypeType t)
{
  switch ((int) t)
    {
    case BooleanType:
    case PureType:
      return 1;

    case IntegerType:
      return 2;

    case LongType:
      return 3;

    case ShortType:
      return 1;

    case RealType:
      return 8;

    case DRealType:
      return 9;

    case StringType:
      return 0;

    case CharType:
      return 4;

    default:
      // case ComplexType: case DComplexType:
      return -1;
    }
}

/**
   It generates the init of MPI with mult-threads:
   \code int r; MPI_Init_thread(NULL,NULL, MPI_THREAD_MULTIPLE, &p );
*/
_VIRTUAL_ void
inteprintc::printCommInit ()
{
  traceCommDirective = string2repstr ("_comm_trace_directive_");
  printstr (string2repstr ("int p;"));
  printSEini ();
  printLP ();
  printkwNULL ();
  printComma ();
  printkwNULL ();
  printComma ();
  printstr (string2repstr (MPI_LEVEL_DESIRED_THREAD_SUPPORT));
  printComma ();
  printstr (string2repstr ("&p"));
  printRPSemiColonLine ();
}

/**
   It generates the finalizing step of MPI:
   \code MPI_Finalize();
*/
_VIRTUAL_ void
inteprintc::printCommFinalize ()
{
  printSE ();
  buf->writeAlfa (kwFinal, lgkwFinal);
  printLP ();
  printRPSemiColonLine ();
}

/**
   It generates the declaration of 2 semaphores for an input/output signal (id).
*/
_VIRTUAL_ void
inteprintc::printDclSemaphoreComm (ptident id)
{
  printkwstatic ();
  printSemaphoreType ();
  printSemaphoreFull (id);
  printComma ();
  printSemaphoreEmpty (id);
  printSemiColonLine ();
}


/** Distributed code.
   It generates the code of the read instruction of a signal (id), received by communication.
   It is synchronized with the "scan" task, using semaphores.
*/
_VIRTUAL_ void
inteprintc::printBodyReceive (ptident id, boolean bref)
{
  blockIndentOpening ();
  printIOSemWait (id, pk_true);
  // memcpy(id, data_id, sizeof(data_id));
  buf->writeAlfa (kwmemcpy, lgkwmemcpy);
  printLP ();
  printident (id);
  printComma ();
  if (bref)
    printkwref ();
  printidentCommModule (id);
  printComma ();
  printkwSizeOf ();
  printLP ();
  printidentCommModule (id);
  printRP ();
  printRPSemiColonLine ();

  printIOSemSignal (id, pk_false);
  printreturn (pk_true);
  printdclfuncend ();
}


/**
   It generates the code of the write instruction of a signal (id), send by communication.
   It is synchronized with the send task, using semaphores.
*/
_VIRTUAL_ void
inteprintc::printBodySend (ptident id, boolean bref)
{
  blockIndentOpening ();
  printIOSemWait (id, pk_false);
  buf->writeAlfa (kwmemcpy, lgkwmemcpy);
  printLP ();
  if (bref)
    printkwref ();
  printidentCommModule (id);
  printComma ();
  if (bref)
    printkwref ();
  printident (id);
  printComma ();
  printkwSizeOf ();
  printLP ();
  printidentCommModule (id);
  printRP ();
  printRPSemiColonLine ();
  printIOSemSignal (id, pk_true);

  printdclfuncend ();
}


/**
   It generates an identifier (id) used in the communication module.
*/
_VIRTUAL_ void
inteprintc::printidentCommModule (ptident id)
{
  printunderscore ();
  buf->writeAlfa (kwio, lgkwio);
  printunderscore ();
  printident (id);
}

_VIRTUAL_ void
inteprintc::printInitSemaphore (ptident id, boolean boutput)
{
  printbeginSemaphoreOp ();
  buf->writeAlfa (kwIni, lgkwIni);
  printLP ();
  printSemaphoreFull (id);
  printchar (',');
  printint (0);
  printRPSemiColonLine ();

  printbeginSemaphoreOp ();
  buf->writeAlfa (kwIni, lgkwIni);
  printLP ();
  printSemaphoreEmpty (id);
  printchar (',');
  //  if (boutput) printint(1); else printint(0);
  printint (1);
  printRPSemiColonLine ();
}


_VIRTUAL_ void
inteprintc::printCreateIOTask (int TaskNumber, ptident id, boolean boutput)
{
  printTaskCreate (TaskNumber, id, pk_true);
}

_VIRTUAL_ void
inteprintc::printNumberOfIOTasks (int nb)
{
  printNumberOfTasks (nb, pk_true);
}


_VIRTUAL_ void
inteprintc::printReadEnvTaskDef (ptident id, TypeType t, ntrees commInfos)
{
  ptident idtask;
  ptrepstr tag, src;

  idtask = printIOTaskHeaderDefHead (id);
  printIOSemWait (id, pk_false);
  getCommInfos (commInfos, 1, &tag, &src);
  printIOSemSignal (id, pk_true);
  printTaskHeaderDefTail (idtask, -1);
}

/**
   wait (sem_<id>_empty);
   Receive instruction;
   signal(sem_<id>_full);

   assumed : elementary type (t)
*/
_VIRTUAL_ void
inteprintc::printReceiveTaskDef (ptident id, TypeType t, ntrees commInfos)
{
  ptident idtask;
  ptrepstr tag, src;

  idtask = printIOTaskHeaderDefHead (id);
  printIOSemWait (id, pk_false);
  getCommInfos (commInfos, 1, &tag, &src);
  printSEReveive (id, t, tag, src, pk_false);
  printIOSemSignal (id, pk_true);
  printTaskHeaderDefTail (idtask, -1);
}


/**
   Assumed : t is a n elementary type.
*/
_VIRTUAL_ void
inteprintc::printReceiveTaskDefSimple (ptident id, TypeType t, ntrees commInfos)
{
  ptident idtask;
  ptrepstr tag, src;

  blockIndentOpening ();
  getCommInfos (commInfos, 1, &tag, &src);
  printSEReveive (id, t, tag, src, pk_true);
  printreturn (pk_true);
  printCloseBlockIndent (0);
}

/**
   wait (sem_<id>_full);
   Send instruction; many dest. Moreover it can be an output of the application (environment) if bEnv.
   signal(sem_<id>_empty);
*/
_VIRTUAL_ void
inteprintc::printSendTaskDef (ptident id, TypeType t, ntrees commInfos, boolean bEnv)
{
  ptident idtask;
  ptrepstr tag, dest;
  int i;

  idtask = printIOTaskHeaderDefHead (id);
  printIOSemWait (id, pk_true);
  printSend (id, t, commInfos, pk_false);
  printIOSemSignal (id, pk_false);
  printTaskHeaderDefTail (idtask,-1);
}

/** xxxxxxx;
   Send instruction; many dest. Moreover it can be an output of the application (environment) if bEnv.
*/
_VIRTUAL_ void
inteprintc::printSendTaskDefSimple (ptident id, TypeType t, ntrees commInfos, boolean bEnv)
{
  blockIndentOpening ();
  printSend (id, t, commInfos, pk_true);
  if (!bEnv)
    printCloseBlockIndent (0);
}


/**
   It generates the specific includes for the distributed code generation.
   wTasks (with tasks, to be completed ).
 */
_VIRTUAL_ void
inteprintc::printIncludeForDistribution (boolean wTasks, boolean withDerivedTypes)
{
  if (wTasks)
    printIncludeForTasks ();
  printinclude (string2repstr ("runConfig.h"), '"', '"');
  printinclude (string2repstr ("pK_distrib.h"), '"', '"');
  printinclude (string2repstr ("string.h"), '<', '>');
  if (withDerivedTypes)
    printinclude (string2repstr (getMPITypeCode[index_MPI_derived_types]), '"', '"');
}


/**
   Communication component.
   It opens (openOut_file()) the file associated with a model (processName) 
   for the generation of IO operations. Moreover, it generates the code of some importations (printincludefiles()).
   For C, with processName=foo, the name of the file is: foo_IO.c
   It generates in foo_IO.c the lines:
   \code 
   // Polychrony banner
   #include "foo_types.h"
   #include "foo_externals.h"
   \endcode 
*/
_VIRTUAL_ void
inteprintc::printdclCommfilehead (ptident processName)
{
  openOut_file (mkNameWithLangsuffix (getCommFileName (processName)), mess_gen_io);
  printincludefiles (processName);
}


/**
   Communication component.

   It generates the beginning of the definition of the initilization of the communication component.
   \code
   For C, it generates   
   EXTERN void foo_OpenIO()
   {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclbegcommfunchead ()
{
  cwriextCvoid ();
  printLinedclfunchead (begcommname);
}


/**
   Communication component.
   It generates the header of the definition of the finalizing of the communication operations of the current graph.
   
   \code
    For C, it generates
    EXTERN void foo_CloseIO()
    {
 \endcode
*/
_VIRTUAL_ void
inteprintc::printdclendcommfunchead ()
{
  cwriextCvoid ();
  printLinedclfunchead (endcommname);
}

_VIRTUAL_ void
inteprintc::printDeclCommunicationCounter (ptident id)
{
  printkwstatic ();
  printkwint ();
  printSpace ();
  printCommCounter (id);
  printequal ();
  printint (0);
  printchar (';');
  printLine ();
}


 _VIRTUAL_ void
 inteprintc::printHeaderCommunicationCounter()
 {
   printexternvoidentrypoint(string2ident (PRINTCOMMCOUNTERS_ID));
 }



_VIRTUAL_ void
inteprintc::printCommunicationCounterBegin ()
{
  printdclrefvoidextCfunc(string2ident (PRINTCOMMCOUNTERS_ID));
  printchar (')');
  printLine ();
  blockIndentOpening ();
  printDirectiveIfDefined (traceCommDirective);
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  buf->writeChar ('(');
  buf->writeAlfa (kwstderr, lgkwstderr);
  buf->writeChar (',');
  printchar ('"');
  printident (string2ident (" ---------------------"));
  printchar (escapeSep);
  printchar ('n');
  printchar ('"');
  printRP ();
  printSemiColonLine ();
}

_VIRTUAL_ void
inteprintc::printNumberCommunicationsBegin ()
{
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  buf->writeChar ('(');
  buf->writeAlfa (kwstderr, lgkwstderr);
  buf->writeChar (',');
  printchar ('"');
  printstr (progName);
  buf->printChar (':');
  printident (string2ident (" Number of  communications from start= %d"));
  printchar (escapeSep);
  printchar ('n');
  printchar ('"');
  printchar (',');
}

_VIRTUAL_ void
inteprintc::printNumberCommunicationsEnd ()
{
  printRP ();
  printSemiColonLine ();
}

_VIRTUAL_ void
inteprintc::genCommunicationNamePlus (ptident id, boolean bplus)
{
  if (bplus)
    printkwplus ();
  printCommCounter (id);
}

_VIRTUAL_ void
inteprintc::printEndIfDirectiveCommunicationCounter ()
{
  printDirectiveEndIf (traceCommDirective);
}

_VIRTUAL_ void
inteprintc::printTailCommunicationCounter ()
{
  printblockclosing ();
}


/**
   It generates printf(" <progName>, nb communications for <id> = %d \n", cpt(<id>));
*/
_VIRTUAL_ void
inteprintc::genCommunicationCounter (ptident id)
{
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  buf->writeChar ('(');
  buf->writeAlfa (kwstderr, lgkwstderr);
  buf->writeChar (',');
  printchar ('"');
  printstr (progName);
  buf->printChar (',');
  printident (string2ident (" number of SEND for "));
  printident (id);
  printequal ();
  printSpace ();
  printCodeFormat (IntegerType);
  printchar (escapeSep);
  printchar ('n');
  printchar ('"');
  printchar (',');
  printCommCounter (id);
  printRP ();
  printSemiColonLine ();
}

_VIRTUAL_ void
inteprintc::printCallToCommunicationCounter ()
{
  printident (string2ident(PRINTCOMMCOUNTERS_ID));
  printchar ('(');
  printRPSemiColonLine ();
}


_VIRTUAL_ void
inteprintc::openConfigFile ()
{
  openOut_file (string2repstr ("runConfig.h"), mess_gen_io);

}

_VIRTUAL_ void
inteprintc::closeConfigFile ()
{
  closefile ();
}

_VIRTUAL_ void
inteprintc::printStepCounter ()
{
  buf->writeAlfa(kwcpt, lgkwcpt);
  printstr(progName);
}


_VIRTUAL_ void
inteprintc::printIncrStepCounter ()
{
  printStepCounter ();
  printkwplus ();
  printkwplus ();
  printchar (';');
  printLine ();
}

_VIRTUAL_ void
inteprintc::DeclStepCounter ()
{
  printkwstatic ();
  printkwint ();
  printSpace ();
  printStepCounter ();
  printequal ();
  printint (0);
  printchar (';');
  printLine ();
}


void inteprintc::printIdentRendezVous(boolean bDecl, ptident nom, ptident idSend, ptident idReceive, ntrees PragmaValuesIdReceive)
{
  ptrepstr tag, src;
  if (bDecl)  cwriextCvoid ();
  printident (nom); 
  printchar ('_');
  printstr(progName); 
  printchar ('_');
  printident (idSend);
  printchar ('_');
  getCommInfos (PragmaValuesIdReceive, 1, &tag, &src);
  printstr(src);
  printchar ('_');
  printident (idReceive);
}


void inteprintc::printSendTo( ptrepstr dest)
{
  printchar ('w');
  printchar ('_');
  printstr (progName);
  printchar ('_');
  printstr (dest);
}
