/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
// ------------ Depending on the support of the execution (MPI)
#define kwSupportExec "MPI"
#define lgkwSupportExec 3
#define kwReceive "Recv"
#define lgkwReceive 4
#define kwSend "Send"
#define lgkwSend 4
#define kwSEIni "Init_thread"
#define lgkwSEIni 11
#define kwFinal "Finalize"
#define lgkwFinal 8
#define counterName_prefix "_n_b_c_o_m_m_s_"
#define MPI_STATUS_IGNORE "MPI_STATUS_IGNORE"
#define MPI_COMM_DEFAULT "MPI_COMM_WORLD"
#define MPI_LEVEL_DESIRED_THREAD_SUPPORT "MPI_THREAD_MULTIPLE"
#define index_MPI_Datatype 13
#define index_MPI_Build_Derived_Type 14
#define index_MPI_Type_free 15
#define index_MPI_derived_types 16
#define PRINTCOMMCOUNTERS_ID "pK_printCommCounters_"
#define kwcpt "_cptSteps"
#define lgkwcpt 9

static const char *getMPITypeCode[17] = {
  "MPI_CHAR",
  "MPI_SHORT",			// WARNING: It is assumed that the boolean/event types are represented in C by a short int. If it is not, the execution will be erroneous.
  "MPI_INT",
  "MPI_LONG",
  "MPI_UNSIGNED_CHAR",
  "MPI_UNSIGNED_SHORT",
  "MPI_UNSIGNED",
  "MPI_UNSIGNED_LONG",
  "MPI_FLOAT",
  "MPI_DOUBLE",
  "MPI_LONG_DOUBLE",
  "MPI_BYTE",
  "MPI_PACKED",
  "MPI_Datatype",		//index_MPI_Datatype
  "MPI_Build_Derived_Type_",	//index_MPI_Build_Derived_Type
  "MPI_Type_free",		//index_MPI_Type_free
  "MPI_derived_types.c"
};

_VIRTUAL_ void
inteprintc::printSE ()
{
  buf->writeAlfa (kwSupportExec, lgkwSupportExec);
  printunderscore ();
}

_VIRTUAL_ void
inteprintc::printSEini ()
{
  printSE ();
  buf->writeAlfa (kwSEIni, lgkwSEIni);
}


_VIRTUAL_ void
inteprintc::printkwReceive ()
{
  printSE ();
  buf->writeAlfa (kwReceive, lgkwReceive);
}


_VIRTUAL_ void
inteprintc::printkwSend ()
{
  printSE ();
  buf->writeAlfa (kwSend, lgkwSend);
}


_VIRTUAL_ void
inteprintc::printSEName (ptrepstr vstr, boolean bref)
{
  if (bref)
    printkwref ();
  printSE ();
  printstr (vstr);
}


_VIRTUAL_ void
inteprintc::printCommType (TypeType t)
{
  int i;
  ptrepstr typeName;

  i = getIndexTypeCode (t);
  if (i >= 0)
    printkwTable (i);
  else if (getTypeName (t, &typeName))
    printSEName (typeName, pk_false);
  else
    printkwnotimpl ();
}

_VIRTUAL_ void
inteprintc::prinkwCommunicator ()
{
  printstr (string2repstr (MPI_COMM_DEFAULT));
}


_VIRTUAL_ void
inteprintc::prinkwCommStatus ()
{
  printstr (string2repstr (MPI_STATUS_IGNORE));
}


void
inteprintc::printkwTable (int vindex)
{
  printstr (string2repstr (getMPITypeCode[vindex]));
}

/** 
    @return 
    - "true" if a type (t) is not a predefined type for MPI, and its name (typeName) is known; 
      in this case it generates the declaration of a derived type(t) for MPI.
    - "false" otherwise.
 */
_VIRTUAL_ boolean
inteprintc::printDeclarationDerivedType (TypeType t, ptrepstr * typeName)
{
  if (!getTypeName (t, typeName))
    return pk_false;
  printkwTable (index_MPI_Datatype);
  printSpace ();
  printSE ();
  printstr (*typeName);
  printSemiColonLine ();
  return pk_true;
}

/** 
    "MPI_Build_derived_type_foo(id, &MPI_foo);"
    where foo is the name of the type t 
*/
_VIRTUAL_ void
inteprintc::printmkDerivedType (TypeType t, ptrepstr typeName, ptident id, boolean bdirect, boolean bsend)
{
  printkwTable (index_MPI_Build_Derived_Type);
  printstr (typeName);
  printLP ();
  printIdComm (id, t, bdirect, bsend);
  printComma ();
  printSEName (typeName, pk_true);
  printRPSemiColonLine ();
}


/*  MPI_Type_free(&MPI_foo); where foo is the name of the type t */
_VIRTUAL_ void
inteprintc::printFinalDerivedType (TypeType t, ptrepstr typeName)
{
  printkwTable (index_MPI_Type_free);
  printLP ();
  printSEName (typeName, pk_true);
  printRPSemiColonLine ();
}

_VIRTUAL_ void
inteprintc::printSEReveive (ptident id, TypeType t, ptrepstr tag, ptrepstr src, boolean bdirect)
{
  boolean b;
  ptrepstr typeName;

  b = printDeclarationDerivedType (t, &typeName);	/* MPI_Datatype MPI_Message_type; */
  if (b)
    printmkDerivedType (t, typeName, id, bdirect, pk_false);	/*   Build_derived_type(response, &MPI_Message_type); */

  printkwReceive ();
  printLP ();
  printIdComm (id, t, bdirect, pk_false);
  printComma ();
  if (t != StringType)
    printint (1);
  else;
  printComma ();
  printCommType (t);
  printComma ();
  printstr (src);
  printComma ();
  printstr (tag);
  printComma ();
  prinkwCommunicator ();
  printComma ();
  prinkwCommStatus ();
  printRPSemiColonLine ();
  if (b)
    printFinalDerivedType (t, typeName);
}

_VIRTUAL_ void
inteprintc::printIdComm (ptident id, TypeType t, boolean bdirect, boolean bsend)
{
  if (bsend)
    {
      if (t != StringType)
	printkwref ();
      if (bdirect)
	printident (id);
      else
	printidentCommModule (id);
    }
  else
    {				// for receive
      if (bdirect)
	printident (id);
      else
	{
	  if (t != StringType)
	    printkwref ();
	  printidentCommModule (id);
	}
    }
}

/* Generating a send of id, of type t, using the canal tag, to dest xxx .\n
   For example, for id="OK", t= "boolean" or "short", tag="9", dest="equationSolving_2" and bdirect="true", it generates the following code:
   MPI_Send(&OK,1,MPI_SHORT,equationSolving_2,9,MPI_COMM_WORLD);
  */
_VIRTUAL_ void
inteprintc::printSESend (ptident id, TypeType t, ptrepstr tag, ptrepstr dest, boolean bdirect)
{
  boolean b;
  ptrepstr typeName;

  b = printDeclarationDerivedType (t, &typeName);
  if (b)
    printmkDerivedType (t, typeName, id, bdirect, pk_true);
  printkwSend ();
  printLP ();
  printIdComm (id, t, bdirect, pk_true);
  printComma ();
  if (t != StringType)
    printint (1);
  printComma ();
  printCommType (t);
  printComma ();
  printstr (dest);
  printComma ();
  printstr (tag);
  printComma ();
  prinkwCommunicator ();
  printRPSemiColonLine ();

  if (b)
    printFinalDerivedType (t, typeName);
}


/**
   @return the string foo_io, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getCommFileName (ptident processName)
{
  ptrepstr name;
  if (makeRepstr_ident_string (processName, COMM_SUFFIX, &name))
    return name;
  return nilstr;
}


/**
   It generates the send operations defined in commInfos (many dest) for the idendifier id of type t.
*/
void
inteprintc::printSend (ptident id, TypeType t, ntrees commInfos, boolean bdirect)
{
  ptrepstr tag, dest;
  int i, n;

  i = 1;
  n = 0;
  while (getCommInfos (commInfos, i, &tag, &dest))
    {
      n++;
      printSESend (id, t, tag, dest, bdirect);
      i = i + 2;
    }
  if (n > 0)
    genIncrCounter (id, n);

}

void
inteprintc::printCommCounter (ptident id)
{
  ptrepstr vstr;
  vstr = string2repstr (counterName_prefix);
  buf->fit (strsize (vstr) + identsize (id, 1));
  printstr (vstr);
  printident (id);
}

void
inteprintc::genIncrCounter (ptident id, integer indexCounter)
{
  printCommCounter (id);
  printkwincr ();
  printchar (';');
  printLine ();
}
