/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_


_VIRTUAL_ boolean
inteprintc::getCommInfos (ntrees tEnonce, int ifrom, ptrepstr * tag, ptrepstr * pnode)
{
  ptrepstr p;
  int nb;

  *tag = nilstr;
  *pnode = nilstr;
  nb = ntreearity (tEnonce);
  if (nb < ifrom)
    return pk_false;
  *tag = getIthPramaStatementWithoutQuotes (tEnonce, ifrom);
  if (nb >= ifrom + 1)
    {
      // src or dest depending on the pragma name.
      *pnode = getIthPramaStatementWithoutQuotes (tEnonce, ifrom + 1);
    }
  return pk_true;
}

/**
   It opens (openOut_file()) the file for the generation of the declaration of the tasks. 
   It generates the include of the header file associated with the scheduler.
*/
  _VIRTUAL_ void
inteprintc::printTaskDeclarations (ptident processName)
{
  openOut_file (getTaskDeclarationFileName (processName), mess_tasks_dcl);
  printIncludeForTasks ();
}

/**
   It generates the include of the header associated with the tasks.
   \code
   #include "FOO_tasks.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printIncludeTaskDeclarations (ptident processName)
{
  printinclude (getTaskDeclarationFileName (processName), '"', '"');
  printLine ();
}

/**
   It prints:
   - the reference to a task (&id_Task) when bref 
   - the name of the task (id_Task) otherwise.
*/
_VIRTUAL_ void
inteprintc::printTaskName (ptident id, boolean bref)
{
  if (bref)
    printkwref ();
  printidentunderscore (id);
  buf->writeAlfa (kwTaskType, lgkwTaskType);
}

/**
   It prints the number of tasks (nb) of the application.
*/
_VIRTUAL_ void
inteprintc::printNumberOfTasks (int nb)
{
  printcomment (string2repstr ("number of tasks"));
  printkwdefine ();
  printSpace ();
  buf->writeAlfa (kwNumberOftasks, lgkwNumberOftasks);
  printSpace ();
  printint (nb);
  printLine ();

  printNumberOfTasks (nb, pk_false);
}

/**
   It prints the number of semaphores (nb) used for the synchronisation of the tasks.
*/
_VIRTUAL_ void
inteprintc::printDclOfSemaphores (int nb)
{
  printcomment (string2repstr ("Declaration of the semaphores."));
  printbeginSemaphoreOp ();
  buf->writeAlfa (kwDcl, lgkwDcl);
  printunderscore ();
  buf->writeAlfa (kwArray, lgkwArray);
  printLP ();
  printint (nb);
  printRPSemiColonLine ();
}


/**
   It generates the code of the creation of the \b TaskNumber-th task called \b idTask.
*/
_VIRTUAL_ void
inteprintc::printTaskCreate (int TaskNumber, ptident idTask, ptrepstr kw)
{
  printTaskCreate (TaskNumber, idTask, pk_false, kw);
}

/**
   It generates the declaration of the header of a task called \b id.
*/
_VIRTUAL_ void
inteprintc::printTaskHeaderDcl (ptident id)
{
  printTaskHeader (id);
  printEndDeclaration ();
}

/**
   It generates the beginning of the code of a task.
   - the header,
   - the "while" loop, with an opening of a block,
   - and some trace informations included in the while block.
*/
_VIRTUAL_ void
inteprintc::printTaskHeaderDef (ptident id, integer numTask)
{
  printTaskHeader (id);
  printLine ();
  blockIndentOpening ();
  printWhileTrueBlock ();

  printTrace(id);
  // printTraceNumTask(numTask, pk_true);
}

/**
   It generates the initialisation of the sem-th semaphore to semIniVal value.
   (ie pK_sem_ini(sem, semIniVal))
*/
_VIRTUAL_ void
inteprintc::printSemaphoreIni (int sem, int semIniVal)
{
  printbeginIndexedSemaphoreOp ();
  buf->writeAlfa (kwIni, lgkwIni);
  printLP ();
  printint (sem);
  printchar (',');
  printint (semIniVal);
  printRPSemiColonLine ();
}

/**
   It generates the wait instruction associated with the TaskNumber-th task.
   (ie pK_sem_Wait(TaskNumber))
*/
_VIRTUAL_ void
inteprintc::printTaskWait (int TaskNumber)
{
  printbeginIndexedSemaphoreOp ();
  buf->writeAlfa (kwWait, lgkwWait);
  printLP ();
  printint (TaskNumber);
  printRPSemiColonLine ();

}

/**
   It generates the code that signals the end of an instant for the TaskNumber_th task.
   (ie pK_sem_Signal(TaskNumber))
*/
_VIRTUAL_ void
inteprintc::printTaskSignal (int TaskNumber, int numTaskPred)
{
  printbeginIndexedSemaphoreOp ();
  buf->writeAlfa (kwSignal, lgkwSignal);
  printLP ();
  printint (TaskNumber);
  printRPSemiColonLine ();
  
  // if(numTaskPred>=0) fprintf(stdout, "%d  -> %d ;\n", numTaskPred, TaskNumber );  // "dot format"
}

/**
   It generates the tail of the code of a task called \b id.
*/
_VIRTUAL_ void
inteprintc::printTaskHeaderDefTail (ptident id, integer numTask)
{
  printTrace(makeIdent_string_ident("End of ", id));
  printTraceNumTask(numTask, pk_false);

  blockclose ();
  blockclose ();
}

/**
   It generates the code of the end of a task (done by itself).
   (ie pK_Task_exit(" ");)
*/
_VIRTUAL_ void
inteprintc::printTaskSuicide ()
{
  printkwPolychrony (pk_true);
  buf->writeAlfa (kwTaskType, lgkwTaskType);
  printunderscore ();
  buf->writeAlfa (kwexit, lgkwexit);

  printLP ();
  printchar ('"');
  printSpace ();
  printchar ('"');
  printRPSemiColonLine ();
}

/**
   It generates the "main" code of the program, it waits the end of the \b indexIterateTask task (the iterate task).
*/
_VIRTUAL_ void
inteprintc::printTasksMainEntry (int indexIterateTask, boolean bdist)
{
  printexternvoidentrypoint (begioname);
  printexternvoidentrypoint (endioname);
  printLine ();
  printexternvoidentrypoint (begcommname);
  printexternvoidentrypoint (endcommname);
  printLine ();

  printMainFunct ();
  printMainDeclaration ();
  printLineidemptycall (begioname);
  if (bdist)
    printLineidemptycall (begcommname);
  printCallIterate (inifuncname);
  // Start
  printLineidemptycall (getIdStart ());

  // Wait for end od application (join).
  printkwPolychrony (pk_true);
  buf->writeAlfa (kwjoin, lgkwjoin);

  buf->printChar ('(');
  printint (indexIterateTask);
  printRPSemiColonLine ();
  printLineidemptycall (endioname);
  if (bdist)
    printLineidemptycall (endcommname);
  printdclfuncend ();
}


_VIRTUAL_ void
inteprintc::printSemaphoreType ()
{
  printbeginSemaphoreOp ();
  buf->writeAlfa (kwtype, lgkwtype);
  printSpace ();
}

/**
   It generates the name of the semaphore (full) associated with id.
*/
_VIRTUAL_ void
inteprintc::printSemaphoreFull (ptident id)
{
  printunderscore ();
  buf->writeAlfa (kwSemaphore, lgkwSemaphore);
  printidentunderscore (id);
  buf->writeAlfa (kwfull, lgkwfull);
}

/**
   It generates the name of the semaphore (empty) associated with id.
*/
_VIRTUAL_ void
inteprintc::printSemaphoreEmpty (ptident id)
{
  printunderscore ();
  buf->writeAlfa (kwSemaphore, lgkwSemaphore);
  printidentunderscore (id);
  buf->writeAlfa (kwempty, lgkwempty);
}




/**
   It generates the specific includes for the distributed code generation.
 */
_VIRTUAL_ void
inteprintc::printIncludeForTasks ()
{
  printinclude (string2repstr ("pK_semaphores.h"), '"', '"');
  printinclude (string2repstr ("pK_threads.h"), '"', '"');
}




/**
   It generates an identifier (id) between underscores.
   \code
   Example: with id=test, it generates 
   _test_
   \endcode
*/
_VIRTUAL_ void
inteprintc::printidentunderscore (ptident id)
{
  buf->fit (identsize (id, 1) + 2);
  printunderscore ();
  printidentUnderscore (id);
}

_VIRTUAL_ void
inteprintc::printTerminationTaskHeaderDef ()
{
  printTaskHeader (finalfuncname);
  printLine ();
  blockIndentOpening ();
}

_VIRTUAL_ void
inteprintc::printkwreturnNull ()
{
  printkwreturn ();
  printSpace ();
  printkwNULL ();
  printSemiColonLine ();
}

/**
   It generates the reference to a primitive (kw) for the \b TaskNumber-th task called \b idTask.
   Example : for kw) cancel, it generates pK_task_cancel(TaskNumber, idTask).
*/
_VIRTUAL_ void
inteprintc::printCallOsPrimitive (int TaskNumber, ptident idTask, ptrepstr kw)
{
  printkwPolychrony (pk_true);
  buf->writeAlfa (kwTask, lgkwTask);
  printunderscore ();
  printstr (kw);
  printLP ();
  printint (TaskNumber);
  printchar (',');
  printidentunderscore (idTask);
  buf->writeAlfa (kwTask, lgkwTask);
  printRPSemiColonLine ();
}


void inteprintc::blockclose ()
{
  backindent (1);
  printblockclosing ();
  printLine ();
}

/**
   It generates the declaration of the interface of an entry point named foo_kw, where foo is the name of the current application, 
   that returns an object with t as type, 
   and without any parameter, followed by the opening of the body statement.
   The declaration is prefixed by EXTERN if bext.
   \code
   Example: with kw=start and t=BooleanType, bext=true it generates
   EXTERN logical foo_start() {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printkwFunctionHeadNoParameters (ptrepstr kw, TypeType t, boolean bext)
{
  printFunctionHeadNoParameters (mkFuncName (kw), t, bext);
}


_VIRTUAL_ void inteprintc::printTraceTaskAfterWait (ptident id, integer numTask)
{
  // printTrace( makeIdent_string_ident(" *** After waiting:", id));
  printTraceNumTask(numTask, pk_true);  
}


_VIRTUAL_ void inteprintc::printCancelTask ()
{
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  buf->writeChar ('(');
  buf->writeAlfa (kwstderr, lgkwstderr);
  buf->writeChar (',');
  buf->writeAlfa (strcancelTasks, lgstrcancelTasks);
  printRPSemiColonLine ();
}
