/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define COMM_SUFFIX "_comm"
#define kwNumberOftasks "NB_TASKS"
#define lgkwNumberOftasks 8
#define kwtype "type"
#define lgkwtype 4
#define kwIndexSemaphore "Isem"
#define lgkwIndexSemaphore 4
#define kwSemaphore "sem"
#define lgkwSemaphore 3
#define kwDcl "dcl"
#define lgkwDcl 3
#define kwArray "array"
#define lgkwArray 5
#define kwIni "ini"
#define lgkwIni 3
#define kwTasksSuffix "_tasks"
#define lgkwTasksSuffix  6
#define kwStart "start"
#define lgkwStart 5
#define kwTaskType "Task"
#define lgkwTaskType 4
#define kwWait "wait"
#define lgkwWait 4
#define kwSignal "signal"
#define lgkwSignal 6
#define kwOs "os"
#define lgkwOs 2
#define kwjoin "join"
#define lgkwjoin 4
#define kwtrace "trace"
#define lgkwtrace 5
#define kwfull "full"
#define lgkwfull 4
#define kwempty "empty"
#define lgkwempty 5
#define kwdecl "decl"
#define lgkwdecl 4
#define kwTask "Task"
#define lgkwTask 4
#define kwmemcpy "memcpy"
#define lgkwmemcpy 6
#define kwio "io"
#define lgkwio 2
#define kwEnd "end"
#define lgkwEnd 3
#define kwtraceInt "traceInt"
#define lgkwtraceInt 8

#define strcancelTasks  "\"Cancelling Tasks\\n\""
#define lgstrcancelTasks  20

#define getIdStart() mkFuncName(kwStart)


/**
   @return the name FOO_tasks where FOO is the value of \b processName.
*/
  ptrepstr inteprintc::getTaskDeclarationFileName (ptident processName)
{
  return makeStr_fileName (processName, kwTasksSuffix, H_SUFFIX);
}


/**
   It generates the header of the declaration of a task named id_Task (when bSuffTask), id otherwise.
   For C language, it generates:
   - static void* id_Task(void*) when bSuffTask,
   - static void* id(void*) when not(bSuffTask).
*/
_VIRTUAL_ void
inteprintc::printTaskHeader (ptident id)
{
  printkwPolychrony (pk_true);
  buf->writeAlfa (kwdecl, lgkwdecl);
  printunderscore ();
  buf->writeAlfa (kwTask, lgkwTask);
  printchar ('(');
  printTaskName (id, pk_false);
  printchar (')');
}

/**
   It generates the code
   \code
   while (pk_true)
   {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printWhileTrueBlock ()
{
  printkwWhile ();
  buf->printChar ('(');
  printkwlogicons (pk_true);
  buf->printChar (')');
  printLine ();
  blockIndentOpening ();
}


_VIRTUAL_ void
inteprintc::printbeginSemaphoreOp ()
{
  printkwPolychrony (pk_true);
  buf->writeAlfa (kwSemaphore, lgkwSemaphore);
  printunderscore ();
}


_VIRTUAL_ void
inteprintc::printbeginIndexedSemaphoreOp ()
{
  printkwPolychrony (pk_true);
  buf->writeAlfa (kwIndexSemaphore, lgkwIndexSemaphore);
  printunderscore ();
}

/**
   It generates the code of the creation of the \b TaskNumber-th task called \b idTask.
   It is a io task if ioTask, else a cluster task.
*/
_VIRTUAL_ void
inteprintc::printTaskCreate (int TaskNumber, ptident idTask, boolean ioTask, ptrepstr kw)
{

  printkwPolychrony (pk_true);
  if (ioTask)
    {
      buf->writeAlfa (kwio, lgkwio);
      printunderscore ();
    }
  buf->writeAlfa (kwTask, lgkwTask);
  printunderscore ();
  printstr (kw);

  printLP ();
  printint (TaskNumber);
  printchar (',');
  if (ioTask)
    {
      printunderscore ();
      buf->writeAlfa (kwio, lgkwio);
    }
  printidentunderscore (idTask);
  buf->writeAlfa (kwTask, lgkwTask);
  printRPSemiColonLine ();
}

_VIRTUAL_ void
inteprintc::printNumberOfTasks (int nb, boolean ioTask)
{
  printcomment (string2repstr ("Declaration of the tasks."));

  printkwPolychrony (pk_false);
  if (ioTask)
    {
      printunderscore ();
      buf->writeAlfa (kwio, lgkwio);
    }
  buf->writeAlfa (kwTasksSuffix, lgkwTasksSuffix);
  printunderscore ();
  buf->writeAlfa (kwDcl, lgkwDcl);
  printunderscore ();
  buf->writeAlfa (kwArray, lgkwArray);

  printLP ();
  printint (nb);
  printRPSemiColonLine ();
}


_VIRTUAL_ void
inteprintc::printIOSemWait (ptident id, boolean bfull)
{
  printbeginSemaphoreOp ();
  buf->writeAlfa (kwWait, lgkwWait);
  printLP ();
  if (bfull)
    printSemaphoreFull (id);
  else
    printSemaphoreEmpty (id);
  printRPSemiColonLine ();
}

_VIRTUAL_ void
inteprintc::printIOSemSignal (ptident id, boolean bfull)
{
  printbeginSemaphoreOp ();
  buf->writeAlfa (kwSignal, lgkwSignal);
  printLP ();
  if (bfull)
    printSemaphoreFull (id);
  else
    printSemaphoreEmpty (id);
  printRPSemiColonLine ();
}

_VIRTUAL_ ptident
inteprintc::getionametask (ptident id)
{
  ptrepstr vstr;
  strconcchastr ('_', ident2repstr (id), &vstr);
  strconc2str (string2repstr (kwio), vstr, pk_false, &vstr);
  return repstr2ident (vstr);
}

_VIRTUAL_ ptident
inteprintc::printIOTaskHeaderDefHead (ptident id)
{
  ptident idtask;

  idtask = getionametask (id);
  printTaskHeader (idtask);
  printLine ();
  blockIndentOpening ();
  printWhileTrueBlock ();
  return idtask;
}



/**
   Generates trace (defines in the pK_threads.h library) for debugging.
   pK_trace(vstr)

 */
void inteprintc::printTrace(ptident vstr)
{
  printkwPolychrony (pk_true);
  buf->writeAlfa (kwtrace, lgkwtrace);
  printLP ();
  printchar ('"');
  printTaskName (vstr, pk_false);
  printchar ('"');
  printRPSemiColonLine ();
}

/**
   Generates trace (defines in the pK_threads.h library) for debugging:
   - pK_trace_task_begin(numTask) if bbegin
   - pK_trace_task_end(numTask) otherwise.
 */
void inteprintc::printTraceNumTask(integer numTask, boolean bbegin)
{
  ptrepstr vstr;
  if (numTask >= 0)
    {
      if (bbegin) vstr= string2repstr("pK_trace_task_begin"); else vstr = string2repstr("pK_trace_task_end");
      printstr (vstr);
      printLP ();
      printchar ('"');
      printstr(progName);
      printchar (':');
      printint(numTask);
      printchar ('"');
      printRPSemiColonLine ();

    }
}
