/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/** 
    Constructor.
*/
inteprintc::inteprintc ()
{
  VERSION_WITH_VCD= pk_false;
}

/**
   Destructor.
*/
_VIRTUAL_ inteprintc::~inteprintc ()
{
}

_VIRTUAL_ void 
inteprintc::printQuote()
{
  printchar ('\'');
}


_VIRTUAL_ void
inteprintc::printPO()
{
  printchar ('(');
}

_VIRTUAL_ void
inteprintc::printPF()
{
  printchar (')');
}


_VIRTUAL_ void
inteprintc::printcolon()
{
  printchar (':');
}

/** 
    It generates the pow operator.
 */
_VIRTUAL_ void
inteprintc::printPow ()
{
  printstr (string2repstr (kwpow));
}



/**
   It writes the complex type.
*/
_VIRTUAL_ void
inteprintc::printkwComplex ()
{
  printstr (string2repstr (kwSignalComplexType));
}



/** It writes the test of equality for arrays when beq, or test of less than for arrays otherwise. */
_VIRTUAL_ void
inteprintc::printArrayTest (boolean beq)
{
  printstr (string2repstr ("Array"));
  if (beq)
    printstr (string2repstr ("Eq"));
  else
    printstr (string2repstr ("Leq"));
}

/** It writes the sizeOf key word.*/
_VIRTUAL_ void
inteprintc::printkwSizeOf ()
{
  printstr (string2repstr (kwSizeOf));
}



/**
   It generates the beginning of a block with an indentation.
   It can be closed by the printCloseBlockIndent() method.
*/
_VIRTUAL_ void
inteprintc::printOpenBlockIndent ()
{
  printLine ();
  printchar (' ');
  printchar (' ');
  printLBracket ();
  printLine ();
  indent ();
}

/** It writes the "structure" key word.*/
_VIRTUAL_ void
inteprintc::printkwStruct ()
{
  printstr (string2repstr (kwStruct));
}

/** It writes the beginning of the declaration of a structure type.*/
_VIRTUAL_ void
inteprintc::printBeginStructure ()
{
  printkwStruct ();
  printOpenBlockIndent ();
}

/**
   It generates the end of a block previoulsy opened with an indentation
   (such that printOpenBlockIndent()).
*/
_VIRTUAL_ void
inteprintc::printCloseBlockIndent (int nbspaces)
{
  int i;
  backindent (1);
  for (i = 0; i < nbspaces; i++)
    printSpace ();
  printblockclosing ();
}

/** It writes the closing of the declaration of a structure type.*/
_VIRTUAL_ void
inteprintc::printEndStructure ()
{
  printCloseBlockIndent (2);
}


/** It writes a line composed of the "not yet implemented" key word. */
_VIRTUAL_ void
inteprintc::printnotimplemented ()
{
  printLine ();
  printkwnotimpl ();
  printLine ();
}


/** It writes an identifier (id).
    For a field identifier the separator of the suffixes is the dot char.
*/
_VIRTUAL_ void
inteprintc::printidentfield (ptident id)
{
  printstr (ident2repstr_sep (id, '.'));
}


/** It writes an identifier, composed by a prefix and a list of suffixes, by printing the prefix followed by the suffixes separated with a separator (vsep);
    p_s1_s2_s3 for example with _ for vsep, p the prefix, si the suffixes.
 */
_VIRTUAL_ void
inteprintc::printlident (ptident id, ptrepstr vsep)
{
  int i, n;

  printident (namidfrl (id, 1));
  n = ntreearity (id);
  for (i = 2; i < n + 1; i++)
    {
      printstr (vsep);
      printident (namidfrl (id, i));
    }
}

/** It prints i as an index (i.e. the reserved code for an index followed by i, I123 for example if I is the reserved code. */
_VIRTUAL_ void
inteprintc::printindex (int i)
{
  ptrepstr str;

  str = integer2repstr (i);
  buf->fit (strsize (str) + 1);
  buf->writeChar (carindice);
  buf->writeRepstr (str);
}

/**
   It prints:
   - a reserved code: index when isindice, or delay-array prefix otherwise. It is used to represent delay trees.
   - the int i
   - the identifier id.
*/
_VIRTUAL_ void
inteprintc::printidfromchaintid (boolean isindice, int i, ptident id)
{
  ptrepstr str1, str2;

  str2 = ident2repstr (id);
  if (i != 0)
    {
      str1 = integer2repstr (i);
      buf->fit (strsize (str1) + strsize (str2) + 1);
    }
  else
    buf->fit (strsize (str2) + 1);
  if (isindice)
    buf->writeChar (carindice);
  else
    buf->writeChar (cardelay);
  if (i != 0)
    cwristr (str1);
  cwristr (str2);
}

/**
   @return "true" if id is the int, float, double cast, "false" otherwise.
*/
_VIRTUAL_ boolean
inteprintc::isCastFunction (ptident nom)
{
  ptrepstr strnom;
  strnom = ident2lowerRepstr (nom);
  return ((strnom == strtypeint) || (strnom == strtypefloat) || (strnom == strtypedouble));
}

/**
   It releases a declaration.
 */
_VIRTUAL_ void
inteprintc::printreleasedcl ()
{
  buf->printChar (';');
  buf->backMark ();
  printLine ();
}

/**
   It prints the define key word (definition of a macro).
*/
_VIRTUAL_ void
inteprintc::printkwdefine ()
{
  buf->printAlfa (kwdefine, lgkwdefine);
}

/**
   It prints the definition of an external symbol.\n
   \code
   For C, it generates
   #define id UNDEF()
   \endcode
*/
_VIRTUAL_ void
inteprintc::printkwdefineExternal (ptident id)
{
  printkwdefine ();
  printSpace ();
  printident (id);
  printSpace ();
  printkwnotimplpopf ();
  printLine ();
}

/**
   It prints the include of a file (nameInclude) with sepBegin as prefix and sepEnd as suffix.\n
   \code
   For C, with sepBegin=", sepEnd=" it generates
   #include "nameInclude"
   with sepBegin=<, sepEnd=> it generates
   #include <nameInclude>
   \endcode
*/
_VIRTUAL_ void
inteprintc::printinclude (ptrepstr nameInclude, char sepBegin, char sepEnd)
{
  printstr(string2repstr (kwinclude));
  printSpace ();
  printchar (sepBegin);
  printstr (nameInclude);
  printchar (sepEnd);
  printLine ();
}



/**
   It prints the or operator.
*/
_VIRTUAL_ void
inteprintc::printkwor ()
{
  buf->print2Chars ('|', '|');
}

/**
   It prints the xor operator.
*/
_VIRTUAL_ void
inteprintc::printkworexclusif ()
{
  buf->printChar ('^');
}

/**
   It prints the and operator.
*/
_VIRTUAL_ void
inteprintc::printkwand ()
{
  buf->print2Chars ('&', '&');
}

/**
   It prints the not operator.
*/
_VIRTUAL_ void
inteprintc::printkwnot ()
{
  buf->printChar ('!');
}

/** It prints the "less than" operator.*/
_VIRTUAL_ void
inteprintc::printkwLessthan ()
{
  buf->printChar ('<');
}

/** It prints the "equal" operator. */
_VIRTUAL_ void
inteprintc::printequal ()
{
  buf->writeChar ('=');
}


/** It prints the "less than or equal" operator. */
_VIRTUAL_ void
inteprintc::printkwLessthanEqual ()
{
  buf->print2Chars ('<', '=');
}

/** It prints the "equal test" operator. */
_VIRTUAL_ void
inteprintc::printkwequal ()
{
  buf->print2Chars ('=', '=');
}

/** It prints the "greater than or equal" operator. */
_VIRTUAL_ void
inteprintc::printkwGreaterthanEqual ()
{
  buf->print2Chars ('>', '=');
}

/** It prints the "modulo" operator. */
_VIRTUAL_ void
inteprintc::printkwmod ()
{
  buf->printChar ('%');
}


/** It prints the "greater" operator. */
_VIRTUAL_ void
inteprintc::printkwGreaterthan ()
{
  buf->printChar ('>');
}

/** It prints the "diff" operator. */
_VIRTUAL_ void
inteprintc::printkwdiff ()
{
  buf->print2Chars ('!', '=');
}


/** It prints the "++" operator. */
_VIRTUAL_ void
inteprintc::printkwincr ()
{
  buf->print2Chars ('+', '+');
}

/** It prints the "--" operator. */
_VIRTUAL_ void
inteprintc::printkwdecr ()
{
  buf->print2Chars ('-', '-');
}

/** It prints the "->" operator. */
_VIRTUAL_ void
inteprintc::printObjectReference ()
{
  buf->print2Chars ('-', '>');
}

/** It prints the "==" Signal operator (equality operator for not prefefined type).
    It prints EQ string.
*/
_VIRTUAL_ void
inteprintc::printkwegeg ()
{
  // == Signal operator
  buf->print2Chars ('E', 'Q');
}

/** It prints the "<<" Signal operator (inclusion test operator for not prefefined type).
    It prints LE string.
*/
_VIRTUAL_ void
inteprintc::printkwiieg ()
{
  // <<= Signal operator
  buf->print2Chars ('L', 'E');
}


/**
   It prints *id if star, id otherwise.
*/
_VIRTUAL_ void
inteprintc::printidentstar (ptident id, boolean star)
{
  if (star)
    printstar ();
  printident (id);
}

/**
   It prints &id if star, id otherwise.
   For a field the separator of the suffixes is dot char otherwise the underscore char.
*/
_VIRTUAL_ void
inteprintc::printidentref (ptident id, boolean cref, boolean bfield)
{
  if (cref)
    printkwref ();
  if (!namislident (id))
    if (bfield)
      printidentfield (id);
    else
      printident (id);
  else
    printlidentDotSeparator (id);
}

/**
   It prints the "static" key word.
*/
_VIRTUAL_ void
inteprintc::printkwstatic ()
{
  buf->printAlfa (kwstatic, lgkwstatic);
  buf->space ();
}


/**
   It prints the boolean constant "true" when cb "false" otherwise,  followed by ';'.
   The line is printed.
*/
_VIRTUAL_ void
inteprintc::printlogiline (boolean cb)
{
  printkwlogicons (cb);
  printchar (';');
  printLine ();
}

/**
   It generates the key-word for a return statement.
*/
_VIRTUAL_ void
inteprintc::printkwreturn ()
{
  buf->printAlfa (kwreturn, lgkwreturn);
}



/**
   It prints the return key-word following by printlogiline().
*/
_VIRTUAL_ void
inteprintc::printreturn (boolean cb)
{
  printkwreturn ();
  buf->space ();
  printlogiline (cb);
}

/**
   It prints the assign operator.
*/
_VIRTUAL_ void
inteprintc::printassign ()
{
  buf->fit (3);
  buf->space ();
  printequal ();
  buf->space ();
}

/**
   It prints the += operator.
*/
_VIRTUAL_ void
inteprintc::printassignincr ()
{
  buf->fit (4);
  buf->space ();
  printkwplus ();
  printequal ();
  buf->space ();
}

/**
   It prints "id + 1" if cplus, "id - 1" otherwise.
*/
_VIRTUAL_ void
inteprintc::printindexinarrayprocess (ptident id, boolean cplus)
{
  printident (id);
  if (cplus)
    printplusOne ();
  else
    buf->print2Chars ('-', '1');
}


/**
   It begins the declaration of a read function (id) of a process (processName as name).
   It calls cwriextCBoolean()  and printReadFunctionLP().
   \code
   For C, it generates, with id=a, processName=foo
      EXTERN logical r_foo_a
   \endcode
*/
_VIRTUAL_ void
inteprintc::printDclEnvReadFuncBegin (ptident id, ptident processName)
{
  cwriextCBoolean ();
  buf->space ();
  printReadFunctionLP (id, processName);
}


/**
   It begins the declaration of a write function (id) of a process (processName as name).
   It calls cwriextCvoid() and printWriteFunctionLP().
    \code
    For C, it generates, with id=a, processName=foo
      EXTERN void w_foo_a(
    \endcode
*/
_VIRTUAL_ void
inteprintc::printDclEnvWriteFuncBegin (ptident id, ptident processName)
{
  cwriextCvoid ();
  printWriteFunctionLP (id, processName);
}

/**
   It prints a "static" function declaration (name) without any parameters, 
   in separated mode when bsep. It is the definition when bdef.
   Moreover, a line is printed when bcr.
   \code
   Example:
   When not bsep, it prints for foo_STEP_initialize as name
   static void foo_STEP_initialize()
   When bsep, it prints for foo as name
   static void foo_STEP_initialize(foo *data)
   \endcode
*/
_VIRTUAL_ void
inteprintc::printstaticfunction (ptident name, boolean bdef, boolean bsep)
{
  printkwstatic ();
  printkwvoid ();
  printident (name);
  printchar ('(');
  if (bsep)
    printClassDataArea ();
  printchar (')');
  if (bdef)
    printLine ();
}

/**
   It prints the begining of the call to the "initialize" entry point of a program named name.\n
   With \e foo as name, it generates: \e foo_initialize(
*/
_VIRTUAL_ void
inteprintc::printInitCallModuleBegin (ptident name)
{
  printidentUnderscore (name);
  printstr (inifuncSuffix);
  printchar ('(');
}

/**
   It prints the call to the "initialize" entry point of a program named name, in separated mode when sep.\n
   With \e foo as name, it generates: \e foo_initialize();  or \e foo_initialize(foo);
*/
_VIRTUAL_ void
inteprintc::printinitcallmodule (ptident name, boolean sep)
{
  printInitCallModuleBegin (name);
  if (sep)
    printident (name);
  printRPSemiColonLine ();
}

/**
   It prints the call to the "initialize" entry point of a program, named name. \n
   Example: foo_iterate();
*/
_VIRTUAL_ void
inteprintc::printstepcallmodule (ptident name)
{
  printidentUnderscore (name);
  printstr (iterfuncSuffix);
  printchar ('(');
  printRPSemiColonLine ();
}


/** It prints the not implemented key word (printkwnotimpl()) followed by "()" string. */
_VIRTUAL_ void
inteprintc::printkwnotimplpopf ()
{
  printkwnotimpl ();
  printLPRP ();
  printSpace ();
}

/**
   It prints the key word for a type declaration (typedef in C).
*/
_VIRTUAL_ void
inteprintc::printkwTypedef ()
{
  buf->writeAlfa (kwtypedef, lgkwtypedef);
  buf->space ();
}

/** 
    It generates the declaration of an identifier (name) as an external type.
    \code
    For C, it generates
    typedef UNDEF() foo;
    \endcode
*/
_VIRTUAL_ void
inteprintc::printdclexterntype (ptrepstr name)
{
  printkwTypedef ();
  printkwnotimplpopf ();
  cwristr (name);
  printreleasedcl ();
}


/**
   It prints the opening of a loop. 
   \code 
   For C, it generates 
   for (
   \endcode 
*/
_VIRTUAL_ void
inteprintc::printForLP ()
{
  printkwfor ();
  buf->space ();
  buf->writeChar ('(');
}


/**
   It prints the opening of a loop with id as indice. 
   \code 
   For C, With id=i, it generates
   for (i= 
   \endcode 
   \see printForLP().
*/
_VIRTUAL_ void
inteprintc::printforident (ptident id)
{
  printForLP ();
  printident (id);
  printassign ();
}

/**
   It prints the key word "else", writes the lines and signals an indentation.
*/
_VIRTUAL_ void
inteprintc::printelseindent ()
{
  printkwelse ();
  printLine ();
  indent ();
}


/**
   It generates the name of the file descriptor associated with and identifier (id). 
   It is obtained by charFile + 'r' or 'w' + id, where charFile is an internal name, 'r' when bread, 'w' when not bread.

   Examples: It generates
   - frFoo for a read of Foo (id=Foo)
   - fwFoo for a write of Foo (id=Foo)
*/
_VIRTUAL_ void
inteprintc::printfilevarname (ptident id, boolean bread)
{
  boolean b;

  b = !namislident (id);
  if (b)
    buf->fit (identsize (id, 1) + 2);
  else;				// must be completed
  buf->writeChar (charFile);
  printReadWriteCode (bread);
  if (b)
    cwristr (ident2repstr (id));
  else
    printlidentUnderscoreSeparator (id);
}

/**
   It prints the beginning of a write operation, on a file depending on id;  
   the definition of the write-format based on the type (idt). \n
   idt is a predefined type or an enumeration type. \n
   \see printfilevarname(), printformat().
    \code 
    For C, it generates with id=error, and idt=realType
    fprintf(fwerror,"%d", 
    \endcode 
*/
_VIRTUAL_ void
inteprintc::printbodywritehead (ptident id, TypeType idt)
{
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  printchar ('(');
  printfilevarname (id, pk_false);
  printchar (',');
  printformat (idt, pk_false);
  printchar (',');
}

/**
   It prints the beginning of a write operation, which calls a function (idWrite) using a file descriptor composed from idpar.
   It is used for composed types. For the elementary types see printbodywritehead().
   \code 
   It generates with idWrite=pK_write_type_test1_Type1, and idpar=test1_w1
   pK_write_type_test1_Type1(fwtest1_w1,
   \endcode 
*/
_VIRTUAL_ void
inteprintc::printbodywriteheadComposedType (ptident idWrite, ptident idpar)
{
  printident (idWrite);
  printLP ();
  printfilevarname (idpar, pk_false);
  printComma ();
}

/**
   It prints the end of a write operation, which calls a function using a file descriptor composed from idpar.
   It is used for composed types. For the elementary types see printbodywritetail().
*/
_VIRTUAL_ void
inteprintc::printbodywritetailComposedType (ptident idpar)
{
  printRP ();
  printbodywritetail (idpar);
}

/**
   It generates the end of the declaration of a function.
   It empties the buffer.
*/
_VIRTUAL_ void
inteprintc::printdclfuncend ()
{
  buf->backTabulation ();
  buf->printChar ('}');
  printLine ();
  printLine ();
}

/**
   It generates the end of the declaration of a read function: it prints a "return pk_true;" statement, and calls
   printdclfuncend().
*/
_VIRTUAL_ void
inteprintc::printtruefuncend ()
{
  printreturn (pk_true);
  printdclfuncend ();
}

/**
   It finalizes the current statement (printRPSemiColonLine())
*/
_VIRTUAL_ void
inteprintc::printbodywritetail (ptident id)
{
  printRPSemiColonLine ();
}

/** 
    For C, the "space" between two values in the generated file, is integrated in the format.
*/
_VIRTUAL_ void
inteprintc::printValueSeparator ()
{
}

_VIRTUAL_ void
inteprintc::printbodywritefinal (ptident id)
{
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  printchar ('(');
  printfilevarname (id, pk_false);
  printchar (',');
  printchar ('"');
  printchar (escapeSep);
  printchar ('n');
  printchar ('"');
  printRPSemiColonLine ();
  printflushOn (id);
}

/**
   It generates the closing of n loops (in C, n '}' characters are generated).
*/
_VIRTUAL_ void
inteprintc::printforloopend (int n)
{
  int i;
  if (n > 0)
    {
      for (i = 1; i <= n; i++)
	printchar ('}');
      printLine ();
    }
}

/**
   It generates the beginning of a loop characterized by an indice number (numind) and an upper bound (bsup).
   \code
   or C, with numind=56 and bsup=n, it generates 
   { int I56; for ( I56=0; I56<n; I56++)
   \endcode
*/
_VIRTUAL_ void
inteprintc::printforloophead (int numind, int bsup)
{
  {
    printchar ('{');
    printkwIntSpace ();
    printindex (numind);
    buf->printChar (';');
  }
  // for (numind = 0 , numind < bsup, numind++)
  printForLP ();
  printindex (numind);
  printassign ();
  buf->printChar ('0');
  buf->printChar (';');
  printindex (numind);
  printkwLessthan ();
  printint (bsup);
  buf->printChar (';');
  printindex (numind);
  printkwincr ();
  printRPlineindent ();
}


/**
   It prints the declaration of an assumed enumerated type (t).
*/
_VIRTUAL_ void
inteprintc::printdclenumtype (TypeType t)
{
  ptrepstr enumName;

  printkwTypedef ();
  printkwEnum ();
  enumName = EnumerationTypeGetName (t);
  printListEnum (t);
  cwristr (enumName);
  printreleasedcl ();
}

/**
   It prints the cast on an identifier id.
*/
_VIRTUAL_ void
inteprintc::printcast (TypeType id)
{
  printchar ('(');
  printtype (id, pk_true);
  printchar (')');
}

/**
   @return "true" if the operation succeeds, "false" otherwise.
   It prints the "code" associated with a type (t).
   For an enumeration type, it calls printkwEnum(), otherwise it calls printtype().
   For an array type, it looks the type of the elements.
*/
_VIRTUAL_ boolean
inteprintc::printCodetype (TypeType t)
{
  if (TypeTestEnumerationClass (t))
    {
      printkwEnum ();
      return pk_true;
    }
  if (TypeTestArrayClass (t))
    {
      t = ArrayMatrixTypeGetTerm (t);
    }
  return printtype (t, pk_false);
}


/** @return "true" if the operation succeeds, "false" otherwise.
   It prints a type (t) followed by a space character when space. 
   It calls  printnotimplemented() when the type is unknown.
*/
_VIRTUAL_ boolean
inteprintc::printtype (TypeType t, boolean space)
{
  ptrepstr typeName;

  t = getTypeToGenerate (t);

  if (t == UndefinedType)
    {
      printkwvoid ();
      return pk_true;
    }

  if (TypeTestPredef (t))
    {
      if (t == ErrorType)
	{
	  ERRORTYPE ();
	  return pk_false;
	}
      printtypePredef (t);
      if (space)
	printSpace ();
      return pk_true;
    }

  if (TypeTestArrayClass (t))
    {
      printtype (ArrayMatrixTypeGetTerm (t), space);
      return pk_true;
    }

  if (space)
    printSpace ();
  if (getTypeName (t, &typeName))
    {
      printstr (typeName);
      if (space)
	printSpace ();
      return pk_true;
    }
  printnotimplemented ();
  return pk_false;
}

/**
   @return "true" if the name (typeName) of a non predefined type (t) is known, "false" otherwise.
*/
_VIRTUAL_ boolean
inteprintc::getTypeName (TypeType t, ptrepstr * typeName)
{
  if (TypeTestPredef (t))
    return pk_false;
  if (TypeTestArrayClass (t))
    return getTypeName (ArrayMatrixTypeGetTerm (t), typeName);
  if (TypeTestExternClass (t))
    {
      *typeName = ExternTypeGetName (t);
      return pk_true;
    }
  if (TypeTestMetaClass (t))
    {
      *typeName = MetaTypeGetName (t);
      return pk_true;
    }
  if (TypeTestEnumerationClass (t))
    {
      *typeName = EnumerationTypeGetName (t);
      return pk_true;
    }
  if (TypeTestNamedClass (t))
    {
      *typeName = NamedTypeGetName (t);
      return pk_true;
    }
  //if (TypeTestStructureClass(t))   { *typeName = StructureTypeGetNameLst(t); b = pk_true; }
  return pk_false;

  /*
     if (TypeTestSubrangeClass(t)) return printnotimplemented();
     if (TypeTestBitsClass(t)) return printnotimplemented();
     if (TypeTestLambdaClass(t)) return printnotimplemented();
     if (TypeTestNodeClass(t)) return printnotimplemented();
     if (TypeTestGenericNodeClass(t)) return printnotimplemented();
     if (TypeTestMetaNodeClass(t)) return printnotimplemented();
     if (TypeTestMatrixClass(t)) return printnotimplemented();;
     if (TypeTestTupleClass(t)) return printnotimplemented();
   */
}

/**
   It prints the beginning of the declaration of an external function with id as name without return status.
   It calls cwriextCvoid().
   \code
   For C,  with id=foo, it generates,
          EXTERN void foo( 
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclrefvoidextCfunc (ptident id)
{
  cwriextCvoid ();
  printident (id);
  printchar ('(');
}

/**
   It prints the beginning of the declaration of an external function with id as name with t as return status type.
   It calls cwrikwexternC().
   \code
   In C,  with id=foo, and t=intType ), it generates,
   EXTERN int foo( 
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclreftypeextfunc (ptident nom, TypeType t)
{
  cwrikwexternC ();
  printtype (t, pk_true);
  printident (nom);
  printchar ('(');
}

/**
   It prints the beginning of the declaration of the initializing entry point, in separated mode when bsep.
   It calls printdclreftypeextfunc(), printClassDataArea() when bsep.
   \code
   In C, it generates, with \b foo is the name of the current model, \b data is a predefined name \n
    - EXTERN logical foo_initialize() {
    - EXTERN logical foo_initialize(foo* data) {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclinitializefunchead (boolean bsep)
{
  printdclreftypeextfunc (inifuncname, BooleanType);
  if (bsep)
    printClassDataArea ();
  printchar (')');
  printLine ();
  blockIndentOpening ();
}

/** Header file.\n
    It generates the declaration of the header of the iterate step of a code generated by Polychrony.
    It calls : cwriextCBoolean() followed by printstepcallmodule().
    Example:(C language)
    - EXTERN logical foo_iterate();
*/
_VIRTUAL_ void
inteprintc::printheaderstepcallmodule (ptident progName)
{
  cwriextCBoolean ();
  printstepcallmodule (progName);
}

/**
   Header file.\n
   It generates the declaration of the header of the initialize step of a code generated by Polychrony.
   Example:(C language)
   - EXTERN logical foo_initialize();  // when not sep
   - EXTERN logical foo_initialize(foo *);  // when sep
*/
_VIRTUAL_ void
inteprintc::printheaderinitialize (ptident progName, boolean sep)
{
  cwriextCBoolean ();
  printInitCallModuleBegin (progName);
  if (sep)
    {
      printident (progName);
      printstar ();
    }
  printRPSemiColonLine ();
}

/**
   Header file.\n
   It generates the declaration of the header of the initializing  function of a step of a code generated by Polychrony.
   Example: (C language)
   - static void foo_STEP_initialize(); // when not bsep
   - static void foo_STEP_initialize(foo *data); // when bsep, data is a predefined name
*/
_VIRTUAL_ void
inteprintc::printheaderentrypointstepinitialize (ptident graphName, boolean bsep)
{
  ptident idStepFinalize;

  idStepFinalize = make_stepInitIdent (graphName);
  printstaticfunction (idStepFinalize, pk_false, bsep);
  printEndDeclaration ();
}

/** Header file.\n
    It generates the declaration of the header of the finalizing function of a step of a code generated by Polychrony.

    Example: (C language)
    - EXTERN logical foo_STEP_finalize(); // when not bsep
    - EXTERN logical foo_STEP_finalize(foo *);// when bsep

*/
_VIRTUAL_ void
inteprintc::printheaderentrypointstepfinalize (ptident progName, boolean bsep)
{
  ptident id;

  id = make_stepFinalIdent (progName);
  printHeaderFunction (id, BooleanType, pk_true);
  printLP ();
  if (bsep)
    printDataParamReference (progName);
  printRP ();
  printchar (';');
  printLine ();
}


/**
   It opens (openOut_file()) the body-header file associated with a model (processName) in separated mode (when sep).
   For C, 
   - with processName=foo, sep=false, the name of the file is foo_body.h 
   - with processName=foo, sep=true, the name of the file is foo_body_separate.h
*/
_VIRTUAL_ void
inteprintc::printdclbodyfileheader (ptident processName, boolean sep)
{
  ptrepstr name;

  if (sep)
    name = getHeaderName_BodySeparate (processName);
  else
    name = getHeaderName_Body (processName);
  openOut_file (name, mess_gen_bodyExternal);
}

/**
   It generates the inclusion of the foo_body.h file, where foo is the value of processName.
   \code
   For C, with processName=foo, it generates 
   #include "foo_body.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printincludeBodyheaderFile (ptident processName)
{
  ptrepstr name;
  name = getHeaderName_Body (processName);
  printinclude (name, '"', '"');
}

/**
   It generates the inclusion of the foo_body_separate.h file, where foo is the value of processName.
   \code
    For C, with processName=foo, it generates
    #include "foo_body_separate.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printincludeBodyheaderFileSeparate (ptident processName)
{
  printinclude (getTypesFileName (processName), '"', '"'); 
  printinclude (getHeaderName_BodySeparate (processName), '"', '"');
}

/**
   It generates the declaration of the iterate entry + the opening of the body statement.
   \code 
   For C, with foo the name of the current model, it generates
   EXTERN logical foo_iterate() {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printiterfunchead ()
{
  printFunctionHeadNoParameters (iterfuncname, BooleanType, pk_true);
}

_VIRTUAL_ void
inteprintc::printiterfuncEnd ()
{
  printtruefuncend ();
}


/**
   It generates the import of the type and external declarations of a graph with graphName as name.
   \code
   For C, it generates, with graphName=foo 
   #include "foo_types.h"
   #include "foo_externals.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printincludefiles (ptident graphName)
{
  printinclude (getTypesFileName (graphName), '"', '"');
  printinclude (getExternalsFileName (graphName), '"', '"');
}

/**
   It generates the beginning of the "step finalize" entry point of a graph  with graphName as name, and in separated mode if bsep.
   \code
   For C, it generates, with graphName=foo and
   - with bsep=false
        EXTERN logical foo_STEP_finalize() {        
   - with bsep=true, data is an internal predefined name
        EXTERN logical foo_STEP_finalize(foo *data) { 
   \endcode
*/
_VIRTUAL_ void
inteprintc::printentrypointstepfinalize (ptident graphName, boolean bsep)
{
  ptident id;

  id = make_stepFinalIdent (graphName);
  printdclreftypeextfunc (id, BooleanType);
  if (bsep)
    printClassDataArea ();
  printchar (')');
  printLine ();
  blockIndentOpening ();
}

/**
   It generates the call to the "step finalize" entry point of a graph  with graphName as name, and in separated mode if bsep.
   \code
   For C, it generates, with graphName=foo and 
   - with bsep=false
         foo_STEP_finalize();
   - with bsep=true, data is an internal predefined name
         foo_STEP_finalize(data);
   \endcode
*/
_VIRTUAL_ void
inteprintc::printcallentrypointstepfinalize (ptident graphName, boolean bsep)
{
  ptident id;
  id = make_stepFinalIdent (graphName);
  if (bsep)
    cgenCallSeparate (id);
  else
    printLineidemptycall (id);
}

/**
   It generates the call to the "step initialize" entry point of a graph  with graphName as name, and in separated mode if bsep.
   \code
   For C, it generates, with graphName=foo and 
   - with bsep=false
       foo_STEP_initialize();
   - with bsep=true, data is an internal predefined name  
       foo_STEP_initialize(data);
   \endcode
*/
_VIRTUAL_ void
inteprintc::printcallentrypointstepinitialize (ptident graphName, boolean bsep)
{
  ptident id;

  id = make_stepInitIdent (graphName);
  if (bsep)
    cgenCallSeparate (id);
  else
    printLineidemptycall (id);
}

/**
   It generates the beginning of the "step initialize" entry point of a graph  with graphName as name, and in separated mode if bsep.
   \code
   For C, it generates, with graphName=foo and 
    - with bsep=false
      static void foo_STEP_initialize() { 
    - with bsep=true, data is an internal predefined name
      static void foo_STEP_initialize(foo *data) { 
   \endcode
*/
_VIRTUAL_ void
inteprintc::printentrypointstepinitialize (ptident graphName, boolean bsep)
{
  ptident idStepFinalize;

  idStepFinalize = make_stepInitIdent (graphName);
  printstaticfunction (idStepFinalize, pk_true, bsep);
  blockIndentOpening ();
}

/**
   It generates the opening of the file for the generation of the body of a model with \e ProcessName as name, in separated mode if  \e separate .
   Moreover, it generates the declaration of some imports (printincludefiles() and printinclude()).
   \code
   For C, it generates, with ProcessName=foo, in the file foo_body.c when not  separate, foo_body_separate.c otherwise, the following lines
   #include "foo_types.h"
   #include "foo_externals.h"
   #include "foo_body.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printbodyfilebegin (ptident ProcessName, boolean separate)
{
  ptrepstr cFileName;
  if (separate)
    cFileName = makeStr_fileName (ProcessName, C_BODY_SEPARATE_SUFFIX, C_SUFFIX);
  else
    cFileName = makeStr_fileName (ProcessName, C_BODY_SUFFIX, C_SUFFIX);
  openOut_file (cFileName, mess_gen_body);
}

/**
   It generates the opening of the file for the generation of the body of a model with \e ProcessName as name, in separated mode if  \e separate .
   Moreover, it generates the declaration of some imports (printincludefiles() and printinclude()).
   \code
   For C, it generates, with ProcessName=foo, in the file foo_body.c when not  separate, foo_body_separate.c otherwise, the following lines
   #include "foo_types.h"
   #include "foo_externals.h"
   #include "foo_body.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printbodyfileIncludes(ptident ProcessName, boolean separate)
{
  ptrepstr hFileName;

  printincludefiles (ProcessName);
  if (separate)
      hFileName = makeStr_fileName (ProcessName, C_BODY_SEPARATE_SUFFIX, H_SUFFIX);
  else
      hFileName = makeStr_fileName (ProcessName, C_BODY_SUFFIX, H_SUFFIX);

  printinclude (hFileName, '"', '"');
  printLine ();
}

/**
   It generates the beginning of the definition of the body of a read of an identifier (idarg) characterized by its type (idt).
   It calls printscancallheadLP().
   \code
   For C, with idarg=a, idt=realType (real), it generates the following sequence
   if (fscanf(fra,"%f",
   \endcode
*/
_VIRTUAL_ void
inteprintc::printbodyreadhead (ptident idarg, TypeType idt)
{
  printkwif ();
  buf->writeChar ('(');
  printscancallheadLP (idarg, codeReadConst, idt);
}

/**
   It generates the end of the definition of the body of a read (elementary type) of an identifier (idarg). 
   The parameter (idarg) is not used for C.
   \code
   It generates the following sequence
   )==EOF)return FALSE;
   return TRUE;
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclreadtail (ptident idarg)
{
  printchar (')');
  printkwequal ();
  buf->printAlfa (kweof, lgkweof);
  printRP ();
  printreturn (pk_false);
}

/**
   It prints the beginning of a read operation, which calls a function (idRead) using a file descriptor composed from idpar.
   It is used for composed types. For the elementary types see printbodyreadhead().
   \code
   For C, with idRead=pK_read_type_test1_Type1, idpar=test1_r1, it generates the following sequence
   if (!pK_read_type_test1_Type1(frtest1_r1,
   \endcode
*/
_VIRTUAL_ void
inteprintc::printbodyreadheadComposedType (ptident idRead, ptident idpar)
{
  printkwif ();
  printLP ();
  printkwnot ();
  printident (idRead);
  printLP ();
  printfilevarname (idpar, pk_true);
  printComma ();
}

/**
   It prints the end of a read operation, which calls a function using a file descriptor composed from idpar.
   It is used for composed types. For the elementary types see printbodyreadtail().
   The parameter (idpar) is not used for C.
   \code
   It generates the following sequence
   )) return FALSE;
   \endcode
*/
_VIRTUAL_ void
inteprintc::printbodyreadtailComposedType (ptident idpar)
{
  printRP ();
  printRP ();
  printSpace ();
  printreturn (pk_false);
}

/**
   It prints 
   - &(idpar->field) if field is defined, 
   - &ipar if field is not defined and bref
   - idpar otherwise.
*/
_VIRTUAL_ void
inteprintc::printReadRefIdField (ptident idpar, ptrepstr field, boolean bref, boolean inCall)
{
  if (field != nilstr)
    {
      printkwref ();
      printLP ();
      printident (idpar);
      printObjectReference ();
      printstr (field);
      printRP ();
    }
  else
    {
      if (bref)
	printkwref ();
      printident (idpar);
    }
}

/**
   It prints idpar.field if field is defined, idpar otherwise.
*/
_VIRTUAL_ void
inteprintc::printWriteRefIdField (ptident idpar, ptrepstr field)
{
  printident (idpar);
  if (field != nilstr)
    {
      printDot ();
      printstr (field);
    }
}

/**
   It generates the opening of a file for a read when bInput, or a write operation otherwise. 
   The name of the file is built from nomfich and bInput.

   \code
   For C, it generates 
   - with nomfich=a, bInput=true
   fra = fopen("Ra.dat","rt");
   if (!fra)
   {
      fprintf(stderr,"Can not open file %s\n","Ra.dat");
      exit(1);
   }

   - with nomfich=x1, bInput=false (it is an output.)
   fwx1 = fopen("Wx1.dat","wt");
    if (!fwx1)
        {
        fprintf(stderr,"Can not open file %s\n","Wx1.dat");
        exit(1);
        }
  \endcode
*/
_VIRTUAL_ void
inteprintc::printfileopen (ptident nomfich, boolean bInput)
{
  // nomfich = identList2ident(nomfich); 
  printfilevarname (nomfich, bInput);
  printassign ();
  buf->fit (lgkwfopen + 1);
  buf->writeAlfa (kwfopen, lgkwfopen);
  buf->writeChar ('(');
  printsysfilename (nomfich, bInput);
  buf->printChar (',');
  printiomode (bInput);
  printRPSemiColonLine ();
  printkwif ();
  buf->writeChar ('(');
  printkwnot ();
  printfilevarname (nomfich, bInput);
  printRPlineindent ();
  printblockopening ();
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  buf->writeChar ('(');
  buf->writeAlfa (kwstderr, lgkwstderr);
  buf->writeChar (',');
  buf->writeAlfa (strcannotopen, lgstrcannotopen);
  printchar (',');
  printsysfilename (nomfich, bInput);
  printRPSemiColonLine ();
  printExit(1);
  printBlockClosingBackIndent ();
  printLine ();
}


_VIRTUAL_  void inteprintc::printExit(int val)
{
  buf->writeAlfa (kwexit, lgkwexit);
  buf->writeChar ('(');
  printint(val);
  buf->writeChar (')');
  printSemiColonLine ();
}

/**
   It prints the header of the declaration of an I/O operation for a composed type (t).
   
   @param t (in) the type of the signal to read/write.
   @param fname (in) the name of the function
   @param id (in) the name of the formal parameter
   @param binput (in) a read function when it is "true", a write function when it is "false".
   \code
   with t=test1_Type1, fname=pK_write_type_test1_Type1, id =X, binput=false, it generates
   static void pK_write_type_test1_Type1(FILE *fwX, test1_Type1 X)
   \endcode

*/
_VIRTUAL_ void
inteprintc::printHeaderFunctionIOType (TypeType t, ptident fname, ptident id, boolean binput)
{
  printHeaderFunction (fname, (binput ? BooleanType : UndefinedType), pk_false);
  printLP ();
  printkwfile ();
  printSpace ();
  printstar ();
  printfilevarname (id, binput);
  printComma ();
  printtype (t, pk_true);
  printidentstar (id, binput);
  printRP ();
  printLine ();
}

/**
   It generates the beginning of the definition of the initilization of the IO component.
   \code
   For C, it generates   
   EXTERN void foo_OpenIO()
   {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclbegiofunchead ()
{
  printFunctionHeadNoParameters (begioname, UndefinedType, pk_true);

}


/**
   It generates the beginning of the definition of the initilization of the IO component.
   \code
   For C, it generates   
   EXTERN void foo_stepIO_begin()
   {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclstepiobeginfunchead ()
{
  printFunctionHeadNoParameters (stepbegioname, UndefinedType, pk_true);
}

/**
   It generates the beginning of the definition of the initilization of the IO component.
   \code
   For C, it generates   
   EXTERN void foo_stepIO_end()
   {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclstepioendfunchead ()
{
  printFunctionHeadNoParameters (stependioname, UndefinedType, pk_true);
}


/**
   IO component.
   It generates the closing of a file in the IO component, previously opened using printfileopen() method.
   The name of the file is built from nomfich and bInput.
   \code
    For C, 
    - with nomfich=a, bInput=true, it generates
           fclose(fra);
    - with nomfich=x1, bInput=false, it generates
           fclose(fwx1);
    \endcode
*/
_VIRTUAL_ void
inteprintc::printfileclose (ptident nomfich, boolean bInput)
{
  buf->writeAlfa (kwfclose, lgkwfclose);
  buf->writeChar ('(');
  printfilevarname (nomfich, bInput);
  printRPSemiColonLine ();
}

/**
   IO component.
   It generates the header of the definition of the finalizing of the IO operations of the current graph.
   
   \code
    For C, it generates
    EXTERN void foo_CloseIO()
    {
 \endcode
*/
_VIRTUAL_ void
inteprintc::printdclendiofunchead ()
{
  printFunctionHeadNoParameters (endioname, UndefinedType, pk_true);
}

_VIRTUAL_ void
inteprintc::printkwfile ()
{
  buf->writeAlfa (kwfile, lgkwfile);
}


/**
   IO component.
   It generates the declaration of a file used in IO operation. 
   The name of the descriptor assigned with the file is built from id and bInput.
   When bfirst, the declaration of the descriptor type is added first and set to "false" value.
   \code
    - with id=a, bInput=true, first=true, it generates
    FILE
    *fra

    - with id=x1, bInput=false, first=false, it generates
    , *fwx1
    \endcode
*/
_VIRTUAL_ void
inteprintc::printdclfile (ptident id, boolean bInput, boolean * bfirst)
{
  if (*bfirst)
    {
      printkwstatic ();
      printkwfile ();
      indent ();
      *bfirst = pk_false;
    }
  else
    {
      printComma ();
      printLine ();
    }
  printstar ();
  printfilevarname (id, bInput);
}

/**
   IO component.
   It opens (openOut_file()) the file associated with a model (processName) 
   for the generation of IO operations. Moreover, it generates the code of some importations (printincludefiles()).
   For C, with processName=foo, the name of the file is: foo_IO.c
   It generates in foo_IO.c the lines:
   \code 
   // Polychrony banner
   #include "foo_types.h"
   #include "foo_externals.h"
   \endcode 
*/
_VIRTUAL_ void
inteprintc::printdcliofilehead (ptident processName)
{
  openOut_file (mkNameWithLangsuffix (getIOFileName (processName)), mess_gen_io);
  printincludefiles (processName);
}

/**
   It opens (openOut_file()) the file associated with a model (processName) 
   for the generation of the type. Moreover, it generates the code of some importations (printincludefileSystem()).
   
   For C, with processName=foo, the name of the file is: foo_types.h
   \code 
   // Polychrony Banner 
   #ifndef foo_types
   #include <stdio.h>
   #include <stdlib.h>
   #include <math.h>
   #include "ansi.h"
   #include "pK_GenCodeC_Predef.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdcltypefile (ptident processName)
{
  ptrepstr name;
  name = getTypesFileName (processName);
  openOut_file (name, mess_gen_types);
  name = ident2repstr (processName);
  makeRepstr_repstr_string (name, C_TYPES_SUFFIX, &name);
  strconcchastr ('_', name, &name);
  printDirectiveIfNotDefined (name);
  printDirectiveDefined (name, pk_true);
  printincludefileSystem ();
}

/**
   It opens (openOut_file()) a file associated with a model (processName) for generation of the external references.
   For C, with processName=foo, the name of the file is: foo_externals.h
*/
_VIRTUAL_ void
inteprintc::printdclexternfile (ptident processName)
{
  ptrepstr name;
  name = getExternalsFileName (processName);
  openOut_file (name, mess_gen_externals);
}

/**
   @return the name of the opened file.
   It opens (openOut_file()) a file associated with a model (processName) for generation of the undefined external references.
   For C, with processName=foo, the name of the file is: foo_externalsUNDEF_LIS.h
*/
_VIRTUAL_ ptrepstr
inteprintc::printdclexternfileUser (ptident processName)
{
  ptrepstr name;
  name = getExternalsFileNameUndef (processName);
  openOut_file (name, mess_gen_externals);
  return name;
}

/**
   @return the name of the opened file.
   It opens (openOut_file()) a file associated with a model (processName) for generation of the external references.
   For C, with processName=foo, the name of the file is: foo_externalsProc.h
*/
_VIRTUAL_ ptrepstr
inteprintc::printdclexternfileProc (ptident processName)
{
  ptrepstr name;
  name = getExternalsFileNameProc (processName);
  openOut_file (name, mess_gen_externals);
  return name;
}

/**
   It finalizes the generation of the file that contains the external references.
   After this call, the file closed.
   This file is previously opened by the method printdclexternfileProc().

   @param progName (in) the name of the application.
*/
_VIRTUAL_ void
inteprintc::printdclexternfileProcEnd (ptident progName)
{
  closefile ();
}

/**
   It initialises the generation of the file that contains the external references (after the user imports).
   No effect for C, but it is redefined for other languages: does not delete it.
   @param progName (in) the name of the application.
*/
_VIRTUAL_ void
inteprintc::printdclexternfileProcBegin (ptident progName)
{
}


/** 
    It generates the beginning of the declaration of a procedure, called name, without any parameters and returned value.
    \code
    In C, with name=foo; it generates \n
    EXTERN void foo();
    \endcode
*/
_VIRTUAL_ void
inteprintc::printexternvoidentrypoint (ptident name)
{
  printdclrefvoidextCfunc (name);
  printRPSemiColonLine ();
}

/**
   Main program.
   It generates the main program for a model (ProcessName).
   It calls openOut_file(), printincludefiles(), printincludeBodyheaderFile(), printexternvoidentrypoint(),printdclmainfunc().
   \code
   For C, with  ProcessName=foo, it generates
   // Polychrony Banner
   #include "foo_types.h"
   #include "foo_externals.h"
   #include "foo_body.h"
   EXTERN void foo_OpenIO();
   EXTERN void foo_CloseIO();
   EXTERN int main()
   {
     logical code;
     foo_OpenIO();
     code = foo_initialize();
     while(code) 
     {
       foo_stepIO_begin();
       code = foo_iterate();
       foo_stepIO_end()
     }
     foo_CloseIO();
   }
   \endcode
*/
_VIRTUAL_ void
inteprintc::printmainfile (ptident ProcessName)
{
  openOut_file (mkNameWithLangsuffix (getMainFileName (ProcessName)), mess_gen_main);
  printincludefiles (ProcessName);
  printincludeBodyheaderFile (ProcessName);
  printexternvoidentrypoint (begioname);
  printexternvoidentrypoint (endioname);
  printexternvoidentrypoint (stepbegioname);
  printexternvoidentrypoint (stependioname);
  printLine ();
  printdclmainfunc ();
  closefile ();
}

/**
   @return the string \<processName\>_externalsUNDEF.h
*/
_VIRTUAL_ ptrepstr
inteprintc::getExternalsFileNameUndefForUser (ptident processName)
{
  ptrepstr name;

  name = getExternalsFileNameUndefPrefix (processName);
  if (sys_filePathInUserOutputDir (name, &name))
    return makeStr_fileName (name, H_SUFFIX);
  return nilstr;
}

/** 
    It generates the import of the file that declares the undefined external references of a model (processName).
    \code
    In C, with processName=foo, it generates
    #include "foo_externalsUNDEF"
    \endcode
*/
_VIRTUAL_ void
inteprintc::printRefExternalInclude (ptident processName)
{
  ptrepstr name;
  name = getExternalsFileNameUndefPrefix (processName);
  name = makeStr_fileName (name, H_SUFFIX);
  printinclude (name, '"', '"');
}

/** 
    It generates the import of the file that declares the external references of a model (processName).
    \code
    In C, with processName=foo, it generates
    #include "foo_externals"
    \endcode
*/
_VIRTUAL_ void
inteprintc::printRefExternalIncludeProc (ptident processName)
{
  ptrepstr name;
  name = getExternalsFileNameProc (processName);
  printinclude (name, '"', '"');
}



/**
   It generates the "string copy" operator.
*/
_VIRTUAL_ void
inteprintc::printkwstrcopy ()
{
  printstr (string2repstr ("strcpy"));
}

/**
   It generates the declaration of the type "integer".
*/
_VIRTUAL_ void
inteprintc::printkwIntSpace ()
{
  printkwint ();
  printSpace ();
}

void
inteprintc::printkwNULL ()
{
  buf->writeAlfa (kwNULL, lgkwNULL);
}

/**
   It generates the beginning of the declaration of the type "class".
   \code
   For C, it generated
   typedef struct {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printClassBeginDeclaration ()
{
  printkwClass ();
  printLine ();
  blockIndentOpening ();
}

/**
   It generates the end of the declaration of the type "class" for the current model, its name is given by getClassName().
   \code
   For C, it generates, where foo is the name of the current model 
   } foo; 
   \endcode
*/
_VIRTUAL_ void
inteprintc::genClassEnd ()
{
  printBlockClosingBackIndent ();
  printstr (getClassName ());
  printEndDeclaration ();
}

/**
   It generates the declaration of the internal predefined data associated with the current model for the separated compling. The name of the 
   current model is given by getClassName().
    \code
    For C,  with foo is the name of the current model, data is a predefined name, it generates
    foo *data
    \endcode
*/
_VIRTUAL_ void
inteprintc::printClassDataArea ()
{
  printstr (getClassName ());
  printSpace ();
  printchar ('*');
  printstr (string2repstr (KWdata));
}

/**
   It generates the externalObjClass followed by externalObjDataName and terminates by an end of statement.
   \code
   Example: 
   The declaration of an instance (foo1) of an external process (Foo), is generated by this method with 
   externalObjClass=foo and externalObjDataName=foo1
   foo foo1; 
   \endcode
*/
_VIRTUAL_ void
inteprintc::printCExternalSignalObjectDeclaration (ptrepstr externalObjClass, ptrepstr externalObjDataName)
{
  printstr (externalObjClass);
  printSpace ();
  printstr (externalObjDataName);
  printSemiColonLine ();
}


/** 
    It initializes the shaper for C language
*/
_VIRTUAL_ void
inteprintc::initialize (ptrepstr processName)
{
  gInitialize (processName);
  UndefinedType = TypeTypeGetUndefinedType ();
}



/**
   It calls printIOname() for a read operation.
*/
_VIRTUAL_ void
inteprintc::printReadFunctionLP (ptident id, ptident processName)
{
  printIOname (id, processName, pk_true);
}


/**
   It calls printIOname() for a write operation.
*/
_VIRTUAL_ void
inteprintc::printWriteFunctionLP (ptident id, ptident processName)
{
  printIOname (id, processName, pk_false);
}


/**
   @return the iterate name of the current application.
*/
_VIRTUAL_ ptident
inteprintc::getIterateName ()
{
  return iterfuncname;
}


/**
   @return the iterate name of the current application.
*/
_VIRTUAL_ ptident
inteprintc::getTerminateName ()
{
  return finalfuncname;
}

/**
   codop : ==, <<=
*/
_VIRTUAL_ void
inteprintc::printHeaderCmp (ptident id, TypeType t, ptident id1, ptident id2, boolean bdef)
{
  printHeaderFunction (id, BooleanType, pk_false);
  printLP ();
  printtype (t, pk_true);
  printident (id1);
  printComma ();
  printtype (t, pk_true);
  printident (id2);
  printRP ();
  printLine ();
  if (bdef)
    blockIndentOpening ();
}

/**
   vType is a predefined one or undefinedTYpe.
*/
_VIRTUAL_ void
inteprintc::printHeaderFunction (ptident id, TypeType vType, boolean bextern)
{
  if (bextern)
    cwrikwexternC ();
  else
    printkwstatic ();
  printtype (vType, pk_false);
  printSpace ();
  printident (id);
}


_VIRTUAL_ void
inteprintc::printDirectiveIfDefined (ptrepstr vstr)
{
  printDirectiveKwStrLine (string2repstr ("ifdef"), vstr);
}

_VIRTUAL_ void
inteprintc::printDirectiveIfNotDefined (ptrepstr vstr)
{
  printDirectiveKwStrLine (string2repstr ("ifndef"), vstr);
}

_VIRTUAL_ void
inteprintc::printDirectiveEndIf (ptrepstr vstr)
{
  printDirectiveKwStrLine (string2repstr ("endif"), nilstr);
}


_VIRTUAL_ void
inteprintc::printDirectiveDefined (ptrepstr vstr, boolean bline)
{
  if (bline)
    printDirectiveKwStrLine (string2repstr ("define"), vstr);
  else
    {
      printDirectiveKwStr (string2repstr ("define"), vstr);
      printSpace ();
    }
}

_VIRTUAL_ void
inteprintc::genIOOpeningVCDFile ()
{
  if ( VERSION_WITH_VCD ) {
    printDirectiveIfDefined (vcdStr);
    printfileopen (idvcd, pk_false);
    printVCDHeader ();
    printDirectiveEndIf (vcdStr);
  }
}


_VIRTUAL_ void
inteprintc::genIOClosingVCDFile ()
{
  if ( VERSION_WITH_VCD ) {
    printDirectiveIfDefined (vcdStr);
    printfileclose (idvcd, pk_false);
    printDirectiveEndIf (vcdStr);
  }
}

_VIRTUAL_ void
inteprintc::printVCDDeclarations ()
{
  ptrepstr name;
  if ( VERSION_WITH_VCD ) { 
  printLine ();
  printBeginComment ();
  printstr (string2repstr ("----The following line must be uncommented for the VCD mode"));
  printEndComment ();
  printLine ();
  printBeginComment ();

  printkwdefine ();
  printSpace ();
  printstr (vcdStr);

  printEndComment ();
  printLine ();
  printLine ();
  printDirectiveIfDefined (vcdStr);
  name = string2repstr (STR_VCD_INCLUDE);
  printinclude (name, '"', '"');

  printkwdefine ();
  printSpace ();
  printstr (vcdStr2);

  printSpace ();
  printint (1);			// vcd step
  printLine ();
  printDirectiveEndIf (vcdStr);
  }
}

_VIRTUAL_ void
inteprintc::printstepiobeginbody ()
{
  ptrepstr cpt;
if ( VERSION_WITH_VCD ) { 
  cpt = string2repstr (STR_VCD_COUNTER);

  printDirectiveIfDefined (vcdStr);
  buf->writeAlfa (kwfprintf, lgkwfprintf);
  printLP ();
  printfilevarname (idvcd, pk_false);
  printchar (',');
  printchar ('"');
  printchar ('#');
  printCodeFormat (IntegerType);
  printchar (escapeSep);
  printchar ('n');
  printchar ('"');
  printchar (',');
  printstr (cpt);
  printRP ();
  printSemiColonLine ();
  printstr (cpt);
  printassign ();
  printstr (cpt);
  printSpace ();
  printkwplus ();
  printSpace ();
  printstr (vcdStr2);
  printSemiColonLine ();

  printflushOn (idvcd);
  printLineidemptycall (string2ident (STR_VCD_SETSTEP));

  printDirectiveEndIf (vcdStr);
 }
}

_VIRTUAL_ void
inteprintc::printStrVCDfile (ptident vcdfile, ptrepstr vstr, boolean beol)
{
  if ( VERSION_WITH_VCD ) {
    buf->writeAlfa (kwfprintf, lgkwfprintf);
    printLP ();
    printfilevarname (vcdfile, pk_false);
    printchar (',');
    printchar ('"');
    printstr (vstr);
    if (!beol)
      printchar (' ');
    else
      {
	printchar (escapeSep);
	printchar ('n');
      }
    printchar ('"');
    printRP ();
    printSemiColonLine ();
  }
}


_VIRTUAL_ void
inteprintc::printdclVCDvar (TypeType t, ptident id, ptident vcdname)
{
  ptrepstr idType;
  int vsize;
  boolean b;
  if ( VERSION_WITH_VCD ) {
  b = pk_true;
  switch ((int) t)
    {
    case BooleanType:
    case PureType:
      idType = string2repstr ("trireg");
      vsize = 1;
      break;

    case IntegerType:
    case LongType:
    case ShortType:
      idType = string2repstr ("integer");
      vsize = 32;
      break;

    case RealType:
    case DRealType:
      idType = string2repstr ("real");
      vsize = 32;
      break;

    default:
      b = pk_false;
      break;
    }
  if (b)
    {
      printDirectiveIfDefined (vcdStr);
      printStrVCDfile (idvcd, string2repstr ("$var"), pk_false);
      printStrVCDfile (idvcd, idType, pk_false);
      printStrVCDfile (idvcd, integer2repstr (vsize), pk_false);
      printStrVCDfile (idvcd, ident2repstr (vcdname), pk_false);
      printStrVCDfile (idvcd, ident2repstr (id), pk_false);
      printStrVCDfile (idvcd, string2repstr ("$end"), pk_true);

      printDirectiveEndIf (vcdStr);
    }
  else;				// I don't know how they are represented in the VCD format.
}
}

_VIRTUAL_ void
inteprintc::printVCDValue (TypeType t, ptident ids, ptident vcdname, boolean bread)
{
  boolean b;

  if ( VERSION_WITH_VCD ) {
  b = (t == BooleanType) || (t == IntegerType) || (t == LongType) || (t == ShortType) || (t == RealType) || (t == DRealType);

  if (b)
    {
      printDirectiveIfDefined (vcdStr);
      printstr (string2repstr ("print_VCD_"));
      printtypePredef (t);
      printLP ();
      if (bread)
	printchar ('*');
      printident (ids);
      printRP ();
      printSemiColonLine ();
      printStrVCDfile (idvcd, ident2repstr (vcdname), pk_true);
      printDirectiveEndIf (vcdStr);
    }
  }
}

_VIRTUAL_ void
inteprintc::printVCDHeaderEnd ()
{
  if ( VERSION_WITH_VCD ) {
    printDirectiveIfDefined (vcdStr);
    printStrVCDfile (idvcd, string2repstr ("$enddefinitions"), pk_false);
    printStrVCDfile (idvcd, string2repstr ("$end"), pk_true);
    printDirectiveEndIf (vcdStr);
  }
}

 _VIRTUAL_ void
inteprintc::printComment (ptrepstr vcomment)
{
  printLangBeginComment();
  buf->writeRepstr (vcomment);
  printLangEndComment();
}



_VIRTUAL_ void
inteprintc::printblockIfClosing ()
{
 printBlockClosingBackIndent ();
}


_VIRTUAL_ void inteprintc::printkwthen ()
{}


_VIRTUAL_ void inteprintc::printkwEndIf ()
{
}

_VIRTUAL_ void inteprintc::OpenBlock()
{
  printblockopening ();
}


_VIRTUAL_ void inteprintc::CloseBlock()
{
  printBlockClosingBackIndent ();
}


_VIRTUAL_ void inteprintc::printkwelseif ()
{
}

_VIRTUAL_ void inteprintc::printSemiColon ()
{
  buf->printChar (';');
}


/**
   it generates "} while(<prefix><id>);"
*/
_VIRTUAL_ void inteprintc::printEndBlockMicroStep(ptrepstr prefix, ptident id)
{
  printRBracket ();
  printSpace ();
  printkwWhile ();
  printSpace ();
  printLP ();
  printstr (prefix);
  printident(id);
  printRP ();
  printSemiColon ();
  printLine ();
}


/**
   it generates "} while(<prefix><id>);"
*/
_VIRTUAL_ void inteprintc::printBeginBlockMicroStep(ptrepstr prefix, ptident id)
{
  printkwDo ();
  printLBracket ();
  printLine ();
}

_VIRTUAL_ void inteprintc::printStepFinalyze(boolean bbox)
{
  if (bbox)
    printdclfuncend (); 
  else
    printtruefuncend ();
}
