/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_

_VIRTUAL_ const char*
inteprintc::getSuffixCodeFile()
{
  return ".c";
}

_VIRTUAL_ const char*
inteprintc::getSuffixBodyFile()
{
  return "_body";
}

_VIRTUAL_ const char*
inteprintc::getkwInclude()
{
  return "#include";
}


_VIRTUAL_ const char *
inteprintc::getHeaderSuffix ()
{
  return ".h";
}

/*
  Management of the form for the C language. 
  Editing is done using the buffer and sysfile utilities.
*/
#define KWdata "data"
#define kwnotimpl       "UNDEF"
#define C_TYPES_SUFFIX  "_types"
#define C_EXTERNAL_SUFFIX  "_externals"
#define PROCTERM "Proc"
#define IO_SUFFIX  "_io"
#define MAIN_SUFFIX  "_main"
#define C_BODY_SUFFIX  getSuffixBodyFile()

#define C_SUFFIX getSuffixCodeFile()
#define H_SUFFIX  getHeaderSuffix ()

#define STR_SEPARATE "_separate"
#define C_BODY_SEPARATE_SUFFIX "_body_separate"
#define LIS_SUFFIX "_LIS"
#define strinit  "initialize"
#define striter  "step"
#define strfinal  "terminate"
#define STR_STEP_FINAL  "step_finalize"
#define STR_STEP_INIT  "step_initialize"
#define stropenio  "OpenIO"
#define strcloseio  "CloseIO"
#define strstepbeginio  "stepIO_begin"
#define strstependio  "stepIO_end"
#define stropencomm  "OpenComm"
#define strclosecomm "CloseComm"
#define vcode  "code"
#define lgcode  4
#define codeReadConst  'r'
#define codeWriteConst 'w'
#define codeUReadConst  'R'
#define codeUWriteConst 'W'
#define carindice  'I'
#define cardelay  'T'
#define charFile  'f'
#define strdatsuff  "dat"
#define strmain  "main"
#define strbegio  "begio"
#define strendio  "endio"
// c, c++
#define kwpow  "pow"
#define kwinclude       getkwInclude()
#define kwexternC       "EXTERN"
#define lgkwexternC     6
#define kwvoid          "void"
#define lgkwvoid        4
#define kwfile          "FILE"
#define lgkwfile        4
#define kwfclose        "fclose"
#define lgkwfclose      6
#define kwfopen         "fopen"
#define lgkwfopen       5
#define kwdefine        "#define"
#define lgkwdefine      7
#define kweof           "EOF"
#define lgkweof         3
#define kwtypedef       "typedef"
#define lgkwtypedef     7
#define kwfscanf        "scanf"
#define lgkwfscanf      5
#define kwfprintf       "fprintf"
#define lgkwfprintf     7
#define kwfflush        "fflush"
#define lgkwfflush      6
#define kwreturn        "return"
#define lgkwreturn      6
#define kwread          "Read"
#define kwwrite         "Write"
#define kwlogical       "boolean"
#define lgkwlogical     7
#define kwchar          "char"
#define lgkwchar        4
#define kwint           "int"
#define lgkwint         3
#define kwlong          "long"
#define kwshort         "short"
#define kwstatic        "static"
#define lgkwstatic      6
#define kwfloat         "float"
#define lgkwfloat       5
#define kwdouble        "double"
#define lgkwdouble      6
#define kwstderr        "stderr"
#define lgkwstderr      6
#define strcannotopen   "\"Can not open file %s\\n\""
#define lgstrcannotopen 24
#define strexit1        "exit(1);"
#define lgstrexit1      8
#define kwexit        "exit"
#define lgkwexit      4
#define strstdio        "stdio.h"
#define strMathInclude  "math.h"
#define strstdlib       "stdlib.h"
#define stransi         "ansi.h"
#define strstring       "string.h"
#define kwSizeOf       "sizeof"
#define kwStruct       "struct "
#define kwMaxStringLength  "_MAXSTRINGLENGTH_"
#define lgkwMaxStringLength  17
#define MaxStringLength 256
#define kwSignalStringType "pK_String "
#define kwStrEq "pK_StringEQ"
#define kwSignalComplexType "pK_complex"
#define kweqeq         "pK_EQEQ_"
#define kwiieg         "pK_IEQEQ_"
#define kwsignalInclude "pK_GenCodeC_Predef.h"
#define kwFirstOp "X"
#define kwSecondOp "Y"
#define kwNULL "NULL"
#define lgkwNULL 4
#define escapeSep   '\\'


// For VCD (Value Change Dump). The definitions are in the file $SignalLibC_incl/pK_vcd.h 
#define STR_VCD_SYMBOL "DUMPFILE_VCD_"
#define STR_VCD_SYMBOL_STEP "VCD_STEP_"
#define STR_VCD_COUNTER "VCD_counter_"
#define STR_VCD_SETSTEP "VCD_setStep_"
#define STR_VCD_INCLUDE "pK_vcd.h"


/**
   @return the name of the current process.
   It is set in gInitialize() method
*/
ptrepstr inteprintc::getProcessName ()
{
  return progName;
}

/**
   @return the name of the class of the process (it is the name of the process).
*/
_VIRTUAL_
  ptrepstr
inteprintc::getClassName ()
{
  return getProcessName ();
}

/**
   @return foo_types.h, for id=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getTypesFileName (ptident id)
{
  return makeStr_fileName (id, C_TYPES_SUFFIX, getHeaderSuffix ());
}

/**
   @return foo_externals.h, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getExternalsFileName (ptident processName)
{
  return makeStr_fileName (processName, C_EXTERNAL_SUFFIX, getHeaderSuffix ());
}

_VIRTUAL_ ptrepstr
inteprintc::getExternalsProcName (ptrepstr processName)
{
  ptrepstr name;

  name = nilstr;
  if (makeRepstr_repstr_string (processName, C_EXTERNAL_SUFFIX, &name))
    makeRepstr_repstr_string (name, PROCTERM, &name);
  return name;
}

/**
   @return foo_externalsProc.h, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getExternalsFileNameProc (ptrepstr processName)
{
  ptrepstr name;

  name = getExternalsProcName (processName);
  return makeStr_fileName (name, getHeaderSuffix ());
}

/**
   @return foo_externalsProc.h, for processName=foo.
   Same as getExternalsFileNameProc(ptrepstr ) for an ident.
*/
_VIRTUAL_ ptrepstr
inteprintc::getExternalsFileNameProc (ptident processName)
{
  return getExternalsFileNameProc (ident2repstr (processName));
}

/**
   @return the string foo_externalsUNDEF.h, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getExternalsFileNameUndefPrefix (ptident processName)
{
  ptrepstr name;

  if (makeRepstr_repstr_string (ident2repstr (processName), C_EXTERNAL_SUFFIX, &name))
    if (makeRepstr_repstr_string (name, kwnotimpl, &name))
      return name;
  return nilstr;
}

/**
   @return the string foo_externalsUNDEF_LIS.h, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getExternalsFileNameUndef (ptident processName)
{
  ptrepstr name;

  name = getExternalsFileNameUndefPrefix (processName);
  return makeStr_fileName (name, LIS_SUFFIX, getHeaderSuffix ());
}

/**
   @return the string foo_io, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getIOFileName (ptident processName)
{
  ptrepstr name;
  if (makeRepstr_repstr_string (ident2repstr (processName), IO_SUFFIX, &name))
    return name;
  return nilstr;
}


/**
   @return the string foo_main, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getMainFileName (ptident processName)
{
  ptrepstr name;
  if (makeRepstr_repstr_string (ident2repstr (processName), MAIN_SUFFIX, &name))
    return name;
  return nilstr;
}

/**
   @return the string foo_body, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getBodyFileName (ptident processName)
{
  ptrepstr name;
  if (makeRepstr_repstr_string (ident2repstr (processName), C_BODY_SUFFIX, &name))
    return name;
  return nilstr;
}

/**
   @return the string foo_body.h, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getHeaderName_Body (ptident id)
{
  return makeStr_fileName (id, C_BODY_SUFFIX, getHeaderSuffix ());
}

/**
   @return the string foo_body_separate.h, for processName=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::getHeaderName_BodySeparate (ptident id)
{
  return makeStr_fileName (id, C_BODY_SEPARATE_SUFFIX, getHeaderSuffix ());
}

/**
   @return the string foo.c, for name=foo
*/
_VIRTUAL_ ptrepstr
inteprintc::mkNameWithLangsuffix (ptrepstr name)
{
  return makeStr_fileName (name, C_SUFFIX);
}

ptident
inteprintc::make_stepInitIdent (ptident aName)
{
  return makeIdent_ident_string (aName, STR_STEP_INIT);
}

ptident
inteprintc::make_stepFinalIdent (ptident aName)
{
  return makeIdent_ident_string (aName, STR_STEP_FINAL);
}

/**
   It generates an error code for a type.
*/
_VIRTUAL_ void
inteprintc::ERRORTYPE ()
{
  printident (string2ident ("ERROR_TYPE"));
}


/**
   It generates foo(); for id=foo.
*/
void
inteprintc::printLineidemptycall (ptident id)
{
  printidemptycall (id);
  buf->printChar (';');
  printLine ();
}


/**
   It generates foo(data); for id=foo, data is a predefined name used for separated compiling.
*/
_VIRTUAL_ void
inteprintc::cgenCallSeparate (ptident id)
{
  printident (id);
  printchar ('(');
  printstr (string2repstr (KWdata));
  printchar (')');
  buf->printChar (';');
  printLine ();
}


/**
   It generates the key-word for a double.
*/
_VIRTUAL_ void
inteprintc::printkwdouble ()
{
  buf->fit (lgkwdouble);
  buf->writeRepstr (strtypedouble);
}


/**
   It generates the key-word for an int.
*/
_VIRTUAL_ void
inteprintc::printkwint ()
{
  buf->fit (lgkwint);
  buf->writeRepstr (strtypeint);
}

/**
   It generates the key-word for a real.
*/
_VIRTUAL_ void
inteprintc::printkwfloat ()
{
  buf->fit (lgkwfloat);
  buf->writeRepstr (strtypefloat);
}

/**
   It generates the key-word for an enumeration.
*/
_VIRTUAL_ void
inteprintc::printkwEnum ()
{
  cwristr (string2repstr ("enum"));
}

/**
   It generates the key-word for a character.
*/
_VIRTUAL_ void
inteprintc::printkwchar ()
{
  buf->fit (lgkwchar);
  buf->writeRepstr (strtypechar);
}

/**
   It generates the key-word for a string (pK_String internal predefined type).
*/
_VIRTUAL_ void
inteprintc::printkwstringtype ()
{
  printstr (string2repstr (kwSignalStringType));
}

/**
   It generates the key-word for a character type.
*/
_VIRTUAL_ void
inteprintc::printkwchartype ()
{
  printkwchar ();
}

/**
   It generates the key-word for a short type.
*/
_VIRTUAL_ void
inteprintc::printkwshort ()
{
  printstr (string2repstr (kwshort));
}

/**
   It generates the key-word for a long type.
*/
_VIRTUAL_ void
inteprintc::printkwlong ()
{
  printstr (string2repstr (kwlong));
}

/**
   It generates the key-word for a boolean type.
*/
_VIRTUAL_ void
inteprintc::printkwlogical ()
{
  buf->printAlfa (kwlogical, lgkwlogical);
}


/**
   It generates the sequence 
   \code 
    EXTERN void
   \endcode
*/
_VIRTUAL_ void
inteprintc::cwriextCvoid ()
{
  cwrikwexternC ();
  printkwvoid ();
}

/** It generates the void key-word. */
_VIRTUAL_ void
inteprintc::printkwvoid ()
{
  buf->writeAlfa (kwvoid, lgkwvoid);
  buf->space ();
}

/** It generates the external key-word. */
_VIRTUAL_ void
inteprintc::cwrikwexternC ()
{
  buf->writeAlfa (kwexternC, lgkwexternC);
  buf->space ();
}

/**
   It generates the sequence 
   \code 
    EXTERN logical 
   \endcode
*/
_VIRTUAL_ void
inteprintc::cwriextCBoolean ()
{
  cwrikwexternC ();
  printkwlogical ();
  printSpace ();
}

/**
   It generates the declaration of the interface of a model characterized by its name (vname) that returns an object with t as type, 
   and without any parameter, followed by the opening of the body statement.
   The declaration is prefixed by EXTERN if bext.
   \code
   Example: with vname=foo and t=BooleanType, it generates
   EXTERN logical foo() {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printFunctionHeadNoParameters (ptident vname, TypeType t, boolean bext)
{
  printHeaderFunction (vname, t, bext);
  buf->print2Chars ('(', ')');
  printLine ();
  blockIndentOpening ();
}

/**
   @return the identifier of "name of the current model" + '_" + s.
   Example: foo_body when foo is the name of the current process, and s="body"

*/
_VIRTUAL_ ptident
inteprintc::mkFuncName (const char *s)
{
  return mkFuncName (string2repstr (s));
}

/**
   @return the identifier of "name of the current model" + '_" + s.
   Example: foo_body when foo is the name of the current process, and s="body"

*/
_VIRTUAL_ ptident
inteprintc::mkFuncName (ptrepstr s)
{
  ptrepstr vstr;
  if (!makeRepstr_repstr_string (progName, "_", &vstr))
    return getIdentNil ();
  if (!strconc2str (vstr, s, pk_false, &vstr))
    return getIdentNil ();
  return repstr2ident (vstr);
}


/**
   It generates the representation of a (assumed) predefined type (ts).
*/
_VIRTUAL_ void
inteprintc::printtypePredef (TypeType ts)
{
  switch ((int) ts)
    {
    case BooleanType:
      printkwlogical ();
      break;
    case PureType:
      printkwlogical ();
      break;

    case IntegerType:
      printkwint ();
      break;
    case LongType:
      printkwlong ();
      break;
    case ShortType:
      printkwshort ();
      break;

    case RealType:
      printkwfloat ();
      break;
    case DRealType:
      printkwdouble ();
      break;

    case StringType:
      printkwstringtype ();
      break;

    case CharType:
      printkwchartype ();
      break;

    case ComplexType:
      printkwComplex ();
      break;

    case DComplexType:
      printkwComplex ();
      printchar ('d');
      break;

    case LabelType: // Does not occur, except when testing of new functionality
      printstr (string2repstr ("LabelType"));
      
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
}




/** It calls printlident() with '.' (dot) as separator. */
_VIRTUAL_ void
inteprintc::printlidentDotSeparator (ptident nameInput)
{
  printlident (nameInput, string2repstr ("."));
}

/** It calls printlident() with '_' (dot) as separator. */
_VIRTUAL_ void
inteprintc::printlidentUnderscoreSeparator (ptident nameInput)
{
  printlident (nameInput, string2repstr ("_"));
}

/**
   It generates the code associated with 
   - the reading of an input (when bin)
   - the writing of an output (when not bin)
*/
_VIRTUAL_ void
inteprintc::printReadWriteCode (boolean bin)
{
  if (bin)
    buf->writeChar (codeReadConst);
  else
    buf->writeChar (codeWriteConst);
}

/**
   It generates the beginning of a read (when bin)/write otherwise, of a identifier (vname)
   of a process (processName) followed by a parenthesis.
   \code
   It generates, with vname=a, processName=foo, bin=true
   r_foo_a(
   and with vname=a, processName=foo, bin=false
   w_foo_a(
   \endcode
*/
_VIRTUAL_ void
inteprintc::printIOname (ptident vname, ptident processName, boolean bin)
{
  boolean b;

  b = !namislident (vname);
  if (b)
    buf->fit (identsize (vname, 1) + identsize (processName, 1) + 1);
  else;				//must be completed (identsize)
  printReadWriteCode (bin);
  printunderscore ();
  printident (processName);
  printunderscore ();
  if (b)
    cwristr (ident2repstr (vname));
  else
    printlidentUnderscoreSeparator (vname);
  printchar ('(');
}

_VIRTUAL_ void
inteprintc::printLangBeginComment()
{
  printBeginComment ();
  printLine ();
}

_VIRTUAL_ void
inteprintc::printLangEndComment()
{
  printLine ();
  printEndComment ();
  printLine ();
}

/**
   It generates the banner of Polychrony as a comment.
   It accesses to the environment (sysGetSignalVersion()).
*/
_VIRTUAL_ void
inteprintc::printPolychronyversion ()
{
  printLangBeginComment();
  buf->writeRepstr (pK_messenger->recupmessStr (comment_polychrony));
  buf->writeRepstr (string2repstr (sysGetSignalVersion ()));
  printLangEndComment();
}

/**
   It generates the closing of a declaration.
*/
_VIRTUAL_ void
inteprintc::printEndDeclaration ()
{
  printSemiColonLine ();
}

/**
   Main component.
   It generates the declaration part of the main entry.
   \code
   It generates
   logical code;
   \endcode
*/
_VIRTUAL_ void
inteprintc::printMainDeclaration ()
{
  printkwlogical ();
  printSpace ();
  buf->writeAlfa (vcode, lgcode);
  printEndDeclaration ();
}

/**
   Main component.
   It generates a call to name without any parameters (printLineidemptycall()).
*/
_VIRTUAL_ void
inteprintc::printEndCallIterate (ptident name)
{
  printLineidemptycall (name);
}

/**
   Main component.
   It generates a call to name without any parameters, with a return value in code (internal name of the main componant).
   \code
   Example:  with name=foo_initialize, it generates
   code = foo_initialize(); 
   \endcode
*/
_VIRTUAL_ void
inteprintc::printCallIterate (ptident name)
{
  buf->writeAlfa (vcode, lgcode);
  printassign ();
  printEndCallIterate (name);
}

/**
   Main component.
   It calls printFunctionHeadNoParameters() for the main name and int as returned type.
   \code
   Example: It generates
   EXTERN int main() {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printMainFunct ()
{
  printFunctionHeadNoParameters (mainname, IntegerType, pk_true);
}


/**
   Main componant.
   It generates the definition of the "main" entry.
   \code
   For foo the name of the current process, it generates
   EXTERN int main()
   {
     logical code;
     foo_OpenIO();
     code = foo_initialize();
     while(code)
     { foo_stepIO_begin();
       code = foo_iterate();
       foo_stepIO_end();
       }
     foo_CloseIO();
   }
   \endcode
*/
_VIRTUAL_ void
inteprintc::printdclmainfunc ()
{
  printMainFunct ();
  printMainDeclaration ();
  printLineidemptycall (begioname);
  printCallIterate (inifuncname);
  printkwWhile ();
  buf->printChar ('(');
  buf->writeAlfa (vcode, lgcode);
  buf->printChar (')');
  printOpenBlockIndent ();
  printLineidemptycall (stepbegioname);
  printCallIterate (iterfuncname);
  printLineidemptycall (stependioname);
  printCloseBlockIndent (2);
  printLineidemptycall (endioname);
  printdclfuncend ();
}

/**
   It generates the importation of the math library.
*/
_VIRTUAL_ void
inteprintc::printPredefinedMathInclude ()
{
  printinclude (string2repstr (strMathInclude), '<', '>');
}

/**
   It generates the importation of the IO library for C.
*/
_VIRTUAL_ void
inteprintc::printIOIncludefileSystem ()
{
  printinclude (string2repstr (strstdio), '<', '>');
}

/**
   It generates the importation of the standard (std) library.
*/
_VIRTUAL_ void
inteprintc::printStdIncludefileSystem ()
{
  printinclude (string2repstr (strstdlib), '<', '>');
}

/**
   It generates the importation of several predefined libraries (C or Polychrony).
   It calls printIOIncludefileSystem(), printStdIncludefileSystem(), printPredefinedMathInclude(), printinclude().
   \code
   #include <stdio.h>
   #include <stdlib.h>
   #include <math.h>
   #include "ansi.h"
   #include "pK_GenCodeC_Predef.h"
   \endcode
*/
_VIRTUAL_ void
inteprintc::printincludefileSystem ()
{
  //  printcomment(pK_messenger->recupmessStr(comm_glob));
  printIOIncludefileSystem ();
  printStdIncludefileSystem ();
  printPredefinedMathInclude ();

  printinclude (string2repstr (stransi), '"', '"');
  printinclude (string2repstr (kwsignalInclude), '"', '"');

}


/**
   It generates a call to (or a declaration of) vname without parameter.
*/
_VIRTUAL_ void
inteprintc::printidemptycall (ptident vname)
{
  printident (vname);
  buf->print2Chars ('(', ')');
}

/**
   It generates a part of the declaration of id without parameter, following by the opening of the body statement.
   \code
   With id=foo, it generates
   foo() {
   \endcode
*/
_VIRTUAL_ void
inteprintc::printLinedclfunchead (ptident id)
{
  printidemptycall (id);
  printLine ();
  blockIndentOpening ();
}


/**
   It generates the name of a file obtained by code + id + ".dat", where code is a character obtained from cario.
   Example:
   - RA.dat for id=A, cario=r
   - WA.dat for id=A, cario=w
*/
_VIRTUAL_ void
inteprintc::printsysfilename (ptident id, boolean bInput)
{
  ptrepstr vstr;
  char cario;

  if (sys_add_typeSuffix2repstr (ident2repstr (id), strdatsuff, &vstr))
    {
      buf->fit (strsize (vstr) + 3);
      buf->writeChar ('"');
      if (bInput)
	cario = codeUReadConst;
      else
	cario = codeUWriteConst;
      buf->writeChar (cario);
      buf->writeRepstr (vstr);
      buf->writeChar ('"');
    }
}

/**
   It generates a declaration of a reference to an object (vname).
   (It generates foo *)
*/
_VIRTUAL_ void
inteprintc::printDataParamReference (ptident vname)
{
  printident (vname);
  printSpace ();
  printchar ('*');
}

/**
   It generates the declaration of the type "class".
   In C, it generated: typedef struct 
*/
_VIRTUAL_ void
inteprintc::printkwClass ()
{
  printkwTypedef ();
  printstrspace (string2repstr ("struct "));
}


/**
  io operation.
  It generates  the io mode opening "w" (when not b) or "r" (when b).
  
*/
_VIRTUAL_ void
inteprintc::printiomode (boolean b)
{
  buf->fit (4);
  buf->writeChar ('"');
  //  buf->writeChar(cario);
  printReadWriteCode (b);
  //  buf->writeChar('t');
  buf->writeChar ('"');
}

/**
   io operation.
   It generates the definition of the format (for io operations) associated with a type (t) without the prefix(% inC).
   It is implemented for predefined types
   (RealType, DRealType, IntegerType, LongType, ShortType, BooleanType, PureType, 
    StringType, CharType, ComplexType,  DComplexType)
    and the enumerate type, similar to IntegerType.
*/
_VIRTUAL_ void
inteprintc::printCodeFormat1 (TypeType t)
{
  switch ((int) t)
    {
    case RealType:
      buf->writeChar ('f');
      break;

    case DRealType:
      buf->writeChar ('l');
      buf->writeChar ('f');
      break;

    case IntegerType:
      buf->writeChar ('d');
      break;

    case LongType:
      buf->writeChar ('l');
      buf->writeChar ('d');
      break;

    case ShortType:
    case BooleanType:
    case PureType:
      buf->writeChar ('h');
      buf->writeChar ('d');
      break;

    case StringType:
      buf->writeChar ('s');
      break;

    case CharType:
      buf->writeChar ('c');
      break;

    case ComplexType:
      printCodeFormat1 (RealType);
      buf->writeChar (' ');
      printCodeFormat (RealType);
      break;

    case DComplexType:
      printCodeFormat1 (DRealType);
      buf->writeChar (' ');
      printCodeFormat (DRealType);
      break;

    default:
      printnotimplemented ();
      break;
    }
}

/**
   io operation.
   It generates the definition of the format (for io operations) associated with a type (t).
   It calls printCodeFormat1() after the generation of the prefix (%)
*/
_VIRTUAL_ void
inteprintc::printCodeFormat (TypeType t)
{
  buf->writeChar ('%');
  if (TypeTestNamedClass (t))
    t = NamedTypeGetTerm (t);
  if (TypeTestEnumerationClass (t))
    t = IntegerType;
  printCodeFormat1 (t);
}

/**
   io operation.
   It generates the string that gives the definition of the format (for io operations) associated with a type (t).
   (Example : "%f " for a real)
*/
_VIRTUAL_ void
inteprintc::printformat (TypeType t, boolean bread)
{
  int lg;
  boolean b;

  if (TypeTestNamedClass (t))
    t = NamedTypeGetTerm (t);
  b = TypeTestArrayClass (t);
  if (b)
    t = ArrayMatrixTypeGetTerm (t);
  if (t == DRealType)
    lg = 5;
  else
    lg = 4;
  buf->fit (lg);
  buf->writeChar ('"');
  printCodeFormat (t);
  if (b || !bread)
    buf->writeChar (' ');
  buf->writeChar ('"');
}

/**
   io operation.
   It generates the beginning of a read operation.
   \code
   scanf(
   \endcode
*/
_VIRTUAL_ void
inteprintc::printkwScanpo ()
{
  buf->writeAlfa (kwfscanf, lgkwfscanf);
  printchar ('(');
}

/**
   It generates the reading of an identifier (id) with idt as type.
   It calls printkwScanpo(), printfilevarname(), printformat().
   \code
   For id=A, minc=r, idt=realType, It generates
   fscanf( frA, "%f",
   \endcode
*/
_VIRTUAL_ void
inteprintc::printscancallheadLP (ptident id, char minc, TypeType idt)
{
  printchar ('f');
  printkwScanpo ();
  printfilevarname (id, minc);
  printchar (',');
  printformat (idt, pk_true);
  printchar (',');
}

/**
   It associates (openfile()) a file to the buffer of the current shaper.
   The fullname of the file (filePathName) is produced by sys_filePathInUserOutputDir().
   The polychrony banner is generated (printPolychronyversion()).   
*/
_VIRTUAL_ boolean
inteprintc::openOut_file (ptrepstr fileName, int ninfo)
{
  ptrepstr filePathName;
  if (open_newBuffer (fileName, 40, 160, &filePathName))
    {
      pK_messenger->messagestr (ninfo, filePathName);
      printLine ();
      printPolychronyversion ();
      return pk_true;
    }
  return pk_false;
}



_VIRTUAL_ void
inteprintc::printkwread ()
{
  printstr (string2repstr (kwread));
}

_VIRTUAL_ void
inteprintc::printkwwrite ()
{
  printstr (string2repstr (kwwrite));
}


/**
   It initializes the shaper for a graph with vname as name.
   All the predefined names are initialized.
*/
_VIRTUAL_ void
inteprintc::gInitialize (ptrepstr vname)
{
  progName = vname;
  inifuncSuffix = string2repstr (strinit);
  iterfuncSuffix = string2repstr (striter);
  mainname = string2ident (strmain);
  strtypechar = string2repstr (kwchar);
  strtypeint = string2repstr (kwint);
  strtypefloat = string2repstr (kwfloat);
  strtypedouble = string2repstr (kwdouble);
  vcdStr = string2repstr (STR_VCD_SYMBOL);
  idvcd = string2ident (STR_VCD_SYMBOL);

  vcdStr2 = string2repstr (STR_VCD_SYMBOL_STEP);


  inifuncname = mkFuncName (strinit);
  iterfuncname = mkFuncName (striter);
  finalfuncname = mkFuncName (strfinal);
  begioname = mkFuncName (stropenio);
  endioname = mkFuncName (strcloseio);

  stepbegioname = mkFuncName (strstepbeginio);
  stependioname = mkFuncName (strstependio);

  begcommname = mkFuncName (stropencomm);
  endcommname = mkFuncName (strclosecomm);

}


/**
   It prints the values of an assumed enumerated type (t).
*/
_VIRTUAL_ void
inteprintc::printListEnum (TypeType t)
{
  boolean Goon;
  NameSubList NSubLst;
  NameList NL;
  ptrepstr velem;

  printchar ('{');
  NL = EnumerationTypeGetNameLst (t);
  NSubLst = NameListCast2SubList (NL);
  Goon = pk_true;
  while (Goon)
    {
      velem = NameSubListGetHead (NSubLst);
      buf->fit (strsize (velem));
      cwristr (velem);
      if (NameSubListTestLast (NSubLst))
	Goon = pk_false;
      else
	{
	  printchar (',');
	  NSubLst = NameSubListGetTail (NSubLst);
	}
    }
  printchar ('}');
}

/**
   @return the type to be generated for a type (t).
   For C/C++ t is returned,  (typedef can be used for the aliases).
*/
_VIRTUAL_ TypeType inteprintc::getTypeToGenerate (TypeType t)
{
  return t;
}

_VIRTUAL_ void
inteprintc::printDirectiveKwStr (ptrepstr kw, ptrepstr vstr)
{
  printchar ('#');
  printstr (kw);
  if (vstr != nilstr)
    {
      printchar (' ');
      printstr (vstr);
    }
}


_VIRTUAL_ void
inteprintc::printDirectiveKwStrLine (ptrepstr kw, ptrepstr vstr)
{
  printDirectiveKwStr (kw, vstr);
  printLine ();
}


_VIRTUAL_ void
inteprintc::printVCDHeader ()
{
  // printStrVCDfile(idvcd, string2repstr("$date"), pk_true);
  // printStrVCDfile(idvcd, string2repstr("$end"), pk_true);
  printStrVCDfile (idvcd, string2repstr ("$version"), pk_true);
  printStrVCDfile (idvcd, pK_messenger->recupmessStr (comment_polychrony), pk_false);
  printStrVCDfile (idvcd, string2repstr (sysGetSignalVersion ()), pk_true);
  printStrVCDfile (idvcd, string2repstr ("$end"), pk_true);
  printStrVCDfile (idvcd, string2repstr ("$timescale"), pk_true);
  printStrVCDfile (idvcd, string2repstr ("    1 ns"), pk_true);
  printStrVCDfile (idvcd, string2repstr ("$end"), pk_true);
}

_VIRTUAL_ void
inteprintc::printflushOn (ptident id)
{
  buf->writeAlfa (kwfflush, lgkwfflush);
  printchar ('(');
  printfilevarname (id, pk_false);
  printRPSemiColonLine ();
}
