/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   \spe of cGenerator::genAssignString(). Here, it is without effect.
*/
  _VIRTUAL_ void
javaGenerator::genAssignString (ptexpcal expg, ptlrepstr lfields, ptexpcal expd, ptlrepstr lfieldsdef)
{
}

/**
   \copydoc cGenerator::isGenAssignString()
   \spe return always "false".
*/
_VIRTUAL_ boolean
javaGenerator::isGenAssignString (ptexpcal expg, ptexpcal expd)
{
  return pk_false;
}


_VIRTUAL_ void
javaGenerator::genIdentDeclaration (ptident id, TypeType t)
{
  TypeType TT;

  if (TypeTestNamedClass (t))
    {
      TT = NamedTypeGetTerm (t);
      if (TypeTestArrayClass (TT))
	t = TT;
    }
  cGenerator::genIdentDeclaration (id, t);
}

/** 
    \copydoc cGenerator::initialize()
    \spe the specific pragma name (PragmaGencodeLanguage) to the predefined PRAGMA_CODEGEN_JAVA name.
*/
_VIRTUAL_ void
javaGenerator::initialize (ptDCG_sch vGraph, boolean genMode, ptrepstr vprefix)
{
  cGenerator::initialize (vGraph, genMode, vprefix);
  isIOmodule = pk_false;
  PragmaGencodeLanguage = getPredefinedPragma (PRAGMA_CODEGEN_JAVA);
  PragmaImportLanguage = getPredefinedPragma (PRAGMA_CODEGEN_JAVA_IMPORT);
  PragmaDeclareLanguage = getPredefinedPragma (PRAGMA_CODEGEN_JAVA_DECLARE);
}

/** 
    \copydoc cGenerator::genBody()
    \spe Code generation of the file foo.java, where foo is the name of the current graph.
    It generates 
    - the definition of the class genBodyClass()
    - the definition of the methods genBodyCode()
*/
_VIRTUAL_ void
javaGenerator::genBody ()
{

  genBodyClass ();/** file <P>.java ( and the corresponding class ) */
  genBodyCode ();
}


/**
   It generates:
   - the standard Signal imports
   - the user imports using the PRAGMA_CODEGEN_JAVA_IMPORT pragma.
*/
_VIRTUAL_ void
javaGenerator::genImports ()
{
  ((inteprintjava *) TextForm)->addpKStdImports ();
  genUserImports ();
}

/** 
    Code generation of the file foo.java, with foo the name of the current graph.
    It calls
    - printdclbodyfileheader(), it opens the file 
    - printClassBeginDeclaration(),
    - printJavaDeclarationIO(), not called in separated compiling
    - cGenerator::gendeclarationsBody(),
    - generateConstructor()
    - printJavaIOMethods() not called in separated compiling
    \code
    Such a code looks like:
    class foo
    {
    private foo_io _io_;
    private KeyboardReader KR = new KeyboardReader();
    //  ==> parameters and indexes
    //  ==> input signals    
    public float a, b, c;
    // ==> output signals 
    public float x2, x1;
    public boolean error;
    //  ==> local signals 
    public boolean OK;
    ....

    public foo(){ _io_=new foo_io (); }
    public void openIO (){ _io_.foo_OpenIO (); }
    public void closeIO (){ _io_.foo_CloseIO (); }
    

    \endcode
*/
_VIRTUAL_ void
javaGenerator::genBodyClass ()
{
  TextForm->printdclbodyfileheader (progName, pk_false);
  genImports ();

  TextForm->printClassBeginDeclaration ();

  if (!SEPARATE)
    {
      ((inteprintjava *) TextForm)->printJavaDeclarationIO ();
      if (externalObjects)
	((inteprintjava *) TextForm)->printJavaDeclarationExternalObject ();
    }
  cGenerator::gendeclarationsBody ();
  generateConstructor ();
  if (externalObjects)
    ((inteprintjava *) TextForm)->printJavaSetExternalObject ();
  if (!SEPARATE)
    ((inteprintjava *) TextForm)->printJavaIOMethods ();
}

/**
   \copydoc cGenerator::genMain()

   Such a code looks like with foo the name of the current graph (see inteprintjava::printmainfile())
   \code
   public class foo_main{
   public static void main ( String[] args )  throws Exception {
   boolean code ;
        foo foo_obj_=new foo();
        foo_obj_.openIO ();
        code =foo_obj_.foo_initialize();
        while (code )code =foo_obj_.foo_step();
        foo_obj_.closeIO ();
    }
    }
    \endcode
*/
_VIRTUAL_ void
javaGenerator::genMain ()
{
  ((inteprintjava *) TextForm)->printmainfile (progName);
}

/** 
    \spe of cGenerator::genBodyCode().
    Code generation of the body of the process (foo.java file, with foo the name of the current graph).
    It calls
    - genInitialisations(), 
    - genSubGraphs(),
    - genEntryPointStepInitialize(),
    - genScheduler(), 
    - genClassEnd(), 
    - closefile().
*/
_VIRTUAL_ void
javaGenerator::genBodyCode ()
{
  genInitialisations ();
  genSubGraphs ();
  genEntryPointStepInitialize ();
  genScheduler ();
  TextForm->genClassEnd ();
  TextForm->closefile ();
}

/** 
   \copydoc cGenerator::genEntryPointStepInitialize()
*/
_VIRTUAL_ void
javaGenerator::genEntryPointStepInitialize ()
{
  TextForm->inteprintc::printentrypointstepinitialize (progName, pk_false);
  set_booleanClocks_undefined (tick);
  TextForm->printdclfuncend ();
}

/** 
   Generation of the keyword "public". 
*/
_VIRTUAL_ void
javaGenerator::genAccessibility ()
{
  if (!INTERNAL_BLOCK)
    ((inteprintjava *) TextForm)->printkwpublic ();
}


/** 
    \copydoc cGenerator::genDclSignalDelayTree()
*/
_VIRTUAL_ void
javaGenerator::genDclSignalDelayTree (ptexpcal exp, ptmemnode mn)
{
  TypeType t1;

  t1 = EM->getType (exp);
  gentype (t1);
  ecrnomrootinddelay (mn);
  TextForm->printLeftSquareBracket ();
  TextForm->printRightSquareBracket ();
  genDclDimensions (EM->getType (ptmemnode_get_signal (mn)));
}

/** 
    \copydoc cGenerator::genDimensionInDeclaration()
*/
_VIRTUAL_ void
javaGenerator::genDimensionInDeclaration (ptexpcal aDim)
{
}

/** 
    \copydoc cGenerator::genTypeBeginDeclaration()
    \spe Generation of the "local" scope key-word then the type t \b without the generating of dimensions. 
 */
_VIRTUAL_ void
javaGenerator::genTypeBeginDeclaration (TypeType t)
{
  boolean todo;

  genAccessibility ();
  todo = pk_false;
  if (TypeTestEnumerationClass (t))
    {
      TextForm->printstr (EnumerationTypeGetName (t));
      TextForm->printSpace ();
    }
  else if (TypeTestNamedClass (t))
    if (TypeTestStructureClass (NamedTypeGetTerm (t)))
      {
	TextForm->printstr (NamedTypeGetName (t));	//zzz
	TextForm->printSpace ();
      }
    else if (TypeTestArrayClass (t))
      t = genTypeBase3 (t);
    else
      todo = pk_true;

  else
    todo = pk_true;
  if (todo)
    genTypeBase (t);
  TextForm->mark ();
  tprec = t;
}


/**
   Generation of the dimensions of the type t
*/
_VIRTUAL_ void
javaGenerator::genConstructorDimensions (TypeType t)
{
  ptexpcal vdim, aDim;

  t = GetUnamedType (t);
  if (!TypeTestArrayClass (t))
    return;

  vdim = ArrayMatrixTypeGetDimLst (t);
  while (EM->getNaryArgs (&vdim, &aDim))
    {
      TextForm->printLeftSquareBracket ();
      genexpression (aDim, opnullary);
      TextForm->printRightSquareBracket ();
    }

}

/**
   Generation of the constructor for a delay signal (exp), node (mn) of a delay tree.
   It calls genDclSignalDelayTreeForConstructor().
*/
_VIRTUAL_ void
javaGenerator::genConstructorDelayTree (ptexpcal exp, ptmemnode mn)
{
  ptlmemnode lmn;

  lmn = ptmemnode_get_equiv (mn);
  do
    {
      if (ptmemnode_satisfies_root (mn))
	{
	  genPrefix ();
	  genDclSignalDelayTreeForConstructor (exp, mn);
	}
      if (!ptlmemnode_getNaryArgs (&lmn, &mn))
	mn = NILMEMNODE;
    }
  while (mn != NILMEMNODE);
}


/** 
    Code generation for the declaration of a delay (exp), node (mn) of a delay tree.

    \code
    timedout=new boolean [63];
    \endcode
*/
_VIRTUAL_ void
javaGenerator::genDclSignalDelayTreeForConstructor (ptexpcal exp, ptmemnode mn)
{
  TypeType t;

  ecrnomrootinddelay (mn);
  ((inteprintjava *) TextForm)->printEqAllocate ();
  t = EM->getType (exp);
  TextForm->printtype (t, pk_true);
  TextForm->printLeftSquareBracket ();
  TextForm->printint (ptmemnode_get_size (mn));
  TextForm->printRightSquareBracket ();
  genConstructorDimensions (t);
  TextForm->printSemiColon ();
  TextForm->printLine ();
}

_VIRTUAL_ void
javaGenerator::genConstructor (ptident id, TypeType t)
{
  genPrefix ();
  TextForm->printident (id);
  ((inteprintjava *) TextForm)->printEqAllocate ();
  genTypeBase2 (t);
  genConstructorDimensions (t);
  TextForm->printSemiColon ();
  TextForm->printLine ();
}

/** 
    Code generation for a signal (exp) of array type. 
*/
_VIRTUAL_ void
javaGenerator::genArrayDclSignal (ptexpcal exp)
{
  genConstructor (EM->getName (exp), EM->getType (exp));
}


/**
   Code Generation for a signal (exp) which is a named type.
*/
_VIRTUAL_ void
javaGenerator::genNamedDclSignal (ptexpcal exp)
{
  TypeType TT;

  TT = EM->getType (exp);
  if (TypeTestNamedClass (TT))
    if (!TypeTestPredef (NamedTypeGetTerm (TT)))
      {
	genPrefix ();
	TextForm->printident (EM->getName (exp));
	((inteprintjava *) TextForm)->printEqAllocate ();
	TextForm->printstr (NamedTypeGetName (TT));
	TextForm->printLP ();
	TextForm->printRP ();
	TextForm->printSemiColon ();
	TextForm->printLine ();
      }
}

/**
   Code generation for the declaration of a signal (exp) assumed not to be a node of a delay tree.
   It calls genArrayDclSignal() for array type, genNamedDclSignal() otherwise:
   only array and named types are considered here, they need the creation of objects.
*/
_VIRTUAL_ void
javaGenerator::genConstructorStd (ptexpcal exp)
{
  if (EM->getNumberOfDimensions (exp) != 0)
    genArrayDclSignal (exp);
  else
    genNamedDclSignal (exp);
}


/** 
    Code generation for the declaration of a signal (exp).
    It calls genConstructorDelayTree() for a node of a delay tree, genConstructorStd() otherwise.
    Only array and named types are considered, they need the creation of objects.
*/
_VIRTUAL_ void
javaGenerator::genConstructor (ptexpcal exp)
{
  ptmemnode mn;

  if (ptexpcal_satisfies_node_tree_delay (exp, &mn))
    genConstructorDelayTree (exp, mn);
  else
    genConstructorStd (exp);
}

/** 
    It generates the declaration of the constructor of the class foo (in file foo.java, assumed opened),
    where foo is the name of the current graph.
*/
_VIRTUAL_ void
javaGenerator::generateConstructor ()
{
  TextForm->printLine ();
  genUserDeclarations ();

  ((inteprintjava *) TextForm)->printkwpublic ();
  TextForm->printident (progName);
  TextForm->printLP ();
  TextForm->printRP ();
  TextForm->blockIndentOpening ();
  gendeclarationsBodyForConstructor ();
  if (!SEPARATE)
    ((inteprintjava *) TextForm)->printJavaCreateIOObject ();
  TextForm->printBlockClosingBackIndent ();
}


/** 
    It generates the allocation of memory for signals (lsig).
    It calls genConstructor() for each useful element.
*/
_VIRTUAL_ void
javaGenerator::genDclSignalsForConstructor (ptlinstances lsig)
{
  ptexpcal sig;
  while (ptlinstances_get_next (&lsig, &sig))
    if (mustBeGenerated (EM->getNode (sig)))
      genConstructor (sig);
}


/** 
    It generates the allocation of memory for non-external constant (lsig). 
    It calls genConstructor() for each element of the list, the type of the element is not a predefined one.
*/
_VIRTUAL_ void
javaGenerator::genDclConstantForConstructor (ptlinstances lsig)
{
  ptexpcal sig;
  while (ptlinstances_get_next (&lsig, &sig))
    if (!EM->IsExternalConstant (sig))
      if (!TypeTestPredef (EM->getUnamedType (sig)))
	genConstructor (sig);
}

/**
   Assumed t is an array type.
   It generates, if the operand signal (exp) is a state variable, the constructor of its "previous" value.
*/
_VIRTUAL_ void
javaGenerator::genConstructorStatevar ()
{
  ptlinstances lexpr;
  ptexpcal exp;
  ptident idPrevious;
  TypeType TT;

  if (EM->getExprList (opvvarin, &lexpr))	// lexpr = Sets of ?x expressions
    while (ptlinstances_get_next (&lexpr, &exp))
      if (EM->getAttribute (exp, AttributeNameGencode, (attributeValue *) & idPrevious))
	{
	  TT = EM->getUnamedType (EM->getUnaryArg (exp));
	  if (TypeTestArrayClass (TT))
	    genConstructor (idPrevious, TT);
	}
}

/** 
    Code generation for the constructor of the class foojava. 
*/
_VIRTUAL_ void
javaGenerator::gendeclarationsBodyForConstructor ()
{
  genDclConstantForConstructor (AGRAPH->getConstants ());
  genDclSignalsForConstructor (AGRAPH->getInputs ());
  genDclSignalsForConstructor (AGRAPH->getOutputs ());
  genDclSignalsForConstructor (AGRAPH->getLocalSignals ());

  genConstructorStatevar ();

  TextForm->printLine ();
}

/* -------------------- IO module ---------------------- */

/** 
    Code generation for the constructor of the class foo_io, where foo is the name of the
    current process (it is genrated in foo_io.java file).
    Example:
    \include ExampleJavaIO
*/
_VIRTUAL_ void
javaGenerator::genIOConstructor ()
{
  ptlinstances lsig;
  ptexpcal sig;

  ((inteprintjava *) TextForm)->printIOConstructorHeader ();

  lsig = AGRAPH->getInputs ();
  while (ptlinstances_get_next (&lsig, &sig))
    ((inteprintjava *) TextForm)->genAllocateFileDescriptorInput (EM->getName (sig));

  lsig = AGRAPH->getOutputs ();
  while (ptlinstances_get_next (&lsig, &sig))
    ((inteprintjava *) TextForm)->genAllocateFileDescriptorOutput (EM->getName (sig));

  TextForm->genClassEnd ();
}


/** 
    \spe of cGenerator::genIOFileDeclarations()
*/
_VIRTUAL_ void
javaGenerator::genIOFileDeclarations ()
{
  ptlinstances lsig;
  ptexpcal sig;

  lsig = AGRAPH->getInputs ();
  while (ptlinstances_get_next (&lsig, &sig))
    {
      ((inteprintjava *) TextForm)->javaprintdclfileInput (EM->getName (sig));
      printptvlinebi (0);
    }

  lsig = AGRAPH->getOutputs ();
  while (ptlinstances_get_next (&lsig, &sig))
    {
      ((inteprintjava *) TextForm)->javaprintdclfileOutput (EM->getName (sig));
      printptvlinebi (0);
    }
  TextForm->printLine ();
}

/** 
   \spe of cGenerator::geniomodule().
   The code is generated in the file foo_io.java, where foo is the name of the current graph.
*/
_VIRTUAL_ void
javaGenerator::geniomodule ()
{
  ptlINode ins, outs;

  NM->getInputs (&ins);		// inputs
  NM->getOutputs (&outs);	// outputs
  isIOmodule = pk_true;
  TextForm->printdcliofilehead (progName);
  ((inteprintjava *) TextForm)->javaPrintIOClassBeginDeclaration ();
  genIOFileDeclarations ();
  genIOConstructor ();
  genIOOpening (ins, outs);
  genIOClosing (ins, outs);
  genIOReadingWriting (ins, outs);

  TextForm->genClassEnd ();
  TextForm->closefile ();

  isIOmodule = pk_false;
}

/**
   \spe of cGenerator::ecrnominddelay().
*/
_VIRTUAL_ void
javaGenerator::ecrnominddelay (ptmemnode mn)
{
  genPrefix ();
  if (isIOmodule)
    ((inteprintjava *) TextForm)->printAccessIOObject ();
  cGenerator::ecrnominddelay (mn);
}

/**
   \spe of cGenerator::genAccessEnumValue().
   For Java, it generates the name of the type, followed by the selector operaotr (.), before the value.
*/
_VIRTUAL_ void
javaGenerator::genAccessEnumValue (TypeType t, ptexpcal exp)
{
  TextForm->printtype (t, pk_false);
  TextForm->printchar ('.');
  genexpression (exp, opnullary);
}


/** 
    \spe of cGenerator::genConstant()
*/
_VIRTUAL_ void
javaGenerator::genConstant (ptexpcal exp)
{
  ptrepstr strval;
  noperators ope;

  ope = EM->getOperator (exp);
  if (ope == optrue)
    TextForm->printkwlogicons (pk_true);
  else if (ope == opfalse)
    TextForm->printkwlogicons (pk_false);
  else
    {
      strval = EM->getConstant (exp);
      switch ((int) ope)
	{
	case opcstdp:
	  strval = strdoubletoCdouble (strval);
	  break;
	case opcstreel:
	  strconcstrcha (strval, 'f', &strval);
	  break;
	}
      TextForm->printstr (strval);
    }
}


/* ---------------- TYPES ----------------- */

/** 
    \spe of cGenerator::genNonExternalTypes.
    It generates the declaration of non predefined types. 
    External Types are generated in a specific file (foo_Externals.java).
*/
_VIRTUAL_ void
javaGenerator::dclTypes ()
{
  genNonExternalTypes ();
}


/** 
    Generates the declaration (genTypeDeclaration()) of non predefined types and non external types referenced in the program. 
*/
_VIRTUAL_ void
javaGenerator::genNonExternalTypes ()
{
  ptlinstances ltypes;
  ptexpcal elem;
  TypeType Tp, TT;
  boolean todo;

  if (AGRAPH->getUserTypesOneLevel (&ltypes))
    while (ptlinstances_get_next (&ltypes, &elem))
      {
	todo = pk_true;
	Tp = EM->getExpType (elem);
	if (TypeTestNamedClass (Tp))
	  {
	    TT = NamedTypeGetTerm (Tp);
	    if (TypeTestPredef (TT))
	      todo = pk_false;
	    else
	      todo = !TypeTestArrayClass (TT);
	  }
	if (todo)
	  if (!TypeTestExternClass (Tp) && !TypeTestMetaClass (Tp))
	    genTypeDeclaration (Tp);
      }
}

/** 
    \spe of cGenerator::genExternalType()
    It calls genUndefClassTypeDeclaration().
*/
_VIRTUAL_ void
javaGenerator::genExternalType (TypeType TT)
{
  ptrepstr typeName;
  if (TypeTypeGetName (TT, &typeName))
    {
      genUndefClassTypeDeclaration (typeName);
      TextForm->printLine ();
    }
}

/** 
    \spe of cGenerator::genTypeDeclaration().
*/
_VIRTUAL_ void
javaGenerator::genTypeDeclaration (TypeType TT)
{
  if (TypeTestPredef (TT))
    {
      return;
    }
  if (TypeTestArrayClass (TT))
    {
      return;
    }
  if (TypeTestEnumerationClass (TT))
    {
      genClassEnumDeclaration (TT);
      return;
    }
  if (TypeTestStructureClass (TT))
    {
      genClassStructDeclaration (TT);
      return;
    }
  if (TypeTestNamedClass (TT))
    GenTypedefDefinition (NamedTypeGetName (TT), NamedTypeGetTerm (TT));
}


/** 
    \spe of cGenerator::GenTypedefDefinition().
    The type (genTypeBase()) is generated in foo_type.java file, with id=foo.
*/
_VIRTUAL_ void
javaGenerator::GenTypedefDefinition (NameClass id, TypeType T)
{
  ((inteprintjava *) TextForm)->javaPrintClassBegin (id, 1, pk_true, pk_true);
  genTypeBase (T);
  ((inteprintjava *) TextForm)->genClassEnd (pk_true);
}

/**
   \spe of javaGenerator::GenTypeDefinition()
*/
_VIRTUAL_ void
javaGenerator::GenTypeDefinition (NameClass id, TypeType T)
{
  T = javaGenerator::genTypeBase2 (T);

  TextForm->printstr (id);
  genDclDimensions (T);

  TextForm->printSemiColonLine ();
  TextForm->printRBracket ();
  TextForm->closefile ();
}


/** 
    Generation of the value of a type (t)
    @return t the value of the TypeType element
*/
_VIRTUAL_ TypeType
javaGenerator::genTypeBase3 (TypeType t)
{
  t = TypeTypeGetUnamedType (t);
  if (!TextForm->printtype (t, pk_true))
    setSomeProblem ();
  return t;
}


/** 
    Generation of the value of a type (t), for structure type and array type, it is applied to the 
    name of the type.
    @return t the value of the TypeType element
*/
_VIRTUAL_ TypeType
javaGenerator::genTypeBase2 (TypeType t)
{
  //  t = TypeTypeGetUnamedType(t);
  if (TypeTestStructureClass (t) || TypeTestArrayClass (t) || TypeTestBundleClass (t))
    {
      t = NamedTypeGetTerm (t);
      return genTypeBase2 (t);
    }

  if (!TextForm->printtype (t, pk_true))
    setSomeProblem ();
  return t;
}


/** 
    It generates the code of the definition for an unknown external type (typename).
    This class contains only the keyword "UNDEF". 
    \code
    For the Signal declaration QueueAddress_type = external;
    It genrates the following Java code

    public class QueueAddress_type{
         UNDEF
    }

    \endcode
*/
void
javaGenerator::genUndefClassTypeDeclaration (ptrepstr typeName)
{
  ((inteprintjava *) TextForm)->javaPrintClassBegin (typeName, 1, pk_false, pk_true);
  TextForm->printkwnotimpl ();
  ((inteprintjava *) TextForm)->genClassEnd (pk_false);
}


/** 
    It generates the declaration of a enumeration type (t).
    \code 
    For the Signal type QueuingDiscipline_type = enum (FIFO, PRIORITY);
    It generates the following sequence (since Java 5.0)
        enum QueuingDiscipline_type {FIFO, PRIORITY};
    \endcode
*/
void
javaGenerator::genClassEnumDeclaration (TypeType t)
{
  ((inteprintjava *) TextForm)->javaPrintClassBegin (EnumerationTypeGetName (t), 1, pk_true, pk_false);
  TextForm->printdclenumtype (t);
  TextForm->closefile ();
}


/** 
    It generates the code of the declaration of a structured type (t) in a file foo.java 
    where foo is the name of the type.
    \code
    For the following Signal declaration 
    type foo = struct ( string Name;
                        real Entry_Point;
			integer_type Stack_Size;
		       );
    it generates the following Java code in the file foo.java

    class foo{
       static String Name;
       static float Entry_Point;
       static int Stack_Size;
       }
    \endcode
*/
_VIRTUAL_ void
javaGenerator::genClassStructDeclaration (TypeType t)
{
  ptrepstr structName;
  if (TypeTypeGetName (t, &structName))
    {
      ((inteprintjava *) TextForm)->javaPrintClassBegin (structName, 1, pk_true, pk_true);
      GenTypeStructureClass (t, pk_true);
      ((inteprintjava *) TextForm)->genClassEnd (pk_true);
    }
}

/**
   Generation of the class foo in foo.java file where foo is a structure type. 
   @param t (in) the structure type  
*/
_VIRTUAL_ void
javaGenerator::GenTypeStructureClass (TypeType t)
{
  GenTypeStructureClass (t, pk_false);
}


/**
   It generates the code of the declaration of a structured type (t),
   Each element is declared  "public static" when bpublic, or "static" otherwise.
*/
_VIRTUAL_ void
javaGenerator::GenTypeStructureClass (TypeType t, boolean bpublic)
{
  NameList NL;
  TypeList fieldsType;
  TypeSubList NSubLstType;
  boolean Goon;
  NameSubList NSubLst;

  NL = StructureTypeGetNameLst (t);
  NSubLst = NameListCast2SubList (NL);
  fieldsType = StructureTypeGetTypeLst (t);
  NSubLstType = TypeListCast2SubList (fieldsType);
  Goon = pk_true;
  while (Goon)
    {
      // ? bpublic : "public static" of "static"
      ((inteprintjava *) TextForm)->printPublicStatic (bpublic);
      cGenerator::GenTypeDefinition (NameSubListGetHead (NSubLst), TypeSubListGetHead (NSubLstType));
      if (NameSubListTestLast (NSubLst))
	Goon = pk_false;
      else
	{
	  NSubLst = NameSubListGetTail (NSubLst);
	  NSubLstType = TypeSubListGetTail (NSubLstType);
	}
    }
}


/** 
    \copydoc cGenerator::genReadEnv()
*/
_VIRTUAL_ void
javaGenerator::genReadEnv (ptexpcal exp, ptident processName, ptexpcal expEff, ptinsth h)
{
  genIncrIndexDelayTreeNode (expEff);
  genBlockOpeningOf (h);
  ((inteprintjava *) TextForm)->printReadEnv (EM->getName (exp), processName);
  genAssignEquivalence (expEff);
}


/**
   \copydoc cGenerator::genWriteEnv()
*/
_VIRTUAL_ void
javaGenerator::genWriteEnv (ptexpcal exp, ptident processName, ptexpcal expEff, ptinsth h)
{
  genBlockOpeningOf (h);
  ((inteprintjava *) TextForm)->printWriteEnv (EM->getName (exp), processName);
}


_VIRTUAL_ boolean
javaGenerator::genProcCallNodeWithCodePragma (ptINode node)
{
  ptpragma valPragma;
  ptexpcal x, linputs, loutputs, par;
  ptident id;
  TypeType vtype;

  if (NM->is_intrinsicfunc (node))
    {
      NM->get_procCall_arg (node, &id, &par, &linputs, &loutputs);
      if (existPredefinedPragma (NM->getPragmas (node), PRAGMA_CONVERSION))
	{
	  // for a conversion of an emuneration, result = X.ordinal().
	  x = EM->getNaryArgs (&linputs);
	  vtype = EM->getType (x);
	  if (TypeTestEnumerationClass (vtype))
	    {
	      genOutputSignal (EM->getNaryArgs (&loutputs), pk_false);
	      TextForm->printassign ();
	      genexpression (x, opnullary);
	      TextForm->printDot ();
	      ((inteprintjava *) TextForm)->printkwOrdinal ();
	      TextForm->printLP ();
	      TextForm->printRPSemiColonLine ();
	      return pk_true;
	    }
	}
    }
  if (!getPragmaStrKeyword (NM->getPragmas (node), PragmaGencodeLanguage, &valPragma))
    return pk_false;
  genPragmaCodeInt (node, valPragma);
  return pk_true;
}


/**
   \spe of cGenerator::genReadHeaderExterncall()
*/
_VIRTUAL_ void
javaGenerator::genReadHeaderExterncall (ptexpcal sig, boolean bdef)
{
  functionDeclaration = pk_true;
  ((inteprintjava *) TextForm)->printReadHeaderExterncall (EM->getName (sig), progName, bdef);
  functionDeclaration = pk_false;
}


/**
   \spe of cGenerator::genWriteHeaderExterncall().
*/
_VIRTUAL_ void
javaGenerator::genWriteHeaderExterncall (ptexpcal sig, boolean bdef)
{
  functionDeclaration = pk_true;
  ((inteprintjava *) TextForm)->printWriteHeaderExterncall (EM->getName (sig), progName, bdef);
  functionDeclaration = pk_false;
}

/**
   \spe of cGenerator::genDefReadingSignal()
*/
_VIRTUAL_ void
javaGenerator::genDefReadingSignal (ptexpcal aSig)
{
  ptident id;

  TextForm->blockIndentOpening ();
  genLocalIndexLstDeclaration (InternalIndexLst, EM->getNumberOfDimensions (aSig));
  genOpenIteration (aSig);
  id = EM->getName (aSig);
  gensigIO (aSig, pk_true);
  ((inteprintjava *) TextForm)->printequal ();
  ((inteprintjava *) TextForm)->printbodyreadhead (id, EM->getType (aSig));
  TextForm->printdclreadtail (id);
  TextForm->printBlockClosingBackIndent (getIndexStackSize ());
  TextForm->printreturn (pk_true);
  TextForm->printdclfuncend ();

  resetTravelIndexes ();
}

/**
   \spe of cGenerator::gensigIO()
*/
_VIRTUAL_ void
javaGenerator::gensigIO (ptexpcal aSig, boolean input)
{
  boolean b;

  b = pk_true;
  if (!input)
    {
      aSig = getExpressionToGenerate (aSig);
      if (!EM->IsSignal (aSig))
	{
	  genexpression (aSig, opnullary);
	  b = pk_false;
	}
    }

  if (b)
    {
      ((inteprintjava *) TextForm)->printAccessIOObject ();
      gensig (aSig, pk_false);
    }
}


/** 
    \spe of cGenerator::genComplexBinaryExpression().
    \code 
    Example: For x:= a+b on complexes, it generates 
    It generatres pK_complex.pK_complexPlus(a,b)
    \endcode
*/
_VIRTUAL_ void
javaGenerator::genComplexBinaryExpression (noperators ope, ptexpcal exp1, ptexpcal exp2, TypeType TT)
{
  TextForm->printkwComplex ();
  TextForm->printDot ();
  cGenerator::genComplexBinaryExpression (ope, exp1, exp2, TT);
}

/* --------------- externals ----------- */

/** 
    Generation of the assignment for an external unknown constant (vconst).
    \code
    Example, with "NUL_CODE" for the name of vconst, it generates
    static final public int NUL_CODE = UNDEF() ;
    \endcode
*/
_VIRTUAL_ void
javaGenerator::genExternalConstant (ptexpcal vconst)
{
  cfindecl ();
  cdebdecl ();
  ((inteprintjava *) TextForm)->printConstantSpec (pk_false);
  genSignalDeclaration (vconst);
  TextForm->printassign ();
  TextForm->printkwnotimplpopf ();
}

/** 
    \spe of cGenerator::genParameters().
    It generates the "external" parameters/constants of the graph.
    \code
    Example: with foo the name of the current graph, NUL_CODE,LIN_CODE, TICK_NUMBER some external parameters,
    it generates the following code
    class foo
    {
     ...
     // ==> parameters and indexes 
     static final public int NUL_CODE = UNDEF() ;
     static final public int LIN_CODE = UNDEF() ;
     static final public int TICK_NUMBER = UNDEF() ;
     ...
     }
   \endcode
 */
_VIRTUAL_ void
javaGenerator::genParameters ()
{
  ptlinstances lconst;
  ptexpcal vconst;

  lconst = AGRAPH->getParameters ();
  if (lconst == NILLINSTANCES)
    return;
  while (ptlinstances_get_next (&lconst, &vconst))
    if (EM->getOperatorClass (vconst) == variabl)
      genExternalConstant (vconst);
}


/**
   \spe of cGenerator::genConstantDeclaration().
   It calls genExternalConstant() for an external constant, genSignalDeclaration() otherwise.
*/
_VIRTUAL_ void
javaGenerator::genConstantDeclaration (ptexpcal exp)
{
  if (mustBeGenerated (EM->getNode (exp)))
    if (!TypeTestPredef (EM->getUnamedType (exp)))
      {
	if (!EM->IsExternalConstant (exp))
	  genSignalDeclaration (exp);
	else
	  genExternalConstant (exp);
      }
}

/** 
    \spe cGenerator::genExternalConstantsTypes()
    Generation of external types.
    For Java, the external constants ARE NOT generated in this file but in the foo.java file (see genConstantDeclaration()). 

    @return "true" if it exists some external types, "false" otherwise.

    The classes are generated in a file named foo_Externals.java, where foo is the name of the current graph.
    \code
    Example: it a type QueueAddress_type is declared in a Signal program, the following code is generated in foo_Externals.java
    public class QueueAddress_type{
    UNDEF
    }
    \endcode

*/
_VIRTUAL_ boolean
javaGenerator::genExternalConstantsTypes ()
{
  if (!TestExternalTypes ())
    return pk_false;

  TextForm->printdclexternfile (progName);
  genComment (comm_paramNotDefined);
  genExternalTypes ();
  TextForm->closefile ();
  return pk_true;
}

/** 
    \spe of cGenerator::genExternalDeclarationsFile()
    For Java, it calls genExternalConstantsTypes()
*/
_VIRTUAL_ void
javaGenerator::genExternalDeclarationsFile ()
{
  mkListOfBlackGreyBoxesDeclarations ();
  genExternalProcs ();
  genExternalConstantsTypes ();
}

/**
   Generation of the headers of reading operations.
   No effect for Java.
*/
_VIRTUAL_ void
javaGenerator::genInputReadFunctions ()
{
}

/**  Generation of  headers of writing operations.
     No effect for Java.
*/
_VIRTUAL_ void
javaGenerator::genOutputWriteFunctions ()
{
}

/**
   Code generation of the files for separated compiling. It is generated in foo.java file, where foo 
   is the name of the current graph.
   It calls genBodyClass() followed by genBodyCodeSeparate().
   An example of generated code:
   \include ExampleJavaSeparate
*/
_VIRTUAL_ void
javaGenerator::genBodySeparate ()
{
  genBodyClass ();
  genBodyCodeSeparate ();
}

/** 
    Code generation of the file foo.java for separated compiling with foo the name of the current graph.
    It calls:
    - gendeclarationsBody(), it generates the declarations
    - genInitialisations(), it generates the initialisations
    - cGenerator::genSubGraphs(), it generates the code of the clusters ("grey box" mode)
    - genBlackBoxCode(), it generates the code in "black box" mode
    - closefile(), closing the associated file, previously opened in genBodyClass().
 */
_VIRTUAL_ void
javaGenerator::genBodyCodeSeparate ()
{
  gendeclarationsBody ();
  genInitialisations ();
  cGenerator::genSubGraphs ();
  genBlackBoxCode ();

  TextForm->closefile ();
}


/**
   \spe of cGenerator::genDefWritingSignal().
   \code 
   For example, it generates, when the identifier of aSig is "x2"
   {
     {
      fwx2.ps.print (_bobj.x2);
      fwx2.ps.print (' ');
      }
      fwx2.ps.println ();
   }
   \endcode

*/
_VIRTUAL_ void
javaGenerator::genDefWritingSignal (ptexpcal aSig)
{
  ptident name, id;
  int n;
  TypeType t;

  name = EM->getName (aSig);
  id = getIOCompleteIdentifier (aSig);
  t = EM->getType (aSig);

  TextForm->blockIndentOpening ();
  genLocalIndexLstDeclaration (InternalIndexLst, EM->getNumberOfDimensions (aSig));
  n = genOpenIteration (aSig);
  TextForm->printLine ();
  ((inteprintjava *) TextForm)->printbodywritehead (id, t);
  gensigIO (aSig, pk_false);
  TextForm->printRPSemiColonLine ();
  ((inteprintjava *) TextForm)->printbodywritehead (id, t);
  TextForm->printstr (string2repstr ("' ')"));
  TextForm->printbodywritetail (name);
  TextForm->printBlockClosingBackIndent (getIndexStackSize ());
  resetTravelIndexes ();

  ((inteprintjava *) TextForm)->printbodywriteln (id);
}

/**
   Generation of the hearder of the global/external procedure or function.
   @return "true" if the header is generated, "false" otherwise (cast function)
   @param name (in) name of the procedure/function
   @param asFunc (in) at pk_true for a function, pk_false for a procedure.
   @param inputs (in) input formal parameters.
   @param outputs (in) output formal parameters.
   @param clusterEntry (in) at pk_false, the name of formal parameters are the names of the signals of the inputs/outputs lists,
           otherwise for a signal X of inputs/outputs, the name of the formal parameter is _X_.
   @param declaration (in) at pk_true, the header is completed by the separator of declarations.
*/
_VIRTUAL_ boolean
javaGenerator::genEntryPoint (ptident name, ptexpcal inputs, ptexpcal outputs, boolean asFunc, boolean clusterEntry, boolean declaration)
{
  outputs = NILEXPCAL;
  return cppGenerator::genEntryPoint (name, inputs, outputs, asFunc, clusterEntry, declaration);
}

/**
   Redefinition of cppGenerator::genTypeBase().
   It calls genTypeBase2() when t is not a structure/bundle type.
*/
_VIRTUAL_ void
javaGenerator::genTypeBase (TypeType t)
{
  if (TypeTestStructureClass (t))
    GenTypeStructureClass (t);
  else if (TypeTestBundleClass (t))
    GenTypeStructureClass (BundleTypeGetTerm (t));
  else
    genTypeBase2 (t);
}

/**
   It generates the string operator associated with a Signal operator (ope).
   It is a call to a method of a predefined class (pK_String) provided with Polychrony.
*/
_VIRTUAL_ void
javaGenerator::genCodeStringOperator (noperators ope)
{
  TextForm->printstr (string2repstr ("pK_String."));
  cGenerator::genCodeStringOperator (ope);
}


/**
   Generation of a list of signals (loutputs) as output parameters of an external call (className).
   The parameter bfirst is unused here.
*/
_VIRTUAL_ void
javaGenerator::genOutputsExternalNode (ptident className, ptexpcal loutputs, boolean bfirst)
{
  ptexpcal exp, saveloutputs;

  saveloutputs = loutputs;
  while (saveloutputs != NILEXPCAL)
    {
      exp = EM->getNaryArgs (&saveloutputs);
      ((inteprintjava *) TextForm)->printSetVariable (EM->getName (exp), className);
    }
}

/** 
    Generation of the code of the expression exp1 codop exp2 with t as type for codop a equal to (==) or smaller/equal to (<<=) operator.
*/
_VIRTUAL_ void
javaGenerator::genCmpOperator (noperators codop, ptexpcal exp1, ptexpcal exp2)
{
  TypeType t;

  t = EM->getUnamedType (exp1);
  if (TypeTestPredef (t) || TypeTestEnumerationClass (t))
    {
      TextForm->printLP ();
      genexpression (exp1, codop);
      if (codop == opegeg)
	genOperator (opegal);
      else
	genOperator (opiegal);
      genexpression (exp2, codop);
      TextForm->printRP ();
    }
  else
    cGenerator::genCmpOperator (codop, exp1, exp2);
}


/**
   Generation of the declaration of a signal (sig) as a formal parameter. 
   An identifier is always generated.
   @param sig (in) a signal.
   @param isoutput (in) must be "true" for an output parameter, "false" otherwise.
   @param bprintIdent (in) when it is "true", a name is printed.
   @param n (in) it is the order of the signal (sig) in the list of parameters.
*/
_VIRTUAL_ void
javaGenerator::genDclFormalParameter (ptexpcal sig, boolean isoutput, boolean bprintIdent, int n)
{
  TypeType t;

  t = EM->getUnamedType (sig);
  gentype (t);
  genFormalParameterIdent (sig, pk_false, n);
  genDclDimensions (t);
}


/** 
    It prints a reference to an external call called id.
    For Java, the identifier is prefixed by the name of the "external object" associated with the application.
*/
_VIRTUAL_ void
javaGenerator::genIdentExternalCall (ptident id)
{
  // It generates foo_eo_.id
  ((inteprintjava *) TextForm)->printJavaExternalObjectName ();
  TextForm->printDot ();
  TextForm->printident (id);
}
