/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   Shaper constructor for C++ language.
*/
  inteprintvhdl::inteprintvhdl ()
{
}

/**
   Destructor.
*/
_VIRTUAL_ inteprintvhdl::~inteprintvhdl ()
{
}

_VIRTUAL_ void
inteprintvhdl::printLangBeginComment()
{
  buf->writeAlfa ("--", 2);
}

_VIRTUAL_ void
inteprintvhdl::printLangEndComment()
{
  printLine (); 
}




_VIRTUAL_ const char*
inteprintvhdl::getSuffixCodeFile()
{
  return ".vhdl";
}

_VIRTUAL_ const char*
inteprintvhdl::getSuffixBodyFile()
{
  return "";
}



/**
   \spe of inteprintc::initialize() for C++ language.
   Moreover, it defines the object associated with the current graph: foo_obj_ is an object of the class foo, where vname=foo.
   This object is returned by getObjectName() method..
*/
_VIRTUAL_ void
inteprintvhdl::initialize (ptrepstr vname)
{
  gInitialize (vname);
  strtypefloat = string2repstr ("real");
  strtypeint = string2repstr ("integer");
}

_VIRTUAL_ void
inteprintvhdl::printkwvariable ()
{
  buf->writeAlfa ("variable", 8);
}


/**
   It prints the assign operator.
*/
_VIRTUAL_ void
inteprintvhdl::printassign ()
{
  buf->fit (4);
  buf->space ();
  printcolon ();
  printequal ();
  buf->space ();
}

_VIRTUAL_ void
inteprintvhdl::printPrefixPort()
{
  buf->writeAlfa ("vhdl_", 5);
}


_VIRTUAL_ void
inteprintvhdl::printEntityBegin()
{
  buf->writeAlfa ("entity", 6);
  buf->space ();
  printstr(progName);
  buf->space ();
  buf->writeAlfa ("is", 2);
  printLine ();
}

_VIRTUAL_ void
inteprintvhdl::printEntityEnd()
{
  buf->writeAlfa ("end", 3);
  buf->space ();
  printstr(progName);
  printSemiColonLine ();
}


_VIRTUAL_ void
inteprintvhdl::printinclude (ptrepstr nameInclude, char sepBegin, char sepEnd)
{
}

_VIRTUAL_ void
inteprintvhdl::printPortsBegin()
{
  buf->writeAlfa ("port", 4);
  buf->space ();
  printPO();
  printLine ();
}

_VIRTUAL_ void
inteprintvhdl::printIn()
{
  buf->writeAlfa ("in",2);
  buf->space ();
}

_VIRTUAL_ void
inteprintvhdl::printOut()
{
  buf->writeAlfa ("out",3 );
  buf->space ();
}

_VIRTUAL_ void
inteprintvhdl::printPortsEnd()
{
  printLine ();
  printRPSemiColonLine ();
}



_VIRTUAL_ void
inteprintvhdl::printArchitectureBegin()
{
  buf->writeAlfa ("architecture", 12);
  buf->space ();
  buf->writeAlfa ("behavior", 8);
  buf->space ();
  buf->writeAlfa ("of", 2);
  buf->space ();
  printstr(progName);
  buf->space ();
  buf->writeAlfa ("is", 2);
  printLine ();
  printkwbegin();
  buf->space ();
  printLine ();
  buf->writeAlfa ("Main:process", 12);
  printLine ();
}

_VIRTUAL_ void
inteprintvhdl::printArchitectureEnd()
{
  buf->writeAlfa ("end", 3);
  buf->space ();
  buf->writeAlfa ("process", 7);
  buf->space ();
  buf->writeAlfa ("Main", 4);
  printSemiColonLine ();
  buf->writeAlfa ("end", 3);
  buf->space ();
  buf->writeAlfa ("behavior", 8);
  printSemiColonLine ();
}

_VIRTUAL_ void
inteprintvhdl::printkwbegin()
{
  buf->writeAlfa ("begin", 5);
}

/** 
   Ggenerate wait until id = true:
 */
_VIRTUAL_ void
inteprintvhdl::printArchitectureBeginBody(ptident id)
{
 printkwbegin();
 printLine ();
 buf->writeAlfa ("wait", 4);
 buf->space ();
 buf->writeAlfa ("until", 5);
 buf->space ();
 printident(id);
 printequal ();
 printkwtrue (); 
 printSemiColonLine ();
}

_VIRTUAL_ void inteprintvhdl::OpenBlock()
{
  printLine ();
}


_VIRTUAL_ void inteprintvhdl::CloseBlock()
{
  printblockIfClosing ();
}

_VIRTUAL_ void
inteprintvhdl::printblockIfClosing ()
{
  backindent (1);
  printkwEndIf ();
  printLine ();
}

_VIRTUAL_ void inteprintvhdl::printkwthen ()
{
  buf->writeAlfa ("then ", 5);
}

_VIRTUAL_ void inteprintvhdl::printkwEndIf ()
{
 buf->writeAlfa ("end if;", 7);
}

/**
   It prints the or operator.
*/
_VIRTUAL_ void
inteprintvhdl::printkwor ()
{
  buf->writeAlfa ("or", 2);
}

/**
   It prints the or operator.
*/
_VIRTUAL_ void
inteprintvhdl::printkwand ()
{
  buf->writeAlfa ("and", 3);
}


/**
   It prints the xor operator.
*/
_VIRTUAL_ void
inteprintvhdl::printkworexclusif ()
{
  buf->writeAlfa ("xor", 3);
}
/**
   It prints the not operator.
*/
_VIRTUAL_ void
inteprintvhdl::printkwnot ()
{
  buf->writeAlfa ("not ", 4);
}

/** It prints the "modulo" operator. */
_VIRTUAL_ void
inteprintvhdl::printkwmod ()
{
 buf->writeAlfa ("mod ", 4);
}

/** It prints the "diff" operator. */
_VIRTUAL_ void
inteprintvhdl::printkwdiff ()
{
  buf->print2Chars ('/', '=');
}
/** It prints the "equal test" operator. */
_VIRTUAL_ void
inteprintvhdl::printkwequal ()
{
  buf->printChar('=');
}

 _VIRTUAL_ void inteprintvhdl::printkwelseif ()
{
 buf->writeAlfa ("elsif ", 6);
}


_VIRTUAL_ void
inteprintvhdl::printkwtrue ()
{
 buf->writeAlfa ("true", 4);
}


_VIRTUAL_ void
inteprintvhdl::printkwfalse ()
{
 buf->writeAlfa ("false", 5);
}


/**
   It generates the boolean constant "TRUE" when cb, "FALSE" otherwise.
*/
_VIRTUAL_ void
inteprintvhdl::printkwlogicons (boolean cb)
{
  if (cb) printkwtrue (); else printkwfalse ();
}

_VIRTUAL_ void inteprintvhdl::printParametersBegin()
{
  buf->writeAlfa ("generic", 7);
  printSpace ();
  printLP ();
  printLine ();
}

_VIRTUAL_ void inteprintvhdl::printParametersEnd()
{
  printRPSemiColonLine ();
}

/* generate:
   library work;
   use work.intrinsics.all;
*/
_VIRTUAL_ void
inteprintvhdl::printdcltypefile (ptident processName)
{
  buf->writeAlfa ("library", 7);
  printSpace ();
  buf->writeAlfa ("work", 4);
  printSemiColonLine ();
  buf->writeAlfa ("use", 3);
  printSpace ();
  buf->writeAlfa ("work", 4);
  buf->writeAlfa (".", 1);
  buf->writeAlfa ("intrinsics", 10);
  buf->writeAlfa (".", 1);
  buf->writeAlfa ("all", 3);
  printSemiColonLine ();
  printLine ();
}


/**
   It generates
   exit when not <prefix><id>;
   end loop <prefix><id>;
*/
_VIRTUAL_ void inteprintvhdl::printEndBlockMicroStep(ptrepstr prefix, ptident id)
{
  // "exit when not <prefix><id>;
  buf->writeAlfa ("exit", 4);
  printSpace ();

  printLoopName(prefix,id );

  printSpace ();
  buf->writeAlfa ("when ",4);
  printSpace ();
  printkwnot ();
  printstr (prefix);
  printident(id);
  printSemiColonLine ();

  // end loop <prefix><id>;
  buf->writeAlfa ("end loop", 8);
  printSpace ();
  printLoopName(prefix,id );
  printSemiColonLine ();
}

_VIRTUAL_ void inteprintvhdl::printkwloop()
{
  buf->writeAlfa ("loop", 4);
}

/**
   it generates "} while(<prefix><id>);"
*/
_VIRTUAL_ void inteprintvhdl::printBeginBlockMicroStep(ptrepstr prefix, ptident id)
{
  printLoopName(prefix,id);
  printSpace ();
  printcolon();
  printkwloop();
  printLine ();
}
