/* 
   Authors: Loic Besnard, Thierry Gautier.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** 
    Entry point of the code generation, with or without clusters, in non-separated mode.
    @param mainFile (in) pk_true if the file foo_main.c must be generated
    @param bodyFile (in) pk_true if the file foo_boby.c must be generated
    @param ioFile (in)  pk_true if the file foo_io.c must be generated
    @param commFile (in) not used
    \see{generate}
    
*/

/**
  Generation of the declaration of the identifier id whose type is t.
*/
_VIRTUAL_ void
vhdlGenerator::genIdentDeclaration (ptident id, TypeType t, boolean bvar)
{
  if (bvar) TextFormVHDL->printkwvariable ();
  TextFormVHDL->printSpace ();
  TextFormVHDL->printident (id);
  TextFormVHDL->printcolon ();
  cGenerator::genTypeBase (t); 
  // gentype (t); genDclDimensions (t); NOT YET IMPLEMENTED
}

/**
   It generates the call to the write function of the signal expEff at a clock (h).
   The form of the call is:  w_<processName_<ident1)> (ident2)), where 
   ident1(resp. ident2) denotes the identifier of the signal exp (resp. expEff).
*/
_VIRTUAL_ void
vhdlGenerator::genWriteEnv (ptexpcal exp, ptident processName, ptexpcal expEff, ptinsth h)
{
  ptexpcal exp1;

  genBlockOpeningOf (h);
  TextFormVHDL->printPrefixPort();
  printidentifier(exp);
  TextFormVHDL->printSpace ();
  TextFormVHDL->printkwLessthanEqual ();
  TextFormVHDL->printSpace ();
  exp = getExpressionToGenerate (exp);
  printidentifier(exp);
  TextFormVHDL->printSemiColonLine ();
}

/**
  Generation of the declaration of the signal exp when it is not a node of a delay tree.
*/
_VIRTUAL_ void
vhdlGenerator::genSignalDeclaration0 (ptexpcal exp, boolean bvar)
{
  genIdentDeclaration (EM->getName (exp), EM->getType (exp), bvar);
  ptexpcal vinit = EM->getInitialValue(exp);
  if (vinit != NILEXPCAL) {
    TextFormVHDL->printassign ();
    genexpression (vinit, opnullary);
  }
}


/**
  Generation of the declaration of the signal exp when it is not a node of a delay tree.
*/
_VIRTUAL_ void
vhdlGenerator::genSignalDeclaration (ptexpcal exp)
{
  genSignalDeclaration0(exp, pk_true);
  TextFormVHDL->printSemiColonLine ();
}


_VIRTUAL_ void vhdlGenerator::genParameters()
{
}

_VIRTUAL_ void  vhdlGenerator::genIdPort( ptident id, TypeType t, boolean bin, boolean blast)
{
  TextFormVHDL->printident (id);
  TextFormVHDL->printSpace ();
  TextFormVHDL->printcolon();
  if (bin) TextFormVHDL->printIn(); else TextFormVHDL->printOut();
  cGenerator::genTypeBase (t); 
  // gentype (t); genDclDimensions (t); NOT YET IMPLEMENTED
  if (!blast) TextFormVHDL->printSemiColonLine ();
}

_VIRTUAL_ void  vhdlGenerator::genPort( ptexpcal sig, boolean bin)
{
  TextFormVHDL->printPrefixPort();
  genIdPort (EM->getName (sig), EM->getType (sig), bin, pk_false);
}

_VIRTUAL_ void  vhdlGenerator::genPorts()
{
  ptexpcal sig;
  ptlinstances lsig;

  TextFormVHDL->printPortsBegin();

  lsig = Inputs;
  while (ptlinstances_get_next (&lsig, &sig)) 
    genPort(sig, pk_true);
  lsig = Outputs;
  while (ptlinstances_get_next (&lsig, &sig))
    genPort (sig, pk_false);

  // MasterClock declared as Port.
  genIdPort (TickId, BooleanType, pk_true, pk_true); 
  
  TextFormVHDL->printPortsEnd ();
}

_VIRTUAL_ void vhdlGenerator::genEntity()
{
  TextFormVHDL->printEntityBegin();
  genVHDLParameters();
  genPorts();
  TextFormVHDL->printEntityEnd();
}

_VIRTUAL_ void vhdlGenerator::genArchitecture()
{
  TextFormVHDL->printArchitectureBegin();
  gendeclarationsBody ();
  TextFormVHDL->printArchitectureBeginBody(TickId);
  //  genInitialisations ();
  genSubGraphs ();
  genEntryPointStepInitialize ();
  genScheduler ();
  genFinalyze ();
  TextFormVHDL->printArchitectureEnd();
}

_VIRTUAL_ void vhdlGenerator::genBody ()
{
  TextFormVHDL->printbodyfilebegin (progName, pk_false); /* opening the file foo.vhdl */
  dclTypes ();
  genEntity();
  genArchitecture();
  TextFormVHDL->closefile ();	/* closing the file foo.vhdl */
}

/**
   Generating the declaration of types referenced in the graph in the file foo_types.h, excluding the 
   external types (see the note below).
   It calls: 
   - printdcltypefile() opening the file
   - genNonExternalTypes() genrating
   - closefile() closing the file.
 
   \note the external types are generated in a specific file with other "external undefined" objects
   ( foo_externalsUNDEF_LIS.h where foo is the name of the current graph).
*/
_VIRTUAL_ void
vhdlGenerator::dclTypes ()
{
  TextFormVHDL->printdcltypefile (progName);
}



_VIRTUAL_ void vhdlGenerator::gencode ()
{
  // genExternalDeclarationsFile ();	/* file foo_externals.h */
  // genMain ();
  genBody ();
}




/** 
   Generation of  the entry foo_STEP_initialize() that contains the constant signals of the program.
*/
_VIRTUAL_ void
vhdlGenerator::genEntryPointStepInitialize ()
{
  set_booleanClocks_undefined (tick);
}


/** 
    Generation of the entry foo_STEP_finalize().
    @param ldelays (in) set of delays of the program;
    @param modeBlackBox (in) at pk_true when the graph is without clusters, "false" otherwise
*/
_VIRTUAL_ void
vhdlGenerator::genStepFinalizeNotTrain (ptlINode ldelays, boolean modeBlackBox)
{
  genMicroStepDef ();
  genOrderedNodes (ldelays, tick, modeBlackBox, pk_false);
  genUpdateStateVariables ();
  genCodeEndochronousWithInternalClockRootEnd ();
}

_VIRTUAL_ void
vhdlGenerator::genSchedulerBegin ()
{

}


_VIRTUAL_ void
vhdlGenerator::genSchedulerEnd ()
{
  
}

_VIRTUAL_ void
vhdlGenerator::genCodeReadEnv (ptexpcal exp, ptident processName, ptexpcal expEff)
{  
  printidentifier(expEff);
  TextFormVHDL->printSpace ();
  TextFormVHDL->printassign ();
  TextFormVHDL->printSpace ();
  TextFormVHDL->printPrefixPort();
  printidentifier(expEff);
  TextFormVHDL->printSemiColonLine (); 
}


/**
   For vhdl, it is generated by a if (COND) expg:=expd else expg:=FALSE
*/
_VIRTUAL_ boolean vhdlGenerator::genLeft_tt(ptinsth h, ptexpcal expg, ptexpcal expd)
{
  ptexpcal exp1, exp2;
  ptinsth vhexp1;

 genBlockOpeningOf (h);

 EM->getBinaryArgs (expd, &exp1, &exp2);
 if (EM->getSynchronousExprHorl (exp1, &vhexp1))
    {
      TextFormVHDL->printkwif ();

      TextFormVHDL->printLP ();
      genClockExpression (vhexp1, opnullary);
      TextFormVHDL->printRP ();
      TextFormVHDL->printkwthen ();
      TextFormVHDL->printSpace ();
      gensig (expg, pk_false);
      TextFormVHDL->printassign ();
      genexpression (exp1, opleft_tt);
      TextFormVHDL->printSemiColon ();
      TextFormVHDL->printSpace ();
      TextFormVHDL->printkwelse ();
      TextFormVHDL->printSpace ();
      gensig (expg, pk_false);
      TextFormVHDL->printassign ();
      TextFormVHDL->printkwlogicons (pk_false);
      TextFormVHDL->printSemiColon ();
      TextFormVHDL->printSpace ();
      TextFormVHDL->printkwEndIf ();
      TextFormVHDL->printLine ();
      return pk_true;
    }
 // pK_messenger->Caseerror (currentLine, currentFile);
  return pk_false;
}



_VIRTUAL_ void
vhdlGenerator::genVHDLParameters()
{
  ptlinstances lconst;
  ptexpcal vconst;
  boolean bfirst;

  bfirst = pk_true;
  lconst = Parameters;
  if (lconst != NILLINSTANCES)
    {
      cdebdecl ();
      while (ptlinstances_get_next (&lconst, &vconst))
	if (EM->getOperatorClass (vconst) == variabl)	// Filtering the type declarations.
	  {
	    if (bfirst) { 
	      TextFormVHDL->printParametersBegin();
	      bfirst = pk_false;
	    }
	    else
	      TextFormVHDL->printSemiColonLine ();
	    genSignalDeclaration0 (vconst, pk_false);
	  }
      if (! bfirst) TextFormVHDL->printParametersEnd();
    }
}


_VIRTUAL_ void
vhdlGenerator::genVHDLConstants()
{
  ptlinstances lconst;
  ptexpcal vconst;

  lconst = Constants;
  while (ptlinstances_get_next (&lconst, &vconst))
    if (EM->IsExternalConstant (vconst))
      if (mustBeGenerated (vconst))
	{
	  TextFormVHDL->printkwdefineExternal (EM->getName (vconst));
	}
}


/**
  Generation of the declaration of the identifier id whose type is t.
*/
_VIRTUAL_ void
vhdlGenerator::genIdentDeclaration (ptident id, TypeType t)
{
  genIdentDeclaration (id, t, pk_true);
  TextFormVHDL->printSemiColonLine ();
}

