/* 
   Authors: Loic Besnard, Thierry Gautier.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** 
    Entry point of the code generation, with or without clusters, in non-separated mode.
    @param mainFile (in) pk_true if the file foo_main.c must be generated
    @param bodyFile (in) pk_true if the file foo_boby.c must be generated
    @param ioFile (in)  pk_true if the file foo_io.c must be generated
    @param commFile (in) not used
    \see{generate}
    
*/


/** 
   This entry point is not generated for VERILOG.
   The call is replaced by the code of the initialisations (for synthetizable tasks).
*/
_VIRTUAL_ void
VLGenerator::genEntryPointStepInitialize ()
{
}


/**
   Generate the identifiers of the signals of a list(lsig).
   The list-separator is generated first if begSep.
 */
_VIRTUAL_ boolean VLGenerator::genIdents (ptlinstances lsig, boolean begSep)
{
  ptexpcal sig;
  boolean b;

  if (begSep) TextFormVL-> printComma ();
  b = pk_false;
  while (ptlinstances_get_next (&lsig, &sig))
    {
      TextFormVL->printident(EM->getName (sig));
      if (lsig != NILLINSTANCES) TextFormVL-> printComma ();
      b = pk_true;
    }
  return b;
}


_VIRTUAL_ void VLGenerator::genDclInputPort (ptexpcal sig)
{
  // inputs are wires, not regs.
  TextFormVL->printInputdecl( EM->getType (sig),  EM->getName (sig));
}


_VIRTUAL_ void VLGenerator::genDclInputPorts ()
{
  ptexpcal sig;
  ptlinstances lsig = Inputs;

   while (ptlinstances_get_next (&lsig, &sig))
     genDclInputPort (sig);

   TextFormVL->printVLDecl();
}

_VIRTUAL_ void VLGenerator::genDclOutputPort (ptexpcal sig)
{
  TextFormVL->printOutputdecl( EM->getType (sig),  EM->getName (sig));
}

_VIRTUAL_ void VLGenerator::genDclOutputPorts ()
{
  ptexpcal sig;
  ptlinstances lsig = Outputs;
  while (ptlinstances_get_next (&lsig, &sig))
    genDclOutputPort (sig);
}

_VIRTUAL_ void VLGenerator::gendeclarationsModule ()
{
  TextFormVL->printBeginModuleHeader(progName);
  boolean b = genIdents (Outputs, genIdents (Inputs, pk_false) );
  if (b) TextFormVL-> printComma ();
  TextFormVL->printTickId();
  if (bSynthetisableCode)
    { 
      TextFormVL-> printComma ();
      TextFormVL->printInitId();
    }

  TextFormVL->printEndModuleHeader(progName);
  
}

_VIRTUAL_ void VLGenerator::gendeclarationsBody ()
{
  ptrepstr savePrefix;

  gendeclarationsModule ();

  savePrefix = getprefix ();
  setPrefix (nilstr);
  genDclParametersAndIndexes ();
  genDclInputPorts ();
  genDclOutputPorts ();

  genDclSignals (Locals);
  genStateVariables ();
  genExternalSignalObjects ();
  genInternalIndexes ();
  genAffineSampleCounterDeclaration ();
  genMicrostepDecl ();
  TextFormVL->printLine ();
  genCmpOperators ();
  setPrefix (savePrefix);
}

/**
   It finalizes the generation of the foo_initialize entry point.
   It is isolated to allow specialisations.
*/
_VIRTUAL_ void
VLGenerator::genInitialisationsFinalyze ()
{
if ( ! bSynthetisableCode  ) 
  {
    TextFormVL->printEndTask();
    TextFormVL->printGenExecuteInitialTask();
  }
}


_VIRTUAL_ void
VLGenerator::StepFinalize(boolean modeBlackBox)
{
  TextFormVL->printStepFinalyze(modeBlackBox);
}

/**
   It generates the call to the write function of the signal expEff at a clock (h).
   The form of the call is:  w_<processName_<ident1)> (ident2)), where 
   ident1(resp. ident2) denotes the identifier of the signal exp (resp. expEff).
*/
_VIRTUAL_ void
VLGenerator::genWriteEnv (ptexpcal exp, ptident processName, ptexpcal expEff, ptinsth h)
{
  ptexpcal exp1;

  if (isNodeAtInitialClock (EM->getNode (exp), &exp1))
    {
      genBlockOpeningOf (h);
      ptident id = EM->getName (exp);
      TextForm->printident(id);
      TextForm->printassign ();
      genexpression (exp1, opnullary);
      TextForm->printSemiColonLine ();
    }
}


_VIRTUAL_ void
VLGenerator::genCodeReadEnv (ptexpcal exp, ptident processName, ptexpcal expEff)
{  }



_VIRTUAL_ void VLGenerator::gencode ()
{
  // genExternalDeclarationsFile ();
  // dclTypes ();
  genBody ();
}

_VIRTUAL_ void VLGenerator::genBody ()
{
  /* opening the file foo.vh */
  TextFormVL->printbodyfilebegin (progName, pk_false); 

  // dclTypes ();
  TextFormVL->printPredefinedTypesIncludes();
  genArchitecture();

  /* closing the file foo.vl */
  TextFormVL->closefile ();	
}

_VIRTUAL_ void
VLGenerator::genSchedulerBegin ()
{
  TextForm->printiterfunchead ();
  if (bSynthetisableCode)
    {
      TextFormVL->printOnResetBegin();
      genInitialisations ();
      TextFormVL->printOnResetEnd();
    }
}


_VIRTUAL_ void VLGenerator::genArchitecture()
{
  gendeclarationsBody ();
  
  TextFormVL->printPredefinedFunctionsIncludes();
  if ( ! bSynthetisableCode )  genInitialisations ();
  genSubGraphs ();
  genEntryPointStepInitialize ();
  genScheduler ();
  genFinalyze ();
}


_VIRTUAL_ void
VLGenerator::genStepFinalizeNotTrain (ptlINode ldelays, boolean modeBlackBox)
{
  genMicroStepDef ();
  genOrderedNodes (ldelays, tick, modeBlackBox, pk_false);
  genUpdateStateVariables ();
  genInitClocks();
  genCodeEndochronousWithInternalClockRootEnd ();
}


_VIRTUAL_ void
VLGenerator::genInitClocks()
{
  set_booleanClocks_undefined (tick);
}
