/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   Shaper constructor for C++ language.
*/
  inteprintVL::inteprintVL ()
{
}

/**
   Destructor.
*/
_VIRTUAL_ inteprintVL::~inteprintVL ()
{
}

/**
   \spe of inteprintc::initialize() for C++ language.
   Moreover, it defines the object associated with the current graph: foo_obj_ is an object of the class foo, where vname=foo.
   This object is returned by getObjectName() method..
*/
_VIRTUAL_ void
inteprintVL::initialize (ptrepstr vname)
{
  gInitialize (vname);
}

_VIRTUAL_ void
inteprintVL::printPrefixPort()
{
  buf->writeAlfa ("verilog_", 8);
}


_VIRTUAL_ void
inteprintVL::printkwtrue ()
{
 buf->writeAlfa ("`TRUE", 5);
}


_VIRTUAL_ void
inteprintVL::printkwfalse ()
{
 buf->writeAlfa ("`FALSE", 6);
}


/**
   It generates the boolean constant "TRUE" when cb, "FALSE" otherwise.
*/
_VIRTUAL_ void
inteprintVL::printkwlogicons (boolean cb)
{
  if (cb) printkwtrue (); else printkwfalse ();
}

_VIRTUAL_ void inteprintVL::printParametersBegin()
{
  buf->writeAlfa ("generic", 7);
  printSpace ();
  printLP ();
  printLine ();
}

_VIRTUAL_ void inteprintVL::printParametersEnd()
{
  printRPSemiColonLine ();
}

_VIRTUAL_ void
inteprintVL::printPredefinedTypesIncludes()
{
  buf->writeAlfa ("`include", 8);
  printSpace ();
  buf->printChar ('"');
  buf->writeAlfa ("VerilogPredefTypes", 18);
  buf->writeAlfa (".", 1);
  buf->writeAlfa ("vl", 2);
  buf->printChar ('"');
  printLine ();
}

_VIRTUAL_ void
inteprintVL::printPredefinedFunctionsIncludes()
{
  buf->writeAlfa ("`include", 8);
  printSpace ();
  buf->printChar ('"');
  buf->writeAlfa ("VerilogPredefFunctions", 22);
  buf->writeAlfa (".", 1);
  buf->writeAlfa ("vl", 2);
  buf->printChar ('"');
  printLine ();
}



/**
   It generates
   exit when not <prefix><id>;
   end loop <prefix><id>;
*/
_VIRTUAL_ void inteprintVL::printEndBlockMicroStep(ptrepstr prefix, ptident id)
{
  printBlockClosingBackIndent ();
}

/**
   Generate "id = true; while(id) begin"
   Replace the do {... id= } while(id) of C language.
*/
_VIRTUAL_ void inteprintVL::printBeginBlockMicroStep(ptrepstr prefix, ptident id)
{
  printident(id);
  printassign();
  printkwtrue ();
  printSemiColonLine ();
  printkwWhile ();
  buf->printChar ('(');
  printident(id);
  buf->printChar (')');
  printOpenBlockIndent ();
  printLine ();
}

// ------------------------  NEW 

_VIRTUAL_ const char*
inteprintVL::getkwInclude()
{
  return "`include";
}



_VIRTUAL_ void
inteprintVL::printBeginTask(ptident id, boolean bsep)
{
  buf->writeAlfa ("task", 4);
  printSpace();
  printident (id);
  printSpace();
  if (bsep) printClassDataArea (); // ??
  printSemiColonLine ();
  buf->writeAlfa ("begin", 5);
  printLine ();
  indent ();
}

_VIRTUAL_ void
inteprintVL::printEndTask ()
{
  backindent (1);
  buf->writeAlfa ("end", 3);
  printLine ();  
  buf->writeAlfa ("endtask", 7);
  printLine ();  
  printLine ();
}


_VIRTUAL_ void
inteprintVL::printdclinitializefunchead (boolean bsep)
{
  if (! bSynthetisableCode )
    printBeginTask(inifuncname, bsep);
}


_VIRTUAL_ void
inteprintVL::printentrypointstepinitialize(ptident graphName, boolean bsep)
{
  ptident idStepFinalize = make_stepInitIdent (graphName);
  printBeginTask(idStepFinalize, bsep);
}



_VIRTUAL_ const char*
inteprintVL::getSuffixCodeFile()
{
  return ".vl";
}

_VIRTUAL_ const char*
inteprintVL::getSuffixBodyFile()
{
  return "";
}

_VIRTUAL_ const char *
inteprintVL::getHeaderSuffix ()
{
  return getSuffixCodeFile();
}

_VIRTUAL_ void
inteprintVL::printbodyfileIncludes(ptident ProcessName, boolean separate)
{
  printLine ();
  printincludefiles (ProcessName);
  printLine ();
}


_VIRTUAL_ void
inteprintVL::printBeginModuleHeader(ptident ProcessName)
{
  buf->writeAlfa ("module", 6);
  printSpace();
  printident(ProcessName);
  printchar ('(');
}


_VIRTUAL_ void
inteprintVL::printEndModuleHeader(ptident ProcessName)
{
  printRPSemiColonLine ();
  indent();
}

_VIRTUAL_ void
inteprintVL::printkwstatic ()
{
}

_VIRTUAL_ void
inteprintVL::printTickId()
{
  printident(TickId);
}

_VIRTUAL_ void
inteprintVL::printInitId()
{
  printident(ResetId);
}


_VIRTUAL_ void
inteprintVL::printiterfunchead ()
{
  buf->writeAlfa ("always", 6);
  printSpace();
  printchar('@');
  printchar ('(');
  buf->writeAlfa ("posedge", 7);
  printSpace();
  printTickId();
  if (bSynthetisableCode)
    {
      buf->writeAlfa (" or ", 4);
      buf->writeAlfa ("posedge", 7);
      printSpace();
      printInitId();
    }
  printchar (')');
  printLine ();
  buf->writeAlfa ("begin", 5);
  printLine ();
  indent();
}


_VIRTUAL_ void
inteprintVL::printiterfuncEnd ()
{
  backindent (1);
  buf->writeAlfa ("end", 3);
  printLine ();
  backindent (1);
  if (bSynthetisableCode)
    {
      buf->writeAlfa ("end", 3);
      printLine ();
    }
  buf->writeAlfa ("endmodule", 9);
  printLine ();
}

_VIRTUAL_ void inteprintVL::printStepFinalyze(boolean bbox)
{
  backindent (1);
  buf->writeAlfa ("end", 3);
  printLine ();
  printLine ();
}



_VIRTUAL_ void
inteprintVL::printentrypointstepfinalize (ptident graphName, boolean bsep)
{
  ptident id;

  id = make_stepFinalIdent (graphName);
  printBeginTask(id, bsep);
}


_VIRTUAL_ void
inteprintVL::printGenExecuteInitialTask()
{
  buf->writeAlfa ("initial", 7);
  printSpace ();
  printident(inifuncname);
  printSemiColonLine ();
  printLine ();
}


_VIRTUAL_ void
inteprintVL::printInputdecl(TypeType t, ptident id)
{
  buf->writeAlfa ("input", 5);
  printSpace ();
  printWireType(t);
  printident(id);
  printSemiColonLine ();
}

_VIRTUAL_ void
inteprintVL::printOutputdecl(TypeType t, ptident id)
{
  buf->writeAlfa ("output", 6);
  printSpace ();
  printRegType(t);
  printSpace ();
  printident(id);
  printSemiColonLine ();
}


/**
   It generates the key-word for a boolean type.
*/
_VIRTUAL_ void
inteprintVL::printkwlogical ()
{
  buf->printAlfa ("reg", 3);
}

/**
   It generates the key-word for an int.
*/
_VIRTUAL_ void
inteprintVL::printkwint ()
{
  buf->printAlfa ("integer", 7);
}

/**
   It generates the key-word for a real.
*/
_VIRTUAL_ void
inteprintVL::printkwfloat ()
{
  buf->printAlfa ("real", 4);
}

_VIRTUAL_ void
inteprintVL::printLBracket ()
{
  buf->printAlfa ("begin",5 );
}

_VIRTUAL_ void
inteprintVL::printRBracket ()
{
  buf->printAlfa ("end",3 );
}


_VIRTUAL_ void inteprintVL::OpenBlock()
{
  buf->printAlfa ("begin",5 );
  printLine ();
}


/** It closes a block statement. */
_VIRTUAL_ void
inteprintVL::printblockclosing ()
{
  buf->printAlfa ("end", 3 );
  printLine ();
}

_VIRTUAL_ void inteprintVL::printVLDecl()
{
  printInputdecl( BooleanType, TickId);
  if ( bSynthetisableCode )
    printInputdecl( BooleanType, ResetId);
}


_VIRTUAL_ void inteprintVL::printOnResetBegin()
{
  printkwif ();
  printLP ();
  printident (ResetId);
  printRP ();
  printSpace();
  printOpenBlockIndent (); //  OpenBlock();
}


_VIRTUAL_ void inteprintVL::printOnResetEnd()
{

  printBlockClosingBackIndent (); // printblockclosing ();

  printkwelse ();
  printOpenBlockIndent (); //  OpenBlock();
}


_VIRTUAL_ void
inteprintVL::init(boolean bs)
{
  bSynthetisableCode =bs;
  TickId = string2ident("VL_TICK");
  ResetId = string2ident("VL_RESET");
}


