/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
  Constructor.\n
  The functionality is applied on a graph (g).\n
*/
  Colouring::Colouring (ptDCG g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  DCplusGraph = isEventClockedFormat (AGRAPH->getFormatLevel ());
  bMarkLastContributionClockDefault = pk_true;
  lboundary = NILLINODE;
  reqDepClocks = pk_true;
}

/**
   Destructor.
 */
_VIRTUAL_ Colouring::~Colouring ()
{
}


/**
   It colors the graph (parameter of the constructor) using a mark (m) and 
   selecting or not the last contribution clock of the default operator (when it is "true", they are marked).
   The operation consists to mark with the mark (m) the clocks/nodes useful for the production of the outputs and the reading of the inputs.
 */
void
Colouring::apply (typeMark m, boolean b, boolean withClockDependence)
{
  bMarkLastContributionClockDefault = b;
  reqDepClocks = withClockDependence;
  aMark = m;
  colouring ();
}

/**
   It colors the nodes of a list (ln) of a graph (parameter of the constructor) using a mark (m) and 
   selecting or not the last contribution clock of the default operator (when "true" they are marked).
   Moreover, a boundary (lb) is given: for a node of this list the operation is NOT applied to its operands.
   It can be use, for example, to mark the nodes of a subgraph.
 */
void
Colouring::apply (ptlINode ln, ptlINode lb, typeMark m, boolean b)
{
  bMarkLastContributionClockDefault = b;
  lboundary = lb;
  aMark = m;
  marklNodes (ln);
}
