/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_ boolean CommManager::existPragma(ptlpragmas lprag, ptrepstr commName, ptexpcal aSig, ptpragma * aPragma)
{
  ptrepstr name;
  ptlinstances objects;
  ptexpcal id;

  while (ptlpragmas_get_next(&lprag, aPragma))
    {
      name = getPragmaName(*aPragma);
      if (streq(name, commName))
	{
	  objects = (ptlinstances) getPragmaObjects(*aPragma);
	  if (ptlinstances_get_next(&objects, &id))
	    if (aSig == id)
	      return pk_true;
	}
    }
  return pk_false;
}

void CommManager::removeEnv(ptDCG g, ptexpcal aSig)
{
  ptpragma vp;
  ptlpragmas l;

  l = g->getPragmas();
  if (existPragma(l, vtrEnv, aSig, &vp))
    if (ptlpragmas_remove_elem(&l, vp))
      g->setPragmas(l);
}

/**
   It adds to a graph (g) a pragma that specifies a communication:
   - "Sending a signal (aSig) to a process (id) with the stamp-th stamp", when boutput
   - "Receiving a signal (aSig) from a process (id) with the stamp-th stamp", when not boutput

   commName is the name associated with the predefined pragma PRAGMA_COMM_RECEIVE or PRAGMA_COMM_SEND
 */
ptpragma CommManager::genComm(ptDCG g, ptexpcal aSig, ptrepstr commName, ptident id, integer stamp, boolean boutput)
{
  ptlinstances objects;
  ptrepstr val1, val2;
  ntrees stmt;
  ptpragma vp;
  boolean done;

  objects = ptlinstances_insert(NILLINSTANCES, aSig);
  val1 = integer2repstr(stamp);
  getDoubleQuotedString(val1, &val1);
  if (id != NILIDENT)
    {
      val2 = ident2repstr(id);
      getDoubleQuotedString(val2, &val2);
    }
  else
    val2 = nilstr;

  done = pk_false;
  if (boutput)
    if (existPragma(g->getPragmas(), commName, aSig, &vp))
      {
	stmt = getPragmaStatement(vp);
	stmt = addEnonceValue(stmt, val1);
	if (!streq(val2, nilstr))
	  stmt = addEnonceValue(stmt, val2);
	done = pk_true;
      }
  if (!done)
    {
      stmt = addEnonceValue(NILTREE, val1);
      if (!streq(val2, nilstr))
	stmt = addEnonceValue(stmt, val2);
      vp = newpragma(commName, (ptObjPragma) objects, stmt);
      g->addPragma(vp);
    }
  return vp;
}

/**
   Goal : Building the containers around the processes.
   It specifies, as pragmas, the communications for two graphs g1 and g2.
   For each signal X, output of g1 and input of g2, it adds 
   - in g1 : output {X} "name(g1), value(stamp)"
   - in g2 : input {X} "name(g2), value(stamp)"

   \see genComm()
*/
boolean CommManager::addCommunicationsFromTo(ptINode aSubGraph1, ptDCG g1, ptINode aSubGraph2, ptDCG g2, integer * stamp)
{
  integer *markg1, *markg2;
  ptlinstances lcomm, lsave;
  ptexpcal aSig, aSig1, aSig2;
  boolean b;

  lcomm = ptlinstances_intersection(NM->getInodelOutputFlows(aSubGraph1), NM->getInodelInputFlows(aSubGraph2));
  if (lcomm == NILLINSTANCES)
    return pk_true;
  lsave = lcomm;
  b = pk_true;
  if (!NM->getAttribute(aSubGraph1, USED_MARK_EXTRACTION, (attributeValue *) & markg1))
    b = pk_false;
  else if (!NM->getAttribute(aSubGraph2, USED_MARK_EXTRACTION, (attributeValue *) & markg2))
    b = pk_false;
  else
    {
      while (b && ptlinstances_get_next(&lcomm, &aSig))
	if (!EM->getAttribute(aSig, *markg1, (attributeValue *) & aSig1))
	  b = pk_false;
	else if (!EM->getAttribute(aSig, *markg2, (attributeValue *) & aSig2))
	  b = pk_false;
	else
	  {
	    (*stamp)++;
	    genComm(g1, aSig1, vtrOutput, g2->getName(), *stamp, pk_true);
	    removeEnv(g2, aSig2);
	    genComm(g2, aSig2, vtrInput, g1->getName(), *stamp, pk_false);
	  }
    }
  ptlinstances_free(lsave);
  return b;
}

boolean CommManager::getRunOnAnnotation(ptlpragmas lp, integer * val)
{
  ptpragma aPragma;
  ptrepstr EnonceValue;
  ntrees tEnonce;

  if (!getPragmaStrKeyword(lp, getPredefinedPragma(PRAGMA_RUN_ON), &aPragma))
    return pk_false;
  tEnonce = getPragmaStatement(aPragma);
  EnonceValue = nstringof(nchild(1, tEnonce));
  *val = ExtractInteger(EnonceValue);
  return pk_true;
}

boolean CommManager::getNextGraphWithSignal(ptexpcal aSig, ptlINode * lSubGraphs, ptlDCG * lsg, ptDCG * g, ptexpcal * aSig1)
{
  ptINode sg;
  integer *markg;

  while (ptlINode_get_elem(lSubGraphs, &sg))
    {
      ptlPointer_get_elemNext(lsg, (pointer *) g);
      if (NM->getAttribute(sg, USED_MARK_EXTRACTION, (attributeValue *) & markg))
	if (EM->getAttribute(aSig, *markg, (attributeValue *) aSig1))
	  return pk_true;
    }
  return pk_false;
}

void CommManager::genCommEnvTo(ptexpcal aSig, integer tag, ptident id)
{
  // environment {aSig} tag "id" ; See the use with "topology".
}

boolean CommManager::addCommIO(ptlINode lSubGraphs, ptlDCG lsg, integer * stamp)
{
  ptlinstances lin, lout;
  ptexpcal aSig, aSig1;
  ptDCG g;
  ptINode sg;
  ptpragma prag;
  ptlDCG lsg1;
  ptlINode lSubGraphs1;
  ptident id;
  boolean bRunOn;

  id = NILIDENT;
  lin = inputs;
  while (ptlinstances_get_next(&lin, &aSig))
    if (EM->IsInputFlow(aSig))
      {
	lsg1 = lsg;
	lSubGraphs1 = lSubGraphs;
	while (getNextGraphWithSignal(aSig, &lSubGraphs1, &lsg1, &g, &aSig1))
	  {
	    (*stamp)++;
	    prag = genComm(g, aSig1, vtrEnv, NILIDENT, *stamp, pk_false);
	    genCommEnvTo(aSig, *stamp, id);
	  }
      }

  lout = outputs;
  while (ptlinstances_get_next(&lout, &aSig))
    if (EM->IsOutputFlow(aSig))
      {
	lsg1 = lsg;
	lSubGraphs1 = lSubGraphs;
	while (getNextGraphWithSignal(aSig, &lSubGraphs1, &lsg1, &g, &aSig1))
	  {
	    (*stamp)++;
	    prag = genComm(g, aSig1, vtrEnv, NILIDENT, *stamp, pk_true);
	    genCommEnvTo(aSig, *stamp, id);
	  }
      }
  return pk_true;
}

/**
   It adds the addressees of an input-signal (aSig) into a pragma (prag).
   For example, 
   environment {b} "2" "P3" specify that the signal (input of the environment) b must be send to the process P3 using the tag 2.
*/
void CommManager::addUsersOfInputExceptIn(ptpragma prag, ptexpcal aSig, ptlINode lSubGraphs, ptlDCG lsg, ptINode sg)
{
  ntrees stmt;
  ptINode aSubGraph;
  ptlinstances l;
  ptDCG g;
  ptrepstr vidstr;

  if (lSubGraphs != NILLINODE)
    {
      stmt = getPragmaStatement(prag);
      while (ptlINode_get_elem(&lSubGraphs, &aSubGraph))
	{
	  ptlPointer_get_elemNext(&lsg, (pointer *) & g);
	  if (sg != aSubGraph)
	    {
	      l = NM->getInodelInputFlows(aSubGraph);
	      if (ptlinstances_isin_exp(l, aSig))
		{
		  getDoubleQuotedString(ident2repstr(g->getName()), &vidstr);
		  stmt = addEnonceValue(stmt, vidstr);
		}
	    }
	  setPragmaStatement(prag, stmt);
	}
    }
}

boolean CommManager::getSubGraph_associatedGraph(integer val, ptlINode lSubGraphs, ptlDCG lsg, ptINode * sg, ptDCG * g)
{
  ptINode aSubGraph;
  boolean found;
  integer i;

  found = pk_false;
  while (!found && ptlINode_get_elem(&lSubGraphs, sg))
    {
      ptlPointer_get_elemNext(&lsg, (pointer *) g);
      if (getRunOnAnnotation(NM->getPragmas(*sg), &i))
	found = (i == val);
    }
  return found;
}

