/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define isEventClockedFormat(t) ((t == DCG_BASIC) ||(t == DCG_POLY) || (t == DCG_ENDO))
#define MYFONTSIZE 20
#define MYWIDTH 0.05
#define MYHEIGHT 0.05

#define SEECLOCKDEPENDENCE pk_false
#include <iterator>

_PROTECTED_ const ptDCG2dot pK_dotPrinter = new DCG2dot ();

/**
   Specifies that the dependence (aDep) is inserted in the exported Signal tree model.
*/
    void
 DCG2dot::setVisited(ptdependence aDep)
{
  printedDependences.insert(aDep);
}

/**
   @return "true" if the dependence (aDep) has been already inserted in the Signal tree model, "false" otherwise.
*/
boolean DCG2dot::isVisited(ptdependence aDep)
{
  return (printedDependences.find(aDep) != printedDependences.end());
}


boolean DCG2dot::alreadyVisited(ptinsth h1, ptinsth h2)
{
  TDependenceClocks::iterator it;
  boolean bin = pk_false;

  it = printedClockDependences.begin();
  while(!bin && (it != printedClockDependences.end()))
    {
      if (it->first == h1) bin= pk_true; else it++;
    }
  if (!bin) return pk_false;
  return  ptlinsth_sorted_isin_elem(it->second, h2);
}

void DCG2dot::insertVisited(ptinsth h1, ptinsth h2)
{
  TDependenceClocks::iterator it;
  boolean b, bin = pk_false;
  ptlinsth lh;

  it = printedClockDependences.begin();
  while (!bin &&  (it != printedClockDependences.end()))
    {
      if (it->first == h1) bin= pk_true; else it++;
    }
  if (!bin) lh = NILLINSTH; else lh =it->second;
  printedClockDependences[h1] = ptlinsth_sorted_insert_elem_bool (lh, h2, &b);
}


/**
   @return "true" if a node (vto) is selected for this operation, "false" otherwise.
*/
boolean DCG2dot::isSelectedNode (ptINode vto)
{
  // if (NM->is_writeNode (vto)) return pk_false;
  if (ptlINode_isIn (lnunused, vto)) return pk_false;
  // return ptlINode_isIn (selectedNodes, vto);
  // Noeuds du meme niveau. 
  return pk_true;
}

/**
   @return "true" if a clock (h) is selected for this operation, "false" otherwise.
*/
boolean DCG2dot::isSelectedClock (ptinsth h)
{
  if (ptlinsth_isin_elem (lhunused, h)) return pk_false;
  return !CKM->satisfiesIsAnEquivalentForm(h);
}


/**
   It resets the internal used marks. The graph is cleaned.
*/
void
DCG2dot::resetMark ()
{
  NM->resetMarkAll (dotAttribute);
}

/**
   Debugging.
*/
void
DCG2dot::TRACE (const char *mess)
{
  printString ("/* ");
  printString (mess);
  printString (" */");
  endLine ();
}

/**
   It opens for writing (using fs file descriptor) the file gname_foo.dot
   where foo is obtained by the following code
   - code=1, foo="_ckm"
   - code=2, foo="_Graph"
   - code=2, foo="_gfd"
*/
void
DCG2dot::openFile (ptident gname, const char* innerSuffix)
{
  ptrepstr strname, name;
  if (makeRepstr_repstr_string (ident2repstr(gname), innerSuffix, &name))
    if (makeRepstr_repstr_string (name, ".dot", &name))
      {
	sys_filePathInUserOutputDir (name, &strname);
	fs.open (repstr2string (strname), ios::out);
	pK_messenger->messagestr (infoDotFile, strname);
      }
}

/**
   It closes fs (the file descriptor) previously opened by DCG2dot::openFile().
*/
void
DCG2dot::closeFile ()
{
  fs.close ();
}

ptlinsth DCG2dot::getlhroots()
{
  if (graph_type == DCG_BASIC) return CKM->getlhroots ();
  ptinsth vhtop = AGRAPH->getRootClock();
  ptlinsth lh = CKM->getChildren(vhtop);
  return lh;
}

ptlinsth DCG2dot::buildClockToSee(ptlinsth lh)
{
  ptlinsth lhaux, lhret;
  boolean b, jecontinue;
  ptinsth vh;

  lhret = ptlinsth_duplicate(lh);
  jecontinue = pk_true;
  while(jecontinue)
    {
      jecontinue = pk_false;
      lhaux = lhret;
      while (ptlinsth_get_next(&lhaux, &vh))
	{
	  if (ptgfd_satisfies_empty (CKM->getClockSubGraph (vh)))
	    if (CKM->getNodeReferencesTo(vh) == NILLINODE)
	      if ( CKM->getClockReferencesTo(vh) == NILLINSTH)
		{
		  jecontinue = pk_true;
		  lhret = ptlinsth_remove_elem (lhret, vh, &b);
		}
	}
    }
  return lhret;
}



/**
   It initializes the printer for a graph (g).
   Internal references to graph managers are initialized.
*/
void
DCG2dot::initialize (ptDCG g)
{
  AGRAPH = g;
  CKM = AGRAPH->getClockManager ();
  EM = AGRAPH->getExpressionManager ();
  NM = AGRAPH->getNodeManager ();
  graph_type = AGRAPH->getFormatLevel ();
  lnunused = AGRAPH->getUnused (&lhunused);
  lhroots = getlhroots ();
  lhplaced = buildClockToSee(CKM->getlhplaced ());
  inputs = AGRAPH->getInputs();
  outputs = AGRAPH->getOutputs();
  selectedNodes = NILLINODE;
}

/**
   It prints an int (v).
 */
void
DCG2dot::printInteger (int v)
{
  fs << v;
}

/**
    It prints a float (v).
*/
void
DCG2dot::printFloat (float v)
{
  fs << v;
}

/**
   It prints a key word (kw) followed by '='..
*/
void
DCG2dot::printKwEq (const char *kw)
{
  printKw (kw);
  printChar ('=');
}

/**
   It prints an end of line.
*/
void
DCG2dot::endLine ()
{
  fs << endl;
}

/**
   It prints the standard parameters for "dotty" for nodes : shape, font size, width, heigh
   (font size, width, heigh are given by MYFONTSIZE, MYWIDTH, MYHEIGHT constants)
   \see printShape(), printKwNode(), printFontSize(), printWidthHeight(), endNodeSpec().

   Example: node[shape=circle,fontsize=20,width=0.05,height=0.05];
*/
void
DCG2dot::printStdParameters ()
{
  printKwNode ();
  printShape ("circle");
  printChar (',');
  printFontSize (MYFONTSIZE);
  printChar (',');
  printWidthHeight (MYWIDTH, MYHEIGHT);
  endNodeSpec ();
}

/**
   It prints "];".
*/
void
DCG2dot::endNodeSpec ()
{
  printChar (']');
  SemicolonEndLine ();
}

/**
   It prints  node[shape=box,style=filled,color=lightblue];
*/
void
DCG2dot::printStdNodeParameters ()
{
  printKwNode ();
  printShape ("box");
  printChar (',');
  printStyle ("filled");
  printChar (',');
  printColor ("lightblue");
  endNodeSpec ();
}

/**
   It prints  node[shape=box,style=filled,color=green];
*/
void
DCG2dot::printStdClockParameters ()
{
  printKwNode ();
  printShape ("box");
  printChar (',');
  printStyle ("filled");
  printChar (',');
  printColor ("green");
  endNodeSpec ();
}

/**
   It prints the begining of the dot code\n
   digraph equationSolving {
   rankdir=TB;
   ratio=fill;
   nodesep=0.15;
   ranksep=0.4;
   center=t; \n
   followed by the generated code by printStdParameters() method.
*/
void
DCG2dot::printGraphBegin (ptident n)
{
  printKw ("digraph ");
  printKw (ident2string (n));
  printChar ('{');
  endLine ();
  printRankDir ("TB"); // LR
  printRatio ("fill");
  printNodeSep (.15);
  printRankSep (.40);
  printCenter ();
  printStdParameters ();
}

/**
   It prints the end specification of a dot graph.
*/
_VIRTUAL_ void
DCG2dot::printGraphEnd ()
{
  endLine ();
  printChar ('}');
  endLine ();
}

/**
    It prints an identifier (id).
*/
void
DCG2dot::printIdent (ptident id)
{
  fs << ident2string (id);
}

/**
   It prints a character (c).
*/
void
DCG2dot::printChar (char c)
{
  fs << c;
}


/**
   It prints a C-string (s).
*/
void
DCG2dot::printString (const char *s)
{
  fs << s;
}

/**
   It prints a C-string (c) as a key-word.
*/
void
DCG2dot::printKw (const char *kw)
{
  fs << kw;
}


/**
   It prints a semicolon followed by a eol.
*/
void
DCG2dot::SemicolonEndLine ()
{
  printChar (';');
  endLine ();
}

/**
   It prints the name of the identifier referenced (assumed) in an expression (e).
*/
void
DCG2dot::printSignalName (ptexpcal e)
{
  printIdent (EM->getName (e));
}


/**
   It prints the dot parameters for the root of the hierarchy (node[color=red];).
*/
void
DCG2dot::printRootHeader ()
{
  printKwNode ();
  printColor ("red");
  endNodeSpec ();
}

/**
   It prints the dot parameters for the non-root of the hierarchy (node[color=green];).
*/
void
DCG2dot::printPlacedHeader ()
{
  printKwNode ();
  printColor ("green");
  endNodeSpec ();
}

/**
   It prints the dot parameters for the representation of the name of a clock (h).
   It is the key-word label folowed by the name of the clock.
*/
void
DCG2dot::printClkIdent (ptinsth h)
{
  ptexpcal vs;
  printKwEq ("label");
  printChar ('"');
  if (CKM->satisfiesUndefClock (h))
    printKw ("UndefClock");
  else if (CKM->satisfiesNullClock (h))
    printKw ("^0");
  else if (isEventClockedFormat (graph_type))
    {
      if (CKM->isClockRepresentedBySignal(h, &vs)) printChar('^');
      printKw (ident2string (CKM->getName (h)));
    }
  else
    {
      if (CKM->satisfiesSub(h))
	{
	  printChar('[');
	  if (CKM->getOperator(h) == nosub)
	    {
	      printChar('!'); 
	      printChar(' ');
	    }
	  ptexpcal c = CKM->getConditionSub(h);
	  printSignalName(c);
	  printChar(']');
	}
      else
	printSignalName (CKM->getBooleanSignal (h));
    }
  printChar ('"');
}

/**
   It prints the dot declaration of a clock (h).\n
   Example c14[label="OK"]; \n
   where OK is the name of the clock.
   It calls printNumClockId(), printClkIdent().
*/
void
DCG2dot::printClockDeclaration (ptinsth h)
{
  printNumClockId (h);
  printChar ('[');
  printClkIdent (h);
  printChar (']');
  SemicolonEndLine ();
}


/**
   It prints 'c' followed by the number associated with the operand clock (h).
   It is the name used for the reference to the clock.
*/
void
DCG2dot::printNumClockId (ptinsth h)
{
  printChar ('c');
  printInteger (CKM->getNum (h));
}

/**
   It prints the information hierarchy about a clock (h): the link from the father clock of h to h is specified.\n
   Example: c14 -> c26, where c26 is the name of h, c14 is the name of it father clock.
*/
void
DCG2dot::printHierachy (ptinsth h)
{
  ptinsth hf;

  if (isSelectedClock(h)) {
    h = CKM->getRepresentativeFormula (h);
    hf = CKM->getFather (h);
    hf = CKM->getRepresentativeFormula (hf);
    printNumClockId (hf);
    printKw ("->");
    printNumClockId (h);
    // [color=red]; 
    printChar ('[');
    printColor("red");
    printChar (']');
    SemicolonEndLine ();
  }
}

/**
   It prints the clocks of a list (lh) as roots of the hierarchy.\n
   The dot parameters are generated before each declaration of roots (root_i).\n
   Example: {rank=min; root_1;...root_n; };
*/
void
DCG2dot::printRoots (ptlinsth lh)
{
  ptinsth h;

  if (lh != NILLINSTH)
    {
      printChar ('{');
      printKwEq ("rank");
      printKw ("min;");
      while (ptlinsth_get_next (&lh, &h))
	if (h == CKM->getRepresentativeFormula (h))
	  {
	    printNumClockId (h);
	    printChar (';');
	  }
      printChar ('}');
      SemicolonEndLine ();
    }
}


/**
   It prints the clocks of a list (lh) as non-roots of the hierarchy: it calls printHierachy() for each element.
*/
void
DCG2dot::printPlaced (ptlinsth lh)
{
  ptinsth h;
  
  while (ptlinsth_get_next (&lh, &h))
    printHierachy (CKM->getRepresentativeFormula (h));
}

/**
   It prints the constraint (vh1=vh2)  by a bi-directionnel link from the 
   references of the implicated nodes.\n
   Example: n1 -> n2 [ color=red, style=bold, dir=both;
*/
void
DCG2dot::printConstraint (ptinsth vh1, ptinsth vh2)
{
  if (vh1 != vh2)
    {
      printNumClockId (vh1);
      printKw ("->");
      printNumClockId (vh2);
      printChar ('[');
      printColor ("red");
      printChar (',');
      printStyle ("bold");
      printChar (',');
      printKwEq ("dir");
      printKw ("both");
      printChar (']');
      SemicolonEndLine ();
    }
}


/**
   It specifies the constraints of placed formula of the system.
   For each constraint, it calls printConstraint().
*/
void
DCG2dot::printConstraints ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  CKM->getConstraintsOnPlacedClocks (&lh1, &lh2);
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vh2 = CKM->getRepresentativeFormula (vh2);
      printConstraint (vh1, vh2);
    }
}

/**
   It prints the rankdir (v) dot parameter. \n
   Example: rankdir= LR; where LR stands for Left to Right.
*/
void
DCG2dot::printRankDir (const char *v)
{
  printKwEq ("rankdir");
  printKw (v);
  SemicolonEndLine ();
}


/**
    It prints the ratio (v) dot parameter. \n
    Example rankdir= fill;
*/
void
DCG2dot::printRatio (const char *v)
{
  printKwEq ("ratio");
  printKw (v);
  SemicolonEndLine ();
}

/**
   It prints the node separator (v) dot parameter. \n
   Example: nodesep=0.15;
*/
void
DCG2dot::printNodeSep (float v)
{
  printKwEq ("nodesep");
  printFloat (v);
  SemicolonEndLine ();
}

/**
   It prints the rank separator (v) dot parameter. \n
   Example: ranksep=0.40;
*/
void
DCG2dot::printRankSep (float v)
{
  printKwEq ("ranksep");
  printFloat (v);
  SemicolonEndLine ();
}

/**
   It prints the center dot parameter (ie center=t;).
*/
void
DCG2dot::printCenter ()
{
  printKwEq ("center");
  printChar ('t');
  SemicolonEndLine ();
}

/**
   It prints the node dot key word (ie node[  ).
*/
void
DCG2dot::printKwNode ()
{
  printKw ("node[");
}

void DCG2dot::printNumNode(ptINode n)
{
  if (isDelayCluster(n)) 
    {
      printChar ('"');
      printChar('n');
      printChar('$');
      printChar ('"');
    }
  else {
      printChar ('n');
      printInteger (NM->get_num (n));
  }
}

/**
   It print the dot name of a node (n). It is composed of the
   char 'n' followed by the number of the node.
*/
void
DCG2dot::printNumNodeId (ptINode n)
{
  if ( NM->is_readNode(n))
    {
      printIO(NM->get_io_arg(n), pk_true);
    }
  else
    if ( NM->is_writeNode(n))
      {
	printIO(NM->get_io_arg(n), pk_false);
      }
    else
    {
      printNumNode(n);
    }
}


/**
   It prints the name of a node (n) as a dot label.
   This name can be 
   - the label for a sub-graph (rintClusterName()),
   - the name of an output of the node, if it exists such an element,
   - the name of external call (if n is an external node) (see printProcCall()),
   - unknown node name otherwise (see printUnknown()).

Example:  [label="pre_next_yy"];
*/
void
DCG2dot::printLabelNode (ptINode n)
{
  boolean first;
  ptlinstances lo;
  ptexpcal e;

  printChar ('[');
  printKwEq ("label");
  printChar ('"');
  if (NM->is_subgraph (n))
    printClusterName (NM->get_num (n));
  else
    {
      lo = NM->getInodelOutputFlows (n);
      if (lo != NILLINSTANCES)
	{
	  first = pk_true;
	  while (ptlinstances_get_next (&lo, &e))
	    {
	      if (first)
		first = pk_false;
	      else
		printChar (',');
	      printSignalName (e);
	    }
	}
      else
	{
	  if (!NM->is_writeNode (n))
	    { 
	      if (NM->isProcessCall (n))
		printProcCall (n);
	      else
		printUnknown (n);
	    }
	}
    }
  printChar ('"');
  printChar (']');
  printChar (';');
}

/**
   It prints informations about a node composed of its dot name (printNumNodeId()) followed by its label ( printLabelNode()).

   n122[label="pre_next_yy"];
*/
void
DCG2dot::printInfoNode (ptINode n)
{
  printNumNodeId (n);
  printLabelNode (n);
}

/**
   It prints the "unknown" identifier for a node (n).
*/
void
DCG2dot::printUnknown (ptINode n)
{
  printNumNodeId (n);
  printKw ("Unknown");
}

/**
   It prints the name of the i-th cluster.
*/
void
DCG2dot::printClusterName (int i)
{
  printKw ("cluster");
  printInteger (i);
}

/**
   It prints the declaration of the i-th sub-graph.\n
   Example: subgraph cluster292{
   
*/
void
DCG2dot::printSubGraphDeclarationHeader (int i)
{
  printKw ("subgraph ");
  printClusterName (i);
  printChar ('{');
  endLine ();
}

/**
   It prints the declaration of the dot parameters for a sub-graph.
   It is composed of:
   - the standard parameters of a graph (printStdNodeParameters())
   followed by specific sub-graph parameters: style=filled; color=red;
*/
void
DCG2dot::printSubGraphDeclarationParameters ()
{
  printStdNodeParameters ();
  printStyle ("filled");
  SemicolonEndLine ();
  // LBesnard, Feb 2018: printColor ("red"); SemicolonEndLine ();
}

/**
   It prints the font size (v) parameter (fontsize=v).
*/
void
DCG2dot::printFontSize (int v)
{
  printKwEq ("fontsize");
  printInteger (v);
}

/**
   It prints the width (w) and  height (h) parameter (width=v, height=h).
*/
void
DCG2dot::printWidthHeight (float w, float h)
{
  printKwEq ("width");
  printFloat (w);
  printChar (',');
  printKwEq ("height");
  printFloat (h);
}

/**
   It prints the style parameter (st).
*/
void
DCG2dot::printStyle (const char *st)
{
  printKwEq ("style");
  printKw (st);
}

/**
   It prints the color parameter (st).
*/
void
DCG2dot::printColor (const char *st)
{
  printKwEq ("color");
  printKw (st);
}

/**
   It prints the shape parameter (sh).
*/
void
DCG2dot::printShape (const char *sh)
{
  printKwEq ("shape");
  printKw (sh);
}


/**
   It prints the internal node associated with a sub-graph (sg).
   It uses specific parameters (shape,style,color) for such an node.
   The node is marked by an internal mark (dotAttribute).
  
   Example: n293[shape=triangle,style=filled,color=purple];
*/
void
DCG2dot::printSubGraphInternalNode (ptINode sg)
{
  ptINode n;

  // Noeud interne
  if (NM->get_subGraphInputNode (sg, &n))
    {
      printNumNodeId (n);
      printChar ('[');
      printShape ("triangle");
      printChar (',');
      printStyle ("filled");
      printChar (',');
      printColor ("purple");
      printChar (']');
      SemicolonEndLine ();
      NM->addMark (n, dotAttribute);
      //      printDependenceNode(n);
    }
}

/**
   It prints some parameters for the reference to a sub-graph using its label in the dot representation.
   
   Example: n292[shape=circle,color=green];
*/
void
DCG2dot::printSubGraphLabelReferenceParameters (ptINode sg)
{
  ptINode n;
  ptexpcal lab;

  if (NM->get_label (sg, &lab))
    {
      n = EM->getNode (lab);
      printNumNodeId (n);
      printChar ('[');
      printShape ("circle");
      printChar (',');
      printStyle ("filled");
      printChar (',');
      printColor ("green");
      printChar (']');
      SemicolonEndLine ();
    }
}

/**
   It prints the end of the declaration of a dot sub-graph.
*/
void
DCG2dot::printSubGraphDeclarationTail ()
{
  printChar ('}');
  endLine ();
}


/**
   It prints the label of the i-th cluster.\n
   Generated code: label="cluster<i>";
*/
void
DCG2dot::printSubGraphLabel (int i)
{
  printKwEq ("label");
  printChar ('"');
  printClusterName (i);
  printChar ('"');
  printChar (';');
  endLine ();
}

/**
   It prints a dot representation of a sub-graph (sg). 
   - printSubGraphDeclarationHeader()
   - printSubGraphDeclarationParameters()
   - printSubGraphLabel()
   - the nodes of the sub-graph are printed (printNodes())
   - printSubGraphDeclarationTail() finalizes the sub-graph definition.

   Example:
   For sub-graph number 292, the dot code will be of the following form. 
   
   \code
   subgraph cluster292 {
   node[shape=box,style=filled,color=lightblue];
   style=filled;
   color=red;
   label="cluster292";
   n292[shape=circle,color=green];
   n293[shape=triangle,style=filled,color=purple];
   ...
   followed by the nodes of the sub-graph.
   }
   \endcode
   \note the node n292 is declared outside of the sub-graph: here, only some parameters are specified.
   
*/
void
DCG2dot::printSubGraph (ptINode sg)
{
  ptlINode ln;
  ptlinsth lh;
  int i;

  i = NM->get_num (sg);
  // subgraph cluster292 {
  printSubGraphDeclarationHeader (i);
  // subgraph cluster292 { node[shape=box,style=filled,color=lightblue]; style=filled; color=red;
  printSubGraphDeclarationParameters ();
  //label="cluster292";
  printSubGraphLabel (i);
  // n292[shape=circle,color=green];
  printSubGraphLabelReferenceParameters (sg);
  // n293[shape=triangle,style=filled,color=purple];
  printSubGraphInternalNode (sg);

  if (NM->get_subGraphDefinitions (sg, &ln, &lh)) printNodes (ln);
  printSubGraphDeclarationTail ();
}

/**
   It prints (printSubGraph()) each sub-graph of the current graph. The 
   sub-graph are marked (dotAttribute mark).
*/
void
DCG2dot::printSubGraphs ()
{
  ptlINode ln;
  ptINode n;

  if (NM->getSubGraphs (&ln))
    while (ptlINode_get_elem (&ln, &n))
      {
	printSubGraph (n);
	NM->addMark (n, dotAttribute);
      }
}

/**
   It prints (printInfoNode()) a node (n) if it is not already marked (dotAttribute mark).
*/
void
DCG2dot::printNode (ptINode n)
{
  if (!NM->isMarked (n, dotAttribute))
    {
      NM->addMark (n, dotAttribute);
      if (! NM->is_writeNode (n))
	{
	  printInfoNode (n);
	  endLine ();
	}
    }
}

/**
   It prints a list of nodes (ln) by applying printNode() method.
*/
void
DCG2dot::printNodes (ptlINode ln)
{
  ptINode n;
  while (ptlINode_get_elem (&ln, &n))
    printNode (n);
}

/**
   It prints a list of nodes (ln).
*/
void
DCG2dot::printPartialDefinitions ()
{
  ptlINode ln;
  if (NM->getGroupsOfPartialEquations (&ln))
    printNodes (ln);
}

/**
   It prints a list of nodes (ln). It calls printNode() for each node.
*/
void
DCG2dot::printProcCall (ptINode n)
{
  ptexpcal l1, l2, par;
  ptident id;

  NM->get_procCall_arg (n, &id, &par, &l1, &l2);
  printIdent (id);
  printChar ('(');
  printChar (')');
}

/**
   It prints (printNodes()) all the unmarked nodes of the graph.
   An internal mark (dotAttribute) is used.
*/
void
DCG2dot::printNotMarked ()
{
  ptlINode ln;
  if (NM->getAllNodes (&ln))
    printNodes (ln);
  endLine ();
}


/**
   It prints the node of the selected graph. It sequentialy calls : 
   printStdNodeParameters(), printSubGraphs(), printPartialDefinitions(), printNotMarked(), resetMark().
*/
void
DCG2dot::printGraphNodes ()
{
  TRACE ("NodeDeclaration");
  printStdNodeParameters ();
  printSubGraphs ();
  printPartialDefinitions ();
  printNotMarked ();
  resetMark ();
}


/**
   It prints, when the internal parameters SEECLOCKDEPENDENCE is set, the clock depClock
   used as the clock of a dependence. Without effect when depClock is the undefined clock.
*/
void
DCG2dot::printDependenceLabel (ptinsth depClock)
{
  if (SEECLOCKDEPENDENCE)
    if (!CKM->satisfiesUndefClock (depClock))
      {
	printChar ('[');
	printClkIdent (depClock);
	printChar (']');
      }
}

boolean DCG2dot::isDelayCluster( ptINode n)
{
  ptlINode ln;
  ptlinsth lh;
  ptINode node;

  if (!NM->is_trainCluster(n)) return pk_false;
  if ( NM->is_delayNodeOrDelayClusterNode(n)) return pk_true;
  return pk_false;

  if (! NM->get_subGraphDefinitions(n, &ln, &lh)) return pk_false;
  while (ptlINode_get_elem (&ln, &node))
    {
      if (! NM->is_delay(node)) return pk_false;
    }
  return pk_true;

}


void DCG2dot::printColorDependence(ptdependence aDep)
{
  // "zero" -> "dix" [color=purple];
  // reinforcementDep  explicitDep  dataFlowDep
  printChar ('[');
  if (aDep-> satisfiesExplicitDependence())
    printColor ("gold");
  else
    if (aDep->  satisfiesReinforcementDependence())
      printColor ("red");
    else
      printColor ("black");
  printChar (']');
}

boolean
DCG2dot::isSelectedForCurrentAlgo(ptINode n)
{
  if (GraphOfClusters) return ptlINode_isIn(selectedNodes, n);
  return pk_true;
}

/**
   It prints a dependence (aDep).
*/
void
DCG2dot::printDependence(ptdependence aDep)
{
  ptinsth h1;
  ptINode vfrom, vto;
  boolean bSwitchDep;

  if ( ! isVisited(aDep))
    {
      setVisited(aDep);
      h1 = aDep->getDependence (&vfrom, &vto);
      if (isSelectedNode(vto) && isSelectedForCurrentAlgo(vto))
	{
	  bSwitchDep = isDelayCluster(vfrom);
	  if (bSwitchDep)
	    {
	      printNumNodeId (vto);
	      printKw ("->");
	      printNumNodeId (vfrom);
	    }
	  else
	    {
	      printNumNodeId (vfrom);
	      printKw ("->");
	      printNumNodeId (vto);
	    }
	  printDependenceLabel (h1);
	  printColorDependence(aDep); // "zero" -> "dix" [color=purple];
	  printChar (';');
	}
    }
}

/**
   It prints a dependence from a node (n) to a clock (h) when the clock is selected.
*/
void
DCG2dot::printNode2ClockDependence (ptINode n, ptinsth h)
{
  h = CKM->getRepresentativeFormula (h);
  if (isSelectedClock (h))
    {
      printNumNodeId (n);
      printKw ("->");
      printNumClockId (h);
      printChar (';');
    }
}


void
DCG2dot::printDependenceNodeToNode (ptINode n)
{
  ptldependence lsucc;
  ptdependence aDep;

  lsucc = NM->get_successors (n);
  while (lsucc->get_next (&aDep, &lsucc))
    printDependence (aDep);
  endLine ();
}


void
DCG2dot::printDependenceNodeToClock (ptINode n)
{
  ptinsth h;
  ptlinsth lh;

  lh = NM->get_clockSuccessors (n);
  while (ptlinsth_get_next (&lh, &h))
    printNode2ClockDependence (n, h);
  endLine ();
}

/**
   It prints the dependences associated with a node (n):
   - it calls printDependence() for each successors,
   - printNode2ClockDependence() for each clock depending on the node.
*/
void
DCG2dot::printDependenceNode (ptINode n)
{
  printDependenceNodeToNode(n);
  printDependenceNodeToClock(n);
}

/**
   It prints the dependences of the graph: for each node of the graph, 
   it calls printDependenceNode() method.
*/
void
DCG2dot::printINodeDependences ()
{
  ptlINode ln;
  ptINode n;

  if (NM->getAllNodes (&ln))
    while (ptlINode_get_elem (&ln, &n))
      if (!NM->isMarked (n, dotAttribute))
	{
	  NM->addMark (n, dotAttribute);
	  printDependenceNode (n);
	}
}



/**
   It prints the dependences from a clock (h) to another one (h1) using the identifiers. The
   dependence is generated if h1 is a selected clock.
   \see isSelectedClock(), printNumClockId().
*/
void
DCG2dot::printClock2ClockDependence (ptinsth h, ptinsth h1)
{
  ptlinsth lh;

  h1 = CKM->getRepresentativeFormula (h1);
  if (isSelectedClock (h1))
    {
      if (! alreadyVisited(h, h1))
	{
	  insertVisited(h,h1);
	  if (CKM->getOperator (h1) == hnot)
	    {
	      lh = CKM->getClockReferencesTo (h1);
	      while (ptlinsth_get_next (&lh, &h1))
		printClock2ClockDependence (h, h1);
	    }
	  else
	    {
	      printNumClockId (h);
	      printKw ("->");
	      printNumClockId (h1);
	      printChar (';');
	    }
	}
    }
}

/**
   It prints the dependence from a clock (h) to a node (n), if h1 is a selected clock using the identifiers.
   \see isSelectedClock(), printNumClockId().
*/
void
DCG2dot::printClock2NodeDependence (ptinsth h, ptINode n)
{
  h = CKM->getRepresentativeFormula (h);
  if (isSelectedNode (n))
    {
      printNumClockId (h);
      printKw ("->");
      printNumNodeId (n);
      printChar (';');
    }
}

/**
   It prints the dependences associated with a clock (h). The successors of h are considered.
   It calls printClock2ClockDependence() for each clock sucessor of h, and printClock2NodeDependence() 
   for each node successor of h.
*/
void
DCG2dot::printClockDependence (ptinsth h)
{
  ptlinsth lh;
  ptinsth hto;
  ptlINode ln;
  ptINode n;

  lh = CKM->getClockReferencesTo (h);
  while (ptlinsth_get_next (&lh, &hto))
    printClock2ClockDependence (h, hto);
  ln = CKM->getNodeReferencesTo (h);
  while (ptlINode_get_elem (&ln, &n))
    printClock2NodeDependence (h, n);
  endLine ();
}

/**
   It prints the dependences associated with each clocks of a list (lh). printClockDependence() is called
   for each element.
*/
void
DCG2dot::printClockDependencesList (ptlinsth lh)
{
  ptinsth h;
  while (ptlinsth_get_next (&lh, &h))
    if (h == CKM->getRepresentativeFormula (h))
      printClockDependence (h);
}


/**
   It prints the dependences associated with the clocks of the system: it calls
   printClockDependencesList() for the root clocks and for the placed clocks.
*/
void
DCG2dot::printClockDependences ()
{
  printClockDependencesList (lhroots);
  printClockDependencesList (lhplaced);
}

/**
   It prints the dependences of the graph: printINodeDependences() for the nodes, printClockDependences() for the clocks.
*/
void
DCG2dot::printDependences ()
{
  TRACE ("printDependences");
  printINodeDependences ();
  printClockDependences ();
  resetMark ();
}

/** It prints the clock as the label of a subgraph */
void
DCG2dot::printLabelPartitionClock (ptinsth h)
{
  printClkIdent (h);
  SemicolonEndLine ();
}

void DCG2dot::printClock2sugGraph(ptinsth h, ptlINode ln)
{
  ptINode n;

  // example : c53 -> n52 [lhead=cluster53];
 printNumClockId (h);
 printKw ("->");
 ptlINode_get_elem (&ln, &n);
 printNumNodeId(n);
 printChar ('[');
 printKwEq ("lhead ");
 printClusterName(CKM->getNum (h));
 printChar (']');
 printChar (';');
 endLine ();

}


/**
   It prints the sub-graph associated with a clock (h) a "dot" sub-graph. It calls the
   following method:
   printSubGraphDeclarationHeader(), printSubGraphDeclarationParameters(), printLabelPartitionClock(),
   printNodes(),printSubGraphDeclarationTail()
*/
void
DCG2dot::printGraphAssignedToClock (ptinsth h)
{
  ptgfd vgfd;
  ptlINode ln;

  vgfd = CKM->getClockSubGraph (h);
  if ( !ptgfd_satisfies_empty (vgfd)) {
    printSubGraphDeclarationHeader (CKM->getNum (h));
    printSubGraphDeclarationParameters ();
    printLabelPartitionClock (h);
    ln = ptgfd_get_nodes (vgfd);
    printNodes (ln);
    printSubGraphDeclarationTail ();
    printClock2sugGraph(h,ln);
  }

}


/**
   It prints the sub-graphs assigned to each clock of a list (lh). 
   It calls printGraphAssignedToClock() for each element of the list.
*/
void
DCG2dot::printGraphAssignedToClocks (ptlinsth lh)
{
  ptinsth h;

  TRACE ("printGraphClock");
  while (ptlinsth_get_next (&lh, &h))
    if (h == CKM->getRepresentativeFormula (h))
      printGraphAssignedToClock (h);
}

/**
   It prints the informations about the hierarchy of the clock system: it calls printRoots()
   for the root clocks, printPlaced() for the placed one.
*/
void
DCG2dot::ClockHierachyInfo ()
{
  printRoots (lhroots);
  printPlaced (lhplaced);
}


/**
   It prints the declarations of a list of clocks (lh): it calls printClockDeclaration() for each element.
*/
void
DCG2dot::clocksDeclarationList (ptlinsth lh)
{
  ptinsth h;
  while (ptlinsth_get_next (&lh, &h))
    {
      h =CKM->getRepresentativeFormula (h);
      if (isSelectedClock(h)) 
	printClockDeclaration (h);
    }
  endLine ();
}

/**
   It prints the declaration of the clocks of the system:
   - printStdClockParameters(), 
   - printRootHeader(),
   - clocksDeclarationList() for the root clocks
   - printPlacedHeader();
   - clocksDeclarationList() for the placed clocks.
*/
void
DCG2dot::printGraphClocks ()
{
  TRACE ("clocksDeclaration");
  printStdClockParameters ();
  printRootHeader ();
  clocksDeclarationList (lhroots);
  printPlacedHeader ();
  clocksDeclarationList (lhplaced);
}

void DCG2dot::init (ptDCG g, const char* innerSuffix)
{
  ptident n = g->getLevelName();
  initialize (g);
  openFile (n, innerSuffix);
  dotAttribute = getMark ();
  printGraphBegin (n);
  GraphOfClusters= pk_false;
}


void DCG2dot::final()
{
  printedDependences.clear();
  resetMark ();
  printGraphEnd ();
  closeFile ();
}

/**
   @return "true" when the graph is a graph of clusters, false otherwise.
*/
boolean DCG2dot::isGraphOfClusters()
{
  ptlINode lsg;
  ptINode aNode;

  if (NM->getSubGraphs(&lsg))
    while (ptlINode_get_elem(&lsg, &aNode))
      if (NM->is_trainCluster(aNode))
	return pk_true;

  return pk_false;
}

ptlINode DCG2dot::printDeclareClusters()
{
  ptlINode lnodes, lsg;
  ptINode aNode;

  lnodes = NILLINODE;
  if (NM->getSubGraphs(&lsg))
    {
      while (ptlINode_get_elem(&lsg, &aNode))
	if (isSelectedNode(aNode))
	  if (NM->is_trainCluster(aNode))
	    {
	      ptlINode_insert_elem(&lnodes, aNode);
	      printSubGraphLabelReferenceParameters(aNode);
	    };
    }
  return lnodes;
}

void DCG2dot::printIO(ptexpcal vsig,  boolean bins)
{
  printChar('"');
  if (bins) printChar('?'); else printChar('!');
  printSignalName (vsig);
  printChar('"');
  // printNumNode( EM->getNode(vsig));
}

void DCG2dot::printDeclareIO(ptlinstances lsig, boolean bins)
{
  ptexpcal vsig;

  while (ptlinstances_get_next(&lsig, &vsig))
    {
      printIO(vsig, bins);
      printChar ('[');
      printStyle ("filled");
      printChar (',');
      printColor ("lightblue");
      endNodeSpec ();
    }
}

void DCG2dot::printClusterDependences(ptldependence lDep, boolean breinforced)
{
  ptdependence  aDep;
  while (lDep->get_next(&aDep, &lDep))
    printDependence(aDep);
}

void DCG2dot::printClusterDependences()
{
  ptldependence lDepExpl, lDepReinf;
  ptlINode ln;
  ptINode aNode;

  ln = selectedNodes;
  while (ptlINode_get_elem(&ln, &aNode))
    printDependenceNodeToNode(aNode);

  lDepExpl = AGRAPH->getExplicitDependences();
  printClusterDependences( lDepExpl, pk_false);
  lDepReinf = AGRAPH->getReinforcedDependences();
  printClusterDependences( lDepReinf, pk_true);
}

void DCG2dot::printInputsDependences(ptlinstances lsig)
{
  ptexpcal vsig;
  while (ptlinstances_get_next(&lsig, &vsig))
    printDependenceNodeToNode(EM->getNode(vsig));
}

void DCG2dot::printGraphOfClusters()
{
  ptlINode ln;

  selectedNodes = printDeclareClusters();
  if (NM->getInputs(&ln)) 
    selectedNodes = ptlINode_fusion(selectedNodes, ln);
  if (NM->getOutputs(&ln)) 
    selectedNodes = ptlINode_fusion(selectedNodes, ln);

  printDeclareIO(inputs, pk_true);
  printDeclareIO(outputs, pk_false);

  printClusterDependences();
  printInputsDependences(inputs);
}
