/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define eqClockPrefix_string "synchro_"
#define isEventClockedFormat(t) ((t == DCG_BASIC) ||(t == DCG_POLY) || (t == DCG_ENDO))
#define indentStr "     "
#define isCommutativeOperator(vop) ((vop == opmult) || (vop == opplus) || (vop == opegal) || (vop == opdiff))
  void
FiacreService::Error ()
{
  bErrCode = pk_true;
}

boolean FiacreService::isSelectedNode (ptINode n)
{
  ptinsth
    h;
  ptexpcal
    s;

  if (ptlINode_isIn (unNodes, n))
    return pk_false;
  if (NM->is_ioNode (n))
    return pk_false;
  s = NM->get_leftOperand (n);
  return !(EM->isSignalUnifiedWithClock (s, &h));
}

boolean FiacreService::isSelectedClock (ptinsth h)
{
  return !ptlinsth_isin_elem (unClocks, h);
}

void
FiacreService::initTable ()
{
  ptrepstr undefStr;
  int i;

  // For constant operand in Fiacre process definition.
  suffcste = string2repstr ("_C");
  suffcsteSecond = string2repstr ("_SC");
  suffcsteFirst = string2repstr ("_CS");
  suffcstAll = string2repstr ("_CC");
  identityId = string2repstr ("Signal_Id");

  clock_operators[0] = string2repstr ("clk_sub");
  clock_operators[1] = string2repstr ("clk_nosub");
  clock_operators[2] = string2repstr ("clk_upper");
  clock_operators[3] = string2repstr ("clk_lower");
  clock_operators[4] = string2repstr ("clk_minus");
  clock_operators[5] = string2repstr ("synchro");

  undefStr = string2repstr ("_Undefined_Operator_");
  // translated Operators
  for (i = 0; i <= SIG_NB_OPERATORS; i++)
    operatorTable[i] = undefStr;

  operatorTable[ophnull] = string2repstr ("nullClock");
  operatorTable[opdiff] = string2repstr ("Signal_diff");
  operatorTable[opegal] = string2repstr ("Signal_eq");
  operatorTable[opiegal] = string2repstr ("Signal_leq");
  operatorTable[opinfe] = string2repstr ("Signal_lt");
  operatorTable[opsegal] = string2repstr ("Signal_geq");
  operatorTable[opsupe] = string2repstr ("Signal_gt");

  operatorTable[opnon] = string2repstr ("Signal_not");
  operatorTable[opet] = string2repstr ("Signal_and");
  operatorTable[opou] = string2repstr ("Signal_or");
  operatorTable[opouex] = string2repstr ("Signal_xor");

  operatorTable[opmod] = string2repstr ("Signal_modulo");
  operatorTable[opnega] = string2repstr ("Signal_uminus");
  operatorTable[opplus] = string2repstr ("Signal_plus");
  operatorTable[opmoin] = string2repstr ("Signal_minus");
  operatorTable[opmult] = string2repstr ("Signal_mult");
  operatorTable[opdivi] = string2repstr ("Signal_divi");

  operatorTable[opextr] = string2repstr ("Signal_when");
  operatorTable[opmele] = string2repstr ("Signal_default");
  operatorTable[opleft_tt] = string2repstr ("Signal_lefttt");
  operatorTable[opret] = string2repstr ("Signal_delay");
  operatorTable[opmemo] = string2repstr ("Signal_cell");

}

boolean
FiacreService::initPrinter ()
{
  EM = vDCGFiacre->getExpressionManager ();
  CKM = vDCGFiacre->getClockManager ();
  NM = vDCGFiacre->getNodeManager ();
  Names = vDCGFiacre->getNameManager ();
  inputs = vDCGFiacre->getInputs ();
  outputs = vDCGFiacre->getOutputs ();
  unNodes = vDCGFiacre->getUnused (&unClocks);
  ProcessNameStr = ident2repstr (vDCGFiacre->getName ());
  shaper = new FiacreShaper ();
  shaper->initialize (ProcessNameStr);
  vpragmaFiacre = string2repstr ("Fiacre");
  idEndOfReaction = Names->mkIdentName (string2ident ("c_eor"));
  idNullClock = Names->mkIdentName (string2ident ("NullClock"));
  idPrefix = string2ident ("xtmp");
  preRefInstance = string2repstr (" * -> ");
  initTable ();
  FiacreAttr = getMark ();
  requireNullClock = pk_false;
  return pk_true;
}

void
FiacreService::finalizePrinter ()
{
  ptlinstances lexp;

  // Cleaning the marked objects.
  if (EM->getExprList (opextr, &lexp))
    EM->resetMark (lexp, FiacreAttr);
  CKM->resetMarkAll (FiacreAttr);
}

void
FiacreService::genDefProcesses ()
{
  int i, n, j, nlp;
  ntrees vPragmaText, ap, obj, lp;
  ptrepstr id;

  // Assumed : vtDefOperators is a module.
  lp = STree_getPragmas (vtDefOperators);
  nlp = ntreearity (lp);
  for (j = 1; j <= nlp; j++)
    {
      ap = nchild (j, lp);
      STree_getPragmaValues (ap, &id, &obj, &vPragmaText);
      // vPragmaText is a lis of chain (opchain : list of opcstcha)
      if (streq (id, vpragmaFiacre))
	{
	  n = ntreearity (vPragmaText);
	  for (i = 1; i <= n; i++)
	    {
	      shaper->printstr (getPragmaTerm (vPragmaText, i));
	      shaper->printLine ();
	    }
	  shaper->printLine ();
	}
    }
}

boolean
FiacreService::genSignalOperatorsInFiacre ()
{
  char *fileNameSrc, *fileNameSrcC;

  fileNameSrc = (char *) "FiacreLib.SIG";
  if (!existFile (fileNameSrc, &fileNameSrcC))
    {
      pK_messenger->globalMessageString (errFileNotFound, fileNameSrc);
      return pk_false;
    }

  if (!sysparser (without_phylum, fileNameSrcC, &vtDefOperators))
    return pk_false;
  vtDefOperators = nchild (1, vtDefOperators);	// list of modules, taking the pragmas of the first one.
  genDefProcesses ();
  return pk_true;
}

void
FiacreService::printClock (ptinsth h)
{
  shaper->printident (getName (h));
}


void
FiacreService::printSignal (ptexpcal s)
{
  shaper->printident (getName (s));
}

/** @return "true" if the operation succeeds, "false" otherwise.
    It prints a type (t) followed by a space character when space. 
    It calls printnotimplemented() when the type is unknown.
*/
boolean FiacreService::printtype (TypeType t)
{
  if (TypeTestPredef (t))
    {
      shaper->printtypePredef (t);
      return pk_true;
    }
  // Revoir pour les enum
  shaper->printkwnotimpl ();
  Error ();
  return pk_false;
}


boolean FiacreService::printtype (ptexpcal vexp)
{
  return printtype (EM->getType (vexp));
}

void
FiacreService::printEndOfReaction (boolean b)
{
  if (b)
    shaper->printComma ();
  shaper->printident (idEndOfReaction);
}

void
FiacreService::printOutEor (ptident id, boolean bComma)
{
  if (bComma)
    shaper->printComma ();
  shaper->printident (id);
  printEndOfReaction (pk_true);
}

void
FiacreService::printOutEor (ptexpcal exp, boolean bComma)
{
  printOutEor (getName (exp), bComma);
}

void
FiacreService::genBinOpSigCst (ptexpcal vout, ptrepstr vop, ptrepstr suff, ptexpcal vsig, ptexpcal vcste)
{
  shaper->printString (vop);
  shaper->printString (suff);
  shaper->printLeftSquareBracket ();
  printSignal (vsig);
  printOutEor (vout, pk_true);
  shaper->printRightSquareBracket ();
  shaper->printLP ();
  genConstant (vcste);
  shaper->printRP ();
}

void
FiacreService::printTypedProcess (ptrepstr vop, TypeType T)
{
  shaper->printString (vop);
  shaper->printunderscore ();
  printtype (T);
}

void
FiacreService::printTypedProcess (ptrepstr vop, TypeType T1, TypeType T2)
{
  shaper->printString (vop);
  shaper->printunderscore ();
  printtype (T1);
  shaper->printunderscore ();
  printtype (T2);
}

void
FiacreService::printTypedProcess (ptrepstr vop, ptexpcal vexp)
{
  printTypedProcess (vop, EM->getType (vexp));
  //  shaper->printString(vop);  shaper->printunderscore();  printtype(vout);
}


ptident
FiacreService::genWhenAttr (ptexpcal vexp)
{
  ptident id;
  ptinsth h;
  ptexpcal e;

  e = EM->getBinaryArgstAt (vexp, 1);
  h = EM->getContributionClock (vexp);
  if (EM->getAttribute (vexp, FiacreAttr, (attributeValue *) & id))
    {
      genWhen (e, h, id);
      shaper->printLine ();
    }
  else				/* pk_true when h */
    id = CKM->getName (h);
  return id;
}

/**
   assumed: vexp is of the form "(E1 when h1) default (E2 when h2)"
*/
void
FiacreService::genDefault (ptexpcal vout, ptexpcal vexp)
{
  ptident id1, id2;
  ptexpcal e1, e2;
  TypeType t1, t2;

  e1 = EM->getBinaryArgstAt (vexp, 1);
  e2 = EM->getBinaryArgstAt (vexp, 2);
  id1 = genWhenAttr (e1);
  id2 = genWhenAttr (e2);

  beginEq ();

  t1 = EM->getType (e1);
  t2 = EM->getType (e2);
  if (t1 == t2)
    if (t1 == PureType)
      shaper->printString (clock_operators[2]);
    else
      printTypedProcess (operatorTable[opmele], vout);
  else
    printTypedProcess (operatorTable[opmele], t1, t2);
  shaper->printLeftSquareBracket ();
  shaper->printident (id1);
  shaper->printComma ();
  shaper->printident (id2);
  printOutEor (vout, pk_true);
  shaper->printRightSquareBracket ();
}


void
FiacreService::genOper2 (ptexpcal vout, ptexpcal vexp)
{
  noperators vop;
  ptexpcal vexpcal1, vexpcal2;
  ptrepstr suff;

  beginEq ();
  vop = EM->getOperator (vexp);
  vexpcal1 = EM->getBinaryArgstAt (vexp, 1);
  vexpcal2 = EM->getBinaryArgstAt (vexp, 2);
  if (LL_is_listOp (vop))
    {
      shaper->printString (operatorTable[vop]);
      Error ();
    }
  else if (EM->IsConstantExpr (vexpcal1))
    if (EM->IsConstantExpr (vexpcal2))
      {
	shaper->printString (operatorTable[vop]);
	shaper->printString (suffcstAll);
	shaper->printLeftSquareBracket ();
	printOutEor (vout, pk_false);
	shaper->printRightSquareBracket ();
	shaper->printLP ();
	genConstant (vexpcal1);
	shaper->printComma ();
	genConstant (vexpcal2);
	shaper->printRP ();
      }
    else
      {
	if (isCommutativeOperator (vop))
	  suff = suffcste;
	else
	  suff = suffcsteFirst;
	genBinOpSigCst (vout, operatorTable[vop], suff, vexpcal2, vexpcal1);
      }
  else if (EM->IsConstantExpr (vexpcal2))
    {
      if (isCommutativeOperator (vop))
	suff = suffcste;
      else
	suff = suffcsteSecond;
      genBinOpSigCst (vout, operatorTable[vop], suff, vexpcal1, vexpcal2);
    }
  else
    {
      shaper->printString (operatorTable[vop]);
      shaper->printLeftSquareBracket ();
      printSignal (vexpcal1);
      shaper->printComma ();
      printSignal (vexpcal2);
      printOutEor (vout, pk_true);
      shaper->printRightSquareBracket ();

    }
}

void
FiacreService::genWhen (ptexpcal vexpcal, ptinsth h, ptident vout)
{
  boolean bconstant;

  bconstant = EM->IsConstantExpr (vexpcal);
  beginEq ();
  printTypedProcess (operatorTable[opextr], vexpcal);
  if (bconstant)
    shaper->printString (suffcste);
  shaper->printLeftSquareBracket ();
  if (!bconstant)
    {
      printSignal (vexpcal);
      shaper->printComma ();
    }
  printClock (h);
  printOutEor (vout, pk_true);
  shaper->printRightSquareBracket ();
  if (bconstant)
    {
      shaper->printLP ();
      genConstant (vexpcal);
      shaper->printRP ();
    }
}


void
FiacreService::genFiacreExpression (ptexpcal vout, ptexpcal vexp)
{
  noperators vop;
  ptexpcal vexpcal1, vexpcal2, vexpcal3;
  ptinsth h;
  ptident id1;

  vop = EM->getOperator (vexp);

  switch ((int) EM->getOperatorClass (vexp))
    {
    case variabl:		// x:=y
      beginEq ();
      shaper->printString (identityId);
      shaper->printLeftSquareBracket ();
      printSignal (vexp);
      printOutEor (vout, pk_true);
      shaper->printRightSquareBracket ();
      break;

    case constant:		// x:= 999
      beginEq ();
      shaper->printString (identityId);
      shaper->printString (suffcste);
      shaper->printLeftSquareBracket ();
      printOutEor (vout, pk_false);
      shaper->printRightSquareBracket ();
      shaper->printLP ();
      genConstant (vexp);
      shaper->printRP ();
      break;

    case oper1:		// opnega, opnot
      beginEq ();
      vexpcal1 = EM->getUnaryArg (vexp);
      if (EM->IsConstantExpr (vexpcal1))
	{
	  shaper->printString (operatorTable[vop]);
	  shaper->printString (suffcste);
	  shaper->printLeftSquareBracket ();
	  printOutEor (vout, pk_false);
	  shaper->printRightSquareBracket ();
	  shaper->printLP ();
	  genConstant (vexpcal1);
	  shaper->printRP ();
	}
      else
	{
	  shaper->printString (operatorTable[vop]);
	  shaper->printLeftSquareBracket ();
	  printSignal (vexpcal1);
	  printOutEor (vout, pk_true);
	  shaper->printRightSquareBracket ();
	}
      break;
    case oper1tp:		// Does not occur
      break;

    case oper2:
      genOper2 (vout, vexp);
      break;

    case oper2tp:		// default
      genDefault (vout, vexp);
      break;

    case oper3:
    case oper3tp:		// opret ($)
      vexpcal1 = EM->getTernaryArgstAt (vexp, 1);	// x init x0
      vexpcal2 = EM->getTernaryArgstAt (vexp, 2);	// vret
      vexpcal3 = EM->getTernaryArgstAt (vexp, 3);	// vfen assumed nil.
      beginEq ();
      printTypedProcess (operatorTable[vop], vout);
      if (vop == opret)
	{
	  shaper->printLeftSquareBracket ();
	  printSignal (EM->getTernaryArgstAt (vexpcal1, 1));
	  printOutEor (vout, pk_true);
	  shaper->printRightSquareBracket ();
	  shaper->printLP ();
	  genConstant (EM->getTernaryArgstAt (vexpcal1, 2));
	  shaper->printRP ();
	}
      else			// cell
	{
	  shaper->printLeftSquareBracket ();
	  printOutEor (vout, pk_false);
	  shaper->printRightSquareBracket ();
	  Error ();

	}
      break;

    case opereh:		// when, cell
      vexpcal1 = EM->getWhenOrCellArg (vexp);
      h = EM->getContributionClock (vexp);
      if (vop == opmemo)	// cell, assumed he operand is not a constant.
	{
	  if (EM->getAttribute (vexpcal1, FiacreAttr, (attributeValue *) & id1))
	    {
	      genWhen (EM->getWhenOrCellArg (vexpcal1), EM->getContributionClock (vexpcal1), id1);
	      shaper->printLine ();
	    }
	  else
	    id1 = EM->getName (vexpcal1);
	  beginEq ();
	  printTypedProcess (operatorTable[opmemo], vout);
	  shaper->printLeftSquareBracket ();
	  shaper->printident (id1);
	  shaper->printComma ();
	  printClock (h);
	  printOutEor (vout, pk_true);
	  shaper->printRightSquareBracket ();
	  shaper->printLP ();
	  genConstant (EM->getInitialValue (vout));
	  shaper->printRP ();
	}
      else
	genWhen (vexpcal1, h, EM->getName (vout));
      break;

    default:
      shaper->printString (operatorTable[vop]);
      shaper->printLeftSquareBracket ();
      printOutEor (vout, pk_false);
      shaper->printRightSquareBracket ();
    }

  shaper->printLine ();

}

/**
   Assumed: expr is the basic expression of upExpr (i.e. e when hUpExpr)
   @return expr if ^e = hUpExpr, upExpr otherwise.
*/
ptexpcal FiacreService::SpecificTranslation (ptexpcal expr, ptexpcal upExpr, ptinsth hUpExpr)
{
  ptinsth
    vh;

  if (EM->getSynchronousExprHorl (expr, &vh))
    {
      vh = CKM->getRepresentativeFormula (vh);
      if (vh == hUpExpr)
	return expr;
    }
  return upExpr;
}

void
FiacreService::genCounter (ptinsth h, ptexpcal vexpr)
{

}

void
FiacreService::genEq (ptINode node)
{
  ptexpcal basicExpr, vopg, vopd;
  ptinsth hd, hg;
  boolean done;
  noperators vop;

  if (!NM->get_rightOperand (node, &vopd))
    Error ();
  else
    {
      vopg = NM->get_leftOperand (node);
      if (EM->IsCellExpr (vopd))
	genFiacreExpression (vopg, vopd);
      else
	{
	  hg = CKM->getRepresentativeFormula (EM->getClock (vopg));
	  hd = CKM->getRepresentativeFormula (EM->getContributionClock (vopd));
	  if (CKM->satisfiesInitClock (hd))
	    genFiacreExpression (vopg, EM->getBasicExpr (vopd));
	  else
	    {
	      if (hg != hd)
		genFiacreExpression (vopg, vopd);
	      else
		{
		  done = pk_false;
		  basicExpr = EM->getBasicExpr (vopd);
		  vop = EM->getOperator (basicExpr);
		  if ((vop == opmele) || (vop == opleft_tt))
		    vopd = basicExpr;
		  else if (vop == opnombm)
		    {
		      done = pk_true;
		      genCounter (hg, EM->getBinaryArgstAt (basicExpr, 2));
		    }
		  else
		    vopd = SpecificTranslation (basicExpr, vopd, hg);
		  if (!done)
		    genFiacreExpression (vopg, vopd);
		}
	    }
	}
    }
}

void
FiacreService::genDelay (ptINode node)
{
  ptexpcal vopd;
  if (NM->get_rightOperand (node, &vopd))
    genFiacreExpression (NM->get_leftOperand (node), vopd);
  else
    Error ();
}

void
FiacreService::genNodes (ptlINode ln)
{
  ptINode node;

  while (ptlINode_get_elem (&ln, &node))
    if (isSelectedNode (node))
      {
	if (NM->is_equation (node))
	  genEq (node);
	else if (NM->is_delay (node))
	  genDelay (node);
	else
	  Error ();
      }
}



ptident
FiacreService::getName (ptexpcal e)
{
  if (EM->getOperatorClass (e) == variabl)
    return EM->getName (e);
  Error ();
  return NILIDENT;
}

ptident
FiacreService::getName (ptinsth vh)
{
  return CKM->getName (vh);
}

void
FiacreService::beginEq ()
{
  shaper->printString (indentStr);
  if (bFirstEquation)
    bFirstEquation = pk_false;
  else
    shaper->printComposition ();
  shaper->printString (preRefInstance);
}

void
FiacreService::endEq ()
{
  shaper->printLine ();
}

void
FiacreService::genClock_unary (ptrepstr op, ptident idop1, ptident id)
{
  beginEq ();
  shaper->printString (op);
  shaper->printLeftSquareBracket ();
  shaper->printident (idop1);
  printOutEor (id, pk_true);
  shaper->printRightSquareBracket ();
  endEq ();
}

void
FiacreService::genClock_binary (ptrepstr op, ptident idop1, ptident idop2, ptident id)
{
  beginEq ();
  shaper->printString (op);
  shaper->printLeftSquareBracket ();
  shaper->printident (idop1);
  shaper->printComma ();
  shaper->printident (idop2);
  printOutEor (id, pk_true);
  shaper->printRightSquareBracket ();
  shaper->printLine ();

}

void
FiacreService::genClockDef (ptinsth vh)
{
  ptinsth h1, h2;
  ptexpcal vcond;
  ptrepstr op;

  switch ((int) (CKM->getOperator (vh)))
    {
    case sub:
      op = clock_operators[0];
      vcond = CKM->getConditionSub (vh);
      genClock_unary (op, getName (vcond), getName (vh));
      break;

    case nosub:
      op = clock_operators[1];
      vcond = CKM->getConditionSub (vh);
      genClock_unary (op, getName (vcond), getName (vh));
      break;

    case hcomp:
      op = clock_operators[4];
      CKM->getUnaryArg (vh, &h1);
      h2 = CKM->getFather (vh);
      genClock_binary (op, getName (h1), getName (h2), getName (vh));
      break;

    case infer:
      if (CKM->getMinusArgs (vh, &h1, &h2))
	op = clock_operators[4];
      else if (CKM->getInfArgs (vh, &h1, &h2))
	op = clock_operators[3];
      else;
      genClock_binary (op, getName (h1), getName (h2), getName (vh));
      break;

    case super:
      CKM->getSupArgs (vh, &h1, &h2);
      op = clock_operators[2];
      genClock_binary (op, getName (h1), getName (h2), getName (vh));
      break;

    case predef:		// hnull, hinit, hundef
      break;

    case vari:
      break;

    case hnot:
      break;

    default:
      pK_messenger->Caseerror (__LINE__, __FILE__);
    }
}

void
FiacreService::genEquations (ptlinsth lh)
{
  ptinsth vh;
  while (ptlinsth_get_next (&lh, &vh))
    if (isSelectedClock (vh))
      {
	genClockDef (vh);
	genNodes (CKM->getNodesAssignedTo (vh));
      }
}

void
FiacreService::genSignalsSynchronizedWithNullClock (ptlinstances lsig)
{
}

void
FiacreService::genConstraintsNullClock (ptlinsth lh)
{
}

void
FiacreService::genAssertionsNullClock (ptlinsth lh)
{
}


/**
   It adds the synchronizations with the null clock (genSignalsSynchronizedWithNullClock())
   and the equivalences not proved to the null clock (genConstraintsNullClock())
   end the assertions with the null clock (genAssertionsNullClock()).
*/
void
FiacreService::insertNullClock ()
{
  ptlinsth lh;
  ptlinstances lsig;
  GraphLevelType graph_type;

  graph_type = vDCGFiacre->getFormatLevel ();
  if (vDCGFiacre->getNullClockSignal (&lsig))
    {
      requireNullClock = pk_true;
      genSignalsSynchronizedWithNullClock (lsig);
    }
  if (vDCGFiacre->getNullClockConstraints (&lh))
    {
      requireNullClock = pk_true;
      genConstraintsNullClock (lh);
    }
  if (isEventClockedFormat (graph_type))
    if (CKM->getNullClockAssertions (&lh))
      {
	genAssertionsNullClock (lh);
	requireNullClock = pk_true;
      }
}

#include "pKio.h"

void
FiacreService::genComponentSynchros (ptinsth vh)
{
  ptlinstances ln, ls, l;
  ptexpcal s;
  ptident id, idComponent;
  ptrepstr vstr;
  boolean bstop;
  ptinsth h;

  l = CKM->getSynchronousSignals (vh);
  ls = l;
  bstop = pk_false;
  while (ptlinstances_get_next (&ls, &s) && !bstop)
    if (EM->isSignalUnifiedWithClock (s, &h))
      {
	bstop = pk_true;
	ln = ptlinstances_copy (l);
	ln = ptlinstances_remove_exp (ln, s);
      }
  if (bstop)
    l = ln;
  TableSynchro[vh] = l;
  if (l != NILLINSTANCES)
    {
      id = getName (vh);
      if (!makeRepstr_string_ident (eqClockPrefix_string, id, &vstr))
	return;
      idComponent = Names->mkIdentName (repstr2ident (vstr));
      CKM->addAttribute (vh, FiacreAttr, (attributeValue) idComponent);
      shaper->printCompoment (ident2repstr (idComponent));
      shaper->printLeftSquareBracket ();
      genDeclaration (l, -1);
      shaper->printComma ();
      genDeclPort (id, PureType);
      shaper->printComma ();
      genDeclPort (idEndOfReaction, PureType);
      shaper->printRightSquareBracket ();
      shaper->printBeginDefinition ();
      BeginComposition ();
      while (ptlinstances_get_next (&l, &s))
	{
	  //  * -> synchro_type(vh)_type(s)[vh, s, eor]
	  beginEq ();
	  shaper->printString (eqClockPrefix);
	  printtype (PureType);
	  shaper->printunderscore ();
	  printtype (s);
	  shaper->printLeftSquareBracket ();
	  shaper->printident (id);
	  printOutEor (s, pk_true);
	  shaper->printRightSquareBracket ();
	  shaper->printLine ();
	}
      EndComposition ();
    }
}

void
FiacreService::genComponentSynchros (ptlinsth lh)
{
  ptinsth vh;

  shaper->printLine ();
  while (ptlinsth_get_next (&lh, &vh))
    if (isSelectedClock (vh))
      genComponentSynchros (vh);
}

void
FiacreService::genConstraintsAndAssertions ()
{

}


void
FiacreService::BeginComposition ()
{
  shaper->printBeginComposition ();
  bFirstEquation = pk_true;
}

void
FiacreService::EndComposition ()
{
  shaper->printEndComposition ();
  shaper->printLine ();
}

/** It generates the definition of the null clock */
void
FiacreService::genSynchronisationsNullClock ()
{
  if (requireNullClock)
    {
      beginEq ();
      shaper->printString (operatorTable[ophnull]);
      shaper->printLeftSquareBracket ();
      printOutEor (idNullClock, pk_false);
      shaper->printRightSquareBracket ();
      endEq ();
    }
}

boolean
FiacreService::genList (ptlinstances l, boolean bpreSep)
{
  ptexpcal s;

  if (l == NILLINSTANCES)
    return pk_false;
  if (bpreSep)
    shaper->printComma ();
  while (ptlinstances_get_next (&l, &s))
    {
      printSignal (s);
      if (l != NILLINSTANCES)
	shaper->printComma ();
    };
  return pk_true;
}

void
FiacreService::genSynchronisations (ptinsth vh)
{
  ptlinstances l;
  ptident id;
  boolean b;

  if (CKM->getAttribute (vh, FiacreAttr, (attributeValue *) & id))
    {
      beginEq ();
      shaper->printident (id);
      shaper->printLeftSquareBracket ();
      l = TableSynchro[vh];
      b = genList (l, pk_false);
      printOutEor (getName (vh), b);
      shaper->printRightSquareBracket ();
      endEq ();
    }
}

void
FiacreService::genSynchronisations (ptlinsth lh)
{
  ptinsth vh;

  shaper->printLine ();
  while (ptlinsth_get_next (&lh, &vh))
    if (isSelectedClock (vh))
      genSynchronisations (vh);

}

void
FiacreService::genSynchronisations ()
{
  shaper->printcomment (string2repstr (" Synchronisations "));
  genSynchronisationsNullClock ();
  genSynchronisations (CKM->getlhroots ());
  genSynchronisations (CKM->getlhplaced ());

}

void
FiacreService::printcomposition ()
{
  BeginComposition ();
  genEquations (CKM->getlhroots ());
  genEquations (CKM->getlhplaced ());
  genSynchronisations ();
  EndComposition ();
}

void
FiacreService::genDeclPort (ptident id, TypeType vtype)
{
  shaper->printSpace ();
  shaper->printident (id);
  shaper->printSpace ();
  shaper->printColon ();
  shaper->printSpace ();
  printtype (vtype);
}

void
FiacreService::genDeclPort (ptexpcal exp)
{
  genDeclPort (EM->getName (exp), EM->getType (exp));
}

boolean
FiacreService::genDeclaration (ptlinstances ldecl, int portAttr)
{
  ptexpcal exp;

  if (ldecl == NILLINSTANCES)
    return pk_false;
  while (ptlinstances_get_next (&ldecl, &exp))
    {
      genDeclPort (exp);
      if (ldecl != NILLINSTANCES)
	shaper->printComma ();
    }
  return pk_true;
}


boolean
FiacreService::printDeclarationInterface ()
{
  if (genDeclaration (inputs, 1) && (outputs != NILLINSTANCES))
    shaper->printComma ();
  genDeclaration (outputs, 2);
  return ((inputs != NILLINSTANCES) || (outputs != NILLINSTANCES));
}

void
FiacreService::printInterface ()
{
  shaper->printLeftSquareBracket ();
  printDeclarationInterface ();
  shaper->printComma ();
  genDeclPort (idEndOfReaction, PureType);
  shaper->printRightSquareBracket ();
}

void
FiacreService::printParameters ()
{
}

void
FiacreService::genConstant (ptexpcal exp)
{
  noperators ope;
  ptrepstr strval;

  switch ((int) EM->getOperatorClass (exp))
    {
    case variabl:
      printSignal (exp);
      break;
    case constant:
      ope = EM->getOperator (exp);
      strval = EM->getConstant (exp);
      switch ((int) ope)
	{
	case optrue:
	  shaper->printkwlogicons (pk_true);
	  break;
	case opfalse:
	  shaper->printkwlogicons (pk_false);
	  break;
	default:
	  shaper->printstr (strval);
	}
      break;
    default:
      Error ();
    }
}

void
FiacreService::genConstants ()
{
  ptlinstances ldecl;
  ptexpcal exp;
  ptINode node;

  ldecl = vDCGFiacre->getConstants ();
  /* const cnt_Cero: int is 0  */
  while (ptlinstances_get_next (&ldecl, &exp))
    {
      shaper->printString ("const");
      genDeclPort (exp);
      shaper->printSpace ();
      shaper->printString ("is");
      shaper->printSpace ();
      node = EM->getNode (exp);
      NM->get_rightOperand (node, &exp);
      exp = EM->getBasicExpr (exp);
      genConstant (exp);
      shaper->printLine ();
    }
}

boolean
FiacreService::mkWhen (ptexpcal s)
{
  ptident id;

  if (EM->getType (s) == PureType)
    return pk_false;
  id = Names->mkIdentName (idPrefix);
  EM->addAttribute (s, FiacreAttr, (attributeValue) id);
  genDeclPortTimeInterval (id, EM->getType (s), pk_false);
  return pk_true;
}

void
FiacreService::genLocalSignals (ptinsth vh)
{
  ptinterface vinter;
  ptexpcal s, vopd;
  ptlinstances ls;
  ptlINode ln;
  ptINode node;
  ptident id;

  vinter = CKM->getInterfaceSubHierarchy (vh);
  ls = ptinterface_get_localSignals (vinter);
  while (ptlinstances_get_next (&ls, &s))
    genDeclPortTimeInterval (EM->getName (s), EM->getType (s), pk_false);

  // Auxillary signals for the defaults, x := (E1 when h1) default (E2 when h2) will be translated in (| x := tmp1 default tmp2 | tmp1 := E1 when h1 | tmp2 := E2 when h2 |)
  ln = CKM->getNodesAssignedTo (vh);
  while (ptlINode_get_elem (&ln, &node))
    if (isSelectedNode (node))
      if (NM->is_equation (node))
	if (NM->get_rightOperand (node, &vopd))
	  {
	    if (EM->IsCellExpr (vopd))
	      {
		id = Names->mkIdentName (idPrefix);
		s = EM->getWhenOrCellArg (vopd);
		EM->addAttribute (s, FiacreAttr, (attributeValue) id);
		genDeclPortTimeInterval (id, EM->getType (s), pk_false);
	      }
	    else
	      {
		vopd = EM->getBasicExpr (vopd);
		if (EM->getOperator (vopd) == opmele)
		  {
		    mkWhen (EM->getBinaryArgstAt (vopd, 1));
		    mkWhen (EM->getBinaryArgstAt (vopd, 2));
		  }
	      }
	  }
}


/**
   It generates the interval [0, ...[
*/
void
FiacreService::genTimeInterval ()
{
  shaper->printSpace ();
  shaper->printkwin ();
  shaper->printLeftSquareBracket ();
  shaper->printint (0);
  shaper->printComma ();
  shaper->printDot ();
  shaper->printDot ();
  shaper->printDot ();
  shaper->printLeftSquareBracket ();
}

/** It generates
    v(id) : v(vtype) [0,...[ ,
*/
void
FiacreService::genDeclPortTimeInterval (ptident id, TypeType vtype, boolean blast)
{
  shaper->printString (indentStr);
  genDeclPort (id, vtype);
  genTimeInterval ();
  if (!blast)
    shaper->printComma ();
  shaper->printLine ();
}

/**
   It generates the declaration (Fiacre ports) for the clock of the operand list (list).
   When bfirst is "true", the declaration is preceded by the keywork "port".
*/
boolean
FiacreService::genDeclClocks (ptlinsth list, boolean bfirst)
{
  ptinsth vh;
  boolean b;
  ptexpcal s;

  while (ptlinsth_get_next (&list, &vh))
    if (isSelectedClock (vh))
      {
	if (CKM->isClockUnifiedWithSignal (vh, &s))
	  b = ptlinstances_isin_exp (inputs, s) || ptlinstances_isin_exp (outputs, s);
	else
	  b = pk_false;
	if (!b)
	  {
	    if (bfirst)
	      {
		shaper->printString ("port");
		shaper->printLine ();
		bfirst = pk_false;
	      }
	    genDeclPortTimeInterval (CKM->getName (vh), PureType, pk_false);
	    genLocalSignals (vh);
	  }
      }
  return bfirst;
}


void
FiacreService::printports ()
{
  ptinsth vhtop;
  ptlinsth lh;

  lh = NILLINSTH;
  vhtop = vDCGFiacre->getRootClock ();
  if (CKM->isFictitiousRootFormula (vhtop))
    lh = ptlinsth_insert_elem (lh, vhtop);
  genDeclClocks (CKM->getlhplaced (), genDeclClocks (CKM->getlhroots (), genDeclClocks (lh, pk_true)));
  genDeclPortTimeInterval (idNullClock, PureType, pk_true);
  ptlinsth_free (lh);
}

void
FiacreService::genComponentSynchros ()
{
  genComponentSynchros (CKM->getlhroots ());
  genComponentSynchros (CKM->getlhplaced ());
  genConstraintsAndAssertions ();
}

void
FiacreService::printMain (ptrepstr vstr)
{
  shaper->printcomment (string2repstr (" Main "));
  shaper->printString (vstr);
  shaper->printLine ();
}

void
FiacreService::genMainComponent ()
{
  genConstants ();

  shaper->printCompoment (ProcessNameStr);
  printInterface ();
  printParameters ();
  shaper->printBeginDefinition ();
  printports ();
  printcomposition ();
}

void
FiacreService::genFiacreComponents ()
{
  genComponentSynchros ();
  shaper->printLine ();
  genMainComponent ();
}

void
FiacreService::genClosedComponent ()
{
  ptrepstr vstrClosedComponent;
  boolean b;

  makeRepstr_repstr_string (ProcessNameStr, "_closed", &vstrClosedComponent);
  shaper->printCompoment (vstrClosedComponent);
  shaper->printBeginDefinition ();
  shaper->printString ("port");
  b = printDeclarationInterface ();
  if (b)
    shaper->printComma ();
  genDeclPort (idEndOfReaction, PureType);
  BeginComposition ();
  beginEq ();
  shaper->printString (ProcessNameStr);
  shaper->printLeftSquareBracket ();
  b = genList (inputs, pk_false);
  b = genList (outputs, b);
  printEndOfReaction (b);
  shaper->printRightSquareBracket ();
  EndComposition ();
  printMain (vstrClosedComponent);
}

void
FiacreService::declareProcesses ()
{
  shaper->openOut_langFile (ProcessNameStr);
  genSignalOperatorsInFiacre ();
  genFiacreComponents ();
  genClosedComponent ();
  shaper->closefile ();
}

void
FiacreService::declareComponents ()
{

}
