
_PRIVATE_
/**
   @return true if a clock (vh) was in the previous system and its name is returned to idprev, false otherwise (idprev is undefined).
*/
    boolean SMTPrinter::getPreviousName(ptinsth vh, ptident * idprev)
{
  ptinsth vhelem;
  ptlinsth lh;
  ptlPointer lids;
  pointer x;

  lids = lPreviousNames;
  lh = lhWithNewNames;
  while (ptlinsth_get_next(&lh, &vhelem))
    {
      ptlPointer_get_elemNext(&lids, &x);
      if (vhelem == vh)
	{
	  *idprev = (ptident &) x;
	  return pk_true;
	}
    }
  return pk_false;
}

/**
   @return the name of a clock in the previous system if the clock was in it, the name in the current system otherwise.
 */
ptident SMTPrinter::getNameInPreviousSystem(ptinsth vh)
{
  ptident id;
  if (!getPreviousName(vh, &id))
    id = getClockId(vh);
  return id;
}

/**
   Generate the declaration of the referenced clocks, elements of lh, but not yet declared.
 */
void SMTPrinter::genDeclReferencedClocks(ptlinsth lh)
{
  ptinsth vh;

  while (ptlinsth_get_next(&lh, &vh))
    if (CKM->satisfiesMarked(vh, vMarkSeen) && !CKM->satisfiesMarked(vh, vMarkDeclared))
      genClockDecl(vh);
}

/**
   Generate the declaration of the referenced clocks but not yet declared.
 */
void SMTPrinter::genDeclReferencedClocks()
{
  genDeclReferencedClocks(CKM->getlhroots());
  genDeclReferencedClocks(CKM->getlhplaced());
  genDeclReferencedClocks(unusedClocks);
}

/**
   Generate the declaration of the referenced signals but not yet declared.
*/
void SMTPrinter::genDeclReferencedSignals(ptlinstances ls)
{
  ptlinstances lsig;
  ptexpcal vsig;

  lsig = lsig_CurrentDCG;
  while (ptlinstances_get_next(&lsig, &vsig))
    {
      if (isSelectedType(vsig))
	if (!EM->isMarked(vsig, vMarkDeclared) && EM->isMarked(vsig, vMarkSeen))
	  genSigDecl(vsig);
    }
  lsig = ls;
  while (ptlinstances_get_next(&lsig, &vsig))
    {
      if (isSelectedType(vsig))
	if (!EM->isMarked(vsig, vMarkDeclared) && EM->isMarked(vsig, vMarkSeen))
	  genSigDecl(vsig);
    }
}

/**
   @return true if the identifier id is already defined, false otherwise.
*/
boolean SMTPrinter::isGeneratedInCurrentSystem(ptident id)
{
  for (std::vector < ptident >::iterator it = generatedIdentifiers.begin(); it != generatedIdentifiers.end(); ++it)
    if (eq_idents(id, *it)) return pk_true;
  return pk_false;
}


void SMTPrinter::genOptimizedMemorizedSignals()
{
  ptlINode ln;
  ptINode node;
  ptexpcal vmemo, vinit;
  
  ln = lMemoizedNodes_PreviousDCG;
  while (ptlINode_get_elem(&ln,&node))
    {
      if ( ! NM->getMemorizedSignal(node, &vmemo))
	{
	  NM->getAttribute(node, SMT_MEMORY_ID, (attributeValue*) &vmemo);
	  NM->getAttribute(node, SMT_MEMORY_INIT, (attributeValue*) &vinit);
	  ptident idMemory = getMemoryId(getSigIdMemory(vmemo, node), -1);
	  genNextValue(EM->getClock(vmemo), idMemory, vmemo);
	  genInitDelay(idMemory, vinit);
	}
    }    
}


/** 
    Generates in (FOO: name of the current graph):
    - the FOO_OPTIMIZED_Defs.smt file the definition of the objects (signals/clocks) that have been optimized from the previous graph.
    - the FOO_OPTIMIZED_Decls.smt file the declarations of the objects (signals/clocks) that have been optimized from the previous graph.
*/
void SMTPrinter::genOptimized()
{
  ptinsth vh;
  ptlinsth OptimizedClocks, lh;
  ptident id_prev;
  ptlinstances lsigOpt;

  OptimizedClocks = getOptimizedClocks();

  forceGeneration = pk_true;
  indent = pk_true;
  if (openFile("OPTIMIZED_Defs"))
    {

      genComment(" >>> Optimized from the previous system.");
      lh = lClocksPreviousSystem;

      while (ptlinsth_get_next(&lh, &vh))
	{
	  if (!getPreviousName(vh, &id_prev))
	    id_prev = getClockId(vh);
	  DEBUG(cout << " previous = " << ident2string(id_prev) << " num clock = " << CKM->getNum(vh) << endl;
	      );
	  if (!isGeneratedInCurrentSystem(id_prev))
	    {
	      if (!CKM->satisfiesVari(vh))
		{
		  DEBUG(cout << " *** WARNING: the clock " << ident2string(id_prev) << " is not defined in the current system, will be reimported." << endl; );
		  genClockDef(vh);
		  genSigDefs();
		}
	    }
	}

      genCommentDebug(" >>> Optimized from OptimizedClocks.");
      genClockDefs(OptimizedClocks);

      lsigOpt = getOptimizedSignals();


      genCommentDebug(" >>> Optimized from lsigOpt.");
      genSigDefs(lsigOpt);

      genCommentDebug(" >>> Optimized memorized signals.");
      genOptimizedMemorizedSignals();

      closeFile();
    }

  // Declarations.
  if (openFile("OPTIMIZED_Decls"))
    {
      indent = pk_false;
      forceGeneration = pk_true;
      genDeclReferencedClocks();
      genDeclReferencedSignals(lsigOpt);
      forceGeneration = pk_false;
      closeFile();
    }
  indent = pk_false;
  forceGeneration = pk_false;
}

/**
   @return the clocks that have been optimized from the previous system.
 */
ptlinsth SMTPrinter::getOptimizedClocks()
{
  ptinsth vhnew, vh;
  ptlinsth lh, lhres;
  boolean b;

  lhres = NILLINSTH;
  lh = lClocksPreviousSystem;
  while (ptlinsth_get_next(&lh, &vh))
    {
      vhnew = CKM->getRepresentativeFormula(vh);
      if (vhnew != vh)		// Excluding renamings.
	{
	  if (!isInNewSystem(vhnew) && (! CKM->satisfiesMarked(vhnew, vMarkSeen) ))
	    lhres = ptlinsth_sorted_insert_elem(lhres, vhnew);
	}
    }
  return lhres;
}

/**
   @return the signals that have been optimized from the previous graph.
 */
ptlinstances SMTPrinter::getOptimizedSignals()
{
  ptlinstances lsig;
  ptexpcal vsig;

  lSigAddedOptimized = NILLINSTANCES;
  lsig = lsig_PreviousDCG;
  while (ptlinstances_get_next(&lsig, &vsig))
    {
      if (!ptlinstances_isin_exp(lsig_CurrentDCG, vsig))
	{
	  lSigAddedOptimized = ptlinstances_insert(lSigAddedOptimized, vsig);
	}
    }
  return lSigAddedOptimized;

}

void SMTPrinter::genClocksWithNewNames()
{
  ptinsth vh;
  ptlinsth lh;
  ptident idnew, idprev;
  ptlPointer lids;
  pointer x;

  if (lhWithNewNames != NILLINSTH)
    {
      lids = lPreviousNames;
      lh = lhWithNewNames;
      indent = pk_true;

      genComment(">>> renamed clocks ");
      while (ptlinsth_get_next(&lh, &vh))
	{
	  ptlPointer_get_elemNext(&lids, &x);
	  ptinsth vhnew = CKM->getRepresentativeFormula(vh);
	  idprev = (ptident &) x;
	  idnew = getClockId(vhnew);
	  if (!isGeneratedInCurrentSystem(idprev))
	    {
	      ofile << "(= " << ident2string(idprev) << " " << ident2string(idnew) << ")";
	      genNewLine();
	      generatedIdentifiers.push_back(idprev);
	    }
	}
      genComment("<<< renamed clocks ");
      indent = pk_false;
    }
}

void SMTPrinter::genMappingOfVariables()
{
  ptinsth vh, vhnew;
  ptident id_vhnew, id_vh;
  ptlinsth lh;
  ptexpcal vsig;

  if (openFile("MappedVariables"))
    {
      lh = lClocksPreviousSystem;
      if (lh != NILLINSTH)
	{
	  forceGeneration = pk_true;
	  indent = pk_true;
	  genString(mapvarmess);
	  genNewLine();
	  while (ptlinsth_get_next(&lh, &vh))
	    {
	      vhnew = CKM->getRepresentativeFormula(vh);
	      if (vhnew != vh)
		{
		  id_vh = getNameInPreviousSystem(vh);
		  id_vhnew = getClockId(vhnew);
		  if (!eq_idents(id_vh, id_vhnew))
		    {
		      ofile << "(= " << ident2string(id_vh) << " " << ident2string(id_vhnew) << ")";
		      genNewLine();
		      genClockDef(vhnew);
		    }
		}
	    }
	  generate();
	  indent = pk_false;
	  forceGeneration = pk_false;
	}
      genClocksWithNewNames();
      closeFile();
    }

  if (ClocksToBeGenerated != NILLINSTH)
    {
      cout << " >>>>>>>>>>>>>>>>> DEBUG: The list ClocksToBeGenerated is not empty !!!" << endl;
    }
}


ptldependence  SMTPrinter::selectIODependences(ptldependence lDep, ptldependence lIODep, ptlinstances ls)
{
  ptdependence aDep;
  ptinsth hdep;
  ptINode vFrom, vTo;

  while (lDep->get_next(&aDep, &lDep))
    if ( NM-> isSelectedDependenceForAbstraction(aDep, unusedNodes, ls ))
      { 
	hdep = aDep->getDependence(&vFrom, &vTo);
	hdep = CKM->getRepresentativeFormula(hdep);
	lIODep = lIODep->insert(new Idependence(vFrom, vTo, hdep));
      }
  return lIODep;
}

ptldependence  SMTPrinter::selectIODependences()
{
  ptldependence lIODep;
  ptlinstances ls, lins, louts;

  ls = ptlinstances_sortedfusion(CurrentDCG->getOutputs(), ptlinstances_copy(CurrentDCG->getInputs()));
  NM->getIOUnsafeExternalCalls(&lins, &louts);
  ls = ptlinstances_sortedfusion(lins, ls);
  ls = ptlinstances_sortedfusion(louts, ls);
  
  lIODep =  NILLDEPENDENCE;
  lIODep = selectIODependences(CurrentDCG->getExplicitDependences(), lIODep, ls);
  lIODep = selectIODependences(CurrentDCG->getReinforcedDependences(), lIODep, ls);
  return lIODep;
}


void  SMTPrinter::genDependencyRefinement(ptldependence lpreviousIODependences, ptldependence lioDep)
{
  ptdependence aDep;
  ptinsth hDepnew, hDepPrevious;
  ptINode vFrom, vTo;

  while (lpreviousIODependences->get_next(&aDep, &lpreviousIODependences))
    {
      hDepPrevious = CKM->getRepresentativeFormula(aDep->getDependence(&vFrom, &vTo));
      if (NM->getClockDependence(lioDep, vFrom, vTo, &hDepnew))
	{
	  hDepnew = CKM->getRepresentativeFormula(hDepnew);
	  genIfClock(hDepnew);
	  genSpace();
	  genClockId(hDepPrevious);
	  genCloseBlock();
	  genNewLine();
	}
      else
	cout << " The clock of an IO Dependence becomes null clock ..." << endl;
    }
}


ptldependence SMTPrinter::genDependencyRefinement()
{
  ptldependence lioDep;

  lioDep = selectIODependences();
  genDependencyRefinement(lpreviousIODependences, lioDep);
  return lioDep;
}

