/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_ 

/**
   Contructor  of the SMTPrinter class.
*/
SMTPrinter::SMTPrinter ()
{}

/**
   Destructor of the SMTPrinter class.
*/
SMTPrinter::~SMTPrinter ()
{}

/**
   Generates the SMT code for a DCG (vDCG) but in certification mode:
   - lhToVisit : the selected clocks  of the current graph (vDCG) to be printed
   - lhPreviousSystem : list of the clocks of the previous graph
   - lhModifiedNames : the list of the clocks that have changed of names in the last functionality.
   - lhPrevNames : list of the names of the clocks of the list lhModifiedNames.
   - lnodes : list of the selected nodes of the current graph to be printed.
   - prevlsig : the list og the signals printed at the previous call.
   - !lClockForNextStep : list of the clocks not in lhToVisit but printed (useful for the next call)
   - !lsigForNextStep : list of the signals that have been printed.(useful for the next call)
 */
boolean SMTPrinter::print(ptDCG vDCG, TypeType TT, ptlinsth lhToVisit, ptlinsth lhPreviousSystem, ptlinsth lhModifiedNames, ptlPointer lhPrevNames, ptldependence liodepprev,  ptlINode lnodes, ptlinstances prevlsig, 
			  ptlINode PrevMemoizedNodes, ptlinsth *lClockForNextStep, ptlinstances *lsigForNextStep, ptldependence *liodep)
{
  modeCert = pk_true;
  PRINTMEMORY=pk_true;

  Clocks_CurrentDCG = lhToVisit;
  lnodestogenerate = lnodes;
  lClocksPreviousSystem = lhPreviousSystem;
  lhWithNewNames = lhModifiedNames;
  lPreviousNames = lhPrevNames;
  lsig_PreviousDCG = prevlsig;
  lMemoizedNodes_PreviousDCG = PrevMemoizedNodes;
  lpreviousIODependences = liodepprev;
  selectedType= TT;

  initialize(vDCG);

  if (openFile("Defs"))
    {
      genSMTFunctionInit(idGraph);
      genDefinitions();
      genSMTFunctionFinal();
      indent = true;
      closeFile();
    }

  if (openFile("Dependences"))
    {
      *liodep = genDependencyRefinement();
      closeFile();
      }

  if (openFile("Decls"))
    {
      genDeclarations();
      closeFile();
    }

  if (openFile("Constants"))
    {
      genConstants();
      closeFile();
    }
  
  genMappingOfVariables();      
  genOptimized();

  finalize();
  *lClockForNextStep = ClocksForTheNextLevel;
  *lsigForNextStep = ptlinstances_sortedfusion (SignalsForTheNextLevel, lSigAddedOptimized);
  if (_ERROR_) cout << " *** ERROR during the SMT print (certification functionality)  !!!" << endl;
  return ! _ERROR_;
}


/**
   Generates the SMT code for a DCG (vDCG), not in certification mode.
   All the elements, type dependent (TO BE COMPLETED xxxxxxxxx) of the graph are printed.
 */
boolean SMTPrinter::print(ptDCG vDCG, TypeType TT)
{
  modeCert = pk_false;
  PRINTMEMORY=pk_true;
  lClocksPreviousSystem = NILLINSTH;
  lhWithNewNames = NILLINSTH;
  lPreviousNames = NILLPOINTER;
  lsig_PreviousDCG = NILLINSTANCES;
  lMemoizedNodes_PreviousDCG = NILLINODE;
  lpreviousIODependences = NILLDEPENDENCE;

  selectedType= TT;

  initialize(vDCG);

  lnodestogenerate = NILLINODE;
  NM->getAllNodes(&lnodestogenerate);

  Clocks_CurrentDCG = ptlinsth_duplicate(CKM->getlhroots());
  Clocks_CurrentDCG = ptlinsth_sorted_fusion (Clocks_CurrentDCG, CKM->getlhplaced());

  if (openFile(""))
    {
      genDeclarations();
      genConstants();
      genSMTFunctionInit(idGraph);
      genDefinitions();   
      genDefinitions(lnodestogenerate);
      genSMTFunctionFinal();
      closeFile();
    }

  if (_ERROR_) cout << " *** ERROR during the SMT STD print !!!" << endl;
  finalize();
  return  ! _ERROR_;
}


