/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_

ptlinsth  SMTService::getClocksToExport(boolean ballclocks)
{
  ptExprManager EM; 
  ptClockManager CKM;
  ptNodeManager NM;
  ptlINode lnodes;
  ptINode node;
  ptlinsth lh;

  CKM = exportedDCG->getClockManager();
  EM = exportedDCG->getExpressionManager();
  NM = exportedDCG->getNodeManager();
  
  lh = EM->getClockOfSignals (exportedDCG->getInputs ());
  lh = EM->addClockSignals(exportedDCG->getOutputs (), lh);
  if (NM->getProcCalls(&lnodes))
    {
      while (ptlINode_get_elem(&lnodes,&node))
	{
	  ptlinstances lsig = NM->getInodelOutputFlows(node);
	  lh = EM->addClockSignals(lsig, lh);
	}
    }
  
  if (NM->getNodeCalls(&lnodes))
    {
      while (ptlINode_get_elem(&lnodes,&node))
	{
	  ptlinstances lsig = NM->getInodelOutputFlows(node);
	  lh = EM->addClockSignals(lsig, lh);
	}
    }
  
  return lh;
}

/**
   IA : integer arithmetic
   IRA : mixed integer Real arithmetic

   L : before IA, IRA for linear fragment of those arithmetics
   N : before IA, IRA for non-linear fragment of those arithmetics
 */
TypeType SMTService::setSelectedTypeAndLogic()
{
  string SMT_Logic = sysGetEnvDef ("SMT_LOGIC");

  if ( SMT_Logic == "QF_BV") return BooleanType ;

  // integer arithmetic
  if (SMT_Logic == "QF_IA")  return LongType;
  if (SMT_Logic == "QF_LIA")  return LongType;
  if (SMT_Logic == "QF_NIA")  return LongType;

  // mixed integer real arithmetic
  if (SMT_Logic == "QF_IRA")  return DRealType;
  if (SMT_Logic == "QF_LIRA")  return DRealType;
  if (SMT_Logic == "QF_NIRA")  return DRealType;

  return ErrorType;
}


boolean  SMTService::isProgramWithSMTRestrictions()
{
  ptExprManager EM; 
  ptClockManager CKM;
  ptNodeManager NM;
  ptexpcal vtype;

  CKM = vDCG->getClockManager();
  EM = vDCG->getExpressionManager();
  NM = vDCG->getNodeManager();

  ptlINode ln;
  if ( NM ->getIterates(&ln))
    {
      // pK_messenger->printString("SMT printer: loops are not yet implemented");
      return pk_true;
    }
  ptlinstances ltypes;
  if (EM->getUserTypes (&ltypes))
    {
      while (ptlinstances_get_next(&ltypes, &vtype))
	{
	  TypeType TT = EM->getExpType(vtype);
	  if (! TypeTestPredef(TT)) 
	    {
	      if (TypeTestNamedClass(TT))
		{
		  TT = NamedTypeGetName(TT);
		  if (! TypeTestPredef(TT)) return pk_true;
		}
	    }
	}
    }
  
  return pk_false;
}


