/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
// The definition of SIGALI_FUNC_NAME must be the same as in the the SIGALI Signal library.
#define SIGALI_FUNC_NAME "Sigali"
// The definition of CONTROLLABLE_FUNC_NAME must be the same as in the the SIGALI Signal library.
#define CONTROLLABLE_FUNC_NAME "Controllable"
#define  cstretat "states"
#define  cstrcond "cond"
#define  cstrletat "states:"
#define  cstrlcond "events:"
#define  cstrlevol "evolutions:"
#define  cstrlinit "initialisations:"
#define  cstrlcontr "constraints:"
#define  cstrlvar "variables:"
#define  cstrdecl "declare"
#define  cstrlcont "controllables:"
#define  cstrlcondlibres "free_cond:"


/**
   Must be called on internal error, for example a not implemented operation (codeOpstr). It 
   occurs on an expression (vsig).
   Global code error BERROR is updated.
*/
void
SigaliService::sendNotimplemented (ptrepstr codeOpstr, ptexpcal vsig)
{
  // BERROR = pk_true;
}

boolean
SigaliService::isSelectedNode (ptINode n)
{
  return !ptlINode_isIn (unNodes, n);
}

boolean
SigaliService::isSelectedClock (ptinsth h)
{
  return !ptlinsth_isin_elem (unClocks, h);
}


/**
   @return the "Z3Z" attribute associated with an expression (vexp).
   The attribute is \ref z3zAttribute() "created" if it does not exist.
*/
ptz3zAttr
SigaliService::getexpcalattr (ptexpcal vsig)
{
  ptz3zAttr attribut;
  if (!EM->getAttribute (vsig, attributeNumber, (attributeValue *) & attribut))
    {
      attribut = new z3zAttribute ();
      EM->addAttribute (vsig, attributeNumber, (attributeValue) attribut);
    }
  return attribut;
}

/**
   @return the "Z3Z" attribute associated with a clock (h).
   The attribute is \ref z3zAttribute() "created" if it does not exist.
*/
ptz3zAttr
SigaliService::getinsthattr (ptinsth h)
{
  ptz3zAttr attribut;
  if (!CKM->getAttribute (h, attributeNumber, (attributeValue *) & attribut))
    {
      attribut = new z3zAttribute ();
      CKM->addAttribute (h, attributeNumber, (attributeValue) attribut);
    }
  return attribut;
}


/** 
    @return the current mark associated with the signal (vsig).
*/
z3zMarkType
SigaliService::expression_get_z3mark (ptexpcal vsig)
{
  ptz3zAttr vattr;
  vattr = getexpcalattr (vsig);
  return vattr->getMark ();
}

/** It assigns a mark (m) to a signal (vsig). */
void
SigaliService::expression_set_z3mark (ptexpcal vsig, z3zMarkType m)
{
  ptz3zAttr vattr;

  vattr = getexpcalattr (vsig);
  vattr->setMark (m);
}

/** @return the polynomial associated with a signal (vsig). */
ptpolynomial
SigaliService::expression_getPolynomial (ptexpcal vsig)
{
  ptz3zAttr vattr;

  vattr = getexpcalattr (vsig);
  return vattr->getPolynomial ();
}

/** It associates a polynomial (p) with a signal (vsig). */
void
SigaliService::expression_setPolynomial (ptexpcal vsig, ptpolynomial p)
{
  ptz3zAttr vattr;

  vattr = getexpcalattr (vsig);
  vattr->setPolynomial (p);
}

/** @return the polynomial associated with a clock (h). */
ptpolynomial
SigaliService::clockAttribute_get_z3poly (ptinsth h)
{
  ptz3zAttr vattr;

  vattr = getinsthattr (h);
  return vattr->getPolynomial ();
}

/** 
    @return the current mark associated with a a clock (h).
*/
z3zMarkType
SigaliService::clockAttribute_get_z3marque (ptinsth h)
{
  ptz3zAttr vattr;
  vattr = getinsthattr (h);
  return vattr->getMark ();
}

/** It assigns a mark (m) to a clock (h). */
void
SigaliService::clockAttribute_set_z3marque (ptinsth h, z3zMarkType m)
{
  ptz3zAttr vattr;
  vattr = getinsthattr (h);
  vattr->setMark (m);
}

/** It associates a polynomial (p) with a clock (h). */
void
SigaliService::clockAttribute_set_z3poly (ptinsth h, ptpolynomial p)
{
  ptz3zAttr vattr;
  vattr = getinsthattr (h);
  vattr->setPolynomial (p);
}

/**
   @return "true" if the clock (vh) of an assumed synchronous expression (pt) is found, "false" otherwise (constant expression, ...).
*/
boolean
SigaliService::getSyncExprClock (ptexpcal pt, ptinsth * vh)
{
  if (!EM->getSynchronousExprHorl (pt, vh)) return pk_false;
  *vh = CKM->getRepresentativeFormula (*vh);
  return pk_true;
}

/**@return the identifier of a clock (aclock).*/
ptident
SigaliService::z3mkhonametree (ptinsth aclock)
{
  return CKM->getName (aclock);
}

/**
   @return  "true" if a clock (h) is \e useful, "false" otherwise.
   A clock is \e useful when it is the clock of a user-signal (not an auxiliary signal created by the compiler).
*/
boolean
SigaliService::isUsefulClock (ptinsth h)
{
  ptlinstances lsig;
  ptexpcal vsig;

  lsig = CKM->getSynchronousSignals (h);
  while (ptlinstances_get_next (&lsig, &vsig))
    if (!EM->IsAuxiliarySignal (vsig)) return pk_true;
  return pk_false;
}

/**
   @return "true" if a signal (vsig) is defined in a SIGALI intrinsic node, "false" otherwise.
*/
boolean
SigaliService::isOutputSigaliIntrinsic (ptexpcal vsig)
{
  ptINode node, vinstr;

  node = EM->getNode (vsig);
  if (NM->is_outputNodeCall (node, &vinstr))
    node = vinstr;
  //  if (!NM->isProcessCall(node)) return pk_false;if (!NM->is_intrinsicfunc(vinstr)) return pk_false;
  return existPredefinedPragma (NM->getPragmas (node), PRAGMA_SIGALI);
}

/**
   @return "true" if all the nodes associated with a clock (h) are SIGALI intrinsic nodes, "false" otherwise
*/
boolean
SigaliService::isClockOfOnlySigaliFunctions (ptinsth h)
{
  ptlinstances lsig;
  ptexpcal vsig;

  lsig = CKM->getSynchronousSignals (h);
  if (lsig == NILLINSTANCES) return pk_false;
  while (ptlinstances_get_next (&lsig, &vsig))
    if (!isOutputSigaliIntrinsic (vsig)) return pk_false;
  return pk_true;

}

/**
   @return  "true" is a signal (e) is \e useful, "false" otherwise.
   It is useful when 
   - it is not unified with a clock (event), 
   - it is not a signal created by the compiler,
   - it is a scalar boolean signal.
*/
boolean
SigaliService::isUsefulSignal (ptexpcal e)
{
  ptinsth h;
  if (EM->isSignalUnifiedWithClock (e, &h)) return pk_false;
  if (!EM->IsSignal (e)) return pk_false;
  if (EM->IsAuxiliarySignal (e)) return EM->isMarked (e, refInSigaliCmd);
  if (!EM->IsBoolean (e)) return pk_false;
  return (EM->getNumberOfDimensions (e) == 0);
}

/** 
    @return  "true" if a signal (vsig) is \e free, "false" otherwise.
    It is free if it is a (scalar) boolean signal, and
    - it is an input of the specification,
    - or, it is defined by a synchronous expression but not a logical one (built on {and, or, not} operators).
*/
boolean
SigaliService::isFreeBoolean (ptexpcal vsig)
{
  ptINode instr;
  noperators oper, t;
  ptexpcal e, vdedef, aux;

  if (!EM->IsBoolean (vsig)) return pk_false;
  if (ptlinstances_isin_exp (vDCGSigali->getInputs (), vsig)) return pk_true;
  instr = EM->getNode (vsig);
  oper = NM->get_oper (instr);
  if (oper == opappe) return pk_true;
  if (oper != opdefsig) return pk_false;
  NM->get_rightOperand (instr, &vdedef);
  if (EM->IsCellExpr (vdedef)) return pk_false;
  e = EM->getBasicExpr (vdedef);
  switch ((int) (EM->getOperatorClass (e)))
    {
    case oper1: return (EM->getOperator (e) != opnon);

    case oper2:
      t = EM->getOperator (e);
      if (t == opet) return pk_false;
      if (t == opou) return pk_false;
      if ((t == opegal) || (t == opdiff))
	{
	  aux = EM->getBinaryArgstAt (e, 1);
	  return !EM->IsBoolean (aux);
	}
      return pk_true;
      // return !((t == opet) || (t == opou));

    default:
      return pk_false;
    }
  return pk_false;
}

/** It adds the definition id : poly  to the list of definitions (listdef)*/
void
SigaliService::mkdefinition (ptident id, ptpolynomial poly)
{
  adddefpoly (listdef, id, poly);
}

/** It adds the constraints p1 = p2 to the list of constraints (listconstraints). */
void
SigaliService::addConstraint (ptpolynomial p1, ptpolynomial p2)
{
  addpolynomial (listconstraints, polyequa (p1, p2));
}

/** It adds the constraints p² = h to the  list of constraints  (listconstraints). */
void
SigaliService::synchro (ptpolynomial p, ptpolynomial h)
{
  addpolynomial (listconstraints, polyequa (polysquare (p), h));
}

/**
   It adds a variable (v) in the list of variables (listvar) and in the list of conditions (listconditions).
*/
void
SigaliService::declarevariable (ptpolynomial v)
{
  addpolynomial (listvar, v);
  addpolynomial (listconditions, v);
}



/** @return the polynomial for 'when e' expression.  */
ptpolynomial
SigaliService::translatesub (ptexpcal e)
{
  return polywhen1 (translateExpression (e));
}

/** @return the polynomial for 'when not e' expression.  */
ptpolynomial
SigaliService::translatenosub (ptexpcal e)
{
  return polywhen1 (polynot (translateExpression (e)));
}

/** @return the polynomial for 'not e' expression. */
ptpolynomial
SigaliService::translateNot (ptexpcal e)
{
  return polynot (translateExpression (e));
}

/** @return the polynomial for 'e1 and e2' expression. */
ptpolynomial
SigaliService::translateAnd (ptexpcal e1, ptexpcal e2)
{
  return polyand (translateExpression (e1), translateExpression (e2));
}

/** @return the polynomial for 'e1 or e2' expression. */
ptpolynomial
SigaliService::translateOr (ptexpcal e1, ptexpcal e2)
{
  return polyor (translateExpression (e1), translateExpression (e2));
}

/** @return the polynomial for 'e1=e2' expression. */
ptpolynomial
SigaliService::translateEqual (ptexpcal e1, ptexpcal e2)
{
  return polyor (polyand (translateExpression (e1), translateExpression (e2)),
		 polyand (polynot (translateExpression (e1)), polynot (translateExpression (e2))));
}

/** @return the polynomial for 'e1 != e2' expression. */
ptpolynomial
SigaliService::translateDiff (ptexpcal e1, ptexpcal e2)
{
  return polyor (polyand (translateExpression (e1), polynot (translateExpression (e2))),
		 polyand (polynot (translateExpression (e1)), translateExpression (e2)));
}

/** @return the polynomial for not(h). */
ptpolynomial
SigaliService::translatehnot (ptinsth h)
{
  return polydifference (polyconstant (1), translateClock (h));
}

/** @return the polynomial for h1^* h2. */
ptpolynomial
SigaliService::translateinfer (ptinsth h1, ptinsth h2)
{
  return polymult (translateClock (h1), translateClock (h2));
}

/** @return the polynomial for h1 default h2. */
ptpolynomial
SigaliService::translatesuper (ptinsth h1, ptinsth h2)
{
  return polydefault (translateClock (h1), translateClock (h2));
}

/** @return the polynomial for h1^- h2. */
ptpolynomial
SigaliService::translatediffer (ptinsth h1, ptinsth h2)
{
  return polymult (translateClock (h1), polydifference (polyconstant (1), translateClock (h2)));
}

ptpolynomial
SigaliService::translateDefault (ptexpcal e)
{
  ptexpcal he1, he2;
  ptpolynomial p1, p2;
  ptinsth h1, h2, hcontr;
  ptexpcal e2;

  EM->getDefaultArgs (e, &he1, &he2);
  p1 = translatewhen (EM->getWhenOrCellArg (he1), EM->getContributionClock (he1));
  hcontr = EM->getContributionClock (he2);
  e2 = EM->getWhenOrCellArg (he2);
  if (CKM->satisfiesNullClock (hcontr))
    return p1;
  if (CKM->getMinusArgs (hcontr, &h1, &h2))
    p2 = translatewhen (e2, h2);
  else
    p2 = translatewhen (e2, hcontr);
  return polydefault (p1, p2);
}

/**  @return the polynomial for 'e when h' expression.*/
ptpolynomial
SigaliService::translatewhen (ptexpcal e, ptinsth h)
{
  ptpolynomial p1, p2;
  ptinsth h1;

  if (EM->getOperatorClass (e) == oper2tp)
    return translateDefault (e);
  if (getSyncExprClock (e, &h1) && (h == h1))
    return translateExpression (e);
  p1 = translateClock (h);
  p2 = translateExpression (e);
  return polymult (p2, p1);
}


/**
   It generates the polynomial definition for a formula (h).
   The case of a "free variable" is not generated here, but filtered in translateClock().
*/
ptpolynomial
SigaliService::genFormulaExpression (ptinsth h)
{
  ptpolynomial poly;
  ptinsth h1, h2;

  switch ((int) (CKM->getOperator (h)))
    {
    case sub:
      poly = translatesub (CKM->getConditionSub (h));
      break;
    case nosub:
      poly = translatenosub (CKM->getConditionSub (h));
      break;
    case hnot:
      CKM->getUnaryArg (h, &h1);
      poly = translatehnot (CKM->getRepresentativeFormula (h1));
      break;
    case infer:
      CKM->getInfArgs (h, &h1, &h2);
      poly = translateinfer (h1, h2);
      break;
    case super:
      CKM->getSupArgs (h, &h1, &h2);
      poly = translatesuper (h1, h2);
      break;
    case vari:
      break;
    case predef:
      if (CKM->satisfiesNullClock (h))
	poly = polyconstant (0);
      else if (CKM->satisfiesInitClock (h))
	poly = polyconstant (1);
      break;
    case hcomp:
      CKM->getUnaryArg (h, &h2);
      poly = translatediffer (CKM->getFather (h), h2);
      break;
    default:;
    }
  return poly;
}


/**
   It generates the definition of a free clock (h):
   - if h is unified with a input signal, it is declared as a variable and an input of the system. 
   The associated polynomial is h².
   - if it exists a signal S, with h as clock, and S is a boolean input of the specification, a polynomial
   variable PS is associated with S. PS is declared as a variable and an input of 
   the system ans the definition h=PS² is added in the system.
   - otherwise, it is declared as a variable and an input of the system. The associated polynomial is h².
*/
void
SigaliService::genFreeClockDefinition (ptinsth h)
{
  ptident nom;
  ptpolynomial poly1, poly2;
  ptlinstances lsig;
  ptexpcal vsig;

  nom = z3mkhonametree (h);
  poly1 = polyvariable (nom);

  // case 1
  if (CKM->isClockUnifiedWithSignal (h, &vsig))
    if (ptlinstances_isin_exp (vDCGSigali->getInputs (), vsig))
      {
	declarevariable (poly1);
	clockAttribute_set_z3marque (h, z3z_defined);
	clockAttribute_set_z3poly (h, polysquare (poly1));
	return;
      }
  // case 2
  lsig = CKM->getSynchronousSignals (h);
  while (ptlinstances_get_next (&lsig, &vsig))
    if (!isOutputSigaliIntrinsic (vsig))
      {
	if (isFreeBoolean (vsig))
	  {
	    poly2 = polyvariable (EM->getName (vsig));
	    if (expression_get_z3mark (vsig) != z3z_defined)
	      {
		declarevariable (poly2);
		expression_set_z3mark (vsig, z3z_defined);
	      };
	    expression_setPolynomial (vsig, poly2);
	    clockAttribute_set_z3marque (h, z3z_defined);
	    clockAttribute_set_z3poly (h, poly1);
	    mkdefinition (nom, polysquare (poly2));
	    return;
	  }
      }
  // otherwise
  declarevariable (poly1);
  clockAttribute_set_z3marque (h, z3z_defined);
  clockAttribute_set_z3poly (h, polysquare (poly1));
}


/**
   If the definition of a clock (h) is 
   - non-recursive: it adds the definition h:poly in the system.
   - recursive: h is declared as a new variable (p1), and input, then a constraint p1=poly is added to the system.

   poly is the polynomial associated with the formula (produced by genFormulaExpression() function.) 

   Assumed: h is not a free variable (see genFreeClockDefinition())

   \note : this method has been modified to fix the bug #8553 (test CKM->isRecursiveFormula()).
*/
void
SigaliService::genClockDefinition (ptinsth h)
{
  ptident nom;
  ptpolynomial poly1, poly2;

  nom = z3mkhonametree (h);
  poly1 = polyvariable (nom);
  clockAttribute_set_z3marque (h, z3z_current);
  if (CKM->isRecursiveFormula (h))
    {
      declarevariable (poly1);
      poly1 = polysquare (poly1);
      clockAttribute_set_z3poly (h, poly1);
      poly2 = genFormulaExpression (h);
      addConstraint (poly1, poly2);
    }
  else
    {
      clockAttribute_set_z3poly (h, poly1);
      poly2 = genFormulaExpression (h);
      mkdefinition (nom, poly2);
    }
}


/**
   It generates the polynomial expression associated with a clock (h) taking into account the
   \ref z3zMarkType attribute associated to each clock. It filters the clock used only to express sigali function,
   in this case, it returns undefined polynomial. Otherwise,
   - if the definition is known (case z3z_defined), it is returned,
   - if the definition is "z3z_current", the definition is a recursive one. The clock is marked "z3z_defined". It will be completly
   processed when the all expression will be explored.
   - if the definition is "z3z_undefined", it is processed. There are several cases:
   - it is a free clock (see genFreeClockDefinition()),
   - it is a useful clock (see isUsefulClock()), it processes by genClockDefinition() that fixes the eventual recursivity.
   - it is not a useful clock, it is an internal formula. The associated polynomial is generated, the recursivity is fixed like 
   previously. The difference with the "useful clock" is that a reference to h in an expression is substituted by the 
   polynomial definition in the expression.

   \note : this method has been modified to fix the bug #8553 (test CKM->isRecursiveFormula()).
*/
ptpolynomial
SigaliService::translateClock (ptinsth h)
{
  ptpolynomial poly, poly2;

  if (isClockOfOnlySigaliFunctions (h)) return NILPOLYNOMIAL;
  switch ((int) (clockAttribute_get_z3marque (h)))
    {
    case z3z_defined:
      return clockAttribute_get_z3poly (h);

    case z3z_undefined:
      if (CKM->satisfiesVari (h))
	{
	  genFreeClockDefinition (h);
	  return clockAttribute_get_z3poly (h);
	}
      if (isUsefulClock (h))
	{
	  genClockDefinition (h);
	  return clockAttribute_get_z3poly (h);
	}
      clockAttribute_set_z3marque (h, z3z_current);
      if (CKM->isRecursiveFormula (h))
	{
	  poly2 = polyvariable (z3mkhonametree (h));
	  declarevariable (poly2);
	  poly2 = polysquare (poly2);
	  clockAttribute_set_z3poly (h, poly2);
	  poly = genFormulaExpression (h);
	  addConstraint (poly2, poly);
	  return poly2;
	}
      poly = genFormulaExpression (h);
      clockAttribute_set_z3marque (h, z3z_defined);
      clockAttribute_set_z3poly (h, poly);
      return poly;


    case z3z_current:		// recursive definition.
      poly = clockAttribute_get_z3poly (h);
      if (poly == NILPOLYNOMIAL)	// never pk_true ???
	{
	  poly = polyvariable (z3mkhonametree (h));
	  clockAttribute_set_z3poly (h, poly);
	}
      clockAttribute_set_z3marque (h, z3z_defined);
      return poly;

    default:
      return NILPOLYNOMIAL;
    }
}


/** @return a new state variable with id as name, associated with the Signal expression: e $1 init init, with h=^e.
    It is added in the list of delays (listDelays). This list is explored  at the end of the translation to
    generates the polynomials that express the dynamical of the system (see \ref translateProgram() method).
    The state variable is added in the list of the variables and the list of the state variables.
    \remark it is used in special context: translateretard(), translatecell().
*/
ptpolynomial
SigaliService::declareStateVariable (ptident id, ptexpcal e, ptinsth h, ptpolynomial init)
{
  ptpolynomial ksi;

  ksi = polyvariable (id);
  addpolynomial (listvar, ksi);
  addpolynomial (liststatevar, ksi);
  addDelay (listDelays, ksi, e, h, init);
  return ksi;
}

/** @return a new state variable associated with the Signal expression: e $1 init init, with h=^e.
 */
ptpolynomial
SigaliService::makeStateVariable (ptexpcal e, ptinsth h, ptpolynomial init)
{
  numetat = numetat + 1;
  return declareStateVariable (makeIdent_ident_int (prefixeetat, numetat), e, h, init);
}

/**
   @return the state variable for a delayed signal (e) with h as clock and initialized to init.
   It searches if such object exists (listDelays), a new one is created if it is not found.
   The state variable (x) allows to express the evolution of the system by: x' = e + (1-e²) x , x0=init, h=e².
*/
ptpolynomial
SigaliService::getStateVariable (ptexpcal e, ptinsth h, ptexpcal init)
{
  ptpolynomial pinit;
  ptrepDelay retard;

  if (init == NILEXPCAL)
    return makeStateVariable (e, h, NILPOLYNOMIAL);
  pinit = translateExpression (init);

  initIteratorlDelay (listDelays);
  while (!isEndlDelay (listDelays))
    {
      retard = nextDelay (listDelays);
      // Optimisation ? if ((getDelaySig(retard) == e) /*&& (getDelayClock(retard) == h) */ && (getDelayInit(retard) == pinit))
      if ((getDelaySig (retard) == e) && (getDelayClock (retard) == h) && (getDelayInit (retard) == pinit))
	return getDelaystatevar (retard);
    }
  return makeStateVariable (e, h, pinit);
}

/** It creates a state variable (x) for a delayed signal (e) with h as clock and initialized to init, such that
    x' = e + (1-e²) x, x0=init, h=^e (e² in z3z).
    @return the polynomial e²* x.
    \remark it is used when a delay doen't verify the "waited" form for the SIMULATION\n
    zx  := x default zzx | zzx := zx $ 1 init x0\n
*/
ptpolynomial
SigaliService::translateDelay (ptexpcal e, ptinsth h, ptexpcal init)
{
  ptpolynomial ksi, ph;

  ph = translateClock (h);
  ksi = getStateVariable (e, h, init);
  return polymult (ksi, ph);
}

/**
   @return the polynomial associated with an initialized (init) cell expression (e).
*/
ptpolynomial
SigaliService::translatecell (ptexpcal e, ptexpcal init)
{
  ptpolynomial ksi, p1, p2;
  ptinsth hcontr;
  ptinsth hinstr;
  // E :: (E1 when HCONTR) cell HINSTR, INIT = initialisation
  hinstr = CKM->getRepresentativeFormula (EM->getContributionClock (e));
  e = EM->getWhenOrCellArg (e);
  hcontr = CKM->getRepresentativeFormula (EM->getContributionClock (e));
  p2 = translateClock (hinstr);
  p1 = translatewhen (e, hcontr);
  e = EM->getWhenOrCellArg (e);
  ksi = getStateVariable (e, hcontr, init);
  return polydefault (p1, polymult (ksi, p2));
}

/**
   @return the polynomial expression associated with the definition of an assumed boolean signal (vsig).
   Only equation and delay nodes are considered:
   - for a delay see translateDelay(),
   - for an equation several cases: see translatecell(), translatewhen(), translateExpression()
*/
ptpolynomial
SigaliService::definitionsignal (ptexpcal vsig)
{
  ptpolynomial result;
  ptINode instr;
  ptinsth hinstr;
  ptexpcal e, x, vret, vfen, vdedef, vinit;
  // long int val;

  result = NILPOLYNOMIAL;
  instr = EM->getNode (vsig);
  hinstr = NM->get_nodeClock (instr);
  switch ((int) (NM->get_oper (instr)))
    {
    case opdefsig:
      NM->get_rightOperand (instr, &vdedef);
      if (EM->IsCellExpr (vdedef))
	result = translatecell (vdedef, EM->getInitialValue (vsig));
      else
	{
	  e = EM->getBasicExpr (vdedef);
	  if (EM->IsSignal (vsig))
	    result = translatewhen (e, hinstr);
	  else
	    result = translateExpression (e);
	}
      break;

    case opret:
      NM->get_delay_arg (instr, &x, &e, &vret, &vfen, &vinit);
      // Restriction : delay 1 assumed.
      //      if (!EM->getIntegerValue(vret, &val)) GdelayOne=false; else if (val != 1) GdelayOne=false; 
      result = translateDelay (e, hinstr, vinit);
      break;

    case opappe:
      sendNotimplemented (string2repstr ("process Call"), vsig);
      break;

    case opnodecall:
      sendNotimplemented (string2repstr ("Node Call"), vsig);
      break;

    case opsubgraph:
      sendNotimplemented (string2repstr ("Sub-Graph reference"), vsig);
      break;

    case opmoti:
      sendNotimplemented (string2repstr ("Array of Processes"), vsig);
      break;

    case opbund:
      sendNotimplemented (string2repstr ("Bundle definition"), vsig);
      break;
    case opldefpart:
    case opdefpart:
      sendNotimplemented (string2repstr ("Partial Definition"), vsig);
      break;

    default:;
    }
  return result;
}

/**
   It defines the polynomial (S) expression associated with an assumed boolean signal (vsig).
   If the signal is a free boolean (isFreeBoolean()) it is declared as an input in the Dynamical System and 
   the constraint ^vsig=S² is added.
   Otherwise, the definition of the signal is translated in \e polyexp (definitionsignal()):
   if the definition is recursive, the signal becomes a new input and the defintion is added as a constraint.
   otherwise S:\e polyexp is added in the definitions. 
*/
void
SigaliService::genPolynomialForSignal (ptexpcal vsig)
{
  ptident nom;
  ptpolynomial poly1, poly2;
  ptinsth hsig;

  nom = EM->getName (vsig);
  poly1 = polyvariable (nom);
  expression_setPolynomial (vsig, poly1);
  if (isFreeBoolean (vsig))
    {
      expression_set_z3mark (vsig, z3z_defined);
      declarevariable (poly1);
      if (getSyncExprClock (vsig, &hsig))
	synchro (poly1, translateClock (hsig));
    }
  else
    {				/* signal intermediaire */
      expression_set_z3mark (vsig, z3z_current);
      poly2 = definitionsignal (vsig);
      if (expression_get_z3mark (vsig) != z3z_current)
	{			/* cas recursif */
	  declarevariable (poly1);
	  addConstraint (poly1, poly2);
	  if (getSyncExprClock (vsig, &hsig))
	    synchro (poly1, translateClock (hsig));	// added November 2008 (L. Besnard)
	}
      else			/* cas non recursif */
	mkdefinition (nom, poly2);
    }
}


/**
   It generates the polynomial expression associated with an assumed boolean signal (e).
   Several cases are possible (recursivity):
   - the signal is marked as "z3z_defined": the associated polynomial is returned.
   - the signal is marked as "z3z_undefined":
   - and unified with a clock (event), the polynomial associated with the clock is returned (translateClock()).
   - and it is a isUsefulSignal() "useful signal", it processes by genPolynomialForSignal() that fixes the eventual recursivity. 
   - it is not a useful signal, it is an internal signal. The associated polynomial is generated, the recursivity is fixed like 
   previously (in genPolynomialForSignal()). 

   - if the signal is marked "z3z_current", the definition is a recursive one. The signal is marked "z3z_defined". It will be completly
   processed when the all expression will be explored.
*/
ptpolynomial
SigaliService::translateSignal (ptexpcal e)
{
  ptinsth h;
  ptpolynomial poly, poly2;

  switch ((int) (expression_get_z3mark (e)))
    {
    case z3z_undefined:
      if (EM->isSignalUnifiedWithClock (e, &h))
	return translateClock (h);
      if (isUsefulSignal (e))
	{
	  genPolynomialForSignal (e);
	  return expression_getPolynomial (e);
	}
      expression_set_z3mark (e, z3z_current);
      poly = definitionsignal (e);
      if (expression_get_z3mark (e) != z3z_current)
	{
	  poly2 = expression_getPolynomial (e);
	  declarevariable (poly2);
	  addConstraint (poly2, poly);
	  return poly2;
	}
      expression_setPolynomial (e, poly);
      expression_set_z3mark (e, z3z_defined);
      return poly;

    case z3z_current:
      poly = expression_getPolynomial (e);
      if (poly == NILPOLYNOMIAL)
	{
	  poly = polyvariable (EM->getName (e));
	  expression_setPolynomial (e, poly);
	}
      expression_set_z3mark (e, z3z_defined);
      return poly;
      break;
    case z3z_defined:
      return expression_getPolynomial (e);

    default:
      return polyconstant (0); // nullPoly
    }
}


/**
   @return a polynomial that represents a free boolean variable (e, predicate on non booleans expressions).
   Here, the freeConditionAttr attribute is used to link the Signal Graph world with the Sigali representation:
   if a "freeConditionAttr" attribute is associated with the expression (e), it is returned. Ohterwise, a new polynomial (free variable)
   is created, it is associated with the expression (e) as an attribute and returned.
*/
ptpolynomial
SigaliService::condition (ptexpcal e)
{
  ptpolynomial cond;
  ptinsth hcond;

  if (EM->getAttribute (e, freeConditionAttr, (attributeValue *) & cond))
    return cond;

  numcond = numcond + 1;
  cond = polyvariable (makeIdent_ident_int (prefixecond, numcond));
  declarevariable (cond);
  addpolynomial (listfreeconditions, cond);
  EM->addAttribute (e, freeConditionAttr, (attributeValue) cond);
  if (!EM->isStateVar (e))
    if (getSyncExprClock (e, &hcond))
      synchro (cond, translateClock (hcond));
  return cond;
}

/**
   It generates the polynomial expression associated with an assumed boolean expression (e) in the following way:
   - for unary (not) and binary synchronous expressions (and, or) it calls specific functions (translateNot(), translateAnd() translateOr()) 
   otherwise a \ref condition() condition is added.
   - the boolean constant "true" is translated by the polynomial (1), and "false" is translated in the polynomial (-1).
   - for a signal it calls translateSignal() function.
   
*/
ptpolynomial
SigaliService::translateExpression (ptexpcal e)
{
  ptexpcal e1, e2;
  noperators op;
  ptexpcal he1, he2;
  ptinsth h1, h2, hcontr;
  ptpolynomial p1, p2;

  if (e == NILEXPCAL)
    return NILPOLYNOMIAL;

  switch ((int) (EM->getOperatorClass (e)))
    {
    case variabl:
      return translateSignal (e);

    case constant:
      if (EM->getValue (e) == 1)
	return polyconstant (1);
      return polyconstant (-1);

    case oper1:
      op = EM->getOperator (e);
      if (op == opvvarin)
	return translateExpression (EM->getUnaryArg (e));
      if (op == opnon)
	return translateNot (EM->getUnaryArg (e));
      return condition (e);

    case oper2:
      EM->getBinaryArgs (e, &e1, &e2);
      op = EM->getOperator (e);
      if (op == opet)
	return translateAnd (e1, e2);
      if (op == opou)
	return translateOr (e1, e2);
      if ((op == opegal) && EM->IsBoolean (e1))
	return translateEqual (e1, e2);
      if ((op == opdiff) && EM->IsBoolean (e1))
	return translateDiff (e1, e2);
      return condition (e);

    case opereh:
      e1 = EM->getWhenOrCellArg (e);
      hcontr = CKM->getRepresentativeFormula (EM->getContributionClock (e));
      return translatewhen (e1, hcontr);

    case oper2tp:
      return translateDefault (e);

    default:			// oper1tp, oper3, oper3tp.
      ;
    }
  return NILPOLYNOMIAL;
}

/**
   Translating a hierarchy with \e hr as root. It realizes a "left to right topdown" traversal of the hierarchy.
   More precisely,
   - for a \ref isUsefulSignal() "useful" \b boolean signal with \e hr as clock, it generates the associated polynomial
   - it generates the constraints associated with the clock (equivalent formula),
   - it recalls recursively the method for each child.
*/
void
SigaliService::translateHierarchy (ptinsth hr)
{
  ptlinstances lsig;
  ptexpcal vsig;
  ptlinsth lClocks;
  ptinsth horl;
  ptpolynomial poly;

  if (isSelectedClock (hr))
    if (!isClockOfOnlySigaliFunctions (hr))
      {
	poly = translateClock (hr);
	lsig = CKM->getSynchronousSignals (hr);
	while (ptlinstances_get_next (&lsig, &vsig))
	  if (!isOutputSigaliIntrinsic (vsig))
	    {
	      if (expression_get_z3mark (vsig) == z3z_undefined)
		if (isUsefulSignal (vsig))
		  genPolynomialForSignal (vsig);
	    }
	// constraints on h : equivalent formula are added as constraints.
	horl = CKM->getNextFormula (hr, pk_true);
	while (horl != hr)
	  {
	    addConstraint (translateClock (horl), poly);
	    horl = CKM->getNextFormula (horl, pk_true);
	  }
	// Recursively, the hierarchy is visited.
	lClocks = CKM->getChildren (hr);
	while (ptlinsth_get_next (&lClocks, &horl))
	  translateHierarchy (horl);
      }
}

/**  It translates the constraints on non root clocks.
     If two formula of the same clock (a clock is a set of equivalent formula) have been placed in the different 
     node of the hierarchy (ie they have not the same canonical form), the constrainst is kept in some lists.
     This method explores the list and generates the associated constraint in Z3Z. 
*/
void
SigaliService::translateConstraintsOnPlacedClocks (void)
{
  ptlinsth lh1, lh2;
  ptinsth h1, h2;

  CKM->getConstraintsOnPlacedClocks (&lh1, &lh2);
  while (ptlinsth_get_next (&lh1, &h1))
    {
      ptlinsth_get_next (&lh2, &h2);
      if (h1 != h2)
	addConstraint (translateClock (h1), translateClock (h2));
    }
}


/**
   It translates the constraints on null clock.
*/
void
SigaliService::translateNullClockConstraints ()
{
  ptlinsth lh;
  ptinsth h;
  ptpolynomial nullClock;

  if (vDCGSigali->getNullClockConstraints (&lh))
    {
      nullClock = polyconstant (0);
      while (ptlinsth_get_next (&lh, &h))
	addConstraint (translateClock (h), nullClock);
    }
}


/**
   It translates the constraints on null clock.
*/
void
SigaliService::translateNullClockAssertions ()
{
  ptlinsth lh;
  ptinsth h;
  ptpolynomial nullClock;

  //  return ; // demander à Hervé.
  if (CKM->getNullClockAssertions(&lh))
    {
      nullClock = polyconstant (0);
      while (ptlinsth_get_next (&lh, &h))
	addConstraint (translateClock (h), nullClock);
    }
}



/**
   For each element \f$ (x, a, ^a, b0)  \f$ of  a list of delays (list), it generates
   the equations that expresses the evolution of the system \f$ x' = a + (1-a^{2}) x, x0= b0  \f$
*/
void
SigaliService::translatelistDelays (ptlrepDelay list)
{
  ptrepDelay retard;
  ptexpcal vsig;
  ptinsth hsig;
  ptpolynomial ksi, pinit, psig, pevol;

  initIteratorlDelay (list);
  while (!isEndlDelay (list))
    {
      retard = nextDelay (list);
      ksi = getDelaystatevar (retard);
      vsig = getDelaySig (retard);
      hsig = getDelayClock (retard);
      pinit = getDelayInit (retard);
      if (pinit != NILPOLYNOMIAL)
	addpolynomial (listinit, polyequa (ksi, pinit));
      else
	addpolynomial (listinit, polyequa (polysquare (ksi), polyconstant (1)));
      // psig = translatewhen(vsig, hsig);
      // September 2009 BUG 
      //   x := ( e when hsig) default x$  
      //   it is necessary to generate (e * hsig) + (1-hsig) * x
      psig = polymult (translateExpression (vsig), translateClock (hsig));
      pevol = polysum (psig, polymult (translatehnot (hsig), ksi));
      addpolynomial (listevolutions, pevol);
    }
}



/**
   It declares a new state variables when a memory is detected (ie zzx := zx $ 1  init x0 | zx  := x default zzx and hx=^x)
*/
void
SigaliService::stateDefinition (ptexpcal zzx, ptexpcal zx, ptexpcal x, ptinsth hx)
{
  ptpolynomial etat;
  ptident id;
  boolean bSignalStateVar;

  bSignalStateVar = EM->isNILEXPCAL (zx);

  if (bSignalStateVar)
    id = EM->getName (zzx);
  else
    id = EM->getName (zx);
  etat = declareStateVariable (id, x, hx, translateExpression (EM->getInitialValue (zzx)));
  if (!bSignalStateVar)
    {
      expression_setPolynomial (zx, etat);
      expression_set_z3mark (zx, z3z_defined);
    }

  /* Marked as "defined" with the associated polynomial. */
  expression_setPolynomial (zzx, etat);
  expression_set_z3mark (zzx, z3z_defined);
}

/**
   @return 
   - "true" if a delay node (delay) satisfies the \e memory following definition (simulator)
   zzx := zx $ 1  init x0 | zx  := x default zzx, and x is a signal (not and expression)
   - "false" otherwise.
   When it is a memory, the same "sigali" object is associated with zx and zzx.
*/
boolean
SigaliService::translateDelayNode (ptINode delay)
{
  ptINode evol_instr;
  ptinsth hx;
  ptexpcal x, zx, zzx, oprdef, evolution, vret, vfen, eh, vdedef, vinit;
  cloper clop;

  /* zzx = zx $ 1 */
  NM->get_delay_arg (delay, &zzx, &zx, &vret, &vfen, &vinit);
  if (!EM->IsBoolean (zx))
    return pk_false;
  if (!EM->IsSignal (zx))
    return pk_false;
  evol_instr = EM->getNode (zx);
  if (NM->get_oper (evol_instr) != opdefsig)
    return pk_false;
  if (!NM->get_rightOperand (evol_instr, &vdedef))
    return pk_false;
  if (EM->IsCellExpr (vdedef))
    return pk_false;
  evolution = EM->getWhenOrCellArg (vdedef);

  /* cas traites zx := x default zzx, zx := constante default zzx, zx := x , zx := constante */
  if (EM->getOperator (evolution) != opmele)
    return pk_false;
  EM->getDefaultArgs (evolution, &eh, &oprdef);
  if (zzx != EM->getWhenOrCellArg (oprdef))
    return pk_false;
  //  if (EM->getOperatorClass(eh) != opereh) return pk_false;
  clop = EM->getOperatorClass (eh);
  if (clop == opereh)
    x = EM->getWhenOrCellArg (eh);
  else if (clop == oper2tp)
    x = eh;
  else
    return pk_false;

  hx = CKM->getRepresentativeFormula (EM->getContributionClock (eh));
  stateDefinition (zzx, zx, x, hx);
  return pk_true;
}

/**
   @return "true" if an signal (id) is a Signal "statevar" identifier defined by an equation, "false" otherwise.
   Similar to \ref translateDelayNode() method.
*/
boolean
SigaliService::translateStateVariable (ptexpcal id)
{
  ptINode node, n;
  ptlINode ln;
  ptexpcal curDef, x;
  ptinsth hx;

  node = EM->getNode (id);
  if (!NM->is_setOfPartialDefinitions (node))
    return pk_false;
  ln = NM->get_setOfPartialDefinitions_arg (node);
  if (ptlINode_get_size (ln) > 1)
    return pk_false;
  ptlINode_get_elem (&ln, &n);
  if (NM->get_rightOperand (n, &curDef))
    {
      x = EM->getWhenOrCellArg (curDef);
      hx = EM->getContributionClock (curDef);
    }
  stateDefinition (id, NILEXPCAL, x, hx);
  return pk_true;
}

/**
   It translates Signal \ref translateDelayNode() "state variables", and \ref translateStateVariable() "signals declared as statevar".
   It is used for the Sigali Simulator that imposes the following syntactic form \n
   zzx := zx $ 1  init x0 | zx  := x default zzx, and x is a signal (not and expression)
   to express the state variables.
*/
void
SigaliService::translateldelays ()
{
  ptlINode ldelays;
  ptINode aDelay, node;
  ptlinstances lids;
  ptexpcal id;

  if (NM->getDelays (&ldelays))
    while (ptlINode_get_elem (&ldelays, &aDelay))
      if (isSelectedNode (aDelay))
	if (!translateDelayNode (aDelay))
	  notStandard = pk_true;
  // State variables.
  if (EM->getExprList (opident, &lids))
    while (ptlinstances_get_next (&lids, &id))
      if (EM->isStateVar (id))
	if (isSelectedNode (EM->getNode (id)))
	  if (!translateStateVariable (id))
	    sendNotimplemented (string2repstr ("StateVar Reference"), id);
}

/** 
    Translating a program:
    - It \ref translateldelays() "translates the state variables of the program" (special syntatic form, used by the Sigali simulator),
    - It \ref translateHierarchy() "translates the hierarchy"  associated to each real root,
    - It \ref translateConstraintsOnPlacedClocks() "translates the constraints on equivalent placed formula",
    - It \ref translateNullClockConstraints() "translates the constraints on null clock",
    - It \ref translatelistDelays() "generates the dynamical informations" from the collected informations during the previous operations
    (managed in the global list called \e listDelays).
*/
void
SigaliService::translateProgram ()
{
  ptlinsth lClocks;
  ptinsth h, htop;

  translateldelays ();

  htop = vDCGSigali->getRootClock ();
  lClocks = CKM->getChildren (htop);
  while (ptlinsth_get_next (&lClocks, &h))
    translateHierarchy (h);
  translateConstraintsOnPlacedClocks ();
  translateNullClockConstraints ();
  translateNullClockAssertions ();
  translatelistDelays (listDelays);
}


/** It initializes the global lists: listDelays,listvar, liststatevar, listcontrolables,  
    listfreeconditions, listconditions,listconstraints, listevolutions, listinit, listdef.  */
void
SigaliService::initlists (void)
{
  listDelays = makeDelayList ();
  listvar = makePolyList ();
  liststatevar = makePolyList ();
  listcontrolables = makePolyList ();
  listfreeconditions = makePolyList ();
  listconditions = makePolyList ();
  listconstraints = makePolyList ();
  listevolutions = makePolyList ();
  listinit = makePolyList ();
  listdef = makeDefinitionList ();
}


/** 
    Initialisation: it creates some specific identifiers (opident(state,nil) and opident(cond,nil))
    and the associated counters (numetat, numcond).
    The translator creates state variables and conditions for booleens built on non-boolean expressions.
*/
void
SigaliService::initident (void)
{
  numetat = 0;
  numcond = 0;
  prefixeetat = string2ident (cstretat);
  prefixecond = string2ident (cstrcond);
}

/**
   Initialisation: it creates global names associated with the different sigali lists (states, events, evolutions, initialisations,...)
*/
void
SigaliService::initstr ()
{
  vstrletat = string2repstr (cstrletat);
  vstrlcond = string2repstr (cstrlcond);
  vstrlcont = string2repstr (cstrlcont);
  vstrlcondlibres = string2repstr (cstrlcondlibres);
  vstrlvar = string2repstr (cstrlvar);
  vstrdecl = string2repstr (cstrdecl);
  vstrlcontr = string2repstr (cstrlcontr);
  vstrlevol = string2repstr (cstrlevol);
  vstrlinit = string2repstr (cstrlinit);
}

/**
   It initializes the Sigali translator for a Signal graph (g).
*/
void
SigaliService::initialize ()
{
  // about the DCG graph.
  EM = vDCGSigali->getExpressionManager ();
  CKM = vDCGSigali->getClockManager ();
  NM = vDCGSigali->getNodeManager ();
  unNodes = vDCGSigali->getUnused (&unClocks);	/* unused objects */
  ProcessNameStr = ident2repstr (vDCGSigali->getName ());

  attributeNumber = getMark ();
  freeConditionAttr = getMark ();	/* internal attribute */
  refInSigaliCmd = getMark ();
  notStandard = pk_false;
  //  GdelayOne = pk_true;
  initconstants ();
  initlists ();
  initident ();
  initstr ();
  vstrsigali = string2ident (SIGALI_FUNC_NAME);
  vstrcontrollable = string2ident (CONTROLLABLE_FUNC_NAME);

  shaper = new z3zShaper ();
  lproperties = getldynamicproperties ();
  SigaliPragma = getPredefinedPragma (PRAGMA_SIGALI);
}


/**
   It finalizes the functionality by cleaning the graph.
   The internal attributes are removed.
*/
void
SigaliService::finalyze ()
{
  EM->resetMarkAll (attributeNumber);
  CKM->resetMarkAll (attributeNumber);
  EM->resetMarkAll (freeConditionAttr);
  if (lproperties != NILLINODE)
    EM->resetMarkAll (refInSigaliCmd);
}


/**
   It prints the text associated to a DCG into a file (name).
*/
void
SigaliService::unparse (ptrepstr name)
{
  if (shaper->open (name))
    {
      shaper->unparse_declaration (listvar, vstrdecl);
      shaper->unparse_listdef (listdef);
      shaper->unparse_listpoly (listconditions, vstrlcond);
      shaper->unparse_listpoly (liststatevar, vstrletat);
      shaper->unparse_listpoly (listinit, vstrlinit);
      shaper->unparse_listpoly (listevolutions, vstrlevol);
      shaper->unparse_listpoly (listconstraints, vstrlcontr);
      shaper->unparse_listpoly (listcontrolables, vstrlcont);
      shaper->unparse_listpoly (listfreeconditions, vstrlcondlibres);
      shaper->close ();
    }
}
