/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/* Add constraint for part definition.
   assumes : ln is a set of part definition nodes, aNode is in ln 
   for  aNode defined by  ::= E, and F the definition of one of the nodes of ln the 
   following constraint is appended :
   E ^* F = when (E when ^F = F when ^E)
   @return none.
*/
  void
intemkIGraph::addPartDefinitionConstraint (ptINode aNode, ptlINode ln)
{
  ptINode n;
  ptinsth h, h1, p, hsub;
  ptexpcal E, F, s1, s2, E1, F1, exp;
  noperators op;
  TypeType t;

  h1 = NM->get_nodeClock (aNode);
  NM->get_rightOperand (aNode, &E);
  E1 = EM->getBasicExpr (E);
  while (ptlINode_get_elem (&ln, &n))
    if (n != aNode)
      if (!NM->is_partialDefinitionDefaultValue (n))
	{
	  NM->get_rightOperand (n, &F);
	  F1 = EM->getBasicExpr (F);
	  if (!EM->isEquivalent (E1, F1, pk_false))
	    {
	      h = NM->get_nodeClock (n);
	      p = CKM->makeInf (h, h1);
	      if (EM->TestExclusiveConstants (E1, F1))
		{
		  p = CKM->makeSynchro (p, the_nullClock_GB_copy);
		}
	      else
		{
		  s1 = NM->createWhen (E, p, pk_false);
		  s2 = NM->createWhen (F, p, pk_false);
		  t = EM->getUnamedType (s1);
		  if (TypeTestPredef (t))
		    op = opegal;
		  else
		    op = opegeg;
		  exp = EM->mkBinaryExp (op, s1, s2, NILBACKANNOTATION);
		  if (exp != EM->getTrue ())
		    {
		      if (exp == EM->getFalse ())
			hsub = the_nullClock_GB_copy;
		      else
			hsub = CKM->makeSub (p, exp, EM->getPredecessors (exp));
		      p = CKM->makeSynchro (p, hsub);
		    }
		}
	    }
	}
}

/**
   Unused 
*/
void
intemkIGraph::addPartDefinitionConstraints ()
{
  ptlINode lNodes, ln;
  ptINode node, n;

  // lNodes : set of nodes defined by a set of ::= nodes
  if (NM->getGroupsOfPartialEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      {
	ln = NM->get_setOfPartialDefinitions_arg (node);
	while (ptlINode_get_elem (&ln, &n))
	  addPartDefinitionConstraint (n, ln);
      }
}

/**
   @return "true" is the signal (exp) is CURRENTLY defined by a partial defintion.
*/
boolean intemkIGraph::isPartialDefinitionOf (ptexpcal exp)
{
  return EM->isMarked (exp, PART_DEFINITION);
}


/**
   It propagates the type of the left part to the right part for equation and delay nodes.
   (See ExprManager::TypePropagate() method)
*/
boolean intemkIGraph::propagateTypes ()
{
  ptlINode lNodes;
  ptINode node;
  ptexpcal expg, expd;

  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      {
	expg = NM->get_leftOperand (node);
	if (NM->get_rightOperand (node, &expd))
	  EM->TypePropagate (EM->getType (expg), expd);
      }
  if (NM->getDelays (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      {
	expg = NM->get_leftOperand (node);
	if (NM->get_rightOperand (node, &expd))
	  EM->TypePropagate (EM->getType (expg), expd);
      }
  return pK_messenger->getErrorCode ();
}





/** 
   Specific to the bundle type.
   Each field becomes an input (the interface of the graph in NOT modified)
   @param s (in) assumed to be a signal identifier expression.
   @return true if it is done, false otherwise.
   Note: used for bundled input Signals. */
boolean intemkIGraph::fieldsInputs (ptexpcal s)
{
  ptlinstances lsig;
  ptINode n, n1, m;
  boolean err;
  ptexpcal aSig;

  n = EM->getNode (s);
  if (NM->is_undefined (n))
    return pk_false;
  if (!TypeTestBundleClass (EM->getUnamedType (s)))
    return pk_false;
  // Bundle
  lsig = mkFieldsDescription (s);
  //  if (!NM->is_lfields(n)) return pk_false; lsig = NM->get_lfields_arg(n);
  // Bundle : each field-signal are defined as inputs signals (internal representation)
  // The interface of the graph in NOT modified
  err = pk_false;
  while (ptlinstances_get_next (&lsig, &aSig))
    if (!fieldsInputs (aSig))
      {
	n1 = EM->getNode (aSig);
	// if (!NM->is_undefined(n1)) err = pk_true;
	addInNode (aSig, n, &m);
      }
  // A virer apres tests
  // if (err)  pK_messenger->messagestr(internalError, string2repstr("Inputs/bundles ***"));

  return pk_true;
}



/** 
   Specific to the bundle type.
   Each field becomes an output (the interface of the graph in NOT modified)
   @param s (in) assumed to be a signal identifier expression.
   @return true if it is done, false otherwise.
   Note : used for bundled output Signals. */
boolean intemkIGraph::fieldsOutputs (ptexpcal s)
{
  ptlinstances lsig;
  ptexpcal aSig;

  if (!TypeTestBundleClass (EM->getUnamedType (s)))
    return pk_false;
  if (!EM->getFieldsOf (s, &lsig))
    return pk_false;
  while (ptlinstances_get_next (&lsig, &aSig))
    if (!fieldsOutputs (aSig))
      addOutNode (aSig, NILINODE);
  return pk_true;
}


/**
   This method is provided for applying particular action for specific class of graph (function, process, node,...).
*/
void
intemkIGraph::graphClassSpecificity ()
{
  noperators op;
  ptlinstances lins, louts;

  op = AGRAPH->getKind ();
  if ((op == opfunc) || (op == opfuncp))
    {
      lins = AGRAPH->getInputs ();
      louts = AGRAPH->getOutputs ();
      NM->addDependences (lins, louts, hundef, pk_false);
    }
}


/**
   @return "true" if an expression (vDelay) is an integer valued expression (v), "false" otherwise.
 */
boolean intemkIGraph::getExpressionValue (ptexpcal vDelay, integer * v)
{
  TypeType TT;
  if (vDelay == NILEXPCAL)
    {
      *v = 0;
      return pk_true;
    }
  if (!EM->IsAssigned (vDelay))
    return pk_false;
  // Blindage
  TT = EM->getType (vDelay);
  if (!TypeTestPredef (TT))
    return pk_false;
  if (!TypeTestIntNumber (TT))
    return pk_false;
  *v = EM->getValue (vDelay);
  return pk_true;
}


/**
   It sets the clock (h) as the clock of the default partial definition (node).
*/
void
intemkIGraph::setPartialDefDefaultValueClock (ptINode node, ptinsth h)
{
  ptexpcal vopg, vdedef, exp, nexp;
  ptinsth hc;

  if (NM->get_oper (node) != opdefpart)
    return;			// no effect
  if (!NM->get_rightOperand (node, &vdedef))
    return;
  vopg = NM->get_leftOperand (node);
  hc = EM->getContributionClock (vdedef);
  NM->removeDependence (hc, node);
  exp = EM->getWhenOrCellArg (vdedef);
  nexp = EM->mkWhenOrCellExp (opextr, exp, h);
  NM->set_def (node, NILEXPCAL);
  NM->becomes_partialDefinition (node, vopg, nexp);
}

/**
   It finalizes the definition of a signal defined by a set of partial definitions (node). 
   In particular, it defines the clock of the default devalue ( see setPartialDefDefaultValueClock()).
*/
void
intemkIGraph::signalpartialDefinitionFinalize (ptINode node)
{
  ptlINode ln;
  ptINode n, defaultNode;
  ptexpcal s, s1;
  ptinsth h, h1, hs, h2, DefaultExpClock;
  boolean bDefaultVal, bDefaultValCur, bMultipleDefaultValue;
  ntrees tdecl;
  ptident id;

  bDefaultVal = pk_false;
  bMultipleDefaultValue = pk_false;
  s = NM->get_leftOperand (node);
  hs = EM->getClock (s);
  h1 = hundef;
  DefaultExpClock = hundef;
  ln = NM->get_setOfPartialDefinitions_arg (node);
  while (ptlINode_get_elem (&ln, &n))
    {				// n is a part definition node, one must be the defaultValue.
      bDefaultValCur = NM->is_partialDefinitionDefaultValue (n);
      if (bDefaultValCur)
	{
	  if (bDefaultVal)
	    {
	      bMultipleDefaultValue = pk_true;
	      bDefaultValCur = pk_false;
	    }
	  else
	    bDefaultVal = pk_true;
	}

      h = NM->get_nodeClock (n);
      if (bDefaultValCur)
	{
	  DefaultExpClock = h;
	  defaultNode = n;
	}
      else
	{
	  addPartDefinitionConstraint (n, ln);
	  h1 = CKM->makeSup (h1, h);
	}
    }

  if (EM->getAttribute (s, PREVIOUSVALUEATTR, (attributeValue *) & s1))
    {
      hs = CKM->makeSynchro (hs, h1);
    }
  else
    {
      tdecl = EM->getTreeOfSignalDeclaration (s);
      id = EM->getOriginalName (s);
      if (bMultipleDefaultValue)
	pK_messenger->errorTreeIdent (errMultipleDefaultValue, tdecl, id);
      // if h1 is undefined clock, no definitions, only the default value ?
      if (CKM->satisfiesUndefClock (h1))
	{
	  pK_messenger->warningaccnom (warNoDefinition, tdecl, id);
	  h = hs;
	  h1 = DefaultExpClock;
	}
      else
	{
	  h2 = CKM->makeSup (h1, hs);
	  hs = CKM->makeSynchro (hs, h2);
	  h = CKM->makeMinus (hs, h1);
	}

      if (CKM->satisfiesUndefClock (DefaultExpClock))
	pK_messenger->warningaccnom (warNoDefaultValue, tdecl, id);
      else
	{
	  h = CKM->makeInf (DefaultExpClock, h);
	  setPartialDefDefaultValueClock (defaultNode, h);
	  NM->moveNodeToClock (defaultNode, h);
	  h2 = CKM->makeSup (h1, DefaultExpClock);
	  h = CKM->makeMinus (hs, h2);
	}

      if (!CKM->satisfiesNullClock (h))
	NM->addPartAff (s, s, h, ptlinstances_push_front (NILLINSTANCES, s));
    }
}

/**
   Automaton.
   @return true if a clock(h) is a test on a state variable (vstate), false otherwise.
   Example [:(_CurState?)==StateType#Q]
 */
boolean  intemkIGraph::getInStateClock(ptinsth h, ptexpcal vstate)
{
  ptexpcal vcond, arg1, arg2;
  noperators vop;
  
  if (! CKM->satisfiesSub(h)) return pk_false;
  vcond = CKM->getConditionSub(h);
  vop = EM->getOperator(vcond);
  if ( ( vop != opegal) && (vop != opegeg)) return pk_false;
  EM->getBinaryArgs(vcond, &arg1, &arg2);
  if ( EM->getOperator(arg1) == opvvarin)
    if (EM->getUnaryArg(arg1) == vstate) return pk_true;
  if ( EM->getOperator(arg2) == opvvarin)
    if (EM->getUnaryArg(arg1) == vstate) return pk_true; 
  return  pk_false;
}

/**
   Automaton.
   Assumed: the operand node (n) is a definition of the next value of a state variable (ie S::= StateType#Q when (S?)==#Q) when Trigger)
   @return the Trigger clock.
 */
ptinsth intemkIGraph::getTrigger(ptINode n)
{
  ptinsth  h, vharg1, vharg2;
  ptexpcal vstate;

  vstate = NM->get_leftOperand (n);
  h = NM->get_contributionClockEquation (n);
  if (! CKM->getInfArgs (h, &vharg1, &vharg2)) return h;
  if (getInStateClock(vharg1, vstate )) return vharg2;
  if (getInStateClock(vharg2, vstate )) return vharg1;
  return h;
}


/**
   Automaton.
   Assumed: the operand node (n) is a definition of the next value of a state variable (ie S::= StateType#Q when (S?)==#Q) when Trigger)
   @return the clock that tests the current state of an automaton ( ie when (S?)==#Q))
*/
boolean intemkIGraph::getInStateClock(ptINode n, ptinsth * InStateClock)
{
  ptinsth  h, vharg1, vharg2;
  ptexpcal vstate;

  vstate = NM->get_leftOperand (n);
  h = NM->get_contributionClockEquation (n);
  if (! CKM->getInfArgs (h, &vharg1, &vharg2)) return pk_false;
  if (getInStateClock(vharg1, vstate )) { *InStateClock = vharg1; return pk_true;}
  if (getInStateClock(vharg2, vstate )) { *InStateClock = vharg2; return pk_true;}
  return pk_false;
}

/**
   Automaton.
   Assumed: The priority attribute PRIORITY is assigned to the operand node (n).
   @return the list l in which the node n is inserted by decreasing priority.
 */
ptlINode intemkIGraph::insert_elem_by_priority (ptlINode l, ptINode  n)
{
  integer vprio, vp;
  ptINode node;
  ptlINode lnodes;

  NM->getAttribute(n, PRIORITY, (attributeValue *) & vprio);
  lnodes = l;
  while (ptlINode_get_elem (&lnodes, &node))
    {
      NM->getAttribute(node, PRIORITY, (attributeValue *) & vp);
      if ( vprio >= vp )
	return ptlINode_append_elem_before(l, n, node);
    }
  return ptlINode_append_elem(l, n);
}

/**
   Automaton.
   @return true if all the partial definitions (lnodes) have the same priority, false otherwise.
 */
boolean intemkIGraph::haveSamePriority(ptlINode lnodes)
{
  integer vprio, vp;
  ptINode n;

  ptlINode_get_elem (&lnodes, &n);
  NM->getAttribute(n, PRIORITY, (attributeValue *) & vprio);
  while (ptlINode_get_elem (&lnodes, &n))
    {
      NM->getAttribute(n, PRIORITY, (attributeValue *) & vp);
      if (vprio != vp) return pk_false;
    }
  return pk_true;
}



/** Automaton.
   Assumed: The nodes of the operand (lnodes) are sorted by decreasing priority. 
            These nodes are the definitions of the transitions from a state of aan automaton.

   This method modifies the partial definitions to take into account the priority assigned to the transitions of the automaton.
   Example:
   \code
     (| Automaton_Transition (Q, R, a, 8)
      | Automaton_Transition (Q, S, b, 7)
      |)
   \endcode
   produces the following definitions (VS is an added state variable):
   \code
     VS ::= #R when (VS? = #Q) when a
     VS ::= #S when (VS? = #Q) when (b ^- a)
   \endcode
 */
void  intemkIGraph::TransitionPriority(ptlINode lnodes)
{
  ptlINode l,ln1;
  ptINode n, vn;
  integer nbTransitions, vp, vprio;
  ptinsth hvn, h;
  boolean baddConstraints;

  nbTransitions = ptlINode_get_size(lnodes);
  if (nbTransitions > 1)
    {
      baddConstraints = pk_false;
      if (haveSamePriority(lnodes)) 
	baddConstraints = pk_true;
      else
	{
	  l = lnodes;
	  while (ptlINode_get_elem (&l, &n))
	    {
	      NM->getAttribute(n, PRIORITY, (attributeValue *) & vprio);
	      h = getTrigger(n);
	      ln1 =l;
	      while (ptlINode_get_elem (&ln1, &vn))
		{
		  NM->getAttribute(vn, PRIORITY, (attributeValue *) & vp);
		  if (vp != vprio)
		    {
		      hvn = NM->get_contributionClockEquation (vn);
		      NM->replace(vn, hvn, CKM->makeMinus(hvn, h));
		    }
		  else
		    baddConstraints = pk_true;
		}
	    }
	}
      if (baddConstraints)
	{
	  while (ptlINode_get_elem (&lnodes, &n))
	    addPartDefinitionConstraint (n, lnodes);
	}
    }
}

/** Automaton.
   Assumed: the operand node (node) is defined by a set of partial definitions (x::= E1, ... x::=En).
   @return true if such nodes are the nodes produced by the compiling of the Automaton_Transition(S,Q,Trigger[, priority]) intrinsics, false otherwise. 

   This method modifies the partial definitions to take into account the priority assigned to the transitions of an automaton.
 */
boolean intemkIGraph::automatonTransition(ptINode node)
{
  // Assumed: automaton transitions are characterized by the PRIORITY attribute.
  typedef std::map<ptinsth, ptlINode> TransitionsTableType;
  ptlINode ln, ln1;
  ptINode n;
  integer vprio;
  ptinsth h;
  TransitionsTableType TransitionsTable;
  TransitionsTableType::iterator it;

  ln = NM->get_setOfPartialDefinitions_arg (node);
  ln1 = ln; 
  if (!ptlINode_get_elem (&ln1, &n)) return pk_false;
  if (! NM->getAttribute(n, PRIORITY, (attributeValue *) & vprio)) return pk_false;
  // sorting the nodes by state, 
  while (ptlINode_get_elem (&ln, &n))
    {
      if (! getInStateClock(n, &h)) return pk_false; // error
      TransitionsTable[h] = insert_elem_by_priority (TransitionsTable[h], n);
    }

  for(it=TransitionsTable.begin(); it != TransitionsTable.end() ; ++it)
    {
      TransitionPriority(it->second);
    }
  return pk_true;
}

/**
   It finalizes the definition of a state variable defined by a set of partial definitions (node).
*/
void
intemkIGraph::statepartialDefinitionFinalize (ptINode node)
{
  ptlINode ln;
  ptINode n;
  ptexpcal s;
  ptinsth hs, fictifClock;
  boolean bErrorState;

  fictifClock = CKM->getFictitiousRoot ();
  bErrorState = pk_false;
  s = NM->get_leftOperand (node);
  hs = EM->getClock (s);
  if (! automatonTransition(node))
    
    { 
      ln = NM->get_setOfPartialDefinitions_arg (node);
      while (ptlINode_get_elem (&ln, &n))
	{				// n is a part definition node, defaultValue not allowed for state variable.
	  if (NM->is_partialDefinitionDefaultValue (n))
	    bErrorState = pk_true;
	  addPartDefinitionConstraint (n, ln);
	}
    }
  if (bErrorState)
    pK_messenger->errorTreeIdent (errStateVar, EM->getTreeOfSignalDeclaration (s), EM->getOriginalName (s));
  // hs should be undefined or fictitious
  if (CKM->satisfiesUndefClock (hs))
    EM->setClock (s, fictifClock);
  else
    // Debugging
    if (!CKM->satisfiesFictitiousRoot (hs))
      {
	pK_messenger->printString (" Debugging : a State variable is not free ");
	pK_messenger->error (errcompilerbug);
      }
}

/**
   @return "true" if it exists an array node in the stack of "label stack", "false" otherwise.
   @param n (out) the first array reached in the stack (from the top).
*/
boolean intemkIGraph::getStackArrayNode (ptINode * n)
{
  ptElemStack nodeLabel;
  ptStack l;

  l = labelsStack;
  while (stackTop (l, &nodeLabel))
    {
      *n = (ptINode) nodeLabel;
      if (NM->is_array (*n))
	return pk_true;
      stackPrevious (l, &l);
    }
  return pk_false;
}

/**
   @param n (in) assumed to be an array node.
   @param s (in) a signal.
   it adds s as an input of the node n.
*/
void
intemkIGraph::addInputArray (ptINode n, ptexpcal s)
{
  ptexpcal v;

  NM->get_rightOperand (n, &v);
  if (!EM->isNILEXPCAL (v))
    {
      if (!EM->satisfiesList (v))
	v = EM->mkNaryExp (oplesign, v, NILBACKANNOTATION);
      s = EM->mkNaryExp (oplesign, s, v, NILBACKANNOTATION);
    }
  NM->set_rightOperand (n, s);
}


/**   
   It adds a signal (s) as an input of all the arrays of the stack of "label stack".
*/
boolean intemkIGraph::addInputArrayNodes (ptexpcal s)
{
  ptElemStack nodeLabel;
  ptStack l;
  ptINode n;

  l = labelsStack;
  while (stackTop (l, &nodeLabel))
    {
      n = (ptINode) nodeLabel;
      if (NM->is_array (n))
	addInputArray (n, s);

      stackPrevious (l, &l);
    }
  return pk_true;
}

/**
   @return true if a signal is not defined, false otherwise.
   For a signal of strutured type, it is undefined if all the fields are undefined.
*/
boolean intemkIGraph::isUndefinedNode (ptexpcal aSig)
{
  ptINode node;
  ptlinstances lfields;
  ptexpcal aField;

  node = EM->getNode (aSig);
  if (!NM->is_undefined (node))
    return pk_false;
  if (!EM->getFieldsOf (aSig, &lfields))
    return pk_true;

  while (ptlinstances_get_next (&lfields, &aField))
    if (!isUndefinedNode (aField))
      return pk_false;
  return pk_true;
}

/**
   @return true if a signal is useful, false otherwise.
   It is useful when it is referenced in the graph.
*/
boolean intemkIGraph::isUseful (ptexpcal aSig)
{
  boolean b;
  ptINode node;
  ptinsth clk;
  ptlinstances lfields;
  ptexpcal aField;

  node = EM->getNode (aSig);
  b = EM->getExtractionLink (aSig, &clk) || NM->isReferenced (node) || (EM->getLuse (aSig) != NILLEXPCAL);
  if (b)
    return b;
  if (!EM->getFieldsOf (aSig, &lfields))
    return b;
  while (ptlinstances_get_next (&lfields, &aField))
    if (isUseful (aField))
      return pk_true;
  return pk_false;
}


/** Adding the definition x:= x for an undefined/referenced signals (aSig)
    @return true if the definition is added, false otherwise.
*/
boolean intemkIGraph::addDefinitionForUndefinedSignals (ptexpcal aSig)
{
  ptINode n;
  boolean added;
  ptinsth clk;

  added = pk_false;
  if (EM->IsSignal (aSig))
    if (isUndefinedNode (aSig) && isUseful (aSig))
      {
	clk = EM->getClock (aSig);
	if (EM->IsEvent (aSig))
	  n = NM->mk_node_equation (aSig, EM->getTrue (), clk, NILLINSTANCES, NILBACKANNOTATION);
	else
	  {
	    n = NM->mk_node_equation (aSig, aSig, clk, EM->getPredecessors (aSig), NILBACKANNOTATION);
	    if (the_nullClock_GB_copy != clk)
	      {
		added = pk_true;
		pK_messenger->warningaccnom (warNotDefined, EM->getTreeOfSignalDeclaration (aSig), EM->getOriginalName (aSig));
	      }
	  }
	NM->insertNode (n, clk);
      }
  return added;
}



/** Adding definition for undefined/referenced signals. 
    The added definition is of the form x:=x.
*/
boolean intemkIGraph::addDefinitionForUndefinedSignals ()
{
  ptlinstances lident;
  ptexpcal aSig;
  boolean added;

  if (AGRAPH->isGraphOfExternalProcess ())
    return pk_false;
  if (!EM->getExprList (opident, &lident))
    return pk_false;
  added = pk_false;
  while (ptlinstances_get_next (&lident, &aSig))
    if (addDefinitionForUndefinedSignals (aSig))
      added = pk_true;

  return added;
}


/**
   This method is without effect when vh is the predefined undefined clock.
   Otherwise, it adds, taking into account the different contexts, a node (n) that defines a signal (x)
   at a provided clock (vh).
*/
void
intemkIGraph::addNodeAt (ptINode n, ptexpcal x, ptinsth vh)
{
  ptINode vinstr;
  if (vh != hundef)
    {
      takeIntoAccountContexts (n);
      vinstr = EM->getNode (x);	// different of n for partial definition.
      if (NM->is_setOfPartialDefinitions (vinstr))
	vh = CKM->makeSup (vh, NM->get_nodeClock (vinstr));
      NM->insertNode (vinstr, vh);
      if (n != vinstr)
	takeIntoAccountContexts (vinstr);
    }
}

/**
   We distinguish several cases:
     - when it is a partial definition of a structured type: the expression is splitted and the method is recall on the splitted expressions.
     The following attributes are used or set in this case PART_DEFAULT_DEFINITION,PART_DEFINITION.
     
     - otherwise, the expression (expcal2 when vhcal) is added to the set of definitions of the signals (expcal1). And the 
       node is inserted in the graph if vh is not the predefined undefined clock (see addNodeAt()).
     
   @param expcal1 (in) the defined signal
   @param expcal2 (in) a synchronous expression component of the definition of the signal (expcal1)
   @param vhcal (in) the clock at which the expcal2 expression defined the signal (expcal1)
   @param instent (in) signals referenced in expcal2
   @param vh (in) the clock of expcal1 or the predefined undefined clock
   @param vtree (in) back attribute to parser abstract tree for error annotations.
*/
void
intemkIGraph::mkEquation (ptexpcal expcal1, ptexpcal expcal2, ptinsth vhcal, ptlinstances instent, ptinsth vh, BackAnnotationType vtree)
{
  ptINode n;
  ptexpcal x, e, vf;
  boolean bSplit;
  TypeType t;
  NameList NL;
  TypeList fieldsType;
  TypeSubList NSubLstType;
  boolean Goon, bpartdefault;
  NameSubList NSubLst;
  ptrepstr vstr;
  integer currentFieldNumber, vprio;

  x = mkFieldAccess (expcal1, vtree);
  t = EM->getUnamedType (x);
  bSplit = isPartialDefinitionOf (expcal1) && TypeTestStructureClass (t);
  if (!bSplit)
    {
      n = NM->mk_node_equation (x, expcal2, vhcal, instent, vtree);
      addNodeAt (n, x, vh);
      if (EM->getAttribute(expcal1, PRIORITY, (attributeValue *) & vprio))
	NM->addAttribute (n, PRIORITY, (attributeValue) vprio);
    }
  else
    {
      bpartdefault = EM->isMarked (expcal1, PART_DEFAULT_DEFINITION);
      // Voir si la suite peut etre adaptee pour utlisation des methodes de la classe splitting
      NL = StructureTypeGetNameLst (t);	// field Name
      NSubLst = NameListCast2SubList (NL);
      fieldsType = StructureTypeGetTypeLst (t);
      NSubLstType = TypeListCast2SubList (fieldsType);
      Goon = pk_true;
      currentFieldNumber = 1;
      while (Goon)
	{
	  vstr = NameSubListGetHead (NSubLst);
	  vf = EM->mkBinaryExp (opcham, x, EM->mkConstantExp (opnom, vstr), NILBACKANNOTATION);	// x.NAME
	  e = splittingObject->mkFieldOfSynchronousExpression (expcal2, currentFieldNumber);
	  instent = EM->getPredecessors (e);

	  EM->addMark (vf, PART_DEFINITION);
	  if (bpartdefault)
	    EM->addMark (vf, PART_DEFAULT_DEFINITION);
	  mkEquation (vf, e, vhcal, instent, vh, vtree);
	  EM->removeMark (vf, PART_DEFINITION);
	  if (bpartdefault)
	    EM->removeMark (vf, PART_DEFAULT_DEFINITION);

	  if (NameSubListTestLast (NSubLst))
	    Goon = pk_false;
	  else
	    {
	      NSubLst = NameSubListGetTail (NSubLst);
	      NSubLstType = TypeSubListGetTail (NSubLstType);
	      currentFieldNumber++;
	    }
	}

    }
}


/** @return "true" if an expression (exp) reference a state variable (x? operator), "false" otherwise.
    When it references a state variable exp1 is the operand expression (exp) where the x? expression is replaced by x.
    \warning It tests only the following expressions: x?  E[],E.field
*/
boolean intemkIGraph::mkExpressionWithoutReferenceToStateVar (ptexpcal exp, ptexpcal * exp1)
{
  noperators op;
  ptexpcal e1;

  op = EM->getOperator (exp);
  // ref to a state variable.
  if (op == opvvarin)
    {
      (*exp1) = EM->getUnaryArg (exp);
      return pk_true;
    }
  if ((op == opcham) || (op == opelem))
    if (mkExpressionWithoutReferenceToStateVar (EM->getBinaryArgstAt (exp, 1), &e1))
      {
	*exp1 = EM->mkBinaryExp (opcham, e1, EM->getBinaryArgstAt (exp, 2), NILBACKANNOTATION);
	return pk_true;
      }
  return pk_false;
}



/**
   @return the clock obtained by the synchronisation of the signals (vTuple) with a given clock (vClock).
*/
ptinsth intemkIGraph::mkSynchronizedTupleClock (ptexpcal vTuple, ptinsth vClock)
{
  ptexpcal vTupleElem;

  while (EM->getNaryArgs (&vTuple, &vTupleElem))
    vClock = getSignalClockSynchronizedWith (vTupleElem, vClock);
  return vClock;
}

/**
   It transforms the reference to the cartesian product operator to an external call.
   @return the clock of the expression.
*/
ptinsth intemkIGraph::mkCartesianProductEquation (ptexpcal vTuple, ptexpcal vDef, ptinsth vhcal, ptlinstances instent, ptinsth vh, BackAnnotationType vtree)
{
  ptINode n;

  if (!EM->satisfiesList (vTuple))
    vTuple = EM->mkNaryExp (oplesign, vTuple, NILBACKANNOTATION);
  vh = mkSynchronizedTupleClock (vTuple, vh);
  EM->TypeCheckingCartesianProductEquation (vTuple, vDef, vtree);
  mkProcCallNode (getCartesianProductId (), NILEXPCAL, vDef, vTuple, vh, opsafe, vtree, NILBACKANNOTATION, pk_true, &n);
  return vh;
}


/**
   It marks (using an internal mark myMark), the signals referenced in the clock system (conditions).
   This mark is used in updateSignalsNames() method.
*/
void
intemkIGraph::markSignalsReferencedInExtractions ()
{
  ptlinsth lh;
  ptinsth vh;
  ptexpcal velem;
  ptlinstances ls;

  lh = CKM->getFormulaList (sub);
  while (ptlinsth_get_next (&lh, &vh))
    {
      ls = CKM->getSignalsOfCondition (vh);
      while (ptlinstances_get_next (&ls, &velem))
	EM->addMark (velem, myMark);
    }
}

/**
   @return "true" if a node (n) is undefined, "false" otherwise.
   For a definition of a structure/bundle, it is undefined iff all the members are undefined, otherwise
   it is undefined iff it is the undefined node.
 */
boolean intemkIGraph::isRecUndefined (ptINode n)
{
  ptlinstances ls;
  ptexpcal s;
  ptINode m;

  if (NM->is_undefined (n))
    return pk_true;
  if (!NM->is_lfields (n))
    return pk_false;
  ls = NM->get_lfields_arg (n);
  while (ptlinstances_get_next (&ls, &s))
    {
      m = EM->getNode (s);
      if (m != n)
	if (!isRecUndefined (m))
	  return pk_false;
    }
  return pk_true;
}

/**
   Specific for an assertion : no node in the graph. It appends the constraints in the <em> clock system </em> :
   Recall : assert{C} specified that the constraint when C = ^C is satisfied.
*/
boolean intemkIGraph::AssertionCallNode (ptident modelName, ptexpcal lentren, ptinsth vh)
{
  ptexpcal vcond;
  boolean bpos;
  ptinsth vhsub, vhnosub;

  if (!isAssertionId (modelName))
    return pk_false;
  if (EM->getNaryArgs (&lentren, &vcond))
    {
      bpos = (EM->getOperator (vcond) != opnon);
      if (!bpos)
	vcond = EM->getUnaryArg (vcond);
      CKM->makeAssert (vh, vcond, bpos, &vhsub, &vhnosub);
      takeIntoAccountContexts (vhsub);
      takeIntoAccountContexts (vhnosub);
    }
  return pk_true;
}


/** 
    @return 
     - the operand (exp) if it is an identifier,
     - the identifier referenced in the left part of a field access.
    For a field access (x.f), it creates  a signal for each field of the (x) structure (see mkFieldsDescription ()).
    If the referenced field (f) is not a field of the struture, an error is signaled but it adds one to continue.
    A back attribute to parser abstract tree (vt) can be provided for error annotations.
    Assumed: exp can be an identifier or an access to a field of a strucutre (not a reference to state variable (x? operator)).

*/
ptexpcal intemkIGraph::getSetField (ptexpcal exp, BackAnnotationType vt)
{
  ptINode n;
  ptexpcal e1, e2, exp1;
  ptident id;
  boolean b;
  ptlinstances ls;

  if (EM->getOperatorClass (exp) == variabl)
    return exp;
  e1 = EM->getBinaryArgstAt (exp, 1);
  exp1 = getSetField (e1, vt);
  n = EM->getNode (exp1);
  b = NM->is_undefined (n);
  if (b)
    {
      ls = splittingObject->mkFieldsDescription (e1);
      EM->setFieldsOf (exp1, ls);
      n = NM->mk_node_lfields (exp1, n, ls, vt);
    }
  if (EM->getSignalAssignedTo (exp, &exp1))
    return exp1;
  if (!b)
    {
      ls = splittingObject->mkFieldsDescription (e1);
      EM->setFieldsOf (exp1, ls);
      NM->nodesMerging (ls, n);
      if (EM->getSignalAssignedTo (exp, &exp1))
	return exp1;
    }
  // The field does not exist : adding one to continue
  pK_messenger->errorTree (errUnexpectedExpr, BACK_ANNOT_GET_ABSTRACT_TREE(vt));
  e2 = EM->getBinaryArgstAt (exp, 2);
  if (EM->getOperatorClass (e2) == constant)
    id = repstr2ident (EM->getConstant (e2));
  else
    id = getIdentNil ();
  exp1 = EM->mkIdentField (id, AnyTypeCreate (), CKM->new_variFormulaRootClock ());
  EM->setSignalAssignedTo (exp, exp1);
  return exp1;
}

/** 
    @return the clock of the expression (vfield).
    @param vfield (in) an expresion assumed to be an identifier (x), a field access(x.f), an array access (x[]).
    @param contextClock (in) context clock used to fix the clock of a reference to a state variable.
    @param instent (out) the referenced signal is added to the list when the expression does not reference a state variable.
    @param treeRef (in) back attribute to parser abstract tree for error annotations.
    It applies mkFieldAccess() on the expression (vfield) first.
*/
ptinsth intemkIGraph::fieldAccess (ptexpcal vfield, ptinsth contextClock, ptlinstances * instent, BackAnnotationType treeRef)
{
  ptexpcal vexpcal1;

  // to do : bundle.
  mkFieldAccess (vfield, treeRef);
  if (EM->getSignalAssignedTo (vfield, &vexpcal1))
    {
      if (!EM->isReferenceToStateVar (vfield))
	{
	  *instent = ptlinstances_insert (*instent, vexpcal1);
	  return EM->getClock (vexpcal1);
	}
      return instContextClock (contextClock, treeRef);
    }

  // Blindage
  *instent = ptlinstances_insert (*instent, EM->getDeclaredSignal (vfield));
  return EM->getClock (vfield);
}


/**
   The affine relation pt=(n, phi, d) is divised in pt1=(1, max(0, -phi), n) and pt2=(1, max(0, phi), d).
*/
void
intemkIGraph::affrmkdivide (ptexpcal n, ptexpcal phi, ptexpcal d, ptAffineRelation * pt1, ptAffineRelation * pt2)
{
  compilerInteger period_n, phase, period_d;
  ptAffineRelation ptcanonic;

  EM->getIntegerValue (n, &period_n);
  if (EM->getOperator (phi) == opnega)
    {
      EM->getIntegerValue (EM->getUnaryArg (phi), &phase);
      phase = -phase;
    }
  else
    EM->getIntegerValue (phi, &phase);
  EM->getIntegerValue (d, &period_d);

  ptcanonic = mkAffineRelation (period_n, phase, period_d);
  if (period_d != 1)
    ptAffineRelation_canonical_form (ptcanonic);
  ptAffineRelation_divide (ptcanonic, pt1, pt2);
}


/**
   It propagates a pragma (aPragma) on a node (node):
   - the pragma is associated with the node.
   - the output signals of the node are added as objects assigned with the pragma.
   
   (example: for X:=E, and the pragma RunOn Y,Z "1", it produces RunOn Y,Z,X "1")
*/
void
intemkIGraph::PragmaPropagateOn (ptpragma aPragma, ptINode node)
{
  ptexpcal outs, elem;
  ptlinstances vl;

  if (aPragma != NILPRAGMA)
    {
      NM->addPragma (node, aPragma);
      outs = NM->get_leftOperand (node);
      // adding the signals to the current pragma.
      if (outs != NILEXPCAL)
	{
	  vl = (ptlinstances) getPragmaObjects (aPragma);
	  if (EM->satisfiesList (outs))
	    while (EM->getNaryArgs (&outs, &elem))
	      vl = ptlinstances_insert (vl, elem);
	  else
	    vl = ptlinstances_insert (vl, outs);
	  setPragmaObjects (aPragma, (ptObjPragma) vl);
	}
    }
}


/** @return "true" if the identifier (vident) of a signal (vsig) can be removed from the
    database of the identifiers, "false" otherwise.
 */
boolean intemkIGraph::removeIdent (ptexpcal vsig, ptident * vident)
{
  ptident idPrefix, id;

  idPrefix = EM->getOriginalName (vsig);
  if (eq_idents (idPrefix, getIdentNil ()))
    return pk_false;
  id = EM->getName (vsig);
  if (!eq_idents (id, idPrefix))
    return pk_false;
  if (!EM->remove (vsig))
    return pk_false;
  *vident = id;
  return pk_true;
}



/**
   @return "true" if some labels (sub-graphs) are eliminated, "false" otherwise.
   A label is removed if it is not referenced in an explicit dependence.
   --UNUSED UNUSED --
*/
boolean intemkIGraph::LabelsEliminating ()
{
  ptlINode ln, ln1, lnodes;
  ptINode n, aNode, n1;
  boolean brm;
  ptldependence ldep;
  ptlinsth lh, lh1;
  ptinsth vhlab, h1;
  ptexpcal vlab;

  brm = pk_false;
  ldep = AGRAPH->getExplicitDependences ();

  // First: cleaning the labels in the arrays.
  if (NM->getIterates (&lnodes))
    while (ptlINode_get_elem (&lnodes, &aNode))
      {
	// Body part.
	NM->get_array_bodyDefinitions (aNode, &ln, &lh);
	while (ptlINode_get_elem (&ln, &n))
	  if (NM->is_subgraph (n))
	    if (!isNodeReferencedInDependences (n, ldep))
	      {
		NM->arrayBodyRemoveNode (aNode, n);
		NM->get_subGraphDefinitions (n, &ln1, &lh1);
		while (ptlINode_get_elem (&ln1, &n1))
		  NM->arrayBodyInsertNode (aNode, n1);
		while (ptlinsth_get_next (&lh1, &h1))
		  NM->arrayBodyInsertClock (aNode, h1);
		NM->removeNode (n);
		brm = pk_true;
	      }
	// Initial part.
	NM->get_array_initDefinitions (aNode, &ln, &lh);
	while (ptlINode_get_elem (&ln, &n))
	  if (NM->is_subgraph (n))
	    if (!isNodeReferencedInDependences (n, ldep))
	      {
		NM->arrayInitRemoveNode (aNode, n);
		NM->get_subGraphDefinitions (n, &ln1, &lh1);
		while (ptlINode_get_elem (&ln1, &n1))
		  NM->arrayInitInsertNode (aNode, n1);
		while (ptlinsth_get_next (&lh1, &h1))
		  NM->arrayInitInsertClock (aNode, h1);
		NM->removeNode (n);
		brm = pk_true;
	      }

      }
  // Second: cleaning in the subgraphs.
  if (NM->getSubGraphs (&lnodes))
    {
      //  ln  = lnodes; while(ptlINode_get_elem(&ln,&n)) NM->subGraphLinkNodes(n, pk_false);
      ln = lnodes;
      while (ptlINode_get_elem (&ln, &aNode))
	if (!isNodeReferencedInDependences (aNode, ldep))
	  {
	    // removing the label
	    if (NM->get_label (aNode, &vlab))
	      {
		vhlab = EM->getClock (vlab);
		if (!CKM->satisfiesNullClock (vhlab))
		  CKM->removeFromClock (vhlab, vlab, pk_false);
	      }
	    // cleaning the graph.
	    NM->get_subGraphDefinitions (aNode, &ln, &lh);
	    if (NM->get_inComposedNode (aNode, &n))
	      {
		if (NM->is_subgraph (n))
		  {
		    while (ptlINode_get_elem (&ln, &n1))
		      NM->subGraphInsertNode (n, n1);
		    while (ptlinsth_get_next (&lh1, &h1))
		      NM->subGraphInsertClock (n, h1);
		  }
		else;		// does not occur, done before.
	      }
	    else
	      NM->linkNodesToClock (ln);	// Nodes of subgraph become visible.
	    NM->subGraphUnlinkNodes (aNode, pk_false);
	    NM->removeNode (aNode);
	    brm = pk_true;
	  }
      // if (NM->getSubGraphs( &ln)) while(ptlINode_get_elem(&ln,&n)) NM->subGraphUnlinkNodes(n, pk_false);
    }

  return brm;
}

boolean intemkIGraph::isNodeReferencedInDependences (ptINode aSubGraph, ptldependence ldep)
{
  ptdependence aDep;
  ptINode vFrom, vTo;
  ptinsth h;

  while (ldep->get_next (&aDep, &ldep))
    {
      h = aDep->getDependence (&vFrom, &vTo);
      if ((vFrom == aSubGraph) || (vTo == aSubGraph))
	return pk_true;
    }
  return pk_false;
}


/**
   It verifies that the delays can be implemented (the delay value is bounded and positive or equal to zero).
   It rewrites  x := y $I init VINIT in (| TX := y windows (N+1) init VINIT | x := TX[N-I] |) where N is the max value of the signal I.
   The expessions are not rewritten here (See delaysRewriting()).
   
*/
void
intemkIGraph::delaysFinalizing ()
{
  ptexpcal resret, sigret, vret, vfen, vinit, vdim, vtab, vsize, newExprDef;
  ptlINode ln;
  ptINode winNode, aDelay;
  compilerInteger bi, bs;
  TypeType TT;
  ptinsth vh;
  ptident id;

  if (NM->getDelays (&ln))
    while (ptlINode_get_elem (&ln, &aDelay))
      {
	NM->get_delay_arg (aDelay, &resret, &sigret, &vret, &vfen, &vinit);
	if (!EM->isNILEXPCAL (vret))
	  if (!EM->IsConstantExpr (vret))
	    {
	      BackAnnotationType ba = NM->getBackAnnotation (aDelay);
	      ntrees t = BACK_ANNOT_GET_ABSTRACT_TREE (ba);
	      if (!EM->getBounds (vret, &bi, &bs))
		pK_messenger->errorTree (errNotBounded, t);
	      else if (bi < 0)
		pK_messenger->errorTree (errctPositiveValue, t);
	      else
		{
		  vh = EM->getClock (resret);
		  if (bs == 0)
		    NM->modifiedDefinition (resret, sigret, vh, vh);
		  else
		    {
		      vsize = EM->mkIntegerConstant (bs + 1);
		      vdim = EM->mkBinaryExp (oplesign, vsize, NILEXPCAL, ba);
		      TT = EM->mkArrayType (vdim, get_IntegerType ());
		      id = getIdentNil ();
		      vtab = EM->mkIdentExp (id, tsignal, TT, NM->mk_node_undefined (), vh, ba);
		      EM->setInitialValue (vtab, vinit);
		      winNode = NM->mk_node_delay (vtab, sigret, NILEXPCAL, vsize, vinit, ba);
		      NM->insertNode (winNode, vh);
		      newExprDef = EM->mkBinaryExp (opelem, vtab, EM->mkBinaryExp (oplesign, EM->mkBinaryExp (opmoin, vsize, vret, ba), NILEXPCAL, ba), ba);
		      NM->modifiedDefinition (resret, newExprDef, vh, vh);
		      EM->setInitialValue (resret, NILEXPCAL);
		    }
		}
	    }
      }
}


/**
   In a delay/window context ( $ <vdelay> window <vwindow> init <vinit>), it adjusts the expression (vinit) according to 
   the values of vdelay and vwindow. For example  $ 2 init 0 is replaced by $2 init [{I from 0 to 1}:0].
   
   @return "true" if the initial value (vinit) is modified, "false" otherwise.
*/
boolean
intemkIGraph::adjustInitialValue (ptexpcal * vinit, ptexpcal vdelay, ptexpcal vwindow, BackAnnotationType t)
{
  noperators vop;
  boolean b;
  compilerInteger bi, bs;
  ptexpcal ubound, id, vexp;
  ntrees vabstree = BACK_ANNOT_GET_ABSTRACT_TREE(t);

  if ((*vinit) == NILEXPCAL) return pk_false;
  if ((vdelay == NILEXPCAL) && (vwindow == NILEXPCAL)) return pk_false;

  vop = EM->getOperator (*vinit);
  if (vop == oplelevec) return pk_false;
  if (vop == oplevect) return pk_false;

  if ((vdelay != NILEXPCAL) && (vwindow != NILEXPCAL))
    {
      pK_messenger->errorTree (errCannotAdjustInitvalue, vabstree);
      return pk_false;
    }

  b = pk_false;
  if (vwindow == NILEXPCAL)	// X := Y $ N init vinit.
    {
      if (EM->getBounds (vdelay, &bi, &bs)) b = (bs > 1); else b = pk_true;
      if (b) ubound = EM->mkBinaryExp (opmoin, vdelay, EM->mkIntegerConstant (1), t);
    }
  else
    // X := Y window M init vinit.
    {
      if (EM->getBounds (vwindow, &bi, &bs)) b = (bs > 2); else b = pk_true;
      if (b) ubound = EM->mkBinaryExp (opmoin, vwindow, EM->mkIntegerConstant (2), t);
    }
  if (b)
    {
      // ubound = EM->mkIntegerConstant(bs-1);
      vexp = EM->mkTernaryExp (opborn, EM->mkIntegerConstant (0), ubound, EM->mkIntegerConstant (1), t);
      id = EM->mkIndexIdent (getIndexId (), ubound, NM->mk_node_undefined (), hundef, t);
      AGRAPH->insertIndex (id);
      vexp = EM->mkBinaryExp (opparc, id, vexp, t);
      vexp = EM->mkNaryExp (oplparc, vexp, t);
      vexp = EM->mkBinaryExp (opensvec, vexp, *vinit, t);
      vexp = EM->mkNaryExp (oplelevec, vexp, t);
      *vinit = vexp;
      return pk_true;
    }
  return pk_false;
}




/**
   @return "true" if the identifier of exp exists, "false" otherwise.
   The returned ideitifier (id) is defined by
   - exp is a signal it is its original (declared) name
   - the name of exp1_exp2 for exp = exp1.exp2 
   - the name of exp1 for exp = exp1[exp2]
   
*/
boolean
intemkIGraph::getOriginalName (ptexpcal exp, ptident * id)
{
  noperators vop;
  ptexpcal vf;

  if (exp == NILEXPCAL) return pk_false;
  if (EM->IsSignal (exp))
    {
      *id = EM->getOriginalName (exp);
      return pk_true;
    }
  vop = EM->getOperator (exp);
  if (vop == opelem) return getOriginalName (EM->getBinaryArgstAt (exp, 1), id);
  if (vop != opcham) return pk_false;
  if (!EM->getSignalAssignedTo (exp, &vf)) return pk_false;
  *id = EM->getOriginalName (vf);
  return pk_true;
}
