
/* Adding (vh1, vh2) in
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   It adds a placed clock (vh) as a constraint to be the null clock.
*/
  void
ClockSystem::addPlacedNullClocks (ptinsth vh)
{
  boolean b;
  lsnullpl = ptlinsth_sorted_insert_elem_bool (lsnullpl, vh, &b);
}

/** 
    Adding (vh1, vh2) in 
    - the set of placed formula to be proved equivalent
    - or in the set of assertions when one of the formula is tagged as an assumption.
*/
void
ClockSystem::addConstraintPlacedClocks (ptinsth vh1, ptinsth vh2)
{
  vh1 = getRepresentativeFormula (vh1);
  vh2 = getRepresentativeFormula (vh2);
  if (vh1 != vh2)
    {
      if (satisfiesIsAnAssumption (vh1))
	hcladdassertion (vh1, vh2);
      else if (satisfiesIsAnAssumption (vh2))
	hcladdassertion (vh2, vh1);
      else if (!assumedWithoutConstraints) 
	addConstraintPlacedClocksNotAssumptions(vh1, vh2);
    }
}

_PROTECTED_
/** Removing a variable.
    Remove a clock (vh) from the list of the clocks without definition(empty class).
*/
  void
ClockSystem::removeVari (ptinsth vh)
{
  removeSortedList (vh, &lvariables, &lvariablesEnd);
  removelhroots (vh);
}

/** Removing NOT formula.
  @return the class vform, in which the formula vform has been removed.
  References and lists of formula are uptdated.
  @param becomesvari (out) at pk_true when the resulting clock is whitout definition, false otherwise.
  @param vform an assumed not-formula to remove.
*/
ptinsth
ClockSystem::removeNot (ptinsth vform, boolean * becomesvari)
{
  boolean bdummy;
  lsnot = ptlinsth_remove_elem (lsnot, vform, &bdummy);
  return removeNotFormulaFromClass (vform, becomesvari);
}

/** Removing a LOWER BOUND formula.
   @return the class vform, in which the formula vform has been removed.
   References and lists of formula are uptdated.
   @param becomesvari (out) at pk_true when the resulting clock is whitout definition, false otherwise.
   @param vform (in) : an assumed inf-formula to remove.
*/
ptinsth
ClockSystem::removeInf (ptinsth vform, boolean * becomesvari)
{
  boolean bdummy;
  if (!satisfiesIsCanonicalRepr (vform))
    {
      lsinfer = ptlinsth_remove_elem (lsinfer, vform, &bdummy);
      removelhroots (vform);
      return removeBinaryFormula (vform, becomesvari);
    }
  return vform;
}

/** Removing UPPER BOUND formula.
    @return the class vform, in which the formula vform has been removed.
    References and lists of formula are uptdated.
    @param vform (in) : an assumed sup-formula to remove.
    @param becomesvari (out) at pk_true when the resulting clock is whitout definition, false otherwise.
*/
ptinsth
ClockSystem::removeSup (ptinsth vform, boolean * becomesvari)
{
  boolean bdummy;
  if (!satisfiesIsCanonicalRepr (vform))
    {
      lssuper = ptlinsth_remove_elem (lssuper, vform, &bdummy);
      removelhroots (vform);
      return removeBinaryFormula (vform, becomesvari);
    }
  return vform;
}

/** Adding equivalence.
   It adds the equivalence VH1 = VH2 to do.
*/
void
ClockSystem::hcladdsynchro (ptinsth vh1, ptinsth vh2)
{
  lterm1equ = ptlinsth_insert_elem (lterm1equ, vh1);
  lterm2equ = ptlinsth_insert_elem (lterm2equ, vh2);
  if (getOperator (vh2) == vari)
    register_variClockRoot (vh2);
}

/** Getting equivalence.
    @return true if it exists some equivalence to do in the system, 
    in this case vh1 and vh2 must be unified.
*/
boolean
ClockSystem::hclsomesynchro (ptinsth * vh1, ptinsth * vh2)
{
  ptlinsth_get_next (&lterm1equ, vh1);
  return ptlinsth_get_next (&lterm2equ, vh2);
}

/** Adding clock to the modified set.
   @param vh  (in) the clock for which the reference must be rewritten.
   @return none.
*/
void
ClockSystem::hcladdmodified (ptinsth vh)
{
  lsmodified = ptlinsth_insert_elem (lsmodified, vh);
  if (getOperator (vh) == vari)
    register_variClockRoot (vh);
}

/** Getting clock from modified set.
   @return true if the set of modified formula is not empty, false otherwise.
   @param vh (out) clock for which the reference to must be rewritten.
*/
boolean
ClockSystem::hclsomehmodified (ptinsth * vh)
{
  return ptlinsth_get_next (&lsmodified, vh);
}

/**
   Adding a clock in the set of roots.
*/
void
ClockSystem::insertlhroots (ptinsth vh)
{
  insertSortedList (vh, &lsroots, &lsrootsEnd);
}

/**
   @return true if the removing of a clock (vh) from the set of roots succeeds, false otherwise.
*/
boolean
ClockSystem::removelhroots (ptinsth vh)
{
  return removeSortedList (vh, &lsroots, &lsrootsEnd);
}

/**
   Adding a clock in the set of placed clocks.
*/
void
ClockSystem::insertlhplaced (ptinsth vh)
{
  pkClockType vop;
  boolean b;

  vop = getOperator (vh);
  if (vop != vari) // NEW NEW October 2013.
    {
      if (vop == sub)
	{
	  removelssub (vh);
	  lsextract = insertlcond (lsextract, vh);
	  removelhroots (vh);
	}
      else
	{
	  removelhroots (vh);
	  if (vop == infer)
	    {
	      lsinfer = ptlinsth_remove_elem (lsinfer, vh, &b);
	      insertSortedList (vh, &lsinfintree, &lsinfintreeEnd);
	    }
	  else if (vop == super)
	    {
	      lssuper = ptlinsth_remove_elem (lssuper, vh, &b);
	      insertSortedList (vh, &lssupintree, &lssupintreeEnd);
	    }
	  else;
	}
      insertSortedList (vh, &lsplaced, &lsplacedEnd);
    }
}

/**
   @return true if the removing of a clock (vh) from the set of placed clocks succeeds, false otherwise.
   When b is pk_true, the clock is also removed from the placed sub-set of formula (one by clock formula operator).
*/
boolean
ClockSystem::removelhplaced (ptinsth vh, boolean b)
{
  pkClockType vop;
  boolean x;

  x = removeSortedList (vh, &lsplaced, &lsplacedEnd);
  if (b)
    {
      vop = getOperator (vh);
      if (vop == infer)
	removeSortedList (vh, &lsinfintree, &lsinfintreeEnd);
      else if (vop == super)
	removeSortedList (vh, &lssupintree, &lssupintreeEnd);
      else if (vop == sub)
	ptlinsth_remove_condition (&lsextract, getArgSub (vh));
    }
  return x;
}

/** Adds vh as a new clock to be merge with the ^0 clock. */
void
ClockSystem::setClockToNullClock (ptinsth vh)
{
  // ptinsth_DEBUG_traceClock("ClockSystem::setClockToNullClock", vh);
  this_nullClockList = ptlinsth_insert_elem (this_nullClockList, vh);
}

/** 
    If a clock (vh) is not already placed, it is inserted as a root in the hierarchy.
*/
void
ClockSystem::addClockAsRoot (ptinsth vh)
{
  vh = getRepresentativeFormula (vh);
  if (!ptinsth_satisfies_predef (vh))
    if (satisfiesRoot (vh))
      {
	placeInClockHierarchy (vh, vh, vh, 0);
	insertlhroots (vh);
      }
}

/** 
    @return the class of the formula op(vharg) where op denotes the "hnot" operator or the "hcomp" operator.
    This formula is added to the vhcl(in) class. A new class is created if vhcl (in) is the undefined clock.
*/
insthFormula
ClockSystem::mkUnaFormula (insthClock * vhresult, insthClock vharg, pkClockType op)
{
  insthFormula result;
  result = insert_newFormula (vhresult, op);
  setUnaryArg (result, vharg);
  return result;
}

/** 
    @return the class of the formula op(vharg1, vharg2) where op denotes a binary operator (^+, ^*).
    This formula is added to the vhcl(in) class. A new class is created if vhcl (in) is the undefined clock.
*/
insthClock
ClockSystem::new_binaryFormulaClock (insthClock vharg1, insthClock vharg2, pkClockType op)
{
  insthClock result;
  result = ptinsth_newClock (op);
  setBinaryArgs (result, vharg1, vharg2);
  // DEBUG_traceClock(" exiting ClockSystem::new_binaryFormulaClock, created clock", result);
  return result;
}

/** 
    @return the class of the formula ^+(vharg1, vharg2)
    ^+(vharg1, vharg2) is added to the vhcl(in) class. A new class is created if vhcl (in) is the undefined clock.
    ^+(vharg1, vharg2) is added in the list of super fromula and as a root
*/
insthFormula
ClockSystem::new_superFormulaRootClock (insthClock * vhresult, insthClock vharg1, insthClock vharg2)
{
  insthFormula vform;
  vform = insert_newFormula (vhresult, super);
  setBinaryArgs (vform, vharg1, vharg2);
  lssuper = ptlinsth_insert_elem (lssuper, vform);
  insertlhroots (vform);
  return vform;
}

/** 
    @return the class of the formula ^*(vharg1, vharg2)
    ^*(vharg1, vharg2) is added to the vhcl(in) class. A new class is created if vhcl (in) is the undefined clock.
    ^*(vharg1, vharg2) is added in the list of infer fromula and as a root
*/
insthFormula
ClockSystem::new_inferFormulaRootClock (insthClock * vhresult, insthClock vharg1, insthClock vharg2)
{
  insthFormula vform;
  vform = insert_newFormula (vhresult, infer);
  setBinaryArgs (vform, vharg1, vharg2);
  lsinfer = ptlinsth_insert_elem (lsinfer, vform);
  insertlhroots (vform);
  return vform;
}

/** 
    It updates the list of the clock system: The clock vh is added to the unplaced clock.
    A "hcomp" formula hcomp(H) is replaced by a complementary formula *(HC, not(H)) where HC is the father clock of H.
*/
void
ClockSystem::addFormulaToListRoot (ptinsth vh)
{
  ptinsth vharg, vfather;
  pkClockType vop;

  vop = getOperator (vh);
  if (vop == hcomp)
    {
      getUnaryArg (vh, &vharg);
      removeReferenceFromTo (vharg, vh);
      vharg = getForce_notClock (vharg);
      vfather = getFather (vh);
      setOperator (vh, infer);
      setBinaryArgs (vh, vfather, vharg);
      vop = infer;
    }

  switch ((int) vop)
    {
    case infer:
      if (!ptlinsth_isin_elem (lsinfer, vh))
	lsinfer = ptlinsth_insert_elem (lsinfer, vh);
      break;
    case super:
      if (!ptlinsth_isin_elem (lssuper, vh))
	lssuper = ptlinsth_insert_elem (lssuper, vh);
      break;

    case sub:
      lssub = insertlcond (lssub, vh, pk_false);
      break;

    case hcomp:		// no occurrence
    case hnot:
    case nosub:
    case vari:
    case predef:
      break;
    default:
      pK_messenger->Caseerror (__LINE__, __FILE__);
    }
}

/**
   Removing an extraction (vh) from the system.
   It assumes that vh is an extraction.
*/
void
ClockSystem::removelssub (ptinsth vh)
{
  removelhroots (vh);
  ptlinsth_remove_condition (&lssub, getArgSub (vh));
}

/**
   @return true if  the output parameter vwhichnull is assigned to a value (-1, 0, 1). The coding is 
   - vwhichnull=-1 if vharg1=^0
   - vwhichnull=0  if vharg1=vharg2
   - vwhichnull=1  if vharg2=^0
   false otherwise.
*/
boolean
ClockSystem::hredinfsup (ptinsth vharg1, ptinsth vharg2, int *vwhichnull)
{
  if (vharg1 == vharg2)
    {
      *vwhichnull = 0;
      return pk_true;
    }
  if (satisfiesNullClock (vharg1))
    {
      *vwhichnull = -1;
      return pk_true;
    }
  if (satisfiesNullClock (vharg2))
    {
      *vwhichnull = 1;
      return pk_true;
    }
  return pk_false;
}

/** 
    @return "true" if (vother ^- vharg) can be rewriten in vother, "false" otherwise.
*/
boolean
ClockSystem::hredhnot (ptinsth vharg, ptinsth vother)
{
  ptinsth vh, vharghnot, vh1;

  if (getOperator (vharg) != hnot)
    return pk_false;
  getUnaryArg (vharg, &vh);
  if (satisfiesNullClock (vh))
    return pk_true;
  if (getMinusArgs (vother, &vharghnot, &vh1))
    return (vh == vharghnot);
  return pk_false;
}

/** 
    @return "true" if some simplification on tick clock can be applied, in this case vh is the result, "false" otherwise.
 */
boolean
ClockSystem::hredupb (ptinsth vharg1, ptinsth vharg2, ptinsth * vh)
{
  if (isFictitiousRootFormula (vharg1))
    {
      if (getOperator (vharg2) == hnot)
	return pk_false;
      *vh = vharg2;
      return pk_true;
    }
  if (!isFictitiousRootFormula (vharg2))
    return pk_false;
  if (getOperator (vharg1) == hnot)
    return pk_false;
  *vh = vharg1;
  return pk_true;
}

/**
   @return true if it exists a null clock (vh) , not yet merge to the null clock.
   The references to this clock must be visited for rewritting.
*/
boolean
ClockSystem::hclsomehnull (ptinsth * vh)
{
  return ptlinsth_get_next (&this_nullClockList, vh);
}

/**
  It adds a not placed clock (vh) as a clock to be merge with the null clock.
  If toinsert is pk_true, it is added in the set of the clocks without definition.
*/
void
ClockSystem::hcladdhnull (ptinsth vh, boolean toinsert)
{
  // ptinsth_DEBUG_traceClock("ClockSystem::hcladdhnull", vh);
  this_nullClockList = ptlinsth_insert_elem (this_nullClockList, vh);
  if (toinsert)
    {
      if (!ptinsth_satisfies_predef (vh))
	{
	  setOperator (vh, vari);
	  register_variClockRoot (vh);
	}
    }
}

/** 
    @return true if it exists not solved constrainsts on placed clocks, false otherwise. 
*/
boolean
ClockSystem::existsynchropl ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  lh1 = lsnullpl;
  while (ptlinsth_get_next (&lh1, &vh1))
    if (!satisfiesIsAnAssumption (vh1))
      return pk_true;

  getConstraintsOnPlacedClocks (&lh1, &lh2);
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = getRepresentativeFormula (vh1);
      vh2 = getRepresentativeFormula (vh2);
      if (vh1 != vh2)
	return pk_true;		/* sinon il faudrait enlever */
    }
  return pk_false;
}

/**
   Adding (vh1, vh2) in the set of assumptions.
*/
void
ClockSystem::hcladdassertion (ptinsth vh1, ptinsth vh2)
{
  if (vh1 == vh2)
    return;
  if (! ptlinsth_search_couple (vh1, vh2, lterm1assertpl, lterm2assertpl))
    {
      lterm1assertpl = ptlinsth_insert_elem (lterm1assertpl, vh1);
      lterm2assertpl = ptlinsth_insert_elem (lterm2assertpl, vh2);
    }
}

/**
   @return true if the formula -(VHARG) exists in the system, false if it is created.
   vhresult (out) contains, all the cases, the class of the formula.
*/
boolean
ClockSystem::findForce_notFormula (ptinsth * vhresult, ptinsth vharg)
{
  ptinsth velem;

  if (ptinsth_get_1arg (vharg, &velem))
    {				/* vharg = not velem, or, vharg = hcomp velem */
      *vhresult = getRepresentativeFormula (velem);
      return pk_true;
    }
  if (find_notClockFormula (vhresult, vharg))
    return pk_true;
  velem = insert_newFormula (vhresult, hnot);
  setUnaryArg (velem, vharg);
  lsnot = ptlinsth_insert_elem (lsnot, velem);
  return pk_false;
}

/**
   @return true if the formula (vharg1 ^* vharg2) exists in the system, false if it is created and not placed.
   @param vharg1, vharg2 (in) : arguments of  the ^* formula.
   @param vhresult (out) : the class of the formula (vharg1 ^* vharg2).
   @param vfinf (out) : the formula (vharg1 ^* vharg2).
*/
boolean
ClockSystem::findForce_inferFormula (ptinsth * vhresult, ptinsth vharg1, ptinsth vharg2, ptinsth * vfinf)
{
  if (sortFormula (&vharg1, &vharg2))
    {
      if (find_binaryClockFormula (lsinfintree, vharg1, vharg2, vhresult))
	{
	  *vfinf = *vhresult;
	  return pk_true;
	}
      setorder_from_num (&vharg1, &vharg2);
    }

  if (rchFormula (infer, vharg1, vharg2, vhresult))
    {
      *vfinf = *vhresult;
      return pk_true;
    }
  /* vhresult is unchanged */
  *vfinf = new_inferFormulaRootClock (vhresult, vharg1, vharg2);
  return pk_false;
}

/**
   @return true if the formula (vharg1 ^+ vharg2) exists in the system, false if it is created and not placed.
   @param vharg1, vharg2 (in) : arguments of  the ^+ formula.
   @param vhresult (out) : the class of the formula (vharg1 ^+ vharg2).
   @param vfsup (out) : the formula (vharg1 ^+ vharg2).
*/
boolean
ClockSystem::findForce_superFormula (ptinsth * vhresult, ptinsth vharg1, ptinsth vharg2, ptinsth * vfsup)
{
  if (sortFormula (&vharg1, &vharg2))
    {
      if (find_binaryClockFormula (lssupintree, vharg1, vharg2, vhresult))
	{
	  *vfsup = *vhresult;
	  return pk_true;
	}
      setorder_from_num (&vharg1, &vharg2);
    }

  if (rchFormula (super, vharg1, vharg2, vhresult))
    {
      *vfsup = *vhresult;
      return pk_true;
    }
  /* vhresult is unchanged */
//  *vfsup = mkBinFormula(vhresult, vharg1, vharg2, super);
//  lssuper = ptlinsth_insert_elem(lssuper, *vfsup);
//  insertlhroots(*vfsup);
  *vfsup = new_superFormulaRootClock (vhresult, vharg1, vharg2);
  return pk_false;
}

/**
  @return true if the condition with cond as expression exists in the system, false otherwise.
  When it exists in the system, when valvrai :
   - is pk_true, vhresult (out) is the  class of [cond] (undersampling by the condition),
   - is pk_false, vhresult (out) is the classe of [not cond] (undersampling by the negation of the condition).
*/
boolean
ClockSystem::rchlsub (boolean valvrai, ptexpcal cond, ptinsth * vhresult)
{
  boolean trouve;
  ptcondition velem;

  trouve = recherche (lssub, cond, &velem);
  if (!trouve)
    trouve = recherche (lsextract, cond, &velem);
  if (!trouve)
    return pk_false;
  if (valvrai)
    *vhresult = velem->get_clksub ();
  else
    *vhresult = velem->get_clknosub ();
  *vhresult = getRepresentativeFormula (*vhresult);
  return pk_true;
}

/**
   It changes the representative formula of a class.
   The representative formula of the class of vhold becomes vhnew.
*/
void
ClockSystem::hclchangform (ptinsth vhold, ptinsth vhnew)
{
  if (!satisfiesRoot (vhold))
    if (getDepth (vhold) == 0)
      removelhroots (vhold);
  modify_clhead (vhold, vhnew);
  addClockAsRoot (vhnew);
}

