/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** print a list of formula terminated by NILLINSTH
    may loop for ever
 */
extern void
ptinsth_DEBUG_printLnFormulaList (char *lisName, ptlinsth aFormulaList)
{
  int numberInLine = -1;
  int i;
  int prefixLength = strlen (lisName) + 4;
  printf ("%s: <<", lisName);
  while (aFormulaList != NILLINSTH)
    {
      if (numberInLine < 0)
	numberInLine = 1;
      else if (numberInLine < 6)
	{
	  printf (", ");
	  numberInLine = numberInLine + 1;
	}
      else
	{
	  printf ("\n");
	  for (i = 0; i < prefixLength; i++)
	    printf (" ");
	  numberInLine = 1;
	}
      ptinsth_DEBUG_printFormula (aFormulaList->elem);
      aFormulaList = aFormulaList->next;
    }
  printf (">>\n");
  fflush (stdout);
}

/** print a list of formula terminated by NILLINSTH
    may loop for ever
 */
extern void
ptinsth_DEBUG_printLnClockResumeList (char *lisName, ptlinsth aFormulaList)
{
  int numberInLine = 0;
  int i, incr;
  int prefixLength = strlen (lisName) + 4;
  insthFormula currentFormula;
  printf ("%s: <<", lisName);
  while (aFormulaList != NILLINSTH)
    {
      if (ptinsth_DEBUG_printClockResume (aFormulaList->elem))
	numberInLine = numberInLine + 2;
      else
	numberInLine = numberInLine + 1;
      aFormulaList = aFormulaList->next;
      if (aFormulaList != NILLINSTH)
	{
	  if (numberInLine >= 6)
	    {
	      printf ("\n");
	      for (i = 0; i < prefixLength; i++)
		printf (" ");
	      numberInLine = incr;
	    }
	  else
	    printf (", ");
	}
    }
  printf (">>\n");
  fflush (stdout);
}

/**
   @return true if the list (lh) is empty.
*/
extern boolean
ptlinsth_isEmptyList (ptlinsth lh)
{
  return (lh == NILLINSTH);
}

/**
   @return the current element of the list assumed not empty.
*/
extern ptinsth
ptlinsth_get_elem (ptlinsth l)
{
  return l->elem;
}

/**
   @return 
   - pk_true if the list  lhtotest is not empty.
   In this case, velem is the current element of the list and the returned list lhtotest is the next.
   - pk_false, the parameters are not modified.
*/
extern boolean
ptlinsth_get_next (ptlinsth * lhtotest, ptinsth * vh)
{
  if ((*lhtotest) == NILLINSTH)
    return pk_false;
  *vh = ptlinsth_removehead_elem (lhtotest);
  return pk_true;
}

/**
   Unsorted list.
   Removing an element.
   @param ptfirsth (in) : head of the list.
   @param velem (in) : element to remove
   @param returncode (out) : at pk_true if velem is in the list, false otherwise.
   @return the list ptfirsth in which the first velem is removed; the list ptfirsth is modified.
*/
extern ptlinsth
ptlinsth_remove_elem (ptlinsth ptfirsth, ptinsth velem, boolean * returncode)
{
  ptlinsth pred, aux;
  boolean found;

  if (ptfirsth == NILLINSTH)
    return NILLINSTH;
  aux = ptfirsth;
  found = pk_false;
  pred = NILLINSTH;
  do
    {
      if (aux->elem == velem)
	found = pk_true;
      else
	{
	  if (pred == NILLINSTH)
	    ptfirsth = aux;
	  pred = aux;
	  aux = aux->next;
	}
    }
  while (!(found || (aux == NILLINSTH)));
  (*returncode) = found;
  if (found)
    {
      if (pred != NILLINSTH)
	pred->next = aux->next;
      else
	ptfirsth = aux->next;
    }
  return ptfirsth;
}

/**
   Unsorted list.
   It adds, the element elem, at the beginning of the list lst.
   @return the new list.
*/
extern ptlinsth
ptlinsth_insert_elem (ptlinsth lst, ptinsth velem)
{
  ptlinsth vli;

  vli = ptlinsth_create ();
  vli->next = lst;
  vli->elem = velem;
  return vli;
}

/**
   Sorted list.
   Removing an element (vh) from a list (deblist).
   @param deblist (in/out) : head of the list, can be modified (out)
   @param vh (in) : element to remove
   @return true if velem is in the list, false otherwise.
*/
extern boolean
ptlinsth_sorted_remove_elem (ptlinsth * deblist, ptinsth vh)
{
  int numero, cnum;
  ptlinsth curlist, predlist, succlist;

  if ((*deblist) == NILLINSTH)
    return pk_false;
  numero = ptinsth_get_num (vh);
  predlist = NILLINSTH;
  succlist = *deblist;
  while (pk_true)
    {
      curlist = succlist;
      cnum = ptinsth_get_num (ptlinsth_removehead_elem (&succlist));
      if (numero < cnum)
	return pk_false;
      if (numero == cnum)
	{
	  if (predlist == NILLINSTH)
	    *deblist = succlist;
	  else
	    predlist->next = succlist;
	  return pk_true;
	}
      if (succlist == NILLINSTH)
	return pk_false;
      predlist = curlist;
    }
}

/**
   Sorted list.
   It adds, the element velem, in a list (deblist) sorted by the number attribute assigned to clocks.
   The ouput parameter binsert is set to the value "true" when the operation succeeds, false when it was already in the list.
   @return the new list.
*/
extern ptlinsth
ptlinsth_sorted_insert_elem_bool (ptlinsth deblist, ptinsth velem, boolean * binsert)
{
  ptlinsth listResult;
  int numero, cnum;
  ptlinsth curlist, laux;

  *binsert = pk_true;
  if (deblist == NILLINSTH)
    return ptlinsth_insert_elem (NILLINSTH, velem);
  numero = ptinsth_get_num (velem);
  curlist = deblist;
  cnum = ptinsth_get_num (ptlinsth_removehead_elem (&deblist));
  if (numero < cnum)
    return ptlinsth_insert_elem (curlist, velem);
  listResult = curlist;
  while (cnum != -1)
    if (cnum == numero)
      {
	cnum = -1;
	*binsert = pk_false;
      }
    else if (deblist == NILLINSTH)
      {
	curlist->next = ptlinsth_insert_elem (NILLINSTH, velem);
	cnum = -1;
      }
    else
      {
	laux = deblist;
	cnum = ptinsth_get_num (ptlinsth_removehead_elem (&deblist));
	if (cnum > numero)
	  {
	    curlist->next = ptlinsth_insert_elem (laux, velem);
	    cnum = -1;
	  }
	else
	  curlist = laux;
      }
  return listResult;
}

/**
   Similar to ptlinsth_sorted_insert_elem_bool (ptlinsth deblist, ptinsth velem, boolean * binsert)
   without the returned boolean.
 */
extern ptlinsth
ptlinsth_sorted_insert_elem (ptlinsth deblist, ptinsth velem)
{
  boolean b;
  return ptlinsth_sorted_insert_elem_bool(deblist, velem, &b);
}

/** @return a copy of the list lst.
 */
extern ptlinsth
ptlinsth_duplicate (ptlinsth lst)
{
  ptinsth vh;

  if (lst == NILLINSTH)
    return NILLINSTH;
  vh = ptlinsth_removehead_elem (&lst);
  return ptlinsth_insert_elem (ptlinsth_duplicate (lst), vh);
}

/**
   Merging two list with filtering.
   @param lhprinc (in) : list where the elements must be inserted, assumed to be sorted.
   @param lhadd  (in) : list of clock to insert in lhprinc.
   @param hexcluded (in) : clock that must not be inserted.
   @return the list ((lhprinc + lhadd ) - {hexcluded}).
*/
extern ptlinsth
ptlinsth_sorted_fusionexcept (ptlinsth lhprinc, ptlinsth lhadd, ptinsth hexcluded)
{
  ptinsth vh;
  ptlinsth curlist;
  boolean b;

  if (lhadd == NILLINSTH)
    return ptlinsth_duplicate (lhprinc);
  curlist = ptlinsth_duplicate (lhprinc);
  while (lhadd != NILLINSTH)
    {
      vh = ptlinsth_removehead_elem (&lhadd);
      if (vh != hexcluded)
	curlist = ptlinsth_sorted_insert_elem_bool (curlist, vh, &b);
    }
  return curlist;
}

/**
   Merging two list.
   @param lhprinc (in) : list where the elements must be inserted, assumed to be sorted.
   @param lhadd  (in) : list of clock to insert in lhprinc.
   @return the list (lhprinc + lhadd).
*/
extern ptlinsth
ptlinsth_sorted_fusion (ptlinsth lhprinc, ptlinsth lhadd)
{
  boolean b;
  ptinsth e;

  while (ptlinsth_get_next (&lhadd, &e))
    lhprinc = ptlinsth_sorted_insert_elem_bool (lhprinc, e, &b);
  return lhprinc;
}

/**
   @return true if velem is an element in the (sorted/unsorted) list l, false otherwise.
*/
extern boolean
ptlinsth_isin_elem (ptlinsth l, ptinsth velem)
{
  int numero;
  ptinsth vh;

  numero = ptinsth_get_num (velem);
  while (ptlinsth_get_next (&l, &vh))
    if (ptinsth_get_num (vh) == numero)
      return pk_true;
  return pk_false;
}

/**
   Sorted list.
   @return true if velem is an element in the sorted list l, false otherwise.
*/
extern boolean
ptlinsth_sorted_isin_elem (ptlinsth l, ptinsth velem)
{
  int numero, cnum;
  ptinsth vh;

  numero = ptinsth_get_num (velem);
  while (ptlinsth_get_next (&l, &vh))
    {
      cnum = ptinsth_get_num (vh);
      if (numero < cnum)
	return pk_false;
      if (cnum == numero)
	return pk_true;
    }
  return pk_false;
}

/**
   @return the list of the last element of the list lst
*/
extern ptlinsth
ptlinsth_get_last (ptlinsth lst)
{
  ptlinsth aux;

  if (lst == NILLINSTH)
    return NILLINSTH;
  aux = lst;
  while (aux->next != NILLINSTH)
    aux = aux->next;
  return aux;
}

/**
   @return the size of the list lst
*/
extern int
ptlinsth_get_size (ptlinsth lst)
{
  int nb;
  ptinsth h;

  nb = 0;
  while (ptlinsth_get_next (&lst, &h))
    nb++;
  return nb;
}

/**
   Removing the element of the list lst marked by the mark m.
*/
extern ptlinsth
ptlinsth_removeMarked (ptlinsth lst, typeMark m)
{
  ptlinsth lh;
  ptinsth h;

  lh = lst;
  while (ptlinsth_get_next (&lst, &h))
    if (ptinsth_satisfies_marked (h, m))
      ptlinsth_sorted_remove_elem (&lh, h);
  return lh;
}

/**
   Inserting an element in a unsorted list.
   It creates a new item-list that becomes the item-successor of the list lst.
*/
extern ptlinsth
ptlinsth_insert_post (ptlinsth lst, ptinsth velem)
{
  ptlinsth vli;

  vli = ptlinsth_create ();
  vli->elem = velem;
  vli->next = NILLINSTH;
  if (lst != NILLINSTH)
    lst->next = vli;
  return vli;
}

/**
   Removing the last element of the list lst.
   The deleted  item-list  is returned  in nlast.
   @return the modified list.
*/
extern ptlinsth
ptlinsth_remove_last_elem (ptlinsth lst, ptlinsth * nlast)
{
  ptlinsth aux;

  if (lst == NILLINSTH)
    {
      *nlast = NILLINSTH;
      return NILLINSTH;
    }
  if (lst->next == NILLINSTH)
    {
      *nlast = NILLINSTH;
      return NILLINSTH;
    }

  aux = lst;
  while (aux->next != NILLINSTH)
    {
      *nlast = aux;
      aux = aux->next;
    }
  (*nlast)->next = NILLINSTH;
  return lst;
}

/**
   It links the lists pred and next. 
   pred is modified, its item-successor becomes next.
*/
extern void
ptlinsth_append (ptlinsth pred, ptlinsth next)
{
  if (pred != NILLINSTH)
    pred->next = next;
}

/**
   It removes (vh1, vh2) from lh1, lh2 if they are at the same rank.
   @return "true" if the operation succeeds, "false" otherwise (elements are not in the lists).
*/
extern boolean
ptlinsth_remove_pair (ptlinsth * lh1, ptlinsth * lh2, ptinsth vh1, ptinsth vh2)
{
  ptlinsth l1, l2, pred1, pred2;

  if (((*lh1)->elem == vh1) && ((*lh2)->elem == vh2))
    {
      (*lh1) = (*lh1)->next;
      (*lh2) = (*lh2)->next;
      return pk_true;
    }

  do
    {
      pred1 = *lh1;
      pred2 = *lh2;
      l1 = (*lh1)->next;
      l2 = (*lh2)->next;
      if (l1 == NILLINSTH)
	return pk_false;
      if (l2 == NILLINSTH)
	return pk_false;
      if ((l1->elem == vh1) && (l2->elem == vh2))
	{
	  pred1->next = l1->next;
	  pred2->next = l2->next;
	  return pk_true;
	}
    }
  while (pk_true);
}

/**
   Garbage collector for a list (lh) of clocks.
*/
extern void
ptlinsth_free (ptlinsth lh)
{
  ptlinsth nlh;

  if (lh != NILLINSTH)
    {
      nlh = lh;
      while (lh->next != NILLINSTH)
	lh = lh->next;
      lh->next = gclinsth;
      gclinsth = nlh;
    }
}

/**
   @return the list l1 in which the elements of l2 are removed.
   The operand lists are not necessary sorted.
*/
extern ptlinsth
ptlinsth_list_remove (ptlinsth l1, ptlinsth l2)
{
  ptinsth h;
  boolean b;

  while (ptlinsth_get_next (&l2, &h))
    l1 = ptlinsth_remove_elem (l1, h, &b);
  return l1;
}

/** Getting the i-th element of a list of clocks.
    @return true if the i-th element exists.
    @param l (in) a list of expressions.
    @param i (in) wanted element.
    @param ie (out) the i-th element if it exists.
*/
extern boolean
ptlinsth_get_ith_element (ptlinsth l, int i, ptinsth * ie)
{
  int n;

  if (i > 0)
    {
      n = 1;
      while (ptlinsth_get_next (&l, ie))
	{
	  if (n == i)
	    return pk_true;
	  n++;
	}
    }
  return pk_false;
}


/**
   @return true if the couple (vh1, vh2) is an element of the list l1, l2. 
   One element by list and at the same rank.
*/
extern boolean ptlinsth_search_couple(ptinsth vh1, ptinsth vh2, ptlinsth l1, ptlinsth l2)
{
  ptinsth vh11, vh22;

  while (ptlinsth_get_next (&l1, &vh11))
    {
      ptlinsth_get_next (&l2, &vh22);
      if ((vh1 == vh11) && (vh2 == vh22)) return pk_true;
      if ((vh2 == vh11) && (vh1 == vh22)) return pk_true;
    }
  return pk_false;
}
