/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/** @return a new interface object initialized with empty lists.
 */
extern ptinterface
ptinterface_create (void)
{
  ptinterface vinterface;

  if (ptinterface_is_undefined (gcinterface))
    vinterface = new interface ();
  else
    {
      vinterface = gcinterface;
      gcinterface = vinterface->next;
    }
  vinterface->inClocks = NILLINSTH;
  vinterface->outClocks = NILLINSTH;
  vinterface->localClocks = NILLINSTH;
  vinterface->lpar = NILLINSTANCES;
  vinterface->lconstants = NILLINSTANCES;
  vinterface->inSignals = NILLINSTANCES;
  vinterface->outSignals = NILLINSTANCES;
  vinterface->localSignals = NILLINSTANCES;
  vinterface->inOutSignals = NILLINSTANCES;
  vinterface->next = NILINTERFACE;
  return vinterface;
}

/**  @return a copy of an interface (vinterref).
*/
extern ptinterface
ptinterface_duplicate (ptinterface vinterref)
{
  ptinterface vinterface;

  if (ptinterface_is_undefined (vinterref))
    return NILINTERFACE;
  vinterface = new interface ();
  // vinterface->hoin = NILLINSTH;
  // vinterface->outClocks = NILLINSTH;
  vinterface->lpar = ptlinstances_copy (vinterref->lpar);
  vinterface->lconstants = ptlinstances_copy (vinterref->lconstants);
  vinterface->inSignals = ptlinstances_copy (vinterref->inSignals);
  vinterface->outSignals = ptlinstances_copy (vinterref->outSignals);
  vinterface->localSignals = ptlinstances_copy (vinterref->localSignals);
  vinterface->inOutSignals = ptlinstances_copy (vinterref->inOutSignals);
  vinterface->inClocks = ptlinsth_duplicate (vinterref->inClocks);
  vinterface->outClocks = ptlinsth_duplicate (vinterref->outClocks);

  return vinterface;
}


/**
   @return the interface with l as input clocks.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_inputClocks (ptinterface vinterface, ptlinsth l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->inClocks = l;
  return vinterface;
}

/**
   @return the interface with l as output clocks.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_outputClocks (ptinterface vinterface, ptlinsth l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->outClocks = l;
  return vinterface;
}

/**
   @return the interface with l as local clocks.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_localClocks (ptinterface vinterface, ptlinsth l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->localClocks = l;
  return vinterface;
}


/**
   @return the interface with l as parameters.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_parameters (ptinterface vinterface, ptlinstances l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->lpar = l;
  return vinterface;
}


/**
   @return the interface with l as constants.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_constants (ptinterface vinterface, ptlinstances l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->lconstants = l;
  return vinterface;
}

/**
   @return the interface with l as input signals.
   An interface is created when vinterface is undefined.  
*/
extern ptinterface
ptinterface_set_inputSignals (ptinterface vinterface, ptlinstances l)
{
  ptexpcal vsig;

  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->inSignals = l;
  return vinterface;
}


/**
   @return the interface with l as output signals.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_outputSignals (ptinterface vinterface, ptlinstances l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->outSignals = l;
  return vinterface;
}


/**
   @return the interface with l as local signals.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_localSignals (ptinterface vinterface, ptlinstances l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->localSignals = l;
  return vinterface;
}

/**
   @return the interface with l as output signals.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_set_inOutSignals (ptinterface vinterface, ptlinstances l)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->inOutSignals = l;
  return vinterface;
}


/**
   @return the interface in which a clock (vh) is added as an input clock and if it is not in the outputs.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_insert_inputClock (ptinterface vinterface, ptinsth vhin)
{
  boolean b;

  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  if (!ptlinsth_isin_elem (vinterface->outClocks, vhin))
    if (!ptlinsth_isin_elem (vinterface->localClocks, vhin))
      vinterface->inClocks = ptlinsth_sorted_insert_elem_bool (vinterface->inClocks, vhin, &b);
  return vinterface;
}



/**
   @return the interface in which  a clock (vh) is added as an output clock. It is removed from the inputs and the locals.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_insert_outputClock (ptinterface vinterface, ptinsth vh)
{
  boolean b;

  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  ptinterface_remove_localClock (vinterface, vh);
  ptinterface_remove_inputClock (vinterface, vh);
  vinterface->outClocks = ptlinsth_sorted_insert_elem_bool (vinterface->outClocks, vh, &b);
  return vinterface;
}

/**
   @return the interface in which  a clock (vh) is added as a local clock if it is not in the outputs.
   An interface is created when vinterface is undefined.
*/
extern ptinterface
ptinterface_insert_localClock (ptinterface vinterface, ptinsth vh)
{
  boolean b;

  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  ptinterface_remove_inputClock (vinterface, vh);
  if (!ptlinsth_isin_elem (vinterface->outClocks, vh))
    vinterface->localClocks = ptlinsth_sorted_insert_elem_bool (vinterface->localClocks, vh, &b);
  return vinterface;
}

/**
   @return the interface in which a signal (vexp) is added as a local signal, if it is not yet in the outputs.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_insert_inputSignal (ptinterface vinterface, ptexpcal vexp)
{
  if (!isArrayLocalSignal (vexp))
    {
      if (ptinterface_is_undefined (vinterface))
	vinterface = ptinterface_create ();
      if (!ptlinstances_isInSortedList (vinterface->outSignals, vexp))
	if (!ptlinstances_isInSortedList (vinterface->localSignals, vexp))
	  vinterface->inSignals = ptlinstances_insert (vinterface->inSignals, vexp);
    }
  return vinterface;
}




/**
   @return the interface in which a signal (vexp) is added as an output signal. It is removed from the inputs and the locals.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_insert_outputSignal (ptinterface vinterface, ptexpcal vexp)
{

  if (!isArrayLocalSignal (vexp))
    if (!ptlinstances_isInSortedList (vinterface->inSignals, vexp))
      {
	if (ptinterface_is_undefined (vinterface))
	  vinterface = ptinterface_create ();
	vinterface->localSignals = ptlinstances_remove_exp (vinterface->localSignals, vexp);
	vinterface->outSignals = ptlinstances_insert (vinterface->outSignals, vexp);
      }
  return vinterface;
}

/**
   @return the interface in which an identifier (vexp) is added as a parameter.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_insert_parameter (ptinterface vinterface, ptexpcal vexp)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->lpar = ptlinstances_insert (vinterface->lpar, vexp);
  return vinterface;
}

/**
   @return the interface in which an identifier (vexp) is added as a constant.
   An interface is created when vinterface is undefined.
*/
extern ptinterface
ptinterface_insert_constant (ptinterface vinterface, ptexpcal vexp)
{
  if (ptinterface_is_undefined (vinterface))
    vinterface = ptinterface_create ();
  vinterface->lconstants = ptlinstances_insert (vinterface->lconstants, vexp);
  return vinterface;
}

/**
   @return the interface in which a signal (vexp) is added as a local signal. It is removed from the inputs and the locals.
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_insert_localSignal (ptinterface vinterface, ptexpcal vexp)
{
  if (!isArrayLocalSignal (vexp))
    {
      if (ptinterface_is_undefined (vinterface))
	vinterface = ptinterface_create ();
      if (!ptlinstances_isInSortedList (vinterface->outSignals, vexp))
	if (!ptlinstances_isInSortedList (vinterface->inSignals, vexp))
	  vinterface->localSignals = ptlinstances_insert (vinterface->localSignals, vexp);
    }
  return vinterface;
}


/** 
    It removes a clock (vh) from the output clocks of an interface (vinterface).
*/
extern void
ptinterface_remove_outputClock (ptinterface vinterface, ptinsth hout)
{
  if (!ptinterface_is_undefined (vinterface))
    ptlinsth_sorted_remove_elem (&vinterface->outClocks, hout);
}

/** 
    It removes a clock (vh) from the input clocks of an interface (vinterface).
*/
extern void
ptinterface_remove_inputClock (ptinterface vinterface, ptinsth hout)
{
  if (!ptinterface_is_undefined (vinterface))
    ptlinsth_sorted_remove_elem (&vinterface->inClocks, hout);
}

/** 
    It removes a clock (vh) from the local clocks of an interface (vinterface).
*/
extern void
ptinterface_remove_localClock (ptinterface vinterface, ptinsth vh)
{
  if (!ptinterface_is_undefined (vinterface))
    ptlinsth_sorted_remove_elem (&vinterface->localClocks, vh);
}



/** 
    It removes a signal (vexp) from the output signals of an interface (vinterface).
*/
extern ptinterface
ptinterface_remove_outputSignal (ptinterface vinterface, ptexpcal vexp)
{
  if (!ptinterface_is_undefined (vinterface))
    vinterface->outSignals = ptlinstances_remove_exp (vinterface->outSignals, vexp);
  return vinterface;
}


/** 
    It removes a signal (vexp) from the input signals of an interface (vinterface).
*/
extern ptinterface
ptinterface_remove_inputSignal (ptinterface vinterface, ptexpcal vexp)
{
  if (!ptinterface_is_undefined (vinterface))
    vinterface->inSignals = ptlinstances_remove_exp (vinterface->inSignals, vexp);
  return vinterface;
}

/** 
    It removes a signal (vexp) from the local signals of an interface (vinterface).
*/
extern ptinterface
ptinterface_remove_localSignal (ptinterface vinterface, ptexpcal vexp)
{
  if (!ptinterface_is_undefined (vinterface))
    vinterface->localSignals = ptlinstances_remove_exp (vinterface->localSignals, vexp);
  return vinterface;
}

/** 
    It removes a signal (vexp) from the input-output signals of an interface (vinterface).
*/
extern ptinterface
ptinterface_remove_inOutSignal (ptinterface vinterface, ptexpcal vexp)
{
  if (!ptinterface_is_undefined (vinterface))
    vinterface->inOutSignals = ptlinstances_remove_exp (vinterface->inOutSignals, vexp);
  return vinterface;
}

/** 
    @return the interface (vinterprinc) that contains the merging of two interfaces (vinterprinc) and (vintersec).
    \warning vinterprinc is modified.
*/
extern ptinterface
ptinterface_union (ptinterface vinterprinc, ptinterface vintersec)
{
  ptlinstances inSignals;
  if (!ptinterface_is_undefined (vintersec))
    {
      if (ptinterface_is_undefined (vinterprinc))
	vinterprinc = ptinterface_create ();
      vinterprinc->lpar = ptlinstances_sortedfusion (vintersec->lpar, vinterprinc->lpar);
      vinterprinc->lconstants = ptlinstances_sortedfusion (vintersec->lconstants, vinterprinc->lconstants);
      inSignals = ptlinstances_copy (vintersec->inSignals);
      vinterprinc->inSignals = ptlinstances_sortedfusion (connectinstances (inSignals, vinterprinc->outSignals),
							  connectinstances (vinterprinc->inSignals, vintersec->outSignals));
      vinterprinc->outSignals = ptlinstances_sortedfusion (vintersec->outSignals, vinterprinc->outSignals);
      vinterprinc->localSignals = ptlinstances_sortedfusion (vintersec->localSignals, vinterprinc->localSignals);
      vinterprinc->inOutSignals = ptlinstances_sortedfusion (vintersec->inOutSignals, vinterprinc->inOutSignals);
      vinterprinc->inClocks = ptlinsth_sorted_fusion (vinterprinc->inClocks, vintersec->inClocks);
      vinterprinc->outClocks = ptlinsth_sorted_fusion (vinterprinc->outClocks, vintersec->outClocks);
      vinterprinc->localClocks = ptlinsth_sorted_fusion (vinterprinc->localClocks, vintersec->localClocks);

    };
  return vinterprinc;
}



/**
   Garbaging: the elements of an interface (vinterface) are removed of this interface.
   The list descriptors become free.
*/
extern void
interface_free (ptinterface vinterface)
{
  if (!ptinterface_is_undefined (vinterface))
    {
      ptlinstances_free (vinterface->inSignals);
      vinterface->inSignals = NILLINSTANCES;
      ptlinstances_free (vinterface->outSignals);
      vinterface->outSignals = NILLINSTANCES;
      ptlinstances_free (vinterface->lpar);
      vinterface->lpar = NILLINSTANCES;
      ptlinstances_free (vinterface->lconstants);
      vinterface->lconstants = NILLINSTANCES;

      ptlinstances_free (vinterface->localSignals);
      vinterface->localSignals = NILLINSTANCES;

      interface_free_clocks (vinterface, pk_false);
    }
}

/**
   Garbaging.
   The clocks of an interface (vinterface) are removed of this interface.
   When bOnlyLocals 
   
   The list descriptors become free.
*/
extern void
interface_free_clocks (ptinterface vinterface, boolean bOnlyLocals)
{
  if (!ptinterface_is_undefined (vinterface))
    {
      if (!bOnlyLocals)
	{
	  ptlinsth_free (vinterface->inClocks);
	  vinterface->inClocks = NILLINSTH;
	  ptlinsth_free (vinterface->outClocks);
	  vinterface->outClocks = NILLINSTH;
	}
      ptlinsth_free (vinterface->localClocks);
      vinterface->localClocks = NILLINSTH;
    }
}

/**
   @return the interface (vinterface) in which a signal (vexp) is added as a input-output signal (shared signal).
   An interface is created when vinterface is undefined. 
*/
extern ptinterface
ptinterface_insert_inOutSignal (ptinterface vinterface, ptexpcal vexp)
{
  if (!isArrayLocalSignal (vexp))
    {
      if (ptinterface_is_undefined (vinterface))
	vinterface = ptinterface_create ();
      vinterface->inOutSignals = ptlinstances_insert (vinterface->inOutSignals, vexp);
    }
  return vinterface;
}
