/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Access to the list of types.
    @param lexp (in) list of expressions
    @param lTypes (out) list of types referenced in the expressions of the operand list (lexp).
    warning : The returned list is sorted using the value of the entry returned by the type manager.
*/
  void
ExprManager::getUserTypes (ptlexpcal lexp, ptlinstances * lTypes)
{
  ptexpcal id;
  TypeType TT;
  TypeList lTypeList;
  boolean Goon;
  TypeSubList SubLst;

  if (lexp != NILLINSTANCES)
    {
      lTypeList = TypeListCreate ();
      while (ptlinstances_get_next (&lexp, &id))
	{
	  if (!isSharedSignalBySeveralGraphs (id))
	    {
	      TT = getType (id);
	      TypeListInsertheadingType (lTypeList, TT);
	    }
	}

      lTypeList = TypeListRegister (lTypeList);
      SubLst = TypeListCast2SubList (lTypeList);
      Goon = pk_true;
      while (Goon)
	{
	  TT = TypeSubListGetHead (SubLst);
	  appendType (TT, lTypes);
	  if (TypeSubListTestLast (SubLst))
	    Goon = pk_false;
	  else
	    SubLst = TypeSubListGetTail (SubLst);
	}
    }
}

/** Access to the list of types.
   @param lTypes (out) list of types referenced in the expressions.
   warning : The returned list is sorted using the value of the entry returned by the type manager.
   @return "true" if such a list is not empty, false otherwise. 
*/
boolean ExprManager::getUserTypes (ptlinstances * lTypes)
{
  ptlexpcal lexp;
  // The returned list is sorted using the value of the entry of the Type Table.
  (*lTypes) = NILLINSTANCES;
  if (getExprList (opident, &lexp))
    getUserTypes (lexp, lTypes);
  // if (getExprList(optype, &lexp))  getUserTypes(lexp, lTypes);  removed Jan 2015
  if (getExprList (opvenum, &lexp))
    getUserTypes (lexp, lTypes);
  return ((*lTypes) != NILLINSTANCES);
}

/** It converts a list of expressions to a combed expression. The order of the element is preserved.
   @param list (in) a list of expressions.
   @return the combed expression
*/
ptexpcal ExprManager::convertLinstances2ListExpcal (ptlinstances list)
{
  ptexpcal
    lresult;
  ptexpcal
    id;

  list = ptlinstances_reverse (list);
  lresult = NILEXPCAL;
  while (ptlinstances_get_next (&list, &id))
    lresult = ptexpcal_RIEN_create_binaryListExp (oplident, oper2, id, lresult);
  return lresult;
}

/**
  @param NotTypedId (out) list of signals which are not typed.
  @return true if the output list is not empty.
*/
boolean ExprManager::getListNotTypedId (ptlinstances * NotTypedId)
{
  ptexpcal
    vexp;
  ptlinstances
    ls;

  *NotTypedId = NILLEXPCAL;

  ls = table[opident];
  while (ptlinstances_get_next (&ls, &vexp))
    if (ptexpcal_IsFreeType (vexp))
      *NotTypedId = ptlinstances_insert (*NotTypedId, vexp);
  return ((*NotTypedId) != NILLEXPCAL);
}


/** Splitting a list of expressions.
   @param linst (in) list of expressions.
   @param lNoEvents (in/out)  elements of linst which have not event type are added to lNoEvents (in)
   @param levents (in/out) elements of linst which have event type are added to levents (in)
*/
void
ExprManager::splitEvents (ptlinstances linst, ptlinstances * lNoEvents, ptlinstances * levents)
{
  ptexpcal exp;
  while (ptlinstances_get_next (&linst, &exp))
    {
      if (ptexpcal_IsEvent (exp))
	(*levents) = ptlinstances_insert (*levents, exp);
      else
	(*lNoEvents) = ptlinstances_insert (*lNoEvents, exp);
    }
}

/**
   @return the merge of two lists of references to type: 
   it adds the elements of lt in lTypesCst.
*/
_VIRTUAL_ ptlinstances
ExprManager::mergeTypesCst (ptlinstances lt, ptlinstances lTypesCst)
{
  ptexpcal elem;
  boolean b;
  int classDEBUG, typeDEBUG;

  // printf("DEBUG: entering ExprManager::mergeTypesCst\n");
  while (ptlinstances_get_next (&lt, &elem))
    {
      classDEBUG = ptexpcal_get_operatorClass (elem);
      if (ptexpcal_get_operatorClass (elem) != opertyp)
	b = pk_true;
      else
	{
	  typeDEBUG = getExpType (elem);
	  b = !TypeTypeIsIn (getExpType (elem), lTypesCst);
	}
      if (b)
	lTypesCst = ptlinstances_insert (lTypesCst, elem);
    }
  // printf("DEBUG: entering ExprManager::mergeTypesCst\n");
  return lTypesCst;
}


/**
   @return the list composed of the clock of each signal of a list of signals (lsig) added in lh.
   The state variable are excluded.
*/
ptlinsth ExprManager::addClockSignals (ptlinstances lsig, ptlinsth lh)
{
  ptexpcal vinst;
  boolean b;

  while (ptlinstances_get_next (&lsig, &vinst))
    if (!isStateVar (vinst))
      {
	lh = ptlinsth_sorted_insert_elem_bool (lh, getClock (vinst), &b);
      }
  return lh;
}



/**
   @return the list composed of the clock of each signal of a list of signals (lsig). 
*/
ptlinsth ExprManager::getClockOfSignals (ptlinstances lsig)
{
  return addClockSignals (lsig, NILLINSTH);
}


/**
   Getting signals referenced in a list of expressions. 
   @returns the list of identifiers (signal) referenced in a list of expression. 
*/
ptlinstances ExprManager::getSignalsReferencedInList (ptlinstances liste)
{
  ptlinstances
    l;
  ptexpcal
    e;

  l = NILLINSTANCES;
  while (ptlinstances_get_next (&liste, &e))
    l = ptlinstances_sortedfusion (getSignalsReferencedIn (e), l);
  return l;
}

/**
   @return true if it exists a signal in the list lsig, false otherwise.
   @param lsig (in) a list of signals.
   @param id (out) the identifier of a signal of the operand list (priority to the first input or output signal, or a local if no input/output)
*/
boolean ExprManager::getident (ptlinstances lsig, ptident * id)
{
  ptexpcal
    vsig,
    localid;
  boolean
    found;

  localid = NILEXPCAL;
  found = pk_false;
  // Recherche un signal d'entrée, ou de sortie
  while (ptlinstances_get_next (&lsig, &vsig) && !found)
    {
      if (IsInputFlow (vsig))
	found = pk_true;
      else if (IsOutputFlow (vsig))
	found = pk_true;
      else
	localid = vsig;
    }
  if (!found)
    {
      if (localid == NILEXPCAL)
	return pk_false;
      vsig = localid;
    }
  *id = getOriginalName (vsig);
  return pk_true;
}
