/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_

/**
   @return  the type of the node (node).
   (\see sort-of-nodes )
*/
    noperators NodeManager::get_oper(ptINode node)
{
  return ptINode_get_oper(node);
}

/** @return  the number associated with a node (aNode). */
int NodeManager::get_num(ptINode aNode)
{
  return ptINode_get_num(aNode);
}

/** @return "true" if the type of a node (aNode) is not set, "false" otherwise. */
boolean NodeManager::is_undefined(ptINode aNode)
{
  return ptINode_is_undefined(aNode);
}

/** @return  "true" if the node (aNode) is a node call node, "false" otherwise.
    \see{get_nodeCall_arg, is_outputNodeCall, get_nodeCall_graph}
*/
boolean NodeManager::isNodeCall(ptINode aNode)
{
  return ptINode_is_nodeCall(aNode);
}

/** @return  "true" if the node (aNode) is a delay node, "false" otherwise.
    \see{get_rightOperand, get_leftOperand, get_delay_arg}
*/
boolean NodeManager::is_delay(ptINode aNode)
{
  return ptINode_is_delay(aNode);
}

/** @return  "true" if the node (aNode) is a composed node (with fields), "false" otherwise.
    \see{get_lfields}
*/
boolean NodeManager::is_lfields(ptINode aNode)
{
  return ptINode_satisfies_lfields(aNode);
}

/** @return  "true" if the node (aNode) is a subgraph node node, "false" otherwise. */
boolean NodeManager::is_subgraph(ptINode aNode)
{
  return ptINode_satisfies_subGraph(aNode);
}

/** Getting the  operands of an equation (for delays, see get_delay_arg()) node (aNode).
    - op1 : expression of the left operand;
    - op2 : expression of the rigth operand;
    \see{get_rightOperand, get_leftOperand}
*/
void NodeManager::get_equation_arg(ptINode aNode, ptexpcal * op1, ptexpcal * op2)
{
  ptINode_get_equation_arg(aNode, op1, op2);
}

/** Getting the operands of a delay (aNode):  # resret := sigret $ vret window vfen init vinit #
    - resret : expression of the left operand;
    - sigret : expression of the delayed signal;
    - vret : expression of the delay;
    - vfen : expression of the window;
    - vinit : expression of the initialisation.

    \see{get_rightOperand, get_leftOperand}
*/
void NodeManager::get_delay_arg(ptINode aNode, ptexpcal * resret, ptexpcal * sigret, ptexpcal * vret, ptexpcal * vfen, ptexpcal * vinit)
{
  ptINode_get_delay_arg(aNode, resret, sigret, vret, vfen, vinit);
}

/**
   @return the signals (or fields) defined in a composed node (aNode) associated with a strutured type.
*/
ptlinstances NodeManager::get_lfields_arg(ptINode aNode)
{
  return ptINode_get_lfields(aNode);
}

/** @return  for a read/write node the referenced signal, "nil" expression otherwise. */
ptexpcal NodeManager::get_io_arg(ptINode node)
{
  return ptINode_get_io_arg(node);
}

/** Getting the operands of a procedure call node (aNode).
    - idCall: the name of the procedure call;
    - OrderedInputs: the set of the input signal expression (left combed expression);
    - OrderedOutputs: the set of the output signals (left combed expression);
*/
void NodeManager::get_procCall_arg(ptINode aNode, ptident * idCall, ptexpcal * params, ptexpcal * OrderedInputs, ptexpcal * OrderedOutputs)
{
  ptINode_get_procCall_arg(aNode, idCall, params, OrderedInputs, OrderedOutputs);
}

/**  Getting the operands of a node call node (aNode).
     - vg : the referenced graph;
     - OrderedParams : the set of the parameters (left combed expression);
     - OrderedInputs : the set of the input signals expressions (left combed expression);
     - OrderedOutputs : the set of the output signals (left combed expression);
*/
void NodeManager::get_nodeCall_arg(ptINode aNode, ptDCG * vg, ptexpcal * OrderedParams, ptexpcal * OrderedInputs, ptexpcal * OrderedOutputs)
{
  ptINode_get_args_nodecall(aNode, vg, OrderedParams, OrderedInputs, OrderedOutputs);
}

/**
   @return  "true" if the node (aNode) is an equation (for a delay \see is_delay), "false" otherwise.
   \see{get_rightOperand, get_leftOperand, get_equation_arg}.
*/
boolean NodeManager::is_equation(ptINode aNode)
{
  return ptINode_satisfies_equation(aNode);
}

/**
   @return 
   - "true" if the node (aNode) is a procedure call, id is its name.
   - "false" otherwise, id is not modified.
*/
boolean NodeManager::get_nameCall(ptINode aNode, ptident * id)
{
  return ptINode_get_nameCall(aNode, id);
}

/** 
    @return  "true" if the node (aNode) is a call to a intrinsic function, "false" otherwise.
    These functions are defined in the SIGNAL module SIGNAL_PREDEF.LIB of the exe.dir directory.
*/
boolean NodeManager::is_intrinsicfunc(ptINode aNode)
{
  return ptINode_satisfies_intrinsicfunc(aNode);
}

/** @return 
    - "true" if a node (aNode) is an input produced by the node call returned in nodecall, 
    - "false" otherwise, nodecall is not modified. 
*/
boolean NodeManager::is_outputNodeCall(ptINode aNode, ptINode * nodecall)
{
  return ptINode_is_outputNodeCall(aNode, nodecall);
}

/**
   @return the graph referenced in a node, assumed to be a node call (aNode).
*/
ptDCG NodeManager::get_nodeCall_graph(ptINode aNode)
{
  return ptINode_get_nodeCall_graph(aNode);
}

/**
   @return "true" if a node (aNode) is a procedure node (i.e. a node call without reference to a graph).
*/
boolean NodeManager::isProcessCall(ptINode aNode)
{
  return ptINode_satisfies_procCall(aNode);
}

/** @return 
    - "true" if the name of a node assumed to be a node of a sub-graph (aSubGraph), is defined; returned in id,
    - "false" otherwise, id is not modified.
*/
boolean NodeManager::get_clusterName(ptINode aSubGraph, ptident * id)
{
  return getAttribute(aSubGraph, attrClusterName, (attributeValue *) id);
}

/**
   @return true if a label (vlabel) is associated with a node (aNode), false otherwise.
*/
boolean NodeManager::get_label(ptINode aNode, ptexpcal * vlabel)
{
  return ptINode_get_label(aNode, vlabel);
}

/**
   @return "true" if a node (aNode) denotes a train cluster, "false" otherwise.
*/
boolean NodeManager::is_trainCluster(ptINode aNode)
{
  return ptINode_is_trainCluster(aNode);
}

/** 
    @return 
    - "true" if a node (aNode) denotes a sub-graph; then, le (resp. ls) is the inputs (resp. outputs) of the sub-graph.
    (le and ls are left combed expressions).
    -"false" otherwise, le and ls are not modified.
*/
boolean NodeManager::get_inOutGraphRef(ptINode aNode, ptexpcal * le, ptexpcal * ls)
{
  return ptINode_get_inOutGraphRef(aNode, le, ls);
}

/** ---Dependences managing-----.
    @return the list of clocks depending on a node (aNode), previously set by set_clocksSuccessors.
*/
ptlinsth NodeManager::get_clockSuccessors(ptINode aNode)
{
  return ptINode_get_clockSuccessors(aNode);
}

/** ---Dependences managing-----
    @return the list of successors of a node (aNode), previously set by set_successors.
*/
ptldependence NodeManager::get_successors(ptINode aNode)
{
  return ptINode_get_successors(aNode);
}

/** ---Dependences managing----- 
    @return the list of predecessors (nodes) of a node (aNode), previously set by set_predecessors.
*/
ptldependence NodeManager::get_predecessors(ptINode aNode)
{
  return ptINode_get_predecessors(aNode);
}

/** ---Dependences managing----- 
    @return the list of predecessors (clocks) of a node (aNode), previously set by set_clockPredecessors.
*/
ptlinsth NodeManager::get_clockPredecessors(ptINode aNode)
{
  ptlinsth lhaux, lh;
  boolean b;
  ptinsth vhrepr, vh;

  lh = ptINode_get_clockPredecessors(aNode);
  lhaux = lh;
  while (ptlinsth_get_next(&lhaux, &vh))
    {
      vhrepr = CKM->getRepresentativeFormula(vh);
      if (vhrepr != vh)
	{
	  ptlinsth_sorted_remove_elem(&lh, vh);
	  lh = ptlinsth_sorted_insert_elem_bool(lh, vhrepr, &b);
	}
    }
  set_clockPredecessors(aNode, lh);
  return lh;
}

/** ---Dependences managing-----
    @return the list of iteration precedence associated with a node (aNode) previously set by set_iterationPrecedence.
    (a node n that reference x[?] induces the iteration dependence node(x) -> n)
*/
ptldependence NodeManager::get_iterationPrecedence(ptINode aNode)
{
  return ptINode_get_iterationPrecedence(aNode);
}

/**
   @return 
   - "true" if a node (aSubGraph) is a node of a sub-graph, the !inputNode is the "fictitious" input node of a subgraph (all the nodes of the sub-graph depend on it).
   - "false" otherwise, inputNode is not modified.
*/
boolean NodeManager::get_subGraphInputNode(ptINode aSubGraph, ptINode * inputNode)
{
  return ptINode_get_subGraphInputNode(aSubGraph, inputNode);
}

/** 
    @return  "true" if the node (aNode) denotes a write access to the environment, "false" otherwise.
    Such a node is assigned to each output signal of the graph. It depends on the definition of the output signal.
    \see{get_io_arg}
*/
boolean NodeManager::is_writeNode(ptINode aNode)
{
  return ptINode_satisfies_writeNode(aNode);
}

/** @return  "true" if the node (aNode) denotes a read access to the environment, "false" otherwise.
    Such a node is assigned to each input signal of the graph.
    \see{get_io_arg}
*/
boolean NodeManager::is_readNode(ptINode aNode)
{
  return ptINode_satisfies_readNode(aNode);
}

/** 
    @return  "true" if the node (aNode) denotes a read or a write access to the environment, "false" otherwise.
    \see{is_readNode, is_writeNode}
*/
boolean NodeManager::is_ioNode(ptINode aNode)
{
  if (ptINode_satisfies_readNode(aNode))
    return pk_true;
  return ptINode_satisfies_writeNode(aNode);
}

/**
   @return  "true" if the node ?node is a cell (i.e X:= E cell c), "false" otherwise.
   It should be generalized for X:= (E when c) default ZX  definition.
*/
boolean NodeManager::is_cellNode(ptINode node)
{
  ptexpcal vdedef;

  if (!ptINode_satisfies_equation(node))
    return pk_false;
  if (!ptINode_get_rightOperand(node, &vdedef))
    return pk_false;
  return EM->IsCellExpr(vdedef);
}

/**
   @return the contribution clock of a node. It is available for equation node (node) (i.e. X:= E when H, x:=(E when H) cell K).
   It is (H) the clock at which the expression E contributes to the definition of the signal X.
   It is always included in the clock of the signal X.
*/
ptinsth NodeManager::get_contributionClockEquation(ptINode node)
{
  ptexpcal exp;

  if (!ptINode_get_rightOperand(node, &exp))
    return NILINSTH;		// ne doit pas se produire.
  if (EM->IsCellExpr(exp))
    exp = EM->getWhenOrCellArg(exp);
  return EM->getContributionClock(exp);
}

/** @return the clock of a node (node). */
ptinsth NodeManager::get_nodeClock(ptINode node)
{
  return CKM->getRepresentativeFormula(ptINode_get_nodeClock(node));
}

/** @return
    - "true" if a node (node) is an equation or a delay (X:=E), then vexp is the expression of definition (E) of the node
    - "false" otherwise, vexp is not modified.
*/
boolean NodeManager::get_rightOperand(ptINode node, ptexpcal * vexp)
{
  return ptINode_get_rightOperand(node, vexp);
}

ptexpcal NodeManager::get_rightOperand(ptINode node)
{
  ptexpcal vexp;
  if (ptINode_get_rightOperand(node, &vexp)) return vexp;
  return NILEXPCAL;
}


/** @return  the left expression (outputs) of a node (node).*/
ptexpcal NodeManager::get_leftOperand(ptINode node)
{
  return ptINode_get_leftOperand(node);
}

/**
   @return  "true" if a node (aNode) denotes a set of partial definition node (set of ::= ), "false" otehrwise.
   \see{get_setOfPartialDefinitions_arg, becomes_setOfPartialDefinitions}
*/
boolean NodeManager::is_setOfPartialDefinitions(ptINode aNode)
{
  return ptINode_satisfies_setOfPartialDefinitions(aNode);
}

/**
   @return  the list of the nodes associated with a node (aNode), assumed to denote a set of partial definition node.
   \see{is_setOfPartialDefinitions, becomes_setOfPartialDefinitions}
*/
ptlINode NodeManager::get_setOfPartialDefinitions_arg(ptINode aNode)
{
  return ptINode_get_SetPartDefinition(aNode);
}

/** @return  "true" if a node (aNode) is a partial definition node (::= operator), "false" otherwise.
    \see{get_rightOperand, get_leftOperand}
*/
boolean NodeManager::is_partialDefinition(ptINode aNode)
{
  return ptINode_satisfies_partialDefinition(aNode);
}

/** @return  "true" if a node (aNode) is a "default value" partial definition node (::= defaultvalue operator), "false" otherwise.
    \see{get_rightOperand, get_leftOperand}
*/
boolean NodeManager::is_partialDefinitionDefaultValue(ptINode aNode)
{
  return isMarked(aNode, PART_DEFAULT_DEFINITION);
}

/**
   @return the safety property associated with a node (safetyPropertyType) (opsafe by default).
*/
safetyPropertyType NodeManager::get_safetyProperty(ptINode n)
{
  return ptINode_get_safetyProperty(n);
}

/**
   @return "true" if a node (node) is a delay or a node of a delay cluster, "false" oterhwise.
*/
boolean NodeManager::is_delayNodeOrDelayClusterNode(ptINode node)
{
  ptexpcal resret, sigret, vret, vfen, vinit;

  if (is_delay(node))
    {
      get_delay_arg(node, &resret, &sigret, &vret, &vfen, &vinit);
      return (vret != NILEXPCAL);
    }
  return existPredefinedPragma(getPragmas(node), PRAGMA_DELAYCLUSTER);
}

/**
   @return "true" if is exists a back annotation (vt) to the abstract tree produced by the parser associated with a node (n), "false" otherwise (vt is not 
   modified).
*/
BackAnnotationType NodeManager::getBackAnnotation(ptINode n)
{
  return ptINode_getBackAnnotation(n);
}

// ARRAY Node

/**
   @return the index of an index-node (left part of I := [ lower bound, upper bound, step ])
*/
ptexpcal NodeManager::get_index(ptINode node)
{
  return ptINode_get_leftOperand(node);
}

/**
   @return the definition part of an index-node (right part of I := [ lower bound, upper bound, step])
*/
ptexpcal NodeManager::get_index_def(ptINode node)
{
  ptexpcal e;
  if (!ptINode_get_rightOperand(node, &e))
    return NILEXPCAL;
  return e;
}

/** @return  "true" if the node (aNode) is a array process node, "false" otherwise..
    \see{get_es_array, get_args_array}
*/
boolean NodeManager::is_array(ptINode aNode)
{
  return ptINode_satisfies_array(aNode);
}

/**
   @return 
   - "true" if a node (aNode) is an array node, intNode is then its "internal" node associated with (all nodes of the array depend on it),
   - "false" otherwise,intNode is not modified.
*/
boolean NodeManager::get_array_inputNode(ptINode aNode, ptINode * intNode)
{
  return ptINode_get_array_inputNode(aNode, intNode);
}

/**
   @return the index node of an assumed array node (node). \see { get_index, get_index_def}.
*/
ptINode NodeManager::get_indexNode(ptINode aNode)
{
  return ptINode_get_array_index(aNode);
}

/**
   Getting the arguments of an assumed array node (aNode).
   - indexNode : its index,
   - vWith : the \b with part of the array
   - vBody : the \b body part of the array
*/
void NodeManager::get_array_arg(ptINode aNode, ptINode * indexNode, ptgfd * vWith, ptgfd * vBody)
{
  ptINode_get_array_arg(aNode, indexNode, vWith, vBody);
}

/**
   Getting the input signals (inputs) and outputs signals (outputs) of an assumed array node (aNode).
   inputs, outputs are comber expressions.
*/
void NodeManager::get_array_io(ptINode aNode, ptexpcal * inputs, ptexpcal * outputs)
{
  ptINode_get_array_io(aNode, inputs, outputs);
}

/**
   @return the body part of an assumed array node (aNode).
*/
ptgfd NodeManager::get_array_bodyGraph(ptINode aNode)
{
  return ptINode_get_array_bodyGraph(aNode);
}

/**
   @return the initial part (with part) of an assumed array node (aNode).
*/
ptgfd NodeManager::get_array_initGraph(ptINode aNode)
{
  return ptINode_get_array_initGraph(aNode);
}

/**
   @return 
   - "true" is a node (n) is an array node, then 
   - ln is the set of nodes of its body part
   - lh is the set of clocks of its body part
   - "false" otherwise, ln and lh are not modified.
*/
boolean NodeManager::get_array_bodyDefinitions(ptINode n, ptlINode * ln, ptlinsth * lh)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_array(n))
    return pk_false;
  vgfd = ptINode_get_bodySubGraph(n);
  (*ln) = ptgfd_get_nodes(vgfd);
  (*lh) = ptgfd_get_clocks(vgfd);
  return pk_true;
}

/**
   @return 
   - "true" is a node (n) is an array node, then 
   - ln is the set of nodes of its initial part
   - lh is the set of clocks of its initial part
   - "false" otherwise, ln and lh are not modified.
*/
boolean NodeManager::get_array_initDefinitions(ptINode n, ptlINode * ln, ptlinsth * lh)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_array(n))
    return pk_false;
  vgfd = ptINode_get_initSubGraph(n);
  (*ln) = ptgfd_get_nodes(vgfd);
  (*lh) = ptgfd_get_clocks(vgfd);
  return pk_true;
}

/**
   @return "true" if a node (aNode) is in a composed node (array, subgraph), "false" otherwise.
*/
boolean NodeManager::is_inComposedNode(ptINode aNode)
{
  return ptINode_satisfies_InComposedNode(aNode);
}

/**
   @return true if the node is in an array (cn), "false" otherwise (cn is assigned to the composed node, or NILINODE).
*/
boolean NodeManager::getInArrayNode(ptINode node, ptINode * cn)
{
  *cn = NILINODE;
  if (!get_inComposedNode(node, cn))
    return pk_false;
  return is_array(*cn);
}

/**
   @return 
   - "true" if a node (aNode) is in a composed node (cNode) (ie array, subgraph),
   - "false" otherwise, cNode is NILINODE.

   \see { getInArrayNode()}
*/
boolean NodeManager::get_inComposedNode(ptINode aNode, ptINode * cNode)
{
  return ptINode_get_inComposedNode(aNode, cNode);
}

/** @return true if the operand node(vnode)  is defined by the intrinsic with "name" as identifier. */
boolean NodeManager::is_intrinsicNode(ptINode vnode, ptident name)
{
  ptident id;
  if (! isProcessCall(vnode)) return pk_false;
  if (! is_intrinsicfunc(vnode)) return pk_false;
  if (! get_nameCall(vnode, &id)) return pk_false;
  return eq_idents(id, name);
}

/**
   @return 
   - "true" is a node (node) is a node associated with a sub-graph, then 
   - ln is the set of nodes of the sub-graph
   - lh is the set of clocks of the sub-graph
   - "false" otherwise, ln and lh are not modified.
*/
boolean NodeManager::get_subGraphDefinitions(ptINode node, ptlINode * ln, ptlinsth * lh)
{
  ptgfd vgfd;
  if (!ptINode_satisfies_subGraph(node))
    return pk_false;
  vgfd = ptINode_get_subGraph(node);
  (*ln) = ptgfd_get_nodes(vgfd);
  (*lh) = ptgfd_get_clocks(vgfd);
  return pk_true;
}

/**
   @return 
   - "true" is a node (n) is a node associated with a sub-graph, 
   then, it returns the projection information(set previously by set_subGraphProjectedSystemInfos)
   - vhemb: the branching clocks of the nodes/clocks of the sub-graph
   - projectedSystemClocks: the list of the clocks used to have an endochronous hierarchy for 
   the sub-graph (parametrization until vhemb, for each node/clock of the sub-graph).
   - "false" otherwise, ln and lh are not modified.
*/
boolean NodeManager::get_subGraphProjectedSystemInfos(ptINode node, ptinsth * vhemb, ptlinsth * projectedSystemClocks)
{
  ptgfd vgfd;

  if (!ptINode_satisfies_subGraph(node))
    return pk_false;
  vgfd = ptINode_get_subGraph(node);
  (*vhemb) = ptgfd_get_projectionInfos(vgfd, projectedSystemClocks);
  return pk_true;
}

/**
   @returns the utility clock associated with a node (n).
*/
boolean NodeManager::getUtilityClock(ptINode n, ptinsth * h)
{
  attributeValue pt;
  if (!getAttribute(n, UTILITYCLOCK, (attributeValue *) h))
    return pk_false;
  (*h) = CKM->getRepresentativeFormula(*h);
  return pk_true;
}

/**
   @return "true" if the utility clock (h) of a signal (vsig) in a subgraph (sg) is computed, "false" otherwise.
*/
boolean NodeManager::getUtilityClockInSubGraph(ptINode sg, ptexpcal vsig, ptinsth * h)
{
  ptTablePointerPointer vtable;
  TablePointerPointerType::iterator it;

  if (!getAttribute(sg, UTILITYCLOCKINSUBGRAPH, (attributeValue *) & vtable))
    return pk_false;
  it = vtable->find(vsig);
  if (it == vtable->end())
    return pk_false;
  (*h) = (ptinsth) (*it).second;	// (*vtable)[vsig];
  return pk_true;
}

/**
   @returns the Assignment clock associated with a node (n).
*/
boolean NodeManager::getAssignmentClock(ptINode n, ptinsth * h)
{
  if (!getAttribute(n, ASSIGNMENTCLOCK, (attributeValue *) h))
    return pk_false;
  (*h) = CKM->getRepresentativeFormula(*h);
  return !CKM->eqClock(*h, hundef);
}


/**
   @return "true" is the operand node is a default value node (x::=defaultvalue E), "false" otherwise.
*/
boolean NodeManager::isDefaultValueNode(ptINode node)
{
  return isMarked(node, PART_DEFAULT_DEFINITION);
}

/**
   @return "true" is the operand node is a default value node such that X::=defaultvalue X$1 init vinit, "false" otherwise.
*/
boolean NodeManager::isDefaultValueNodeDelay(ptINode node, ptexpcal * vinit)
{
  ptexpcal vRpart;

  if (!isMarked(node, PART_DEFAULT_DEFINITION))
    return pk_false;
  if (!ptINode_get_rightOperand(node, &vRpart))
    return pk_false;
  if (EM->IsCellExpr(vRpart))
    return pk_false;
  vRpart = EM->getBasicExpr(vRpart);
  if (EM->getOperator(vRpart) == opmele)
    return pk_false;
  return EM->isDelayOneOf(vRpart, ptINode_get_leftOperand(node), vinit);
}

/** 
    Code generation.

    @return "true" if the node node must be generated, "false" otherwise.
    @param node (in) can be annotated by one of the attribute  : GencodeMark, substituteGenCodeMark
*/
boolean NodeManager::mustBeGenerated (ptINode node)
{
  ptexpcal x, vce;

  if (!isMarked (node, GencodeMark)) return pk_false;
  if (isMarked (node, substituteGenCodeMark)) return pk_false;
  if (is_equation (node))
    {
      x = get_leftOperand (node);
      if (EM->getAttribute (x, OPTIMIZED_ATTRIBUTE, (attributeValue *) & vce)) return pk_false;
      if (get_rightOperand (node, &vce))
	{ // case x = x (by unification)
	  vce = EM->getBasicExpr (vce);
	  if (x == EM->getExpressionToGenerate (vce)) return pk_false;
	}
      return !EM->isAggregatedSignal (x);
    }
  if (is_delay (node)) return !EM->getAttribute (get_leftOperand (node), OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *) & vce);
  return pk_true;
}



/**
   @return true if a node (node) is internal to another one (nodeup), false otherwise.
   Remark: it returns true if node is nodeup.
 */
boolean NodeManager::isInternalNodeOf(ptINode node, ptINode nodeup)
{
  ptINode cNode;
  if ( node == nodeup) return pk_true;
  while ( get_inComposedNode(node, &cNode))
    {
      if (cNode == nodeup) return pk_true;
      node=cNode;
    }
  return pk_false;
}



boolean NodeManager::isSendNode(ptINode node)
{
  return is_intrinsicNode(node,getSendId());
}

boolean NodeManager::isReceiveNode(ptINode node)
{
  return is_intrinsicNode(node,getReceiveId());
}


boolean NodeManager::getMemorizedSignal(ptINode aNode, ptexpcal *vsig)
{
  ptexpcal sigret, vret, vfen, vinit;

  if (is_delay(aNode))
    {
      get_delay_arg(aNode, &sigret, vsig, &vret, &vfen, &vinit);
      return pk_true;
    }
  if (is_cellNode(aNode))
    {
      *vsig = get_leftOperand(aNode);
      return pk_true;
    }
  return pk_false;
}
