/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PUBLIC_
/**
   Constructor.
*/
    NodeManager::NodeManager()
{

  GdependencesList = NILLDEPENDENCE;

  // mettre en oeuvre du hash-coding pour la gestion des listes triées suivantes.
  // Actuellement : utilisation de tête/queue de liste.
  AllNodes = NILLINODE;
  AllNodesEnd = NILLINODE;
  deletedNodes = NILLINODE;
  deletedNodesEnd = NILLINODE;	// voir utilité
  lread = NILLINODE;
  lreadEnd = NILLINODE;
  lwrite = NILLINODE;
  lwriteEnd = NILLINODE;
  ldelay = NILLINODE;
  ldelayEnd = NILLINODE;
  ldefsig = NILLINODE;
  ldefsigEnd = NILLINODE;
  lArray = NILLINODE;
  lArrayEnd = NILLINODE;
  listrefextern = NILLINODE;
  listrefexternEnd = NILLINODE;
  lNodeCall = NILLINODE;
  lNodeCallEnd = NILLINODE;
  lSubGraph = NILLINODE;
  lSubGraphEnd = NILLINODE;
  lBundStruct = NILLINODE;
  lBundStructEnd = NILLINODE;
  ldefpart = NILLINODE;
  ldefpartEnd = NILLINODE;
}

void
 NodeManager::initialize(ptClockManager ckm, ptExprManager em, ptTraceabilityManager tm)
{
  EM = em;
  CKM = ckm;
  hundef = CKM->getUndefClock();
  the_nullClock_NM_copy = CKM->getNullClock();
  TM = tm;
}

/**
   Destructor.
*/
NodeManager::~NodeManager()
{
  ptlINode ln;
  ptINode node;
  ptldependence ldep;

  // printf(" removing a dcg in... ~NodeManager()\n"); fflush(stdout);
  free_dependences(GdependencesList);
  GdependencesList->ptldependence_free();

  ln = AllNodes;
  while (ptlINode_get_elem(&ln, &node))
    {
      ptlinsth_free(ptINode_get_clockPredecessors(node));
      ptlinsth_free(ptINode_get_clockSuccessors(node));
      ptINode_free(node);
    }
}

/** @return true if the output list of nodes is not empty.
   @param list (out) list of all the nodes of the graph. */
boolean NodeManager::getAllNodes(ptlINode * list)
{
  *list = AllNodes;
  return (AllNodes != NILLINODE);
}

/** @return the list of the nodes taged by a pragma (pragmaName) */
ptlINode NodeManager::getTaggedNodes(int pragmaName)
{
  ptlINode ls;
  ptINode aNode;
  ptlINode ltaggednodes;

  // we look at the node call/function call nodes. 
  ltaggednodes = NILLINODE;
  if (getProcCalls(&ls))
    while (ptlINode_get_elem(&ls, &aNode))
      if (existPredefinedPragma(getPragmas(aNode), pragmaName))
	ptlINode_insert_elem(&ltaggednodes, aNode);

  if (getNodeCalls(&ls))
    while (ptlINode_get_elem(&ls, &aNode))
      if (existPredefinedPragma(getPragmas(aNode), pragmaName))
	ptlINode_insert_elem(&ltaggednodes, aNode);

  return ltaggednodes;
}

/**
   @return the list of the external calls.
   The caller may free the returned list.
*/
boolean NodeManager::getExternalCalls(ptlINode * list)
{
  ptlINode ln;
  ptINode n;

  *list = NILLINODE;
  if (getProcCalls(&ln))
    *list = ptlINode_Copy(ln);
  if (getNodeCalls(&ln))
    while (ptlINode_get_elem(&ln, &n))
      if (get_nodeCall_graph(n) != NILDCG)
	ptlINode_insert_elem(list, n);
  return ((*list) != NILLINODE);
}

/**
   @return the dependences of the graph. 
   The returned list must not be modified.
*/
ptldependence NodeManager::getDependences()
{
  return GdependencesList;
}

/**
   @return true if the list ln of the equations (X:=E, excluding delays) of the graph is not empty, false otherwise.
*/
boolean NodeManager::getEquations(ptlINode * ln)
{
  return getNodesList(opdefsig, ln);
}

/**
   @return "true" if the list ln of the delays (X:=Y$N window M) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getDelays(ptlINode * ln)
{
  return getNodesList(opret, ln);
}

/**
   @return "true" if the list ln of the external calls (node call without spec) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getProcCalls(ptlINode * ln)
{
  return getNodesList(opappe, ln);
}

/**
   @return "true" if the list ln of the external calls (node call with spec) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getNodeCalls(ptlINode * ln)
{
  return getNodesList(opnodecall, ln);
}

/**
   @return "true" if the list ln of the input nodes (node assigned to input signals) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getInputs(ptlINode * ln)
{
  return getNodesList(oplread, ln);
}

/**
   @return "true" if the list ln of the output nodes (node for producing the values of the ouputs signals) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getOutputs(ptlINode * ln)
{
  return getNodesList(oplwrite, ln);
}

/**
   @return "true" if the list ln of the iterate nodes (iterate I (|  |) structure ) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getIterates(ptlINode * ln)
{
  return getNodesList(opmoti, ln);
}

/**
   @return "true" if the list ln of the subgraph nodes ( label:: (|  |) structure ) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getSubGraphs(ptlINode * ln)
{
  return getNodesList(opsubgraph, ln);
}

/**
   @return "true" if the list ln of the bundle nodes ( bundle (|  |) spec C  structure ) of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getBundles(ptlINode * ln)
{
  return getNodesList(opbund, ln);
}

/**
   @return "true" if the list ln of the groups of partial equation nodes ( a node is composed of a list of partial definitions: x::=E1, x::=E2,...x::=En assigned to a signal x) 
   of the graph is not empty, "false" otherwise.
*/
boolean NodeManager::getGroupsOfPartialEquations(ptlINode * ln)
{
  return getNodesList(opldefpart, ln);
}

void NodeManager::printTracebilityInfos(ptident graphname)
{
  ofstream ofile;
  ptrepstr vfilename;
  ptident vid;
  ptINode node;
  ptlINode ln;

  vid = makeIdent_ident_string(graphname, "NM");
  if (sys_suffixedFilePathInUserOutputDir(ident2repstr(vid), "txt", &vfilename))
    {
      ofile.open(repstr2string(vfilename));
      ln = AllNodes;
      while (ptlINode_get_elem(&ln, &node))
	{
	  ofile << " Num Node: " << get_num(node);
	  BackAnnotationType ba = getBackAnnotation(node);
	  if (ba == NILBACKANNOTATION)
	    {
	      ofile << " No BACKANNOTATION, oper = " << get_oper(node);
	    }
	  else
	    {
	      ntrees vAbsTree = ba->getAbsTree();
	      ntrees vCallTree = ba->getCallTree();
	      ofile << "( " << vAbsTree << ", " << vCallTree << " )" << "oper = " << get_oper(node);
	    }
	  ofile << endl;
	}
      ofile.close();
    }
}

void NodeManager::getIOUnsafeExternalCalls(ptlinstances *linputs, ptlinstances *loutputs)
{
  ptlINode ln;
  ptINode node;
  ptexpcal lparam, lent, lsort;
  ptDCG vg;
  ptident id;
  ptlinstances lin, lout;

  lin = NILLINSTANCES;
  lout = NILLINSTANCES;

  if (getProcCalls(&ln))
    while (ptlINode_get_elem(&ln, &node))
      {
	if (! is_readNode(node))
	  if (! is_writeNode(node))
	    if (get_safetyProperty (node) != opsafe)
	      {
		lin = ptlinstances_sortedfusion (getInodelInputFlows(node), lin);
		lout = ptlinstances_sortedfusion (getInodelOutputFlows(node), lout);
	      }
      }
  if (getNodeCalls(&ln))
    while (ptlINode_get_elem(&ln, &node))
      if (get_safetyProperty (node) != opsafe)
	{
	  lin = ptlinstances_sortedfusion (getInodelInputFlows(node) , lin);
	  lout = ptlinstances_sortedfusion (getInodelOutputFlows(node), lout);
	}
  *linputs = lin;
  *loutputs = lout;
}
