/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   @return the current element of a list (vlist)
*/
extern ptINode ptlINode_getElem(ptlINode vlist)
{
  return vlist->elem;
}

/**
   @return the previous list-element of vlist.
*/
extern ptlINode ptlINode_getPred(ptlINode vlist)
{
  return vlist->pred;
}

/**
   @return the tail of a list (vlist)
*/
extern ptlINode ptlINode_getNext(ptlINode vlist)
{
  return vlist->next;
}

/**
   @param vlist (in) a list of nodes
   @param velem (out) the current node pointed by vlist
   @param pr (out) the previous element of vlist
*/
extern void ptlINode_get_pred(ptlINode vlist, ptINode * velem, ptlINode * pr)
{
  *velem = vlist->elem;
  *pr = vlist->pred;
}

/**
   @return the last list-element of ln.
*/
extern ptlINode ptlINode_get_last(ptlINode ln)
{
  ptlINode pred;
  ptINode x;

  pred = ln;
  while (ptlINode_get_elem(&ln, &x))
    if (ln != NILLINODE)
      pred = ln;
  return pred;
}

/**
   @return "true" if the operand list is not empty, "false" otherwise.
   @param vlist (out)  set to next element of the list.
   @param node (out)  current element of the list
*/
extern boolean ptlINode_get_elem(ptlINode * vlist, ptINode * node)
{
  if ((*vlist) == NILLINODE)
    return pk_false;
  (*node) = (*vlist)->elem;
  *vlist = (*vlist)->next;
  return pk_true;
}

/** 
    @return the list {aNode} + y, element is inserted in head position. 
*/
extern ptlINode ptlINode_append_head(ptINode aNode, ptlINode y)
{
  ptlINode vsucc;

  vsucc = new linstr();
  vsucc->elem = aNode;
  vsucc->next = y;
  vsucc->pred = NILLINODE;
  if (y != NILLINODE)
    y->pred = vsucc;
  return vsucc;
}

/**
   It inserts in an assumed sorted list (deblist) an element (velem).
   @return "true" if the element is inserted in the list, "false" when it is already in it.
*/
extern boolean ptlINode_insert_elem(ptlINode * deblist, ptINode velem)
{
  boolean vres;
  int numero, cnum;
  ptlINode curlist, succ;

  if (*deblist == NILLINODE)
    {
      *deblist = ptlINode_append_head(velem, NILLINODE);
      return pk_true;
    }
  numero = ptINode_get_num(velem);
  cnum = ptINode_get_num((*deblist)->elem);
  if (numero < cnum)
    {
      *deblist = ptlINode_append_head(velem, *deblist);
      curlist = *deblist;
      return pk_true;
    }
  curlist = *deblist;
  while (cnum > -1)
    if (cnum == numero)
      {
	cnum = -1;
	vres = pk_false;
      }
    else
      {
	succ = curlist->next;
	if (succ == NILLINODE)
	  {
	    succ = ptlINode_append_head(velem, NILLINODE);
	    curlist->next = succ;
	    succ->pred = curlist;
	    cnum = -1;
	    vres = pk_true;
	  }
	else
	  {
	    cnum = ptINode_get_num(succ->elem);
	    if (cnum > numero)
	      {
		succ = ptlINode_append_head(velem, succ);
		curlist->next = succ;
		succ->pred = curlist;
		cnum = -1;
		vres = pk_true;
	      }
	  }
	curlist = succ;
      }
  return vres;
}

/**
   @return "true" if an element (i) is in a list (li), "false" otherwise.
   The element is removed form the list
*/
extern boolean ptlINode_remove_elem(ptlINode * li, ptINode i)
{
  boolean res;
  ptlINode linst;

  linst = *li;
  res = pk_false;
  while (linst != NILLINODE)
    if (linst->elem == i)
      {
	res = pk_true;
	if (linst == (*li))
	  {
	    *li = (*li)->next;
	    if ((*li) != NILLINODE)
	      (*li)->pred = NILLINODE;
	  }
	else
	  {
	    linst->pred->next = linst->next;
	    if (linst->next != NILLINODE)
	      linst->next->pred = linst->pred;
	  };
	linst = NILLINODE;
      }
    else
      linst = linst->next;
  return res;
}

/** It removes the head element of a list (l). l is modified. */
extern void ptlINode_removehead(ptlINode * l)
{
  if ((*l) != NILLINODE)
    {
      *l = (*l)->next;
      if ((*l) != NILLINODE)
	(*l)->pred = NILLINODE;
    }
}

/** @return the fusion of two sorted list l1 and l2 with side effects on l1 list. */
extern ptlINode ptlINode_fusion(ptlINode l1, ptlINode l2)
{
  ptINode n;
  if (l1 == NILLINODE)
    return l2;
  while (ptlINode_get_elem(&l2, &n))
    ptlINode_insert_elem(&l1, n);
  return l1;
}

/** @return "true" if an element (node) is in a list (vlist), "false" otherwise. */
extern boolean ptlINode_isIn(ptlINode vlist, ptINode node)
{
  ptINode x;

  while (ptlINode_get_elem(&vlist, &x))
    if (x == node)
      return pk_true;
  return pk_false;
}

/**
   @return 
   - "true" if an element (node) is in a list (vlist), pn is the previous element of the list
   - "false" otherwise, pn is undefined.
*/
extern boolean ptlINode_isIn_getPrevious(ptlINode vlist, ptINode node, ptINode * pn)
{
  ptINode x;
  ptlINode l;
  boolean bfirst;

  if (vlist == NILLINODE)
    return pk_false;
  l = vlist;
  bfirst = pk_true;
  while (ptlINode_get_elem(&vlist, &x))
    {
      if (x == node)
	{
	  if (!bfirst)
	    *pn = l->elem;
	  else
	    *pn = NILINODE;
	  return pk_true;
	}
      if (!bfirst)
	l = l->next;
      bfirst = pk_false;
    }
  return pk_false;
}

/** @return "true" if te intersection of two lists (l1, l2) is empty, "false" otherwise. */
extern boolean ptlINode_intersection_isEmpty(ptlINode l1, ptlINode l2)
{
  ptINode n1;
  while (ptlINode_get_elem(&l1, &n1))
    if (ptlINode_isIn(l2, n1))
      return pk_false;
  return pk_true;
}

/**
   It appends an element (elem) at the end of a list (ln).
   @return the new list.
*/
extern ptlINode ptlINode_append_elem(ptlINode ln, ptINode elem)
{
  ptlINode l, last;

  if (ln == NILLINODE)
    return ptlINode_append_head(elem, NILLINODE);
  last = ptlINode_get_last(ln);
  l = ptlINode_append_head(elem, NILLINODE);
  last->next = l;
  l->pred = last;
  return ln;
}

/** 
   @return the result of the concatenating of a list (lnSec) to another one (ln).
   The elements are appended at the head of ln (side effect).
*/
extern ptlINode ptlINode_catenate_pre(ptlINode ln, ptlINode lnSec)
{
  ptINode n;
  while (ptlINode_get_elem(&lnSec, &n))
    ln = ptlINode_append_head(n, ln);
  return ln;
}

/**
   @return the result of the concatenating of a list (lnSec) to another one (ln).
   The elements are appended at the end of ln (side effect).
*/
extern ptlINode ptlINode_catenate_post(ptlINode ln, ptlINode lnSec)
{
  ptlINode lCopy, last;

  if (lnSec == NILLINODE)
    return ln;
  lCopy = ptlINode_Copy(lnSec);
  if (ln == NILLINODE)
    return lCopy;
  last = ptlINode_get_last(ln);
  last->next = lCopy;
  lCopy->pred = last;
  return ln;
}

/**
   @return the number of elements of a list (ln)
*/
extern int ptlINode_get_size(ptlINode ln)
{
  int nb;
  ptINode x;

  nb = 0;
  while (ptlINode_get_elem(&ln, &x))
    nb++;
  return nb;
}

/**
   @return a list (ln) is which an element (elem) is inserted before another one (before). When the before element  is not 
   in the list, elem is appended at the end of ln.
*/
extern ptlINode ptlINode_append_elem_before(ptlINode ln, ptINode elem, ptINode before)
{
  ptlINode list, ln1, pred;
  boolean done;
  ptINode n;

  list = ptlINode_append_head(elem, NILLINODE);
  if (ln == NILLINODE)
    return list;
  ln1 = ln;
  pred = ln;
  done = pk_false;
  while (ptlINode_get_elem(&ln, &n) && !done)
    if (n == before)
      {
	ptlInode_link(list, pred);
	done = pk_true;
      }
    else if (ln == NILLINODE)
      {
	ptlInode_link(pred, list);
	done = pk_true;
      }
    else
      pred = ln;

  if (list->pred == NILLINODE)
    return list;
  return ln1;
}

/** @return a copy of the list l. */
extern ptlINode ptlINode_Copy(ptlINode l)
{
  ptlINode lcopy;
  ptINode x;

  lcopy = NILLINODE;
  while (ptlINode_get_elem(&l, &x))
    lcopy = ptlINode_append_elem(lcopy, x);	// if (ptlINode_insert_elem(&lcopy, x));
  return lcopy;
}

/**
   @return a list (ln) in which all the elements MARKED by m are removed.
*/
extern ptlINode ptlINode_removeMarked(ptlINode ln, typeMark m)
{
  ptlINode ln1;
  ptINode n;

  ln1 = ln;
  while (ptlINode_get_elem(&ln, &n))
    if (ptINode_satisfies_marked(n, m))
      ptlINode_remove_elem(&ln1, n);
  return ln1;
}

/**
   @return a list (ln) in which all the elements UNMARKED by m are removed.
*/
extern ptlINode ptlINode_remove_unmarked(ptlINode ln, typeMark m)
{
  ptlINode ln1;
  ptINode n;

  ln1 = ln;
  while (ptlINode_get_elem(&ln, &n))
    if (!ptINode_satisfies_marked(n, m))
      ptlINode_remove_elem(&ln1, n);
  return ln1;
}

/** Garbage collector.
    Operand list (l) becomes free.
*/
extern void ptlINode_free(ptlINode l)
{
  ptlINode ngclinodes;

  if (l != NILLINODE)
    {
      ngclinodes = l;
      while (l->next != NILLINODE)
	l = l->next;
      l->next = gclinodes;
      gclinodes = ngclinodes;
    }
}


