/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_ 

ptident TraceabilityManager::getClockId(ptinsth vh)
{
  ptexpcal vsig;
  ptident id;
  ptrepstr repstrResult;

  id = CKM->getName(vh);
  if (! CKM->isClockRepresentedBySignal (vh, &vsig) && ! CKM->isClockUnifiedWithSignal(vh, &vsig) ) return id;
  if (EM->IsEvent(vsig))  return EM->getName(vsig); // id;
  makeRepstr_string_ident ("^", id, &repstrResult);
  return repstr2ident(repstrResult);
}

ptinsth TraceabilityManager::getClassOf(ptinsth h, ptlinsth lh)
{
  ptinsth vh;
  while (ptlinsth_get_next (&lh, &vh))
    if ( CKM->getRepresentativeFormula(vh) == CKM->getRepresentativeFormula(h))
      return vh;
  
  return hundef;
}


ptexpcal TraceabilityManager::getClassOf(ptexpcal s, ptlinstances lexp)
{
  ptexpcal newcurExp, curExp;
  while (ptlinstances_get_next (&lexp, &curExp))
    if (EM->getIndirection(curExp, &newcurExp))
      if (newcurExp == s) return curExp;
  return NILEXPCAL;
}

ptlinsth TraceabilityManager::mkClockFormulaToKeep(ptlinsth lhTovisit, ptlinsth lh, ptClockNamesTable clocknametable, ptlinsth lUnusedClocks)
{
  ptinsth nextFormula, vh;

  while (ptlinsth_get_next(&lhTovisit, &vh))
    if (! ptlinsth_isin_elem(lUnusedClocks, vh)) 
      {
	lh = ptlinsth_insert_elem (lh, vh);
	(*clocknametable)[vh] = getClockId(vh);
	nextFormula = CKM->getNextFormula(vh, pk_true);
	while (nextFormula != vh)
	  {
	    lh = ptlinsth_insert_elem (lh, nextFormula);
	    nextFormula = CKM->getNextFormula(nextFormula, pk_true);
	  }
      }
  return lh;
}


boolean TraceabilityManager::getIdentifiersTablePrevious(ptClockNamesTable *idtablePrevious)
{
  pointer x;
  ptStack S;

  if (!modeKeepSubstitutions) return pk_false;
  if (ClockNamesTableStackRef == NILSTACK) return pk_false;
  if (ClockNamesTableStack != ClockNamesTableStackRef) return pk_false;
  if (! stackPrevious(ClockNamesTableStack, &S)) return pk_false;
  if (! stackTop(S, &x)) return pk_false;
  (*idtablePrevious) = (ptClockNamesTable) x;
  return pk_true;

}


// DEBUG
void TraceabilityManager::printInfoClockRew(ptinsth vh, ptinsth vhprev)
{
  ptident id_vhprev, id_vh;

  id_vh = CKM->getName(vh);
  if (vhprev == hundef) 
    cout << " Previous clock of " << vh << " ( " << ident2string(id_vh) << " )" << " is undefined !" << endl;
  else
    {
      id_vhprev = CKM->getName(vhprev);
      if ( ! eq_idents(id_vh, id_vhprev))
	cout << " " << ident2string(id_vh) << " was " << ident2string(id_vhprev) << endl;
      /**
	 if (vhprev == vh)
	 cout << " No modification for " << vh << " ( " << ident2string(id_vh) << " )" << endl;
	 else
	 cout << vh << "( " << ident2string(id_vh) << " )" << " was " << vhprev << "( " << ident2string(id_vhprev) << " )"<< endl;
      */
    }
}

// DEBUG
void TraceabilityManager::printInfoClockRew2(ptinsth vh, ptinsth vhnew)
{
  ptident id_vhnew, id_vh;

  if ( vhnew != vh )
    {
      id_vh = CKM->getName(vh);
      id_vhnew = CKM->getName(vhnew);
      if ( ! eq_idents(id_vh, id_vhnew))
	cout << "(= " << ident2string(id_vh) << " " << ident2string(id_vhnew) << ")"  << endl;
    }
}


// DEBUG
void TraceabilityManager::printmodifiedSinceLastMarkedFunction()
{
  ptinsth vhnew, vhprev, vh;
  ptlinsth lh;
  pointer x;

  // For debug.  
  return;

  if (modeKeepSubstitutions)
    {
      if (ClockSystemsStack == NILSTACK)
	cout << " >>>>>> Empty stack !!!" << endl;
      else
	{
	  if (ClockSystemsStackRef == NILSTACK)
	    cout << " >>>>>> No ClockSystem Reference " << endl;
	  else
	    if (ClockSystemsStack == ClockSystemsStackRef)
	      cout << " >>>>>>  No modifications" << endl;
	    else
	      {
		if (stackTop(ClockSystemsStackRef, &x))
		  {
		    lh = (ptlinsth) x;
		    ptlinsth lh_save = lh;
		    cout << "; Mapping of variables" << endl;
		    while (ptlinsth_get_next (&lh, &vh))
		      {
			printInfoClockRew2(vh, CKM->getRepresentativeFormula(vh));
		      }
		  }
	      }
	}
    }
}

void TraceabilityManager::addIdentToTable(ptlinsth lhTovisit, ptClockNamesTable clocknametable)
{
  ptinsth nextFormula, vh;

 while (ptlinsth_get_next(&lhTovisit, &vh))
    {
      (*clocknametable)[vh] = getClockId(vh);

      /* nextFormula = CKM->getNextFormula(vh, pk_true);
	 while (nextFormula != vh) {
	 (*clocknametable)[nextFormula] = getClockId(nextFormula);
	 nextFormula = CKM->getNextFormula(nextFormula, pk_true);
	 } */
      }
}


ptlINode TraceabilityManager::getMemorizedNodes(ptlinstances lsig)
{
  ptlINode ln;
  ptexpcal vsig, vdedef, vmemo;
  ptINode node;

  ln = NILLINODE;
  
  while (ptlinstances_get_next (&lsig, &vsig))
    {
      node = EM->getNode (vsig);
      if (NM->getMemorizedSignal(node, &vmemo))
	ptlINode_insert_elem(&ln, node);
    }
  return ln;
}


ptlinstances TraceabilityManager::removeUnused(ptlinstances lsig, ptlINode lUnusedNodes)
{
  ptlinstances ls;
  ptexpcal vsig;
  ptINode node;

  return lsig;

  ls = lsig;
  while (ptlinstances_get_next (&ls, &vsig))
    {
      node = EM->getNode (vsig);
      if (ptlINode_isIn(lUnusedNodes, node)) lsig = ptlinstances_remove_exp ( lsig, vsig);
    }
  return lsig;
}
