/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/***
    |**
    |**                                                    File TypeGarbage_i.c
    |**						       	            **|
    |**						       	            **|
    |**						       	            **|
    |**	            THIS FILE CONTAINS Type Garbage                 **|
    |**		           LOCAL FUNCTIONS                          **|
    |**								    **|
    |*********|*********|*********|*********|*********|*********|*********/

#define macroTytestmarked(tyty) ((macroTygetclass(tyty)) < 0)
#define macroTysetmarked(tyty)\
 ( macroTygetclass(tyty) = (-(macroTygetclass(tyty)+1)))
#define macroTysetunmarked(tyty)\
 ( macroTygetclass(tyty) = (-(macroTygetclass(tyty)+1)))
#define macroTygetunmarkedclass(tyty) (-(macroTygetclass(tyty)+1))

/* The following macros are used to avoid double marking of type in shared
   type lists */
#define macrotestmarkeduserType(tyty) (tyty < (-PredefNumber))
#define macrogetmarkType(tyty) (-(tyty + PredefNumber+1))
#define macrogetunmarkType(tyty) (-(tyty + PredefNumber+1))


/*********|*********|*********|*********|*********|*********|
	  |**			       	                          **|
	  |**	       	Initial Marking     	                  **|
	  |**			       	                          **|
	  |*********|*********|*********|*********|*********|*********/

void RecursiveMarkTypeTerm (TypeType TT);

static void
MarkTypeList (TypeList TLst)
  /*---------------************------------------------*/
{
  if (!(macroTypeListtestempty (TLst)))
    {
      BoolType Goon;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  RecursiveMarkTypeTerm (macroTypeSubListgethead (SubLst));
	  if (macroTypeSubListtestlast (SubLst))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
    };
}

/*********|*********|*********|*********|*********|*********|
	  |**			       	                          **|
	  |**    	    Next index computing for marked types         **|
	  |**			       	                          **|
	  |** static int XXXSetFuture(TypeType TT, int *AvailableIndex)
	  |**	returns the future index for type TT of class XXX
	  |**			       	                          **|
	  |**			       	                          **|
	  |**			       	                          **|
	  |*********|*********|*********|*********|*********|*********/

static int RecursiveSetFutureTypeTerm (TypeType TT, int *AvailableIndex);

static void
ListofTypeSetFuture (TypeList TLst, int *AvailableIndex)
{
  if (!(macroTypeListtestempty (TLst)))
    {
      BoolType Goon;
      TypeType TT;
      TypeSubList SubLst;
      SubLst = macroTypeListgetfirst (TLst);
      Goon = pk_true;
      do
	{
	  TT = RecursiveSetFutureTypeTerm (macroTypeSubListgethead (SubLst), AvailableIndex);
	  macroTypeSubListsethead (SubLst, TT);
	  if (macroTypeSubListtestlast (SubLst))
	    Goon = pk_false;
	  else
	    SubLst = macroTypeSubListgettail (SubLst);
	}
      while (Goon);
    };
}

static int
AnoTestandSetFut (AnonymousTypeRepr * Descr, int *AvailableIndex)
{
  TypeType Future;
  Future = (macroDegetmyself (Descr));
  if (macrotestmarkeduserType (Future))
    return (macrogetunmarkType (Future));
  else
    {
      Future = (*AvailableIndex);
      macroDesetmyself (Descr, macrogetmarkType (*AvailableIndex));
      *AvailableIndex = (*AvailableIndex) + 1;
      return (Future);
    };
}

static BoolType
AnonymousTestUnsetFuture (AnonymousTypeRepr * Descr, TypeType * RTT)
{
  *RTT = (macroDegetmyself (Descr));
  if (macrotestmarkeduserType (*RTT))
    {
      *RTT = macrogetunmarkType (*RTT);
      return pk_false;
    }
  else
    return pk_true;
}

static int
AnonymousSetFuture (AnonymousTypeRepr * Descr, int *AvailableIndex)
{
  TypeType Future;
  Future = (*AvailableIndex);
  macroDesetmyself (Descr, macrogetmarkType (*AvailableIndex));
  *AvailableIndex = (*AvailableIndex) + 1;
  return (Future);
}

static int
ExternSetFuture (TypeType TT, int *AvailableIndex)
{
  return (AnoTestandSetFut ((AnonymousTypeRepr *) (macroTyExternSTree_getDescriptionr (TT)), AvailableIndex));
}

static int
EnumerationSetFuture (TypeType TT, int *AvailableIndex)
{
  return (AnoTestandSetFut ((AnonymousTypeRepr *) (macroTyEnumerationSTree_getDescriptionr (TT)), AvailableIndex));
}

static int
SubrangeSetFuture (TypeType TT, int *AvailableIndex)
{
  return (AnoTestandSetFut ((AnonymousTypeRepr *) (macroTySubrangeSTree_getDescriptionr (TT)), AvailableIndex));
}

static int
BitsSetFuture (TypeType TT, int *AvailableIndex)
{
  return (AnoTestandSetFut ((AnonymousTypeRepr *) (macroTyBitsSTree_getDescriptionr (TT)), AvailableIndex));
}

static int
ArrayMatrixSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  ArrayMatrixTypeRepr *Descr;
  Descr = macroTyarraymatrixSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroArraygetterm (Descr), AvailableIndex);
      macroArraysetterm (Descr, Future);
      Current = macroArraygetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroArraysetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroArraysetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
TupleSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  TupleTypeRepr *Descr;

  Descr = macroTytupleSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      ListofTypeSetFuture (macroTuplegettypelst (Descr), AvailableIndex);
      Current = macroTuplegetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroTuplesetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroTuplesetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	}
    }
  return Future;
}

static int
StructureSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  StructureTypeRepr *Descr;
  Descr = macroTystructureSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroStructuregetargument (Descr), AvailableIndex);
      macroStructuresetargument (Descr, Future);
      Current = macroStructuregetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroStructuresetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroStructuresetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
BundleSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  BundleTypeRepr *Descr;
  Descr = macroTybundleSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroBundlegetfields (Descr), AvailableIndex);
      macroBundlesetfields (Descr, Future);
      Current = macroBundlegetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroBundlesetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroBundlesetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
LambdaSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  LambdaTypeRepr *Descr;
  Descr = macroTylambdaSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroLambdagetterm (Descr), AvailableIndex);
      macroLambdasetterm (Descr, Future);
      Current = macroLambdagetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroLambdasetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroLambdasetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
AnySetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  AnyTypeRepr *Descr;
  Descr = macroTyanySTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Current = macroAnygetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroAnysetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroAnysetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
NamedSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  MetaNamedTypeRepr *Descr;
  Descr = macroTymetanamedSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroNamedgetterm (Descr), AvailableIndex);
      macroNamedsetterm (Descr, Future);
      Current = macroMetaNamedgetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroMetaNamedsetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroMetaNamedsetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
MetaSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  MetaNamedTypeRepr *Descr;
  Descr = macroTymetanamedSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Current = macroMetaNamedgetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroMetaNamedsetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroMetaNamedsetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
NodeSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  NodeTypeRepr *Descr;
  Descr = macroTynodeSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroNodegetdomain (Descr), AvailableIndex);
      macroNodesetdomain (Descr, Future);
      Future = RecursiveSetFutureTypeTerm (macroNodegetimage (Descr), AvailableIndex);
      macroNodesetimage (Descr, Future);
      Current = macroNodegetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroNodesetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroNodesetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
GenericNodeSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  GenericNodeTypeRepr *Descr;
  Descr = macroTygenericnodeSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroGenericNodegetargument (Descr), AvailableIndex);
      macroGenericNodesetargument (Descr, Future);
      Future = RecursiveSetFutureTypeTerm (macroGenericNodegetterm (Descr), AvailableIndex);
      macroGenericNodesetterm (Descr, Future);
      Current = macroGenericNodegetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroGenericNodesetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroGenericNodesetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
MetaNodeSetFuture (TypeType TT, int *AvailableIndex)
{
  TypeType Future, Current;
  MetaNodeTypeRepr *Descr;
  Descr = macroTymetanodeSTree_getDescriptionr (TT);
  if (AnonymousTestUnsetFuture ((AnonymousTypeRepr *) (Descr), &Future))
    {
      Future = RecursiveSetFutureTypeTerm (macroMetaNodegetargument (Descr), AvailableIndex);
      macroMetaNodesetargument (Descr, Future);
      Future = RecursiveSetFutureTypeTerm (macroMetaNodegetterm (Descr), AvailableIndex);
      macroMetaNodesetterm (Descr, Future);
      Current = macroMetaNodegetcurrent (Descr);
      if (Current == TT)
	{
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	  macroMetaNodesetcurrent (Descr, Future);
	}
      else
	{
	  Future = RecursiveSetFutureTypeTerm (Current, AvailableIndex);
	  macroMetaNodesetcurrent (Descr, Future);
	  Future = AnonymousSetFuture ((AnonymousTypeRepr *) (Descr), AvailableIndex);
	};
    };
  return Future;
}

static int
RecursiveSetFutureTypeTerm (TypeType TT, int *AvailableIndex)
  /*--************------------------------
    Recursively set type Index to their future values;
    Returns the next value of TT.
    Returns in AvailableIndex the next free place in the type table
    This procedure has important side effect and cannot be used without a
    following CleanTypeTable
  */
{
  TypeClass ClCl;

  if (macroTytestpredef (TT))
    return TT;
  if (macrotestmarkeduserType (TT))
    return macrogetunmarkType (TT);
  if macroTytestmarked
    (TT) ClCl = macroTygetunmarkedclass (TT);
  else
    ClCl = macroTygetclass (TT);
  switch (ClCl)
    {
    case ExternClass:
      return (ExternSetFuture (TT, AvailableIndex));
    case EnumerationClass:
      return (EnumerationSetFuture (TT, AvailableIndex));
    case SubrangeClass:
      return (SubrangeSetFuture (TT, AvailableIndex));
    case BitsClass:
      return (BitsSetFuture (TT, AvailableIndex));
    case ArrayClass:
    case MatrixClass:
      return (ArrayMatrixSetFuture (TT, AvailableIndex));
    case NamedClass:
      return (NamedSetFuture (TT, AvailableIndex));
    case LambdaClass:
      return (LambdaSetFuture (TT, AvailableIndex));
    case AnyClass:
      return (AnySetFuture (TT, AvailableIndex));
    case MetaClass:
      return (MetaSetFuture (TT, AvailableIndex));
    case TupleClass:
      return (TupleSetFuture (TT, AvailableIndex));
    case StructureClass:
      return (StructureSetFuture (TT, AvailableIndex));
    case BundleClass:
      return (BundleSetFuture (TT, AvailableIndex));
    case NodeClass:
      return (NodeSetFuture (TT, AvailableIndex));
    case GenericNodeClass:
      return (GenericNodeSetFuture (TT, AvailableIndex));
    case MetaNodeClass:
      return (MetaNodeSetFuture (TT, AvailableIndex));
    }
  return TT;
}


static void
SetFutureWhoamIType (int *AvailableIndex, int max)
  /*--------- *******************----------------------------*/
{
  int k, aux;

  for (k = 0; k < max; k++)
    if (macroTytestmarked (k))
      aux = RecursiveSetFutureTypeTerm (k, AvailableIndex);
}

/*********|*********|*********|*********|*********|*********|
	  |**			       	                          **|
	  |**	       	Type Table compression                    **|
	  |**			       	                          **|
	  |*********|*********|*********|*********|*********|*********/

static BoolType
AnonymouTesttoSetFuture (TypeType TT, TypeType * Res)
{
  *Res = macroDegetmyself (macroTySTree_getDescriptionr (AnonymousTypeRepr, TT));
  if (macrotestmarkeduserType (*Res))
    {
      (*Res) = macrogetunmarkType (*Res);
      return pk_true;
    }
  else
    return (pk_false);
}

static void
RecursiveGarbageLocal (TypeType Origine)
  /*---------------************------------------------*/
{
  TypeType target;
  TypeClass ClCl;
  AnonymousTypeRepr *Descr;
  if (macroTytestmarked (Origine))
    {
      /* Origine will be kept in the table */
      if (AnonymouTesttoSetFuture (Origine, &target))
	{
	  /* Its next index is target and it has to be set in the correct slot */
	  Descr = macroTySTree_getDescriptionr (AnonymousTypeRepr, Origine);
	  if (Origine != target)
	    {
	      /* the descriptor must be moved */
	      ClCl = macroTygetclass (Origine);
	      macroTysetclass (Origine, macroTygetunmarkedclass (Origine));
	      RecursiveGarbageLocal (target);
	      macroTysetdescr (target, Descr);
	      macroTysetclass (target, ClCl);
	    };
	  macroDesetmyself (Descr, target);
	}
    }
}


static void
TypeGarbageLocal (int max)
  /*---------------************------------------------*/
{
  int k;
  for (k = 0; k < max; k++)
    RecursiveGarbageLocal (k);
}


/*********|*********|*********|*********|*********|*********|
	  |**			       	                          **|
	  |**	             Type Table cleaning                  **|
	  |**			       	                          **|
	  |*********|*********|*********|*********|*********|*********/

static void
TypeListCleanRegistered (int max)
  /*---------------************------------------------*/
{
  int k;
  for (k = 0; k < max; k++)
    macroTysetunmarked (k);
  macroTTDsetnext (max);
}
