/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/***
    |**
    |**
    |**                                                    File TypeManager_g.c
    |**    							            **|
    |**    							            **|
    |**	      THIS FILE CONTAINS T Y P E REPRESENTATION             **|
    |**		     GLOBAL FUNCTIONS                               **|
    |**	       						            **|
    |*********|*********|*********|*********|*********|*********|*********/

/*********|*********|*********|*********|*********|*********|
	  |**                 GLOBAL TEST FUNCTIONS                 **|
	  |*********|*********|*********|*********|*********|*********/

/***ITM Delivers an access to the Category of x type */
/*
 */
extern TypeClass
TypeGetClass (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return SimpleClass;
  else
    return (macroTygetclass (TT));
}

extern BoolType
TypeTestParameter (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    return (macroCltestParameter (macroTygetclass (TT)));
}

extern BoolType
TypeRecTestandGetCurrent (TypeType TT, TypeType * RTT)
  /*----------***********************--------------------------*/
{
  *RTT = TT;
  if (macroTytestpredef (TT))
    return pk_false;
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	case MatrixClass:
	  return (ArrayMatrixRecTestandGetCurrent (TT, RTT));
	case NamedClass:
	  return (MetaNamedRecTestandGetCurrent (TT, RTT));
	case LambdaClass:
	  return (LambdaRecTestandGetCurrent (TT, RTT));
	case AnyClass:
	  return (AnyRecTestandGetCurrent (TT, RTT));
	case MetaClass:
	  return (MetaNamedRecTestandGetCurrent (TT, RTT));
	case TupleClass:
	  return (TupleRecTestandGetCurrent (TT, RTT));
	case StructureClass:
	  return (StructureRecTestandGetCurrent (TT, RTT));
	case BundleClass:
	  NotImplemented ("TestandGetCurrent for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	  return (NodeRecTestandGetCurrent (TT, RTT));
	case GenericNodeClass:
	  return (GenericNodeRecTestandGetCurrent (TT, RTT));
	case MetaNodeClass:
	  return (MetaNodeRecTestandGetCurrent (TT, RTT));
	}
    }
  return pk_false;
}

extern BoolType
TypeTestGeneric (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return pk_false;
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	case MatrixClass:
	  return TypeTestGeneric (ArrayMatrixTypeGetTerm (TT));
	case NamedClass:
	  return TypeTestGeneric (NamedTypeGetTerm (TT));
	case LambdaClass:
	case AnyClass:
	case MetaClass:
	  return pk_true;
	case TupleClass:
	  return (TypeListTestGeneric (TupleTypeGetTypeLst (TT)));
	case StructureClass:;
	  return (TypeListTestGeneric (StructureTypeGetTypeLst (TT)));
	case BundleClass:
	  return (TypeListTestGeneric (StructureTypeGetTypeLst (BundleTypeGetTerm (TT))));
	  NotImplemented ("Test boolean for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	  {
	    TypeList TLst1, TLst2;
	    TLst1 = NodeTypeGetInputTypeLst (TT);
	    TLst2 = NodeTypeGetOutputTypeLst (TT);
	    return ((TypeListTestGeneric (TLst1)) || (TypeListTestGeneric (TLst2)));
	  };
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_true;
	}
    };
  return pk_false;
}

extern BoolType
TypeTestLogical (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftestlogical (TT));
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	case MatrixClass:
	  return TypeTestLogical (ArrayMatrixTypeGetTerm (TT));
	case NamedClass:
	  return TypeTestLogical (NamedTypeGetTerm (TT));
	case LambdaClass:
	  return TypeTestLogical (LambdaTypeGetTerm (TT));
	case AnyClass:
	  if (AnyTypeTestDefined (TT))
	    return TypeTestLogical (AnyTypeGetCurrent (TT));
	  else
	    return pk_false;
	case MetaClass:
	  if (MetaTypeTestDefined (TT))
	    return TypeTestLogical (MetaTypeGetCurrent (TT));
	  else
	    return pk_false;
	case TupleClass:
	  return (TypeListTestLogical (TupleTypeGetTypeLst (TT)));
	case StructureClass:;
	  return (TypeListTestLogical (StructureTypeGetTypeLst (TT)));
	case BundleClass:
	  // Lbesnard Nov 2000
	  return (TypeListTestLogical (StructureTypeGetTypeLst (BundleTypeGetTerm (TT))));
	  NotImplemented ("Test boolean for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_false;
	}
    };
  return pk_false;
}

extern BoolType
TypeTestIntNumber (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftestintnumber (TT));
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	  return pk_false;
	case SubrangeClass:
	  return pk_true;
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	case MatrixClass:
	  return TypeTestIntNumber (ArrayMatrixTypeGetTerm (TT));
	case NamedClass:
	  return TypeTestIntNumber (NamedTypeGetTerm (TT));
	case LambdaClass:
	  return TypeTestIntNumber (LambdaTypeGetTerm (TT));
	case AnyClass:
	  if (AnyTypeTestDefined (TT))
	    return TypeTestIntNumber (AnyTypeGetCurrent (TT));
	  else
	    return pk_false;
	case MetaClass:
	  if (MetaTypeTestDefined (TT))
	    return TypeTestIntNumber (MetaTypeGetCurrent (TT));
	  else
	    return pk_false;
	case TupleClass:
	  return (TypeListTestIntNumber (TupleTypeGetTypeLst (TT)));
	case StructureClass:;
	  return (TypeListTestIntNumber (StructureTypeGetTypeLst (TT)));
	case BundleClass:
	  // lbesnard Nov 2000
	  return (TypeListTestIntNumber (StructureTypeGetTypeLst (BundleTypeGetTerm (TT))));
	  NotImplemented ("Test integer for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_false;
	}
    };
  return pk_false;
}

extern BoolType
TypeTestRealNumber (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftestrealnumber (TT));
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	case MatrixClass:
	  return TypeTestRealNumber (ArrayMatrixTypeGetTerm (TT));
	case NamedClass:
	  return TypeTestRealNumber (NamedTypeGetTerm (TT));
	case LambdaClass:
	  return TypeTestRealNumber (LambdaTypeGetTerm (TT));
	case AnyClass:
	  if (AnyTypeTestDefined (TT))
	    return TypeTestRealNumber (AnyTypeGetCurrent (TT));
	  else
	    return pk_false;
	case MetaClass:
	  if (MetaTypeTestDefined (TT))
	    return TypeTestRealNumber (MetaTypeGetCurrent (TT));
	  else
	    return pk_false;
	case TupleClass:
	  return (TypeListTestRealNumber (TupleTypeGetTypeLst (TT)));
	case StructureClass:;
	  return (TypeListTestRealNumber (StructureTypeGetTypeLst (TT)));
	case BundleClass:
	  // lbesnard Nov 2000
	  return (TypeListTestRealNumber (StructureTypeGetTypeLst (BundleTypeGetTerm (TT))));
	  NotImplemented ("Test boolean for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_false;
	}
    };
  return pk_false;
}

extern BoolType
TypeTestComplexNumber (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftestcomplexnumber (TT));
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	case MatrixClass:
	  return TypeTestComplexNumber (ArrayMatrixTypeGetTerm (TT));
	case NamedClass:
	  return TypeTestComplexNumber (NamedTypeGetTerm (TT));
	case LambdaClass:
	  return TypeTestComplexNumber (LambdaTypeGetTerm (TT));
	case AnyClass:
	  if (AnyTypeTestDefined (TT))
	    return TypeTestComplexNumber (AnyTypeGetCurrent (TT));
	  else
	    return pk_false;
	case MetaClass:
	  if (MetaTypeTestDefined (TT))
	    return TypeTestComplexNumber (MetaTypeGetCurrent (TT));
	  else
	    return pk_false;
	case TupleClass:
	  return (TypeListTestComplexNumber (TupleTypeGetTypeLst (TT)));
	case StructureClass:;
	  return (TypeListTestComplexNumber (StructureTypeGetTypeLst (TT)));
	case BundleClass:
	  return (TypeListTestComplexNumber (StructureTypeGetTypeLst (BundleTypeGetTerm (TT))));
	  NotImplemented ("Test boolean for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_false;
	}
    };
  return pk_false;
}

extern BoolType
TypeTestNumerical (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftestnumerical (TT));
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	  return pk_false;
	case SubrangeClass:
	  return pk_true;
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	case MatrixClass:
	  return TypeTestNumerical (ArrayMatrixTypeGetTerm (TT));
	case NamedClass:
	  return TypeTestNumerical (NamedTypeGetTerm (TT));
	case LambdaClass:
	  return TypeTestNumerical (LambdaTypeGetTerm (TT));
	case AnyClass:
	  if (AnyTypeTestDefined (TT))
	    return TypeTestNumerical (AnyTypeGetCurrent (TT));
	  else
	    return pk_false;
	case MetaClass:
	  if (MetaTypeTestDefined (TT))
	    return TypeTestNumerical (MetaTypeGetCurrent (TT));
	  else
	    return pk_false;
	case TupleClass:
	  return (TypeListTestNumerical (TupleTypeGetTypeLst (TT)));
	case StructureClass:;
	  return (TypeListTestNumerical (StructureTypeGetTypeLst (TT)));
	case BundleClass:
	  // lbesnard Nov 2000
	  return (TypeListTestNumerical (StructureTypeGetTypeLst (BundleTypeGetTerm (TT))));
	  NotImplemented ("Test integer for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_false;
	}
    };
  return pk_false;
}


extern BoolType
TypeTestText (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftesttext (TT));
  switch (macroTygetclass (TT))
    {
    case ExternClass:
    case EnumerationClass:
      return pk_false;
    case SubrangeClass:
      return pk_true;		// ?
    case BitsClass:
      return pk_false;
    case ArrayClass:
    case MatrixClass:
      return TypeTestText (ArrayMatrixTypeGetTerm (TT));
    case NamedClass:
      return TypeTestText (NamedTypeGetTerm (TT));
    case LambdaClass:
      return TypeTestText (LambdaTypeGetTerm (TT));

    case AnyClass:
      if (AnyTypeTestDefined (TT))
	return TypeTestText (AnyTypeGetCurrent (TT));
      return pk_false;

    case MetaClass:
      if (MetaTypeTestDefined (TT))
	return TypeTestText (MetaTypeGetCurrent (TT));
      return pk_false;

    case TupleClass:
      return pk_false;		// must be completed
    case StructureClass:;
      return pk_false;		// must be completed
    case BundleClass:
      NotImplemented ("Test Text for Bundle. Returns false");
      return pk_false;
    case NodeClass:
    case GenericNodeClass:
    case MetaNodeClass:
      return pk_false;
    }
  return pk_false;
}


extern BoolType
TypeTestPartOrder (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftestpartorder (TT));
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	  return pk_false;
	case EnumerationClass:
	case SubrangeClass:
	case BitsClass:
	  return pk_true;
	case ArrayClass:
	case MatrixClass:
	  return TypeTestPartOrder (ArrayMatrixTypeGetTerm (TT));
	case NamedClass:
	  return TypeTestPartOrder (NamedTypeGetTerm (TT));
	case LambdaClass:
	  return TypeTestPartOrder (LambdaTypeGetTerm (TT));
	case AnyClass:
	  if (AnyTypeTestDefined (TT))
	    return TypeTestPartOrder (AnyTypeGetCurrent (TT));
	  else
	    return pk_false;
	case MetaClass:
	  if (MetaTypeTestDefined (TT))
	    return TypeTestPartOrder (MetaTypeGetCurrent (TT));
	  else
	    return pk_false;
	case TupleClass:
	  return (TypeListTestPartOrder (TupleTypeGetTypeLst (TT)));
	case StructureClass:;
	  return (TypeListTestPartOrder (StructureTypeGetTypeLst (TT)));
	case BundleClass:
	  // lbesnard Nov 2000
	  return (TypeListTestPartOrder (StructureTypeGetTypeLst (BundleTypeGetTerm (TT))));
	  NotImplemented ("Test Partial ordering for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_false;
	}
    };
  return pk_false;
}

extern BoolType
TypeTestOrder (TypeType TT)
  /*---------------************------------------------*/
{
  if (macroTytestpredef (TT))
    return (macroPredeftestorder (TT));
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	  return pk_false;
	case EnumerationClass:
	  return pk_true;
	case SubrangeClass:
	  return pk_true;
	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  return pk_false;
	case NamedClass:
	  return TypeTestOrder (NamedTypeGetTerm (TT));
	case LambdaClass:
	  return TypeTestOrder (LambdaTypeGetTerm (TT));
	case AnyClass:
	  if (AnyTypeTestDefined (TT))
	    return TypeTestOrder (AnyTypeGetCurrent (TT));
	  else
	    return pk_false;
	case MetaClass:
	  if (MetaTypeTestDefined (TT))
	    return TypeTestOrder (MetaTypeGetCurrent (TT));
	case TupleClass:
	case StructureClass:
	case BundleClass:
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return pk_false;
	}
    };
  return pk_false;
}

extern BoolType
TypeReduction (TypeType TT, TypeType * RTT)
  /*---------------************------------------------*/
{
  *RTT = TT;
  if (macroTytestpredef (TT))
    return pk_false;
  else
    {
      switch (macroTygetclass (TT))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	case BitsClass:
	  return pk_false;
	case ArrayClass:
	  return (ArrayReduction (TT, RTT));
	case MatrixClass:
	  return (MatrixReduction (TT, RTT));
	case NamedClass:
	  return (NamedReduction (TT, RTT));
	case LambdaClass:
	  return (LambdaReduction (TT, RTT));
	case AnyClass:
	  return (AnyReduction (TT, RTT));
	case MetaClass:
	  return (MetaReduction (TT, RTT));
	case TupleClass:
	  return (TupleReduction (TT, RTT));
	case StructureClass:
	  return (StructureReduction (TT, RTT));
	case BundleClass:
	  NotImplemented ("Type Reduction for Bundle. Returns false");
	  return pk_false;
	case NodeClass:
	  return (NodeReduction (TT, RTT));
	case GenericNodeClass:
	  return (GenericNodeReduction (TT, RTT));
	case MetaNodeClass:
	  return (MetaNodeReduction (TT, RTT));
	}
    };
  return pk_false;
}

extern int
TypeCastType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both Reduced
    returns:  
    -1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (macroTytestpredef (Target))
    return CastType2PredefType (Source, Target);
  switch (macroTygetclass (Target))	// switch (macroTygetclass(Source)) (Juillet 2001 LB)
    {
    case ExternClass:
      if (TypeEqAdr (Source, Target))
	return 1;
      else
	return -1;
    case EnumerationClass:
      return CastType2EnumerationType (Source, Target);
    case SubrangeClass:
      return CastType2SubrangeType (Source, Target);
    case BitsClass:
      return CastType2BitsType (Source, Target);
    case ArrayClass:
      return CastType2ArrayType (Source, Target);
    case MatrixClass:
      return CastType2MatrixType (Source, Target);
      //    case NamedClass: return TypeCastType(NamedTypeGetTerm(Source), Target); (Juillet 2001 LB)
    case NamedClass:
      return TypeCastType (Source, NamedTypeGetTerm (Target));
    case LambdaClass:
      return 0;			/* Type should be non generic */
    case AnyClass:
      return 0;
    case MetaClass:
      return 0;
    case TupleClass:
      return CastType2TupleType (Source, Target);
    case StructureClass:
      return CastType2StructureType (Source, Target);
    case BundleClass:
      NotImplemented ("Type Cast for Bundle. Returns -1");
      return -1;
    case NodeClass:
    case GenericNodeClass:
    case MetaNodeClass:
      return -1;
    }
  return -1;
}

extern TypeType
TypeUpb (TypeType TT1, TypeType TT2)
  /*---------------************------------------------*/
{
  NotImplemented ("TypeUpb. Returns first argument");
  return TT1;
}



/**
   @returns the predefined undefined type
*/
extern TypeType
TypeTypeGetUndefinedType ()
{
  return UndefinedType;
}

/** @return 
    - pk_true if the type (t) is a named type (vstr)
    - false otherwise (vstr not modified).
*/
extern boolean
TypeTypeGetName (TypeType t, ptrepstr * vstr)
{
  if (TypeTestExternClass (t))
    {
      *vstr = ExternTypeGetName (t);
      return pk_true;
    }
  if (TypeTestMetaClass (t))
    {
      *vstr = MetaTypeGetName (t);
      return pk_true;
    }
  if (TypeTestEnumerationClass (t))
    {
      *vstr = EnumerationTypeGetName (t);
      return pk_true;
    }
  if (TypeTestNamedClass (t))
    {
      *vstr = NamedTypeGetName (t);
      return pk_true;
    }
  return pk_false;
}

/**
   @return 
   - for a named type (t), the associated type definition 
   - t otherwise.
 */
extern TypeType
TypeTypeGetUnamedType (TypeType t)
{
  if (!TypeTestNamedClass (t))
    return t;
  return TypeTypeGetUnamedType (NamedTypeGetTerm (t));
}

extern const char *
TypeTypePredefTypeStringValue (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a predef type */
{
  if ((((long int) TT) > 0) || (((long int) TT) <= -nbPredefinedType))
    return "Not a predefined type";
  return SigTypeStr[-TT];
}

extern boolean
TypeType_hasAnyType (TypeType TT)
{
  return ((macroTygetclass (TT)) == AnyClass);
}

extern const char *
TypeTypeStringCategory (TypeType TT)
  /*---------------************------------------------
    TT is assumed to be a predef type */
{
  // printf("DEBUG entering TypeTypeStringCategory, TT is %li, nbPredefinedType is %li\n", (long int)TT,  nbPredefinedType); fflush(stdout);
  if (((long int) TT) <= 0)
    return TypeTypePredefTypeStringValue (TT);
  switch (macroTygetclass (TT))
    {
    case ExternClass:
      return "extern";
    case EnumerationClass:
      return "enum";
    case SubrangeClass:
      return "subrange";
    case BitsClass:
      return "bits";
    case ArrayClass:
      return "array";
    case MatrixClass:
      return "matrix";
    case NamedClass:
      return "named";
    case LambdaClass:
      return "lambda";
    case AnyClass:
      return "any";
    case MetaClass:
      return "meta";
    case TupleClass:
      return "tuple";
    case StructureClass:
      return "struct";
    case BundleClass:
      return "bundle";
    case NodeClass:
      return "node";
    case GenericNodeClass:
      return "generic node";
    case MetaNodeClass:
      return "meta node";
    default:
      return "unknown category";
    }
}

extern void
TypeManagerReset ()
{
  ClassTableInitiate ();
  TypeListInitiate ();
  NameListInitiate ();
  UndefinedType = AnyTypeCreate ();
}

extern void
TypeManagerInit (int InitialSize, int MaxSize, int ResizingFactor)
{
  TypeTableInitiate (&RefTypeTableDescriptor, InitialSize, MaxSize, ResizingFactor, sizeof (TypeItemType), TypeTableName);
  TypeManagerReset ();
}
