/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
/***                                                     File TypeManager_i.c
							 |**						       	 ====================
							 |**						       	            **|
							 |**						       	            **|
							 |**						       	            **|
							 |**						       	            **|
							 |**						       	            **|
							 |**	      	THIS FILE CONTAINS T Y P E REPRESENTATION           **|
							 |**		           LOCAL FUNCTIONS                          **|
							 |**								    **|
							 |*********|*********|*********|*********|*********|*********|*********/

BoolType TypeRecTestandGetCurrent (TypeType TT, TypeType * RTT);
int TypeCastType (TypeType Target, TypeType Source);
static int CastType2EnumerationType (TypeType Source, TypeType Target);

static int
CastType2PredefType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is a predefined Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can always be converted to Target 
  */
{
  if (macroTytestpredef (Source))
    {
      if (macroPredefTestCast (Source, Target)) return 1;
      return -1;
    }
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	  return -1;

	case EnumerationClass:
	  if (macroPredefTestCast (Target, ShortType)) return 1; /* Il faudrait vérifier le nombre d'éléments */
	  return -1;

	case SubrangeClass:
	  if (macroPredefTestCast (Source, IntegerType)) return 1; /* Il faudrait vérifier le nombre d'éléments */
	  return -1;

	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  return -1;

	case NamedClass:
	  return CastType2PredefType (NamedTypeGetTerm (Source), Target);

	case LambdaClass:
	  return 0;		/* Types should be concrete */

	case AnyClass:
	  if (AnyTypeTestDefined (Source)) return CastType2PredefType (AnyGetDefinition (Source), Target);
	  return -1;

	case MetaClass:
	  if (MetaTypeTestDefined (Source)) return CastType2EnumerationType (MetaTypeGetCurrent (Source), Target);
	  return -1;

	case TupleClass:
	case StructureClass:
	case BundleClass:
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return -1;
	}
    }
  return -1;
}

static int
CastType2EnumerationType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is an Enumeration Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (macroTytestpredef (Source))
    if (macroPredeftestintnumber (Source))
      return 0;
    else
      return -1;
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	  return -1;
	case EnumerationClass:
	  {
	    int Res;
	    NameList LstS;
	    NameList LstT;
	    LstS = EnumerationTypeGetNameLst (Source);
	    LstT = EnumerationTypeGetNameLst (Target);
	    Res = NameListUnsortedCompare (LstS, LstT);
	    if (Res == -1)
	      return 1;
	    else if (Res == 0)
	      return 1;
	    else if (Res == 1)
	      return 0;
	    else
	      return -1;
	  }
	case SubrangeClass:
	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  return -1;
	case NamedClass:
	  return CastType2EnumerationType (NamedTypeGetTerm (Source), Target);
	case LambdaClass:
	  return 0;		/* Types should be concrete */
	case AnyClass:
	  if (AnyTypeTestDefined (Source))
	    return CastType2EnumerationType (AnyGetDefinition (Source), Target);
	  else
	    return -1;
	case MetaClass:
	  if (MetaTypeTestDefined (Source))
	    return CastType2EnumerationType (MetaTypeGetCurrent (Source), Target);
	  else
	    return -1;
	case TupleClass:
	case StructureClass:
	case BundleClass:
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	default:
	  return -1;
	}
    }
}

int
CastType2SubrangeType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is an Subrange Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (macroTytestpredef (Source))
    if (macroPredeftestintnumber (Source))
      return 0;
    else
      return -1;
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	  return -1;
	case EnumerationClass:
	case SubrangeClass:
	  NotImplemented ("Casting to Subrange not implemented. Returns -1");
	  return -1;
	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  return -1;
	case NamedClass:
	  return CastType2SubrangeType (NamedTypeGetTerm (Source), Target);
	case LambdaClass:
	  return 0;		/* Types should be concrete */
	case AnyClass:
	  if (AnyTypeTestDefined (Source))
	    return CastType2SubrangeType (AnyGetDefinition (Source), Target);
	  else
	    return -1;
	case MetaClass:
	  if (MetaTypeTestDefined (Source))
	    return CastType2SubrangeType (MetaTypeGetCurrent (Source), Target);
	  else
	    return -1;
	case TupleClass:
	case StructureClass:
	case BundleClass:
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	default:
	  return -1;
	}
    }
}

int
CastType2BitsType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is an Bits Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (macroTytestpredef (Source))
    {
      if (macroPredeftestintnumber (Source)) return 0;
      return -1;
    }
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	  return -1;
	case BitsClass:
	  NotImplemented ("Casting to Bits not implemented. Returns -1");
	  return -1;
	case ArrayClass:
	case MatrixClass:
	  return -1;
	case NamedClass:
	  return CastType2BitsType (NamedTypeGetTerm (Source), Target);
	case LambdaClass:
	  return 0;		/* Types should be concrete */
	case AnyClass:
	  if (AnyTypeTestDefined (Source))
	    return CastType2BitsType (AnyGetDefinition (Source), Target);
	  else
	    return -1;
	case MetaClass:
	  if (MetaTypeTestDefined (Source))
	    return CastType2BitsType (MetaTypeGetCurrent (Source), Target);
	  else
	    return -1;
	case TupleClass:
	case StructureClass:
	case BundleClass:
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return -1;
	}
    }
  return -1;
}

int
CastType2ArrayType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is an Array Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (Source == Target)
    return 1;
  if (macroTytestpredef (Source))
    {
      if (macroPredeftestintnumber (Source)) return 0;
      return -1;
    }
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	  return -1;

	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  NotImplemented ("Casting to Array not implemented. Returns -1");
	  return -1;

	case NamedClass:
	  return CastType2ArrayType (NamedTypeGetTerm (Source), Target);

	case LambdaClass:
	  return 0;		/* Types should be concrete */

	case AnyClass:
	  if (AnyTypeTestDefined (Source))
	    return CastType2ArrayType (AnyGetDefinition (Source), Target);
	  else
	    return -1;
	case MetaClass:
	  if (MetaTypeTestDefined (Source))
	    return CastType2ArrayType (MetaTypeGetCurrent (Source), Target);
	  else
	    return -1;
	case TupleClass:
	case StructureClass:
	case BundleClass:
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	  return -1;
	}
    }
  return -1;
}

int
CastType2MatrixType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is an Matrix Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (Source == Target)
    return 1;
  if (macroTytestpredef (Source))
    if (macroPredeftestintnumber (Source))
      return 0;
    else
      return -1;
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	  return -1;
	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  NotImplemented ("Casting to Matrix not implemented. Returns -1");
	  return -1;
	case NamedClass:
	  return CastType2MatrixType (NamedTypeGetTerm (Source), Target);
	case LambdaClass:
	  return 0;		/* Types should be concrete */
	case AnyClass:
	  if (AnyTypeTestDefined (Source))
	    return CastType2MatrixType (AnyGetDefinition (Source), Target);
	  else
	    return -1;
	case MetaClass:
	  if (MetaTypeTestDefined (Source))
	    return CastType2MatrixType (MetaTypeGetCurrent (Source), Target);
	  else
	    return -1;
	case TupleClass:
	case StructureClass:
	case BundleClass:
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	default:
	  return -1;
	}
    }
}

int
CastTupleType2TupleType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target and Source are Tuple Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  TypeList l1, l2;
  TypeType t1, t2;
  TypeSubList SubLst1, SubLst2;

  l1 = TupleTypeGetTypeLst (Source);
  l2 = TupleTypeGetTypeLst (Target);
  SubLst1 = macroTypeListgetfirst (l1);
  SubLst2 = macroTypeListgetfirst (l2);
  while (pk_true)
    {
      t1 = macroTypeSubListgethead (SubLst1);
      t2 = macroTypeSubListgethead (SubLst2);
      if (TypeCastType (t1, t2) != 1)
	return -1;
      if (macroTypeSubListtestlast (SubLst1))
	{
	  if (macroTypeSubListtestlast (SubLst2))
	    return 1;
	  return -1;
	}
      else if (macroTypeSubListtestlast (SubLst2))
	return -1;
      SubLst1 = macroTypeSubListgettail (SubLst1);
      SubLst2 = macroTypeSubListgettail (SubLst2);
    }
  return -1;
}

int
CastType2TupleType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is an Tuple Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (Source == Target)
    return 1;

  if (macroTytestpredef (Source))
    if (macroPredeftestintnumber (Source))
      return 0;
    else
      return -1;
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	  return -1;
	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  return -1;
	case NamedClass:
	  return CastType2TupleType (NamedTypeGetTerm (Source), Target);
	case LambdaClass:
	  return 0;		/* Types should be concrete */
	case AnyClass:
	  if (AnyTypeTestDefined (Source))
	    return CastType2TupleType (AnyGetDefinition (Source), Target);
	  else
	    return -1;
	case MetaClass:
	  if (MetaTypeTestDefined (Source))
	    return CastType2TupleType (MetaTypeGetCurrent (Source), Target);
	  else
	    return -1;

	  // Febrary 2003 (LB)
	case TupleClass:
	  return CastTupleType2TupleType (Source, Target);

	  // Febrary 2003 (LB) case TupleClass:
	case StructureClass:
	case BundleClass:
	  NotImplemented ("Casting to Tuple not implemented. Returns -1");
	  return -1;

	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	default:
	  return -1;
	}
    }
}

int
CastType2StructureType (TypeType Source, TypeType Target)
  /*--************------------------------
    assumes: Target and Source have been both reduced
    Target is an Structure Type
    returns:-1 if Source to Target conversion is impossible
    0 if  Source to Target conversion is not decidable due
    to meta or generic current types or absence of value
    1 if Source can be converted to Target without explicit
    internal conversions
  */
{
  if (Source == Target)
    return 1;

  if (macroTytestpredef (Source))
    if (macroPredeftestintnumber (Source))
      return 0;
    else
      return -1;
  else
    {
      switch (macroTygetclass (Source))
	{
	case ExternClass:
	case EnumerationClass:
	case SubrangeClass:
	  return -1;
	case BitsClass:
	case ArrayClass:
	case MatrixClass:
	  return -1;
	case NamedClass:
	  return CastType2StructureType (NamedTypeGetTerm (Source), Target);
	case LambdaClass:
	  return 0;		/* Types should be concrete */
	case AnyClass:
	  if (AnyTypeTestDefined (Source))
	    return CastType2StructureType (AnyGetDefinition (Source), Target);
	  else
	    return -1;
	case MetaClass:
	  if (MetaTypeTestDefined (Source))
	    return CastType2StructureType (MetaTypeGetCurrent (Source), Target);
	  else
	    return -1;
	case TupleClass:
	  return CastType2TupleType (Source, StructureTypeGetArgument (Target));

	case BundleClass:
	case StructureClass:
	  NotImplemented ("Casting to Structure not implemented. Returns -1");
	  return -1;
	case NodeClass:
	case GenericNodeClass:
	case MetaNodeClass:
	default:
	  return -1;
	}
    }
}
