/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PROTECTED_
/**
   It creates a table (SubGraphTable) with \e nbSets elements of elemTableGenCode type.
   - SubGraphTable, nbSets are attributes member of the class.
   - nbSets = 1 for a graph without clusters 
   - nbSets = number of Clusters + 1 (scheduller) + 1 (for a particular "delay" cluster added at the end of the translation).
   - In all the cases, SubGraphTable[0] is the scheduller associated to the sbDC+ graph.
   - nbSets will be decreased when some cluster becomes empty by optimisations.

   Moreover, it initializes the following global lists:
   - lNodesHinit is initialized to the empty list. It will contain the list of the nodes that can be 
   defined at the initial time (constant signal).
   - lClocksNotToBeInitiliazed is initialized to the empty list. It will contain the list of the boolean clocks
   that depend on memories (they must not be initialized at each instant)
*/
  void
DCGBool2DCGSchTranslator::setsInitialize ()
{
  integer i;
  ptlINode lSubGraph, ln;
  ptINode SubGraph;

  lNodesHinit = NILLINODE;
  lClocksNotToBeInitiliazed = NILLINSTANCES;

  if (!NM->getSubGraphs (&lSubGraph))
    lSubGraph = NILLINODE;
  else
    {
      ln = lSubGraph;
      while (ptlINode_get_elem (&ln, &SubGraph))
	if (!NM->is_trainCluster (SubGraph))
	  ptlINode_remove_elem (&lSubGraph, SubGraph);
    }
  i = ptlINode_get_size (lSubGraph);
  nbSets = i + 1;		/* 1 for the Scheduler */
  if (i != 0)
    nbSets = nbSets + 1;	/* +1 for (future) delay cluster. */
  SubGraphTable = new elemTableGenCode *[nbSets];

  for (i = 0; i < nbSets; i++)
    SubGraphTable[i] = new elemTableGenCode ();

  /* Scheduler : index 0 */
  SubGraphTable[0]->aSubGraph = NILINODE;
  SubGraphTable[0]->ln = NILLINODE;
  SubGraphTable[0]->ldelays = NILLINODE;

  // For a graph of clusters only.
  if (nbSets > 1)
    {
      i = 1;
      while (ptlINode_get_elem (&lSubGraph, &SubGraph))
	{
	  SubGraphTable[i]->aSubGraph = SubGraph;
	  NM->subGraphLinkNodes (SubGraph, pk_true);
	  SubGraphTable[i]->ln = NILLINODE;
	  SubGraphTable[i]->ldelays = NILLINODE;
	  i++;
	}
      // Delay Cluster
      SubGraphTable[i]->aSubGraph = NILINODE;
      SubGraphTable[i]->ln = NILLINODE;
      SubGraphTable[i]->ldelays = NILLINODE;
    }
}

ptlINode DCGBool2DCGSchTranslator::addDirective(ptlINode ln, ptINode node, ptinsth h, boolean bOpen)
{
  // ajouter en tête pour open, en queue pour close
  ptexpcal arg2, arg1, vins;
  ptident id;

  arg1 = CKM->getConditionSub (h);
  if (CKM->getOperator (h) == nosub)  arg2 = EM->getTrue(); else arg2 = EM->getFalse();
  if (bOpen) id=getIdOpenBlock(); else id = getIdCloseBlock();

  vins = EM->mkNaryExp(oplesign, arg2, NILBACKANNOTATION);
  vins = EM->mkNaryExp(oplesign, arg1, vins, NILBACKANNOTATION);
  
  ptINode vdirective = NM->mk_procCall(id,NILEXPCAL,vins, NILEXPCAL, NILBACKANNOTATION);
  NM->set_intrinsicfunc(vdirective);

  if (node == NILINODE)
    ln = ptlINode_append_elem(ln, vdirective);
  else
    if (bOpen) ln = ptlINode_append_elem_before(ln, vdirective, node); else ln = ptlINode_append_elem_before(ln, vdirective, node); 
  return ln;
}

ptlINode DCGBool2DCGSchTranslator::genCloseBlock(ptlINode ln, ptINode node, ptinsth hFrom, ptinsth hTo)
{
  while (hFrom != hTo)
    {
      ln = addDirective(ln , node, hFrom, pk_false);
      hFrom = CKM->getFather (hFrom);
    }
  return ln;
}


ptlINode DCGBool2DCGSchTranslator::genOpenBlock(ptlINode ln, ptINode node, ptinsth hFrom, ptinsth hTo)
{
  if (hTo != hFrom)
    {
      ln = genOpenBlock(ln, node, hFrom, CKM->getFather (hTo));
      ln = addDirective(ln, node, hTo, pk_true);
    }
  return ln;
}


ptlINode DCGBool2DCGSchTranslator::mkOpenCloseDirectives(ptlINode ln, ptINode node, ptinsth currentClock, ptinsth nextCurrentClock)
{
  ptinsth vhemb;

  if ( nextCurrentClock != currentClock )
    {
      vhemb = CKM->getBranching (nextCurrentClock, currentClock);
      if (vhemb == nextCurrentClock)
	ln = genCloseBlock (ln, node, currentClock, vhemb);
      else if (vhemb == currentClock)
	ln = genOpenBlock (ln, node, vhemb, nextCurrentClock);
      else
	{
	  ln = genCloseBlock (ln, node, currentClock, vhemb);
	  ln = genOpenBlock (ln, node, vhemb, nextCurrentClock);
	}
    }
  return ln;
}

ptinsth DCGBool2DCGSchTranslator::getClockGenCode(ptINode node)
{
  return NM->get_nodeClock (node);
  /*   if (! NM->is_cellNode(node)) return NM->get_nodeClock (node);
       return NM->get_contributionClockEquation(node); */
}

ptlINode DCGBool2DCGSchTranslator::addOpenCloseBlockDirectives(ptlINode ln)
{
  ptINode node;
  ptinsth currentClock, h, vhemb;
  ptlINode laux;

  laux = ln;
  if (! ptlINode_get_elem(&laux, &node) ) return ln;

  currentClock = CKM->getTickClock();
  laux = ln;
  while (ptlINode_get_elem(&laux, &node))
    {
      h = getClockGenCode(node);
      ln = mkOpenCloseDirectives(ln, node, currentClock, h);
      currentClock = h; 
    }
  ln = genCloseBlock (ln, NILINODE, currentClock, CKM->getTickClock());
  return ln;
}

void DCGBool2DCGSchTranslator::addOpenCloseBlockDirectives()
{
  for (integer i=0; i < nbSets; i++)
    {
      SubGraphTable[i]->ln = addOpenCloseBlockDirectives(SubGraphTable[i]->ln);
      SubGraphTable[i]->ldelays = addOpenCloseBlockDirectives(SubGraphTable[i]->ldelays);
    }     
}

/**
   It finalizes DCGBool to DCGSch translation.
   The global informations (nbSets, SubGraphTable, lNodesHinit, lClocksNotToBeInitiliazed) are 
   associated with the graph.
   @return always "true".
*/
boolean DCGBool2DCGSchTranslator::finalize ()
{
  addOpenCloseBlockDirectives();
  AGRAPH->setDCGSchInfos (nbSets, SubGraphTable, lNodesHinit, lClocksNotToBeInitiliazed);
  return pk_true;
}






/**
   It adds a node (node) in the ordered list of node associated with an array (nodeArray).
   Here, a global predefined attribute (ARRAY_ORDER) is used (see CompilerAttributes.h).
*/
void
DCGBool2DCGSchTranslator::appendInOrderedSetArray (ptINode node, ptINode nodeArray)
{
  ptlINode ln;

  if (!NM->getAttribute (nodeArray, ARRAY_ORDER, (attributeValue *) & ln))
    ln = NILLINODE;
  ln = ptlINode_append_elem (ln, node);
  NM->addAttribute (nodeArray, ARRAY_ORDER, (attributeValue) ln);
}


/** 
    It orders the nodes with initial clock as contribution clock. Such a node defined a constant signal.
    \see orderingGraph().
 */
void
DCGBool2DCGSchTranslator::orderingInitClockNodes ()
{
  ptlsucc linst;
  ptINode inst;
  ptexpcal expg, expd;
  ptmemnode mn;

  PARC->initTravelForInitialClock ();
  orderingGraph (NILINODE, pk_true);
  if (AGRAPH->getNodesAtInitClock (&linst))
    while (ptlINode_get_elem (&linst, &inst))
      if (!ptlINode_isIn (unNodes, inst))
	if (!PARC->isNodeAlreadySelected (inst))
	  if (!NM->isMarked (inst, InitArrayMark))
	    {
	      NM->get_equation_arg (inst, &expg, &expd);
	      if (!EM->isAggregatedSignal (expg))
		if (!CKM->satisfiesNullClock (NM->get_nodeClock (inst)))
		  if (!ptexpcal_satisfies_node_tree_delay (expg, &mn))
		    {
		      storeNodeInSet (inst, NILINODE, pk_true);
		      PARC->ptINodeAttr_setSelected (inst);
		      PARC->signalSelectedNode (inst);
		    }
		  else
		    PARC->NodeReachedByAPredecessor (inst);
	    }
}

/**
  It travels the nodes/clocks of the graph, for a block context (blockNode) and at the initial clock when bhinit, using the \ref GraphTravel API.
  When a node is selected, it calls nodeProcessing(), for a clock it calls clockProcessing().
*/
void
DCGBool2DCGSchTranslator::orderingGraph (ptINode blockNode, boolean bhinit)
{
  ptINode inst;
  ptinsth horl;
  boolean bIsNode;

  if (blockNode != NILINODE)
    PARC->openBlock (blockNode);

  while (PARC->getNextNode (&bIsNode, &inst, &horl))
    if (bIsNode)
      nodeProcessing (inst, blockNode, bhinit);
    else
      clockProcessing (horl, blockNode);

  if (blockNode != NILINODE)
    PARC->closeBlock ();
}

/**
   @return "true" if the sub-graph (sg) is an entry of SubGraphTable, "false" otherwise.
   The first entry of SubGraphTable is reserved for sg is undefined.
   \see DCGBool2DCGSchTranslator::setsInitialize().
*/
boolean DCGBool2DCGSchTranslator::getSet (ptINode sg, integer * n)
{
  boolean
    found;
  integer
    i;

  found = pk_false;
  i = 0;
  while (!found && i <= nbSets)
    {
      found = (SubGraphTable[i]->aSubGraph == sg);
      if (!found)
	i++;
      else
	*n = i;
    }
  return found;
}

/**
   It stores a delay node (inst) in the iSet-th delay set.
   Delays are sorted: 
   for zzx := zx $ N and zx is a memory then zzx is inserted before zx.
   If the clock (boolean-clock) of zzx is a delay, zzx is inserted before its clock.  
*/
void
DCGBool2DCGSchTranslator::storeNodeInDelaySet (ptINode inst, integer iSet)
{
  ptexpcal resret, signalRetarde, x, y, vinit;
  boolean app;
  ptINode node;
  ptinsth h;

  app = pk_true;
  if (NM->is_delay (inst))
    {
      NM->get_delay_arg (inst, &resret, &signalRetarde, &x, &y, &vinit);
      if (EM->IsSignal (signalRetarde))
	{
	  node = EM->getNode (signalRetarde);
	  app = !NM->is_delay (node);
	}
    }
  if (app)
    {
      h = NM->get_nodeClock (inst);
      if (CKM->satisfiesSub (h))
	{
	  x = CKM->getConditionSub (h);
	  node = EM->getNode (x);
	  app = !NM->is_delay (node);
	}
    }
  if (app)
    SubGraphTable[iSet]->ldelays = ptlINode_append_elem (SubGraphTable[iSet]->ldelays, inst);	// at the end.
  else
    SubGraphTable[iSet]->ldelays = ptlINode_append_elem_before (SubGraphTable[iSet]->ldelays, inst, node);
}


boolean DCGBool2DCGSchTranslator::isMemoryNode(ptINode node)
{
  ptident name;

  if (NM->is_delayNodeOrDelayClusterNode (node)) return pk_true;
  if (NM->is_intrinsicfunc (node) && NM->get_nameCall (node, &name)) if (isStatusId (name)) return pk_true;
  return pk_false;
  }

/**
   The node (inst) is inserted:
   - in the global list (lNodesHinit) if bhinit
   - in an array block (sg) when sg is a array node (appendInOrderedSetArray())
   - in a delay set when it is a delay node (storeNodeInDelaySet())
   - in the i-th entry of the global table SubGraphTable. The i-th sub-graph is given by getSet() method applied to sg.
*/
void
DCGBool2DCGSchTranslator::storeNodeInSet (ptINode inst, ptINode sg, boolean bhinit)
{
  integer i;
  boolean done;

  if (!NM->isMarked (inst, InitArrayMark))
    {
      DEBUGGING (inst, sg, bhinit);
      if (bhinit)
	lNodesHinit = ptlINode_append_elem (lNodesHinit, inst);
      else
	{
	  done = pk_false;
	  if (sg != NILINODE)
	    if (NM->is_array (sg) || (NM->is_subgraph (sg) && !NM->is_trainCluster (sg)))
	      {
		done = pk_true;
		appendInOrderedSetArray (inst, sg);
	      }
	  if (!done)
	    {
	      if (!getSet (sg, &i))	// Blindage
		{
		  pK_messenger->error (errDCGBool2DCGSchInternalError);
		  pK_messenger->error (errcompilerbug);
		}
	      /* Delays are isolated. */
	      if (isMemoryNode(inst)) // xxxxx here 
		storeNodeInDelaySet (inst, i);
	      else
		SubGraphTable[i]->ln = ptlINode_append_elem (SubGraphTable[i]->ln, inst);
	    }
	}
    }
}

/**
   It signals that a clock (h) has been visited (signalSelectedClock()).
*/
void
DCGBool2DCGSchTranslator::clockProcessing (ptinsth h, ptINode sg)
{
  PARC->signalSelectedClock (h);
}

/**
   - It inserts a node (node) in an entry of the table of nodes (storeNodeInSet()).
   - For composed nodes (array, sub-graph) the travelling is recalled with a new context (to visit the internal nodes of the compsoed node).
   - It signals that the node has been visited (signalSelectedNode()).
*/
void
DCGBool2DCGSchTranslator::nodeProcessing (ptINode node, ptINode sg, boolean bhinit)
{
  ptINode indexNode, inputNode;

  storeNodeInSet (node, sg, bhinit);
  // For composed nodes (array, sub-graph) the travelling is recalled with a new context
  switch ((int) (NM->get_oper (node)))
    {
    case opmoti:
      PARC->signalArraySelectedNode (node);
      orderingGraph (node, bhinit);
      break;

    case opsubgraph:
      PARC->signalSubGraphSelectedNode (node);
      orderingGraph (node, bhinit);
      break;

    default:;
    }
  PARC->signalSelectedNode (node);
}


/**
   It removes a node (sg) from the scheduller.
*/
void
DCGBool2DCGSchTranslator::removeScheduller (ptINode sg)
{
  ptlINode ln;

  NM->dependencesPropagate(sg); // added Aug 2016 LBesnard.
  NM->removeNode (sg);
  ln = SubGraphTable[0]->ln; // Nodes of the scheduller
  if (ptlINode_remove_elem (&ln, sg))
    SubGraphTable[0]->ln = ln;
}

/**
   It synthetizes the i/o signals of a sub-graph (aSubGraph).
*/
void
DCGBool2DCGSchTranslator::setSubGraphIOReferences (ptINode aSubGraph)
{
  ptlinstances lsig, entrees, sorties;
  ptexpcal lentrees, lsorties;

  lsig = NM->getInodelInputFlows (aSubGraph);
  entrees = ptlinstances_intersection (lsig, AGRAPH->getInputs ());
  lentrees = EM->convertLinstances2ListExpcal (entrees);
  lsig = NM->getInodelOutputFlows (aSubGraph);
  sorties = ptlinstances_intersection (lsig, AGRAPH->getOutputs ());
  lsorties = EM->convertLinstances2ListExpcal (sorties);
  NM->set_inOutGraphRef (aSubGraph, lentrees, lsorties);
}


/**
   It synthetizes the interface of a cluster (aCluster).
*/
void
DCGBool2DCGSchTranslator::buildClusterInterface (ptINode aCluster)
{
  NM->set_interface (aCluster, NILINTERFACE);
  NM->SubGraphInterfaceCalculusUndersamplingClocks (aCluster);
  NM->moveOutputs2localsExcept (aCluster, AGRAPH->getOutputs (), NILLINSTH);
  setSubGraphIOReferences (aCluster);
}

/**
   For a sub-graph (aSubGraph):
   - It builds its interface
   - it inserts the sub-graph at the branching clock of the nodes of the sub-graph (ie the information must be available)
   - the branching clock becomes the clock of the label associated with the sub-graph. 
*/
void
DCGBool2DCGSchTranslator::insertSubGraph (ptINode aSubGraph)
{
  ptlinsth lh;
  ptinsth vhemb;
  ptexpcal vlab;

  NM->getSet_interface (aSubGraph);
  NM->get_subGraphProjectedSystemInfos (aSubGraph, &vhemb, &lh);
  NM->insertNode (aSubGraph, vhemb);
  if (NM->get_label (aSubGraph, &vlab))
    EM->setClock (vlab, vhemb);
}

/**
   It is called when the graph is "clusterized" (input train clusters for example).

   It defines a new cluster for delay nodes (ldelays) when the list is not empty.
   This particular cluster is annotated with the global pragmas PRAGMA_DELAYCLUSTER and PRAGMA_CLUSTER and inserted in the graph.
   Its references is added in the scheduler.
   A dependence from this cluster to each other cluster is added.

   @return "true" if the operand list (ldelays) is not empty, "false" otherwise.
*/
boolean
DCGBool2DCGSchTranslator::appendDelayCluster (ptlINode ldelays)
{
  ptINode aCluster;
  ptexpcal vexpcal;
  ptINode aNode;
  integer i;
  ptinsth h;

  if (ldelays == NILLINODE && (AGRAPH->getStatesVar () == NILLINSTANCES))
    {
      nbSets = nbSets - 1;
      return pk_false;
    } // No delay Cluster

  SubGraphTable[nbSets - 1]->ldelays = ldelays;

  vexpcal = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), CKM->getUndefClock (), NILBACKANNOTATION, pk_true);
  aCluster = NM->mk_node_subgraph (vexpcal, NILBACKANNOTATION);
  NM->set_trainClusterTag (aCluster);
  while (ptlINode_get_elem (&ldelays, &aNode))
    NM->subGraphInsertNode (aCluster, aNode);
  insertSubGraph (aCluster);
  buildClusterInterface (aCluster);
  // --- NO I/O for delay cluster.
  h = CKM->getUndefClock ();	// To be fixed
  for (i = 1; i < nbSets - 1; i++)
    NM->addDependence (aCluster, SubGraphTable[i]->aSubGraph, h);

  SubGraphTable[nbSets - 1]->aSubGraph = aCluster;
  // Append to scheduller
  SubGraphTable[0]->ln = ptlINode_append_elem (SubGraphTable[0]->ln, aCluster);
  NM->subGraphLinkNodes (aCluster, pk_true);
  NM->addPragma (aCluster, newPredefinedPragma (PRAGMA_DELAYCLUSTER, NILOBJPRAGMA, NILENONCEPRAGMA));
  NM->addPragma (aCluster, newPredefinedPragma (PRAGMA_CLUSTER, NILOBJPRAGMA, NILENONCEPRAGMA));
  return pk_true;
}


	  /* if (SubGraphTable[i]->ln == NILLINODE)
	    {
	      for (j = i; j < nbSets; j++) { SubGraphTable[j] = SubGraphTable[j + 1];}
	      nbSets = nbSets - 1;
	      removeScheduller (sg);
	    }
	  else
	    i = i + 1;
	  */

boolean DCGBool2DCGSchTranslator::moveEmptyEntry(int i, boolean bclean)
{
  ptINode sg;
  int j;
  if ((SubGraphTable[i]->ln == NILLINODE) && (SubGraphTable[i]->ldelays == NILLINODE ))
    {
      sg = SubGraphTable[i]->aSubGraph;
      for (j = i; j < nbSets; j++) { SubGraphTable[j] = SubGraphTable[j + 1];}
      nbSets = nbSets - 1;
      removeScheduller (sg);
      if (bclean) NM->removeMark (sg, GencodeMark);
      return pk_true;
    }
  return pk_false;
}

/**
   @return "true" if graph is "clusterized", "false" otherwise.

   When the graph is "clusterized" (input train clusters for example):
   - It builds the delay cluster (appendDelayCluster()), it is the last element of the SubGraphTable table.
   - It cleans the empty clusters (induced by the previous operation).
   - It updates (buildClusterInterface()) the interfaces of the clusters (induced by the previous operations).

   <em> This operation modifies the graph. </em>
*/
boolean
DCGBool2DCGSchTranslator::delayClustering ()
{
  integer i, j;
  ptINode sg, n;
  ptlINode ldelays, ld, l;
  boolean b;

  if (nbSets == 1) return pk_false; // No Clusters
  // It cleans the empty clusters
  ldelays = NILLINODE;
  b = pk_false;
  i = 1;
  while (i < (nbSets - 1))
    {
      ld = SubGraphTable[i]->ldelays;
      if (ld != NILLINODE)
	{
	  SubGraphTable[i]->ldelays = NILLINODE;
	  sg = SubGraphTable[i]->aSubGraph;
	  // Unlink the nodes of the sub-graph.
	  l = ld;
	  while (ptlINode_get_elem (&l, &n))
	    {
	      NM->subGraphRemoveNode (sg, n);
	      NM->subGraphUnlinkNodes (sg, pk_true);
	    }
	  ldelays = ptlINode_catenate_post (ldelays, ld);
	  b = pk_true;
	  if (! moveEmptyEntry(i, pk_false)) i++;
	}
      else
	i = i + 1;
    }

  //  It updates the interfaces of the clusters 
  if (b)
    for (i = 1; i < nbSets - 1; i++)
      buildClusterInterface (SubGraphTable[i]->aSubGraph);

  // Appending the delay Cluster if ldelays is not empty
  return appendDelayCluster (ldelays);
}


/**
   It assigns name to clusters if the graph is a graph of clusters with a delay cluster is \e withDelayCluster.
   The name is built from the name of the graph (\e foo)
   \e foo_Cluster_i for i in [1, nbSets-1], and \e foo_Cluster_delays for the delay cluster.
*/
void
DCGBool2DCGSchTranslator::setClusterName (boolean withDelayCluster)
{
  integer i, n;
  ptident id, clusterName;

  if (nbSets == 1)
    return;			// No Clusters
  clusterName = newIdent_stringSuffixed (AGRAPH->getName (), "Cluster");
  if (withDelayCluster)
    n = nbSets - 1;
  else
    n = nbSets;
  for (i = 1; i < n; i++)
    {
      id = makeIdent_ident_int (clusterName, i);
      NM->set_clusterName (SubGraphTable[i]->aSubGraph, id);
    }
  if (withDelayCluster)
    {
      id = newIdent_stringSuffixed (AGRAPH->getName (), "STATES");
      NM->set_clusterName (SubGraphTable[i]->aSubGraph, id);
    }
}

/**
   It marks (GencodeMark global attribute) the cluster for code generation.
*/
void
DCGBool2DCGSchTranslator::clusterMarkGenCode ()
{
  integer i;

  if (nbSets > 1)
    for (i = 1; i < nbSets; i++)
      {
	ptINode SubGraph = SubGraphTable[i]->aSubGraph;
	NM->addMark (SubGraph, GencodeMark);
	// NM->subGraphUnlinkNodes(SubGraph, pk_true);  removed Nov 2013 L.Besnard.
      }
}

/** 
   This method builds the list of the boolean clocks (lClocksNotToBeInitiliazed) which must not be initialized at each instant.
   Such clocks are built on delay variables. This list is, at the end of the translation, associated with the graph (DCGBool2DCGSchTranslator::finalize()).

   It references markClocksNotToBeInitiliazed(ptlinstances) method.
*/
_VIRTUAL_ void
DCGBool2DCGSchTranslator::markClocksNotToBeInitiliazed ()
{
  lClocksNotToBeInitiliazed = NILLINSTANCES;
  markClocksNotToBeInitiliazed (AGRAPH->getOutputs ());
  markClocksNotToBeInitiliazed (AGRAPH->getLocalSignals ());
}

/** 
   This method appends to the global list (lClocksNotToBeInitiliazed) 
   the boolean signals of the list lsig which are built on delay variables (isMemoryEquation()).
*/
_VIRTUAL_ void
DCGBool2DCGSchTranslator::markClocksNotToBeInitiliazed (ptlinstances lsig)
{
  ptexpcal aSig;
  ptINode node;
  while (ptlinstances_get_next (&lsig, &aSig))
    if (EM->IsBoolean (aSig))
      if (!ptlinstances_isin_exp (lClocksNotToBeInitiliazed, aSig))
	{
	  node = EM->getNode (aSig);
	  if (isMemoryEquation (node))
	    {
	      lClocksNotToBeInitiliazed = ptlinstances_insert (lClocksNotToBeInitiliazed, aSig);
	      if (EM->isAggregatedSignal (aSig))
		{
		  aSig = EM->getAggregation (aSig);
		  lClocksNotToBeInitiliazed = ptlinstances_insert (lClocksNotToBeInitiliazed, aSig);
		}
	    }
	}
}

/**
   @return "true" if a node (node) is a memory node (delay, cell ,status, constant signal), "false" otherwise. 
*/
_VIRTUAL_ boolean
DCGBool2DCGSchTranslator::isMemoryEquation (ptINode node)
{
  ptexpcal expg, expd;
  ptident name;

  if (NM->isMarked (node, BELONG_MEMORY))
    return pk_true;		// by optimisation
  if (NM->is_delay (node))
    return pk_true;
  if (NM->is_cellNode (node))
    return pk_true;

  if (NM->is_intrinsicfunc (node) && NM->get_nameCall (node, &name)) return isStatusId (name);

  if (!NM->is_equation (node))
    return pk_false;
  if (!CKM->satisfiesInitClock (NM->get_contributionClockEquation (node)))
    return pk_false;
  NM->get_equation_arg (node, &expg, &expd);
  return (EM->getNumberOfDimensions (expg) == 0);
}


/** It associates a label (vlabel) to a node(n) if the node is without any outputs.
    The clock of the label is the clock of the node.
    @return true if a label is created, false otherwise.
*/
boolean DCGBool2DCGSchTranslator::addLabel (ptINode n)
{
  ptinsth
    h;
  ptexpcal
    vl;

  if (NM->get_leftOperand (n) != NILEXPCAL)
    return pk_false;
  if (NM->get_label (n, &vl))
    return pk_false;
  if (NM->is_ioNode (n))
    return pk_false;
  h = NM->get_nodeClock (n);
  vl = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), h, NM->getBackAnnotation (n), pk_true);
  NM->set_label (n, vl);
  EM->setNode (vl, n);
  return pk_true;
}

boolean DCGBool2DCGSchTranslator::getnextIO (ptlINode * ln, ptINode * ionode)
{
  boolean
    found;
  ptINode
    n;

  if ((*ln) == NILLINODE)
    return pk_false;
  found = pk_false;
  while (!found && ptlINode_get_elem (ln, &n))
    if (NM->is_ioNode (n))
      {
	*ionode = n;
	ptlINode_get_elem (ln, &n);
	found = pk_true;
      }
  return found;
}

/**
   It reinforces the dependences n(i)->n(i+1) for n(i) an element of the operand list(lnodes).
*/
_VIRTUAL_ void
DCGBool2DCGSchTranslator::addReinforcingDependences (ptlINode lnodes)
{
  ptINode n, n1;
  ptinsth hdep;
  ptdependence dep;
  ptlINode ln;

  ln = lnodes;
  // The "writing nodes" are filtered.
  if (ptlINode_get_elem (&ln, &n1))
    {
      hdep = CKM->getUndefClock ();
      // hdep = AGRAPH->getRootClock ();
      while (ptlINode_get_elem (&ln, &n))
	//  boolean b; if (!NM->is_writeNode(n)) b = pk_true; else b = NM->is_delay( EM->getNode(NM->get_io_arg(n)));
	if (! NM->is_writeNode (n))
	  {
	    addLabel (n1);
	    addLabel (n);
	    NM->addReinforcingDependence (n1, n, hdep);
	    n1 = n;
	  }
    }

  /* // March 2014 Test LBesnard.
    ln = lnodes; 
    if (getnextIO (&ln, &n1))
    while (getnextIO (&ln, &n)) { NM->addReinforcingDependence (n1, n, hundef); n1 = n; }
  */
}

/**
   It reinforces the dependences induces by the static schedulling.
   A complete static scheduling when btotally.
*/
_VIRTUAL_ void
DCGBool2DCGSchTranslator::addReinforcingDependences (boolean btotally)
{
  int i0, i;

  if (nbSets == 1)
    {
      addReinforcingDependences (SubGraphTable[0]->ln);
      addReinforcingDependences (SubGraphTable[0]->ldelays);
    }
  else				// Clusters
    {
      if (btotally) i0 = 0; else i0 = 1;  
      for (i = i0; i < nbSets; i++)
	{
	  addReinforcingDependences (SubGraphTable[i]->ln);
	  addReinforcingDependences (SubGraphTable[i]->ldelays);
	}
    }
}

/** 
    It finalizes the ordering operation: 
   - addReinforcingDependences(): dependence reinforcing. When btotally is pk_true, the graph is totally ordered (static scheduling) otherwise only the clusters are ataticaly scheduled (not the graph of the clusters).
   - delayClustering(): for a graph of clusters, the memories are isolated in a particular cluster.
   - setClusterName(): for a graph of clusters, names are associated with the clusters.
   - clusterMarkGenCode(): for a graph of clusters, it marks the clusters for code generation.
   - markClocksNotToBeInitiliazed(): it marks the clocks built on memories to inhibit optimizations
*/
_VIRTUAL_ void
DCGBool2DCGSchTranslator::finalizeOrderingGraph (boolean btotally)
{
  addReinforcingDependences (btotally);
  setClusterName (delayClustering ());
  optimize();
  clusterMarkGenCode ();
  markClocksNotToBeInitiliazed ();
}

ptlINode DCGBool2DCGSchTranslator::optimizeNodes(ptlINode ln)
{
  ptlINode lnprim;
  ptINode aNode;

  if (ln == NILLINODE) return NILLINODE;
  lnprim = ln;
  while (ptlINode_get_elem (&ln, &aNode))
    {
      if (! NM->mustBeGenerated(aNode))
	{
	  ptlINode_remove_elem (&lnprim, aNode);
	  NM->removeMark (aNode, GencodeMark);
	}
    }
  return lnprim;
}

void  DCGBool2DCGSchTranslator::optimize(integer indexSubGraphTable)
{
  SubGraphTable[indexSubGraphTable]->ln = optimizeNodes (SubGraphTable[indexSubGraphTable]->ln);
  // SubGraphTable[indexSubGraphTable]->ldelays = optimizeNodes (SubGraphTable[indexSubGraphTable]->ldelays);
}


void DCGBool2DCGSchTranslator::optimize()
{
  int i,j;
  ptINode sg;
  
  if (nbSets == 1)
    optimize (0);
  else // Clusters
    {
      for (i = 1; i < nbSets; i++) { optimize (i);}
      // The references to the empty clusters are remove from the scheduler.
      i = 1;
      while (i < nbSets)
	{
	  if (! moveEmptyEntry(i, pk_true)) i++;
	}
    }
}
