/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** 
    Constructor of a DCGPoly to DCGBool translator.
    The functionality is applied to a graph (g), the attribute members (graph, managers,...) are imported from it.
*/
DCGEndo2DCGBoolTranslator::DCGEndo2DCGBoolTranslator (ptDCG g):InterFormatTranslator (g)
{
}

/**
   Destructor.
*/
_VIRTUAL_ DCGEndo2DCGBoolTranslator::~DCGEndo2DCGBoolTranslator ()
{
}

/**
   It translates a  DCGPoly (selected in the construtor) into a DCGBool.
   Clocks (events) are represented by boolean signals in DCGBool.
   @param lsig (in/out) the inputs "moved to outputs" are added to this list.

   It is achieved by the following operations:
   - tickCoding()
   - InterFormatTranslator::clocksCoding()
   - defaults2lefttts()
   - assertions2nodes()
   - removeInputEvents()
   - event2boolean()
   - updateDependencesContributionClock()
   - NodeClockUpdate()
   - endOp()  

   We assume here that the hierarchy is reduced to ONE tree (the program is ENDOCHONOUS).
   The hierarchy has been completed before.
*/
void
DCGEndo2DCGBoolTranslator::apply (boolean mkSubForNosub, ptlinstances * lsig)
{
  if (isEventClockedFormat (oriFormat))
    {
      modeCreateSubForNosub = mkSubForNosub;
      local_init ();
      tickCoding ();
      clocksCoding ();
      defaults2lefttts ();
      removeInputEvents (lsig);
      assertions2nodes ();
      events2booleans ();
      updateDependencesContributionClock ();
      nodeClockUpdate ();
      endOp ();
      local_final ();
    }
}
