/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

// #include <iostream>#include <fstream>using namespace std;


_PROTECTED_
#define isEventClockedFormat(t) ((t == DCG_BASIC) || (t == DCG_POLY) || (t == DCG_ENDO))

/** 
    The identifier associated with nexpr is replaced by the identifier associated with oexpr, if 
    - oexpr and nexpr are two instances of the same flow 
    - and oexpr is without suffix.
*/
  void
DCGEndo2DCGBoolTranslator::changeName (ptexpcal oexpr, ptexpcal nexpr)
{
  ptident nameOldInput, nameOrigOldInput, nameOrigNewInput;

  // LB: Revoir utilite
  nameOldInput = EM->getName (oexpr);
  nameOrigOldInput = EM->getOriginalName (oexpr);
  if (nameOldInput == nameOrigOldInput)
    {
      nameOrigNewInput = EM->getOriginalName (nexpr);
      if (nameOrigNewInput == nameOldInput)
	EM->setOriginalName (nexpr);
    }
}


/**
   @return the boolean that represents the clock (clk) at the clock hAT.
   \note It assumes that it exists: the boolean clock of clk has been defined before.
*/
_VIRTUAL_ ptexpcal
DCGEndo2DCGBoolTranslator::getBooleanSignalAt (ptinsth clk, ptinsth hAT)
{
  ptinsth vh;
  pkClockType th;
  ptexpcal e;

  if (!CKM->getBooleanClockAt (clk, hAT, &vh))
    return NILEXPCAL;		/* error */
  th = CKM->getOperator (vh);
  if (th == sub)
    return CKM->getConditionSub (vh);
  if (th == nosub)
    {
      e = CKM->getConditionSub (vh);
      return EM->mkUnaryExp (opnon, e, EM->getBackAnnotation (e));
    }
  return NILEXPCAL;		/* error (debugging)  */
}

/**
   @return "true" if it exists an positive extraction(sub) equivalent to the clock (clk) in its class, "false" otherwise.
 */
_VIRTUAL_ boolean
DCGEndo2DCGBoolTranslator::getSignalEquiv (ptinsth clk, ptinsth hAT, ptinsth * vhresult)
{
  boolean btick;
  ptinsth vh1;

  clk = CKM->getRepresentativeFormula (clk);
  hAT = CKM->getRepresentativeFormula (hAT);
  btick = CKM->satisfiesTickClock (clk);

  if ((clk == hAT) && (!btick))
    {				/* cas bizare de ^- (avec bdd) xxxxx A REVOIR */
      if (CKM->getOperator (clk) == sub)
	{
	  *vhresult = clk;
	  return pk_true;
	}
    }
  else
    {
      vh1 = clk;
      do
	{
	  if (CKM->getOperator (vh1) == sub)
	    {
	      if ((CKM->getClockSub (vh1) == hAT) || btick)
		if (EM->IsSignal (CKM->getConditionSub (vh1)))
		  {
		    *vhresult = vh1;
		    return pk_true;
		  }
	    }
	  vh1 = CKM->getNextFormula (vh1, pk_false);
	}
      while (vh1 != clk);
    }
  return pk_false;
}



/** 
    @return the branching of all the references to the clock (vh) in the graph.
*/
_VIRTUAL_ ptinsth
DCGEndo2DCGBoolTranslator::OccurrencesBranchingClock (ptinsth vh)
{
  ptlinsth luse, luse1, lh1, lh2;
  ptinsth vhemb, use, use1, k;
  ptINode aNode;
  ptlINode lnodes;

  if (CKM->satisfiesTickClock (vh))
    return vh;
  if (satisfiesIOstepClock (vh))
    return tickClock;

  vhemb = CKM->getFather (vh);
  luse = CKM->getClockReferencesTo (vh);
  while (ptlinsth_get_next (&luse, &use))
    if (!CKM->satisfiesIsAnEquivalentForm (use))
      if (CKM->getOperator (use) != hnot)
	{
	  k = CKM->getBranching (vhemb, use);
	  if (k != hundef)
	    vhemb = k;
	}
      else
	{
	  luse1 = CKM->getClockReferencesTo (use);
	  while (ptlinsth_get_next (&luse1, &use1))
	    if (CKM->getOperator (use1) != hnot)
	      {
		k = CKM->getBranching (vhemb, use1);
		if (k != hundef)
		  vhemb = k;
	      }
	}

  lnodes = CKM->getNodeReferencesTo (vh);
  while (ptlINode_get_elem (&lnodes, &aNode))
    vhemb = CKM->getBranching (vhemb, NM->get_nodeClock (aNode));

  if (CKM->getConstraintsOnPlacedClocks (&lh1, &lh2))
    while (ptlinsth_get_next (&lh1, &use))
      {
	ptlinsth_get_next (&lh2, &use1);
	if ((use == vh) || (use1 == vh))
	  vhemb = CKM->getBranching (vhemb, CKM->getBranching (use, use1));
      }

  if (CKM->getAssertionsOnPlacedClocks (&lh1, &lh2))
    while (ptlinsth_get_next (&lh1, &use))
      {
	ptlinsth_get_next (&lh2, &use1);
	if ((use == vh) || (use1 == vh))
	  vhemb = CKM->getBranching (vhemb, CKM->getBranching (use, use1));
      }

  if (CKM->getOperator (vhemb) == troot)
    return CKM->getFather (vh);
  return vhemb;
}

_VIRTUAL_ void DCGEndo2DCGBoolTranslator::setBooleanDefinition(ptinsth clk, ptexpcal boolExp, ptinsth clkFather)
{
  if (modeCreateSubForNosub) 
    endDefinition (clk, boolExp, clkFather);
  else
    {
      pkClockType vop = CKM->getOperator (clk);
      if ((vop == sub) || (vop == nosub) || (vop == hcomp))
	CKM->setClockAsBoolean (clk, boolExp, EM->getPredecessors (boolExp), clkFather); 
      else
	endDefinition (clk, boolExp, clkFather);
    }
}


_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::nosubForm2bDC (ptinsth vh)
{
  ptexpcal exp, id, exp1;
  ptinsth vexpClk, vh1, vhpos;
  ptinterface vstru;

  vexpClk = CKM->getFather(vh);
  CKM->getComplExtraction (vh, &vhpos);
  // ok because the clock "when not E" is always visited after the clock "when E"
  vhpos = CKM->getRepresentativeFormula (vhpos);
  exp = CKM->getConditionSub (vhpos);
  exp1 = EM->mkUnaryExp (opnon, exp, EM->getBackAnnotation (exp));
  setBooleanDefinition(vh, exp1, vexpClk);
}


/** 
    DCGPoly2DCGBool coding for an affine sample clock (vh).
*/
_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::affine_sample2bDC (ptinsth vh)
{
  ptexpcal parameters, inputs, outputs, exp;
  ptAffineRelation affr;
  AffineParam period_n, phase, period_d;
  ptinsth clkFather;
  ptINode node;

  if (CKM->satisfiesSub (vh))
    {
      exp = CKM->getConditionSub (vh);
      if (EM->isNILEXPCAL (exp))
	if (CKM->getAffineRelation (vh, &affr))
	  {
	    BackAnnotationType ba = NILBACKANNOTATION;
	    ptAffineRelation_get_elems (affr, &period_n, &phase, &period_d);
	    parameters = EM->mkNaryExp (oplesign, EM->mkIntegerConstant ((int) period_d), NILEXPCAL, ba);
	    parameters = EM->mkNaryExp (oplesign, EM->mkIntegerConstant ((int) phase), parameters, ba);
	    clkFather = CKM->getFather (vh);
	    inputs = EM->mkNaryExp (oplesign, CKM->getBooleanSignal (clkFather), NILEXPCAL, ba);

	    exp = EM->mkIdentExp (choiceNameBooleanClock (vh), tsignal, get_BooleanType (), NM->mk_node_undefined (), vh, ba);
	    outputs = EM->mkNaryExp (oplesign, exp, NILEXPCAL, NILBACKANNOTATION);
	    node = NM->mk_procCall (getSampleId (), parameters, inputs, outputs, ba);
	    NM->set_intrinsicfunc (node);
	    NM->insertNode (node, clkFather);
	    CKM->setConditionSub (vh, exp, ptlinstances_insert (NILLINSTANCES, exp));
	  }
    }
}

/**
   DCGPoly2DCGBool coding for a positive extraction (vh).
   A signal is created when the condition assigned to the operand clock (vh) is not reduced to a signal.
*/
_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::subForm2bDC (ptinsth vh)
{
  ptexpcal exp;
  ptinsth vexpClk, vh1;
  ptexpcal id;

  affine_sample2bDC (vh);
  exp = CKM->getConditionSub (vh);
  vexpClk = CKM->getClockSub (vh);
  if (EM->IsSignal (exp))
    CKM->addChild (vexpClk, vh);
  else if (!getSignalEquiv (vh, vexpClk, &vh1))
    endDefinition (vh, exp, vexpClk);
  else if (vh != vh1)
    {
      id = CKM->getConditionSub (vh1);
      NM->modifiedDefinition (id, exp, vexpClk, vexpClk);
      CKM->switchFormula (vh, vh1, vexpClk);
    }
}

/**
   @return "true" if the operation succeeds, "false" otherwise (internal error).
   It produces the boolean representation (boolExp) of a clock (clk), child of clkFather in the hierachy.    
*/
_VIRTUAL_ boolean
DCGEndo2DCGBoolTranslator::clk2Boolean (ptinsth clk, ptinsth clkFather, ptexpcal * boolExp)
{
  ptexpcal vop1, vop2, no_vop1;
  ptinsth vh1, vh2, branchingClk1, branchingClk2;

  switch (CKM->getOperator (clk))
    {
    case vari:			/* le Tick du programme bDC+ */
      return pk_false;
      break;
    case sub:
      (*boolExp) = CKM->getConditionSub (clk);
      break;
    case nosub:
      vop1 = CKM->getConditionSub (clk);
      (*boolExp) = EM->mkUnaryExp (opnon, vop1, EM->getBackAnnotation (vop1));
      break;
    case infer:
      if (CKM->getMinusArgs (clk, &vh1, &vh2))
	{
	  vh1 = CKM->getRepresentativeFormula (vh1);
	  branchingClk1 = CKM->getBranching (clkFather, vh1);
	  vop1 = getBooleanSignalAt (vh1, branchingClk1);
	  no_vop1 = EM->mkUnaryExp (opnon, vop1, EM->getBackAnnotation (vop1));
	  vh2 = CKM->getRepresentativeFormula (vh2);
	  if (vh2 == clkFather)
	    (*boolExp) = no_vop1;
	  else
	    {
	      branchingClk2 = CKM->getBranching (clkFather, vh2);
	      vop2 = getBooleanSignalAt (vh2, branchingClk2);
	      (*boolExp) = EM->mkBinaryExp (opet, no_vop1, vop2, NILBACKANNOTATION);
	    }
	}
      else
	{
	  CKM->getBinaryArgs (clk, &vh1, &vh2);
	  vh1 = CKM->getRepresentativeFormula (vh1);
	  vh2 = CKM->getRepresentativeFormula (vh2);
	  branchingClk1 = CKM->getBranching (clkFather, vh1);
	  vop1 = getBooleanSignalAt (vh1, branchingClk1);
	  branchingClk2 = CKM->getBranching (clkFather, vh2);
	  vop2 = getBooleanSignalAt (vh2, branchingClk2);
	  (*boolExp) = EM->mkBinaryExp (opet, vop1, vop2, NILBACKANNOTATION);
	};
      break;

    case super:
      CKM->getBinaryArgs (clk, &vh1, &vh2);
      vh1 = CKM->getRepresentativeFormula (vh1);
      branchingClk1 = CKM->getBranching (clkFather, vh1);
      vop1 = getBooleanSignalAt (vh1, branchingClk1);

      vh2 = CKM->getRepresentativeFormula (vh2);
      branchingClk2 = CKM->getBranching (clkFather, vh2);
      vop2 = getBooleanSignalAt (vh2, branchingClk2);

      (*boolExp) = EM->mkBinaryExp (opou, vop1, vop2, NILBACKANNOTATION);
      break;

    case hcomp:
      CKM->getUnaryArg (clk, &vh1);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vop1 = getBooleanSignalAt (vh1, clkFather);
      /*      vh2 = clkFather; */
      /*      vop2 = CKM->getBooleanSignal(vh2); */
      /*      no_vop1 = EM->mkUnaryExp(opnon, vop1, EM->getBackAnnotation(vop1)); */
      /*      (*boolExp) = EM->mkBinaryExp(opet, vop2, no_vop1, NILTREE);  */
      (*boolExp) = EM->mkUnaryExp (opnon, vop1, EM->getBackAnnotation (vop1));
      break;

    default:
      return pk_false;
    }
  return pk_true;
}

/** 
    It translates a free formula (vh) assumed to be an undersampling poisitive (when C) or negative (when not C). 
    It calls subForm2bDC() fore a positive form,  nosubForm2bDC() for a negative one.
*/
_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::freeForm2bDC (ptinsth vh)
{
  if (CKM->getOperator (vh) == sub)
    subForm2bDC (vh);
  else
    nosubForm2bDC (vh);
}

/** 
    It translates a formula (vh) assumed to be a lower bound, or upper bound.
    The boolean form is added to the class of  vh, and becomes the representative formula.
    \see getSignalEquiv(), clk2Boolean(), endDefinition().
*/
_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::noFreeForm2bDC (ptinsth clk)
{
  ptexpcal vres, boolExp;
  ptinsth vh1, clkFather;

  clkFather = CKM->getFather (clk);
  if (getSignalEquiv (clk, clkFather, &vh1))
    {
      /* Redefinition du booleen vres par l'expression booleenne de la forme canonique */
      if (CKM->getOperator (vh1) == sub)	/* blindage */
	{
	  vres = CKM->getConditionSub (vh1);
	  if (clk2Boolean (clk, clkFather, &boolExp))
	    NM->modifiedDefinition (vres, boolExp, clkFather, clkFather);
	  else
	    clk_trace ("ERROR in noFreeForm2bDC :", clk);
	}
      else
	clk_trace ("ERROR in noFreeForm2bDC :", clk);
      CKM->setClockAsBoolean (clk, vres, EM->getPredecessors (vres), clkFather);
    }
  else if (clk2Boolean (clk, clkFather, &boolExp))
    setBooleanDefinition(clk, boolExp, clkFather);
  else
    clk_trace ("ERROR in noFreeForm2bDC :", clk);
}



/** 
    It is a specialisation of InterFormatTranslator::formulaCoding().

    It produces the boolean representation of a clock formula (clk).
    The boolean form is added to the class of clk, and becomes the representative formula.
    It calls freeForm2bDC(), noFreeForm2bDC().
 */
_VIRTUAL_ ptinsth
DCGEndo2DCGBoolTranslator::formulaCoding (ptinsth clk)
{
if (CKM->satisfiesSub (clk))
    freeForm2bDC (clk);
  else
    noFreeForm2bDC (clk);
 ptinsth vhnew = CKM->getRepresentativeFormula (clk);
 //  cout <<  " formulaCoding " << ident2string ( CKM->getName(clk) ) <<", num = " << CKM->getNum(clk) << " replaced by " << ident2string ( CKM->getName(vhnew) ) << ", num = " << CKM->getNum(vhnew) << endl;
 return vhnew;
}

/** 
    It links a clock (vclk) to an array-node (ArrayNode).
*/
void
DCGEndo2DCGBoolTranslator::linkClockToArray (ptinsth vclk, ptINode ArrayNode)
{
  if (ArrayNode != NILINODE)
    {
      NM->arrayBodyInsertClock (ArrayNode, vclk);
      CKM->setInComposedNode (vclk, ArrayNode);
    }
}

/** 
    It links a node (node) to an array-node (ArrayNode).
*/
void
DCGEndo2DCGBoolTranslator::linkNodeToArray (ptINode node, ptINode ArrayNode)
{
  if (ArrayNode != NILINODE)
    {
      NM->arrayBodyInsertNode (ArrayNode, node);
      NM->set_inComposedNode (node, ArrayNode);
    }
}

/**
   @return an extraction clock (undersampling by an expression) that represents a clock (clk) at another clock (hAt).
    
   The following order is used:
   - searching an equivalent at hAt clock.
   - searching an equivalent placed above hAt (hAt is included in such a clock), then creating a when. 
   - Creating a signal defined by a left_tt that represent the clock to hAt.

  \note It assumes that clk and hAT are on the same branch and hAT is the banching of clk and hAt. 
*/
ptinsth
DCGEndo2DCGBoolTranslator::mkBooleanClockScopeExtand (ptinsth clk, ptinsth hAT)
{
  ptinsth vhresult;
  ptexpcal vsig, bclk;
  integer rcode;

  rcode = CKM->getBooleanClockScopeExtandAt (clk, hAT, &vhresult);
  if (rcode == 0) return vhresult;
  bclk = CKM->getBooleanSignal (clk);
  if (rcode == 1)
    vsig = NM->createWhen (bclk, hAT, pk_false);
  else
    vsig = NM->createLeftt (bclk, CKM->getBooleanSignal (vhresult), hAT);
  return CKM->makeEquivForm (clk, vsig, EM->getPredecessors (vsig), hAT);
}


/** 
   It propagates to the references of clk, the boolean definition of clk.
   Booleans (equivalent to clk) are defined for each clock where clk is referenced.

   It references for that the following methods: OccurrencesBranchingClock(), mkBooleanClockScopeExtand().
   Moereover, the array blocks are managed linkClockToArray() if clk is a internal clock of an array.

   \note It assumes that the clock clk has been translated into xxxxbDC+ format.
*/
_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::propagateClock (ptinsth clk)
{
  ptinsth vhemb, vh, clkFrom;
  boolean b;
  ptINode vblockarray;

  clk_trace ("propagate clock <begin> :", clk);
  vhemb = OccurrencesBranchingClock (clk);
  clk = CKM->getRepresentativeFormula (clk);

  b = CKM->getInComposedNode (clk, &vblockarray);

  vhemb = CKM->getBranching (clk, vhemb);
  if (vhemb != clk)
    {
      clkFrom = CKM->getFather (clk);
      while (clkFrom != vhemb)
	{
	  vh = mkBooleanClockScopeExtand (clk, clkFrom);
	  if (b)
	    linkClockToArray (vh, vblockarray);
	  clkFrom = CKM->getFather (clkFrom);
	}
      vh = mkBooleanClockScopeExtand (clk, vhemb);
      if (b)
	linkClockToArray (vh, vblockarray);
    }
  clk_trace ("propagate clock <end> :", clk);
}

/**
   It codes the tick clock of the program (assumed endochronous).
   The definition is given by clk2Boolean() followed by endDefinition().
   When a clock is visited, it is marked with an internal mark (CST_MYMARK).

   Assumed: the program is assumed to be endochonous.
*/
_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::tickCoding_endo ()
{
  ptexpcal vtick, boolExp;
  ptinsth clk;

  clk = tickClock;
  boolExp = EM->getTrue ();
  endDefinition (clk, boolExp, clk);
  CKM->addMark (clk, CST_MYMARK);
  if (CKM->isClockRepresentedBySignal (clk, &vtick))
    NM->addMark (EM->getNode (vtick), DontOptimize);
}


/**
   April 2013...
*/
_VIRTUAL_ void
DCGEndo2DCGBoolTranslator::tickCoding ()
{
  if (CKM->getNbRoots () == 1)
    tickCoding_endo ();
}



/** 
   Removing input events of the graph.
   @param lsig (out) the inputs moved to outputs are added to this list.
   @return
   -"true" if the interface of the graph is modified (inputs moved to outputs); in this case the parameter (lsig) is updated, 
   - "false" otherwise.

   Assumed: the boolean clocks are available.
*/
boolean
DCGEndo2DCGBoolTranslator::removeInputEvents (ptlinstances * lsig)
{
  ptlinstances alreadySelected, list, inputs;
  ptexpcal vsig, exp, vtrue;
  ptinsth vh, vh1, vh2;
  ptINode n, aNode;
  boolean modified, b;

  inputs = AGRAPH->getInputs ();
  list = inputs;
  vtrue = EM->getTrue ();
  modified = pk_false;
  alreadySelected = NILLINSTANCES;
  while (ptlinstances_get_next (&list, &vsig))
    if (EM->IsEvent (vsig))
      {
	vh = EM->getClock (vsig);
	if (!CKM->satisfiesNullClock (vh))
	  {
	    b = pk_false;
	    if (!CKM->satisfiesTickClock (vh))
	      {
		exp = CKM->getConditionSub (vh);
		if (EM->IsInputFlow (exp) && (CKM->getOperator (vh) == sub))
		  if (!ptlinstances_isin_exp (alreadySelected, exp))
		    {
		      b = pk_true;
		      //  EM->setName (exp, EM->getName (vsig));
		      alreadySelected = ptlinstances_insert (alreadySelected, exp);
		    }
	      }

	    if (b)
	      {
		// mkDefAssertion (exp, vh);
		/* Removing the event (vsig) from the inputs. */
		inputs = ptlinstances_remove_exp (inputs, vsig);
		aNode = EM->getNode (vsig);
		NM->removeNode (aNode);
		NM->becomes_undefined (aNode);
		EM->substitute (vsig, vtrue, vh);
		CKM->removeFromClock (vh, vsig, pk_false);
	      }
	    else
	      {
		EM->setType (vsig, BooleanType);
		mkDefAssertion (vsig, vh);
	      }

	    modified = pk_true;
	  }
      }
  ptlinstances_free (alreadySelected);
  AGRAPH->setInputs (inputs);
  return modified;
}

/**
   It finalizes the translation.
*/
void
DCGEndo2DCGBoolTranslator::endOp ()
{
}


/*
  @return "true" if the a signal is rewritten by the following rules, "false" otherwise.
  Rules:
  - if x:= (true when c) default (false when K) then x:=c
  - if x:= (y when c) default (false when K)  with ^y = [c] then x:= leftt(y, ^x) 

  The reules are defined in the isLeftttDefinition() method.
*/
boolean
DCGEndo2DCGBoolTranslator::defaults2lefttts ()
{
  ptlINode lNodes;
  ptINode aNode;
  ptexpcal vopg, vopd, nvopd;
  ptinsth clkinstr;
  boolean b;

  b = pk_false;
  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      {
	vopg = NM->get_leftOperand (aNode);
	if (EM->IsBoolean (vopg))
	  if (NM->get_rightOperand (aNode, &vopd))
	    {
	      clkinstr = NM->get_nodeClock (aNode);
	      if (isLeftttDefinition (vopd, clkinstr, &nvopd))
		{
		  // if (vopg == nvopd) nvopd = EM->getTrue();  /*c := c   ==> c devient une entree ??? */
		  NM->modifiedDefinition (vopg, nvopd, clkinstr, clkinstr);
		  b = pk_true;
		}
	    }
      }
  return b;
}


/**
   @return "true" if the following rules can be applied, "false" otherwise.
   Rules (with x stands for vexp, clkinstr stands for ^x):
  - if x:= (true when c) default (false when K) then lefttExp=c
  - if x:= (y when c) default (false when K)  with ^y = [c] then lefttExp=leftt(y, ^x)
*/
boolean
DCGEndo2DCGBoolTranslator::isLeftttDefinition (ptexpcal vexp, ptinsth clkinstr, ptexpcal * lefttExp)
{
  ptexpcal vexpg, vexpd, vexp1, expFather;
  ptinsth clk_vexp1, hAt;
  noperators opd, opg;

  if (EM->IsCellExpr (vexp))
    return pk_false;
  vexp = EM->getBasicExpr (vexp);
  if (EM->getOperator (vexp) != opmele)
    return pk_false;
  vexpg = EM->getDefaultArgstAt (vexp, 1);
  vexpd = EM->getDefaultArgstAt (vexp, 2);
  /* simple default */
  opg = EM->getOperator (vexpg);
  if (opg == opmele)
    return pk_false;
  opd = EM->getOperator (vexpd);
  if (opd == opmele)
    return pk_false;

  if (EM->getWhenOrCellArg (vexpd) != EM->getFalse ())
    return pk_false;
  clk_vexp1 = EM->getContributionClock (vexpg);
  if (!CKM->getBooleanClockAt (clk_vexp1, CKM->getBranching (clk_vexp1, clkinstr), &hAt))
    return pk_false;
  clk_vexp1 = hAt;
  if (!CKM->satisfiesSub (clk_vexp1))
    return pk_false;
  if (CKM->getFather (clk_vexp1) != clkinstr)
    return pk_false;
  if (EM->getWhenOrCellArg (vexpg) == EM->getTrue ())
    {
      /* Case (true when H) default (false when K) */
      *lefttExp = CKM->getConditionSub (clk_vexp1);
      return pk_true;
    }
  vexp1 = EM->getWhenOrCellArg (vexpg);
  if (!EM->IsSignal (vexp1))
    return pk_false;
  /* case (y when c) default (false when K) */
  if (EM->getClock (vexp1) != clk_vexp1)
    return pk_false;
  expFather = CKM->getBooleanSignal (clk_vexp1);
  if (expFather == NILEXPCAL)
    return pk_false;
  *lefttExp = EM->mkBinaryExp (opleft_tt, vexp1, expFather, NILBACKANNOTATION);
  return pk_true;
}


/**
   For a when expression (whenExpr) such that E when C, it adds a dependence
   from the boolean contribution clock C at hnode clock, to the node (node).

   \note It assumes that 
   - the clock of node is hnode.
   - whenExpr is a when expression, a sub expression of the node.
*/
void
DCGEndo2DCGBoolTranslator::addContributionClockDependence (ptexpcal whenExpr, ptinsth hnode, ptINode node)
{
  ptinsth hAt, h, k;
  ptexpcal cond;

  h = EM->getContributionClock (whenExpr);
  k = CKM->getBranching (h, hnode);
  if (CKM->getBooleanClockAt (h, k, &hAt))
    {
      // Sept 2016 LBesnard --the 2 lines:
      // hAt = CKM->getRepresentativeFormula (hAt); 
      // if (!CKM->eqClock (h, hAt))
      // are replaced by: 
      if (h != hAt)
	{
	  cond = CKM->getConditionSub (hAt);
	  if (EM->IsSignal (cond)) 
	    {
	      NM->addDependence ( EM->getNode (cond), node, h);
	    }
	}
    }
}

/**
   It updates the contribution clocks of the equations (by construction X:= (E1 when C1) default ...(En when Cn)
   It calls addContributionClockDependence() method for each when sub expression.
*/
void
DCGEndo2DCGBoolTranslator::updateDependencesContributionClock ()
{
  ptlINode listNodes;
  ptINode node;
  ptexpcal vhorlexpcal, vexpcal;
  ptinsth hnode;

  if (NM->getEquations (&listNodes))
    while (ptlINode_get_elem (&listNodes, &node))
      {
	NM->get_rightOperand (node, &vhorlexpcal);
	if (!EM->isNILEXPCAL (vhorlexpcal))
	  {
	    hnode = NM->get_nodeClock (node);
	    if (EM->IsCellExpr (vhorlexpcal))
	      vhorlexpcal = EM->getWhenOrCellArg (vhorlexpcal);
	    vexpcal = EM->getWhenOrCellArg (vhorlexpcal);
	    while (EM->getOperator (vexpcal) == opmele)
	      {
		addContributionClockDependence (EM->getDefaultArgstAt (vexpcal, 2), hnode, node);
		vhorlexpcal = EM->getDefaultArgstAt (vexpcal, 1);
		vexpcal = EM->getWhenOrCellArg (vhorlexpcal);
	      }
	    addContributionClockDependence (vhorlexpcal, hnode, node);
	  }
      }

}


/**
   - It modifies the clocks of the array nodes: an array node is assigned now to the branching clock of its internal nodes.
   - It links the node defining internal (boolean) clocks of the arrays to these arrays.
*/
void
DCGEndo2DCGBoolTranslator::nodeClockUpdate ()
{
  ptINode aNode, n, n1;
  ptlINode lNodes, nln, ln;
  ptlinsth lh, lh1, nlh;
  boolean b;
  ptinsth vhemb, vh;
  ptexpcal vsig;

  // an array node is assigned now to the branching clock of its nodes.
  if (NM->getIterates (&lNodes))
    while (ptlINode_get_elem (&lNodes, &aNode))
      {
	NM->get_array_bodyDefinitions (aNode, &ln, &lh);
	nln = ln;
	lh1 = NILLINSTH;
	while (ptlINode_get_elem (&ln, &n))
	  lh1 = ptlinsth_sorted_insert_elem_bool (lh1, NM->get_nodeClock (n), &b);

	vhemb = CKM->getBranchingClocks (lh1);
	if (vhemb != NM->get_nodeClock (aNode))
	  NM->moveNodeToClock (aNode, vhemb);

	// Updating boolean clocks.
	while (ptlinsth_get_next (&lh, &vh))
	  if (CKM->satisfiesSub (vh))
	    {
	      vsig = CKM->getConditionSub (vh);
	      if (EM->IsSignal (vsig))
		{
		  n1 = EM->getNode (vsig);
		  linkNodeToArray (n1, aNode);
		}
	    }
      }
}

/**
   @return the expression for xxx
*/
ptexpcal DCGEndo2DCGBoolTranslator::mkExprAssertion (ptinsth vh)
{
  ptexpcal exp, vdef;

  exp = getBooleanSignalAt (vh, tickClock);
  if (EM->getSyncExprDefSig (exp, &vdef) && (EM->getOperator (vdef) == opnon))
    return EM->getUnaryArg (vdef);
  return EM->mkUnaryExp (opnon, exp, EM->getBackAnnotation (exp));
}


/**
   It translates each assertion (ie h1=h2 for placed clocks, assumed satisfied)
   by an intrinsic unsafe external call assert(b), specifying that b is always pk_true when it is present.

   The assert node is defined at

   - the tick clock if h1 (resp.h2) is the null clock, and b := not E, where E is the boolean signal of h2 (resp.h1) at tick 
   - the branching clock (k) of h1,h2, b:= E1=E2, where E is the boolean signal of h2 (resp.h1) at k.

   \note The boolean signal of a clock at another one is given by the getBooleanSignalAt() method.
*/
void
DCGEndo2DCGBoolTranslator::assertions2nodes ()
{
  ptlinsth lh1, lh2;
  ptinsth vhemb, vh1, vh2, htick;
  ptexpcal e, e1, e2, exp;

  htick = tickClock;
  lh1 = CKM->getPlacedAssertionNullClocks ();	// assertions on null clock.
  while (ptlinsth_get_next (&lh1, &vh1))
    if (!CKM->satisfiesIsAnEquivalentForm (vh1))
      {
	e = mkExprAssertion (vh1);
	mkDefAssertion (e, htick);
      }
  // assertions on placed clocks.
  if (CKM->getAssertionsOnPlacedClocks (&lh1, &lh2))
    while (ptlinsth_get_next (&lh1, &vh1))
      {
	ptlinsth_get_next (&lh2, &vh2);
	if (CKM->satisfiesNullClock (vh1))	// assertions on null clock. already seen
	  {
	    e = mkExprAssertion (vh2);
	    vhemb = htick;
	  }
	else if (CKM->satisfiesNullClock (vh2))	// assertions on null clock. already seen
	  {
	    e = mkExprAssertion (vh1);
	    vhemb = htick;
	  }
	else
	  {
	    vhemb = CKM->getBranching (vh1, vh2);
	    if (vhemb == vh1)
	      e = getBooleanSignalAt (vh2, vhemb);
	    else if (vhemb == vh2)
	      e = getBooleanSignalAt (vh1, vhemb);
	    else
	      {
		e1 = getBooleanSignalAt (vh1, vhemb);
		e2 = getBooleanSignalAt (vh2, vhemb);
		e = EM->mkBinaryExp (opegal, e1, e2, NILBACKANNOTATION);
	      }
	  }
	mkDefAssertion (e, vhemb);
      }
  // Added but they should be in getPlacedAssertionNullClocks() !!!
  if (CKM->getNullClockAssertions (&lh1)) 
    while (ptlinsth_get_next (&lh1, &vh1))
      {
	e = mkExprAssertion (vh1);
	mkDefAssertion (e, htick);
      }
}


/**
   It adds the assertion node assert(e) at the operand clock (vh) to the graph.
*/
void
DCGEndo2DCGBoolTranslator::mkDefAssertion (ptexpcal e, ptinsth vh)
{
  ptINode n;
  ptexpcal inputs;

  inputs = EM->mkNaryExp (oplident, e, NILEXPCAL, NILBACKANNOTATION);
  n = NM->mk_procCall (getAssertionId (), NILEXPCAL, inputs, NILEXPCAL, NILBACKANNOTATION);
  NM->set_intrinsicfunc (n);
  NM->set_safetyProperty (n, opunsafe);
  NM->insertNode (n, vh);
}


/** 
    Events are replaced by booleans except the tick of the program.
 */
void
DCGEndo2DCGBoolTranslator::events2booleans ()
{
  ptexpcal vTick;
  ptinsth hTick;

  EM->events2booleans ();
  hTick = tickClock;
  if (CKM->isClockRepresentedBySignal (hTick, &vTick))
    EM->setType (vTick, PureType);
  // Other solution : if the boolean type is assigned to the signal (B) that represents the tick, add assert(B) in the program.
}

void
DCGEndo2DCGBoolTranslator::local_init ()
{
  ptlINode lArrayNodes;
  ptINode aNode;

  if (NM->getIterates (&lArrayNodes))
    while (ptlINode_get_elem (&lArrayNodes, &aNode))
      NM->arrayLinkNodes (aNode, pk_false);
}

void
DCGEndo2DCGBoolTranslator::local_final ()
{
  ptlINode lArrayNodes;
  ptINode aNode;
  if (NM->getIterates (&lArrayNodes))
    while (ptlINode_get_elem (&lArrayNodes, &aNode))
      NM->arrayUnlinkNodes (aNode, pk_false);
}

