/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
   @copydoc AbstractHomomorphism::endHomomorphismNode
   Finalizing the translation of a node (n1) of the orignal graph in a node (n) of the new graph.
   It adds a link from n1 to n (see setNodeAttribute()).
*/
  _VIRTUAL_ void
Duplicator::endHomomorphismNode (ptINode n1, ptINode n)
{
  ptlINode ln;
  ptlinsth lh;
  ptINode elemnodearray;
  ptexpcal vlabel, nvlabel;
  ptinsth h;

  h = getAttribute (fNM->get_nodeClock (n1));
  sNM->insertNode (n, h);
  igc->takeIntoAccountContexts (n);
  sNM->addPragmas (n, fNM->getPragmas (n1)); // Il faudra dupliquer (contexte)
  setNodeAttribute (n1, n);
  sNM->set_safetyProperty (n, fNM->get_safetyProperty (n1));
  // if (fNM->is_intrinsicfunc (n1)) sNM->set_intrinsicfunc (n);

  if (sNM->is_array (n))
    {
      if (sNM->get_array_bodyDefinitions (n, &ln, &lh))
	{
	  while (ptlINode_get_elem (&ln, &elemnodearray))
	    sNM->set_inComposedNode (elemnodearray, n);
	}
    }
  if (! fNM->is_subgraph(n1) && fNM->get_label(n1, &vlabel))
    {
      nvlabel = mkCloneIdentifier(vlabel, n, h);
      sNM->set_label(n, nvlabel);
    }
}

/**
   @copydoc AbstractHomomorphism::homomorphismWrite
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismWrite (ptINode node)
{
  ptexpcal vsig1, vsig;
  ptINode n;

  if (getNodeAttribute (node, &n)) return n;
  vsig = fNM->get_io_arg (node);
  vsig1 = homomorphismExp (vsig);
  return sNM->mk_node_output (vsig1, fNM->getBackAnnotation(node));
}

/**
   @copydoc AbstractHomomorphism::homomorphismRead
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismRead (ptINode node)
{
  ptexpcal vsig1, vsig;
  ptINode n;

  if (getNodeAttribute (node, &n)) return n;
  vsig = fNM->get_io_arg (node);
  vsig1 = getAttribute (vsig);
  return sNM->mk_node_input (vsig1, NILINODE, fNM->getBackAnnotation(node));
}

/**
   @copydoc AbstractHomomorphism::homomorphismEqu
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismEqu (ptINode node)
{
  ptexpcal vopg, vopd, vopdn;
  ptinsth h;
  ptINode n;

  if (getNodeAttribute (node, &n)) return n;

  if (fNM->get_rightOperand (node, &vopd))
    {
      vopg = fNM->get_leftOperand (node);
      if (fEM->IsWhenOrCellExpr (vopd))
	{
	  vopd = fEM->getWhenOrCellArg (vopd);
	  if (fEM->IsWhenOrCellExpr (vopd))
	    vopd = fEM->getWhenOrCellArg (vopd);	// case : (e when h) cell k
	}
      vopdn = homomorphismExp (vopd);
      h = getAttribute (fNM->get_contributionClockEquation (node));
      return sNM->mk_node_equation (igc->mkFieldAccess (homomorphismExp (vopg), NILBACKANNOTATION), vopdn, h, NILLINSTANCES, NILBACKANNOTATION);
    }
  return sNM->mk_node_undefined ();
}

/**
   @copydoc AbstractHomomorphism::homomorphismDelay
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismDelay (ptINode node)
{
  ptexpcal vopg, sigret, vret, vfen, vinit;
  ptINode n;

  if (getNodeAttribute (node, &n)) return n;
  vopg = fNM->get_leftOperand (node);
  fNM->get_delay_arg (node, &vopg, &sigret, &vret, &vfen, &vinit);
  return sNM->mk_node_delay (igc->mkFieldAccess (homomorphismExp (vopg), NILBACKANNOTATION),
			     homomorphismExp (sigret), homomorphismExp (vret), homomorphismExp (vfen), homomorphismExp (vinit), NILBACKANNOTATION);
}

/**
   @copydoc AbstractHomomorphism::homomorphismNodecall
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismNodecall (ptINode node)
{
  ptexpcal lent, lsort, lparam;
  ptDCG vg;
  ptINode n;

  if (getNodeAttribute (node, &n)) return n;
  fNM->get_nodeCall_arg (node, &vg, &lparam, &lent, &lsort);
  n = sNM->mk_node_call (vg, homomorphismExp (lparam), homomorphismExp (lent), igc->mkFieldAccess (homomorphismExp (lsort), NILBACKANNOTATION), fNM->getBackAnnotation (node));
  return n;
}


/**
   @copydoc AbstractHomomorphism::homomorphismProccall
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismProccall (ptINode node)
{
  ptexpcal lin, lout, lpar;
  ptident nom;
  ptINode n;

  if (getNodeAttribute (node, &n)) return n;
  fNM->get_procCall_arg (node, &nom, &lpar, &lin, &lout);
  n = sNM->mk_procCall (nom, homomorphismExp (lpar), homomorphismExp (lin), igc->mkFieldAccess (homomorphismExp (lout), NILBACKANNOTATION), fNM->getBackAnnotation (node));
  if (fNM->is_intrinsicfunc (node)) sNM->set_intrinsicfunc (n);

  return n;
}

/**
   @copydoc AbstractHomomorphism::homomorphismSubGraph
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismSubGraph (ptINode node)
{
  ptinsth h, nh;
  ptlinsth lh;
  ptlINode ln;
  ptINode n, nsg, aNode;
  ptexpcal vlabel, vlabelCopy;

  if (getNodeAttribute (node, &n)) return n;
  fNM->get_label (node, &vlabel);
  vlabelCopy = getAttribute (vlabel);
  nsg = sNM->mk_node_subgraph (vlabelCopy, fNM->getBackAnnotation (node));
  fNM->get_subGraphDefinitions (node, &ln, &lh);
  while (ptlINode_get_elem (&ln, &aNode))
    if (homomorphismNode (aNode, &n))
      sNM->subGraphInsertNode (nsg, n);

  while (ptlinsth_get_next (&lh, &h))
    if (isSelectedClock (h))
      {
	homomorphismClock (h);
	getClockAttribute (h, &nh);
	sNM->subGraphInsertClock (nsg, nh);
      }

  sNM->unlinkNodesFromClock (nsg);
  return nsg;
}

_VIRTUAL_ ptINode
Duplicator::homomorphismIndex (ptINode node)
{
  ptINode indexNode;
  ptexpcal vIndex, indexDef;

  indexNode = fNM->get_indexNode (node);
  vIndex = fNM->get_index (indexNode);
  indexDef = fNM->get_index_def (indexNode);
  return sNM->mk_node_index (homomorphismExp (vIndex), homomorphismExp (indexDef));
}


_VIRTUAL_ ptINode
Duplicator::homomorphismArray (ptINode node)
{
  ptlINode ln;
  ptlinsth lh;
  ptINode m_node, n, m_n;
  ptexpcal inputs, outputs;

  if (getNodeAttribute (node, &m_node))
    return m_node;

  fNM->get_array_io (node, &inputs, &outputs);
  m_node = sNM->mk_node_array (homomorphismIndex (node), homomorphismExp (inputs), homomorphismExp (outputs), fNM->getBackAnnotation (node));
  if (fNM->get_array_bodyDefinitions (node, &ln, &lh))
    {
      while (ptlINode_get_elem (&ln, &n))
	if (homomorphismNode (n, &m_n))
	  sNM->arrayBodyInsertNode (m_node, m_n);
      // voir pour les horloges ...; 
      // arrayBodyInsertClock( ptINode node, ptinsth vh)
    }
  if (fNM->get_array_initDefinitions (node, &ln, &lh))
    {
      while (ptlINode_get_elem (&ln, &n))
	if (homomorphismNode (n, &m_n))
	  sNM->arrayInitInsertNode (m_node, m_n);
    }
  return m_node;
}

/**
   @copydoc AbstractHomomorphism::homomorphismPartialDefs
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismPartialDefs (ptINode node)
{
  ptINode n;
  if (getNodeAttribute (node, &n))
    return n;
  return sNM->mk_node_undefined ();
}

/**
   @copydoc AbstractHomomorphism::homomorphismPartialDef
*/
_VIRTUAL_ ptINode
Duplicator::homomorphismPartialDef (ptINode node)
{
  ptINode n, setofDefs;
  ptexpcal vopg, nvopg, vopd, vopdn, x;
  ptinsth h;

  if (getNodeAttribute (node, &n))
    return n;

  vopg = fNM->get_leftOperand (node);
  nvopg = homomorphismExp (vopg);
  sEM->addMark (nvopg, PART_DEFINITION);
  fNM->get_rightOperand (node, &vopd);
  vopdn = homomorphismExp (fEM->getBasicExpr (vopd));
  h = getAttribute (fNM->get_contributionClockEquation (node));
  x = igc->mkFieldAccess (nvopg, NILBACKANNOTATION);
  n = sNM->mk_node_equation (x, vopdn, h, NILLINSTANCES, NILBACKANNOTATION);
  setofDefs = sEM->getNode (x);	// != n for partial definition.
  endHomomorphismNode (fEM->getNode (vopg), setofDefs);
  sEM->removeMark (nvopg, PART_DEFINITION);
  return n;
}

/**
   @copydoc AbstractHomomorphism::isSelectedSignal
*/
_VIRTUAL_ boolean
Duplicator::isSelectedSignal (ptexpcal x)
{
  if (selectMark == DUPLICATEALL)
    return !ptlINode_isIn (unNodes, fEM->getNode (x));
  return fEM->isMarked (x, selectMark);
}

/**
   @copydoc AbstractHomomorphism::isSelectedNode
*/
_VIRTUAL_ boolean
Duplicator::isSelectedNode (ptINode n)
{
  if (selectMark == DUPLICATEALL)
    return !ptlINode_isIn (unNodes, n);
  return fNM->isMarked (n, selectMark);
}


/**
   @copydoc AbstractHomomorphism::isSelectedClock
*/
_VIRTUAL_ boolean
Duplicator::isSelectedClock (ptinsth h)
{
  ptinsth hc;

  if (fCKM->satisfiesSub (h))
    {
      fCKM->getComplExtraction (h, &hc);
      if (ptlinsth_isin_elem (unClocks, h) && ptlinsth_isin_elem (unClocks, hc))
	return pk_false;
    }
  else
    if (ptlinsth_isin_elem (unClocks, h)) return pk_false;

  if (selectMark == DUPLICATEALL) return pk_true;
  return fCKM->satisfiesMarked (h, selectMark);
}

_VIRTUAL_ void
Duplicator::propagateAttributes (ptexpcal vsig, ptexpcal nsig)
{
  if (fEM->isMarked (vsig, DONT_CHANGE_TYPE))
    sEM->addMark (nsig, DONT_CHANGE_TYPE);
}

ptinsth Duplicator::assignCloneClock(ptexpcal vsig, ptexpcal nsig)
{
  ptinsth clkCopy, clk;
  clk = fEM->getClock (vsig);
  clkCopy = getAttribute (clk);
  sEM->setClock (nsig, clkCopy);
  return clkCopy;
}

_VIRTUAL_ void
Duplicator::homomorphismSignal (ptexpcal vsig, ptexpcal nsig)
{
  ptinsth clkCopy;
  ptexpcal vClone, nvClone;

  clkCopy = assignCloneClock(vsig, nsig);
  fEM->setInitialValue (nsig, homomorphismExp (fEM->getInitialValue (vsig)));
  propagateAttributes (vsig, nsig);

  // The clones (array structure)
  if (fEM->getCloneClassHead (vsig) == vsig)
    {
      while (fEM->getClone (vsig, &vClone) && !fEM->isCloneClassHead (vClone))
	{
	  if (!getSigAttribute (vClone, &nvClone))
	    {
	      nvClone = sEM->mkIdentClone (vsig, sNM->mk_node_undefined ());
	      setSigAttribute (vClone, nvClone);
	      sEM->setClock (nvClone, clkCopy);
	    }
	  vsig = vClone;
	}
    }
}

ptexpcal Duplicator::mkCloneIdentifier(ptexpcal vsig, ptINode vnode, ptinsth vclk)
{
  ptexpcal nsig;
  ptlpragmas lprag;
  ptpragma vpragma;
  
  nsig = sEM->mkIdentCopy (vsig, vnode, vclk);
  sEM->setTreeOfSignalDeclaration (nsig, fEM->getTreeOfSignalDeclaration (vsig));
  lprag = fEM->getPragmas (vsig);
  while (ptlpragmas_get_next (&lprag, &vpragma))
    sEM->addPragma (nsig, vpragma); // assumed: such a pragma is without assigned objects.
  setSigAttribute (vsig, nsig);
  return nsig;
}

/**
   @return a created copy of a signal (vsig)
*/
_VIRTUAL_ ptexpcal
Duplicator::mkAttribute (ptexpcal vsig)
{
  ptexpcal nsig;
  typinst t;

  nsig = mkCloneIdentifier(vsig, sNM->mk_node_undefined (), sCKM->getUndefClock ());
  t = fEM->getClassIdentifier (vsig);
  switch (t)
    {
    case tsignal:
      homomorphismSignal (vsig, nsig);
      break;
    case tparam:
      homomorphismParameter (vsig, nsig);
      break;
    case tconstant:
      homomorphismConstant (vsig, nsig, pk_true);
      break;
    case tindex:
      sGraph->insertIndex (nsig);
      break;
      // case tlabel: case ttyp: 
    default:
      ;
    }
  return nsig;
}

/**
   @copydoc AbstractHomomorphism::homomorphismDclIdentifier
*/
_VIRTUAL_ void
Duplicator::homomorphismDclIdentifier (ptexpcal vsig)
{
  if (isSelectedSignal (vsig)) getAttribute (vsig);
}

/**
   @return the copy of the operand (vsig). When it does not exist, it is created.
*/
_VIRTUAL_ ptexpcal
Duplicator::getAttribute (ptexpcal vsig)
{
  ptexpcal nsig;

  if (getSigAttribute (vsig, &nsig)) return nsig;
  return mkAttribute (vsig);
}

/**
   @return the copy of an operand clock (clk). When it does not exist, it is created.
*/
_VIRTUAL_ ptinsth
Duplicator::getAttribute (ptinsth clk)
{
  ptinsth nclk;

  clk = fCKM->getRepresentativeFormula (clk);
  // Attention au "tick"
  if (fCKM->satisfiesNullClock (clk)) return sCKM->getNullClock ();
  if (fCKM->satisfiesUndefClock (clk)) return sCKM->getUndefClock ();
  if (fCKM->satisfiesFictitiousRoot (clk)) return sCKM->getFictitiousRoot ();	// state variables.
  if (!getClockAttribute (clk, &nclk))
    {
      nclk = sCKM->new_variFormulaRootClock ();
      setClockAttribute (clk, nclk);
      ClocksNotCloned = ptlinsth_sorted_insert_elem (ClocksNotCloned, clk);
    }
  return sCKM->getRepresentativeFormula (nclk);
}

/**
   It assigns the copy (nclk) to a clock (clk). When such a copy is already assigned with it 
   a synchronization is added.
*/
_VIRTUAL_ void
Duplicator::setAttribute (ptinsth clk, ptinsth nclk)
{
  ptinsth h;

  fCKM->addMark (clk, aMark);
  clk = fCKM->getRepresentativeFormula (clk);
  if (getClockAttribute (clk, &h))
    nclk = sCKM->makeSynchro (sCKM->getRepresentativeFormula (h), nclk);
  setClockAttribute (clk, nclk);
}

/**
   @copydoc AbstractHomomorphism::getCurrentInstances
*/
_VIRTUAL_ ptlinstances
Duplicator::getCurrentInstances (ptlinstances ls)
{
  ptexpcal vsig;
  ptlinstances lsn;

  lsn = NILLINSTANCES;
  while (ptlinstances_get_next (&ls, &vsig))
    if (isSelectedSignal (vsig))
      lsn = ptlinstances_insert (lsn, getAttribute (vsig));
  return lsn;
}

/**
   It duplicates the definition of a parameter (vsig) in the new graph,
   the copy of vsig is nsig in the new graph.
*/
_VIRTUAL_ void
Duplicator::homomorphismParameter (ptexpcal vsig, ptexpcal nsig)
{
  homomorphismConstant (vsig, nsig, pk_false);
}


/**
   It duplicates the definition of a constant (vsig) in the new graph,
   the copy of vsig is nsig in the new graph.
*/
_VIRTUAL_ void
Duplicator::homomorphismConstant (ptexpcal vsig, ptexpcal nsig, boolean bconst)
{
  ptlinstances curlist;
  ptINode n;

  if (bconst)
    ptlinstances_isin_insert_exp (&lconstants, &curlist, nsig);
  else
    ptlinstances_isin_insert_exp (&lparameters, &curlist, nsig);
  n = fEM->getNode (vsig);
  setNodeAttribute (n, nodeHomomorphism (n));
}

/**
   It propagates the pragmas and the attributes from exp (element of the current graph) to nexp (element of the copy).
*/
_VIRTUAL_ void
Duplicator::homomorphismCallExpAnnotations (ptexpcal exp, ptexpcal nexp)
{
  ptlpragmas lprag;
  ptpragma vpragma;
  ntrees vlang;
  // Called for function call in an expression.
  sEM->setBackAnnotation (nexp, fEM->getBackAnnotation (exp));
  lprag = fEM->getPragmas (exp);
  while (ptlpragmas_get_next (&lprag, &vpragma))
    sEM->addPragma (nexp, vpragma);
  if (fEM->getAttribute (exp, EXTERN_LANGUAGE_IMPL, (attributeValue *) & vlang))
    sEM->addAttribute (nexp, EXTERN_LANGUAGE_IMPL, (attributeValue) vlang);
  if (fEM->isMarked (exp, INTRINSIC))
    sEM->addMark (nexp, INTRINSIC);
}

/**
   @return the representation of an expression (exp)  of the original graph in the new one.
*/
_VIRTUAL_ ptexpcal
Duplicator::homomorphismExp (ptexpcal exp)
{
  noperators op;
  BackAnnotationType t;
  ptexpcal nexp, lp, inputs;
  ptident id;

  if (exp == NILEXPCAL) return exp;
  op = fEM->getOperator (exp);
  t = fEM->getBackAnnotation (exp);
  switch (fEM->getOperatorClass (exp))
    {
    case variabl:
      return getAttribute (exp);
    case constant:
      return sEM->mkConstantExp (op, fEM->getConstant (exp));
    case oper1:
      return sEM->mkUnaryExp (op, homomorphismExp (fEM->getUnaryArg (exp)), t);
    case oper1tp:
      return sEM->mkUnaryExpTp (op, homomorphismExp (fEM->getUnaryArg (exp)), t);
    case oper2:
      return sEM->mkBinaryExp (op, homomorphismExp (fEM->getBinaryArgstAt (exp, 1)), homomorphismExp (fEM->getBinaryArgstAt (exp, 2)), t);
    case oper2tp:
      {
	nexp = sEM->mkBinaryExpTp (op, homomorphismExp (fEM->getBinaryArgstAt (exp, 1)), homomorphismExp (fEM->getBinaryArgstAt (exp, 2)), t);
	fEM->setContributionClock (nexp, getAttribute (fCKM->getRepresentativeFormula (fEM->getContributionClock (exp))));
	return nexp;
      }

    case oper3:
    case oper3tp:
      return sEM->mkTernaryExp (op, homomorphismExp (fEM->getTernaryArgstAt (exp, 1)),
				homomorphismExp (fEM->getTernaryArgstAt (exp, 2)), homomorphismExp (fEM->getTernaryArgstAt (exp, 3)), t);
    case opereh:
      return sEM->mkWhenOrCellExp (op, homomorphismExp (fEM->getWhenOrCellArg (exp)),
				   getAttribute (fCKM->getRepresentativeFormula (fEM->getContributionClock (exp))));

    case opertyp:
      return sEM->mkExpType (fEM->getExpType (exp));

    case call:
      fEM->getFunctionCall (exp, &id, &lp, &inputs);
      nexp = sEM->mkCallExp (id, homomorphismExp (lp), homomorphismExp (inputs), fEM->getType (exp), getAttribute (fEM->getClock (exp)));
      homomorphismCallExpAnnotations (exp, nexp);
      return nexp;

    default:
      pK_messenger->errorMethodFile ("homomorphismExp", "Duplicator_m.c");
    }

  return NILEXPCAL;
}


/**
   @return  the representation of an extraction  (formula) of the original graph in the new one.
*/
_VIRTUAL_ ptinsth
Duplicator::formulaHomomorphismExtract (ptinsth formula)
{
  ptexpcal cond, nCond;
  ptinsth nf, hc, hcn, clkCond, nclkCond;
  noperators op;

  clkCond = fCKM->getClockSub (formula);
  cond = fCKM->getConditionSub (formula);
  nCond = homomorphismExp (cond);
  nclkCond = getAttribute (clkCond);
  fCKM->getComplExtraction (formula, &hc);
  if (fEM->IsEvent (nCond))
    {
      nf = nclkCond;
      hcn = sCKM->getNullClock ();
    }
  else
    {
      op = fEM->getOperator (nCond);
      if (op == opfalse)
	{
	  nf = sCKM->getNullClock ();
	  hcn = nclkCond;
	}
      else
	{
	  nf = sCKM->makeSub (nclkCond, nCond, sEM->getPredecessors (nCond));
	  hcn = sCKM->makeNoSub (nCond);
	  
	}
    }
  if (!fCKM->satisfiesIsAnEquivalentForm (hc))
    {
      setAttribute (hc, hcn);
      unificationSignalClock(hc, hcn);
    }
  return nf;
}

/** 
    @return  the representation of a formula  (formula) of the original graph in the new one.
 */
_VIRTUAL_ ptinsth
Duplicator::formulaHomomorphism (ptinsth formula)
{
  ptinsth nf, h1, h2, vhemb;
  ptlpragmas lprag;
  ptpragma vpragma;

  if (fCKM->satisfiesNullClock (formula))
    nf = sCKM->getNullClock ();
  else if (fCKM->satisfiesUndefClock (formula))
    nf = sCKM->getUndefClock ();
  else
    {
      switch (fCKM->getOperator (formula))
	{
	case sub:
	  nf = formulaHomomorphismExtract (formula);
	  break;
	case nosub:
	  nf = getAttribute (formula);
	  break;
	case infer:
	  if (fCKM->getMinusArgs (formula, &h1, &h2))
	    nf = sCKM->makeMinus (getAttribute (h2), getAttribute (h1));
	  else
	    {
	      fCKM->getBinaryArgs (formula, &h1, &h2);
	      nf = sCKM->makeInf (getAttribute (h1), getAttribute (h2));
	    }
	  break;
	case super:
	  fCKM->getBinaryArgs (formula, &h1, &h2);
	  nf = sCKM->makeSup (getAttribute (h1), getAttribute (h2));
	  break;

	case hcomp:
	  fCKM->getUnaryArg (formula, &h1);
	  vhemb = fCKM->getBranching (formula, h1);
	  if (bExportHierarchy)	// added Sep 2014
	    nf = sCKM->new_compClockInTree (getAttribute (h1));
	  else
	    nf = sCKM->makeMinus (getAttribute (vhemb), getAttribute (h1));
	  break;
	case vari:
	  nf = getAttribute (formula);
	  break;
	case hnot:		// pas d'occurrences
	case predef:		// pas d'occurrences
	default:
	  pK_messenger->Caseerror (currentLine, currentFile);
	}

      setAttribute (formula, nf);
    }

  // revoir gestion homogeme des pragmas.
  lprag = fCKM->getPragmas (formula);
  while (ptlpragmas_get_next (&lprag, &vpragma))
    sCKM->addPragma (nf, vpragma);
  return nf;
}


/**
   It duplicates the formula of a class (h) in a new one (nh).
   h is not duplicated here, its duplication is assumed to be nh.
*/
_VIRTUAL_ void
Duplicator::homomorphismFormulaClass (ptinsth h, ptinsth nh)
{
  ptinsth nh1, h1;
  boolean b, bNullClock;

  bNullClock = fCKM->satisfiesNullClock(h);
  if (bEventClockedFormat)
    {
      h1 = fCKM->getNextFormula (h, pk_true);
      while (h1 != h)
	{
	  if (bNullClock) b = ptlinsth_isin_elem (lConstraintsNullClock, h1); else b=true;
	  if (b)
	    {
	      nh1 = formulaHomomorphism (h1);
	      setClockAttribute(h1, nh1);
	      nh = sCKM->makeSynchro (nh, nh1);
	      h1 = fCKM->getNextFormula (h1, pk_true);
	    }
	}
      h1 = h;
      while (fCKM->getNextAssertion (h1, &h1))
	{
	  if (bNullClock) b = ptlinsth_isin_elem (lAssertionsNullClock, h1); else b=true;
	  if (b)
	    {
	      nh1 = formulaHomomorphism (h1);
	      setClockAttribute(h1, nh1);
	      nh = sCKM->makeAssumption (nh1, nh);
	    }
	}
    }
}

_VIRTUAL_ boolean
Duplicator::isExportedSignal (ptexpcal vsig)
{
  return pk_true;
}

/**
   @copydoc AbstractHomomorphism::homomorphismClock
*/
_VIRTUAL_ void
Duplicator::homomorphismClock (ptinsth h)
{
  ptinsth nh;
  boolean b;

  h = fCKM->getRepresentativeFormula (h);
  ClocksNotCloned = ptlinsth_remove_elem (ClocksNotCloned, h, &b);
  if (isSelectedClock (h))
    if (!fCKM->satisfiesMarked (h, aMark))
      {
	fCKM->addMark (h, aMark);
	if (h != rootClock)
	  {
	    nh = formulaHomomorphism (h);
	    homomorphismFormulaClass (h, nh);
	    unificationSignalClock(h, nh);
	    ClocksNotCloned = ptlinsth_remove_elem (ClocksNotCloned, h, &b);
	  }
      }
}


_VIRTUAL_ void
Duplicator::unificationSignalClock(ptinsth h, ptinsth nh)
{
  ptexpcal nvsig, vsig;
  
  // Unification Signal/clock 
  if (fCKM->isClockUnifiedWithSignal (h, &vsig))
    if (!fEM->isSharedSignalBySeveralGraphs (vsig))
      {
	nvsig = getAttribute (vsig);
	sCKM->setClockUnifiedWithSignal (nh, nvsig);
	sEM->setSignalUnifiedWithClock (nvsig, nh);
      }
  // Signal that represents the clock.
  if (fCKM->isClockRepresentedBySignal (h, &vsig))
    if (!fEM->isSharedSignalBySeveralGraphs (vsig))
      {
	nvsig = getAttribute (vsig);
	sCKM->setClockRepresentedBySignal (nh, nvsig);
      }
}


/**
   It adds in the new graph (copy) a dependence 
   - from nNode to each element of lh
   - from nNode to each element of larcs   
*/
_VIRTUAL_ void
Duplicator::homomorphismDependence (ptINode nNode, ptlinsth lh, ptldependence larcs)
{
  ptinsth h, nh;
  ptdependence arc;
  ptINode vto, vfrom, nvto;

  while (ptlinsth_get_next (&lh, &h))
    if (getClockAttribute (h, &nh))
      sNM->addDependence (nNode, nh);

  while (larcs->get_next (&arc, &larcs))
    {
      h = arc->getDependence (&vfrom, &vto);	// vfrom :: aNode
      h = fCKM->getRepresentativeFormula (h);
      if (getNodeAttribute (vto, &nvto))
	{
	  if (!getClockAttribute (h, &nh))
	    {
	      TRACE ("Duplicator::homomorphismDependence, clock not imported !");
	      nh = hundef;
	    }
	  sNM->addDependence (nNode, nvto, nh, arc->getDependenceKind ());
	}
    }
}

/**
   @copydoc AbstractHomomorphism::homomorphismDependence
*/
_VIRTUAL_ void
Duplicator::homomorphismDependence (ptINode aNode)
{
  ptINode nNode;
  ptldependence larcs;
  ptlinsth lh;

  if (getNodeAttribute (aNode, &nNode))
    {
      lh = fNM->get_clockSuccessors (aNode);
      larcs = fNM->get_successors (aNode);
      homomorphismDependence (nNode, lh, larcs);
    }
}

/**
   @copydoc AbstractHomomorphism::homomorphismDependence
*/
_VIRTUAL_ void
Duplicator::homomorphismDependence (ptinsth h)
{
  ptINode aNode, nNode;
  ptinsth nh, nsucc, succ;
  ptlinsth lhsucc;
  ptlINode ln;

  if (getClockAttribute (h, &nh))
    {
      lhsucc = fCKM->getClockReferencesTo (h);
      while (ptlinsth_get_next (&lhsucc, &succ))
	if (getClockAttribute (succ, &nsucc))
	  if ( ! ( ( nsucc == nh ) && (h != succ))) // Added Jan 2016 LBesnard.
	    sNM->addDependence (nh, nsucc);

      ln = fCKM->getNodeReferencesTo (h);
      while (ptlINode_get_elem (&ln, &aNode))
	if (getNodeAttribute (aNode, &nNode))
	  {
	    sNM->addDependence (nh, nNode);
	  }
    }
}


/**
   @copydoc AbstractHomomorphism::finalize
*/
_VIRTUAL_ void
Duplicator::finalize ()
{
  ptlinstances ls;
  ptexpcal vsig;

  // sGraph ->computeControlClocks(); // compute step/iostep of the new graph.
  sEM->rewritingExprEH ();	// update clock of cell/when expression
  if (ClocksNotCloned != NILLINSTH)
    pK_messenger->printString ("INTERNAL ERROR: Duplicator::finalize() : The list ClocksNotCloned is not empty ");
}

/**
   It assigns the name of a clock (aClock) to its copy (aClockCopy).
*/
_VIRTUAL_ void
Duplicator::setClockName (ptinsth aClock, ptinsth aClockCopy)
{
  sCKM->setName (aClockCopy, fCKM->getName (aClock));	// Copy of the original name
}

/**
   @copydoc AbstractHomomorphism::homomorphismHierarchy
*/
_VIRTUAL_ void
Duplicator::homomorphismHierarchy (ptlinsth lh)
{
  ptinsth aClock, fatherClock, aClockCopy, fatherClockCopy, vh1, vh2, vhc;
  ptlinsth lhsave;

  lhsave = lh;
  while (ptlinsth_get_next (&lh, &aClock))
    {
      vh1 = fCKM->getRepresentativeFormula (aClock);
      if (getClockAttribute (vh1, &aClockCopy))
	{
	  setClockName (aClock, aClockCopy);
	  if (aClock != rootClock)
	    if (!fCKM->satisfiesRoot (aClock))
	      {
		fatherClock = fCKM->getFather (aClock);
		fatherClockCopy = getAttribute (fatherClock);
		aClockCopy = sCKM->addPlacedClone (aClockCopy, fatherClockCopy, aClock);

		if (fCKM->satisfiesSub (vh1))
		  {
		    fCKM->getComplExtraction (vh1, &vh2);
		    if (!ptlinsth_isin_elem (lhsave, vh2))
		      {
			if (getClockAttribute (vh2, &aClockCopy))
			  {
			    setClockName (vh2, aClockCopy);
			    aClockCopy = sCKM->addPlacedClone (aClockCopy, fatherClockCopy, vh2);
			  }
			else
			  {
			    sCKM->getComplExtraction (aClockCopy, &vhc);
			    aClockCopy = sCKM->addPlacedClone (vhc, fatherClockCopy, vh2);
			  }
		      }
		  }
	      }
	}
    }
}


/**
   @copydoc AbstractHomomorphism::homomorphismConstraint
*/
_VIRTUAL_ void
Duplicator::homomorphismConstraint (ptinsth vh1, ptinsth vh2)
{
  ptinsth vhn1, vhn2;

  vh1 = fCKM->getRepresentativeFormula (vh1);
  vh2 = fCKM->getRepresentativeFormula (vh2);
  if (vh1 != vh2)
    if (getClockAttribute (vh1, &vhn1))
      {
	if (getClockAttribute (vh2, &vhn2))
	  sCKM->addConstraintPlacedClocks (vhn1, vhn2);
	else
	  printf("Duplicator::homomorphismConstraint no object in duplicated graph\n");
      }
    else
      printf("Duplicator::homomorphismConstraint no object in duplicated graph\n");
}


/**
   @copydoc AbstractHomomorphism::initialize
*/
_VIRTUAL_ void
Duplicator::initialize (ptDCG g1, ptDCG g2, boolean bExportHierarchy)
{
  AbstractHomomorphism::initialize (g1, g2);
  lparameters = NILLINSTANCES;
  lconstants = NILLINSTANCES;
  setAttribute (fCKM->getInitClock (), sCKM->getInitClock ());
  aMark = getMark ();
  rootClock = fCKM->getUndefClock ();
  bEventClockedFormat = isEventClockedFormat ();
  ClocksNotCloned = NILLINSTH;
  setExportHierarchy (bExportHierarchy);
}

/**
   @copydoc AbstractHomomorphism::homomorphismAttribute
*/
_VIRTUAL_ void
Duplicator::homomorphismAttribute (ptexpcal vsig)
{
  ptexpcal sigCopy;
  ptinsth h, clkCopy;
  ptINode n;

  if (!fEM->isSharedSignalBySeveralGraphs (vsig))
    if (fEM->isSignalUnifiedWithClock (vsig, &h))
      {
	clkCopy = getAttribute (h);
	sigCopy = getAttribute (vsig);
	sEM->setSignalUnifiedWithClock (sigCopy, clkCopy);
	sCKM->setClockUnifiedWithSignal (clkCopy, sigCopy);
      }

  n = fEM->getNode (vsig);
  if (fNM->isMarked (n, DontOptimize))
    sNM->addMark (n, DontOptimize);
}

/**
   @return a copy of g.
*/
ptDCG Duplicator::mkCopy (ptDCG g, boolean bExportHierarchy)
{
  ptDCG gcopy;

  gcopy = new DCG ();
  gcopy->setName (g->getName ());
  copyIn (g, gcopy, bExportHierarchy);
  return gcopy;
}

/**
   It defines the interface of the new graph.
   See AbstractHomomorphism::homomorphismInterface()
*/
_VIRTUAL_ void
Duplicator::homomorphismInterface ()
{
  sGraph->setParameters (lparameters);
  sGraph->setConstants (lconstants);
  AbstractHomomorphism::homomorphismInterface ();
}

/**
   This method defines the copy of a original graph to a new one assumed here available (the initialize() method has been applied).
   It applies sequentially the following methods:
   - homomorphismDclIdentifiers() it assigns a copy in the new graph for all the identifiers of the original one
   - homomorphismNodes() for all the nodes of the original graph
   -  homomorphismClocks()  for all the clocks of the original graph
   - homomorphismHierarchy()  for all the clocks of the original graph
   - homomorphismUndefinedIdentifiers(ls)  for all the signals of the original graph
   - homomorphismDependences() for all the clocks of the original graph
   - homomorphismDependences() for all the nodes of the original graph
   - homomorphismPragmas() 
   - homomorphismAttributes()  
   - homomorphismConstraints()
   - homomorphismInterface()
*/
_VIRTUAL_ void
Duplicator::copyGraph ()
{
  ptlinsth lh;
  ptlINode ln;
  ptlinstances ls;

  // Getting nodes/clocks of original graph.
  fNM->getAllNodes (&ln);
  ln = ptlINode_Copy (ln);

  lh = ptlinsth_duplicate (fCKM->getlhroots ());
  lh = ptlinsth_sorted_fusion (lh, fCKM->getlhplaced ());

  homomorphismDclIdentifiers ();
  homomorphismNodes (ln);
  homomorphismClocks (lh);
  if (getExportHierarchy ())
    homomorphismHierarchy (lh);
  if (fEM->getExprList (opident, &ls))
    homomorphismUndefinedIdentifiers (ls);
  homomorphismDependences (lh);
  homomorphismDependences (ln);
  homomorphismPragmas ();
  homomorphismAttributes ();
  homomorphismConstraints ();
  homomorphismInterface ();
}

/**
   @copydoc AbstractHomomorphism::homomorphismUndefinedIdentifier
*/
_VIRTUAL_ void
Duplicator::homomorphismUndefinedIdentifier (ptexpcal vsig)
{
  ptexpcal nsig;
  ptINode n, n1;

  if (isSelectedSignal (vsig))
    {
      n = sEM->getNode (vsig);
      if (!getNodeAttribute (n, &n1))
	{
	  nsig = getAttribute (vsig);
	  n1 = sEM->getNode (nsig);
	  endHomomorphismNode (n, n1);
	}
    }
}


void
Duplicator::setExportHierarchy (boolean b)
{
  bExportHierarchy = b;
}

boolean
Duplicator::getExportHierarchy ()
{
  return bExportHierarchy;
}

void Duplicator::replaceInterface(ptDCG ng, ptinterface vinter)
{
  ptlinstances lin, lout, lin1, lout1;

  lin1 = ptinterface_get_inputSignals(vinter);
  lin = getCurrentInstances (lin1);
  ng->setInputs (lin);
  lout1 = ptinterface_get_outputSignals(vinter);
  lout = getCurrentInstances (lout1);
  ng->setOutputs (lout);
}
