/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/* Constructor.
   The functionality is applied to a graph (GraphInitial).
*/
 InputsTrainClusters::InputsTrainClusters(ptDCG GraphInitial):TrainClusters(GraphInitial)
{
}

/**
   Destructor.
*/
_VIRTUAL_ InputsTrainClusters::~InputsTrainClusters()
{
}

/**
   @return the list of sub-graphs, a sub-graph is a cluster such that 
   n and m are in the same sub-graph is they depend on the same sub-set on inputs of the graph.
   The algorithm is the following:
   - First, a colouring of the node/clocks of the graph is applied:
     for each input, a new mark is taken, and (transitively) all the successors of the input are marked with this mark.
   - then the nodes with the same associated marks are inserted in the same partition (sub-graph).

   It applies sequentially the following methods:
   - selectNodes()
   - colouring()
   - partitionning()
   - attributesSetting()
   - graphCleaning()
*/
ptlINode InputsTrainClusters::apply()
{
  pK_messenger->globalMessageIdent(trainesmess, AGRAPH->getName());
  selectNodes();
  colouring();
  partitionning();
  attributesSetting();
  return lSubGraph;
}

