/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define mess1 " *** ERROR: Clocks are not all localizated."
#define mess2 " *** ERROR: Nodes are not all localizated."
#define mess3 " *** ERROR: A clock is not an extraction."
#define trace1 "              Done for :"

_PRIVATE_
/** It specifies that the the m-th set references the clock (clk).
    It uses the referencedInMark attribute associated with a list of marks to keep the information.
    It is the places of consomation of the clock and it will be used to fix the place of production.
 */
    boolean RunOnMark::ptinsthAttr_referencedIn_append_mark(ptinsth clk, typeMark m)
{
  ptlMark lm;

  if (!CKM->getAttribute(clk, referencedInMark, (attributeValue *) & lm))
    lm = NILLMARK;
  if (!ptlMark_insert_elem(&lm, m))
    return pk_false;
  CKM->addAttribute(clk, referencedInMark, lm);
  return pk_true;
}

/**
   It specifies that the m-th set does not reference the clock (clk).  
*/
void RunOnMark::ptinsthAttr_referencedIn_removeMark(ptinsth clk, typeMark m)
{
  ptlMark lm;

  if (CKM->getAttribute(clk, referencedInMark, (attributeValue *) & lm))
    if (ptlMark_removeMark(&lm, m))
      CKM->addAttribute(clk, referencedInMark, lm);
}

/**
   @return "true" if a clock (clk) is referenced in the val-th set, "false" otherwise.
*/
boolean RunOnMark::ptinsthAttr_referencedIn_satisfies_marked(ptinsth clk, typeMark val)
{
  ptlMark lm;

  if (!CKM->getAttribute(clk, referencedInMark, (attributeValue *) & lm))
    return pk_false;
  return ptlMark_IsIn(lm, val);
}

/**
   @return the list of marks associated with a clock (clk). It is the sets that reference this clock.
   It can be used to fix the localization of the production of the clock.
*/
ptlMark RunOnMark::ptinsthAttr_referencedIn_get_marks(ptinsth clk)
{
  ptlMark lm;

  if (!CKM->getAttribute(clk, referencedInMark, (attributeValue *) & lm))
    return NILLMARK;
  return lm;
}

/**
   It assigns a list of marks (lm) to a clock (vh). The clock localization can be fixed with this information.
   @return "true" if the list (lm) is not empty, "false" otherwise.
*/
boolean RunOnMark::setClockLocalization(ptinsth vh, ptlMark lm)
{
  typeMark m;

  if (ptlMark_satifies_isEmpty(lm))
    return pk_false;
  while (ptlMark_get_elemNext(&lm, &m))
    ptinsthAttr_append_mark(vh, m);
  return pk_true;
}

/**
   It assigns a list of marks (lm) to a node (aNode) assumed to be a node without location.
   @return "true" if the list (lm) is not empty, "false" otherwise.
   If this list is not empty, the node is removed from the set of "unlocated" nodes (LNodesWithoutPragma).
   This list must be empty at the end of the partitionning.
*/
boolean RunOnMark::setNodeLocalization(ptINode aNode, ptlMark lm)
{
  typeMark m;

  if (ptlMark_satifies_isEmpty(lm))
    return pk_false;
  while (ptlMark_get_elemNext(&lm, &m))
    ptINodeAttr_append_mark(aNode, m);
  ptlINode_remove_elem(&LNodesWithoutPragma, aNode);
  return pk_true;
}

/**
   @return "true" if the place of production of a clock (vh) is fixed, "false" otherwise.
   It calls setClockLocalization() to assign the places of production of the clock (vh).
   If it returns "true", the list (lm) is also associated with the node (aNode).
   \note vh is assumed to be an extraction (when C), and aNode the node of C. 
*/
boolean RunOnMark::setClockAndNodeLocalization(ptinsth vh, ptINode aNode, ptlMark lm)
{
  if (!setClockLocalization(vh, lm))
    return pk_false;
  return setNodeLocalization(aNode, lm);
}

/**
   It associates the list of marks (lm) to a node (aNode). lm is the places of production of
   the node. If the list (lm) is not empty, 
   it is also associated with each clock when C, where C is a signal produced by aNode; this 
   is not applied if the clock is already localized.
*/
boolean RunOnMark::setNodeAndClockLocalization(ptINode aNode, ptlMark lm)
{
  ptlMark lm1;
  ptlinstances lsig;
  ptexpcal sig;
  ptinsth vh;
  boolean b;

  if (!setNodeLocalization(aNode, lm))
    return pk_false;
  lsig = NM->getInodelOutputFlows(aNode);
  b = pk_false;
  while (ptlinstances_get_next(&lsig, &sig))
    if (EM->getExtractionLink(sig, &vh))
      {
	lm1 = ptinsthAttr_get_marks(vh);
	// Attention !!!!
	if (ptlMark_satifies_isEmpty(lm1))
	  setClockLocalization(vh, lm);
	b = pk_true;
      }
  return b;
}

/**
   @return "true" if the place of production of a node of the global list LNodesWithoutPragma is fixed, "false" otherwise. 
   For a node (N) of the list LNodesWithoutPragma, it looks at if all the inputs are located in the same set (producedInSamePartition())
   in this case, such a node is also located to this set (setNodeAndClockLocalization()).
*/
boolean RunOnMark::localizedNodesWithoutPragmas()
{
  ptlINode lNodes;
  boolean traite;
  ptlMark lm;
  ptINode aNode;
  ptlinstances lsig;

  lNodes = LNodesWithoutPragma;
  traite = pk_false;
  while (ptlINode_get_elem(&lNodes, &aNode))
    {
      lsig = NM->getInodelInputFlows(aNode);
      if (producedInSamePartition(lsig, &lm))
	if (setNodeAndClockLocalization(aNode, lm))
	  traite = pk_true;
    }
  return traite;
}

/** @return "true" if a node without a localization (no PRAGMA_RUN_ON pragma) is assigned here, "false" otherwise.
    This method tries to assign a set to a node (X) of the global list (LNodesWithoutPragma).
    Such a choice is done in setNodeLocalization(). 
    It depends on the location of the nodes that produce the inputs of the node X.
*/
boolean RunOnMark::localizedNodesWithoutPragmasChoice()
{
  ptlINode lNodes;
  ptlMark lm;
  ptINode aNode, aNode1, n1, n2;
  ptexpcal sig;
  ptlinstances lsig;
  ptldependence ldep;
  ptdependence aDep;

  lNodes = LNodesWithoutPragma;
  while (ptlINode_get_elem(&lNodes, &aNode))
    {
      lsig = NM->getInodelInputFlows(aNode);
      while (ptlinstances_get_next(&lsig, &sig))
	{
	  aNode1 = EM->getNode(sig);
	  lm = ptINodeAttr_get_marks(aNode1);
	  if (setNodeLocalization(aNode, lm))
	    return pk_true;
	}
      // Added Aug 2014 LBesnard. x := f(y1,...yn) and th yi are inputs (not localized)--> using the references.
      ldep = NM->get_successors(aNode);
      while (ldep->get_next(&aDep, &ldep))
	{
	  aDep->getDependence(&n1, &n2);
	  lm = ptINodeAttr_get_marks(n2);
	  if (setNodeLocalization(aNode, lm))
	    return pk_true;
	}
    }
  return pk_false;
}

/**
   It stores a node (aNode) in the list of unassigned node (LNodesWithoutPragma).
   LNodesWithoutPragma is a meber attribute of the class. A node without PRAGMA_RUN_ON pragma 
   will be localized at the end of the colouring (localizedNodesWithoutPragmas()).
*/
void RunOnMark::storeNodeWithoutPragma(ptINode aNode)
{
  ptlINode_insert_elem(&LNodesWithoutPragma, aNode);
}

/**
   It marks (by Mark) a clock (h). This mark specifies that the clock is referenced in the Mark-th set.
*/
void RunOnMark::colouringCLock(ptinsth h, typeMark Mark)
{
  h = CKM->getRepresentativeFormula(h);
  if (!CKM->satisfiesPredefinedClock(h))
    if (!CKM->satisfiesTickClock(h))
      {
	ptinsthAttr_referencedIn_append_mark(h, Mark);
	if (bDCplusFormat && !CKM->satisfiesSub(h))
	  pK_messenger->printString(mess3);
      }
}

/**
   It marks (by Mark) the clock of the node (aNode)  and the clocks referenced in aNode. 
   This mark specifies that the clock is referenced in the Mark-th set.
   \see RunOnMark::colouringCLock().
*/
void RunOnMark::colouringClocksReferencedIn(ptINode aNode, typeMark Mark)
{
  ptlinsth lh;
  ptinsth h;

  h = NM->get_nodeClock(aNode);	/* horloge du noeud */
  colouringCLock(h, Mark);
  lh = NM->getInodelInputClocks(aNode);	/* horloges referencees dans le noeud */
  while (ptlinsth_get_next(&lh, &h))
    colouringCLock(h, Mark);
}

/**
   @return the index of a string (EnonceValue) in the internal table.
   It is more general than the previous integer value.
   For example : RunOn { } "ILOT1"
   The name (ILOTS) will be also used for fixing of the name of the partition.
*/
typeMark RunOnMark::GetMarkOf(ptrepstr EnonceValue)
{
  ptrepstr str;
  RunOnNameTableType::iterator i;
  integer index;

  str = EnonceValue;
  i = RunOnNameTable.begin();
  index = 1;
  while (i != RunOnNameTable.end())
    {
      if (streq(*i, str))
	return (typeMark) index;
      index++;
      i++;
    }
  RunOnNameTable.push_back(str);
  return (typeMark) index;
}

/**
   It marks a node (aNode) annotated by the pragma (aPragma) with name PRAGMA_RUN_ON (assumed).
   The pragma is of the following form: PRAGMA_RUN_ON "i1" "i2" ... where ij is an integer (ij stands for ij-th partition).
   The number of sub sets is managed the method setMinMaxMark().
   It marks also the clocks referenced in the definition (colouringClocksReferencedIn()).
   \note restriction: the multi-values are accept in the value of the pragma, but there is no code duplication, currently.
*/
void RunOnMark::colouringNodeWithPragma(ptINode aNode, ptpragma aPragma)
{
  ptrepstr EnonceValue;
  typeMark Mark;
  integer i, nbEnonce;
  ptexpcal vsig;
  ptlinstances lsig, louts;
  ntrees tEnonce;

  tEnonce = getPragmaStatement(aPragma);
  if (IS_NILTREE(tEnonce))
    return;
  if (NM->isNodeCall(aNode))
    louts = NM->getInodelOutputFlows(aNode);
  else
    louts = NILLINSTANCES;
  nbEnonce = ntreearity(tEnonce);
  for (i = 1; i <= nbEnonce; i++)
    {
      EnonceValue = nstringof(nchild(i, tEnonce));
      Mark = GetMarkOf(EnonceValue);
      ptINodeAttr_append_mark(aNode, Mark);
      colouringClocksReferencedIn(aNode, Mark);
      lsig = louts;
      while (ptlinstances_get_next(&lsig, &vsig))
	ptINodeAttr_append_mark(EM->getNode(vsig), Mark);
      setMinMaxMark(Mark);
      colouringComposedNode(aNode, Mark);
    }
}

/**
  It marks a node (aNode) by applying colouringNodeWithPragma(ptINode) if it is annotated 
  by the pragma PRAGMA_RUN_ON, otherwise storeNodeWithoutPragma().
*/
_VIRTUAL_ void RunOnMark::colouringNode(ptINode aNode)
{
  ptlpragmas LPrag;
  ptpragma aPrag;

  LPrag = NM->getPragmas(aNode);
  if (getPragmaKeyword(LPrag, PRAGMA_RUN_ON, &aPrag))
    colouringNodeWithPragma(aNode, aPrag);
  else
    storeNodeWithoutPragma(aNode);	/* booleen representant une horloge ? */
}

/** @returns pk_true if the signals of the list ?lsig are produced on the same processor.
   Such nodes are colored with the marks !lm. 
*/
boolean RunOnMark::producedInSamePartition(ptlinstances lsig, ptlMark * lm)
{
  ptexpcal sig;
  ptlMark lmark, lmark1;
  ptINode aNode;

  if (lsig == NILLINSTANCES)
    return pk_false;

  if (ptlinstances_get_next(&lsig, &sig))
    {
      aNode = EM->getNode(sig);
      lmark = ptINodeAttr_get_marks(aNode);
      (*lm) = lmark;
    };
  while (ptlinstances_get_next(&lsig, &sig))
    {
      aNode = EM->getNode(sig);
      lmark1 = ptINodeAttr_get_marks(aNode);
      if (!ptlMark_satisfies_equal(lmark, lmark1))
	return pk_false;
    }
  return pk_true;
}

/** 
    @return "true" it the production localization of a clock (vh) is fixed, "false" otherwise.
    To achieve this, if the localization is not yet fixed, 
    The clock is of the form when C.
    - if C is localized, the clock is located to the same places
    - if C is referenced only on a set, the clock and the node of C are located to the same place
    - if C is not referenced and if ^C is localized, the clock is located to the same places
    - otherwise if C is referenced on several sets, it choices a localization

    The localization is always done by the method setClockAndNodeLocalization().
*/
boolean RunOnMark::localizatingClockLocality(ptinsth vh)
{
  ptexpcal vsig;
  ptlinstances lsig;
  ptlMark lm;
  ptINode aNode;
  typeMark vmark;
  integer sizelm;
  ptinsth vh1, vhFather;

  /* Already assigned */
  lm = ptinsthAttr_get_marks(vh);
  if (!ptlMark_satifies_isEmpty(lm))
    return pk_true;
  /* h = when C et C Already assigned */
  vsig = CKM->getBooleanSignal(vh);
  aNode = EM->getNode(vsig);
  lm = ptINodeAttr_get_marks(aNode);
  if (setClockAndNodeLocalization(vh, aNode, lm))
    return pk_true;
  /* h = when C et C referencee sur un seul processeur k => localisation sur k. */
  lm = ptinsthAttr_referencedIn_get_marks(vh);
  sizelm = ptlMark_get_size(lm);
  if (sizelm == 1)
    {
      if (setClockAndNodeLocalization(vh, aNode, lm))
	return pk_true;
    }

  if (NM->is_readNode(aNode))	/* Input without Pragma. */
    {
      vsig = NM->get_io_arg(aNode);
      vh1 = EM->getClock(vsig);
      lm = ptinsthAttr_get_marks(vh1);
      if (setClockAndNodeLocalization(vh, aNode, lm))
	return pk_true;
    }
  /* h = when C et C := E et qqsoit x de E, x localise sur le meme processeur.     */
  lsig = NM->getInodelInputFlows(aNode);
  if (producedInSamePartition(lsig, &lm))
    {
      if (setClockAndNodeLocalization(vh, aNode, lm))
	return pk_true;
    }
  /* h = when C et C referencee sur un seul processeur k => localisation sur k. */
  lm = ptinsthAttr_referencedIn_get_marks(vh);
  sizelm = ptlMark_get_size(lm);
  /* Pas de references, mais l'horloge mere est localisee => meme localisation que la mere. */
  if (sizelm == 0)
    {
      vhFather = CKM->getFather(vh);
      lm = ptinsthAttr_get_marks(vhFather);
      if (setClockAndNodeLocalization(vh, aNode, lm))
	return pk_true;
    }
  else				/* choice  */
    {
      ptlMark_get_elemNext(&lm, &vmark);
      lm = NILLMARK;
      ptlMark_insert_elem(&lm, vmark);
      if (setClockAndNodeLocalization(vh, aNode, lm))
	return pk_true;
    }

  return pk_false;
}

/**
   @return "true" as soons as the place of production of a clock (aClock) of lh is fixed, "false" otherwise (no such clok in lh).
   For a clock (such that vh = [c], c:=E), if one of the signals referenced in E is localized, the clock is localized to the same set.
*/
boolean RunOnMark::localizatingClockChoice(ptlinsth lh, ptinsth * aClock)
{
  ptexpcal vsig;
  ptlinstances lsig;
  ptlMark lm;
  ptINode aNode, CondNode;
  ptinsth vh;

  while (ptlinsth_get_next(&lh, &vh))
    {
      vsig = CKM->getBooleanSignal(vh);
      CondNode = EM->getNode(vsig);
      lsig = NM->getInodelInputFlows(CondNode);
      /* vh = [c], c:=E */
      /* nb procecceurs referencant ?vh > 1, et marques differentes pour les signaux ?x de E */
      /* Choix : marque du premier signal */
      while (ptlinstances_get_next(&lsig, &vsig))
	{
	  aNode = EM->getNode(vsig);
	  lm = ptINodeAttr_get_marks(aNode);
	  if (setClockAndNodeLocalization(vh, CondNode, lm))
	    {
	      clockTracing(trace1, vh);
	      (*aClock) = vh;
	      return pk_true;
	    }
	}
    }
  return pk_false;
}

/**
   For each clock of the list of clocks (lh), it tries to fix (localizatingClockLocality()) 
   the place of production of this clock. If the operation succeeds, the clock is removed from the list.
*/
boolean RunOnMark::localityAlgorithm(ptlinsth * lh)
{
  ptlinsth lhSave;
  ptinsth vh;
  boolean bcontinue, apply, b;

  apply = pk_false;
  do
    {
      bcontinue = pk_false;
      lhSave = *lh;
      while (ptlinsth_get_next(lh, &vh))
	if (localizatingClockLocality(vh))
	  {
	    bcontinue = pk_true;
	    clockTracing(trace1, vh);
	    lhSave = ptlinsth_remove_elem(lhSave, vh, &b);
	    apply = pk_true;
	  }
      (*lh) = lhSave;
    }
  while (bcontinue);
  return apply;
}

/** It tries to fix the place of production of each  boolean clocks.
    It calls:
    - localityAlgorithm()
    - localizedNodesWithoutPragmas()
    - localizatingClockChoice()
    \todo to be cleaned.
*/
boolean RunOnMark::colouringBooleanCLocks()
{
  ptlinsth lh, lh1;
  ptinsth vh;
  boolean bcontinue, Done, AllNodesAssigned, b;
  ptexpcal vtick;
  ptlINode ln;
  ptINode anode;
  // removing the Tick Signal.
  vh = CKM->getTickClock();
  if (!CKM->satisfiesFictitiousRoot(vh))
    if (CKM->isClockRepresentedBySignal(vh, &vtick))
      ptlINode_remove_elem(&LNodesWithoutPragma, EM->getNode(vtick));

  // Filtering the selecting clocks. Only the positive extractions are kept.
  lh1 = selectedClockList;
  lh = NILLINSTH;
  while (ptlinsth_get_next(&lh1, &vh))
    if (CKM->getOperator(vh) == sub)
      lh = ptlinsth_sorted_insert_elem_bool(lh, vh, &Done);

  // fixing the place of the production of each boolean clocks (lh)
  bcontinue = pk_true;
  while (bcontinue)
    {
      while (localityAlgorithm(&lh)) ;
      if (!localizedNodesWithoutPragmas())
	if (localizatingClockChoice(lh, &vh))
	  lh = ptlinsth_remove_elem(lh, vh, &b);
	else
	  bcontinue = localizedNodesWithoutPragmasChoice();
    }
  
  // Verifying
  // AllNodesAssigned =(LNodesWithoutPragma == NILLINODE) .Sept 2016 LBesnard, replaced by 
  AllNodesAssigned = true; 
  ln = LNodesWithoutPragma;
  while(AllNodesAssigned && ptlINode_get_elem(&ln, &anode))
    {
     AllNodesAssigned = NM->is_readNode(anode);
    }
  // ----
  if (!AllNodesAssigned)
    {
      pK_messenger->printString(mess2);
      AGRAPH->addAttribute(RUNON_NODES_NOT_LOCALIZED, (attributeValue) LNodesWithoutPragma);
    }

  Done = ptlinsth_get_size(lh) == 0;
  if (!Done && AllNodesAssigned)
    pK_messenger->printString(mess1);

  return Done && AllNodesAssigned;
}

/**
   For a partial definition node (aNode), it marks by  mark all the partial assignments.
*/
void RunOnMark::colouringComposedNode(ptINode aNode, typeMark mark)
{
  ptlINode ln;
  ptINode n;
  ptlinsth lh;

  if (NM->is_setOfPartialDefinitions(aNode))
    {
      ln = NM->get_setOfPartialDefinitions_arg(aNode);
      // nodes of ln are set in same partition than aNode!!!
      while (ptlINode_get_elem(&ln, &n))
	{
	  ptINodeAttr_append_mark(n, mark);
	  colouringClocksReferencedIn(n, mark);
	}
    }
  else if (NM->get_array_bodyDefinitions(aNode, &ln, &lh))
    {
      while (ptlINode_get_elem(&ln, &n))
	{
	  ptINodeAttr_append_mark(n, mark);
	  // colouringClocksReferencedIn(n, mark);
	}
    }
}

/**
   Restriction: (bDC+) graph.
   
   @return the selected clocks of the (bDC+) graph: it is the list of the 
   positive extractions (when C) of the graph.
*/
_VIRTUAL_ ptlinsth RunOnMark::selectingClocks()
{
  ptlinsth lh, lh1;
  ptinsth vh;
  boolean bdummy;

  /* Selecting the extractions (boolean clocks) */
  lh1 = CKM->getlhplaced();
  lh = NILLINSTH;
  while (ptlinsth_get_next(&lh1, &vh))
    {
      vh = CKM->getRepresentativeFormula(vh);
      if (!ptlinsth_isin_elem(unusedClocks, vh))
	if (CKM->satisfiesSub(vh))	// sub, nosub
	  lh = ptlinsth_sorted_insert_elem_bool(lh, vh, &bdummy);
    }
  return lh;
}

/**
   It marks the nodes of a list of nodes (lNodes) by applying colouringNodeWithPragma(ptINode) for each node
   annotated by the pragma PRAGMA_RUN_ON, otherwise storeNodeWithoutPragma()
*/
void RunOnMark::colouringNodes(ptlINode lNodes)
{
  ptINode aNode;

  while (ptlINode_get_elem(&lNodes, &aNode))
    if (!ptlINode_isIn(unusedNodes, aNode))
      colouringNode(aNode);
}

_PROTECTED_ ptlinsth RunOnMark::getSelectedClocks()
{
  return selectedClockList;
}

/**
   It marks the nodes of the graph by applying colouringNode(ptlINode) for each list of nodes.
*/
_VIRTUAL_ void RunOnMark::colouringNodes()
{
  ptlINode lNodes;
  // getInputs() Added Sept 2016  (LBesnard) when assert(x) and x is an input
  if (NM->getInputs(&lNodes))  
    colouringNodes(lNodes);    
  if (NM->getEquations(&lNodes))
    colouringNodes(lNodes);
  if (NM->getDelays(&lNodes))
    colouringNodes(lNodes);
  if (NM->getProcCalls(&lNodes))
    colouringNodes(lNodes);
  if (NM->getNodeCalls(&lNodes))
    colouringNodes(lNodes);
  if (NM->getGroupsOfPartialEquations(&lNodes))
    colouringNodes(lNodes);
  if (NM->getIterates(&lNodes))
    colouringNodes(lNodes);
}

/**
   Restriction: bDC+ graph.
   It marks the clocks of the graph, calling colouringBooleanCLocks() for each clock of the system.
   An error is detected if there is a node/clock without production localization.
*/
_VIRTUAL_ void RunOnMark::colouringCLocks()
{
  error = !colouringBooleanCLocks();
}
