/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/* Constructeur.
   The functionality is applied to a graph (g).
*/
 typeSelection::typeSelection(ptDCG g):separate_selected_others(g)
{
  lSelectedNodes = NILLINODE;
  initSelectedTypes();
}

/**
   Destructor.
*/
typeSelection::~typeSelection()
{
}

/**
   It separates the nodes of the graph in two sub-sets:
   - those which produce selected type signals
   - the others nodes
   
   @return the two sub-sets (selected type nodes, other nodes).
   \see typeSelection::colouring(), typeSelection::partitionning()
*/
ptlINode typeSelection::apply()
{
  ptlINode ln;

  selectTypes();
  colouring();
  ln = partitionning();
  finalyzeInterface(ln);
  return ln;
}

/**
   This method defines the types that must be selected in the "colouring" method.
   It is the first operation of this class of partitionning.
*/
_VIRTUAL_ void typeSelection::selectTypes()
{
}

void typeSelection::addSelected(ptlINode ln)
{
  lSelectedNodes = ln;
}


ptlINode typeSelection::getSelectedNodes()
{
  ptlINode ln;
  selectTypes();
  colouring();
  ln = separate_selected_others::getSelectedNodes();
  graphCleaning();  
  return ln;
}
