/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __DCG2SignalTree__
#define __DCG2SignalTree__

#include <set>

typedef class DCG2SignalTree * ptDCG2SignalTree;
typedef std::set<ptdependence> DependenceSet;

class DCG2SignalTree
{
#include "DCG2SignalTree_m.h"
#include "DCG2SignalTreeAbstraction_i.h"
#include "DCG2SignalTreeDependences_i.h"
#include "DCG2SignalTreeSubGraph_i.h"
#include "DCG2SignalTreeDecl_i.h"
#include "DCG2SignalTreeNode_i.h"
#include "DCG2SignalTreeType_i.h"
#include "DCG2SignalTreeClock_i.h"
#include "DCG2SignalTreeExpr_i.h"
#include "DCG2SignalTreeHierarchy_i.h"
#include "DCG2SignalTreeExclusiveClocks_i.h"
#include "DCG2SignalTreeAutomaton_i.h"
#include "DCG2SignalTree_g.h"

protected:
  ptDCG AGRAPH;    ///< the functionality is applied to this graph.
  ptExprManager EM;   ///< operand graph info: expression manager.
  ptClockManager CKM; ///< operand graph info: clock manager
  ptNodeManager NM;   ///< operand graph info: node manager.
  ptNameManager NameM; ///< operand graph info: the name manager.
  ptinsth vRootClock; ///< the root clock of the clock hierarchy.
  boolean isEndochronous; ///< graph info: the status (endo/poly) of the graph.
  GraphLevelType graph_type; ///< graph info: Graph Level of the graph.
  ptlINode unNodes;         ///< graph info: Unused Nodes
  ptlinsth unClocks;        ///< graph info: Unused Clocks
  ptlinstances inputs,      ///< graph info: inputs of the graph.
    outputs;                ///< graph info: outputs of the graph.
  ptlpragmas lpragmas; ///< graph info: list of the pragmas.
  ptINode CurrentBlock;  ///< block context: current array/subgraph node.
  ntrees lassertions,  ///< Signal abstract tree representation of the assertions (expression of processes)
    lsigAssertions, ///< Signal abstract tree representation of the assertions on signals
    lequnotprov, ///< Signal abstract tree representation of the not proved constraints
    lMemOptimizedAttribute, ///< Signal abstract tree representation of the memory optimisations (not implemented)
    lDefForAssertions, ///< list of the clocks used in assertions and not defined as a node of the hierarchy.
    ClockParameterTree;///< list of the clock parameters of the clock hierarchy.
    ptlDCG lIGraph; ///< list of the graphs to be printed (referenced in the node call nodes).

  ptrepstr traSuffix, ///< trace suffix, composed with the name of the graph and the graph level to produce the name of the Signal model.
    abstractSuffix, ///< abstract suffix  composed with the name of the graph to produce the name of the abtraction of a Signal model.
    ecSuffix;///< abstract suffix  composed with the name of the graph to produce the exclusive clocks of a Signal model.

  // For added identifiers and renaming i/o models.
  boolean mkNewIdentTag;  ///< context information: "false" by default, set to "true" to make the declaration of an external model from a call.
  ptident identPrefix;    ///< used when mkNewIdentTag to create identifiers using this prefix composed with cpt_names counter.
  ptident identPrefixKernel; ///< Specific prefix used for the identifiers created in "kernel mode".
  integer cpt_names;      ///< used when mkNewIdentTag, counter of created names with identPrefix as prefix.

  boolean flatMode;        ///< Flat mode: the external Signal view of the hierarchy is flatened (default false).
  boolean KernelExprMode;  ///< Kernel Expresson mode: the expresssions are decomposed when it is true (default false).
    boolean SpecMode,      ///< The specification is added to the tree representing the graph when it is true (default true)
      bDependenceWithClock; ///< Set to true when the clocks of the dependences are exported.
  integer nb_max_list_elem; ///< size max of a tree-list (there is a limit in the Signal syntactic analyzer).
  ntrees G_TreeGraph;    ///< Flat mode: global Signal abstract tree representation of the program 
  ptlinstances G_LocSig;   ///< Flat mode: it constains all the local signals of the program (union of the locals of each node of the hierarchy)
  ptlinsth G_Clocks;      ///< Flat mode: it constains all the local clocks of the program (union of the locals of each node of the hierarchy)

    ptlinsth G_ClockListToDeclareProjection;
    typeMark VisitedMark, SignalTreeAttribute, subgraphAlreadyInserted, DontDeclareInAbstaction; ///< local attributes, cleaned at the end of the functionality.

  boolean genInterface; ///< When "true" the I/O of the sub-models are generated. It is "false" by default.
  boolean DEBUG_MODE;     ///< debugging tag
  boolean innerModel; 
  boolean TraceabilityMode; ///< Tracability tag
  boolean  noDoubleSynchroSameId; ///< misc.
  DependenceSet  printedDependences;
  boolean G_TickIsReferenced, notInitialized;
  boolean RunOnSetOnSubgraphs, simplyfyExpr;
  ptlinstances labelsToDeclare; ///< list of the labels to be declare at the master clock.
  typeMark genScheduledGraphMark;
  boolean DontGenerateTheLastClockContribution;
  ntrees vtreeClockScheduledGraph;
  boolean actionsAutomaton;
};

extern ptDCG2SignalTree pK_Graph2SignalTreeBuilder; ///< Predefined object.

#endif
