/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/** @return dependences + the Signal Abstract tree representation of the dependences from a sub-graph (aSubGraph)
    to other sub-graphs.
    \warning if aSubGraph is a "delay cluster" sub-graph (particular cluster annotated with the PRAGMA_DELAYCLUSTER pragma), the dependences are reversed.
*/
    ntrees DCG2SignalTree::mkTreeAbstractDependencesSubGraph(ptINode aSubGraph, ntrees dependences)
{
  ptldependence ldep;
  ptdependence aDep;
  ptINode vFrom, vTo;
  ptinsth h;
  ntrees tFrom, tTo, tClock, taDep;
  boolean bSwitchDep;

  ldep = NM->get_successors(aSubGraph);
  bSwitchDep = existPredefinedPragma(NM->getPragmas(aSubGraph), PRAGMA_DELAYCLUSTER);
  // bSwitchDep a vrai : inversion des dependances (lignees des retards)
  while (ldep->get_next(&aDep, &ldep))
    {
      h = aDep->getDependence(&vFrom, &vTo);
      if (NM->get_oper(vTo) == opsubgraph)
	{
	  tClock = STree_getNilTree();	/* tClock = mkhonametree(h); (A REVOIR) */
	  tFrom = getLabelTree(vFrom);
	  tTo = getLabelTree(vTo);
	  if (bSwitchDep)
	    taDep = STree_mkDependenceTree(tTo, tFrom, tClock);
	  else
	    taDep = STree_mkDependenceTree(tFrom, tTo, tClock);
	  STree_addExprToComposition_Post(&dependences, taDep);
	}
    }
  if (bSwitchDep)
    return dependences;
  // ajout des dependences entrees-> lignees
  ldep = NM->get_predecessors(aSubGraph);
  tClock = STree_getNilTree();	// REVOIR
  while (ldep->get_next(&aDep, &ldep))
    {
      h = aDep->getDependence(&vFrom, &vTo);
      if (NM->is_readNode(vFrom))
	{
	  tFrom = mkNameNode(vFrom);
	  taDep = STree_mkDependenceTree(tFrom, getLabelTree(vTo), tClock);
	  STree_addExprToComposition_Post(&dependences, taDep);
	}
    }
  return dependences;
}

boolean DCG2SignalTree::isNextDependenceCompatibleWith(pticlosure iclos, ptinsth hexp, ptexpcal * expFrom)
{
  ptinsth hdep;

  if (iclos == NILICLOSURE)
    return pk_false;
  hdep = CKM->getRepresentativeFormula(pticlosure_get_hdep(iclos));
  if (CKM->satisfiesNullClock(hdep))
    return pk_true;
  if (hdep != hexp)
    return pk_false;
  *expFrom = pticlosure_get_from(iclos);

  return pk_true;
}

/**
  It adds to the Signal Abstract tree representation of the closures (composition of dependences) the 
  closure informations (lclos).
  For each element <em> (Y,H) </em> of lclos, it adds to transclosure 
  the dependence <em> (Y ---> vsig) \b when H </em>.

  It assumes that when expg is defined hg is undefined,and when expg is undefined hg is defined,.
*/
ntrees DCG2SignalTree::solveclosure(ntrees transclosure, ptlclosure lclos, ptexpcal vsig)
{
  ptinsth hexp, hdep;
  pticlosure iclos;
  ntrees vdeptree, htree, vte1, vte2;
  ptexpcal expFrom;

  iclos = ptlclosure_get_iclos(lclos);

  hexp = EM->getClock(vsig);
  while (iclos != NILICLOSURE)
    {
      SET_NILTREE(vdeptree);
      expFrom = pticlosure_get_from(iclos);
      if (vsig != expFrom)
	{
	  hdep = CKM->getRepresentativeFormula(pticlosure_get_hdep(iclos));
	  if (!CKM->satisfiesNullClock(hdep))
	    {
	      vte1 = ptexpcal2ntrees(expFrom);
	      if (hdep == hexp)
		SET_NILTREE(htree);
	      else
		{
		  htree = mkhonametree(hdep, pk_true);
		  if (noper(htree) == opwhen1)
		    htree = nchild(1, htree);
		}
	      vte2 = ptexpcal2ntrees(vsig);
	      vdeptree = STree_mkDependenceTree(vte1, vte2, htree);
	      while (isNextDependenceCompatibleWith(pticlosure_get_next(iclos), hdep, &expFrom))
		{
		  if (vsig != expFrom)
		    STree_addDependencePre(vdeptree, ptexpcal2ntrees(expFrom));
		  iclos = pticlosure_get_next(iclos);
		}
	    }
	}
      if (ISNOT_NILTREE(vdeptree))
	transclosure = STree_addDependenceToDependences(vdeptree, transclosure);
      iclos = pticlosure_get_next(iclos);
    }
  return transclosure;
}

/**
  It adds to the Signal Abstract tree representation of the closures (composition of dependences) the 
  closure informations (lclos).
  For each element <em> (Y,H) </em> of lclos, it adds to transclosure 
  the dependence <em> (Y ---> hg) \b when H  </em>

  It assumes that when expg is defined hg is undefined,and when expg is undefined hg is defined,.
*/
ntrees DCG2SignalTree::solveclosure(ntrees transclosure, ptlclosure lclos, ptinsth hg)
{
  ptinsth hexp, hdep;
  pticlosure iclos;
  ntrees vdeptree, htree, vte1, vte2;
  ptexpcal expFrom;

  /* remarque :   expg a nil => hg != nil et reciproquement. */
  iclos = ptlclosure_get_iclos(lclos);
  hexp = NILINSTH;

  while (iclos != NILICLOSURE)
    {
      SET_NILTREE(vdeptree);
      expFrom = pticlosure_get_from(iclos);
      hdep = CKM->getRepresentativeFormula(pticlosure_get_hdep(iclos));
      if (!CKM->satisfiesNullClock(hdep))
	{
	  vte1 = ptexpcal2ntrees(expFrom);
	  if (hdep == hexp)
	    SET_NILTREE(htree);
	  else
	    htree = mkhonametree(hdep);
	  if (hg != NILINSTH)
	    vte2 = mkhonametree(CKM->getRepresentativeFormula(hg));
	  vdeptree = STree_mkDependenceTree(vte1, vte2, htree);
	}
      if (ISNOT_NILTREE(vdeptree))
	STree_addExprToComposition_Post(&transclosure, vdeptree);
      iclos = pticlosure_get_next(iclos);
    }
  return transclosure;
}

/**
   If the operand (aDep) is a i/o dependence, its Signal representation is added to the operand tree (transclosure).
*/
void DCG2SignalTree::mkIOClosureTree(ptdependence aDep, ntrees * transclosure)
{
  ptinsth hdep;
  ptINode vFrom, vTo;

  if ( NM-> isSelectedDependenceForAbstraction(aDep, unNodes))
    {
      hdep = CKM->getRepresentativeFormula(aDep->getDependence(&vFrom, &vTo));
      STree_addExprToComposition_Post(transclosure, mkTreeDependence(vFrom, vTo, hdep));
    }
}

/**
   @return the Signal Abstract tree representation of the transitive closure of the current graph
   reduced to the input/output signals of the graph.
   For each selected node, it calls solveclosure().
*/
ntrees DCG2SignalTree::mkIOClosureTree()
{
  ntrees transclosure;
  ptdependence aDep;
  ptldependence lDep;

  SET_NILTREE(transclosure);
  lDep = AGRAPH->getExplicitDependences();
  while (lDep->get_next(&aDep, &lDep))
    mkIOClosureTree(aDep, &transclosure);

  lDep = AGRAPH->getReinforcedDependences();
  while (lDep->get_next(&aDep, &lDep))
    mkIOClosureTree(aDep, &transclosure);

  /* lDep = AGRAPH->getDependences(); while (lDep->get_next(&aDep, &lDep)) mkIOClosureTree(aDep, &transclosure ); */
  return transclosure;
}

/**
   @return "true" if a clock (vh) is the clock of an input/output signal, "false" oterhwise.
*/
boolean DCG2SignalTree::isIOClock(ptinsth vh)
{
  ptexpcal aSig;

  if (!CKM->isClockUnifiedWithSignal(vh, &aSig))
    return pk_false;
  return isIOSignal(aSig);
  // removed 12/09/2013 LBesnard.
  //  if (ptlinstances_isin_exp(inputs, aSig)) return pk_true; return ptlinstances_isin_exp(outputs, aSig);
}

/**
   @return the Signal Abstract tree representation of a list of the identifiers 
   of a list of clocks (lhorl) excluding the clocks of input/output signal.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkNamesClocksWithoutIO(ptlinsth lhorl)
{
  ntrees vtl;
  ptinsth vh;
  boolean first;
  ptexpcal vsig;
  boolean b0, b1, b2;

  first = pk_true;
  vtl = STree_getNilTree();
  while (ptlinsth_get_next(&lhorl, &vh))
    if (!CKM->satisfiesMarked(vh, DontDeclareInAbstaction))
      {
	b0 = CKM->isClockRepresentedBySignal(vh, &vsig);
	b1 = !b0 && !isIOClock(vh);
	b2 = b0 && !EM->IsInputFlow(vsig) && !EM->IsOutputFlow(vsig);
	if (b1 || b2)
	  {
	    if (first)
	      {
		first = pk_false;
		vtl = LL_make_lident_emptyList();
	      }
	    vtl = ntreeList_insertFirst(mkhonametree1(vh), vtl);
	  }
      }

  return vtl;
}

/**
   @return the Signal Abstract tree representation of a declaration of 
   a list of clocks (lhorl) excluding the clocks of input/output signals.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkDeclNotExternalClocks(ptlinsth lhorl)
{
  return mkClockDeclarationFinalize(mkNamesClocksWithoutIO(lhorl));
}

/** 
    @return the Signal Abstract tree representation of the projected system of the graph.
    It takes h as root if withTick (used for clusters).
 */
ntrees DCG2SignalTree::mkProjectionTree(ptinsth h, boolean withTick, ntrees * vtDeclSig)
{
  ptlinsth lh2, lh;
  ptlinstances ladded, lsig;
  ntrees vtDeclSig1, vt, vtdecl;

  lh = CKM->getProjectedSystem(&lh2, &lsig);
  ladded = lsig;
  lsig = ptlinstances_sortedfusion(AGRAPH->getInputs(), lsig);
  lsig = ptlinstances_sortedfusion(AGRAPH->getOutputs(), lsig);

  vt = buildClockTable(lh, lh2, lsig, &vtDeclSig1);
  SET_NILTREE(vtdecl);
  if (isEventClockedFormat(graph_type))
    {
      vtdecl = LL_make_ldecl_list(mkDeclNotExternalClocks(G_ClockListToDeclareProjection));
      vtdecl = addSignalsDeclaration(vtdecl, lsig, pk_true, pk_false, pk_false);
      if (!nisniltree(vtDeclSig1))
	vtdecl = ntreeList_appendLast(vtdecl, vtDeclSig1);
    }
  else if (ISNOT_NILTREE(vtDeclSig1))
    vtdecl = LL_make_ldecl_list(vtDeclSig1);

  *vtDeclSig = vtdecl;
  return vt;
}

/**
   @return the Signal Abstract tree representation of the 
   declaration of a list of signals represented as an combed expression (lInsGreyBox).
   The initial value is added to the declaration is binit.
*/
ntrees DCG2SignalTree::mkTreeDeclarations(ptexpcal lInsGreyBox, boolean binit)
{
  ptexpcal asig;
  ntrees linTree;

  if (EM->isNILEXPCAL(lInsGreyBox))
    return NILTREE;
  linTree = LL_make_ldecs_emptyList();
  while (EM->getNaryArgs(&lInsGreyBox, &asig))
    linTree = ntreeList_appendLast(linTree, mkdecsig(asig, binit));
  return linTree;
}

/**
  It adds to a Signal Abstract tree representation of a model (process)
  the pragma PRAGMA_BLACK_BOX "ident"  
*/
void DCG2SignalTree::addPragmaAbstractionBlackBox(ntrees process, ptident PragmaStatement)
{
  ntrees vtPragmaStatement, aPragmaTree;
  ptident id;

  vtPragmaStatement = repstr2PragmaStatement(ident2repstr(PragmaStatement));
  id = repstr2ident(getPredefinedPragma(PRAGMA_BLACK_BOX));
  aPragmaTree = LL_make_prag_tree(id, STree_mkAtomNilTree(), vtPragmaStatement);
  STree_addPragma(process, aPragmaTree);
}

/**
  It adds to a Signal Abstract tree representation of a model (process)
  the pragma PRAGMA_GREY_BOX "ident"  
*/
void DCG2SignalTree::addPragmaAbstractionGreyBox(ntrees process, ptident idGraph)
{
  ntrees vtPragmaStatement, aPragmaTree;
  ptident id;

  vtPragmaStatement = repstr2PragmaStatement(ident2repstr(idGraph));
  id = repstr2ident(getPredefinedPragma(PRAGMA_GREY_BOX));
  aPragmaTree = LL_make_prag_tree(id, STree_mkAtomNilTree(), vtPragmaStatement);
  STree_addPragma(process, aPragmaTree);
}

/**
   @return the Signal Abstract tree representation of a sub-graph (aSubGraph).
   It is a skeleton of a model with 
   - the name of the sub-graph as name of the process
   - the inputs/outputs of the sub-graph as inputs/outpust of the model
   - the pragmas of the sub-graph 
   - when isolated, the clock table is added, otherwise the model is considered as an external model implemented in "vlang" language.
*/
ntrees DCG2SignalTree::mkAbstractTreeGreyBoxSubGraphs(ptINode aSubGraph, boolean isolated, ntrees vlang)
{
  ntrees process;
  ptexpcal lins, louts;
  ntrees inputsTree, outputsTree;

  if (isolated)
    process = STree_mkSkeletonModel(mknameSubGraph(aSubGraph), opact);
  else
    process = STree_mkSkeletonExternProcess(mknameSubGraph(aSubGraph), vlang, opact);
  // interface
  NM->get_inOutGraphRef(aSubGraph, &lins, &louts);
  inputsTree = mkTreeDeclarations(lins, pk_false);
  outputsTree = mkTreeDeclarations(louts, pk_true);
  STree_replaceInterface(process, STree_mkInterface(NILTREE, inputsTree, outputsTree, NILTREE));
  addPragmas(NM->getPragmas(aSubGraph), process);
  if (isolated)
    addSubGraphClocksTable(aSubGraph, process);
  return process;
}

/** @return nil tree currently.
    Not implemented. Explicit dependences.
*/
ntrees DCG2SignalTree::mkTreeExplicitDependencesAbstraction()
{
  ntrees tdependences;

  tdependences = STree_getNilTree();
  /*  
     ptldependence lexplicitDep; ptdependence aDep; ptINode vFrom, vTo; ptinsth h;
     ntrees tFrom, tTo, tClock, taDep;
     ptlinstances inputs, outputs, l1, l2, ls;
     inputs = AGRAPH->getInputs();
     outputs = AGRAPH->getOutputs();

     lexplicitDep = AGRAPH->getExplicitDependences();
     while (lexplicitDep->get_next( &aDep, &lexplicitDep))
     {
     h = aDep->getDependence(&vFrom, &vTo);
     lsortFrom = NM->getInodelOutputFlows(vFrom);
     lsortTo = NM->getInodelOutputFlows(vTo);

     l1 = ptlinstances_intersection(inputs,lsortFrom );
     l2 = ptlinstances_intersection(inputs,lsortTo );
     tdependences = mkTreeDependencesSignals(l1, l2, tdependences);

     }
   */
  return tdependences;
}

/** 
    @return the Signal Abstract tree representation of the synchronisation of 
    the label associated with a sub-graph (aSubGraph) to the clock of this sub-graph 
*/
ntrees DCG2SignalTree::mkClocklabelTree(ptINode aSubGraph)
{
  ntrees vLabel;
  ptinsth clk;

  vLabel = getLabelTree(aSubGraph);
  clk = NM->get_nodeClock(aSubGraph);
  return STree_addSynchro(STree_mkSynchroIni(vLabel), mkWhenClock(clk));
}

/**
   It adds the Signal Abstract tree representation of the GreyBox abstraction of a list of sub-graphs (subGraphs)
   to a model (process). The Implementation Language (vlang) is used to define the expression of definition of the
   Signal Abstract tree representation of a sub-graph.

   For each element of subGraphs, 
   - the model associated with the sub-graph is produced (mkAbstractTreeGreyBoxSubGraphs()),
   - the model is annotated with the "Black Box" pragma (addPragmaAbstractionBlackBox()),
   - a reference to the sub-graph is inserted in the (upper) model (process) mkReferenceToSubGraphDefinition(),
   - the dependences from the sub-graph to other sub-graphs are built (mkTreeAbstractDependencesSubGraph())

   It finalizes the abstraction by the inserting (in process)
   - the inter sub-graph dependences
   - the declaration of the label associated with each sub-graph.
*/
void DCG2SignalTree::mkAbstractTreeGreyBoxSubGraphs(ntrees process, ptlINode subGraphs, ntrees vlang)
{
  ntrees SubProcess, dependences, llabels;
  ptINode aSubGraph;

  /* Ne pas oublier de specifier l'horloge des lignees */
  /* Les dependences entre lignees (a prendre au niveau du scheduler) */
  dependences = STree_getNilTree();
  llabels = STree_getNilTree();
  while (ptlINode_get_elem(&subGraphs, &aSubGraph))
    {
      SubProcess = mkAbstractTreeGreyBoxSubGraphs(aSubGraph, pk_false, vlang);
      addPragmaAbstractionBlackBox(SubProcess, NILIDENT);
      if (IS_NILTREE(llabels))
	llabels = LL_make_lident_emptyList();
      llabels = ntreeList_appendLast(llabels, getLabelTree(aSubGraph));
      STree_addProcessDeclaration(process, SubProcess);
      STree_addExpressionOfProcessToaModel(process, mkReferenceToSubGraphDefinition(aSubGraph, pk_true));
      //      STree_addExpressionOfProcessToaModel(process, mkClocklabelTree(aSubGraph));
      dependences = mkTreeAbstractDependencesSubGraph(aSubGraph, dependences);
    }
  STree_addExpressionOfProcessToaModel(process, dependences);
  dependences = mkTreeExplicitDependencesAbstraction();
  STree_addExpressionOfProcessToaModel(process, dependences);
  STree_addLabelsInModel(llabels, process);
}

/**
   It declares in the model (process) the boolean representation of the tick of the hierarchy.
*/
void DCG2SignalTree::dclBooleanTick(ntrees process)
{
  ptexpcal vtick;
  ntrees vtl, vtype, treeTick;
  TypeType t;
  ptinsth hmax;

  if (!isEventClockedFormat(graph_type))
    {
      hmax = CKM->getTickClock();
      vtick = CKM->getBooleanSignal(hmax);
      //  t = event2booleanConversion(get_PureType());
      t = get_PureType();
      vtype = STree_mkNullaryTree(predefTypeType2SignalType(t));
      vtl = LL_make_lident_emptyList();
      vtl = ntreeList_appendLast(vtl, ptexpcal2ntrees(vtick));
      treeTick = LL_make_decsig_tree(vtype, vtl);
      STree_addLocalDeclaration(process, treeTick);
    }
}

/**
   Cleaning the graph (local marks,...)
 */
void DCG2SignalTree::AbstractionFinalize()
{
  ptlinstances lsig;
  ptlinsth lh;

  G_ClockListToDeclareProjection = NILLINSTH;
  CKM->resetMarklClocks(CKM->getProjectedSystem(&lh, &lsig), DontDeclareInAbstaction);
  CKM->resetMarklClocks(lh, DontDeclareInAbstaction);
}

/**
   Cleans the specification part of a model (vmodel).
*/
void DCG2SignalTree::cleanTreeSpec(ntrees vmodel)
{
  ntrees vtspec, body, vt;
  vtspec = STree_getSpec(vmodel);
  if (!STree_isUndefinedTree(vtspec))
    {
      body = nchild(2, vtspec);
      vt = SimplifyComposition(SimplifyComposition(body, pk_false), pk_false);
      if (vt != body)
	{
	  if (IS_NILTREE(vt))
	    vt = STree_mkAtomNilTree();
	  nchangchild(vtspec, vt, 2);
	}
    }
}

/**
   @return the Signal Abstract tree representation of the abstraction of a graph (g).

   When the graph is a graph of clusters, the abstraction is a Grey Box abstraction otherwise it is a black box one.

   - For a black box abstraction, it is a model that specify the i/o data dependencies and the i/o clock relations.
   - For a Grey Box abstraction, the model is completed with the black box abstraction of the clusters and the 
   dependences beteween clusters are added.   

   \see mkOIClosureTree(), mkProjectionTree(),getTrainList(), mkAbstractTreeGreyBoxSubGraphs()
   addPragmaAbstractionGreyBox(), addPragmaAbstractionBlackBox(), dclBooleanTick().
*/
ntrees DCG2SignalTree::IGraph2AbtractionTree(ptDCG g)
{
  ntrees vtDeclSpec, transclosure, treeProjection, process, processGreyBox, x;
  ptinsth htop;
  ptlINode liste;
  ptident id;
  ptinsth tick;
  ntrees vlang;
  boolean o_flatMode;

  DontDeclareInAbstaction = getMark();
  DCG2SignalTreeini(g);
  //  b = pK_messenger->gmessSetTypeTree(pk_true); /* To allow the trees annotation */
  tick = CKM->getTickClock();
  id = g->getName();
  vlang = g->getImplementationLanguage();
  process = STree_mkSkeletonExternProcess(createSuffIdent(id, abstractSuffix), vlang, opprocess);
  o_flatMode = flatMode;
  flatModeInit(process, pk_false);
  addPragmaAbstractionBlackBox(process, id);
  transclosure = mkIOClosureTree();
  treeProjection = mkProjectionTree(tick, pk_false, &vtDeclSpec);
  STree_addSpecification(process, transclosure);
  STree_addSpecification(process, treeProjection);
  STree_addSpecificationDeclaration(process, vtDeclSpec);

  // STree_addSpecificationSafetyProperty(g->get_safetyProperty());
  htop = g->getRootClock();
  mtInterfaceTree(htop, process);
  AssertionValue(process);

  if (getTrainList(&liste))	// Grey Box
    {
      // ATTENTION : pour boite grise l'abstration n'est pas définie dans la zone "spec";
      // Il faudra adapter la premiere passe de compilation pour accepter les 
      // declarations de modele dans toutes les zones "where".
      processGreyBox = STree_mkSkeletonModel(createSuffIdent(id, abstractSuffix), opprocess);
      transclosure = STree_scopy(transclosure);
      treeProjection = mkProjectionTree(tick, pk_true, &vtDeclSpec);
      STree_addExpressionOfProcessToaModel(processGreyBox, treeProjection);
      STree_replaceInterface(processGreyBox, STree_duplicateInterfaceWithoutSpec(process));
      mkAbstractTreeGreyBoxSubGraphs(processGreyBox, liste, vlang);
      addPragmaAbstractionGreyBox(processGreyBox, AGRAPH->getName());
      dclBooleanTick(processGreyBox);
      x = processGreyBox;
    }
  else
    x = process;
  // addAbstractionPragmas(g, x);
  AbstractionFinalize();
  cleanTreeSpec(x);
  flatModeInit(process, o_flatMode);
  return x;
}
