/*
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**  */
#define getNameLabelState(S) STree_scopy(((ptDCG) S)->getLabelStateName())

_PROTECTED_
/**
    Initializing the translator for an automaton(vautomaton).
    Quite similar than for a graph, but some informations are not available.
*/
    void
 DCG2SignalTree::initTranslatorFor(ptAutomaton vautomaton)
{
  ptinterface vinter;

  CKM = vautomaton->getClockManager();
  EM = vautomaton->getExpressionManager();
  NM = vautomaton->getNodeManager();
  NameM = vautomaton->getNameManager();
  lpragmas = vautomaton->getPragmas();

  vinter = vautomaton->getInterface();
  inputs = ptinterface_get_inputSignals(vinter);
  outputs = ptinterface_get_outputSignals(vinter);

  // isEndochronous = AGRAPH->isEndochronous();  // if all the States are endo
  // graph_type = AGRAPH->getFormatLevel(); // according to the level of the States.
  // unNodes = AGRAPH->getUnused(&unClocks); replaced by
  unNodes = NILLINODE;
  unClocks = NILLINSTH;
}

/**
   @return the Signal Abstract Tree representation of a transition (vtransition).
 */
ntrees DCG2SignalTree::mkTreeTransition(ptTransition vtransition)
{
  State s1, s2;
  ptinsth vh;
  ntrees lin;

  vtransition->get(&s1, &s2, &vh);
  lin = LL_make_lesign_list(getClockExpr0(vh));
  lin = ntreeList_insertFirst(getNameLabelState(s2), lin);
  lin = ntreeList_insertFirst(getNameLabelState(s1), lin);
  return LL_make_appe_tree(STree_scopy(getTransitionId()), lin);
}

/**
   Define the Signal Abstract Tree representation of the body of an automaton:
   - the states are represented by a composition of S_P_i:: P_i(), where S_P_i is a label used to define the transitions (via an enumerated type).
   - a transition is representeted by a call to a predefined model: S_P_1 -[C]-> S_P_2 is represented by  Transition(S_P_1, S_P_2, C)
   - the initial state is representeted by a call to a predefined model: Initial_State(S_P_1).

\code
Example:
   (| (| (| on S_DIGITS_1 :: DIGITS_1() |)
       | (| on S_DIGITS_2 :: DIGITS_2() |)
       |)
    | (| Initial_State(S_DIGITS_1)
       | Transition(S_DIGITS_1, S_DIGITS_1, TENS)
       | Transition(S_DIGITS_1, S_DIGITS_2, not TENS)
       | Transition(S_DIGITS_2, S_DIGITS_2, not TENS)
       | Transition(S_DIGITS_2, S_DIGITS_1, TENS)
       |)
    |)
\endcode
 */
void DCG2SignalTree::mkBodyAutomaton(ptAutomaton vautomaton, ntrees vt_automaton)
{
  ptlINode ln = NILLINODE;
  ptINode node;
  ptlTransitions lTransitions;
  ntrees vTreeInit, vtaux, vt_transitions, vt_body;
  pointer elem;
  ptexpcal TheCurrentState;

  // States
  if (NM->getNodeCalls(&ln))
    {
      vt_body = STree_getNilTree();

      actionsAutomaton = pk_true;
      STree_addExprToComposition_Post(&vt_body, inodes2ntrees(ln, NILTREE));
      actionsAutomaton = pk_false;

      // labels
      while (ptlINode_get_elem(&ln, &node))
	{
	  vtaux = getLabelTree(node);
	  if (ISNOT_NILTREE(vtaux))
	    STree_addLabelInModel(vtaux, vt_automaton);
	}
    }

  vt_transitions = STree_getNilTree();
  // InitialState, Current State variable.
  vTreeInit = (ntrees) vautomaton->getInitialState();	// init is an identifier (label name).
  vtaux = LL_make_appe_tree(STree_scopy(getInitialStateId()), vTreeInit);
  STree_addExprToComposition_Post(&vt_transitions, vtaux);

  TheCurrentState = vautomaton->getCurrentStateVar();
  if (TheCurrentState != NILEXPCAL)
    {
      vtaux = LL_make_appe_tree(STree_scopy(getSetCurrentStateId()), ptexpcal2ntrees(TheCurrentState));
      STree_addExprToComposition_Post(&vt_transitions, vtaux);
    }

  // Transitions.
  lTransitions = vautomaton->getTransitions();
  while (ptlPointer_get_elemNext(&lTransitions, &elem))
    {
      STree_addExprToComposition_Post(&vt_transitions, mkTreeTransition((ptTransition) elem));
    }
  STree_addExprToComposition_Post(&vt_body, vt_transitions);

  // Definitions of the outputs.
  if (!NM->getEquations(&ln))
    ln = NILLINODE;
  STree_addExprToComposition_Post(&vt_body, inodes2ntrees(ln, NILTREE));
  STree_setBody(vt_automaton, vt_body);
}

/**
   It adds, in tha local declaration of the Signal Tree model (vt_automaton),
   the declaration of each program (state) of an automaton (vautomaton)
*/
void DCG2SignalTree::mkStatesDeclaration(ptAutomaton vautomaton, ntrees vt_automaton)
{
  pointer velem;
  ptDCG vg;
  ntrees vt;
  boolean save_innerModel;
  ptlStates lStates;

  lStates = vautomaton->getStates();
  save_innerModel = innerModel;
  innerModel = pk_true;
  while (ptlPointer_get_elemNext(&lStates, &velem))
    {
      vg = (ptDCG) velem;
      vt = graph2SignalTree(vg, vg->getName());
      if (ISNOT_NILTREE(vt))
	STree_addProcessDeclaration(vt_automaton, vt);
    }
  innerModel = save_innerModel;
}
