/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
/** 
    @return the SIGNAL Abstract Tree representation of a clock formula.
 */
    ntrees DCG2SignalTree::getClockExpr0(ptinsth vh)
{
  ntrees vt;
  boolean bdummy;

  if (CKM->satisfiesTickClock(vh))
    G_TickIsReferenced = pk_true;
  if (CKM->isSubstitutable(vh))
    if (mkClockExpression(vh, pk_false, &bdummy, &vt))
      return vt;
  return mkhonametree(vh);
}

/** 
    @return the SIGNAL Abstract Tree representation of a clock formula taking 
    into account the attrNotSubstituableClock attibute.
 */
ntrees DCG2SignalTree::getClockExprInWhenCell(ptinsth vh)
{
  ntrees vt;

  vt = getClockExpr0(vh);
  if (noper(vt) == opwhen1)
    vt = nchild(1, vt);
  return vt;
}

/** 
    @return the SIGNAL Abstract Tree representation of a clock formula.
 */
ntrees DCG2SignalTree::getClockExpr(ptinsth vh)
{
  return mac_STRee_get_argIfEvent(getClockExpr0(vh));
}

/** 
    @return the name of the clock without the ^.

    Similar to mkhonametree()
 */
ntrees DCG2SignalTree::mkhonametree1(ptinsth h)
{
  return mac_STRee_get_argIfEvent(mkhonametree(h));
}

/** @return the SIGNAL Abstract Tree representation of
    - the name of a clock (vh) for a DC+ program,
    - the extraction by a condition otherwise (clock is represented by a boolean)
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkWhenClock(ptinsth vh)
{
  if (isEventClockedFormat(graph_type))
    return mkhonametree(vh);
  vh = CKM->getRepresentativeFormula(vh);
  return mkExtraction(CKM->getBooleanSignal(vh), pk_false);
}

/**
   Similar to mkWhenClock(), but when the clock is "^X", it returns "X".
*/
ntrees DCG2SignalTree::mkWhenClock1(ptinsth vh)
{
  ntrees vt;
  boolean bdummy;
  if (CKM->isSubstitutable(vh))
    if (mkClockExpression(vh, pk_false, &bdummy, &vt))
      return vt;
  return mac_STRee_get_argIfEvent(mkWhenClock(vh));
}

/** @return the SIGNAL Abstract Tree representation of a clock formula <em> vharg1 vop vharg2  </em>
    where \em vop denotes an operator of the clock lattice: ^+, ^*, ^-.
 */
_VIRTUAL_ ntrees DCG2SignalTree::mkClockLatticeExpression(noperators vop, ptinsth vharg1, ptinsth vharg2)
{
  return nmakebinary(vop, getClockExpr(vharg1), getClockExpr(vharg2));
}

/**
   @return the SIGNAL Abstract Tree representation of a clock formula vharg ^- vhargnot.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkComplementaryClock(ptinsth vharghnot, ptinsth vharg)
{
  return mkClockLatticeExpression(opcomp, vharg, vharghnot);
}

/**
   @return the SIGNAL Abstract Tree representation of an extraction by a condition (vtform), negative if neg.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkExtraction(ptexpcal vtform, boolean neg)
{
  ntrees vtt;

  vtt = ptexpcal2ntrees(vtform);
  if (neg) vtt = LL_make_non_tree(vtt);
  return LL_make_when1_tree(vtt);
  
}

/**
   @return "true" if a clock (vform) is a affine clock, "false" otherwise.
   For an affine clock, vt is its SIGNAL Abstract Tree representation.
   It is defined by "when affine_sample{phi, d} (h)" with phi the phase, d the period and h the clock of the condition.
*/
_VIRTUAL_ boolean DCG2SignalTree::affineclock2tree(ptinsth vform, ntrees * vt)
{
  ntrees vtreeclk, vtreephase, vtreeperiodD, vtreeparam, vtaux;
  ptAffineRelation affr;
  AffineParam period_n, phase, period_d;
  ptinsth vh;

  if (!EM->isNILEXPCAL(CKM->getConditionSub(vform)))
    return pk_false;
  if (CKM->getAffineRelation(vform, &affr))
    {
      vtreeclk = LL_make_lesign_list(mkhonametree(CKM->getClockSub(vform)));
      ptAffineRelation_get_elems(affr, &period_n, &phase, &period_d);

      vtreephase = LL_make_ent_atom(integer2repstr(phase));
      vtreeperiodD = LL_make_ent_atom(integer2repstr(period_d));
      vtreeparam = LL_make_lepara_list(vtreephase);
      vtreeparam = ntreeList_appendLast(vtreeparam, vtreeperiodD);

      vtaux = LL_make_inst_tree(STree_scopy(getSampleId()), vtreeparam);
      vtaux = LL_make_appe_tree(vtaux, vtreeclk);
      *vt = vtaux;
    }
  else
    {
      CKM->getComplExtraction(vform, &vh);
      *vt = mkComplementaryClock(vh, CKM->getFather(vform));
    }
  return pk_true;
}

/**
   @return the SIGNAL Abstract Tree representation of a negative extraction 
   (affine clock or by an extraction by a condition).
*/
_VIRTUAL_ ntrees DCG2SignalTree::nosub2tree(ptinsth vform, boolean babstraction)
{
  ntrees vt;

  if (affineclock2tree(vform, &vt))
    return vt;
  vt = mkExtraction(vform, babstraction, pk_true);
  return vt;

}

/**
   @return the SIGNAL Abstract Tree representation of an positive extraction 
   (affine clock or an extraction by a condition).
*/
_VIRTUAL_ ntrees DCG2SignalTree::sub2tree(ptinsth vform, boolean babstraction)
{
  ntrees vt;

  if (affineclock2tree(vform, &vt))
    return vt;
  return mkExtraction(vform, babstraction, pk_false);
}

ntrees DCG2SignalTree::mkExtraction(ptinsth vform, boolean babstraction, boolean neg)
{
  boolean bpos;
  ptexpcal exp;

  exp = CKM->getConditionSub(vform);
  if (!babstraction)
    return mkExtraction(exp, neg);

  exp = EM->getExprExportBooleanAbtraction(exp, NILLINSTANCES, &bpos);
  CKM->addMark(vform, DontDeclareInAbstaction);
  if (neg && !bpos)
    neg = pk_false;
  else if (!neg && !bpos)
    neg = pk_true;
  return mkExtraction(exp, neg);
}

/**
   @return "true" if the formula (vform) is not a predefined clock or a variable, "false" otherwise.
   For a positive answer, 
   - vt is a the SIGNAL Abstract Tree representation of a formula
   - isaextract is "true" for an extraction (positive/negative), "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::mkClockExpression(ptinsth vform, boolean babstraction, boolean * isaextract, ntrees * vt)
{
  ptinsth vharg1, vharg2;

  switch ((int)(CKM->getOperator(vform)))
    {
    case sub:
      *vt = sub2tree(vform, babstraction);
      *isaextract = pk_true;
      return pk_true;
      break;
    case nosub:
      *vt = nosub2tree(vform, babstraction);
      *isaextract = pk_true;
      return pk_true;
      break;
    case hcomp:
      *isaextract = pk_false;
      CKM->getUnaryArg(vform, &vharg1);
      *vt = mkComplementaryClock(vharg1, CKM->getFather(vform));
      return pk_true;
      break;
    case infer:
      *isaextract = pk_false;
      if (CKM->getMinusArgs(vform, &vharg1, &vharg2))
	*vt = mkComplementaryClock(vharg1, vharg2);
      else if (CKM->getInfArgs(vform, &vharg1, &vharg2))
	*vt = mkClockLatticeExpression(opbinf, vharg1, vharg2);
      return pk_true;
      break;
    case super:
      *isaextract = pk_false;
      CKM->getSupArgs(vform, &vharg1, &vharg2);
      *vt = mkClockLatticeExpression(opbsup, vharg1, vharg2);
      return pk_true;
      break;
    case vari:
    case predef:
      SET_NILTREE(*vt);
      *isaextract = pk_false;
      return pk_false;
      break;
    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
  return pk_false;
}

boolean DCG2SignalTree::getClockNameRepresentedBySignal(ptinsth vh, ntrees * vid)
{
  ptexpcal vsig;

  if (!CKM->isClockRepresentedBySignal(vh, &vsig)) return pk_false;
  (*vid) = ptexpcal2ntrees(vsig);
  return pk_true;
}

/**  
     @return the SIGNAL Abstract Tree representation of the name of a clock (vclock).
*/
ntrees DCG2SignalTree::mkhonametree(ptinsth vclock)
{
  ntrees id;

  if (CKM->satisfiesIsAnEquivalentForm(vclock))
    vclock = CKM->getRepresentativeFormula(vclock);
  if (CKM->satisfiesUndefClock(vclock))
    return STree_mkAtomNilTree();
  if (CKM->satisfiesNullClock(vclock))
    return LL_make_hnull_atom();
  if (isEventClockedFormat(graph_type))
    {
      if (getClockNameRepresentedBySignal(vclock, &id))
	return LL_make_evnt_tree(id);
      return copy_ident(CKM->getName(vclock));
    }
  id = ptexpcal2ntrees(CKM->getBooleanSignal(vclock));
  if (id == NILIDENT)
    id = copy_ident(CKM->getName(vclock));	/* when a clock is removed from the graph and referenced in the unused signals. */
  return id;
}

/**
   @return the SIGNAL Abstract Tree representation (a composition) of a list of clocks (lh),
   added to the operand list (vtreereached).
 */
ntrees DCG2SignalTree::clocks2ntrees(ptlinsth lh, ntrees vtreereached)
{
  ptinsth h;

  while (ptlinsth_get_next(&lh, &h))
    if (!CKM->isSubstitutable(h))
      STree_addExprToComposition_Post(&vtreereached, clock2Tree(h));
  return vtreereached;
}

/** 
    @return the SIGNAL Abstract Tree representation of the operand formula (h). 
*/
_VIRTUAL_ ntrees DCG2SignalTree::clock2Tree(ptinsth h)
{
  boolean bdummy;
  ntrees vt;

  vt = formula2ntrees(h, NILTREE, &bdummy);
  if (IS_NILTREE(vt))
    mkClockExpression(h, pk_false, &bdummy, &vt);
  mkClockDebugInfos(h, vt);
  return vt;
}

/**
   @return the Signal abstract tree of the tick clock (h) taking into account the level format.
*/
_VIRTUAL_ ntrees DCG2SignalTree::__tick2Tree(ptinsth h)
{
  ptexpcal vtick;

  // -- (Aug 2013) UNUSED. The Tick is not printed in BOOLEAN format, it is an event Signal.
  if (isEventClockedFormat(graph_type))
    return LL_make_defsig_tree(mkhonametree(h), LL_make_evnt_tree(mkhonametree(h)));
  vtick = CKM->getBooleanSignal(h);
  if (flatMode)
    G_LocSig = ptlinstances_insert(G_LocSig, vtick);
  return LL_make_defsig_tree(ptexpcal2ntrees(vtick), STree_mkNullaryTree(optrue));
}

/** 
    @return the Signal Tree Abstraction representation of a formula of the clock calculus.
    @param vform (in): the formula to translate.
    @param signame (in): used when the formula is a free variable. 
    In this case, if signame is defined,  it is used to defined the formula ( <em> identifier of vform := ^ signame </em>),
    otherwise (signame is undefined)
    the returned tree is <em> identifier of vform := ^ identifier of vform> </em> 
    @param isaextract (out): set to pk_true when vform is an extraction.
*/
ntrees DCG2SignalTree::formulaTontrees(ptinsth vform, ntrees signame, boolean * isaextract)
{
  ptexpcal aSig;
  ntrees vt, vt1, vname;
  boolean b;
  ptident id;

  if (!isEventClockedFormat(graph_type))
    {
      // if (CKM->satisfiesTickClock(vform)) return tick2Tree(vform); remove Aug 2013 (LBesnard)
      *isaextract = pk_true;
      return NILTREE;
    }
  if (CKM->isClockUnifiedWithSignal(vform, &aSig))
    b = ptlinstances_isin_exp(inputs, aSig);
  else
    b = pk_false;

  vname = mkhonametree(vform);
  *isaextract = pk_false;
  // Null clock.
  if (CKM->satisfiesNullClock(vform))
    {
      vt = LL_make_hnull_atom();
      if (!b)
	return vt;
      if (IS_NILTREE(signame))
	signame = STree_scopy(vname);
      vt1 = mac_STree_make_initSynchro(signame);
      vt1 = ntreeList_appendLast(vt1, signame);
      return vt1;
    }
  // Others
  if (!mkClockExpression(vform, pk_false, isaextract, &vt))
    return NILTREE;		// Clock Parameter.
  if (!b)
    {
      if (IS_NILTREE(signame))
	id = STree_scopy(vname);
      else
	id = signame;
      vt1 = mac_STree_make_initSynchro(id);
      return ntreeList_appendLast(vt1, vt);
    }
  if (KernelExprMode && noDoubleSynchroSameId)
    {
      // Added for Julio Perralta: he does not want the expression X ^= X (a signal must be referenced once in a synchronisation)
      id = mkNewIdentKernelMode();
      vt = LL_make_defsig_tree(STree_scopy(id), vt);
      STree_addExpressionOfProcessToaModel(G_TreeGraph, vt);
      STree_addLocalDeclaration(G_TreeGraph, STree_mkSignalDeclaration(STree_scopy(id), TypeType2ntrees(PureType), NILTREE));
      vt1 = mac_STree_make_initSynchro(STree_scopy(id));
      vt1 = ntreeList_appendLast(vt1, STree_scopy(vname));
      STree_addExpressionOfProcessToaModel(G_TreeGraph, vt1);
      return NILTREE;
    }
  vt1 = mac_STree_make_initSynchro(vt);
  vt1 = ntreeList_appendLast(vt1, STree_scopy(vname));
  return vt1;
}

ntrees DCG2SignalTree::formula2ntrees(ptinsth vform, ntrees signame, boolean * isaextract)
{
  ntrees vt;

  vt = formulaTontrees(vform, signame, isaextract);
  if (DEBUG_MODE)
    if (ISNOT_NILTREE(vt))
      STree_setVMAnnotations(vt, mknumtree(CKM->getNum(vform)));
  return vt;
}

/**
   For debugging:
   The tree (t) is annotated with the dependence informations associated with a clock (clk).
*/
_VIRTUAL_ void DCG2SignalTree::mkClockDebugInfos(ptinsth clk, ntrees t)
{
  ntrees vattr;
  ptINode n;
  ptlINode ln;

  if (DEBUG_MODE)
    if (ISNOT_NILTREE(t))
      {
	ln = CKM->getNodeReferencesTo(clk);
	vattr = LL_make_lident_emptyList();
	while (ptlINode_get_elem(&ln, &n))
	  {
	    vattr = ntreeList_appendLast(vattr, LL_make_nom_atom(integer2repstr(NM->get_num(n))));
	  }
	if (ntreearity(vattr) > 0)
	  {
	    vattr = LL_make_egal_tree(LL_make_nom_atom(string2repstr("Node Succ")), vattr);
	    STree_setVMAnnotations(t, vattr);
	  }
      }
}

ntrees DCG2SignalTree::mkhonametree(ptinsth vh, boolean babstract)
{
  ntrees vtClock;
  ptexpcal aSig;

  return getClockExpr(vh);

  if (CKM->isClockRepresentedBySignal(vh, &aSig))
    return ptexpcal2ntrees(aSig);
  if (CKM->isClockUnifiedWithSignal(vh, &aSig))
    return ptexpcal2ntrees(aSig);

  if (!babstract)
    vtClock = mkhonametree(vh);
  else if (CKM->satisfiesSub(vh))
    if (CKM->getOperator(vh) == sub)
      vtClock = sub2tree(vh, babstract);
    else
      vtClock = nosub2tree(vh, babstract);
  else
    vtClock = mkhonametree(vh);
  return vtClock;
}

ntrees DCG2SignalTree::mkClockSynchro(ptlinstances lsig, boolean bwithTick)
{
  ptexpcal vsig;
  ptinsth h;
  ntrees synchros, vtli;
  boolean first, b;
  ptlinstances lsigSave;
  ptlinsth lh, lhSave;

  lsigSave = lsig;
  lh = EM->getClockOfSignals(lsig);
  lhSave = lh;
  SET_NILTREE(synchros);
  while (ptlinsth_get_next(&lh, &h))
    {
      first = pk_true;
      SET_NILTREE(vtli);
      lsig = CKM->getSynchronousSignals(h);
      while (ptlinstances_get_next(&lsig, &vsig))
	if (ptlinstances_isin_exp(lsigSave, vsig))
	  addSynchroClockSignal(h, vsig, &first, &vtli, pk_true);
      vtli = ClockSynchroClean(vtli);
      if (!STree_isSignalNilTree(vtli))
	{
	  if (ntreearity(vtli) > 1)
	    {
	      if (!bwithTick && CKM->satisfiesTickClock(h))
		{
		  ndestroyListElem(vtli, 1);
		  b = (ntreearity(vtli) > 1);
		}
	      else
		b = pk_true;
	    }
	  if (b)
	    STree_addExprToComposition_Post(&synchros, vtli);
	}
    }
  ptlinsth_free(lhSave);
  return synchros;
}

boolean DCG2SignalTree::getIOSignalIn(ptlinstances ls, ptexpcal * aSig)
{
  while (ptlinstances_get_next(&ls, aSig))
    if (EM->IsInputFlow(*aSig) || EM->IsOutputFlow(*aSig))
      return pk_true;
  return pk_false;
}

ntrees DCG2SignalTree::addClockRelation(ptlinsth lhToCompare, ptlinsth lh)
{
  ntrees vtres, vt;
  ptinsth vh, vh1;
  ptlinsth lhs;
  ptexpcal aSig1, aSig2;

  vtres = STree_getNilTree();
  lhs = lh;
  while (ptlinsth_get_next(&lhToCompare, &vh))
    {
      while (ptlinsth_get_next(&lhs, &vh1))
	{
	  if (!CKM->eqClock(vh, vh1))
	    if (CKM->isIncluded(vh, vh1))
	      /* For an i/o signal (X) with vh1 as clock, and an i/o signal (Y) with vh as clock, adding "X ^< Y" */
	      if (getIOSignalIn(CKM->getSynchronousSignals(vh), &aSig1) && getIOSignalIn(CKM->getSynchronousSignals(vh1), &aSig2))
		{
		  vt = LL_make_lcrohor_list(ptexpcal2ntrees(aSig1));
		  vt = ntreeList_appendLast(vt, ptexpcal2ntrees(aSig2));
		  STree_addExprToComposition_Post(&vtres, vt);
		}
	}
      lhs = lh;
    }
  return vtres;
}
