/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
#define MayBeInputOutput(vexp) (! EM->isSharedSignalBySeveralGraphs(vexp) && ! EM->isStateVar(vexp) && ! EM->isSharedSignal(vexp))
/**
   In "flat mode", it adds the declarations of the signals (G_LocSig) and the clocks (G_Clocks) 
   in the global model (G_TreeGraph).
   The signals, clocks are stored in the list during the graph travel.
*/
    _VIRTUAL_ void
 DCG2SignalTree::insertDeclarationsFlatMode()
{
  if (flatMode)
    {
      appdeclsig(G_TreeGraph, G_LocSig, 0, pk_false);
      if (G_Clocks != NILLINSTH)
	STree_addLocalDeclaration(G_TreeGraph, mkDeclarationOfClocks(G_Clocks));
    }
}

/**
   @return "true" if a graph (aGraph) is an \em external graph with a back annotation (SIGNALTreeDecl) to its SIGNAL abstract tree,
   "false" otherwise.
*/
_VIRTUAL_ boolean DCG2SignalTree::isKnown_SIGNAL_tree(ptDCG aGraph, ntrees * SIGNALTreeDecl)
{
  ntrees vtcurrent;

  if (!aGraph->isGraphOfExternalProcess())
    return pk_false;
  vtcurrent = aGraph->getSIGNALTreeDeclaration(); /** reference a l'arbre de la declaration du process (SIGNAL only) */
  if (IS_NILTREE(vtcurrent))
    return pk_false;
  *SIGNALTreeDecl = STree_scopy(vtcurrent);
  return pk_true;
}

/**
   @return SIGNAL abstract tree representation of a new identifier built on identPrefix (attribute member of the class) and a counter
   of such identifiers (cpt_names: attribute member of the class).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkNewIdent()
{
  cpt_names = cpt_names + 1;
  return makeIdent_ident_int(identPrefix, cpt_names);
}

_VIRTUAL_ void DCG2SignalTree::mktreeinterface(ptinterface vinterface, ntrees vtl)
{
  mktreeparameters(vinterface, vtl);
  mktreeinput(vinterface, vtl);
  mktreeoutput(vinterface, vtl);
}

/** 
    @return the SIGNAL abstract tree representation of an interface (vinterface)
    It is composed of the input signals, outputs signals of vinterface.
 */
_VIRTUAL_ ntrees DCG2SignalTree::mktreeinterface(ptinterface vinterface)
{
  ntrees vtl;
  vtl = STree_mkSkeletonModel(NILIDENT, opprocess);
  mktreeinterface(vinterface, vtl);
  return nchild(2, vtl);
}

/**
   @return the Signal abstract tree representation of a definition of a constant (aConst).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkConstantDefinition(ptexpcal aConst)
{
  ptexpcal vLeftPart, vRightPart;
  ntrees lconst, vtdef, treeType, vt;
  ptINode vnode;

  if (EM->IsExternalConstant(aConst))
    {
      vtdef = LL_make_cons_tree(notKernptexpcal2ntrees(aConst), STree_mkUndefinedExternalValue());
      vLeftPart = aConst;
    }
  else
    {
      vnode = EM->getNode(aConst);
      NM->get_equation_arg(vnode, &vLeftPart, &vRightPart);
      vtdef = LL_make_cons_tree(ptexpcal2ntrees(vLeftPart), notKernptexpcal2ntrees(vRightPart));
    }

  lconst = LL_make_lcons_list(vtdef);
  treeType = typeExp2ntrees(vLeftPart);
  vt = LL_make_deccon_tree(treeType, lconst);
  if (STree_isImplicitTypeTree(treeType))
    pK_messenger->warningaccForce(warUntypedSignal, vt);
  return vt;
}

/**
   For each element of a list of assumed constants (lsig),
   it adds, to a model (aModelTree), the declaration (STree_addLocalDeclaration()) of the definition of the constant (mkConstantDefinition()).
*/
_VIRTUAL_ void DCG2SignalTree::mktreeconstants(ntrees aModelTree, ptlinstances lsig)
{
  ptexpcal aConst;

  while (ptlinstances_get_next(&lsig, &aConst))
    STree_addLocalDeclaration(aModelTree, mkConstantDefinition(aConst));
}

/**
   @return the Signal abstract tree representation of a declaration of a signal (vexp) that can be initialized (withInit).
   \see mkIdentWithInit().
 */
_VIRTUAL_ ntrees DCG2SignalTree::mkdecsig(ptexpcal vexp, boolean withInit)
{
  ntrees vlsig, t, t1;

  vlsig = LL_make_lsign_list(mkIdentWithInit(vexp, withInit));
  t = typeExp2ntrees(vexp);
  if (EM->isStateVar(vexp))
    t1 = LL_make_decvar_tree(t, vlsig);
  else if (EM->isSharedSignal(vexp))
    t1 = LL_make_decpar_tree(t, vlsig);
  else
    t1 = LL_make_decsig_tree(t, vlsig);
  if (STree_isImplicitTypeTree(t))
    pK_messenger->warningaccForce(warUntypedSignal, vlsig);
  return t1;
}

/**
   @return 
   - a new identifier (mkNewIdent()) when mkNewIdentTag (attribute member of the class); it is used when the context 
   operation is a declaration of a external model (mkDeclarationExternProc()) from a reference to the model.
   - the Signal abstract tree representation of the expression (ptexpcal2ntrees()).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkIdentTree(ptexpcal vexp)
{
  if (mkNewIdentTag)
    return mkNewIdent();
  return ptexpcal2ntrees(vexp);
}

_VIRTUAL_ boolean DCG2SignalTree::getInfoValue(ptexpcal vexp, ntrees * t)
{
  compilerInteger n, m;
  TypeType TT;
  ntrees t1, t2;

  TT = EM->getType(vexp);
  if (TT == PureType) return pk_false;
  if (TT == BooleanType)
    {
      if (! EM->IsSignal(vexp)) return pk_false;
      if (EM->eq_trueExpcal(vexp)) 
	{
	  (*t) = ptexpcal2ntrees(vexp);
	  return pk_true;
	}
      
      if (EM->eq_trueExpcal(vexp)) 
	{
	  t1 = ptexpcal2ntrees(vexp);
	  t1 =  LL_make_non_tree(t1);
	  (*t) = t1;
	  return pk_true;
	}
      return pk_false;
    }
  if (!EM->getBounds(vexp, &n, &m)) return pk_false;
  if (n == m)
    {
      t1 = STree_mkCstTree((int)n);
      t1 = LL_make_egal_tree(ptexpcal2ntrees(vexp), t1);
    }
  else
    {
      t1 = STree_mkCstTree((int)n);
      t2 = STree_mkCstTree((int)m);
      t1 = LL_make_iegal_tree(t1, ptexpcal2ntrees(vexp));
      t2 = LL_make_iegal_tree(ptexpcal2ntrees(vexp), t2);
      t1 = LL_make_et_tree(t1, t2);
    }
  (*t) = t1;
  return pk_true;
}

/**
   @return the SIGNAL abstract tree representation of the identifier of a signal (vexp) + its initialization (if available)
   when vexp is a state variable or when \em withInit.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkIdentWithInit(ptexpcal vexp, boolean withInit)
{
  ptexpcal vexpini;
  ntrees t, t1;

  if (EM->IsSignal(vexp) || EM->IsParameter(vexp))
    {
      t = mkIdentTree(vexp);
      if (EM->isStateVar(vexp) || withInit)
	{
	  vexpini = EM->getInitialValue(vexp);
	  if (!EM->isNILEXPCAL(vexpini))
	    t = LL_make_init_tree(t, notKernptexpcal2ntrees(vexpini));
	}
      if (TraceabilityMode)
	if (getInfoValue(vexp, &t1))
	  STree_setVMAnnotations(t, t1);
      return t;
    }
  // For external call with expressions (declaration of a model from a reference to this model).
  return mkNewIdent();
}

/**
   In "flat mode", the signals (ls) are added to G_LocSig (all the locals of all the nodes of the hierarchy will be declared at the tick clock),
   otherwise  the signals of ls are added to the locals of a model (aModelTree).
*/
_VIRTUAL_ void DCG2SignalTree::localStoreOrDeclare(ptlinstances ls, ntrees aModelTree)
{
  if (ls == NILLINSTANCES)
    return;
  if (!flatMode)
    appdeclsig(aModelTree, ls, 0, pk_false);
  else
    G_LocSig = ptlinstances_sortedfusion(ls, G_LocSig);
}

/**
   In "flat mode", the clocks (lh) are added to G_Clocks (all the locals of all the nodes of the hierarchy will be declared at the tick clock),
   otherwise  the signals of ls are added to the locals of a model (aModelTree).
*/
_VIRTUAL_ void DCG2SignalTree::localStoreOrDeclare(ptlinsth lh, ntrees aModelTree)
{
  if (lh == NILLINSTH)
    return;
  if (!flatMode)
    STree_addLocalDeclaration(aModelTree, mkDeclarationOfClocks(lh));
  else
    G_Clocks = ptlinsth_sorted_fusion(G_Clocks, lh);
}

/**
   It adds the declaration of the local of an interface (vinter) 
   to a SIGNAL abstract tree representation of a model (process).

   It calls localStoreOrDeclare(ptlinstances) for the signals and localStoreOrDeclare(ptlinsth) for the clocks.
*/
_VIRTUAL_ void DCG2SignalTree::appendLoc(ptinterface vinter, ntrees process)
{
  ptlinsth lhorl;
  ptlinstances lsig, lsig1, curlist;
  ptinsth vh1;
  ptexpcal vsig;

  lsig = ptinterface_get_localSignals(vinter);
  localStoreOrDeclare(lsig, process);
  lhorl = ptinterface_get_localClocks(vinter);
  if (isEventClockedFormat(graph_type))
    localStoreOrDeclare(lhorl, process);
  else
    {
      /** bidouille pour ne pas avoir les doubles declarations ; devrait etre calcule dans mkInterface. */
      lsig1 = NILLINSTANCES;
      while (ptlinsth_get_next(&lhorl, &vh1))
	{
	  vh1 = CKM->getRepresentativeFormula(vh1);
	  if (!CKM->satisfiesPredefinedClock(vh1))
	    {
	      vsig = CKM->getBooleanSignal(vh1);
	      if (vsig != NILEXPCAL)	// blindage
		if (!ptlinstances_isin_exp(lsig, vsig))
		  ptlinstances_isin_insert_exp(&lsig1, &curlist, vsig);
	    }
	}
      localStoreOrDeclare(lsig1, process);
    }
}

/** 
    @return vtdecl in which a declaration (aDecl) of a signal/type (vexp) is added.
    @param loc (in): must be "true" for a list of local declaration, "false" otherwise.
*/
_VIRTUAL_ ntrees DCG2SignalTree::addSignalDeclaration(ntrees vldecl, ntrees aDecl, boolean loc)
{
  if (loc)
    return STree_addDeclarationToList(vldecl, aDecl);
  return STree_addEndDeclarationToList(vldecl, aDecl);	// ntreeList_appendLast(vldecsig, vt);
}

/** 
    @return vldecsig + the declaration of a signal/type (vexp).
    @param vexp (in) : a signal or a reference to a type.
    @param vldecsig (in) : a list of declarations (niltree is accepted)
    @param locaux (in) : must be "true" for a list of local declaration, "false" otherwise.
    @param withInit (in) : must be "true" when the initial value (for a signal) must be generated in the declaration, "false" otherwise.
*/
_VIRTUAL_ ntrees DCG2SignalTree::insertdecsig(ptexpcal vexp, ntrees vldecsig, boolean locaux, boolean withInit)
{
  noperators op, vop;
  ntrees vt;
  boolean btype;

  // Filtering: the shared/state variables are not inserted if an interface.
  if (!locaux && (EM->isStateVar(vexp) || EM->isSharedSignal(vexp)))
    return vldecsig;
  vop = EM->getOperator(vexp);

  btype = (vop == optype);
  if (btype)
    {
      TypeTypeUser2ntrees(EM->getType(vexp), &vt);
      vt = LL_make_ltype_list(vt);
    }
  else
    vt = mkdecsig(vexp, withInit);

  if (IS_NILTREE(vldecsig))
    {
      if (locaux)
	op = opldecl;
      else
	op = opldecs;
      return nmakelist(op, vt);
    }
  if (btype)
    return ntreeList_appendLast(vldecsig, vt);
  return addSignalDeclaration(vldecsig, vt, locaux);
}

/**
   @return vt + the declarations of the element of lexp.
   For each element of a list of declarations (lsig), it calls insertdecsig().

   @param lexp (in) : list of expressions. An element is assumed to be a signal or a reference to a type.
   @param vt (in) : a list of declarations (niltree is accepted)
   @param locaux (in) : must be "true" for a list of local declaration "false" otherwise. When it is pk_true, the io signals are filtered.
   @param withInit (in) : must be "true" when the initial value (for a signal) must be generated in the declaration, "false" otherwise.
*/
_VIRTUAL_ ntrees DCG2SignalTree::addSignalsDeclaration(ntrees vt, ptlinstances lexp, boolean locaux, boolean withInit, boolean bExcludeSharedSignal)
{
  ptexpcal vaux, vsig;
  boolean b;

  while (ptlinstances_get_next(&lexp, &vsig))
    {
      if (locaux)
	{
	  if (isIOSignal(vsig))
	    b = pk_false;
	  else
	    b = !bExcludeSharedSignal;
	}
      else
	b = MayBeInputOutput(vsig);
      if ( isScheduledGraph())
	{
	  vaux = getExpressionToGenerate(vsig);
	  b = (vaux == vsig);
	  if (b) 
	    {
	    ptINode n = EM->getNode (vsig);
	    b = EM->IsOutputFlow (vsig) || EM->IsInputFlow (vsig) || NM->mustBeGenerated(n);
	  }
	}
      if (b)
	{
	  vt = insertdecsig(vsig, vt, locaux, withInit);
	  if ( isScheduledGraph()) addClockScheduledGraph(vsig);
	}
    }
  return vt;
}

/*
  @return vt + the declarations of the element of lh
   For each element of a list of declarations (lh), it calls insertdecsig().

   @param lexp (in) : list of expressions. An element is assumed to be a signal or a reference to a type.
   @param vt (in) : a list of declarations (niltree is accepted)
   @param locaux (in) : must be "true" for a list of local declaration "false" otherwise. When it is pk_true, the io signals are filtered.
   @param withInit (in) : must be "true" when the initial value (for a signal) must be generated in the declaration, "false" otherwise.
*/
_VIRTUAL_ ntrees DCG2SignalTree::addClocksDeclaration(ntrees vt, ptlinsth lh, boolean locaux, boolean withInit, ntrees * vtlclocks)
{
  ptlinstances lsig;
  boolean bdone;
  ptinsth vh;
  ntrees vtl;

  SET_NILTREE(vtl);
  while (ptlinsth_get_next(&lh, &vh))
    {
      if (CKM->isSubstitutable(vh))
	if (CKM->satisfiesSub(vh))
	  {
	    lsig = EM->getSignalsReferencedIn(CKM->getConditionSub(vh));
	    vt = addSignalsDeclaration(vt, lsig, locaux, withInit, pk_true);
	    bdone = pk_true;
	  }
      if (!bdone)
	vtl = insertClockIdentifierToList(vtl, vh);
    }
  if (ISNOT_NILTREE(vtl))
    vtl = mkClockDeclarationFinalize(vtl);
  *vtlclocks = vtl;
  return vt;
}

/**
  It adds the declarations of the signals/types of lexp to a SIGNAL abstract tree representation of a model (aModelTree).
  The inserting area is given by where: 
  - 0: locals 
  - 1: parameters
  - 2: inputs 
  - 3: outputs.
*/
_VIRTUAL_ void DCG2SignalTree::appdeclsig(ntrees aModelTree, ptlinstances lsig, int where, boolean bExcludeSharedSignal)
{
  ntrees vt, vtchild, vtres;
  int vindex;

  if (lsig != NILLINSTANCES)
    {
      vt = STree_getDeclaration(aModelTree, where, &vindex);
      vtchild = nchild(vindex, vt);
      vtres = addSignalsDeclaration(vtchild, lsig, (where == 0), (vindex == 3), bExcludeSharedSignal);
      if (IS_NILTREE(vtchild))
	nchangchild(vt, vtres, vindex);
    }
}

/**
   It adds to a SIGNAL abstract tree representation of a model (aModelTree) the 
   declaration of the external models referenced by the model.
   It explores the list of the visited graphs (stored in lIGraph, an attribute memeber of the class).
*/
void DCG2SignalTree::appDeclExternNodeCall(ntrees aModelTree)
{
  pointer elem;
  ptlPointer lg;
  ptDCG aGraph;
  ntrees vtcurrent;
  boolean done;

  lg = lIGraph;
  while (ptlPointer_get_elemNext(&lg, &elem))
    {
      aGraph = (ptDCG) elem;
      done = isKnown_SIGNAL_tree(aGraph, &vtcurrent);
      if (!done)
	{
	  vtcurrent = STree_mkSkeletonModel(aGraph->getName(), aGraph->getKind());
	  IGraph2Process(aGraph, vtcurrent);
	}
      STree_addProcessDeclaration(aModelTree, vtcurrent);
    }
}

/**
   XXXXXXXXXXXXXXXXXXXxxxx   
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkDeclarationExternProc(ptINode vnode, noperators op, boolean bmkNewIdentTag)
{
  ptident saveId, id;
  ptexpcal entrees, sorties, elem, par;
  ntrees vt, vlang;
  boolean s;

  NM->get_procCall_arg(vnode, &id, &par, &entrees, &sorties);
  if (!NM->getAttribute(vnode, EXTERN_LANGUAGE_IMPL, (attributeValue *) & vlang))
    SET_NILTREE(vlang);
  vt = STree_mkSkeletonExternProcess(id, vlang, op);
  s = mkNewIdentTag;
  mkNewIdentTag = bmkNewIdentTag;
  cpt_names = 0;		// names can be created.
  saveId = identPrefix;
  identPrefix = string2ident("i");
  while (entrees != NILEXPCAL)
    {
      EM->getBinaryArgs(entrees, &elem, &entrees);
      appdeclsig(vt, ptlinstances_push_front(NILLINSTANCES, elem), 2, pk_false);
    }
  identPrefix = string2ident("o");
  while (sorties != NILEXPCAL)
    {
      EM->getBinaryArgs(sorties, &elem, &sorties);
      appdeclsig(vt, ptlinstances_push_front(NILLINSTANCES, elem), 3, pk_false);
    }
  addPragmas(NM->getPragmas(vnode), vt);
  mkNewIdentTag = s;
  identPrefix = saveId;

  return vt;
}

/** 
    It adds to a SIGNAL abstract tree representation of a model (aModelTree) the 
    declaration of the external models referenced in lProcCalls (procedure calls) and lfuncs (functions).
    The objects marked by INTRINSIC and DONTDECLARE (compiler attributes) are filtered.
    - For a procedure call, the declaration is produced by mkDeclarationExternProc()
    - For a function call, the declaration is assumed available by a back annotation to the SIGNAL abstract tree representation produced
    by the parser (getBackAnnotation()).
*/
_VIRTUAL_ void DCG2SignalTree::appDeclExternProcedures(ntrees aModelTree, ptlINode lProcCalls, ptlinstances lfuncs)
{
  ptexpcal aCall;
  ptINode vnode;
  ntrees t;
  ptlPointer ls;

  while (ptlINode_get_elem(&lProcCalls, &vnode))
    if (!NM->is_intrinsicfunc(vnode) && !NM->isMarked(vnode, DONTDECLARE))
      {
	t = BACK_ANNOT_GET_ABSTRACT_TREE(NM->getBackAnnotation(vnode));
	if (ISNOT_NILTREE(t))
	  STree_addProcessDeclaration(aModelTree, mkDeclarationExternProc(vnode, noper(t), pk_false));
      }

  ls = NILLPOINTER;
  while (ptlinstances_get_next(&lfuncs, &aCall))
    if (!EM->isMarked(aCall, INTRINSIC) && !EM->isMarked(aCall, DONTDECLARE))
      {
	t = BACK_ANNOT_GET_ABSTRACT_TREE(EM->getBackAnnotation(aCall));
	if (!ptlPointer_IsIn(ls, t))
	  {
	    ls = ptlPointer_add(ls, t);
	    STree_addProcessDeclaration(aModelTree, STree_scopy(t));
	  }
      }
  ptlPointer_free(ls);
}

/**  
     It adds the declarations of the constants/types (ltc) to a SIGNAL abstract tree representation of a model (aModelTree).
     A constant/type is added (testCanBeGenerated()) when all its references are already declared in aModelTree.
 */
_VIRTUAL_ void DCG2SignalTree::mkDeclCstTypes(ptlinstances ltc, ntrees aModelTree)
{
  ptexpcal elem;
  ntrees vt;
  typeMark aMark;
  ptlinstances lsave;
  boolean b;

  if (ltc != NILLINSTANCES)
    {
      aMark = getMark();
      lsave = ltc;
      b = pk_true;
      while (b)
	{
	  b = pk_false;
	  ltc = lsave;
	  while (ptlinstances_get_next(&ltc, &elem))
	    if (!EM->isMarked(elem, aMark))
	      if (testCanBeGenerated(elem, aMark, lsave))
		{
		  vt = mkDeclCstType(elem);
		  EM->addMark(elem, aMark);
		  if (ISNOT_NILTREE(vt))
		    STree_addLocalDeclaration(aModelTree, vt);
		  b = pk_true;
		}
	}
      EM->resetMark(lsave, aMark);
    }
}

/**
   @return the SIGNAL abstract tree representation of a declaration of a list of clock identifiers (vtl).
   When the graph is a DC+ graph, the "event" type is used, "boolean" otherwise.   
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkClockDeclarationFinalize(ntrees vtl)
{
  ntrees vt;

  if (IS_NILTREE(vtl))
    return NILTREE;
  if (isEventClockedFormat(graph_type))
    vt = STree_mkNullaryTree(opevent);
  else
    vt = STree_mkNullaryTree(oplogical);
  return LL_make_decsig_tree(vt, vtl);
}

ntrees DCG2SignalTree::insertClockIdentifierToList(ntrees vtl, ptinsth vh)
{
  ptexpcal vsig;

  if (!CKM->isClockRepresentedBySignal(vh, &vsig))
    {
      if (IS_NILTREE(vtl))
	vtl = LL_make_lident_emptyList();
      vtl = ntreeList_insertFirst(mkhonametree(vh), vtl);
    }
  return vtl;
}

/** 
    @return the Signal abstract tree representation of a list of identifiers of clocks (lClocks).
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkClockIdentifierList(ptlinsth lClocks)
{
  ntrees vtl;
  ptinsth vh;

  SET_NILTREE(vtl);
  while (ptlinsth_get_next(&lClocks, &vh))
    vtl = insertClockIdentifierToList(vtl, vh);
  return vtl;
}

/** 
    @return the Signal abstract tree representation of a declaration of a list of clocks (lClocks).
    It calls mkClockDeclarationFinalize(), mkClockIdentifierList().
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkDeclarationOfClocks(ptlinsth lClocks)
{
  return mkClockDeclarationFinalize(mkClockIdentifierList(lClocks));
}

/**
   Adds the parameters of an inteface (vinterface) to the Signal model (aModelTree).
*/
_VIRTUAL_ void DCG2SignalTree::mktreeparameters(ptinterface vinterface, ntrees aModelTree)
{
  appdeclsig(aModelTree, ptinterface_get_parameters(vinterface), 1, pk_false);
}

/** 
    Adds the parameters of the inteface of a clock (vh) to the Signal model (aModelTree).
*/
_VIRTUAL_ void DCG2SignalTree::mktreeparameters(ptinsth vh, ntrees aModelTree)
{
  mktreeparameters(CKM->getInterfaceSubHierarchy(vh), aModelTree);
}

_VIRTUAL_ void DCG2SignalTree::mktreeinput(ptinterface vinterface, ntrees aModelTree)
{
  ptlinstances lsig;
  ptlinsth lClocks;

  lsig = ptinterface_get_inputSignals(vinterface);
  appdeclsig(aModelTree, lsig, 2, pk_true);
  lClocks = ptinterface_get_inputClocks(vinterface);
  if (lClocks != NILLINSTH)
    STree_addDeclarationToInterface(aModelTree, mkDeclarationOfClocks(lClocks), 2);
}

/** It adds to the Signal abstract tree representation of the model (aModelTree) associated with a clock (vh)
    the declaration of the inputs of the model.
*/
_VIRTUAL_ void DCG2SignalTree::mktreeinput(ptinsth vh, ntrees aModelTree)
{
  mktreeinput(CKM->getInterfaceSubHierarchy(vh), aModelTree);
}

_VIRTUAL_ void DCG2SignalTree::mktreeoutput(ptinterface vinterface, ntrees aModelTree)
{
  ptlinstances lsig;
  ptlinsth lClocks;

  lsig = ptinterface_get_outputSignals(vinterface);
  appdeclsig(aModelTree, lsig, 3, pk_true);
  lClocks = ptinterface_get_outputClocks(vinterface);
  if (lClocks != NILLINSTH)
    STree_addDeclarationToInterface(aModelTree, mkDeclarationOfClocks(lClocks), 3);
}

/** It adds to the Signal abstract tree representation of the model (aModelTree) associated with a clock (vh)
    the declaration of the outputs of the model.
*/
_VIRTUAL_ void DCG2SignalTree::mktreeoutput(ptinsth vh, ntrees aModelTree)
{
  mktreeoutput(CKM->getInterfaceSubHierarchy(vh), aModelTree);
}

/** 
    It adds to the Signal abstract tree representation of the model (aModelTree) associated with a clock (vh)
    the declaration of the locals of the model. It takes into account the "flat mode" (appendLoc()).
*/
_VIRTUAL_ void DCG2SignalTree::mktreeloc(ptinsth vh, ntrees aModelTree)
{
  appendLoc(CKM->getInterfaceSubHierarchy(vh), aModelTree);
}

/**
   It adds to the Signal abstract tree representation of the model (aModelTree) associated with a clock (vh)
   the declaration of the interface (parameters, inputs, outputs) of the model.    
   It calls:
   - mktreeparameters() for the parameters,
   - mktreeinput() for the inputs,
   - mktreeoutput() for the outputs.
*/
_VIRTUAL_ void DCG2SignalTree::mtInterfaceTree(ptinsth aClock, ntrees aModelTree)
{
  if (!flatMode)
    mktreeparameters(aClock, aModelTree);
  if (CKM->satisfiesFictitiousRoot(aClock))
    {
      if (flatMode)
	mktreeparameters(aClock, aModelTree);
      mktreeinput(aClock, aModelTree);
      mktreeoutput(aClock, aModelTree);
    }
  else if (genInterface)
    {
      mktreeinput(aClock, aModelTree);
      mktreeoutput(aClock, aModelTree);
    }
}

/**
   @return the Signal Abstract tree representation of the declaration of a type or constant (elem).
   - for a constant, it calls mkConstantDefinition().
   - for a type, it calls  TypeTypeUser2ntrees() with the type referenced in elem.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkDeclCstType(ptexpcal elem)
{
  TypeType TT;
  ntrees vt;
  noperators vop;

  vt = STree_getNilTree();
  vop = EM->getOperator(elem);
  if (vop == optype)
    {
      TT = EM->getExpType(elem);
      if (!TypeTestMetaClass(TT))
	if (TypeTypeUser2ntrees(TT, &vt))
	  return LL_make_ltype_list(vt);
      return NILTREE;
    }
  // Constant
  return mkConstantDefinition(elem);
}

/**
   @return "true" if it exists at least one undeclared model referenced in some external references, "false" otherwise.
   @param lProcCalls (in) a list of procedure calls
   @param lfuncs (in) a list of function calls (currently ignored)
   @param ldecl (out) a Signal Abstract tree representation of a list of model declarations. 

   For each element of lProcCalls, it calls mkDeclarationExternProc(), if the back annotation to a model is not available.
   The objects marked by DONTDECLARE (compiler attributes) are filtered.
*/
_VIRTUAL_ boolean DCG2SignalTree::mkUndeclaredDeclarationExternProcedures(ptlINode lProcCalls, ptlinstances lfuncs, ntrees * ldecl)
{
  ptINode vnode;
  ntrees x, t, vlist;

  SET_NILTREE(vlist);
  while (ptlINode_get_elem(&lProcCalls, &vnode))
    {
      x = BACK_ANNOT_GET_ABSTRACT_TREE(NM->getBackAnnotation(vnode));
      if (IS_NILTREE(x))
	if (!NM->isMarked(vnode, DONTDECLARE))
	  {
	    t = mkDeclarationExternProc(vnode, opprocess, pk_true);
	    if (IS_NILTREE(vlist))
	      vlist = LL_make_ldecl_list(t);
	    else
	      vlist = ntreeList_appendLast(vlist, t);
	  }
    }
  if (IS_NILTREE(vlist))
    return pk_false;
  *ldecl = vlist;
  return pk_true;
}


void DCG2SignalTree::addClockScheduledGraph(ptexpcal vsig)
{
  ntrees vt1 = mac_STree_make_initSynchro(getClockExpr0(EM->getClock(vsig)));
  ntreeList_appendLast(vt1, ptexpcal2ntrees(vsig));
  STree_addExprToComposition_Post(&vtreeClockScheduledGraph, vt1);
}
