/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_

ptexpcal DCG2SignalTree::getExpressionToGenerate (ptexpcal exp)
{
  ptexpcal exp1;
  //  return exp;

  if (exp == NILEXPCAL)
    return exp;
  exp = EM->getCloneClassHead (exp);
  if (EM->getAttribute (exp, OPTIMIZED_ATTRIBUTE, (attributeValue *) & exp1))
    return getExpressionToGenerate (exp1);
  // if (EM->getAttribute (exp, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *) & exp1)) return getExpressionToGenerate (exp1);
  exp1 = EM->getAggregation (exp);  
  if (exp1 != exp) return getExpressionToGenerate (exp1);
  if ( isOptimizedSignal(exp, &exp1)) return exp1;
  return exp;
}

/**
  @return  "true" if the signal cond is an optimized one, "false" otherwise.
  When it is optimized, it is replaced by !vexp.
*/
boolean
DCG2SignalTree::isOptimizedSignal (ptexpcal vsig,  ptexpcal *vexp)
{
  if (!EM->IsSignal (vsig)) return pk_false;
  return isOptimizedNode (EM->getNode (vsig), vexp);
}

/**
   @return "true" if a node (aNode) is marked by substituteGenCodeMark, "false" otherwise.
   When it is optimized, it is replaced by !vexp.
*/
boolean
DCG2SignalTree::isOptimizedNode (ptINode aNode, ptexpcal *vexp)
{
  ptexpcal expd;
  if (!NM->isMarked (aNode, substituteGenCodeMark)) return pk_false;
  NM->get_rightOperand (aNode, &expd);
  *vexp = EM->getBasicExpr (expd);
  return pk_true;
}




/**
   @return the Abstract tree Signal representation of an expression (vexp).
   A new signal is created for the expression when kernelExprMode is set.
*/
    _VIRTUAL_ ntrees DCG2SignalTree::ptexpcal2ntrees_operand(ptexpcal vexp)
{
  ntrees vt, vtype;
  ptident id;
  noperators op;
  ptinsth h;

  vt = ptexpcal2ntrees(vexp);
  if (!KernelExprMode)
    return vt;

  if (IS_NILTREE(vt))
    return vt;
  op = noper(vt);
  if ((op == opident) || (LL_is_atomOp(op)))
    return vt;

  id = mkNewIdentKernelMode();
  vt = LL_make_defsig_tree(STree_scopy(id), vt);
  STree_addExpressionOfProcessToaModel(G_TreeGraph, vt);
  // type de vexp.
  vtype = typeExp2ntrees(vexp);
  STree_addLocalDeclaration(G_TreeGraph, STree_mkSignalDeclaration(STree_scopy(id), vtype, NILTREE));
  if (EM->getExprClock(vexp, &h))
    {
      vt = STree_mkSynchroIni(mkhonametree(h));
      vt = STree_addSynchro(vt, STree_scopy(id));
      STree_addExpressionOfProcessToaModel(G_TreeGraph, vt);
    }
  return id;
}

/**
   @return the Abstract tree Signal representation of an expression (vexp).
*/
_VIRTUAL_ ntrees DCG2SignalTree::ptexpcal2ntrees(ptexpcal vexp)
{
  ptident id;
  noperators op, operateur;
  ptexpcal vexpcal1, vexpcal2, vexpcal3, nexp;
  ptinsth hexp, h;
  ntrees vt, t1, t2;
  TypeType TT;

  if ( isScheduledGraph() )
    vexp = getExpressionToGenerate(vexp);
  if (vexp == NILEXPCAL)
    return NILTREE;
  if (EM->getIndirection(vexp, &nexp))
    vexp = nexp;

  switch ((int)(EM->getOperatorClass(vexp)))
    {
    case variabl:
      vt = STree_scopy(EM->getName(vexp));
      setInfosTracabilty(vexp, vt);
      return vt;

    case constant:
      operateur = EM->getOperator(vexp);
      if (operateur == opcstenum)
	operateur = opnom;
      return natomstring(operateur, EM->getConstant(vexp));

    case oper1:
    case oper1tp:
      operateur = EM->getOperator(vexp);
      vexpcal1 = EM->getUnaryArg(vexp);
      t1 = ptexpcal2ntrees_operand(vexpcal1);
      return nmakeunary(operateur, t1);

    case oper2:
    case oper2tp:
      operateur = EM->getOperator(vexp);
      if (LL_is_listOp(operateur))
	{
	  vt = nmakemptylist(operateur);
	  while (vexp != NILEXPCAL)
	    vt = ntreeList_appendLast(vt, ptexpcal2ntrees_operand(EM->getNaryArgs(&vexp)));
	  return vt;
	}
      if (operateur == opvenum)
	if (EM->IsFreeType(vexp))
	  t1 = STree_mkAtomNilTree();
	else
	  t1 = typeExp2ntrees(vexp);
      else
	t1 = ptexpcal2ntrees_operand(EM->getBinaryArgstAt(vexp, 1));
      vexpcal2 = EM->getBinaryArgstAt(vexp, 2);
      if (operateur == opelem)
	t2 = notKernptexpcal2ntrees(vexpcal2);
      else
	{
	  t2 = ptexpcal2ntrees_operand(vexpcal2);
	  if ((operateur == opmele) && EM->IsEvent(vexp))
	    operateur = opbsup;
	}
      return nmakebinary(operateur, t1, t2);

    case oper3:
    case oper3tp:
      operateur = EM->getOperator(vexp);
      vexpcal1 = EM->getTernaryArgstAt(vexp, 1);
      vexpcal2 = EM->getTernaryArgstAt(vexp, 2);
      vexpcal3 = EM->getTernaryArgstAt(vexp, 3);

      if (operateur == opret)
	return mkTreeDelay(EM->getTernaryArgstAt(vexpcal1, 1), vexpcal2, vexpcal3, EM->getTernaryArgstAt(vexpcal1, 2), NILEXPCAL);
      return nmaketernary(operateur, ptexpcal2ntrees_operand(vexpcal1), ptexpcal2ntrees_operand(vexpcal2), ptexpcal2ntrees_operand(vexpcal3));

    case opereh:
      operateur = EM->getOperator(vexp);
      vexpcal1 = EM->getWhenOrCellArg(vexp);
      h = EM->getContributionClock(vexp);
      // cell
      if (operateur == opmemo)
	return nmaketernary(opmemo, ptexpcal2ntrees_operand(vexpcal1), getClockExprInWhenCell(h), NILTREE);

      if (EM->isMarked(vexp, genScheduledGraphMark ))
	return ptexpcal2ntrees_operand(vexpcal1);

      // For E when H and E ^= H, E is returned.
	if (simplyfyExpr)
	if (EM->getSynchronousExprHorl(vexpcal1, &hexp))
	  if (CKM->eqClock(h, hexp))
	    return ptexpcal2ntrees_operand(vexpcal1);
      if (isEventClockedFormat(graph_type))
	{
	  op = EM->getOperator(vexpcal1);
	  if (op == optrue)
	    return getClockExpr0(h);
	}
      return nmakebinary(operateur, ptexpcal2ntrees_operand(vexpcal1), getClockExprInWhenCell(h));

    case opertyp:
      TT = EM->getExpType(vexp);
      return TypeType2ntrees(TT);

    case call:
      EM->getFunctionCall(vexp, &id, &vexpcal2, &vexpcal1);
      if (existPredefinedPragma(EM->getPragmas(vexp), PRAGMA_CONVERSION))
	{
	  vt = STree_mkNullaryTree(identType2SignalType(ident2repstr(id)));	// Type predefini
	  return LL_make_conv_tree(vt, mkListIdentCall(vexpcal1, 2));
	}
      return mkRightPartCallTree(id, vexpcal2, vexpcal1);

    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
  return NILTREE;
}

/** 
    @return the Signal abstract tree of  the equation <em> vLeftPart := vtreeopd </em>.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkdefsigtree(ptexpcal vLeftPart, ntrees vtreeopd)
{
  boolean b;
  ntrees vt;

  b = TraceabilityMode;
  vt = LL_make_defsig_tree(ptexpcal2ntrees(vLeftPart), vtreeopd);
  TraceabilityMode = b;
  return vt;
}

/**
   @return the Signal abstract tree representation of a list of expression (ls) used in a call.
   The following code is used : 
   - inOutParam=1, ls is a list of parameters, 
   - inOutParam=2, ls is a list of inputs 
   - inOutParam=3, ls is a list of outputs.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkListIdentCall(ptexpcal ls, int inOutParam)
{
  ntrees outTree;
  noperators vop;

  outTree = ptexpcal2ntrees(ls);
  // -- should be in the Tree part.
  if (IS_NILTREE(outTree))
    return outTree;
  if (inOutParam == 1)
    vop = oplepara;
  else if (inOutParam == 2)
    vop = oplesign;
  else if (inOutParam == 3)
    {
      if (ntreearity(outTree) == 1)
	return nchild(1, outTree);
      vop = opleflot;
    }
  nsetop(outTree, vop);
  return outTree;
}

/**
   @return the Signal abstract tree representation of a call to a process with \em nom as name,  \em lparam as parameters, and  \em lent as inputs.\n
   Signal syntax: "<callName> {<params>} (<inputs>)"
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkRightPartCallTree(ptident callName, ptexpcal lparam, ptexpcal lent)
{
  ntrees vt, lin, lpar;

  vt = STree_scopy(callName);
  lpar = mkListIdentCall(lparam, 1);
  if (ISNOT_NILTREE(lpar))
    vt = LL_make_inst_tree(vt, lpar);
  lin = mkListIdentCall(lent, 2);
  return LL_make_appe_tree(vt, lin);
}

/**  
     @return the Signal abstract tree representation of a call defined by "<outputs> := <callName> {<params>} (<inputs>)"
     or  "<callName> {<params>} (<inputs>)", when outputs is an empty list.
*/
ntrees DCG2SignalTree::mkCallTree(ptident callName, ptexpcal params, ptexpcal inputs, ptexpcal outputs)
{
  ntrees vt, lout;

  vt = mkRightPartCallTree(callName, params, inputs);
  lout = mkListIdentCall(outputs, 3);
  if (IS_NILTREE(lout))
    return vt;
  return LL_make_defsig_tree(lout, vt);
}

/** 
    @return the Signal abstract tree representation of the expression "<callName> {<params>}".
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkInstTree(ptident callName, ptexpcal params)
{
  ntrees vt, lpar;

  vt = STree_scopy(callName);
  lpar = mkListIdentCall(params, 1);
  if (IS_NILTREE(lpar))
    lpar = STree_mkAtomNilTree();
  return LL_make_inst_tree(vt, lpar);
}

/** 
    @return the Abstract tree Signal representation of a the type of an expression (vexp).
*/
ntrees DCG2SignalTree::typeExp2ntrees(ptexpcal vexp)
{
  return TypeType2ntrees(EM->getType(vexp));
}

/** 
    @return the Abstract tree Signal representation of a list, characterized by its operator (opl), of expressions (lsig) catenating (post) with \em vtree.
    Each element is translated by ptexpcal2ntrees().
*/
ntrees DCG2SignalTree::mklnamessig(ptlinstances lsig, ntrees vtree)
{
  ptexpcal vsig;
  if (!ptlinstances_get_next(&lsig, &vsig))
    return vtree;
  if (IS_NILTREE(vtree))
    vtree = LL_make_lident_emptyList();
  vtree = ntreeList_appendLast(vtree, ptexpcal2ntrees(vsig));
  while (ptlinstances_get_next(&lsig, &vsig))
    vtree = ntreeList_appendLast(vtree, ptexpcal2ntrees(vsig));
  return vtree;
}

/**
   @return The translation of the operand expression (vexp) without decomposition.
 */
_VIRTUAL_ ntrees DCG2SignalTree::notKernptexpcal2ntrees(ptexpcal vexp)
{
  ntrees vtree;
  boolean b;

  b = KernelExprMode;
  KernelExprMode = pk_false;
  vtree = ptexpcal2ntrees(vexp);
  KernelExprMode = b;
  return vtree;
}

/**
   @return the Signal abstract tree representation of a delay expression:
   - <em> delayedSignal \b when contextClock \b $ delayExpr \b windows winExpr \b init initExpr </em>
   when \em delayedSignal is a constant expression and
   \em contextClock is not undefined, 
   - <em> delayedSignal \b $ delayExpr  \b windows winExpr \b init initExpr </em> othewise.
*/
_VIRTUAL_ ntrees DCG2SignalTree::mkTreeDelay(ptexpcal delayedSignal, ptexpcal delayExpr, ptexpcal winExpr, ptexpcal initExpr, ptexpcal vdelay)
{
  ntrees vtreeret, vtreefen, vtreesigret, vtreeInit;

  vtreeInit = notKernptexpcal2ntrees(initExpr);
  vtreeret = notKernptexpcal2ntrees(delayExpr);
  vtreefen = notKernptexpcal2ntrees(winExpr);

  vtreesigret = ptexpcal2ntrees(delayedSignal);

  if (EM->IsConstantExpr(delayedSignal) && (vdelay != NILEXPCAL))
    vtreesigret = LL_make_extr_tree(vtreesigret, LL_make_evnt_tree(ptexpcal2ntrees(vdelay)));

  if (IS_NILTREE(vtreefen))
    return LL_make_ret_tree(vtreesigret, vtreeret, vtreeInit);	/* delay without window */
  if (IS_NILTREE(vtreeret))
    return LL_make_fen_tree(vtreesigret, vtreefen, vtreeInit);	/* window without delay */
  return LL_make_fen_tree(LL_make_ret_tree(vtreesigret, vtreeret, NILTREE), vtreefen, vtreeInit);	/* delay + window */
}


void DCG2SignalTree::markDefaultLastExpression(ptexpcal vexp, boolean bmark)
{
  ptexpcal varg2 = EM->getBinaryArgstAt(vexp, 2);

  if ( EM->getOperator(varg2) == opmele)
    markDefaultLastExpression(varg2, bmark);
  else
    {
      if (bmark) 
	EM->addMark(varg2, genScheduledGraphMark); 
      else 
	EM->removeMark(varg2, genScheduledGraphMark); 
    }
}
