/*
 * Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic. Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2
 * License - 2010.
 * 
 * Contact: polychronycontact@listes.irisa.fr
 * 
 * This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).
 * 
 * Polychrony Signal Toolbox is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 * 
 * Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2 for more details.
 * 
 * You should have received a copy of the GNU General Public License Version 2 along with Polychrony Signal Toolbox; see the file COPYING. If not,
 * consult the url http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA. 
 */

_PUBLIC_
/** 
    Constructor. 
*/
DCG::DCG ()
{
  initialyze (new NameManager ());
  setnum (pK_GraphManager->addGraph (this));
}

DCG::DCG (ptNameManager nm)
{
  initialyze (nm);
  setnum (pK_GraphManager->addGraph (this));
}

ptDCG
DCG::getDCG ()
{
  return this;
}

/**
   Destructor.
*/
DCG::~DCG ()
{
  ptlINode ln;
  ptINode n;
  ptinsth hinit;

  hinit = aClockManager->getInitClock ();
  if (aNodeManager->getAllNodes (&ln))
    while (ptlINode_get_elem (&ln, &n))
      aNodeManager->removeDependence (hinit, n);

  ptlINode_free (list_extern_proc);
  ptlINode_free (unusedNodes);
  ptlinsth_free (unusedClocks);
  ptlmemnode_free (delays_tree);

  ptlinstances_free (list_extern_funcs);
  ptlinstances_free (list_indices);

  ptlinstances_free (lSigNullClock);

  ptlINode_free (lNodesAtInitClock);
  ptlinsth_free (lConstraintsNullClock);

  ptlPointer_free (cycles);
  ptlPointer_free (PseudoCycles);

  ptlinstances_free (lssiglocgc);
  ptlinstances_free (lClocksNotToBeInitiliazed);
  ptlINode_free (lNodesHinit);

  pK_GraphManager->removeGraph (this);	// removing from the database.
}

/**
   @return the current format GraphLevelType of the graph.
*/
GraphLevelType
DCG::getFormatLevel ()
{
  return vgraph_level;
}


/**
   It sets the current format of the graph to currentlevel value.
*/
void
DCG::setFormatLevel (GraphLevelType currentlevel)
{
  /*   if (vgraph_level != currentlevel)
    {
      ptTraceabilityManager tm; 
      tm = getTraceabilityManager ();
      tm->beginFunctionality();
      tm->printmodifiedSinceLastMarkedFunction();
       tm->setCurrentAsReference();
    }
  */
  vgraph_level = currentlevel;
}

/**
   @return "true" if the graph is a graph of external process, "false" otherwise.
*/
boolean
DCG::isGraphOfExternalProcess ()
{
  return externalNode;
}

/**
   @return "true" if the graph is an empty one, "false" otherwise.
*/
boolean
DCG::isEmptyGraph ()
{
  ptlDCG lg;
  if (externalNode)
    return pk_false;
  if (getModuleGraphs (&lg))
    return (lg == NILLDCG);
  return aClockManager->isAnEmptySystem ();
}

/**
   @return the list of the detected cycles computed by hasStaticCycleOneLevel().
*/
ptlPointer
DCG::getCycles ()
{
  return cycles;
}

/**
   It sets to lc the list of the detected cycles of the graph.
*/
void
DCG::setCycles (ptlPointer lc)
{
  cycles = lc;
}

/**
   @return the list of the detected "false cycles" computed by hasStaticCycleOneLevel().
*/
ptlPointer
DCG::getPseudoCycles ()
{
  return PseudoCycles;
}

/**
   It sets to lfc the list of the detected "false cycles" of the graph.
*/
void
DCG::setPseudoCycles (ptlPointer lfc)
{
  PseudoCycles = lfc;
}

/**
   Storing informations for graph of an external process.
   @param SIGNALDeclaration (in) the Signal tree declaration builts by the parser.
   @param vlang (in) implementation language assigned to the external process.
*/
void
DCG::setGraphOfExternalProcess (ntrees SIGNALDeclaration, ntrees vlang)
{
  externalNode = pk_true;
  SIGNAL_Tree = SIGNALDeclaration;
  setImplementationLanguage (vlang);
}

/**
   Manage only for external graph.
   @return the Signal tree declaration (original tree builts by the parser)
*/
ntrees
DCG::getSIGNALTreeDeclaration ()
{
  return SIGNAL_Tree;
}

/**
   It sets the implementation language (codeLanguage) for a graph.
   (used for the graph of an external process)
*/
void
DCG::setImplementationLanguage (ntrees codeLanguage)
{
  addAttribute (EXTERN_LANGUAGE_IMPL, (attributeValue) codeLanguage);
}

/**
   @return the implementation language assigned to the graph.
*/
ntrees
DCG::getImplementationLanguage ()
{
  ntrees cl;

  if (!getAttribute (EXTERN_LANGUAGE_IMPL, (attributeValue *) & cl))
    return NILTREE;
  return cl;
}

ptinterface
DCG::getInterface ()
{
  return GET_ROOT_INTERFACE ();
}

/**
   @return the "fictitious root" of the clock hierarchy forest.
   This root can be used for a "left to right topdown" traversal of the hierarchy.
*/
ptinsth
DCG::getRootClock ()
{
  return aClockManager->getFictitiousRoot ();
}

/**
   It sets the list (lc) of the constants of the program.
*/
void
DCG::setConstants (ptlinstances lc)
{
  ptinterface_set_constants (GET_ROOT_INTERFACE (), lc);
}

/** 
    @return the list of the constants of the graph.
*/
ptlinstances
DCG::getConstants ()
{
  return ptinterface_get_constants (GET_ROOT_INTERFACE ());
}

/**
   It adds e in the set of constants.
*/
void
DCG::addConstant (ptexpcal e)
{
  ptlinstances l;
  l = getConstants ();
  if (!ptlinstances_isin_exp (l, e))
    setConstants (ptlinstances_push_end (l, e));
}

/**
   It removes e from the set of constants.
*/
void
DCG::removeConstant (ptexpcal e)
{
  setConstants (ptlinstances_remove_exp (getConstants (), e));
}

/** 
    @return the list of the external procedure of the graph.
*/
ptlINode
DCG::getListExternalProcedure (ptlinstances * lfuncs)
{
  *lfuncs = list_extern_funcs;
  return list_extern_proc;
}

/**
   Setting external functions and procedures.
   @param lprocs (in): the list of the external procedures;
   @param lfuncs (in): the list of the external functions.
*/
void
DCG::setListExternalProc (ptlINode lprocs, ptlinstances lfuncs)
{
  list_extern_proc = lprocs;
  list_extern_funcs = lfuncs;
}

/** @return the list of delays tree of the graph.  */
ptlmemnode
DCG::getDelaysTree ()
{
  return delays_tree;
}

/** It defines the list of delays tree of the graph by dt.  */
void
DCG::setDelaysTree (ptlmemnode dt)
{
  delays_tree = dt;
}

/** @return the list of the index identifiers of the graph.*/
ptlinstances
DCG::getListIndexes ()
{
  return list_indices;
}

/** It inserts the index identifier vexp in the list of indexes of the graph.
 */
void
DCG::insertIndex (ptexpcal vexp)
{
  ptlinstances liste, dummy;

  liste = list_indices;
  if (ptlinstances_isin_insert_exp (&liste, &dummy, vexp))
    list_indices = liste;
}

/** 
    It stores the unused nodes  and the unused clocks of the graph.
    These objects are not removed.
*/
void
DCG::setUnused (ptlINode unNodes, ptlinsth unClocks)
{
  unusedNodes = unNodes;
  unusedClocks = unClocks;
}

/** 
    @return the unused nodes and the unused clocks (lh) unused nodes of the graph.
*/
ptlINode
DCG::getUnused (ptlinsth * lh)
{
  *lh = unusedClocks;
  return unusedNodes;
}

/** @return the name of the node of the graph.
    This name is FOO_XX where FOO is the name of the graph and XX is provided by some pragmas or the internal number associated with the node.
*/
ptident
DCG::getName (ptINode vnode)
{
  ptident id;
  ptlpragmas lp;
  ptpragma aPrag;
  ptrepstr EnonceValue;

  require_node_manager (NM);
  lp = NM->getPragmas (vnode);
  if (getPragmaKeyword (lp, PRAGMA_RUN_ON, &aPrag) || getPragmaKeyword (lp, PRAGMA_SELECTED_OTHERS_PARTITION, &aPrag))
    id = createSuffIdent (getName (), strFirstStmtsValue (aPrag));
  else
    id = makeIdent_ident_int (getName (), NM->get_num (vnode));
  return id;
}

/** @return the name of the graph, previously set by setName(). */
ptident
DCG::getLevelName ()
{
  switch (getFormatLevel ())
    {
    case DCG_BASIC:
      return makeIdent_ident_string (vname, "BASIC");
    case DCG_POLY:
      return makeIdent_ident_string (vname, "POLY");
    case DCG_ENDO:
      return makeIdent_ident_string (vname, "ENDO");
    case DCG_BOOL:
      return makeIdent_ident_string (vname, "BOOL");
    case DCG_FLAT:
      return makeIdent_ident_string (vname, "FLAT");
    case DCG_SCH:
      return makeIdent_ident_string (vname, "SCH");
    default:
      return vname;
    }
}

boolean
DCG::isModule ()
{
  return (getKind () == opmodu);
}

/**
   @return "true" if the current graph is a graph of a module, "false" otherwise.
   @param lg (out): the list of the internal graphs of the module.
*/
boolean
DCG::getModuleGraphs (ptlDCG * lg)
{
  (*lg) = getLocalGraphs ();
  if ((*lg) != NILLDCG)
    return pk_true;
  return isModule ();
}

/**
   @return "true" if it exists some signals with null clock, "false" otherwise.
   lsig (out) : the list of signals at null clock.
*/
boolean
DCG::getNullClockSignal (ptlinstances * lsig)
{
  (*lsig) = lSigNullClock;
  return (lSigNullClock != NILLINSTANCES);
}

/**
   @return "true" if it exists some formula constraints with null clock, "false" otherwise.

   lh (out) : the list of clocks (set of formula) not proved equivalent to null clock.
   \see addNullClockInfo, storeNullClockInfo
*/
boolean
DCG::getNullClockConstraints (ptlinsth * lh)
{
  (*lh) = lConstraintsNullClock;
  return (lConstraintsNullClock != NILLINSTH);
}

/**
   It sets the kind of graph tag :
   @param op (in) : Signal operator
   - opmodu for a graph of a module,
   - opprocess for a graph of a  process,
   - opfunction for a graph of a function,
   - opnode for a graph of a node.
*/
void
DCG::setKind (noperators op)
{
  kind = op;
}

/**
   @return the kind of graph tag.
   \see setKind for the values.
*/
noperators
DCG::getKind ()
{
  return kind;
}

/**
   It defines the list of inputs of the graph by l. 
*/
void
DCG::setInputs (ptlinstances l)
{
  ptexpcal vsig;
  ptinterface vinterface;

  vinterface = GET_ROOT_INTERFACE ();
  getNodeManager ()->inSplipListToInterface (vinterface, l);
  while (ptlinstances_get_next (&l, &vsig))
    getExpressionManager ()->belongsInFlow (vsig);
}

/**
   It defines the list of outputs of the graph by l. 
*/
void
DCG::setOutputs (ptlinstances l)
{
  ptexpcal vsig;
  ptinterface vinterface;

  vinterface = GET_ROOT_INTERFACE ();
  getNodeManager ()->outSplipListToInterface (vinterface, l);
  while (ptlinstances_get_next (&l, &vsig))
    getExpressionManager ()->belongsOutFlow (vsig);
}

/**
   It defines the list of parameters of the graph by l. 
*/
void
DCG::setParameters (ptlinstances l)
{
  ptinterface_set_parameters (GET_ROOT_INTERFACE (), l);
}

/** @return the list of the input signals of the graph.*/
ptlinstances
DCG::getInputs ()
{
  return ptinterface_get_inputSignals (GET_ROOT_INTERFACE ());
}

/** @return the list of the input clocks of the graph.*/
ptlinsth
DCG::getImportedClocks ()
{
  return ptinterface_get_inputClocks (GET_ROOT_INTERFACE ());
}


/** @return the list of the output signals of the graph.*/
ptlinstances
DCG::getOutputs ()
{
  return ptinterface_get_outputSignals (GET_ROOT_INTERFACE ());
}

/** @return the list of the input-output signals of the graph (shared signals).*/
ptlinstances
DCG::getInputsOutputs ()
{
  return ptinterface_get_inOutSignals (GET_ROOT_INTERFACE ());
}

/** @return the list of the output clocks of the graph.*/
ptlinsth
DCG::getExportedClocks ()
{
  return ptinterface_get_outputClocks (GET_ROOT_INTERFACE ());
}

/** @return the list of the parameters of the graph.*/
ptlinstances
DCG::getParameters ()
{
  return ptinterface_get_parameters (GET_ROOT_INTERFACE ());
}

/** It adds e in the set of inputs.*/
void
DCG::addInput (ptexpcal e)
{
  ptlinstances l;
  l = getInputs ();
  if (!ptlinstances_isin_exp (l, e))
    setInputs (ptlinstances_insert (l, e));
}

/** It adds e in the set of outputs.*/
void
DCG::addOutput (ptexpcal e)
{
  ptlinstances l;
  l = getOutputs ();
  if (!ptlinstances_isin_exp (l, e))
    setOutputs (ptlinstances_push_end (l, e));
}

/** It adds e in the set of parameters.*/
void
DCG::addParameter (ptexpcal e)
{
  ptlinstances l;
  l = getParameters ();
  if (!ptlinstances_isin_exp (l, e))
    setParameters (ptlinstances_push_end (l, e));
}





/**
   @return "true" if it exists some signals with null clock, "false" otherwise.
   lsig (out) : the list of signals at null clock.
*/
boolean
DCG::getNodesAtInitClock (ptlINode * ln)
{
  (*ln) = lNodesAtInitClock;
  return (lNodesAtInitClock != NILLINODE);
}

/**
   @return the list of the explicit dependences of the graph.
   The returned list MUST NOT be deleted by the user.
*/
ptldependence
DCG::getDependences ()
{
  return getNodeManager ()->getDependences ();
}

/**
   @return the list of the explicit dependences of the graph.
   The returned list may be deleted by the user.
*/
ptldependence
DCG::getExplicitDependences ()
{
  return getTypedDependences (explicitDep, pk_false);
}

/**
   @return the list of the dependences induced by the reinforcing of the graph.
   The returned list may be deleted by the user.
*/
ptldependence
DCG::getReinforcedDependences ()
{
  return getTypedDependences (reinforcementDep, pk_true);
}



/**
   It assigns the default initial value to the non initialized signals needing such value.
*/
void
DCG::setInitialValues ()
{
  verifyinitialisations_delays ();
  verifyinitialisations_cells ();
  verifyinitialisations_statevars ();
}


/** id becomes the name of the object. */
void
DCG::setName (ptident id)
{
  Common_DCG_Automaton::setName (id);
}

/** @return the name of the graph, previously set by setName(). */
ptident
DCG::getName ()
{
  return Common_DCG_Automaton::getName ();
}

/** @return the name of the graph when it is used as a state of an automaton. */
ptident
DCG::getLabelStateName ()
{
  ptident id = Common_DCG_Automaton::getName ();
  return makeIdent_string_ident ("S_", id);
}

/** 
    @return the clock system manager of the graph.
*/
ptClockManager
DCG::getClockManager ()
{
  return Common_DCG_Automaton::getClockManager ();
}

/** @return the node manager of the graph.*/
ptNodeManager
DCG::getNodeManager ()
{
  return Common_DCG_Automaton::getNodeManager ();
}

/** @return the expresion manager of the graph.*/
ptExprManager
DCG::getExpressionManager ()
{
  return Common_DCG_Automaton::getExpressionManager ();
}

/** @return the name manager of the graph.*/
ptNameManager
DCG::getNameManager ()
{
  return aNameManager;
}

/** @return the traceability manager of the graph.*/
ptTraceabilityManager
DCG::getTraceabilityManager ()
{
  return Common_DCG_Automaton::getTraceabilityManager ();
}

/** @return the set of state variables of the program.*/
ptlinstances
DCG::getStatesVar ()
{
  return Common_DCG_Automaton::getStatesVar ();
}

/** l (in): set of state variables. */
void
DCG::setStatesVar (ptlinstances l)
{
  Common_DCG_Automaton::setStatesVar (l);
}

/** It adds vsig in the set of state variables.*/
void
DCG::addStatesVar (ptexpcal vsig)
{
  Common_DCG_Automaton::addStatesVar (vsig);
}

/** 
    It assigns an automaton (vauto) to the graph.
*/
void
DCG::setAutomaton (ptAutomaton vauto)
{
  vautomaton = vauto;
}

/**
   @return true if an automaton (vautomat) is assigned to the graph, false otherwise.
*/
boolean
DCG::getAutomaton (ptAutomaton * vauto)
{
  if (vautomaton == NILAUTOMATON)
    return pk_false;
  *vauto = vautomaton;
  return pk_true;
}

/**
   It defines the list of local signals (lsigloc) of the graph.
   Used for module, and for a graph which is not a graph of a module, it is used by code generators.
*/
void
DCG::setLocalSignals (ptlinstances lsigloc)
{
  lssiglocgc = lsigloc;
}

/** 
    @return the list of all the local signals of the graph.
*/
ptlinstances
DCG::getLocalSignals ()
{
  return lssiglocgc;
}

/**
   Setting ordering informations (sbDCplus format).
   These informations are available after the \ref totallyOrder method. 
   @param nbsets (in) : the number of elements of the Table parameter;
   @param Table (in) : one entry by cluster and Table[0] contains a sequential scheduller.
   @param lnodeshinit (in) : set of nodes defining constant signals. They can be generated as initializations.
   @param lclocksnottobeinitiliazed (in) : set of clocks build on state variables. They must not to be intialized each step.
*/
void
DCG::setDCGSchInfos (int nbsets, ptelemTableGenCode * Table, ptlINode lnodeshinit, ptlinstances lclocksnottobeinitiliazed)
{
  SubGraphTable = Table;
  lNodesHinit = lnodeshinit;
  nbSets = nbsets;
  lClocksNotToBeInitiliazed = lclocksnottobeinitiliazed;
}

/** 
    These informations are available after the \ref totallyOrder method. 
    @param nbsets (out) : the number of elements of the Table parameter;
    @param Table (out) : one entry by cluster and Table[0] contains a sequential scheduller.
    @param lnodeshinit (out) : set of nodes defining constant signals. They can be generated as initializations.
    @param lclocksnottobeinitiliazed (out) : set of clocks build on state variables. They must not to be intialized each step.
*/
void
DCG::getDCGSchInfos (int *nbsets, ptelemTableGenCode ** Table, ptlINode * lnodeshinit, ptlinstances * lclocksnottobeinitiliazed)
{
  *Table = SubGraphTable;
  *lnodeshinit = lNodesHinit;
  *nbsets = nbSets;
  *lclocksnottobeinitiliazed = lClocksNotToBeInitiliazed;
}



/**
   @return the initial value assigned with a signal (s).
   It creates n default initial values and assigns them to a signal (s)
*/
ptexpcal
DCG::setInitialValueDefault (ptexpcal s, long int n)
{
  ptrepstr vstr;
  ptexpcal vinit, index;

  require_managers (EM, NM, CKM);
  if (n != 0)
    {
      vstr = ident2repstr (EM->getName (s));
      if (EM->mkDefaultInit (s, n, &index, &vinit))
	{
	  pK_messenger->messagestr (defaultInit1, vstr);
	  if (!EM->isNILEXPCAL (index))
	    insertIndex (index);
	  EM->setInitialValue (s, vinit);
	  return vinit;
	}
      else
	pK_messenger->messagestr (defaultInit2, vstr);
    }
  return NILEXPCAL;
}

/**
   It verifies the cells of the current graph by adding "default" initial values (defined in the Signal reference manual)
   to the uninitialized ones.
*/
void
DCG::verifyinitialisations_cells ()
{
  ptlINode lNodes;
  ptINode node;
  ptexpcal expg, sr;

  require_managers (EM, NM, CKM);
  if (NM->getEquations (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      if (NM->is_cellNode (node))
	{
	  expg = NM->get_leftOperand (node);
	  sr = EM->getInitialValue (expg);
	  if (!EM->isNILEXPCAL (sr))
	    EM->TypePropagate (EM->getType (expg), sr);
	  else
	    setInitialValueDefault (expg, 1);
	}
}

/**
   It verifies the state variables of the current graph by adding "default" initial values (defined in the Signal reference manual)
   to the uninitialized ones.
*/
void
DCG::verifyinitialisations_statevars ()
{
  ptexpcal r, sr;
  ptlinstances ls;

  require_managers (EM, NM, CKM);
  ls = getStatesVar ();
  while (ptlinstances_get_next (&ls, &r))
    {
      sr = EM->getInitialValue (r);
      if (EM->isNILEXPCAL (sr))
	setInitialValueDefault (r, 1);
    }
}

/**
   It verifies the delay node of the current graph by adding "default" initial values (defined in the Signal reference manual)
   to the uninitialized ones.
*/
void
DCG::verifyinitialisations_delays ()
{
  ptlINode lNodes;
  ptINode node;
  ptexpcal r, sr, vret, vfen, x, vinit;
  boolean b, b0;
  compilerInteger nd, nw;
  ptinsth h;

  require_managers (EM, NM, CKM);
  if (NM->getDelays (&lNodes))
    {
      b0 = pk_false;
      while (ptlINode_get_elem (&lNodes, &node))
	{
	  NM->get_delay_arg (node, &r, &sr, &vret, &vfen, &vinit);

	  if (EM->IsEvent (r))
	    {
	      // a event delay: type error ???. A warning is sent to the user. It is replaced by pk_true.
	      h = EM->getClock (r);
	      NM->modifiedDefinition (r, EM->getTrue (), h, h);
	      b0 = pk_true;
	    }
	  else
	    {
	      b = pk_true;
	      if (EM->isNILEXPCAL (vret))
		{
		  nd = 0;
		}
	      else
		b = EM->getIntegerValue (vret, &nd);
	      if (EM->isNILEXPCAL (vfen))
		{
		  nw = 0;
		}
	      else
		b = b && EM->getIntegerValue (vfen, &nw);
	      if (b && (nw != 0))
		nw = nw - 1;
	      if (EM->isNILEXPCAL (vinit))
		{
		  if (!b)
		    pK_messenger->messageIdent (defaultInit2, EM->getName (r));
		  else
		    {
		      vinit = setInitialValueDefault (r, nd + nw);
		      if (!EM->isNILEXPCAL (vinit))
			{
			  x = EM->mkBinaryExp (opinit, sr, vinit, NILBACKANNOTATION);
			  x = EM->mkTernaryExp (opret, x, vret, vfen, NILBACKANNOTATION);
			  NM->set_def (node, NILEXPCAL);
			  NM->set_def (node, x);
			}
		    }
		}
	      else
		{
		  if (b) EM->TypeCkeckingInitialValue (sr, vinit, NM->getBackAnnotation (node));
		  EM->TypePropagate (EM->getType (r), vinit);
		}
	    }
	}
      if (b0)
	pK_messenger->message (infoEventDelay);
    }
}

/**
   It sets the name of the clocks of the system.
   The operation can be inhibited, for a clock, using the global attribute called NO_NAME_FOR_OBJ.
   \see setClockName()
*/
void
DCG::setClocksName (boolean breprBySignal)
{
  if (!bSuspendClockNameAction)
    {
      require_clock_manager (CKM);
      // The predefined ones.
      setClockName (CKM->getNullClock (), breprBySignal);
      setClockName (CKM->getInitClock (), breprBySignal );
      setClockName (CKM->getFictitiousRoot (), breprBySignal);
      setClocksName(CKM->getlhroots (), breprBySignal); // Roots
      setClocksName(CKM->getlhplaced (), breprBySignal); // Placed
    }
}

/**
   The operation is applied if the operand clock (clk) is not marked by NO_NAME_FOR_OBJ
   (global attribute defined in CompilerAttributes.h file).
   
   It sets the name of a clock (clk):
   - Specific names are used for predefined clock (null, init, fictitious root),
   - if it exists an event signal declared by the user, its name is used.
   - otherwise, a suffixed name is built.
   Then, a name is assigned to each formula of the class, and to the assertions of the class.

   In this operation, a signal (s) and a clock (clk) may be unified.
*/
void
DCG::setClockName (ptinsth clk, boolean breprBySignal)
{
  ptinsth clk1;
  ptlinstances lident, lid;
  ptexpcal aSig, aSig1;
  boolean first, b, bfixed, found;
  ptINode node;
  ptident id, id1;

  require_managers (EM, NM, CKM);
  clk = CKM->getRepresentativeFormula (clk);
  if (!CKM->satisfiesMarked (clk, NO_NAME_FOR_OBJ))
    {
      id = CKM->getName (clk);
      if (eq_idents (id, NILIDENT))
	{
	  if (CKM->isClockUnifiedWithSignal (clk, &aSig))
	    id = EM->getName (aSig);
	  else
	    {
	      if (CKM->satisfiesNullClock (clk)) id = getNullClockId ();
	      else if (CKM->satisfiesFictitiousRoot (clk)) id = getUpbClockId ();
	      else if (CKM->satisfiesInitClock (clk)) id = getInitClockId ();
	      else
		{
		  lident = CKM->getSynchronousSignals (clk);
		  lid = lident;
		  aSig1 = NILEXPCAL;
		  bfixed = pk_true;
		  if (!ptlinstances_get_next (&lid, &aSig1))
		    {
		      if (CKM->eqClock (clk, getIOstep ()))
			{
			  id = getIOStepClockId ();
			  id = getNameManager ()->exclusiveIdent (makeIdent_ident_ident (id, getName ())); // added Jan 2015 (automata)
			}
		      else
			{
			  lident = CKM->getSynchronousSignalsKept (clk);
			  if (ptlinstances_get_next (&lident, &aSig)) id = EM->getName (aSig); else bfixed = pk_false;
			}
		    }
		  else
		    {
		      if (breprBySignal)
			{
			  found = pk_false;
			  aSig1 = NILEXPCAL;
			  while (!found && ptlinstances_get_next (&lident, &aSig))
			    if (EM->IsSignal (aSig))	// filtering the identifier of struct/bundle
			      {
				if (EM->IsEvent (aSig))
				  {
				    node = EM->getNode (aSig);
				    b = EM->IsInputFlow (aSig) || NM->is_equation (node);
				  }
				else
				  b = pk_true;
				if (b)
				  { 
				    if (!EM->IsAuxiliarySignal (aSig)) found = pk_true; else aSig1 = aSig;
				  }
			      }
			  if (found)
			    {
			      id = EM->getName (aSig);
			      if (EM->IsEvent (aSig))
				{
				  CKM->setClockUnifiedWithSignal (clk, aSig);
				  EM->setSignalUnifiedWithClock (aSig, clk);
				}
			      else
				CKM->setClockRepresentedBySignal (clk, aSig);
			    }
			  else if (aSig1 != NILEXPCAL)
			    {
			      id = EM->getName (aSig1);
			      CKM->setClockRepresentedBySignal (clk, aSig1);
			    }
			  else
			    bfixed = pk_false;
			}	
		      else
			{
			  found = pk_false;
			  aSig1 = NILEXPCAL;
			  first = pk_true;
			  while (!found && ptlinstances_get_next(&lident, &aSig))
			    if (EM->IsSignal(aSig)) 	// filtering the identifier of struct/bundle
			      {
				if (EM->IsEvent(aSig))
				  {
				    node = EM->getNode(aSig);
				    found = EM->IsInputFlow(aSig) || NM->is_equation(node);
				  }
				if (!found && first) { aSig1 = aSig; first = false;}
			      }  
			  
			  if (found) 
			    {
			      id = EM->getName(aSig);
			      CKM->setClockUnifiedWithSignal (clk, aSig);
			      EM->setSignalUnifiedWithClock (aSig, clk);
			    }
			  else 
			    if (first)
			      id = getNameManager()->new_pureClockIdent(-1);
			    else
			      {
				id = newIdent_concIdentIdent(getClockId (), EM->getName(aSig1));
				id = getNameManager()->exclusiveIdent(id);
			      }
			}
		    }
		  if (!bfixed)
		    {
		      if (CKM->satisfiesSub (clk))
			{
			  // for "when C", the clock is named TT_C ; FF_C for "when not C "
			  aSig = CKM->getConditionSub (clk);
			  if (aSig != NILEXPCAL)	// Affine Clock
			    if (EM->IsSignal (aSig))
			      {
				id = getNameManager ()->new_clockSamplingIdent (EM->getName (aSig), (CKM->getOperator (clk) == sub));
				bfixed = pk_true;
			      }
			}
		    }
		  if (!bfixed)
		    {
		      id = getNameManager ()->new_pureClockIdent (CKM->getNum (clk));
		      //  std::cout << " ---- new_pureClockIdent ( " << CKM->getNum (clk) << ')' << std::endl;
		    }
		}
	    }
	  CKM->setName (clk, id);
	}
      setNamesConstraintsAssertions(clk);
    }
}


/**
   It assigns h as the IO step clock of the program.
*/
void
DCG::setIOstep (ptinsth h)
{
  iostep = h;
  if (aClockManager->satisfiesUndefClock (complete))
    setCompleteClock(h);
}

/**
   @return the IO clock..
*/
ptinsth
DCG::getIOstep ()
{
  iostep = aClockManager->getRepresentativeFormula (iostep);
  return iostep;
}



/**
   It assigns h as the IO step clock of the program.
*/
void
DCG::setStep (ptinsth h)
{
  step = h;
}

/**
   @return the Step clock.
*/
ptinsth
DCG::getStep ()
{
  ptinsth h;
  h = aClockManager->getRepresentativeFormula (step);
  step = h;
  return step;
}


/**
   It assigns h as the IO step clock of the program.
*/
void
DCG::setCompleteClock (ptinsth h)
{
  complete = h;
}

/**
   @return the Complete clock.
*/
ptinsth
DCG::getCompleteClock ()
{
  return aClockManager->getRepresentativeFormula (complete);
}



