/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_ static ptDCG_endo gw = NILDCG_ENDO;

#include "DCGs_Macros.c"

void
DCG_endo::init (ptDCG g)
{
  if (gw == NILDCG_ENDO) gw = new DCG_endo ();
  gw->setDCG (g);
}


/**
   DCG_endo graph to a DCG_bool one.

   This method performs the interformat DCGPoly to DCGBool translation for the current graph. 
   The operation is not applied to the referenced sub-graphs. 
   \warning {the referenced graphs must be endochronous.}
   
   \ref InterFormats "For more details..."
   
*/
void
DCG_endo::transformToBooleanControlOneLevel (boolean mkSubForNosub)
{
  ptDCGEndo2DCGBoolTranslator x;
  ptlinstances lsig;

  //undersamplingOnSignal(pk_false, pk_false);
  x = new DCGEndo2DCGBoolTranslator (getDCG ());
  lsig = NILLINSTANCES; // set of inputs moved to ouptuts.
  x->apply (mkSubForNosub, &lsig);
  delete x;
  setFormatLevel (DCG_BOOL);
  addHierarchyPragmas ();
  unmarkObjects (CLKSIGUNIFICATION);	// It removes the links clock <-> signal
  updatePragmas ();
}

/** 

    It adds exceptions for clock constraints. 
    It allows to have code generating with clock constraints.
    The equivalence of clock formula will be, for example, verified dynamically during simulation (at run time).
*/
void
DCG_endo::addExceptionsForConstraintsOneLevel ()
{
  pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("DCG_endo::addExceptionsForConstraintsOneLevel"));
}
