/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Compiling klExprSig1:= P( klExprSig2 ) or list_of_signals::= P( list_of_signals)
   where P is a reference to a model.

   @param code (in): signal code operator := or::= .
   @param klExprSig1, klExprSig2: lists of expressions on signals (see grammar).
   @return the upper bound clock of the expression.
*/
  ptinsth kernel2graph::instCallDefinition (ptklExprSig klExprSig1, ptklExprSig klExprSig2, noperators code)
{
  ptinsth vh, vh1, vhdis;
  ptexpcal vlaux, leftExpr, e, e1;
  ptlinstances instent;
  boolean b, bPartDef, bPartDefDefault;

  bPartDef = (code != opdefsig);
  b = pk_false;
  vh = instLeftPartDefinition (bPartDef, klExprSig1, pk_true, &leftExpr, &b);
  if (b) return the_nullClock_K2G_copy;
  // Creation de signaux intermediaires pour une definition partielle.
  if (bPartDef) e = NILEXPCAL; else e = leftExpr;
  instent = NILLINSTANCES;
  vh = instCall (klExprSig2, e, NILEXPCAL, &vlaux, &instent);
  if (bPartDef)
    {
      bPartDefDefault = (code == opdefpartdef);
      while (EM->getNaryArgs (&vlaux, &e))
	{
	  EM->getNaryArgs (&leftExpr, &e1);
	  if (bPartDefDefault)
	    EM->addMark (e1, PART_DEFAULT_DEFINITION);
	  vhdis = EM->getClock (e);
	  if (igc->topContextClock (&vh1))
	    vhdis = CKM->makeInf (vhdis, vh1);

	  EM->addMark (e1, PART_DEFINITION);
	  vhdis = igc->createEquation (e1, e, vhdis, vhdis, ptlinstances_push_front (NILLINSTANCES, e), back_annot_obj);
	  EM->removeMark (e1, PART_DEFINITION);

	  if (bPartDefDefault)
	    EM->removeMark (e1, PART_DEFAULT_DEFINITION);
	}
    }

  return vh;
}

/**
   Compiling klExprSig1:= klExprSig2, or  klExprSig1::= klExprSig2, or  klExprSig1:= defaultvalue klExprSig2.
   where klExprSig2 is not a reference to a model.
   @param code (in): signal code operator (:=, opdefsig) or (::=, opdefpart), or (::= defaultvalue, opdefpartef).
   @return the clock of the defined signal.
*/
ptinsth
kernel2graph::instSimpleOrTupleDefinition (ptklExprSig klExprSig1, ptklExprSig klExprSig2, noperators code)
{
  ptexpcal vexpcal;
  ptinsth vh, contextClock;
  boolean b, bpartdef;

  b = pk_false;
  bpartdef = (code != opdefsig);
  contextClock = instLeftPartDefinition (bpartdef, klExprSig1, pk_false, &vexpcal, &b);
  //  traceability(vexpcal);
  if (b) return the_nullClock_K2G_copy;	// double definition
  if (bpartdef)
    vh = instPartialDefinitionEqu (contextClock, vexpcal, klExprSig2, code);
  else
    vh = instEquation (contextClock, vexpcal, klExprSig2, code);
  return vh;
}



/**
  It instanciates a reference (vklExprProcappe) to a model (vfi) when the model is a external function.
  @param vh (out) the clock of the resulting expression;
  @param vklExprProcappe (in) a kernel-tree reference to a model;
  @param vfi (in) the referenced model;
  @param vexpcal (out) the instanciated expression of vklExprProcappe;
  @param instent (in/out) vexpcal is added to the instent (in) list
  @param isfree (in/out) if "true" vhcste must be used to fix the clock of the expression. It is set to "false" value
  when the expression has its own clock.
  @param vhcste (in)  clock context to fix the clock of the constant, parameter.
  @return "true" if the model is an external model, "false" otherwise.
*/
boolean
  kernel2graph::instExternalFunction (ptklExprSig vklExprSig, ptinsth vhcste, boolean * isfree,
				      ptinsth * vh, ptexpcal * vexpcal, ptlinstances * instent)
{
  ptklExprProc corps;
  ptident id;
  ptklExprSig vklExprSiglarg, vali;
  ntrees savetree, vlang, arb;
  ptlrepid param, entree, sortie;
  boolean bTypeConversion;
  noperators vop;
  boolean b;
  categoryType vcat;
  ptreptyp vtyp;
  ptklExprProc vklExprProc;
  ptlinstances instent1;
  ptexpcal linstentpos, lp, x;
  TypeType TT;
  ptinsth vhcal;
  ptrepfi vfi;
  ptklExprProc vklExprProcappe;

  vklExprProcappe = kl_getFlowExprModelProcCall (vklExprSig);
  vfi = getModelFromExpflot (vklExprSig);

  vfi = getModelOf (vfi, &bTypeConversion);
  savetree = kl_getBodyModel (vfi, &corps);
  if (kl_getExprProcType (corps) != opexte) return pk_false;

  sortie = kl_getInsOutsModel (vfi, &entree);
  if (ptlrepid_size (sortie) != 1)
    return pk_false;
  if (!bTypeConversion)
    {
      vop = noper (savetree);
      if (!LL_isFunctionFamilyOp (vop)) return pk_false;
    }

  ptlrepid_getIdInfosAt (sortie, 1, &vali, &vtyp, &vcat, &arb);
  TT = sigTreeType2TypeType (vtyp);

  instent1 = NILLINSTANCES;
  vklExprProc = kl_getModelProcCall (vklExprProcappe, &vklExprSiglarg, &currentTree);
  NEW_BACK_ANNOT(currentTree);
  param = kl_getParametersModel (vfi);
  instdclparameters (param, kl_getEffParameterCall (vklExprProc));
  linstentpos = instposit (entree, vklExprSiglarg, &instent1, hundef);
  if (!igc->isInConstantContext ())
    *vh = igc->synchronizeIdentifiers (instent1);
  else
    *vh = hinit;
  b = pk_true;
  vhcal = *vh;
  if (CKM->satisfiesUndefClock (*vh))
    {
      *vh = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
      if (CKM->satisfiesUndefClock (*vh))
	{
	  b = pk_false;
	  vhcal = hinit;
	}
    }
  *instent = ptlinstances_sortedfusion (instent1, *instent);
  param = kl_getParametersModel (vfi);
  lp = mkOrderedListOfInstances (param, NILEXPCAL);
  id = kl_getNameModel (vfi);
  *vexpcal = EM->mkCallExp (id, lp, linstentpos, TT, vhcal);
  vlang = kl_getLangExternal (corps);
  addPragmasAndAttributesProcCall (vfi, *vexpcal, vlang);
  deplinst (param);
  if (kl_isIntrinsicModel (vfi))
    {
      if (EM->satisfiesList (*vexpcal))
	x = *vexpcal;
      else
	x = EM->mkNaryExp (oplesign, *vexpcal, NILEXPCAL, back_annot_obj);
      EM->TypeCheckingIntrinsic (id, lp, linstentpos, x, back_annot_obj);
    }
  if (b) (*isfree) = pk_false;
  return pk_true;
}

/**
   Compiling a reference to a model of an external process: a procCall node is added to the graph.
   Assumed: the context of the signals is well defined.
   @param vklExprProc (in) the expression definning a reference to a model of an external process.
   @param vh (out) the clock of the expression.
   @return "true" if the expression is a reference to a external model, "false" otherwise.
   Local to instmacro() method.
*/
boolean
kernel2graph::externalNode (ptklExprProc vklExprProc, ptinsth * vh)
{
  ptrepfi vfi;
  ptklExprProc corps;
  ptident nomfi;
  ptklExprSig lparam;
  ntrees arbrefi, savetree, vlang;
  ptlrepid varid, param, entree, sortie;
  ptexpcal lentren, lsortren, effParam;
  ptlinstances lent, lsort;
  ptDCG g;
  ptinsth vh1;
  boolean AutomatonModelSave, SignalThreadSave, baprocedure, baFunction, bspec, baNode, banAction, bTypeConversion, b;
  ptINode vinstr, n;
  ptpragma vRunOnSave, vGreyBoxSave, vBlackBoxSave;
  noperators vop;

  kl_getMacro (vklExprProc, &varid, &lparam, &arbrefi);
  vfi = getModelOf (varid, &bTypeConversion);
  savetree = kl_getBodyModel (vfi, &corps);
  if (kl_getExprProcType (corps) != opexte)
    return pk_false;
  if (!bTypeConversion)
    vop = noper (savetree);
  else
    vop = opfunc;
  nomfi = kl_getNameModel (vfi);

  vlang = kl_getLangExternal (corps);
  sortie = kl_getInsOutsModel (vfi, &entree);
  param = kl_getParametersModel (vfi);

  baFunction = LL_isFunctionFamilyOp (vop);
  baNode = LL_isNodeFamilyOp (vop);
  banAction = LL_isActionFamilyOp (vop);
  baprocedure = LL_isProcedureFamilyOp (vop);

  lentren = ctexternemacro (entree, pk_true, currentTree, NILEXPCAL);
  lsortren = ctexternemacro (sortie, pk_false, currentTree, NILEXPCAL);
  currentTree = arbrefi;
  NEW_BACK_ANNOT(currentTree);
  lent = getCurrentInstances (entree, NILLINSTANCES, pk_true);
  lsort = getCurrentInstances (sortie, NILLINSTANCES, pk_false);
  effParam = mkOrderedListOfInstances (param, NILEXPCAL);

  pushContext (vfi, &vRunOnSave, &vGreyBoxSave, &vBlackBoxSave, &AutomatonModelSave, &SignalThreadSave);
  traceability (effParam, lentren, lsortren);

  if (!instintrinsicfunc (vfi, nomfi, lent, lsort, effParam, lentren, lsortren, baFunction, vh))
    { /* process, node, Function, action */
      bspec = kl_getSpecModel (vfi, &corps);
      if (banAction && !bspec)
	{
	  *vh = igc->getLabelClock (currentTree);
	  b = pk_true;
	}
      else if (bspec || !baFunction)	// node, action, process, procedure, "function without spec"
	{
	  vh1 = instUnexpandedModel (vfi, effParam, vop, &g, &n);
	  g->setGraphOfExternalProcess (savetree, vlang);
	  *vh = instexpandedModel (vfi, effParam, lentren, lsortren);
	  if (baprocedure)
	    *vh = CKM->makeSup (igc->upperBoundClockOfIdentifiers (lent), igc->upperBoundClockOfIdentifiers (lsort));
	  else
	    *vh = CKM->makeSup (*vh, vh1);
	  b = pk_false;
	}
      else
	/* function without Spec  */
	{
	  *vh = CKM->makeSynchro (igc->synchronizeIdentifiers (lent), igc->synchronizeIdentifiers (lsort));
	  if (CKM->satisfiesUndefClock (*vh))
	    *vh = igc->getLabelClock (currentTree);
	  b = pk_true;
	}

      if (b)
	{
	  if (igc->mkProcCallNode (nomfi, effParam, lentren, lsortren, *vh, kl_getSafetyProperty (vfi), back_annot_obj, MK_BACK_ANNOT( kl_getTreeReference (vfi)), pk_false, &vinstr))
	    addPragmasAndAttributesProcCall (vfi, vinstr, vlang);
	}

    }
  if (baFunction || baNode || banAction || baprocedure)
    {
      NM->addDependences (lent, lsort, hundef, pk_false);
    }
  deplinst (param);
  popContext (vfi, vRunOnSave, vGreyBoxSave, vBlackBoxSave, AutomatonModelSave, SignalThreadSave);
  ptlinstances_free (lent);
  ptlinstances_free (lsort);
  return pk_true;
}

/**
   It pushes a call expression (vtree) in the stack of calls (used for tracability).
*/
void
kernel2graph::callStackPush (ntrees vtree)
{
  TM->pushCallTree (vtree);	// Traceability
  callsStack = stackPush (callsStack, vtree);
}

/**
   It pops the top of the stack of calls (used for tracability).
*/
void
kernel2graph::callStackPop (ntrees vtree)
{
  TM->popCallTree (vtree);	// Traceability
  callsStack = stackPop (callsStack);
}

/**
   Compiling (macroexpansion) a reference to a model of a process.
   Assumed: the context of the signals is well defined.
   @return the upper-bound clock of the expression.
*/
ptinsth
kernel2graph::instmacro (ptklExprProc vklExprProc)
{
  ptpragma vpragma, vRunOnSave, vGreyBoxSave, vBlackBoxSave;
  ptlpragmas lprag;
  ptrepfi vfi;
  ptident save_id1, save_id2;
  ptklExprSig lparam;
  ntrees arbrefi, savetree, t;
  ptlrepid varid, param, entree, sortie;
  ptklExprProc corps, spec;
  ptinsth vClockAutomatonSave, vh;
  boolean AutomatonSave, SignalThreadSave, bIsClockedAutomatonSave, /* bseparate,*/ bTypeConversion, bunexpanded, bexpanded, bsignalThread;
  noperators opl, vop;
  ptexpcal vCurrentStateAutomatonSave, effparam, lentren, lsortren;
  ptDCG g;
  ptlinstances l1, l2;
  ptINode n;
  BackAnnotationType save_back_annot_obj;
  
  kl_getMacro (vklExprProc, &varid, &lparam, &arbrefi);
  vfi = getModelOf (varid, &bTypeConversion);

  // Restriction : a root model with formal process is excluded.
  if (kl_isFormalModel (vfi))
    return the_nullClock_K2G_copy;

  if (!externalNode (vklExprProc, &vh))
    {
      savetree = currentTree; 
      save_back_annot_obj = back_annot_obj;
      t = kl_getBodyModel (vfi, &corps);
      NEW_BACK_ANNOT(t);
      if (!stackIsIn (callsStack, arbrefi))
	{
	  callStackPush (arbrefi);
	  sortie = kl_getInsOutsModel (vfi, &entree);
	  lprag = kl_getPragmaModel (vfi);
	  // bseparate = getPragmaKeyword (lprag, PRAGMA_SEPARATE, &vpragma);
	  // if (bseparate) pK_messenger->printString (messDeprecatedSeparate);

	  bunexpanded = getPragmaKeyword (lprag, PRAGMA_UNEXPANDED, &vpragma);
	  bsignalThread = getPragmaKeyword (lprag, PRAGMA_SIGNAL_THREAD, &vpragma);
	  bexpanded = /* !bseparate &&*/ !bunexpanded && !bsignalThread;
	  if (bexpanded) opl = opnil; else opl = oplident;
	  pushContext (vfi, &vRunOnSave, &vGreyBoxSave, &vBlackBoxSave, &AutomatonSave, &SignalThreadSave);
	  param = kl_getParametersModel (vfi);
	  NEW_BACK_ANNOT(arbrefi);

	  vop = noper (t);
	  if (LL_isFunctionFamilyOp (vop))
	    {
	      l1 = getCurrentInstances (entree, NILLINSTANCES, pk_true);
	      l2 = getCurrentInstances (sortie, NILLINSTANCES, pk_false);
	      vh = CKM->makeSynchro (igc->synchronizeIdentifiers (l1), igc->synchronizeIdentifiers (l2));
	      ptlinstances_free (l1);
	      ptlinstances_free (l2);
	    }

	  if (bexpanded)
	    {
	      lentren = ctexternemacro (entree, pk_true, currentTree, NILEXPCAL);
	      lsortren = ctexternemacro (sortie, pk_false, currentTree, NILEXPCAL);
	      effparam = mkOrderedListOfInstances (param, NILEXPCAL);
	      traceability (effparam, lentren, lsortren);
	      save_id1 = pushModelAndContextId (vfi, &save_id2, &vCurrentStateAutomatonSave, &vClockAutomatonSave, &bIsClockedAutomatonSave);
	      // vhOversamplingSignalThread = SignalThreadInitialize(entree, sortie);
	      vh = compilingEproc (corps);
	      kl_getSpecModel (vfi, &spec);
	      vh = CKM->makeSup (vh, compilingEproc (spec));

	      automatonFinalize(entree);
	      // SignalThreadFinalize(entree, sortie);
	      popModelAndContextId (vfi, save_id1, save_id2, vCurrentStateAutomatonSave, vClockAutomatonSave, bIsClockedAutomatonSave);
	    }
	  else
	    if (bsignalThread ) 
	      vh = instSignalThreadModel (vfi, mkOrderedListOfInstances (param, NILEXPCAL), noper (t), &g, &n);
	    else
	      vh = instUnexpandedModel (vfi, mkOrderedListOfInstances (param, NILEXPCAL), noper (t), &g, &n);
	  deplinst (param);
	  popContext (vfi, vRunOnSave, vGreyBoxSave, vBlackBoxSave, AutomatonSave, SignalThreadSave);
	  callStackPop (arbrefi);
	}
      else
	{
	  pK_messenger->errorTree (errcycle, arbrefi);
	  vh = hundef;
	}
      currentTree = savetree;
      back_annot_obj = save_back_annot_obj;
    }
  return vh;
}

/**
   It manages parameters for the model references
    - (x,,...) := P{p1,...pn} (e1, ...ep)
    - or P{p1,...pn} 
    Effective parameters are associated with the formal ones.
*/
void
kernel2graph::pushParameters (ptklExprProc vklExprProc)
{
  ptrepfi vfi;
  ptklExprSig lparam;
  ntrees arbrefi;
  ptlrepid varid, param;
  boolean b;

  kl_getMacro (vklExprProc, &varid, &lparam, &arbrefi);
  vfi = getModelOf (varid, &b);
  param = kl_getParametersModel (vfi);
  NEW_BACK_ANNOT(arbrefi);
  instdclparameters (param, lparam);
}


/**
   Compiling (macroexpansion) a reference (vklExprProcappe) to a model of a process.
   It prepares the stack of the i/o identifiers of the model before calling \ref instmacro method.
   @return the upper-bound clock of the expression.
   
*/
ptinsth
kernel2graph::instappe (ptklExprProc vklExprProcappe, ptexpcal ls, ptexpcal vexp, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptrepfi vfi;
  ptinsth hconstantcontext, vh;
  ptklExprProc vklExprProc;
  ptklExprSig vklExprSiglarg, lparam;
  ntrees arbreappe;
  ptlrepid varid, lsorties, lentrees;
  ptexpcal linstentpos;
  ptlinstances instent1;
  boolean bAutomatonWithLabelArg_old, bAutomatonWithLabelArg, bintrinsic, b, b1, bTypeConversion, bappel;
  ptident id;

  numappel++;
  vklExprProc = kl_getModelProcCall (vklExprProcappe, &vklExprSiglarg, &currentTree);
  NEW_BACK_ANNOT(currentTree);
  instent1 = NILLINSTANCES;
  kl_getMacro (vklExprProc, &varid, &lparam, &arbreappe);
  vfi = getModelOf (varid, &bTypeConversion);
  bintrinsic = kl_isIntrinsicModel (vfi);
  id = kl_getNameModel (vfi);
  if (bintrinsic && isNilProcId (id)) return hundef; // Nil_Process <=> (||) without spec.

  lsorties = kl_getInsOutsModel (vfi, &lentrees);
  // A state variable is accepted for the current state of an automaton.
  b = bintrinsic && isSetCurrentStateId (id);
  if (b) b1 = igc->getSetStateVarContext (pk_true);

  // Context clock: Transition_Automaton(S1,S2, trigger), constant context clock is the clock of the automaton.
  if ( bintrinsic && isTransitionId (id)) hconstantcontext = igc->getClockAutomaton (); else hconstantcontext = hundef;

  // Don't assign a clock to a label used for in automaton intrinsics.
  bAutomatonWithLabelArg = (bintrinsic && ( isTransitionId (id) || isInitialStateId (id) || isAutomatonIsIn(id) ));
  if (bAutomatonWithLabelArg) bAutomatonWithLabelArg_old = igc->setInAutomatonIntrinsicContext(true);
  linstentpos = instposit (lentrees, vklExprSiglarg, &instent1, hconstantcontext );
  if (bAutomatonWithLabelArg) bAutomatonWithLabelArg_old = igc->setInAutomatonIntrinsicContext(bAutomatonWithLabelArg_old);
  if (b) b1 = igc->getSetStateVarContext (b1);

  pushParameters (vklExprProc);
  // l = linstentpos;
  bappel = (linstentpos != NILEXPCAL);	/* bappel: at pk_true for (x,y):= P{n}(e1,e2) */
  if (bappel) setCurrentInstances (lentrees, linstentpos);
  *vexpcalcul = pushModelOutputs (ls, vfi, vexp, instent);
  vh = instmacro (vklExprProc);
  popModelOutputs (vfi);
  if (bappel)
    {
      deplinst (lentrees);
      EM->free (linstentpos);
    }
  return vh;
}

/**
   Compiling the operand expression (vklExprProc) defined by E1 op E2 ... op En,
   where op is an operator of the boolean lattice (^<, ^>, ^#).
   The clock relations are not added in the system. They must be 
   added after the calling to this method.
   @return the instanciated expression of the operand expression.
   
*/
ptexpcal
kernel2graph::instLatticeOpCommon (ptklExprProc vklExprProc, ptlinstances * instent, ptident idNameAssociatedWithOperator)
{
  ptklExprSig lsync;
  ptexpcal vexp;

  *instent = NILLINSTANCES;
  lsync = kl_getlExprClockOp (vklExprProc, &currentTree);
  NEW_BACK_ANNOT(currentTree);
  vexp = instposit (NILLREPID, lsync, instent, hundef);
  return vexp;
}

/**
    Compiling  ^= expression.
    @param vklExprProc: an assumed "E1^= .... ^= En"  expression
    @return the synchronized clock of the expressions.
*/
ptinsth
kernel2graph::instClockEqu (ptklExprProc vklExprProc)
{
  ptlinstances instent;
  ptexpcal vexpcal, vleft, vright;
  ptinsth vh, h, hleft;
  ptident id;

  getNameAssociatedWithOperator (oplequhor, &id);
  vexpcal = instLatticeOpCommon (vklExprProc, &instent, id);

  if (!EM->getNaryArgs (&vexpcal, &vleft))
    return the_nullClock_K2G_copy;

  if (EM->getExprClock (vleft, &hleft))
    vh = hleft;
  else
    hleft = hundef;

  while (EM->getNaryArgs (&vexpcal, &vright))
    if (EM->getExprClock (vright, &h))
      {
	if (hleft == hundef)
	  {
	    hleft = h;
	    vh = hleft;
	  }
	else
	  {
	    if (inAssertBlock)
	      h = CKM->makeAssumption (h, hleft);
	    else
	      h = CKM->makeSynchro (h, hleft);
	    vh = CKM->makeSup (vh, h);
	  }
      }
    else if (!EM->IsConstantExpr (vright))
      pK_messenger->errorTree (errCannotDefineClock, currentTree);
  ptlinstances_free (instent);
  return vh;
}


/**
   Compiling  ^< or ^> expression.
   @param vklExprProc: an "E1 < op > .... < op > En"  expression.
   @param binf (in): specify < op > operator; at pk_true for ^<, pk_false for ^>.
   @return the upper-bound clock of the expression.
*/
ptinsth
kernel2graph::instClockSupInf (ptklExprProc vklExprProc, boolean binf)
{
  ptlinstances instent;
  ptexpcal vexpcal, vleft, vright;
  ptinsth vh, h, hleft, hright;
  ptident id;

  //  if (binf) op = opldechor; else op = oplcrohor;
  getNameAssociatedWithOperator (binf ? opldechor : oplcrohor, &id);
  vexpcal = instLatticeOpCommon (vklExprProc, &instent, id);
  if (!EM->getNaryArgs (&vexpcal, &vleft))
    return the_nullClock_K2G_copy;
  EM->getExprClock (vleft, &hleft);	// hleft =  EM->getClock(vleft);
  vh = hleft;
  while (EM->getNaryArgs (&vexpcal, &vright))
    {
      EM->getExprClock (vright, &hright);	// hright = EM->getClock(vright);
      if (binf)
	h = CKM->makeInf (hleft, hright);
      else
	h = CKM->makeSup (hleft, hright);
      if (inAssertBlock)
	h = CKM->makeAssumption (h, hleft);
      else
	h = CKM->makeSynchro (h, hleft);
      vh = CKM->makeSup (vh, h);
      hleft = hright;
    }
  ptlinstances_free (instent);
  return vh;
}

/*
  Compiling  ^< expression.
  @param vklExprProc: an "E1 ^< e .... ^< e En"  expression.
  @return the upper-bound clock of the expression.
*/
ptinsth
kernel2graph::instClockIncreasing (ptklExprProc vklExprProc)
{
  return instClockSupInf (vklExprProc, pk_true);
}


/*
  Compiling  ^> expression.
  @param vklExprProc: an "E1 ^> .... ^> En"  expression.
  @return the upper-bound clock of the expression.
*/
ptinsth
kernel2graph::instClockDecreasing (ptklExprProc vklExprProc)
{
  return instClockSupInf (vklExprProc, pk_false);
}



/*
  Compiling  ^# expression.
  For each element x of the list (list), it adds clock(x) ^* h = ^0 in the clock system.
  @param h (in) context clock.
  @param list (in) list of expressions on signals.
  
  @return none.
*/
void
kernel2graph::instClockExlcusive (ptinsth h, ptexpcal list)
{
  ptexpcal velem;
  ptinsth helem, h1;

  while (EM->getNaryArgs (&list, &velem))
    {
      EM->getExprClock (velem, &helem);	//  helem = EM->getClock(velem);
      h1 = CKM->makeInf (h, helem);
      if (inAssertBlock)
	h1 = CKM->makeAssumption (h1, the_nullClock_K2G_copy);
      else
	h1 = CKM->makeSynchro (h1, the_nullClock_K2G_copy);
    }
}

/*
  Compiling  ^# expression.
  @param vklExprProc (in) an "E1 ^# .... ^# En"  expression.
  @return the upper-bound clock of the expression.
*/
ptinsth
kernel2graph::instClockExlcusive (ptklExprProc vklExprProc)
{
  ptlinstances instent;
  ptexpcal vexpcal, vleft;
  ptinsth vh, h;
  ptident id;

  getNameAssociatedWithOperator (oplexchor, &id);
  vexpcal = instLatticeOpCommon (vklExprProc, &instent, id);
  vh = the_nullClock_K2G_copy;
  while (EM->getNaryArgs (&vexpcal, &vleft))
    {
      EM->getExprClock (vleft, &h);
      instClockExlcusive (h, vexpcal);
      vh = CKM->makeSup (vh, h);
    }
  return vh;
}

/** Finalizing the array node (vArrayNode).

    For each flow (f) of the list (vlrepid), the current signal (x) associated with f is popped, 
    the array node is assigned to the top of stack associated with f, taking into account the fields of the
    structures.
*/
void
kernel2graph::popCopyOutputSignals (ptlrepid vlrepid, ptINode vArrayNode)
{
  ptexpcal sig, vcopy, vsignal1, vsignal2;
  ptlinstances lfields1, lfields2;
  int n, i;
  ptINode node2, node1;

  n = ptlrepid_size (vlrepid);
  for (i = 1; i <= n; i++)
    {
      vcopy = getCurrentInstance (vlrepid);	// the clone of the output of the array.
      kl_popInstance (nameId (vlrepid));
      sig = getCurrentInstance (vlrepid);	// the output of the array.

      if (EM->getFieldsOf (vcopy, &lfields1))
	// Some fields of a structure/bundle are defined inside of the array of process (vArrayNode).
	{
	  lfields2 = igc->mkFieldsDescription (sig);
	  while (ptlinstances_get_next (&lfields1, &vsignal1))
	    {
	      ptlinstances_get_next (&lfields2, &vsignal2);
	      node1 = EM->getNode (vsignal1);
	      if (!NM->is_undefined (node1))
		EM->setNode (vsignal2, vArrayNode);
	      else
		{
		  // Merging the nodes.
		  node2 = EM->getNode (vsignal2);
		  NM->dependencesPasteOn (node1, node2);
		  NM->set_successors (node1, NILLDEPENDENCE);
		  NM->removeNode (node1);
		  EM->setNode (vsignal1, node2);
		}
	    }
	}
      else
	EM->setNode (sig, vArrayNode);

      vlrepid = succid (vlrepid);
    }
}


/** 
    Compiling the bodypart/withpart of a array/iterate structure.
        
    @param corpsdo (in): the body of the arrary/iterate process.
    @param withPart (in): the body of the arrary/iterate process.
    @return the clock defined by the upper bound  of the expression of the compiled expression.
*/
ptinsth
kernel2graph::CommonArrayIterate (ptklExprProc corpsdo, ptklExprProc withPart, ptlrepid sortie, ptINode vArrayNode, ptexpcal vIndex)
{
  ptinsth vh, vhEproc;

  igc->pushArray (vArrayNode);
  // With Part
  setCopyCurrentInstances (sortie);
  igc->setCurrentSubgraph (pk_false);
  vhEproc = compilingEproc (withPart);
  deplinst (sortie);

  // Body part
  igc->setCurrentSubgraph (pk_true);
  setCopyCurrentInstances (sortie);
  vhEproc = compilingEproc (corpsdo);
  popCopyOutputSignals (sortie, vArrayNode);
  igc->popArray ();

  vh = igc->getSignalClockSynchronizedWith (vIndex, vhEproc);
  NM->unlinkNodesFromClock (vArrayNode);
  igc->takeIntoAccountContexts (vArrayNode);
  NM->insertNode (vArrayNode, vh);
  return vh;
}


/** 
    Compiling array of processes expression (array strucutre).

    @param vklExprProc (in): the compiled expression assumed to be an array of processes (array i to N of ...).
    @return the clock defined by the upper bound  of the expression of the compiled expression.
*/
ptinsth
kernel2graph::instArrayOfProcesses (ptklExprProc vklExprProc)
{
  ptlrepid indice;
  ptklExprSig bsup, indexExpr;
  ptklExprProc corpsdo, withPart;
  ptlrepid param, entree, sortie;
  ptINode vArrayNode, indexNode;
  ptexpcal vIndexdef, vIndex, outputs, inputs, upperBound;
  ptinsth vh;
  boolean isfree;
  ptlinstances instent;
  TypeType ContextType;

  kl_getArray (vklExprProc, &indexExpr, &bsup, &withPart, &corpsdo, &currentTree);
  NEW_BACK_ANNOT(currentTree);
  kl_getInsOutsParams (vklExprProc, &entree, &sortie, &param);
  instent = NILLINSTANCES;
  isfree = pk_true;
  ContextType = get_IntegerType ();
  vh = instOperand (hundef, ContextType, &isfree, bsup, &upperBound, &instent, opmoti);
  indice = kl_getIdentifierExpr (indexExpr);
  vIndex = mkNewIndex (indice, upperBound);
  vh = igc->setClockVariable (vIndex);
  vIndexdef = EM->mkTernaryExp (openum, ptexp0, upperBound, ptexp1, back_annot_obj);
  indexNode = NM->mk_node_index (vIndex, vIndexdef);

  outputs = mkOrderedListOfInstances (sortie, NILEXPCAL);
  inputs = mkOrderedListOfInstances (entree, NILEXPCAL);

  vArrayNode = NM->mk_node_array (indexNode, inputs, outputs, back_annot_obj);
  vh = CommonArrayIterate (corpsdo, withPart, sortie, vArrayNode, vIndex);
  return vh;
}


/** 
    Compiling array of processes (iterate structure).
    @param vklExprProc (in): the compiled expression assumed to be an iteration of processes (iterate (i1,...,in) of ).
    @return the clock defined by the upper bound  of the expression of the compiled expression.
*/
ptinsth
kernel2graph::instIterateProcess (ptklExprProc vklExprProc)
{
  ptklExprSig bsup, indexExpr;
  ptklExprProc corpsdo, withPart;
  ptlrepid param, entree, sortie;
  ptINode vArrayNode, indexNode, vnode;
  ptexpcal vIndexdef, vIndex, outputs, inputs, upperBound, vx, aux, vind, vlindex;
  ptinsth vh;
  boolean isfree;
  ptlinstances instent;
  TypeType ContextType;

  kl_getArray (vklExprProc, &indexExpr, &bsup, &withPart, &corpsdo, &currentTree);
  NEW_BACK_ANNOT(currentTree);
  kl_getInsOutsParams (vklExprProc, &entree, &sortie, &param);
  instent = NILLINSTANCES;
  isfree = pk_true;
  ContextType = get_IntegerType ();
  vh = instOperand (hundef, ContextType, &isfree, indexExpr, &vind, &instent, opiter);

  // Setting upper bound.
  upperBound = EM->TypeCheckingGetUpperBoundIterationIndex (vind, back_annot_obj);

  // Setting indices
  vIndex = mkNewIndex (NILLREPID, upperBound);
  vh = igc->setClockVariable (vIndex);
  vIndexdef = EM->mkTernaryExp (openum, ptexp0, upperBound, ptexp1, back_annot_obj);
  indexNode = NM->mk_node_index (vIndex, vIndexdef);
  vlindex = EM->mkNaryExp (oplesign, vIndex, NILEXPCAL, back_annot_obj);

  push_Iterate_Indexes (indexExpr, vind, vlindex);

  if (EM->satisfiesList (vind))
    inputs = vind;
  else
    inputs = EM->mkNaryExp (oplesign, vind, NILEXPCAL, back_annot_obj);

  outputs = mkOrderedListOfInstances (sortie, NILEXPCAL);
  vArrayNode = NM->mk_node_array (indexNode, inputs, outputs, back_annot_obj);

  if (EM->satisfiesList (vind))
    {
      aux = vind;
      while (EM->getNaryArgs (&aux, &vx))
	{
	  vnode = EM->getNode (vx);
	  NM->addDependence (vnode, vArrayNode, hundef);
	}
    }
  else
    {
      vnode = EM->getNode (vind);
      NM->addDependence (vnode, vArrayNode, hundef);
    }

  vh = CommonArrayIterate (corpsdo, withPart, sortie, vArrayNode, vIndex);
  pop_Iterate_Indexes (indexExpr);

  return vh;
}


/** Compiling "dependence" expression.

    @param vklExprProc (in) a dependence definition.
    @return the clock defined by the upper bound of the expression of the compiled expression.
*/
ptinsth
kernel2graph::instDependences (ptklExprProc vklExprProc)
{
  ptklExprSig vfrom, vto, vat;
  ptlinstances lsig1, lsig2, instent;
  ptexpcal vexpcond;
  boolean isfree;
  ptinsth vhCond, vhexp;

  kl_getDependence (vklExprProc, &vfrom, &vto, &vat);
  isfree = pk_false;
  lsig1 = instSignalsLabels (vfrom);
  lsig2 = instSignalsLabels (vto);
  isfree = pk_false;
  instent = NILLINSTANCES;
  if (vat != NILEXPRSIG)
    vhCond = instargwhen1 (hundef, pk_false, vat, &vexpcond, pk_false, &vhexp);	//condition 
  else
    vhCond = hundef;
  NM->addDependences (lsig1, lsig2, vhCond, pk_true);
  return the_nullClock_K2G_copy;
}

/**
   Compiling "case" structure.
   @param vklExprProc (in): a case definition.
   @return the clock defined by the upper bound  of the expression of the compiled expression.
*/
ptinsth
kernel2graph::instCase (ptklExprProc vklExprProc)
{
  ptklExprSig id, val;
  ptlcase l_case;
  ptklExprProc p_else, vexpr;
  ptexpcal testSig, condExpr, valInst, lexp;
  ptlinstances lval, instent, instent1, vdummy;
  boolean pelse, isfree;
  ptinsth helse, contextClock, hresult, vhsub, h2, h1, hleft, h, vh;
  ntrees savetree;
  BackAnnotationType save_back_annot_obj;
  
  savetree = currentTree;
  save_back_annot_obj = back_annot_obj;
  
  kl_getCase (vklExprProc, &id, &l_case, &p_else, &currentTree);
  NEW_BACK_ANNOT(currentTree);
  pelse = (p_else != NILEXPFIL);
  isfree = pk_true;
  instent = NILLINSTANCES;
  contextClock = instOperand (hundef, UndefinedType, &isfree, id, &testSig, &instent, oppcas);
  if ( EM->getOperatorClass (testSig) != variabl)
    {
      ptexpcal idselector = EM->mkIdentExp (string2ident ("_case_selector"), tsignal, AnyTypeCreate (), NM->mk_node_undefined (), hundef,  MK_BACK_ANNOT(NILTREE));
      if (contextClock == hundef) vh = CKM->new_variFormulaRootClock (); else vh = contextClock;
      EM->setClock (idselector, vh);
      vh = igc->createEquation (idselector, testSig, vh, vh, instent, MK_BACK_ANNOT(currentTree));
      testSig = idselector;
    }
  lval = NILLINSTANCES;
  helse = hundef;
  hleft = helse;
  hresult = the_nullClock_K2G_copy;
  while (kl_ptlcase_get_next (&l_case, &val, &vexpr))
    {
      NEW_BACK_ANNOT(kl_getAbstractTree (vexpr));
      isfree = pk_true;
      instent1 = NILLINSTANCES;
      h1 = instIntegerScalarListExpression (contextClock, &isfree, val, &lexp, &instent1);
      h1 = CKM->makeSynchro (contextClock, h1);
      vhsub = the_nullClock_K2G_copy;
      while (EM->getNaryArgs (&lexp, &valInst))
	{
	  if (!ptlinstances_isin_insert_exp (&lval, &vdummy, valInst))
	    pK_messenger->errorTree (errDuplCaseValue, currentTree);
	  condExpr = EM->mkBinaryExp (opegal, testSig, valInst, back_annot_obj);
	  h = h1;
	  h2 = instargwhen1_setboolwhen1 (contextClock, &h, &condExpr);
	  // Tester h2 == h1, h2 == the_nullClock_K2G_copy 
	  vhsub = CKM->makeSup (h2, vhsub);
	  {
	    condExpr = EM->mkUnaryExp (opnon, condExpr, back_annot_obj);
	    h = h1;
	    h2 = instargwhen1_setboolwhen1 (contextClock, &h, &condExpr);
	    helse = CKM->makeInf (helse, h2);
	  }
	}
      // To facilitate the computation of exclusive clocks.
      vhsub = CKM->makeInf (vhsub, hleft);
      h2 = instExprCase (vexpr, vhsub, pk_true);
      hresult = CKM->makeSup (hresult, h2);
      hleft = helse;
    }
  // else part
  if (pelse)
    {
      helse = CKM->makeInf (helse, contextClock);
      h2 = instExprCase (p_else, helse, pk_true);
      hresult = CKM->makeSup (hresult, h2);
    }
  currentTree = savetree;
  back_annot_obj = save_back_annot_obj;
  return hresult;
}


/**
   Compiling 
   - a "case" structure (local to \ref  instCase method), when bInCase
   - a state of an automaton, when not bInCase.
*/
ptinsth
kernel2graph::instExprCase (ptklExprProc eCase, ptinsth vhsub, boolean bInCase)
{
  ptinsth vhp, vh;
  ptlrepid inFlows, outFlows;

  if (CKM->satisfiesNullClock (vhsub)) return vhsub;
  inFlows = kl_getInputs(eCase);
  outFlows = kl_getOutputs( eCase );
  pushContextCase(inFlows, outFlows, vhsub, &vhp);
  if (bInCase) vh = compilingEproc (eCase); else vh = instDefinition (eCase);
  popContextCase(inFlows, outFlows, vhp);
  return vh;
}

/**
   Compiling "case" structure or controlled process.
   Remove the undersampling of flows of the list l. (set previuously by Undersampling())
*/
void
kernel2graph::popUndersampling (ptlrepid l)
{
  int n, i;
  ptexpcal currentIn;

  n = ptlrepid_size (l);
  for (i = 1; i <= n; i++)
    {
      if (isFlowId (l))
	{
	  currentIn = getCurrentInstance (l);
	  if (!EM->isStateVar (currentIn))
	    kl_popInstance (nameId (l));
	  else
	    EM->removeAttribute (currentIn, REF_STATEVAR_ATTR);
	}
      l = succid (l);
    }
}

/** Compiling "case" structure (local to \ref instExprCase method).
    Undersampling of flows of the list l.
*/
void
kernel2graph::pushUndersampling (ptlrepid l, ptinsth vhsub)
{
  ptinsth vh;
  int n, i;
  ptexpcal vexp, currentIn, nv;
  ptlinstances instent;

  n = ptlrepid_size (l);
  for (i = 1; i <= n; i++)
    {
      currentIn = getCurrentInstance (l);
      nv = EM->mkIdentCopy (currentIn, NM->mk_node_undefined (), hundef);
      if (! EM->isStateVar (currentIn))
	{
	  EM->setSharedSignal (nv, pk_false); // added Dec 2016 LBesnard.
	  setCurrentInstance (l, nv);
	  vh = CKM->makeInf (igc->setClockVariable (currentIn), vhsub);
	  if (CKM->satisfiesNullClock (vh))
	    vh = igc->getSignalClockSynchronizedWith (nv, the_nullClock_K2G_copy);
	  else
	    {
	      instent = ptlinstances_push_front (NILLINSTANCES, currentIn);
	      vh = igc->createEquation (nv, currentIn, vh, vh, instent, back_annot_obj);
	    }
	}
      else
	{
	  // x1:= x? when vhsub is created. x? will be replaced by x1 in the expression (using the attribute REF_STATEVAR_ATTR).
	  EM->setStateTag (nv, pk_false);
	  vexp = EM->mkUnaryExp (opvvarin, currentIn, back_annot_obj);
	  vh = igc->createEquation (nv, vexp, vhsub, vhsub, NILLINSTANCES, back_annot_obj);
	  EM->addAttribute (currentIn, REF_STATEVAR_ATTR, (attributeValue) nv);
	}
      l = succid (l);
    }
}



/** Compiling an expression of process (local to \ref compilingEproc method).
    @param vklExprProc (in): an expression of process: equation, partial equation, process call, iterate, case....
    @return the clock defined by the upper bound of the expression of the compiled expression.
 */
ptinsth
kernel2graph::instDefinition (ptklExprProc vklExprProc)
{
  ptklExprProc vklExprProc1, vklExprProc2;
  ptinsth vUpperBoundClock, vh1;
  noperators code;
  ptklExprSig klExprSig1, klExprSig2;
  boolean save_inAssertBlock;
  ptexpcal x;
  ptlinstances l;

  save_inAssertBlock = inAssertBlock;
  if (!inAssertBlock) inAssertBlock = kl_isCompositionAssertion (vklExprProc);

  code = kl_getExprProcType (vklExprProc);
  switch ((int) code)
    {
    case opnullproc:
    case opnil:
      vUpperBoundClock = hundef;
      break;

    case opdefpartdef:
    case opdefpart:
    case opdefsig:
      kl_getEquation (vklExprProc, &klExprSig1, &klExprSig2, &currentTree);
      NEW_BACK_ANNOT(currentTree);
      if (kl_getFlowExprClassType (klExprSig2) == call)
	vUpperBoundClock = instCallDefinition (klExprSig1, klExprSig2, code);
      else
	vUpperBoundClock = instSimpleOrTupleDefinition (klExprSig1, klExprSig2, code);
      break;

    case opappe:		/* P{p1,,...pn} (x1,...,xm) */
      vUpperBoundClock = instappe (vklExprProc, NILEXPCAL, NILEXPCAL, &x, &l);
      break;

    case opinst:		/* P{p1,....pn} */
      pushParameters (vklExprProc);
      vUpperBoundClock = instmacro (vklExprProc);
      break;

    case oplfeed:		/* composition */
      kl_getComposition (vklExprProc, &vklExprProc1, &vklExprProc2, &currentTree);
      NEW_BACK_ANNOT(currentTree);
      vh1 = compilingEproc (vklExprProc1);
      vUpperBoundClock = compilingEproc (vklExprProc2);
      vUpperBoundClock = CKM->makeSup (vh1, vUpperBoundClock);
      break;

    case opmoti:		/* array (from H2 version) */
      vUpperBoundClock = instArrayOfProcesses (vklExprProc);
      break;

    case opiter:		/* iterate */
      vUpperBoundClock = instIterateProcess (vklExprProc);
      break;

    case oppcas:		/* case */
      vUpperBoundClock = instCase (vklExprProc);
      break;

    case oplequhor:		/*  "^=" (egalite) */
      vUpperBoundClock = instClockEqu (vklExprProc);
      break;

    case oplcrohor:		/* ^<  */
      vUpperBoundClock = instClockIncreasing (vklExprProc);
      break;

    case opldechor:		/* ^> */
      vUpperBoundClock = instClockDecreasing (vklExprProc);
      break;

    case oplexchor:		/* ^# (exclusive clocks) */
      vUpperBoundClock = instClockExlcusive (vklExprProc);
      break;

    case opdepd:		/* Dependences x1,...xn -->y1,...yp when H */
      vUpperBoundClock = instDependences (vklExprProc);
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
      vUpperBoundClock = hundef;
    }

  inAssertBlock = save_inAssertBlock;
  return vUpperBoundClock;
}

/** 
    It manages the process expression with label.
    When a label is associated with an expression, 
      - it is pushed to the stack of labels (this stack is managed in igc object) 
      - each object created during the compiling of the expression will be a node/clock of a sub-graph.
      The popLabel() must be called to complete such a sub graph.

   @param vklExprProc (in) a process expression.
   @param node (out) the node associated with the label (instance of vklExprProc), if it exists.
   @param vlabel (out) the name of the label (instance), if it exists.
   @param blabel (out) pk_true when expression is labelled and it is not a state of an automaton, false otherwise.
   @return 
   - "true" if the process expression is a state of an automaton
   - "false" otherwise
*/
boolean
kernel2graph::pushLabel (ptklExprProc vklExprProc, ptINode * node, ptexpcal * vexpcal, boolean * blabel)
{
  ptklExprSig etiq;
  ptinsth clkLabel;
  ptlrepid vrepid;
  boolean bres;
  
  *blabel = pk_false;
  if (!kl_getLabelExpr (vklExprProc, &etiq)) return pk_false;
  vrepid = kl_getIdentifierExpr (etiq);
  (*vexpcal) = getCurrentInstance (vrepid);
  clkLabel = igc->setClockVariable (*vexpcal);

  if (ptrepproto_isLabelStateAutomaton (getIdRepid (vrepid)))
    {
      // labelled expression, referenced in a transition: vklExprProc is a state of an automaton.
      // Not allowed now, use on S :: (| ...|)
      pK_messenger->errorTree (errStateAsLabelNotAllowed, currentTree);
      *node = NILINODE;
      bres = pk_true;
    }
  else
    {
      // labelled expression, vklExprProc is not a state of an automaton.
      *blabel = pk_true;
      *node = NM->mk_node_subgraph (*vexpcal, back_annot_obj);
      bres = pk_false;
    }
  igc->pushLabel (*node, clkLabel);	// node is NILINODE for a state of an automaton.( lab :: STATE() )
  return bres;
}

/**
   @return "true" if a kernel tree representation (vklExprProc) is a reference to a model (vfi), "false" otherwise.
*/
boolean
kernel2graph::getModelExpfil (ptklExprProc vklExprProc, ptrepfi * vfi)
{
  noperators code;
  ptklExprSig klExprSig1, klExprSig2;
  ntrees vt;
  ptlrepid varid;
  boolean b;

  code = kl_getExprProcType (vklExprProc);
  if (code == opdefsig)
    {
      kl_getEquation (vklExprProc, &klExprSig1, &klExprSig2, &vt);
      if (kl_getFlowExprClassType (klExprSig2) != call)	return pk_false;
      *vfi = getModelFromExpflot (klExprSig2);
      return pk_true;
    }
  if (code == opappe)
    {
      vklExprProc = kl_getModelProcCall (vklExprProc, &klExprSig1, &vt);
      kl_getMacro (vklExprProc, &varid, &klExprSig1, &vt);
      *vfi = getModelOf (varid, &b);
      return pk_true;
    }
  if (code == opinst)
    {
      kl_getMacro (vklExprProc, &varid, &klExprSig1, &vt);
      *vfi = getModelOf (varid, &b);
      return pk_true;
    }
  return pk_false;
}

/**
   @return "true" if a kernel tree representation (vklExprProc) is synchronized by an other expression, "false" otherwise.
   It is the case for a model with the predefined pragma "Cluster". 
   Such a pragma is generated by Polychrony in abstractions for separated compiling.
boolean
kernel2graph::testSynchronizedCall (ptklExprProc vklExprProc)
{
  ptlpragmas lp;
  ptrepfi vfi;

  if (!getModelExpfil (vklExprProc, &vfi)) return pk_false;
  lp = kl_getPragmaModel (vfi);
  return existPredefinedPragma (lp, PRAGMA_CLUSTER);
}
*/

/** 
    @return "true" if the compiled expression (vklExprProc) is labelled, "false" otherwise.
    @param vklExprProc (in) a labelled expression of process.
    @param vh (in) the clock of the expression.
    When it is a labelled expression, it assumes that the pushLabel() method has been applied: 
    each object created during the compiling of the expression has been added to a sub-graph.
    When the expression is a reference to a "cluster abstraction", the clock of the sub graph is given by vh.
    Otherwise, the clock of the sub graph is the upper bound of vh and the clock of the inputs/outputs signals of the sub-graph.    
*/
boolean
kernel2graph::popLabel (ptklExprProc vklExprProc, ptinsth vh)
{
  ptklExprSig etiq;
  ptexpcal vlabel;
  ptElemStack node;
  ptlrepid entree, sortie, par;
  ptlinstances lent, lsort;
  ptinsth vh1;
  ptINode aSubGraph;
  ptlpragmas lp;
  ptrepfi vfi;
  boolean bClusterOrTransition;

  if (!kl_getLabelExpr (vklExprProc, &etiq)) return pk_false;
  if (!igc->topLabel (&node)) return pk_false;
  aSubGraph = (ptINode) node;

  if (aSubGraph == NILINODE) // It is associated with a state of an automaton.
    {
      // removed Sept 2015 LBesnard.
      // vlabel = getCurrentInstance (kl_getIdentifierExpr (etiq));
      // vh = igc->getSignalClockSynchronizedWith (vlabel, vh);
      igc->popLabel ();
    }
  else
    {
      if (getModelExpfil (vklExprProc, &vfi))
	{
	  lp = kl_getPragmaModel (vfi);
	  if (existPredefinedPragma (lp, PRAGMA_CLUSTER))
	    bClusterOrTransition = true;
	  else
	    bClusterOrTransition = kl_isIntrinsicModel (vfi) && isTransitionId (kl_getNameModel(vfi));
	}
      else
	bClusterOrTransition = false;
      
      NM->get_label (aSubGraph, &vlabel);
      if (bClusterOrTransition)
	{
	  vh = igc->getSignalClockSynchronizedWith (vlabel, vh);
	}
      else
	{
	  /* The clock of the label is the upper bound of the clock of the I/O signals.
	     NB: - The clock (vh) upper bound of the clock of all the signals of the subgraph cannot be the clock of the label (for oversampling).
	     - Only the clock of assigment for a partial definition must be considered for the clock of the label.
	   */
	  kl_getInsOutsParams (vklExprProc, &entree, &sortie, &par);
	  lent = getCurrentInstances (entree, NILLINSTANCES, pk_true);
	  lsort = getCurrentInstances (sortie, NILLINSTANCES, pk_false);
	  vh1 = CKM->makeSup (igc->upperBoundClockOfIdentifiers (lent), igc->upperBoundClockOfIdentifiers (lsort));
	  // vh = CKM->makeSup(vh, vh1); vh = CKM->makeSynchro(vh, igc->getSignalClockSynchronizedWith(vlabel, vh));  
	  ptlinstances_free (lent);
	  ptlinstances_free (lsort);
	}

      NM->insertNode ((ptINode) node, vh);
      NM->set_subGraphProjectedSystemInfos (aSubGraph, vh, NILLINSTH);	/* revoir */
      NM->unlinkNodesFromClock (aSubGraph);
      igc->popLabel ();
      igc->takeIntoAccountContexts (aSubGraph);
    }
  return pk_true;
}

/** Compiling an expression of process.
    @param vklExprProc (in): an expression of process (equation, partial equation, process call, iterate, case....).
    It can have local declarations (where).
    @return the clock defined by the upper bound of the expression of the compiled expression.
 */
ptinsth
kernel2graph::compilingEproc (ptklExprProc vklExprProc)
{
  ntrees savetree;
  ptinsth vhContext, vhp, vhEproc;
  ptlrepid const_typ_loc, sig_loc, llabels;
  ptINode vlabelNode;
  ptpragma runon;
  boolean bControlldedProcess, bStateAutomaton, blabel, brunon;
  ptexpcal vlabelname;
  ptident idmodel, ids, id;
  BackAnnotationType save_back_annot_obj;

  if (vklExprProc == NILEXPFIL) return hundef;
  if (kl_getExprProcType (vklExprProc) == opexte) return hundef;

  savetree = currentTree;
  save_back_annot_obj = back_annot_obj;

  bStateAutomaton = pushLabel (vklExprProc, &vlabelNode, &vlabelname, &blabel);
  bControlldedProcess = pushControlClock(vklExprProc, &vhp);
  brunon = pk_false;
  if (blabel)
    {
      brunon = pushPragma (NM->getPragmas (vlabelNode), PRAGMA_RUN_ON, &runon);
      id = EM->getOriginalName (vlabelname);
      idmodel = nameManager->setCurrentModelName (id);
      ids = nameManager->setContextSubExpr (newIdent_concIdentIdent (id, getStrVar ()));
    }
  /* Where */
  kl_getLocals (vklExprProc, &const_typ_loc, &sig_loc, &llabels);
  compiling_locals (const_typ_loc, sig_loc, llabels);
  if (bStateAutomaton)
    {
      vhContext = hundef;
      vhEproc = instExprCase (vklExprProc, mkStateAutomatonUndersamplingClock (vlabelname, vhContext), pk_false);
      vhEproc = CKM->makeSynchro (vhEproc, vhContext);
    }
  else
    vhEproc = instDefinition (vklExprProc);
  /* Postlude */
  popLabel (vklExprProc, vhEproc);
  if (bControlldedProcess) popControlClock(vklExprProc, vhp);

  if (brunon) igc->setRunOnPragma (runon);
  if (blabel)
    {
      propagatePragmas (vklExprProc, vlabelname);
      nameManager->setCurrentModelName (idmodel);
      nameManager->setContextSubExpr (ids);
    }
  compiling_locals_finalize (const_typ_loc, sig_loc, llabels);
  currentTree = savetree;
  back_annot_obj = save_back_annot_obj;
  return vhEproc;
}

boolean kernel2graph::pushControlClock(ptklExprProc vklExprProc, ptinsth *PreviousClockContext)
{
  ptklExprSig vOnExprFlot;
  ptinsth vhon;
  boolean vhonFixed;

  if (! kl_getControlExpr(vklExprProc, &vOnExprFlot)) return pk_false;
  vhonFixed = pk_false;
  if (getTypeExpflot(vOnExprFlot) == variabl)  // on S:: (|  |) and S is the name of a state.
    {
      ptlrepid vrepid = kl_getIdentifierExpr (vOnExprFlot);
      if ( ptrepproto_isLabelStateAutomaton (getIdRepid (vrepid)))
	{
	  ptexpcal vlabelname = getCurrentInstance ( vrepid );
	  vhon = mkStateAutomatonUndersamplingClock (vlabelname, hundef /* ? */);
	  vhonFixed = pk_true;
	}
    }
  if (! vhonFixed ) vhon = instevent (hundef, vOnExprFlot);
  pushContextCase(kl_getInputs(vklExprProc), kl_getOutputs(vklExprProc), vhon, PreviousClockContext ); 
  return pk_true;
}


void kernel2graph::popControlClock(ptklExprProc vklExprProc, ptinsth PreviousClockContext)
{
 ptklExprSig vOnExprFlot;

 if (kl_getControlExpr(vklExprProc, &vOnExprFlot))
   popContextCase(kl_getInputs(vklExprProc), kl_getOutputs( vklExprProc ), PreviousClockContext);
}

void  kernel2graph::pushContextCase( ptlrepid inFlows, ptlrepid outFlows, ptinsth vhon, ptinsth *PreviousContextClock)
{
  pushUndersampling (inFlows, vhon);
  igc->pushContextClock (vhon);
  *PreviousContextClock = igc->pushCaseContextClock (vhon);
}


void  kernel2graph::popContextCase( ptlrepid inFlows,  ptlrepid outFlows,  ptinsth PreviousContextClock)
{
  igc->popContextClock ();
  igc->pushCaseContextClock (PreviousContextClock);
  popUndersampling (inFlows);
}

void
kernel2graph::pushCellAt(ptlrepid l, ptinsth vhCell)
{
  ptinsth vh;
  int n, i;
  ptexpcal currentIn, nv;
  ptlinstances instent;

  n = ptlrepid_size (l);
  for (i = 1; i <= n; i++)
    {
      currentIn = getCurrentInstance (l);
      nv = EM->mkIdentCopy (currentIn, NM->mk_node_undefined (), hundef);
      setCurrentInstance (l, nv);
      instent = ptlinstances_push_front (NILLINSTANCES, currentIn);
      vh = igc->createEquation (nv, currentIn, igc->mkClockInstance(currentIn), vhCell, instent,  MK_BACK_ANNOT(NILTREE));
      l = succid (l);
    }
}

