/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
   It adds (head) the current instance of a identifier (vexp) to a list of expression (lsig)
   @return the result list.
*/
ptlinstances
kernel2graph::addSignalToList (ptklExprSig vexp, ptlinstances lsig)
{
  ptlinstances instent;
  ptexpcal vsig;
  boolean isfree;
  ptinsth vh;
  TypeType ContextType;

  isfree = pk_false;
  instent = NILLINSTANCES;
  ContextType = UndefinedType;
  vh = instOperand (hundef, ContextType, &isfree, vexp, &vsig, &instent, opdepd);
  if (vsig == vexptrue)
    getCurrentInstance (kl_getIdentifierExpr (vexp), 1, tsignal, &vsig);
  return ptlinstances_insert (lsig, vsig);
}

/**
   @return the list of the instanciated signals  of the operand list (vexp)
   @param vexp (in) list of flows or labels (dependence, object of a pragma)
*/
ptlinstances
kernel2graph::instSignalsLabels (ptklExprSig vexp)
{
  ptlinstances lsig;
  ptklExprSig varg;

  lsig = NILLINSTANCES;
  if (vexp != NILEXPRSIG)
    {
      if (kl_getFlowExprClassType (vexp) == variabl)
	lsig = addSignalToList (vexp, lsig);
      else
	// list : restriction a des signaux/etiquettes
	while (vexp != NILEXPRSIG)
	  {
	    varg = kl_getNextExpr (&vexp);
	    if (kl_getFlowExprClassType (varg) != variabl)
	      {
		pK_messenger->errorTree (errnotimplemented, currentTree);
	      }
	    else
	      lsig = addSignalToList (varg, lsig);
	  }
    }
  return lsig;
}


/**
   @return the list of the instanciated signals  of the operand list (vexp)
   @param vexp (in) list of flows or labels (dependence, object of a pragma)

   \note: unreferenced.

*/
ptexpcal
kernel2graph::mkSignalsLabelsExp (ptklExprSig vexp)
{
  ptexpcal vsig, lsig;
  ptklExprSig varg;
  boolean isfree;
  ptlinstances instent;

  if (vexp == NILEXPRSIG) return NILEXPCAL;

  varg = kl_getNextExpr (&vexp);
  lsig = mkSignalsLabelsExp (vexp);
  if (kl_getFlowExprClassType (varg) != variabl)
    pK_messenger->errorTree (errnotimplemented, currentTree);
  isfree = pk_false;
  instent = NILLINSTANCES;
  instOperand (hundef, UndefinedType, &isfree, varg, &vsig, &instent, oplopra);
  if (vsig == vexptrue)
    getCurrentInstance (kl_getIdentifierExpr (vexp), 1, tsignal, &vsig);
  return EM->mkNaryExp (oplesign, vsig, lsig, back_annot_obj);
}


/**
   @return the list of the instanciated signals  of the operand list (vexp)
   @param vexp (in) list of flows or labels (dependence, object of a pragma)

   \note: unreferenced.
*/
void
kernel2graph::verifSignalsLabels (ptklExprSig vexp)
{
  ptklExprSig varg;

  /*list : restriction a des signaux/etiquettes */
  while (vexp != NILEXPRSIG)
    {
      varg = kl_getNextExpr (&vexp);
      if (kl_getFlowExprClassType (varg) != variabl)
	{
	  pK_messenger->errorTree (errnotimplemented, currentTree);
	}
    }
}

/**
   @return "false " if the operand expression (vali) is undefined, "true" otherwise.
   @param vali expression to compile, assumed to be an expression of initialisation.
   @param vexp current instance of the vali expression
   @param vh (out) clock of the expression.
*/
boolean
kernel2graph::instExprInit (ptklExprSig vali, ptexpcal * vexp, ptinsth * vh)
{
  ptlinstances instent;
  boolean isfree, b;
  TypeType ContextType;
  noperators op;

  *vh = hundef;
  if (vali == NILEXPRSIG)
    {
      *vexp = NILEXPCAL;
      return pk_false;
    }
  op = kl_getFlowExprOperator (vali);
  if (op == opexte)
    {
      *vexp = NILEXPCAL;
      return pk_false;
    }
  isfree = pk_true;
  instent = NILLINSTANCES;
  b = igc->getSetConstantContext (pk_true);
  ContextType = UndefinedType;
  if (getTypeExpflot (vali) == oper2)
    // to avoid to create identifiers
    *vh = instoper2 (hinit, ContextType, &isfree, vali, vexp, &instent);
  else
    *vh = instOperand (hinit, ContextType, &isfree, vali, vexp, &instent, op);
  b = igc->getSetConstantContext (b);
  if ((*vh) == hinit)
    *vh = hundef;
  return pk_true;
}


/**
   @return the instanciated constant expression denoted by the operand expression (pt).
*/
ptexpcal
kernel2graph::instconst (ptklExprSig pt)
{
  noperators operateur;

  operateur = kl_getFlowExprOperator (pt);
  return EM->mkConstantExp (operateur, kl_getConstExpr (pt));
}

boolean kernel2graph::getClockExprWhen (ptexpcal vexpcal, boolean * isfree, ptinsth * vh)
{
  noperators op;
  ptexpcal arg;

  op = EM->getOperator (vexpcal);
  if (op != opextr) return pk_false;
  arg = EM->getWhenOrCellArg (vexpcal);
  if (EM->getOperatorClass (arg) != constant) return pk_false;
  *isfree = pk_false;
  (*vh) = EM->getContributionClock (vexpcal);
  return pk_true;
}


/**
   It instanciates the name vklExprSig in vexpcal expression.
   @return the clock of the signal ; a new clock is associated with this signal if it is undefined.
   The undefined clock is retuned when there is no signal associated with the kernel-tree expression (vklExprSig).
   @param vklExprSig (in) a kernel-tree identifier;
   @param vexpcal (out) the instanciated expression (name) of vexflot;
   @param instent (in/out) vexpcal is added to the instent (in) list
   @param isfree (in/out) if "true" vhcste must be used to fix the clock of the expression. It is set to "false" value
   when the expression has its own clock.
   @param vhcste (in)  clock context to fix the clock of the constant, parameter.
*/
ptinsth
kernel2graph::instsigin (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcal, ptlinstances * instent)
{
  ptinsth vhCaseContext, vh;

  if (!getCurrentInstance (kl_getIdentifierExpr (vklExprSig), 1, tsignal, vexpcal)) return hundef;

  if (!getClockExprWhen (*vexpcal, isfree, &vh))
    {
      vh = igc->getSetIdClock (*vexpcal, isfree, vhcste, back_annot_obj, instent);
      // Added Dec 2016 LBesnard (for signal that are IN/OUT in a system. (x ::= E | u := f(x) | ....) , 
      // x is a IO flow, but not managed currently in the Trees2kernel translation...
      if ( EM->isSharedSignal(*vexpcal))
	{ 
	  vhCaseContext = igc->getCaseContextClock();
	  if (! CKM->satisfiesUndefClock (vhCaseContext)) vh = CKM->makeInf (vhCaseContext, vh);
	}
    }
  return vh;
}

ptinsth
kernel2graph::instCall (ptklExprSig vklExprSig, ptexpcal l, ptexpcal vexp, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  return instappe (kl_getFlowExprModelProcCall (vklExprSig), l, vexp, vexpcalcul, instent);
}


/** Recursive call for an expression. 
    @param vklExprSig (in) expression to instanciate.
    @param ContextType (in) type for added signals during the compiling of vklExprSig (context type).
    @param vhcste (in) clock of the constants (context clock).
    @param isfree (in/out) at pk_true (in) if constants are allowed, at pk_true (out) if the expression is a constant one.
    @param vexpcalcul (out) instanciated expression of vklExprSig.
    @param instent (in/out) instent (in) is completed with the list of signals referenced in vexpvalcul.
    @return clock of the operand (can be undefined clock).
*/
ptinsth
kernel2graph::instOperand (ptinsth vhcste, TypeType ContextType, boolean * isfree, ptklExprSig vklExprSig,
			   ptexpcal * vexpcalcul, ptlinstances * instent, noperators fatherOp)
{
  ptinsth vhresult;
  ptinsth vh;
  noperators operateur;
  noperators contextId;

  operateur = kl_getFlowExprOperator (vklExprSig);


  switch ((int) (getTypeExpflot (vklExprSig)))
    {
    case call:
      if (!instExternalFunction (vklExprSig, vhcste, isfree, &vhresult, vexpcalcul, instent))
	{
	  *isfree = pk_false;
	  vh = instCall (vklExprSig, NILEXPCAL, NILEXPCAL, vexpcalcul, instent);
	  *vexpcalcul = EM->getNaryArgs (vexpcalcul);
	  vhresult = igc->setClockVariable (*vexpcalcul);
	}
      break;

    case constant:
      *vexpcalcul = instconst (vklExprSig);
      vhresult = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
      break;
    case variabl:
      vhresult = instsigin (vhcste, isfree, vklExprSig, vexpcalcul, instent);
      break;
    case oper1:
      if (operateur == optran)
	{
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);
	  vhresult = instEquationOp1 (vhcste, isfree, *vexpcalcul, vklExprSig);
	}
      else
	vhresult = instoper1 (vhcste, ContextType, isfree, vklExprSig, vexpcalcul, instent);
      break;
    case oper1tp:
      *isfree = pk_false;
      if (operateur == ophnull)
	{
	  *vexpcalcul = vexptrue;
	  vhresult = the_nullClock_K2G_copy;
	}
      else if (operateur == opevnt)
	{
	  *vexpcalcul = vexptrue;
	  vhresult = instevent (vhcste, vklExprSig);
	}
      else if (operateur == opwhen1)
	{
	  *vexpcalcul = vexptrue;
	  vhresult = instwhen1 (hundef, pk_false, vklExprSig);
	}
      break;
    case oper2:
      if ((fatherOp != opsequ)
	  && ((operateur == opmmul) || (operateur == oprepe) || (operateur == oplevect)
	      || (operateur == oplelevec) || (operateur == oprest) || (operateur == opsequ)))
	{
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);
	  vhresult = instEquationOp2 (vhcste, isfree, *vexpcalcul, vklExprSig);
	}
      else
	vhresult = instoper2 (vhcste, ContextType, isfree, vklExprSig, vexpcalcul, instent);
      break;

    case oper2tp:
      contextId = nameManager->setContextSubExpr (operateur);
      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
      nameManager->setContextSubExpr (contextId);
      vhresult = instEquationOp2tp (vhcste, isfree, *vexpcalcul, vklExprSig);

      break;

    case oper3:
      contextId = nameManager->setContextSubExpr (operateur);
      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
      nameManager->setContextSubExpr (contextId);
      vhresult = instEquationOp3 (vhcste, isfree, *vexpcalcul, vklExprSig);

      break;

    case oper3tp:
      contextId = nameManager->setContextSubExpr (operateur);
      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
      nameManager->setContextSubExpr (contextId);
      vhresult = instEquationOp3tp (vhcste, isfree, *vexpcalcul, vklExprSig);
      //      *isfree = pk_false;
      break;

    case opertyp:
      *vexpcalcul = EM->mkExpType (mkInstType (NILLREPID, kl_getTypeExpr (vklExprSig)));
      vhresult = hundef;
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return vhresult;
}


/** @return true if vklExprSig is a reference to a state variable.
    @param vhcste (in) : constant constant clock.
    @param vexpcalcul (out): instance of input expression.
    @param vh (out) : clock of the expression.
*/
boolean
kernel2graph::instStateVariable (ptinsth vhcste, ptklExprSig varg, ptexpcal * vexpcalcul, ptinsth * vh)
{
  boolean isfree, b;
  ptlinstances l;
  ptexpcal vexpcal;
  ptklExprSig vklExprSig;

  if (kl_getFlowExprOperator (varg) != opvvarin) return pk_false;
  kl_getUnaryExpr (varg, &vklExprSig);
  if (getTypeExpflot (vklExprSig) != variabl)
    {
      pK_messenger->errorTree (errRefStateVar, currentTree);
      *vh = vhcste;
      vexpcal = NILEXPCAL;
      (*vexpcalcul) = EM->mkUnaryExp (opvvarin, vexpcal, back_annot_obj);
      return pk_true;
    }
  b = igc->getSetStateVarContext (pk_true);
  isfree = pk_true;
  l = NILLINSTANCES;
  *vh = instsigin (vhcste, &isfree, vklExprSig, &vexpcal, &l);
  if (EM->getAttribute (vexpcal, REF_STATEVAR_ATTR, (attributeValue *) & vexpcal))
    {
      (*vexpcalcul) = vexpcal;
      *vh = EM->getClock (vexpcal);
    }
  else if (!EM->isStateVar (vexpcal))
    {
      if (!EM->isMarked (vexpcal, INPUT_ADDED_FOR_UNEXPANDED))
	pK_messenger->errorTree (errRefStateVar, currentTree);
      (*vexpcalcul) = vexpcal;
    }
  else
    (*vexpcalcul) = EM->mkUnaryExp (opvvarin, vexpcal, back_annot_obj);
  b = igc->getSetStateVarContext (b);
  return pk_true;
}

/**
   @param vklExprSig (in) expression to instanciate; assumed to be a list of int scalar expressions.
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at "true" (out) if the expression is a constant one.
   @param vexpcalcul (out) instanciated expression of vklExprSig249.
   @param instent (in/out) instent (in) is completed with the list of signals referenced in vexpvalcul.
   @return clock of the synchronised operand (can be undefined clock).
*/
ptinsth
kernel2graph::instIntegerScalarListExpression (ptinsth vhcste, boolean * isfree,
					       ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vhResult;
  ptklExprSig v1klExprSig;
  ptexpcal vexpcal1;
  boolean locfree, isappel;
  ptinsth vh1, vh2;
  ptlinstances instent1;
  TypeType ContextType;

  ContextType = UndefinedType;
  if (_isalist_ (kl_getFlowExprOperator (vklExprSig)))
    v1klExprSig = kl_getNextExpr (&vklExprSig);
  else
    {
      v1klExprSig = vklExprSig;
      vklExprSig = NILEXPRSIG;
    }
  if (vklExprSig == NILEXPRSIG)
    if (getTypeExpflot (v1klExprSig) == call)
      {
	instent1 = NILLINSTANCES;
	if (instExternalFunction (v1klExprSig, vhcste, isfree, &vh1, &vexpcal1, &instent1))
	  *vexpcalcul = EM->mkNaryExp (oplesign, vexpcal1, back_annot_obj);
	else
	  {
	    *isfree = pk_false;
	    vh1 = instCall (v1klExprSig, NILEXPCAL, NILEXPCAL, vexpcalcul, &instent1);
	    *vexpcalcul = EM->concatenate (*vexpcalcul, NILEXPCAL, back_annot_obj);
	  }
	vhResult = igc->synchronizeIdentifiers (instent1);
	*instent = ptlinstances_sortedfusion (instent1, *instent);
      }
    else
      {
	vhResult = instOperand (hundef, ContextType, isfree, v1klExprSig, &vexpcal1, instent, oplesign);
	*vexpcalcul = EM->mkNaryExp (oplesign, vexpcal1, back_annot_obj);
      }
  else
    {
      isappel = (boolean) (getTypeExpflot (v1klExprSig) == call);
      if (isappel)
	{
	  locfree = pk_false;
	  instent1 = NILLINSTANCES;
	  if (instExternalFunction (v1klExprSig, vhcste, isfree, &vh1, &vexpcal1, &instent1))
	    {
	      isappel = pk_false;
	      vh1 = CKM->makeSynchro (vh1, igc->synchronizeIdentifiers (instent1));
	    }
	  else
	    {
	      vh1 = instCall (v1klExprSig, NILEXPCAL, NILEXPCAL, &vexpcal1, &instent1);
	      vh1 = igc->synchronizeIdentifiers (instent1);
	    }
	  *instent = ptlinstances_sortedfusion (instent1, *instent);
	}
      else
	{
	  locfree = pk_true;
	  vh1 = instOperand (hundef, ContextType, &locfree, v1klExprSig, &vexpcal1, instent, oplesign);
	}
      if (locfree)
	{
	  vh2 = instIntegerScalarListExpression (hundef, isfree, vklExprSig, vexpcalcul, instent);
	  vhResult = igc->instop2syncfree (vhcste, vh1, vh2, *isfree);
	}
      else
	{
	  locfree = pk_true;
	  vh2 = instIntegerScalarListExpression (vh1, &locfree, vklExprSig, vexpcalcul, instent);
	  vhResult = CKM->makeSynchro (vh1, vh2);
	  *isfree = pk_false;
	}
      if (isappel)
	*vexpcalcul = EM->concatenate (vexpcal1, *vexpcalcul, back_annot_obj);
      else
	*vexpcalcul = EM->mkNaryExp (oplesign, vexpcal1, *vexpcalcul, back_annot_obj);
    }
  return vhResult;
}



/**
   @param vklExprSig (in) expression to instanciate; assumed to be a not expression.
   @param vhcste (in) clock of the constants (context clock).
   @param ContextType (in) type for added signals during the compiling of the expression (context type).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at "true" (out) if the expression is a constant one.
   @param vexpcalcul (out) instanciated expression of vklExprSig.
   @param instent (in/out) instent (in) is completed with the list of signals referenced in vexpvalcul.
   @return clock of the synchronised operand (it can be undefined clock).
*/
ptinsth
kernel2graph::instnot (ptinsth vhcste, TypeType ContextType, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul,
		       ptlinstances * instent)
{
  ptinsth vh;
  kl_getUnaryExpr (vklExprSig, &vklExprSig);
  vh = instOperand (vhcste, ContextType, isfree, vklExprSig, vexpcalcul, instent, opnon);
  *vexpcalcul = EM->mkUnaryExp (opnon, *vexpcalcul, back_annot_obj);
  return vh;
}


/**
   @param vklExprSig (in) expression to instanciate; assumed to be a unary expression (not, unary minus,...)
   @param ContextType (in) type for added signals during the compiling of the expression (context type).
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at "true" (out) if the expression is a constant one.
   @param vexpcalcul (out) instanciated expression of vklExprSig.
   @param instent (in/out) instent (in) is completed with the list of signals referenced in vexpvalcul.
   @return clock of the synchronised operand (it can be undefined clock).
*/
ptinsth
kernel2graph::instoper1 (ptinsth vhcste, TypeType ContextType, boolean * isfree, ptklExprSig vklExprSig,
			 ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vh;
  ptexpcal vexpcal1;
  noperators op;
  ptklExprSig vklExprSig1;
  ptlinstances instent1;

  if (instStateVariable (vhcste, vklExprSig, vexpcalcul, &vh)) return vh;
  op = kl_getFlowExprOperator (vklExprSig);
  if (op == opnon)
    return instnot (vhcste, UndefinedType, isfree, vklExprSig, vexpcalcul, instent);
  kl_getUnaryExpr (vklExprSig, &vklExprSig1);

  if (op == opsdec)		// x[?]
    {
      instent1 = NILLINSTANCES;
      vh = instOperand (vhcste, UndefinedType, isfree, vklExprSig1, &vexpcal1, &instent1, op);
      *vexpcalcul = EM->mkUnaryExp (op, vexpcal1, back_annot_obj);
      igc->linkToCurrentIndexArray (*vexpcalcul);
      *instent = ptlinstances_insert (*instent, *vexpcalcul);
      return vh;
    }
  // opnega (- unaire) , optran (tranposee de matrice), Cartesian product
  if (op == oppcar)		// Cartesian product
    vh = instIntegerScalarListExpression (vhcste, isfree, vklExprSig1, &vexpcal1, instent);
  else
    {
      if (op != opnega)
	ContextType = UndefinedType;
      vh = instOperand (vhcste, ContextType, isfree, vklExprSig1, &vexpcal1, instent, op);
    }
  *vexpcalcul = EM->mkUnaryExp (op, vexpcal1, back_annot_obj);
  return vh;
}



/**
   Local to instoper2() method.
   @param vklExprSig (in) expression to instanciate; assumed to be a binary expression (selected in instoper2() method).
   @param ContextType (in) type for added signals during the compiling of the expression (context type).
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at "true" (out) if the expression is a constant one.
   @param vexpcal1 (out) instanciated expression of the first arg of  the compiled expression.
   @param vexpcal2 (out) instanciated expression of the second arg of   the compiled expression.
   @param instent (in/out) instent (in) is completed with the list of signals referenced in vexpvalcul.
   @return clock of the synchronised operand (it can be undefined clock).
*/
ptinsth
kernel2graph::instoper2_instcal (ptinsth vhcste, TypeType ContextType, ptklExprSig vklExprSig, boolean * isfree,
				 /* ptexpcal * vexpcal1, ptexpcal *vexpcal2, */ ptlinstances * instent, ptexpcal * vexpcalcul)
{
  ptklExprSig v1klExprSig, v2klExprSig;
  ptinsth hres, vhh, vh1, vh2;
  boolean locfree1, locfree2, b;
  TypeType ContextType1, ContextType2;
  noperators op;
  ptexpcal vexpcal1, vexpcal2;
  int tcOp1;
  
  op = kl_getFlowExprOperator (vklExprSig);
  switch ((int) (op))
    {

    case opet:
    case opou:
    case opouex:
      ContextType1 = get_BooleanType ();
      ContextType2 = get_BooleanType ();
      break;

    case opdivi:
    case opmoin:
    case opmult:
    case opplus:
      ContextType1 = ContextType;
      ContextType2 = ContextType;
      break;

    case oppuis:
    case opmod:
      ContextType1 = ContextType;
      ContextType2 = get_IntegerType ();
      break;

    default:
      /* case opegeg: case opiieg: case opdiff: case opegal: case opinfe: case opiegal: case opsegal:
         case opsupe: case opconc: case oprecu: case opmmul: case oprepe: case opsequ:
      */
      ContextType1 = UndefinedType;
      ContextType2 = UndefinedType;
      break;
    }
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  locfree1 = pk_true;
  if (op == oprecu)
    {
      // The type checking is done but no errors for the bounds.
      b = EM->setTypeCheckingOutOfBounds (pk_false, &tcOp1);
      vh1 = instOperand (vhcste, ContextType1, &locfree1, v1klExprSig, &vexpcal1, instent, op);
      EM->setTypeCheckingOutOfBounds (b, &tcOp1);
    }
  else
    vh1 = instOperand (vhcste, ContextType1, &locfree1, v1klExprSig, &vexpcal1, instent, op);
  if (locfree1)
    {
      vh2 = instOperand (vhcste, ContextType2, isfree, v2klExprSig, &vexpcal2, instent, op);
      if (!*isfree) vh1 = vh2;
      hres = igc->instop2syncfree (vhcste, vh1, vh2, *isfree);
    }
  else
    {
      locfree2 = pk_true;
      vhh = instOperand (vh1, ContextType2, &locfree2, v2klExprSig, &vexpcal2, instent, op);
      *isfree = pk_false;
      hres = CKM->makeSynchro (vhh, vh1);
    }
  if ((op == oprecu) && (tcOp1 > 0))
    {
      if (tcOp1 == 1) *vexpcalcul = vexpcal1; else *vexpcalcul = vexpcal2;
    }
  else
    *vexpcalcul = EM->mkBinaryExp (op, vexpcal1, vexpcal2, back_annot_obj);
  return hres;
}


/**
   Local to instElemAccess() method.
*/
boolean
kernel2graph::instLeftElemOperand (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent,
				   ptinsth * vh)
{
  noperators op;
  noperators contextId;

  switch ((int) (getTypeExpflot (vklExprSig)))
    {
    case constant:
      return pk_false;

    case variabl:
      *vh = instsigin (vhcste, isfree, vklExprSig, vexpcalcul, instent);
      break;

    default:
      op = kl_getFlowExprOperator (vklExprSig);
      if (op == opsdec)
	*vh = instoper1 (hundef, UndefinedType, isfree, vklExprSig, vexpcalcul, instent);
      else if (op == opcham)
	*vh = instoperField (hundef, isfree, vklExprSig, vexpcalcul, instent);
      else
	{
	  if (!instStateVariable (vhcste, vklExprSig, vexpcalcul, vh))
	    {
	      contextId = nameManager->setContextSubExpr (op);
	      *vexpcalcul = igc->mkVariable (back_annot_obj, instent);
	      nameManager->setContextSubExpr (contextId);
	      *vh = instEquation (hundef, *vexpcalcul, vklExprSig, opdefsig);
	    }
	  *isfree = pk_false;
	}
    }

  return pk_true;
}

/** Compiling the element access expression (E[]).
    Local to instoper2() method.
*/
ptinsth
kernel2graph::instElemAccess (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vhResult;
  ptklExprSig v1klExprSig, v2klExprSig;
  ptexpcal vexpcal1;
  ptinsth vhh, vh2;
  boolean locfree;

  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  locfree = pk_true;
  if (kl_getFlowExprOperator (v1klExprSig) == opelem)
    vhh = instElemAccess (vhcste, &locfree, v1klExprSig, &vexpcal1, instent);
  else if (!instLeftElemOperand (vhcste, &locfree, v1klExprSig, &vexpcal1, instent, &vhh))
    {
      pK_messenger->errorTree (errnotimplemented, currentTree);
      return the_nullClock_K2G_copy;
    }

  if (locfree)
    {
      vh2 = instIntegerScalarListExpression (hundef, isfree, v2klExprSig, vexpcalcul, instent);
      vhResult = igc->instop2syncfree (vhcste, vhh, vh2, *isfree);
    }
  else
    {
      locfree = pk_true;
      vhResult = CKM->makeSynchro (vhh, instIntegerScalarListExpression (vhh, &locfree, v2klExprSig, vexpcalcul, instent));
      *isfree = pk_false;
    }
  *vexpcalcul = EM->mkBinaryExp (opelem, vexpcal1, *vexpcalcul, back_annot_obj);
  return vhResult;
}

/** Compiling the vectorial element expression.
    Local to instlelevecRec() and instEnsvec() methods.
*/
ptinsth
kernel2graph::instElevec (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vexpresult;
  ptexpcal vexpcal1;
  ptklExprSig v1klExprSig;
  boolean locfree;
  ptinsth vh1, vh2, vhcontext;
  TypeType ContextType;

  ContextType = UndefinedType;
  vhcontext = getUndefOrInitClock (vhcste);
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &vklExprSig);
  locfree = pk_true;

  vh1 = instIntegerScalarListExpression (vhcontext, &locfree, v1klExprSig, &vexpcal1, instent);
  if (locfree)
    {
      vh2 = instOperand (vhcontext, ContextType, isfree, vklExprSig, vexpcalcul, instent, opelevec);
      vexpresult = igc->instop2syncfree (vhcste, vh1, vh2, *isfree);
    }
  else
    {
      locfree = pk_true;
      vh2 = instOperand (vh1, ContextType, &locfree, vklExprSig, vexpcalcul, instent, opelevec);
      vexpresult = CKM->makeSynchro (vh1, vh2);
      *isfree = pk_false;
    }
  *vexpcalcul = EM->mkBinaryExp (opelevec, vexpcal1, *vexpcalcul, back_annot_obj);
  return vexpresult;
}



/**
   Local to instRestructuration() method.
*/
ptinsth
kernel2graph::instLindexes (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vh, vhcontext;
  TypeType ContextType;
  if (kl_getFlowExprOperator (vklExprSig) == opleflot)
    vh = instIntegerScalarListExpression (vhcste, isfree, vklExprSig, vexpcalcul, instent);
  else
    {
      ContextType = UndefinedType;
      vhcontext = getUndefOrInitClock (vhcste);
      vh = instOperand (vhcontext, ContextType, isfree, vklExprSig, vexpcalcul, instent, oplesign);
      *vexpcalcul = EM->mkNaryExp (oplesign, *vexpcalcul, back_annot_obj);
    }
  return vh;
}

/**
   Compiling the expression (vklExprSig) defined by I1,..,In : Expr (oprest operator).
*/
ptinsth
kernel2graph::instRestructuration (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vexpresult;
  ptexpcal vexpcal1;
  ptklExprSig vklExprSig1, vklExprSig2;
  boolean locfree;
  ptinsth vh1, vh2, vhcontext;
  TypeType ContextType;
  ContextType = UndefinedType;
  vhcontext = getUndefOrInitClock (vhcste);
  kl_getBinaryExpr (vklExprSig, &vklExprSig1, &vklExprSig2);
  locfree = pk_true;
  vh1 = instLindexes (vhcontext, &locfree, vklExprSig1, &vexpcal1, instent);
  if (locfree)
    {
      vh2 = instOperand (vhcontext, ContextType, isfree, vklExprSig2, vexpcalcul, instent, oprest);
      vexpresult = igc->instop2syncfree (vhcste, vh1, vh2, *isfree);
    }
  else
    {
      locfree = pk_true;
      vh2 = instOperand (vh1, ContextType, &locfree, vklExprSig2, vexpcalcul, instent, oprest);
      vexpresult = CKM->makeSynchro (vh1, vh2);
      *isfree = pk_false;
    }
  *vexpcalcul = EM->mkBinaryExp (oprest, vexpcal1, *vexpcalcul, back_annot_obj);
  return vexpresult;
}

/** Local to instEnsvec() method.
 */
ptinsth
kernel2graph::instEnsvec_parc (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul,
			       ptlinstances * instent, ptlrepid * indices)
{
  ptinsth vexpresult;
  ptklExprSig v1klExprSig, v2klExprSig;
  ptexpcal vexpcal1, vexpcal2, vexpcal3;
  ptinsth lvhcste;
  boolean locfree;
  ptlrepid vrb;
  TypeType ContextType;
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &vklExprSig);
  if (v1klExprSig == NILEXPRSIG)
    vexpcal1 = mkNewIndex (NILLREPID, NILEXPCAL);
  else
    {
      vrb = kl_getIdentifierExpr (v1klExprSig);
      vexpcal1 = mkNewIndex (vrb, NILEXPCAL);
      *indices = insertlist (vrb, *indices);
    }

  ContextType = UndefinedType;
  if (vklExprSig == NILEXPRSIG)
    *vexpcalcul = EM->mkBinaryExp (opparc, vexpcal1, NILEXPCAL, back_annot_obj);
  else
    {
      kl_getTernaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig, &vklExprSig);
      if (v1klExprSig == NILEXPRSIG)
	{
	  *vexpcalcul = ptexp0;	// NILEXPCAL;
	  lvhcste = hundef;
	}
      else
	{
	  locfree = pk_true;
	  lvhcste = instOperand (hundef, ContextType, &locfree, v1klExprSig, vexpcalcul, instent, opparc);
	  *isfree = (boolean) ((*isfree) && locfree);
	}
      if (v2klExprSig == NILEXPRSIG)
	vexpcal2 = NILEXPCAL;
      else
	{
	  locfree = pk_true;
	  lvhcste = CKM->makeSynchro (lvhcste, instOperand (lvhcste, ContextType, &locfree, v2klExprSig, &vexpcal2, instent, opparc));
	  *isfree = (boolean) ((*isfree) && locfree);
	}
      if (vklExprSig == NILEXPRSIG)
	vexpcal3 = ptexp1;
      else
	{
	  locfree = pk_true;
	  lvhcste = CKM->makeSynchro (lvhcste, instOperand (lvhcste, ContextType, &locfree, vklExprSig, &vexpcal3, instent, opparc));
	  *isfree = (boolean) ((*isfree) && locfree);
	}
      *vexpcalcul = EM->mkBinaryExp (opparc, vexpcal1, EM->mkTernaryExp (opborn, *vexpcalcul, vexpcal2, vexpcal3, back_annot_obj), back_annot_obj);
      if (*isfree)
	vhcste = CKM->makeSynchro (lvhcste, vhcste);
      else
	vhcste = lvhcste;
    }
  vexpresult = igc->getSignalClockSynchronizedWith (vexpcal1, vhcste);
  return vexpresult;
}

/** 
    Local to instensec method.
*/
ptinsth
kernel2graph::instEnsvec_lparc (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent,
				ptlrepid * indices)
{
  ptinsth vexpresult;
  ptklExprSig v1klExprSig;
  ptexpcal v1expcal;
  ptinsth vh1, vhh, vhcontext;
  boolean locfree;
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &vklExprSig);
  if (vklExprSig == NILEXPRSIG)
    {
      vexpresult = instEnsvec_parc (vhcste, isfree, v1klExprSig, &v1expcal, instent, indices);
      *vexpcalcul = EM->mkNaryExp (oplparc, v1expcal, back_annot_obj);
    }
  else
    {
      locfree = pk_true;
      vhcontext = getUndefOrInitClock (vhcste);
      vh1 = instEnsvec_parc (vhcontext, &locfree, v1klExprSig, &v1expcal, instent, indices);
      if (locfree)
	{
	  vhh = instEnsvec_lparc (vhcontext, isfree, vklExprSig, vexpcalcul, instent, indices);
	  vexpresult = igc->instop2syncfree (vhcste, vh1, vhh, *isfree);
	}
      else
	{
	  locfree = pk_true;
	  vhh = instEnsvec_lparc (vh1, &locfree, vklExprSig, vexpcalcul, instent, indices);
	  *isfree = pk_false;
	  vexpresult = CKM->makeSynchro (vhh, vh1);
	}
      *vexpcalcul = EM->mkNaryExp (oplparc, v1expcal, *vexpcalcul, back_annot_obj);
    }
  return vexpresult;
}


/**
   local to instlelevecRec() method.
*/
ptinsth
kernel2graph::instEnsvec (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vexpresult;
  ptexpcal vexpcal1;
  ptklExprSig v1klExprSig;
  ptinsth vh1, vhh, vhcontext;
  boolean locfree;
  ptlinstances linstbidon;
  ptlrepid indices;
  noperators op;
  TypeType ContextType = UndefinedType;
  vhcontext = getUndefOrInitClock (vhcste);
  op = kl_getFlowExprOperator (vklExprSig);
  if ((op == opelevec) || (op == opensvec))	// added Dec 2006 (LB)
    {
      kl_getBinaryExpr (vklExprSig, &v1klExprSig, &vklExprSig);
      linstbidon = NILLINSTANCES;
      indices = NILLREPID;
      locfree = pk_true;
      vh1 = instEnsvec_lparc (vhcontext, &locfree, v1klExprSig, &vexpcal1, instent, &indices);
      if (locfree)
	{
	  if (kl_getFlowExprOperator (vklExprSig) == opelevec)
	    vhh = instElevec (vhcontext, isfree, vklExprSig, vexpcalcul, instent);
	  else
	    vhh = instOperand (vhcontext, ContextType, isfree, vklExprSig, vexpcalcul, instent, opensvec);
	  vexpresult = igc->instop2syncfree (vhcste, vhh, vh1, *isfree);
	}
      else
	{
	  locfree = pk_true;
	  if (kl_getFlowExprOperator (vklExprSig) == opelevec)
	    vhh = instElevec (vh1, &locfree, vklExprSig, vexpcalcul, instent);
	  else
	    vhh = instOperand (vh1, ContextType, &locfree, vklExprSig, vexpcalcul, instent, opensvec);
	  vexpresult = CKM->makeSynchro (vhh, vh1);
	  *isfree = pk_false;
	}
      deplinst (indices);
      *vexpcalcul = EM->mkBinaryExp (opensvec, vexpcal1, *vexpcalcul, back_annot_obj);
    }
  else
    vexpresult = instOperand (vhcontext, ContextType, isfree, vklExprSig, vexpcalcul, instent, opensvec);
  return vexpresult;
}


/**
   @return the combed list of instanciated expressions of a list of expressions.
   @param list (in) the left part of a definition (signal, vector element).
*/
ptexpcal
kernel2graph::instlidentifiers (noperators op, ptklExprSig list, boolean * berror)
{
  ptexpcal vexpcal, vlst;
  ptklExprSig e;
  boolean b;
  if (list == NILEXPRSIG)
    {
      *berror = pk_false;
      return NILEXPCAL;
    }
  e = kl_getNextExpr (&list);
  instElemLeftPartDefinition (pk_false, e, &vexpcal, &b);
  if (b) *berror = pk_true;
  if (list == NILEXPRSIG) return EM->mkNaryExp (op, vexpcal, back_annot_obj);
  vlst = instlidentifiers (op, list, berror);
  return EM->mkNaryExp (op, vexpcal, vlst, back_annot_obj);
}

/**
   Compiling a reference to a string constant denoted by vepxflot.
   @return an instance of the operand expression.
*/
ptexpcal
kernel2graph::instString (ptklExprSig vklExprSig)
{
  ptexpcal vexpcal1;
  ptklExprSig v1klExprSig, v2klExprSig;
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  vexpcal1 = instconst (v1klExprSig);
  if (v2klExprSig == NILEXPRSIG) return EM->mkNaryExp (opchai, vexpcal1, back_annot_obj);
  return EM->mkNaryExp (opchai, vexpcal1, instString (v2klExprSig), back_annot_obj);
}

/**
   local to instlelevecRec() method.
*/
boolean
kernel2graph::instRestructurationSimple (ptklExprSig vklExprSig)
{
  noperators op;
  ptklExprSig v1, v2, varg;
  cloper x;

  op = kl_getFlowExprOperator (vklExprSig);
  if (op != oprest) return pk_false;
  // Looking at constant: E, or [constant or ident]:E
  kl_getBinaryExpr (vklExprSig, &v1, &v2);
  if (kl_getFlowExprClassType (v1) == constant) return pk_true;
  op = kl_getFlowExprOperator (v1);
  if (op != oplevect) return pk_false;
  varg = kl_getNextExpr (&v1);
  if (v1 != NILEXPRSIG) return pk_false;
  x = kl_getFlowExprClassType (varg);
  return pk_true;
}

/**
   Local to instlelevec() method.
*/
ptinsth
kernel2graph::instlelevecRec (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vexpresult;
  ptexpcal vexpcal1;
  ptklExprSig v1klExprSig, v2klExprSig;
  boolean locfree;
  ptinsth vh1, vh2, vhcontext;
  noperators op;
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  vhcontext = getUndefOrInitClock (vhcste);
  op = kl_getFlowExprOperator (v1klExprSig);
  if (v2klExprSig == NILEXPRSIG)
    {
      if ((op == opelevec) || instRestructurationSimple (v1klExprSig))
	vexpresult = instElevec (vhcontext, isfree, v1klExprSig, &vexpcal1, instent);
      else
	vexpresult = instEnsvec (vhcontext, isfree, v1klExprSig, &vexpcal1, instent);
      *vexpcalcul = EM->mkNaryExp (oplelevec, vexpcal1, back_annot_obj);
    }
  else
    {
      locfree = pk_true;
      if ((op == opelevec) || instRestructurationSimple (v1klExprSig))
	vh1 = instElevec (vhcontext, &locfree, v1klExprSig, &vexpcal1, instent);
      else
	vh1 = instEnsvec (vhcontext, &locfree, v1klExprSig, &vexpcal1, instent);
      if (locfree)
	{
	  vh2 = instlelevecRec (vhcontext, isfree, v2klExprSig, vexpcalcul, instent);
	  vexpresult = igc->instop2syncfree (vhcste, vh1, vh2, *isfree);
	}
      else
	{
	  locfree = pk_true;
	  vh2 = instlelevecRec (vh1, &locfree, v2klExprSig, vexpcalcul, instent);
	  vexpresult = CKM->makeSynchro (vh1, vh2);
	  *isfree = pk_false;
	}
      *vexpcalcul = EM->mkNaryExp (oplelevec, vexpcal1, *vexpcalcul, back_annot_obj);
    }
  return vexpresult;
}

/**
   Compiling an expression  such that ["travel",[0]: EXPRESSION,[E2]: EXPRRESSION, ...]
   @warning { the first element is necessary a travel (syntactic rule).
   For [[0]: EXPRESSION, [E2]: EXPRESSION, ...] see oplelevect operator.}
*/
ptinsth
kernel2graph::instlelevec (ptinsth vhcste, TypeType ContextType, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul,
			   ptlinstances * instent)
{
  ptinsth vh;
  vh = instlelevecRec (vhcste, isfree, vklExprSig, vexpcalcul, instent);
  if (ContextType != UndefinedType)
    EM->TypeCheckingForceType (*vexpcalcul, ContextType, back_annot_obj);
  return vh;
}

/**
   compiling [ EXPRESSSION1, EXPRESSSION2,....] (without index).
*/
ptinsth
kernel2graph::instlistElemvector (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent,
				  int vindex)
{
  ptinsth vexpresult;
  ptexpcal vexpcal1, vexpIndex;
  ptklExprSig v1klExprSig, v2klExprSig;
  boolean locfree;
  ptinsth vh1, vh2, vhcontext;
  TypeType ContextType;

  ContextType = UndefinedType;
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  vexpIndex = EM->mkNaryExp (oplesign, EM->mkIntegerConstant (vindex), back_annot_obj);
  vhcontext = getUndefOrInitClock (vhcste);
  if (v2klExprSig == NILEXPRSIG)
    {
      vexpresult = instOperand (vhcontext, ContextType, isfree, v1klExprSig, &vexpcal1, instent, oplevect);
      vexpcal1 = EM->mkBinaryExp (opelevec, vexpIndex, vexpcal1, back_annot_obj);
      *vexpcalcul = EM->mkNaryExp (oplelevec, vexpcal1, back_annot_obj);
    }
  else
    {
      locfree = pk_true;
      vindex = vindex + 1;
      vh1 = instOperand (vhcontext, ContextType, &locfree, v1klExprSig, &vexpcal1, instent, oplevect);
      vexpcal1 = EM->mkBinaryExp (opelevec, vexpIndex, vexpcal1, back_annot_obj);
      if (locfree)
	{
	  vh2 = instlistElemvector (vhcontext, isfree, v2klExprSig, vexpcalcul, instent, vindex);
	  vexpresult = igc->instop2syncfree (vhcste, vh1, vh2, *isfree);
	}
      else
	{
	  locfree = pk_true;
	  vh2 = instlistElemvector (vh1, &locfree, v2klExprSig, vexpcalcul, instent, vindex);
	  vexpresult = CKM->makeSynchro (vh1, vh2);
	  *isfree = pk_false;
	}
      *vexpcalcul = EM->mkNaryExp (oplelevec, vexpcal1, *vexpcalcul, back_annot_obj);
    }
  // printf("       exiting kernel2graph::instlistElemvector\n"); fflush(stdout);
  return vexpresult;
}


/**
   Compiling [ EXPRESSSION1, EXPRESSSION2,....] (list of restructuration, scalar, not a "travel").
   The expression EXPRESSIONi is a restructuration or a scalar.
   @warning { See  instlelevec()  for other forms (travel).}
*/
boolean
kernel2graph::instlevector (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent,
			    ptinsth * vhresult)
{
  ptklExprSig v1klExprSig, v2klExprSig;

  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  if (kl_getFlowExprOperator (v1klExprSig) == oprest)
    return pk_false;
  *vhresult = instlistElemvector (vhcste, isfree, vklExprSig, vexpcalcul, instent, 0);
  return pk_true;
}

/** Compiling a reference to an enumeration value : T#"value" or #"value"( T is "Any Type") expression.
    @return an expression (constant(T), constant(value))
*/
ptexpcal kernel2graph::instRefEnumValue (ptklExprSig vklExprSig)
{
  ptklExprSig v1klExprSig, v2klExprSig;
  ptexpcal vexp1;
  TypeType TT;
  ptrepstr val2;
  /* Creating an expcal structure (constant(nameType), constant(val)) */
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  val2 = kl_getConstExpr (v2klExprSig);	// PLG WARNING should be kl_getIdent ?
  if (v1klExprSig != NILEXPRSIG)
    {
      TT = getCurrentType (kl_getIdentifierExpr (v1klExprSig));
      EM->TypeCheckingControlValEnum (TT, val2, back_annot_obj);
    }
  else
    TT = AnyTypeCreate ();
  vexp1 = EM->mkBinaryExp (opvenum, EM->mkExpType (TT), EM->mk_enumPtexpcal (val2), back_annot_obj);
  return vexp1;
}

/**
   Compiling the expression ?vklExprSig 
   Assumed: operator of ?vklExprSig is "opleflot" (list of expressions).
*/
ptinsth
kernel2graph::instoperTuple (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul,
			     ptlinstances * instent, boolean bSynchro)
{
  ptklExprSig v1klExprSig;
  ptexpcal ve1, ve2;
  ptinsth vh1, vh2;
  TypeType ContextType;
  v1klExprSig = kl_getNextExpr (&vklExprSig);
  if (vklExprSig == NILEXPRSIG)
    {
      ContextType = UndefinedType;
      vh1 = instOperand (vhcste, ContextType, isfree, v1klExprSig, &ve1, instent, opleflot);
      *vexpcalcul = EM->mkNaryExp (opleflot, ve1, back_annot_obj);
      return vh1;
    }
  vh2 = instoperTuple (vhcste, isfree, vklExprSig, &ve2, instent, bSynchro);
  ContextType = UndefinedType;
  vh1 = instOperand (vhcste, ContextType, isfree, v1klExprSig, &ve1, instent, opleflot);
  *vexpcalcul = EM->mkNaryExp (opleflot, ve1, ve2, back_annot_obj);

  if (bSynchro) return CKM->makeSynchro (vh2, vh1);
  return CKM->makeSup (vh2, vh1);
}


/**
   Compiling the expression ?vklExprSig 
   Assumed: operator of ?vklExprSig is "opcham" (access to a element of struct/bundle)
   @return the clock of the expression.
   @param vexpcalcul (out) the instanciated expression vklExprSig.
*/
ptinsth
kernel2graph::instoperField (ptinsth vhcste, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptklExprSig v1klExprSig, v2klExprSig;
  ptinsth vh1;
  ptexpcal vexpcal1, vexpcal2;
  ptlinstances instent1;
  TypeType ContextType;

  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  ContextType = UndefinedType;
  instent1 = NILLINSTANCES;
  vh1 = instOperand (vhcste, ContextType, isfree, v1klExprSig, &vexpcal1, &instent1, opcham);
  vexpcal2 = instconst (v2klExprSig);
  *vexpcalcul = EM->mkBinaryExp (opcham, vexpcal1, vexpcal2, back_annot_obj);
  return igc->getSetIdClock (*vexpcalcul, isfree, vhcste, back_annot_obj, instent);
}


/**
   Compiling a synchronous binary expression.
   @param vklExprSig (in) expression to instanciate.
   @param ContextType (in) type for added signals during the compiling of vklExprSig (context type).
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at pk_true (out) if the expression is a constant one.
   @param vexpcalcul (out) instanciated expression of vklExprSig.
   @param instent (in/out) instent (in) is completed with the list of signals referenced in vexpvalcul.
   @return clock of the operand (can be undefined clock).
*/
ptinsth
kernel2graph::instoper2 (ptinsth vhcste, TypeType ContextType, boolean * isfree, ptklExprSig vklExprSig,
			 ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vhresult;
  noperators operateur;
  ptlinstances instent1;
  operateur = kl_getFlowExprOperator (vklExprSig);

  switch ((int) operateur)
    {
    case opcstdp:
    case opcstreel:
      *vexpcalcul = instconst (vklExprSig);
      vhresult = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
      break;

    case opvenum:
      *vexpcalcul = instRefEnumValue (vklExprSig);
      vhresult = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
      break;

    case opcplx:
    case opegeg:		/*  ==  */
    case opiieg:		/*  <<= */
    case opdiff:
    case opegal:
    case opinfe:
    case opiegal:
    case opsegal:
    case opsupe:
    case opet:
    case opou:
    case opdivi:
    case opmoin:
    case opmult:
    case opplus:
    case oppuis:
    case opconc:
    case opmod:
    case opouex:
    case opmmul:
    case oprepe:		// e1 |* e2 , repeat e1 e2 times
    case opsequ:		// E1 next E2
    case oprecu:		//  e1 \\ e2 
      vhresult = instoper2_instcal (vhcste, ContextType, vklExprSig, isfree, instent, vexpcalcul);
      break;

    case opelem:
      instent1 = NILLINSTANCES;
      vhresult = instElemAccess (vhcste, isfree, vklExprSig, vexpcalcul, &instent1);
      *instent = ptlinstances_insert (*instent, *vexpcalcul);
      break;

    case oplevect:		// [<expr_1>,<expr_2>, <expr_3>] <expr_i> is a restructuration or a scalar.
      if (!instlevector (vhcste, isfree, vklExprSig, vexpcalcul, instent, &vhresult))
	vhresult = instlelevec (vhcste, ContextType, isfree, vklExprSig, vexpcalcul, instent);
      break;

    case oprest:		// I1,..,In : Expr
      vhresult = instRestructuration (vhcste, isfree, vklExprSig, vexpcalcul, instent);
      break;

    case oplelevec:		/* (inherited from H2)
				   Warning: the first element is necessary a travel "i to n ",[ {i to n}:E, [0]: <expr>, ...]
				   An element cannot be a scalar.
				   It is dangerous to replace [1,2,3] by [[0]:1, [1]:2, [2]:3] (the re-compiling should be not equivalent).
				*/
      vhresult = instlelevec (vhcste, ContextType, isfree, vklExprSig, vexpcalcul, instent);
      break;

    case opchai:		// list of strings.
      *vexpcalcul = instString (vklExprSig);
      vhresult = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
      break;

    case opcham:		// Access to a field of a structure.
      vhresult = instoperField (vhcste, isfree, vklExprSig, vexpcalcul, instent);
      break;

    case opleflot:		// --> Creating a Tuple
      vhresult = instoperTuple (vhcste, isfree, vklExprSig, vexpcalcul, instent, !TypeTestBundleClass (ContextType));
      break;

    case oppcar:
      // xxxxx;
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return vhresult;
}


/** Local to instargwhen1() method.
 */
ptinsth
kernel2graph::instargwhen1_setokwhen1 (ptinsth * vhh, ptexpcal * vexpcalcul)
{
  ptinsth vhsub, vhnosub;
  ptexpcal exp;
  noperators op;

  EM->TypeCheckingVerifyTypeCondition (*vexpcalcul, back_annot_obj);
  op = EM->getOperator (*vexpcalcul);
  if (op == opnon)
    {
      exp = EM->getUnaryArg (*vexpcalcul);
      vhsub = CKM->makeSub (*vhh, exp, EM->getPredecessors (exp));
      igc->takeIntoAccountContexts (vhsub);
      vhnosub = CKM->makeNoSub (exp);
      igc->takeIntoAccountContexts (vhnosub);
      return vhnosub;
    }
  if (op == opdiff)
    {
      exp = EM->mkBinaryExp (opegal, EM->getBinaryArgstAt (*vexpcalcul, 1), EM->getBinaryArgstAt (*vexpcalcul, 2), back_annot_obj);
      vhsub = CKM->makeSub (*vhh, exp, EM->getPredecessors (exp));
      igc->takeIntoAccountContexts (vhsub);
      vhnosub = CKM->makeNoSub (exp);
      igc->takeIntoAccountContexts (vhnosub);
      return vhnosub;
    }

  if (EM->IsEvent (*vexpcalcul))
    {
      *vexpcalcul = vexptrue;
      {
	return *vhh;
      }
    }
  vhsub = CKM->makeSub (*vhh, *vexpcalcul, EM->getPredecessors (*vexpcalcul));
  igc->takeIntoAccountContexts (vhsub);
  return vhsub;
}

/**
   @return "true" if an expression (vexp) is the constant boolean value (b), "false" otherwise.
*/
boolean
kernel2graph::isBooleanConstant (ptexpcal vexp, boolean * b)
{
  TypeType T;

  if (vexp == vexptrue)
    {
      *b = pk_true;
      return pk_true;
    }
  if (vexp == vexpfalse)
    {
      *b = pk_false;
      return pk_true;
    }
  if (EM->IsAssigned (vexp))
    {
      T = EM->getUnamedType (vexp);
      if (T == PureType)
	{
	  *b = pk_true;
	  return pk_true;
	}
      if (T == BooleanType)
	{
	  *b = (EM->getValue (vexp) == 1);
	  return pk_true;
	}
    }
  return pk_false;
}

/** Local to instargwhen1() method.
 */
ptinsth
kernel2graph::instargwhen1_setboolwhen1 (ptinsth vhcste, ptinsth * vhh, ptexpcal * vexpcalcul)
{
  boolean b;
  if (isBooleanConstant (*vexpcalcul, &b))
    {
      if (b)
	return *vhh;		// when(True)
      if ((*vhh) == vhcste)
	*vhh = the_nullClock_K2G_copy;
      return the_nullClock_K2G_copy;	//when(False)
    }
  if (CKM->satisfiesUndefClock (*vhh))
    *vhh = vhcste;
  return instargwhen1_setokwhen1 (vhh, vexpcalcul);
}

/**  Local to instargwhen1 method.
 */
ptinsth
kernel2graph::instargwhen1_seterreurwhen1 (ptexpcal * vexpcalcul, ptinsth * vhh)
{ 
  EM->TypeCheckingCondition (*vexpcalcul, back_annot_obj);
  *vexpcalcul = vexptrue;
  *vhh = CKM->new_variFormulaRootClock ();
  return *vhh;
}


/**
   @return the undersampling of the clock of the expression (vexpcalcul) by this condition.
   @param vklExprSig (in) expression of the condition.
   @param isfree (in) context info xxxx.
   @param vhcste (in) clock context info used to fix the clock of constant expressions.
   @param vexpcalcul (out) expression of the condition associated with vklExprSig. 
   @param iscond (in) => appele pour la compilation de la  condition d'un IF 
   @param vhh (out) horloge de l'expr. non sous-echantillonnee utilisee pour la compilation des IFs dans le cas
   d'expressions comme : if not (event A) else ...
*/
ptinsth
kernel2graph::instargwhen1 (ptinsth vhcste, boolean isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, boolean iscond, ptinsth * vhh)
{
  ptinsth vhresult, vhcal;
  ptlinstances instent;
  noperators operateur;
  TypeType ContextType;
  noperators contextId;
  ContextType = get_BooleanType ();
  instent = NILLINSTANCES;
  operateur = kl_getFlowExprOperator (vklExprSig);
  switch ((int) (getTypeExpflot (vklExprSig)))
    {
    case call:
      *vhh = instCall (vklExprSig, NILEXPCAL, NILEXPCAL, vexpcalcul, &instent);
      *vexpcalcul = EM->getNaryArgs (vexpcalcul);
      *vhh = igc->setClockVariable (*vexpcalcul);
      vhresult = instargwhen1_setboolwhen1 (vhcste, vhh, vexpcalcul);
      break;
    case constant:
      *vexpcalcul = instconst (vklExprSig);
      *vhh = igc->instsiginvhcste (&isfree, vhcste, back_annot_obj);
      vhresult = instargwhen1_setboolwhen1 (vhcste, vhh, vexpcalcul);
      break;
    case variabl:
      *vhh = instsigin (vhcste, &isfree, vklExprSig, vexpcalcul, &instent);
      vhresult = instargwhen1_setboolwhen1 (vhcste, vhh, vexpcalcul);
      break;
    case oper1:
      *vhh = instoper1 (vhcste, ContextType, &isfree, vklExprSig, vexpcalcul, &instent);
      vhresult = instargwhen1_setboolwhen1 (vhcste, vhh, vexpcalcul);
      break;
    case oper1tp:
      if (operateur == ophnull)
	{
	  *vexpcalcul = vexptrue;
	  *vhh = the_nullClock_K2G_copy;
	  vhresult = *vhh;
	}
      else if (operateur == opevnt)
	{
	  *vexpcalcul = vexptrue;
	  *vhh = instevent (vhcste, vklExprSig);
	  vhresult = *vhh;
	}
      else if (operateur == opwhen1)
	{
	  *vexpcalcul = vexptrue;
	  *vhh = instwhen1 (vhcste, isfree, vklExprSig);
	  vhresult = *vhh;
	}
      else
	{
	  kl_getUnaryExpr (vklExprSig, &vklExprSig);
	  *vhh = instargwhen1 (hundef, pk_false, vklExprSig, vexpcalcul, pk_false, &vhcal);
	  vhresult = instargwhen1_seterreurwhen1 (vexpcalcul, vhh);
	}
      break;
    case oper2:
      *vhh = instoper2 (vhcste, ContextType, &isfree, vklExprSig, vexpcalcul, &instent);
      vhresult = instargwhen1_setboolwhen1 (vhcste, vhh, vexpcalcul);
      break;

    case oper2tp:
      if ((operateur == opextr) && !iscond)
	{
	  *vhh = instclextr (&vhcste, ContextType, pk_true, isfree, vklExprSig, vexpcalcul, &instent);
	  vhresult = *vhh;
	}
      else
	{
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, &instent);
	  nameManager->setContextSubExpr (contextId);
	  *vhh = instEquationOp2tp (vhcste, &isfree, *vexpcalcul, vklExprSig);
	  vhresult = instargwhen1_setokwhen1 (vhh, vexpcalcul);
	}
      break;

    case oper3:
      contextId = nameManager->setContextSubExpr (operateur);
      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, &instent);
      nameManager->setContextSubExpr (contextId);
      *vhh = instEquationOp3 (vhcste, &isfree, *vexpcalcul, vklExprSig);

      vhresult = instargwhen1_setokwhen1 (vhh, vexpcalcul);
      break;

    case oper3tp:
      contextId = nameManager->setContextSubExpr (operateur);
      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, &instent);
      nameManager->setContextSubExpr (contextId);
      *vhh = instEquationOp3tp (vhcste, &isfree, *vexpcalcul, vklExprSig);

      vhresult = instargwhen1_setokwhen1 (vhh, vexpcalcul);
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return vhresult;
}

/**
   Local to instEquationIf() method. 
   Compiling the first argument (condition) of the if operator.
*/
ptinsth
kernel2graph::instoperCond (ptinsth vhcste, TypeType ContextType, boolean * isfree, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vhcal;
  if (kl_getFlowExprOperator (vklExprSig) == opextr)
    {
      vhcal = vhcste;
      return instclextr (&vhcal, ContextType, pk_false, pk_true, vklExprSig, vexpcalcul, instent);
    }
  return instOperand (vhcste, ContextType, isfree, vklExprSig, vexpcalcul, instent, opcond);
}

/**
   Adding the equation vexpcal := compiling(vklExprSig) in the graph.
   @param vklExprSig (in) expression to instanciate; assumed to be an "if" expression (synchronous ternary operator).
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at pk_true (out) if the expression is a constant one.
   @param vexpcal (in) the left part of the equation.
   @return clock of vexpcal synchronized to the clock of the right part expression.
*/
ptinsth
kernel2graph::instEquationIf (ptinsth vhcste, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vhresult, vhexp, vh1, vh2, lvhcste, vhnosub, vhconstante;
  ptklExprSig v1klExprSig, v2klExprSig, v3klExprSig;
  ptexpcal vexpcal1, vexpcal2, vexpcond;
  boolean locfree, locfreecond;
  ptlinstances instent1, instent2, instentcond, lindices;
  TypeType ContextType;

  if (kl_getFlowExprOperator (vklExprSig) != opcond)	// Blindage
    {
      pK_messenger->errorTree (errnotimplemented, currentTree);
      return the_nullClock_K2G_copy;
    }

  kl_getTernaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig, &v3klExprSig);
  instentcond = NILLINSTANCES;
  locfreecond = pk_true;
  ContextType = get_BooleanType ();
  vhexp = instOperand (vhcste, ContextType, &locfreecond, v1klExprSig, &vexpcond, &instentcond, opcond);
  EM->TypeCheckingCondition (vexpcond, back_annot_obj);
  lindices = NILLINSTANCES;
  lvhcste = instargwhen1_setboolwhen1 (vhcste, &vhexp, &vexpcond);
  ContextType = EM->getType (vexpcal);
  if (lvhcste == vhexp)
    {
      locfree = pk_true;
      instent1 = NILLINSTANCES;
      vh1 = instoperCond (lvhcste, ContextType, &locfree, v2klExprSig, &vexpcal1, &instent1);
      vhresult = CKM->makeSynchro (vhexp, vh1);
      vhresult = igc->createEquation (vexpcal, vexpcal1, vhresult, vhresult, instent1, back_annot_obj);
    }
  else if (lvhcste == the_nullClock_K2G_copy)
    {
      instent2 = NILLINSTANCES;
      vh2 = instoperCond (vhexp, ContextType, isfree, v3klExprSig, &vexpcal2, &instent2);
      vhresult = CKM->makeSynchro (vhexp, vh2);
      vhresult = igc->createEquation (vexpcal, vexpcal2, vhresult, vhresult, instent2, back_annot_obj);
    }
  else
    {				/* lvhcste est un SUB */
      instent1 = NILLINSTANCES;
      locfree = pk_true;
      vhconstante = vhexp;
      vh1 = instoperCond (vhconstante, ContextType, &locfree, v2klExprSig, &vexpcal1, &instent1);

      instent2 = NILLINSTANCES;
      locfree = pk_true;
      if (!CKM->getComplInFather (lvhcste, &vhnosub))
	vhnosub = hundef;

      CKM->makeSynchro (vhexp, CKM->makeSup (lvhcste, vhnosub));
      vhconstante = vhexp;
      vh2 = instoperCond (vhconstante, ContextType, &locfree, v3klExprSig, &vexpcal2, &instent2);

      vhresult = CKM->makeSynchro (vhexp, CKM->makeSynchro (vh1, vh2));
      igc->addSubEquation (vexpcal, vexpcal1, lvhcste, instent1, back_annot_obj);
      vhresult = igc->createEquation (vexpcal, vexpcal2, vhnosub, vhresult, instent2, back_annot_obj);
    }
  EM->TypeCheckingExpcalNLExtensions (vexpcal, back_annot_obj);
  return vhresult;
}


/**
   Adding the equation vexpcal := compiling(vklExprSig) in the graph.
   @param vklExprSig (in) expression to instanciate; assumed to be an enumeration expression (synchronous ternary operator) such as (e1 .. e2 step e3).
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at pk_true (out) if the expression is a constant one.
   @param vexpcal (in) the left part of the equation.
   @return clock of vexpcal synchronized to the clock of the right part expression.
*/
ptinsth
kernel2graph::instEquationIndexExpr (ptinsth vhcste, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vh, vh1, vh2, vhcal;
  ptklExprSig v1klExprSig, v2klExprSig, v3klExprSig;
  ptexpcal vbinf, vbsup, vstep;
  boolean locfree;
  ptlinstances instent;
  TypeType ContextType;
  ContextType = get_IntegerType ();
  kl_getTernaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig, &v3klExprSig);

  vh = igc->getSignalClockSynchronizedWith (vexpcal, vhcste);
  instent = NILLINSTANCES;
  // lower bound
  locfree = pk_true;
  vh1 = instOperand (vhcste, ContextType, &locfree, v1klExprSig, &vbinf, &instent, openum);
  if (v1klExprSig == NILEXPRSIG)
    vbinf = ptexp0;
  vh = CKM->makeSynchro (vh, vh1);
  *isfree = (*isfree) && locfree;

  // upper bound
  locfree = pk_true;
  vh2 = instOperand (vhcste, ContextType, &locfree, v2klExprSig, &vbsup, &instent, openum);
  vh = CKM->makeSynchro (vh, vh2);
  *isfree = (*isfree) && locfree;
  //step
  if (v3klExprSig == NILEXPRSIG)
    vstep = ptexp1;
  else
    {
      locfree = pk_true;
      vh = CKM->makeSynchro (vh, instOperand (vhcste, ContextType, &locfree, v3klExprSig, &vstep, &instent, openum));
      *isfree = (*isfree) && locfree;
    }
  if (*isfree)
    vhcal = hinit;
  else
    vhcal = vh;
  vh = igc->getSignalClockSynchronizedWith (vexpcal, vh);
  return igc->createEquation (vexpcal, EM->mkTernaryExp (openum, vbinf, vbsup, vstep, back_annot_obj), vhcal, vh, instent, back_annot_obj);
}

/**
   Adding the equation vexpcal := compiling(vklExprSig) in the graph.
   @param vklExprSig (in) expression to instanciate; assumed to be a synchronous ternary operator (if, enumeration).
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at pk_true (out) if the expression is a constant one.
   @param vexpcal (in) the left part of the equation.
   @return clock of vexpcal synchronized to the clock of the right part expression.
*/
ptinsth
kernel2graph::instEquationOp3 (ptinsth vhcste, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  noperators operateur;

  operateur = kl_getFlowExprOperator (vklExprSig);
  if (operateur == opcond)
    return instEquationIf (vhcste, isfree, vexpcal, vklExprSig);
  if (operateur == openum)
    return instEquationIndexExpr (vhcste, isfree, vexpcal, vklExprSig);	// e1 .. e2 step e3
  // traite avec operateur opparc if (operateur == opborn) ;
  return NILINSTH;
}


/** Local to instevent() method.
 */
ptinsth
kernel2graph::instargevent (ptinsth vhcste, boolean strong, boolean * isfree, ptklExprSig vklExprSig, ptlinstances * instent)
{
  ptinsth vhresult, vh1, vh2;
  boolean locfree;
  ptklExprSig v1klExprSig, v2klExprSig, v3klExprSig;
  ptexpcal vexpcalcul;
  ptinsth vhexp, lvhcste, vh;
  noperators operateur;
  switch ((int) (getTypeExpflot (vklExprSig)))
    {
    case call:
      *isfree = pk_false;
      vhexp = instCall (vklExprSig, NILEXPCAL, NILEXPCAL, &vexpcalcul, instent);
      vexpcalcul = EM->getNaryArgs (&vexpcalcul);
      vhresult = igc->setClockVariable (vexpcalcul);
      break;
    case constant:
      vhresult = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
      break;
    case variabl:
      vhresult = instsigin (vhcste, isfree, vklExprSig, &vexpcalcul, instent);
      break;
    case oper1:
      kl_getUnaryExpr (vklExprSig, &v1klExprSig);
      vhresult = instargevent (vhcste, pk_false, isfree, v1klExprSig, instent);
      break;
    case oper1tp:
      operateur = kl_getFlowExprOperator (vklExprSig);
      if (operateur == ophnull)
	vhresult = the_nullClock_K2G_copy;
      else if (operateur == opevnt)
	vhresult = instevent (vhcste, vklExprSig);
      else if (operateur == opwhen1)
	vhresult = instwhen1 (vhcste, pk_false, vklExprSig);
      else
	{
	  kl_getUnaryExpr (vklExprSig, &v1klExprSig);
	  vhresult = instargwhen1 (vhcste, pk_false, v1klExprSig, &vexpcalcul, pk_false, &vhexp);
	}
      *isfree = pk_false;
      break;
    case oper2:
      operateur = kl_getFlowExprOperator (vklExprSig);
      if (operateur == opcham)
	vhresult = instoperField (vhcste, isfree, vklExprSig, &vexpcalcul, instent);
      else if (operateur == opvenum)
	vhresult = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
      else
	{
	  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
	  if (v2klExprSig == NILEXPRSIG)
	    vhresult = instargevent (vhcste, pk_false, isfree, v1klExprSig, instent);
	  else
	    {
	      locfree = pk_true;
	      vh = instargevent (vhcste, pk_false, &locfree, v1klExprSig, instent);
	      if (locfree)
		{
		  lvhcste = instargevent (vh, pk_false, isfree, v2klExprSig, instent);
		  if (CKM->satisfiesUndefClock (lvhcste))
		    vhresult = vhcste;
		  else
		    vhresult = CKM->makeSynchro (lvhcste, vh);
		}
	      else
		{
		  locfree = pk_true;
		  lvhcste = instargevent (vh, pk_false, &locfree, v2klExprSig, instent);
		  vhresult = CKM->makeSynchro (lvhcste, vh);
		  *isfree = pk_false;
		}
	    }
	}
      break;


    case oper2tp:
      kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
      switch ((int) (kl_getFlowExprOperator (vklExprSig)))
	{
	case opmele:
	  locfree = (boolean) ((*isfree) && strong);
	  lvhcste = instargevent (vhcste, strong, &locfree, v1klExprSig, instent);
	  if (*isfree)
	    if (CKM->satisfiesUndefClock (vhcste))
	      {
		vhcste = CKM->new_variFormulaRootClock ();
		vhresult = CKM->makeSynchro (vhcste, CKM->makeSup (lvhcste, instargevent (vhcste, strong, isfree, v2klExprSig, instent)));
	      }
	    else
	      vhresult = CKM->makeSup (lvhcste, instargevent (vhcste, strong, isfree, v2klExprSig, instent));
	  else
	    vhresult = CKM->makeSup (lvhcste, instargevent (vhcste, pk_false, isfree, v2klExprSig, instent));
	  break;
	case opextr:
	  locfree = strong;
	  lvhcste = instargwhen1 (vhcste, locfree, v2klExprSig, &vexpcalcul, pk_false, &vhexp);
	  if (strong)
	    lvhcste = CKM->makeInf (lvhcste, vhcste);
	  *isfree = pk_true;
	  vhresult = CKM->makeInf (lvhcste, instargevent (lvhcste, pk_true, isfree, v1klExprSig, instent));
	  *isfree = pk_false;
	  break;

	case opvar:
	  *isfree = pk_false;
	  vhresult = igc->instsiginvhcste (isfree, vhcste, back_annot_obj);
	  break;

	case opnombm:
	  *isfree = pk_true;
	  vhresult = instargevent (vhcste, pk_false, isfree, v1klExprSig, instent);
	  break;

	case opnomba:
	case opnombd:
	  *isfree = pk_true;
	  vhresult = CKM->makeSup (instargevent (vhcste, pk_false, isfree, v1klExprSig, instent),
				   instargevent (vhcste, pk_false, isfree, v2klExprSig, instent));
	  break;
	default:
	  pK_messenger->Caseerror (currentLine, currentFile);
	}
      break;
    case oper3:
      kl_getTernaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig, &v3klExprSig);
      switch ((int) (kl_getFlowExprOperator (vklExprSig)))
	{
	case opcond:		//if
	case openum:		// e1 .. e2 step e3
	  *isfree = pk_false;
	  vh1 = instargevent (vhcste, pk_false, isfree, v1klExprSig, instent);
	  vh2 = instargevent (vh1, pk_false, isfree, v2klExprSig, instent);
	  vh1 = CKM->makeSynchro (vh1, vh2);
	  vh2 = instargevent (vh1, pk_false, isfree, v3klExprSig, instent);
	  vhresult = CKM->makeSynchro (vh1, vh2);
	  break;
	}
      break;

    case oper3tp:
      kl_getTernaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig, &v3klExprSig);
      switch ((int) (kl_getFlowExprOperator (vklExprSig)))
	{
	case opmemo:
	  *isfree = pk_false;
	  vhresult = CKM->makeSup (instargevent (vhcste, pk_false, isfree, v1klExprSig, instent),
				   instargwhen1 (vhcste, pk_false, v2klExprSig, &vexpcalcul, pk_false, &vhexp));
	  break;
	default:		// delay
	  locfree = pk_true;
	  lvhcste = instargevent (vhcste, pk_false, &locfree, v1klExprSig, instent);
	  if (v2klExprSig == NILEXPRSIG)
	    if (v3klExprSig == NILEXPRSIG)
	      vhresult = TestSetFreeClock (locfree, lvhcste /* vhresult */ );
	    else
	      {
		locfree = (boolean) (!locfree);
		vhresult = CKM->makeSynchro (lvhcste, instargevent (lvhcste, pk_false, &locfree, v3klExprSig, instent));
	      }
	  else if (v3klExprSig == NILEXPRSIG)
	    {
	      locfree = (boolean) (!locfree);
	      vhresult = CKM->makeSynchro (lvhcste, instargevent (lvhcste, pk_false, &locfree, v2klExprSig, instent));
	    }
	  else
	    {
	      *isfree = locfree;
	      locfree = pk_true;
	      lvhcste = CKM->makeSynchro (lvhcste, instargevent (lvhcste, pk_false, &locfree, v2klExprSig, instent));
	      locfree = (boolean) (!(locfree || (*isfree)));
	      vhresult = CKM->makeSynchro (lvhcste, instargevent (lvhcste, pk_false, &locfree, v3klExprSig, instent));
	    }
	}
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return vhresult;
}

/**
   Compiling the ^E expression.
   @param vklExprSig (in) expression to instanciate; assumed to be the ^ E expression.
   @param vhcste (in) clock of the constants (context clock).
   @return the clock of the expression.
*/
ptinsth
kernel2graph::instevent (ptinsth vhcste, ptklExprSig vklExprSig)
{
  ptklExprSig v1klExprSig;
  boolean locfree;
  ptinsth vh;
  ptlinstances instent;

  if (kl_getFlowExprOperator (vklExprSig) == opevnt) kl_getUnaryExpr (vklExprSig, &v1klExprSig); else  v1klExprSig = vklExprSig;
  instent = NILLINSTANCES;
  locfree = pk_true;
  vh = instargevent (vhcste, pk_false, &locfree, v1klExprSig, &instent);
  return TestSetFreeClock (locfree, vh);
}

/**
   Compiling the when E expression.
   @param vklExprSig (in) expression to instanciate; assumed to be the unary when expression.
   @param isfree (in) at "true" when the constants are allowed.
   @param vhcste (in) clock of the constants (context clock).
   @return the clock of the expression.
*/
ptinsth
kernel2graph::instwhen1 (ptinsth vhcste, boolean isfree, ptklExprSig vklExprSig)
{
  ptklExprSig v1klExprSig;
  ptexpcal vexpcalcul;
  ptinsth vh, vhexp;
  ptident id;

  kl_getUnaryExpr (vklExprSig, &v1klExprSig);
  id = nameManager->setContextSubExprCondition ();
  vh = instargwhen1 (vhcste, isfree, v1klExprSig, &vexpcalcul, pk_false, &vhexp);
  nameManager->setContextSubExpr (id);
  return vh;
}

/** Local to instclmele() method. 
 */
ptinsth
kernel2graph::instargmele (ptinsth vhleft, ptinsth vhcste, boolean strong, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  boolean locfree;
  ptinsth vh, vhresult;
  ptexpcal vexpcalcul;
  ptlinstances instent;
  noperators operateur;
  TypeType ContextType;
  noperators contextId;

  ContextType = EM->getType (vexpcal);
  operateur = kl_getFlowExprOperator (vklExprSig);
  if (operateur == opmele)
    vhresult = instclmele (vhleft, vhcste, strong, isfree, vexpcal, vklExprSig);
  else
    {
      instent = NILLINSTANCES;
      switch ((int) (getTypeExpflot (vklExprSig)))
	{
	case call:
	  locfree = pk_false;
	  if (!instExternalFunction (vklExprSig, vhcste, &locfree, &vh, &vexpcalcul, &instent))
	    {
	      vh = instCall (vklExprSig, NILEXPCAL, NILEXPCAL, &vexpcalcul, &instent);
	      vexpcalcul = EM->getNaryArgs (&vexpcalcul);
	      vh = igc->setClockVariable (vexpcalcul);
	    }
	  if (EM->IsEvent (vexpcalcul))
	    {
	      vexpcalcul = vexptrue;
	      instent = NILLINSTANCES;
	    }
	  break;
	case constant:
	  vexpcalcul = instconst (vklExprSig);
	  locfree = pk_true;
	  vh = vhcste;
	  break;
	case variabl:
	  locfree = *isfree;
	  vh = instsigin (vhcste, &locfree, vklExprSig, &vexpcalcul, &instent);
	  break;
	case oper1:
	  locfree = *isfree;
	  vh = instoper1 (vhcste, ContextType, &locfree, vklExprSig, &vexpcalcul, &instent);
	  break;
	case oper1tp:
	  locfree = pk_false;
	  if (operateur == ophnull)
	    {
	      vexpcalcul = vexptrue;
	      vh = the_nullClock_K2G_copy;
	    }
	  else if (operateur == opevnt)
	    {
	      vexpcalcul = vexptrue;
	      vh = instevent (vhcste, vklExprSig);
	    }
	  else if (operateur == opwhen1)
	    {
	      vexpcalcul = vexptrue;
	      vh = instwhen1 (vhcste, strong, vklExprSig);
	    }
	  break;
	case oper2:
	  locfree = *isfree;
	  vh = instoper2 (vhcste, ContextType, &locfree, vklExprSig, &vexpcalcul, &instent);
	  break;
	case oper2tp:
	  switch ((int) operateur)
	    {
	    case opextr:
	      vh = instclextr (&vhcste, ContextType, pk_false, strong, vklExprSig, &vexpcalcul, &instent);
	      break;
	    case opvar:
	      contextId = nameManager->setContextSubExpr (operateur);
	      vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, &instent);
	      nameManager->setContextSubExpr (contextId);
	      vh = instEquationVar (vhcste, vexpcalcul, vklExprSig);

	      break;
	    case opnombm:
	      contextId = nameManager->setContextSubExpr (operateur);
	      vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, &instent);
	      nameManager->setContextSubExpr (contextId);
	      vh = instEquationCount (vhcste, vexpcalcul, vklExprSig);
	      break;
	    default:
	      pK_messenger->Caseerror (currentLine, currentFile);
	    }
	  locfree = pk_false;
	  break;

	case oper3:
	  contextId = nameManager->setContextSubExpr (operateur);
	  vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, &instent);
	  nameManager->setContextSubExpr (contextId);
	  vh = instEquationOp3 (hundef, isfree, vexpcalcul, vklExprSig);
	  locfree = pk_false;
	  break;

	case oper3tp:
	  contextId = nameManager->setContextSubExpr (operateur);
	  vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, &instent);
	  nameManager->setContextSubExpr (contextId);
	  if (operateur == opmemo)
	    vh = instEquationCell (vhcste, vexpcalcul, vklExprSig);
	  else // delay
	    vh = instEquationOp3tp (vhcste, isfree, vexpcalcul, vklExprSig);
	  locfree = pk_false;
	  break;
	default:
	  pK_messenger->Caseerror (currentLine, currentFile);
	}

      if (CKM->satisfiesUndefClock (vh))
	vh = igc->getSetIdClock (vexpcal, isfree, vhcste, back_annot_obj, &instent);
      if (CKM->satisfiesNullClock (vhleft))
	igc->addSubEquation (vexpcal, vexpcalcul, vh, instent, back_annot_obj);
      else
	igc->addSubEquation (vexpcal, vexpcalcul, CKM->makeMinus (vh, vhleft), instent, back_annot_obj);

      if (locfree)
	vh = igc->instsiginvhcste (isfree, vh, back_annot_obj);
      else
	*isfree = pk_false;
      vhresult = CKM->makeSup (vhleft, vh);
    }
  return vhresult;
}


/**
   Compiling a default expression.
   @param strong (in) context for constant.
   @param vhleft (in) clock of the compiled part (context) or the null clock.
   @param vhcste (in) clock of the constants (context clock).
   @param isfree (in/out) at pk_true (in) if constants are allowed, at pk_true (out) if the expression is a constant one.
   @param vklExprSig (in) expression to compile; assumed to be a default expression.
   @param vexpcal (out) instanciated expression associated with the compiled expression. 
   @return clock of the expression.
*/
ptinsth
kernel2graph::instclmele (ptinsth vhleft, ptinsth vhcste, boolean strong, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vhresult;
  ptklExprSig v1klExprSig, v2klExprSig;
  boolean locfree;

  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  locfree = strong;
  vhresult = instargmele (vhleft, vhcste, strong, &locfree, vexpcal, v1klExprSig);

  if (*isfree)
    {
      vhresult = instargmele (vhresult, vhcste, strong, isfree, vexpcal, v2klExprSig);
      *isfree = (boolean) ((*isfree) || locfree);
    }
  else
    vhresult = instargmele (vhresult, vhcste, strong, isfree, vexpcal, v2klExprSig);

  return vhresult;
}

/**
   Macro expansion of a when-expression.
   @param vklExprSig (in): a when-expression.
   @param vexpcalcul (out): expansed expression of vklExprSig expression according to the contexts.
   @param vhcal (in/out): context clock for constants (in),  (out) computation clock (it can be init-clock).
   @param ContextType (in) type for added signals during the compiling of the expression (context type).
   @param instent (in/out): list instent (in) in which the signals of vexpcalcul expression are added.
   @param bextract (in) : at pk_true for compiling when(klExprSig), false otherwise.
   @param strong (in) : context for constants; it must be at pk_true value when constant are not allowed, false otherwise.
   @return the clock of the expansed expression.
*/
ptinsth
kernel2graph::instclextr (ptinsth * vhcal, TypeType ContextType, boolean bextract, boolean strong, ptklExprSig vklExprSig,
			  ptexpcal * vexpcalcul, ptlinstances * instent)
{
  ptinsth vhresult, vh1;
  ptklExprSig arg1flot, v2klExprSig;
  ptinsth vhexp, vhcste;
  ptexpcal vexpcond;
  noperators operateur;
  noperators contextId;
  ptident id;

  kl_getBinaryExpr (vklExprSig, &arg1flot, &v2klExprSig);
  // The second operand
  id = nameManager->setContextSubExprCondition ();
  vhcste = instargwhen1 (*vhcal, strong, v2klExprSig, &vexpcond, pk_false, &vhexp);
  nameManager->setContextSubExpr (id);
  // The first operand
  if (CKM->satisfiesUndefClock (*vhcal))
    *vhcal = vhcste;
  else
    *vhcal = CKM->makeInf (*vhcal, vhcste);
  if (bextract)
    vhresult = CKM->makeInf (vhcste, instargwhen1 (vhcste, pk_true, arg1flot, &vexpcond, pk_false, &vhexp));
  else
    {
      operateur = kl_getFlowExprOperator (arg1flot);
      strong = pk_true;
      switch ((int) (getTypeExpflot (arg1flot)))
	{
	case call:
	  if (!instExternalFunction (arg1flot, vhcste, &strong, &vh1, vexpcalcul, instent))
	    {
	      *vhcal = instCall (arg1flot, NILEXPCAL, NILEXPCAL, vexpcalcul, instent);
	      *vexpcalcul = EM->getNaryArgs (vexpcalcul);
	      vh1 = igc->setClockVariable (*vexpcalcul);
	    }
	  *vhcal = CKM->makeInf (vhcste, vh1);
	  vhresult = *vhcal;
	  break;
	case constant:
	  *vexpcalcul = instconst (arg1flot);
	  *vhcal = hinit;
	  vhresult = vhcste;
	  break;
	case variabl:
	  *vhcal = instsigin (vhcste, &strong, arg1flot, vexpcalcul, instent);
	  if (strong)
	    {
	      *vhcal = hinit;
	      vhresult = vhcste;
	    }
	  else
	    {
	      *vhcal = CKM->makeInf (vhcste, *vhcal);
	      vhresult = *vhcal;
	    }
	  break;
	case oper1:
	  *vhcal = instoper1 (hundef, ContextType, &strong, arg1flot, vexpcalcul, instent);

	  if (CKM->satisfiesUndefClock (*vhcal))
	    {
	      vhresult = vhcste;
	      *vhcal = hinit;
	    }
	  else
	    {
	      *vhcal = CKM->makeInf (vhcste, *vhcal);
	      vhresult = *vhcal;
	    }
	  break;
	case oper1tp:
	  if (operateur == ophnull)
	    {
	      *vexpcalcul = vexptrue;
	      *vhcal = the_nullClock_K2G_copy;
	      vhresult = *vhcal;
	    }
	  else if (operateur == opevnt)
	    {
	      *vexpcalcul = vexptrue;
	      *vhcal = hinit;
	      vhresult = CKM->makeInf (vhcste, instevent (vhcste, arg1flot));
	    }
	  else if (operateur == opwhen1)
	    {
	      *vexpcalcul = vexptrue;
	      *vhcal = hinit;
	      vhresult = CKM->makeInf (vhcste, instwhen1 (vhcste, strong, arg1flot));
	    }
	  break;
	case oper2:
	  *vhcal = instoper2 (hundef /* vhcste */ , ContextType, &strong, arg1flot, vexpcalcul, instent);
	  if (CKM->satisfiesUndefClock (*vhcal))
	    {
	      vhresult = vhcste;
	      *vhcal = hinit;
	    }
	  else
	    {
	      *vhcal = CKM->makeInf (vhcste, *vhcal);
	      vhresult = *vhcal;
	    }
	  break;
	case oper2tp:
	  switch ((int) operateur)
	    {
	    case opmele:
	      contextId = nameManager->setContextSubExpr (operateur);
	      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	      nameManager->setContextSubExpr (contextId);
	      *vhcal = CKM->makeInf (vhcste, instEquationmele (vhcste, pk_true, &strong, *vexpcalcul, arg1flot));
	      vhresult = *vhcal;
	      break;
	    case opextr:
	      *vhcal = vhcste;	// assert( x:= ( b=bb) when (5=5) when ^b)
	      vhcste = CKM->makeInf (vhcste, instclextr (vhcal, ContextType, pk_false, strong, arg1flot, vexpcalcul, instent));
	      if (!CKM->satisfiesInitClock (*vhcal))
		*vhcal = vhcste;
	      vhresult = vhcste;
	      break;
	    case opvar:
	      contextId = nameManager->setContextSubExpr (operateur);
	      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	      nameManager->setContextSubExpr (contextId);
	      *vhcal = instEquationVar (vhcste, *vexpcalcul, arg1flot);

	      vhresult = *vhcal;
	      break;

	    case opnombm:
	      contextId = nameManager->setContextSubExpr (operateur);
	      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	      nameManager->setContextSubExpr (contextId);

	      *vhcal = CKM->makeInf (vhcste, instEquationCount (vhcste, *vexpcalcul, arg1flot));
	      vhresult = *vhcal;
	      break;
	    default:
	      pK_messenger->Caseerror (currentLine, currentFile);
	    }
	  break;

	case oper3:
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);
	  *vhcal = CKM->makeInf (vhcste, instEquationOp3 (hundef, &strong, *vexpcalcul, arg1flot));

	  vhresult = *vhcal;
	  break;

	case oper3tp:
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);
	  if (operateur == opmemo)
	    *vhcal = CKM->makeInf (vhcste, instEquationCell (vhcste, *vexpcalcul, arg1flot));
	  else
	    *vhcal = CKM->makeInf (vhcste, instEquationOp3tp (vhcste, &strong, *vexpcalcul, arg1flot));
	  vhresult = *vhcal;
	  break;
	default:
	  pK_messenger->Caseerror (currentLine, currentFile);
	}
    }
  return vhresult;
}


/** INSTEQUATIONX for a "var" expression: var E init V0. */
ptinsth
kernel2graph::instEquationVar (ptinsth vhcste, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vhcal, vhvar, vh1, vhDummy;
  ptexpcal vCell, vinit, vexpcalcul;
  ptklExprSig arg1flot, vklExprSiginit;
  ptlinstances instent, instent1;
  noperators contextId;

  kl_getBinaryExpr (vklExprSig, &arg1flot, &vklExprSiginit);
  instExprInit (vklExprSiginit, &vinit, &vhDummy);
  instent = NILLINSTANCES;
  contextId = nameManager->setContextSubExpr (opvar);
  vCell = igc->mkVariable (back_annot_obj, EM->getType (vexpcal), &instent);
  nameManager->setContextSubExpr (contextId);
  vinit = igc->verifAssignInit (vCell, vinit);
  instent1 = NILLINSTANCES;
  vh1 = instCellFirstArg (vhcste, vCell, arg1flot, &vexpcalcul, &vhcal, &instent1);
  if (CKM->satisfiesUndefClock (vhcste))
    vhcste = CKM->new_variFormulaRootClock ();
  vhvar = CKM->makeSup (vhcste, vh1);
  vh1 = igc->createEquation (vCell, vexpcalcul, vhcal, vhvar, instent1, back_annot_obj);
  igc->setAuxiliarySignalName (vexpcal, vexpcalcul, opvar);
  return igc->createEquation (vexpcal, vCell, vhcste, vhcste, instent, back_annot_obj);

}

/**
   Adding the equation vexpcal := compiling(vklExprSig) for a default expression (INSTEQUATIONX for a default expression).
   @param vhcste (in) clock of the constants (context clock).
   @param strong (in) context for constant.
   @param isfree (in/out) at pk_true (in) if constants are allowed, at pk_true (out) if the expression is a constant one.
   @param vklExprSig (in) expression to compile; assumed to be a default expression.
   @param vexpcal (in) the left part of the equation.
   @return clock of vexpcal synchronized to the clock of the right part expression.
*/
ptinsth
kernel2graph::instEquationmele (ptinsth vhcste, boolean strong, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vhresult;
  ptINode vinstr;

  vhresult = instclmele (the_nullClock_K2G_copy, vhcste, strong, isfree, vexpcal, vklExprSig);
  vhresult = igc->getSignalClockSynchronizedWith (vexpcal, vhresult);
  vinstr = EM->getNode (vexpcal);
  NM->insertNode (vinstr, vhresult);
  igc->takeIntoAccountContexts (vinstr);

  return vhresult;
}




/** INSTEQUATIONX for an extraction (when operator).
    @param vexpcal (in) the left part of the equation.
    @param vhcste (in) clock of the constants (context clock).
    @param vklExprSig (in) expression to compile; assumed to be a when expression.
    @return clock of vexpcal synchronized to the clock of the right part expression.
*/
ptinsth
kernel2graph::instEquationextr (ptinsth vhcste, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptexpcal vexpcalcul;
  ptlinstances instent;
  ptinsth vhcal, vhvar;
  instent = NILLINSTANCES;
  vhcal = vhcste;
  vhvar = instclextr (&vhcal, EM->getType (vexpcal), pk_false, pk_false, vklExprSig, &vexpcalcul, &instent);
  return igc->createEquation (vexpcal, vexpcalcul, vhcal, vhvar, instent, back_annot_obj);
}

/**
   Compiling the first arg of a cell operator or the arg of the var operator.
   Local to instCell() and instEquationVar() methods.
*/
ptinsth
kernel2graph::instCellFirstArg (ptinsth vhcste, ptexpcal vexpcal, ptklExprSig arg1flot, ptexpcal * vexpcalcul, ptinsth * vhcal,
				ptlinstances * instent)
{
  ptinsth vhresult, vhextr;
  boolean isfree;
  noperators operateur;
  TypeType ContextType;
  noperators contextId;
  ContextType = EM->getType (vexpcal);
  isfree = pk_false;
  operateur = kl_getFlowExprOperator (arg1flot);
  switch ((int) (getTypeExpflot (arg1flot)))
    {
    case call:
      if (!instExternalFunction (arg1flot, vhcste, &isfree, vhcal, vexpcalcul, instent))
	{
	  *vhcal = instCall (arg1flot, NILEXPCAL, NILEXPCAL, vexpcalcul, instent);
	  *vexpcalcul = EM->getNaryArgs (vexpcalcul);
	  *vhcal = igc->setClockVariable (*vexpcalcul);
	}
      vhresult = CKM->makeSup (vhcste, *vhcal);
      break;
    case constant:
      pK_messenger->errorTree (errcsteint, currentTree);
      *vexpcalcul = instconst (arg1flot);
      *vhcal = the_nullClock_K2G_copy;
      vhresult = CKM->new_variFormulaRootClock ();
      break;
    case variabl:
      *vhcal = instsigin (vhcste, &isfree, arg1flot, vexpcalcul, instent);
      vhresult = CKM->makeSup (vhcste, *vhcal);
      break;
    case oper1:
      *vhcal = instoper1 (vhcste, ContextType, &isfree, arg1flot, vexpcalcul, instent);
      vhresult = CKM->makeSup (vhcste, *vhcal);
      break;
    case oper1tp:
      if (operateur == ophnull)
	{
	  *vexpcalcul = vexptrue;
	  *vhcal = the_nullClock_K2G_copy;
	  vhresult = *vhcal;
	}
      else if (operateur == opevnt)
	{
	  *vexpcalcul = vexptrue;
	  *vhcal = hinit;
	  vhresult = CKM->makeSup (vhcste, instevent (vhcste, arg1flot));
	}
      else if (operateur == opwhen1)
	{
	  *vexpcalcul = vexptrue;
	  *vhcal = hinit;
	  vhresult = CKM->makeSup (vhcste, instwhen1 (vhcste, pk_false, arg1flot));
	}
      break;
    case oper2:
      *vhcal = instoper2 (vhcste, ContextType, &isfree, arg1flot, vexpcalcul, instent);
      vhresult = CKM->makeSup (vhcste, *vhcal);
      break;
    case oper2tp:
      switch ((int) operateur)
	{
	case opmele:
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);
	  *vhcal = instEquationOp2tp (vhcste, &isfree, *vexpcalcul, arg1flot);

	  vhresult = CKM->makeSup (vhcste, *vhcal);
	  break;

	case opextr:
	  vhextr = vhcste;
	  *vhcal = instclextr (&vhextr, ContextType, pk_false, pk_false, arg1flot, vexpcalcul, instent);
	  vhresult = CKM->makeSup (vhcste, *vhcal);
	  if (CKM->satisfiesInitClock (vhextr))
	    *vhcal = vhextr;
	  break;

	case opvar:
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);
	  *vhcal = instEquationVar (vhcste, *vexpcalcul, arg1flot);

	  vhresult = *vhcal;
	  break;

	case opnombm:
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);

	  *vhcal = instEquationCount (vhcste, *vexpcalcul, arg1flot);
	  vhresult = *vhcal;
	  break;
	default:
	  pK_messenger->Caseerror (currentLine, currentFile);
	}
      break;

    case oper3:
      contextId = nameManager->setContextSubExpr (operateur);
      *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
      nameManager->setContextSubExpr (contextId);
      *vhcal = instEquationOp3 (vhcste, &isfree, *vexpcalcul, arg1flot);	// ??? false

      vhresult = CKM->makeSup (vhcste, *vhcal);
      break;

    case oper3tp:
      if (operateur == opmemo)
	vhresult = CKM->makeSup (vhcste, instCell (vhcste, vexpcal, arg1flot, vexpcalcul, vhcal, instent));
      else
	{
	  // delay
	  contextId = nameManager->setContextSubExpr (operateur);
	  *vexpcalcul = igc->mkVariable (back_annot_obj, ContextType, instent);
	  nameManager->setContextSubExpr (contextId);
	  *vhcal = instEquationOp3tp (vhcste, &isfree, *vexpcalcul, arg1flot);

	  vhresult = CKM->makeSup (vhcste, *vhcal);
	}
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  igc->setAuxiliarySignalName (vexpcal, *vexpcalcul, opmemo);
  return vhresult;
}

/**
   Local to instEquationCell() method.
   Compiling a cell expression.
*/
ptinsth
kernel2graph::instCell (ptinsth vhcste, ptexpcal vexpcal, ptklExprSig vklExprSig, ptexpcal * vexpcalcul, ptinsth * vhcal, ptlinstances * instent)
{
  ptinsth vhresult, vhexp, lvhcste;
  ptexpcal vexpcond;
  ptklExprSig arg1flot, v2klExprSig, vklExprSiginit;
  ptident id;

  kl_getTernaryExpr (vklExprSig, &arg1flot, &v2klExprSig, &vklExprSiginit);
  id = nameManager->setContextSubExprCondition ();
  lvhcste = instargwhen1 (vhcste, pk_false, v2klExprSig, &vexpcond, pk_false, &vhexp);	//condition
  nameManager->setContextSubExpr (id);
  vhresult = instCellFirstArg (lvhcste, vexpcal, arg1flot, vexpcalcul, vhcal, instent);	// first argument.
  return vhresult;
}


/**
   INSTEQUATIONX for a cell.
*/
ptinsth
kernel2graph::instEquationCell (ptinsth vhcste, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vhresult;
  ptexpcal vexpcalcul, vinit;
  ptlinstances instent;
  ptinsth vhcal, vhvar, vh;
  ptklExprSig arg1, arg2, vklExprSiginit;

  instent = NILLINSTANCES;
  vexpcalcul = NILEXPCAL;
  vhvar = instCell (vhcste, vexpcal, vklExprSig, &vexpcalcul, &vhcal, &instent);
  kl_getTernaryExpr (vklExprSig, &arg1, &arg2, &vklExprSiginit);
  instExprInit (vklExprSiginit, &vinit, &vh);
  vinit = igc->verifAssignInit (vexpcal, vinit);
  vhresult = igc->createEquation (vexpcal, vexpcalcul, vhcal, vhvar, instent, back_annot_obj);
  return vhresult;
}


/**
   UNUSED: local to instEquationnomb.
   @return the clock when OK, with (| OK := vh1 default OK$1 init false | OK ^= vh |)
*/
ptinsth kernel2graph::mkFilterCounter (ptinsth vh1, ptinsth vh, ptexpcal vexpcal)
{
  ptlinstances instent;
  ptexpcal vdims, vret, ok;
  TypeType TT, vtype;

  instent = NILLINSTANCES;
  EM->getTypeAndDim (EM->getUnindexedSignal (vexpcal), &vtype, &vdims);

  TT = get_BooleanType ();
  if (vdims != NILEXPCAL)
    TT = mkArraytype (vdims, TT);
  ok = igc->mkVariable (back_annot_obj, TT, &instent);
  EM->setClock (ok, vh);

  if (EM->getOperator (vexpcal) == opelem)
    ok = EM->mkBinaryExp (opelem, ok, EM->getBinaryArgstAt (vexpcal, 2), back_annot_obj);
  vret = EM->mkBinaryExp (opinit, ok, vexpfalse, back_annot_obj);
  vret = EM->mkTernaryExp (opret, vret, ptexp1, NILEXPCAL, back_annot_obj);
  igc->addSubEquation (ok, vexptrue, vh1, NILLINSTANCES, back_annot_obj);
  igc->createEquation (ok, vret, CKM->makeMinus (vh, vh1), vh, NILLINSTANCES, back_annot_obj);
  CKM->makeSub (vh, ok, instent);
  return CKM->makeNoSub (ok);
}

/** UNUSED: tThe counters from/after are replaced in the first step of the compiling 
    by a reference to a model of the predefined Signal Library. 
    
    INSTEQUATIONX for a "counter from" if first or "counter after" otherwise.
    - x := A from B is replaced by
    (| x:= cn when ^A | cn := (1 when ^A when ^B) default (0 when ^B) default (cn$1 init 0)+1 | cn ^= a ^+ b |)
    - x := A after B
    (| x:= cn when ^A | cn := (0 when ^B) default (cn$1 init 0)+1 | cn ^= a ^+ b |)
*/
ptinsth
kernel2graph::_instEquationnomb (boolean first, ptinsth vhcste, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptexpcal valexpcal, vexpcal1, vexpcal2, vret, cn;
  ptinsth vh, vh1, vh2, vhWhenNotOk, vh3, vh4;
  ptklExprSig v1klExprSig, v2klExprSig;
  boolean isfree;
  ptlinstances instent;
  TypeType ContextType, vtype;

  // Recall: the clock of the counter is the clock of the first operand.
  ContextType = get_PureType ();
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  isfree = pk_true;
  instent = NILLINSTANCES;
  vh1 = instOperand (vhcste, ContextType, &isfree, v1klExprSig, &vexpcal1, &instent, opnomba);
  vh2 = instOperand (vhcste, ContextType, &isfree, v2klExprSig, &vexpcal2, &instent, opnomba);
  vh = CKM->makeSup (vh1, vh2);

  instent = NILLINSTANCES;
  // May 2011: the type must be the same than the type of the result counter (array context).
  vtype = EM->getType (EM->getUnindexedSignal (vexpcal));
  cn = igc->mkVariable (back_annot_obj, vtype, &instent);
  EM->setClock (cn, vh);
  EM->setInitialValue (cn, ptexp0);
  if (EM->getOperator (vexpcal) == opelem)
    cn = EM->mkBinaryExp (opelem, cn, EM->getBinaryArgstAt (vexpcal, 2), back_annot_obj);

  if (first)			/* counter from */
    {
      igc->addSubEquation (cn, ptexp1, CKM->makeInf (vh1, vh2), NILLINSTANCES, back_annot_obj);
      igc->addSubEquation (cn, ptexp0, CKM->makeMinus (vh2, vh1), NILLINSTANCES, back_annot_obj);
    }
  else
    igc->addSubEquation (cn, ptexp0, vh2, NILLINSTANCES, back_annot_obj);

  EM->setInitialValue (vexpcal, ptexp0);
  vret = EM->mkBinaryExp (opinit, cn, ptexp0, back_annot_obj);
  vret = EM->mkTernaryExp (opret, vret, ptexp1, NILEXPCAL, back_annot_obj);
  valexpcal = EM->mkBinaryExp (opplus, vret, ptexp1, back_annot_obj);

  vhWhenNotOk = mkFilterCounter (vh2, vh, vexpcal);
  vh3 = CKM->makeInf (vh1, vhWhenNotOk);
  igc->addSubEquation (cn, ptexp0, CKM->makeMinus (vh3, vh2), NILLINSTANCES, back_annot_obj);
  vh4 = CKM->makeSup (vh2, vh3);
  igc->createEquation (cn, valexpcal, CKM->makeMinus (vh1, vh4), vh, NILLINSTANCES, back_annot_obj);

  return igc->createEquation (vexpcal, cn, vh1, vh1, instent, back_annot_obj);
}


/**
   INSTEQUATIONX for count operator (Exp1 count Exp2).
*/
ptinsth
kernel2graph::instEquationCount (ptinsth vhcste, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptexpcal valexpcal, vexpcal1, vexpcal2;
  ptinsth vh1, vh2;
  ptklExprSig v1klExprSig, v2klExprSig;
  boolean isfree;
  ptlinstances instent1, instent2;
  TypeType ContextType;

  // type(Exp1)  = event ; Exp2 must be a int constant expression.
  kl_getBinaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig);
  isfree = pk_true;

  instent1 = NILLINSTANCES;
  ContextType = get_PureType ();
  vh1 = instOperand (vhcste, ContextType, &isfree, v1klExprSig, &vexpcal1, &instent1, opnombm);

  instent2 = NILLINSTANCES;
  ContextType = get_IntegerType ();
  vh2 = instOperand (vh1, ContextType, &isfree, v2klExprSig, &vexpcal2, &instent2, opnombm);
  if (instent2 != NILLINSTANCES)
    pK_messenger->errorTree (errNotConstantExpr, currentTree);

  if (vexpcal2 != NILEXPCAL)
    EM->setInitialValue (vexpcal, EM->mkBinaryExp (opmoin, vexpcal2, ptexp1, back_annot_obj));
  else
    pK_messenger->errorTree (errNotConstantExpr, currentTree);
  valexpcal = EM->mkBinaryExp (opnombm, vexpcal1, vexpcal2, back_annot_obj);
  vh1 = igc->createEquation (vexpcal, valexpcal, vh1, vh1, NILLINSTANCES, back_annot_obj);
  NM->addMark (EM->getNode (vexpcal), DontOptimize);
  return vh1;
}



/**
   INSTEQUATIONX for an synchronous unary operand.  
*/
ptinsth
kernel2graph::instEquationOp1 (ptinsth vhcste, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vhresult, vhcal;
  ptlinstances instent;
  ptexpcal vexpcalcul;
  TypeType ContextType;
  ContextType = EM->getType (vexpcal);
  instent = NILLINSTANCES;
  vhcste = instoper1 (vhcste, ContextType, isfree, vklExprSig, &vexpcalcul, &instent);
  if (*isfree) vhcal = hinit; else vhcal = vhcste;
  vhresult = igc->createEquation (vexpcal, vexpcalcul, vhcal, vhcste, instent, back_annot_obj);
  return vhresult;
}


/** INSTEQUATIONX for an unary temporal operand (vklExprSig)
    operators : nullclock, clock, unary when.
*/
ptinsth
kernel2graph::instEquationOp1tp (ptinsth vhcste, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  noperators op;
  ptinsth vh;

  op = kl_getFlowExprOperator (vklExprSig);
  if (op == ophnull)
    return igc->createEquation (vexpcal, vexptrue, the_nullClock_K2G_copy, the_nullClock_K2G_copy, NILLINSTANCES, back_annot_obj);
  if (op == opevnt)
    {
      vh = igc->createEquation (vexpcal, vexptrue, hinit, instevent (vhcste, vklExprSig), NILLINSTANCES, back_annot_obj);
      if (!EM->isMarked (vexpcal, PART_DEFINITION)) EM->setType (vexpcal, get_PureType ());
      return vh;
    }
  if (op == opwhen1)
    {
      vh = igc->createEquation (vexpcal, vexptrue, hinit, instwhen1 (vhcste, pk_false, vklExprSig), NILLINSTANCES, back_annot_obj);
      if (!EM->isMarked (vexpcal, PART_DEFINITION)) EM->setType (vexpcal, get_PureType ());
      return vh;
    }
  return NILINSTH;
}


/**  INSTEQUATIONX for an synchronous binary operand.  */
ptinsth
kernel2graph::instEquationOp2 (ptinsth vhcste, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vh, vh1;
  ptlinstances instent;
  ptexpcal vexpcalcul;
  TypeType ContextType;

  instent = NILLINSTANCES;
  ContextType = EM->getType (vexpcal);
  vh1 = instoper2 (vhcste, ContextType, isfree, vklExprSig, &vexpcalcul, &instent);
  if (!(*isfree))
    {
      vh = igc->createEquation (vexpcal, vexpcalcul, vh1, vh1, instent, back_annot_obj);
    }
  else
    {
      if (CKM->satisfiesUndefClock (vh1))
	{
	  if (!CKM->satisfiesUndefClock (vhcste))
	    {
	      vh1 = vhcste;
	    }
	  else
	    {
	      vh1 = igc->setClockVariable (vexpcal);
	      if (CKM->satisfiesUndefClock (vh1))
		{
		  pK_messenger->errorTree (errcsteint, currentTree);
		  vh1 = CKM->new_variFormulaRootClock ();
		}
	    }
	}
      vh = igc->createEquation (vexpcal, vexpcalcul, hinit, vh1, instent, back_annot_obj);
    }
  igc->setAuxiliarySignalName (vexpcal, vexpcalcul, EM->getOperator (vexpcalcul));
  return vh;
}


/**
   INSTEQUATIONX for an binary temporal operand.
*/
ptinsth
kernel2graph::instEquationOp2tp (ptinsth vhcste, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptinsth vhresult;
  noperators operateur;

  operateur = kl_getFlowExprOperator (vklExprSig);
  switch ((int) operateur)
    {
    case opmele:
      return instEquationmele (vhcste, pk_false, isfree, vexpcal, vklExprSig);

    case opvar:
      return instEquationVar (vhcste, vexpcal, vklExprSig);

    case opextr:
      vhresult = instEquationextr (vhcste, vexpcal, vklExprSig);
      break;

    case opnombm:
      vhresult = instEquationCount (vhcste, vexpcal, vklExprSig);
      break;
      /*
	case opnomba:
	vhresult = instEquationnomb(pk_false, vhcste, vexpcal, vklExprSig);
	break;
	case opnombd:
	vhresult = instEquationnomb (pk_true, vhcste, vexpcal, vklExprSig);
	break;
      */
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  *isfree = pk_false;
  return vhresult;
}


/** @return an instance of the operand expression.
    Assumed: operand expression is a list of scalar expressions (bound expressions).
*/
ptexpcal
kernel2graph::mkDimExpcal (ptklExprSig trep)
{
  boolean isfree, b, sb;
  ptlinstances instent;
  ptexpcal vexpcal;
  ptinsth vh;

  if (trep == NILEXPRSIG)
    return NILEXPCAL;
  isfree = pk_true;
  instent = NILLINSTANCES;
  sb = EM->setArithmeticRuleMode (pk_true);
  b = igc->getSetConstantContext (pk_true);
  vh = instIntegerScalarListExpression (hundef, &isfree, trep, &vexpcal, &instent);
  b = igc->getSetConstantContext (b);
  EM->setArithmeticRuleMode (sb);
  return vexpcal;
}


/**
   INSTEQUATIONX for a ternary temporal operand (delay, window).
   Constants are not allowed for delayed signal.
*/
ptinsth
kernel2graph::instEquationOp3tp (ptinsth vhcste, boolean * isfree, ptexpcal vexpcal, ptklExprSig vklExprSig)
{
  ptklExprSig v1klExprSig, v2klExprSig, v3klExprSig, vinit, vinitDelay;
  ptlinstances instent, instent2;
  ptexpcal sv, vexpcal1, vexpcal2, vexpinit;
  ptinsth lvhcste, vh;
  boolean locfree, loc2free;
  cloper vcl;
  noperators op;
  TypeType ContextType;
  noperators contextId;

  
  vhcste = hundef;		// constants are not allowed for operand of the $ operator
  op = kl_getFlowExprOperator (vklExprSig);

  if (op == opmemo)
    {
      *isfree = pk_false;
      return instEquationCell (vhcste, vexpcal, vklExprSig);
    }

  kl_getTernaryExpr (vklExprSig, &v1klExprSig, &v2klExprSig, &vinit);
  if (op == opfen)
    {
      v3klExprSig = v2klExprSig;
      if (kl_getFlowExprOperator (v1klExprSig) == opret)
	{
	  kl_getTernaryExpr (v1klExprSig, &v1klExprSig, &v2klExprSig, &vinitDelay);
	  pK_messenger->warningacc (warIgnoreInit, currentTree);
	}
      else
	v2klExprSig = NILEXPRSIG;
    }
  else
    v3klExprSig = NILEXPRSIG;

  instExprInit (vinit, &vexpinit, &vh);	/* initial value */

  instent = NILLINSTANCES;
  instent2 = NILLINSTANCES;
  locfree = pk_true;
  vcl = getTypeExpflot (v1klExprSig);
  if (vcl == variabl)
    {
      lvhcste = instsigin (vhcste, &locfree, v1klExprSig, &vexpcal1, &instent);
      if (locfree || (vexpcal1 == vexptrue))
	getCurrentInstance (kl_getIdentifierExpr (v1klExprSig), 1, tsignal, &vexpcal1);
      sv = vexpcal1;
    }
  else if (vcl == call)
    {
      locfree = pk_false;
      if (!instExternalFunction (v1klExprSig, hundef, &locfree, &lvhcste, &sv, &instent))
	{
	  lvhcste = instCall (v1klExprSig, NILEXPCAL, NILEXPCAL, &sv, &instent);
	  sv = EM->getNaryArgs (&sv);
	  lvhcste = igc->setClockVariable (sv);
	}
    }
  else if (vcl == constant)
    {
      sv = instconst (v1klExprSig);
      lvhcste = igc->getSignalClockSynchronizedWith (vexpcal, hundef);
    }
  else
    {
      contextId = nameManager->setContextSubExpr (op);
      sv = igc->mkVariable (back_annot_obj, &instent);
      nameManager->setContextSubExpr (contextId);
      switch ((int) vcl)
	{
	case oper1:
	  lvhcste = instEquationOp1 (vhcste, &locfree, sv, v1klExprSig);
	  break;
	case oper1tp:
	  lvhcste = instEquationOp1tp (vhcste, sv, v1klExprSig);
	  locfree = pk_true;
	  break;
	case oper2:
	  lvhcste = instEquationOp2 (vhcste, &locfree, sv, v1klExprSig);
	  break;
	case oper2tp:
	  lvhcste = instEquationOp2tp (vhcste, &locfree, sv, v1klExprSig);
	  break;
	case oper3:
	  lvhcste = instEquationOp3 (vhcste, &locfree, sv, v1klExprSig);
	  break;
	case oper3tp:
	  lvhcste = instEquationOp3tp (vhcste, &locfree, sv, v1klExprSig);
	  break;
	default:
	  pK_messenger->Caseerror (currentLine, currentFile);
	}
    }

  if (!locfree)
    vhcste = lvhcste;
  lvhcste = igc->getSignalClockSynchronizedWith (vexpcal, lvhcste);
  CKM->makeSynchro (lvhcste, vh);

  vexpcal2 = NILEXPCAL;
  vexpcal1 = NILEXPCAL;
  ContextType = get_IntegerType ();
  if (v3klExprSig == NILEXPRSIG)
    if (v2klExprSig == NILEXPRSIG)
      if (locfree)
	pK_messenger->errorTree (errCannotDefineClock, currentTree);
      else
	lvhcste = igc->createEquation (vexpcal, sv, lvhcste, lvhcste, instent, back_annot_obj);
    else
      {
	loc2free = pk_true;
	lvhcste = CKM->makeSynchro (lvhcste, instOperand (vhcste, ContextType, &loc2free, v2klExprSig, &vexpcal1, &instent2, op));
	igc->mkDelayWindowNode (vexpcal, sv, vexpcal1, NILEXPCAL, vexpinit, lvhcste, back_annot_obj);
	locfree = (boolean) (locfree && loc2free);
      }
  else
    {
      loc2free = pk_true;
      lvhcste = CKM->makeSynchro (lvhcste, instOperand (vhcste, ContextType, &loc2free, v3klExprSig, &vexpcal2, &instent2, op));
      locfree = (boolean) (locfree && loc2free);
      if (v2klExprSig == NILEXPRSIG)
	if (locfree)
	  pK_messenger->errorTree (errCannotDefineClock, currentTree);
	else
	  {
	    instent2 = ptlinstances_sortedfusion (instent, instent2);
	    igc->mkDelayWindowNode (vexpcal, sv, NILEXPCAL, vexpcal2, vexpinit, lvhcste, back_annot_obj);
	  }
      else
	{
	  loc2free = pk_true;
	  lvhcste = CKM->makeSynchro (lvhcste, instOperand (vhcste, ContextType, &loc2free, v2klExprSig, &vexpcal1, &instent2, op));
	  igc->mkDelayWindowNode (vexpcal, sv, vexpcal1, vexpcal2, vexpinit, lvhcste, back_annot_obj);
	  locfree = (boolean) (locfree && loc2free);
	}
    }
  *isfree = locfree;

  if ((vexpcal1 != NILEXPCAL) || (vexpcal2 != NILEXPCAL))
    igc->setAuxiliarySignalName (vexpcal, sv, (vexpcal2 != NILEXPCAL) ? opfen : opret);

  if (locfree)
    pK_messenger->errorTree (errcsteint, currentTree);
  return lvhcste;
}

/**
   Compiling an equation or a partial definition or a default-value expression.
   @param vexpcal (in) the instanciated left part of the equation.
   @param vhdis (in) context clock to fix the clock of the constants.
   @param vklExprSig (in) expression to compile;
   @param codeop (in)  equation, partial definition or default-value code operator.
   @return clock of vexpcal synchronized to the clock of the right part expression (ie instanciated expression of vklExprSig).
*/
ptinsth
kernel2graph::instEquation (ptinsth vhdis, ptexpcal vexpcal, ptklExprSig vklExprSig, noperators codeop)
{
  boolean isfree;
  ptlinstances instent;
  ptexpcal vexpcalcul, linstsort;
  boolean bPartDef, x;
  ptinsth vh1;
  noperators contextId;
  ptident id;

  if (isTuple (vexpcal))
    // Restriction for tuples; only cartesian product allowed.
    if (kl_getFlowExprOperator (vklExprSig) != oppcar)
      {
	pK_messenger->errorTree (errnotimplemented, currentTree);
	return the_nullClock_K2G_copy;
      }



  x = setAccessMode (pk_true);
  bPartDef = (codeop != opdefsig);
  isfree = pk_true;
  switch ((int) (getTypeExpflot (vklExprSig)))
    {
    case call:
      instent = NILLINSTANCES;
      if (bPartDef) vexpcalcul = NILEXPCAL; else vexpcalcul = EM->mkNaryExp (oplident, vexpcal, back_annot_obj);
      vhdis = instCall (vklExprSig, vexpcalcul, NILEXPCAL, &linstsort, &instent);
      if (bPartDef)
	{
	  vexpcalcul = EM->getNaryArgs (&linstsort);
	  if (igc->topContextClock (&vh1))
	    vhdis = CKM->makeInf (vhdis, vh1);
	  vhdis = igc->createEquation (vexpcal, vexpcalcul, vhdis, vhdis, instent, back_annot_obj);
	}
      break;
    case constant:
      vhdis = igc->instContextClock (vhdis, back_annot_obj);
      vhdis = igc->createEquation (vexpcal, instconst (vklExprSig), hinit, vhdis, NILLINSTANCES, back_annot_obj);
      break;
    case variabl:
      instent = NILLINSTANCES;
      vhdis = instsigin (vhdis, &isfree, vklExprSig, &vexpcalcul, &instent);
      if (isfree)
	{
	  vhdis = igc->setClockVariable (vexpcal);
	  vhdis = igc->createEquation (vexpcal, vexpcalcul, hinit, vhdis, instent, back_annot_obj);
	}
      else
	vhdis = igc->createEquation (vexpcal, vexpcalcul, vhdis, vhdis, instent, back_annot_obj);
      break;
    case oper1:
      id = setContextSubExpr (vexpcal);
      vhdis = instEquationOp1 (vhdis, &isfree, vexpcal, vklExprSig);
      nameManager->setContextSubExpr (id);
      break;
    case oper2:
      id = setContextSubExpr (vexpcal);
      vhdis = instEquationOp2 (vhdis, &isfree, vexpcal, vklExprSig);
      nameManager->setContextSubExpr (id);
      break;

    case oper3:
      vhdis = instEquationOp3 (vhdis, &isfree, vexpcal, vklExprSig);
      break;

    case oper1tp:
      vhdis = instEquationOp1tp (vhdis, vexpcal, vklExprSig);
      break;

    case oper2tp:		// for X := E1 default E2, similar to X::=E1, X::=E2 for the naming rules.
      id = setContextSubExpr (vexpcal);
      vhdis = instEquationOp2tp (vhdis, &isfree, vexpcal, vklExprSig);
      nameManager->setContextSubExpr (id);
      break;

    case oper3tp:
      if (bPartDef)
	{
	  // A revoir : induit des modifs dans (mk_node_delay)
	  instent = NILLINSTANCES;
	  contextId = nameManager->setContextSubExpr (kl_getFlowExprOperator (vklExprSig));
	  vexpcalcul = igc->mkVariable (back_annot_obj, &instent);
	  nameManager->setContextSubExpr (contextId);
	  vhdis = instEquationOp3tp (vhdis, &isfree, vexpcalcul, vklExprSig);
	  if (igc->topContextClock (&vh1))
	    vhdis = CKM->makeInf (vhdis, vh1);
	  vhdis = igc->createEquation (vexpcal, vexpcalcul, vhdis, vhdis, instent, back_annot_obj);
	}
      else
	vhdis = instEquationOp3tp (vhdis, &isfree, vexpcal, vklExprSig);
      break;

    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  setAccessMode (x);
  return vhdis;
}


/**
   @return the next element of the operand list (lentrees), nil if it does not exist.
   Local to instposit() method.
*/
ptlrepid
kernel2graph::getlentcour (ptlrepid lentrees, ptklExprSig vklExprSigarg)
{
  ptlrepid lsort, laux;
  cloper t;

  if (lentrees == NILLREPID) return NILLREPID;
  t = getTypeExpflot (vklExprSigarg);
  if (t != call) return succid (lentrees);
  lsort = kl_getInsOutsModel (getModelFromExpflot (vklExprSigarg), &laux);
  laux = lentrees;
  while (lsort != NILLREPID)
    {
      laux = succid (laux);
      lsort = succid (lsort);
    }
  return laux;
}

/** 
    @return "true" if l is not an empty list, "false" otherwise.
    @param l (in) a list of declaration of flows (kernel representation).
    @param vrepid (out) next element of l.
    @param TT (out) the type of the first element of l.
*/
boolean
kernel2graph::getTypeFirstElement (ptlrepid l, ptlrepid * vrepid, TypeType * TT)
{
  ntrees xtrees;
  ptreptyp vtyp;
  ptklExprSig vinit;
  categoryType vcat;

  if (!ptlrepid_getRepidAt (l, 1, vrepid)) return pk_false;
  if (!ptlrepid_getIdInfosAt (l, 1, &vinit, &vtyp, &vcat, &xtrees)) return pk_false;
  *TT = sigTreeType2TypeType (vtyp);
  return pk_true;
}

/**
   @return a new signal associated with the firt element of the list (lentrees) with vh as clock.
   This signal is added to the list instent. It is defined by the instanciated expression of vklExprSigarg expression.
   Local to instposit() method.
*/
ptexpcal
kernel2graph::instpositCreateAff (ptlrepid lentrees, ptklExprSig vklExprSigarg, ptinsth vh, ptlinstances * instent)
{
  ptexpcal vexpcal;
  vexpcal = igc->mkVariable (back_annot_obj, instent);
  if (vh == hundef) vh = igc->getSignalClockSynchronizedWith (vexpcal, vh);
  vh = instEquation (vh, vexpcal, vklExprSigarg, opdefsig);
  return vexpcal;
}

/**
   @return the instantiated expression associated with the operand expression (vklExprSiglarg). 
   The returned list is the output signals created during the instanciation. 
   It eventually takes the element of lentrees as model (if it is not empty).
   The creatred signals are added to the list instent.
*/
ptexpcal
kernel2graph::instposit (ptlrepid lentrees, ptklExprSig vklExprSiglarg, ptlinstances * instent, ptinsth hcontext)
{
  ptklExprSig vklExprSigarg;
  ptexpcal vexpcal, linstentpos;
  ptinsth vh;
  boolean isfree;
  noperators op;
  if (vklExprSiglarg == NILEXPRSIG) return NILEXPCAL;

  vklExprSigarg = kl_getNextExpr (&vklExprSiglarg);
  linstentpos = instposit (getlentcour (lentrees, vklExprSigarg), vklExprSiglarg, instent, hcontext );

  switch ((int) (getTypeExpflot (vklExprSigarg)))
    {
    case call:
      isfree = pk_true;
      vh = instCall (vklExprSigarg, NILEXPCAL, linstentpos, &linstentpos, instent);
      break;
    case variabl:
      isfree = pk_true;
      vh = instsigin (hundef, &isfree, vklExprSigarg, &vexpcal, instent);
      if (vexpcal == vexptrue)
	{
	  getCurrentInstance (kl_getIdentifierExpr (vklExprSigarg), 1, tsignal, &vexpcal);
	  *instent = ptlinstances_insert (*instent, vexpcal);
	}
      linstentpos = EM->mkNaryExp (oplesign, vexpcal, linstentpos, back_annot_obj);
      break;
    case oper2:
      op = kl_getFlowExprOperator (vklExprSigarg);
      if (op == opvenum)
	vexpcal = instRefEnumValue (vklExprSigarg);
      else if ((op == opcham) || (op == opelem))
	{
	  isfree = pk_true;
	  vh = instOperand (hundef, UndefinedType, &isfree, vklExprSigarg, &vexpcal, instent, oplesign);
	}
      else
	vexpcal = instpositCreateAff (lentrees, vklExprSigarg, hcontext, instent);
      linstentpos = EM->mkNaryExp (oplesign, vexpcal, linstentpos, back_annot_obj);
      break;
    case oper1tp:
      op = kl_getFlowExprOperator (vklExprSigarg);
      if (op == ophnull)
	vexpcal = EM->mkWhenOrCellExp (opextr, vexptrue, the_nullClock_K2G_copy);
      else // ^ 
	vexpcal = instpositCreateAff (lentrees, vklExprSigarg, hcontext, instent);
      linstentpos = EM->mkNaryExp (oplesign, vexpcal, linstentpos, back_annot_obj);
      break;
    case constant:
    case oper1:
    case oper3:
    case oper2tp:
    case oper3tp:
      vexpcal = instpositCreateAff (lentrees, vklExprSigarg, hcontext, instent);
      linstentpos = EM->mkNaryExp (oplesign, vexpcal, linstentpos, back_annot_obj);
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return linstentpos;
}

/**  
     Type Checking for a formal expression (lsigfor) and the effective expression (sigeff) associated with, for a call.
     Type checking for a formal signal (lsigfor) and an effective signal (sigeff)
     @param lsigfor (in) a formal signal.
     @param sigeff (in) an effective expression
     @param inStack (in) must be "true" when it is an input signal of the call, "false" for an output. It is used for type checking.
     @param vexplsigeff (in) a list of expression in which the current element (sigeff) will be added.
     @param decltree (in) trace information (parsing), used to add error/warning messages.
     @return the list vexplsigeff in which the effective expression (sigeff) is added at the head.
     
     Local to ctexternemacro() method.
*/
ptexpcal
kernel2graph::formalEffectiveTypeChecking (ptlrepid lsigfor, boolean inStack, ptexpcal sigeff, ptexpcal vexplsigeff, ntrees decltree)
{
  ptexpcal vinit;
  ptklExprSig vali;
  ptinsth vh;
  categoryType vcat;
  ntrees arb;
  ptreptyp vtyp;
  TypeType TT;

  ptlrepid_getIdInfosAt (lsigfor, 1, &vali, &vtyp, &vcat, &arb);
  TT = sigTreeType2TypeType (vtyp);
  EM->TypeCheckingFormalEffective (TT, sigeff, inStack, MK_BACK_ANNOT(decltree));

  if (instExprInit (vali, &vinit, &vh))
    {
      if (EM->getInitialValue (sigeff) == NILEXPCAL)
	EM->setInitialValue (sigeff, vinit);	/* else le signaler a l'utilisateur */
    }
  return EM->mkNaryExp (oplident, sigeff, vexplsigeff, MK_BACK_ANNOT(decltree));
}


/**
   @retrun the expression // x? == #vLabelName, where x is the name of the current state of the compiled automaton.
*/
ptexpcal  kernel2graph::mkTestStateIsIn(ptexpcal vLabelName)
{
  ptexpcal vexp, vCurrentState;
  // x? == #lab
  vCurrentState = igc->getAutomatonCurrentStateVar ();
  vexp = EM->mkUnaryExp (opvvarin, vCurrentState, back_annot_obj);
  return EM->mkBinaryExp (opegal, vexp, mkExprLabelEnumValue (vCurrentState, vLabelName), back_annot_obj);
}

/**
   Assumed: vklExprProc is a state of an automaton (e::P(), and e is referenced in a 
   transition, detected in the firt step (Trees2Kernel translation) of the compiler.)
*/
ptinsth
kernel2graph::mkStateAutomatonUndersamplingClock (ptexpcal vLabelName, ptinsth vhContext)
{
  ptinsth vhsub;

  //  hsub = [: x? == #lab], where lab is the name of the labelled associated with a state of an automaton.
  vhsub = CKM->makeSub (igc->getClockAutomaton (),  mkTestStateIsIn(vLabelName) , NILLINSTANCES);
  return CKM->makeInf (vhsub, vhContext);
}

ptexpcal kernel2graph::mkExprLabelEnumValue (ptexpcal vCurrentStateVar, ptexpcal vlabel)
{
  boolean ok;
  TypeType TT;
  ptident id;
  NameList NL;
  ptexpcal vexpType;
  int i;
  ptrepstr vidstr;

  if (!EM->IsLabel (vlabel)) return NILEXPCAL;
  // boolean baux = EM->IsAuxiliarySignal(vCurrentStateVar);
  // baux => on peut modifier le type du signal (ajout d'une valeur d'enum )
  TT = EM->getType (vCurrentStateVar);
  id = EM->getName (vlabel);
  vidstr = ident2repstr (id);
  ok = pk_true;
  if (TypeTestAnyClass (TT))
    {
      NL = NameListCreate ();
      NL = NameListRegister (NL);
      TT = EnumerationTypeCreate (ident2repstr (nameManager->mkNameType (string2ident ("StateType"))), NL);
      EM->setType (vCurrentStateVar, TT);
    }
  else if (!TypeTestEnumerationClass (TT))
    {
      pK_messenger->errorTree (errElemType, currentTree);
      ok = pk_false;
      vexpType = NILEXPCAL;
    }

  if (ok)
    {
      vexpType = EM->mkExpType (TT);
      // adding the value if it does not exist in TT
      if (!EnumTypeOrdinal (TT, vidstr, &i))
	{
	  NL = EnumerationTypeGetNameLst (TT);
	  NameListAppendName (NL, (NameClass) vidstr);
	}
    }
  return EM->mkBinaryExp (opvenum, vexpType, EM->mk_enumPtexpcal (vidstr), back_annot_obj);
}

boolean
kernel2graph::isintrinsicSignalThread (ptident name, ptexpcal parameters, ptexpcal inputs, ptexpcal outputs, ptinsth * vh)
{
  ptexpcal out, in; 
 if (isThreadBooleanClockId (name))
    { 
      if (isSignalThreadModel)
	{
	 out = EM->getBinaryArgstAt (outputs, 1);
	 in  = EM->getBinaryArgstAt (inputs, 1);
	 (*vh) = genDefinitionsOutputBooleanClocks(out, in);
	 // out := assignmentClock(in) default (false when Dispatch) default (out $1 init false)
	 return pk_true;
	}
      else ; // ERROR 
    }

    if (isThreadClockId (name))
      {
	if (isSignalThreadModel) 
	  {
	   out = EM->getBinaryArgstAt (outputs, 1);
	   *vh = igc->createEquation (out, vexptrue,  vhOversamplingSignalThread,  vhOversamplingSignalThread, NILLINSTANCES, back_annot_obj);
	    return pk_true;
	  } 
	else ; // ERROR
      }
    return pk_false;
}


/**
   Compiling a predefined automaton intrinsic (local to \ref instintrinsicfunc method), predefined in Signal library:
   Such an intrinsic is used only in a automaton model (verified before). 
   Some of  them use the label context managed for an automaton. (See kernel2graph::pushLabel())

   @param parameters (in) paramaters of the call.
   @param inputs (in) inputs of the call.
   @param outputs (in) outputs of the call.
   For these, the clock of the expression (vh) is returned

   @return "true" if name is the predefined automaton intrinsic model, "false" otherwise.
*/
boolean
kernel2graph::isintrinsicAutomaton (ptident name, ptexpcal parameters, ptexpcal inputs, ptexpcal outputs, ptinsth * vh)
{
  ptexpcal vout, vexp, vinit, lab1, lab2, vTriggerExp;
  ptinsth vhcal, vTriggerTransition;
  ptexpcal partialDefExp, vCurrentState;

  *vh = hundef;
  //  Automaton_SetCurrentStateId(CurState), CurState is the state variable that specifies the current state of the automaton.
  if (isSetCurrentStateId (name)) return pk_true;

  // Automaton_Initial_State( Statelabel), the initial state of the automaton, Statelabel is a label.
  if (isInitialStateId (name))
    {
      vCurrentState = igc->getAutomatonCurrentStateVar ();
      vinit = EM->getInitialValue (vCurrentState);
      if (vinit != NILEXPCAL)
	pK_messenger->warningaccForce (warIgnoreInit, EM->getTreeOfSignalDeclaration (vCurrentState));

      EM->setInitialValue (vCurrentState, mkExprLabelEnumValue (vCurrentState, EM->getBinaryArgstAt (inputs, 1)));
      return pk_true;
    }

  if (isAutomatonIsIn (name))
    {
      vexp = mkTestStateIsIn(EM->getBinaryArgstAt (inputs, 1));
      vout = EM->getBinaryArgstAt (outputs, 1);
      vhcal = igc->getClockAutomaton ();
      *vh = igc->createEquation (vout, vexp, vhcal, vhcal, NILLINSTANCES, back_annot_obj);
      return pk_true;
    }

  if (isAutomatonSetClockId (name))
    {
      if (igc->isClockedAutomaton())
	{
	  pK_messenger->errorTree (errClockAutomatonAlreadyFixed, currentTree);
	}
      else
	{
	  igc->setClockedAutomaton(true);
	  // vexp = EM->getBinaryArgstAt (inputs, 1);
	  if (EM->getExprClock( EM->getBinaryArgstAt (inputs, 1) , vh))
	    {
	      CKM->makeSynchro ( igc->getClockAutomaton (), *vh);
	      //  CKM->makeSynchro ( EM->getClock(vexp),  *vh);
	    }
	  else
	    pK_messenger->errorTree (errCannotDefineClock, currentTree);
	}
      return pk_true;
    }

  if (isAutomatonClockId (name))
    {
      vhcal = igc->getClockAutomaton ();
      *vh = igc->createEquation (EM->getBinaryArgstAt (outputs, 1), vexptrue,  vhcal,  vhcal, NILLINSTANCES, back_annot_obj);
      return pk_true;
    }

   if (!isTransitionId (name)) return pk_false;

   // Automaton_Transition
   lab1 = EM->getBinaryArgstAt (inputs, 1);
   ptexpcal vaux = EM->getBinaryArgstAt (inputs, 2);
   lab2 = EM->getBinaryArgstAt (vaux, 1);
   vaux = EM->getBinaryArgstAt (vaux, 2);
   vTriggerExp = EM->getBinaryArgstAt (vaux, 1);

   ptexpcal vPrio = EM->getBinaryArgstAt (vaux, 2); 
   vPrio = EM->getBinaryArgstAt (vPrio, 1);
   compilerInteger val = 0;
   if (! EM->IsConstant (vPrio))
     {
       if ( ! EM->getIntegerValue(vPrio, &val))
	 pK_messenger->errorTree (errIntValueExpected, currentTree);
     }

   EM->getExprClock (vTriggerExp, &vTriggerTransition);
   vCurrentState = igc->getAutomatonCurrentStateVar ();
   
   vhcal = mkStateAutomatonUndersamplingClock (lab1, vTriggerTransition);
   
   partialDefExp = mkExprLabelEnumValue (vCurrentState, lab2);

   EM->addAttribute (vCurrentState, PRIORITY, (attributeValue) val);
   *vh = igc->createPartialEquation(vCurrentState, partialDefExp, vhcal, vhcal, NILLINSTANCES, back_annot_obj);
   EM->removeAttribute(vCurrentState, PRIORITY);

   // If lab2 is an element of the igc->getCompleteStatesAutomaton (), a partial definition of the complete signal is added
   return pk_true;
}

/**
   Compiling a predefined "affine clock" model (local to \ref instintrinsicfunc method) when name is the name of one of the 
   following process (predefined in Signal library):
   - process affine_sample={int phi, d} (? INE; ! OUT;)
   - process affine_clock_relation={int n, phi, d} (? IN1, IN2;)
   - process affine_unsample={int n, phi} (? INE1, INE2; ! OUT;)
   @param parameters (in) paramaters of the call.
   @param inputs (in) inputs of the call.
   @param outputs (in) outputs of the call.
   For these, the clock of the expression (vh) is returned according to the semantic of the affine operators.

   @return "true" if name is an affine clock predefined intrinsic function?.
   "false" otherwise.
*/
boolean
kernel2graph::isintrinsicAffine (ptident name, ptexpcal parameters, ptexpcal inputs, ptexpcal outputs, ptinsth * vh)
{
  ptinsth vh1, vh2, vh3;
  ptexpcal x, y, z;
  ptinsth vhx;
  ptexpcal n, phi, d;
  compilerInteger val;
  boolean b, bval;

// affine_clock_relation
  if (isAffineClkId (name))  
    {
      /*  parameters={n, phi, d} with n>0, d>0.  */
      x = EM->getNaryArgs (&inputs);
      y = EM->getNaryArgs (&inputs);
      n = EM->getNaryArgs (&parameters);
      EM->hasPositiveIntegerType (n, back_annot_obj, pk_true, &bval, pk_true);
      b = bval;

      phi = EM->getNaryArgs (&parameters);
      if (!EM->getIntegerValue (phi, &val))
	b = pk_false;

      d = EM->getNaryArgs (&parameters);
      EM->hasPositiveIntegerType (d, back_annot_obj, pk_true, &bval, pk_true);
      if (!bval)
	b = pk_false;

      if (!b)
	{
	  pK_messenger->errorTree (errRestrictionAffineClockValue, currentTree);
	  return pk_false;
	}

      igc->makeAffineClock (n, phi, d, &vh2, &vh3);
      vh2 = igc->getSignalClockSynchronizedWith (x, vh2);
      vh3 = igc->getSignalClockSynchronizedWith (y, vh3);
      *vh = the_nullClock_K2G_copy;
      return pk_true;
    }
  
  // affine_sample
  if (isSampleId (name))
    {
      /*  parameters={phi, d} with phi>=0, d>0. */
      x = parameters;
      phi = EM->getNaryArgs (&x);
      EM->hasPositiveIntegerType (phi, back_annot_obj, pk_false, &bval, pk_true);
      b = bval;

      d = EM->getNaryArgs (&x);
      EM->hasPositiveIntegerType (d, back_annot_obj, pk_true, &bval, pk_true);
      if (!bval)
	b = pk_false;
      if (!b)
	{
	  pK_messenger->errorTree (errRestrictionAffineClockValue, currentTree);
	  return pk_false;
	}

      x = EM->getNaryArgs (&inputs);
      vhx = igc->setClockVariable (x);
      vh1 = igc->makeSampleClock (parameters, vhx);
      y = EM->getNaryArgs (&outputs);
      vh1 = igc->getSignalClockSynchronizedWith (y, vh1);
      igc->mkSampleNode (y, x);
      *vh = vh1;
      return pk_true;
    }

  // affine_unsample
  if (isUnsampleId (name))
    {
      /*  parameters={n, phi} with n>0. */
      x = parameters;
      n = EM->getNaryArgs (&x);
      EM->hasPositiveIntegerType (n, back_annot_obj, pk_true, &bval, pk_true);
      b = bval;

      phi = EM->getNaryArgs (&x);
      if (!EM->getIntegerValue (phi, &val))
	b = pk_false;
      if (!b)
	{
	  pK_messenger->errorTree (errRestrictionAffineClockValue, currentTree);
	  return pk_false;
	}

      vh1 = igc->makeUnsampleClock (parameters, &vh2);
      x = EM->getNaryArgs (&inputs);
      vh2 = igc->getSignalClockSynchronizedWith (x, vh2);
      z = EM->getNaryArgs (&inputs);
      vh1 = igc->getSignalClockSynchronizedWith (z, vh1);
      y = EM->getNaryArgs (&outputs);
      vh1 = igc->getSignalClockSynchronizedWith (y, vh1);
      igc->mkUnsampleNode (y, x, z);
      *vh = vh1;
      return pk_true;
    }
  return pk_false;
}


/**
   Compiling a predefined model, except the SIGALI ones.
   @return "true" if the model is a predefined one, "false" otherwise.
*/
boolean
kernel2graph::instintrinsicfunc (ptrepfi vfi, ptident nom, ptlinstances lent, ptlinstances lsort,
				 ptexpcal effParam, ptexpcal lentren, ptexpcal lsortren, boolean bFunction, ptinsth * vh)
{
  ptINode vinstr;
  ptlpragmas lprag;
  ptpragma vpragma;
  boolean b;
  ptklExprProc vspec;
  lprag = kl_getPragmaModel (vfi);
  if (getPragmaKeyword (lprag, PRAGMA_SIGALI, &vpragma)) return pk_false;
  if (!kl_isIntrinsicModel (vfi)) return pk_false;
  if (isintrinsicAffine (nom, effParam, lentren, lsortren, vh)) return pk_true;
  if (isintrinsicAutomaton (nom, effParam, lentren, lsortren, vh)) return pk_true;
  if (isintrinsicSignalThread (nom, effParam, lentren, lsortren, vh)) return pk_true;
  b = pk_true;
  if (kl_getSpecModel (vfi, &vspec))
    *vh = compilingEproc (vspec);
  else
    *vh = CKM->makeSup (igc->upperBoundClockOfIdentifiers (lent), igc->upperBoundClockOfIdentifiers (lsort));
  if (bFunction)
    *vh = CKM->makeSynchro (igc->synchronizeIdentifiers (lent), igc->synchronizeIdentifiers (lsort));
  if (igc->mkProcCallNode (nom, effParam, lentren, lsortren, *vh, kl_getSafetyProperty (vfi), back_annot_obj,  MK_BACK_ANNOT(kl_getTreeReference (vfi)), b, &vinstr))
    while (ptlpragmas_get_next (&lprag, &vpragma))
      NM->addPragma (vinstr, vpragma);
  return pk_true;
}

/**
   It adds to a node (vinstr):
   - the pragmas associated with the model (vfi),  
   - the external language implementation (vlang).
*/
void
kernel2graph::addPragmasAndAttributesProcCall (ptrepfi vfi, ptINode vinstr, ntrees vlang)
{
  ptlpragmas lprag;
  ptpragma vpragma;

  lprag = kl_getPragmaModel (vfi);
  while (ptlpragmas_get_next (&lprag, &vpragma))
    NM->addPragma (vinstr, vpragma);
  NM->addAttribute (vinstr, EXTERN_LANGUAGE_IMPL, (attributeValue) vlang);
}

/**
   It adds to function call expression (e):
   - the pragmas associated with the model (vfi),  
   - the external language implementation (vlang).
*/
void
kernel2graph::addPragmasAndAttributesProcCall (ptrepfi vfi, ptexpcal e, ntrees vlang)
{
  ptlpragmas lprag;
  ptpragma vpragma;

  lprag = kl_getPragmaModel (vfi);
  while (ptlpragmas_get_next (&lprag, &vpragma))
    EM->addPragma (e, vpragma);
  EM->addAttribute (e, EXTERN_LANGUAGE_IMPL, (attributeValue) vlang);
  EM->setBackAnnotation (e, MK_BACK_ANNOT(kl_getTreeReference (vfi)));
  if (kl_isIntrinsicModel (vfi))
    EM->addMark (e, INTRINSIC);
}

/** 
    @return the combed list of instanciated expressions of a list of expressions.
    @param list (in) the left part of a definition (signal, vector element).
*/
ptinsth
kernel2graph::instlident (boolean bpartdef, ptklExprSig list, ptexpcal * vres, boolean * berror)
{
  ptexpcal vexpcal, vlst;
  ptklExprSig e;
  ptinsth vh, h;
  boolean b;
  if (list == NILEXPRSIG)
    {
      *vres = NILEXPCAL;
      *berror = pk_false;
      return hundef;
    }

  e = kl_getNextExpr (&list);
  h = instElemLeftPartDefinition (bpartdef, e, &vexpcal, &b);
  if (b)
    *berror = pk_true;

  if (list == NILEXPRSIG)
    {
      *vres = EM->mkNaryExp (oplident, vexpcal, back_annot_obj);
      return h;
    }
  vh = instlident (bpartdef, list, &vlst, berror);
  *vres = EM->mkNaryExp (oplident, vexpcal, vlst, back_annot_obj);
  return CKM->makeSup (vh, h);
}

/**
   @return the clock of the resulting instanciated expression.
   It instanciates exp in vexpr when exp is a left part of an equation ( exp :=) or a partial definition (exp ::=).
   The boolean bpartdef is "true" for a partial definition, "false" for an equation.
   Some controls are done during this method (double definition,..). 
   The returned code (berror) is set to "true" when anomalies are detected, not modified otherwise.
*/
ptinsth
kernel2graph::instElemLeftPartDefinition (boolean bpartdef, ptklExprSig exp, ptexpcal * vexpr, boolean * berror)
{
  boolean isfree;
  ptlinstances instent;
  ptinsth vh, vh1;

  isfree = pk_true;
  instent = NILLINSTANCES;
  vh = instOperand (hundef, UndefinedType, &isfree, exp, vexpr, &instent, opdefsig);
  if (*vexpr == vexptrue)
    *vexpr = getCurrentInstance (kl_getIdentifierExpr (exp));
  (*berror) = !igc->verifyDefSig (*vexpr, back_annot_obj);
  ptlinstances_free (instent);
  if (bpartdef) if (igc->topContextClock (&vh1)) return vh1;
  return vh;
}


/**
   Compiling of the left part of an equation or a partial equation.
   @return the clock of the resulting instanciated expression. 
   @param exp (in) the left part of a definition (signal, vector element). It can be a list or single element.
   @param vres (out) the combed list of instanciated expressions of a list of expressions (i.e. when blist), or the instanciated expression.
   @param blist (in) must be pk_true when the expression must be inserted in a list (vres output parameter), "false" otherwise.
   @param bpartdef (in) assumed to be "true" for a partial definition, "false" for an equation.
*/
ptinsth
kernel2graph::instLeftPartDefinition (boolean bpartdef, ptklExprSig exp, boolean blist, ptexpcal * vres, boolean * berror)
{
  boolean b, x;
  ptinsth vh;
  noperators op;
  x = setAccessMode (pk_false);
  b = igc->getSetStateVarContext (pk_true);
  op = kl_getFlowExprOperator (exp);
  if (_isalist_ (op))
    {
      vh = instlident (bpartdef, exp, vres, berror);
    }
  else
    {
      vh = instElemLeftPartDefinition (bpartdef, exp, vres, berror);
      if (blist) (*vres) = EM->mkNaryExp (oplident, *vres, back_annot_obj);
    }

  b = igc->getSetStateVarContext (b);
  x = setAccessMode (x);
  return vh;
}

/**
   Compiling a partial definition or a default-value expression (i.e. ::= or ::= defaultvalue )
   @param vexpcal (in) the instanciated left part of the equation.
   @param vklExprSig (in) expression to compile: the right part of the expression.
   @param code (in) the code of the node (operator of ::=, operator of ::= defaultvalue)
   @param contextClock (in) context clock to fix the clock of the constants.
   @return clock of vexpcal synchronized to the clock of the right part expression (ie instanciated expression of vklExprSig).
*/
ptinsth
kernel2graph::instPartialDefinitionEqu (ptinsth contextClock, ptexpcal vexpcal, ptklExprSig vklExprSig, int code)
{
  ptinsth clk;
  boolean bPartDefDefault;

  // ::= or ::= defaultvalue 
  bPartDefDefault = (code == opdefpartdef);
  if (bPartDefDefault) 
    EM->addMark (vexpcal, PART_DEFAULT_DEFINITION);
  else if (!igc->topContextClock (&contextClock))
    contextClock = hundef;

  EM->addMark (vexpcal, PART_DEFINITION);
  clk = instEquation (contextClock, vexpcal, vklExprSig, code);
  EM->removeMark (vexpcal, PART_DEFINITION);

  if (bPartDefDefault) EM->removeMark (vexpcal, PART_DEFAULT_DEFINITION);
  return clk;
}


/** 
    Instanciation of a list of labels (vlrepid).
    To each element of the list is associated with a new label identifier.
    This new identifier becomes the top of the stack (current instance) associated with the element.
*/
void
kernel2graph::instlabels (ptlrepid vlrepid)
{
  int numero;
  boolean notEmpty;
  ptexpcal vexpcal;
  ptlrepid vrepid;
  ptinsth vh;

  numero = 1;
  notEmpty = pk_true;
  while (notEmpty)
    {
      notEmpty = ptlrepid_getRepidAt (vlrepid, numero, &vrepid);
      if (notEmpty)
	{
	  if (ptrepproto_isLabelStateAutomaton (getIdRepid (vrepid)))  vh = igc->getClockAutomaton (); else vh = hundef;
	  vexpcal = EM->mkLabelIdent (identpile (nameId (vrepid)), NM->mk_node_undefined (), vh, back_annot_obj, pk_false);
	  setCurrentInstance (vrepid, vexpcal);
	  numero = numero + 1;
	}
    }
}

/**
   Instanciation of a list of declaration of signals (vlrepid).  
   A new signal is created in the top of the stack for each element of the operand list.
   When the boolean isExternRef is  "true", a new signal, associated with  a model declared as "statevar",
   is not declared as a statevar (used in unexpanded compiling).
   @return none.
*/
void
kernel2graph::sigTreeFlows2DCG (ptlrepid vlrepid, boolean isExternRef)
{
  int numero;
  boolean notEmpty;
  ntrees savetree;
  ptexpcal vexpcal;
  BackAnnotationType save_back_annot_obj;

  savetree = currentTree;
  save_back_annot_obj = back_annot_obj;

  numero = 1;
  notEmpty = pk_true;
  while (notEmpty)
    {
      notEmpty = mkAndPushInstance (vlrepid, numero, tsignal, isExternRef, &vexpcal);
      numero = numero + 1;
    }
  currentTree = savetree;
  back_annot_obj = save_back_annot_obj;
}


/**
   Instantiation of a list of generic parameters from the pos-th element.
   It assumes that the types declared as parameters are in param[1, pos-1] (\ref sigTreeTypeParam2DCG): if a type is found, an error is detected.
   A new parameter identifier is created in the top of the stack for each element of the operand list.
   Local to sigTreeInterface2DCG().
*/
void
kernel2graph::sigTreeGenericParam2DCG (ptlrepid param, integer pos)
{
  boolean notEmpty;
  ptexpcal vexpcal;
  ptlrepid vrepid;
  TypeType TT;

  notEmpty = pk_true;
  while (notEmpty)
    {
      notEmpty = ptlrepid_getRepidAt (param, pos, &vrepid);
      if (notEmpty)
	{
	  if (isTypeId (vrepid))	// type in parameters
	    {
	      pK_messenger->errorTree (errTypeParamNotAllowed, currentTree);
	      TT = makeMetaType (vrepid);
	    }
	  else if (isModelId (vrepid))	// formal process in parameters
	    {
	      pK_messenger->errorTree (errMainModelWithFormalProcess, currentTree);
	    }
	  else
	    {
	      notEmpty = mkAndPushInstance (param, pos, tparam, pk_false, &vexpcal);
	      TT = EM->getType (vexpcal);
	    }
	  pos++;
	}
    }
}

/**
   Instantiation of a list of types declared as parameters.
   It assumes that the types declared as parameters are in param[1, pos-1]: it stops on the first non type parameter.
   A new parameter identifier is created in the top of the stack for each element of the operand list.
   @return the position in param of the no type parameter element.
   Local to sigTreeInterface2DCG().
*/
int
kernel2graph::sigTreeTypeParam2DCG (ptlrepid param)
{
  ptlrepid vrepid;
  TypeType TT;
  int pos;
  //  ntrees savetree; savetree = currentTree;
  pos = 1;
  while (pk_true)
    {
      if (!ptlrepid_getRepidAt (param, pos, &vrepid)) return pos;
      if (!isTypeId (vrepid)) return pos;
      TT = makeMetaType (vrepid);
      pos++;
    }
}

/**
   @return a generic node type for a list of parameters (param), starting from the pos-th element of param followed by TNode. 
   The result will be of the form (GenericTypeNode(Type(param(1)),..., (GenericTypeNode(Type(param(pos)), TNode))
   Local to sigTreeInterface2DCG().
*/
TypeType
kernel2graph::buildGenericNodes (ptlrepid param, int pos, TypeType TNode)
{
  int nb;
  boolean notEmpty, last;
  ptexpcal vexpcal;
  ptlrepid vrepid;
  TypeType TT1, TT;

  notEmpty = pk_true;
  nb = pos;
  while (notEmpty)
    {
      notEmpty = ptlrepid_getRepidAt (param, nb, &vrepid);
      if (notEmpty)
	nb = nb + 1;
    }
  if (nb > pos)
    nb = nb - 1;
  if (nb == pos)
    return TNode;
  last = pk_true;
  while (nb != pos)
    {
      notEmpty = ptlrepid_getRepidAt (param, nb, &vrepid);
      if (notEmpty)
	{
	  if (!isTypeId (vrepid))
	    {
	      getCurrentInstance (param, nb, tparam, &vexpcal);
	      TT1 = EM->getType (vexpcal);
	      if (last)
		{
		  TT = TNode;
		  last = pk_false;
		}
	      TT = GenericNodeTypeCreate (TT1, TT);
	    }
	  nb = nb - 1;
	}
    }
  return TT;
}

/**
   @return a meta node type associated with a list of types (param) and a oprand list (TN).
   It assumes that the types declared as parameters are in param[1, pos-1]: it stops on the first non type parameter.
   Each type is added (in pre position) to the operand list TN.
*/
TypeType
kernel2graph::buildMetaNode (ptlrepid param, TypeType TN)
{
  boolean bcontinue, first;
  ptlrepid vrepid;
  TypeList TLMeta;
  TypeType TT;
  int pos;
  //  ntrees savetree; savetree = currentTree;
  pos = 1;
  bcontinue = pk_true;
  first = pk_true;
  while (bcontinue)
    {
      bcontinue = ptlrepid_getRepidAt (param, pos, &vrepid);
      if (bcontinue)
	if (isTypeId (vrepid))
	  {
	    if (first)
	      {
		TLMeta = TypeListCreate ();
		first = pk_false;
	      }
	    TT = getCurrentType (vrepid);
	    TypeListAppendType (TLMeta, TT);
	  }
      pos = pos + 1;
    }
  if (first)
    return TN;
  TLMeta = TypeListRegister (TLMeta);
  return MetaNodeTypeCreate (TupleTypeCreate (TLMeta), TN);
}

/** 
    Compiling an interface composed of parameters (param), inputs (entrees) and outputs (sorties).
    \note : A Meta node type is built but not used currently.
*/
void
kernel2graph::sigTreeInterface2DCG (ptlrepid param, ptlrepid entrees, ptlrepid sorties)
{
  int numero;
  ntrees savetree;
  TypeType TSInputs, TSOutputs, TNode, DCGType, TT;
  ClkSystClass Clock;
  GraphClass Graph;
  BackAnnotationType save_back_annot_obj;

  savetree = currentTree;
  save_back_annot_obj = back_annot_obj;
  numero = sigTreeTypeParam2DCG (param);
  sigTreeGenericParam2DCG (param, numero);
  sigTreeFlows2DCG (entrees, pk_false);
  TSInputs = mkStrucuredTypeFromFlows (entrees, pk_true);
  sigTreeFlows2DCG (sorties, pk_false);
  TSOutputs = mkStrucuredTypeFromFlows (sorties, pk_false);
  Clock = NILCLKSYSTCLASS;
  Graph = (GraphClass) 0;
  TNode = NodeTypeCreate (TSInputs, TSOutputs, Clock, Graph);
  TT = buildGenericNodes (param, numero, TNode);
  DCGType = buildMetaNode (param, TT);
  currentTree = savetree;
  back_annot_obj = save_back_annot_obj;
}


/** 
    @return "true" if the association of effective parameters (vklExprSig) with formal parameters (vlrepid) succeeds, "false" for 
    a call without parameters.
    Assume: verification on the number of effectice/formal parameters have been done during Trees2kernel translation.
*/
boolean
kernel2graph::instdclparameters (ptlrepid vlrepid, ptklExprSig vklExprSig)
{
  ptklExprSig aParam;
  int n, i;
  boolean isfree, b;
  ptlinstances instent;
  ptexpcal vexpcal;
  ptinsth vh;
  ntrees savetree;
  noperators ope, op;
  ptlrepid varid;
  ptrepfi vfi, vfiformal;
  BackAnnotationType save_back_annot_obj;

  if ((vklExprSig == NILEXPRSIG) && (vlrepid != NILLREPID)) return pk_false;
  if (vklExprSig != NILEXPRSIG)
    {
      b = igc->getSetConstantContext (pk_true);
      savetree = currentTree;
      save_back_annot_obj = back_annot_obj;
      op = kl_getFlowExprOperator (vklExprSig);
      n = ptlrepid_size (vlrepid);
      for (i = 1; i <= n; i++)
	{
	  aParam = kl_getNextExpr (&vklExprSig);
	  if (aParam != NILEXPRSIG)
	    {
	      if (isTypeId (vlrepid))
		{
		  isfree = pk_true;
		  instent = NILLINSTANCES;
		  vh = instOperand (hundef, UndefinedType, &isfree, aParam, &vexpcal, &instent, op);
		  setCurrentType (vlrepid, EM->getExpType (vexpcal));
		}
	      else if (isModelId (vlrepid))	// formal process
		{
		  varid = kl_getIdentifierExpr (aParam);
		  vfi = getModelRepid (varid);
		  vfiformal = getModelRepid (vlrepid);
		  kl_setEffectiveModel (vfiformal, vfi);
		}
	      else
		{
		  ope = kl_getFlowExprOperator (aParam);
		  if ((getTypeExpflot (aParam) == constant) || (ope == opvenum) || (ope == opnom) || (ope == opident))
		    {
		      isfree = pk_true;
		      instent = NILLINSTANCES;
		      vh = instOperand (hundef, UndefinedType, &isfree, aParam, &vexpcal, &instent, op);
		    }
		  else
		    {
		      mkInstance (vlrepid, tconstant, pk_false, &vexpcal);
		      vh = instEquation (hinit, vexpcal, aParam, opdefsig);
		    }
		  setCurrentInstance (vlrepid, vexpcal);

		  /*  ope = kl_getFlowExprOperator(aParam);
		      switch ((int) ope)
		      {
		      case opvenum:
		      case opnom:
		      case opident:
		      isfree = pk_true; 
		      instent = NILLINSTANCES;
		      vh = instOperand(hundef, UndefinedType, &isfree, aParam, &vexpcal, &instent, op);
		      setCurrentInstance(vlrepid, vexpcal); 
		      break;
		      default:
		      mkInstance(vlrepid, tconstant, pk_false, &vexpcal);
		      vh = instEquation(hinit, vexpcal, aParam, opdefsig);
		      setCurrentInstance(vlrepid, vexpcal); 
		      } */
		}
	      vlrepid = succid (vlrepid);
	    }
	}
      currentTree = savetree;
      back_annot_obj = save_back_annot_obj;
      b = igc->getSetConstantContext (b);
    }
  return pk_true;
}

/**
   @return the combed expression (list of expression) associated with the list of declaration of identifiers (vlrepid) added to the previous 
   combed expression (combedExp). Back information to tree syntactic tree is given by savetree.
   Each identifier x characterized by two stacks (one for input, one for output). 
   When inStack is "true" the inserted signal is the top of the input stack, the top of the output stack otherwise.
*/
ptexpcal
kernel2graph::ctexternemacro (ptlrepid vlrepid, boolean inStack, ntrees savetree, ptexpcal combedExp)
{
  ptexpcal vexp1, vexpcal;

  if (!getCurrentInstance (vlrepid, 1, tsignal, &vexpcal))
    return combedExp;
  vexp1 = ctexternemacro (succid (vlrepid), inStack, savetree, combedExp);
  return formalEffectiveTypeChecking (vlrepid, inStack, vexpcal, vexp1, savetree);
}


/**
   Local to instUnexpandedModel() method.
*/
ptexpcal
kernel2graph::mkAddedOutputsUnexpanded (ptlrepid l, ptinsth vh)
{
  ptexpcal combedExp, s, s1;
  int i, n;
  ptlrepid vrepid;
  boolean b;

  if (l == NILLREPID)
    return NILEXPCAL;
  combedExp = NILEXPCAL;
  n = ptlrepid_size (l);
  for (i = n; i > 0; i--)
    {
      ptlrepid_getRepidAt (l, i, &vrepid);
      s = getCurrentInstance (vrepid);
      b = EM->isStateVar (s);
      if (b)
	{
	  s1 = s;
	  mkAndPushInstance (l, i, tsignal, pk_true, &s);
	  EM->addMark (s, XMARK);	// must be popped.
	  igc->setClockVariable (s);	// free variable !!!

	  // s1 ::= s when vh  and s will be used for the call to the process.
	  vh = igc->createPartialEquation (s1, s, vh, vh, ptlinstances_push_front (NILLINSTANCES, s), back_annot_obj);
	}
      combedExp = EM->mkNaryExp (oplesign, s, combedExp, MK_BACK_ANNOT(NILTREE));
    }
  return combedExp;
}

/**
   Local to instUnexpandedModel() method.
*/
void
kernel2graph::popAddedOutputsUnexpanded (ptlrepid l)
{
  ptexpcal s;
  int i, n;
  ptlrepid vrepid;

  n = ptlrepid_size (l);
  for (i = n; i > 0; i--)
    {
      ptlrepid_getRepidAt (l, i, &vrepid);
      s = getCurrentInstance (vrepid);
      if (EM->isMarked (s, XMARK))
	{
	  kl_popInstance (nameId (vrepid));
	  EM->removeMark (s, XMARK);
	}
    }
}

/**
   @return {B(x), the boolean clock of x (x element of l) synchronized with vh}.
 */
ptexpcal kernel2graph::genBooleanClocksSignalThread( ptlrepid l, ptinsth vh)
{
  ptexpcal vboolsig, vsig, ladded;
  ptident id; 
  ptlrepid vrepid;
  int i, n;

  ladded = NILEXPCAL;
  n = ptlrepid_size (l);
  for (i = n; i > 0; i--)
    {
      ptlrepid_getRepidAt (l, i, &vrepid);
      vsig = getCurrentInstance (vrepid);
      id = EM->getName (vsig);
      id = makeIdent_ident_string(id, "bclk");
      vboolsig = EM->mkIdentExp (id, tsignal, get_BooleanType (), NM->mk_node_undefined (), hundef,  MK_BACK_ANNOT(NILTREE));
      EM->setClock(vboolsig, vh);
      ladded = EM->mkNaryExp(oplesign, vboolsig, ladded, back_annot_obj);
    }
  return ladded;
}

/**
   Compiling a model as an unexpanded one.
   @param vfi (in) the model of process.
   @param parameters (in) the effective parameters (nilexpcal for compiling without parameters).
   @param vop (in) type of model (function, node, action, process,...).
   @param g (out) the built graph associated with the model.
   @param vcall (out) a node that references the built graph (ie nodecall(g))
   @return the clock of the vcall node.
*/
ptinsth
kernel2graph::instUnexpandedModel (ptrepfi vfi, ptexpcal parameters, noperators vop, ptDCG * g, ptINode * vcall)
{
  ptexpcal inputs, outputs;
  ptDCG vgraph;
  ptlinstances lent, lsort;
  ptINode vnodeCall;
  ptinsth vh, h1;
  ptexpcal vsig;
  ptpragma vpragma;
  ptlrepid entree, sortie, l;
  ptlpragmas lprag;

  vgraph =buildGraphOf(vfi, false);
  sortie = kl_getInsOutsModel (vfi, &entree);

  h1 = CKM->new_variFormulaRootClock ();

  setAccessMode (pk_true);
  l = kl_getExternalReferencesIn (vfi);
  lent = NILLINSTANCES;
  inputs = unexpdandedModelExternalReferencesIn (l, NILEXPCAL, h1, &lent);
  lent = getCurrentInstances (entree, lent, pk_true);
  inputs = ctexternemacro (entree, pk_true, currentTree, inputs);

  setAccessMode (pk_false);
  l = kl_getExternalReferencesOut (vfi);
  outputs = mkOrderedListOfInstances (l, NILEXPCAL);
  outputs = ctexternemacro (sortie, pk_false, currentTree, outputs);

  lsort = EM->getSignalsReferencedIn (outputs);

  if ((vop == opact) || (vop == opactp))	// Action
    vh = igc->getLabelClock (currentTree);
  else
    vh = CKM->makeSup (igc->upperBoundClockOfIdentifiers (lent), igc->upperBoundClockOfIdentifiers (lsort));

  vh = CKM->makeSynchro (vh, h1);
  if (l != NILLREPID)
    {
      outputs = mkAddedOutputsUnexpanded (l, vh);
      outputs = ctexternemacro (sortie, pk_false, currentTree, outputs);
    }

  /* Building the node call to the graph. */
  ptexpcal vouts = igc->mkFieldAccess (outputs,  back_annot_obj);
  vnodeCall = NM->mk_node_call (vgraph, parameters, inputs, vouts, back_annot_obj);
  if (vgraph->isGraphOfExternalProcess ())
    NM->set_safetyProperty (vnodeCall, kl_getSafetyProperty (vfi));
  NM->insertNode (vnodeCall, vh);
  lprag = kl_getPragmaModel (vfi);
  if (getPragmaKeyword (lprag, PRAGMA_DELAYCLUSTER, &vpragma))
    NM->addPragma (vnodeCall, vpragma);
  if (getPragmaKeyword (lprag, PRAGMA_SIGALI, &vpragma))
    NM->addPragma (vnodeCall, vpragma);

  /* Taking account the context/pragmas. */
  igc->takeIntoAccountContexts (vnodeCall);
  while (ptlinstances_get_next (&lsort, &vsig))
    igc->takeIntoAccountContexts (EM->getNode (vsig));

  if (l != NILLREPID) popAddedOutputsUnexpanded (l);
  GRAPHS_MANAGER->incrNbReferences (vgraph);
  *g = vgraph;
  *vcall = vnodeCall;
  return vh;
}

/**
   @return the graph of a model (vfi).
 */
ptDCG kernel2graph::buildGraphOf(ptrepfi vfi, boolean bglobalNameManager)
{
  ptDCG vgraph;
  ptkernel2graph translator;
  ptklExprSig param;
  boolean b;

  // already compiled ?
  if (GRAPHS_MANAGER->getDCGOfTree ((pointer) vfi, &vgraph)) return vgraph;
  // otherwise it is built.
  param = NILEXPRSIG;
  if (bglobalNameManager)
    translator = new kernel2graph (GRAPHS_MANAGER, nameManager,  NILSTACK, btm);
  else
    translator = new kernel2graph (GRAPHS_MANAGER, NILSTACK, btm);
  b = translator->kernelTree2DCG (vfi, param, pk_false);
  vgraph = translator->getDCG ();
  GRAPHS_MANAGER->setLinkGraph (vgraph, (pointer) vfi);
  delete translator;
  return vgraph;
}


ptexpcal kernel2graph::genNewOutputSignalThread(ptlrepid l, ptinsth vhDispatch)
{
  int n, i;
  ptexpcal vouts, vsig, vcopy;
  ptlrepid vrepid;
  
  vouts = NILEXPCAL;
  n = ptlrepid_size (l);  
  for (i = n; i > 0; i--)
    {
      ptlrepid_getRepidAt (l, i, &vrepid);
      vsig = kl_topInstance (nameId (vrepid));
      vcopy = EM->mkIdentCopy (vsig, NM->mk_node_undefined (), hundef);
      // EM->setClock (vcopy, vhDispatch);
      vouts = EM->mkNaryExp (oplesign, vcopy, vouts, back_annot_obj);
    }
  return vouts;
}

/** It generates l1[i] :=  l2[i] when lconds[i] */
void kernel2graph::genAssignOutputsSignalThreads(ptexpcal l1, ptexpcal l2, ptexpcal lconds)
{
  ptexpcal vsig, vsig1, vcond;
  ptinsth vhsub;

  while (EM->getNaryArgs (&l1, &vsig))
    {
      EM->getNaryArgs (&l2, &vsig1);
      EM->getNaryArgs (&lconds, &vcond);
      vhsub = CKM->makeSub (EM->getClock(vcond), vcond, ptlinstances_insert(NILLINSTANCES, vcond));
      igc->createEquation (vsig, vsig1, vhsub, vhsub, ptlinstances_insert(NILLINSTANCES, vsig1), back_annot_obj);
    }
}

/**
   Compiling a model as a "SignalThread" one.
   @param vfi (in) the model of process.
   @param parameters (in) the effective parameters (nilexpcal for compiling without parameters).
   @param vop (in) type of model (function, node, action, process,...).
   @param g (out) the built graph associated with the model.
   @param vcall (out) a node that references the built graph (ie nodecall(g))
   @return the clock of the vcall node.
*/
ptinsth
kernel2graph::instSignalThreadModel (ptrepfi vfi, ptexpcal parameters, noperators vop, ptDCG * g, ptINode * vcall)
{
  ptexpcal inputs, outputs, vDispatch, vsig;
  ptlinstances lent, lsort;
  ptINode vnodeCall;
  ptinsth vhDispatch;
  ptlrepid entree, sortie;
  ptpragma vpragma_separate;
  ptDCG vgraph;

  vgraph = buildGraphOf(vfi, true);
  // PRAGMA_SEPARATE is added to the graph. replaced by UNEXPANDED
  // vpragma_separate = newPredefinedPragma (PRAGMA_SEPARATE, NILOBJPRAGMA, NILTREE);
  vpragma_separate = newPredefinedPragma (PRAGMA_UNEXPANDED, NILOBJPRAGMA, NILTREE);

  vgraph->addPragma (vpragma_separate); 
  sortie = kl_getInsOutsModel (vfi, &entree);

  setAccessMode (pk_true);
  lent = getCurrentInstances (entree, NILLINSTANCES, pk_true);
  inputs = ctexternemacro (entree, pk_true, currentTree, NILEXPCAL);

  setAccessMode (pk_false);
  outputs = ctexternemacro (sortie, pk_false, currentTree, NILEXPCAL);
  lsort = EM->getSignalsReferencedIn (outputs);

  // Convention: The first input is the Dispatch signal (assumed) for a Signal Thread.
  vDispatch = EM->getBinaryArgstAt(inputs, 1);
  vhDispatch = igc->mkClockInstance(vDispatch);

  /* Building the node call to the graph. */
  ptexpcal vouts = igc->mkFieldAccess (outputs, back_annot_obj);
  vnodeCall = NM->mk_node_call (vgraph, parameters, inputs, vouts, back_annot_obj);
  NM->insertNode (vnodeCall, vhDispatch);
  NM->addMark (vnodeCall, DontOptimize);
  NM->set_safetyProperty (vnodeCall, opunsafe);

  /* Taking account the context/pragmas. */
  igc->takeIntoAccountContexts (vnodeCall);
  while (ptlinstances_get_next (&lsort, &vsig))
    igc->takeIntoAccountContexts (EM->getNode (vsig));

  GRAPHS_MANAGER->incrNbReferences (vgraph);
  *g = vgraph;
  *vcall = vnodeCall;
  return vhDispatch;
}


/**
   Compiling a model by macro expansion. All objects are added to the current graph (AGRAPH attribute of the class).
   Called for (lo) := reference to the model (vfi) {lp} (li)
   @param vfi (in) the model of process.
   @param lp (in) the effective parameters 
   @param li (in) the effective inputs
   @param lo (in) the effective outputs
   @return the clock of the added part in the graph.
*/
ptinsth
kernel2graph::instexpandedModel (ptrepfi vfi, ptexpcal lp, ptexpcal li, ptexpcal lo)
{
  ptlrepid sortie, entree, param;
  ptklExprProc corps, spec;
  ptinsth vhCorps, vhSpec;
  ntrees savetree;

  sortie = kl_getInsOutsModel (vfi, &entree);
  param = kl_getParametersModel (vfi);
  setCurrentInstances (param, lp);
  setCurrentInstances (sortie, lo);
  setCurrentInstances (entree, li);
  savetree = kl_getBodyModel (vfi, &corps);

  pushModel (vfi);
  vhCorps = compilingEproc (corps);
  kl_getSpecModel (vfi, &spec);
  vhSpec = compilingEproc (spec);
  popModel (vfi);
  deplinst (entree);
  deplinst (sortie);
  deplinst (param);

  return CKM->makeSup (vhCorps, vhSpec);
}
