/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define NILCLKSYSTCLASS (ClkSystClass)0
/**
   @return 
   - "true" if the current type associated with a declaration (vrepid) exists, TT is this type.
   - "false" otherwise, TT is unchanged.
 */
  boolean kernel2graph::getCurrentType (ptlrepid vrepid, TypeType * TT)
{
  return kl_topTypeInstance (nameId (vrepid), TT);
}

/** @return the current type associated with a declaration  (vrepid).
    If such a type does not exists, an "anyType" type is created and returned.
*/
TypeType kernel2graph::getCurrentType (ptlrepid vrepid)
{
  TypeType
    TT;
  if (!getCurrentType (vrepid, &TT))
    TT = AnyTypeCreate ();
  return TT;
}

/**
   @return 
   pk_true if the type (nom_typ) is a virtual one, then TT is the current value.
   false otherwise, TT is unchanged.
*/
boolean kernel2graph::getTypeVirtualType (ptlrepid nom_typ, TypeType * TT)
{
  if (nom_typ == NILLREPID)
    return pk_false;		// extyernal type
  // virtual : textern and (nom_typ != NILLREPID) 
  *TT = getCurrentType (nom_typ);
  return pk_true;
}

/**
   @return the type (internal representation of the types) associated with an operator of a predefined type (ope).
   For an unknown predefined type, it returns a "any" type.
*/
TypeType
kernel2graph::sigTreePredefType2TypeType (noperators ope)
{
  switch ((int) ope)
    {
    case opchar:
      return get_CharType ();
    case opcomplex:
      return get_ComplexType ();
    case opstring:
      return get_StringType ();
    case opinteger:
      return get_IntegerType ();
    case oplong:
      return get_LongType ();
    case opshort:
      return get_ShortType ();
    case opreal:
      return get_RealType ();
    case opdpreal:
      return get_DRealType ();
    case opcplx:
      return get_ComplexType ();
    case opdcomplex:
      return get_DComplexType ();
    case opevent:
      return get_PureType ();
    case oplogical:
      return get_BooleanType ();
    case opimplicit:
      return UndefinedType;
    default:
      return AnyTypeCreate ();
    }
}

/**
   @return a list of name composed of the element of a list of enumerated values (lvalenum).
 */
NameList
kernel2graph::sigTreeEnumType2NameList (ptlEnum lvalenum)
{
  NameList NL;
  NameClass Nom;
  pointer elem;

  NL = NameListCreate ();

  while (ptlPointer_get_elemNext (&lvalenum, &elem))
    {
      Nom = (NameClass) elem;
      NameListAppendName (NL, Nom);
    }
  NL = NameListRegister (NL);
  return NL;
}

/**
   @return the type associated with a list of the fields of a structure (champs).
 */
TypeType
kernel2graph::sigTreeStructType2TypeType (ptrepfield champs)
{
  ptrepstr nom;
  ptreptyp vtyp;
  NameList NL;
  TypeList TL;
  TypeType TS, TT;

  TL = TypeListCreate ();	/* types des champs */
  NL = NameListCreate ();	/* noms des champs */
  while (kl_get_elemNext_Field (&champs, &nom, &vtyp))
    {
      NameListAppendName (NL, (NameClass) nom);
      TT = sigTreeType2TypeType (vtyp);
      if (isBundleType (TT))
	pK_messenger->errorTree (errBundleType2, kl_getTypeTreeDecl (vtyp));
      TypeListAppendType (TL, TT);
    }
  NL = NameListRegister (NL);
  TL = TypeListRegister (TL);
  TS = StructureTypeCreate (NL, TupleTypeCreate (TL));
  return TS;
}

/**
   @return true if a type (aType) is a bundle one, false otherwise.
*/
boolean kernel2graph::isBundleType (TypeType aType)
{
  return TypeTestBundleClass (TypeTypeGetUnamedType (aType));
}

/**
   @return the array type [vexpcalDim] aType if vexpcalDim is defined, aType otherwise.
*/
TypeType
kernel2graph::mkArraytype (ptexpcal vexpcalDim, TypeType aType)
{
  if (vexpcalDim == NILEXPCAL)
    return aType;
  if (aType == UndefinedType)
    aType = AnyTypeCreate ();
  return EM->mkArrayType (vexpcalDim, aType);
}

/**
   @return the bundle type (tbundle) with t the definition of the fields, and without clock relations 
   The link between the with the model of the declation is kept in a global table (bundleTable) used for the instanciation of the clock  relations
   when a signal is created.
*/
TypeType
kernel2graph::mkBundleType (ptreptyp vtyp, TypeType t)
{
  t = BundleTypeCreate (t, NILCLKSYSTCLASS);
  BundleTable[t] = vtyp;
  return t;
}


/** 
    @return a type representation (in graph) from a model (vtyp) of type(kernel-tree representation).
*/
TypeType
kernel2graph::sigTreeType2TypeType (ptreptyp vtyp)
{
  ptklExprSig trep;
  noperators ope;
  typ_type t;
  TypeType t1;
  ptrepfield champs;
  ptlrepid nom_typ;
  ptexpcal vexpcalDim;
  ptlEnum lvalenum;

  kl_getFieldType (vtyp, &t, &ope, &trep, &nom_typ, &champs, &lvalenum);
  vexpcalDim = mkDimExpcal (trep);
  switch ((int) t)
    {
    case tbase:
      t1 = sigTreePredefType2TypeType (ope);
      break;
      /* case textern: return getCurrentType(nom_typ); it does not occur ! */
    case tenum:
      NameClass name;
      NameList NL;
      NL = sigTreeEnumType2NameList (lvalenum);
      if (!EnumerationExist (NL, &t1))
	{
	  name = ident2repstr (nameManager->mkNameType (getTypeId ()));
	  t1 = EnumerationTypeCreate (name, NL);
	}
      break;

    case tnom:
      t1 = getCurrentType (nom_typ);
      if ((vexpcalDim != NILEXPCAL) && isBundleType (t1))
	{
	  pK_messenger->errorTree (errBundleType, kl_getTypeTreeDecl (vtyp));
	  vexpcalDim = NILEXPCAL;
	}
      break;

    case tstruct:
      t1 = sigTreeStructType2TypeType (champs);
      break;

    case tbund:
      t1 = sigTreeStructType2TypeType (champs);
      if (vexpcalDim != NILEXPCAL)
	pK_messenger->errorTree (errBundleType, kl_getTypeTreeDecl (vtyp));
      return mkBundleType (vtyp, t1);

    default:
      printf ("DEBUG kernel2graph::sigTreeType2TypeType not yet 1\n");
      fflush (stdout);
      pK_messenger->errorTree (errnotimplemented, currentTree);
      return AnyTypeCreate ();
    }
  return mkArraytype (vexpcalDim, t1);
}


/**
   @return the representation of the type of the identifier designed by vrepid  defined by the kernel tree expression (vtyp).
 */
TypeType
kernel2graph::mkInstType (ptlrepid vrepid, ptreptyp vtyp)
{
  ptklExprSig trep;
  noperators ope;
  typ_type t;
  TypeType t1;
  ptrepfield champs;
  ptlrepid nom_typ;
  ptexpcal vexpcalDim;
  ptlEnum lvalenum;
  NameClass name, NClass;
  ptident id;

  kl_getFieldType (vtyp, &t, &ope, &trep, &nom_typ, &champs, &lvalenum);
  vexpcalDim = mkDimExpcal (trep);

  if (vrepid == NILLREPID)
    id = getTypeId ();
  else
    id = identpile (nameId (vrepid));
  // modif LBesnard Jul 2010 : id = nameManager -> mkNameType(id); replaced by
  if (t != textern)
    id = nameManager->mkNameType (id);
  NClass = ident2repstr (id);

  switch ((int) t)
    {
    case tbase:
      t1 = sigTreePredefType2TypeType (ope);
      t1 = mkArraytype (vexpcalDim, t1);
      return NamedTypeCreate (NClass, t1);
      break;

    case tenum:
      NameList NL;
      NL = sigTreeEnumType2NameList (lvalenum);
      if (!EnumerationExist (NL, &t1))
	{
	  if (vexpcalDim == NILEXPCAL)
	    name = NClass;
	  else
	    name = ident2repstr (nameManager->mkNameType (getTypeId ()));
	  t1 = EnumerationTypeCreate (name, NL);
	}
      if (vexpcalDim == NILEXPCAL)
	return t1;
      t1 = mkArraytype (vexpcalDim, t1);
      return NamedTypeCreate (NClass, t1);
      break;

    case textern:
      // The name of an external type does not change.
      // virtual : textern and (nom_typ != NILLREPID) 
      if (!getTypeVirtualType (nom_typ, &t1))
	t1 = ExternTypeCreate (NClass);
      return mkArraytype (vexpcalDim, t1);
      break;

    case tnom:
      t1 = getCurrentType (nom_typ);
      if ((vexpcalDim != NILEXPCAL) && isBundleType (t1))
	{
	  pK_messenger->errorTree (errBundleType, kl_getTypeTreeDecl (vtyp));
	  vexpcalDim = NILEXPCAL;
	}
      return mkArraytype (vexpcalDim, t1);
      break;

    case tstruct:
      t1 = sigTreeStructType2TypeType (champs);
      t1 = mkArraytype (vexpcalDim, t1);
      return NamedTypeCreate (NClass, t1);
      break;

    case tbund:
      t1 = sigTreeStructType2TypeType (champs);
      if (vexpcalDim != NILEXPCAL)
	pK_messenger->errorTree (errBundleType, kl_getTypeTreeDecl (vtyp));
      t1 = mkBundleType (vtyp, t1);
      return NamedTypeCreate (NClass, t1);
      break;

    default:
      printf ("DEBUG kernel2graph::mkInstType not yet 2\n");
      fflush (stdout);
      pK_messenger->errorTree (errnotimplemented, currentTree);
      return AnyTypeCreate ();
    }
}


/**
   @return a Meta type for the identifier referenced by its declaration (vrepid).
 */
TypeType
kernel2graph::makeMetaType (ptlrepid vrepid)
{
  TypeType TT;

  TT = MetaTypeCreate (ident2repstr (identpile (nameId (vrepid))));
  setCurrentType (vrepid, TT);
  EM->mkExpType (TT);
  return TT;
}

/**  @return the Structured Type associated with a list of declarations (list).
     For each element of list, it takes the current instance in the input-stack if inStack is "true", in the output-stack otherwise.
*/
TypeType
kernel2graph::mkStrucuredTypeFromFlows (ptlrepid list, boolean inStack)
{
  int numero;
  boolean notEmpty;
  ptexpcal aFlow;
  TypeType TT;
  TypeList TL;
  NameList NL;

  TL = TypeListCreate ();	/* types des champs */
  NL = NameListCreate ();	/* noms des champs */
  numero = 1;
  notEmpty = pk_true;
  while (notEmpty)
    {
      notEmpty = getCurrentInstance (list, numero, tsignal, &aFlow);
      if (notEmpty)
	{
	  NameListAppendName (NL, ident2repstr (EM->getName (aFlow)));
	  TT = EM->getType (aFlow);
	  TypeListAppendType (TL, TT);
	  numero = numero + 1;
	}
    }
  NL = NameListRegister (NL);
  TL = TypeListRegister (TL);
  return StructureTypeCreate (NL, TupleTypeCreate (TL));
}
