/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/



#define TRACE(s)		//s

// #define messDeprecatedSeparate  ">>>>> Separate is deprecated, use Unexpanded"

#define MK_BACK_ANNOT(T) new BackAnnotation(T, TM->getCurrentCallTree())
#define NEW_BACK_ANNOT(T) currentTree = T; back_annot_obj = new BackAnnotation(T, TM->getCurrentCallTree())

_PRIVATE_ ptStack  kernel2graph::callsStack = NILSTACK;	///< Class variable.
BundleTableType kernel2graph::BundleTable;	///< Class variable.

_PRIVATE_

void
kernel2graph::printTraceabilityInfos()
{
  if (btm)
    {
      ptident id = AGRAPH->getName();
      TM->print(id);
      EM->printTracebilityInfos(id);
      NM->printTracebilityInfos(id);
    }
}
   
int
kernel2graph::traceability (ptexpcal vexpcal, int pos)
{
  ptexpcal e;

  if (btm)
    if (vexpcal != NILEXPCAL)
      {
	if (EM->satisfiesList (vexpcal))
	  while (EM->getNaryArgs (&vexpcal, &e))
	    {
	      TM->add ((ptObjectCommonInfos) e, pos);
	      pos++;
	    }
	else
	  {
	    TM->add ((ptObjectCommonInfos) vexpcal, pos);
	    pos++;
	  }
      }
  return pos;
}

void
kernel2graph::traceability (ptexpcal effparam, ptexpcal lentren, ptexpcal lsortren)
{
  int n;
  n = traceability (effparam, 1);
  n = traceability (lentren, n);
  n = traceability (lsortren, n);
}


/**
   @return vfi when it is not a conversion model,  the model associated with the conversion model otherwise.
   @param bTypeConversion (out) set to the vlaue "true" when vfi is a conversion model.
*/
ptrepfi
kernel2graph::getModelOf (ptrepfi vfi, boolean * bTypeConversion)
{
  ptlrepid vrepidNameType;
  ptrepfi nvfi;
  ptident nomfi;
  noperators vop;
  TypeType TT;

  if (kl_getEffectiveModel (vfi, &nvfi) && (nvfi != NILREPFI))
    return nvfi;
  *bTypeConversion = kl_isTypeConversionModel (vfi, &vrepidNameType);
  if (*bTypeConversion)
    {
      // conversion
      TT = getCurrentType (vrepidNameType);
      TT = TypeTypeGetUnamedType (TT);
      if (TypeTestPredef (TT))
	{
	  vop = predefTypeType2SignalType (TT);
	  nomfi = repstr2ident (predefinedType2ident (vop));
	  if (klGetModelOf (kl_getRefToIntrinsics (vfi), nomfi, &nvfi))
	    {
	      vfi = nvfi;
	      *bTypeConversion = pk_false;
	    }
	}
    }
  return vfi;
}

/**
   @return the model with varid as name and set bTypeConversion to the vlaue "true"
   when it is a conversion.
*/
ptrepfi
kernel2graph::getModelOf (ptlrepid varid, boolean * bTypeConversion)
{
  ptrepfi vfi;
  vfi = getModelRepid (varid);
  return getModelOf (vfi, bTypeConversion);
}

/**
   @return true if vexpcal is a list-expression. 
*/
boolean
kernel2graph::isTuple (ptexpcal vexpcal)
{
  return EM->satisfiesList (vexpcal);
}

/**
   Access Mode management to signals: v must be "true" for input access, false otherwise. 
*/
boolean
kernel2graph::setAccessMode (boolean v)
{
  boolean x;
  x = ACCESS_MODE_INPUT;
  ACCESS_MODE_INPUT = v;
  return x;
}

#include "pKio.h"

/** Pops the element assigned to the list of identifiers vlrepid.*/
void
kernel2graph::deplinst (ptlrepid vlrepid)
{
  int n, i;

  n = ptlrepid_size (vlrepid);
  for (i = 1; i <= n; i++)
    {
      if (isTypeId (vlrepid))
	kl_popTypeInstance (nameId (vlrepid));
      else if (isFlowId (vlrepid))
	kl_popInstance (nameId (vlrepid));
      else			// Formal model (parameter)
	kl_setEffectiveModel (getModelRepid (vlrepid), NILREPFI);

      vlrepid = succid (vlrepid);
    }

}

/**
   Pops the locals, constants and types, labels resp. referenced by the list dcllocalsig, dclconstantesTypes and llabels.
   The poped instance was previously pushed by compiling_locals() method.
*/
void
kernel2graph::compiling_locals_finalize (ptlrepid dclconstantesTypes, ptlrepid dcllocalsig, ptlrepid llabels)
{
  deplinst (dcllocalsig);
  deplinst (dclconstantesTypes);
  deplinst (llabels);
}


/**
   Compiling (stack management) local declarations.
   @param ConstantTypes (in) a list of constants and types,
   @param locSig (in) a list of local signals,
   @param llabels (in) a list of labels.
*/
void
kernel2graph::compiling_locals (ptlrepid ConstantTypes, ptlrepid locSig, ptlrepid llabels)
{
  sigTreeDeclConstantsTypes2DCG (ConstantTypes);
  sigTreeFlows2DCG (locSig, pk_false);
  instlabels (llabels);
}

/**
   It pushes the declaration of the constants, types, locals labels and the import modules of a model (vfi) 
*/
void
kernel2graph::pushModel (ptrepfi vfi)
{
  ptlrepid locales, dclConstantesTypes, labels;

  pushModules (kl_getModules (vfi));
  kl_getLocalDeclarations (vfi, &dclConstantesTypes, &locales, &labels);
  compiling_locals (dclConstantesTypes, locales, labels);
}



/**
   It pushes (calling pushModel()) the declaration of the constants, types, locals labels and the import modules of a model (vfi).
   and sets the context Identifier to the name of te model.
*/
ptident
kernel2graph::pushModelAndContextId (ptrepfi vfi, ptident * idmodel, ptexpcal *pStatevarAutomaton, ptinsth *pAutomatonClock, boolean *bAutomatonClockAlreadyFixed)
{
  ptident id;
  ptlrepid vstavar;
  ptexpcal  vCurStateVar;
  
  *pStatevarAutomaton = igc->getAutomatonCurrentStateVar ();
  *pAutomatonClock = igc->getClockAutomaton ();
  *bAutomatonClockAlreadyFixed  = igc->isClockedAutomaton();

  pushModel (vfi);

  if (LL_isAutomatonFamilyOp (noper (kl_getTreeReference (vfi))))
    {
      // the model is an automaton: setting the current state variable.
      if (kl_getCurrentStateVarAutomaton (vfi, &vstavar)) // a state var has been declared.
	vCurStateVar = getCurrentInstance (vstavar);
      else
	{ // a new state var is created 
	  vCurStateVar = EM->mkIdentExp (string2ident ("_CurState"), tsignal, AnyTypeCreate (), NM->mk_node_undefined (), hundef,  MK_BACK_ANNOT(NILTREE));
	  EM->setStateTag (vCurStateVar, pk_true);
	}
      igc->setAutomatonCurrentStateVar (vCurStateVar);
      ptinsth hauto = CKM->new_variFormulaRootClock ();
      igc->setClockAutomaton (hauto);
    }

  id = kl_getNameModel (vfi);
  *idmodel = nameManager->setCurrentModelName (id);
  return nameManager->setContextSubExpr (newIdent_concIdentIdent (id, getStrVar ()));
}

/**
   Push a list of modules (modules). For each of them, it pushes:
      - Local declarations (constants, types)
      - list of models
*/
void
kernel2graph::pushModules (ptlrepid modules)
{
  while (modules != NILLREPID)
    {
      pushModel (getModelRepid (modules));
      modules = succid (modules);
    }
}

/**
   It pops the declaration of the constants, types, locals labels and the import modules of a model (vfi) 
*/
void
kernel2graph::popModel (ptrepfi vfi)
{
  ptlrepid locales, dclConstantesTypes, labels;

  kl_getLocalDeclarations (vfi, &dclConstantesTypes, &locales, &labels);
  compiling_locals_finalize (dclConstantesTypes, locales, labels);
  popModules (kl_getModules (vfi));
}

/**
   It pops (calling popmodel()) the declaration of the constants, types, locals labels and the import modules of a model (vfi) and sets the 
   current contextual identifier to id.
*/
void
kernel2graph::popModelAndContextId (ptrepfi vfi, ptident id, ptident idmodel, ptexpcal vStateVarAutomaton, ptinsth vClockAutomaton, boolean bIsClockedAutomaton)
{
  popModel (vfi);
  nameManager->setContextSubExpr (id);
  nameManager->setCurrentModelName (idmodel);
  igc->setAutomatonCurrentStateVar (vStateVarAutomaton);
  igc->setClockAutomaton (vClockAutomaton);
  igc->setClockedAutomaton(bIsClockedAutomaton);
}

/**
   Pops the list of modules (modules) previously pushed with pushModules().
*/
void
kernel2graph::popModules (ptlrepid modules)
{
  if (modules == NILLREPID)
    return;
  popModules (succid (modules));
  popModel (getModelRepid (modules));
}

/**
   @return the model referenced in the expression (vexp) assumed to be a call.
*/
ptrepfi
kernel2graph::getModelFromExpflot (ptklExprSig vexp)
{
  ptlrepid varid;
  boolean b;
  varid = kl_getFlowExprModelProcCallname (vexp);
  return getModelOf (varid, &b);
}


/**
   It pushes for each identifier of lsorties a signal given by
   - either the list (lexpnvsort) when it is not empty, 
   - or a new signal is created.
   
   @return a list of expression defined by the new outputs followed by the operand list (lexpsortadd).
*/
ptexpcal
kernel2graph::pushAndMakeList (ptexpcal lexpnvsort, ptlrepid lsorties, ptexpcal lexpsortadd, ptlinstances * instent)
{
  ptexpcal vexpcal;

  if (lsorties == NILLREPID) return lexpsortadd;
  if (lexpnvsort == NILEXPCAL) 
    {
      mkAndPushInstance (lsorties, 1, tsignal, pk_false, &vexpcal);
    }
  else
    {
      vexpcal = EM->getNaryArgs (&lexpnvsort);
      setCurrentInstance (lsorties, vexpcal);
    }
  *instent = ptlinstances_insert (*instent, vexpcal);
  return EM->mkNaryExp (oplesign, vexpcal, pushAndMakeList (lexpnvsort, succid (lsorties), lexpsortadd, instent), back_annot_obj);
}

/**
  For a model (vfi), it pushes new signals for the outputs.
  If the list (lexpnvsort) is not empty, it gives the new signals, otherwise new signals are created.
  @return a list of expression defined by the new outputs followed by the operand list (lexpsortadd).
*/
ptexpcal
kernel2graph::pushModelOutputs (ptexpcal lexpnvsort, ptrepfi vfi, ptexpcal lexpsortadd, ptlinstances * instent)
{
  ptlrepid li, lo;

  lo = kl_getInsOutsModel (vfi, &li);
  return pushAndMakeList (lexpnvsort, lo, lexpsortadd, instent);
}

/**
   Pops the outputs of a model (vfi) previously pushed with pushModelOutputs().
*/
void
kernel2graph::popModelOutputs (ptrepfi vfi)
{
  ptlrepid lentrees, sorties;
  sorties = kl_getInsOutsModel (vfi, &lentrees);
  deplinst (sorties);
}

/** 
    Iterations of processes.
    @param vklExprSig list of indexes vklExprSig of an iteration of processes,
    @param vind an instance of vklExprSig,
    @param vlindex the index associated with the travel of the elements of vind.
    of For each index (x) of a list of indexes vklExprSig of an iterate of processes,
    it assigns to its current instance (c(x) element of vind) the access to its vlindex-value.

    For x, element of vklExprSig, x_k the associated element in vind, x_k[I] is pushed in the stack of x.
    I is the index (vlindex) 0 to size(x_k)-1.
    The references to x_k will be replaced by x_k[I].
 */
void
kernel2graph::push_Iterate_Indexes (ptklExprSig vklExprSig, ptexpcal vind, ptexpcal vlindex)
{
  ptlrepid indice;
  ptexpcal vx;

  if (EM->satisfiesList (vind))
    {
      // Tuple (I1,...,In)
      while (EM->getNaryArgs (&vind, &vx))
	{
	  vx = EM->mkBinaryExp (opelem, vx, vlindex, back_annot_obj);
	  indice = kl_getIdentifierExpr (kl_getNextExpr (&vklExprSig));
	  setCurrentInstance (indice, vx);
	}
    }
  else
    {
      vx = EM->mkBinaryExp (opelem, vind, vlindex, back_annot_obj);
      indice = kl_getIdentifierExpr (vklExprSig);
      setCurrentInstance (indice, vx);
    }
}

/**
   Iterations of processes.
   It pops the indexes associated with an iteration index (vklExprSig), previously pushed in push_Iterate_Indexes() method.
*/
void
kernel2graph::pop_Iterate_Indexes (ptklExprSig vklExprSig)
{
  ptlrepid indice;

  if (_isalist_ (kl_getFlowExprOperator (vklExprSig)))
    while (vklExprSig != NILEXPRSIG)	// Tuple (I1,...,In)
      {
	indice = kl_getIdentifierExpr (kl_getNextExpr (&vklExprSig));
	kl_popInstance (nameId (indice));
      }
  else
    {
      indice = kl_getIdentifierExpr (vklExprSig);
      kl_popInstance (nameId (indice));
    }
}

/** 
    @return the subclass of an kernel expression (e).
    Error message is send when such an expression is not accepted for constant expression.
*/
cloper
kernel2graph::getTypeExpflot (ptklExprSig e)
{
  cloper cl_e;

  cl_e = kl_getFlowExprClassType (e);
  if (igc->isInConstantContext ())
    switch ((int) cl_e)
      {
	//    case constant:  case variabl: --Must be tested after--
	//    case oper1: case oper2:  case oper3:  case opertyp:    break;    
	//      case call: Function allowed.
      case oper1tp:
      case oper2tp:
      case oper3tp:
	pK_messenger->errorTree (errNotConstantExpr, currentTree);
	break;
      }
  return cl_e;
}


/**
   It annotates (warning) a tree (arbre) with the warnum-th  warning message, this message is parametrized by the name of identifier (vrepid).
 */
void
kernel2graph::warning (int warnum, ntrees arbre, ptlrepid vrepid)
{
  pK_messenger->warningaccnom (warnum, arbre, getextidentrepid (vrepid));
}


/**
   @return vh if locfree is "false", a new free-variable of clock (without definition) otherwise and an 
   signals an error (context in with constant are not allowed)
*/
ptinsth
kernel2graph::TestSetFreeClock (boolean locfree, ptinsth vh)
{
  if (locfree)
    {
      vh = CKM->new_variFormulaRootClock ();
      pK_messenger->errorTree (errCannotDefineClock, currentTree);
    }
  return vh;
}

/** It associates a new instqnce (vsig) to a model (vrepid): the top of the stack assigned to
    the identifier becomes vsig.
*/
void
kernel2graph::setCurrentInstance (ptlrepid vrepid, ptexpcal vsig)
{
  kl_pushInstance (nameId (vrepid), vsig);
}

/** 
    @return the signal at a given depth (n) of the stack associated with a declaration (vrepid).
*/
ptexpcal kernel2graph::getStackTopMinus (ptlrepid vrepid, int n)
{
  return kl_topInstanceMinus (nameId (vrepid), n);
}


/** @return current signal (top) assoicated with a declaration (vrepid).
*/
ptexpcal kernel2graph::getCurrentInstance (ptlrepid vrepid)
{
  ptexpcal
    x;

  x = kl_topInstance (nameId (vrepid));
  if (x != NILEXPCAL)
    if (ACCESS_MODE_INPUT)
      if (EM->isMarked (x, OUTPUT_ADDED_FOR_UNEXPANDED))
	x = getStackTopMinus (vrepid, 1);
  return x;
}



/**
   @return the "init" clock if vh is the init clock, "undefined" clock otherwise.
*/
ptinsth
kernel2graph::getUndefOrInitClock (ptinsth vh)
{
  if (vh == hinit) return hinit;
  return hundef;
}

/**
  @return a context clock (used for declaration)
      - undefined clock for a state variable (bstate)
      - undefined clock when the current compiling is not in assertion mode
      - a new clock if it is in assertion mode and t is tsignal
      - undefined clock otherwise.
*/
ptinsth
kernel2graph::getDclContextClock (boolean bstate, typinst t)
{
  if (bstate) return hundef;
  if (!bWithAssertions) return hundef;
  if (t == tsignal) return CKM->new_variFormulaRootClock ();
  return hundef;
}

/**
   @return "true" if the l-th identifier of the list l exists, "false" otherwise.
   When it exists, vexpcal is the current instance associated with it, if such an element does not exist, one is created and assigned to it.
*/
boolean
kernel2graph::getCurrentInstance (ptlrepid l, int numero, typinst t, ptexpcal * vexpcal)
{
  ptlrepid vrepid;
  TypeType TT;

  if (!ptlrepid_getRepidAt (l, numero, &vrepid)) return pk_false;
  *vexpcal = getCurrentInstance (vrepid);
  if ((*vexpcal) == NILEXPCAL)
    {
      TT = AnyTypeCreate (); /*opimplicit */
      *vexpcal = EM->mkIdentExp (identpile (nameId (vrepid)), t, TT, NM->mk_node_undefined (), getDclContextClock (pk_false, t), back_annot_obj);
      setCurrentInstance (vrepid, *vexpcal);
      warning (warcreateinst, currentTree, vrepid);
    }
  return pk_true;
}

/**  
     @return the list of the "current" instances, associated with a list of declaration (list), inserted in a operand list (l).
     The stack (input or ouput) to select is given by the mode (true for inputs, pk_false for outputs).
*/
ptlinstances
kernel2graph::getCurrentInstances (ptlrepid list, ptlinstances l, boolean mode)
{
  int n;
  boolean notEmpty, x;
  ptexpcal e;

  x = setAccessMode (mode);
  n = 1;
  notEmpty = pk_true;
  while (notEmpty)
    {
      notEmpty = getCurrentInstance (list, n, tsignal, &e);
      if (notEmpty)
	{
	  l = addInstOf (l, e);
	  n++;
	}
    }
  setAccessMode (x);
  return l;
}

/**  @return the list of expressions current instances associated with a list (list) of declarations (type/signals).
 */
ptlinstances
kernel2graph::getCurrentExpcal (ptlrepid list)
{
  int n, i;
  ptexpcal vexp;
  ptlinstances listr;
  ptlrepid vrepid;

  listr = NILLINSTANCES;
  n = ptlrepid_size (list);
  for (i = n; i > 0; i--)
    {
      ptlrepid_getRepidAt (list, i, &vrepid);
      if (isTypeId (vrepid))
	vexp = EM->mkExpType (getCurrentType (vrepid));
      else
	vexp = getCurrentInstance (vrepid);
      listr = ptlinstances_push_front (listr, vexp);
    }
  return listr;
}

/** Pushs a new type (aType) for a declaration (vrepid). */
void
kernel2graph::setCurrentType (ptlrepid vrepid, TypeType aType)
{
  kl_pushTypeInstance (nameId (vrepid), aType);
}


/**   
   @return 
   - "true" if a pragma, with code name i, in a list of pragmas (lp) exists, then it is pushed in the context. vp is the pragma (named vname) before this call.
   - "false" otherwise (vp is not modified).
 */
boolean
kernel2graph::pushPragma (ptlpragmas lp, int i, ptpragma * vp)
{
  ptpragma vr;

  if (!getPragmaKeyword (lp, i, &vr))
    return pk_false;
  *vp = igc->getRunOnPragma ();
  igc->setRunOnPragma (vr);
  return pk_true;
}

/**
   It pushes (managed in igc object) the pragmas RUN_ON, GREY_BOX, BLACK_BOX of a model (vfi), if such a pragma exists.
   @param vptSave (out) the previous value of the RUN_ON pragma.
   @param vGBSave (out) the previous value of the GREY_BOX pragma.
   @param vBBSave (out) the previous value of the BLACK_BOX pragma.
   @param vSignalthread (out) the previous value of the isSignalThreadModel context variable.
*/
void
kernel2graph::pushContext (ptrepfi vfi, ptpragma * vptSave, ptpragma * vGBSave, ptpragma * vBBSave, boolean *vAutomaton, boolean *vSignalthread)
{
  *vptSave = igc->getRunOnPragma ();
  *vGBSave = igc->getGreyBoxPragma ();
  *vBBSave = igc->getBlackBoxPragma ();
  *vSignalthread = isSignalThreadModel; // the new value of isSignalThreadModel is set in instPragmas.
  *vAutomaton = isAutomatonModel;

  ntrees t = kl_getTreeReference(vfi);
  isAutomatonModel = LL_isAutomatonFamilyOp (noper(t));
  // setting the new pragmas.
  instPragmas (vfi); 
}

/**
   It pops (managed in igc object) the pragmas RUN_ON, GREY_BOX, BLACK_BOX of a model (vfi), if such a pragma exists.
   It fixes to (previous values)
   - vRunOnSave the pragma RUN_ON, 
   - the pragma GREY_BOX to vGreyBoxSave, 
   - the pragma BLACK_BOX  to vBlackBoxSave.
   These values have been previously returned by pushContext().
*/
void
kernel2graph::popContext (ptrepfi vfi, ptpragma vRunOnSave, ptpragma vGreyBoxSave, ptpragma vBlackBoxSave, boolean vAutomaton, boolean vSignalThread)
{
  ptpragma vr, vrunon;
  ptlpragmas lp;

  lp = kl_getPragmaModel (vfi);
  if (getPragmaKeyword (lp, PRAGMA_RUN_ON, &vr))
    if (getPragmaObjects (vr) == NILOBJPRAGMA)
      {
	vrunon = igc->getRunOnPragma ();
	if (vrunon != NILPRAGMA)
	  AGRAPH->addPragma (vrunon);
      }

  if (getPragmaKeyword (lp, PRAGMA_RESTRICTED_MODES, &vr))
    igc->popPragma (vr);

  igc->setRunOnPragma (vRunOnSave);
  igc->setGreyBoxPragma (vGreyBoxSave);
  igc->setBlackBoxPragma (vBlackBoxSave);
  isSignalThreadModel = vSignalThread;
  isAutomatonModel = vAutomaton;
}

/**
   @return true if the constant (vrepidconst) is a virtual one, "false otherwise.
   For a virtual one, the current top of thestack assigned to vrepid is selected.
*/
boolean kernel2graph::isVirtualConstant (ptlrepid vrepidconst)
{
  ptklExprSig    init;
  ptreptyp vtyp;
  categoryType vcat;
  ntrees vtree;
  ptexpcal vexpcal;
  TypeType TT;

  getInfoProto (vrepidconst, &init, &vtyp, &vcat, &vtree);
  if (init != NILEXPRSIG)
    return pk_false;		// for an external constant, it exists a value.
  vexpcal = getCurrentInstance (vrepidconst);
  TT = sigTreeType2TypeType (vtyp);
  // TT and type(vexpcal) must be compatible, or TT must be implicit ?
  setCurrentInstance (vrepidconst, vexpcal);
  return pk_true;
}

/** Instantiation of the list of declarations (Constants and types): 
    For each element of list, a new instance is created.
*/
void
kernel2graph::sigTreeDeclConstantsTypes2DCG (ptlrepid list)
{
  int numero;
  boolean notEmpty;
  ntrees savetree;
  ptreptyp vtyp;
  ptlrepid vrepid;
  TypeType TT2;
  ptexpcal vexpcal;
  BackAnnotationType save_back_annot_obj;

  savetree = currentTree;
  save_back_annot_obj = back_annot_obj;

  numero = 1;
  notEmpty = pk_true;
  while (notEmpty)
    {
      notEmpty = ptlrepid_getRepidAt (list, numero, &vrepid);
      if (notEmpty)
	{
	  if (isTypeId (vrepid))
	    {
	      ptlrepid_getIdInfosAttype (list, numero, &vtyp);
	      TT2 = mkInstType (vrepid, vtyp);
	      setCurrentType (vrepid, TT2);
	    }
	  else if (isFlowId (vrepid))
	    if (!isVirtualConstant (vrepid))
	      {
		mkInstance (vrepid, tconstant, pk_false, &vexpcal);
		igc->createConstantDefinition (vexpcal, EM->getInitialValue (vexpcal), MK_BACK_ANNOT(flowIdGetTree (vrepid)));
		setCurrentInstance (vrepid, vexpcal);
	      }
	  numero = numero + 1;
	}
    }
  currentTree = savetree;
  back_annot_obj = save_back_annot_obj;
}

/**
   Compiling the spec of a bundle type (TT) for a signal (vsig) typed TT.
   The link between a bundle type (TT) and its declaration is available by the global table (BundleTable).
   This link is set in mkBundleType(), when the declaration of the type is compiled.
   @return true when TT is a bundle type, false otherwise.
*/
boolean
kernel2graph::specBundleType (TypeType TT, ptexpcal vsig)
{
  ptreptyp vtyp;
  ptklExprProc bundleSpec;
  ptlrepid lrepidfields, listrepid;
  ptlinstances l;
  ptinsth vh;
  ptexpcal exp;

  TT = TypeTypeGetUnamedType (TT);
  if (!TypeTestBundleClass (TT))
    return pk_false;
  vtyp = BundleTable[TT];
  //  if (kl_getSpecBundleType(vtyp, &bundleSpec))
  {
    kl_getRepidFields (vtyp, &lrepidfields);
    l = igc->mkFieldsDescription (vsig);
    listrepid = lrepidfields;
    while (listrepid != NILLREPID)
      {
	ptlinstances_get_next (&l, &exp);	// Assumed: order of linstances = order of entree
	setCurrentInstance (listrepid, exp);
	listrepid = succid (listrepid);
      }
    if (kl_getSpecBundleType (vtyp, &bundleSpec))
      vh = compilingEproc (bundleSpec);
    deplinst (lrepidfields);
  }
  return pk_true;
}

/**
    @return "true" if the numero-th identifier of ther list l exists, "false" otherwise.
    When numero-th identifier of ther list l exists:
      - it creates a new instance (vexpcal) with t as type (signal, param,index,...),
      - vrepid (out) is the declartion of the numero-th identifier.
    When the boolean isExternRef is "true", a new signal, associated with  a model declared as "statevar",
    is not declared as a statevar (used in unexpanded compiling).
*/
void
kernel2graph::mkInstance (ptlrepid vrepid, typinst t, boolean isExternRef, ptexpcal * vexpcal)
{
  ptklExprSig vinit;
  ntrees arb;
  ptreptyp vtyp;
  TypeType TT;
  ptexpcal vexpcalInit;
  ptinsth vh;
  boolean bstate, b;
  categoryType vcat;
  BackAnnotationType save_back_annot_obj;

  arb = currentTree;
  save_back_annot_obj = back_annot_obj;

  getInfoProto (vrepid, &vinit, &vtyp, &vcat, &currentTree);
  NEW_BACK_ANNOT(currentTree);
  
  TT = sigTreeType2TypeType (vtyp);

  if (isExternRef)
    bstate = pk_false;
  else
    bstate = (vcat == statevarSignal);
  instExprInit (vinit, &vexpcalInit, &vh);
  *vexpcal = EM->mkIdentExp (identpile (nameId (vrepid)), t, TT, vexpcalInit, NM->mk_node_undefined (), getDclContextClock (bstate, t), back_annot_obj);
  b = specBundleType (TT, *vexpcal);
  if (bstate)
    {
      if (b)
	pK_messenger->errorTree (errTypeStateVar, currentTree);
      EM->setStateTag (*vexpcal, pk_true);
      lstateVar = ptlinstances_push_front (lstateVar, *vexpcal);
    }
  else if (vcat == sharedSignal)
    EM->setSharedSignal (*vexpcal, pk_true);
  if (!CKM->satisfiesUndefClock (vh))
    igc->getSignalClockSynchronizedWith (*vexpcal, vh);
  currentTree = arb;
  back_annot_obj = save_back_annot_obj;
}

/**
    @return "true" if the numero-th identifier of ther list l exists, "false" otherwise.
    When numero-th identifier of ther list l exists:
      - it creates a new instance (vexpcal) with t as type (signal, param,index,...),
      - vrepid (out) is the declartion of the numero-th identifier.
    When the boolean isExternRef is "true", a new signal, associated with  a model declared as "statevar",
    is not declared as a statevar (used in unexpanded compiling).
*/
boolean
kernel2graph::mkInstance (ptlrepid l, int numero, ptlrepid * vrepid, typinst t, boolean isExternRef, ptexpcal * vexpcal)
{
  if (!ptlrepid_getRepidAt (l, numero, vrepid))
    return pk_false;
  mkInstance (*vrepid, t, isExternRef, vexpcal);
  return pk_true;
}


/**
   It creates a new instance (cf mkInstance() method) associated with the numero-th element of a list of declarations (l)
   and, when the operation succeeds, pushes it.
*/
boolean
kernel2graph::mkAndPushInstance (ptlrepid l, int numero, typinst t, boolean isExternRef, ptexpcal * vexpcal)
{
  ptlrepid vrepid;

  if (!mkInstance (l, numero, &vrepid, t, isExternRef, vexpcal))
    return pk_false;
  setCurrentInstance (vrepid, *vexpcal);
  return pk_true;
}

/**
   @return the list of the instances of expressions associated with a list of declarations (list) followed by a given list (combedExp).
   The order of list is kept.
*/
ptexpcal
kernel2graph::mkOrderedListOfInstances (ptlrepid list, ptexpcal combedExp)
{
  int n, i;
  ptexpcal vexp;
  ptlrepid vrepid;
  ptrepfi vfi;
  ptrepinterface vinter;
  noperators vop;

  n = ptlrepid_size (list);
  for (i = n; i > 0; i--)
    {
      ptlrepid_getRepidAt (list, i, &vrepid);
      if (isTypeId (vrepid))
	vexp = EM->mkExpType (getCurrentType (vrepid));
      else if (isModelId (vrepid))
	{ // Pour test.
	  vfi = getModelRepid (vrepid);
	  vexp = EM->mkConstantExp (opcstcha, ident2repstr (kl_getNameModel (vfi)));
	  vinter = kl_getInterfaceModel (vfi);
	  vop = kl_getInterfaceOperator (vinter);
	  if (vop == opprocess) vop = opformproc; else if (vop == opnode) vop = opformnode; else if (vop == opact) vop = opformact; else vop = opformfunc;
	  vexp = EM->mkUnaryExp (vop, vexp, MK_BACK_ANNOT(NILTREE) );	// op
	}
      else
	vexp = getCurrentInstance (vrepid);
      combedExp = EM->mkNaryExp (oplesign, vexp, combedExp, MK_BACK_ANNOT(NILTREE));
    }
  return combedExp;
}


/**
   Builting the list of expressions  associated with a list of declarations of external references (referenced as inputs) for an unexpanded model.
   @return the list of the instances of expressions associated with a list of declarations (list) followed by a given list (combedExp). The order of list is kept.
   When an element is a "statevar" signal, a new signal is created with vh as clock: it is defined by ?x where x is the stae variable.
   The list l (out) contains the signals of referenced in the returned list.   
*/
ptexpcal
kernel2graph::unexpdandedModelExternalReferencesIn (ptlrepid list, ptexpcal combedExp, ptinsth vh, ptlinstances * l)
{
  int n, i;
  ptexpcal vexp, v1;
  ptlrepid vrepid;
  ptINode node;
  TypeType TT;

  n = ptlrepid_size (list);
  for (i = n; i > 0; i--)
    {
      ptlrepid_getRepidAt (list, i, &vrepid);

      if (isTypeId (vrepid))
	vexp = EM->mkExpType (getCurrentType (vrepid));
      else
	{
	  vexp = getCurrentInstance (vrepid);
	  *l = ptlinstances_insert (*l, vexp);
	  if (EM->isStateVar (vexp))
	    {
	      // v1 = EM->mkCopyIdentExp(vexp, NM->mk_node_undefined(), vh);
	      TT = EM->getType (vexp);
	      v1 = EM->mkIdentExp (NILIDENT, EM->getClassIdentifier (vexp), TT, NM->mk_node_undefined (), vh, back_annot_obj);
	      specBundleType (TT, v1);
	      vexp = EM->mkUnaryExp (opvvarin, vexp, back_annot_obj);
	      node = NM->mk_node_equation (v1, vexp, vh, NILLINSTANCES, back_annot_obj);
	      NM->insertNode (node, vh);
	      vexp = v1;
	    }
	}
      combedExp = EM->mkNaryExp (oplesign, vexp, combedExp, back_annot_obj);
    }
  return combedExp;
}


/**
   @return the list (l) in which the signals referenced in an expression (e) are inserted.
*/
ptlinstances
kernel2graph::addInstOf (ptlinstances l, ptexpcal e)
{
  cloper c;
  ptident id;
  ptexpcal le, lp;
  noperators op;

  // Note de LB: voir si on peut remplacer par getSignalsReferencedIn()
  c = EM->getOperatorClass (e);
  if (c == variabl)
    l = ptlinstances_insert (l, e);
  else if (c == call)
    {
      EM->getFunctionCall (e, &id, &lp, &le);
      while (EM->getNaryArgs (&le, &e))
	l = addInstOf (l, e);
    }
  else
    {
      op = EM->getOperator (e);
      if ((op == opvvarin) /* ref to a state variable x? */  || (op == opsdec) /* x[?] */ )
	l = addInstOf (l, EM->getUnaryArg (e));
      else if (op == opelem)
	{
	  l = addInstOf (l, EM->getBinaryArgstAt (e, 1));
	  l = addInstOf (l, EM->getBinaryArgstAt (e, 2));
	}
    }
  return l;
}


/**
   It creates a index identifier, associates it upper bound value (ubound) and declares it in the current graph.
   When the declaration (vexpind) is undefined
       - its name is used to define the name of the new instqnce of index.
       - the created index is pushed in the stack associated with vexpind.
   otherwise, the name is built from a predefined (internal) name.
   */
ptexpcal
kernel2graph::mkNewIndex (ptlrepid vexpind, ptexpcal ubound)
{
  ptexpcal vexpcal;
  ptident id;

  if (vexpind == NILLREPID)
    {
      id = getIndexId ();
      vexpcal = EM->mkIndexIdent (id, ubound, NM->mk_node_undefined (), hundef, back_annot_obj);
    }
  else
    {
      id = identpile (nameId (vexpind));
      vexpcal = EM->mkIndexIdent (id, ubound, NM->mk_node_undefined (), hundef, back_annot_obj);
      setCurrentInstance (vexpind, vexpcal);
    }
  AGRAPH->insertIndex (vexpcal);
  return vexpcal;
}


/**
   It associates the i-th element of linstances with the i-th element of listrepid.
*/
void
kernel2graph::setCurrentInstances (ptlrepid listrepid, ptexpcal linstances)
{
  ptexpcal arg;

  while (listrepid != NILLREPID)
    {
      EM->getNaryArgs (&linstances, &arg);
      setCurrentInstance (listrepid, arg);
      listrepid = succid (listrepid);
    }
}


/**
   For each element of a list of declarations (listrepid), it associates a copy of the current associated value.
*/
void
kernel2graph::setCopyCurrentInstances (ptlrepid listrepid)
{
  ptexpcal arg, nexp;

  while (listrepid != NILLREPID)
    {
      arg = getCurrentInstance (listrepid);
      igc->setClockVariable (arg);
      nexp = EM->mkIdentClone (arg, NM->mk_node_undefined ());
      setCurrentInstance (listrepid, nexp);
      listrepid = succid (listrepid);
    }
}

/**
   The pragma vp is associated with the signals of a list (l).
   @return "true" if l is not an empty list, "false" otherwise.
*/
boolean
kernel2graph::assignPragmaToSignals (ptlinstances l, ptpragma vp)
{
  ptexpcal aSig;
  ptrepstr name;
  ptINode node;
  ntrees stmt;

  if (l == NILLINSTANCES)
    return pk_false;
  name = getPragmaName (vp);
  stmt = getPragmaStatement (vp);
  while (ptlinstances_get_next (&l, &aSig))
    {
      node = EM->getNode (aSig);
      // if (NM->removePragma(node, name)); // warning();
      if (!NM->addPragma (node, vp));	// warning() : already associated with. The first is kept!!.
      EM->addPragma (aSig, newpragma (name, NILOBJPRAGMA, stmt));
    }
  return pk_true;
}

boolean  kernel2graph::getSignal(ptlrepid l, ptident id, ptexpcal *vsig)
{
  ptlrepid vrep;

  if (!searchId(l, id, &vrep)) return pk_false;
  *vsig = getCurrentInstance (vrep);
  return pk_true;
    
}

/**
   Assumed : vfi is a SIGNAL thread model.
   It verifies that the operand model (vfi) has the Complete(output) and Dispatch (input) signals in it interface.
   @return true if the test is satisfied, false otherwise.
 */
boolean kernel2graph::verify_SIGNAL_Thread(ptrepfi vfi)
{
  ptlrepid vrep, entrees, sorties;
  boolean b;
  ptident id;

  b = pk_true;
  sorties = kl_getInsOutsModel (vfi, &entrees);
  id = getDispatchId();
  if (!searchId(entrees, id, &vrep))
    {
      pK_messenger->errorTreeIdent(errUndefSignalSignalThread, kl_getTreeReference(vfi), id);
      b = pk_false;
    }
    
  id = getCompleteId();
  if (!searchId(sorties, id, &vrep))
    {
      pK_messenger->errorTreeIdent(errUndefSignalSignalThread, kl_getTreeReference(vfi), id);
      b = pk_false;
    }
  return b;
}

/** Compiling pragmas: they are transfered from the model to the current graph. 
 */
void
kernel2graph::instPragmas (ptrepfi vfi)
{
  ptlpragmas lprag, lp;
  ptrepstr name;
  ptObjPragma obj;
  ntrees enonce;
  ptpragma v;
  ptlinstances e;
  boolean bSignalThreadModel, b;

  if (!kl_isIntrinsicModel (vfi))
    {
      lprag = kl_getPragmaModel (vfi);
      lp = NILLPOINTER;
      bSignalThreadModel =false;
      while (ptlpragmas_get_next (&lprag, &v))
	{
	  getPragma (v, &name, &obj, &enonce);
	  e = instSignalsLabels ((ptklExprSig) obj);
	  b = pk_false;
	  v = newpragma (name, (ptObjPragma) e, enonce);
	  if (assignPragmaToSignals (e, v))
	    b = pk_true;
	  else if (isPragmaKeyword (v, PRAGMA_RUN_ON))
	    igc->setRunOnPragma (v);
	  else if (isPragmaKeyword (v, PRAGMA_GREY_BOX))
	    igc->newGreyBoxPragma (v);
	  else if (isPragmaKeyword (v, PRAGMA_BLACK_BOX))
	    igc->setBlackBoxPragma (v);
	  else if (isPragmaKeyword (v, PRAGMA_RESTRICTED_MODES))
	    {
	      igc->pushPragma (v);
	      b = pk_true;
	    }
	  else
	    b = (this->vfi == vfi);

	  if (b) 
	    {
	      AGRAPH->addPragma (v);
	      if (isPragmaKeyword (v, PRAGMA_SIGNAL_THREAD))
		bSignalThreadModel = verify_SIGNAL_Thread(vfi);
	    }
	}
      isSignalThreadModel = bSignalThreadModel;
    }
}

/**
   @return lio + the i/o signals of the list l that are in lwhere.
*/
ptlinstances
kernel2graph::getIOElemsOf (ptlrepid l, ptlinstances lwhere, ptlinstances lio)
{
  int n, i;
  ptexpcal vexp;

  n = ptlrepid_size (l);
  for (i = 1; i <= n; i++)
    {
      vexp = kl_topInstance (nameId (l));
      if (ptlinstances_isInSortedList (lwhere, vexp))
	lio = ptlinstances_insert (lio, vexp);
      l = succid (l);
    }
  return lio;
}

/**
 The pragma
   - BindingIO {e1} "xxx"
  where e1 is the label (vabelname) is replaced by
   - BindingIO  {s1,...sn} "xxx" where the identifiers si are input/output signals of the current graph.
*/
void
kernel2graph::propagatePragmas (ptklExprProc vklExprProc, ptexpcal vabelname)
{
  ptlpragmas lprag;
  ptlrepid entrees, sorties, params;
  ptpragma vpragma;
  ptlinstances lio;
  boolean found;
  ptrepstr name;
  ptObjPragma obj;
  ntrees enonce;

  lprag = AGRAPH->getPragmas ();
  found = pk_false;
  while (!found && ptlpragmas_get_next (&lprag, &vpragma))
    if (isPragmaKeyword (vpragma, PRAGMA_BINDINGIO))
      {
	getPragma (vpragma, &name, &obj, &enonce);
	found = ptlinstances_isin_exp ((ptlinstances) obj, vabelname);
      }
  if (found)
    {
      kl_getInsOutsParams (vklExprProc, &entrees, &sorties, &params);
      lio = ptlinstances_remove_exp ((ptlinstances) obj, vabelname);
      lio = getIOElemsOf (entrees, linputs, lio);
      lio = getIOElemsOf (sorties, loutputs, lio);
      setPragmaObjects (vpragma, (ptObjPragma) lio);
      assignPragmaToSignals (lio, vpragma);
    }
}

/**
   Local to sigTreeFlowsExternalReferences()
*/
void
kernel2graph::marklrepid (ptlrepid l, typeMark m, boolean bOutput)
{
  int n, i;
  ptexpcal vexp, vin;

  n = ptlrepid_size (l);
  for (i = 1; i <= n; i++)
    {
      vexp = kl_topInstance (nameId (l));
      EM->addMark (vexp, m);
      if (bOutput)
	{
	  vin = getStackTopMinus (l, 1);	// input
	  EM->addAttribute (vexp, PREVIOUSVALUEATTR, (attributeValue) vin);
	}
      l = succid (l);
    }
}

/**
   Specific to unexpanded model compiling: references to external declarations are replaced by inputs (element of externRefIn list) 
   or outputs (element of externRefIn list).
   These new signals are tagged by specific attributes (INPUT_ADDED_FOR_UNEXPANDED, OUTPUT_ADDED_FOR_UNEXPANDED).
*/
void
kernel2graph::sigTreeFlowsExternalReferences (ptlrepid externRefIn, ptlrepid externRefOut)
{
  sigTreeFlows2DCG (externRefIn, pk_true);
  marklrepid (externRefIn, INPUT_ADDED_FOR_UNEXPANDED, pk_false);
  sigTreeFlows2DCG (externRefOut, pk_true);
  marklrepid (externRefOut, OUTPUT_ADDED_FOR_UNEXPANDED, pk_true);
}

/** Initialisations : it assigns the name of the graph by id, it gets predefined objects, and initialises internal variables.
    Assumed: managers are known (getManager() method has been previously applied).
    It is referenced in sigProduceModule() and sigProduceGraph() methods.
*/
void
kernel2graph::produceGraphInit (ptident id)
{
  numappel = 0; 
  AGRAPH->setName (id);

  // Getting predefined Clocks
  hinit = CKM->getInitClock ();
  the_nullClock_K2G_copy = CKM->getNullClock ();
  hundef = CKM->getUndefClock ();

  // Getting/Cretaing predefined expressions
  vexptrue = EM->getTrue ();
  vexpfalse = EM->getFalse ();
  ptexp0 = EM->mkIntegerConstant (0);
  ptexp1 = EM->mkIntegerConstant (1);

  // Getting predefined Types
  UndefinedType = TypeTypeGetUndefinedType ();

  // constants are managed in intemkIGraph object.
  lparameters = NILLINSTANCES;
  lstateVar = NILLINSTANCES;
  bWithAssertions = pk_false;
  inAssertBlock = pk_false;
}

/**
   The following actions are applied:
   - For a node, a function or an action, I/O dependencies are added.
   - For a function, the I/O are synchronous.
   @return the synchronized clock of the I/O for a function, undefined clock otherwise.
*/
ptinsth kernel2graph::addingInterfaceRelations (ptlinstances linputs, ptlinstances loutputs)
{
  noperators vop;
  ptinsth vh0;

  // i/o are synchronous for a function.
  vop = AGRAPH->getKind ();
  if (LL_isFunctionFamilyOp (vop))
    vh0 = CKM->makeSynchro (igc->synchronizeIdentifiers (linputs), igc->synchronizeIdentifiers (loutputs));
  else
    vh0 = hundef;
  // Adding dependencies Inputs to Outputs for function, node, action.
  if (LL_isFunctionFamilyOp (vop) || LL_isNodeFamilyOp (vop) || LL_isActionFamilyOp (vop))
    NM->addDependences (linputs, loutputs, hundef, pk_true);
  return vh0;
}



ptinsth kernel2graph::genDefinitionsOutputBooleanClocks(ptexpcal vboolsig, ptexpcal vsig)
{
  ptexpcal zvboolsig;
  ptident ids, id;
  ptinsth vassClock, vhInf, vh1;
  
  vassClock = EM->getSetAssignmentClock(vsig);
  if (CKM->satisfiesNullClock(vassClock))
    return  igc->createEquation (vboolsig, vexpfalse, vhOversamplingSignalThread, vhOversamplingSignalThread, NILLINSTANCES, MK_BACK_ANNOT(NILTREE));

  ids = EM->getName (vboolsig);
  id = makeIdent_string_ident("Z",ids);
  zvboolsig = EM-> mkIdentExp(id, tsignal, get_BooleanType (), NM->mk_node_undefined (), vhOversamplingSignalThread, MK_BACK_ANNOT(NILTREE));
  igc->mkDelayWindowNode (zvboolsig, vboolsig, ptexp1, NILEXPCAL, vexpfalse, vhOversamplingSignalThread, MK_BACK_ANNOT(NILTREE));

  vhInf =  CKM->makeInf (vassClock, vhDispatchSignalThread);
  igc->addSubEquation (vboolsig, vexptrue, vhInf, NILLINSTANCES, MK_BACK_ANNOT(NILTREE));
  vh1 =  CKM->makeMinus (vhDispatchSignalThread, vassClock);
  igc->addSubEquation (vboolsig, vexpfalse, vh1, NILLINSTANCES, MK_BACK_ANNOT(NILTREE));
  vh1 = CKM->makeMinus (vassClock, vhDispatchSignalThread);
  igc->addSubEquation (vboolsig, vexptrue, vh1, NILLINSTANCES, MK_BACK_ANNOT(NILTREE));
  return igc->createEquation (vboolsig, zvboolsig, CKM->makeMinus (vhOversamplingSignalThread, CKM->makeSup(vassClock, vhDispatchSignalThread)), 
			      vhOversamplingSignalThread, ptlinstances_insert (NILLINSTANCES, zvboolsig), MK_BACK_ANNOT(NILTREE));
}

/**
   Signal Thead model.

   It adds the definition of the boolean clock of the signals of l.
   For x, the current instance of l(i), it creates a new signal x_bclk. It is defined by:
    - x_bclk := (true when H) default false when Dispatch_Thread default (x_bclk $1 init false), where H is the "assignment clock" of x
    - synchronized with the master clock of the Thread ( vhOversamplingSignalThread).
 */
void kernel2graph::genDefinitionsOutputBooleanClocks(ptlrepid l)
{
  int n, i;
  ptexpcal vsig, vboolsig;
  ptident ids, id;
  ptinsth vh;

  n = ptlrepid_size (l);
  for (i = 1; i <= n; i++)
    {
      vsig = kl_topInstance (nameId (l));
      ids = EM->getName (vsig);
      id = makeIdent_ident_string(ids, "iBclk"); /* internal name of the boolean clock */
      vboolsig = EM->mkIdentExp (id, tsignal,get_BooleanType (), NM->mk_node_undefined (), vhOversamplingSignalThread,  MK_BACK_ANNOT(NILTREE));
      vh = genDefinitionsOutputBooleanClocks( vboolsig, vsig);
      igc->addOutNode (vboolsig, NILINODE);
      loutputs = ptlinstances_insert(loutputs, vboolsig);  // Output is added to the current graph.
      l = succid (l);
    }
}


/** Finalize a Signal thread:
    - set the Complete Clock of the model
*/
void kernel2graph::SignalThreadFinalize(ptlrepid inFlows, ptlrepid outFlows)
{
  ptexpcal vCompleteSig;

  if (isSignalThreadModel)
    {
      // genDefinitionsOutputBooleanClocks(outFlows);
      getSignal(outFlows, getCompleteId(), &vCompleteSig);
      AGRAPH->setCompleteClock(EM->getClock (vCompleteSig));
      popUndersampling(succid (inFlows)); // Except the Dispatch that is the first one.
    }
}

/**
   For a Signal thread, the signal Dispatch and Complete exist in the interface.
   This function creates an oversampling (the internal clock of the thread): ^bstop.
   (| bstop = Complete default (not Dispatch) default zbstop
    | zbstop := bstop $1 init true
    | e1 ^= .... ^= en ^= Dispatch  % the inputs are synchronous %
    | Dispatch ^= when zbtop  % accepts new inputs when Complete is true %
    | % new internal signals are used for inputs. The inputs are memorized at the clock of the thread.%
    | ei' := ei cell ^bstop
    |) 
    Moreover, it the model is an automaton, the clock of the automaton is the  clock of the thread.
    @return the oversampling clock or hundef (when it is not the compiling of a Signal Thread).
 */
ptinsth kernel2graph::SignalThreadInitialize(ptident id, ptlrepid inFlows, ptlrepid outFlows)
{
  ptexpcal zStop, Stop, vComplete, vDispatch;
  ptinsth vhsub, vhStop, vhDispatch, vhComplete, vh;
  ptlinstances linst;

  if (isSignalThreadModel)
    {
      Stop = EM-> mkIdentExp(newIdent_stringSuffixed(id,"bstop"), tsignal, get_BooleanType (), NM->mk_node_undefined (), hundef, MK_BACK_ANNOT(NILTREE));
      vhStop = igc->mkClockInstance(Stop);
      zStop = EM-> mkIdentExp(newIdent_stringSuffixed(id,  "zbstop"), tsignal, get_BooleanType (), NM->mk_node_undefined (), vhStop, MK_BACK_ANNOT(NILTREE));
      igc->mkDelayWindowNode (zStop, Stop, ptexp1, NILEXPCAL, vexptrue, vhStop, MK_BACK_ANNOT(NILTREE));


      getSignal(inFlows, getDispatchId(), &vDispatch);
      vhDispatch = igc->mkClockInstance(vDispatch);
      vhDispatchSignalThread = vhDispatch; // Global
      getSignal(outFlows, getCompleteId(), &vComplete);
      vhComplete = igc->mkClockInstance(vComplete);
      igc->addSubEquation (Stop, vexptrue, CKM->makeInf(vhDispatch, vhComplete), ptlinstances_insert (NILLINSTANCES, vDispatch ), MK_BACK_ANNOT(NILTREE));
      igc->addSubEquation (Stop, vexpfalse, CKM->makeMinus (vhDispatch, vhComplete) , ptlinstances_insert (NILLINSTANCES, vDispatch ), MK_BACK_ANNOT(NILTREE));
      igc->addSubEquation (Stop, vComplete, CKM->makeMinus (vhComplete, vhDispatch), ptlinstances_insert (NILLINSTANCES, vComplete ), MK_BACK_ANNOT(NILTREE));
      vh = igc->createEquation (Stop, zStop, CKM->makeMinus (vhStop, CKM->makeSup(vhComplete, vhDispatch)), vhStop, ptlinstances_insert (NILLINSTANCES, zStop), MK_BACK_ANNOT(NILTREE));

      // The inputs are synchronized with "when zbStop."
      linst = getCurrentInstances (inFlows, NILLINSTANCES, pk_true);
      vh = igc->synchronizeIdentifiers(linst);
      ptlinstances_free (linst);
      vhsub = CKM->makeSub (vhStop, zStop, ptlinstances_insert (NILLINSTANCES, zStop));
      CKM->makeSynchro(vh, vhsub);
      
      // The inputs, except the dispatch, are memorized at the oversampling (master) clock and pushed.
      pushCellAt(succid(inFlows), vhStop);

      // For a automaton thread, the clock of the automaton is the thread clock.
      if (isAutomatonModel) 
	{
	  CKM->makeSynchro(igc->getClockAutomaton(), vhStop);
	  igc->setClockedAutomaton(true);
	}
      return vhStop;
    }
  return hundef;
}

/**
   Finalizing an automaton:
   - If the clock of the automaton is not fixed, it is defined by the upper bound of the clocks of the signals of lsig.
   - It verifies that the initial state is fixed.
 */
void kernel2graph::automatonFinalize(ptlrepid l)
{
  ptlinstances lsig;

  if (isAutomatonModel)
    {
      AGRAPH->setKind (opprocess);// the automaton macros have been substituted.
      // Fixing the clock of an automaton if it is not yet fixed.
      if (! isSignalThreadModel && ! igc->isClockedAutomaton())
	{
	  lsig = getCurrentInstances (l, NILLINSTANCES, pk_true);
	  CKM->makeSynchro (igc->getClockAutomaton (), igc->upperBoundClockOfIdentifiers (lsig));
	}
      
      // The initial state must be fixed.
      ptexpcal vCurrentState = igc->getAutomatonCurrentStateVar ();
      ptexpcal vinit = EM->getInitialValue (vCurrentState);
      if (vinit == NILEXPCAL) pK_messenger->errorTree (errNotInitializedAutomaton, currentTree);
    }
}

/**
   Producing a graph from an instance of a model (vfi attribute of the kernel2graph class).
*/
void
kernel2graph::sigProduceGraph ()
{
  ptlrepid entrees, sorties, param, externRefIn, externRefOut;
  ptklExprProc corps, spec;
  ptinsth vClockAutomatonSave, vh0, vh1, vh2;
  ptident idmodel, save_id, id;
  ptpragma vRunOnSave, vGreyBoxSave, vBlackBoxSave;
  ptexpcal vCurrentStateAutomatonSave;
  boolean  vAutomatonModelSave, SignalThreadSave, bIsClockedAutomaton;

  /*
    TODO :
    - for a node, the model must be endochronous (to be verified) without clock calculus (no synthesis).
    - for a function, the model must be safe. 
  */
  id = kl_getNameModel (vfi);
  pK_messenger->messageIdent (graphGeneration_subAnnounce, id);
  produceGraphInit (id);
  NEW_BACK_ANNOT( kl_getBodyModel (vfi, &corps));

  bWithAssertions = kl_getAssertionsFlag (vfi);
  CKM->setModeSynchroSolving (!bWithAssertions);
  param = kl_getParametersModel (vfi);
  sorties = kl_getInsOutsModel (vfi, &entrees);
  if (!instdclparameters (param, parglobal))	/* Producing an instance of the model. */
    {
      sigTreeInterface2DCG (param, entrees, sorties);
      lparameters = getCurrentExpcal (param);
    }
  else
    {
      sigTreeFlows2DCG (entrees, pk_false);
      sigTreeFlows2DCG (sorties, pk_false);
    }
  if (corps != NILEXPFIL)
    if (kl_getExprProcType (corps) == opexte)
      AGRAPH->setGraphOfExternalProcess (currentTree, kl_getLangExternal (corps));

  linputs = getCurrentInstances (entrees, NILLINSTANCES, pk_true);
  loutputs = getCurrentInstances (sorties, NILLINSTANCES, pk_false);

  externRefIn = kl_getExternalReferencesIn (vfi);
  externRefOut = kl_getExternalReferencesOut (vfi);
  sigTreeFlowsExternalReferences (externRefIn, externRefOut);
  linputs = getCurrentInstances (externRefIn, linputs, pk_true);
  loutputs = getCurrentInstances (externRefOut, loutputs, pk_false);

  traceability (mkOrderedListOfInstances (param, NILEXPCAL), mkOrderedListOfInstances (entrees, NILEXPCAL), mkOrderedListOfInstances (sorties, NILEXPCAL));

  vh0 = addingInterfaceRelations (linputs, loutputs);

  save_id = pushModelAndContextId (vfi, &idmodel, &vCurrentStateAutomatonSave, &vClockAutomatonSave, &bIsClockedAutomaton);
  pushContext (vfi, &vRunOnSave, &vGreyBoxSave, &vBlackBoxSave, &vAutomatonModelSave, &SignalThreadSave);
  
  ptinsth save_vhOversamplingSignalThread = vhOversamplingSignalThread;
  vhOversamplingSignalThread = SignalThreadInitialize(id, entrees, sorties);

  igc->addInNodes (linputs);
  vh1 = compilingEproc (corps);
  kl_getSpecModel (vfi, &spec);
  vh2 = compilingEproc (spec);
  igc->addOutNodes (loutputs);
  igc->updateSignalsNames ();

  automatonFinalize(entrees);
  SignalThreadFinalize(entrees, sorties);
  vhOversamplingSignalThread= save_vhOversamplingSignalThread;

  popModelAndContextId (vfi, save_id, idmodel, vCurrentStateAutomatonSave, vClockAutomatonSave, bIsClockedAutomaton);

  deplinst (param);
  deplinst (entrees);
  deplinst (sorties);

  deplinst (externRefOut);
  deplinst (externRefIn);

  popContext (vfi, vRunOnSave, vGreyBoxSave, vBlackBoxSave, vAutomatonModelSave, SignalThreadSave);
  AGRAPH->setParameters (lparameters);
  AGRAPH->setInputs (linputs);
  AGRAPH->setOutputs (loutputs);
  AGRAPH->setStatesVar (lstateVar);
}

/**
   Getting managers associated with the current graph.
*/
void
kernel2graph::getManagers ()
{
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  nameManager = AGRAPH->getNameManager ();
  TM = AGRAPH->getTraceabilityManager ();
  TM->setMode (btm);
}


/**
   Initialisations called by the constructors to fix :
    - the graph manager (gm),
    - the context (LabelsStack) of the labels stack (propagated for modules),
    - the traceability mode (traceabilityMode).
*/
void
kernel2graph::kernel2graphInit (ptGraphManager gm, ptStack LabelsStack, boolean traceabilityMode)
{
  GRAPHS_MANAGER = gm;
  igc = new intemkIGraph (AGRAPH);
  igc->setLabelStack (LabelsStack);
  INPUT_ADDED_FOR_UNEXPANDED = getMark ();
  OUTPUT_ADDED_FOR_UNEXPANDED = getMark ();
  XMARK = getMark ();
  REF_STATEVAR_ATTR = getMark ();
  btm = traceabilityMode;
  isAutomatonModel = false;
  // Signal Thread
  isSignalThreadModel = false;
  vhOversamplingSignalThread = hundef;
  vhDispatchSignalThread = hundef;
}


/**
   The ContextSubExpr identifier is set to FOO_subexpr, where FOO is the name of the identifier vexpcal.
   The operand (vexpcal) is assumed to represent an identifier.
   The ContextSubExpr identifier is used to created auxiliary expression, by default it is the NIL identifier.
   @return the value of the ContextSubExpr identifier before the call.
 */
ptident kernel2graph::setContextSubExpr (ptexpcal vexpcal)
{
  ptident id;

  id = EM->getOriginalName (EM->getDeclaredSignal (vexpcal));
  return nameManager->setContextSubExpr (makeIdent_ident_ident (id, getSubexprId ()));
}
