/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

static ptident vidnil, vidempty;
#define suffSeparator '_'
#define cstrempty "XZX"

#include <stdio.h>
#include <string.h>

/**
  @return "true" if the tree (res) built from the string contained in the register (s) is created? false otherwise.
static boolean regMake_ident(ptreg s, ptident * res)
{
  ptrepstr str;

  if (!strregstore(s, &str))
    return pk_false;
  *res = LL_make_ident_tree(LL_make_nom_atom(str), LL_make_lsuff_emptyList());
  return pk_true;
}
*/

/**
   It changes the string, in pos position (pos >1) in id, by an integer (str).
*/
static void identChange_inSuffix(ptident id, ptrepstr str, int pos)
{
  nchangchild(nchild(2, id), LL_make_ent_atom(str), pos - 1);
}

/**
  @return the string associated with the main part of an identifier (id).
  id = main_part + a set of suffixes.
*/
static ptrepstr identGet_strPrefix(ptident id)
{
  return nstringof(nchild(1, id));
}

/**
   @return the string associated with the i-th child of a tree (vl).
   (assumed: vl has at least i children, and the i-th is a string).
*/
static ptrepstr listGet_strAtIndex(int i, ntrees vl)
{
  ptrepstr vstr = nstringof(nchild(i, vl));
  return vstr;
}

/**
   @return the string associated with the i-th suffix of an identifier (id).
*/
static ptrepstr identGet_strAtIndexInsuffix(int i, ntrees id)
{
  return listGet_strAtIndex(i - 1, nchild(2, id));
}

/**
   Append aString as the last suffix of id
   Side effect on id
*/
static void identAppend_stringSuffix(ptident id, const char *aString)
{
  ntrees lsuff = nchild(2, id);
  ntrees lastSuff = LL_make_nom_atom(string2repstr(aString));

  if (noper(id) == opnom) {
    fprintf(stdout, " ERREUR **************** NOPER(VNAME) = OPNOM \n");
    fflush(stdout);
  }

  if (nisniltree(lsuff) || nisemptyList(lsuff)) {
    lsuff = nmakelist(oplsuff, lastSuff);
    nchangchild(id, lsuff, 2);
  }
  else
    (void)ntreeList_appendLast(lsuff, lastSuff);
}

/**  UNUSED
  append lsuffSuff to the end of lsuf
  assume lsuf not NIL

static void lsuffAppend_lsuffCopy(ntrees lsuf, ntrees lsuffSuff)
{
  int i;
  int lsuffSuffNumber;
  if (nisniltree(lsuffSuff) || nisemptyList(lsuffSuff))
    return;
  lsuffSuffNumber = ntreearity(lsuffSuff);
  for (i = 1; i <= lsuffSuffNumber; i++)
    (void)ntreeList_appendLast(lsuf, ncopytree(nchild(i, lsuffSuff), pk_true));
}
*/

/* UNUSED static void identAppend_identCopy(ptident id, ptident idSuff)
{
  ntrees lsuff = nchild(2, id);
  ntrees lsuffSuff = nchild(2, idSuff);
  ntrees newIdSuffPrefix = ncopytree(nchild(1, idSuff), pk_true);
  if (nisniltree(lsuff) || nisemptyList(lsuff)) {
    lsuff = nmakelist(oplsuff, newIdSuffPrefix);
    nchangchild(id, lsuff, 2);
  }
  else
    (void)ntreeList_appendLast(lsuff, newIdSuffPrefix);
  if (nisniltree(lsuffSuff) || nisemptyList(lsuffSuff))
    return;
  (void)lsuffAppend_lsuffCopy(lsuff, nchild(2, idSuff));
  } */

/**
   create an ident made of newPrefix followed by the suffix part in id
*/
static ptident identMake_identSuffix(ptrepstr newPrefix, ptident id)
{
  return LL_make_ident_tree(LL_make_nom_atom(newPrefix), ncopytree(nchild(2, id), pk_true));
}

/**
   @return 
   - 0 if the oper(p)=opnil, in this case nbempty is increased.
   - the size of the string designed by p (assumed to be a string).
*/
static int inclng(ntrees p, int *nbempty)
{
  noperators opera;

  opera = noper(p);
  if (opera == opnil) {
    *nbempty = (*nbempty) + 1;
    return 0;
  }
  return strsize(nstringof(p));
}

/**
  append to reg the lowercase prefix of id
  @param id (in,out) : an identifier, out the suffixes of id.
  @param reg (in, out) a string register, the prefix of id is catenated to reg in lower cases.
  @return false when internal error occurs (memory or register overflow)
*/
static boolean identLoad_lowerCasePrefix(ptident id, ptreg reg)
{
  ptrepstr vstr;

  if (strGet_lowerCaseStr(identGet_strPrefix(id), &vstr)) {
    /*
    if (vstr == nilstr) { ptrepstr x;
    if (strGet_lowerCaseStr(identGet_strPrefix(vidempty), &x))
    return strregconcstr(reg, x, pk_false);
    return pk_false;
    } 
    */
    return strregconcstr(reg, vstr, pk_false);
  }
  return pk_false;
}

/**
  @return "true" if the set of suffixes of an identifier (id) is not empty, "false" otherwise.
  @param id (in,out) : in an identifier, out the suffixes of id.
  @param reg (in, out) a string register, the prefix of id is catenated to reg.
*/
static boolean locidtoreg(ptident * id, ptreg reg)
{
  ptrepstr vstr;

  vstr = identGet_strPrefix(*id);
  /*
  if (vstr == nilstr)
    if (strregconcstr(reg, identGet_strPrefix(vidempty), pk_false)) ;
    else; 
    else 
  */
    strregconcstr(reg, vstr, pk_false);
  *id = nchild(2, *id);
  return (boolean) (!nisemptyList(*id));
}

/**
   @return the string composed of the string contained in a register (reg) to whom is 
   catenated the alphabetical representation (in lower cases) of an identifier (id).
   retunr nilstr in case of internal error (memory or register overflow)
*/
static ptrepstr regConcStore_lowerCaseId(ptreg reg, ptident id, char sep)
{
  if (identLoad_lowerCasePrefix(id, reg))
  {				// failure :internal error (memory or register overflow)
    ntrees suffList = nchild(2, id);
    if (!nisemptyList(suffList)) {
      ptrepstr vstr;
      ptrepstr x;
      int i;
      for (i = 1; i <= ntreearity(suffList); i++) {
	if (!strregconccha(reg, sep))
	  return nilstr;
	if (!strGet_lowerCaseStr(listGet_strAtIndex(i, suffList), &x))
	  return nilstr;
	if (!strregconcstr(reg, x, pk_false))
	  return nilstr;
      }
      if (strregstore(reg, &vstr))
	return vstr;
    }
  }
  return nilstr;
}

/**
  @return the string composed of the string contained in a register (reg)  to whom is 
  catenated the alphabetical representation of an identifier (id).
  @note similar to regConcStore_lowerCaseId().
*/
static ptrepstr appendidtoreg(ptreg reg, ptident id, char sep)
{
  ptrepstr vstr, xx;
  int i;
  ntrees son;

  son = id;
  if (locidtoreg(&son, reg)) {
    for (i = 1; i <= ntreearity(son); i++) {
      strregconccha(reg, sep);
      xx = listGet_strAtIndex(i, son);
      // printf("vstr = %d \n", xx);      printf("vstr = %s \n", repstr2string(xx));
      strregconcstr(reg, xx, pk_false);
    }
  }
  if (strregstore(reg, &vstr))
    return vstr;
  return nilstr;
}

/**
   It inserts a suffix (suff) in pos suffix in an identifier (id).
   @return (modified) id.
*/
static ptident identInsert_suffix(ptident id, ntrees suff, int pos)
{
  ninsertleft(pos - 1, suff, nchild(2, id));
  return id;
}

/**
   It inserts an integer element (str) in pos suffix in an identifier (id).
   @return (modified) id.
*/
static ptident identInsert_intSuffix(ptident id, int num, int pos)
{
  // Modified May 2019, LBesnard.
  // return identInsert_suffix(id, LL_make_ent_atom(integer2repstr(num)), pos);
  return identInsert_suffix(id, LL_make_nom_atom(integer2repstr(num)), pos);
}

ptrepstr make_localRepstr(int num)
{
  ptrepstr suffixe;
  strconcchastr('L', integer2repstr(num), &suffixe);
  return suffixe;
}

ptident make_localIdent(int num)
{
  ptrepstr suffixe;
  suffixe = make_localRepstr(num);
  return repstr2ident(suffixe);
}
