/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#ifndef __OptionManager_H__
#define __OptionManager_H__

typedef class OptionManager * ptOptionManager;

class OptionManager
{
#include "OptionManager_m.h"
#include "OptionManager_g.h"
  
 private:
  /* ====> input file options */
  boolean in_isf;
  char * commandFile; /* -com=FILE_NAME */
  char *srcFile, *paramFile; /* -par=FILE_NAME */
  boolean srcFileDefined, readParamFile;
  boolean paramFileDefined; /* true iff readParamFile is true and paramFile got a value */

  /* ====> output file options */
  boolean out_isf;
  boolean outputDirDefined; /// -d=DIRNAME: files are created in subdirectory FOO of outputDir
  boolean targetDirDefined; /// -D=DIRNAME: files are created in outputDir
  char * outputDir; /// -d=DIRNAME
    boolean opt_warn, opt_verbose, opt_cosimulation, opt_keeprew;
  boolean opt_trace_samplingAsWhen; ///< printing the unary when as "when" when it is true, [: ] otherwise.
  boolean opt_trace_unstructuredProcess; // print without subprocess.

/* ====> graph class transformation options */
  boolean opt_basicGraph, opt_tracebasic;
  boolean opt_polyGraph, opt_tracepoly;
  boolean opt_endoGraph, opt_traceendo;
  boolean opt_boolGraph, opt_tracebool;
  boolean opt_sbdcGraph, opt_tracesbdc; /* sequential code */
  boolean opt_flatGraph, opt_traceflat;
  boolean opt_monoGraph, opt_tracemono;
//  boolean opt_dcGraph;

  /* ====> kind of output */
  boolean opt_traceAllowed; /* if false no _TRA file is being generated whatever the other command options are*/
  boolean opt_genFile;
  boolean opt_list, opt_traceALL, opt_traceNone;
  boolean opt_sme, opt_ssme, opt_geneauto ;
  boolean opt_abstraction, opt_header;  
  boolean opt_Exclusive_Clock;  
  integer ecLevel;

 /* ====> graph options */
  integer opt_prof;
  boolean opt_retiming;
  boolean opt_crew_bool_ctrl, opt_crew_bool_all, opt_crew_definitions, opt_crew_statics;
  boolean opt_us, opt_default_splitting, opt_partialdef_merging, opt_default_normal;
  integer opt_substitute_value;
  boolean opt_delayMultiplexing, opt_delayOptimizing;
  boolean opt_kernel;
  boolean opt_splitDataStructure;
  boolean opt_enum2Integer;
  boolean opt_utility_clock; ///< Optimisations using the utility clock.
  boolean opt_check; ///< dynamically check clock constraints of the Poly; sub option of -force ??
  boolean opt_distrib; ///< distributed code option.
  boolean opt_polyendo; ///< testing polyendochronous hierarchy.

 /* ====> partition options */
  boolean opt_states_others, opt_bool_others, opt_finiteTypes_others, opt_clustering;
  boolean opt_tagBased_partition;

 /* ====> kind of exported file */
  boolean opt_filedot;
  boolean opt_alwaysCode;

  boolean opt_exportFile; /* true iff one of the following export is true */
  boolean opt_filez3z;
  boolean opt_tracez3z;
  boolean opt_filesmt;
  boolean opt_tracesmt;
  boolean opt_filelustre;
  boolean opt_tracelustre;
  boolean opt_c_modubody, opt_c_modumain, opt_c_moduio;
  boolean opt_tracec;
  boolean opt_cpp_modubody, opt_cpp_modumain, opt_cpp_moduio;
  boolean opt_tracecpp;
  boolean opt_java_modubody, opt_java_modumain, opt_java_moduio, opt_java;
  boolean opt_tracejava;
  boolean opt_threads;
  boolean opt_filesyndex;
  boolean opt_tracesyndex;
  boolean opt_export_profiling;
  boolean opt_vhdl;
  boolean opt_verilog;

/* ====> listing options */
  boolean opt_stat;
  boolean opt_debug; /* (as WAR) detailed debug information in TRA */

/* ====> wizzard option */
  boolean opt_test_function; /* Loic */
  char * testName;  /* Loic */ /// -test=string

  boolean opt_synoptic; ///< synoptic representation of the signal program.


};


#endif // __OptionManager_H__


