/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Notify incompatibility of previousOption and newOption; newOption is ignored
   assign pk_true to this_erroredParsing 
*/
  void
BatchCompiler::notifyIncompatibleOptions (const char *previousOption, const char *newOption)
{
  pK_messenger->messageStringString (infoIncompatibleOptions, previousOption, newOption);
  this_erroredParsing = pk_true;
}

/**
   Notify incompatibility of previous export option and newOption; newOption is ignored
   assign pk_true to this_erroredParsing 
*/
void
BatchCompiler::notifyIncompatibleExportOptions (const char *newOption)
{
  if (OM->req_export_lustre_opt () && (!(eqStrings (CODELUSTRE, newOption))))
    notifyIncompatibleOptions (CODELUSTRE, newOption);	/* assign pk_true to this_erroredParsing */
  else if (OM->req_export_c_modules_opt () && (!(eqStrings (CODEC, newOption))))
    notifyIncompatibleOptions (CODEC, newOption);	/* assign pk_true to this_erroredParsing */
  else if (OM->req_export_cpp_modules_opt () && (!(eqStrings (CODECPP, newOption))))
    notifyIncompatibleOptions (CODECPP, newOption);	/* assign pk_true to this_erroredParsing */
  else if (OM->req_export_java_modules_opt () && (!(eqStrings (CODEJAVA, newOption))))
    notifyIncompatibleOptions (CODEJAVA, newOption);	/* assign pk_true to this_erroredParsing */
  else if (OM->req_export_syndex_opt () && (!(eqStrings (CODESYNDEX, newOption))))
    notifyIncompatibleOptions (CODESYNDEX, newOption);	/* assign pk_true to this_erroredParsing */
  else if (OM->req_export_profiling_opt () && (!(eqStrings (PROFILING, newOption))))
    notifyIncompatibleOptions (PROFILING, newOption);	/* assign pk_true to this_erroredParsing */
}

/** return a new allocated str copy of STRING in strRes for an option defined by -OPTION_NAME=STRING
   @return true when strRes is defined, false otherwise.
   error: assign pk_true to this_erroredParsing 
*/
boolean
BatchCompiler::getOptionStringValue (char **strRes)
{
  (*strRes) = allocate_string (strlen (this_commandArgument) + 1);
  if ((*strRes) == (char *) 0)
    {
      printf ("ERROR: Internal error when allocating string space in BatchCompiler::getOptionStringValue\n");
      this_erroredParsing = pk_true;
      return pk_false;
    }
  strcpy (*strRes, this_commandArgument);
  return pk_true;
}

/** return int value of INTEGER in intRes for an option defined by -OPTION_NAME=INTEGER
   @return true when intRes is defined, false otherwise.
   error: assign pk_true to this_erroredParsing 
*/
boolean
BatchCompiler::parse_integerValue (int *intRes, const char *commandTag)
{
  (*intRes) = -1;
  sscanf (this_commandArgument, "%d", intRes);
  if ((*intRes) < 0)
    {
      pK_messenger->errorString (commandArgument_VALUE_ERROR, commandTag);
      this_erroredParsing = pk_true;
      return pk_false;
    }
  return pk_true;
}

/** Analysis of print option; 
   @return true if analysis is correct, false otherwise.
   error: assign pk_true to this_erroredParsing 
*/
void
BatchCompiler::parse_traOptions ()
{
  if ((this_commandArgument[0] == 'w') || (this_commandArgument[0] == 'W'))
    {
      OM->set_gen_whenFormat_subopt (pk_true);
      if (this_commandArgument[1] == '\0')
	return;
      if ((this_commandArgument[1] == 'f') || (this_commandArgument[1] == 'F'))
	{
	  OM->set_gen_flatFormat_subopt (pk_true);
	  if (this_commandArgument[2] == '\0')
	    return;
	}
    }
  else if ((this_commandArgument[0] == 'f') || (this_commandArgument[0] == 'F'))
    {
      OM->set_gen_flatFormat_subopt (pk_true);
      if (this_commandArgument[1] == '\0')
	return;
      if ((this_commandArgument[1] == 'w') || (this_commandArgument[1] == 'W'))
	{
	  OM->set_gen_whenFormat_subopt (pk_true);
	  if (this_commandArgument[2] == '\0')
	    return;
	}
    }
  pK_messenger->errorString (ERR_NOTAVAILABLE_OPTION, "trace option");
  this_erroredParsing = pk_true;
}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_LustreOptions (boolean trace)
{
  if (OM->req_export_opt ())
    notifyIncompatibleExportOptions (CODELUSTRE);	/* error: assign pk_true to this_erroredParsing */
  OM->setON_export_lustre_opt (trace);
}

/**
   Analysis of c,c++,java suboptions
   @return -1: error; this_erroredParsing is assigned true
   @return 0: only body generation is required
   @return 1: body and io code generation are required
   @return 2: body and main code generation are required
   @return 3: body, io and main code generation are required
 */
integer
BatchCompiler::get_export_Options ()
{
  if (this_commandArgument[0] == '\0')
    return 0;
  if ((this_commandArgument[0] == 'm') || (this_commandArgument[0] == 'M'))
    {
      if (this_commandArgument[1] == '\0')
	return 2;
      if ((this_commandArgument[1] == 'i') || (this_commandArgument[1] == 'I'))
	if (this_commandArgument[2] == '\0')
	  return 3;
    }
  else if ((this_commandArgument[0] == 'i') || (this_commandArgument[0] == 'I'))
    {
      if (this_commandArgument[1] == '\0')
	return 1;
      if ((this_commandArgument[1] == 'm') || (this_commandArgument[1] == 'M'))
	{
	  if (this_commandArgument[2] == '\0')
	    return 3;
	}
    }
  pK_messenger->errorString (commandArgument_VALUE_ERROR, "export code option");
  this_erroredParsing = pk_true;
  return -1;
}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_CModulesOptions (boolean trace)
{
  if (OM->req_export_opt ())
    notifyIncompatibleExportOptions (CODEC);	/* error: assign pk_true to this_erroredParsing */
  OM->set_export_c_modules_opt  (pk_true, trace);
}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_C_Options (boolean trace)
{
  int code;
  if (OM->req_export_opt ())
    {
      notifyIncompatibleExportOptions (CODEC);	/* error: assign pk_true to this_erroredParsing */
      return;
    }
  OM->set_export_c_modules_opt (pk_false, trace);
  code = get_export_Options ();
  if (code < 0)			/* error: assign pk_true to this_erroredParsing */
    return;
  OM->set_export_c_body_opt (pk_true);
  if (code == 0)
    return;
  if (code == 1)
    OM->set_export_c_io_opt (pk_true);
  else if (code == 2)
    OM->set_export_c_main_opt (pk_true);
  else
    {
      OM->set_export_c_io_opt (pk_true);
      OM->set_export_c_main_opt (pk_true);
    }
  return;
}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_CPPModulesOptions (boolean trace)
{
  if (OM->req_export_opt ())
    notifyIncompatibleExportOptions (CODECPP);	/* error: assign pk_true to this_erroredParsing */
  OM->set_export_cpp_modules_opt  (pk_true, trace);

}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_CPP_Options (boolean trace)
{
  int code;
  if (OM->req_export_opt ())
    {
      notifyIncompatibleExportOptions (CODECPP);	/* error: assign pk_true to this_erroredParsing */
      return;
    }
  OM->set_export_cpp_modules_opt (pk_false, trace);
  code = get_export_Options ();
  if (code < 0)			/* error: assign pk_true to this_erroredParsing */
    return;
  OM->set_export_cpp_body_opt (pk_true);
  if (code == 0)
    return;
  if (code == 1)
    OM->set_export_cpp_io_opt (pk_true);
  else if (code == 2)
    OM->set_export_cpp_main_opt (pk_true);
  else
    {
      OM->set_export_cpp_io_opt (pk_true);
      OM->set_export_cpp_main_opt (pk_true);
    }
  return;
}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_JAVAModulesOptions (boolean trace)
{
  if (OM->req_export_opt ())
    notifyIncompatibleExportOptions (CODEJAVA);	/* error: assign pk_true to this_erroredParsing */
  OM->set_export_java_modules_opt  (pk_true, trace);
}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_JAVA_Options (boolean trace)
{
  int code;
  if (OM->req_export_opt ())
    {
      notifyIncompatibleExportOptions (CODEJAVA);	/* error: assign pk_true to this_erroredParsing */
      return;
    }
  OM->set_export_java_modules_opt (pk_false, trace);
  code = get_export_Options ();
  if (code < 0)			/* error: assign pk_true to this_erroredParsing */
    return;
  OM->set_export_java_body_opt (pk_true);
  if (code == 0)
    return;
  if (code == 1)
    OM->set_export_java_io_opt (pk_true);
  else if (code == 2)
    OM->set_export_java_main_opt (pk_true);
  else
    {
      OM->set_export_java_io_opt (pk_true);
      OM->set_export_java_main_opt (pk_true);
    }
  return;
}

/**
   error: assign pk_true to this_erroredParsing 
 */
void
BatchCompiler::set_SyndexOptions (boolean isValued, boolean trace)
{
  if (isValued)
    {
      printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
      this_erroredParsing = pk_true;
    }
  else
    {
      if (OM->req_single2multipleDefinition_opt ())
	notifyIncompatibleOptions (PARTIALDEF_MERGING, CODESYNDEX);	/* assign pk_true to this_erroredParsing */
      if (OM->req_export_opt ())
	notifyIncompatibleExportOptions (CODESYNDEX);	/* error: assign pk_true to this_erroredParsing */
      OM->setON_export_syndex_opt (trace);
    }
}

/** Analysis of the depth of the clock resolution option;
   @return true when analysis is correct, false otherwise.
   error: assign pk_true to this_erroredParsing 
*/
void
BatchCompiler::depthClockCalculusOptionContinuation ()
{
  int prof;

  if (parse_integerValue (&prof, "clock resolution tuning"))
    OM->set_solvingDepth_opt (prof);	/* error: assign pk_true to this_erroredParsing */
}

/** Analysis of exclusive clock option
   error: assign pk_true to this_erroredParsing 
*/
void
BatchCompiler::exclusiveClocksOptionContinuation ()
{
  OM->set_gen_exclusiveClock_opt (pk_false, -1);
  if (this_commandArgument[0] == '\0')
    OM->set_gen_exclusiveClock_opt  (pk_true, 0);
  else if (this_commandArgument[0] == '0')
    {				/* 0 X Y */
      OM->set_gen_exclusiveClock_opt  (pk_true, 0);
      if (this_commandArgument[1] == '\0')
	return;
      if (this_commandArgument[1] == '1')
	{			/* 0 1 Y */
	  OM->set_gen_exclusiveClock_opt  (pk_true, 1);	// Signals + clocks referenced in expressions.
	  if (this_commandArgument[2] == '\0')
	    return;
	  if (this_commandArgument[2] == '2')
	    {			/* 0 1 2 */
	      OM->set_gen_exclusiveClock_opt  (pk_true, 2);	// all the clocks of the system.
	      if (this_commandArgument[2] == '\0')
		return;
	    }
	}
      else if (this_commandArgument[1] == '2')
	{			/* 0 2 Y */
	  OM->set_gen_exclusiveClock_opt  (pk_true, 2);
	  {			// all the clocks of the system.
	    if (this_commandArgument[2] == '\0')
	      return;
	  }
	  if (this_commandArgument[2] == '1')
	    {			/* 0 2 1 */
	      OM->set_gen_exclusiveClock_opt  (pk_true, 1);	// Signals + clocks referenced in expressions.
	      if (this_commandArgument[2] == '\0')
		return;
	    }
	}
    }
  else if (this_commandArgument[0] == '1')
    {				/* 1 X Y */
      OM->set_gen_exclusiveClock_opt  (pk_true, 0);
      if (this_commandArgument[1] == '\0')
	return;
      if (this_commandArgument[1] == '0')
	{			/* 1 0 Y */
	  OM->set_gen_exclusiveClock_opt  (pk_true, 0);	// Signals + clocks referenced in expressions.
	  if (this_commandArgument[2] == '\0')
	    return;
	  if (this_commandArgument[2] == '2')
	    {			/* 1 0 2 */
	      OM->set_gen_exclusiveClock_opt  (pk_true, 2);	// all the clocks of the system.
	      if (this_commandArgument[2] == '\0')
		return;
	    }
	}
      else if (this_commandArgument[1] == '2')
	{			/* 1 2 Y */
	  OM->set_gen_exclusiveClock_opt  (pk_true, 2);
	  {			// all the clocks of the system.
	    if (this_commandArgument[2] == '\0')
	      return;
	  }
	  if (this_commandArgument[2] == '0')
	    {			/* 1 2 0 */
	      OM->set_gen_exclusiveClock_opt  (pk_true, 0);	// Signals + clocks referenced in expressions.
	      if (this_commandArgument[2] == '\0')
		return;
	    }
	}
    }
  else if (this_commandArgument[0] == '2')
    {				/* 2 X Y */
      OM->set_gen_exclusiveClock_opt  (pk_true, 0);
      if (this_commandArgument[1] == '\0')
	return;
      if (this_commandArgument[1] == '0')
	{			/* 2 0 Y */
	  OM->set_gen_exclusiveClock_opt  (pk_true, 0);	// Signals + clocks referenced in expressions.
	  if (this_commandArgument[2] == '\0')
	    return;
	  if (this_commandArgument[2] == '1')
	    {			/* 2 0 1 */
	      OM->set_gen_exclusiveClock_opt  (pk_true, 1);	// all the clocks of the system.
	      if (this_commandArgument[2] == '\0')
		return;
	    }
	}
      else if (this_commandArgument[1] == '1')
	{			/* 2 1 Y */
	  OM->set_gen_exclusiveClock_opt  (pk_true, 1);
	  {			// all the clocks of the system.
	    if (this_commandArgument[2] == '\0')
	      return;
	  }
	  if (this_commandArgument[2] == '0')
	    {			/* 2 1 0 */
	      OM->set_gen_exclusiveClock_opt  (pk_true, 0);	// Signals + clocks referenced in expressions.
	      if (this_commandArgument[2] == '\0')
		return;
	    }
	}
    }
  pK_messenger->errorString (ERR_NOTAVAILABLE_OPTION, "exclusive clock option");
  this_erroredParsing = pk_true;
}

/** Analysis of condition rewriting
   error: assign pk_true to this_erroredParsing 
*/
void
BatchCompiler::conditionRewritingOptionContinuation ()
{
  if ((this_commandArgument[0] == 'b') || (this_commandArgument[0] == 'B'))
    {
      OM->setON_run_ctrl2event_opt ();
      if (this_commandArgument[1] == '\0')
	return;
      if ((this_commandArgument[1] == 'c') || (this_commandArgument[1] == 'C'))
	{
	  OM->setON_crewDefinition_opt ();
	  if (this_commandArgument[1] == '\0')
	    return;
	}
    }
  else if ((this_commandArgument[0] == 'c') || (this_commandArgument[0] == 'C'))
    {
      OM->setON_crewDefinition_opt ();
      if (this_commandArgument[1] == '\0')
	return;
      if ((this_commandArgument[1] == 'b') || (this_commandArgument[1] == 'B'))
	{
	  OM->setON_run_ctrl2event_opt ();
	  if (this_commandArgument[1] == '\0')
	    return;
	}
    }
  pK_messenger->errorString (ERR_NOTAVAILABLE_OPTION, "exclusive clock option");
  this_erroredParsing = pk_true;
}

/** Analysis of output file option completion;
    uppercase id pk_true for a -D option completion, pk_false for a -d option completion
    @return true when an output file option has been correctly analysed
    this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
    */
void
BatchCompiler::outputFileOptionCompletion (boolean uppercase)
{
  char *dirName;

  if (getOptionStringValue (&dirName))	/* error: assign pk_true to this_erroredParsing */
    if (OM->is_outputDirectoryDefined ())
      {
	pK_messenger->errorString (commandArgument_ATMOSTONE_ERROR, "output directory");
	this_erroredParsing = pk_true;
      }
  OM->set_outputDirectory (dirName, uppercase);
}

/* ========================= OPTIONS 
 */
/** Analysis of input file options; this_commandOption is the string to be analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
void
BatchCompiler::inputFileName (const char *fileNameOption)
{
  if (OM->is_sourceFile_defined ())
    {
      pK_messenger->errorString (commandArgument_ATMOSTONE_ERROR, "input file");
      this_erroredParsing = pk_true;
    }
  else
    {
      char *fileName;
      fileName = allocate_string (strlen (fileNameOption) + 1);
      if (fileName == (char *) 0)
	{
	  printf ("ERROR: Internal error when allocating string space in BatchCompiler::inputFileName\n");
	  this_erroredParsing = pk_true;
	  return;
	}
      strcpy (fileName, fileNameOption);
      OM->setSourceFile (fileName);
    }
}

/** Analysis of input file options; this_commandOption is the string to be analyzed
   @return true when an input file option has been analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::inputFileOptions (boolean isValued)
{
  char *argValue;
  if (eqStrings (INISF, this_lwCommandOption))
    {				// -isf=<filename>
      if (isValued)
	{
	  OM->set_inIsf_opt (pk_true);
	  inputFileName (this_commandArgument);
	}
      else
	{
	  printf ("ERROR: missing expected file argument in -%s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
    }
  else if (eqStrings (COMMANDFILE, this_lwCommandOption))
    {				// -com=<filename>
      if (isValued)
	{
	  if (getOptionStringValue (&argValue))
	    {
	      OM->setCommandFile (argValue);
	      this_erroredParsing = pk_true;
	      pK_messenger->errorString (ERR_NOTAVAILABLE_OPTION, this_commandOption);
	    }
	}
      else
	{
	  printf ("ERROR: missing expected file argument in -com option\n");
	  this_erroredParsing = pk_true;
	}
    }
  /* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (PARAMREAD, this_lwCommandOption))
    {				// -par{=<filename>}
      if (isValued)
	{
	  if (getOptionStringValue (&argValue))
	    {			/* error: assign pk_true to this_erroredParsing */
	      if (OM->is_parameterFileDefined ())
		{
		  pK_messenger->errorString (commandArgument_ATMOSTONE_ERROR, "parameter file");
		  this_erroredParsing = pk_true;
		}
	      else
		OM->set_parameterFile (argValue);
	    }
	}
      else
	OM->set_readParameters ();
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of output file options;  this_commandOption is the string to be analyzed
   @return true when an output file option has been  analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::outputFileOptions (boolean isValued)
{
  if (eqStrings (OUTISF, this_lwCommandOption))
    {				// -com=<filename>
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->set_outIsf_opt (pk_true);
    }
  else if (eqStrings (OUTPUTROOTUC, this_commandOption))
    {				// -D=<directory> file are generated in <directory>
      if (isValued)
	outputFileOptionCompletion (pk_true);	/* error: assign pk_true to this_erroredParsing */
      else
	{
	  printf ("ERROR: missing expected argument in -%s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
    }
  else if (eqStrings (OUTPUTROOT, this_lwCommandOption))
    {				// -d=<directory>, file are generated in the subdirectory FOO of <directory>
      if (isValued)
	outputFileOptionCompletion (pk_false);	/* error: assign pk_true to this_erroredParsing */
      else
	{
	  printf ("ERROR: missing expected argument in -%s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
    }
  else if (eqStrings (WARNINGS, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_warn_opt ();
    }
  else if (eqStrings (VERBOSE, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_verbose_opt ();
    }
  else if (eqStrings (COSIMULATIONMODE, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_gen_cosimuTable_opt ();
    }
 else if (eqStrings (KEEPTRACEREW, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->set_keeprew_opt(pk_true);
    }
 else if (eqStrings (deprecated_PRINTOPTIONS, this_lwCommandOption))
    {				// -print= how to print when
      if (isValued)
	{
	  printf ("WARNING -%s=%s is deprecated: use -tra=%s\n", this_commandOption, this_commandArgument, this_commandArgument);
	  parse_traOptions ();
	}			/* error: assign pk_true to this_erroredParsing */
      else
	{
	  printf ("ERROR: missing expected argument in -%s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of class transformation options; this_commandOption is the string to be analyzed
   @return true when a partition option has been analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::graphClassTransformationOptions (boolean isValued)
{
  if (eqStrings (MAKEGENBASIC, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_basic_opt (pk_true);
    }
  else if (eqStrings (MAKEBASIC, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_basic_opt (pk_false);
    }
  else if (eqStrings (MAKEGENPOLY, this_commandOption))
    {
      if (isValued)
	{
	  depthClockCalculusOptionContinuation ();
	  OM->setON_make_poly_opt (pk_true);
	}
      else
	OM->setON_make_poly_opt (pk_true);
    }
  else if (eqStrings (MAKEPOLY, this_lwCommandOption))
    {
      if (isValued)
	{
	  depthClockCalculusOptionContinuation ();
	  OM->setON_make_poly_opt (pk_false);
	}
      else
	OM->setON_make_poly_opt (pk_false);
    }
  else if (eqStrings (deprecated_PROF, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("WARNING -%s=n is deprecated: use -poly=n\n", this_commandOption);
	  depthClockCalculusOptionContinuation ();	/* error: assign pk_true to this_erroredParsing */
	}
      else
	{
	  printf ("ERROR missing required argument of %s\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
    }
  else if (eqStrings (MAKEGENENDO, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_endo_opt (pk_true);
    }
  else if (eqStrings (MAKEENDO, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_endo_opt (pk_false);
    }
  else if (eqStrings (MAKEGENBOOL, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_bool_opt (pk_true);
    }
  else if (eqStrings (MAKEBOOL, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_bool_opt (pk_false);
    }
  else if (eqStrings (deprecated_MAKEBOOL, this_lwCommandOption))
    {
      printf ("WARNING -%s is deprecated: use -bool\n", this_commandOption);
      OM->setON_make_bool_opt (pk_false);
      return pk_true;
    }
  else if (eqStrings (MAKEGENSEQUENTIAL, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_sequential_opt (pk_true);
    }
  else if (eqStrings (MAKESEQUENTIAL, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_sequential_opt (pk_false);
    }
  else if (eqStrings (deprecated_MAKESEQUENTIAL, this_lwCommandOption))
    {
      printf ("WARNING -%s is deprecated: use -seq\n", this_commandOption);
      OM->setON_make_sequential_opt (pk_false);
    }
  else if (eqStrings (MAKEGENFLAT, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_flat_opt (pk_true);
    }
  else if (eqStrings (MAKEFLAT, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_flat_opt (pk_false);
    }
  else if (eqStrings (MAKEGENMONO, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_mono_opt (pk_true);
    }
  else if (eqStrings (MAKEMONO, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_make_mono_opt (pk_false);
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of kind of output file options; this_commandOption is the string to be analyzed
   @return true when a kind output file option has been  analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::kindOfOutputOptions (boolean isValued)
{
  if (eqStrings (LIST, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_Signal_list_opt ();
    }
  else if (eqStrings (NOTRACE, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setOFF_tra_opt ();
    }
  else if (eqStrings (TRACEUC, this_commandOption))
    {
      if (isValued)
	parse_traOptions ();	/* error: assign pk_true to this_erroredParsing */
      OM->setON_TRACE_opt ();
    }
  else if (eqStrings (TRACE, this_commandOption))
    {
      if (isValued)
	parse_traOptions ();	/* error: assign pk_true to this_erroredParsing */
      OM->setON_trace_opt ();
    }
  else if (eqStrings (SME, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_gen_sme_opt ();
    }
  else if (eqStrings (SSME, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_gen_ssme_opt ();
    }

  else if (eqStrings (GENEAUTO, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_gen_geneauto_opt ();
    }

  else if (eqStrings (ABSTRACTION, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_run_abstraction_opt ();
    }
  else if (eqStrings (EXCLUSIVECLOCK, this_lwCommandOption))
    {
      if (isValued)
	exclusiveClocksOptionContinuation ();	/* error: assign pk_true to this_erroredParsing */
      else
	OM->set_gen_exclusiveClock_opt  (pk_true, 0);
    }
  else if (eqStrings (MKMODULEHEADER, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_gen_header_opt ();
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of graph options; this_commandOption is the string to be analyzed
   @return true when a graph option has been analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::graphOptions (boolean isValued)
{
  if (eqStrings (DRETIMING, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_run_retiming_opt ();
    }
  /* LRETIMING            "-lr"  NOT IMPLEMENTED */
  else if (eqStrings (REWCOND, this_lwCommandOption))
    {
      if (isValued)
	conditionRewritingOptionContinuation ();
      else
	{
	  OM->setON_run_ctrl2event_opt ();
	  OM->setON_crewDefinition_opt ();
	}
    }
  else if (eqStrings (US, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_run_signalUnification_opt ();
    }
  else if (eqStrings (DEFAULT_SPLITTING, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	{
	  OM->set_single2multipleDefinition_opt (pk_true);
	  if (OM->req_multiple2singleDefinition_opt ())
	    notifyIncompatibleOptions (DEFAULT_SPLITTING, this_commandOption);
	}
    }				/* assign pk_true to this_erroredParsing */
  else if (eqStrings (PARTIALDEF_MERGING, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	{
	  OM->setON_multiple2singleDefinition_opt ();
	  if (OM->req_single2multipleDefinition_opt ())
	    notifyIncompatibleOptions (DEFAULT_SPLITTING, this_commandOption);
	}
    }				/* assign pk_true to this_erroredParsing */
  else if (eqStrings (SPLITDATA, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_splitDataStructure_opt ();
    }
  else if (eqStrings (ENUM2INT, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_enum2Integer_opt ();
    }
  else if (eqStrings (SUBSTITUTE, this_lwCommandOption))
    {
      if (isValued)
	{
	  int intRes;
	  if (parse_integerValue (&intRes, "substitution tuning"))
	    OM->set_run_signalSubstitution_opt (intRes);
	}
      else
	{
	  printf ("ERROR missing required argument of %s\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
    }
  else if (eqStrings (KERNEL, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_kernel_opt ();
    }
  else if (eqStrings (UTILITY_CLOCK_OPT, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_run_utilityClock_opt ();
    }
  else if (eqStrings (PROFILING, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	{
	  OM->setON_export_profiling_opt ();
	  if (OM->req_export_opt ())
	    notifyIncompatibleExportOptions (PROFILING);
	}
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (POLYENDO, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	{
	  OM->setON_polyendo_opt ();
	  OM->setON_force_subopt (); // polyendo => force
	  if (OM->req_run_tagBasedSeparation_opt())
	    notifyIncompatibleOptions (TAGBASEDPARTITION, this_commandOption);
	}
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of partition options; this_commandOption is the string to be analyzed
   @return true when a partition option has been   analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::partitionOptions (boolean isValued)
{
  if (eqStrings (SEP_STATES_OTHER, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_run_stateSeparation_opt ();
    }
  else if (eqStrings (SEP_BOOL_OTHER, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_run_booleanSeparation_opt ();
    }
  else if (eqStrings (SEP_FINITE_TYPES_OTHER, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_run_finiteTypeSeparation_opt ();
    }
  else if (eqStrings (CLUSTERING, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_clustering_opt ();
    }
  else if (eqStrings (TAGBASEDPARTITION, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	{
	  OM->setON_run_tagBasedSeparation_opt ();
	  if (OM->req_polyendo_opt ())
	    notifyIncompatibleOptions (POLYENDO, this_commandOption);
	}
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of kind of exported file options; this_commandOption is the string to be analyzed
   @return true when a kind exported file option has been  analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::kindOfExportedFileOptions (boolean isValued)
{
  if (eqStrings (Z3ZUC, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_export_z3z_opt (pk_true);
    }
  else if (eqStrings (Z3Z, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_export_z3z_opt (pk_false);
    }
  else
  if (eqStrings (SMTUC, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_export_smt_opt (pk_true);
    }
  else if (eqStrings (SMT, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_export_smt_opt (pk_false);
    }

  else if (eqStrings (CODELUSTREUC, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	set_LustreOptions (pk_true);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODELUSTRE, this_commandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	set_LustreOptions (pk_false);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODECUC, this_commandOption))
    {
      if (isValued)
	set_C_Options (pk_true);
      else
	set_CModulesOptions (pk_true);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODEC, this_commandOption))
    {
      if (isValued)
	set_C_Options (pk_false);
      else
	set_CModulesOptions (pk_false);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODETHREADS, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	{
	  OM->setON_clustering_opt ();
	  OM->setON_export_threads_subopt ();
	}
    }
  else if (eqStrings (CODECPPUC, this_commandOption))
    {
      if (isValued)
	set_CPP_Options (pk_true);
      else
	set_CPPModulesOptions (pk_true);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODECPP, this_lwCommandOption))
    {
      if (isValued)
	set_CPP_Options (pk_false);
      else
	set_CPPModulesOptions (pk_false);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODEJAVAUC, this_commandOption))
    {
      if (isValued)
	set_JAVA_Options (pk_true);
      else
	set_JAVAModulesOptions (pk_true);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODEJAVA, this_lwCommandOption))
    {
      if (isValued)
	set_JAVA_Options (pk_false);
      else
	set_JAVAModulesOptions (pk_false);
    }				/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODESYNDEXUC, this_commandOption))
    set_SyndexOptions (isValued, pk_true);	/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (CODESYNDEX, this_commandOption))
    set_SyndexOptions (isValued, pk_false);	/* error: assign pk_true to this_erroredParsing */
  else if (eqStrings (VHDLCODEUC, this_commandOption))
    OM->setON_export_vhdl_opt (pk_true); 
  else if (eqStrings (VHDLCODE, this_commandOption))
    OM->setON_export_vhdl_opt (pk_false);  
  else if (eqStrings (VERILOGCODEUC, this_commandOption))
    OM->setON_export_verilog_opt (pk_true);
  else if (eqStrings (VERILOGCODE, this_commandOption))
    OM->setON_export_verilog_opt (pk_false); 
  else if (eqStrings (DOTCODE, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_export_dot_opt ();
    }
  else if (eqStrings (OPTFORCE, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_force_subopt ();
    }
  else if (eqStrings (CHECK, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_export_check_subopt ();
    }
  else if (eqStrings (UNDELAYMULTIPLEXING, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->set_run_delayMultiplexing_opt (pk_false);
    }
  else if (eqStrings (UNDELAYOPTIMIZING, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->set_run_delayOptimizing_opt (pk_false);
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of listing options; this_commandOption is the string to be analyzed
   @return true when a listing option has been   analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::listingOptions (boolean isValued)
{
  if (eqStrings (STATISTICS, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_stat_opt ();
    }
  else if (eqStrings (DEBUG, this_lwCommandOption))
    {
      if (isValued)
	{
	  printf ("ERROR: unexpected argument in %s option\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      else
	OM->setON_debug_opt ();
    }
  else
    return pk_false;
  return pk_true;
}

/** Analysis of wizard test function options; this_commandOption is the string to be analyzed
   @return true when a listing option has been   analyzed
   this_erroredParsing is set to pk_true if an error occurred; it remains unchanged otherwise
*/
boolean
BatchCompiler::testFunctionOptions (boolean isValued)
{
  char *strRes;
  // -test=<string> 
  if (eqStrings (TESTFUNC, this_lwCommandOption))
    {
      if (isValued)
	{
	  if (getOptionStringValue (&strRes))
	    {			/* error: assign pk_true to this_erroredParsing */
	      OM->set_wiz_testFunctionality_opt (pk_true);
	      OM->setTestName (strRes);
	    }
	}
      else
	{
	  printf ("ERROR missing required argument of %s\n", this_commandOption);
	  this_erroredParsing = pk_true;
	}
      return pk_true;
    }
  return pk_false;
}

/* ========================= OPTIONS END
 */
/**
   strArray is an array of strings
   The private shared variable this_commandIndex denotes the position of the current argument in strArray
   @return -2 if the current element is an empty string
   @return -1 if there is no more argument in strArray
   @return 0 if the current element does not start with '-'
        return the string in private shared variable this_commandOption
   @return 1 if the current element is a not valued option
        return the option name in private shared variable this_commandOption
   @return 2 if the current element is a valued option
        return the option name in private shared variable this_commandOption
        return the (possibly empty) value argument in private shared variable this_commandArgument
   @return the position of the next argument in private shared variable this_commandIndex
 */
int
BatchCompiler::parseArgument (char *strArray[])
{
  char *auxString;
  if (this_commandIndex == this_strArrayLength)
    return -1;
  this_commandOption = strArray[this_commandIndex];
  this_commandIndex++;
  if (this_commandOption[0] == '\0')
    {
      printf ("ERROR: unexpected empty option\n");
      return -2;
    }
  if (this_commandOption[0] == '-')
    {
      if (this_commandOption[1] == '\0')
	{
	  printf ("ERROR: missing expected option name after '-'\n");
	  return -2;
	}
      if (this_commandOption[1] == '=')
	{
	  printf ("ERROR: missing expected option name between '-' and '='\n");
	  return -2;
	}
      this_commandOption++;
      this_commandArgument = strchr (this_commandOption, (int) '=');
      if (this_commandArgument == (char *) 0)
	{			/* '=' does not occur in current strArray element */
	  this_lwCommandOption = safeLowerString (this_commandOption);
	  if (this_commandIndex == this_strArrayLength)
	    return 1;
	  this_commandArgument = strArray[this_commandIndex];
	  if (this_commandArgument[0] != '=')
	    return 1;
	  this_commandIndex++;
	}
      else
	{			/* '=' occurs in this_commandOption */
	  this_commandArgument[0] = '\0';	/* remove '=' in the this_commandOption */
	  this_lwCommandOption = safeLowerString (this_commandOption);
	}
      /* The option is a valued option */
      this_commandArgument++;
      if (this_commandArgument[0] == 0)
	{
	  if (this_commandIndex == this_strArrayLength)
	    {
	      this_commandArgument[0] = '\0';	/* empty string value */
	      return 2;
	    }
	  auxString = strArray[this_commandIndex];
	  if (auxString[0] == '-')
	    {
	      this_commandArgument[0] = '\0';	/* empty string value */
	      return 2;
	    }
	  this_commandArgument = auxString;
	  this_commandIndex++;
	}
      return 2;
    }
  else				/* not a starting '-' character */
    return 0;
}

/**
  addArgument analyzes an option and its suboptions of the batch compiler. The option is in strArray[*i].
  The options are defined in polychronyOptions.h file.
  @return true if the analysis is not completed
*/
boolean
BatchCompiler::addArgument (char *strArray[])
{
  int optionType;

  optionType = parseArgument (strArray);
  if (optionType == -1)
    return pk_false;
  if (optionType == -2)
    this_erroredParsing = pk_true;
  else if (optionType == 0)
    {
      appendCommandLineSimpleOption (this_commandOption);
      inputFileName (this_commandOption);
    }
  else
    {
      boolean isValued;
      if (optionType == 1)
	{
	  isValued = pk_false;
	  appendCommandLineSimpleOption (this_commandOption);
	}
      else
	{
	  isValued = pk_true;
	  appendCommandLineValuedOption (this_commandOption, this_commandArgument);
	}
      if (inputFileOptions (isValued))
	return pk_true;
      if (outputFileOptions (isValued))
	return pk_true;
      if (graphClassTransformationOptions (isValued))
	return pk_true;
      if (kindOfOutputOptions (isValued))
	return pk_true;
      if (graphOptions (isValued))
	return pk_true;
      if (partitionOptions (isValued))
	return pk_true;
      if (kindOfExportedFileOptions (isValued))
	return pk_true;
      if (listingOptions (isValued))
	return pk_true;
      if (testFunctionOptions (isValued))
	return pk_true;
      pK_messenger->errorString (ERR_UNKNOWN_OPTION, this_commandOption);
      this_erroredParsing = pk_true;
    }
  return pk_true;
}

void
BatchCompiler::set_implicit_opt ()
{
  if (!(OM->req_genFile_opt () || OM->req_basicGraph ()))
    {
      if (OM->test_traceNone ())
	OM->setON_make_poly_opt (OM->test_traceAllowed ());
      else
	OM->setON_make_poly_opt (pk_true);
    }
}
