/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_

/**
   For Aoste team. Link Signal/microSignal. Dec 2010.
   Called by the -test=aoste option.
*/
  integer BatchCompiler::genMicroSignalAoste (ptDCG_poly polyGraph)
{
  ptINode finiteTypesSubGraph, othersSubGraph;
  ptDCG_basic g1;
  ntrees vmicroSignal;
  ptDCG_poly loc_gPoly;

  polyGraph->expressionExpand ();
  polyGraph->separateFiniteTypesOthers (&finiteTypesSubGraph, &othersSubGraph);
  g1 = polyGraph->cloneSubgraph2Graph (finiteTypesSubGraph, pk_true, pk_true);
  g1->setName (newIdent_stringSuffixed (polyGraph->getName (), "FiniteTypes"));
  loc_gPoly = new DCG_poly (g1);
  loc_gPoly->setIntrinsicClocks ();
  // Graph2Tree in Kernel mode and flat mode.
  vmicroSignal = produceFlat (loc_gPoly, pk_true);
  // Removing the pragmas
  STree_setPragmas (vmicroSignal, NILTREE);
  printSignal (vmicroSignal);
  return polychronyNoError;
}

/**
   For testing the phylum for a poly graph.
*/
void
BatchCompiler::polyPhylums (ptDCG_poly polyGraph, char *str)
{
  polyGraph->nonProductiveSignalElimination ();
  polyGraph->breakCPG ();	// Flatenning the subgraphs of the graph
  polyGraph->setClockDependences ();	// isOrderRelation(gBool, &bx, balwaysCode, pk_false);
  if (polyGraph->phylumClustering ())
    {
      polyGraph->nonProductiveSignalElimination ();
      // isOrderRelation(gBool, &bx, balwaysCode, pk_false); // verifying
      genCodeSignal (polyGraph, pk_true, "polyPhylum");
    }
}

/** Entry point for test of functionality over bool graphs
 @return true if some bool_wizard_testFunctionality has been selected */
boolean
BatchCompiler::bool_wizard_testFunctionality (ptDCG_bool gBool, integer * res)
{
  char *str = OM->getTestName ();
  if (eqStrings (str, "iodep"))
    {
      computeIOTransitiveClosure (gBool);
      genCodeSignal (gBool, pk_true, "ioclosure");
      (*res) = polychronyNoError;
      return pk_true;
    }

  return pk_false;
}

/** Entry point for test of functionality over endo graphs
 @return true if some endo_wizard_testFunctionality has been selected */
boolean
BatchCompiler::endo_wizard_testFunctionality (ptDCG_endo gEndo, integer * res)
{
  char *str = OM->getTestName ();
  if (eqStrings (str, "pseudoCycles"))
    {
      analyzeStaticCycles (gEndo);
      (*res) = polychronyNoError;
      return pk_true;
    }

  if (eqStrings (str, "iodep"))
    {
      computeIOTransitiveClosure (gEndo);
      genCodeSignal (gEndo, pk_true, "ioclosure");
      (*res) = polychronyNoError;
      return pk_true;
    }

  return pk_false;
}

/**
   Producing automata on state part.
*/
boolean
BatchCompiler::buildAutomatonOnStatePart (ptDCG_poly polyGraph)
{
  ptINode statesSubGraph, othersSubGraph;
  ptDCG_basic basicOtherGraph, basicStateGraph;
  ptDCG_poly polyOtherGraph, polyStateGraph;

  polyGraph->separateStatesOthers (&statesSubGraph, &othersSubGraph);
  basicStateGraph = polyGraph->cloneSubgraph2Graph (statesSubGraph, pk_true, pk_true);
  if (!DCGBasic2DCGPoly (basicStateGraph, OM->get_solvingDepth_opt (), &polyStateGraph))
    {
      genCodeSignal (polyStateGraph, pk_true, "erroredPoly");
      return pk_false;
    }
  genCodeSignal (polyStateGraph, pk_true, "StatePoly");

  basicOtherGraph = polyGraph->cloneSubgraph2Graph (othersSubGraph, pk_true, pk_true);
  if (!DCGBasic2DCGPoly (basicOtherGraph, OM->get_solvingDepth_opt (), &polyOtherGraph))
    {
      genCodeSignal (polyOtherGraph, pk_true, "erroredPoly");
      return pk_false;
    }
  genCodeSignal (polyOtherGraph, pk_true, "OtherPoly");


  polyStateGraph->buildAutomaton ();
  return pk_true;
}


/** Entry point for test of functionality over poly graphs
    @return true if some poly_wizard_testFunctionality has been selected
 */
boolean
BatchCompiler::poly_wizard_testFunctionality (ptDCG_poly polyGraph, integer * res)
{
  char *str = OM->getTestName ();
  printf ("DEBUG BatchCompiler::poly_wizard_testFunctionality, str is <<%s>>\n", str);
  fflush (stdout);
  
  if (eqStrings (str, "automaton"))
    {
      polyGraph->buildAutomaton ();
      return pk_true;
    }

  if (eqStrings (str, "automaton1"))
    {
      return buildAutomatonOnStatePart (polyGraph);
    }

  if (eqStrings (str, "aoste"))
    {
      genMicroSignalAoste (polyGraph);
      (*res) = polychronyNoError;
      return pk_true;
    }
  if (eqStrings (str, "pclu"))
    {
      polyPhylums (polyGraph, str);
      (*res) = polychronyNoError;
      return pk_true;
    }

  if (eqStrings (str, "fiacre"))
    {
      FiacrePrintExtraction (polyGraph);
      (*res) = polychronyNoError;
      return pk_true;
    }

  if (eqStrings (str, "iodep"))
    {
      computeIOTransitiveClosure (polyGraph);
      genCodeSignal (polyGraph, pk_true, "ioclosure");
      (*res) = polychronyNoError;
      return pk_true;
    }

 return pk_false;
}

/** Entry point for test of functionality over basic graphs
    @return true if some basic_wizard_testFunctionality has been selected */
boolean
BatchCompiler::basic_wizard_testFunctionality (ptDCG_basic gBasic, integer * res)
{
  char *str = OM->getTestName ();
  Testing_reducingCommunications_assClock = pk_false;
  Testing_reducingCommunications_b = pk_false;
  Testing_reducingCommunications_assClock = eqStrings (str, "rc_ac");
  Testing_reducingCommunications_b = eqStrings (str, "rc") || Testing_reducingCommunications_assClock;


  if (eqStrings (str, "Loic"))
    {
      // test used to prove that the availibility clock cannot be relaced by the utility clock for the solving. It may introduce cycles.
      // Example : The model S_PACE_MAKER defined in the $SignalLib_Graphical/CONTEXT.SIG library.
      gBasic->computeProductiveClock ();
      gBasic->rewUsingProductiveClock ();
      gBasic->setUtilityClocksAsClock ();
    }

  // return pk_false;

  if (eqStrings (str, "iodep"))
    {
      computeIOTransitiveClosure (gBasic);
      genCodeSignal (gBasic, pk_true, "ioclosure");
      SMTPrintExtraction (gBasic, pk_true, pk_true); 
      (*res) = polychronyNoError;
      return pk_true;
    }

  return pk_false;
}



/** Entry point for test of functionality over source file
 @return true if some syntax_wizard_testFunctionality has been selected */
boolean
BatchCompiler::syntax_wizard_testFunctionality (ntrees termAbstractTree, ntrees paramAbstractTree, integer * res)
{
  return pk_false;
}

/** Entry point for test of functionality over bool graphs
 @return true if some bool_wizard_testFunctionality has been selected */
boolean
BatchCompiler::flat_wizard_testFunctionality (ptDCG_flat gFlat, integer * res)
{
  // char *str = OM->getTestName ();
  return pk_false;
}


/**
   Certification.
   It generates SMT code for the certification of  the compiler.
*/
void BatchCompiler::certification( ptDCG_basic vgraph )
{
  // Parameters of the certification.
  boolean bwithIODep = pk_false; 
  boolean certification_DEBUG = pk_false;

  if (OM->req_export_certification()) 
    {
      GraphLevelType glevel = vgraph->getFormatLevel ();
      if (bwithIODep)
	{
	  computeIOTransitiveClosure (vgraph);
	  if ((glevel == DCG_POLY) || (glevel == DCG_ENDO))
	    {
	      clockCalculus((ptDCG_poly) vgraph);
	      clockCalculus((ptDCG_poly) vgraph);
	    }
	  if (certification_DEBUG)
	    cout << "  ===== DONE ===== computeIOTransitiveClosure ===== " << endl;
	}

      if (glevel == DCG_ENDO)
	vgraph->nonProductiveSignalElimination (pk_true);
      
       if (certification_DEBUG) 
	 genCodeSignal (vgraph, pk_true, "certification");

      SMTPrintExtraction (vgraph, pk_true, pk_true); 
    }
}
