/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor of an option manager object. The default values are set (setopt_init()).
*/
  OptionManager::OptionManager ()
{
  setopt_init ();
}

/**
   Destructor.
*/
OptionManager::~OptionManager ()
{
}

/* ====> input file options */
/** It sets the command file to f. Option not yet implemented.  */
void
OptionManager::setCommandFile (char *f)
{
  commandFile = f;
}

/** @return the name of the command file. Option not yet implemented.*/
char *
OptionManager::getCommandFile ()
{
  return commandFile;
}

/** It sets to f the source file  to compile. */
void
OptionManager::setSourceFile (char *f)
{
  srcFileDefined = pk_true;
  srcFile = f;
}

/** It sets the name of the directory to f. It is used to generate code produced by the compiler. */
void
OptionManager::set_inIsf_opt (boolean b)
{
  in_isf = b;
}

boolean OptionManager::is_inIsf ()
{
  return in_isf;
}

/** It sets to f the source file  to compile. */
boolean OptionManager::is_sourceFile_defined ()
{
  return srcFileDefined;
}

/** @return the name of source file to compile, previously set by setSourceFile().  */
char *
OptionManager::getSourceFile ()
{
  return srcFile;
}

/** parameters have to be read but no file parameter is provided */
void
OptionManager::set_readParameters ()
{
  readParamFile = pk_true;
  paramFileDefined = pk_false;
}

/** It sets the parameter file (f) to compile */
void
OptionManager::set_parameterFile (char *f)
{
  paramFile = f;
  readParamFile = pk_true;
  paramFileDefined = pk_true;
}

/** return pk_true if parameter file with standard naming rule should be read. */
boolean OptionManager::req_readParamFile ()
{
  return readParamFile;
}

/** return pk_true if the parameter file is defined. */
boolean OptionManager::is_parameterFileDefined ()
{
  return paramFileDefined;
}

/** @return the name of parameter file to compile, previously set by set_parameterFile().  */
char *
OptionManager::get_parameterFile ()
{
  return paramFile;
}



/* ====> output file options */
/** It sets the name of the directory to f. It is used to generate code produced by the compiler. */
void
OptionManager::set_outIsf_opt (boolean b)
{
  out_isf = b;
}

boolean OptionManager::req_outIsf ()
{
  return out_isf;
}

/** It sets the name of the directory to f. It is used to generate code produced by the compiler. */
void
OptionManager::set_outputDirectory (char *f, boolean targetDir)
{
  outputDirDefined = pk_true;
  outputDir = f;
  targetDirDefined = targetDir;
}

/** return pk_true if the output directory is defined. */
boolean OptionManager::is_outputDirectoryDefined ()
{
  return outputDirDefined;
}

/** return pk_true if the output directory is defined. */
boolean OptionManager::is_targetDirectoryDefined ()
{
  return targetDirDefined;
}

/** @return the name of output directory, previously set by set_OutputDirectory(), used by the compiler to generate files. */
char *
OptionManager::get_outputDirectory ()
{
  return outputDir;
}

/**
   It enables/disables(default) the warning message when b/not b in output file FOO_LIS.SIG
*/
void
OptionManager::setON_warn_opt ()
{
  opt_warn = pk_true;
  pK_messenger->setOptionMessenger (2, pk_true);
}

/**
   @return "true" if the warning are visible, "false" otherwise.
*/
boolean OptionManager::req_warn_opt ()
{
  return opt_warn;
}

/**
   It enables/disables(default) the verbose mode when b/not b in file DIR/FOO.log
*/
void
OptionManager::setON_verbose_opt ()
{
  opt_verbose = pk_true;
  pK_messenger->setOptionMessenger (3, pk_true);
  setON_stat_opt ();
}

/**
   @return "true" if the verbose mode is enable, "false" otherwise.
*/
boolean OptionManager::req_verbose_opt ()
{
  return opt_verbose;
}

/**
   Enable the generation ofa  table for cosimulation mode.
   In cosimulation mode, a file FOO_ASSO.txt contains a dictionnary associating generated signals to source code
   TODO
*/
void
OptionManager::setON_gen_cosimuTable_opt ()
{
  opt_cosimulation = pk_true;
}

/** switch cosimulation  requirement
*/
void
OptionManager::switch_gen_cosimuTable_opt ()
{
  opt_cosimulation = !opt_cosimulation;
}


/**
   @return the current mode for the cosimulation.
*/
boolean OptionManager::req_gen_cosimuTable_opt ()
{
  return opt_cosimulation;
}

/**
   @return the current mode for the keeping traces of rewritings.
*/
boolean OptionManager::req_keeprew_opt ()
{
  return opt_keeprew;
}

/**
   Enable (when b)/Disable (when not b) the current mode for the keeping traces of rewritings.
*/
void
OptionManager::set_keeprew_opt (boolean b)
{
  opt_keeprew = b;
}


/**
   Trace option -tra=w
   It enables/disables(default) the printing of the unary when operator (true : it prints "when", otherwise [:]).
*/
void
OptionManager::set_gen_whenFormat_subopt (boolean whenOp)
{
  opt_trace_samplingAsWhen = whenOp;
}

/**
   Trace option
   @return "true" if the the printing of the printing of the unary when operator is "when", "false" otherwise.
*/
boolean OptionManager::req_gen_whenFormat_subopt ()
{
  return opt_trace_samplingAsWhen;
}

/**
   Trace option (-tra=f)
   It disables the printing of structured code
*/
void
OptionManager::set_gen_flatFormat_subopt (boolean isFlat)
{
  opt_trace_unstructuredProcess = isFlat;
}

/**
   Trace option
   @return "true" if the the printing of the printing of the unary when operator is "when", "false" otherwise.
*/
boolean OptionManager::req_gen_flatFormat_subopt ()
{
  return opt_trace_unstructuredProcess;
}

/* ====> graph class transformation options */
/**
   Require to build the basic Signal graph
   Printing the result of the transformation is required iff b is pk_true
*/
void
OptionManager::setON_make_basic_opt (boolean b)
{
  opt_basicGraph = pk_true;
  opt_tracebasic = opt_tracebasic || b;
  if (opt_tracebasic)
    opt_traceNone = pk_false;
}

/**
   Require to unparse the basic Signal graph if it is built
*/
void
OptionManager::setON_unparse_basic_opt ()
{
  opt_tracebasic = pk_true;
  opt_traceNone = pk_false;
}

/**
   @return "true" if  building the basic Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_basicGraph ()
{
  return opt_basicGraph;
}

/**
   @return "true" if printing the basic Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_trace_basic ()
{
  return opt_traceAllowed && opt_tracebasic;
}

/**
   Require to build the poly Signal graph
   Printing the result of the transformation is required if b is pk_true
*/
void
OptionManager::setON_make_poly_opt (boolean b)
{
  opt_polyGraph = pk_true;
  opt_tracepoly = opt_tracepoly || b;
  if (opt_tracepoly)
    opt_traceNone = pk_false;
  setON_make_basic_opt (pk_false);
}

/**
   Require to unparse the poly Signal graph if it is built
*/
void
OptionManager::setON_unparse_poly_opt ()
{
  opt_tracepoly = pk_true;
  opt_traceNone = pk_false;
}

/**
   @return "true" if  building the poly Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_polyGraph ()
{
  return opt_polyGraph;
}

/**
   @return "true" if printing the poly Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_trace_poly ()
{
  return opt_traceAllowed && opt_tracepoly;
}

/**
   Require to build the endo Signal graph
   Printing the result of the transformation is required if b is pk_true
*/
void
OptionManager::setON_make_endo_opt (boolean b)
{
  opt_endoGraph = pk_true;
  opt_traceendo = opt_traceendo || b;
  if (opt_traceendo)
    opt_traceNone = pk_false;
  setON_make_poly_opt (pk_false);
}

/**
   Require to unparse the endo Signal graph if it is built
*/
void
OptionManager::setON_unparse_endo_opt ()
{
  opt_traceendo = pk_true;
  opt_traceNone = pk_false;
}

/**
   @return "true" if  building the endo Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_endoGraph ()
{
  return opt_endoGraph;
}

/**
   @return "true" if printing the endo Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_trace_endo ()
{
  return opt_traceAllowed && opt_traceendo;
}

/**
   Require to build the bool Signal graph
   Printing the result of the transformation is required if b is pk_true
*/
void
OptionManager::setON_make_bool_opt (boolean b)
{
  opt_boolGraph = pk_true;
  opt_tracebool = opt_tracebool || b;
  if (opt_tracebool)
    opt_traceNone = pk_false;
  setON_make_endo_opt (pk_false);
}

/**
   Switch requirement to build the bool Signal graph
   Printing becomes equal to the new requirement
*/
void
OptionManager::switch_make_bool_opt ()
{
  opt_boolGraph = !opt_boolGraph;
  opt_tracebool = opt_boolGraph;
}

/**
   Require to unparse the bool Signal graph if it is built
*/
void
OptionManager::setON_unparse_bool_opt ()
{
  opt_tracebool = pk_true;
  opt_traceNone = pk_false;
}

/**
   @return "true" if  building the bool Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_boolGraph ()
{
  return opt_boolGraph;
}

/**
   @return "true" if printing the bool Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_trace_bool ()
{
  return opt_traceAllowed && opt_tracebool;
}

/**
   Require to build the sequential Signal graph
   Printing the result of the transformation is required if b is pk_true
*/
void
OptionManager::setON_make_sequential_opt (boolean b)
{
  opt_sbdcGraph = pk_true;
  opt_tracesbdc = opt_tracesbdc || b;
  if (opt_tracesbdc)
    opt_traceNone = pk_false;
  setON_make_bool_opt (pk_false);
}

/**
   Switch requirement to build the sequential Signal graph
   Printing becomes equal to the new requirement
*/
void
OptionManager::switch_make_sequential_opt ()
{
  opt_sbdcGraph = !opt_sbdcGraph;
  opt_tracesbdc = opt_sbdcGraph;
  if (opt_tracesbdc)
    opt_traceNone = pk_false;
}

/**
   @return "true" if building the sequential Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_sequentialGraph ()
{
  return opt_sbdcGraph;
}

/**
   @return "true" if printing the sequential Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_trace_sequential ()
{
  return opt_traceAllowed && opt_tracesbdc;
}

/**
   Require to build the flat Signal graph
   Printing the result of the transformation is required if b is pk_true
*/
void
OptionManager::setON_make_flat_opt (boolean b)
{
  opt_flatGraph = pk_true;
  opt_traceflat = opt_traceflat || b;
  if (opt_traceflat)
    opt_traceNone = pk_false;
  setON_make_bool_opt (pk_false);
}

/**
   Switch requirement to build the flat Signal graph
   Printing becomes equal to the new requirement
*/
void
OptionManager::switch_make_flat_opt ()
{
  opt_flatGraph = !opt_flatGraph;
  opt_traceflat = opt_flatGraph;
  if (opt_traceflat)
    opt_traceNone = pk_false;
}

/**
   Require to unparse the flat Signal graph if it is built
*/
void
OptionManager::setON_unparse_flat_opt ()
{
  opt_traceflat = pk_true;
  opt_traceNone = pk_false;
}

/**
   @return "true" if building the flat Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_flatGraph ()
{
  return opt_flatGraph;
}

/**
   @return "true" if printing the flat Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_trace_flat ()
{
  return opt_traceAllowed && opt_traceflat;
}


/**
   Require to build the mono clocked Signal graph
   Printing the result of the transformation is required if b is pk_true
*/
void
OptionManager::setON_make_mono_opt (boolean b)
{
  opt_monoGraph = pk_true;
  opt_tracemono = opt_tracemono || b;
  if (opt_tracemono)
    opt_traceNone = pk_false;
  setON_make_flat_opt (pk_false);
}

/**
   Switch requirement to build the mono Signal graph
   Printing becomes equal to the new requirement
*/
void
OptionManager::switch_make_mono_opt ()
{
  opt_monoGraph = !opt_monoGraph;
  opt_tracemono = opt_monoGraph;
  if (opt_tracemono)
    opt_traceNone = pk_false;
}

/**
   Require to unparse the mono Signal graph if it is built
*/
void
OptionManager::setON_unparse_mono_opt ()
{
  opt_tracemono = pk_true;
  opt_traceNone = pk_false;
}

/**
   @return "true" if building the mono Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_monoGraph ()
{
  return opt_monoGraph;
}

/**
   @return "true" if printing the mono Signal graph is required
   "false" (default) otherwise.
*/
boolean OptionManager::req_trace_mono ()
{
  return opt_tracemono;
}

/* ====> kind of output */
/**
   @return "true" if some output file that is neither a trace file nor an exported file is generated
*/
boolean OptionManager::req_genFile_opt ()
{
  return opt_genFile;
}

/**
   Trace option (-lis)
   Enable the result of the parser analysis printing of the result (abstract tree possibly annotated) of the parser analysis, when b/not b.
*/
void
OptionManager::setON_Signal_list_opt ()
{
  opt_list = pk_true;
  opt_traceNone = pk_false;
}

/**
   Trace option (-lis)
   @return "true" if the the printing of the result (abstract tree possibly annotated) of the parser analysis option is 
   enable, "false" (default) otherwise.
*/
boolean OptionManager::req_listFile ()
{
  return opt_list;
}

/**
   Trace option (-notra)
   Disable every unparsing in TRA file 
*/
void
OptionManager::setOFF_tra_opt ()
{
  opt_traceAllowed = pk_false;
}

/**
   set trace file generation option "on"
*/
void
OptionManager::setON_trace_opt ()
{
  setON_unparse_basic_opt ();
  setON_unparse_poly_opt ();
  setON_unparse_endo_opt ();
  setON_unparse_bool_opt ();
  setON_unparse_flat_opt ();
  setON_unparse_mono_opt ();
}

/** 
    switch trace file generation requirement
*/
void
OptionManager::switch_trace_opt ()
{
  opt_traceALL = !opt_traceALL;
  if (opt_traceALL)
    opt_traceNone = pk_false;
}

/**
   set trace file generation option "on"
*/
void
OptionManager::setON_TRACE_opt ()
{
  opt_traceNone = pk_false;
  opt_traceALL = pk_true;
}

/**
   Trace option (-notra)
   Disable every unparsing in TRA file 
*/
boolean OptionManager::test_traceAllowed ()
{
  return opt_traceAllowed;
}

/**
   Trace option 
   @return "false" if the printing of the graph in Signal textual form option is 
   enable, "true" (default) otherwise.
*/
boolean OptionManager::test_traceNone ()
{
  return opt_traceNone;
}

/**
   Trace option (-tra)
   @return "true" if the printing of the graph in Signal textual form option is 
   enable, "false" (default) otherwise.
*/
boolean OptionManager::req_traceALL ()
{
  return opt_traceAllowed && opt_traceALL;
}


/**
   SME generation. The SME form of a Signal program is produced. The program is not compiled.
   Enable the SME format generation.
*/
void
OptionManager::setON_gen_sme_opt ()
{
  opt_sme = pk_true;
  opt_genFile = pk_true;
}

/**
   SME generation.
   @return "true" if the SME code generation option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_gen_sme_opt ()
{
  return opt_sme;
}

/**
   SSME generation. The ssme form of a Signal program is produced. The program is not compiled.
   It enables/disables(default) the distributed code generation, when b/not b.
*/
void
OptionManager::setON_gen_ssme_opt ()
{
  opt_ssme = pk_true;
  opt_genFile = pk_true;
}

/**
   SSME generation.
   @return "true" if the ssme code generation option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_gen_ssme_opt ()
{
  return opt_ssme;
}

/**
   GENEAUTO generation. The ssme form of a Signal program is produced. The program is not compiled.
   It enables/disables(default) the distributed code generation, when b/not b.
*/
void
OptionManager::setON_gen_geneauto_opt ()
{
  opt_geneauto = pk_true;
  opt_genFile = pk_true;
}

/**
   GENEAUTO generation.
   @return "true" if the ssme code generation option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_gen_geneauto_opt ()
{
  return opt_geneauto;
}

/**
   It enables/disables(default) the kernel translation option when b/not b.
*/
void
OptionManager::setON_kernel_opt ()
{
  opt_kernel = pk_true;
  setON_make_basic_opt (pk_false);
}

/** 
    @return "true" if the kernel translation option is enable, "false" otherwise.    
*/
boolean OptionManager::req_kernel_opt ()
{
  return opt_kernel;
}

/**
   It enables/disables(default) the abstraction option when b/not b.
   When the abstraction is built, it is printed
*/
void
OptionManager::setON_run_abstraction_opt ()
{
  setON_make_basic_opt (pk_false);
  opt_abstraction = pk_true;
}

/**
   switch the abstractiongeneration  requirement
*/
void
OptionManager::switch_run_abstraction_opt ()
{
  opt_abstraction = !opt_abstraction;
}

/** 
    @return "true" if the abtraction option is enable, "false" otherwise.    
*/
boolean OptionManager::req_run_abstraction_opt ()
{
  return opt_abstraction;
}


/**
   Enable the header Signal code generation
*/
void
OptionManager::setON_gen_header_opt ()
{
  opt_header = pk_true;
  opt_genFile = pk_true;
}

/**
   @return the current mode for the header.
*/
boolean OptionManager::req_gen_header_opt ()
{
  return opt_header;
}

/**
   @param b (in) option that enables (b pk_true)/disables (b pk_false) the computation of the "exclusive clocks"
   @param vlevel (in) allows to select the set of clocks on which the calculus is applied:
        - 0 , only the clocks of the signals are considered.
        - 1 , the clocks of the signals are considered + the clocks referenced in the expressions.
	- 2 , all the clocks are considered.
   \see req_gen_exclusiveClock_opt.
   A Signal process FOO_excl.SIG is generated (TOBEDONE)
*/
void
OptionManager::set_gen_exclusiveClock_opt (boolean b, integer vlevel)
{
  opt_Exclusive_Clock = b;
  ecLevel = vlevel;
  setON_make_poly_opt (pk_false);
}

/**
   @return true if the exclusive clocks of the program must be computed, false (default) otherwise.
   @param vlevel (out) the selected set of clocks on which the calculus is applied.

   \see set_gen_exclusiveClock_opt()
*/
boolean OptionManager::req_gen_exclusiveClock_opt (integer * vlevel)
{
  *vlevel = ecLevel;
  return opt_Exclusive_Clock;
}

/* ====> graph options */

/* algorithm tuning */
/**
   Clock calculus option.\n
   It sets the depth of solving to vdepth.
   -depth=0 is equivalent to -basic alone
*/
void
OptionManager::set_solvingDepth_opt (int vdepth)
{
  opt_prof = vdepth;
}

/**
   Clock calculus option.\n
   It sets the depth of solving to vdepth.
   -depth=0 is equivalent to -basic alone
*/
void
OptionManager::setstd_solvingDepth_opt ()
{
  opt_prof = 5;
}

/**
   Switch the Clock calculus option.
*/
void
OptionManager::switch_solvingDepth_opt ()
{
  if (opt_prof != 0)
    opt_prof = 0;
  else
    setstd_solvingDepth_opt ();
}

/** 
    Clock calculus option.\n
    @return the current value of the depth of solving.
*/
integer OptionManager::get_solvingDepth_opt ()
{
  return opt_prof;
}

/**
   Clock calculus option.\n
   @return "true" if the clock calculus is enable (default), "false" otherwise.
   (It is disable by a setting the depth of solving to 0, set_solvingDepth_opt())
*/
boolean OptionManager::req_solving_opt ()
{
  return (opt_prof > 0);
}

/**
   It enables/disables(default) the retiming option when b/not b.
*/
void
OptionManager::setON_run_retiming_opt ()
{
  opt_retiming = pk_true;
  setON_make_basic_opt (pk_false);
}

/**
   Switch the retiming option
*/
void
OptionManager::switch_run_retiming_opt ()
{
  opt_retiming = !opt_retiming;
}

/** 
    @return "true" if the retiming option is enable, "false" otherwise.    
*/
boolean OptionManager::req_run_retiming_opt ()
{
  return opt_retiming;
}


/**
   Enable the "rewriting of control boolean to events " option (b)..
*/
void
OptionManager::setON_run_ctrl2event_opt ()
{
  opt_crew_bool_ctrl = pk_true;
  setON_make_basic_opt (pk_false);
}

/**
   Switch the boolean to event conversion option
*/
void
OptionManager::switch_ctrl2event_opt ()
{
  opt_crew_bool_ctrl = !opt_crew_bool_ctrl;
}

/**
   @return "true" if the "rewriting of booleans to events " is enable, "false" otherwise.
*/
boolean OptionManager::req_ctrl2event_opt ()
{
  return opt_crew_bool_ctrl;
}


/**
   Enable the "rewriting of booleans to events for all booleans
*/
void
OptionManager::setON_crewDefinition_opt ()
{
  opt_crew_definitions = pk_true;
  setON_make_poly_opt (pk_false);
}

/**
   @return "true" if the "rewriting of booleans to events " is enable, "false" otherwise.
*/
boolean OptionManager::req_crewDefinition_opt ()
{
  return opt_crew_definitions;
}

/**
   Enable the "rewriting of statical expressions"
*/
void
OptionManager::setON_crew_statics_opt ()
{
  opt_crew_statics = pk_true;
}

/**
   @return "true" if the "rewriting of statical expressions " is enable, "false" otherwise.
*/
boolean OptionManager::req_crew_statics_opt ()
{
  return opt_crew_statics;
}

/**
   Enable the signal unification option
   When ON, the signals x and y,  defined by x:=E and y :=E are unified. 
   \see req_run_signalUnification_opt()
*/
void
OptionManager::setON_run_signalUnification_opt ()
{
  opt_us = pk_true;
}

/**
   Enable the signal unification option
   When ON, the signals x and y,  defined by x:=E and y :=E are unified. 
   \see req_run_signalUnification_opt()
*/
void
OptionManager::switch_run_signalUnification_opt ()
{
  opt_us = !opt_us;
}

/**
   @return "true" if the signal unification option is enable, "false" (default) otherwise.
   \see setON_run_signalUnification_opt()
*/
boolean OptionManager::req_run_signalUnification_opt ()
{
  return opt_us;
}


/**
   split x:= E1 default E2 (x::=E1 default E2) into x::=E1 | x::=E2 when (E1 ^-E2)
*/
void
OptionManager::set_single2multipleDefinition_opt (boolean b)
{
  opt_default_splitting = b;
  opt_default_normal = b;
}

/**
   @return true if spliting default is required
*/
boolean OptionManager::req_single2multipleDefinition_opt ()
{
  return opt_default_splitting;
}

/**
   group x::=E1 | x::=E2 into x::=E1 default E2 | E1^*E2^=^0
*/
void
OptionManager::setON_multiple2singleDefinition_opt ()
{
  opt_partialdef_merging = pk_true;
  opt_default_normal = pk_true;
  setON_make_basic_opt (pk_false);
}

/**
   @return true if grouping partial definitions is required 
*/
boolean OptionManager::req_multiple2singleDefinition_opt ()
{
  return opt_partialdef_merging;
}

/**
   @return true if default must be normalized
*/
boolean OptionManager::req_normalDefault_opt ()
{
  return opt_default_normal;
}

/** 
    It defines by val, the value such that a signal with less than or equal to \b val 
    reference will be substituted by its definition.
    It can be considered as a partial inverse function of -kernel
    \see req_run_signalSubstitution_opt()
*/
void
OptionManager::set_run_signalSubstitution_opt (integer val)
{
  opt_substitute_value = val;
}

/** 
*/
void
OptionManager::set_run_stdSignalSubstitution_opt ()
{
  opt_substitute_value = 1;
}

/** 
    Switch the substituion of variables option
*/
void
OptionManager::switch_run_signalSubstitution_opt ()
{
  if (opt_substitute_value == 0)
    set_run_stdSignalSubstitution_opt ();
  else
    opt_substitute_value = 0;
}

/** 
    @return
    - pk_true:  a signal with less than or equal to \b val references will be substituted by its definition.
    - false: (default) signals are not substituted ( renaming), val=0.
*/
boolean OptionManager::req_run_signalSubstitution_opt (integer * val)
{
  *val = opt_substitute_value;
  return (opt_substitute_value != 0);
}


/** Optimisation (delays)
    It enables/disables the delay multiplexing option  when b/not b.
    \see req_run_delayMultiplexing_opt.
*/
void
OptionManager::set_run_delayMultiplexing_opt (boolean b)
{
  opt_delayMultiplexing = b;
}

/** Switch the delay multiplexing option
*/
void
OptionManager::switch_run_delayMultiplexing_opt ()
{
  opt_delayMultiplexing = !opt_delayMultiplexing;
}

/**
   @return the current value of the "delay multiplexing" option:
   - pk_true: enable (default), the delay built on a signal are multiplexed (managed) in the same array (code genration), using indices for each signal,
   - false: disable: x and x$ for a scalar signal are managed by two variables.
*/
boolean OptionManager::req_run_delayMultiplexing_opt ()
{
  return opt_delayMultiplexing;
}


/** Optimisation (delays)
    It enables/disables the delay Optimizing option  when b/not b.
    \see req_run_delayOptimizing_opt.
*/
void
OptionManager::set_run_delayOptimizing_opt (boolean b)
{
  opt_delayOptimizing = b;
}

/** Switch the delay Optimizing option
    \see req_run_delayOptimizing_opt.
*/
void
OptionManager::switch_run_delayOptimizing_opt ()
{
  opt_delayOptimizing = !opt_delayOptimizing;
}

/** Optimisation (delays):
    @return the current value of the "delay optimizing" option:
   - pk_true: enable (default): x and x$  are managed by only one variable, if possible.
   - false: disable: x and x$ for a scalar signal are managed by two variables
*/
boolean OptionManager::req_run_delayOptimizing_opt ()
{
  return opt_delayOptimizing;
}

/**
   It enables/disables(default) the kernel translation option when b/not b.
*/
void
OptionManager::setON_splitDataStructure_opt ()
{
  opt_splitDataStructure = pk_true;
  setON_make_basic_opt (pk_false);
}

/** 
    @return "true" if the splitDataStructure translation option is enable, "false" otherwise.    
*/
boolean OptionManager::req_splitDataStructure_opt ()
{
  return opt_splitDataStructure;
}

/**
   It enables/disables(default) the kernel translation option when b/not b.
*/
void
OptionManager::setON_enum2Integer_opt ()
{
  opt_enum2Integer = pk_true;
  setON_make_basic_opt (pk_false);
}

/** 
    @return "true" if the enum2Integer translation option is enable, "false" otherwise.    
*/
boolean OptionManager::req_enum2Integer_opt ()
{
  return opt_enum2Integer;
}

/**
   It enables/disables(default) the utility_clock option when b/when not b.
   When this otpion is on, for all XX signal a signal XX_reduced is created and defined as XX awhen util_clk
*/
void
OptionManager::setON_run_utilityClock_opt ()
{
  opt_utility_clock = pk_true;
  setON_make_poly_opt (pk_false);
}

/**
   @return the current mode for the utility_clock.
*/
boolean OptionManager::req_run_utilityClock_opt ()
{
  return opt_utility_clock;
}

/**
   C code generation: Enable the generation of code for checking the satisfaction of assertions.
*/
void
OptionManager::setON_export_check_subopt ()
{
  opt_check = pk_true;
}

/**
   Switch the check option 
*/
void
OptionManager::switch_export_check_subopt ()
{
  opt_check = !opt_check;
}

/** 
    @return "true" if the check option is enable (code is generated for assertions), "false" otherwise.    
*/
boolean OptionManager::req_export_check_subopt ()
{
  return opt_check;
}


/** PLG: which setting ??? is it related to userPartitionning and otion -dist ?
   Distributed code generation.
   It sets enabled the distributed code generation option
*/
void
OptionManager::setON_export_distrib_subopt ()
{
  opt_distrib = pk_true;
}

/** PLG: which setting ???
   Distributed code generation.
   @return "true" if the Distributed code generation option is enable, "false" (default) otherwise.
*/
void
OptionManager::switch_export_distrib_subopt ()
{
  opt_distrib = !opt_distrib;
}

/** PLG: which setting ???
   Distributed code generation.
   @return "true" if the Distributed code generation option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_distrib_subopt ()
{
  return opt_distrib;
}


/* ====> partition options */

/**
   Partitionning option.
   Enable the separating the state nodes isolation.
*/
void
OptionManager::setON_run_stateSeparation_opt ()
{
  opt_states_others = pk_true;
  setON_make_basic_opt (pk_false);
}

/**
   Partitionning option.
   @return "true" if the separating of the states/others nodes option is enable, "false" otherwise.
*/
void
OptionManager::switch_run_stateSeparation_opt ()
{
  opt_states_others = !opt_states_others;
}

/**
   Partitionning option.
   @return "true" if the separating of the states/others nodes option is enable, "false" otherwise.
*/
boolean OptionManager::req_run_stateSeparation_opt ()
{
  return opt_states_others;
}


/**
   Partitionning option.
   Enable the separation of boolean nodes.
*/
void
OptionManager::setON_run_booleanSeparation_opt ()
{
  opt_bool_others = pk_true;
  setON_make_basic_opt (pk_false);
}

/**
   Switch the booleans nodes isolation option
*/
void
OptionManager::switch_run_booleanSeparation_opt ()
{
  opt_bool_others = !opt_bool_others;
}

/**
   Partitionning option.
   @return "true" if the separating of the booleans/others nodes option is enable, "false" otherwise.
*/
boolean OptionManager::req_run_booleanSeparation_opt ()
{
  return opt_bool_others;
}


/**
   Partitionning option.
   It enables/disables(default) the separating of the finite types/others nodes option when b/not b.
*/
void
OptionManager::setON_run_finiteTypeSeparation_opt ()
{
  opt_finiteTypes_others = pk_true;
  setON_make_basic_opt (pk_false);
}

/**
   Partitionning option.
   @return "true" if the separating of the  finite types/others nodes option is enable, "false" otherwise.
*/
boolean OptionManager::req_run_finiteTypeSeparation_opt ()
{
  return opt_finiteTypes_others;
}

/**
   set on the IO based clustering requirement
   This operation is a partitionning graph operation: two nodes depending on the same sub-set of inputs are set in the same sub-graph.
*/
void
OptionManager::setON_clustering_opt ()
{
  opt_clustering = pk_true;
  setON_make_basic_opt (pk_false);
}

/**
   Switch the input train clustering option
   \see req_clustering_opt()
*/
void
OptionManager::switch_clustering_opt ()
{
  opt_clustering = !opt_clustering;
}

/**
   @return "true" if the input train clustering operation is enable, "false" (default) otherwise.
   This operation is a partitionning graph operation: two nodes depending on the same sub-set of inputs are set in the same sub-graph.
*/
boolean OptionManager::req_clustering_opt ()
{
  return opt_clustering;
}


/**
   Partitionning option: User partitionning.\n
   Enable the graph separation based on user index defined in RunOn pragma.
*/
void
OptionManager::setON_run_tagBasedSeparation_opt ()
{
  opt_tagBased_partition = pk_true;
  setON_make_poly_opt (pk_false);
}

/**
   Switch the color based graph partitionning
*/
void
OptionManager::switch_run_tagBasedSeparation_opt ()
{
  opt_tagBased_partition = !opt_tagBased_partition;
}

/**
   Partitionning option: User partitionning.\n
   @return "true" if the the partitionning of the graph using the pragma RunOn option is enable, "false" otherwise.
*/
boolean OptionManager::req_run_tagBasedSeparation_opt ()
{
  return opt_tagBased_partition;
}

/** It sets enabled the test of poly endochronous programs */
void
OptionManager::setON_polyendo_opt ()
{
  opt_polyendo = pk_true;
  setON_force_subopt ();
  setON_make_poly_opt (pk_false);
}

/**
   Switch the color based graph partitionning
*/
void
OptionManager::switch_polyendo_opt ()
{
  opt_polyendo = ! opt_polyendo;
}


/** It sets enabled the test of poly-endochronous programs */
boolean
OptionManager::req_polyendo_opt ()
{
  return opt_polyendo;
}

/* ====> kind of exported file */
/** Dot generator. 
    It enables the dot code generator option.
    \see req_export_dot_opt() 
*/
void
OptionManager::setON_export_dot_opt ()
{
  opt_filedot = pk_true;
  opt_genFile = pk_true;
}

/** @return the current value of the dot option:
    - pk_true: enable, 
    - false: disable (default).
*/
boolean OptionManager::req_export_dot_opt ()
{
  return opt_filedot;
}

/** Code generator sub-option
    Enable the dynamical test of constraints.
    \see req_force_subopt().
*/

void
OptionManager::setON_force_subopt ()
{
  opt_alwaysCode = pk_true;
}

/** Switch the code generator force otpion
*/
void
OptionManager::switch_force_subopt ()
{
  opt_alwaysCode = !opt_alwaysCode;
}

/** @return the value assigned with the "-force" value.
    This option when it is enable can be used to generate the code even if the graph contains constraints, 
    in this case exceptions are generated (the constraints are dynamically tested).
*/
boolean OptionManager::req_force_subopt ()
{
  return opt_alwaysCode;
}

/**
   If this option is ON: some export action is scheduled
        
*/
void
OptionManager::setON_export_opt ()
{
  opt_exportFile = pk_true;
  setON_make_basic_opt (pk_false);
}

/** 
    @return "true" if some export action is required
*/
boolean OptionManager::req_export_opt ()
{
  return opt_exportFile;
}

/**
   Sigali code generation.
   It enables the exportation of the graph to Sigali
   if trace is pk_true, the Signal graph is unparsed
*/
void
OptionManager::setON_export_z3z_opt (boolean trace)
{
  opt_filez3z = pk_true;
  opt_tracez3z = trace;
  if (trace)
    opt_traceNone = pk_false;
  setON_export_opt ();
  setON_multiple2singleDefinition_opt ();
}

/**
   Switch the Sigali code generation requirement
*/
void
OptionManager::switch_export_z3z_opt ()
{
  if (opt_filez3z)
    opt_filez3z = pk_false;
  else
    setON_export_z3z_opt (req_trace_z3z_opt ());
}

/**
   Sigali code generation.
   @return "true" if the exportation of the graph for Sigali tool option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_z3z_opt ()
{
  return opt_filez3z;
}

/**
   Sigali code generation.
   @return "true" if the unparsing of the exported graph is required
*/
boolean OptionManager::req_trace_z3z_opt ()
{
  return opt_traceAllowed && opt_tracez3z;
}

/**
   SMT code generation.
   It enables the exportation of the graph to SMT solvers
   if trace is pk_true, the Signal graph is unparsed
*/
void
OptionManager::setON_export_smt_opt (boolean trace)
{
  opt_filesmt = pk_true;
  opt_tracesmt = trace;
  if (trace)
    opt_traceNone = pk_false;
  setON_export_opt ();
  setON_multiple2singleDefinition_opt ();
}

/**
   Switch the SMT code generation requirement
*/
void
OptionManager::switch_export_smt_opt ()
{
  if (opt_filesmt)
    opt_filesmt = pk_false;
  else
    setON_export_smt_opt (req_trace_smt_opt ());
}

/**
   SMT code generation.
   @return "true" if the exportation of the graph for SMT tool option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_smt_opt ()
{
  return opt_filesmt;
}

/**
   SMT code generation.
   @return "true" if the unparsing of the exported graph is required
*/
boolean OptionManager::req_trace_smt_opt ()
{
  return opt_traceAllowed && opt_tracesmt;
}

/**
   Lustre code generation.
   It enables the exportation of the graph to Lustre tool
*/
void
OptionManager::setON_export_lustre_opt (boolean trace)
{
  opt_filelustre = pk_true;
  opt_tracelustre = trace;
  if (trace)
    opt_traceNone = pk_false;
  setON_export_opt ();
  setON_multiple2singleDefinition_opt ();
  setON_splitDataStructure_opt ();
  setON_enum2Integer_opt ();
  setON_make_bool_opt (pk_false);
}

/**
   Lustre code generation.
   It enables the exportation of the graph to Lustre tool
*/
void
OptionManager::switch_export_lustre_opt ()
{
  opt_filelustre = !opt_filelustre;
}

/**
   Lustre code generation.
   @return "true" if the printing of the graph for Lustre tool option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_lustre_opt ()
{
  return opt_filelustre;
}

/**
   Lustre code generation.
   @return "true" if the unparsing of the exported graph is required
*/
boolean OptionManager::req_trace_lustre_opt ()
{
  return opt_traceAllowed && opt_tracelustre;
}

/**
   C code generator.
   It enables/disables the generation of the BODY part for C language, when b/not b.
*/
void
OptionManager::set_export_c_body_opt (boolean b)
{
  opt_c_modubody = b;
  if (b)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** C code generator.
    @return "true" if the C code generation (BODY part) option is enable, "false" (default) otherwise.
    \see set_export_c_body_opt()
*/
boolean OptionManager::req_export_c_body_opt ()
{
  return opt_c_modubody;
}

/**
   C code generator.
   It enables/disables the generation of the MAIN part for C language, when b/not b.
*/
void
OptionManager::set_export_c_main_opt (boolean genmain)
{
  opt_c_modumain = genmain;
  if (genmain)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** C code generator.
    @return "true" if the C code generation (MAIN part) option is enable, "false" (default) otherwise.
    \see set_export_c_main_opt()
*/
boolean OptionManager::req_export_c_main_opt ()
{
  return opt_c_modumain;
}

/**
   C code generator.
   It enables/disables the generation of the IO file for C language, when b/not b.
*/
void
OptionManager::set_export_c_io_opt (boolean genio)
{
  opt_c_moduio = genio;
  if (genio)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** C code generator.
    @return "true" if the C code generation (IO part) option is enable, "false" (default) otherwise.
    \see set_export_c_io_opt()
*/
boolean OptionManager::req_export_c_io_opt ()
{
  return opt_c_moduio;
}

/** C code generator.
    It enables/disables the generation of the BODY/MAIN/IO parts for C++ language, when b/not b.
*/
void
OptionManager::set_export_c_modules_opt (boolean b, boolean trace)
{
  opt_tracec = trace;
  if (trace)
    opt_traceNone = pk_false;
  set_export_c_body_opt (b);
  set_export_c_main_opt (b);
  set_export_c_io_opt (b);
}

/** Switch the C code generation requirement
*/
void
OptionManager::switch_export_c_modules_opt ()
{
  set_export_c_modules_opt (!req_export_c_modules_opt (), req_trace_c_opt ());
}

/** C code generator.
    @return "true" if the C code generation option is enable, "false" (default) otherwise.
    It is the result of req_export_c_body_opt() or req_export_c_main_opt() or req_export_c_io_opt().
*/
boolean OptionManager::req_export_c_modules_opt ()
{
  return req_export_c_body_opt () || req_export_c_main_opt () || req_export_c_io_opt ();
}

/**
   C code generation.
   @return "true" if the unparsing of the exported graph is required
*/
boolean OptionManager::req_trace_c_opt ()
{
  return opt_traceAllowed && opt_tracec;
}

/**
   C++ code generator.
   It enables/disables the generation of the BODY part for C++ language, when b/not b.
*/
void
OptionManager::set_export_cpp_body_opt (boolean b)
{
  opt_cpp_modubody = b;
  if (b)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** C++ code generator.
    @return "true" if the C++ code generation (BODY part) option is enable, "false" (default) otherwise.
    \see set_export_cpp_body_opt()
*/
boolean OptionManager::req_export_cpp_body_opt ()
{
  return opt_cpp_modubody;
}

/**
   C++ code generator.
   It enables/disables the generation of the MAIN part for C++ language, when b/not b.
*/
void
OptionManager::set_export_cpp_main_opt (boolean b)
{
  opt_cpp_modumain = b;
  if (b)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** C++ code generator.
    @return "true" if the C code generation (MAIN part) option is enable, "false" (default) otherwise.
    \see set_export_cpp_main_opt()
*/
boolean OptionManager::req_export_cpp_main_opt ()
{
  return opt_cpp_modumain;
}

/**
   C++ code generator.
   It enables/disables the generation of the IO file for C++ language, when b/not b.
*/
void
OptionManager::set_export_cpp_io_opt (boolean b)
{
  opt_cpp_moduio = b;
  if (b)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** C++ code generator.
    @return "true" if the C++ code generation (IO part) option is enable, "false" (default) otherwise.
    \see set_export_cpp_io_opt()
*/
boolean OptionManager::req_export_cpp_io_opt ()
{
  return opt_cpp_moduio;
}

/** C++ code generator.
    It enables/disables the generation of the BODY/MAIN/IO parts for C++ language, when b/not b.
*/
void
OptionManager::set_export_cpp_modules_opt (boolean b, boolean trace)
{
  opt_tracec = trace;
  if (trace)
    opt_traceNone = pk_false;
  set_export_cpp_body_opt (b);
  set_export_cpp_main_opt (b);
  set_export_cpp_io_opt (b);
}

/** Switch the CPP code generation requirement
*/
void
OptionManager::switch_export_cpp_modules_opt ()
{
  set_export_cpp_modules_opt (!req_export_cpp_modules_opt (), req_trace_cpp_opt ());
}

/** C++ code generator.
    @return "true" if the C++ code generation option is enable, "false" (default) otherwise.
    (result of req_export_cpp_body_opt() or req_export_cpp_main_opt() or req_export_cpp_io_opt() )
*/
boolean OptionManager::req_export_cpp_modules_opt ()
{
  return req_export_cpp_body_opt () || req_export_cpp_main_opt () || req_export_cpp_io_opt ();
}

/**
   C++ code generation.
   @return "true" if the unparsing of the exported graph is required
*/
boolean OptionManager::req_trace_cpp_opt ()
{
  return opt_traceAllowed && opt_tracecpp;
}

/**
   Java code generator.
   It enables/disables the generation of the BODY part for Java language, when b/not b.
*/
void
OptionManager::set_export_java_body_opt (boolean b)
{
  opt_java_modubody = b;
  if (b)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** 
    Java code generator.
    @return "true" if the Java code generation (BODY part) option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_java_body_opt ()
{
  return opt_java_modubody;
}

/**
   Java code generator.
   It enables/disables the generation of the MAIN part for Java language, when b/not b.
*/
void
OptionManager::set_export_java_main_opt (boolean b)
{
  opt_java_modumain = b;
  if (b)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** 
    Java code generator.
    @return "true" if the Java code generation (MAIN part) option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_java_main_opt ()
{
  return opt_java_modumain;
}

/**
   Java code generator.
   It enables/disables the generation of the IO file for Java language, when b/not b.
*/
void
OptionManager::set_export_java_io_opt (boolean b)
{
  opt_java_moduio = b;
  if (b)
    {
      setON_export_opt ();
      setON_make_bool_opt (pk_false);
    }
}

/** 
    Java code generator.
    @return "true" if the Java code generation (IO part) option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_java_io_opt ()
{
  return opt_java_moduio;
}

/**
   Java code generator.
   It enables/disables the Java code generation, when b/not b.
*/
void
OptionManager::set_export_java_opt (boolean b)
{
  opt_java = b;
}

/** Java code generator.
    It enables/disables the generation of the BODY/MAIN/IO parts for C++ language, when b/not b.
*/
void
OptionManager::set_export_java_modules_opt (boolean b, boolean trace)
{
  opt_tracejava = trace;
  if (trace)
    opt_traceNone = pk_false;
  set_export_java_body_opt (b);
  set_export_java_main_opt (b);
  set_export_java_io_opt (b);
  set_export_java_opt (b);
}

/** Switch the java code generation requirement
*/
void
OptionManager::switch_export_java_modules_opt ()
{
  set_export_java_modules_opt (!req_export_java_modules_opt (), req_trace_java_opt ());
}

/** Java code generator.
    @return "true" if the Java code generation option is enable, "false" (default) otherwise.
    (result of req_export_java_body_opt() or req_export_java_main_opt() or req_export_java_io_opt() )
*/
boolean OptionManager::req_export_java_modules_opt ()
{
  return req_export_java_body_opt () || req_export_java_main_opt () || req_export_java_io_opt ();
}

/**
   java code generation.
   @return "true" if the unparsing of the exported graph is required
*/
boolean OptionManager::req_trace_java_opt ()
{
  return opt_traceAllowed && opt_tracejava;
}

/** 
    Java code generator.
    @return "true" if the Java code generation option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_java_opt ()
{
  return opt_java;
}

/**
   Code generator mode using threads.
   implies clusterisation.
   One thread is generated for each cluster (currently, C, C++)
   It enables the code generation with threads.
*/
void
OptionManager::setON_export_threads_subopt ()
{
  opt_threads = pk_true;
  setON_export_opt ();
  setON_clustering_opt ();
  setON_make_bool_opt (pk_false);
}

/**
   Switch the thread generation option
*/
void
OptionManager::switch_export_threads_subopt ()
{
  opt_threads = !opt_threads;
}

/** 
    Code generator using threads.
    @return "true" if the code generation with threads option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_threads_subopt ()
{
  return opt_threads;
}


/**
   SynDEx code generation.
   It enables the exportation of the graph to SynDEx tool
   Requires kernel, multiple2singleDefinition
*/
void
OptionManager::setON_export_syndex_opt (boolean trace)
{
  opt_filesyndex = pk_true;
  opt_tracesyndex = trace;
  if (trace)
    opt_traceNone = pk_false;
  setON_export_opt ();
  setON_multiple2singleDefinition_opt ();
  setON_kernel_opt ();
  setON_make_flat_opt (pk_false);
}

/**
   Switch the SynDEx code generation.option
*/
void
OptionManager::switch_export_syndex_opt ()
{
  opt_filesyndex = !opt_filesyndex;
}

/**
   SynDEx code generation.
   @return "true" if the printing of the graph for SynDEx tool option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_syndex_opt ()
{
  return opt_filesyndex;
}

/**
   SynDEx code generation.
   @return "true" if the unparsing of the exported graph is required
*/
boolean OptionManager::req_trace_syndex_opt ()
{
  return opt_traceAllowed && opt_tracesyndex;
}

/**
   It enables/disables(default) the profiling (morphism of program for evaluation of performance) option when b/not b.
   If this option is ON: it generates
        FOO_simu.SIG  that contains FOO.SIG solved and enhanced with synchronization boolean signals
        FOO_type_context.SIG that contains a list of notations for referenced types
        FOO_profiling.SIG  that contains morphism(FOO)
        
*/
void
OptionManager::setON_export_profiling_opt ()
{
  opt_export_profiling = pk_true;
  setON_export_opt ();
}

/**
  Switch the profiling generation option
        
*/
void
OptionManager::switch_export_profiling_opt ()
{
  opt_export_profiling = !opt_export_profiling;
}

/** 
    @return "true" if the profiling option is enable, "false" otherwise.    
*/
boolean OptionManager::req_export_profiling_opt ()
{
  return opt_export_profiling;
}


/* ====> listing options */

/**
   Enable the statistics mode option.
*/
void
OptionManager::setON_stat_opt ()
{
  setON_make_poly_opt (pk_false);
  opt_stat = pk_true;
  pK_messenger->setOptionMessenger (1, pk_true);
}

/**
   @return "true" if the statistic mode option is enable, "false" otherwise.
*/
boolean OptionManager::req_stat_opt ()
{
  return opt_stat;
}

/**
   Enable the debug mode.
*/
void
OptionManager::setON_debug_opt ()
{
  opt_debug = pk_true;
}

/**
   @return "true" if the debug mode is ebnable, "false" otherwise.
*/
boolean OptionManager::req_debug_opt ()
{
  return opt_debug;
}

/* ====> wizzard option */

/** PLG: Not a compiler option
   Used by developpers for testing their own functionalities.
   It enables/disables the test functionality option when b/not b.
   \see req_wiz_testFunctionality_opt()
*/
void
OptionManager::set_wiz_testFunctionality_opt (boolean b)
{
  opt_test_function = b;
}

/**
   Used by developpers for testing their own functionalities.
   @return true if the test functionality is enable, false (default) otherwise.
*/
boolean OptionManager::req_wiz_testFunctionality_opt ()
{
  return opt_test_function;
}

/**
   Used by developpers for testing their own functionalities.
   It sets the name of the test.
   \see getTestName()
*/
void
OptionManager::setTestName (const char *vstr)
{
  testName = (char *) vstr;
}

/**
   Used by developpers for testing their own functionalities.
   @return the name of the test, sets by setTestName()
*/
char *
OptionManager::getTestName ()
{
  return testName;
}

void
OptionManager::reset ()
{
  setopt_init ();
}


/**
   VHDL code generation.
   It enables the exportation of the graph to VHDL code
*/
void
OptionManager::setON_export_vhdl_opt (boolean trace)
{
  opt_vhdl = pk_true;
  setON_export_opt ();
  setON_make_bool_opt (pk_false);
  // opt_tracevhdl = trace;

}


/**
   VHDL code generation.
   @return "true" if the exportation of the graph for VHDL option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_vhdl_opt ()
{
  return opt_vhdl;
}

/**
   Verilog code generation.
   It enables the exportation of the graph to Verilog code
*/
void
OptionManager::setON_export_verilog_opt (boolean trace)
{
  opt_verilog = pk_true;
  setON_export_opt ();
  setON_make_bool_opt (pk_false);
  // opt_traceverilog = trace;
}


/**
   Verilog code generation.
   @return "true" if the exportation of the graph for Verilog option is enable, "false" (default) otherwise.
*/
boolean OptionManager::req_export_verilog_opt ()
{
  return opt_verilog;
}


/**
   Certification options.
*/
boolean OptionManager::req_export_certification()
{
  // -smt -kt
  return( req_export_smt_opt () && req_keeprew_opt ());
}
