/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/** Specific method for GUI.
    @return the father process of a process (p) when it exists, NILTREE otherwise.
*/
  ntrees Trees2Kernel::getprocenglobant (ntrees p)
{
  p = nfather (p);
  if (IS_NILTREE (p))
    return NILTREE;
  if (noper (p) != opldecl)
    return NILTREE;
  p = nfather (nfather (nfather (p)));
  return p;
}

/** Specific method for GUI.
    @return the list of the identifiers (abstract tree representation) from a list of declarations (kernel tree representation)
*/
ntrees Trees2Kernel::mkklmklident (ptlrepid listerepid)
{
  ptidentl
    lident;

  if (listerepid == NILLREPID)
    return namlfridempty ();
  lident = namlfrid (identpile (nameId (listerepid)));
  listerepid = succid (listerepid);
  while (listerepid != NILLREPID)
    {
      namlisput (lident, identpile (nameId (listerepid)), 1);
      listerepid = succid (listerepid);
    }
  return lident;
}

/** Specific method for GUI.
 */
ntrees Trees2Kernel::mkklmkduplident (ptlrepid listerepidtest, ptlrepid listerepid)
{
  ptidentl
    lident;

  if (listerepid == NILLREPID)
    return namlfridempty ();
  if (ptlrepid_remove_Unsorted (&listerepidtest, nameId (listerepid)))
    lident = namlfrid (copy_ident (identpile (nameId (listerepid))));
  else
    lident = namlfrid (identpile (nameId (listerepid)));
  listerepid = succid (listerepid);
  while (listerepid != NILLREPID)
    {
      if (ptlrepid_remove_Unsorted (&listerepidtest, nameId (listerepid)))
	namlisput (lident, copy_ident (identpile (nameId (listerepid))), 1);
      else
	namlisput (lident, identpile (nameId (listerepid)), 1);
      listerepid = succid (listerepid);
    }
  return lident;
}

/** Specific method for GUI.
 */
boolean Trees2Kernel::parlistproc (ntrees p, ntrees p_model, ptlrepid * listglob)
{
  int
    i,
    n;
  boolean
    b;
  ptrepfi
    vfiglob;
  ptIdStack
    varid;

  listpro = nfather (p);
  if (IS_NILTREE (listpro))
    return pk_true;
  if (noper (listpro) != opldecl)
    return pk_true;
  n = ntreearity (listpro);
  if (n == 0)
    return pk_true;
  b = pk_true;
  for (i = 1; i <= n; i++)
    if (isChild (i, listpro))
      if (operande != p_model)
	if (STree_isModel (operande))
	  {
	    cSpecGUI (operande, &varid, &vfiglob);
	    if (!appendProcess (listglob, varid, vfiglob))
	      if (!STree_isOperatorModel (operande))
		b = pk_false;
	  }
  return b;
}

_PUBLIC_
/** Specific method for GUI.
    Similar to modelCompiling() method in GUI context (it produces list of identifiers in abstract tree representation).
    It produces the inputs/outputs of an expression (x) which is an expression referenced in the model (vfi).
    vfi is characterized by the stack (varidproc) associated with the name of the model.
*/
  void
Trees2Kernel::modelCompilingGUI (ptrepfi vfi, ptIdStack varidproc, ntrees x, ntrees * lidententree, ntrees * lidentsortie)
{
  ptklExprProc vklExprProc;
  ntrees p, p_model;
  ptIdStack varid;
  ptrepfi vfiloc;
  ptlrepid inputs, outputs, ins, outs, listglob, locfi;
  integer i, nb_liste;
  ptrepfi save_process;
  boolean nonempiler, empile, todo;
  // boolean b;

  GUIContext = pk_true;
  save_process = G_process;
  G_process = vfi;

  if (vfi != NILREPFI)
    {
      empile = pk_true;
      p_model = kl_getBodyModel (vfi, &vklExprProc);
      /* Added lbesnard, Sep 2008 */
      outs = kl_getInsOutsModel (vfi, &ins);
      pushFlows (ins);
      pushFlows (outs);
      /* --------- */
    }
  else
    empile = pk_false;

  todo = pk_true;
  inputs = NILLREPID;
  outputs = NILLREPID;
  if (empile)
    {
      listglob = NILLREPID;
      nonempiler = pk_false;
      p = p_model;
      while (ISNOT_NILTREE (p))
	{
	  /* b =*/ parlistproc (p, p_model, &listglob);
	  p = getprocenglobant (p);
	}
      if (ptlrepid_remove_Unsorted (&listglob, varidproc))
	pushModels (listglob);
      else
	nonempiler = pk_true;
      if (isChild (3, p_model))
	{
	  if (isoChild (2, operande))
	    if (voperator == oppconf)
	      {
		listpro = nchild (2, operande);
		locfi = NILLREPID;
		nb_liste = ntreearity (listpro);
		for (i = 1; i <= nb_liste; i++)
		  if (isChild (i, listpro))
		    if (STree_isModel (operande))
		      {
			cSpecGUI (operande, &varid, &vfiloc);
			/* b = */ appendProcess (&locfi, varid, vfiloc);
		      }
		operande = x;
		voperator = noper (x);
		if (locfi == NILLREPID)
		  /*b =*/ klparcours (&inputs, &outputs, &vklExprProc);
		else
		  {
		    kl_setLocModels (vfi, locfi);
		    pushModels (locfi);
		    /*b = */klparcours (&inputs, &outputs, &vklExprProc);
		    popModels (locfi);
		  }
	      }
	  if (!nonempiler)
	    popModels (listglob);
	  popFlows (kl_getParametersModel (vfi));
	  popFlows (ins);	// added  Sep 2008
	  popFlows (outs);	// added  Sep 2008
	  todo = pk_false;
	}
    }

  if (todo)
    {
      operande = x;
      voperator = noper (x);
      /* b = */ klparcours (&inputs, &outputs, &vklExprProc);
    }

  *lidentsortie = mkklmklident (outputs);
  *lidententree = mkklmkduplident (outputs, inputs);
  G_process = save_process;
}


/** Specific method for GUI.
    Similar to the cSpec() method in GUI context: from an abstract tree representation of a model (p), it produces 
    a kernel tree representation (vmodel).
    @param p (in) Signal model (process, node,...) in abstract tree representation (produced by the Signal Parser),
    @param varid (out) the stack associated with the name of the model.
    @param vmodel (out) Signal model in  kernel tree representation.
 */
void
Trees2Kernel::cSpecGUI (ntrees p, ptIdStack * varid, ptrepfi * vmodel)
{
  ptident id;
  ptlrepid param, inList, outList;
  //  ntrees vfather 
  noperators op;

  if (IS_NILTREE (p))
    *vmodel = NILREPFI;
  else
    {
      GUIContext = pk_true;
      if (unnamedStack == NILPILE)
	unnamedStack = get_stackOf (getIdentNil ());
      *varid = locateModelName (p);
      // vfather = nfather (p);
      id = identpile (*varid);
      op = noper (p);
      if (op == opprocess)
	op = optypeproc;
      else if (op == opact)
	op = optypeact;
      else if (op == opnode)
	op = optypenode;
      else
	op = optypefunc;

      *vmodel = kl_mkModel (p, id, op);
      if (isChild (2, p))
	if (voperator == opinte)
	  {
	    p = operande;
	    parametersInputsOutputsCompiling (p, &param, &inList, &outList);
	    /* vio = nchild(2, p);
	       param = cDecFlowsInterface(1, p, pk_true);
	       inList = cDecFlowsInterface(1, vio, pk_false);
	       outList = cDecFlowsInterface(2, vio, pk_false);
	     */
	    p = nfather (p);
	  }
      kl_setParamsInputsOutputs (*vmodel, param, inList, outList);
    }
}




/** For an (abstract tree) expression (x), it synthetizes the inputs (lidententree) and the output (lidentsortie).
    The following actions are achieved, for a valid tree:
     - It searches up in x tree, the model in which x is, 
     - the specification of this model is produced (cSpecGUI() method),
     - the interface of x is then produced in this context (modelCompilingGUI() method).
*/
void
Trees2Kernel::Trees2KernelGUI (ntrees x, ntrees * lidententree, ntrees * lidentsortie)
{
  ptIdStack varid;
  ntrees vfather;
  ptrepfi vfi;

  SET_NILTREE (*lidententree);
  SET_NILTREE (*lidentsortie);
  if (ISNOT_NILTREE (x))
    {
      vfather = x;
      while (!STree_isModel (vfather))
	vfather = nfather (vfather);
      cSpecGUI (vfather, &varid, &vfi);
      modelCompilingGUI (vfi, varid, x, lidententree, lidentsortie);
    }
}
