/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PUBLIC_
/**
   Constructor.
*/
  Trees2Kernel::Trees2Kernel ()
{
  {
    modeConstTypeCompiling = pk_false;
    currentTypeConst = NILLREPID;
    lConstTypesCurrentBlock = NILLREPID;
  }
  outputMode = pk_false;
  intrinsicLoading = pk_false;
  SET_NILTREE (listpro);	/* liste des declarations locList du processus courant */
  G_process = NILREPFI;
  unnamedStack = NILPILE;
  SET_NILTREE (operande);
  voperator = opnil;
  lindex = NILLREPID;
  moduleInstrinsics = NILREPFI;
  toppilefc = NILSTACK;
  bWarSafeRestict = pk_false;
  G_ExternalOutputsModel = NILLREPID;
  G_ExternalInputsModel = NILLREPID;
}


/**
   Initializing Trees2Kernel translator.
   It loads Signal predefined libraries defined in SIGNAL.PREDEF Signal program, using the SIGNAL_LIBRARY_PATH environment
   shell variable.
   @return "true" if the operation succeeds, "false" otherwise.
 */
boolean Trees2Kernel::initialize ()
{
  ntrees
    vtree;
  ptrepstr
    dirName,
    name,
    fullPathName;
  const char *
    stringFileName = "SIGNAL_INTRINSICS.SIG";

  // NB: the models are pushed during the compiling (in moduleCompiling() method).
  if (moduleInstrinsics != NILREPFI)
    return pk_true;
  name = string2repstr (stringFileName);
  if (!sysGetDirnameLibrary (name, &dirName, &fullPathName))
    {
      pK_messenger->globalMessageString (error_openingFile, stringFileName);
      return pk_false;
    }
  /* the full path name dirName/name exists */
  if (!parseLibrary (fullPathName, &vtree))
    return pk_false;
  loadIntrinsicLibrary (vtree);
  AttributeIntrinsics ();
  return pk_true;
}

/**
   Finalizing Trees2Kernel translator.
   It pops the predefined Signal libraries previously loaded in initialize() method.
 */
void
Trees2Kernel::finalize ()
{
  popIntrinsics ();
  atpilereset ();
  loadedLibraries.clear ();
  lpredConstTypes.clear ();
}

/**
   Translation of a tree (x) to a kernel tree object.
   @param x (in) : a abstract-tree of a model of process (module, list of modules, process, node, ...,operator).
   @param cc (out) : at "true" if some erros have been detected, "false" otherwise
   @param idproc (out) : name of the model 
   @return  the kernel-Tree representation of the model.
*/
ptrepfi Trees2Kernel::Trees2KernelStep1 (ntrees x, boolean * cc, ptident * idproc, boolean announce)
{
  ptIdStack
    varid;
  noperators
    op;
  boolean
    b;
  ptrepfi
    vfi;

  GUIContext = pk_false;
  if (IS_NILTREE (x))
    return NILREPFI;
  op = noper (x);

  // Process or a list of modules
  *cc = pk_true;
  pK_messenger->setErrorCode (*cc);
  if (announce)
    pK_messenger->message (subAnnounce_reduction2basic);
  unnamedStack = getidnil ();
  switch ((int) op)
    {
    case opmodu:		// accepted for the GUI.
      vfi = moduleCompiling (x, &varid);
      kl_setLocModels (vfi, selectMainProcesses (kl_getLocModels (vfi), x));	// "mains" are selected or all models
      *cc = kl_getErrorCode (vfi);
      popModule (vfi);
      break;

    case oplmodu:
      // Restriction : application à un seul module.
      ptlrepid lvfi;
      lvfi = modulesListCompiling (x, &b);
      if (!b)
	mkklWarning (warlmodurestrict);
      vfi = getModelRepid (lvfi);	// le premier de la liste.
      kl_setLocModels (vfi, selectMainProcesses (kl_getLocModels (vfi), x));	// "mains" are selected or all models
      *cc = kl_getErrorCode (vfi);
      popModule (vfi);
      break;
    case opprocess:
    case opnode:
    case opfunc:
    case opautomaton:
    case opprocedure:
    case opoper:		// operator 
      cSpec (x, &varid, &vfi);
      modelCompiling (vfi);
      *cc = pK_messenger->getErrorCode ();
      kl_setError (vfi, *cc);
      break;
    default:
      notYetImplemented (x);
    }
  popModules (kl_getModules (vfi));
  *idproc = kl_getNameModel (vfi);
  return vfi;
}


/**
   Translation of a tree (x) to a kernel tree object. Producing a parametrized instance of a model.
   @param x (in) : a instance of process ( P {p1,...pn} )
   @param vfi (in) : representation du modele du processus,
   @param parglobal (out) : representation of the list of effective parameters.
   @param cc (in/out)  : error code.
   @return true if no errors, false otherwise.
*/
boolean Trees2Kernel::Trees2KernelStep2 (ntrees x, ptrepfi vfi, ptklExprSig * parglobal, boolean * cc)
{
  ptlrepid
    laux,
    laux2;
  ptklExprProc
    expfi;

  GUIContext = pk_false;
  if (IS_NILTREE (x))
    {
      *parglobal = NILEXPRSIG;
      return pk_false;
    }
  if (!*cc)
    {
      *parglobal = NILEXPRSIG;
      return pk_false;
    }
  if (isoChild (2, x))
    {
      laux2 = NILLREPID;
      laux = NILLREPID;
      *parglobal = cExpEffectiveParam (operande, vfi, &laux, &laux2);
      if (ptlrepid_size (laux) > 0)
	mkklErrorAnnotTree (errparam, kl_getBodyModel (vfi, &expfi));
      verifParameters (vfi, *parglobal, operande);
    }
  else
    *parglobal = NILEXPRSIG;
  *cc = pK_messenger->getErrorCode ();
  kl_setError (vfi, *cc);
  return *cc;
}

/**
   @return the loaded libraries. Each element of the table is a pair(filename, internal representation of the module).
   It must be called after a call to Trees2KernelStep1(), otherwise it is an empty table.
 */
libTableType
Trees2Kernel::getLoadedLib ()
{
  return loadedLibraries;
}


void
Trees2Kernel::compilingDeclPush (ptrepfi vfi, ntrees list_decl, boolean csp)
{
  ntrees p_model;
  ptlrepid vlrepid, vparameters, locList, inList, outList, locfi, const_typ_loc, llabels, lvfiModules, modelTypes;
  ptrepfi vfiLoc;
  ptklExprProc vklExprProc;

  if (!STree_isUndefinedTree (list_decl))
    {
      p_model = kl_getBodyModel (vfi, &vklExprProc);
      outList = kl_getInsOutsModel (vfi, &inList);
      vparameters = kl_getParametersModel (vfi);

      operande = list_decl;

      pushParameters (vparameters);	/* warning: types/processes are allowed in parameters. */
      lvfiModules = cDecImports (list_decl);
      kl_setModules (vfi, lvfiModules);
      const_typ_loc = cDecLocalConstantsTypes (list_decl);	// ... and they are pushed.

      verifDoubleDec (&const_typ_loc, vparameters, pk_true, p_model);
      verifDoubleDec (&const_typ_loc, outList, pk_true, p_model);
      verifDoubleDec (&const_typ_loc, inList, pk_true, p_model);

      llabels = cDecLabels (list_decl);
      verifDoubleDec (&llabels, vparameters, pk_true, p_model);
      verifDoubleDec (&llabels, outList, pk_true, p_model);
      verifDoubleDec (&llabels, inList, pk_true, p_model);
      verifDoubleDec (&llabels, const_typ_loc, pk_true, p_model);
      kl_setLabelsModel (vfi, llabels);

      const_typ_loc = cDecLocalConstantsTypesSet (const_typ_loc);	// ordered (types/constants precedences)

      modelTypes = cDecModelType (list_decl);
      verifDoubleDec (&modelTypes, outList, pk_true, p_model);
      verifDoubleDec (&modelTypes, inList, pk_true, p_model);
      verifDoubleDec (&modelTypes, vparameters, pk_true, p_model);
      verifDoubleDec (&modelTypes, const_typ_loc, pk_true, p_model);
      verifDoubleDec (&modelTypes, llabels, pk_true, p_model);
      kl_setModelTypes (vfi, modelTypes);

      locList = cLocalSignals (list_decl);
      verifDoubleDec (&locList, outList, pk_true, p_model);
      verifDoubleDec (&locList, inList, pk_true, p_model);
      verifDoubleDec (&locList, vparameters, pk_true, p_model);
      verifDoubleDec (&locList, const_typ_loc, pk_true, p_model);
      verifDoubleDec (&locList, llabels, pk_true, p_model);

      locfi = cLocalModels (list_decl);

      verifDoubleDec (&locList, locfi, pk_true, p_model);
      kl_setLocalFlowModel (vfi, locList);
      kl_setConstantsTypesModel (vfi, const_typ_loc);
      pushFlows (inList);
      pushFlows (outList);
      pushFlows (locList);
      if (locfi != NILLREPID)
	{ /* Warning:the labels are not visible in the local models. */
	  verifDoubleDec (&const_typ_loc, locfi, pk_true, p_model);
	  verifDoubleDec (&modelTypes, locfi, pk_true, p_model);
	  kl_setLocModels (vfi, locfi);
	  pushModels (locfi);
	  if (csp)
	    {
	      vlrepid = locfi;
	      do
		{
		  vfiLoc = getModelRepid (vlrepid);
		  modelCompiling (vfiLoc);
		  popModules (kl_getModules (vfiLoc));
		  vlrepid = succid (vlrepid);
		}
	      while (vlrepid != NILLREPID);
	    }
	}
      pushLabels (llabels);
    }
}


void
Trees2Kernel::compilingDeclPop (ptrepfi vfi)
{
  ptlrepid vparameters, locList, inList, outList, locfi, modelTypes, const_typ_loc, llabels;

  outList = kl_getInsOutsModel (vfi, &inList);
  vparameters = kl_getParametersModel (vfi);
  kl_getLocalDeclarations (vfi, &const_typ_loc, &locList, &llabels);
  locfi = kl_getLocModels (vfi);
  modelTypes = kl_getModelTypes (vfi);

  popFlows (llabels);
  popModelTypes (modelTypes);
  popModels (locfi);
  popFlows (locList);
  _POP_INTERFACE (vparameters, inList, outList);
  popConstantsTypes (const_typ_loc);
}
