/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* OPERATORS NOT YET IMPLEMENTED 
   opltypep // "private" "type"  <l_def_types> 
   opdecconp // "private" "constant"  <type> <l_def_constantes> 
*/


#define _PUSH_INTERFACE(vparameters,inList, outList) { pushParameters(vparameters); pushFlows(inList); pushFlows(outList);}
#define _POP_INTERFACE(vparameters,inList, outList) { popFlows(outList); popFlows(inList); popFlows(vparameters); }
#define OPIS_SIGNALIDENTIFIER(voperator) ( (voperator == opnom) || (voperator == opident) )

Trees2Kernel *pK_Tree2Kernel = new Trees2Kernel ();	// Predefined object

_PROTECTED_


boolean Trees2Kernel::IsAnIdentifier (ptklExprSig p1, ntrees p)
{
  if (kl_getFlowExprClassType (p1) == variabl) return pk_true;
  mkklErrorAnnotTree (errconflid, p);
  return pk_false;
}


boolean
Trees2Kernel::verifExprSigIsLabel (ptklExprSig p1, ntrees p, ptrepproto * videntlabel)
{
  ptIdStack name;
  ptlrepid vrepid;

  if (! IsAnIdentifier(p1,p)) return pk_false;

  vrepid = kl_getIdentifierExpr (p1);
  name = nameId (vrepid);
  if (isTopaLabel (name))
    {
      (*videntlabel) = getIdRepid (vrepid);
      return pk_true;
    }
  mkklErrorAnnotTreeName (errconflid, p, name);
  return pk_false;
}

boolean
Trees2Kernel::isAutomatonModel (ptrepfi vmodel)
{
  noperators op;
  op = noper (kl_getTreeReference (vmodel));
  return LL_isAutomatonFamilyOp (op);
}

/**
   It verifies that the specific instrinsics for automaton are used in the good context.
   @return true if it is correct, false otherwise.
*/
boolean
Trees2Kernel::verifAutomatonIntrinsics (ptklExprSig inputs1, ptident id, ntrees p, boolean *ok)
{
  ptklExprSig p1, p2;
  ptrepproto vlab;

  *ok = pk_true;
  if (isAutomatonModel (G_process))  // current model.
    {
      if (isTransitionId (id))
	{
	  // Transition(lab1, lab2,..) : the two first inputs must be labels
	  kl_getBinaryExpr (inputs1, &p1, &p2);
	  if (verifExprSigIsLabel (p1, p, &vlab)) ptrepproto_setLabelStateAutomaton (vlab, pk_true); else *ok = pk_false;
	  kl_getBinaryExpr (p2, &p1, &p2);
	  if (verifExprSigIsLabel (p1, p, &vlab)) ptrepproto_setLabelStateAutomaton (vlab, pk_true); else *ok = pk_false;
	  return pk_true;
	}

      if (isInitialStateId (id))
	{
	  // Initial_State(lab1) : the input must be a label.
	  kl_getBinaryExpr (inputs1, &p1, &p2);
	  if (verifExprSigIsLabel (p1, p, &vlab)) ptrepproto_setLabelStateAutomaton (vlab, pk_true); else *ok = pk_false;
	  return pk_true;
	}

      if (isSetCurrentStateId (id))
	{
	  kl_getBinaryExpr (inputs1, &p1, &p2);
	  kl_setCurrentStateVarAutomaton (G_process, kl_getIdentifierExpr (p1));
	  return pk_true;
	}
      
      if (isAutomatonIsIn (id))
	{
	  // Automaton_IsIn(lab1): the input must be a label.
	  kl_getBinaryExpr (inputs1, &p1, &p2);
	  if (verifExprSigIsLabel (p1, p, &vlab)) ptrepproto_setLabelStateAutomaton (vlab, pk_true); else *ok = pk_false;
	  return pk_true;
	}

      if (isAutomatonClockId (id))  // Getting the clock of the automaton.
	{
	  return pk_true;
	}

      if (isAutomatonSetClockId (id))  // Automaton_SetClock(vh, Expr) fixing the name of the clock automaton and its definition.
	{
	  // kl_getBinaryExpr (inputs1, &p1, &p2); 
	  kl_getUnaryExpr(inputs1, &p1); 
	  // *ok = IsAnIdentifier(p1, p);
	  return pk_true;
	}
    }

  // Intrinsics used only in an automaton model.
  if (isTransitionId (id) || isInitialStateId (id) || isSetCurrentStateId (id) || isAutomatonIsIn (id) || isAutomatonClockId (id) || isAutomatonSetClockId (id))
    {
      mkklErrorAnnotTree (errIntrinsicAutomaton, p);
      *ok = pk_false;
    }

  return pk_false;
}

/**
   Verifies that the element of l are declared as state variable or shared signal.
   The operand tree (t) is annotated in case of error.
*/
void
Trees2Kernel::shouldBeStatevOrShared (ptlrepid l, boolean bpartAss, ntrees t)
{
  while (l != NILLREPID)
    {
      if (bpartAss)
	{
	  if (!isStateVariable (l) && !isSharedVariable (l))
	    mkklErrorAnnotTree (errSharedOrStatevarExpected, t);
	}
      else if (isStateVariable (l) || isSharedVariable (l))
	mkklErrorAnnotTree (errSharedOrStatevarNotExpected, t);

      l = succid (l);
    }
}

/**
   @return 
   - pk_true if a tree (p) is associated with a stack (identifier, parameter, type, model,...), in this case 
   its declaration  (vdecl) is set. Moreover, if the object referenced in vdecl is declared in a module, this module
   is (vmodule) also returned.
   - false otherwise, vdecl and vmodule are not modified.
*/
boolean Trees2Kernel::getTreeDecl (ntrees p, ntrees * vdecl, ntrees * vmodule, typid * proto, typval * IdClass)
{
  ptIdStack name;
  ptIdentProp topId;
  ptrepfi vfi, vfiModule;

  if (!locateOrError (p, &name))
    return pk_false;
  SET_NILTREE (*vmodule);
  // ref to a signal/param/constant identifier
  if (getSignalClassTop (name, &topId, IdClass))
    *vdecl = ptrepproto_getTree (ptIdentProp_getIdent (topId));
  // ref to a model
  else if (kl_getTopIfModel (name, &vfi))
    {
      *vdecl = kl_getTreeReference (vfi);
      topId = kl_top (name);
    }
  // ref to a named type
  else if (getTypeClassTop (name, &topId))
    *vdecl = kl_getTypeTreeDecl (ptIdentProp_getType (topId));
  else
    return pk_false;
  (*proto) = ptIdentProp_getIdClassType (topId);	//  protoIdent, protoModel, protoType;
  if (ptIdentProp_getModule (topId, &vfiModule))
    (*vmodule) = kl_getTreeReference (vfiModule);
  return pk_true;
}

/**
   @return "true" if the operarion succeeds, "false" otherwise.
   It produces the type (vtypid) of the no-th parameter of the process model (vfi). It is 
   - protoIdent for a identifier of a parameter
   - protoType for a type used as parameter
*/
_VIRTUAL_
  boolean
Trees2Kernel::getParameterTypeId (ptrepfi vfi, integer no, typid * vtypid)
{
  ptlrepid vlrepid;
  vlrepid = kl_getParametersModel (vfi);
  return ptlrepid_getIdTypeOf (vlrepid, no, vtypid);
}

/**
   @return "true" if vt is undefined object (nil tree, tree with opnil operator or opnullproc operator), "false" otherwise.
 */
boolean Trees2Kernel::isUndefinedBodyTree (ntrees vt)
{
  if (STree_isUndefinedTree (vt))
    return pk_true;
  return (noper (vt) == opnullproc);
}

/** 
    @return "true" if a tree (vt) is an empty system of equations, "false" otherwise.
    Assumed : vt is the tree of the body of a process model.
*/
boolean Trees2Kernel::isNullProc (ntrees vt)
{
  noperators
    op;
  int
    i,
    nb;

  if (STree_isUndefinedTree (vt))
    return pk_false;
  op = noper (vt);
  if (op == opnullproc)
    return pk_true;
  if (op != oplfeed)
    return pk_false;
  nb = ntreearity (vt);
  for (i = 1; i <= nb; i++)
    if (!isNullProc (nchild (i, vt)))
      return pk_false;
  return pk_true;
}

/** @return 
    - "true" if oper(p) is in { oplesign, oplepara, opleflot, opdefsig } or (opdefpart and oper(first child of (p))=opleflot)
    - "false" otherwise
*/
_VIRTUAL_
  boolean
Trees2Kernel::isInSet (ntrees p)
{
  noperators t;

  t = noper (p);
  return ((t == oplesign) || (t == oplepara) || (t == opleflot) || (((t == opdefsig) || (t == opdefpart)) && (noper (nchild (1, p)) == opleflot)));
}

/** the construction (p) is not yet implemented.
    An error is signaled.
*/
_VIRTUAL_ void
Trees2Kernel::notYetImplemented (ntrees p)
{
  mkklErrorAnnotTree (errnotimple, p);
}

/** A tree (p) is annotated with the errind-th error. */
_VIRTUAL_ void
Trees2Kernel::mkklErrorAnnotTree (integer errind, ntrees p)
{
  if (!GUIContext)
    pK_messenger->errorTree (errind, p);
}

/** A tree (p) is annotated with the errind-th error, parametrized by name. */
_VIRTUAL_ void
Trees2Kernel::mkklErrorAnnotTreeName (integer errind, ntrees p, ptIdStack name)
{
  mkklErrorAnnotTreeIdent (errind, p, identpile (name));
}


/** A tree (p) is annotated with the errind-th error, parametrized by name. */
_VIRTUAL_ void
Trees2Kernel::mkklErrorAnnotTreeIdent (integer errind, ntrees p, ptident name)
{
  if (!GUIContext)
    pK_messenger->errorTreeIdent (errind, p, name);
}


/** A tree (p) is annotated with the warind-th warning, parametrized by name. */
_VIRTUAL_ void
Trees2Kernel::mkklWarningAnnotTreeIdent (integer warind, ntrees p, ptident name)
{
  if (!GUIContext)
    pK_messenger->warningaccnom (warind, p, name);
}

/** A tree (p) is annotated with the warind-th warning, parametrized by name. */
_VIRTUAL_ void
Trees2Kernel::mkklWarningAnnotTreeName (integer warind, ntrees p, ptIdStack name)
{
  mkklWarningAnnotTreeIdent (warind, p, identpile (name));
}

/** A tree (p) is annotated with the n-th warning. */
_VIRTUAL_ void
Trees2Kernel::mkklWarningAnnotTree (integer n, ntrees p)
{
  if (!GUIContext)
    pK_messenger->warningacc (n, p);
}


/** It sends the n-th warning. */
_VIRTUAL_ void
Trees2Kernel::mkklWarning (integer n)
{
  if (!GUIContext)
    pK_messenger->warning (n);
}

/**
   It sends a warning to signal that "Safe is implemented as Unsafe".
*/
_VIRTUAL_ void
Trees2Kernel::mkklWarningSafeRestict ()
{
  if ((!GUIContext) && !bWarSafeRestict)
    {
      bWarSafeRestict = pk_true;
      pK_messenger->warning (warSafeRestict);
    }
}

/**
   It removes from a list (lmasquee) the elements of another list (linterne). It is used to detect the double declaration of the same identifier.
   When an element is removed, it attributes the tree (p) by an error message when isfatale, or a warning otherwise.
*/
_VIRTUAL_ void
Trees2Kernel::verifDoubleDec (ptlrepid * lmasquee, ptlrepid linterne, boolean isfatale, ntrees p)
{
  ptIdStack varid;

  while (linterne != NILLREPID)
    {
      varid = nameId (linterne);
      if (ptlrepid_remove_Unsorted (lmasquee, varid))
	{
	  if (isfatale)
	    mkklErrorAnnotTreeName (errdoubledecl, p, varid);
	  else
	    mkklWarningAnnotTreeName (wardbldec, p, varid);
	}
      linterne = succid (linterne);
    }
}


/** It calls verifDoubleDec() for lists of constants.  */
_VIRTUAL_ void
Trees2Kernel::verifDoubleDecConstantSelect (ptlrepid * lmasquee, ptlrepid linterne, boolean isfatale, ntrees p)
{
  verifDoubleDec (lmasquee, linterne, isfatale, p);
}


/**
   @return "true" if the top of the stack associated with varid is a conversion 
*/
_VIRTUAL_ boolean
Trees2Kernel::mkklConversionModel (ptIdStack varid, ptrepfi * vfi)
{
  ptlrepid vrepid, xlrepid, ins, outs;
  ptreptyp vreptyp;
  ptIdentProp topId;
  ptrepproto vrepproto;
  ptident id;

  if (!getTypeClassTop (varid, &topId))
    return pk_false;
  //---- A modifier: Creation d'un modele à chaque reference!!!! ------
  vreptyp = ptIdentProp_getType (topId);
  vrepid = NILLREPID;
  appendProtoTyp (&vrepid, varid, vreptyp, &xlrepid);
  (*vfi) = kl_mkTypeConversionModel (identpile (varid), vrepid);
  kl_setRefToIntrinsics (*vfi, moduleInstrinsics);	// utilise en deuxieme passe

  // input
  vrepproto = mkrepproto (kl_mkBasicType (opimplicit), NILEXPRSIG, undefCategory, NILTREE, NILTREE);
  ins = NILLREPID;
  appendProtoId (&ins, getidnil (), vrepproto, &xlrepid);
  // output
  vrepproto = mkrepproto (vreptyp, NILEXPRSIG, undefCategory, NILTREE, NILTREE);
  outs = NILLREPID;
  id = identpile (varid);
  id = createSuffIdent (id, string2repstr ("_"));
  appendProtoId (&outs, get_stackOf (id), vrepproto, &xlrepid);
  kl_setParamsInputsOutputs (*vfi, NILLREPID, ins, outs);
  kl_setBodyModel (*vfi, kl_mkExternal (NILTREE, STree_mkUndefinedExternalValue ()));
  return pk_true;
}

/** @return 
    - "true" if a process model (vfi) is the top of the stack associated with varid, the conversion are allowed (mkklConversionModel())
    - "false" otherwise, vfi is not modified.
*/
_VIRTUAL_ boolean
Trees2Kernel::getTopIfModel (ptIdStack varid, ptrepfi * vfi)
{
  if (kl_getTopIfModel (varid, vfi))
    return pk_true;
  // on accepte une conversion de type
  return mkklConversionModel (varid, vfi);
}


/**
   @return
   - "true" if p is the current top of the toppilefc stack or it is the NIL tree, and it pops it
   - "false" otherwise (it does not pop)
*/
_VIRTUAL_ boolean
Trees2Kernel::popFirstChild (ntrees p)
{
  pointer x;
  ntrees el;

  if (!stackTop (toppilefc, &x))
    return pk_false;
  el = (ntrees) x;
  if ((p == el) || (IS_NILTREE (p)))
    {
      toppilefc = stackPop (toppilefc);
      return pk_true;
    }
  return pk_false;
}


/**
   It pushes in toppilefc the operand tree (p).
*/
_VIRTUAL_ void
Trees2Kernel::pushFirstChild (ntrees p)
{
  toppilefc = stackPush (toppilefc, (ptElemStack) p);
}



/**
   @return 
   - "true" if a tree (p) has a non nil n-th child, \b operande (attribute member of the class) is set to this child, 
   and \b operator (attribute member of the class) is the operator of this child.
   - "false" otherwise, an error message is stick to p.
*/
_VIRTUAL_ boolean
Trees2Kernel::isChild (integer n, ntrees p)
{
  operande = nchild (n, p);
  if (IS_NILTREE (operande))
    {
      mkklErrorAnnotTree (errargmanq, p);
      voperator = opnil;
      return pk_false;
    }
  voperator = noper (operande);
  if (voperator == opnil)
    {
      mkklErrorAnnotTree (errargmanq, p);
      voperator = opnil;
      return pk_false;
    }
  return pk_true;
}


/**
   @return 
   - "true" if a tree (p) has a non nil n-th child, \b operande (attribute member of the class) is set to this child, 
   and \b operator (attribute member of the class) is the operator of this child. 
   - "false" otherwise, when the n-th child is the nil object, operande is set to it and  voperator to opnil
*/
_VIRTUAL_ boolean
Trees2Kernel::isoChild (integer n, ntrees p)
{
  operande = nchild (n, p);
  if (IS_NILTREE (operande))
    {
      voperator = opnil;
      return pk_false;
    }
  voperator = noper (operande);
  //  return (boolean) ((voperator != opnil) && (voperator != opnullproc)); nov 2011
  return (boolean) (voperator != opnil);
}

/**
   Assumed: p is a tree-list.
   - It defines \b operator(attribute member of the class) by the operator of operande (attribute member of the class)
   - it pushes operande (pushFirstChild())
   - and defines p by the right child of operande.
   @return "true" if the selected tree (operande) with operator as operator is defined (non null tree), "false" otherwise.
*/
_VIRTUAL_ boolean
Trees2Kernel::finalizeFirstChild (ntrees * p)
{
  voperator = noper (operande);
  pushFirstChild (operande);
  *p = ntrees_getNext (operande);
  return (boolean) (voperator != opnil);
}


/**
   @return 
   - "true" if the the first child of a tree-list (p) exists and it is not the nil tree object, in this case
      - \b operande (attribute member of the class) is defined by this element and 
      - \b operator (attribute member of the class) is the operator of this element
      .
   Then p is then set to the right child of the selected element
   - "false" otherwise.

   It calls finalizeFirstChild().
*/
_VIRTUAL_ boolean
Trees2Kernel::firstChild (ntrees * p)
{
  operande = nchild (1, *p);
  if (ISNOT_NILTREE (operande))
    return finalizeFirstChild (p);
  mkklErrorAnnotTree (errargmanq, *p);
  return pk_false;
}


/**
   Similar to firstChild(), but here the element of the list may be the null object (the list can be empty).
*/
_VIRTUAL_ boolean
Trees2Kernel::firstChild2 (ntrees * p)
{
  operande = nchild (1, *p);
  if (ISNOT_NILTREE (operande))
    return finalizeFirstChild (p);
  return pk_false;
}


/**
   Access to the first child of a sub-list (p).
   @return
   - "true" if the element exists, \b operande and \b voperator are defined by this element/its operator.
   p is modified: p = p - the (selected) first child.
   - "false" otherwise.
*/
_VIRTUAL_ boolean
Trees2Kernel::nextChild (ntrees * p)
{
  if (popFirstChild (*p))
    return pk_false;
  operande = *p;
  voperator = noper (operande);
  *p = ntrees_getNext (*p);
  return (boolean) (voperator != opnil);
}



/** @return the number of outputs for a list of expressions.
    (when an element is a reference to a model, it counts the number of outputs of the model).
 */
_VIRTUAL_ integer
Trees2Kernel::numberOutputsOfExprSig (ptklExprSig liste)
{
  integer nb;
  ptklExprSig op1;
  ptrepfi vfi;
  ptlrepid laux;

  nb = 0;
  while (liste != NILEXPRSIG)
    {
      kl_getBinaryExpr (liste, &op1, &liste);
      if (kl_getFlowExprOperator (op1) == opappe)
	{
	  if (getTopIfModel (nameId (kl_getFlowExprModelProcCallname (op1)), &vfi))
	    nb = nb + ptlrepid_size (kl_getInsOutsModel (vfi, &laux));
	  else
	    nb = nb + 1;
	}
      else
	nb = nb + 1;
    }
  return nb;
}

/**
   @return 
   - "true" if the operand tree(p) is a well formed identifier, then vpile is its associated stack,
   - "false" otherwise, vpile is not modified.
*/
_VIRTUAL_ boolean
Trees2Kernel::locate (ntrees p, ptIdStack * vpile)
{
  ptident identname;

  identname = copy_ident (p);
  if (identname == NILIDENT)
    return pk_false;
  // opnom, opident
  *vpile = get_stackOf (identname);
  return pk_true;
}

/**
   @copydoc locate()
   When it returns "false" an error message is stick to the tree.
*/
_VIRTUAL_ boolean
Trees2Kernel::locateOrError (ntrees p, ptIdStack * vpile)
{
  if (locate (p, vpile))
    {
      addPredTypeConst (*vpile);
      return pk_true;
    }
  mkklErrorAnnotTree (errargmanq, operande);
  return pk_false;
}


/**
   It calls locateOrError() for the n-th child of p.
*/
_VIRTUAL_ boolean
Trees2Kernel::locateChild (ntrees p, integer n, ptIdStack * vpile)
{
  if (isChild (n, p))
    return locateOrError (operande, vpile);
  return pk_false;
}

/**
   @return the stack associated with the name of a model (p), when the abstract tree is not well formed
   the stack associated with the nil identifier is returned.

   It calls locateOrError(), locateChild() to compute the result.
*/
_VIRTUAL_ ptIdStack
Trees2Kernel::locateModelName (ntrees p)
{
  ptIdStack varid;
  ptident id;
  integer posid;
  ntrees s;
  noperators so;

  posid = 1;
  if (noper (p) == opoper)
    {
      s = operande;
      so = voperator;		// contexte
      if (isChild (posid, p))
	if (voperator == opop)
	  {
	    id = repstr2ident (nstringof (operande));
	    if (locateOrError (id, &varid))
	      return varid;
	  }
      operande = s;
      voperator = so;		// contexte
    }
  if (locateChild (p, posid, &varid))
    return varid;
  id = identpile (getidnil ());
  return kl_mkIdStack (id);
}


/**
   @return 
   - "true" if the stack (Id) is not empty, and if the top element is an identifier of a signal/constant/index/parameter/label (typval type). 
   Then,
       - topId is the top of the stack
       - IdClass is the kind of identifier 
   - "false" otherwise, topId is the top of the stack; and IdClass is not modified.
*/
_VIRTUAL_ boolean
Trees2Kernel::getSignalClassTop (ptIdStack Id, ptIdentProp * topId, typval * IdClass)
{
  if (kl_isEmptyIdStack (Id))
    return pk_false;
  *topId = kl_top (Id);
  if (!ptIdentProp_equal (*topId, protoIdent))
    return pk_false;
  *IdClass = ptIdentProp_getIdClass (*topId);
  return pk_true;
}


/**
   @return "true" if the top element of a stack (Id) is an identifier of a constant. Then vproto is set (ptIdentProp_getIdent()).
   "false" otherwise, vproto is not modified.

   It applies getSignalClassTop() on Id, then verifies that the top element is an identifier of a constant.
*/
_VIRTUAL_ boolean
Trees2Kernel::getConstantClassTop (ptIdStack Id, ptIdentProp * topId)
{
  typval IdClass;

  if (!getSignalClassTop (Id, topId, &IdClass))
    return pk_false;
  if (IdClass != cste)
    return pk_false;
  /*  // virtual: only if it is declare in a module
     if (!ptIdentProp_getModule(topId, &vmodule)) return pk_false;
     *vproto =  ptIdentProp_getIdent(topId);
     mkklWarningAnnotTreeIdent(wardclinmodule, p,  kl_getNameModel(vmodule));
   */
  return pk_true;
}


/**
   @return "true" if the top element of a stack (Id) is an identifier of a label, "false" otherwise.
   
   It applies getSignalClassTop() on Id, then verifies that the top element is an identifier of a label.
*/
_VIRTUAL_ boolean
Trees2Kernel::isTopaLabel (ptIdStack Id)
{
  ptIdentProp topId;
  typval IdClass;

  if (!getSignalClassTop (Id, &topId, &IdClass))
    return pk_false;
  return (IdClass == eti);
}

/**
   @return "true" if the top element of a stack (Id) is an identifier of a type, "false" otherwise.
   
   topId is the top of the stack
*/
_VIRTUAL_ boolean
Trees2Kernel::getTypeClassTop (ptIdStack Id, ptIdentProp * topId)
{
  if (kl_isEmptyIdStack (Id))
    return pk_false;
  *topId = kl_top (Id);
  return ptIdentProp_equal (*topId, protoType);
}

/**
   @return "true" if the top element of a stack (Id) is an identifier of a model type (ie an interface), "false" otherwise.
   
   topId is the top of the stack
*/
_VIRTUAL_ boolean
Trees2Kernel::getModelTypeClassTop (ptIdStack Id, ptIdentProp * topId)
{
  if (kl_isEmptyIdStack (Id))
    return pk_false;
  *topId = kl_top (Id);
  return ptIdentProp_equal (*topId, protoModelType);
}


/**
   @return
   - "true" if the top element of a stack (Id) is an identifier of a type (vreptyp) defined in a module,
   - "false" otherwise, vreptyp is not modified.
*/
_VIRTUAL_ boolean
Trees2Kernel::getTypeClassTopInModule (ntrees p, ptIdStack Id, ptreptyp * vreptyp)
{
  ptIdentProp topId;
  ptrepfi vmodule;

  // UNUSED.  
  if (!getTypeClassTop (Id, &topId))
    return pk_false;
  if (!ptIdentProp_getModule (topId, &vmodule))
    return pk_false;
  *vreptyp = ptIdentProp_getType (topId);
  mkklWarningAnnotTreeIdent (wardclinmodule, p, kl_getNameModel (vmodule));
  return pk_true;
}

/** @return 
    -"true"  if the model pointed by p_decl is an external model defined in a module,
    the stack associated of the identifier (varid) is returned, and the associated model (vfiloc).
    -"false" otherwise, the output parameters are not modified.
*/
_VIRTUAL_ boolean
Trees2Kernel::getModelTop (ntrees p_decl, ptIdStack * varid, ptrepfi * vfiloc)
{
  ntrees corps;
  ptIdentProp topId;
  ptrepfi vmodule;

  //Il FAUDRA verifier les interfaces; et l'inclusion des types de modeles (process, node, function)
  corps = nchild (2, nchild (3, p_decl));
  if (!STree_isUndefinedTree (corps))
    if (noper (corps) != opexte)
      return pk_false;		// excluding null proc.
  *varid = locateModelName (p_decl);
  if (kl_isEmptyIdStack (*varid))
    return pk_false;
  topId = kl_top (*varid);
  if (!ptIdentProp_equal (topId, protoModel))
    return pk_false;
  if (!ptIdentProp_getModule (topId, &vmodule))
    return pk_false;
  *vfiloc = ptIdentProp_getProcess (topId);
  operande = p_decl;
  mkklWarningAnnotTreeIdent (wardclinmodule, p_decl, kl_getNameModel (vmodule));
  mkklWarningAnnotTree (warRestrictiondclinmodule, p_decl);
  return pk_true;
}

/**
   @return a new representation of an identifier  (p) (signal/parameter/indice/label), or NILEXPRSIG  when the identifier is not well formed.
   A new model is inserted (taking into account the
   top of the associated stack) in a (sorted) list:
    - les: when the identifier is a signal or a label
    - lparam: when the identifier is a parameter or a constant
    - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExpName (ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vklExprSig_res;
  ptIdStack name;
  ptIdentProp topId;
  typval IdClass;
  ptlrepid xlrepid;

  if (!locateOrError (p, &name))
    return NILEXPRSIG;
  if (!getSignalClassTop (name, &topId, &IdClass))
    {
      ptrepproto vrepproto = mkrepproto (kl_mkBasicType (opimplicit), NILEXPRSIG, undefCategory, operande, NILTREE);
      appendProtoId (les, name, vrepproto, &xlrepid);
      mkklErrorAnnotTreeName (errdecl, p, name);
      return NILEXPRSIG;
    }
  vklExprSig_res = kl_mkIdentifierExpr (mkRepidsigRef (name, topId), p);
  /* il y a quelque chose d'empile */
  switch ((int) IdClass)
    {
    case ind:
      ptlrepid_insert (indexList, name);
      break;
    case par:
    case cste:
      /* les constantes sont rendues dans lparam */
      ptlrepid_insert (lparam, name);
      break;
    case sig:
      ptlrepid_insert (les, name);
      break;
    case eti:
      /* les etiquettes sont rendues dans les */
      ptlrepid_insert (les, name);
      break;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return vklExprSig_res;
}

/**
   @return a new representation of a list of expressions (p), with ope as operator.   
   - it calls cExprSig() for each element of the list.
   - p is niltree at the end of the oepration.
   - the reached identifiers are added in the following (sorted) lists
     - les: when the identifier is a signal or a label,
     - lparam: when the identifier is a parameter or a constant,
     - indexList: when the identifier is an indice

     p is assumed to be on the first element of the list
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::creclist (ntrees * p, noperators ope, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig v_res;
  ptklExprSig vklExprSig1, vklExprSig2;
  ntrees x;

  if (!nextChild (p))
    return NILEXPRSIG;
  x = operande;
  vklExprSig1 = cExprSig (operande, les, lparam, indexList);
  if (vklExprSig1 == NILEXPRSIG)
    v_res = creclist (p, ope, les, lparam, indexList);
  else
    {
      vklExprSig2 = creclist (p, ope, les, lparam, indexList);
      if (vklExprSig2 == NILEXPRSIG)
	v_res = kl_mkListExpr (ope, vklExprSig1, x);
      else
	v_res = kl_mkInsertHeadList (ope, vklExprSig1, vklExprSig2, x);
    }
  return v_res;
}


/**
   Similar to creclist(): it handles the first element before calling creclist(), p is set to the first element before calling it.

   @return a new representation of a list of expressions (p), with ope as operator.   
   - it calls cExprSig() for each element of the list.
   - p is niltree at the end of the oepration.
   - the reached identifiers are added in the following (sorted) lists 
     - les: when the identifier is a signal or a label,
     - lparam: when the identifier is a parameter or a constant,
     - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExpList (ntrees * p, noperators ope, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vexp_res, vklExprSig1, vklExprSig2;
  ntrees x;

  if (firstChild (p))
    {
      x = operande;
      vklExprSig1 = cExprSig (operande, les, lparam, indexList);
      if (vklExprSig1 == NILEXPRSIG)
	vexp_res = creclist (p, ope, les, lparam, indexList);
      else
	{
	  vklExprSig2 = creclist (p, ope, les, lparam, indexList);
	  if (vklExprSig2 == NILEXPRSIG)
	    vexp_res = kl_mkListExpr (ope, vklExprSig1, x);
	  else
	    vexp_res = kl_mkInsertHeadList (ope, vklExprSig1, vklExprSig2, x);
	}
    }
  else
    vexp_res = NILEXPRSIG;
  return (vexp_res);
}

/** 
    E1 after E2, E1 from E2 are replaced by a reference to the model of their expansion.
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::expansionCounterFromAfter (noperators ope, ptklExprSig vklExprSig1, ptklExprSig vklExprSig2, ntrees p)
{
  ptlrepid inputs, outputs, l1, l2;
  ptklExprProc vklExprProc;
  ptident id;
  ptIdStack varid;
  ptklExprSig vklExprSig;

  vklExprSig = NILEXPRSIG;
  getNameAssociatedWithOperator (ope, &id);
  if (locateOrError (id, &varid))
    {
      if (verifCall (p, NILEXPRSIG, varid, &inputs, &outputs, &vklExprProc, &l1, &l2))
	{
	  vklExprSig = kl_mkInsertHeadList (oplesign, vklExprSig1, kl_mkListExpr (oplesign, vklExprSig2, p), p);
	  vklExprProc = kl_mkProcCall (inputs, outputs, NILLREPID, p, vklExprProc, vklExprSig);
	  vklExprSig = kl_mkProcCallExpr (vklExprProc, p);
	}
    }
  return vklExprSig;
}

/**
   @return a new representation of a binary expression (p) synchronous when typoper=oper2, or polychronous when typoper=oper2tp, with ope as operator.
   The clock operators (^+, ^* ^- ) are handled in cExprSigClockBinExpr() method.
   - It calls cExprSig() for each element of the expression.
   - the reached identifiers are added in the following (sorted) lists 
     - les: when the identifier is a signal or a label,
     - lparam: when the identifier is a parameter or a constant,
     - indexList: when the identifier is an indice
*/
_VIRTUAL_
  ptklExprSig Trees2Kernel::cExprSigBinExpr (cloper typoper, ntrees p, noperators ope, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig
    vklExprSig1,
    vklExprSig2;
  ptlrepid
    laux;

  if (isChild (1, p))
    vklExprSig1 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig1 = NILEXPRSIG;

  if (isChild (2, p))
    if ((ope == opelem) && outputMode)
      // For x[I]:= only x must appear in the output-list les.
      {
	laux = NILLREPID;
	vklExprSig2 = cExprSig (operande, &laux, lparam, indexList);
      }
    else
      vklExprSig2 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig2 = NILEXPRSIG;
  if ((vklExprSig1 == NILEXPRSIG) || (vklExprSig2 == NILEXPRSIG))
    mkklErrorAnnotTree (errargmanq, p);
  if ((ope == opnomba) || (ope == opnombd))
    return expansionCounterFromAfter (ope, vklExprSig1, vklExprSig2, p);

  return kl_mkBinaryExpr (typoper, ope, vklExprSig1, vklExprSig2, p);
}


/**
   @return a new representation of a binary expression (p) synchronous when typoper=oper2, or polychronous when typoper=oper2tp, 
   with ope an clock operator { ^+, ^* ^- }.
   - It calls cExprSig() for each element of the expression.
   - the reached identifiers are added in the following (sorted) lists 
     - les: when the identifier is a signal or a label,
     - lparam: when the identifier is a parameter or a constant,
     - indexList: when the identifier is an indice
*/
_VIRTUAL_
  ptklExprSig
Trees2Kernel::cExprSigClockBinExpr (ntrees p, noperators ope, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vexp, vklExprSig1, vklExprSig2, vexp_res;

  if (isChild (1, p))
    vklExprSig1 = kl_mkUnaryExpr (oper1tp, opevnt, cExprSig (operande, les, lparam, indexList), operande);
  else
    vklExprSig1 = NILEXPRSIG;
  if (isChild (2, p))
    vklExprSig2 = kl_mkUnaryExpr (oper1tp, opevnt, cExprSig (operande, les, lparam, indexList), operande);
  else
    vklExprSig2 = NILEXPRSIG;
  if ((vklExprSig1 == NILEXPRSIG) || (vklExprSig2 == NILEXPRSIG))
    mkklErrorAnnotTree (errargmanq, p);
  switch ((int) ope)
    {
    case opcomp:
      // TODO :should be kept for the second pass.
      vexp = kl_mkUnaryExpr (oper1, opnon, vklExprSig2, operande);
      vexp_res = kl_mkUnaryExpr (oper1tp, opwhen1, kl_mkBinaryExpr (oper2tp, opmele, vexp, vklExprSig1, p), p);
      break;
    case opbinf:
      vexp_res = kl_mkBinaryExpr (oper2tp, opextr, vklExprSig1, vklExprSig2, p);
      break;
    case opbsup:
      vexp_res = kl_mkBinaryExpr (oper2tp, opmele, vklExprSig1, vklExprSig2, p);
      break;
    default:
      notYetImplemented (p);
    }
  return (vexp_res);
}

/**
   @return "true" if the call to a model (m) is correct for an effective parameter list (eParam), "false" otherwise.
*/
_VIRTUAL_ boolean
Trees2Kernel::verifParameters (ptrepfi m, ptklExprSig eParam, ntrees p)
{
  ptlrepid fParam;
  boolean bok;
  ptklExprSig aParam;
  cloper clexp;
  noperators vop1, vop2;

  fParam = kl_getParametersModel (m);
  bok = pk_true;
  while ((fParam != NILLREPID) && bok)
    if (eParam == NILEXPRSIG)
      {
	mkklErrorAnnotTree (errparam, p);
	bok = pk_false;
      }
    else
      {
	aParam = kl_getNextExpr (&eParam);
	if (aParam != NILEXPRSIG)
	  {
	    clexp = kl_getFlowExprClassType (aParam);
	    if (isTypeId (fParam))
	      bok = (clexp == opertyp);
	    else		// ident (formal process, constant,...)
	      {
		bok = (clexp != opertyp);
		if (bok && isModelId (fParam))
		  {
		    vop1 = kl_getInterfaceOperator (kl_getInterfaceModel (getModelRepid (fParam)));
		    vop2 = kl_getInterfaceOperator (kl_getInterfaceModel (getModelRepid (kl_getIdentifierExpr (aParam))));
		    if (vop1 != vop2)
		      mkklErrorAnnotTree (errConflictEffModelFormalModel, p);
		  }
	      }
	  }
	if (!bok)
	  mkklErrorAnnotTree (errUnexpectedExpr, p);
	else
	  fParam = succid (fParam);
      }

  return bok;
}

/**
   @return
   - "true" if the model (varid) is declared, ie the stack (varid) is not empty and the top is a model of a process.
   In this case, it verifies the coherence between the effective and the formal parameters (verifParameters())
      - inputs: a copy of the inputs of the model
      - outputs: a copy of the outputs of the model
      - lexternalsOuts a copy of the (external) outputs of the model (ie not declared in the inteface of the model)
      - lexternalsIns: a copy of the (external) inputs of the model (ie not declared in the inteface of the model)
      - vklExprProc: a reference to the model (varid) with effParam as parameters, inputs and outputs (kl_mkMacro()).

   - "false" otherwise, the ouput parameters are not modified.
*/
_VIRTUAL_
  boolean
  Trees2Kernel::verifCall (ntrees p, ptklExprSig effParam, ptIdStack varid,
			   ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc, ptlrepid * lexternalsIns, ptlrepid * lexternalsOuts)
{
  ptrepfi vfi;
  ptlrepid xlrepid, lins;

  if (!getTopIfModel (varid, &vfi))
    {
      mkklErrorAnnotTreeName (errprocnondec, p, varid);
      return pk_false;
    }
  verifParameters (vfi, effParam, p);
  *outputs = insertunsort (kl_getInsOutsModel (vfi, &lins), NILLREPID);
  *lexternalsOuts = kl_getExternalReferencesOut (vfi);

  *inputs = insertunsort (lins, NILLREPID);
  *lexternalsIns = kl_getExternalReferencesIn (vfi);

  *vklExprProc = kl_mkMacro (*inputs, *outputs, p, varid, vfi, effParam);
  if (!GUIContext)
    appendrepfi (&G_listOfReferencedModels, varid, vfi, &xlrepid);
  return pk_true;
}

/** 
    @return a new representation of a conversion expression (p).
    - It produces a procedure call to a name given by predefinedType2ident() function.
    - the reached identifiers are added in the following (sorted) lists 
       - les: when the identifier is a signal or a label,
       - lparam: when the identifier is a parameter or a constant,
       - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExpConversion (ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vklExprSig;
  ptlrepid inputs, lparameters, outputs, l1, l2;
  ptklExprProc vklExprProc;
  ptIdStack varid;
  ntrees x;

  if (isChild (1, p))
    locateOrError ((ntrees) repstr2ident (predefinedType2ident (voperator)), &varid);	// type predefini
  inputs = NILLREPID;
  lparameters = NILLREPID;
  if (isChild (2, p))
    {
      x = operande;
      vklExprSig = kl_mkListExpr (oplesign, cExprSig (operande, les, lparam, indexList), x);
    }
  else
    vklExprSig = NILEXPRSIG;
  if (verifCall (p, NILEXPRSIG, varid, &inputs, &outputs, &vklExprProc, &l1, &l2))
    {
      vklExprProc = kl_mkProcCall (inputs, outputs, lparameters, p, vklExprProc, vklExprSig);
      return kl_mkProcCallExpr (vklExprProc, p);
    }
  return NILEXPRSIG;
}

/**
   @return a list of expressions from a list of descriptor of identifiers (list).
   p is used for annotations (errors, warnings).
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::getCurrentIdent (ptlrepid list, ntrees p)
{
  ptklExprSig vexp, v_res, vklExprSig1;
  ptIdentProp topId;
  ptIdStack name;

  if (list == NILLREPID)
    return NILEXPRSIG;
  vklExprSig1 = getCurrentIdent (succid (list), p);
  name = nameId (list);
  if (kl_isEmptyIdStack (name))
    {
      mkklErrorAnnotTreeName (errdecl, p, name);
      vexp = NILEXPRSIG;
    }
  else
    {
      topId = kl_top (name);
      if (!ptIdentProp_IdClassEqual (topId, sig))
	mkklErrorAnnotTreeName (errconflid, p, name);
      vexp = kl_mkIdentifierExpr (mkRepidsigRef (name, topId), p);
    }
  if (vklExprSig1 == NILEXPRSIG)
    v_res = kl_mkListExpr (opleflot, vexp, p);
  else
    v_res = kl_mkInsertHeadList (opleflot, vexp, vklExprSig1, p);
  return v_res;
}


/**
   @return a copy of the elements of list selecting the identifier declared as signals.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::getCurrentIdentRepids (ptlrepid list)
{
  ptlrepid vl;
  ptIdentProp topId;
  ptIdStack name;

  if (list == NILLREPID)
    return NILLREPID;
  vl = getCurrentIdentRepids (succid (list));
  name = nameId (list);
  if (kl_isEmptyIdStack (name))
    return vl;
  topId = kl_top (name);
  if (ptIdentProp_IdClassEqual (topId, sig))
    vl = ptlrepid_add_elem (vl, mkRepidsigRef (name, topId));
  return vl;
}

/**
   Compiling instances of processes (p) such that P{n1, ..., n2} where n1...n2 are parameters.
   @return "true" if the compiling succeeds, "false" otherwise (expression not well formed).
   It produces 
     - vklExprProc the kernel representation of an instance (procedure call)
     - inputs the list of the inputs of the model of p
     - outputs the list of the outputs of the model of p
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpRefToModel (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ptklExprSig klExprSigOut, vklExprSig;
  ptIdStack varid;
  ptrepfi vfi;
  ptlrepid outs, ins, lparam, indexList, lexternalIns, lexternalOuts, ls, le;
  integer n;
  boolean b;

  lparam = NILLREPID;
  indexList = NILLREPID;
  if (!cExpRefToModelloc (p, inputs, &lparam, &indexList, outputs, vklExprProc, &varid, &lexternalIns, &lexternalOuts))
    return pk_false;
  getTopIfModel (varid, &vfi);
  outs = kl_getInsOutsModel (vfi, &ins);

  *vklExprProc = kl_mkProcCall (*inputs, *outputs, lparam, p, *vklExprProc, getCurrentIdent (ins, p));
  klExprSigOut = getCurrentIdent (outs, p);
  *outputs = getCurrentIdentRepids (outs);
  *inputs = getCurrentIdentRepids (ins);

  n = kl_getSizeExpr (klExprSigOut);
  if (n != 0)
    {
      ls = *outputs;
      le = *inputs;
      vklExprSig = kl_mkProcCallExpr (*vklExprProc, p);
      if (n > 1)
	*vklExprProc = kl_mkEquation (opdefsig, le, ls, lparam, p, klExprSigOut, vklExprSig);
      else
	*vklExprProc = kl_mkEquation (opdefsig, le, ls, lparam, p, kl_getNextExpr (&klExprSigOut), vklExprSig);
    }
  b = pk_false;
  if (lexternalIns != NILLREPID)
    {
      *inputs = insertunsort (getCurrentIdentRepids (lexternalIns), *inputs);
      b = pk_true;
    }
  if (lexternalOuts != NILLREPID)
    {
      *outputs = insertunsort (getCurrentIdentRepids (lexternalOuts), *outputs);
      b = pk_true;
    }

  if (b)
    {
      diffLists (*outputs, inputs);
      klset_inout (*vklExprProc, *inputs, *outputs);
    }
  return pk_true;
}

/**
   It finalyzes the compiling of a reference to a model (x1,...,xn):= P {p1,...,pp} (i1,...,im)
   with 
   - lparam= {p1,...,pp} the parameters
   - inputs= (i1,...,im) the inputs
   - outputs= (x1,...,xn) the outputs
  
   @return from vklExprProc defined by P {p1,...,pp} (i1,...,im),
   - an equation if the list of outputs is not empty
   - vklExprProc otherwise.
*/
_VIRTUAL_ ptklExprProc
Trees2Kernel::cExpCallFinalyze (ntrees p, ptlrepid lparam, ptlrepid inputs, ptlrepid outputs, ptklExprProc vklExprProc)
{
  integer n;
  ptklExprSig klExprSigOut, vklExprSig;

  klExprSigOut = getCurrentIdent (outputs, p);
  n = kl_getSizeExpr (klExprSigOut);
  if (n == 0)
    return vklExprProc;
  vklExprSig = kl_mkProcCallExpr (vklExprProc, p);
  if (n > 1)
    return kl_mkEquation (opdefsig, inputs, outputs, lparam, p, klExprSigOut, vklExprSig);
  return kl_mkEquation (opdefsig, inputs, outputs, lparam, p, kl_getNextExpr (&klExprSigOut), vklExprSig);
}


/**
   @return a new representation of a "IF" expression (p)
   The reached identifiers are added in the following (sorted) lists 
       - les: when the identifier is a signal or a label,
       - lparam: when the identifier is a parameter or a constant,
       - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExpIf (ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vklExprSig1, vklExprSig2, vklExprSig3;

  if (isChild (1, p))
    vklExprSig1 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig1 = NILEXPRSIG;
  if (isChild (2, p))
    vklExprSig2 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig2 = NILEXPRSIG;
  if (isChild (3, p))
    vklExprSig3 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig3 = NILEXPRSIG;
  return kl_mkTernaryExpr (oper3, opcond, vklExprSig1, vklExprSig2, vklExprSig3, p);
}


/**
   @return a new representation of an access to a value of a enum type ("\#rouge" or "T\#rouge")  expression (p)
   The reached identifiers are added in the following (sorted) lists 
       - les: when the identifier is a signal or a label,
       - lparam: when the identifier is a parameter or a constant,
       - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExpAccessEnumValue (ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ntrees p_typ, p_val;
  ptklExprSig vres, v1, v2;
  ptIdStack name;

  p_typ = nchild (1, p);
  p_val = nchild (2, p);
  v1 = NILEXPRSIG;
  if (!STree_isUndefinedTree (p_typ))	/*  ex: T#rouge */
    {
      locateOrError (p_typ, &name);	/* on verifie que le type existe */
      if (kl_isTopaType (name))
	v1 = kl_mkIdentifierExpr (mkRepidsigRef (name, kl_top (name)), p_typ);
      else
	mkklErrorAnnotTreeName (errtypnondec, p_typ, name);
    }
  v2 = kl_mkConstExpr (opcstcha, nstringof (p_val), p_val);
  vres = kl_mkBinaryExpr (oper2, opvenum, v1, v2, p);
  return (vres);
}

/**
   @return 
   - "true" if it exists a model (vfi) associated with an identifier given by p. p must be in positional form, ie foo{}() or foo().
   - "false" otherwise.
*/
_VIRTUAL_ boolean
Trees2Kernel::locateCalledProcess (ntrees p, ptrepfi * vfi)
{
  ntrees saveOperande, arbreaux;
  noperators savevoperator;
  boolean b;
  ptIdStack name;

  saveOperande = operande;
  savevoperator = voperator;

  b = isoChild (1, p);
  if (b)
    {
      if (voperator == opinst)
	arbreaux = operande;
      else
	arbreaux = p;
      locateChild (arbreaux, 1, &name);
      b = getTopIfModel (name, vfi);
    }
  operande = saveOperande;
  voperator = savevoperator;
  return b;
}


/**
   @return 
   - "true" if vfi is not undefined and its number of outputs is equal to nbOutCall
   - "false" otehrwise, an error is sticked to where.
*/
_VIRTUAL_ boolean
Trees2Kernel::verifCalledProcess (ptrepfi vfi, integer nbOutCall, ntrees where)
{
  ptlrepid vlaux;

  if (vfi == NILREPFI)
    mkklErrorAnnotTree (errprocnondec, where);
  else
    {
      if (ptlrepid_size (kl_getInsOutsModel (vfi, &vlaux)) == nbOutCall) return pk_true;
      ptident name= kl_getNameModel (vfi);
      if (isIndicesId (name)) return pk_true;
      if (isRendezVousId(name)) return pk_true;
      mkklErrorAnnotTree (errdefmul2, where);
    }
  return pk_false;
}

/**
   @return a new representation of a ternary expression (p) synchronous when clop=oper3, or polychronous when clop=oper3tp, 
   with ope as operator.
   - It calls cExprSig() for each child of the expression.
   - the reached identifiers are added in the following (sorted) lists 
     - les: when the identifier is a signal or a label,
     - lparam: when the identifier is a parameter or a constant,
     - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExprSigTernary (ntrees p, noperators ope, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList, cloper clop)
{
  ptklExprSig vklExprSig1, vklExprSig2, vklExprSig3;

  if (isChild (1, p))
    vklExprSig1 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig1 = NILEXPRSIG;
  if (isChild (2, p))
    vklExprSig2 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig2 = NILEXPRSIG;
  if (isoChild (3, p))
    vklExprSig3 = cExprSig (operande, les, lparam, indexList);
  else
    vklExprSig3 = NILEXPRSIG;
  if ((vklExprSig1 == NILEXPRSIG) || (vklExprSig2 == NILEXPRSIG))
    {
      mkklErrorAnnotTree (errargmanq, p);
      return NILEXPRSIG;
    }
  return kl_mkTernaryExpr (clop, ope, vklExprSig1, vklExprSig2, vklExprSig3, p);
}

/**
   @return a new representation of an expression (p) used as left part of an equation. 
   It pushes the outputMode (attribute member of the class), then it sets it to pk_true before calling cExprSig().
   It restores the output mode context.
   The reached identifiers are added in the following (sorted) lists 
     - les: when the identifier is a signal or a label,
     - lparam: when the identifier is a parameter or a constant,
     - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExprSigLeftExpr (ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  boolean b;
  ptklExprSig e;

  b = outputMode;
  outputMode = pk_true;
  e = cExprSig (p, les, lparam, indexList);
  outputMode = b;
  return e;
}


/**
   @return a new representation of an expression (p) on signals/parameters/constants/indexes (recursively computed).
   The reached identifiers are added in the following (sorted) lists 
     - les: when the identifier is a signal or a label,
     - lparam: when the identifier is a parameter or a constant,
     - indexList: when the identifier is an indice
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExprSig (ntrees exprTree, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vklExprSig1, vklExprSig2, vklExprSig3, vexpres;
  ptklExprProc vklExprProc1;
  ptlrepid l1, l2, letiq;
  noperators opecourant;
  ptIdStack name;
  ptrepfi vfi;
  boolean b /*, bdepiler */ ;
  ptrepproto vrepproto;
  ptIdentProp topId;
  ntrees x;

  if (IS_NILTREE (exprTree))
    return NILEXPRSIG;
  opecourant = noper (exprTree);
  x = operande;
  switch ((int) (opecourant))
    {
      /* Nullary operators
         ----------------- */
    case opcstent:
    case opcstdp:
    case opcstreel:
    case opcstcar:		/* example:   'a'     */
    case opcstcha:		/* example:  "toto"   */
    case opexte:		/* "external" or "external <string>" */
      vexpres = kl_mkConstExpr (opecourant, nstringof (exprTree), exprTree);
      break;
    case opfalse:
    case optrue:
      vexpres = kl_mkConstExpr (opecourant, nilstr, exprTree);
      break;
    case opretun:
      vexpres = kl_mkConstExpr (opcstent, char2repstr ('1'), exprTree);
      break;
    case opnil:
      vexpres = NILEXPRSIG;
      break;
      /* Unary expressions
         ----------------- */
    case opnom:
    case opident:
      vexpres = cExpName (exprTree, les, lparam, indexList);
      break;

    case opevnt:
    case opwhen1:
      if (isChild (1, exprTree))
	{
	  l1 = NILLREPID;
	  l2 = NILLREPID;
	  vexpres = kl_mkUnaryExpr (oper1tp, opecourant, cExprSig (operande, les, &l1, &l2), exprTree);
	}
      else
	vexpres = NILEXPRSIG;
      break;
    case opvvarin:		// x? (x is a state variable)

    case optran:		//  "tr" <expr_composite> 
    case opnega:		// - Expr
    case opnon:		        // not Expr
    case oppcar:		// cartesian product: <produit_index> << l_expr_signal >> 
      if (isChild (1, exprTree))
	vexpres = kl_mkUnaryExpr (oper1, opecourant, cExprSig (operande, les, lparam, indexList), exprTree);
      else
	vexpres = NILEXPRSIG;
      break;
    case ophnull:
      vexpres = kl_mkUnaryExpr (oper1tp, opecourant, NILEXPRSIG, exprTree);
      break;
      
    case opsdec:		// <fact_flow> [-]
      if (isChild (1, exprTree))
	{
	  if (OPIS_SIGNALIDENTIFIER (voperator)) //  || (voperator==opcham)
	    vexpres = kl_mkUnaryExpr (oper1, opecourant, cExprSig (operande, les, lparam, indexList), exprTree);
	  else
	    {
	      notYetImplemented (operande);
	      vexpres = NILEXPRSIG;
	    }
	}
      break;

      /* Binary expressions
         ------------------ */
    case opcham:		/* acces au field d'un flow de type structure:  A.x */
      if (isChild (1, exprTree))
	{
	  if (OPIS_SIGNALIDENTIFIER (voperator) || (voperator == opcham) || (voperator == opelem) || (voperator == opvvarin))
	    vklExprSig1 = cExprSig (operande, les, lparam, indexList);
	  else
	    {
	      mkklErrorAnnotTree (errnosens, exprTree);
	      vklExprSig1 = NILEXPRSIG;
	    }
	}
      // if (locateOrError(nchild(2, p), &name)) vklExprSig2 = kl_mkIdentifierExpr(mkRepidsigRef(name, kl_top(name)));
      vklExprSig2 = kl_mkConstExpr (opnom, ident2repstr (nchild (2, exprTree)), nchild (2, exprTree));
      vexpres = kl_mkBinaryExpr (oper2, opcham, vklExprSig1, vklExprSig2, exprTree);
      break;

    case opinst:
      vexpres = NILEXPRSIG;
      mkklErrorAnnotTree (errprocesscall, operande);
      break;

    case opappe:
      if (isInSet (nfather (exprTree)))
	b = pk_true;
      else if (locateCalledProcess (exprTree, &vfi))
	b = verifCalledProcess (vfi, 1, exprTree);
      else
	b = pk_false;

      operande = exprTree;
      voperator = noper (exprTree);

      vexpres = NILEXPRSIG;
      letiq = kl_getLabelsModel (G_process);
      l1 = NILLREPID;
      l2 = NILLREPID;
      if (klparcours (&l1, &l2, &vklExprProc1))
	if (b)
	  {
	    vexpres = kl_mkProcCallExpr (vklExprProc1, exprTree);
	    b = intrinsicSettings (exprTree, NILEXPRSIG, vexpres);
	  }

      *les = insertlist (l1, *les);
      break;
    case opconv:		/* Conversion de type  <type predefini>( Expression ) */
      vexpres = cExpConversion (exprTree, les, lparam, indexList);
      break;
    case opextr:
    case opmele:
    case opnombm:		/* counter modulo */
    case opnomba:		/* counter after. */
    case opnombd:		/* counter from. */
      vexpres = cExprSigBinExpr (oper2tp, exprTree, opecourant, les, lparam, indexList);
      break;

    case opcomp:		/* h1 ^- h2 */
    case opbinf:		/* h1 ^* h2 */
    case opbsup:		/* h1 ^+ h2 */
      vexpres = cExprSigClockBinExpr (exprTree, opecourant, les, lparam, indexList);
      break;
    case opegeg:		/*  ==  */
    case opiieg:		/*  <<= */
    case opmod:		/* modulo */
    case oprecu:		/* recuperation "\\" */
    case opconc:
    case opcplx:
    case opdiff:
    case opegal:
    case opiegal:
    case opinfe:
    case opsegal:
    case opsupe:
    case opet:
    case opou:
    case opouex:
    case opdivi:
    case opmoin:
    case opmult:
    case opmmul:		// Matrix Mult.
    case opplus:
    case oppuis:
    case opelem:
    case opelevec:
    case oprepe:		// e1 |* e2 , repeat e1 e2 times
    case opsequ:		// e1 next e2 (sequential definition)
      vexpres = cExprSigBinExpr (oper2, exprTree, opecourant, les, lparam, indexList);
      break;
    case opintvff:		/* enumerations du case: intervalles */
    case opintvfo:
    case opintvof:
    case opintvoo:
      vexpres = cExprSigBinExpr (oper2, exprTree, opecourant, les, lparam, indexList);
      notYetImplemented (exprTree);
      break;
    case oprest:
      if (noper (nfather (exprTree)) != opensvec)
	vexpres = cExprSigBinExpr (oper2, exprTree, opecourant, les, lparam, indexList);
      else			/* cas du oprest a l'interieur d'un opensvec: on le traite comme un opelevec */
	{
	  if (isChild (1, exprTree))
	    vklExprSig1 = cExprSig (operande, les, lparam, indexList);
	  else
	    vklExprSig1 = NILEXPRSIG;
	  if (isChild (2, exprTree))
	    vklExprSig2 = cExprSig (operande, les, lparam, indexList);
	  else
	    vklExprSig2 = NILEXPRSIG;
	  if ((vklExprSig1 == NILEXPRSIG) || (vklExprSig2 == NILEXPRSIG))
	    mkklErrorAnnotTree (errargmanq, exprTree);
	  vexpres = kl_mkBinaryExpr (oper2, opelevec, vklExprSig1, vklExprSig2, exprTree);
	}
      break;
    case opensvec:
      if (isChild (1, exprTree))
	{
	  lindex = NILLREPID;
	  vklExprSig1 = cExprSig (operande, les, lparam, indexList);
	  if (isChild (2, exprTree))
	    vklExprSig2 = cExprSig (operande, les, lparam, indexList);
	  else
	    vklExprSig2 = NILEXPRSIG;
	  popFlows (lindex);
	  diffLists (lindex, indexList);
	  if ((vklExprSig1 == NILEXPRSIG) || (vklExprSig2 == NILEXPRSIG))
	    vexpres = NILEXPRSIG;
	  else
	    vexpres = kl_mkBinaryExpr (oper2, opensvec, vklExprSig1, vklExprSig2, exprTree);
	}
      else
	{
	  if (isChild (2, exprTree))
	    vklExprSig2 = cExprSig (operande, les, lparam, indexList);
	  vexpres = NILEXPRSIG;
	}
      break;
    case opparc:
      // bdepiler = pk_false;
      if (isoChild (1, exprTree))
	if (locateOrError (operande, &name))
	  {
	    vrepproto = mkrepproto (kl_mkBasicType (opinteger), NILEXPRSIG, undefCategory, operande, NILTREE);
	    topId = mkIdentProp (vrepproto);
	    kl_pushIndex (name, topId);	// depilé via lindex
	    vklExprSig1 = kl_mkIdentifierExpr (mkRepidsigRef (name, topId), operande);
	    ptlrepid_insert (&lindex, name);
	    letiq = kl_getLabelsModel (G_process);
	    if (ptlrepid_isIn_Unsorted (letiq, name))
	      mkklErrorAnnotTreeName (errconflid, operande, name);
	    // bdepiler = pk_true;
	  }
	else
	  vklExprSig1 = NILEXPRSIG;
      else
	vklExprSig1 = NILEXPRSIG;
      if (isChild (2, exprTree))
	vklExprSig2 = cExprSig (operande, les, lparam, indexList);
      else
	vklExprSig2 = NILEXPRSIG;
      vexpres = kl_mkBinaryExpr (oper2, opparc, vklExprSig1, vklExprSig2, exprTree);
      break;

    case opvenum:		/* acces aux valeurs d'un type enumere: #rouge ou T#rouge  */
      vexpres = cExpAccessEnumValue (exprTree, les, lparam, indexList);
      break;

    case opvar:		// var E init E0
      if (isChild (1, exprTree))
	vklExprSig1 = cExprSig (operande, les, lparam, indexList);
      else
	vklExprSig1 = NILEXPRSIG;
      if (isoChild (2, exprTree))
	vklExprSig2 = cExprSig (operande, les, lparam, indexList);
      else
	vklExprSig2 = NILEXPRSIG;
      vexpres = kl_mkBinaryExpr (oper2tp, opecourant, vklExprSig1, vklExprSig2, exprTree);
      break;

      /* The lists */
    case opchai:		/* list of strings */
    case opleflot:
    case oplesign:
    case oplepara:
    case oplevect:
    case oplelevec:
    case oplparc:
    case oplsign:
    case oplcrohor:
    case opldechor:
    case oplequhor:
    case oplexchor:
    case oplident:
    case oplopra:		// objects assigned to a pragma:  name {id1, ..., idn} text
      vexpres = cExpList (&exprTree, opecourant, les, lparam, indexList);
      break;

      /* Ternary operators
         ----------------- */
    case opborn:
      if (isoChild (1, exprTree))
	vklExprSig1 = cExprSig (operande, les, lparam, indexList);
      else
	vklExprSig1 = NILEXPRSIG;
      if (isoChild (2, exprTree))
	vklExprSig2 = cExprSig (operande, les, lparam, indexList);
      else
	vklExprSig2 = NILEXPRSIG;
      if (isoChild (3, exprTree))
	vklExprSig3 = cExprSig (operande, les, lparam, indexList);
      else
	vklExprSig3 = NILEXPRSIG;
      vexpres = kl_mkTernaryExpr (oper3, opborn, vklExprSig1, vklExprSig2, vklExprSig3, x);
      break;
    case opmemo:
    case opret:
    case opfen:
      vexpres = cExprSigTernary (exprTree, opecourant, les, lparam, indexList, oper3tp);
      break;
    case openum:		// ".." <term_index> "step" <term_index> ;
      vexpres = cExprSigTernary (exprTree, opecourant, les, lparam, indexList, oper3);
      break;

    case opcond:
      vexpres = cExpIf (exprTree, les, lparam, indexList);
      break;

    default:
      notYetImplemented (operande);
    }
  return vexpres;
}



/**
   Compiling an expression of synchronisation (p) , vop in {^<, ^>, ^=, ^#}.
   - it produces a new representation (vklExprProc)  of the expression of synchronisation (kl_mkClockOp()).
   - the reached identifiers are added in the following (sorted) lists 
       - inputs: when the identifier is a signal or a label,
       - outputs: always an empty list,
       - lparam: when the identifier is a parameter or a constant,
       - indexList: when the identifier is an indice.
       
   @return "true" if the compiling succeeds, "false" otherwise (not well formed expression).
*/
_VIRTUAL_
  boolean
  Trees2Kernel::cExpSynchro (ntrees p, noperators vop, ptlrepid * inputs, ptlrepid * outputs,
			     ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  ptklExprSig vklExprSig;

  *inputs = NILLREPID;
  *lparam = NILLREPID;
  *indexList = NILLREPID;
  vklExprSig = cExprSig (operande, inputs, lparam, indexList);
  if (vklExprSig == NILEXPRSIG)
    return pk_false;
  *vklExprProc = kl_mkClockOp (*inputs, p, vklExprSig, vop);
  *outputs = NILLREPID;
  return pk_true;
}

/**
   Verifications for an expression (p) that reference an instrinsic model.
   The "kernel" representation of p is given by e1:=e2, when e1 are the list of outputs and e2 the reference to the model.
   \code
   function indices=(?integer x0, x1, x2,x3,x4, x5,x6, x7, x8, x9;  !o0, o1, o2,o3,o4, o5,o6, o7, o8,o9)
      pragmas 
      COMMENT " WARNING: input is list of integers. output is a list of indices with same cardinality of the inputs."
     end pragmas ;	
     \endcode
   @return 
   - "false" when p is a reference to the intrinsic model "indices" and the number of inputs is not equal to the number of outputs
   - "true" otherwise. When it is a reference to "indices", a new model is created to take into account the number (variable) of 
   effective inputs (= number of outputs). The reference to t he model is modified in e2 expression.
*/
_VIRTUAL_ boolean
Trees2Kernel::intrinsicSettings (ntrees p, ptklExprSig e1, ptklExprSig e2)
{
  ptklExprProc x;
  ptklExprSig eff;
  ntrees xtree;
  ptlrepid xoutputs, xinputs, l1, l2;
  integer i, nbsigdef;
  ptrepfi vfi, vfi1;
  ptIdStack varid;
  ptident id;

  if (IS_NILTREE (p)) return pk_true;
  if (noper (p) != opappe) return pk_true;
  if (!locateCalledProcess (p, &vfi)) return pk_true;
  if (!kl_isIntrinsicModel (vfi)) return pk_true;

  id = kl_getNameModel (vfi);
  if (isIndicesId (id))
    {
      x = kl_getFlowExprModelProcCall (e2);
      /* ptklExprProc y = */ kl_getModelProcCall (x, &eff, &xtree);
      nbsigdef = kl_getSizeExpr (eff);
      if (e1 != NILEXPRSIG)
	if (kl_getSizeExpr (e1) != nbsigdef)
	  {
	    mkklErrorAnnotTree (errnbparfonc, p);
	    return pk_false;
	  }
      vfi1 = kl_DuplicateModel (vfi);
      xoutputs = kl_getInsOutsModel (vfi, &xinputs);
      l1 = NILLREPID;
      l2 = NILLREPID;
      for (i = 1; i <= nbsigdef; i++)
	{
	  l1 = addRepid (l1, xinputs);
	  xinputs = succid (xinputs);
	  l2 = addRepid (l2, xoutputs);
	  xoutputs = succid (xoutputs);
	}
      kl_setParamsInputsOutputs (vfi1, NILLREPID, l1, l2);
      varid = nameId (kl_getFlowExprModelProcCallname (e2));
      kl_setmodel (kl_getExpfilAppel (e2), varid, vfi1);
      return pk_true;
    }

  return pk_true;
}


/**
   Compiling an expression (p) such that (X1,...,Xn):= E, or (X1,...,Xn)::=E
   with vop in {opdefsig (:=), opdefpart (::=) } and such that the left part is a tuple (not reduced to an element, see cExpEquation()).

   @return "true" if the compiling succeeds, "false" otherwise (expression not well formed, or not yet implemented).
   It produces 
     - vklExprProc the kernel representation of the expression ( kl_mkEquation())
     - inputs the list of the inputs of the model of p
     - outputs the list of the outputs of the model of p
     - lparam: the list of parameters/constants,
     - indexList: the list of indices.

   \note { only (x, y):= P{}()  and (I1, ... In):= << vector1, ...., vector_n >> are implemented.}
}
*/
_VIRTUAL_
  boolean
  Trees2Kernel::cTupleDef (noperators vop, ntrees p, ptlrepid * inputs, ptlrepid * outputs,
			   ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  ntrees arbreaux;
  integer nbsigdef;
  ptklExprSig vklExprSig1, vklExprSig2;
  boolean b;
  ptrepfi vfi;

  b = pk_false;
  nbsigdef = 0;
  *lparam = NILLREPID;
  *outputs = NILLREPID;
  *indexList = NILLREPID;
  *inputs = NILLREPID;
  if (isChild (1, p))
    {
      arbreaux = operande;
      vklExprSig1 = cExprSigLeftExpr (operande, outputs, lparam, indexList);
      if ((*outputs == NILLREPID) || (*lparam != NILLREPID))
	{
	  mkklErrorAnnotTree (errconflid, arbreaux);
	  *lparam = NILLREPID;
	  *indexList = NILLREPID;
	}
      else
	{
	  nbsigdef = ptlrepid_size (*outputs);
	  if (nbsigdef != kl_getSizeExpr (vklExprSig1))
	    mkklErrorAnnotTree (errsortid, arbreaux);
	  else
	    b = isIdentifierForm (arbreaux, pk_true, pk_true);
	}

    }
  if (!isChild (2, p))
    return pk_false;
  // (x, y):= APPEL(...) accepté 
  if ((voperator == opappe) || (voperator == opinst))
    {
      arbreaux = operande;
      vklExprSig2 = cExprSig (operande, inputs, lparam, indexList);
      if (locateCalledProcess (arbreaux, &vfi))
	b = verifCalledProcess (vfi, nbsigdef, arbreaux);
      if (b)
	b = intrinsicSettings (arbreaux, vklExprSig1, vklExprSig2);
    }
  else if (voperator == oppcar)
    {				// (I1, ... In):= << vector1, ...., vector_n >> 
      arbreaux = operande;
      vklExprSig2 = cExprSig (operande, inputs, lparam, indexList);
    }
  else
    {
      notYetImplemented (p);	// tuples are not yet fully implemented.
      return pk_false;
    }
  b = b && (vklExprSig2 != NILEXPRSIG);
  if (!b)
    return pk_false;
  shouldBeStatevOrShared (*outputs, (vop == opdefpart) || (vop == opdefpartdef), p);
  *vklExprProc = kl_mkEquation (vop, *inputs, *outputs, *lparam, p, vklExprSig1, vklExprSig2);
  return pk_true;
}

/**
   @return true if the operand (aTree) is an identifier (x, x.y) or a list of such elements, false otherwise.
   restriction: y[] not allowed.
*/
_VIRTUAL_ boolean
Trees2Kernel::testAcceptIdentifierForm (ntrees aTree, boolean bListAccept, boolean brefArray)
{
  // return pk_true;
  noperators vop;
  ntrees p;

  if (IS_NILTREE (aTree))
    return pk_false;
  vop = noper (aTree);
  if (LL_is_listOp (vop))
    if (bListAccept)
      {
	p = aTree;
	if (firstChild (&p))
	  do
	    if (!isIdentifierForm (operande, pk_false, brefArray))
	      return pk_false;
	  while (nextChild (&p));
	return pk_true;
      }
    else;
  else
    {
      if (OPIS_SIGNALIDENTIFIER (vop))
	return pk_true;
      if ((vop == opcham) || ((vop == opelem) && brefArray))
	{
	  if (!isChild (1, aTree))
	    return pk_false;
	  return isIdentifierForm (operande, pk_false, brefArray);
	}
    }
  mkklErrorAnnotTree (errUnexpectedExpr, aTree);
  return pk_false;
}

/**
   @returns
   - "true" if an expression (aTree) is a reference to a identifier (x, x.y, x[]) without access to array element if not brefArray.
     It is generalized to a list of such expressions (when bListAccept) and each element satisfies the previous property.
   - "false" otherwise.
*/
_VIRTUAL_ boolean
Trees2Kernel::isIdentifierForm (ntrees aTree, boolean bListAccept, boolean brefArray)
{
  noperators vop;
  ntrees save;
  boolean b;

  // Sauvegarde due aux effets de bords des méthodes locList de manipulation d'arbres (isChild, ...)
  vop = voperator;
  save = operande;
  b = testAcceptIdentifierForm (aTree, bListAccept, brefArray);
  operande = save;
  voperator = vop;
  return b;
}

/** 
    Compiling an equation (p) X:= E or X::= E , op in {:=,::=}, the left part is not a tuple (see cTupleDef()).
    @return "true" if the compiling succeeds, "false" otherwise (expression not well formed, or not yet implemented).
    It produces 
    - vklExprProc the kernel representation of the expression ( kl_mkEquation())
    - inputs the list of the inputs of the model of p
    - outputs the list of the outputs of the model of p
    - lparam: the list of parameters/constants,
    - indexList: the list of indices.
 */
_VIRTUAL_
  boolean
  Trees2Kernel::cExpEquation (noperators op, ntrees p, ptlrepid * inputs, ptlrepid * outputs,
			      ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  ptklExprSig vklExprSig1, vklExprSig2;
  boolean b;
  ntrees x;

  *lparam = NILLREPID;
  *inputs = NILLREPID;
  *outputs = NILLREPID;
  *indexList = NILLREPID;
  b = pk_false;
  if (isChild (1, p))
    {
      if (!isIdentifierForm (operande, pk_false, pk_true))
	return pk_false;
      vklExprSig1 = cExprSigLeftExpr (operande, outputs, lparam, indexList);
      if ((*outputs) == NILLREPID)
	mkklErrorAnnotTree (errconflid, operande);
      else
	b = pk_true;
      shouldBeStatevOrShared (*outputs, (op == opdefpart) || (op == opdefpartdef), operande);
    }
  if (!isChild (2, p))
    return pk_false;
  x = operande;
  G_ExternalOutputsModel = NILLREPID;
  G_ExternalInputsModel = NILLREPID;
  vklExprSig2 = cExprSig (operande, inputs, lparam, indexList);
  if (vklExprSig2 == NILEXPRSIG)
    return pk_false;
  intrinsicSettings (x, vklExprSig1, vklExprSig2);

  *inputs = insertunsort (G_ExternalInputsModel, *inputs);
  *outputs = insertunsort (G_ExternalOutputsModel, *outputs);

  diffLists (*outputs, inputs);
  if (!b)
    return pk_false;
  *vklExprProc = kl_mkEquation (op, *inputs, *outputs, *lparam, p, vklExprSig1, vklExprSig2);
  return pk_true;
}

/**
   @return "true" if the compiling succeeds, "false" otherwise (the model is not declared, number of effective inputs not equals
   to the formal ones,...).
   
   It translates a reference to a model (p):  foo (e1, e2, ..., en) or foo{p1,...,pp} (e1, e2, ..., en)
   where 
   - foo the name of the model or a
   - the ei are the effective input signals
   - p1,...pp the effective parameters
   
   It produces:
   - vklExprProc the kernel representation of the expression
   - inputs the list of the inputs of the model of p
   - outputs the list of the outputs of the model of p
   - lparam: the list of parameters/constants,
   - indexList: the list of indices.

   It calls cExpRefToModel(), cExpInstInAppe()
*/
_VIRTUAL_
  boolean
  Trees2Kernel::cExpCall (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  ptklExprProc vklExprProc1;
  ptklExprSig vklExprSig, p1, p2, vaux;
  boolean ok, b;
  ptlrepid outs, ins;
  ptrepfi vfi;
  integer n;
  ptident id;
  ptIdStack varid;
  ptlrepid lexternalIns, lexternalOuts;
  ptlrepid inputs1;
  noperators ope;

  ope = noper (nfather (p));
  b = ((ope == oplfeed) || (ope == oppetiq) || (ope == oppcontrol));

  if (b)
    // [on] label :: P() is replaced by [on] label :: P{} and  | P() is replaced by  | P{}
    {
      if (!isoChild (2, p)) // no inputs
	{
	  if (!isChild (1, p)) return pk_false;
	  if (voperator == opinst) return cExpRefToModel (operande, inputs, outputs, vklExprProc);
	  if (OPIS_SIGNALIDENTIFIER (voperator))
	    {
	      mkklWarningAnnotTree (warSyntaxModif, p);
	      nsetop (p, opinst);
	      return cExpRefToModel (p, inputs, outputs, vklExprProc);
	    }
	  return pk_false;
	}
    }

  ok = pk_false;
  if (!isChild (1, p))
    return ok;
  *inputs = NILLREPID;
  *lparam = NILLREPID;
  *indexList = NILLREPID;
  inputs1 = NILLREPID;
  if (cExpInstInAppe (&inputs1, outputs, &vklExprProc1, &varid, &lexternalIns, &lexternalOuts))
    {
      if (isoChild (2, p))
	vklExprSig = cExprSig (operande, inputs, lparam, indexList);
      else
	vklExprSig = NILEXPRSIG;

      G_ExternalInputsModel = getCurrentIdentRepids (lexternalIns);
      G_ExternalOutputsModel = getCurrentIdentRepids (lexternalOuts);

      locateCalledProcess (p, &vfi);
      id = kl_getNameModel (vfi);
      n = numberOutputsOfExprSig (vklExprSig);
      if (verifAutomatonIntrinsics (vklExprSig, id, p, &ok))
	{ 
	  if (ok) 
	    {
	      int nform = ptlrepid_size (inputs1);
	      ok = (nform == n);
	      if (! ok)  /* Testing Transition(S1, S2, Trigger [,priority]) */
		if (isTransitionId (id)) 
		  {
		    ok = (n == nform-1);
		    if (ok)
		      {
			// Adding 0 for the priority (default value).
			kl_getBinaryExpr (vklExprSig, &p1, &vaux); // S1
			kl_getBinaryExpr (vaux, &p1, &vaux); // S2
			kl_getBinaryExpr (vaux, &p1, &p2); // Trigger
			ptklExprSig vzero = kl_mkListExpr (oplesign, kl_mkConstExpr (opcstent, string2repstr ("0"), p), p);
			kl_setBinaryExprChild2(vaux,vzero);
		      }
		  }
	    }
	}
      else
	if (ptlrepid_size (inputs1) != n)
	  {
	    if (kl_isIntrinsicModel (vfi))
	      {
		if (isRendezVousId(id)) // implementation of polyendochrony.
		  ok = (n >1);
		else
		  if (isIndicesId (id))
		    // indices(e1, ..., en) produces n outputs with [0..e1-1, ..., 0..en-1] integer as type
		    ok = (n != 0);
		  else if (isSizeId (id))
		    {
		      if (n == 1)
			{
			  // size(t) is replaced by size(t, 1)
			  ok = pk_true;
			  kl_getBinaryExpr (vklExprSig, &p1, &p2);
			  p2 = kl_mkListExpr (oplesign, kl_mkConstExpr (opcstent, string2repstr ("1"), p), p);
			  vklExprSig = kl_mkInsertHeadList (oplesign, p1, p2, p);
			}
		    }
	      }
	  }
      if (ok)
	{
	  *vklExprProc = kl_mkProcCall (*inputs, *outputs, *lparam, p, vklExprProc1, vklExprSig);
	  if (b)
	    {
	      locateCalledProcess (p, &vfi);
	      outs = kl_getInsOutsModel (vfi, &ins);
	      if (isAssertionId (kl_getNameModel (vfi)))
		kl_setAssertionsFlag (G_process, pk_true);
	      *vklExprProc = cExpCallFinalyze (p, *lparam, *inputs, outs, *vklExprProc);
	    }
	}
      else
	mkklErrorAnnotTree (errnbparfonc, p);
    }
  else if (isoChild (2, p))
    vklExprSig = cExprSig (operande, inputs, lparam, indexList);
  else
    vklExprSig = NILEXPRSIG;

  return (ok);
}

/**
   Internal to cExpCall() method. The current tree is assumed to be a reference to a model (foo or foo{p1,...,pp}).
   @return "true" if the compiling succeeds, "false" otherwise.
   It calls verifCall() on the current operande (operande tree) when it is a reference without parameters, 
   or cExpRefToModelloc() otherwise.

   It produces:
   - vklExprProc the kernel representation of the expression (kl_mkMacro())
   - varid the stack assigned to the name of the model
   - inputs the formal inputs of the model
   - outputs the formal outputs of the model
   - lexternalIns the formal inputs of the model not declared in the model
   - lexternalOuts the formal outputs of the model not declared in the model
*/
_VIRTUAL_
  boolean
  Trees2Kernel::cExpInstInAppe (ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc, ptIdStack * varid,
				ptlrepid * lexternalIns, ptlrepid * lexternalOuts)
{
  ptlrepid lparam, indexList;

  if (OPIS_SIGNALIDENTIFIER (voperator))
    {
      if (!locateOrError (operande, varid))
	return pk_false;
      return verifCall (operande, NILEXPRSIG, *varid, inputs, outputs, vklExprProc, lexternalIns, lexternalOuts);
    }
  if (voperator == opinst)
    return cExpRefToModelloc (operande, inputs, &lparam, &indexList, outputs, vklExprProc, varid, lexternalIns, lexternalOuts);
  return pk_false;
}


/**
   @return "true" if the compiling succeeds, "false" otherwise.

   It compiles a reference to a model such that foo{p1,...,pp} where pi are parameters.

   It produces:
   - vklExprProc the kernel representation of the expression (kl_mkMacro())
   - varid the stack assigned to the name of the model
   - lparam the formal parameters of the model
   - inputs the formal inputs of the model
   - outputs the formal outputs of the model
   - lexternalIns the formal inputs of the model not declared in the model
   - lexternalOuts the formal outputs of the model not declared in the model

   It calls cExpEffectiveParam(), verifCall().
*/
_VIRTUAL_
  boolean
  Trees2Kernel::cExpRefToModelloc (ntrees p, ptlrepid * inputs, ptlrepid * lparam, ptlrepid * indexList, ptlrepid * outputs,
				   ptklExprProc * vklExprProc, ptIdStack * varid, ptlrepid * lexternalIns, ptlrepid * lexternalOuts)
{
  ptklExprSig deblistparam;
  ptrepfi vfi;
  boolean b;

  if (!locateChild (p, 1, varid))
    return pk_false;
  b = getTopIfModel (*varid, &vfi);
  if (b && isoChild (2, p))
    deblistparam = cExpEffectiveParam (operande, vfi, lparam, indexList);
  else
    deblistparam = NILEXPRSIG;
  return verifCall (p, deblistparam, *varid, inputs, outputs, vklExprProc, lexternalIns, lexternalOuts);
}

/**
   @return a new representation of a list of effective parameters (p).
   It compiles a list of effective parameters (p) when there are used to call a model (vfi).
   - lparam the list of the parameters referenced in the expression
   - indexList the list of the indices referenced in the expression
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::cExpEffectiveParam (ntrees p, ptrepfi vfi, ptlrepid * lparam, ptlrepid * indexList)
{
  ntrees aux, x;
  integer no_param;
  ptklExprSig v, vklExprSig1, vklExprSig, deblistparam;
  typid vtypid;
  ptlrepid inputs1;
  ptIdStack name;
  ptrepfi vfil;

  no_param = 1;
  aux = p;
  inputs1 = NILLREPID;
  *lparam = NILLREPID;
  *indexList = NILLREPID;
  vklExprSig1 = NILEXPRSIG;
  if (firstChild (&aux))
    do
      {
	x = operande;
	if (OPIS_SIGNALIDENTIFIER (voperator))
	  {
	    /* the parameter is an identifier */
	    if (getParameterTypeId (vfi, no_param, &vtypid))
	      if (vtypid == protoIdent)
		{
		  /* the identifier must be a name of a parameter. */
		  vklExprSig = cExprSig (operande, &inputs1, lparam, indexList);
		  if ((vklExprSig == NILEXPRSIG) || (inputs1 != NILLREPID))
		    mkklErrorAnnotTreeName (errparnondecl, p, nameId (inputs1));
		}
	      else if (vtypid == protoModel)
		{
		  if (locateOrError (operande, &name))
		    {
		      if (kl_getTopIfModel (name, &vfil))
			{
			  vklExprSig = kl_mkIdentifierExpr (mkRepidsigRef (name, kl_top (name)), operande);
			  ptlrepid_insert (lparam, name);
			}
		      else
			mkklErrorAnnotTreeName (errparnondecl, p, name);
		    }
		}
	      else		// (vtypid == protoType)
		/* the identifier must be a name of a type. */
		vklExprSig = kl_mkTypeExpr (voperator, cExpType (operande), x);
	    else
	      mkklErrorAnnotTree (errparam, operande);
	  }
	else /* the parameter is not  an identifier */ if (LL_isTypeOp (noper (operande)))	/* The parameter  is a type. */
	  vklExprSig = kl_mkTypeExpr (voperator, cExpType (operande), x);
	else			/* cas ou le parametre est une expression de vparameters */
	  {
	    vklExprSig = cExprSig (operande, &inputs1, lparam, indexList);
	    while (inputs1 != NILLREPID)
	      {
		mkklErrorAnnotTreeName (errparnondecl, p, nameId (inputs1));
		inputs1 = succid (inputs1);
	      }
	  }
	no_param++;
	/* creating of the list of the parameters */
	v = kl_mkBinaryExpr (oper2, oplepara, vklExprSig, NILEXPRSIG, x);
	if (vklExprSig1 == NILEXPRSIG)
	  {
	    vklExprSig1 = v;
	    deblistparam = v;
	  }
	else
	  {
	    kl_setBinaryExprChild2 (vklExprSig1, v);
	    vklExprSig1 = v;
	  }
      }
    while (nextChild (&aux));
  return deblistparam;
}




/**
   @return the list of the identifiers (stack associated with a string identifier) referenced in a list of identifiers (string).
   Assumed: lsig is the list (s1,...sn) of the signals of a masking P / s1,...,sn.
   The type associated with an unknown(*) identifier is the implicit type (default), otherwise the context gives the information.

   (*) unknown stands for: empty stack, or the top is not a signal.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::getLocalSignalsOfMasking (ntrees lsig)
{
  ntrees pred, arbre;
  ptIdStack varid;
  ptlrepid list, xlrepid;
  ptrepproto vrepproto, vproto;
  ptreptyp deftyp_default, vtyp;
  ptklExprSig vinit_default, vinit;
  ptIdentProp topId;
  typval IdClass;
  boolean done;
  categoryType vcatDecl;

  pred = lsig;
  if (firstChild (&pred))
    {
      list = NILLREPID;
      deftyp_default = kl_mkBasicType (opimplicit);
      vinit_default = NILEXPRSIG;
      do
	{
	  if (locateOrError (operande, &varid))
	    {
	      done = pk_false;
	      // If the context provides a model, it is chosen.
	      if (getSignalClassTop (varid, &topId, &IdClass))
		if (IdClass == sig)
		  {
		    vproto = ptIdentProp_getIdent (topId);
		    ptrepproto_getInfos (vproto, &vinit, &vtyp, &vcatDecl, &arbre);
		    vrepproto = mkrepproto (vtyp, vinit, vcatDecl, operande, NILTREE);
		    done = pk_true;
		  }
	      if (!done)
		vrepproto = mkrepproto (deftyp_default, vinit_default, undefCategory, operande, NILTREE);
	      if (!appendProtoId (&list, varid, vrepproto, &xlrepid))
		mkklErrorAnnotTreeName (errdoubledecl, operande, varid);
	    }
	}
      while (nextChild (&pred));
    }
  return list;
}

/** 
    It compiles a masking process expression (p): EPROC / x1, ...., xn 
    @return "true" if the compiling succeeds, "false" otherwise.
    
    It produces:
    - vklExprProc the kernel representation of the expression. It is the representation of the left part completed with the local signals 
    returned by the getLocalSignalsOfMasking() method.
    - inputs the inputs of the EPROC expression (signals, labels)
    - outputs the outputs of the EPROC expression
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpMasking (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ntrees l_decl;
  ptlrepid sig_loc;
  boolean ok;

  if (!isChild (1, p))
    return pk_false;
  l_decl = nchild (2, p);
  sig_loc = getLocalSignalsOfMasking (l_decl);
  //   verifDoubleDec(&sig_loc/* , ldef_etiq */, pk_true, l_decl);
  pushFlows (sig_loc);
  operande = nchild (1, p);
  voperator = noper (operande);
  ok = klparcours (inputs, outputs, vklExprProc);
  if (ok)
    {
      kl_setLocals (*vklExprProc, NILLREPID, sig_loc, NILLREPID);
      diffLists (sig_loc, outputs);
    }
  popFlows (sig_loc);
  return ok;
}



/**
   It compiles a dependence process expression (p): X, Y->Z, W when H
   @return "true" if the compiling succeeds, "false" otherwise.
   
   It produces:
   - vklExprProc the kernel representation of the expression ( kl_mkDependence()).
   - inputs the inputs of the expression (signals, labels)
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpDependence (ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  ntrees pg, pd, p_when;
  ptklExprSig v1, v2, v3;
  ptlrepid indexList, lparam, laux;
  boolean ok, b;

  laux = NILLREPID;
  lparam = NILLREPID;
  indexList = NILLREPID;

  pg = nchild (1, p);
  ok = isIdentifierForm (pg, pk_true, pk_false);

  v1 = cExprSig (pg, &laux, &lparam, &indexList);
  b = pk_true;
  if ((lparam != NILLREPID) || (indexList != NILLREPID))
    {
      mkklErrorAnnotTree (errconflid, pg);
      b = pk_false;
    }
  pd = nchild (2, p);
  ok = ok && isIdentifierForm (pd, pk_true, pk_false);
  v2 = cExprSig (pd, &laux, &lparam, &indexList);
  if ((lparam != NILLREPID) || (indexList != NILLREPID))
    {
      mkklErrorAnnotTree (errconflid, pd);
      b = pk_false;
    }
  p_when = nchild (3, p);
  v3 = cExprSig (p_when, &laux, &lparam, &indexList);
  if (!b)
    return pk_false;
  *inputs = insertlist (laux, *inputs);
  *vklExprProc = kl_mkDependence (laux, p, v1, v2, v3);
  return ok;
}


/**
   It compiles a list of dependence process expression (p): X, Y->Z, W -> .... -> E,F.
   @return "true" if the compiling succeeds, "false" otherwise.
   
   It produces:
   - vklExprProc the kernel representation of the expression (a composition of dependences: kl_mkDependence())
   - inputs the inputs of the expression (signals, labels)
*/
_VIRTUAL_ boolean
Trees2Kernel::cExplDependences (ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  ntrees pg, pd;
  ptklExprSig v1, v2;
  ptlrepid indexList, lparam, laux;
  boolean ok, b;
  integer i, nb;
  ptklExprProc vklExprProc1;

  *vklExprProc = NILEXPFIL;
  ok = pk_true;
  nb = ntreearity (p);
  for (i = 1; i < nb; i++)
    {
      pg = nchild (i, p);
      laux = NILLREPID;
      lparam = NILLREPID;
      indexList = NILLREPID;
      ok = ok && isIdentifierForm (pg, pk_true, pk_false);
      v1 = cExprSig (pg, &laux, &lparam, &indexList);
      b = pk_true;
      if ((lparam != NILLREPID) || (indexList != NILLREPID))
	{
	  mkklErrorAnnotTree (errconflid, pg);
	  b = pk_false;
	}
      pd = nchild (i + 1, p);
      ok = ok && isIdentifierForm (pd, pk_true, pk_false);
      v2 = cExprSig (pd, &laux, &lparam, &indexList);
      if ((lparam != NILLREPID) || (indexList != NILLREPID))
	{
	  mkklErrorAnnotTree (errconflid, pd);
	  b = pk_false;
	}
      if (b)
	{
	  *inputs = insertlist (laux, *inputs);
	  vklExprProc1 = kl_mkDependence (laux, p, v1, v2, NILEXPRSIG);
	  if (*vklExprProc == NILEXPFIL)
	    *vklExprProc = vklExprProc1;
	  else
	    *vklExprProc = kl_mkComposition (laux, NILLREPID, p, *vklExprProc, vklExprProc1);
	}
      else
	ok = pk_false;
    }
  return ok;
}


/**
   It compiles (p) a labelled process such that e:: P{}. 
   @return "true" if the compiling succeeds, "false" otherwise.
   It produces:
    - vklExprProc the kernel representation of the expression. It is the representation of the right part, attributed with a label (kl_setLabelExpr()).
    - inputs the inputs of the expression (signals, labels)
    - outputs the outputs of the expression
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpLabelledExp (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ptIdStack name;
  ptklExprSig vexpetiq;
  boolean ok;
  ntrees vt;
  ptIdentProp topId;
  typval IdClass;

  vt = nchild (1, p);
  if (!locateOrError (vt, &name)) return pk_false;
  if (!getSignalClassTop (name, &topId, &IdClass))
    {
      mkklErrorAnnotTreeName (errdecl, vt, name);
      return pk_false;
    }
  if (IdClass != eti)
    {
      mkklErrorAnnotTree (errUnexpectedExpr, vt);
      return pk_false;
    }
  vexpetiq = kl_mkIdentifierExpr (mkRepidsigRef (name, topId), vt);
  if (!isChild (2, p)) return pk_false;
  ok = klparcours (inputs, outputs, vklExprProc);
  if (ok)
    {
      ptrepproto vrepproto = ptIdentProp_getIdent (topId);
      // A label must be used for only one Expression
      if (ptrepproto_getLabelledExpr (vrepproto) != NILEXPFIL)
	{
	  mkklErrorAnnotTree (errLabelExpr, vt);
	}
      else
	{
	  ptrepproto_setLabelledExpr (vrepproto, *vklExprProc);
	  kl_setLabelExpr (*vklExprProc, vexpetiq);
	}
    }
  return ok;
}




/**
   It compiles (p) a process such that
              on h:: (| ...|) 
   @return "true" if the compiling succeeds, "false" otherwise.
   It produces:
    - vklExprProc the kernel representation of the expression. It is the representation of the right part, attributed with a label (kl_setLabelExpr()).
    - inputs the inputs of the expression (signals, labels)
    - outputs the outputs of the expression
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpControlledExp (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ptklExprSig vOnExpr;
  boolean ok;
  ntrees vt;
  ptlrepid lparam, indexList;

  vt = nchild (1, p);
  lparam = NILLREPID;
  indexList = NILLREPID;
  vOnExpr = cExprSig (vt, inputs, &lparam, &indexList);
  if (!isChild (2, p)) return pk_false;
  ok = klparcours (inputs, outputs, vklExprProc);
  if (ok) kl_setControlExpr (*vklExprProc, vOnExpr);
  return ok;
}


/**
   It compiles (recursively) a element (elemlist) of a list of expressions on processes (rootList),
   and cur_p is the right of elemlist.
   Presently, only the composition of processes is defined (i.e. op=oplfeed).
   @return "true" if the compiling succeeds, "false" otherwise.
   It calls 
   It produces:
    - vklExprProc the kernel representation of the expression (kl_mkComposition())
    - inputs the inputs of the expression (signals, labels)
    - outputs the outputs of the expression
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpListExpressionsOnProcessesRec (ntrees * cur_p, ntrees rootList, noperators op, ntrees elemlist,	/* ntrees current, */
						 ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean result;
  ptlrepid inputs1, outputs1, aux;
  ptklExprProc vklExprProc_loc;
  ptIdStack elem;

  if (nextChild (cur_p))
    // result = cExpListExpressionsOnProcessesRec(cur_p, current, op, operande, *cur_p, inputs, outputs, vklExprProc); June 2006
    result = cExpListExpressionsOnProcessesRec (cur_p, rootList, op, operande, inputs, outputs, vklExprProc);
  else
    result = pk_false;
  operande = elemlist;
  voperator = noper (operande);
  inputs1 = NILLREPID;
  outputs1 = NILLREPID;
  vklExprProc_loc = NILEXPFIL;
  if (klparcours (&inputs1, &outputs1, &vklExprProc_loc))
    {
      if (result)
	{
	  switch ((int) op)
	    {
	    case oplfeed:
	      aux = outputs1;
	      while (aux != NILLREPID)
		{
		  elem = nameId (aux);
		  if (!isStateVariable (aux))
		    ptlrepid_remove_Unsorted (inputs, elem);
		  aux = succid (aux);
		}
	      aux = *outputs;
	      while (aux != NILLREPID)
		{
		  elem = nameId (aux);
		  if (!isStateVariable (aux))
		    ptlrepid_remove_Unsorted (&inputs1, elem);
		  aux = succid (aux);
		}
	      break;
	    default:
	      notYetImplemented (operande);
	    }
	  *inputs = insertlist (inputs1, *inputs);
	  *outputs = insertlist (outputs1, *outputs);
	  *vklExprProc = kl_mkComposition (*inputs, *outputs, rootList, vklExprProc_loc, *vklExprProc);
	}
      else
	{
	  *inputs = inputs1;
	  *outputs = outputs1;
	  *vklExprProc = vklExprProc_loc;
	  result = pk_true;
	}
    }
  return result;
}

/** Temporary.
    It moves the array processes at the end of the composition (help fo the type checking, that it must be completed).
*/
void
Trees2Kernel::reorgComposition (ntrees p)
{
  boolean move, b;
  integer nb, i;
  ntrees s;
  noperators op;

  nb = ntreearity (p);
  move = pk_false;
  for (i = nb; i > 0; i--)
    {
      s = nchild (i, p);
      op = noper (s);
      b = pk_false;
      if (op == opiter)
	b = pk_true;
      else if (op == oppetiq)
	if (isChild (2, s))
	  b = (noper (operande) == opiter);

      if (!b)
	move = pk_true;
      else if (move)
	{
	  ndestroyListElem (p, i);
	  p = ntreeList_appendLast (p, s);
	}
    }
}

/**
   It compiles an operation on a list of expressions on processes (p). 
   Presently, only the composition of processes is defined (i.e. op=oplfeed).
   It calls crecbinproc()
   @return "true" if the compiling succeeds, "false" otherwise.

   It produces:
    - vklExprProc the kernel representation of the expression (kl_mkComposition())
    - inputs the inputs of the expression (signals, labels)
    - outputs the outputs of the expression
*/
_VIRTUAL_
  boolean Trees2Kernel::cExpListExpressionsOnProcesses (ntrees p, noperators op, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ntrees
    cur_p;

  cur_p = p;
  reorgComposition (p);
  if (!firstChild (&cur_p))
    return pk_false;
  return cExpListExpressionsOnProcessesRec (&cur_p, p, op, operande, /* cur_p, */ inputs, outputs, vklExprProc);
}


/**
   It compiles (p) an array process (array i to n of P with Q end ). 
   @return "true" if the compiling succeeds, "false" otherwise.
   It produces:
    - vklExprProc the kernel representation of the expression (kl_mkArray()).
    - inputs the inputs of the expression (signals, labels)
    - outputs the outputs of the expression

    Quite similar to cExpArray().
*/
_VIRTUAL_
  boolean
Trees2Kernel::cExpArray (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ntrees x, vstree;
  ptklExprProc corpsdo, pw;
  ptIdStack indexId;
  ptlrepid outputs1, inputs1, vindex, laux, indexList;
  ptklExprSig bsup;
  boolean result;
  ptrepproto vrepproto;
  ptIdentProp topId;

  // "array" <indice_motif> "of" <expr_processus> <o_init_motif>
  bsup = NILEXPRSIG;
  if (isChild (1, p))		// <indice_motif> cad: ( indice , borne sup)
    {
      vstree = operande;
      if (locateChild (operande, 1, &indexId))
	{
	  x = operande;
	  if (isTopaLabel (indexId))
	    mkklErrorAnnotTreeName (errconflid, p, indexId);
	  else;
	}
      else
	indexId = unnamedStack;
      if (isChild (2, vstree))
	{
	  indexList = NILLREPID;
	  inputs1 = NILLREPID;
	  laux = NILLREPID;
	  bsup = cExprSig (operande, &inputs1, &laux, &indexList);
	  if ((inputs1) != NILLREPID)
	    mkklErrorAnnotTree (errconflid, operande);
	}
    }
  else
    indexId = unnamedStack;

  // "of" <expr_processus>
  if (!isChild (2, p))
    return pk_false;
  if (indexId == unnamedStack)
    x = operande;
  vrepproto = mkrepproto (kl_mkBasicType (opinteger), NILEXPRSIG, undefCategory, x /*operande */ , NILTREE);
  topId = mkIdentProp (vrepproto);
  kl_pushIndex (indexId, topId);
  vindex = mkRepidsigRef (indexId, topId);
  result = pk_true;
  corpsdo = NILEXPFIL;
  inputs1 = NILLREPID;
  outputs1 = NILLREPID;
  if (!klparcours (&inputs1, &outputs1, &corpsdo))
    result = pk_false;
  else
    {
      *inputs = insertlist (inputs1, *inputs);
      *outputs = insertlist (outputs1, *outputs);
      // <o_init_motif>
      pw = NILEXPFIL;
      if (isoChild (3, p))
	{
	  inputs1 = NILLREPID;
	  outputs1 = NILLREPID;
	  result = klparcours (&inputs1, &outputs1, &pw);
	  *inputs = insertlist (inputs1, *inputs);
	  *outputs = insertlist (outputs1, *outputs);
	}
      *vklExprProc = kl_mkArray (*inputs, *outputs, laux, p, kl_mkIdentifierExpr (vindex, p), bsup, pw, corpsdo);
    }
  popIndexRepid (vindex);
  return result;
}

/**
   It compiles (p) an iterate process (iterate i of P with Q end). 
   @return "true" if the compiling succeeds, "false" otherwise.
   It produces:
    - vklExprProc the kernel representation of the expression (kl_mkArray()).
    - inputs the inputs of the expression (signals, labels)
    - outputs the outputs of the expression

    Quite similar to cExpArray().
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpIterate (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ptklExprProc corpsdo, pw;
  ptlrepid lparam, lind;
  boolean result;
  ptklExprSig indexExpr;

  // "iterate" <expression_indice> "of" <expr_processus> <o_init_motif>
  if (!isChild (1, p))
    return pk_false;
  result = pk_true;
  // ntrees vstree = operande;
  lparam = NILLREPID;
  lind = NILLREPID;
  indexExpr = cExprSig (operande, inputs, &lparam, &lind);
  // Restriction
  // if (kl_getFlowExprOperator(indexExpr) == opleflot) { notYetImplemented(vstree); result = pk_false; }
  // "of" <expr_processus>
  if (!isChild (2, p))
    return pk_false;
  corpsdo = NILEXPFIL;
  if (!klparcours (inputs, outputs, &corpsdo))
    return pk_false;
  // <o_init_motif>
  pw = NILEXPFIL;
  if (isoChild (3, p))
    result = klparcours (inputs, outputs, &pw);
  *vklExprProc = kl_mkIterate (*inputs, *outputs, lparam, p, indexExpr, pw, corpsdo);
  return result;
}


/**
   It compiles (p) a case expression. 
   @return "true" if the compiling succeeds, "false" otherwise.
   It produces:
   - vklExprProc the kernel representation of the expression (kl_mkCase(), kl_ptlcase_add()).
   - inputs the inputs of the expression (signals, labels)
   - outputs the outputs of the expression
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpCase (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ptlrepid lparam, indexList, l1, l2, l3;
  ptklExprSig vname, val;
  ptklExprProc vexp_elem, vexp_else;
  ptlcase l_case;
  ntrees p_case, p_elem;
  int i, nb;
  boolean ok, b;

  *inputs = NILLREPID;
  *outputs = NILLREPID;
  lparam = NILLREPID;

  // Sept 2019. Accept an expression.
  vname = cExprSig /* cExpName */ (nchild (1, p), inputs, &lparam, &indexList);

  p_case = nchild (2, p);
  l_case = NILLCASE;
  nb = ntreearity (p_case);
  ok = pk_true;
  for (i = 1; i <= nb; i++)
    {
      /* list of elements such that  value: expression of process */
      p_elem = nchild (i, p_case);
      l1 = NILLREPID;
      l3 = NILLREPID;
      val = cExprSig (nchild (1, p_elem), &l1, &lparam, &l3);
      if ((l1 != NILLREPID) || (l3 != NILLREPID))
	{
	  mkklErrorAnnotTree (errconflid, operande);
	  val = NILEXPRSIG;
	  ok = pk_false;
	}
      if (isChild (2, p_elem))
	{
	  l1 = NILLREPID;
	  l2 = NILLREPID;
	  vexp_elem = NILEXPFIL;
	  b = klparcours (&l1, &l2, &vexp_elem);
	  *inputs = insertlist (l1, *inputs);
	  *outputs = insertlist (l2, *outputs);
	  if (b)
	    l_case = kl_ptlcase_add (l_case, val, vexp_elem);
	  else
	    ok = pk_false;
	}
      else
	ok = pk_false;
    }
  if (isoChild (3, p))		/* else part */
    {
      l1 = NILLREPID;
      l2 = NILLREPID;
      if (!klparcours (&l1, &l2, &vexp_else))
	ok = pk_false;
      *inputs = insertlist (l1, *inputs);
      *outputs = insertlist (l2, *outputs);
    }
  else
    vexp_else = NILEXPFIL;

  if (ok)
    *vklExprProc = kl_mkCase (*inputs, *outputs, lparam, p, vname, l_case, vexp_else);
  return (ok);
}

/**
   Restriction applied for the local declarations of a confined process.
   Presently, the declaration of sub-processes and imports in this context are not allowed.
*/
_VIRTUAL_ void
Trees2Kernel::restrictionsConfinedProcess (ntrees liste_decl)
{
  ntrees liste;

  if (!STree_isUndefinedTree (liste_decl))
    {
      liste = liste_decl;
      if (firstChild2 (&liste))
	do
	  {
	    if ((voperator == oplimpo) || LL_isModelFamilyOp (voperator))
	      mkklWarningAnnotTree (wardeclnotused, operande);
	  }
	while (nextChild (&liste));
    }
}

/** 
    It compiles a masking process expression (p):  EXPR_PROC where ...end. 
    Presently, there are some restrictions (See restrictionsConfinedProcess())
    @return "true" if the compiling succeeds, "false" otherwise.
    
    It produces:
    - vklExprProc the kernel representation of the expression. It is the representation of the left part completed with the local signals 
    returned by the getLocalSignalsOfMasking() method.
    - inputs the inputs of the EPROC expression (signals, labels)
    - outputs the outputs of the EPROC expression
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpConfinedExp (ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ntrees l_decl;
  ptlrepid const_typ_loc, sig_loc, letiq;
  boolean ok;

  l_decl = nchild (2, p);
  restrictionsConfinedProcess (l_decl);	// verif: pas de modeles de processus.
  const_typ_loc = cDecLocalConstantsTypes (l_decl);
  const_typ_loc = cDecLocalConstantsTypesSet (const_typ_loc);	// ordered (types/constants precedences)
  sig_loc = cLocalSignals (l_decl);
  letiq = cDecLabels (l_decl);
  verifDoubleDecConstantSelect (&const_typ_loc, sig_loc, pk_true, l_decl);
  verifDoubleDecConstantSelect (&const_typ_loc, letiq, pk_true, l_decl);
  verifDoubleDec (&sig_loc, letiq, pk_true, l_decl);
  pushLabels (letiq);
  pushFlows (sig_loc);
  operande = nchild (1, p);
  voperator = noper (operande);
  ok = klparcours (inputs, outputs, vklExprProc);
  diffLists (sig_loc, outputs);
  diffLists (letiq, inputs);

  *vklExprProc = kl_mkComposition (*inputs, *outputs, p, *vklExprProc, NILEXPFIL);
  if (ok)
    kl_setLocals (*vklExprProc, const_typ_loc, sig_loc, letiq);

  //  popModelTypes(modelTypes);
  popConstantsTypes (const_typ_loc);
  popFlows (letiq);
  popFlows (sig_loc);
  return ok;
}

/**
   Added for the specialisations.
 */
_VIRTUAL_ void
Trees2Kernel::initklparcours (ntrees p, noperators op)
{

}


/**
   Added for the specialisations.
 */
_VIRTUAL_ void
Trees2Kernel::finalklparcours (ntrees p, noperators op)
{

}

/**
   It compiles an assertion such that assert(c) where c is a boolean.
   @return "true" if the compiling succeeds, "false" otherwise.
   
   It produces:
   - vklExprProc the kernel representation of the expression (a call to an internal model called "_assert")
   - inputs the inputs of the expression (signals, labels)
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpAssert (ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  ptklExprSig vklExprSig;
  ptlrepid indexList, lparam, laux;
  ptrepfi vfi;
  boolean ok;
  ptIdStack vstack;
  ptlrepid lins;

  ok = pk_false;
  if (isChild (1, p))
    {
      locateOrError (getAssertionId (), &vstack);
      if (!getTopIfModel (vstack, &vfi))	/* internal error */
	mkklErrorAnnotTreeName (errprocnondec, p, vstack);
      else
	{
	  ok = pk_true;
	  /* ptlrepid louts = */ kl_getInsOutsModel (vfi, &lins);
	  laux = insertunsort (lins, NILLREPID);
	  lparam = NILLREPID;
	  indexList = NILLREPID;
	  vklExprSig = cExprSig (operande, inputs, &lparam, &indexList);
	  // list : vklExprSig
	  vklExprSig = kl_mkListExpr (opleflot, vklExprSig, operande);
	  *vklExprProc = kl_mkMacro (laux, NILLREPID, p, vstack, vfi, NILEXPRSIG);
	  *vklExprProc = kl_mkProcCall (*inputs, NILLREPID, lparam, p, *vklExprProc, vklExprSig);
	}
    }
  kl_setAssertionsFlag (G_process, pk_true);
  return ok;
}

/**
   The composition associated with an assert has some restrictions. 
   The accepted operators are 
   - the clock operators (^<, ^>, ^= and ^#)
   - the :=: operator. (opdefegal)
*/
boolean Trees2Kernel::verifyOperatorsAssert (ntrees p)
{
  noperators vop;
  boolean b;
  integer nb, i;

  b = pk_true;
  vop = noper (p);
  if (vop == oplfeed)
    {
      nb = ntreearity (p);
      for (i = 1; i <= nb; i++)
	if (!verifyOperatorsAssert (nchild (i, p)))
	  b = pk_false;
    }
  else if ((vop != opdefegal) && (vop != oplcrohor) && (vop != opldechor) && (vop != oplequhor) && (vop != oplexchor))
    {
      mkklErrorAnnotTree (errUnexpectedExpr, p);
      b = pk_false;
    }
  return b;
}

/**
   It compiles an assertion such that assert(|...|) with some restrictions for the expression of the composition.
   The accepted operators are : ^=, ^<, ^>, ^#, and :=:.
   @return "true" if the compiling succeeds, "false" otherwise.
   
   It produces:
   - vklExprProc the kernel representation of the expression (kl_mkDependence()).
   - inputs the inputs of the expression (signals, labels)
*/
_VIRTUAL_
  boolean
Trees2Kernel::cExpAssertlfeed (ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  ptlrepid outputs;
  boolean ok;
  ntrees voperand;

  ok = pk_false;
  if (isChild (1, p))
    {
      voperand = operande;
      if (verifyOperatorsAssert (operande))
	{
	  operande = voperand;
	  voperator = noper (operande);
	  outputs = NILLREPID;
	  ok = klparcours (inputs, &outputs, vklExprProc);
	  if (ok)
	    {
	      if (outputs == NILLREPID)
		kl_setCompositionAssertion (*vklExprProc);
	      else
		{
		  mkklErrorAnnotTree (errsortid, p);
		  ok = pk_false;
		}
	    }
	}
    }
  kl_setAssertionsFlag (G_process, pk_true);
  return ok;
}

/**
   Compiling  a tree (p) such that E1 :=: E2.
   @return "true" if the compiling succeeds, "false" otherwise.
   
   It produces:
   - vklExprProc the kernel representation of the expression 
   - inputs the inputs of the expression (signals, labels)
*/
_VIRTUAL_ boolean
Trees2Kernel::cExpDefEgal (ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  /* 
     --->>>>  Note for LBesnard : See the method XmiTranslator::cExpDefEgal()
     boolean b;
     ptlrepid indexList, lparam;

     indexList = lparam = NILLREPID;
     if (!isChild(1, p)) return pk_false;
     vklExprSig1 = cExprSig(operande, inputs, &lparam, &indexList);
     if (vklExprSig1 == NILEXPRSIG) return pk_false;
     if (!isChild(2, p)) return pk_false;
     vklExprSig2 = cExprSig(operande, inputs, &lparam, &indexList);
     if (vklExprSig2 == NILEXPRSIG) return pk_false;
     *vklExprProc = kl_mkEquation(opdefegal, *inputs, NILLREPID, *lparam, p, vklExprSig1, vklExprSig2);
   */
  notYetImplemented (p);
  return pk_false;
}

/**
   It translates into the kernel tree representation the current tree (operande) with voperator as operator.
   (operande, voperator are attribute members of the Trees2Kernel class). 
   It is assumed that operande is an expression on process or a list of such expressions.
   @return "true" if the compiling succeeds, "false" otherwise (errors, warnings are sticked to the abstract tree)

   It calls 
   - cExpLabelledExp label expression on process
   - cExplDependences list of dependences (l1 -> l2 -> ... -> ln)
   - cExpDependence for a dependence 
_VIRTUAL_
   - cExpSynchro  for an expression of syncrhonization (^<, ^>,  ^=, ^#)
_VIRTUAL_
   - cTupleDef for a definition of a tuple  ( (x1,...xn):= E, (x1,...xn)::= E)
   - cExpEquation  for a definition of an equation or a delay  ( x:=E, x::=E)
   - cExpRefToModel  for a reference to a model P{}
   - cExpCall  for a reference to a model P{} (e1,...,en)
   - cExpCase  for a case process 
   - cExpConfinedExp for a confined process ( P where local_declarations end)
   - cExpMasking  for a masking process ( P / x1,...,xn )
   - cExpArray for an array process  (array I to n of P with Q end)
   - cExpIterate for an iterate process (iterate I of P with Q end) 

   It produces: 
   - vklExprProc the kernel representation of the expression.
   - inputs the inputs of the expression (signals, labels)
   - outputs the outputs of the expression
*/
_VIRTUAL_ boolean
Trees2Kernel::klparcours (ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ptlrepid lparam, indexList;
  boolean b;
  noperators op, vop;
  ntrees p;

  if (STree_isUndefinedTree (operande)) return pk_true;
  p = operande;
  vop = voperator;
  initklparcours (p, vop);
  switch ((int) vop)
    {
    case oppetiq: // lab :: (|  |)
      b = cExpLabelledExp (operande, inputs, outputs, vklExprProc);
      break;

    case oppcontrol:  // on trigger ::  (| |)
       b = cExpControlledExp (operande, inputs, outputs, vklExprProc);
      break;

    case opldep:
      b = cExplDependences (operande, inputs, vklExprProc);
      break;
    case opdepd:
      b = cExpDependence (operande, inputs, vklExprProc);
      break;
    case oplcrohor:		// list of "^<"
    case opldechor:		// list of "^>"
    case oplequhor:		// list of ^=
    case oplexchor:		// list of "^#"
      b = cExpSynchro (operande, voperator, inputs, outputs, &indexList, &lparam, vklExprProc);
      break;

    case opdefpart:
      //  "::="
      if (noper (nchild (1, operande)) == opleflot)	// tuple
	b = cTupleDef (opdefpart, operande, inputs, outputs, &indexList, &lparam, vklExprProc);
      else
	b = cExpEquation (opdefpart, operande, inputs, outputs, &indexList, &lparam, vklExprProc);
      break;
    case opdefpartdef:		// <expr_signal> "::=" "defaultvalue" <expr_signal> ;
      // A verifier: endroit de la declaration du signal
      b = cExpEquation (opdefpartdef, operande, inputs, outputs, &indexList, &lparam, vklExprProc);
      break;

    case opdefsig:
      //:=
      if (noper (nchild (1, operande)) == opleflot)	// tuple
	b = cTupleDef (opdefsig, operande, inputs, outputs, &indexList, &lparam, vklExprProc);
      else
	b = cExpEquation (opdefsig, operande, inputs, outputs, &indexList, &lparam, vklExprProc);
      break;

    case opinst:
      // <production>::= <identificateur> "{" <o_l_expr_param> "}" ;
      op = noper (nfather (operande));
      if ((op != oplfeed) && (op != oppetiq) && (op != oppcontrol))
	{
	  mkklErrorAnnotTree (errprocesscall, operande);
	  b = pk_false;
	}
      else
	b = cExpRefToModel (operande, inputs, outputs, vklExprProc);
      break;

    case opappe:
      // <appel>::= <reference_processus> "(" <o_l_expr_signal> ")" ;
      //  <reference_processus>::= <production> ;
      // <reference_processus>::= <identificateur> ;
      b = cExpCall (operande, inputs, outputs, &indexList, &lparam, vklExprProc);
      break;

    case opobli:		// masquage --> traitement processus confiné, mais ...
      // <expr_profil>::= <term_processus> "/" <l_identificateurs> ;
      b = cExpMasking (operande, inputs, outputs, vklExprProc);
      break;

    case oplfeed:
      // <l_composition>::= <l_composition> "|" <expr_processus> ;
      b = cExpListExpressionsOnProcesses (operande, voperator, inputs, outputs, vklExprProc);
      break;

    case oppcas:
      // <expr_case>::= "case" <identificateur> "in" <l_cas> <o_cas_sinon> "end" ;
      b = cExpCase (operande, inputs, outputs, vklExprProc);
      break;

    case oppconf:
      // <processus_confine>::= <primaire_processus> <bloc_declarations> ;
      b = cExpConfinedExp (operande, inputs, outputs, vklExprProc);
      break;

    case opmoti:
      // <motif>::= "array" <indice_motif> "of" <expr_processus> <o_init_motif> "end" ;
      b = cExpArray (operande, inputs, outputs, vklExprProc);
      break;

    case opiter:
      // <motif>::= "iterate" <expr_signal> "of" <expr_processus> <o_init_motif> "end" ;
      b = cExpIterate (operande, inputs, outputs, vklExprProc);
      break;

    case opappeassert:
      // assert(c)
      b = cExpAssert (operande, inputs, vklExprProc);
      break;
    case oppassert:
      //  assert(|  |)
      b = cExpAssertlfeed (operande, inputs, vklExprProc);
      break;

      /* ---Not implemeted--- */
    case opdefegal:		//:=:
      b = cExpDefEgal (operande, inputs, vklExprProc);
      break;

    default:
      notYetImplemented (operande);
      b = pk_false;
    }
  finalklparcours (p, vop);
  return b;
}

/**
   @return "true" if the current tree (operande) is an identifier of a type, "false" otherwise
   \note the predefined types are allowed here, but it is not in the SIGNAL V4 definition.
*/
_VIRTUAL_ boolean
Trees2Kernel::mkklGetTypeIdStack (ptIdStack * varid)
{
  ptident id;
  ntrees s;
  noperators so;

  // The name of a type may be an operator of a predefined type (for example type real = )
  s = operande;
  so = voperator;
  if (locate (operande, varid))
    return pk_true;		// opnom, opident
  operande = s;
  voperator = so;
  // type predefini 
  id = repstr2ident (predefinedType2ident (voperator));
  if (id == NILIDENT)
    {
      mkklErrorAnnotTree (errargmanq, operande);
      return pk_false;
    }
  return locateOrError (id, varid);
}

/**
   New New ... to be documented
*/
_VIRTUAL_ void
Trees2Kernel::parametersInputsOutputsCompiling (ntrees p, ptlrepid * param, ptlrepid * inList, ptlrepid * outList)
{
  ntrees vio;

  *param = cDecFlowsInterface (1, p, pk_true);	// parameters: they are pushed
  vio = nchild (2, p);
  *inList = cDecFlowsInterface (1, vio, pk_false);	// inputs
  *outList = cDecFlowsInterface (2, vio, pk_false);	// outputs.
  popFlows (*param);		// parameters: they are popped
}

/**
  It compiles a list of declarations of types assumed to be available in operande (attribute member of the class).
  @return deblist + the representation of the declarations 
  (the identifiers (string in trees structure) are replaced by a reference to the stack 
   associated with the identifier, i.e. the references are solved.)

  type T:= external, type T:= A , type T:= struct ...
  If is_param is "true", it specifies that the current operation is internal to the compiling of parameters (types are allowed); 
  in this case, it is possible to have a list of names.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecType (ptlrepid deblist, boolean is_param)
{
  ntrees pred, currentDecl;
  boolean result, ok;
  ptIdStack varid;
  ptreptyp vreptyp;
  ptlrepid xlrepid;
  ptIdentProp topId;

  pred = operande;
  if (firstChild (&pred))
    do
      {
	currentDecl = operande;
	ok = pk_true;
	switch (voperator)
	  {
	  case opnom:
	  case opident:
	    // Virtual type 
	    result = mkklGetTypeIdStack (&varid);
	    xlrepid = NILLREPID;
	    if (result)
	      {
		if (getTypeClassTop (varid, &topId))
		  xlrepid = mkRepidsigRef (varid, topId);
		else if (!is_param)
		  mkklErrorAnnotTree (errNoDefVirtualType, currentDecl);
	      }
	    // virtual = textern + xlrepid != NILLREPID
	    vreptyp = kl_mkType (textern, NULLOP, NILEXPRSIG, xlrepid, NILREPFIELD, NILLPOINTER);
	    break;

	  case optype:
	    if (is_param)
	      {
		mkklErrorAnnotTree (errdecl, operande);
		ok = pk_false;
	      }
	    else
	      {
		// ntrees aux = operande;
		if (isChild (1, operande))
		  result = mkklGetTypeIdStack (&varid);
		// if (isChild(2, aux)) vreptyp = cExpType(operande);replaced by (Oct 2011)
		vreptyp = kl_mkType (textern, NULLOP, NILEXPRSIG, NILLREPID, NILREPFIELD, NILLPOINTER);
	      }
	    break;
	  case optypeproc:
	  case optypeact:
	  case optypefunc:
	  case optypenode:
	  case optypeprocedure:
	  case optypeautomaton:
	    // See cDecModelType()
	    ok = pk_false;
	    break;

	  default:
	    notYetImplemented (operande);
	    ok = pk_false;
	  }
	if (ok)
	  {
	    kl_setTypeTreeDecl (vreptyp, currentDecl);
	    if (!appendProtoTyp (&deblist, varid, vreptyp, &xlrepid))
	      mkklErrorAnnotTreeName (errdoubledecl, currentDecl, varid);
	    kl_pushType (varid, ptlrepid_getIdentProp (xlrepid));
	  }
      }
    while (nextChild (&pred));

  return deblist;
}

/**
  It compiles a list of declarations of types assumed to be available in operande (attribute member of the class).
  @return deblist + the representation of the declarations 
  (the identifiers (string in trees structure) are replaced by a reference to the stack 
   associated with the identifier, i.e. the references are solved.)

  type T:= external, type T:= A , type T:= struct ...
  If is_param is "true", it specifies that the current operation is internal to the compiling of parameters (types are allowed); 
  in this case, it is possible to have a list of names.
*/
_VIRTUAL_ void
Trees2Kernel::cDecTypeDef (ptlrepid vlrepid, boolean is_param)
{
  ntrees p;
  ptreptyp vreptypdef, vreptyp;

  // ptIdStack varid = nameId (vlrepid);
  vreptyp = getTypeRepid (vlrepid);
  p = kl_getTypeTreeDecl (vreptyp);
  operande = p;
  voperator = noper (p);
  if (voperator == optype)
    if (isChild (2, operande))
      {
	vreptypdef = cExpType (operande);
	if (vreptypdef != NILREPTYP)
	  kl_typereplace (vreptyp, vreptypdef);
      }
}

/*
  It verifies that 
  - a node/function model has at least one output.
  - a procedure has at least one input-output.
  @return "true" if no detected error, "false" otherwise.
*/
_VIRTUAL_ boolean
Trees2Kernel::verifOutputsModel (ptlrepid inList, ptlrepid outList, noperators currentOp, ntrees currentDecl)
{

  if (outList != NILLREPID) return pk_true;
  if (((currentOp == optypefunc) || (currentOp == opfunc) || (currentOp == opfuncp))
      || ((currentOp == optypenode) || (currentOp == opnode) || (currentOp == opnodep))
      || (((currentOp == optypeproc) || (currentOp == opprocedure) || (currentOp == opprocedurep)) && (inList == NILLREPID)) )
    {
      mkklErrorAnnotTree (errModelNbIO, currentDecl);
      return pk_false;
    }
  return pk_true;
}

/**
 It compiles the "type interface" members of a list of declarations.
 It exists two forms, where OP denotes an element of { process, function, action, node}
   - type OPERATOR name = name; OPERATOR
   - type OPERATOR name = (identificateur, def_interface, o_l_pragmas)

  @return deblist + the representation of the reached declarations 
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecTypeInterface (ntrees p, ptlrepid deblist)
{
  ntrees p_graph_extern, interface, pred, currentDecl;
  ptIdStack varid;
  ptlrepid xlrepid, param, inList, outList;
  ptrepinterface vinterface;
  boolean ok;
  noperators currentOp;

  pred = p;
  if (firstChild (&pred))
    do
      {
	currentDecl = operande;
	currentOp = voperator;
	if (LL_isModelType (voperator))	// type XXX ID= INTERFACE;
	  {
	    // type CLASSINTERFACE NAME (<identificateur>, <def_interface>, <o_l_pragmas>), where CLASSINTERFACE is process, node, function or action.  
	    ok = pk_true;
	    varid = locateModelName (currentDecl);
	    vinterface = kl_mkInterface (currentDecl, identpile (varid), currentOp);
	    if (isChild (2, currentDecl))
	      {
		if (voperator == opinte)	// param, inputs, outputs, spec, pragmas.
		  {
		    interface = operande;
		    parametersInputsOutputsCompiling (interface, &param, &inList, &outList);
		    kl_setInterfaceParameters (vinterface, param);
		    kl_setInterfaceInputs (vinterface, inList);
		    kl_setInterfaceOutputs (vinterface, outList);
		    // third child is the external graph.
		    if (isoChild (3, interface))
		      {
			p_graph_extern = operande;
			_PUSH_INTERFACE (param, inList, outList);
			cExternalGraph (p_graph_extern, vinterface, currentDecl, inList, outList, NILLREPID, NILLREPID, pk_true);
			_POP_INTERFACE (param, inList, outList);
		      }
		    verifOutputsModel (inList, outList, currentOp, currentDecl);
		  }
		else		// ident : a name of a model type. type process T = I  begin pragmas <o_l_pragmas> end_pragmas.
		  {
		    notYetImplemented (operande);
		    ok = pk_false;
		  }
	      }
	    // Pragmas
	    if (isoChild (3, currentDecl))
	      cPragmas (operande, vinterface);

	    if (ok)
	      {
		if (!appendProtoModelType (&deblist, varid, vinterface, &xlrepid))
		  mkklErrorAnnotTreeName (errdoubledecl, currentDecl, varid);
		kl_pushModelType (varid, ptlrepid_getIdentProp (xlrepid));
	      }
	  }
      }
    while (nextChild (&pred));
  return deblist;
}


/**
  It compiles a list of declarations of constants(p)
  @return deblist + the representation of the declarations.
  (the identifiers (string in trees structure) are replaced by a reference to the stack 
  associated with the identifier, i.e. the references are solved.)
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecConstant (ntrees p, ptlrepid deblist)
{
  ntrees vtyptree, l_cst, p_cst, p_name;
  ptrepproto vrepproto;
  ptIdStack varid;
  integer i, nb;
  noperators op;
  ptlrepid xlrepid;
  boolean ok;
  ptIdentProp topId;

  l_cst = nchild (2, p);
  nb = ntreearity (l_cst);
  if (nb <= 0)
    return deblist;
  vtyptree = nchild (1, p);
  for (i = 1; i <= nb; i++)
    {
      p_cst = nchild (i, l_cst);
      op = noper (p_cst);
      if (OPIS_SIGNALIDENTIFIER (op))
	/*  virtual constant */
	{
	  p_name = p_cst;
	  ok = locateOrError (p_name, &varid);
	  if (ok)
	    if (!getConstantClassTop (varid, &topId))
	      mkklErrorAnnotTree (errNoDefVirtualConst, p_cst);
	}
      else
	{
	  p_name = nchild (1, p_cst);
	  ok = locateOrError (p_name, &varid);
	}
      if (ok)
	{
	  vrepproto = mkrepproto (NILREPTYP, NILEXPRSIG, undefCategory, p_cst, vtyptree);
	  if (!appendProtoId (&deblist, varid, vrepproto, &xlrepid))
	    mkklErrorAnnotTreeName (errdoubledecl, p_cst, varid);
	  kl_pushConstant (varid, ptlrepid_getIdentProp (xlrepid));
	}
    }
  return deblist;
}


/**
  It compiles a list of declarations of constants(p)
  @return deblist + the representation of the declarations.
  (the identifiers (string in trees structure) are replaced by a reference to the stack 
  associated with the identifier, i.e. the references are solved.)
*/
_VIRTUAL_ void
Trees2Kernel::cDecConstantDef (ptlrepid vlrepid)
{
  ptlrepid l1, l2, l3;
  ptrepproto vrepproto;
  ptklExprSig val;
  ntrees vtypetree, p, p_val;

  // ptIdStack varid = nameId (vlrepid);
  vrepproto = getIdRepid (vlrepid);
  vtypetree = ptrepproto_getTypeTree (vrepproto);

  operande = vtypetree;
  setRepprotoType (vrepproto, cExpType (vtypetree));

  p = ptrepproto_getTree (vrepproto);
  voperator = noper (p);
  if (!(OPIS_SIGNALIDENTIFIER (voperator)))
    {
      p_val = nchild (2, p);	/* value of the constant */
      l1 = NILLREPID;
      l2 = NILLREPID;
      l3 = NILLREPID;
      val = cExprSig (p_val, &l1, &l2, &l3);
      while (l1 != NILLREPID)
	{
	  mkklErrorAnnotTreeName (errparnondecl, p_val, nameId (l1));
	  l1 = succid (l1);
	}
      setRepprotoValue (vrepproto, val);
    }
  else;				/*  virtual constant */

}



/**
   @return the expression dim (assumed to be a list)  in which trep is added at the end.
*/
_VIRTUAL_ ptklExprSig
Trees2Kernel::glueListExprDim (ptklExprSig dim, ptklExprSig trep)
{
  integer i, n;
  ptklExprSig aux;
  
  if (dim == NILEXPRSIG)
    return trep;
  aux = dim;
  n = kl_getSizeExpr (dim);
  for (i = 1; i < n; i++) 
    { 
      kl_getNextExpr (&aux);
    }
  kl_setBinaryExprChild2 (aux, trep);
  return dim;
}

/**
   COmpiling a bundle type (p) in a context of an the array type with the given dimensions (vdim).
*/
_VIRTUAL_ ptreptyp
Trees2Kernel::cBundleType (ntrees p, ptklExprSig vdim)
{
  ptrepfield lfields;
  ptreptyp deftyp;
  ptlrepid lrepidfields;

  lfields = cExpFields (p, &lrepidfields);
  deftyp = kl_mkType (tbund, NULLOP, vdim, NILLREPID, lfields, NILLPOINTER);
  kl_setSpecBundleType (deftyp, cExpBundleSpec (p, lfields, lrepidfields), lrepidfields);
  return deftyp;
}

/** 
    It compiles a definition of a type (p) when it is a vectorial type, struct, bundle.
    It produces the dimension (dim) of the type (vectorial type.)

    @return the representation of the type in Basic Signal (see kl_mkType()).    
*/
_VIRTUAL_ ptreptyp
Trees2Kernel::cExpTyperec (ntrees p, ptklExprSig * dim)
{
  ptklExprSig trep;
  ptreptyp deftyp;
  ntrees aux;
  ptlrepid laux1, laux2, laux3, vlrepid;
  ptIdStack name;

  aux = p;
  trep = NILEXPRSIG;
  deftyp = NILREPTYP;
  switch (voperator)
    {
    case optvect:
      if (isChild (1, aux))
	{
	  laux1 = NILLREPID;
	  laux2 = NILLREPID;
	  laux3 = NILLREPID;
	  trep = cExprSig (operande, &laux1, &laux2, &laux3);
	  if (laux1 != NILLREPID)
	    mkklErrorAnnotTreeName (errparnondecl, operande, nameId (laux1));
	  trep = glueListExprDim (*dim, trep);
	}
      else
	trep = *dim;
      if (isChild (2, aux))
	{
	  switch (voperator)
	    {
	    case optvect:
	    case opstru:
	    case opbund:
	      deftyp = cExpTyperec (operande, &trep);
	      break;
	    case opnom:
	    case opident:
	      locateOrError (operande, &name);
	      vlrepid = mkRepidsigRef (name, kl_top (name));
	      if (kl_isTopaType (name))
		deftyp = kl_mkType (tnom, NULLOP, trep, vlrepid, NILREPFIELD, NILLPOINTER);
	      else
		mkklErrorAnnotTreeName (errtypnondec, operande, name);

	      break;
	    default:
	      deftyp = kl_mkType (tbase, voperator, trep, NILLREPID, NILREPFIELD, NILLPOINTER);
	      break;
	    }
	}
      else
	deftyp = kl_mkType (tbase, opimplicit, trep, NILLREPID, NILREPFIELD, NILLPOINTER);
      break;

    case opstru:
      laux1 = NILLREPID;
      deftyp = kl_mkType (tstruct, NULLOP, *dim, NILLREPID, cExpFields (operande, &laux1), NILLPOINTER);
      break;
    case opbund:
      deftyp = cBundleType (aux, *dim);
      break;
    default:
      notYetImplemented (operande);
    }
  *dim = trep;
  return deftyp;
}


/**
   It compiles a field (p) of a structure type and produces a new representation in the kernel tree (kl_mkFieldType).
   It is added 
    - to the list of fields (field_prec) 
    - to the list of the names (l_name_fields). This list is used to verify the unicity of the names of the fields of the structure.
    
    @return the list field_prec in which the current compiled element is added.
*/
_VIRTUAL_ ptrepfield
Trees2Kernel::cDecFields (ntrees p, ptrepfield field_prec, ptlrepid * l_name_fields)
{
  ntrees pred;
  ptreptyp deftyp;
  ptrepfield vfield, premier_field;
  ptIdStack varid;
  boolean premier_tour;
  ptrepproto vrepproto;
  ptlrepid xlrepid;

  premier_field = NILREPFIELD;
  pred = p;
  if (noper (p) != opdecsig)
    {
      mkklErrorAnnotTree (errdecl, p);
      return premier_field;
    }

  if (isoChild (1, operande))
    deftyp = cExpType (operande);
  else
    deftyp = kl_mkBasicType (opimplicit);

  premier_tour = pk_true;
  if (isChild (2, pred))	/* Several names are allowed for a field */
    {
      field_prec = kl_getLastField (field_prec);
      pred = operande;
      if (firstChild (&pred))
	do
	  {
	    if (OPIS_SIGNALIDENTIFIER (voperator))
	      {
		if (locateOrError (operande, &varid))
		  {
		    vrepproto = mkrepproto (deftyp, NILEXPRSIG, undefCategory, pred, NILTREE);
		    if (!appendProtoId (l_name_fields, varid, vrepproto, &xlrepid))
		      mkklErrorAnnotTreeName (errdoubledecl, operande, varid);
		    vfield = kl_mkFieldType (ident2repstr (operande), deftyp, field_prec);	// modif LBesnard Mar 2015 nstringof(operande) replaced by ident2repstr(operande)
		    field_prec = vfield;
		    if (premier_tour)
		      {
			premier_field = vfield;
			premier_tour = pk_false;
		      }
		  }
	      }
	    else
	      mkklErrorAnnotTree (errdecl, operande);
	  }
	while (nextChild (&pred));
    }
  return (premier_field);
}

/**
   It compiles the "spec" part of a bundle type (vBundleDcl).
   It is not yet implemented.
   @return a null expression (presently).
*/
_VIRTUAL_ ptklExprProc
Trees2Kernel::cExpBundleSpec (ntrees vBundleDcl, ptrepfield lfields, ptlrepid lrepidfields)
{
  ptklExprProc vklExprProc;
  ptlrepid inputs, outputs;

  if (!isoChild (2, vBundleDcl))
    return NILEXPFIL;
  // ntrees vspec = operande;
  inputs = NILLREPID;
  outputs = NILLREPID;
  vklExprProc = NILEXPFIL;
  pushFlows (lrepidfields);
  /* boolean ok =*/ klparcours (&inputs, &outputs, &vklExprProc);
  popFlows (lrepidfields);
  // outputs must be empty, and inputs = lrepidfields
  /* Already detected in klparcours()
     #define erroutbundlespec 23 ; and in messignal.2.txt : 23%e The identifier &1 is not a field of the bundle%
     while (outputs != NILLREPID)
     {
     varid = nameId(outputs);
     mkklErrorAnnotTreeName(erroutbundlespec, vBundleDcl, varid);
     outputs = succid(outputs);
     }
     while (inputs != NILLREPID)
     { 
     varid = nameId(inputs);
     if ( !ptlrepid_isIn_Unsorted(lrepidfields, varid)) mkklErrorAnnotTreeName(erroutbundlespec, vBundleDcl, varid);
     inputs = succid(inputs);
     } 
   */
  return vklExprProc;
}

/**
   It compile a declaration of a \b structure type (p).
   It calls cDecFields() for each field of the structure type.
   @return the list of the fields in the kernel representation (see kl_mkFieldType() for the representation of a structure type in the 
   kernel representation).
*/
_VIRTUAL_ ptrepfield
Trees2Kernel::cExpFields (ntrees p, ptlrepid * l_name_fields)
{
  ntrees p_field;
  ptrepfield vfield, field_prec, premChamps;

  (*l_name_fields) = NILLREPID;
  if (!isChild (1, p))
    return NILREPFIELD;
  premChamps = NILREPFIELD;
  p_field = operande;
  field_prec = NILREPFIELD;
  if (firstChild (&p_field))
    do
      {
	vfield = cDecFields (operande, field_prec, l_name_fields);
	if (field_prec == NILREPFIELD)
	  premChamps = vfield;
	field_prec = vfield;
      }
    while (nextChild (&p_field));
  else
    mkklErrorAnnotTree (errargmanq, p);
  return premChamps;
}


/**
   It compile the expression definition of a type (p).
   It calls 
   - cExpTyperec() for a vectorial type
   - kl_mkType() for external, named, and predefined type, and enumeration (with the list of the values)
   - cExpFields() for bundle, structure
   @return the kernel representation of the type expression (kl_mkType())
*/
_VIRTUAL_ ptreptyp
Trees2Kernel::cExpType (ntrees p)
{
  int i, n;
  ptrepstr name_enum;
  ptlPointer l_enum;
  ptreptyp deftyp;
  ptlrepid vlrepid;
  ptklExprSig dim;
  ntrees p_fils;
  ptIdStack name;

  deftyp = NILREPTYP;
  voperator = noper (p);
  switch (voperator)
    {
    case optvect:
      dim = NILEXPRSIG;
      deftyp = cExpTyperec (p, &dim);
      break;

    case opnom:
    case opident:
      locateOrError (p, &name);
      vlrepid = mkRepidsigRef (name, kl_top (name));
      if (kl_isTopaType (name))
	deftyp = kl_mkType (tnom, NULLOP, NILEXPRSIG, vlrepid, NILREPFIELD, NILLPOINTER);
      else
	mkklErrorAnnotTreeName (errtypnondec, p, name);
      break;

    case opexte:		/* occurrence when it is called from cDecType() */
      deftyp = kl_mkType (textern, NULLOP, NILEXPRSIG, NILLREPID, NILREPFIELD, NILLPOINTER);
      break;

    case optenum:
      n = ntreearity (p);
      l_enum = NILLPOINTER;
      for (i = 1; i <= n; i++)
	{
	  p_fils = nchild (i, p);
	  name_enum = STree_getIdentStr (p_fils);
	  if (!ptlPointer_insert_elem (&l_enum, (pointer) name_enum))
	    {
	      locateOrError (p_fils, &name);
	      mkklErrorAnnotTree (errdoubledecl, p_fils);
	    }
	}
      deftyp = kl_mkType (tenum, NULLOP, NILEXPRSIG, NILLREPID, NILREPFIELD, l_enum);
      break;

    case opstru:
      deftyp = kl_mkType (tstruct, NULLOP, NILEXPRSIG, NILLREPID, cExpFields (p, &vlrepid), NILLPOINTER);
      break;

    case opbund:
      deftyp = cBundleType (p, NILEXPRSIG);
      break;

    default:
      deftyp = kl_mkType (tbase, voperator, NILEXPRSIG, NILLREPID, NILREPFIELD, NILLPOINTER);
    }
  if (deftyp != NILREPTYP)
    {
      kl_setTypeTreeDecl (deftyp, p);
    }
  return (deftyp);
}


/**
   It compiles a declaration of flows (an optional type followed by a list of identifiers) assumed to be available in 
   operande (attribute member of the class).
   
   If is_param, it is a declaration of some parameters (they are pushed here), 
   otherwise it is
   - a declaration of flows (when opContext=opdecsig)
   - a declaration of state variables (when opContext=opdecvar)
   - or, a declaration of shared variables (when opContext=opdecpar).

   It verifies the unicity of the declaration of a identifier (errors, warnings are sticked to the abstract tree).
   @return deblist + the identifiers reached in the declaration.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecFlows (noperators opContext, ptlrepid deblist, boolean is_param)
{
  ptlrepid laux1, laux2, laux3, xlrepid;
  ntrees arbreid;
  ptklExprSig vali;
  ptreptyp deftyp;
  ptrepproto vrepproto;
  ntrees aux, pred;
  ptIdStack varid;
  boolean result;
  ptident id;
  categoryType vcat;

  pred = operande;
  //  arbreid = operande;
  // type
  if (isoChild (1, operande))
    deftyp = cExpType (operande);
  else
    deftyp = kl_mkBasicType (opimplicit);
  // signal and initialisation
  if (isChild (2, pred))	/* on compile le signal et son initialisation */
    {
      pred = operande;
      if (firstChild (&pred))
	do
	  {
	    if (voperator == opinit)
	      {
		aux = operande;
		if (isChild (2, operande))
		  {
		    laux1 = NILLREPID;
		    laux2 = NILLREPID;
		    laux3 = NILLREPID;
		    vali = cExprSig (operande, &laux1, &laux2, &laux3);
		    while (laux1 != NILLREPID)
		      {
			mkklErrorAnnotTreeName (errparnondecl, aux, nameId (laux1));
			laux1 = succid (laux1);
		      }
		  }
		else
		  vali = NILEXPRSIG;
		result = locateChild (aux, 1, &varid);
		arbreid = operande;
	      }
	    else
	      {
		result = locateOrError (operande, &varid);
		vali = NILEXPRSIG;
		arbreid = operande;
	      }
	    if (opContext == opdecvar)
	      vcat = statevarSignal;
	    else if (opContext == opdecpar)
	      vcat = sharedSignal;
	    else
	      vcat = undefCategory;
	    vrepproto = mkrepproto (deftyp, vali, vcat, arbreid, NILTREE);
	    if (is_param)	/* -- parameter-- */
	      {
		if (result)
		  {
		    if (!appendProtoId (&deblist, varid, vrepproto, &xlrepid))
		      mkklErrorAnnotTreeName (errdoubledecl, operande, varid);
		  }
		else
		  {
		    id = identpile (getidnil ());
		    varid = kl_mkIdStack (id);
		    appendProtoId (&deblist, varid, vrepproto, &xlrepid);
		  }
		kl_pushParameter (varid, ptlrepid_getIdentProp (xlrepid));
	      }
	    else if (result)	/* cas des signaux, indices */
	      if (!appendProtoId (&deblist, varid, vrepproto, &xlrepid))
		mkklErrorAnnotTreeName (errdoubledecl, operande, varid);
	  }
	while (nextChild (&pred));
    }
  return deblist;
}

/** Compiling a formal parameter when it is a formal process.
    {..., process TypeInterface Name, ...}
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecFormalProcess (noperators vop, ntrees aFormalProcess, ptlrepid lparam)
{
  ptrepinterface vinterface;
  ptIdStack idInterface, idModel;
  ptlrepid xlrepid;
  ptIdentProp topId;
  ptrepfi vfi;

  // Name of the type interface.
  if (isChild (1, aFormalProcess))
    {
      if (locateOrError (operande, &idInterface))
	{
	  if (getModelTypeClassTop (idInterface, &topId))
	    {
	      vinterface = ptIdentProp_getModelType (topId);
	      vop = LL_getFormalModelType (vop);
	      if (vop != kl_getInterfaceOperator (vinterface))
		mkklErrorAnnotTree (errConflictTypeModelFormalModel, aFormalProcess);
	    }
	  else
	    mkklErrorAnnotTreeName (errdecl, aFormalProcess, idInterface);
	}
      else
	vinterface = NILREPINTERFACE;
    }
  // Name of the formal parameter.
  if (locateChild (aFormalProcess, 2, &idModel));
  else
    idModel = kl_mkIdStack (identpile (getidnil ()));
  vfi = kl_mkModel (aFormalProcess, identpile (idModel), vop);
  kl_setInterfaceModel (vfi, vinterface);
  kl_belongsFormalModel (vfi);
  if (!appendrepfi (&lparam, idModel, vfi, &xlrepid))
    mkklErrorAnnotTreeName (errdoubledecl, aFormalProcess, idModel);
  kl_pushModel (idModel, ptlrepid_getIdentProp (xlrepid));	// It IS PUSHED
  return lparam;
}


/**
   It compiles the n-th element of a list of declarations of flows parameters or types (p).
   It is called to compile the interface of a model. It is the parameters (when is_param), the inputs/outputs otherwise.

   It calls cDecFlows(), cDecType()

   @return the identifiers reached in the list of declarations.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecFlowsInterface (integer n, ntrees p, boolean is_param)
{
  ptlrepid deblist;

  if (!isoChild (n, p))
    return NILLREPID;
  p = operande;
  if (!firstChild (&p))
    return NILLREPID;
  deblist = NILLREPID;
  do
    {
      switch (voperator)
	{
	case opdecsig:
	  deblist = cDecFlows (voperator, deblist, is_param);
	  break;
	case opltype:
	case opltypep:
	  deblist = cDecType (deblist, is_param);
	  break;

	case opdecpar:		// shared
	case opdecvar:		// statevar
	case opref:
	case oplab:
	case opdeccon:
	case opdecconp:
	  /* declaration de constantes, de variable d'etat, d'étiquette de reference interdite dans l'interface */
	  mkklErrorAnnotTree (errnosens, operande);
	  break;

	case opformfunc:
	case opformnode:
	case opformproc:
	case opformact:
	case opformprocedure:
	case opformautomaton:
	  deblist = cDecFormalProcess (voperator, operande, deblist);
	  break;
	}
    }
  while (nextChild (&p));
  return deblist;
}


/**
   It compiles a list of declarations (p), selecting only the constants/types? i.e the declarations of flows, imports are excluded.
   It applies cDecType() or cDecConstant() for each element of the list.
   @return the identifiers reached in the list of declarations.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecLocalConstantsTypes (ntrees p)
{
  ntrees list;
  ptlrepid deblist;

  list = p;
  if (STree_isUndefinedTree (list))
    return NILLREPID;

  deblist = NILLREPID;
  if (firstChild2 (&list))
    do
      switch (voperator)
	{
	case opltype:
	case opltypep:
	  deblist = cDecType (deblist, pk_false);
	  break;
	case opdeccon:
	case opdecconp:
	  deblist = cDecConstant (operande, deblist);
	  break;
	case opdecsig:
	case opdecvar:		// statevar
	case opdecpar:		// shared
	case oplimpo:
	  break;

	}
    while (nextChild (&list));

  return deblist;
}

void
Trees2Kernel::initBuildPredTypeConstBlock (ptlrepid vlrepid)
{
  modeConstTypeCompiling = pk_true;
  lConstTypesCurrentBlock = vlrepid;
  lpredConstTypes.clear ();
}

void
Trees2Kernel::removeReferencesTo (ptlrepid velem)
{
  ptlrepid vc;
  TablePredConstTypeType::iterator i;

  i = lpredConstTypes.begin ();
  while (i != lpredConstTypes.end ())
    {
      vc = (*i).second;
      if (ptlrepid_remove_Unsorted (&vc, nameId (velem)))
	(*i).second = vc;
      i++;
    }
}

ptlrepid
Trees2Kernel::finalBuildPredTypeConstBlock ()
{
  ptlrepid  vc, sortedList;
  TablePredConstTypeType::iterator i, j;
  boolean goon;
  ntrees p;

  sortedList = NILLREPID;
  i = lpredConstTypes.begin ();
  while (i != lpredConstTypes.end ())
    {
      vc = (*i).first;
      setsuccid (vc, NILLREPID);
      i++;
    }
  goon = pk_true;
  while (goon)
    {
      goon = pk_false;
      i = lpredConstTypes.begin ();
      while (i != lpredConstTypes.end ())
	{
	  if ((*i).second == NILLREPID)
	    {
	      vc = (*i).first;
	      appendrepid (&sortedList, vc);
	      removeReferencesTo (vc);
	      j = i;
	      i++;
	      lpredConstTypes.erase (j);
	      goon = pk_true;
	    }
	  else
	    i++;
	}
    }
  if (lpredConstTypes.size () != 0)
    {
      i = lpredConstTypes.begin ();
      vc = (*i).first;
      if (isTypeId (vc))
	p = kl_getTypeTreeDecl (getTypeRepid (vc));
      else
	p = ptrepproto_getTree (getIdRepid (vc));
      mkklErrorAnnotTreeName (errCycleDefConstType, p, nameId (vc));
    }

  modeConstTypeCompiling = pk_false;
  lpredConstTypes.clear ();
  return sortedList;
}

void
Trees2Kernel::setBuildPredTypeConst (ptlrepid vTypeConst)
{
  if (modeConstTypeCompiling)
    {
      lpredConstTypes[vTypeConst] = NILLREPID;
      currentTypeConst = vTypeConst;
    }
}

void
Trees2Kernel::addPredTypeConst (ptIdStack elem)
{
  ptlrepid l;
  ptIdentProp topId;
  boolean b;

  if (modeConstTypeCompiling)
    {
      if (getConstantClassTop (elem, &topId))
	b = pk_true;
      else
	b = getTypeClassTop (elem, &topId);
      if (b && ptlrepid_isIn_Unsorted (lConstTypesCurrentBlock, elem))
	{
	  l = lpredConstTypes[currentTypeConst];
	  if (ptlrepid_insert (&l, elem))
	    lpredConstTypes[currentTypeConst] = l;
	}
    }
}

/**
   It compiles a list of declarations (p), selecting only the constants/types? i.e the declarations of flows, imports are excluded.
   It applies cDecType() or cDecConstant() for each element of the list.
   @return the identifiers reached in the list of declarations.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecLocalConstantsTypesSet (ptlrepid vlrepid)
{
  int n, i;

  if (vlrepid != NILLREPID)
    {
      n = ptlrepid_size (vlrepid);
      initBuildPredTypeConstBlock (vlrepid);
      for (i = 1; i <= n; i++)
	{
	  setBuildPredTypeConst (vlrepid);
	  if (isTypeId (vlrepid))
	    cDecTypeDef (vlrepid, pk_false);
	  else if (isFlowId (vlrepid))
	    cDecConstantDef (vlrepid);
	  vlrepid = succid (vlrepid);
	}
      vlrepid = finalBuildPredTypeConstBlock ();
    }
  return vlrepid;		// ordered (precedence)
}

/**
   It compiles a list of declarations of interface type (p).
   It applies cDecTypeInterface() for each selected element of the list.
   @return the deblist + the model types reached in the list of declarations.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecModelType (ntrees p)
{
  ntrees list;
  ptlrepid deblist;

  list = p;
  if (STree_isUndefinedTree (list))
    return NILLREPID;

  deblist = NILLREPID;
  if (firstChild2 (&list))
    do
      switch (voperator)
	{
	case opltype:
	case opltypep:
	  deblist = cDecTypeInterface (operande, deblist);
	  break;
	}
    while (nextChild (&list));
  return deblist;
}

/**
   It compiles a list of declarations (p), selecting only the imports (declaration of flows, constants, types are excluded).
   It calls importOfModules() for such an element (use M1,...Mn).
   A warning is send to the user when the "use" are not at the end of the list.
   @return the list of the modules, in kernel tree representation.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecImports (ntrees p)
{
  ntrees list;
  ptlrepid lvfiModules;
  boolean bmodule;

  list = p;
  if (STree_isUndefinedTree (list))
    return NILLREPID;
  bmodule = pk_false;
  lvfiModules = NILLREPID;
  if (firstChild2 (&list))
    do
      {
	switch (voperator)
	  {
	  case opref:
	  case oplab:
	  case opdecsig:
	  case opdecpar:	// shared
	  case opdecvar:	// statevar
	  case opltype:
	  case opltypep:
	  case opdeccon:
	  case opdecconp:
	    if (bmodule)
	      mkklWarningAnnotTree (warModuleEndOfLocalDeclarations, operande);
	    break;
	  case oplimpo:	// importation use "M1, M2, ....Mn"
	    importOfModules (operande, &lvfiModules);
	    bmodule = pk_true;
	    break;

	  }
      }
    while (nextChild (&list));
  return lvfiModules;
}



/**
   It compiles a list of declarations (p), selecting only the flows/state variables.
   It calls cDecFlows() for such an element.
   @return the identifiers reached in the list of declarations.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cLocalSignals (ntrees p)
{
  ntrees list;
  ptlrepid deblist;

  list = p;
  deblist = NILLREPID;
  if (STree_isUndefinedTree (list))
    return (deblist);
  if (firstChild2 (&list))
    do
      switch (voperator)
	{
	case opdecsig:
	case opdecvar:		// statevar
	case opdecpar:		// shared
	  deblist = cDecFlows (voperator, deblist, pk_false);
	  break;
	case opref:
	  notYetImplemented (operande);
	  break;
	}
    while (nextChild (&list));
  return deblist;
}


/**
   It compiles the declaration of a label (labeldecl).
   @return letiq + the reached label.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cDecLabel (ntrees labeldecl, ptlrepid letiq)
{
  ptlrepid xlrepid;
  ptIdStack name;
  ptrepproto vrepproto;

  if (locateOrError (labeldecl, &name))
    {
      vrepproto = mkrepprotoLabel (labeldecl);
      if (!appendProtoId (&letiq, name, vrepproto, &xlrepid))
	mkklErrorAnnotTreeName (errdoubledecl, labeldecl, name);
    }
  return letiq;
}

/**
   It compiles a list of declarations (p) of labels.
   @return ldecletiq + the labels reached in the declaration.
*/
ptlrepid Trees2Kernel::cDecListLabels (ntrees ldecletiq, ptlrepid letiq)
{
  ntrees p;
  p = ldecletiq;
  if (firstChild (&p))
    do
      letiq = cDecLabel (operande, letiq);
    while (nextChild (&p));
  return letiq;
}

/**
   It compiles a list of declarations (p), selecting only the declaration of the labels.
   It calls cDecListLabels() for such an element.
   @return the identifiers reached in the list of declarations.
*/
_VIRTUAL_
  ptlrepid
Trees2Kernel::cDecLabels (ntrees list_decl)
{
  ntrees list;
  ptlrepid letiq;

  list = list_decl;
  if (STree_isUndefinedTree (list))
    return NILLREPID;
  letiq = NILLREPID;

  if (firstChild2 (&list))
    do
      if (voperator == oplab)
	letiq = cDecListLabels (nchild (1, operande), letiq);
    while (nextChild (&list));
  return letiq;
}

/**
   It compiles the parameters/inputs/outputs of a model of a process (p).
   It produces
   - varid: the stack associated with the name of the process
   - vfiloc: representation of the model (kl_mkModel())

   It verifies the unicity of the names (errors are sticked to the abstract tree).

   \todo { take into account a reference to an interface in the definition of a model. }
*/
_VIRTUAL_ void
Trees2Kernel::cSpec (ntrees p, ptIdStack * varid, ptrepfi * vfiloc)
{
  ptlrepid param, inList, outList;
  ntrees arbraux;
  noperators vop1, vopModel, op;
  ptIdStack idInterface;
  ptIdentProp topId;
  ptrepinterface vinterface;

  *varid = locateModelName (p);
  op = noper (p);
  vopModel = LL_getModelType (op);
  *vfiloc = kl_mkModel (p, identpile (*varid), vopModel);
  if (isChild (2, p))		/* P represente le modele de processus */
    {
      arbraux = operande;
      p = nchild (3, p);	// body
      op = noper (arbraux);
      if (OPIS_SIGNALIDENTIFIER (op))
	{			// reference to a model type (interface)
	  if (locateOrError (arbraux, &idInterface))
	    {
	      if (getModelTypeClassTop (idInterface, &topId))
		{
		  vinterface = ptIdentProp_getModelType (topId);
		  vop1 = kl_getInterfaceOperator (vinterface);
		  kl_setInterfaceModel (*vfiloc, vinterface);	// dupliquer ?
		  if (vop1 != vopModel)
		    mkklErrorAnnotTree (errConflictInterfaceModel, p);
		}
	      else
		mkklErrorAnnotTreeName (errdecl, arbraux, idInterface);
	    }
	}
      else
	{
	  parametersInputsOutputsCompiling (arbraux, &param, &inList, &outList);
	  if (ptlrepid_size (outList) > 0)
	    {
	      verifDoubleDec (&param, outList, pk_true, arbraux);
	      if (ptlrepid_size (inList) > 0)
		verifDoubleDec (&inList, outList, pk_true, arbraux);
	    }
	  if (ptlrepid_size (inList) > 0)
	    verifDoubleDec (&param, inList, pk_true, arbraux);

	  kl_setParamsInputsOutputs (*vfiloc, param, inList, outList);
	}
    }
}

/**
   It produces 
   - lExternalRefIn: the external references used as inputs
   - lExternalRefOut: the external references used as outputs)

   from the external references of each element of a list of models (lmodels):
   - An external reference x used as an input, is added to lExternalRefIn 
   if it is not an element of locList (local), outList (output) and not in inList (input).

   - An external reference x used as an output, is added to lExternalRefOut 
   if it is not an element of locList (local) and not in outList (output).
*/
_VIRTUAL_ void
Trees2Kernel::mkExternalReferences (ptlrepid lmodels,
				    ptlrepid inList, ptlrepid outList, ptlrepid locList, ptlrepid * lExternalRefIn, ptlrepid * lExternalRefOut)
{
  ptlrepid l;
  ptIdStack varid;
  ptrepfi vfi;

  if (!GUIContext)
    while (lmodels != NILLREPID)
      {
	vfi = getModelRepid (lmodels);
	l = kl_getExternalReferencesIn (vfi);
	/* x in l, added to lExternalRefIn if it is not local or output or input */
	while (l != NILLREPID)
	  {
	    varid = nameId (l);
	    if (!ptlrepid_isIn_Unsorted (inList, varid))
	      if (!ptlrepid_isIn_Unsorted (outList, varid))
		if (!ptlrepid_isIn_Unsorted (locList, varid))
		  if (!ptlrepid_isIn_Unsorted (*lExternalRefIn, varid))
		    *lExternalRefIn = addRepid (*lExternalRefIn, l);
	    l = succid (l);
	  }
	l = kl_getExternalReferencesOut (vfi);
	/* x in l, added to lExternalRefOut if it is not local or output */
	while (l != NILLREPID)
	  {
	    varid = nameId (l);
	    if (!ptlrepid_isIn_Unsorted (outList, varid))
	      if (!ptlrepid_isIn_Unsorted (locList, varid))
		if (!ptlrepid_isIn_Unsorted (*lExternalRefOut, varid))
		  {
		    *lExternalRefOut = addRepid (*lExternalRefOut, l);
		  }
	    l = succid (l);
	  }

	lmodels = succid (lmodels);
      }
}

/**
   It compiles the (body) expression of process (p_extr) of a process (p_model).
   It is the "spec" part when is_g_ext. The controls are not the same ones at the end of the compiling (?).

   It applies klparcours() followed by some verifications about the declarations of the model 
   (ptlrepid inList, ptlrepid outList, ptlrepid locList, ptlrepid llabels) and the synthetized information returned by
   klparcours() method.

   @return the kernel representation of the p_extr expression. When an error is detected, 
   a message is sticked to the abstract tree.
*/
_VIRTUAL_ void
Trees2Kernel::cExprProc (ntrees p_expr, ntrees p_model, ptlrepid inList, ptlrepid outList, ptlrepid locList,
			 ptlrepid llabels, ptklExprProc * vklExprProc, boolean is_g_ext, boolean toplevel)
{
  boolean ok;
  ptIdStack varid;
  ptlrepid inputs, outputs, vlrepid, lExternalRefOut, lExternalRefIn;
  ptIdentProp topId;
  typval IdClass;
  ntrees InterfaceTree;

  G_listOfReferencedModels = NILLREPID;
  lExternalRefIn = NILLREPID;
  lExternalRefOut = NILLREPID;

  // to be completed for virtual.
  if (STree_isUndefinedTree (p_expr))
    {
      if (is_g_ext)
	return;
      *vklExprProc = kl_mkExternal (p_model, STree_mkUndefinedExternalValue ());
      ok = pk_true;
    }
  else if (noper (p_expr) == opexte)
    {
      *vklExprProc = kl_mkExternal (p_model, p_expr);
      ok = pk_true;
    }
  else if (isNullProc (p_expr))
    {
      *vklExprProc = kl_mkEmptySystem (p_model);
      ok = pk_true;
    }
  else
    {
      InterfaceTree = nchild (2, p_model);
      /* ptlrepid param = */ kl_getParametersModel (G_process);
      inputs = NILLREPID;
      outputs = NILLREPID;
      voperator = noper (p_expr);

      operande = p_expr;
      ok = klparcours (&inputs, &outputs, vklExprProc);
      if (ok)
	{
	  if (is_g_ext)
	    {			/* en principe, pour un graphe externe, il faudrait aussi verifier que les
				   noms de vparameters, constantes, types, etiquettes et modeles utilises sont locaux
				   au graphe externe ou definis dans l'interface du modele -> les synthetiser ?
				   - en l'etat actuel, la verification n'est faite que pour les signaux */
	      while (inputs != NILLREPID)
		{
		  varid = nameId (inputs);
		  if (!isEventFlow (inputs))
		    if (!ptlrepid_isIn_Unsorted (llabels, varid)	/* ne devraient etre autorisees que les etiquettes du graphe externe */
			&& !ptlrepid_isIn_Unsorted (inList, varid) && !ptlrepid_isIn_Unsorted (outList, varid))
		      {
			mkklErrorAnnotTreeName (errentspe, InterfaceTree, varid);	// except for the event!!!
			ok = pk_false;
		      }
		  inputs = succid (inputs);
		}
	      while (outputs != NILLREPID)
		{
		  varid = nameId (outputs);
		  if ( /* !ptlrepid_isIn_Unsorted(varid, llabels) ??? && */ 
		      ! ptlrepid_isIn_Unsorted (inList, varid) && !ptlrepid_isIn_Unsorted (outList, varid))
		    {
		      mkklErrorAnnotTreeName (errentspe, InterfaceTree, varid);
		      ok = pk_false;
		    }
		  outputs = succid (outputs);
		}
	    }
	  else
	    { // ---------les inList------------
	      vlrepid = inList;
	      while ((vlrepid) != NILLREPID)
		{
		  varid = nameId (vlrepid);
		  if (!ptlrepid_remove_Unsorted (&inputs, varid))
		    mkklWarningAnnotTreeName (warentcal, InterfaceTree, varid);
		  vlrepid = succid (vlrepid);
		}

	      vlrepid = inputs;
	      while (inputs != NILLREPID)
		{
		  varid = nameId (inputs);
		  if (ptlrepid_isIn_Unsorted (llabels, varid))
		    ptlrepid_remove_Unsorted (&vlrepid, varid);
		  else if (kl_isEmptyIdStack (varid))
		    {
		      mkklErrorAnnotTreeName (errentspe, InterfaceTree, varid);
		      ok = pk_false;
		    }
		  else if (ptlrepid_isIn_Unsorted (locList, varid))
		    {
		      if (!isStateVariable (inputs) && !isEventFlow (inputs))
			mkklWarningAnnotTreeName (warsignlocnotdefined, InterfaceTree, varid);
		      ptlrepid_remove_Unsorted (&vlrepid, varid);
		    }
		  else
		    /* input not localy declared. */
		    if (!getSignalClassTop (varid, &topId, &IdClass))
		      mkklErrorAnnotTreeName (errentspe, InterfaceTree, varid);
		    else if (!isEventFlow (inputs))
		      mkklWarningAnnotTreeName (warsignlocdecl, InterfaceTree, varid);

		  inputs = succid (inputs);
		}
	      lExternalRefIn = vlrepid;
	      // ---------Les outList------------
	      vlrepid = outList;
	      while (vlrepid != NILLREPID)
		{
		  varid = nameId (vlrepid);
		  if (!ptlrepid_remove_Unsorted (&outputs, varid))	/* output not computed */
		    if (!isEventFlow (vlrepid))
		      mkklWarningAnnotTreeName (warsorcal, InterfaceTree, varid);
		  vlrepid = succid (vlrepid);
		}
	      vlrepid = outputs;
	      while (outputs != NILLREPID)
		{
		  varid = nameId (outputs);
		  // if (!ptlrepid_isIn_Unsorted(locList, varid)) 
		  if (ptlrepid_isIn_Unsorted (locList, varid))
		    ptlrepid_remove_Unsorted (&vlrepid, varid);
		  else
		    {
		      /* input not localy declared. */
		      if (!getSignalClassTop (varid, &topId, &IdClass))
			mkklErrorAnnotTreeName (errdecl, InterfaceTree, varid);
		      else
			mkklWarningAnnotTreeName (warsignlocdecl, InterfaceTree, varid);
		    }
		  outputs = succid (outputs);
		}
	      lExternalRefOut = vlrepid;
	      /** Added October 2008, an element of lExternalRefIn cannot be also an output of the model.*/
	      vlrepid = lExternalRefIn;
	      while (vlrepid != NILLREPID)
		{
		  varid = nameId (vlrepid);
		  if (ptlrepid_isIn_Unsorted (outList, varid))
		    ptlrepid_remove_Unsorted (&lExternalRefIn, varid);
		  vlrepid = succid (vlrepid);
		}
	    }
	}
    }
  if (ok)
    {
      if (is_g_ext)
	kl_setInterfaceSpec (kl_getInterfaceModel (G_process), *vklExprProc);
      else
	{
	  kl_setBodyModel (G_process, *vklExprProc);
	  mkExternalReferences (G_listOfReferencedModels, inList, outList, locList, &lExternalRefIn, &lExternalRefOut);
	  if (toplevel)
	    {
	      vlrepid = lExternalRefIn;
	      while (vlrepid != NILLREPID)
		{
		  varid = nameId (vlrepid);
		  mkklErrorAnnotTreeName (errentspe, InterfaceTree, varid);
		  vlrepid = succid (vlrepid);
		}

	      vlrepid = lExternalRefOut;
	      while (vlrepid != NILLREPID)
		{
		  varid = nameId (vlrepid);
		  mkklErrorAnnotTreeName (errentspe, InterfaceTree, varid);
		  vlrepid = succid (vlrepid);
		}
	    }
	  lExternalRefIn = declareExternalOutputsAsInputs (lExternalRefIn, lExternalRefOut);
	  kl_setExternalReferencesOut (G_process, lExternalRefOut);
	  kl_setExternalReferencesIn (G_process, lExternalRefIn);
	}
    }
}

/**
   Each external output (element of lExternalRefOut) is added to the list of external input (lExternalRefIn).
   (partial definitions management).
   @return the modified lExternalRefIn list.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::declareExternalOutputsAsInputs (ptlrepid lExternalRefIn, ptlrepid lExternalRefOut)
{
  while (lExternalRefOut != NILLREPID)
    {
      lExternalRefIn = addRepid (lExternalRefIn, lExternalRefOut);
      lExternalRefOut = succid (lExternalRefOut);
    }
  return lExternalRefIn;
}

/**
   Restriction about actions.\n
   An action model (p)  cannot have an expression of definition; It is a \b restriction.
   An error is sticked to the abstract tree when the condition are not satisfied.
*/
_VIRTUAL_ void
Trees2Kernel::verifExternalAction (ntrees p)
{
  ntrees corps, descr;
  noperators op;

  descr = nchild (3, p);	// description
  corps = nchild (2, descr);
  if (!STree_isUndefinedTree (corps))
    {
      op = noper (corps);
      if ((op != opexte) && (op != oppconf))
	mkklErrorAnnotTree (errAction, p);
    }
  operande = p;
}

/**
   It compiles a list of declarations (p), selecting only the declarations of models.
   It calls cSpec() for such an element. It applies verifExternalAction() for an action (restriction).
   @return the list of the models (appendProcess()) reached in the list of declarations.
*/
_VIRTUAL_ ptlrepid
Trees2Kernel::cLocalModels (ntrees p)
{
  ntrees list, p_decl;
  ptrepfi vfiloc;
  ptlrepid locfi;
  ptIdStack varid;
  noperators op;
  boolean b;

  list = p;
  locfi = NILLREPID;
  if (firstChild2 (&list))
    do
      {
	op = voperator;
	p_decl = operande;
	if (LL_isModelFamilyOp (op))
	  {
	    if (LL_isActionFamilyOp (op))
	      {
		verifExternalAction (p_decl);	/* Restriction: une action est sans expression de definition. */
		b = pk_false;
	      }
	    else
	      b = getModelTop (p_decl, &varid, &vfiloc);
	    if (!b)
	      if (isChild (3, p_decl))
		cSpec (p_decl, &varid, &vfiloc);
	    // plusieurs modeles "operator  ident = " du même symbole ident sont acceptés
	    if (!appendProcess (&locfi, varid, vfiloc))
	      if (op != opoper)
		mkklErrorAnnotTreeName (errdoubledecl, p_decl, varid);
	  }
      }
    while (nextChild (&list));
  return (locfi);
}

/** 
    It compiles the pragmas (p) and associates them to an interface (vinterface).
*/
_VIRTUAL_ boolean
Trees2Kernel::cPragmas (ntrees p, ptrepinterface vinterface)
{
  boolean ok;
  int i, n;
  ptrepstr name;
  ntrees enonce;
  ptklExprSig obj;
  ptlPointer l_prag;
  ntrees p_obj, aPragma;
  ptlrepid les, lparam, indexList;

  if (STree_isUndefinedTree (p))
    return pk_true;
  ok = pk_true;
  l_prag = NILLPOINTER;
  n = ntreearity (p);
  les = NILLREPID;
  lparam = NILLREPID;
  indexList = NILLREPID;
  for (i = 1; i <= n; i++)
    {
      aPragma = nchild (i, p);
      name = getTreePragmaName (aPragma);
      p_obj = nchild (2, aPragma);
      obj = cExprSig (p_obj, &les, &lparam, &indexList);
      enonce = nchild (3, aPragma);
      ptlPointer_insert_elem (&l_prag, newpragma (name, (pointer) obj, enonce));
    }
  kl_setInterfacePragmas (vinterface, l_prag);
  return (ok);
}

/** 
    It compiles the pragmas (p) and associates them to a model (vfi).
*/
_VIRTUAL_ boolean
Trees2Kernel::cPragmas (ntrees p, ptrepfi vfi)
{
  return cPragmas (p, kl_getInterfaceModel (vfi));
}


void
Trees2Kernel::verifSafetyProperty (ptrepinterface vinterface, ntrees p_graph_extern, ntrees p_model)
{
  noperators opcodeattr, processCode;
  boolean berr;

  processCode = noper (p_model);
  opcodeattr = kl_getInterfaceProperty (vinterface);
  if (opcodeattr == opnil)
    {
      // Default
      switch (processCode)
	{
	case opfunc:
	case opfuncp:
	  opcodeattr = opsafe;
	  break;
	case opnode:
	case opnodep:
	  opcodeattr = opdeterm;
	  break;
	case opprocess:
	case opprocessp:
	  opcodeattr = opunsafe;
	  break;
	case opact:
	case opactp:
	  opcodeattr = opunsafe;
	  break;
	}
      kl_setInterfaceProperty (vinterface, opcodeattr);
    }
  else
    {
      // verifications
      berr = pk_false;
      switch (processCode)
	{
	case opfunc:
	case opfuncp:
	  berr = (opcodeattr != opsafe);
	  break;
	case opnode:
	case opnodep:
	  berr = (opcodeattr == opunsafe);
	  break;
	}
      if (berr)
	mkklErrorAnnotTree (errUnexpectedAttr, p_graph_extern);
    }
}

/**
   It compiles the external graph ("spec" part)  represented by p_graph_extern of 
   a model (p_model for its abstract tree, vfi for its kernel representation).
   The lists inList, outList, locList, and llabels are respectively the inputs of the model, the outputs, the local flow declarations,
   and the label declarations of the model.

   \note { Restriction Presently, 
   - the declarations of the external graph are considered not visible in in the body of the graph.
   - the declarations of sub-models in the external graph is not allowed, as in confined process.
   .
   But, the declarations of the modele are visible in the external graph.
   It must be modified (not corresponding with the SIGNAL V4 REFERENCE MANUAL). }
*/
_VIRTUAL_ void
Trees2Kernel::cExternalGraph (ntrees p_graph_extern, ptrepinterface vinterface, ntrees p_model,
			      ptlrepid inList, ptlrepid outList, ptlrepid locList, ptlrepid llabels, boolean toplevel)
{
  ntrees p_expr, p_attr;
  ptklExprProc vklExprProc;
  noperators opcodeattr;


  opcodeattr = opnil;
  if (!STree_isUndefinedTree (p_graph_extern))
    {
      p_attr = nchild (1, p_graph_extern);	// niltree, safe, deterministic, unsafe
      if (!STree_isUndefinedTree (p_attr))
	{
	  if (ntreearity (p_attr) > 1)
	    mkklWarningAnnotTree (warOneAttribute, p_graph_extern);
	  p_attr = nchild (1, p_attr);	// Only the first attribute is kept.
	  opcodeattr = noper (p_attr);
	}

      if ((opcodeattr == opsafe) && !intrinsicLoading)
	{			// Safe implemented as Unsafe             !!!! revoir !!!!
	  mkklWarningSafeRestict ();
	  mkklWarningAnnotTree (warSafeRestict, p_graph_extern);
	}
      p_expr = nchild (2, p_graph_extern);
      if (!isNullProc (p_expr))
	cExprProc (p_expr, p_model, inList, outList, locList, llabels, &vklExprProc, pk_true, toplevel);
      else
	kl_setInterfaceSpec (vinterface, kl_mkNilExprProc (p_graph_extern));
    }

  kl_setInterfaceProperty (vinterface, opcodeattr);
}



/**
   It compiles the external graph ("spec" part)  represented by p_graph_extern of 
   a model (p_model for its abstract tree, vfi for its kernel representation).
   The lists inList, outList, locList, and llabels are respectively the inputs of the model, the outputs, the local flow declarations,
   and the label declarations of the model.

   \note { Restriction Presently, 
   - the declarations of the external graph are considered not visible in in the body of the graph.
   - the declarations of sub-models in the external graph is not allowed, as in confined process.
   .
   But, the declarations of the modele are visible in the external graph.
   It must be modified (not corresponding with the SIGNAL V4 REFERENCE MANUAL). }
*/
_VIRTUAL_ void
Trees2Kernel::cExternalGraph (ntrees p_graph_extern, ptrepfi vfi, ntrees p_model,
			      ptlrepid inList, ptlrepid outList, ptlrepid locList, ptlrepid llabels, boolean toplevel)
{
  ptrepinterface vinterface;
  vinterface = kl_getInterfaceModel (vfi);
  cExternalGraph (p_graph_extern, vinterface, p_model, inList, outList, locList, llabels, toplevel);
  verifSafetyProperty (vinterface, p_graph_extern, p_model);
}


/**
   It compiles a SIGNAL model (vfi) previously created in cSpec() method. So, the interface (input, output, parameters) is already available in vfi.
   The definition is completed by the compiling of the external graph ("spec" part) and the body part.
   The following actions are applied:
   - pushing the parameters of the model
   - compiling of the declarations of the local flows (it can be a reference to modules, See cDecImports())
   - compiling of the interfaces of the local models (verifications of double declarations)
   - pushing of the inputs, outputs, local flows and local models of the model 
   - compiling (recursively) of the body of the local models
   - pushing the labels
   - compiling the expressions of processus of the model, it synthetizes the signals/parameters/labels/..of the expression
   - compiling the external graph part of the model, it synthetizes the signals/parameters/labels/..of the expression ( cExternalGraph())
   - verifing the correpondance of the synthetized lists with the declaration list (verifDoubleDec())
   - popping the labels, local models, local signals, outputs, inputs of the model
   - popping of the parameters of the model
*/
_VIRTUAL_ void
Trees2Kernel::modelCompiling (ptrepfi vfi)
{
  ptklExprProc vklExprProc;
  ntrees p_model, p_expr, p_conf, p_graph_extern, vpragmaTree;
  ptlrepid vparameters, locList, inList, outList, llabels;
  ptrepfi save_process;
  boolean toplevel;
  noperators OpModel, op;

  save_process = G_process;
  toplevel = (save_process == NILREPFI);
  G_process = vfi;
  // ptlrepid const_typ_loc = NILLREPID;
  locList = NILLREPID;
  llabels = NILLREPID;
  p_model = kl_getBodyModel (G_process, &vklExprProc);
  OpModel = noper (p_model);
  if (LL_isModelFamilyOp (OpModel))
    {
      p_graph_extern = nchild (3, nchild (2, p_model));
      outList = kl_getInsOutsModel (G_process, &inList);
      verifOutputsModel (inList, outList, OpModel, p_model);
      vparameters = kl_getParametersModel (G_process);
      if (isChild (3, p_model))
	{			/* description */
	  vpragmaTree = nchild (1, operande);
	  if (isoChild (2, operande))
	    {			/* There is a body definition that can be "external" or null proc (| |) */
	      p_expr = operande;
	      op = noper (operande);
	      if (op == oppconf)
		{
		  p_conf = operande;
		  operande = nchild (2, p_conf);
		  if (!STree_isUndefinedTree (operande))
		    {
		      compilingDeclPush (vfi, operande, pk_true);
		      llabels = kl_getLabelsModel (vfi);
		      locList = kl_getLocalFlowsModel (vfi);
		      cExternalGraph (p_graph_extern, vfi, p_model, inList, outList, locList, llabels, toplevel);
		      cPragmas (vpragmaTree, vfi);

		      if (isoChild (1, p_conf))
			p_expr = operande;
		      else
			SET_NILTREE (p_expr);
		      cExprProc (p_expr, p_model, inList, outList, locList, llabels, &vklExprProc, pk_false, toplevel);
		      compilingDeclPop (vfi);
		    }
		  else		/* Model without locals (empty list) */
		    {
		      _PUSH_INTERFACE (vparameters, inList, outList);
		      cExternalGraph (p_graph_extern, vfi, p_model, inList, outList, locList, llabels, toplevel);
		      cPragmas (vpragmaTree, vfi);
		      cExprProc (p_expr, p_model, inList, outList, locList, llabels, &vklExprProc, pk_false, toplevel);
		      _POP_INTERFACE (vparameters, inList, outList);
		    }
		}
	      else		/* Model without locals */
		{
		  _PUSH_INTERFACE (vparameters, inList, outList);
		  cExternalGraph (p_graph_extern, vfi, p_model, inList, outList, locList, llabels, toplevel);
		  cPragmas (vpragmaTree, vfi);
		  cExprProc (p_expr, p_model, inList, outList, locList, llabels, &vklExprProc, pk_false, toplevel);
		  _POP_INTERFACE (vparameters, inList, outList);
		}
	    }
	  else /* A revoir : Ne devrait plus être accepté avec nouvelle syntaxe cas d'un processus externe. */
	    {
	      _PUSH_INTERFACE (vparameters, inList, outList);
	      cExternalGraph (p_graph_extern, vfi, p_model, inList, outList, locList, llabels, toplevel);
	      cPragmas (vpragmaTree, vfi);
	      kl_setBodyModel (vfi, kl_mkExternal (p_model, STree_mkUndefinedExternalValue ()));
	      _POP_INTERFACE (vparameters, inList, outList);
	    }
	}
      else /* A revoir : Ne devrait plus être accepté avec nouvelle syntaxe
              cas d'un processus externe sans pragmas (ne se produit que si on peut detruire la "description" d'un modele) */
	{
	  _PUSH_INTERFACE (vparameters, inList, outList);
	  cExternalGraph (p_graph_extern, vfi, p_model, inList, outList, locList, llabels, toplevel);
	  kl_setBodyModel (vfi, kl_mkExternal (p_model, STree_mkUndefinedExternalValue ()));
	  _POP_INTERFACE (vparameters, inList, outList);
	}
    }
  else
    notYetImplemented (operande);
  // The "assertion context" is propagated, for the graph building.
  if (save_process != NILREPFI) 
    kl_setAssertionsFlag (save_process, kl_getAssertionsFlag (G_process));
  G_process = save_process;
}
