/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** @return the size of a list (vlist). */
extern integer
ptlPointer_get_size (ptlPointer vlist)
{
  pointer e;
  integer nb;

  nb = 0;
  while (ptlPointer_get_elemNext (&vlist, &e))
    nb++;
  return nb;
}

/** @return "true" if a list (vlist) is an empty list, "false" otherwise. */
extern boolean
ptlPointer_satisfies_isEmpty (ptlPointer vlist)
{
  return PTLPOINTER_SATISFIES_ISEMPTY (vlist);
}

/** 
    @return "true" if a list (list) is not empty, "false" otherwise.
    @param vlist (in/out) the head of a list, it is set to the next element (if such an element exists).
    @param elem (out) the value of the head of the (in) list.
*/
extern boolean
ptlPointer_get_elemNext (ptlPointer * vlist, pointer * elem)
{
  if (PTLPOINTER_SATISFIES_ISEMPTY (*vlist))
    return pk_false;
  (*elem) = (*vlist)->elem;
  (*vlist) = (*vlist)->next;
  return pk_true;
}



/**
  It inserts an element (elem) in a list of pointers (deblist). The unicity is guaranteed.
  @return "false" if the element is yet in the list, "true" otherwise.
*/
extern boolean
ptlPointer_insert_elem (ptlPointer * deblist, pointer elem)
{
  ptlPointer curlist, preclist;
  boolean trouve;

  if (PTLPOINTER_SATISFIES_ISEMPTY (*deblist))
    {
      (*deblist) = ptlPointer_RIEN_create (elem, NILLPOINTER);
      return pk_true;
    }
  trouve = pk_false;
  curlist = *deblist;
  while (!trouve && curlist != NILLPOINTER)
    {
      trouve = (elem == curlist->elem);
      preclist = curlist;
      curlist = curlist->next;
    }
  if (trouve)
    return (pk_false);
  preclist->next = ptlPointer_RIEN_create (elem, NILLPOINTER);
  return (pk_true);
}

/**
   It inserts a pointer (elem) at the END of a list (deblist).
   @return the new list.
*/
extern ptlPointer
ptlPointer_add (ptlPointer deblist, pointer elem)
{
  ptlPointer curlist, preclist;

  if (PTLPOINTER_SATISFIES_ISEMPTY (deblist))
    return ptlPointer_RIEN_create (elem, NILLPOINTER);
  curlist = deblist;
  while (curlist != NILLPOINTER)
    {
      preclist = curlist;
      curlist = curlist->next;
    }
  preclist->next = ptlPointer_RIEN_create (elem, NILLPOINTER);
  return deblist;
}


/** It removes a pointer (elem) from a list (li).
    @return "true" if the pointer has been removed from the list, "false" otherwise.
*/
extern boolean
ptlPointer_remove_elem (ptlPointer * li, pointer elem)
{
  ptlPointer linst, pred;

  if (PTLPOINTER_SATISFIES_ISEMPTY (*li))
    return pk_false;
  if ((*li)->elem == elem)
    {
      (*li) = (*li)->next;
      return pk_true;
    }
  pred = *li;
  linst = (*li)->next;
  while (!PTLPOINTER_SATISFIES_ISEMPTY (linst))
    {
      if (linst->elem == elem)
	{
	  pred->next = linst->next;
	  return pk_true;
	}
      pred = linst;
      linst = linst->next;
    }
  return pk_false;
}

/**
   @return "true" if an element (elem) is in a list (li), "false" otherwise.
 */
extern boolean
ptlPointer_IsIn (ptlPointer li, pointer elem)
{
  while (pk_true)
    {
      if (PTLPOINTER_SATISFIES_ISEMPTY (li))
	return pk_false;
      if (li->elem == elem)
	return pk_true;
      li = li->next;
    }
}

/** It merges two lists of marks (lprinc,listToInsert). It modifies lprinc.
    @return lprinc, the modified list
*/
extern ptlPointer
ptlPointer_fusion (ptlPointer lprinc, ptlPointer listToInsert)
{
  pointer elem;
  ptlPointer listres;

  listres = lprinc;
  while (ptlPointer_get_elemNext (&listToInsert, &elem))
    ptlPointer_insert_elem (&listres, elem);
  return listres;
}


/**
   Garbage collector for a list (l) of pointers.
*/
extern void
ptlPointer_free (ptlPointer l)
{
  ptlPointer last, laux;

  if (!PTLPOINTER_SATISFIES_ISEMPTY (l))
    {
      if (PTLPOINTER_SATISFIES_ISEMPTY (gclPointer))
	gclPointer = l;
      else
	{
	  laux = gclPointer;
	  while (!PTLPOINTER_SATISFIES_ISEMPTY (laux))
	    {
	      last = laux;
	      laux = laux->next;
	    }
	  last->next = l;
	}
    }
}


extern ptlPointer
ptlPointer_conc (ptlPointer l1, ptlPointer l2)
{
  ptlPointer l;

  if (PTLPOINTER_SATISFIES_ISEMPTY (l1))
    return l2;
  if (PTLPOINTER_SATISFIES_ISEMPTY (l2))
    return l1;
  l = l1;
  while (l1->next != NILLPOINTER)
    l1 = l1->next;
  l1->next = l2;
  return l;
}
/**
   It inserts a pointer (elem) at the END of a list (deblist).
   @return the new list.
*/
extern ptlPointer
ptlPointer_add_front (ptlPointer deblist, pointer elem)
{
  return ptlPointer_RIEN_create (elem, deblist);
}

extern void ptlPointer_replace_value(ptlPointer elemlist, pointer elem)
{
  elemlist->elem = elem; 
}
