/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/** It adds an elem (velem) at the head of a list (aStack).
    @return the new list.
*/
extern ptStack
stackPush (ptStack aStack, ptElemStack velem)
{
  return (ptStack) ptlPointer_RIEN_create ((pointer) velem, (ptlPointer) aStack);
}

/** @return the list (pt) without its first element. */
extern ptStack
stackPop (ptStack aStack)
{
  ptStack n;
  if (stackIsEmpty (aStack))
    return aStack;
  n = (ptStack) ptlPointer_get_next ((ptlPointer) aStack);
  ptlPointer x = (ptlPointer) aStack;
  x->next = gclPointer;
  gclPointer = x;
  return n;

}

/** 
    @return "true" if a stack (aStack) is not empty, "false" otherwise.
    @param aStack (in) a stack.
    @param  pStack (out) aStack without it head element.
*/
extern boolean
stackPrevious (ptStack aStack, ptStack * pStack)
{
  if (stackIsEmpty (aStack))
    return pk_false;
  *pStack = (ptStack) ptlPointer_get_next ((ptlPointer) aStack);
  return pk_true;
}


/** @return "true" if a stack (aStack) is empty, "false" otherwise. */
extern boolean
stackIsEmpty (ptStack aStack)
{
  return ptlPointer_satisfies_isEmpty ((ptlPointer) aStack);
}

/**
   @return "true" if a stack (aStack) is not empty, "false" otherwise.
   The top of the stack is returned (velem).
 */
extern boolean
stackTop (ptStack aStack, ptElemStack * velem)
{
  ptlPointer l;
  pointer x;

  l = (ptlPointer) aStack;
  if (!ptlPointer_get_elemNext (&l, &x)) return pk_false;
  (*velem) = (pointer) x;
  return pk_true;
}

/**
   @return "true" is an element (el) is yet pushed in a stack (aStack).
*/
extern boolean
stackIsIn (ptStack aStack, ptElemStack el)
{
  return ptlPointer_IsIn (aStack, el);
}


/**
   @return the depth of the stack.
*/
extern integer
stackDepth (ptStack aStack)
{
  return ptlPointer_get_size (aStack);
}

extern void StackReplaceValue(ptStack S, ptElemStack elem)
{
  ptlPointer_replace_value(S, elem);
}
