/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define isIdentOperator(op) (op == opnom) || (op == opident)
#define getSMEKeyword(k) strTableGet(SMEbasekeyword, k+1)
#define getikwFlag(v) getSMEKeyword(FirstFlag+v)
#define setVisibility(ikw) visibilityKW=ikw;

#define XmlVersion "1.0"
#define SMEVersion "2.0"
#define xmlns_xmi  "http://www.omg.org/XMI"
#define xmlns_xsi  "http://www.w3.org/2001/XMLSchema-instance"
#define xmlns_sme  "http://www.irisa.fr/espresso/sme.ecore"
#define UTF_str    "UTF-8"

#define tkwundef -1

#define X1_indent 5
#define smallindent 3
#define actindent 2

#define comBegin "<!--"
#define comEnd "-->"
#define comincom " "
#define comincomend " "
#define errSymbol "~"
#define cstSymbolHolo "..."

_PROTECTED_
/**
   Global initialisation of the printer.
   Some of the initialisations are done in reset().
*/
    void
 SME_Printer::GeneralInit()
{
  undefSMEId = string2repstr("UndefinedSMEId_");
  currentName = nilstr;
  SET_NILTREE(currentModule);
  assertContext = pk_false;	// pk_true for the operand of assert(| |).
  reset();
}

/**
   @return "true" if the i-th child (s) of a tree (vtree) exists,
   "false" otherwise.
*/
boolean SME_Printer::getchild(ntrees vtree, int i, ntrees * s)
{
  noperators op;
  if (IS_NILTREE(vtree))
    return pk_false;
  op = noper(vtree);
  if (LL_is_atomOp(op))
    return pk_false;
  (*s) = nchild(i, vtree);
  if (noper(*s) == opevnt)
    (*s) = nchild(1, *s);
  if (IS_NILTREE(*s))
    return pk_false;
  return (boolean) (noper(*s) != opnil);
}

/**
   Specialisation of SignalPrinter::embedproc()
   It recalls markunparse() on the current tree.
*/
_VIRTUAL_ void SME_Printer::embedproc(boolean embed)
{
  markunparse(vholoph);
}

/**
   Specialisation of SignalPrinter::embedincproc().
   It recalls markunparse() on the current tree.
*/
_VIRTUAL_ void SME_Printer::embedincproc(boolean embed)
{
  markunparse(vholoph);
}

/**
   @return "true" if a tree (t) is annotated by the SME_DeclATTRIBUTENUMBER-th attribute.
   It gives the link between a reference (t) to an identifier and its declaration (vdecl).
*/
boolean SME_Printer::getDeclarationAttribute(ntrees t, ntrees * vdecl)
{
  noperators op;
  if (!nattrexist(t, SME_DeclATTRIBUTENUMBER))
    return pk_false;
  (*vdecl) = (ntrees) nattrptr(t, SME_DeclATTRIBUTENUMBER);
  op = noper(*vdecl);
  // Reference to an object of a module ?
  if ((op == opSMEImportType) || (op == opSMEImportCons))
    return pk_true;

  while (nattrexist(*vdecl, SME_DeclATTRIBUTENUMBER))
    (*vdecl) = (ntrees) nattrptr(*vdecl, SME_DeclATTRIBUTENUMBER);
  return pk_true;
}

void SME_Printer::printConstant(noperators op)
{
  ntrees t;
  boolean b;

  t = treeunparseTree;
  printBeginBlock(IKWLiterals);
  printSMEId(t);
  printConnectionFrom(t);
  b = (op == opcstcar) || (op == opcstcha) || (op == opvenum);
  printKwQuotedSignalText(IKWValue, t, b);
  closeBlockLine();
}

/**
   It prints the ikw-th key word, following by '='.
*/
void SME_Printer::printSMEKwEq(int ikw)
{
  printSMEKw(ikw);
  sncharpri('=');
}

/**
   It prints the ikw1-th keyword = "the ikw2-th keyword".
   Example:
   \verbatim
   printSMEKwEqQuotedKey(IKWmodelType, IKWprocess);
   prints 
   modelType="process"
   \endverbatim
*/
void SME_Printer::printSMEKwEqQuotedKey(int ikw1, int ikw2)
{
  printSMEKwEq(ikw1);
  printQuotedKey(ikw2);
}

void SME_Printer::printSMEKwKwEq(int ikwpredfix, int ikw)
{
  sncharpri(' ');
  printSMEKwKw(ikwpredfix, ikw);
  sncharpri('=');
}

/**
   It initialises the printer for printing a XML string (some characters needs codification).
   \see setCharEncodingMode().
*/
void SME_Printer::beginString()
{
  sncharpri('"');
  setCharEncodingMode(pk_false);	// XML characters.
}

/**
   It finalyzes the printer for printing a XML string, the XML mode is previously opened in beginString().
   \see setCharEncodingMode().
*/
void SME_Printer::endString()
{
  sncharpri('"');
  setCharEncodingMode(pk_true);	// std characters.
}

/**
   It prints an attribute tree (p) (ie a tree that attribute a Signal tree) used to code the 
   connections (src_DataflowConnections, dst_DataflowConnections,...) assigned with an object.
*/
void SME_Printer::printSMEattr(ntrees p)
{
  noperators op;
  int i, ikw;
  ntrees s;

  op = noper(p);
  switch ((int)(op))
    {
    case opsrc_dfc:
      ikw = IKWsrc_DataflowConnections;
      break;
    case opdst_dfc:
      ikw = IKWdst_DataflowConnections;
      break;
    case opsrc_condc:
      ikw = IKWsrc_conditions;
      break;
    case opdst_condc:
      ikw = IKWdst_condition;
      break;

    case opsrc_clockc:
      ikw = IKWsrc_clockRelation;
      break;
    case opdst_clockc:
      ikw = IKWdst_clockRelation;
      break;

    case opsrc_cstaff:
      ikw = IKWsrc_cstaff;
      break;
    case opdst_cstaff:
      ikw = IKWdst_cstaff;
      break;

    case opsrc_case:
      ikw = IKWsrc_case;
      break;

    case opdst_case:
      ikw = IKWdst_case;
      break;

    case opsrc_dependences:
      ikw = IKWsrc_Dependences;
      break;

    case opdst_dependences:
      ikw = IKWdst_Dependences;
      break;
    }

  printSMEKwEq(ikw);

  beginString();
  for (i = 1; i <= ntreearity(p); i++)
    {
      if (i > 1)
	sncharpri(' ');
      s = nchild(i, p);
      printSMEIdValue(s);
    }
  endString();
}

/**
   It prints the ikw-th keyword of the SMEKEYWORDS.kw table (defined in the directory of the binaries of Polychrony).
*/
void SME_Printer::printSMEKw(int ikw)
{
  if (ikw >= 0)
    sigpristr(getSMEKeyword(ikw));
}

/**
   It prints the ikw-th keywords (see printSMEKw()) prefixed by the ikwpredfix-th, if it is not undefined.
   Example:
   \verbatim
   printSMEKwKw(IKWsme, IKWModelDeclaration); produces the code
   sme:ModelDeclaration
   \endverbatim
*/
void SME_Printer::printSMEKwKw(int ikwpredfix, int ikw)
{
  if (ikwpredfix != IKWundef)
    {
      printSMEKw(ikwpredfix);
      sncharpri(':');
    }
  printSMEKw(ikw);
}

void SME_Printer::printTree(ntrees vt)
{
  inteprint_initialize();	// no space
  unparse_subtree(vt, 99999);
}

void SME_Printer::printQuotedTree(ntrees vt)
{
  beginString();
  printTree(vt);
  endString();
}

void SME_Printer::printKwId(ntrees vt, int ikwpredfix, int ikw)
{
  printSMEKwKwEq(ikwpredfix, ikw);
  printQuotedTree(vt);
}

/**
   A set of strings is generated as only one string.
*/
void SME_Printer::printStmtPragma(int ikw, ntrees enonce)
{
  int j;
  ntrees s;
  if (!STree_isUndefinedTree(enonce))
    {
      snspace();
      printSMEKwEq(ikw);
      beginString();
      for (j = 1; j <= ntreearity(enonce); j++)
	{
	  s = nchild(j, enonce);	// opcstcha
	  sigpristr(ExtractString(nstringof(s)));
	  sncharpri(' ');
	}
      endString();
    }
}

void SME_Printer::printPragmasInHeader(ntrees p)
{
  int vindex;
  int i, n;
  ptrepstr name;
  ntrees aPragma;
  boolean b;

  if (!STree_isUndefinedTree(p))
    {
      n = ntreearity(p);
      for (i = 1; i <= n; i++)
	{
	  aPragma = nchild(i, p);
	  if (!nattrexist(aPragma, SME_AlreadyPrintedATTRIBUTENUMBER))
	    {
	      name = STree_getIdentStr(nchild(1, aPragma));
	      if (isPredefinedPragmaName(name, &vindex))
		{
		  b = pk_true;
		  switch (vindex)
		    {
		    case PRAGMA_COMMENT:
		      // p_obj = nchild(2, aPragma); // empty I hope !
		      printStmtPragma(IKWcomment, nchild(3, aPragma));
		      break;
		    case PRAGMA_MAIN:
		      printFlagtrue(mainFlag);
		      break;
		    case PRAGMA_SIGALI:
		      printFlagtrue(sigaliFlag);
		      break;
		    case PRAGMA_PROCESSOR_TYPE:
		      printStmtPragma(IKWprocessorTypePragma, nchild(3, aPragma));
		      break;

		      /*modified June 2010 : RunOn value is not now an int.
		         case PRAGMA_RUN_ON:
		         printStmtPragma(IKWrunOnPragma, nchild(3, aPragma));
		         break;
		       */

		    case PRAGMA_BLACK_BOX:
		      printKwQuotedStr(IKWcompilationDirectivePragma, getikwFlag(BlackBoxFlag));
		      break;
		    case PRAGMA_GREY_BOX:
		      printKwQuotedStr(IKWcompilationDirectivePragma, getikwFlag(GreyBoxFlag));
		      break;
		    case PRAGMA_DELAYCLUSTER:
		      printKwQuotedStr(IKWcompilationDirectivePragma, getikwFlag(DelayClusterFlag));
		      break;
		    case PRAGMA_UNEXPANDED:
		      printKwQuotedStr(IKWcompilationDirectivePragma, getikwFlag(unexpandedFlag));
		      break;
		      /* case PRAGMA_SEPARATE:
		      printKwQuotedStr(IKWcompilationDirectivePragma, getikwFlag(separateFlag));
		      break; */

		    case PRAGMA_CODEGEN_C:
		      printStmtPragma(IKWcCodePragma, nchild(3, aPragma));
		      break;

		    case PRAGMA_CODEGEN_CPP:
		      printStmtPragma(IKWcppCodePragma, nchild(3, aPragma));
		      break;

		    case PRAGMA_CODEGEN_JAVA:
		      printStmtPragma(IKWjavaCodePragma, nchild(3, aPragma));	// printKwQuotedStr(IKWcompilationDirectivePragma, getikwFlag(codeJavaPragmaFlag));
		      break;

		    default:
		      b = pk_false;
		    }
		  // to avoid the printing in the method decpragmas()
		  if (b)
		    nmakeattrntrees(aPragma, SME_AlreadyPrintedATTRIBUTENUMBER, STree_mkNullaryTree(optrue));
		}
	    }
	}
    }
}

void SME_Printer::printVisibilityAttribute()
{
  sncharpri(' ');
  printSMEKwEqQuotedKey(IKWvisibility, visibilityKW);
}

/**
   It prints the model type associated with an operator (op)
   -   op = opprocess  modelType="process"
   -   op = opnode  modelType="node"
*/
void SME_Printer::printModelType(noperators op)
{
  snspace();
  switch ((int)(op))
    {
    case opprocessp:
      printSMEKwEqQuotedKey(IKWmodelType, IKWprocess);
      printVisibilityAttribute();
      break;

    case opprocess:
      printSMEKwEqQuotedKey(IKWmodelType, IKWprocess);
      break;

    case opfunc:
      printSMEKwEqQuotedKey(IKWmodelType, IKWfunction);
      break;

    case opfuncp:
      printSMEKwEqQuotedKey(IKWmodelType, IKWfunction);
      printVisibilityAttribute();
      break;
    case opnode:
      printSMEKwEqQuotedKey(IKWmodelType, IKWnode);
      break;
    case opnodep:
      printSMEKwEqQuotedKey(IKWmodelType, IKWnode);
      printVisibilityAttribute();
      break;

    case opact:
      printSMEKwEqQuotedKey(IKWmodelType, IKWaction);
      break;

    case opactp:
      printSMEKwEqQuotedKey(IKWmodelType, IKWaction);
      printVisibilityAttribute();
      break;

    case opoper:		// ??
      break;
    }
}

void SME_Printer::printExternalGraphAttributes(ntrees p)
{
  ntrees lattr;

  return;
  if (STree_getAttributeSpec(p, &lattr))
    {
      snspace();
      treeunparseTree = lattr;
      voperator = oplattr;
      operande = lattr;
      decl0(0, -2);
    }
}

/**
   It prints a model or a subprocess (m) with id as name.
   op gives the kind of object to print (model, subprocess).
   It prints the connections associated with m.
 */
int SME_Printer::beginModel(ntrees m, noperators op, ntrees id)
{
  int i, XsiType;
  ntrees vdecl;
  boolean isamodel;

  isamodel = (op != oppconf) && (op != opiter) && (op != oppetiq) && (op != opspec) && (op != opmoti);

  if (FirstHeader)
    {
      if (op == opmodu)
	i = IKWModule;
      else
	i = IKWModelDeclaration;	/* but modelType must be set. */

      sncharpri('<');
      printSMEKwKw(IKWsme, i);
      printSMEKwKwEq(IKWxmi, IKWversion);
      printQuotedStr(string2repstr(SMEVersion));
      printSMEKwKwEq(IKWxmlns, IKWxmi);
      printQuotedStr(string2repstr(xmlns_xmi));

      printSMEKwKwEq(IKWxmlns, IKWxsi);
      printQuotedStr(string2repstr(xmlns_xsi));

      printSMEKwKwEq(IKWxmlns, IKWsme);
      printQuotedStr(string2repstr(xmlns_sme));
      FirstHeader = pk_false;
    }
  else
    {
      XsiType = IKWundef;
      if (op == oppconf)
	i = IKWConfinedProcess;
      else if (op == oppetiq)
	i = IKWConfinedProcess;
      else if (op == opiter)
	{
	  i = IKWConfinedProcess;
	  XsiType = IKWIterate;
	}
      else if (op == opmoti)
	{
	  i = IKWConfinedProcess;
	  XsiType = IKWBasicIterate;
	}
      else if (op == opspec)
	i = IKWspec;
      else
	i = IKWModelDeclarations;	/* but modelType must be set. */

      printBeginBlock(i, XsiType, pk_false);
    }

  snspace();
  printSMEId(m);
  if (id != NILIDENT)
    printKwId(id, IKWundef, IKWname);
  if (isamodel)
    {
      printModelType(op);
      if (op != opmodu)
	printExternalGraphAttributes(m);
      printPragmasInHeader(STree_getPragmas(m));
    }
  printConnectionFrom(m);
  if (op == oppetiq)
    if (getDeclarationAttribute(id, &vdecl))
      printConnectionFrom(vdecl);
    else;
  else if (op == opmoti)
    printKwQuotedSignalText(IKWupperBound, nchild(2, nchild(1, m)), pk_false);
  printCloseHeader();
  if (isamodel)
    decpragmas(STree_getPragmas(m));
  sntab(4);
  printConnections(m);
  printReferences(m);
  printImports(m);
  return i;
}

void SME_Printer::endModel(int ikw1, int ikw2, ntrees id)
{
  snbacktab();
  printSMEBracketedKw(ikw1, ikw2, pk_true);
  if (ISNOT_NILTREE(id))
    {
      this_intePrinter->intedebcomprint();
      identornameprint(id);
      this_intePrinter->intefincomprint();
    }
  snline();
}

/**
   It prints a list of declarations of an interface (p), parametrized (pio). 
   For each element of the list, it calls printdeclaration().
   @param p (in) list of declarations. 
   @param pio (in) index of the associated keyword defined in the global table SMEKwTable (param, input, output,...)
*/
void SME_Printer::printInterfaceDeclarations(ntrees p, int pio)
{
  int i;
  for (i = 1; i <= ntreearity(p); i++)
    printdeclaration(nchild(i, p), pio, opinte);
}

void SME_Printer::printBaseType(ntrees vtype)
{
  while (noper(vtype) == optvect)
    vtype = nchild(2, vtype);
  printTree(vtype);
  // unparse_subtree(vtype, 99999);
}

void SME_Printer::printDimType(ntrees vtype)
{
  ntrees d;
  int i;

  snline();
  while (noper(vtype) == optvect)
    {
      d = nchild(1, vtype);
      // <arrayDimensions>L</arrayDimensions> for each dim
      for (i = 1; i <= ntreearity(d); i++)
	printSignalText(nchild(i, d), IKWArrayDimensions);
      vtype = nchild(2, vtype);
    }
}

boolean SME_Printer::printType(ntrees vtype)
{
  noperators op;
  boolean barray;

  op = noper(vtype);
  barray = pk_false;
  if (LL_is_atomOp(op))		// Predefined Type + reference to a named type.
    printKwId(vtype, IKWundef, IKWtype);
  else if (op == opstru) ;
  else if (op == opbund) ;
  else if (op == optenum) ;	// enum
  else
    {				// array, ident, 
      printSMEKwEq(IKWtype);
      beginString();
      if (op == optvect)
	{
	  barray = pk_true;
	  printBaseType(vtype);
	}
      else if (isIdentOperator(op))
	printSignalText(vtype);
      endString();
    }
  return barray;
}

void SME_Printer::printPragmaAttribute(ntrees aDecl)
{
  if (nattrexist(aDecl, SME_PragmaATTRIBUTENUMBER))
    printPragmasInHeader(nattrntrees(aDecl, SME_PragmaATTRIBUTENUMBER));
}

void SME_Printer::printStatusKind(noperators opcontext)
{
  snspace();
  if (opcontext == opdecvar)
    printSMEKwEqQuotedKey(IKWStatusKind, IKWStatevar);
  else if (opcontext == opdecpar)
    printSMEKwEqQuotedKey(IKWStatusKind, IKWShared);
}

/**
   @param p (in) a declaration (type, list of signals)
   @param pio (in) index of the associated keyword defined in the global table SMEKwTable (param, input, output,...)
   @param opcontext (in) the operator of the context ( declaration of statevars,...,.... (TO BE COMPLETED))
*/
void SME_Printer::printdeclaration(ntrees p, int pio, int opcontext)
{
  noperators op;
  ntrees vtype, id, ldecl, aDecl, vinit;
  int i;
  int IK;
  boolean b;

  if (noper(p) == opltype)
    {
      unparse_subtree(p, 99999);	// a type as parameter.
    }
  else
    {
      op = STreeDecl_getType(p, &vtype);	// op = oper(vtype), opinteger,..., opimplicit.

      ldecl = nchild(2, p);
      for (i = 1; i <= ntreearity(ldecl); i++)
	{
	  aDecl = nchild(i, ldecl);
	  op = noper(aDecl);
	  SET_NILTREE(id);
	  SET_NILTREE(vinit);
	  if ((op == opcons) || (op == opinit))
	    {
	      id = nchild(1, aDecl);
	      vinit = nchild(2, aDecl);
	    }
	  else if (isIdentOperator(op))
	    id = aDecl;

	  if (ISNOT_NILTREE(id))
	    {
	      printBeginBlock(pio);
	      printSMEId(id);
	      printConnectionFrom(id);

	      printKwId(id, IKWundef, IKWname);
	      printStatusKind(opcontext);
	      if (pio == IKWconstants)
		printVisibilityAttribute();
	      sncharpri(' ');
	      b = printType(vtype);
	      printPragmaAttribute(aDecl);
	      if (!b && (IS_NILTREE(vinit)))
		closeBlockLine();
	      else
		{
		  sncharpri('>');
		  if (b)
		    printDimType(vtype);
		  if (ISNOT_NILTREE(vinit))
		    {
		      if (op == opcons)
			IK = IKWValue;
		      else
			IK = IKWinitValue;
		      snline();
		      printSignalText(vinit, IK);
		    }
		  printTailDef(pio);
		}
	    }
	}
    }
}

void SME_Printer::closeBlockLine()
{
  sncharpri('/');
  sncharpri('>');
  snline();
}

void SME_Printer::printSignalText(ntrees t)
{
  pK_SignalPrinter->unparse1(t);
}

void SME_Printer::printSignalText(ntrees t, int ikw)
{
  printSMEBracketedKw(ikw, pk_false);
  printSignalText(t);
  printSMEBracketedKw(ikw, pk_true);
}

void SME_Printer::printFlag(ntrees vflag)
{
  int i;
  ntrees valueflag;

  i = (int)nintof(nchild(1, vflag));
  if (i >= 0)
    sigpristr(getikwFlag(i));
  else
    sigpristr(undefSMEId);
  sncharpri('=');
  valueflag = nchild(2, vflag);
  if (noper(valueflag) == opSMEvalue)
    {
      i = (int)nintof(valueflag);
      if (i >= 0)
	printQuotedStr(getikwFlag(i));
      else
	sigpristr(undefSMEId);
    }
  else
    // Conversion, or a counter "from, after, count"
  if ((i == castTypeFlag) || (i == counterKindFlag) /*|| (i == arrayRecoveryFlag) */ )
    printQuotedTree(valueflag);
  else
    sigpristr(undefSMEId);
}

/**
   the access to the previous iteration [?] is translated by a flag.
*/
boolean SME_Printer::printLastIterationValueFlag(ntrees s)
{
  noperators op;

  while (pk_true)
    {
      op = noper(s);
      if (op == opsdec)
	{
	  printFlagtrue(useLastIterationValueFlag);
	  return pk_true;
	}
      if (op != opelem)
	return pk_false;
      s = nchild(1, s);
    }
}

ntrees SME_Printer::getIdentOf(ntrees t, boolean bsdec)
{
  noperators op;

  while (pk_true)
    {
      op = noper(t);
      if ((op == opcham) || (op == opelem) /* || (op  == oprecu) */  || (!bsdec && (op == opsdec)))
	t = nchild(1, t);
      else
	return t;
    }
}

void SME_Printer::printFieldsValues(int IKW, ntrees vtree, boolean bsdec)
{
  noperators op, i;
  ntrees t, f;

  op = noper(vtree);
  if (op == opsdec)
    {
      vtree = nchild(1, vtree);
      op = noper(vtree);
    }
  if ((op == opelem) || (op == opcham))
    {
      vtree = STree_copy(vtree, pk_false);
      t = vtree;
      if (bsdec)
	{
	  t = getIdentOf(vtree, pk_true);
	  i = nposit(t);
	  f = nfather(t);
	  nchangchild(f, nchild(1, t), i);
	  t = f;
	}
      t = getIdentOf(t, pk_false);
      if (t != vtree)
	{
	  i = nposit(t);
	  nchangchild(nfather(t), string2ident(" "), i);
	  printKwQuotedSignalText(IKW, vtree, pk_true);
	}
    }
}

void SME_Printer::printFields(int IKW, ntrees vtree)
{
  ntrees s;
  noperators op;

  // recup, Fields and array accesses are particular cases!!! 
  // they are not general (x[] \\ 5) [3] : to be clarified
  // moreover, X[I] [?] and  X [?][I] ???
  if (nattrexist(vtree, SME_BackAnnotationATTRIBUTENUMBER))
    {
      s = (ntrees) nattrptr(vtree, SME_BackAnnotationATTRIBUTENUMBER);
      op = noper(s);
      // if (op == oprecu) s = nchild(1,s); // E \\ E2
      printFieldsValues(IKW, s, printLastIterationValueFlag(s));
    }
}

void SME_Printer::printFlags(ntrees vtree)
{
  ntrees vflags;
  int i;

  if (nattrexist(vtree, SME_FlagATTRIBUTENUMBER))
    {
      sncharpri(' ');
      vflags = nattrntrees(vtree, SME_FlagATTRIBUTENUMBER);
      // attr is a binary tree with operator = opSMEtag, sons =(opinteger, int)
      for (i = 1; i <= ntreearity(vflags); i++)
	printFlag(nchild(i, vflags));
    }

}

void SME_Printer::printCaseSelection(ntrees vconnection)
{
  ntrees s;
  noperators op;
  boolean ok;
  int ikw;

  if (!nattrexist(vconnection, SME_BackAnnotationATTRIBUTENUMBER))
    {
      // else part of the case.
      closeBlockLine();
      printSMEKwEqQuotedKey(IKWcaseKind, IKWElseCase);	// printSMEKwEq(IKWcaseKind); printQuotedKey(IKWElseCase);
    }
  else
    {
      snspace();
      s = (ntrees) nattrptr(vconnection, SME_BackAnnotationATTRIBUTENUMBER);
      op = noper(s);
      if (op == oplesign)
	{
	  printSMEKwEqQuotedKey(IKWcaseKind, IKWEnumeration);	//  printSMEKwEq(IKWcaseKind); printQuotedKey(IKWEnumeration);
	  printCloseHeader();
	  printList(s, IKWenumeration);
	  printSMEBracketedKw(IKWconnections, pk_true);
	}
      else
	{
	  ok = pk_true;
	  if (op == opintvff)
	    ikw = IKWClosedClosedInterval;
	  else if (op == opintvfo)
	    ikw = IKWClosedOpenedInterval;
	  else if (op == opintvof)
	    ikw = IKWOpenedClosedInterval;
	  else if (op == opintvoo)
	    ikw = IKWOpenedOpenedInterval;
	  else
	    ok = pk_false;
	  if (ok)
	    {
	      printSMEKwEqQuotedKey(IKWcaseKind, ikw);	// printSMEKwEq(IKWcaseKind); printQuotedKey(ikw);
	      printKwQuotedSignalText(IKWlowerBound, nchild(1, s), pk_false);
	      printKwQuotedSignalText(IKWupperBound, nchild(2, s), pk_false);
	    }
	  else
	    this_intePrinter->printerror();
	  closeBlockLine();
	}
    }
}

/**
   It genrates the text of a connection (vconnection).
   A connection is a list (name, from, to) and it can be attributed by some flags. 
   The flags are printed by the method printFlags().
*/
void SME_Printer::printConnection(ntrees vconnection)
{
  int op, IKW;
  ntrees s;

  op = noper(vconnection);
  if (op == opDataflowConnection)
    IKW = IKWDataflowConnection;
  else if (op == opConditionConnection)
    IKW = IKWCondition;
  else if (op == opClockRelationConnection)
    IKW = IKWClockRelation;
  else if (op == opConstantAffectation)
    IKW = IKWConstantAffectation;
  else if (op == opCaseConnection)
    IKW = IKWCaseConnection;
  else if (op == opDependenceConnection)
    IKW = IKWDependence;

  printBeginBlock(IKWconnections, IKW, pk_true);

  printKwId(nchild(1, vconnection), IKWxmi, IKWid);
  s = nchild(2, vconnection);
  printKwId(s, IKWundef, IKWsrc);
  printFields(IKWSrcField, s);
  s = nchild(3, vconnection);
  printKwId(s, IKWundef, IKWdst);
  printFields(IKWDstField, s);
  printFlags(vconnection);
  if (op != opCaseConnection)
    closeBlockLine();
  else
    printCaseSelection(vconnection);
}

void SME_Printer::printConnections(ntrees m)
{
  int i;
  ntrees lconnections;

  if (nattrexist(m, SME_ConnectionATTRIBUTENUMBER))
    {
      lconnections = nattrntrees(m, SME_ConnectionATTRIBUTENUMBER);
      for (i = 1; i <= ntreearity(lconnections); i++)
	printConnection(nchild(i, lconnections));
    }
}

void SME_Printer::printReference(ntrees t)
{
  noperators op;
  ntrees x;

  op = noper(t);
  if (op == opSMESignalRef)
    {
      printHeader(IKWSignalRef, t, tkwundef, tkwundef, pk_true);
      x = (ntrees) nattrptr(t, SME_ReferenceDeclATTRIBUTENUMBER);
      printKwSMEId(x, IKWundef, IKWsignal);
      closeBlockLine();
    }
  else if (op == opSMEConstantRef)
    {
      printHeader(IKWConstantRef, t, tkwundef, tkwundef, pk_true);
      x = (ntrees) nattrptr(t, SME_ReferenceDeclATTRIBUTENUMBER);
      // printKwId(x, IKWundef, IKWCstName);
      printKwSMEId(x, IKWundef, IKWconstant);
      closeBlockLine();
    }
  else if (op == opSMEModelRef) /*{printHeaderDef(IKWModelRef, t, tkwundef, tkwundef, pk_true); } */ ;
  else
    this_intePrinter->printerror();
}

void SME_Printer::printReferences(ntrees m)
{
  int i;
  ntrees l;

  if (nattrexist(m, SME_ReferencesATTRIBUTENUMBER))
    {
      l = nattrntrees(m, SME_ReferencesATTRIBUTENUMBER);
      for (i = 1; i <= ntreearity(l); i++)
	printReference(nchild(i, l));
    }
}

void SME_Printer::printImport(ntrees m)
{
  noperators op;
  ntrees vdecl, vmodule;
  int ikw;

  op = noper(m);		//  opSMEImportType (type) || opSMEImportCons (constant)
  if (op == opSMEImportCons)
    {
      /* import of a constant:  
         <constantRef SME:id="_vwCFgO-KEdyObL5zFqHJ6w">
         <constant xsi:type="sme:ConstantValue" href="ModuleEx.sme#_zjESQO-KEdyObL5zFqHJ6w"/>
         </constantRef>
       */
      ikw = IKWConstantRef;
      printBeginBlock(ikw);

      printSMEId(m);
      printConnectionFrom(m);

      getDeclarationAttribute(m, &vdecl);
      // printKwId(vdecl, IKWundef, IKWCstName);
      printCloseHeader();

      if (getAttributeModule(vdecl, &vmodule))
	{
	  printBeginBlock(IKWconstant, IKWconstantValue, pk_false);
	  printModuleReference(vmodule, vdecl);
	}
      closeBlockLine();
      printTailDef(ikw);
    }
  else if (op == opSMEImportType)
    {
      // import of a type: <types xsi:type="sme:SubstitutionType" SME:id="_NjE3gO-LEdyObL5zFqHJ6w" name="SubstitutionType1" type="Type1"/>
      ikw = IKWtype;
      printHeader(IKWtypes, m, tkwundef, IKWSubstitutionType, pk_false);
      getDeclarationAttribute(m, &vdecl);
      printKwId(nchild(1, vdecl), IKWundef, IKWname);	// changed July 2010-- printKwId(vdecl, IKWundef, IKWname);
      closeBlockLine();
    }

}

void SME_Printer::printImports(ntrees m)
{
  int i;
  ntrees l;

  if (nattrexist(m, SME_ImportATTRIBUTENUMBER))
    {
      l = nattrntrees(m, SME_ImportATTRIBUTENUMBER);
      for (i = 1; i <= ntreearity(l); i++)
	printImport(nchild(i, l));
    }
}

/**
   Model declaration printing without first keyword.
   (action, process, node, private node,...)
*/
void SME_Printer::processUnparsing()
{
  ntrees nomproc, m;
  noperators op;
  int ikw, ikw1;
  boolean b;

  SET_NILTREE(nomproc);
  op = voperator;
  m = treeunparseTree;
  b = FirstHeader;
  snline();
  if (ischildprint(1))
    {
      nomproc = operande;
      ikw = beginModel(m, op, operande);
    }
  else
    this_intePrinter->printerror();

  //body
  treeunparseTree = m;
  wline(vholoph - actindent);
  if (ischildprint(3))
    unparse_subtree(operande, vholoph - 1);
  else
    this_intePrinter->printerror();

  // interface
  treeunparseTree = m;
  wline(vholoph - actindent);
  if (ischildprint(2))
    unparse_subtree(operande, vholoph + 1);
  else
    this_intePrinter->printerror();
  // Connections

  // final
  if (b)
    ikw1 = IKWsme;
  else
    ikw1 = IKWundef;
  endModel(ikw1, ikw, nomproc);
}

boolean SME_Printer::getAttributeModule(ntrees vdecl, ntrees * module)
{
  if (!nattrexist(vdecl, SME_ModuleATTRIBUTENUMBER))
    return pk_false;
  (*module) = (ntrees) nattrptr(vdecl, SME_ModuleATTRIBUTENUMBER);
  return ((*module) != currentModule);
  //  return pk_true;
}

void SME_Printer::printRefModule(ntrees p)
{
  if (!STree_isUndefinedTree(p))
    {
      printTree(STreeModelModule_getName(p));
      sncharpri('.');
      printSMEKw(IKWsme);
      sncharpri('#');
    }
}

void SME_Printer::printQuotedSMEIdValue(ntrees module, ntrees p)
{
  beginString();
  printRefModule(module);
  printSMEIdValue(p);
  endString();
}

void SME_Printer::printQuotedStr(ptrepstr vstr)
{
  beginString();
  sigpristr(vstr);
  endString();
}

void SME_Printer::printSMEIdValue(ntrees p)
{
  sigpristr(nstringof(p));
}

boolean SME_Printer::getSMEId(ntrees p, ntrees * v)
{
  if (!nattrexist(p, SME_idATTRIBUTENUMBER))
    return pk_false;
  *v = nattrntrees(p, SME_idATTRIBUTENUMBER);
  return pk_true;
}

boolean SME_Printer::getSMEIdOperand(ntrees p, ntrees * v)
{
  if (!nattrexist(p, SME_idoperandATTRIBUTENUMBER))
    return pk_false;
  *v = nattrntrees(p, SME_idoperandATTRIBUTENUMBER);
  return pk_true;
}

void SME_Printer::printSMEId(ntrees p)
{
  printKwSMEId(p, IKWxmi, IKWid);
}

void SME_Printer::printKwKwSMEId(int kw1, int kw2, ntrees module, ntrees v)
{
  printSMEKwKwEq(kw1, kw2);
  printQuotedSMEIdValue(module, v);
}

// "Undefined SME Id detected, add one : UndefinedId_<value(cpt)>
void SME_Printer::UndefinedSMEId(int kw1, int kw2)
{
  cpt++;
  printSMEKwKwEq(kw1, kw2);
  beginString();
  sigpristr(undefSMEId);
  sigpristr(integer2repstr(cpt));
  endString();
}

ntrees SME_Printer::printSMEIdOperand(ntrees op)
{
  ntrees v, res;
  boolean b;

  res = op;
  if (getSMEIdOperand(op, &v))
    {
      b = pk_true;
      res = v;
    }
  else
    b = getSMEId(op, &v);
  if (b)
    printKwKwSMEId(IKWxmi, IKWid, NILTREE, v);
  else
    UndefinedSMEId(IKWxmi, IKWid);

  return res;
}

void SME_Printer::printKwSMEId(ntrees p, int kw1, int kw2)
{
  ntrees v;

  if (getSMEId(p, &v))
    printKwKwSMEId(kw1, kw2, NILTREE, v);
  else
    UndefinedSMEId(kw1, kw2);
}

_VIRTUAL_ void SME_Printer::printOperator(int tkw)
{
  if (tkw != tkwundef)
    {
      sncharpri(' ');
      printSMEKwEq(IKWoperator);
      beginString();
      this_intePrinter->printkey(tkw);
      endString();
    }
}

/**
   currentName = getSMEKeyword(xsiType) + Names_Table[xsiType]
*/
void SME_Printer::setCurrentName(int xsiType)
{
  if (Names_Table.find(xsiType) == Names_Table.end())
    Names_Table[xsiType] = 1;
  else
    Names_Table[xsiType]++;
  if (!strconc2str(getSMEKeyword(xsiType), integer2repstr(Names_Table[xsiType]), pk_false, &currentName))
    currentName = undefSMEId;
}

void SME_Printer::printName(int xsiType, int opIndex, int ikwop)
{
  snspace();
  printSMEKwEq(IKWname);	// name=
  beginString();
  sigpristr(currentName);
  if (opIndex > 0)
    {
      sncharpri('_');
      printSMEKw(ikwop);
      sigpristr(integer2repstr(opIndex));
    }
  endString();
}

boolean SME_Printer::isNamedExpression(int xsiType)
{
  if (xsiType == IKWnullClock)
    return pk_false;
  return pk_true;
}

_VIRTUAL_ void SME_Printer::printXsiType(int xsiType, boolean addName)
{
  if (xsiType > 0)
    {
      // xsi:type="sme:xsiType"
      printSMEKwKwEq(IKWxsi, IKWtype);

      beginString();
      printSMEKwKw(IKWsme, xsiType);
      endString();

      if (addName)
	{
	  setCurrentName(xsiType);
	  printName(xsiType, -1, IKWundef);
	}
    }
}

void SME_Printer::printBeginBlock(int ikind)
{
  sncharpri('<');
  printSMEKw(ikind);
  sncharpri(' ');
}

void SME_Printer::printBeginBlock(int ikind, int xsiType, boolean addname)
{
  printBeginBlock(ikind);
  printXsiType(xsiType, addname);
}

_VIRTUAL_ void SME_Printer::printSMEBracketedKw(int ikw1, int ikw2, boolean b)
{
  sncharpri('<');
  if (b)
    sncharpri('/');
  printSMEKwKw(ikw1, ikw2);
  sncharpri('>');
  if (b)
    snline();
}

_VIRTUAL_ void SME_Printer::printSMEBracketedKw(int ikw, boolean b)
{
  printSMEBracketedKw(IKWundef, ikw, b);
}

_VIRTUAL_ void SME_Printer::printConnectionFrom(ntrees p)
{
  ntrees vt;
  int i;

  if (nattrexist(p, SME_ATTRIBUTENUMBER))
    {
      vt = nattrntrees(p, SME_ATTRIBUTENUMBER);
      for (i = 1; i <= ntreearity(vt); i++)
	{
	  sncharpri(' ');
	  printSMEattr(nchild(i, vt));
	}
    }
}

boolean SME_Printer::selectedOpUnparse(ntrees * t)
{
  noperators op;

  op = noper(*t);
  if (op == opconv)
    {
      (*t) = nchild(2, *t);
      op = noper(*t);
    }
  return !(isIdentOperator(op));	// (op != opident) && (op != opnom);
}

int SME_Printer::getPrefixOperand(int xsiType, int ith)
{

  if ((xsiType == IKWComparisonOperator) || (xsiType == IKWCounter) || (xsiType == IKWrepetition)
      || (xsiType == IKWComplexOperator) || (xsiType == IKWarrayRestructuration) || (xsiType == IKWArrayRecovery))
    {
      if (ith == 1)
	return IKWfirstOperand;
      if (ith == 2)
	return IKWSecondOperand;
    }

  if (xsiType == IKWIf)
    {
      if (ith == 2)
	return IKWfirstOperand;
      if (ith == 3)
	return IKWSecondOperand;
    }

  if (ith == 1)
    {
      if (xsiType == IKWExtraction)
	return IKWin;
      if (xsiType == IKWVar)
	return IKWin;
      if (xsiType == IKWWindow)
	return IKWin;
      if (xsiType == IKWnot)
	return IKWin;
      if (xsiType == IKWDelay)
	return IKWin;
      if (xsiType == IKWCell)
	return IKWin;
      if (xsiType == IKWUminus)
	return IKWin;
      if (xsiType == IKWTransposition)
	return IKWin;
    }
  return IKWins;

}

void SME_Printer::beginOperand(int ith, int xsiType)
{
  int kw;

  sncharpri(' ');
  sncharpri(' ');
  sncharpri('<');
  kw = getPrefixOperand(xsiType, ith);
  printSMEKw(kw);
}

void SME_Printer::endOperand(int xsiType, int opIndex)
{
  printName(xsiType, opIndex, IKWin);
  closeBlockLine();
}

ntrees SME_Printer::getSMETreeOf(ntrees op)
{
  noperators vop;

  vop = noper(op);
  if (vop == opevnt)
    return getSMETreeOf(nchild(1, op));
  if (vop == opconv)
    return getSMETreeOf(nchild(2, op));
  if (vop == opelem)
    return getSMETreeOf(nchild(1, op));
  if (vop == opcham)
    return getSMETreeOf(nchild(1, op));
  //  if (vop == oprecu) return getSMETreeOf(nchild(1,op));
  if (vop == opsdec)
    return getSMETreeOf(nchild(1, op));
  return op;
}

_VIRTUAL_ void SME_Printer::opUnparse(ntrees op, ntrees father, int ith, int xsiType)
{
  beginOperand(ith, xsiType);
  op = getSMETreeOf(op);
  op = printSMEIdOperand(op);
  printConnectionFrom(op);
  printKwSMEId(father, IKWundef, IKWoperatorParent);
  endOperand(xsiType, ith);
}

void SME_Printer::printKwQuotedSignalText(int IKW, ntrees vtree, boolean bxmlChar)
{
  sncharpri(' ');
  printSMEKwEq(IKW);

  sncharpri('"');
  if (bxmlChar)
    setCharEncodingMode(pk_false);	// XML characters.
  printSignalText(vtree);
  sncharpri('"');
  if (bxmlChar)
    setCharEncodingMode(pk_true);	// std characters.
}

void SME_Printer::printHeaderAttribute(int IKW, ntrees vtree, int iSon)
{
  ntrees s;
  if (getchild(vtree, iSon, &s))
    printKwQuotedSignalText(IKW, s, pk_true);
}

void SME_Printer::printHeader(int kwkind, ntrees t, int tkw, int xsiType, boolean bname)
{
  if (bname)
    bname = isNamedExpression(xsiType);

  printBeginBlock(kwkind, xsiType, bname);
  printSMEId(t);
  printConnectionFrom(t);

  printOperator(tkw);
  printFlags(t);
}

void SME_Printer::printCloseHeader()
{
  sncharpri('>');
  snline();
}

void SME_Printer::printHeaderDef(int kwkind, ntrees t, int tkw, int xsiType, boolean named)
{
  printHeader(kwkind, t, tkw, xsiType, named);
  printCloseHeader();
}

void SME_Printer::printTailDef(int ikwkind)
{
  printSMEBracketedKw(ikwkind, pk_true);
}

void SME_Printer::printSelectedTree(ntrees p)
{
  if (selectedOpUnparse(&p))
    unparse_subtree(p, 9999);
}

void SME_Printer::printOperands(ntrees vtree)
{
  int i, nb;
  ntrees p;

  nb = ntreearity(vtree);
  for (i = 1; i <= nb; i++)
    if (getchild(vtree, i, &p))
      printSelectedTree(p);

}

void SME_Printer::printOpOperator(ntrees vtree, ntrees father, int xsiType)
{
  int i, nb;
  ntrees p;

  nb = ntreearity(vtree);
  for (i = 1; i <= nb; i++)
    if (getchild(vtree, i, &p))
      opUnparse(p, father, i, xsiType);
}

_VIRTUAL_ void SME_Printer::printBinOperator(int classOp, int tkw, int op, int xsiType)
{
  ntrees vtree;

  vtree = treeunparseTree;
  printHeaderDef(classOp, vtree, tkw, xsiType, pk_true);
  printOpOperator(vtree, vtree, xsiType);
  printTailDef(classOp);
  printOperands(vtree);
}

ntrees SME_Printer::getFormalId(int ikw, ntrees model, int rank)
{
  ntrees formalId;

  if (ikw == IKWinputs)
    STree_getIthInput(model, rank, &formalId);
  else if (ikw == IKWoutputs)
    STree_getIthOutput(model, rank, &formalId);
  else if (ikw == IKWparameters)
    STree_getIthParameter(model, rank, &formalId);
  else if (ikw == IKWouts)
    STree_getIthOutput(model, rank, &formalId);	// cartesian product.
  else
    SET_NILTREE(formalId);
  return formalId;
}

int SME_Printer::getFormalKw(int ikw)
{
  if (ikw == IKWinputs)
    return IKWinput;
  if (ikw == IKWoutputs)
    return IKWoutput;
  if (ikw == IKWparameters)
    return IKWparameter;
  return IKWundef;
}

/**
   it prints the SME-identifier of an input/output/parameter element, selected by ikw,  of a model (model)
*/
void SME_Printer::printRefObj(ntrees model, int ikw, int rank)
{
  ntrees formalId;

  formalId = getFormalId(ikw, model, rank);
  sncharpri(' ');
  printKwSMEId(formalId, IKWundef, getFormalKw(ikw));
}

/**
   it prints the SME-identifier of an input/output/parameter element, selected by ikw,  of a model (model)
*/
void SME_Printer::printRefObjModule(ntrees module, ntrees model, int ikw, int rank)
{
  ntrees formalId;
  // <input href="ModuleEx.sme#_G-8joOuPEdyxR_OpGHu7Fg"/>
  formalId = getFormalId(ikw, model, rank);
  printBeginBlock(getFormalKw(ikw), IKWundef, pk_false);
  printModuleReference(module, formalId);
  closeBlockLine();
  snline();
}

/**
   It prints a string built from the name of the (rank)-th input/output/parameter (ikw) of a model (model).
   For a name foo, it prints ~foo followed by an int if it is not the first reference to the name.
 */
void SME_Printer::printNameValue(ntrees id)
{
  int n;
  ptrepstr vstr;

  vstr = ident2repstr(id);	// unicity of the string (ptrepstr).
  if (formalIdNames_Table.find(vstr) == formalIdNames_Table.end())
    formalIdNames_Table[vstr] = 0;
  else
    formalIdNames_Table[vstr]++;
  n = formalIdNames_Table[vstr];
  snspace();
  printSMEKwEq(IKWname);	// name=
  beginString();
  sncharpri('_');
  printTree(id);
  if (n > 0)
    sigpristr(integer2repstr(n));
  endString();
}

void _VIRTUAL_ SME_Printer::printEffectiveValue(ntrees model, ntrees op, int ikw, int rank, ntrees module)
{
  boolean bm;

  op = getSMETreeOf(op);
  //  op = printSMEIdOperand(op);  

  sncharpri(' ');
  printBeginBlock(ikw);
  op = printSMEIdOperand(op);
  printConnectionFrom(op);

  bm = pk_true;
  if (ISNOT_NILTREE(model))
    if (noper(model) == oppcar)
      {
	printKwSMEId(model, IKWundef, IKWoperatorParent);	// cartesian product.
	bm = pk_false;
      }
    else if (IS_NILTREE(module))
      {
	printRefObj(model, ikw, rank);	// a ref to a model.
	bm = pk_false;
      }
  printNameValue(getFormalId(ikw, model, rank));
  if (!bm)
    closeBlockLine();
  else
    {
      printCloseHeader();
      printRefObjModule(module, model, ikw, rank);
      printSMEBracketedKw(ikw, pk_true);
    }
}

void _VIRTUAL_ SME_Printer::printEffectiveValues(ntrees model, ntrees leff, int IKW, ntrees module)
{
  int i;
  ntrees s;

  if (!STree_isUndefinedTree(leff))
    if (STree_isList(leff))
      {
	for (i = 1; i <= ntreearity(leff); i++)
	  if (getchild(leff, i, &s))
	    printEffectiveValue(model, s, IKW, i, module);
      }
    else
      printEffectiveValue(model, leff, IKW, 1, module);
}

boolean SME_Printer::getOPorts(ntrees t, ntrees * vdecl)
{
  if (!nattrexist(t, SME_OPortsATTRIBUTENUMBER))
    return pk_false;
  (*vdecl) = nattrntrees(t, SME_OPortsATTRIBUTENUMBER);
  return pk_true;
}

_VIRTUAL_ void SME_Printer::printEffectiveOutputs(ntrees model, ntrees rm, int ikw, ntrees module)
{
  ntrees s, port, id, l;
  int i;

  if (getOPorts(rm, &l))
    {
      for (i = 1; i <= ntreearity(l); i++)
	{
	  s = nchild(i, l);
	  port = nchild(1, s);
	  id = nchild(2, s);
	  printEffectiveValue(model, id, ikw, (int)nintof(port), module);
	}
    }
}

_VIRTUAL_ void SME_Printer::unparse_subtree(ntrees p, int holoph)
{
  SignalPrinter::unparse_subtree(p, holoph);
  snglue(pk_true);
}

/**
   Printer initialisations:
   - An \ref intePrinter object is associated with the printer, if it is not yet done.
   - A "printer mode" is defined in this interface. It allows to print or not the comments.
   
   For a graphical mode, the comments are not printed, and for a textual one, they are printed.
   By default, the "textual mode" is set. To change it, use intePrinter::setPrinterContext() method.

   It loads:
   - the Signal-SME table of keywords, called SIGNALSME.kw, defined in $Signal_bin directory.
   This table contains the redefined key words of SIGNAL.
   - a specific table (SMEKEYWORDS.kw defined in $Signal_bin directory)

   @return "true" if the operations succeed, "false" otherwise.
*/
_VIRTUAL_ boolean SME_Printer::initPrinter()
{
  if (notannounced)
    {
      notannounced = pk_false;
      this_intePrinter = new intePrinter(string2repstr(comBegin), string2repstr(comEnd), string2repstr(comincom), string2repstr(comincomend), string2repstr(errSymbol), string2repstr(cstSymbolHolo));
      /* Loading tables (SIGNALSME.kw, SMEKEYWORDS.kw defined in the $Signal_bin directory). */
      if (!this_intePrinter->load_KWLangTable("SIGNALSME"))
	return pk_false;
      if (!tabkwload("SMEKEYWORDS", &SMEbasekeyword))
	return pk_false;
      this_intePrinter->setPrinterContext(pk_true);
    }
  return pk_true;
}

void SME_Printer::printKwQuotedStr(int ikw, ptrepstr str)
{
  sncharpri(' ');
  printSMEKwEq(ikw);
  printQuotedStr(str);
}

/**
   it generates <? xml version="1.0" encoding="UTF-8" ?>
*/
_VIRTUAL_ void SME_Printer::SMEBegin()
{
  setCharEncodingMode(pk_true);	// STD char encoding mode.
  sncharpri('<');
  sncharpri('?');
  printSMEKw(IKWxml);
  printKwQuotedStr(IKWversion, string2repstr(XmlVersion));
  printKwQuotedStr(IKWencoding, string2repstr(UTF_str));
  sncharpri('?');
  sncharpri('>');
}

_VIRTUAL_ void SME_Printer::SMEEnd()
{
  setCharEncodingMode(pk_true);	// reset XML mode (set in SME_Printer::SMEBegin())
}

/* ---------- Nullary operators -----------*/

/**
   Printing a nullary tree: opnom operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opnom()
{
  inteconsalphaprint(treeunparseTree);
}

/**
   Printing a nullary tree: opchar operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opchar()
{
  this_intePrinter->printalphakey(kwchar);
}

/**
   Printing a nullary tree: opcomplex operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opcomplex()
{
  this_intePrinter->printalphakey(kwcomplex);
}

/**
   Printing a nullary tree: opcstcar operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opcstcar()
{
  printConstant(opcstcar);	// inteconsalphaprint(treeunparseTree); 
}

/**
   Printing a nullary tree: opcstcha operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opcstcha()
{
  printConstant(opcstcha);	// inteconsalphaprint(treeunparseTree); 
}

/**
   Printing a nullary tree: opcstdp operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opcstdp()
{
  printConstant(opcstdp);	// inteconsalphaprint(treeunparseTree); 
}

/**
   Printing a nullary tree: opcstent operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opcstent()
{
  printConstant(opcstent);	//  inteconsalphaprint(treeunparseTree); 
}

/**
   Printing a nullary tree: opcstreel operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opcstreel()
{
  printConstant(opcstreel);	// inteconsalphaprint(treeunparseTree); 
}

/**
   Printing a nullary tree: opdcomplex operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opdcomplex()
{
  this_intePrinter->printalphakey(kwdcomplex);
}

/**
   Printing a nullary tree: opdeterm operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opdeterm()
{
  printSMEKwEqQuotedKey(IKWmodelClass, IKWdeterministic);
}

/**
   Printing a nullary tree: opdpreal operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opdpreal()
{
  this_intePrinter->printalphakey(kwdreal);
}

/**
   Printing a nullary tree: opevent operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opevent()
{
  this_intePrinter->printalphakey(kwevent);
}

/**
   Printing a nullary tree: ophnull operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_ophnull()
{
  printBinOperator(IKWConstraints, tkwundef, ophnull, IKWnullClock);
}

/**
   Printing a nullary tree: opfalse operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opfalse()
{
  printConstant(opfalse);	// this_intePrinter->printalphakey(kwfalse);
}

/**
   Printing a nullary tree: opimplicit operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opimplicit()
{
}

/**
   Printing a nullary tree: opinteger operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opinteger()
{
  this_intePrinter->printalphakey(kwinteger);
}

/**
   Printing a nullary tree: oplogical operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_oplogical()
{
  this_intePrinter->printalphakey(kwboolean);
}

/**
   Printing a nullary tree: oplong operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_oplong()
{
  this_intePrinter->printalphakey(kwlong);
}

/**
   Printing a nullary tree: opnil operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opnil()
{
}

/**
   Printing a nullary tree: opnullproc operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opnullproc()
{
  /*
     this_intePrinter->printkey(kwpouv);
     this_intePrinter->printkey(kwvert);
     snspace();
     this_intePrinter->printkey(kwvert);
     this_intePrinter->printkey(kwpfer);
   */
}

/**
   Printing a nullary tree: opop  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opop()
{
  inteconsalphaprint(treeunparseTree);
}

/**
   Printing a nullary tree: opreal  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opreal()
{
  this_intePrinter->printalphakey(kwreal);
}

/**
   Printing a nullary tree: opretun  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opretun()
{
}

/**
   Printing a nullary tree: opsafe  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opsafe()
{
  printSMEKwEqQuotedKey(IKWmodelClass, IKWsafe);
}

/**
   Printing a nullary tree: opshort  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opshort()
{
  this_intePrinter->printalphakey(kwshort);
}

/**
   Printing a nullary tree: opstring  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opstring()
{
  this_intePrinter->printalphakey(kwstring);
}

/**
   Printing a nullary tree: optabvide  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_optabvide()
{
  this_intePrinter->printcspkey(kwcouv);
  this_intePrinter->printalphakey(kwcfer);
}

/**
   Printing a nullary tree: optrue  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_optrue()
{
  printConstant(optrue);
  // this_intePrinter->printalphakey(kwtrue);
}

/**
   Printing a nullary tree: opunsafe  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opunsafe()
{
  printSMEKwEqQuotedKey(IKWmodelClass, IKWunsafe);
}

/**
   Printing a nullary tree: opcomment  operator. 
*/
_VIRTUAL_ void SME_Printer::decta0_opcomment()
{
  inteconsprint(treeunparseTree);
}

/* ---------- Unary operators -----------*/

/**
	Printing an unary tree: opevnt operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opevnt()
{
  printConnectionFrom(treeunparseTree);
  printOperands(treeunparseTree);
}

/**
	Printing an unary tree: opexte operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opexte()
{
  this_intePrinter->printalphakey(kwexternal);
  recosonunparse(1);
}

/**
	Printing an unary tree: oplab operator.
*/
_VIRTUAL_ void SME_Printer::decta1_oplab()
{
  /*
     if (ischildprint(1))
     {
     markkeyunparse(kwlabel, vholoph);
     this_intePrinter->printkeysp(kwpvir);
     }
   */
}

/**
	Printing an unary tree: opnega operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opnega()
{
  printBinOperator(IKWoperators, tkwundef, opnega, IKWUminus);
}

/**
   Printing an unary tree: opnon operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opnon()
{
  printBinOperator(IKWoperators, tkwundef, opnon, IKWnot);
}

/**
   Printing an unary tree: oppcar operator.
*/
_VIRTUAL_ void SME_Printer::decta1_oppcar()
{
  ntrees vtree;

  vtree = treeunparseTree;
  printHeaderDef(IKWoperators, vtree, tkwundef, IKWCartesianProduct, pk_true);
  printOpOperator(nchild(1, vtree), vtree, IKWCartesianProduct);
  printEffectiveOutputs(vtree, vtree, IKWouts, NILTREE);
  printTailDef(IKWoperators);
  printOperands(vtree);
}

/**
   Printing an unary tree: opref operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opref()
{
  if (ischildprint(1))
    {
      markkeyunparse(kwref, vholoph);
      this_intePrinter->printkeysp(kwpvir);
    }
}

/**
	Printing an unary tree: opsdec operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opsdec()
{
  ntrees s = nchild(1, treeunparseTree);
  printSelectedTree(s);
}

/**
	Printing an unary tree: opstru operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opstru()
{
  // this_intePrinter->printalphakeysp(kwstruct);
  wline(vholoph - actindent);
  if (ischildprint(1))
    unparse_subtree(operande, 9999);
  // markkeyunparsekey(kwpouv, kwpfer, vholoph);
}

/**
	Printing an unary tree: optran operator.
*/
_VIRTUAL_ void SME_Printer::decta1_optran()
{
  printBinOperator(IKWoperators, tkwundef, optran, IKWTransposition);
}

/**
	Printing an unary tree: opvvarin operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opvvarin()
{
  // does not occur.
  recsonunparse(1);
  this_intePrinter->printkey(kwinte);
}

/**
	Printing an unary tree: opwhen1 operator.
*/
_VIRTUAL_ void SME_Printer::decta1_opwhen1()
{
  ntrees vtree;

  vtree = treeunparseTree;
  printHeaderDef(IKWoperators, vtree, tkwundef, IKWExtraction, pk_true);
  beginOperand(1, IKWExtraction);
  UndefinedSMEId(IKWxmi, IKWid);
  endOperand(IKWExtraction, 1);
  printTailDef(IKWoperators);
  printOperands(vtree);

}

/**
   Printing an unary tree: oppassert operator: assert(| ... |).
*/
_VIRTUAL_ void SME_Printer::decta1_oppassert()
{
  ntrees vtree;
  // Currently, it is unuseful to manage a stack of such a context.
  assertContext = pk_true;
  vtree = treeunparseTree;
  printOperands(vtree);
  // pop context
  assertContext = pk_false;
}

/**
   Printing an unary tree: opappeassert operator: assert(C).
*/
_VIRTUAL_ void SME_Printer::decta1_opappeassert()
{
  ntrees vtree;

  vtree = treeunparseTree;
  printHeaderDef(IKWassertions, vtree, tkwundef, IKWundef, pk_true);
  printTailDef(IKWassertions);
  printOperands(vtree);
}

/* ---------- Binary operators -----------*/

void SME_Printer::printModuleReference(ntrees module, ntrees objModule)
{
  ntrees v;
  if (getSMEId(objModule, &v))
    printKwKwSMEId(IKWundef, IKWhref, module, v);
  else
    UndefinedSMEId(IKWundef, IKWhref);
}

/**
   Printing a binary tree: opappe operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opappe()
{
  ntrees vmodule, vtree, id, ins, lp, model;
  boolean b;

  vtree = treeunparseTree;
  id = STree_getNameCalledModel(vtree);
  ins = nchild(2, vtree);
  SET_NILTREE(lp);

  printBeginBlock(IKWRefModel);
  printSMEId(vtree);
  sncharpri(' ');

  if (getDeclarationAttribute(id, &model))
    {
      b = getAttributeModule(model, &vmodule);
      if (!b)
	{
	  printKwSMEId(model, IKWundef, IKWRefInterface);
	  SET_NILTREE(vmodule);
	}

      /* setCurrentName(IKWRefModel);
         printName(IKWRefModel, -1, IKWundef);   name="containedInstances1"
         replaced by, 23/07/2008
       */
      printNameValue(id);
      printCloseHeader();
      if (b)
	{
	  printBeginBlock(IKWRefInterface, IKWModelDeclaration, pk_false);
	  printModuleReference(vmodule, model);
	  closeBlockLine();
	}

      // outputs
      printEffectiveOutputs(model, vtree, IKWoutputs, vmodule);
      // inputs

      printEffectiveValues(model, ins, IKWinputs, vmodule);
      // parameters
      lp = nchild(1, vtree);
      if (noper(lp) == opinst)
	lp = nchild(2, lp);
      else
	SET_NILTREE(lp);
      printEffectiveValues(model, lp, IKWparameters, vmodule);

    }
  else
    this_intePrinter->printerror();

  printSMEBracketedKw(IKWRefModel, pk_true);

  printOperands(lp);
  printOperands(ins);

}

/**
   Printing a binary tree: opbinf operator. (^*)
*/
_VIRTUAL_ void SME_Printer::decta2_opbinf()
{
  printBinOperator(IKWConstraints, kwchapetoi, opbinf, IKWClockRelationOperator);
}

/**
   Printing a binary tree: opbsup operator.(^+)
*/
_VIRTUAL_ void SME_Printer::decta2_opbsup()
{
  printBinOperator(IKWConstraints, kwchapplus, opbsup, IKWClockRelationOperator);
}

/**
   Printing a binary tree: opcomp operator.(^-)
*/
_VIRTUAL_ void SME_Printer::decta2_opcomp()
{
  printBinOperator(IKWConstraints, kwchapminu, opcomp, IKWClockRelationOperator);
}

/**
   Printing a binary tree: opbund operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opbund()
{
  this_intePrinter->printalphakeysp(kwbundle);
  if (ischildprint(1))
    markkeyunparsekey(kwpouv, kwpfer, vholoph);
  if (ischildprint(2))
    {
      wline(vholoph - actindent);
      sigargprocunp();
    };
}

void SME_Printer::printSubProcess(ntrees operand)
{
  boolean sp;
  int ikw;

  sp = (noper(operand) != oppconf);
  if (sp)
    {
      setCurrentName(IKWConfinedProcess);
      ikw = beginModel(operand, oppconf, repstr2ident(currentName));
    }
  unparse_subtree(operand, 9999);
  snline();
  if (sp)
    endModel(IKWundef, ikw, NILTREE);
}

/**
   Printing a binary tree: opcas operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opcas()
{
  ntrees operand;

  /* son 1 : opintvff opintvfo opintvof opintvoo + oplesign
     Son 1, not generated here, it is associated with the switch selector.
     if (getchild(treeunparseTree, 1, &operand)) unparse_subtree(operand, 9999);
   */

  if (getchild(treeunparseTree, 2, &operand))
    printSubProcess(operand);
}

/**
   Printing a binary tree: opcham operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opcham()
{
  ntrees s = nchild(1, treeunparseTree);
  printSelectedTree(s);
}

/**
   Printing a binary tree: opconc operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opconc()
{
  printBinOperator(IKWoperators, tkwundef, opconc, IKWConcatenation);
}

/**
   Printing a binary tree: opcons operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opcons()
{
  recsonunparse(1);
  if (!marktalphakeyunparse(kwequl, vholoph - 1, 2))
    this_intePrinter->printerror();
}

/**
   Printing a binary tree: opconv operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opconv()
{
  // translated in a flag caseType="predefinedType"
  ntrees s = nchild(2, treeunparseTree);
  printSelectedTree(s);
}

/**
   Printing a binary tree: opcplx operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opcplx()
{
  printBinOperator(IKWoperators, tkwundef, opcplx, IKWComplexOperator);	//  binsigexp(kwarro, 0);
}

/**
   Printing a binary tree: opdeccon operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdeccon()
{
  setVisibility(IKWpublic);
  printdeclaration(treeunparseTree, IKWconstants, opdeccon);
}

/**
   Printing a binary tree: opdecconp operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdecconp()
{
  setVisibility(IKWprivate);
  printdeclaration(treeunparseTree, IKWconstants, opdeccon);
}

/**
   Printing a binary tree: opdecsig operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdecsig()
{
  printdeclaration(treeunparseTree, IKWlocals, opdecsig);
}

/**
   Printing a binary tree: opdecvar operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdecvar()
{
  printdeclaration(treeunparseTree, IKWlocals, opdecvar);
}

/**
   Printing a binary tree: opdecpar operator. (shared)
*/
_VIRTUAL_ void SME_Printer::decta2_opdecpar()
{
  printdeclaration(treeunparseTree, IKWlocals, opdecpar);
}

/**
   Printing a binary tree: opdefegal operator (:=:)
*/
_VIRTUAL_ void SME_Printer::decta2_opdefegal()
{
  printClockRelation(kwdpinequldpin);
}

/**
   Printing a binary tree: opdefpart operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdefpart()
{
  decta2_opdefsig();

  /*
     recsonunparse(1);
     if (!marktalphakeyunparse(kwdpindpinequl, vholoph - 1, 2))
     this_intePrinter->printerror();
   */
}

/**
   Printing a binary tree: opdefpartdef operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdefpartdef()
{
  decta2_opdefsig();
}

/**
   Printing a binary tree: opdefsig operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdefsig()
{
  ntrees s;

  // X:= E , first child is not printed (connection), 
  // second child is printed if it is not an identifier.
  s = nchild(2, treeunparseTree);
  printSelectedTree(s);
}

/**
   Printing a binary tree: opdesc operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdesc()
{
  if (ischildprint(1))
    unparse_subtree(operande, vholoph);
  if (ischildprint(2))
    sigargincprimprocunp();
}

/**
   Printing a binary tree: opdescmodu operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdescmodu()
{
  if (ischildprint(1))
    unparse_subtree(operande, vholoph);
  if (ischildprint(2))
    unparse_subtree(operande, vholoph);
}

/**
   Printing a binary tree: opdiff operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdiff()
{
  printBinOperator(IKWoperators, kwslahequl, opdiff, IKWComparisonOperator);
}

/**
   Printing a binary tree: opdivi operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opdivi()
{
  printBinOperator(IKWoperators, kwslah, opdivi, IKWArithmeticOperator);
}

/**
   Printing a binary tree: opegal operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opegal()
{
  printBinOperator(IKWoperators, kwequl, opegal, IKWComparisonOperator);
}

/**
   Printing a binary tree: opegeg operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opegeg()
{
  printBinOperator(IKWoperators, kwequlequl, opegeg, IKWComparisonOperator);
}

/**
   Printing a binary tree: opelem operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opelem()
{
  ntrees s = nchild(1, treeunparseTree);
  printSelectedTree(s);
}

/**
   Printing a binary tree: opelevec operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opelevec()
{
  reccoper(1, pk_true);
  this_intePrinter->printkey(kwdpin);
  recsondecunparse(2);
}

/**
   Printing a binary tree: opensvec operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opensvec()
{
  recaoper(1, pk_false);
  this_intePrinter->printkey(kwdpin);
  recsondecunparse(2);
}

/**
   Printing a binary tree: opentsor operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opentsor()
{
  this_intePrinter->printkeymark(kwpouv, vholoph - actindent);
  if (marktkeyunparse(kwinte, vholoph, 1))
    wline(vholoph - actindent - 1);
  if (marktkeyunparse(kwxcla, vholoph, 2))
    wline(vholoph - actindent - 1);
  this_intePrinter->printkeyrelease(kwpfer, vholoph - actindent);
}

/**
   Printing a binary tree: opet operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opet()
{
  printBinOperator(IKWoperators, tkwundef, opet, IKWand);
}

/**
   Printing a binary tree: opou operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opou()
{
  printBinOperator(IKWoperators, tkwundef, opou, IKWor);
}

/**
   Printing a binary tree: opouex operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opouex()
{
  printBinOperator(IKWoperators, tkwundef, opouex, IKWxor);
}

/**
   Printing a binary tree: opextr operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opextr()
{
  ntrees vtree, p;

  vtree = treeunparseTree;
  printHeaderDef(IKWoperators, vtree, tkwundef, IKWExtraction, pk_true);
  if (getchild(vtree, 1, &p))
    opUnparse(p, vtree, 1, IKWExtraction);
  //  if (getchild(vtree, 2, &p))  opUnparse(p, vtree, 2, IKWExtraction); // --> condition
  printTailDef(IKWoperators);
  printOperands(vtree);
}

/**
   Printing a binary tree: opformact operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opformact()
{
  this_intePrinter->printalphakeysp(kwaction);
  recsonunparse(1);
  recsonunparse(2);
}

/**
   Printing a binary tree: opformfunc operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opformfunc()
{
  this_intePrinter->printalphakeysp(kwfunction);
  recsonunparse(1);
  recsonunparse(2);
}

/**
   Printing a binary tree: opformnode operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opformnode()
{
  this_intePrinter->printalphakeysp(kwnode);
  recsonunparse(1);
  recsonunparse(2);
}

/**
   Printing a binary tree: opformproc operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opformproc()
{
  this_intePrinter->printalphakeysp(kwprocess);
  recsonunparse(1);
  recsonunparse(2);
}

/**
   Printing a binary tree: opiegal operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opiegal()
{
  printBinOperator(IKWoperators, kwinfeequl, opiegal, IKWComparisonOperator);
}

/**
   Printing a binary tree: opiieg operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opiieg()
{
  printBinOperator(IKWoperators, kwinfeinfeequl, opiieg, IKWComparisonOperator);
}

/**
   Printing a binary tree: opindmot operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opindmot()
{
  recsondecunparse(1);
  this_intePrinter->printspalphakey(kwto);
  recsondecunparse(2);
}

/**
   Printing a binary tree: opinfe operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opinfe()
{
  printBinOperator(IKWoperators, kwinfe, opinfe, IKWComparisonOperator);
}

/**
   Printing a binary tree: opinit operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opinit()
{
  recsonunparse(1);
  this_intePrinter->printspalphakey(kwinit);
  recsondecunparse(2);
}

/**
   Printing a binary tree: opinst operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opinst()
{
  recopinst();
}

/**
   Printing a binary tree: opintvff operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opintvff()
{
  this_intePrinter->printkey(kwcouvpoin);
  recsondecunparse(1);
  this_intePrinter->printkey(kwvirg);
  recsondecunparse(2);
  this_intePrinter->printkey(kwpoincfer);
}

/**
   Printing a binary tree: opintvfo operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opintvfo()
{
  this_intePrinter->printkey(kwcouvpoin);
  recsondecunparse(1);
  this_intePrinter->printkey(kwvirg);
  recsondecunparse(2);
  this_intePrinter->printkey(kwcouvpoin);
}

/**
   Printing a binary tree: opintvof operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opintvof()
{
  this_intePrinter->printkey(kwpoincfer);
  recsondecunparse(1);
  this_intePrinter->printkey(kwvirg);
  recsondecunparse(2);
  this_intePrinter->printkey(kwpoincfer);
}

/**
   Printing a binary tree: opintvoo operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opintvoo()
{
  this_intePrinter->printkey(kwpoincfer);
  recsondecunparse(1);
  this_intePrinter->printkey(kwvirg);
  recsondecunparse(2);
  this_intePrinter->printkey(kwcouvpoin);
}

/**
   Printing a binary tree: opmele operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opmele()
{
  printBinOperator(IKWoperators, tkwundef, opmele, IKWMerge);
}

/**
   Printing a binary tree: opmmul operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opmmul()
{
  printBinOperator(IKWoperators, tkwundef, opmmul, IKWArrayProduct);
}

/**
   Printing a binary tree: opmod operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opmod()
{
  printBinOperator(IKWoperators, kwmodulo, opmod, IKWArithmeticOperator);
}

/**
   Printing children of a module (operande tree)
*/
_VIRTUAL_ void SME_Printer::moduleunparse()
{
  ntrees nomproc, m, save;
  int ikw;
  boolean b;

  save = currentModule;		// push
  SET_NILTREE(nomproc);
  m = treeunparseTree;
  currentModule = m;
  b = FirstHeader;
  snline();
  if (ischildprint(1))
    {
      ikw = beginModel(m, opmodu, operande);
      nomproc = operande;
    }
  else
    this_intePrinter->printerror();
  wline(vholoph - actindent);
  if (ischildprint(2))
    tabunparse(vholoph - 1, smallindent);
  else
    this_intePrinter->printerror();

  // final
  endModel(IKWsme, ikw, nomproc);
  currentModule = save;		// pop
}

/**
   Printing a binary tree: opmodu operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opmodu()
{
  // this_intePrinter->printalphakeysp(kwmodule);
  moduleunparse();
}

/**
   Printing a binary tree: opmoin operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opmoin()
{
  printBinOperator(IKWoperators, kwminu, opmoin, IKWArithmeticOperator);
}

/**
   Printing a binary tree: opmult operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opmult()
{
  printBinOperator(IKWoperators, kwetoi, opmult, IKWArithmeticOperator);
}

/**
   Printing a binary tree: opnomba operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opnomba()
{
  printBinOperator(IKWoperators, tkwundef, opnomba, IKWCounter);
}

/**
   Printing a binary tree: opnombd operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opnombd()
{
  printBinOperator(IKWoperators, tkwundef, opnombd, IKWCounter);
}

/**
   Printing a binary tree: opnombm operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opnombm()
{
  printBinOperator(IKWoperators, tkwundef, opnombm, IKWCounter);
}

/**
   Printing a binary tree: opobli operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opobli()
{
  recunproexp(kwslah);
}

/**
   Printing a binary tree: oppconf operator.
*/
_VIRTUAL_ void SME_Printer::decta2_oppconf()
{
  int ikw;
  boolean b;
  ntrees m;
  noperators opf;

  m = operande;

  opf = noper(nfather(operande));
  b = (opf == opdesc) || (opf == opspec);
  if (!b)
    {				// confined process.
      setCurrentName(IKWConfinedProcess);
      ikw = beginModel(m, oppconf, repstr2ident(currentName));	// a revoir il faut un nom en 3ieme arg.
    }

  if (ischildprint(1))
    sigargprocunp();
  if (ischildprint(2))
    if (ntreearity(operande) != 0)	/*la liste des declarations peut etre vide */
      {
	wline(vholoph - actindent);
	unparse_subtree(operande, vholoph - 1);
      }
  if (!b)
    endModel(IKWundef, ikw, NILTREE);
}

/**
   Printing a binary tree: oppetiq operator.
*/
_VIRTUAL_ void SME_Printer::decta2_oppetiq()
{
  int ikw;
  //  recsondecunparse(1); the label is associated with the subprocess (2th child).
  ikw = beginModel(treeunparseTree, oppetiq, nchild(1, treeunparseTree));
  if (ischildprint(2))
    unparse_subtree(operande, vholoph - 1);
  endModel(IKWundef, ikw, NILTREE);
  // if (!marktalphakeyunparse(kwdpindpin, vholoph - 1, 2)) this_intePrinter->printerror();
}

/**
   Printing a binary tree: opparc operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opparc()
{
  if (ischildprint(1))
    unparse_subtree(operande, vholoph - 1);
  recsondecunparse(2);
}

/**
   Printing a binary tree: opplus operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opplus()
{
  printBinOperator(IKWoperators, kwplus, opplus, IKWArithmeticOperator);
}

/**
   Printing a binary tree: oppuis operator.
*/
_VIRTUAL_ void SME_Printer::decta2_oppuis()
{
  printBinOperator(IKWoperators, kwetoietoi, oppuis, IKWArithmeticOperator);
}

/**
   Printing a binary tree: oprecu operator.
*/
_VIRTUAL_ void SME_Printer::decta2_oprecu()
{
  printBinOperator(IKWoperators, tkwundef, oprecu, IKWArrayRecovery);
}

/**
   Printing a binary tree: oprepe operator.
*/
_VIRTUAL_ void SME_Printer::decta2_oprepe()
{
  printBinOperator(IKWoperators, tkwundef, oprepe, IKWrepetition);	// binalphasigexp(kwvertetoi, 0);
}

/**
   Printing a binary tree: oprest operator.
*/
_VIRTUAL_ void SME_Printer::decta2_oprest()
{
  printBinOperator(IKWoperators, tkwundef, oprest, IKWarrayRestructuration);
}

/**
   Printing a binary tree: opsegal operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opsegal()
{
  printBinOperator(IKWoperators, kwsupeequl, opsegal, IKWComparisonOperator);
}

/**
   Printing a binary tree: opsequ operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opsequ()
{
  printBinOperator(IKWoperators, tkwundef, opsequ, IKWnext);
}

/**
   Printing a binary tree: opspec operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opspec()
{
  int ikw;
  ntrees x;

  // The attributes of the spec are integrated in the header of the model (beginModel()).
  if (ischildprint(2))
    {
      x = operande;
      ikw = beginModel(operande, opspec, NILIDENT);
      operande = x;
      unparse_subtree(x, 9999);
      endModel(IKWundef, ikw, NILTREE);
    }
}

/**
   Printing a binary tree: opsupe operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opsupe()
{
  printBinOperator(IKWoperators, kwsupe, opsupe, IKWComparisonOperator);
}

/**
   Printing a binary tree: optvect operator.
*/
_VIRTUAL_ void SME_Printer::decta2_optvect()
{
  reccoper(1, pk_true);
  recsonunparse(2);
}

/**
   Printing a binary tree: opvenum operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opvenum()
{
  ntrees t, s;

  //  printConstant(opvenum); replaced Nov 2008 by:
  t = treeunparseTree;
  printHeader(IKWLiterals, t, tkwundef, IKWRefEnumerationValue, pk_false);
  if (getchild(t, 1, &s))
    printKwQuotedSignalText(IKWTypeValue, s, pk_true);
  if (getchild(t, 2, &s))
    printKwQuotedSignalText(IKWValue, s, pk_true);
  closeBlockLine();
}

/**
   Printing a binary tree: opvar operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opvar()
{
  ntrees s1, s2, vtree;

  //current expression is : var E1 init E0
  vtree = treeunparseTree;
  printHeaderDef(IKWoperators, vtree, tkwundef, IKWVar, pk_true);
  if (getchild(vtree, 1, &s1))
    opUnparse(s1, vtree, 1, IKWVar);
  if (getchild(vtree, 2, &s2))
    printSignalText(s2, IKWInitialValue);
  printTailDef(IKWoperators);
  printSelectedTree(s1);
}

/**
   Printing a binary tree: opident operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opident()
{
  separate();
  identprint(treeunparseTree);
}

/**
   Printing a binary tree: opleft_tt operator.
*/
_VIRTUAL_ void SME_Printer::decta2_opleft_tt()
{
  this_intePrinter->printalphakey(kwlefttt);
  this_intePrinter->printalphakey(kwpouv);
  binsigexp(kwvirg, 0);
  this_intePrinter->printalphakey(kwpfer);
}

/* ---------- Ternary operators -----------*/

/**
  Printing a ternary tree :  opact operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opact()
{
  setVisibility(IKWpublic);
  processUnparsing();
}

/**
   Printing a ternary tree :  opactp operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opactp()
{
  setVisibility(IKWprivate);
  processUnparsing();
}

/**
   Printing a ternary tree :  opborn operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opborn()
{
  if (ischildprint(1))
    {
      this_intePrinter->printspalphakey(kwin);
      sigargunp();
    }
  if (ischildprint(2))
    {
      this_intePrinter->printspalphakey(kwto);
      sigargunp();
    }
  if (ischildprint(3))
    {
      this_intePrinter->printspalphakey(kwstep);
      sigargunp();
    }
}

/**
   Printing a ternary tree :  opcond operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opcond()
{
  ntrees vtree, p;

  vtree = treeunparseTree;
  printHeaderDef(IKWoperators, vtree, tkwundef, IKWIf, pk_true);
  // if (getchild(vtree, 1, &p)) opUnparse(p, vtree, 1, IKWExtraction); ---> condition
  if (getchild(vtree, 2, &p))
    opUnparse(p, vtree, 2, IKWIf);
  if (getchild(vtree, 3, &p))
    opUnparse(p, vtree, 3, IKWIf);
  printTailDef(IKWoperators);
  printOperands(vtree);
}

/**
   Printing a ternary tree :  opdepd operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opdepd()
{
  ntrees s3;
  if (!getchild(treeunparseTree, 3, &s3))
    SET_NILTREE(s3);
  printDependence(treeunparseTree, s3);
}

/**
   Printing a ternary tree :  openum operator.
*/
_VIRTUAL_ void SME_Printer::decta3_openum()
{
  ntrees vtree;

  vtree = treeunparseTree;
  printHeader(IKWindexes, vtree, tkwundef, IKWIndex, pk_true);
  printHeaderAttribute(IKWfrom, vtree, 1);
  printHeaderAttribute(IKWto, vtree, 2);
  if (noper(nchild(3, vtree)) != opnil)
    printHeaderAttribute(IKWstep, vtree, 3);
  closeBlockLine();
}

/**
   Printing a ternary tree :  opfen operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opfen()
{
  ntrees s1, s3, vtree;

  vtree = treeunparseTree;
  printHeader(IKWoperators, vtree, tkwundef, IKWWindow, pk_true);
  printHeaderAttribute(IKWWindowSize, vtree, 2);
  printCloseHeader();
  //current expression is : E Window E  {init E}
  if (getchild(vtree, 1, &s1))
    opUnparse(s1, vtree, 1, IKWWindow);
  // initial value
  if (getchild(vtree, 3, &s3))
    printSignalText(s3, IKWInitialValue);
  printTailDef(IKWoperators);
  printSelectedTree(s1);
}

/**
   Printing a ternary tree :  opfunc operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opfunc()
{
  setVisibility(IKWpublic);
  processUnparsing();
}

/**
   Printing a ternary tree :  opfuncp operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opfuncp()
{
  setVisibility(IKWprivate);
  processUnparsing();
}

/**
   Printing a ternary tree :  opinte operator. (interface)
*/
_VIRTUAL_ void SME_Printer::decta3_opinte()
{
  ntrees s;

  s = treeunparseTree;
  if (ischildprint(1))
    printInterfaceDeclarations(operande, IKWparameters);	// parameters
  wline(vholoph - actindent);
  treeunparseTree = s;
  if (ischildprint(2))
    {
      treeunparseTree = operande;
      if (ischildprint(1))
	printInterfaceDeclarations(operande, IKWinputs);	// inputs
      if (ischildprint(2))
	printInterfaceDeclarations(operande, IKWoutputs);	// inputs
    }
  treeunparseTree = s;
  wline(vholoph - actindent);
  if (ischildprint(3))
    unparse_subtree(operande, vholoph);	// spec external graph
  treeunparseTree = s;
}

/**
   Printing a ternary tree :  opiter operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opiter()
{
  ntrees vtree, s1, s2, s3;
  int ikw;

  vtree = treeunparseTree;

  setCurrentName(IKWIterate);
  ikw = beginModel(vtree, opiter, repstr2ident(currentName));
  if (getchild(vtree, 2, &s2))
    unparse_subtree(s2, 9999);
  if (getchild(vtree, 3, &s3))
    {
      printHeaderDef(IKWiterationInit, s3, tkwundef, IKWundef, pk_false);
      unparse_subtree(s3, 9999);
      printTailDef(IKWiterationInit);
    }
  // indexes
  if (getchild(vtree, 1, &s1))
    if (STree_isList(s1))
      printOpOperator(s1, vtree, IKWIterate);
    else
      opUnparse(s1, vtree, 1, IKWIterate);
  endModel(IKWundef, ikw, NILTREE);
}

/**
   Printing a ternary tree :  opmemo operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opmemo()
{
  ntrees s1, s2, s3, vtree;

  vtree = treeunparseTree;
  printHeaderDef(IKWoperators, vtree, tkwundef, IKWCell, pk_true);
  //current expression is : E cell C init E0
  if (getchild(vtree, 1, &s1))
    opUnparse(s1, vtree, 1, IKWCell);
  //  if (getchild(vtree, 2, &s2)) opUnparse(s2, vtree, 2, IKWCell);  // --> condition
  if (getchild(vtree, 3, &s3))
    printSignalText(s3, IKWInitialValue);	// initial value
  printTailDef(IKWoperators);

  printSelectedTree(s1);
  if (getchild(vtree, 2, &s2))
    printSelectedTree(s2);
}

/**
   It prints the iterator of an "array" structure.
   Example:
   \verbatim
   <iterator SME:id="_n6WT8e-QEdyUTZkRdBkPqQ" src_DataflowConnections="_egW1kPAbEdycOeCAsJSThA" name="I"/>
   \endverbatim
*/
void SME_Printer::printIteratorArray(ntrees m)
{
  ntrees s;
  if (getchild(m, 1, &s))
    if (getchild(s, 1, &s))
      {
	printHeader(IKWIterator, s, tkwundef, IKWundef, pk_false);
	printKwId(s, IKWundef, IKWname);	// the name
	closeBlockLine();
      }
}

/**
   Printing a ternary tree :  opmoti operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opmoti()
{
  ntrees vtree, s2, s3;
  int ikw;

  vtree = treeunparseTree;

  setCurrentName(IKWBasicIterate);
  ikw = beginModel(vtree, opmoti, repstr2ident(currentName));
  if (getchild(vtree, 2, &s2))
    unparse_subtree(s2, 9999);
  if (getchild(vtree, 3, &s3))
    {
      printHeaderDef(IKWiterationInit, s3, tkwundef, IKWundef, pk_false);
      unparse_subtree(s3, 9999);
      printTailDef(IKWiterationInit);
    }
  printIteratorArray(vtree);
  endModel(IKWundef, ikw, NILTREE);
}

/**
   Printing a ternary tree :  opnode operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opnode()
{
  setVisibility(IKWpublic);
  processUnparsing();
}

/**
   Printing a ternary tree :  opnodep operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opnodep()
{
  setVisibility(IKWprivate);
  processUnparsing();
}

/**
   Printing a ternary tree :  opoper operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opoper()
{
  //  this_intePrinter->printalphakeysp(kwoperator);
  processUnparsing();
}

/**
   Printing a ternary tree :  oppcas operator.
*/
_VIRTUAL_ void SME_Printer::decta3_oppcas()
{
  ntrees vtree, s1;

  // Example : <operators  xsi:type="sme:Switch" name="Switch1" SME:id="363" src_CaseConnections="367 380" signal="364" />
  // the signal is given by the first son of the current tree.
  vtree = treeunparseTree;
  s1 = nchild(1, vtree);
  printHeader(IKWoperators, vtree, tkwundef, IKWSwitch, pk_true);
  getDeclarationAttribute(s1, &s1);
  printKwSMEId(s1, IKWundef, IKWsignal);
  closeBlockLine();
  // the list of cases.
  if (ischildprint(2))
    markunparse(vholoph - 1);
  else
    this_intePrinter->printerror();
  // else part.
  if (ischildprint(3))
    printSubProcess(operande);
}

/**
   Printing a ternary tree :  opprag operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opprag()
{
  /*  recsondecunparse(1);
     if (ischildprint(2))
     {
     snspace();
     unparse_subtree(operande, vholoph - 1);
     }
     if (ischildprint(3))
     {
     snspace();
     markunparse(vholoph - 1);
     }
   */
}

/**
   Printing a ternary tree :  opprocess operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opprocess()
{
  setVisibility(IKWpublic);
  processUnparsing();
}

/**
   Printing a ternary tree :  opprocessp operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opprocessp()
{
  setVisibility(IKWprivate);
  processUnparsing();
}

boolean SME_Printer::isTreeOne(ntrees t)
{
  noperators op;
  ptrepstr vstr;

  op = noper(t);
  if (op == opretun)
    return pk_true;
  if (op != opcstent)
    return pk_false;
  vstr = nstringof(t);
  return (repstr2integer(vstr) == 1);
}

/**
   Printing a ternary tree :  opret operator.
*/
_VIRTUAL_ void SME_Printer::decta3_opret()
{
  ntrees s1, s2, s3, vtree;

  //current expression is : s1 $ s2 init s3
  vtree = treeunparseTree;

  printHeader(IKWoperators, vtree, tkwundef, IKWDelay, pk_true);
  // A revoir : flag nbInstants="expression"
  // en fait ceci est faux : on peut avoir une expression sur signaux.
  if (getchild(vtree, 2, &s2))
    if (!isTreeOne(s2))
      printHeaderAttribute(IKWNbInstants, vtree, 2);
  printCloseHeader();
  if (getchild(vtree, 1, &s1))
    opUnparse(s1, vtree, 1, IKWDelay);
  // initial value
  if (getchild(vtree, 3, &s3))
    printSignalText(s3, IKWInitialValue);
  printTailDef(IKWoperators);

  printSelectedTree(s1);
}

void SME_Printer::printQuotedKey(int kw)
{
  printQuotedStr(getSMEKeyword(kw));
}

void SME_Printer::printFlagTypeBundle()
{
  // kind="bundle"
  snspace();
  sigpristr(getikwFlag(kindFlag));
  sncharpri('=');
  printQuotedKey(IKWbundle);
}

/**
   It generate the assignment to the value "true" for the aFlag-th of the table of flag.
*/
void SME_Printer::printFlagtrue(int aFlag)
{
  snspace();
  sigpristr(getikwFlag(aFlag));
  sncharpri('=');
  printQuotedStr(getikwFlag(trueFlag));
}

/**
   Printing a ternary tree :  optype operator.
*/
_VIRTUAL_ void SME_Printer::decta3_optype()
{
  ntrees s1, s2, s3, vtree;
  int xsiType;
  noperators op;
  boolean b;

  // type <identificateur> "=" <description_type> "init" <expr_signal> ;
  // (<identificateur>, <description_type>, <expr_signal>)
  vtree = treeunparseTree;
  s1 = nchild(1, vtree);	// name
  if (getchild(vtree, 2, &s2))
    {
      op = noper(s2);
      if (op == optenum)
	xsiType = IKWEnumerationType;
      else if (op == opstru)
	xsiType = IKWTupleType;
      else if (op == opbund)
	xsiType = IKWTupleType;
      else
	xsiType = IKWSubstitutionType;
    }
  else
    xsiType = IKWundef;

  printHeader(IKWtypes, s1, tkwundef, xsiType, pk_false);
  printKwId(s1, IKWundef, IKWname);	// the name

  printVisibilityAttribute();

  b = pk_false;
  if (op == opexte)
    printFlagtrue(externalFlag);
  else
    {
      if (op == opbund)
	printFlagTypeBundle();
      if ((op == opstru) || (op == opbund) || (op == optenum))
	{
	  sncharpri('>');
	  printSelectedTree(s2);
	  b = pk_true;
	}
      else
	{
	  sncharpri(' ');
	  if (printType(s2))
	    {
	      sncharpri('>');
	      printDimType(s2);
	      b = pk_true;
	    }
	}
    }
  if (getchild(vtree, 3, &s3)) ;	// not used.
  if (b)
    printTailDef(IKWtypes);
  else
    closeBlockLine();
}

/**
   Printing a ternary tree :  optypeact operator.
*/
_VIRTUAL_ void SME_Printer::decta3_optypeact()
{
  this_intePrinter->printalphakeysp(kwaction);
  processtypeunp();
}

/**
   Printing a ternary tree :  optypefunc operator.
*/
_VIRTUAL_ void SME_Printer::decta3_optypefunc()
{
  this_intePrinter->printalphakeysp(kwfunction);
  processtypeunp();
}

/**
   Printing a ternary tree :  optypenode operator.
*/
_VIRTUAL_ void SME_Printer::decta3_optypenode()
{
  this_intePrinter->printalphakeysp(kwnode);
  processtypeunp();
}

/**
   Printing a ternary tree :  optypeproc operator.
*/
_VIRTUAL_ void SME_Printer::decta3_optypeproc()
{
  this_intePrinter->printalphakeysp(kwprocess);
  processtypeunp();
}

/* ---------- Nary operators (lists) -----------*/

/**
   Printing a list of composition with seplst as separator.
*/
_VIRTUAL_ void SME_Printer::printClockRelation(int kwclockop)
{
  ntrees vtree;
  int IKW;

  if (assertContext)
    IKW = IKWAssertionClockConstraintOperator;
  else
    IKW = IKWClockConstraintOperator;

  vtree = treeunparseTree;
  printHeaderDef(IKWConstraints, vtree, kwclockop, IKW, pk_true);
  //  printOpOperator(vtree, IKWClockConstraintOperator);
  printTailDef(IKWConstraints);
  printOperands(vtree);
}

/**
   Printing a list tree: opcomments operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opcomments()
{
  decl0(0, -1);
}

/**
   Printing a list tree: opchai operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opchai()
{
  decl0(0, -1);
}

/**
   Printing a list tree: oplattr operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplattr()
{
  //  decl0(0, -2);
}

/**
   Printing a list tree: oplcas operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplcas()
{
  decl0(0, -1);
}

/**
   Printing a list tree: oplcons operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplcons()
{
  decl0(kwvirg, 1);
}

/**
   Printing a list tree: oplcrohor operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplcrohor()
{
  int IKW;
  if (assertContext)
    IKW = IKWAssertionClockSpeedOperator;
  else
    IKW = IKWClockSpeedOperator;
  printBinOperator(IKWConstraints, kwchapinfe, oplcrohor, IKW);
}

/**
   Printing a list tree: opldechor operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opldechor()
{
  int IKW;
  if (assertContext)
    IKW = IKWAssertionClockSpeedOperator;
  else
    IKW = IKWClockSpeedOperator;
  printBinOperator(IKWConstraints, kwchapsupe, opldechor, IKW);
}

/**
   Printing a list tree: opldecs operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opldecs()
{
  decl0(0, -1);
}

/**
   Printing a list tree: opldecl operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opldecl()
{
  decl0(0, -1);
}

void SME_Printer::printDependence(ntrees vtree, ntrees condition)
{
  printHeaderDef(IKWConstraints, vtree, tkwundef, IKWDependenceOperator, pk_true);
  printTailDef(IKWConstraints);
  if (ISNOT_NILTREE(condition))
    printSelectedTree(condition);
}

/**
   Printing a list tree: opldep operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opldep()
{
  printDependence(treeunparseTree, NILTREE);
}

/**
   Printing a list tree: opleflot operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opleflot()
{
  printBinOperator(IKWoperators, tkwundef, opleflot, IKWTupleOperator);
  //  this_intePrinter->printgluekeymark(kwpouv, vholoph - actindent);
  //  decl0(kwvirg, 0);
  //  this_intePrinter->printgluekeyrelease(kwpfer, vholoph - actindent);
}

/**
   Printing a list tree: oplepara operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplepara()
{
  decl0(kwvirg, 0);
}

/**
   Printing a list tree: oplesign operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplesign()
{
  this_intePrinter->printgluekeymark(kwpouv, vholoph - actindent);
  decl0(kwvirg, 0);
  this_intePrinter->printgluekeymark(kwpouv, vholoph - actindent);
}

/**
   Printing a list tree: oplevect operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplevect()
{
  printBinOperator(IKWoperators, tkwundef, oplelevec, IKWarrayEnumeration);
  /* this_intePrinter->printgluekeymark(kwcouv, vholoph - actindent);
     decl0(kwvirg, 0);
     this_intePrinter->printgluekeyrelease(kwcfer, vholoph - actindent);
   */
}

/**
   Printing a list tree: oplelevec operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplelevec()
{

  this_intePrinter->printgluekeymark(kwcouv, vholoph - actindent);
  decl0(kwvirg, 0);
  this_intePrinter->printgluekeyrelease(kwcfer, vholoph - actindent);

}

/**
   Printing a list tree: oplequhor operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplequhor()
{
  printClockRelation(kwchapequl);
}

/**
   Printing a list tree: oplexchor operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplexchor()
{
  printClockRelation(kwchapdies);
}

boolean SME_Printer::needSubprocess(ntrees p)
{
  int i, n;
  ntrees t;

  if (noper(p) != oplfeed)
    return pk_false;
  n = ntreearity(p);
  for (i = 1; i <= n; i++)
    if (getchild(p, i, &t))
      if (!selectedOpUnparse(&t))
	n--;
      else if (noper(t) == opdefsig)
	{
	  t = nchild(2, t);
	  if (noper(t) == opevnt)
	    if (getchild(t, 1, &t))
	      if (!selectedOpUnparse(&t))
		n--;
	}

  return (n > 1);
}

/**
   Printing a list of composition with seplst as separator.
*/
_VIRTUAL_ void SME_Printer::declfeed()
{
  int ikw, i;
  ntrees operand;
  boolean sp;

  for (i = 1; i <= ntreearity(treeunparseTree); i++)
    if (getchild(treeunparseTree, i, &operand))
      {
	if (needSubprocess(operand))
	  {
	    sp = pk_true;
	    setCurrentName(IKWConfinedProcess);
	    ikw = beginModel(operand, oppconf, repstr2ident(currentName));
	  }
	else
	  sp = pk_false;
	unparse_subtree(operand, 9999);
	snline();
	if (sp)
	  endModel(IKWundef, ikw, NILTREE);
      }
}

/**
   Printing a list tree: oplfeed operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplfeed()
{
  declfeed();
}

/**
   Printing a list tree: oplident operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplident()
{
  decl0(kwvirg, 1);
}

/**
   Printing a list tree: oplimpo operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplimpo()
{
  ntrees vdecl, m, v, t;
  int i;

  /* this_intePrinter->printspalphakey(kwuse); decl0(kwvirg, 1); this_intePrinter->printkeysp(kwpvir);  */
  /*
     <useModule SME:id="_ZB7n8OuPEdyxR_OpGHu7Fg">
     <modules href="ModuleEx.sme#_EKwisOn4EdyMlu3osg99Ig"/>
     </useModule>
   */
  m = treeunparseTree;
  printBeginBlock(IKWuse, IKWundef, pk_false);
  printSMEId(m);
  printCloseHeader();
  for (i = 1; i <= ntreearity(m); i++)
    {
      t = nchild(i, m);
      if (getDeclarationAttribute(t, &vdecl))
	{
	  printBeginBlock(IKWmodules);
	  printSMEId(t);
	  if (!getSMEId(vdecl, &v))
	    v = vdecl;
	  printKwKwSMEId(IKWundef, IKWhref, vdecl, v);
	  closeBlockLine();
	}
    }
  printTailDef(IKWuse);
}

/**
   Printing a list tree: oplmodu operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplmodu()
{
  decl0(0, -1);
}

/**
   Printing a list tree: oplopra operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplopra()
{
  this_intePrinter->printgluekeymark(kwbouv, vholoph - actindent);
  decl0(kwvirg, 0);
  this_intePrinter->printgluekeyrelease(kwbfer, vholoph - actindent);
}

/**
   Printing a list tree: oplparc operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplparc()
{
  decl0(kwvirg, 1);
}

/**
   Printing a list of pragmas (operande) with  an holophrase if holoph=0.
*/
_VIRTUAL_ void SME_Printer::decpragmas(ntrees p)
{
  int j, i, n;
  ntrees name, s, v, p_obj, vdecl, aPragma;

  if (!STree_isUndefinedTree(p))
    {
      n = ntreearity(p);
      for (i = 1; i <= n; i++)
	{
	  aPragma = nchild(i, p);
	  if (!nattrexist(aPragma, SME_AlreadyPrintedATTRIBUTENUMBER))
	    {
	      name = nchild(1, aPragma);
	      p_obj = nchild(2, aPragma);
	      printHeader(IKWpragmas, aPragma, tkwundef, tkwundef, pk_true);
	      printKwId(name, IKWundef, IKWpragma);
	      // printNameValue(name);
	      if (!STree_isUndefinedTree(p_obj))
		{
		  snspace();
		  printSMEKwEq(IKWtargets);
		  beginString();
		  for (j = 1; j <= ntreearity(p_obj); j++)
		    {
		      s = nchild(j, p_obj);
		      if (getDeclarationAttribute(s, &vdecl))
			if (getSMEId(vdecl, &v))
			  printSMEIdValue(v);
			else
			  this_intePrinter->printerror();
		      else
			this_intePrinter->printerror();
		      sncharpri(' ');
		    }
		  endString();
		}
	      printStmtPragma(IKWexpression, nchild(3, aPragma));
	      closeBlockLine();
	    }
	}
    }
}

/**
   Printing a list tree: oplprag operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplprag()
{
  // decpragmas(vholoph);
}

/**
   Printing a list tree: oplsign operator. 
*/
_VIRTUAL_ void SME_Printer::declist_oplsign()
{
  decl0(kwvirg, 1);
}

/**
   Printing a list tree: opltype operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opltype()
{
  setVisibility(IKWpublic);
  decl0(kwvirg, 1);
}

/**
   Printing a list tree: opltypep operator. 
*/
_VIRTUAL_ void SME_Printer::declist_opltypep()
{
  setVisibility(IKWprivate);
  decl0(kwvirg, 1);
}

/** Printing a list tree using a separator (seplst) to separate each element and a glue parameter (coller) to shape of the text.
    A final separator is added at the end of the list.
 */
_VIRTUAL_ void SME_Printer::decl1(int seplst, int coller)
{
  decl0(seplst, coller);
  // this_intePrinter->printkey(seplst);  
}

/**
   Printing a list tree: optenum operator. 
*/
_VIRTUAL_ void SME_Printer::declist_optenum()
{
  printList(treeunparseTree, IKWEnumerationValue);
}

void SME_Printer::printList(ntrees d, int kw)
{
  int i;

  snline();
  for (i = 1; i <= ntreearity(d); i++)
    printSignalText(nchild(i, d), kw);
}
