/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
*/
    SME_SyntaxRestriction::SME_SyntaxRestriction()
{
  GUIContext = pk_false;
  idNil = string2ident("_XMI_Id");
  cptIds = 0;
  SET_NILTREE(lids);
  SET_NILTREE(lfeed);
  Ok = pk_true;
}

/**
   Destructor.
*/
SME_SyntaxRestriction::~SME_SyntaxRestriction()
{

}

/**
   Specialisation of kernel::cExpCall().
   It calls  kernel::cExpCall() then mkXmiCallTree().
*/
_VIRTUAL_ boolean SME_SyntaxRestriction::cExpCall(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  boolean bResult;

  bResult = Trees2Kernel::cExpCall(p, inputs, outputs, indexList, lparam, vklExprProc);
  if (bResult)
    mkXmiCallTree(p, *vklExprProc);
  return bResult;
}

/**
   Specialisation of kernel::cExpRefToModel().
   It calls  kernel::cExpCall() then mkXmiCallTree().
*/
_VIRTUAL_ boolean SME_SyntaxRestriction::cExpRefToModel(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean bResult;
  bResult = Trees2Kernel::cExpRefToModel(p, inputs, outputs, vklExprProc);
  if (bResult)
    mkXmiCallTree(p, *vklExprProc);
  return bResult;
}

/**
   operators :
   - oplcrohor: // list of "^<"
   - opldechor: // list of "^>"
   -  case oplequhor: // list of ^=
   -  case oplexchor: // list of "^#"
   The standard expressions are not allowed (SME aspects), so new identifiers are introduced.
   \see kl_mkClockOp(), xmiSyntaxClockExpression()

   \verbatim
   Example:
   a+b ^= x default z 
   is 
   replaced by
   (| x1 := a+b | x2 := x default z | x1 ^= x2 |) where x1,x2; end
   \endverbatim
   
*/
_VIRTUAL_ boolean SME_SyntaxRestriction::cExpSynchro(ntrees p, noperators vop, ptlrepid * inputs, ptlrepid * outputs, ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  ptklExprSig vklExprSig;

  vklExprSig = Trees2Kernel::cExprSig(p, inputs, lparam, indexList);
  if (vklExprSig == NILEXPRSIG)
    return pk_false;
  *vklExprProc = kl_mkClockOp(*inputs, p, vklExprSig, vop);
  *outputs = NILLREPID;

  xmiSyntaxClockExpression(p);
  return pk_true;
}

/**
   Specialisation of kernel::cExprSig().
   It calls  kernel::cExprSig() then xmiSyntaxExpression().
   The expression is modified when a sub expression is not allowed in a context (SME aspects).
*/
_VIRTUAL_ ptklExprSig SME_SyntaxRestriction::cExprSig(ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vklExprSig;

  if (IS_NILTREE(p))
    return NILEXPRSIG;
  vklExprSig = Trees2Kernel::cExprSig(p, les, lparam, indexList);
  xmiSyntaxExpression(p);
  return vklExprSig;
}

/**
   @return the status of the translation:
   - "true"  the translation succeeds
   - "false" the translation fails
 */
boolean SME_SyntaxRestriction::getStatusCode()
{
  return Ok;
}
