/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define isClockRelationOperator_local(op) (op == oplequhor) || (op == oplexchor)
#define isClockSpeedRelationOperator(op) (op == oplcrohor) || (op == opldechor)

#define PushSMEMode() boolean save_modeSME = modeSME;
#define PopSMEMode() modeSME = save_modeSME;

#define PushContext() PushSMEMode() \
 ntrees save_currentBlock = currentBlock;  \
 int scptName = cptName; \
 ptTypeRefTable save_RefTable= RefTable;

#define PopContext() PopSMEMode() \
  currentBlock = save_currentBlock ; \
  cptName = scptName; \
  RefTable = save_RefTable;

_PROTECTED_
/**
   When an operator is not yet implemented, the translator will return the "false" value.
   Currently, only some "manipulations of arrays" are not implemented. They can be rewritten handmade.
 */
    void
 SME_Translator::notOK(ntrees e)
{
  Ok = pk_false;
  Trees2Kernel::notYetImplemented(e);
}

/**
   Specialisation of Trees2Kernel::notYetImplemented().
   The construction (p) is not yet implemented, but it is accepted by the SME translator.
*/
_VIRTUAL_ void SME_Translator::notYetImplemented(ntrees p)
{
  //  mkklErrorAnnotTree(errnotimple, p);
}

/**
   @return
   - "true" if the list of the connections (lconnections) of the current global model is not empty.
   - "false" otherwise; the parameter lconnections is not modified.
*/
boolean SME_Translator::getModelConnections(ntrees * lconnections)
{
  if (!nattrexist(currentBlock, SME_ConnectionATTRIBUTENUMBER))
    return pk_false;
  (*lconnections) = nattrntrees(currentBlock, SME_ConnectionATTRIBUTENUMBER);
  return pk_true;
}

/**
   @return "true" if it exists a declaration associated with a tree (t), "false" otherwise.
   The declaration is not necessary a syntactic Signal declaration. 
   For example, for a label expression e:: EXPR , the associated declaration with the label is EXPR (a label is not declared in the sme model).
   Moreover, it may have several indirections (lab --> declaration of the label --> sub-process prefixed by the label).
 */
boolean SME_Translator::getDeclarationAttribute(ntrees t, ntrees * vdecl)
{
  noperators op;
  if (!nattrexist(t, SME_DeclATTRIBUTENUMBER))
    return pk_false;
  (*vdecl) = (ntrees) nattrptr(t, SME_DeclATTRIBUTENUMBER);
  op = noper(*vdecl);
  // Reference to an object of a module ?
  if ((op == opSMEImportType) || (op == opSMEImportCons))
    return pk_true;

  while (nattrexist(*vdecl, SME_DeclATTRIBUTENUMBER))
    (*vdecl) = (ntrees) nattrptr(*vdecl, SME_DeclATTRIBUTENUMBER);
  return pk_true;
}

/**
   It adds a flag to a tree (vtree). A flag f is a binary-tree with opSMEtag as operator. 
   It is assigned as an attribute (SME_FlagATTRIBUTENUMBER) to the tree.
   The first operand of f is an int value (iFlag), and the second is the associated value (valueFlag).
   A tree can be characterized by a list of flags (a list tree with opSMEList as operator).
   The operators opSMEList and opSMEtag, SME_FlagATTRIBUTENUMBER are declared in the file SMEDefs.h.
*/
void SME_Translator::addFlag(ntrees vtree, int iFlag, ntrees valueFlag)
{
  ntrees attr;

  if (!nattrexist(vtree, SME_FlagATTRIBUTENUMBER))
    {
      attr = LL_make_SMEList_emptyList();
      nmakeattrntrees(vtree, SME_FlagATTRIBUTENUMBER, attr);
    }
  else
    attr = nattrntrees(vtree, SME_FlagATTRIBUTENUMBER);
  attr = ntreeList_appendLast(attr, nmakebinary(opSMEtag, natomint(opSMEvalue, iFlag), valueFlag));
}

/**
   It adds a module (vmodule) in the list of the referenced modules of the current compiled program (topModel global variable).
   The module is added to the list of the referenced module managed as an attribute SME_ReferencedModuleATTRIBUTENUMBER associated with the topModel.
   This list can be used for the printer to produce only the sme code of the referenced modules.
   SME_ReferencedModuleATTRIBUTENUMBER is declared in the file SMEDefs.h.
*/
void SME_Translator::addImportModule(ntrees vmodule)
{
  ptlPointer l;

  if ((ISNOT_NILTREE(vmodule)) && (vmodule != currentModule))
    {
      if (nattrexist(topModel, SME_ReferencedModuleATTRIBUTENUMBER))
	l = (ptlPointer) nattrptr(topModel, SME_ReferencedModuleATTRIBUTENUMBER);
      else
	l = NILLPOINTER;
      ptlPointer_insert_elem(&l, (pointer) vmodule);
      nmakeattrptr(topModel, SME_ReferencedModuleATTRIBUTENUMBER, (pointer) l);
    }
}

/**
   It links a declaration (vdecl) to a module (xmodule). This link is an attribute, number SME_ModuleATTRIBUTENUMBER,
   associated with the declaration.
   It is done only if the module (xmodule) is not the current compiled module (it is not an imported object).
   @return "true" if the links is set, "false" otherwise.
*/
boolean SME_Translator::setModuleAttribute(ntrees vdecl, ntrees xmodule)
{
  if ((ISNOT_NILTREE(vdecl)) && (ISNOT_NILTREE(xmodule)) && (xmodule != currentModule))
    {
      nmakeattrptr(vdecl, SME_ModuleATTRIBUTENUMBER, (pointer) xmodule);
      addImportModule(xmodule);
      return pk_true;
    }
  return pk_false;
}

/**
   It links an object (t) to its declaration (tdecl), declared in a module (vmodule). 
   The link (t, tdecl) is an attribute, number SME_DeclATTRIBUTENUMBER, associated with t.
   The link (tdecl, vmodule) is set by the method setModuleAttribute().
*/
boolean SME_Translator::setDeclarationAttribute(ntrees t, ntrees tdecl, ntrees vmodule)
{
  if ((ISNOT_NILTREE(t)) && (ISNOT_NILTREE(tdecl)))
    {
      nmakeattrptr(t, SME_DeclATTRIBUTENUMBER, (pointer) tdecl);
      return setModuleAttribute(tdecl, vmodule);
    }
  return pk_false;
}

/**
   @return a new identifier built from the predefined name (addedNameStr :member of the SME_Translator class) at which is added the current
   value of an internal counter (cptName : member of the SME_Translator class).   

   \remark {addeNameStr must be set before the call to this method.}
*/
ntrees SME_Translator::getNewName()
{
  ptrepstr v;

  cptName++;
  if (makeRepstr_repstr_int(addedNameStr, cptName, &v))
    return repstr2ident(v);
  return NILIDENT;
}

/**
   @return a new SME identifier, given in our implementation, by a counter (cpt : member of the class).
   Such identifier will appear as SME:id="89" in the produced sme text, in the definition of the connections,...
*/
ntrees SME_Translator::newSMEId()
{
  cpt++;
  return natomstring(opSMEid, integer2repstr(cpt));
}

/**
   It creates a new SME identifier and it assigns it with a tree (vc).
   When b is "true", SME_idATTRIBUTENUMBER, SME_idoperandATTRIBUTENUMBER otherwise. xxxxxxxxxxx A COMPLETER xxxxxxxxx.
   \see newSMEId()
*/
ntrees SME_Translator::mkSMEId(ntrees vc, boolean b)
{
  ntrees x;

  x = newSMEId();
  if (b)
    nmakeattrntrees(vc, SME_idATTRIBUTENUMBER, x);
  else
    nmakeattrntrees(vc, SME_idoperandATTRIBUTENUMBER, x);
  return x;
}

/**
   @return the SME identifier associated with a tree (vc).
   When it does not exist, it is created.
   \see mkSMEId().
*/
ntrees SME_Translator::setSMEId(ntrees vc)
{
  if (nattrexist(vc, SME_idATTRIBUTENUMBER))
    return nattrntrees(vc, SME_idATTRIBUTENUMBER);
  return mkSMEId(vc, pk_true);
}

/**
   @return a copy of the SME identifier associated with a tree (vc).
   When it does not exist, it is created.
   \see setSMEId().
*/
ntrees SME_Translator::getSMEId(ntrees vc)
{
  return STree_scopy(setSMEId(vc));
}

/**
   @return 
   - "true" if it exists a tree (vattr) with (op) as operator in a set (vdef).
   - "false" otherwise, vattr is not modified.
   It is used to manage the references between objects. For example, if an object is the source of some data connections these
   informations will be stored in a attribute (element of vdef here) printed in sme \n
         src_DataflowConnections="221 256 345" \n
   The information is managed as a tree with op=opsrc_dfc as operator, it is a list tree composed of a copy of the SME identifiers (221, 256, 345).
*/
boolean SME_Translator::getAttributeWithOp(ntrees vdef, int op, ntrees * vattr)
{
  boolean found;
  int i;
  ntrees t;

  found = pk_false;
  for (i = 1; !found && (i <= ntreearity(vdef)); i++)
    {
      t = nchild(i, vdef);
      if (noper(t) == op)
	{
	  found = pk_true;
	  *vattr = t;
	}
    }
  return found;
}

/** 
    It adds a reference from n1 to n2. The reference is specialized according to an operator (vop) and a boolean (bsrc):
     - vop : an element of { opDataflowConnection, opConditionConnection, opClockRelationConnection, opConstantAffectation, opDependenceConnection,
    opCaseConnection}
     - bsrc : pk_true for the "source", pk_false for the addressee.
     It calls addAttributeWithOp()
*/
_VIRTUAL_ void SME_Translator::linkObject(ntrees n1, ntrees n2, noperators vop, boolean bsrc)
{
  noperators op;

  if (vop == opDataflowConnection)
    if (bsrc)
      op = opsrc_dfc;
    else
      op = opdst_dfc;
  else if (vop == opConditionConnection)
    if (bsrc)
      op = opsrc_condc;
    else
      op = opdst_condc;
  else if (vop == opClockRelationConnection)
    if (bsrc)
      op = opsrc_clockc;
    else
      op = opdst_clockc;
  else if (vop == opConstantAffectation)
    if (bsrc)
      op = opsrc_cstaff;
    else
      op = opdst_cstaff;
  else if (vop == opDependenceConnection)
    if (bsrc)
      op = opsrc_dependences;
    else
      op = opdst_dependences;
  else if (vop == opCaseConnection)
    if (bsrc)
      op = opsrc_case;
    else
      op = opdst_case;

  addAttributeWithOp(n1, n2, op);
}

/** 
    It adds a reference from n1 to n2. The reference is specialized according to an operator (vop)
    - vop : an element of { opsrc_dfc, opdst_dfc, opsrc_condc, opdst_condc, opsrc_clockc, opdst_clockc,
    opsrc_cstaff, opdst_cstaff,  opsrc_dependences,  opdst_dependences, opsrc_case, opdst_case}.
    The codification is:
    - opsrc_xxx stands for the source of xxx, 
    - opdst_xxx stands for the addressee of xxx.
    .
    with xxx in
    - dfc : data flow connection
    - condc : condition connection
    - clockc : clock connection
    - cstaff : constant affectation
    - dependences : dependence
    - case : case connection.
*/
void SME_Translator::addAttributeWithOp(ntrees n1, ntrees n2, int op)
{
  boolean done;
  ntrees vdef, t;

  done = pk_false;
  if (!nattrexist(n1, SME_ATTRIBUTENUMBER))
    {
      vdef = LL_make_SMEdef_emptyList();
      nmakeattrntrees(n1, SME_ATTRIBUTENUMBER, vdef);
    }
  else
    {
      vdef = nattrntrees(n1, SME_ATTRIBUTENUMBER);
      if (getAttributeWithOp(vdef, op, &t))
	{
	  t = ntreeList_appendLast(t, n2);
	  done = pk_true;
	}
    }
  if (!done)
    vdef = ntreeList_appendLast(vdef, nmakelist(op, n2));
}

/**
   @return a tree defining a connection vop(from, to) built from
   - the SME identifier of from, 
   - the SME identifier of from when b, or a copy of to otherwise (in this case, to is already a SME identifier.
   - vop gives the class of the connection (dataflowconnection, condition connection, ...).
   The connection is added to the current model (currentBlock attribute member): the list of connections is managed via the attribute
   number SME_ConnectionATTRIBUTENUMBER.

   These informations will produce the lines such that
   \verbatim
   <connections  xsi:type="sme:DataflowConnection" name="DataflowConnection1101" SME:id="44" src="38" dst="42"/>
   <connections  xsi:type="sme:DataflowConnection" name="DataflowConnection1102" SME:id="46" src="47" dst="41"/>
   <connections  xsi:type="sme:Condition" name="Condition238" SME:id="49" src="43" dst="45"/>
   <connections  xsi:type="sme:DataflowConnection" name="DataflowConnection1103" SME:id="52" src="37" dst="50"/>
   <connections  xsi:type="sme:DataflowConnection" name="DataflowConnection1104" SME:id="55" src="45" dst="54"/>
   <connections  xsi:type="sme:Condition" name="Condition239" SME:id="57" src="51" dst="53"/>

   \endverbatim
   in a sme file.
*/
ntrees SME_Translator::mkConnection(ntrees from, ntrees to, noperators vop, boolean b)
{
  ntrees vc;
  ntrees SMEId, SMEIdFrom, SMEIdTo;

  // adding the connection to the model.
  vc = nmakemptylist(vop);
  // defining the connection.
  SMEId = getSMEId(vc);
  SMEIdFrom = getSMEId(from);
  if (b)
    SMEIdTo = getSMEId(to);
  else
    SMEIdTo = STree_scopy(to);	// to is already an attibute
  vc = ntreeList_appendLast(vc, SMEId);
  vc = ntreeList_appendLast(vc, SMEIdFrom);
  vc = ntreeList_appendLast(vc, SMEIdTo);
  ntreeList_appendLast(nattrntrees(currentBlock, SME_ConnectionATTRIBUTENUMBER), vc);
  return vc;
}

/**
   It adds a connection vop(from, to) parametrized by b : see the method mkConnection().
   And it adds
   - the link between from and this connection (repaired by its SME identifier)
   - the link between to and this connection (repaired by its SME identifier)
   \verbatim
   <connections  xsi:type="sme:DataflowConnection" name="DataflowConnection1101" SME:id="44" src="38" dst="42"/>
   <locals  SME:id="38" src_DataflowConnections="44" ... name="full"/>
   <operators  xsi:type="sme:Not" name="Not90" SME:id="43" src_Conditions="49">
   <in SME:id="42" dst_DataflowConnections="44" operatorParent="43" name="Not90_in1"/>
   </operators>
   \endverbatim
*/
ntrees SME_Translator::addConnection1(ntrees from, ntrees to, noperators vop, boolean b)
{
  ntrees vc;
  ntrees SMEId;

  vc = mkConnection(from, to, vop, b);
  SMEId = nchild(1, vc);
  // link the from,to objects with the connection
  linkObject(from, STree_scopy(SMEId), vop, pk_true);
  linkObject(to, STree_scopy(SMEId), vop, pk_false);
  return vc;
}

/**
   @return the tree of the operand tree (from) used for the connection.
   Because, the meta model does not defined all the operator of the Signal language (Grrr!!!)
   For example : for the expression x[I,J,K], x is used for the connection, the expression "[I,J,K]" will be an attribute of the connection.
*/
ntrees SME_Translator::getTreeForConnection(ntrees from)
{
  noperators opfrom;

  opfrom = noper(from);

  if (opfrom == opvvarin)
    from = nchild(1, from);
  else if (opfrom == opevnt)
    from = nchild(1, from);
  else if (opfrom == opconv)
    from = nchild(2, from);
  else if ((opfrom == opcham) || (opfrom == opelem) /*|| (opfrom  == oprecu) */  || (opfrom == opsdec))
    from = STree_getIdentOf(from);
  return from;
}

/** 
    @return a tree representing the connection vop(X, Y) or NILTREE.
    - NILTREE when the operator(to) is a reference to a model (call, cartesian product):  particular cases induced by the meta model (Grrr!!!).
    - the connection is defined otherwise:
      it takes into account the getTreeForConnection() of  the operand (from, to).
      - case 1: 
         If 'from' is associated with a declaration attribute (getDeclarationAttribute()) called 'Dfrom', a connection 
	 is added (addConnection1()) from 'Dfrom' to 'Y', where 'Y' is
         - the declaration of 'to' when it is also characterized by a declaration attribute
	 - 'from' otherwise.
      - case 2:
       if 'to' is associated with a declaration attribute (getDeclarationAttribute()) called 'Dto', a connection 
      is added (addConnection1()) from 'from' to 'Dto'.
      - case 3: a connection vop(from, Y) is added to the model, where Y denotes the operand attibute associated with from.

      \verbatim  
      Examples:
       case 1: x:=y from="y", to="x"
       case 2  x:= a+b with from ="a+b", to ="x"
       case 3: (a+b) +  c :from= "a+b", to="(a+b) + c"
       \endverbatim
    \see addConnection1().

    Moreover, 
    - the links from the connection to the abstract compiled tree is kept: it is done using an attribute number SME_BackAnnotationATTRIBUTENUMBER.
    - flags are added some particular cases (from operand). The connection is marked by
          - useSrcClockFlag for an event
	  - castTypeFlag for a conversion
	  - arrayRecoveryFlag for the exception operator E\\E1.
*/
_VIRTUAL_ ntrees SME_Translator::addConnection(ntrees to, ntrees from, noperators vop)
{
  ntrees sFrom, sTo, x1, x2, vc;
  noperators opto, opfrom;

  // filters for particular cases (Grrr!!!)
  opto = noper(to);
  if (opto == opappe)
    return NILTREE;
  if (opto == oppcar)
    return NILTREE;
  if (opto == oplopra)
    return NILTREE;		// For a pragma, the objects are not connected with the declaration.

  sTo = to;
  if ((opto == opcham) || (opto == opelem))
    to = STree_getIdentOf(to);

  sFrom = from;
  opfrom = noper(from);
  from = getTreeForConnection(from);
  // ----------
  if (getDeclarationAttribute(from, &x1))
    {
      if (!getDeclarationAttribute(to, &x2))
	x2 = from;
      vc = addConnection1(x1, x2, vop, pk_true);
    }
  else if (getDeclarationAttribute(to, &x2))
    vc = addConnection1(from, x2, vop, pk_true);
  else
    {
      if (nattrexist(from, SME_idoperandATTRIBUTENUMBER))
	x1 = nattrntrees(from, SME_idoperandATTRIBUTENUMBER);
      else
	x1 = mkSMEId(from, pk_false);
      vc = addConnection1(from, x1, vop, pk_false);
    }

  // For all, we keep the link with the tree that produce the connection.
  nmakeattrptr(nchild(2, vc), SME_BackAnnotationATTRIBUTENUMBER, (pointer) sFrom);
  nmakeattrptr(nchild(3, vc), SME_BackAnnotationATTRIBUTENUMBER, (pointer) sTo);
  // some particular cases.
  if (opfrom == opevnt)
    addFlag(vc, useSrcClockFlag, natomint(opSMEvalue, TrueFlag));
  else if (opfrom == opconv)
    addFlag(vc, castTypeFlag, STree_scopy(nchild(1, sFrom)));
  //  else if (opfrom == oprecu) addFlag(vc, arrayRecoveryFlag,  STree_scopy(nchild(2, sFrom)));  
  return vc;
}

/**
   It adds a data flow connection to the current model.
   See addConnection() 
*/
_VIRTUAL_ ntrees SME_Translator::addDataflowConnection(ntrees to, ntrees from)
{
  return addConnection(to, from, opDataflowConnection);
}

/**
   It adds a data flow connection to the current model from the ithChild child of p to p.
   See addConnection() 
*/
_VIRTUAL_ ntrees SME_Translator::addDataflowConnection(ntrees p, int ithChild)
{
  if (isoChild(ithChild, p))
    return addDataflowConnection(p, operande);
  return NILTREE;
}

/**
   It adds a condition connection to the current model.
   See addConnection() 
*/
_VIRTUAL_ void SME_Translator::addConditionConnection(ntrees to, ntrees from)
{
  addConnection(to, from, opConditionConnection);
}

/**
   It adds a condition connection to the current model from the ithChild child of p to p.
   Used for the second operator of a when or cell.
   It calls addConnection(), followed by moveAttribute() : the second operand in the meta model is managed as a parameter, not as an operand.
   \verbatim
   The following SME code:
   <locals  SME:id="39" dst_DataflowConnections="78" src_Conditions="228 233 238 243" name="OK_perform"  type=""/>
   ...
   <connections  xsi:type="sme:Condition" name="Condition246" SME:id="228" src="39" dst="226"/>
   ...
   <operators  xsi:type="sme:Extraction" name="Extraction218" SME:id="226" dst_Condition="228" src_DataflowConnections="246">
   <in SME:id="224" dst_DataflowConnections="227" operatorParent="226" name="Extraction218_in1"/>
   </operators>
   expresses that the condition (second operand of a when) of the Extraction218 is the signal OK_perform 
   (node "39", source of the condition connection number "228").
   \verbatim
*/
_VIRTUAL_ void SME_Translator::addConditionConnection(ntrees p, int ithChild)
{
  ntrees t;
  if (isoChild(ithChild, p))
    {
      t = operande;
      addConditionConnection(p, t);
      moveAttribute(p, t, opsrc_condc, opdst_condc);
    }
}

/**
   It adds a clock relation connection to the current model.
   See addConnection().
   Used for the ^=, ^<, ^#, ^>, ^+, ^*, ^- SIGNAL operators. 
*/
_VIRTUAL_ ntrees SME_Translator::addClockRelationConnection(ntrees to, ntrees from)
{
  return addConnection(to, from, opClockRelationConnection);
}

/**
   It adds a clock relation connection to the current model from the ithChild child of p to p.
   See addConnection() 
*/
_VIRTUAL_ ntrees SME_Translator::addClockRelationConnection(ntrees p, int ithChild)
{
  if (isoChild(ithChild, p))
    return addClockRelationConnection(p, operande);
  return NILTREE;
}

/**
   Specialisation of Trees2Kernel::cDecFlows().
   
   Moreover, a SME identifier is associated with each identifier of the list of declaration.
*/
_VIRTUAL_ ptlrepid SME_Translator::cDecFlows(noperators opContext, ptlrepid deblist, boolean is_param)
{
  ptlrepid x;
  ntrees p;
  ntrees pred;

  p = operande;
  x = Trees2Kernel::cDecFlows(opContext, deblist, is_param);

  operande = p;
  pred = operande;
  // type : if (isoChild(1, operande)) deftyp = cExpType(operande); else deftyp = kl_mkBasicType(opimplicit);
  // signal and initialisation
  if (isChild(2, pred))
    {
      pred = operande;
      if (firstChild(&pred))
	do
	  {
	    if (voperator == opinit)
	      operande = nchild(1, operande);
	    getSMEId(operande);
	  }
	while (nextChild(&pred));
    }
  return x;
}

/**
   @return 
   - "true" if a declaration (vdecl), characterized by its operator (op), is already imported (vimport) in the current compiling unit, 
   - "false" otherwise, vimport is not modified.
*/
boolean SME_Translator::existImportOf(ntrees importList, ntrees vdecl, noperators op, ntrees * vimport)
{
  int i;
  ntrees s, sdecl;

  for (i = 1; i <= ntreearity(importList); i++)
    {
      s = nchild(i, importList);
      if (noper(s) == op)
	if (getDeclarationAttribute(s, &sdecl))
	  if (vdecl == sdecl)
	    {
	      *vimport = s;
	      return pk_true;
	    }
    }
  return pk_false;
}

/**
   Local to setIdentAttribute().
   It adds to the current model the import of the declaration of the object pointed by vdecl (constant/type).
   The list of the imports is managed as an attribute number=SME_ImportATTRIBUTENUMBER associated with currentModel. 

   \verbatim
   In the following example, the object 211 of the module APEX_ARINC653_TYPES.sme is imported in the current model using the SME identifier 776.
   <constantRef  SME:id="776" src_DataflowConnections="779 797">
      <constant  xsi:type="sme:ConstantValue" href="APEX_ARINC653_TYPES.sme#211"/>
   </constantRef>
   \endverbatim

   \note Currently, the meta model does not managed the types like the constants... (to be completed).
*/
void SME_Translator::setIdentAttributeImport(ntrees p, ntrees vdecl, ntrees xmodule)
{
  ntrees vimport, l, x;
  boolean stop;
  noperators op;

  // vdecl is a constant or type declaration
  x = vdecl;
  do
    {
      x = nfather(x);
      if (IS_NILTREE(x))
	{
	  stop = pk_true;
	  op = opnil;
	}
      else
	{
	  op = noper(x);
	  stop = (op == opdeccon) || (op == opdecconp) || (op == opltype);
	}
    }
  while (!stop);

  if (op != opnil)
    {
      if (op == opltype)
	op = opSMEImportType;
      else
	op = opSMEImportCons;	// encoding for the attribute. 

      if (nattrexist(currentModel, SME_ImportATTRIBUTENUMBER))
	l = nattrntrees(currentModel, SME_ImportATTRIBUTENUMBER);
      else
	{
	  l = LL_make_SMEImportList_emptyList();
	  nmakeattrntrees(currentModel, SME_ImportATTRIBUTENUMBER, l);
	}

      if (!existImportOf(l, vdecl, op, &vimport))
	{
	  vimport = STree_mkNullaryTree(op);
	  setDeclarationAttribute(vimport, vdecl, NILTREE);
	  l = ntreeList_appendLast(l, vimport);
	  setSMEId(vimport);
	  setSMEId(vdecl);
	}
      setDeclarationAttribute(p, vimport, NILTREE);
    }
}

/**
   @return "true" if a tree (vdecl) is an sub-tree, using the visibility rules of the language, 
   of the current block (currentBlock), "false" otherwise.
*/
boolean SME_Translator::isInCurrentBlock(ntrees vdecl)
{
  noperators op;
  while (pk_true)
    {
      if (IS_NILTREE(vdecl))
	return pk_false;
      if (vdecl == currentBlock)
	return pk_true;
      op = noper(vdecl);
      if (op == opinte)
	{
	  if (nfather(vdecl) == currentBlock)
	    return pk_true;
	  else
	    return pk_false;
	}
      if (op == opiter)
	return pk_false;
      if (op == oppcas)
	return pk_false;
      if (op == opmoti)
	return pk_false;
      if (op == oppcas)
	return pk_false;
      if (op == oppconf)
	if (noper(nfather(vdecl)) != opdesc)
	  return pk_false;

      vdecl = nfather(vdecl);
    }

}

/**
   @param vdecl (in) identifier tree of a declaration.
   @param idClass (in) type of identifier (constant, signal, ...)
   @return "true" if res is fixed , false otherwise (vdecl is a label, a index ,...)
   @param res (out) a tree representing a reference to a constant/parameter/signal (depending on idClass) designed by vdecl.

   \verbatim
   In the following example, the object numbered 1 is referenced in the current model using the SME identifier 19.
   <signalRef  SME:id="19" src_DataflowConnections="24" signal="1"/>
   \endverbatim  
*/
boolean SME_Translator::mkReferenceObjet(ntrees vdecl, typval idClass, ntrees * res)
{
  ntrees x;
  noperators op;

  if ((idClass == cste) || (idClass == par))
    op = opSMEConstantRef;
  else if (idClass == sig)
    op = opSMESignalRef;
  else
    return pk_false;		// label, index,

  x = STree_mkNullaryTree(op);
  nmakeattrptr(x, SME_ReferenceDeclATTRIBUTENUMBER, (pointer) vdecl);
  setSMEId(x);
  (*res) = x;
  return pk_true;
}

/**
   Assumed : p is an identifier.

   For an identifier (p), it searches for its declaration (reference solving defined in getTreeDecl()).
   Then, it assignes this declaration to p (setDeclarationAttribute()).
   Moreover, it adds to the model the imports of the object p in the current model (setIdentAttributeImport().)

   
*/
_VIRTUAL_ void SME_Translator::setIdentAttribute(ntrees p)
{
  ntrees vdecl, xmodule;
  typid proto;
  typval idClass;
  ntrees x;

  if (getTreeDecl(p, &vdecl, &xmodule, &proto, &idClass))
    {				// proto in { protoIdent, protoType, protoModel}
      /*  and when  proto= protoIdent, 
         idClass in {  sig (signal), ind (index), par (parameter), cste (constant), eti (label)}
       */
      if (IS_NILTREE(xmodule))
	if (proto == protoIdent)
	  {
	    setSMEId(vdecl);
	    if (RefTable->find(vdecl) != RefTable->end())
	      {
		x = (*RefTable)[vdecl];
		setDeclarationAttribute(p, x, NILTREE /*xmodule */ );
		return;
	      }
	    else
	      {
		if ((idClass != ind) && (idClass != eti))
		  if (!isInCurrentBlock(vdecl))
		    {
		      // Construire le x = f(signalRef/constantRef, vdecl)
		      if (mkReferenceObjet(vdecl, idClass, &x))
			{
			  (*RefTable)[vdecl] = x;
			  setDeclarationAttribute(p, x, NILTREE /*xmodule */ );
			}
		      return;
		    }
	      }
	  }
      if (setDeclarationAttribute(p, vdecl, xmodule))
	setIdentAttributeImport(p, vdecl, xmodule);
    }
}

/**
   Translating an unary tree (p).
   In the general case, it call addDataflowConnection() applied on the operand of p.

   \remarks{
   - Cartesian product (oppcar operator is a particular case in the meta model).
   - Assumed : p is not a unary when, "when C" is supposed rewritten in "true when C"
   - event (x) is managed as an attribute in the meta model (useSrcClock=)
   - x [?] : [?] is managed as an attribute of a connection (the connection is build on x with [?] as attribute)
   }
*/
_VIRTUAL_ void SME_Translator::SMEUnary(ntrees p)
{
  noperators opecourant;

  opecourant = noper(p);
  switch ((int)(opecourant))
    {

    case opvvarin:		// x? (x is a state variable)
      break;

    case optran:		//   transposee de matrice  "tr" <expr_composite> 
    case opnega:		// - Expr
    case opnon:		// not Expr

    case oppcar:		// cartesian product: <produit_index> << l_expr_signal >> 
      addDataflowConnection(p, 1);
      break;

    case opwhen1:		// it does not occur ( "when C" is rewritten in "true when C")
      addConditionConnection(p, 1);
      break;

    case ophnull:		// ^0
      break;

    case opevnt:
      break;

    case opsdec:		// <fact_flow> [?]
      // addDataflowConnection( p, 1);
      break;
    }
}

/**
   Local to moveAttribute() method.
   It moves the flags (attribute number SME_FlagATTRIBUTENUMBER) associated with a connection (vconnection).
   It replaces the useSrcClockFlag of the connection by the conditionKindFlag flag. It is used to express when ^C.
*/
void SME_Translator::moveFlags(ntrees vconnection)
{
  int i, nb;
  ntrees vcodeFlag, vflags, vflag, v;
  boolean done;

  done = pk_false;
  if (nattrexist(vconnection, SME_FlagATTRIBUTENUMBER))
    {
      vflags = nattrntrees(vconnection, SME_FlagATTRIBUTENUMBER);
      for (i = 1; !done && i <= ntreearity(vflags); i++)
	{
	  vflag = nchild(i, vflags);
	  vcodeFlag = nchild(1, vflag);
	  nb = nintof(vcodeFlag);
	  if (nb == useSrcClockFlag)
	    {
	      done = pk_true;
	      nmodifvalint(vcodeFlag, conditionKindFlag);
	      v = nchild(2, vflag);
	      nmodifvalint(v, presentFlag);
	    }
	}
    }
  if (!done && (noper(vconnection) == opConditionConnection))
    addFlag(vconnection, conditionKindFlag, natomint(opSMEvalue, TrueFlag));

}

/**
   It moves the attributes associated with a tree (vt) to an another one (p), when vt is:
   - a condition of a when, cell 
   - a child of a clock relation (^=, ^#, ?).
   Why? : because the SME model is not homogenous.
   Some flags are also modified : see moveFlags().
*/
_VIRTUAL_ void SME_Translator::moveAttribute(ntrees p, ntrees vt, int opsrc, int opdst)
{
  ntrees vdef, t;
  int i;
  ntrees idConnection, cDst, s, lconnections, dst;
  boolean found;

  vt = getTreeForConnection(vt);
  if (!getDeclarationAttribute(vt, &t))
    t = vt;			// for identifiers.
  if (nattrexist(t, SME_ATTRIBUTENUMBER))
    {
      vdef = nattrntrees(t, SME_ATTRIBUTENUMBER);
      if (getAttributeWithOp(vdef, opsrc, &t))
	if (getModelConnections(&lconnections))
	  {
	    // Changing the definition of the connection.
	    if (nattrexist(vt, SME_idoperandATTRIBUTENUMBER))
	      dst = nattrntrees(vt, SME_idoperandATTRIBUTENUMBER);
	    else
	      dst = nattrntrees(vt, SME_idATTRIBUTENUMBER);
	    found = pk_false;
	    for (i = 1; !found && i <= ntreearity(lconnections); i++)
	      {
		s = nchild(i, lconnections);
		idConnection = nchild(1, s);
		cDst = nchild(3, s);
		if (STree_isTreeEq(cDst, dst))
		  {
		    nchangchild(s, getSMEId(p), 3);
		    addAttributeWithOp(p, STree_scopy(idConnection), opdst);
		    moveFlags(s);
		    found = pk_true;
		  }
	      }
	  }

    }
}

/**
   Translating a binary tree (p).
   In the general case, it call addDataflowConnection() applied on the two operands of p.

   \remarks{
   - for an identifier (opident) it calls setIdentAttribute().
   - for a when addConditionConnection() is called for the second operand (particular case).
   - for the clock operator (^+ ^* ^-) addClockRelationConnection() is called for each operand.
   - E\\F is not considered here (F is an textual attribute of E in the meta model).
   - counterKindFlag flag is set for the counters (after, from, count).
   - E |* E, var E init E0 : the second operand will be a textual attribute of the first one.
   - for a reference to a model P{X} (I1,...,In), the reference to the model is solved in locateCalledProcess().
   - A reference to a field, or an element of an array is not considered here. The fields/[] will be a textual attribute of the referenced signal.
   - a reference to a value of an enumerated type is filtered before (in isSelectedExpression()), it becomes a literal in SME.
}
*/
_VIRTUAL_ void SME_Translator::SMEBinary(ntrees p)
{
  noperators opecourant;

  opecourant = noper(p);
  switch ((int)(opecourant))
    {
    case opident:
      setIdentAttribute(p);	// an attribute is set at the declaration, it is propagated here
      break;

    case opextr:		// when
      addDataflowConnection(p, 1);
      addConditionConnection(p, 2);
      break;

    case opbsup:		// ^+ ^* ^-
    case opbinf:
    case opcomp:
      addClockRelationConnection(p, 1);
      addClockRelationConnection(p, 2);
      break;

      /* case oprecu: break; */

      // case oprepe: // E |* E
    case opvar:		//var E init E0
      addDataflowConnection(p, 1);
      // the second will be a "Signal text".
      break;

    case opappe:
      // Done in locateCalledProcess for the referenced process...Caution for assert().
      addDataflowConnection(p, 2);
      break;

    case opconv:
      // it is a special case!!! the connection is attributed with the attribute "castType"
      break;

    case opcham:
    case opelem:
      // it is a special case!!! the connection is attributed with the attribute "srcFields" or dstFields
      break;

    default:
      addDataflowConnection(p, 1);
      addDataflowConnection(p, 2);
      // SME default: if (op == opnomba) addFlag(p, counterKindFlag, string2ident("after")); else
      if (opecourant == opnombd)
	addFlag(p, counterKindFlag, string2ident("from"));
      else if (opecourant == opnombm)
	addFlag(p, counterKindFlag, string2ident("count"));
    }
}

/**
   Translating a ternary tree (p).
   In the general case, it call addDataflowConnection() applied on the three operands of p.

   \remarks{
   - for a delay/windows the second and the third operands becomes "textual" attributes of the delay.
   - for a cell, the third operand (initial value) becomes a "textual" attributes of the memory.
   - for a if operator, addConditionConnection() is applied for the first operand.
   - for an enumeration "lb .. up step s" the 3 operands are textual fields in SME.
   }
*/
_VIRTUAL_ void SME_Translator::SMETernary(ntrees p)
{
  noperators op;

  // OPERATORS: opborn, opmemo, opret, opfen, openum, opcond.
  op = noper(p);
  switch ((int)(op))
    {
    case opmemo:
      addDataflowConnection(p, 1);
      addConditionConnection(p, 2);
      // 3th child (initial value) is printed as a SIGNAL Text.
      break;
    case opret:
    case opfen:
      addDataflowConnection(p, 1);
      // 2th child (nb instants) becomes a flag (but it can be a signal expression ; it is an error!!)
      // 3th child (initial value) is printed as a SIGNAL Text.
      break;
    case opcond:
      addConditionConnection(p, 1);
      addDataflowConnection(p, 2);
      addDataflowConnection(p, 3);
      break;

    case openum:
      // The 3 operands are fields in SME.
      break;

    default:			// case opfen: case opborn: case opfen:
      addDataflowConnection(p, 1);
      addDataflowConnection(p, 2);
      addDataflowConnection(p, 3);
    }
}

/**
   Translating a list of effecive parameters (p).
   It calls addConnection() for each parameter to add a "constant affectation connection" (opConstantAffectation operator).
   
   Assumed : p is a list.
*/
_VIRTUAL_ void SME_Translator::SMEListParams(ntrees p)
{
  int i;
  ntrees s;

  for (i = 1; i <= ntreearity(p); i++)
    {
      s = nchild(i, p);
      addConnection(p, s, opConstantAffectation);
    }
}

/**
   Translating a list tree (p):
   - for a list, operand of a dependence or an iterate (list of index) : this method is without effect.
   - for a clock relation operator (^=, ^#): addClockRelationConnection(), followed by moveAttribute(), are applied for each son.
   - for a clock relation operator  ^< ^>: addClockRelationConnection() is applied for each son.
   - in the general case, it call addDataflowConnection() applied for each son of p.
*/
_VIRTUAL_ void SME_Translator::SMEList(ntrees p)
{
  noperators opecourant, opf;
  int i;
  boolean b1, b2;
  ntrees vc, f;

  opecourant = noper(p);
  f = nfather(p);

  if (!STree_isUndefinedTree(f))
    {
      opf = noper(f);
      if ((opecourant == oplesign) && ((opf == opldep) || (opf == opdepd)))
	return;
      if ((opecourant == opleflot) && (opf == opiter))
	return;
      // if ((opecourant == oplesign) && (opf == oprest)) return;
    }

  b1 = isClockRelationOperator_local(opecourant);	// ^= ^#
  if (!b1)
    b2 = isClockSpeedRelationOperator(opecourant);	// ^< ^>
  else
    b2 = pk_false;

  for (i = 1; i <= ntreearity(p); i++)
    if (b1 || b2)
      {
	vc = addClockRelationConnection(p, i);
	if (b1)
	  moveAttribute(p, nchild(i, p), opsrc_clockc, opdst_clockc);
      }
    else
      addDataflowConnection(p, i);
}

/**
 Translating a nullary tree (p).
 - for an identifier : see setIdentAttribute().
 - for constants : see  SMEConstant().
 - the other SIGNAL operators (opexte, opnil) are ignored.
*/
_VIRTUAL_ void SME_Translator::SMENullary(ntrees p)
{
  if (noper(p) == opnom)
    setIdentAttribute(p);	// an attribute is set at the declaration, it is propagated here

  /* Nullary operators
     -----------------
     opnom opcstent opcstdp opcstreel opcstcar opcstcha 
     opfalse optrue opexte opretun opnil 
   */
}

/**
   Translating an expression (p):
    - a SME identifier is assigned to p, 
    - then, SMENullary(), SMEUnary(), SMEBinary(), SMEList() is called depending on the arity of p.
*/
_VIRTUAL_ void SME_Translator::cExprSigSME(ntrees p)
{
  noperators voperator;

  getSMEId(p);
  voperator = noper(p);
  if (LL_is_listOp(voperator))
    SMEList(p);
  else if (LL_is_ternaryOp(voperator))
    SMETernary(p);
  else if (LL_is_binaryOp(voperator))
    SMEBinary(p);
  else if (LL_is_unaryOp(voperator))
    SMEUnary(p);
  else if (LL_is_atomOp(voperator))
    SMENullary(p);
  else
    return;			/* PLG internal error */
}

/**
   Local to connectOutputCall().
   
*/
boolean SME_Translator::getOPorts(ntrees t, ntrees * vdecl)
{
  if (!nattrexist(t, SME_OPortsATTRIBUTENUMBER))
    return pk_false;
  (*vdecl) = nattrntrees(t, SME_OPortsATTRIBUTENUMBER);
  return pk_true;
}

/**
   The rank-th output of output, is added as an output port of the call (calltree).
   A new SME identifier ID is created to realize the connection.
   To a reference of a model (calltree) is associated  a list of ports (a tree with oploports as operator).
   An element of this list is a binary tree (opport, rank, ID).
   The link calltree-list of ports is managed by the attribute called SME_OPortsATTRIBUTENUMBER.
*/
void SME_Translator::connectOutputCall(ntrees calltree, ntrees output, int rank)
{
  ntrees op, vp, id;

  // adding as attribute
  if (!getOPorts(calltree, &op))
    {
      op = LL_make_loports_emptyList();
      nmakeattrntrees(calltree, SME_OPortsATTRIBUTENUMBER, op);
    }
  id = newSMEId();
  addDataflowConnection(output, id);
  vp = nmakebinary(opport, natomint(opSMEvalue, rank), id);
  op = ntreeList_appendLast(op, vp);
}

/** 
    It adds the connections of the outputs (outputs) of a reference to a model (calltree).
    It calls connectOutputCall(ntrees, ntrees, int) for each element of outputs.
*/
void SME_Translator::connectOutputsCall(ntrees outputs, ntrees calltree)
{
  int i;
  ntrees s;

  if (!STree_isUndefinedTree(outputs))
    if (STree_isList(outputs))
      {
	for (i = 1; i <= ntreearity(outputs); i++)
	  {
	    s = nchild(i, outputs);
	    connectOutputCall(calltree, s, i);
	  }
      }
    else
      connectOutputCall(calltree, outputs, 1);
}

/**
   @return "true" if an expression managed in the meta model, 
   "false" otherwise, because there are some restrictions in the meta-model.
   
   It can also return "false" for some expression not yet implemented.
*/
boolean SME_Translator::isSelectedExpression(ntrees e)
{
  noperators op, opf;
  ntrees f;

  op = noper(e);
  if ((op == opelevec) || (op == opensvec))
    notOK(e);

  //  if (op == opevnt) return pk_false; // it becomes an attribute in a connection.
  if (op == openum)
    return pk_false;		// all the operands become fields in SME.
  if (op == opvenum)
    return pk_false;		// it becomes a literal in SME.

  f = nfather(e);
  if (!STree_isUndefinedTree(f))
    {
      opf = noper(f);
      if ((opf == opret) || (opf == opfen))
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	  if (nchild(3, f) == e)
	    return pk_false;
	}
      else if (opf == opcham)
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	}
      else if (opf == oprepe)
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	}
      else if (opf == opelem)
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	}
      else if (opf == opmemo)
	{
	  if (nchild(3, f) == e)
	    return pk_false;
	}
      else if (opf == opinit)
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	}
      else if (opf == opcons)
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	}
      else if (opf == opvar)
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	}
      //      else if (opf == oprecu) { if (nchild(2,f) == e) return pk_false; } // operateur E1 \\ E2
      else if (opf == opcas)
	{
	  if (nchild(1, f) == e)
	    return pk_false;
	}
      else if (opf == opindmot)
	{
	  if (nchild(2, f) == e)
	    return pk_false;
	}			// upper bound of an index (array structure)
    }
  return pk_true;
}

/**
   specialisation of Trees2Kernel::locateCalledProcess().

   Specialisation: for a reference (p to a model (vfi), 
   the links 
   - from p to its declaration, 
   - from its declaration to the module in which the model is declared.

   are kept (see setDeclarationAttribute()).
*/
_VIRTUAL_ boolean SME_Translator::locateCalledProcess(ntrees p, ptrepfi * vfi)
{
  ntrees saveOperande, arbreaux, vmodule, vdecl;
  noperators savevoperator;
  boolean b;
  ptIdStack name;
  ptrepfi vfiModule;

  saveOperande = operande;
  savevoperator = voperator;

  b = isoChild(1, p);
  if (b)
    {
      if (voperator == opinst)
	arbreaux = operande;
      else
	arbreaux = p;
      locateChild(arbreaux, 1, &name);
      b = getTopIfModel(name, vfi);
      vdecl = kl_getTreeReference(*vfi);
      if (ptIdentProp_getModule(kl_top(name), &vfiModule))
	vmodule = kl_getTreeReference(vfiModule);
      else
	SET_NILTREE(vmodule);
      setDeclarationAttribute(STree_getNameCalledModel(p), vdecl, vmodule);
    }
  operande = saveOperande;
  voperator = savevoperator;
  return b;
}

/**
   It initializes the translator for a new block:
   - the tree becomes the current block, vt is 
          - a model (process, node, function, ....), 
          - a confined expression, 
          - a labelled expression,
	  - a case of a switch,
          - an iterate,
	  
   - the set of added constants is reset,
   - the counter of the new identifier is reset,
   - the SME mode is set.

   The translation of a block must be finalized by a call to finalBlock().
*/
void SME_Translator::initBlock(ntrees vt)
{
  ntrees lab, vdecl;

  currentBlock = vt;
  if (noper(vt) == oppetiq)	// A label is not declared in the SME.
    {
      // the id of the declaration of the label is assigned to the process.
      lab = nchild(1, vt);
      setIdentAttribute(lab);
      if (getDeclarationAttribute(lab, &vdecl))
	nmakeattrntrees(vt, SME_idATTRIBUTENUMBER, getSMEId(vdecl));
    }
  else
    setSMEId(currentBlock);
  cptName = 0;
  nmakeattrntrees(currentBlock, SME_ConnectionATTRIBUTENUMBER, LL_make_SMEList_emptyList());
  modeSME = pk_true;
  RefTable = new TypeRefTable();
}

/**
   It finalizes the translation of the current block, initialized in initBlock().
   The added constant are declared for the current block (a confined expression can be created).
*/
void SME_Translator::finalBlock()
{
  ntrees attr;
  TypeRefTable::iterator it;

  if (!RefTable->empty())
    {
      attr = LL_make_SMEList_emptyList();
      nmakeattrntrees(currentBlock, SME_ReferencesATTRIBUTENUMBER, attr);
      for (it = RefTable->begin(); it != RefTable->end(); it++)
	attr = ntreeList_appendLast(attr, (*it).second);
    }
  delete RefTable;
}

/**
   specialisation of Trees2Kernel::cExprSig().

   Specialisation:
    - the SME mode is pushed,
    - the expression is compiled by the standard method: Trees2Kernel::cExprSig(),
    - it is revisited if the current SME mode is set (cExprSigSME()).
      The SME mode is unset when an expression is printed as textual form in the SME external representation.
    - the SME mode is then poped.    
*/
_VIRTUAL_ ptklExprSig SME_Translator::cExprSig(ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig vexpres;

  PushSMEMode();
  if (modeSME)
    modeSME = isSelectedExpression(p);
  vexpres = Trees2Kernel::cExprSig(p, les, lparam, indexList);
  if (modeSME)
    cExprSigSME(p);
  PopSMEMode();
  return vexpres;
}

/**
   
   Compiling a model (vfi), specialisation of Trees2Kernel::modelCompiling().
   The following actions are applied:   
   - The current model, the main model are managed
   - the different contexts are pushed
   - a new block is initialized, see initBlock()
   - the classical compiling is called, see Trees2Kernel::modelCompiling()
   - the block is finalized, see finalBlock()
   - the different contexts are poped.
*/
_VIRTUAL_ void SME_Translator::modelCompiling(ptrepfi vfi)
{
  ntrees save;
  boolean b;

  save = currentModel;
  currentModel = kl_getTreeReference(vfi);
  if (IS_NILTREE(topModel))
    {
      topModel = currentModel;
      b = pk_true;
    }
  else
    b = pk_false;
  PushContext();
  initBlock(currentModel);
  Trees2Kernel::modelCompiling(vfi);
  finalBlock();
  PopContext();
  if (b)
    SET_NILTREE(topModel);
  currentModel = save;
}

/**
   It finalizes the compiling of an equation X := E, X::=E, (X1,...) := P{} (),...
   The outputs are connected: for a reference to a model, it calls connectOutputsCall(), otherwise a 
   standard connection is added (see addConnection()).

   For ::= , the flag "partialDefFlag" is added to the connection.
   For ::= defaultvalue, the flag "defaultValuepartialDefFlag" is added to the connection.
*/
void SME_Translator::SMEFinalizeEquation(ntrees p, noperators op)
{
  ntrees vc, t1, t2;
  noperators opt2;

  t1 = nchild(1, p);
  t2 = nchild(2, p);
  opt2 = noper(t2);

  if ((opt2 == opappe) || (opt2 == opinst) || (opt2 == oppcar))
    // x,y := P (a,b) (x,y) := <<A,B,...>>
    connectOutputsCall(t1, t2);
  else
    {
      vc = addConnection(t1, t2, opDataflowConnection);
      if (op == opdefpart)	// ::=
	;			// remove December 2008, (shared, statevar signals): addFlag(vc, partialDefFlag,  natomint(opSMEvalue, TrueFlag));
      else if (op == opdefpartdef)	// ::= defaultvalue
	addFlag(vc, defaultValuepartialDefFlag, natomint(opSMEvalue, TrueFlag));
    }
}

/**
   Compiling an equation X:=E or X::=E or X::= defaultvalue E, specialisation of Trees2Kernel::cExpEquation().
   The standard compiling (Trees2Kernel::cExpEquation()) is followed by SMEFinalizeEquation() to connect the output.
*/
boolean SME_Translator::cExpEquation(noperators op, ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  boolean b;

  b = Trees2Kernel::cExpEquation(op, p, inputs, outputs, indexList, lparam, vklExprProc);
  if (b)
    SMEFinalizeEquation(p, op);
  return b;
}

/**
   Compiling an equation (X1,...Xn):=E, or (X1,...Xn)::=E, specialisation of Trees2Kernel::cExpEquation().
   The standard compiling (Trees2Kernel::cExpEquation()) is followed by SMEFinalizeEquation() to connect the output.
*/
_VIRTUAL_ boolean SME_Translator::cTupleDef(noperators op, ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  boolean b;
  b = Trees2Kernel::cTupleDef(op, p, inputs, outputs, indexList, lparam, vklExprProc);
  if (b)
    SMEFinalizeEquation(p, op);
  return b;
}

/**
   Compiling a confined expression (| ... |) where ... end;, specialisation of Trees2Kernel::cExpConfinedExp().
   The following actions are applied:
   - the different contexts are pushed
   - a new block is initialized, see initBlock()
   - the classical compiling is called, see Trees2Kernel::cExpConfinedExp()
   - the block is finalized, see finalBlock()
   - the different contexts are poped.
*/
_VIRTUAL_ boolean SME_Translator::cExpConfinedExp(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean b;

  PushContext();
  initBlock(p);
  b = Trees2Kernel::cExpConfinedExp(p, inputs, outputs, vklExprProc);
  finalBlock();
  PopContext();
  return b;
}

/**
   Local to importType().
   Currently, it is without effect.
*/
void SME_Translator::importConstant(ntrees p)
{
}

/**
   It marks:
   - the referenced types in a type (p) to have the imported type.
   - the referenced constants in a type (p) to have the imported constants (see importConstant()).
*/
void SME_Translator::importType(ntrees p)
{
  ntrees s, t;
  int i;

  voperator = noper(p);
  switch (voperator)
    {
    case opnom:
    case opident:
      setIdentAttribute(p);
      break;

    case optvect:
      if (isChild(1, p))
	importConstant(operande);	// il faudrait marquer les objects (constantes) à importer.
      if (isChild(2, p))
	importType(operande);
      break;

    case opstru:
    case opbund:
      if (isChild(1, p))	// access to the fields
	{
	  t = operande;
	  for (i = 1; i <= ntreearity(t); i++)
	    {
	      s = nchild(i, t);
	      if (isoChild(1, s))
		importType(operande);
	    }
	}
      break;

    default:			// basic type + case opexte: case optenum:
      break;

    }
}

/**
   Compiling a type (p) : specialisation of Trees2Kernel::cExpType(p).
   The following action are applied:
   - the SME mode is pushed and unset,
   - the classical compiling is called, see Trees2Kernel::cExpType()
   - the specific compiling is called, see importType().
   - the SME mode is poped.
*/
_VIRTUAL_ ptreptyp SME_Translator::cExpType(ntrees p)
{
  ptreptyp x;

  PushSMEMode();
  modeSME = pk_false;
  x = Trees2Kernel::cExpType(p);
  importType(p);
  PopSMEMode();
  return x;
}

/**
   Compiling a list of effective parameters (p) : specialisation of Trees2Kernel::cExpEffectiveParam(p).
   The following action are applied:
   - the SME mode is pushed and set,
   - the classical compiling is called, see Trees2Kernel::cExpEffectiveParam()
   - the specific compiling is called, see SMEListParams().
   - the SME mode is poped.
*/
_VIRTUAL_ ptklExprSig SME_Translator::cExpEffectiveParam(ntrees p, ptrepfi vfi, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig x;

  PushSMEMode();
  modeSME = pk_true;
  x = Trees2Kernel::cExpEffectiveParam(p, vfi, lparam, indexList);
  SMEListParams(p);
  PopSMEMode();
  return x;
}

/**
   Compiling a labelled expression : specialisation of Trees2Kernel::cExpLabelledExp().
   
   For SME, the label is associated with a subprocess, and the lable is not declared.
   (1) e::(| SP |) and e is declared somewhere (2) label e.
   We attribute "e" in 1 by its declaration address (2) like for the other identifiers.
   We also attribute (| SP |) by this same attribute.

   Because the labels are not declared in the meta model, the pretty-printer will print the dependences 
   assigned with the declaration of the label in the definition of the sub-process.
*/
_VIRTUAL_ boolean SME_Translator::cExpLabelledExp(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean b;

  PushContext();
  initBlock(p);
  b = Trees2Kernel::cExpLabelledExp(p, inputs, outputs, vklExprProc);
  finalBlock();
  PopContext();
  return b;
}

/**
   for a selector and a sub-process (sp) of a case of a switch,
   a case connection is added from the selector to the sub-process.
   The induced links are also added (addAttributeWithOp()).
*/
ntrees SME_Translator::linkCaseOperands(ntrees selector, ntrees sp)
{
  ntrees SMEId, vc, x1;

  if (nattrexist(sp, SME_idoperandATTRIBUTENUMBER))
    x1 = nattrntrees(sp, SME_idoperandATTRIBUTENUMBER);
  else
    x1 = mkSMEId(sp, pk_true);
  vc = mkConnection(selector, x1, opCaseConnection, pk_false);
  SMEId = nchild(1, vc);
  addAttributeWithOp(selector, STree_scopy(SMEId), opsrc_case);
  addAttributeWithOp(sp, STree_scopy(SMEId), opdst_case);
  return vc;
}

/**
   Compiling a case of a switch.
   aCase is a tree such that {  } : p where p is a composition.
   
   The composition is compiled as a block 
   - the different contexts are pushed
   - a new block is initialized, see initBlock()
   - the classical compiling for a composition is called, see Trees2Kernel::cExpListExpressionsOnProcesses()
   - the block is finalized, see finalBlock()
   - the different contexts are poped. 

   then the links (case connection) between the selector of the switch and the current case are added, see linkCaseOperands().
*/
boolean SME_Translator::caseConnection(ntrees aCase, ntrees p, noperators op, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  ntrees vc, selector, s1, s2;
  boolean b;

  PushContext();
  initBlock(p);
  b = Trees2Kernel::cExpListExpressionsOnProcesses(p, op, inputs, outputs, vklExprProc);
  finalBlock();
  PopContext();
  if (b)
    {
      // link between the selector of the switch and the current case.
      selector = nfather(nfather(aCase));
      getSMEId(selector);
      cExprSigSME(nchild(1, selector));
      s1 = nchild(1, aCase);
      s2 = nchild(2, aCase);	// s2 is p
      getSMEId(s2);
      vc = linkCaseOperands(selector, s2);
      nmakeattrptr(vc, SME_BackAnnotationATTRIBUTENUMBER, (pointer) s1);	// useful for the SME pretty-printer.
    }
  return b;
}

/**
   Compiling a list of expressions on processes (| ...|), specilaisation of Trees2Kernel::cExpListExpressionsOnProcesses().

   Specialisation: if p is the second child of a case expression (example: {2}: (|s3 := 2 * z |)), the method caseConnection() is applied,
   otherwise the classical compiling (Trees2Kernel::cExpListExpressionsOnProcesses()).
*/
_VIRTUAL_ boolean SME_Translator::cExpListExpressionsOnProcesses(ntrees p, noperators op, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean b;
  ntrees f;

  f = nfather(p);
  if (noper(f) == opcas)	// { } : p
    b = caseConnection(f, p, op, inputs, outputs, vklExprProc);
  else
    b = Trees2Kernel::cExpListExpressionsOnProcesses(p, op, inputs, outputs, vklExprProc);
  return b;
}

/**
   Compiling a case expression : specialisation of Trees2Kernel::cExpCase().
   
   Specialisation:
   For the "else part" it calls linkCaseOperands(), the other cases of a switch are managed in cExpListExpressionsOnProcesses().
*/
_VIRTUAL_ boolean SME_Translator::cExpCase(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean b;

  b = Trees2Kernel::cExpCase(p, inputs, outputs, vklExprProc);
  if (b && isoChild(3, p))
    linkCaseOperands(p, operande);	/* else part */
  return b;
}

/**
   For each element (X) of lids, a dependence connection (X, dep) is added in the current model.
*/
void SME_Translator::addConnectionDependenceIn(ntrees lids, ntrees dep)
{
  int i;
  ntrees s, x1;

  if (STree_isList(lids))
    for (i = 1; i <= ntreearity(lids); i++)
      {
	s = nchild(i, lids);
	if (!getDeclarationAttribute(s, &x1))
	  x1 = s;
	addConnection1(x1, dep, opDependenceConnection, pk_true);
      }
  else
    {
      if (!getDeclarationAttribute(lids, &x1))
	x1 = lids;
      addConnection1(x1, dep, opDependenceConnection, pk_true);
    }
}

/**
   For each element (X) of lids, a dependence connection (dep, X) is added in the current model.
*/
void SME_Translator::addConnectionDependenceOut(ntrees lids, ntrees dep)
{
  int i;
  ntrees s, x1;

  if (STree_isList(lids))
    for (i = 1; i <= ntreearity(lids); i++)
      {
	s = nchild(i, lids);
	if (!getDeclarationAttribute(s, &x1))
	  x1 = s;
	addConnection1(dep, x1, opDependenceConnection, pk_true);
      }
  else
    {
      if (!getDeclarationAttribute(lids, &x1))
	x1 = lids;
      addConnection1(dep, x1, opDependenceConnection, pk_true);
    }
}

/**
   Compiling a dependence ( E --> F when H ): specialisation of Trees2Kernel::cExpDependence().

   Specialisation: 
   - Connections are added in the current model from each signal referenced in the left part of the dependence 
   to the dependence operator : see addConnectionDependenceIn().

   - Connections are added in the current model from each signal referenced in the right part of the dependence 
   to the dependence operator : see addConnectionDependenceOut().

   - the clock of the dependence is added in addConditionConnection().
*/
_VIRTUAL_ boolean SME_Translator::cExpDependence(ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  boolean b;

  b = Trees2Kernel::cExpDependence(p, inputs, vklExprProc);
  if (b)
    {
      nmakeattrntrees(p, SME_idoperandATTRIBUTENUMBER, getSMEId(p));
      if (isoChild(1, p))
	addConnectionDependenceIn(operande, p);
      if (isoChild(2, p))
	addConnectionDependenceOut(operande, p);
      addConditionConnection(p, 3);
    }
  return b;
}

/**
   Compiling a dependence (E --> F): specialisation of Trees2Kernel::cExplDependences().

   Specialisation: 
    - Connections are added in the current model from each signal referenced in the left part of the dependence 
   to the dependence operator : see addConnectionDependenceIn()

   - Connections are added in the current model from each signal referenced in the right part of the dependence 
   to the dependence operator : see addConnectionDependenceOut()

   Assumed : E --> F --> G does not occured, here (rewritten before in E --> F | F --> G).
*/
_VIRTUAL_ boolean SME_Translator::cExplDependences(ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  boolean b;

  b = Trees2Kernel::cExplDependences(p, inputs, vklExprProc);
  if (b)
    {
      nmakeattrntrees(p, SME_idoperandATTRIBUTENUMBER, getSMEId(p));
      if (isoChild(1, p))
	addConnectionDependenceIn(operande, p);
      if (isoChild(2, p))
	addConnectionDependenceOut(operande, p);
    }

  return b;
}

/**
   Compiling the left part of an equation ( :=), partial equation (::=), or the default value of a partial equation (::=defaultvalue),
   specialisation of Trees2Kernel::cExprSigLeftExpr().

   Specialisation:
   the expression is classicaly compiled (Trees2Kernel::cExprSigLeftExpr()) in a unset SME mode if it is set in the calling context.
   If in the current calling context, the SME mode is set, the references to the signals in the compiled expression are solved.

*/
_VIRTUAL_ ptklExprSig SME_Translator::cExprSigLeftExpr(ntrees p, ptlrepid * les, ptlrepid * lparam, ptlrepid * indexList)
{
  ptklExprSig x;
  int i;
  // very strange the managing of the SME mode.
  PushSMEMode();
  if (modeSME)
    modeSME = pk_false;
  x = Trees2Kernel::cExprSigLeftExpr(p, les, lparam, indexList);
  PopSMEMode();
  if (modeSME)
    if (STree_isList(p))
      for (i = 1; i <= ntreearity(p); i++)
	cExprSigSME(STree_getIdentOf(nchild(i, p)));
    else
      cExprSigSME(STree_getIdentOf(p));
  return x;
}

/**
   Compiling an iterate, specialisation of Trees2Kernel::cExpArray().
   
   Specialisation:
   - the different contexts are pushed
   - a new block is initialized, see initBlock()
   - the classical compiling for an iterate is applied, see Trees2Kernel::cExpArray()
   - the block is finalized, see finalBlock()
   - the different contexts are poped. 
   
   then the connections of the indexes are added in the current model.
*/
_VIRTUAL_ boolean SME_Translator::cExpArray(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean b;

  PushContext();
  initBlock(p);
  b = Trees2Kernel::cExpArray(p, inputs, outputs, vklExprProc);
  finalBlock();
  PopContext();
  return b;
}

/**
   Compiling an iterate, specialisation of Trees2Kernel::cExpIterate().
   
   Specialisation:
   - the different contexts are pushed
   - a new block is initialized, see initBlock()
   - the classical compiling for an iterate is applied, see Trees2Kernel::cExpIterate()
   - the block is finalized, see finalBlock()
   - the different contexts are poped. 
   
   then the connections of the indexes are added in the current model.

*/
_VIRTUAL_ boolean SME_Translator::cExpIterate(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptklExprProc * vklExprProc)
{
  boolean b;
  int i;
  ntrees from, to, s, x1;

  PushContext();
  initBlock(p);
  b = Trees2Kernel::cExpIterate(p, inputs, outputs, vklExprProc);
  finalBlock();
  PopContext();
  // Connection of the indexes.
  if (b)
    if (isChild(1, p))
      {
	from = operande;
	to = operande;
	// ---lbesnard:  it seems to be like a "left part" of a dependence with opDataflowConnection for a connection.
	if (STree_isList(from))
	  for (i = 1; i <= ntreearity(from); i++)
	    {
	      s = nchild(i, from);
	      if (!getDeclarationAttribute(s, &x1))
		x1 = s;
	      addConnection1(x1, s, opDataflowConnection, pk_true);
	    }
	else
	  {
	    if (!getDeclarationAttribute(from, &x1))
	      x1 = from;
	    addConnection1(x1, to, opDataflowConnection, pk_true);
	  }
      }
  return b;
}

_VIRTUAL_ boolean SME_Translator::cExpAssert(ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{
  boolean b;

  b = Trees2Kernel::cExpAssert(p, inputs, vklExprProc);
  if (b)
    addConditionConnection(p, 1);
  return b;
}

/**
   Translating a reference to a model (p):  foo (e1, e2, ..., en) or foo{p1,...,pp} (e1, e2, ..., en), specialisation of Trees2Kernel::cExpCall().
   Specialisation:
   - classical compiling : Trees2Kernel::cExpCall()
   - particular case for a reference to assert(C) : p is marked with an attribute, number SMEAssertionAttribute,and a 
   condition connection is added from the argument of the assertion to the tree p (assert is an operator in the meta model).
*/
_VIRTUAL_ boolean SME_Translator::cExpCall(ntrees p, ptlrepid * inputs, ptlrepid * outputs, ptlrepid * indexList, ptlrepid * lparam, ptklExprProc * vklExprProc)
{
  boolean b;
  ptrepfi vfi;

  b = Trees2Kernel::cExpCall(p, inputs, outputs, indexList, lparam, vklExprProc);
  if (b)
    {
      locateCalledProcess(p, &vfi);
      /*
	ntrees x1, arg;
	if (isAssertionId(kl_getNameModel(vfi)))
         {
         setSMEId(p);
         nmakeattrint(p, SMEAssertionAttribute, 1); //SMEAssertionAttribute is NOT DEFINED 
         arg = nchild( 1, nchild(2, p));
         if (!getDeclarationAttribute(arg, &x1)) x1=arg;
         addConnection1(x1, p, opConditionConnection, pk_true);
         }
       */
    }
  return b;
}

/**
   It adds a pragma (aPragma) to a tree (id).
   It is used to propagate a pragma to the specified signal.
*/
void SME_Translator::addPragmaAttribute(ntrees id, ntrees aPragma)
{
  ntrees obj, attr;

  if (!getDeclarationAttribute(id, &obj))
    obj = id;
  if (!nattrexist(obj, SME_PragmaATTRIBUTENUMBER))
    {
      attr = LL_make_lprag_emptyList();
      nmakeattrntrees(obj, SME_PragmaATTRIBUTENUMBER, attr);
    }
  else
    attr = nattrntrees(obj, SME_PragmaATTRIBUTENUMBER);
  attr = ntreeList_appendLast(attr, aPragma);
}

/** 
    For a pragma with associated objects, the pragma is propagated to these objects, then removed.
    This is done only for the predefined pragmas:  it is \e not \e homogeneous in the meta model.
*/
_VIRTUAL_ boolean SME_Translator::cPragmas(ntrees p, ptrepfi vfi)
{
  int i, j, m, n;
  int nb;
  ntrees lp, stmt, obj, np, p_obj, aPragma, name;
  boolean b;

  b = Trees2Kernel::cPragmas(p, vfi);
  if (b)
    if (!isUndefinedBodyTree(p))
      {
	// setSMEId(p);
	n = ntreearity(p);
	lp = nfather(p);
	for (i = 1; i <= n; i++)
	  {
	    aPragma = nchild(i, p);
	    setSMEId(aPragma);
	    p_obj = nchild(2, aPragma);
	    if (!isUndefinedBodyTree(p_obj))
	      {
		name = nchild(1, aPragma);
		if (isPredefinedPragmaName(ident2repstr(name), &nb))
		  {
		    stmt = nchild(3, aPragma);
		    m = ntreearity(p_obj);
		    for (j = 1; j <= m; j++)
		      {
			obj = nchild(j, p_obj);
			np = STree_mkPragmaIdent(STree_scopy(name), NILTREE, STree_scopy(stmt));
			addPragmaAttribute(obj, np);
		      }
		    nmakeattrntrees(aPragma, SME_AlreadyPrintedATTRIBUTENUMBER, STree_mkNullaryTree(optrue));
		  }
	      }
	  }
      }
  return b;
}

/**
   Compiling a module (aModule), specialisation of Trees2Kernel::moduleCompiling().
   Specialisation:
   - push and set some context informations : current module, counter used to create the SME id is reset, the main model is managed.
   - classical compiling : Trees2Kernel::moduleCompiling()
   - pop the context informations.
*/
_VIRTUAL_ ptrepfi SME_Translator::moduleCompiling(ntrees aModule, ptIdStack * varid)
{
  ntrees vmodule, save;
  ptrepfi vfi;
  int s_cpt;
  boolean b;

  //push
  s_cpt = cpt;
  save = currentModule;
  //work
  cpt = 0;
  currentModule = aModule;
  if (IS_NILTREE(topModel))
    {
      topModel = aModule;
      b = pk_true;
    }
  else
    b = pk_false;
  vfi = Trees2Kernel::moduleCompiling(aModule, varid);
  vmodule = kl_getTreeReference(vfi);	// vmodule is aModule ?
  setSMEId(vmodule);
  // pop
  currentModule = save;
  cpt = s_cpt;
  if (b)
    SET_NILTREE(topModel);
  return vfi;
}

/**
   Compiling a library (aImportedModule) : specialisation of Trees2Kernel::loadLibraryFile().
   The result is returned in aModule. Three cases can occur:
   - (1) If the library has been already translated (value available in the SME_Translator::loadedLibraries table, the stored model is returned.
   - (2) If the library has been already parsed, but not translated, 
     the associated module is translated and the result is stored in the SME_Translator::loadedLibraries table.
   - (3) Otherwise, the file is parser, and the previous case is applied.

   See the constructor of the class SME_Translator to have the explanation of why the case (2), can occur. 
*/
_VIRTUAL_ boolean SME_Translator::loadLibraryFile(ntrees aImportedModule, ptrepfi * aModule)
{
  ptlrepid lvfi;
  ntrees treeModule;
  boolean b;
  ptrepstr fileName;
  ptIdStack varid;

  if (!getFileNameLibrary(aImportedModule, &fileName))
    return pk_false;
  if (loadedLibraries.find(fileName) == loadedLibraries.end())
    {
      if (FirstStepLoadedLibraries.find(fileName) == FirstStepLoadedLibraries.end())
	{
	  if (!parseLibrary(fileName, &treeModule))
	    return pk_false;
	  lvfi = modulesListCompiling(treeModule, &b);
	  *aModule = getModelRepid(lvfi);	// The fisrt of the list.
	}
      else
	{
	  *aModule = FirstStepLoadedLibraries[fileName];
	  *aModule = moduleCompiling(kl_getTreeReference(*aModule), &varid);
	}
      loadedLibraries[fileName] = *aModule;
    }
  else
    *aModule = loadedLibraries[fileName];
  return pk_true;
}

/**
   Compiling a import of a module (aImportedModule) : specialisation of Trees2Kernel::importOfModule().

   Specialisation:
   - the classical compiling is applied : see Trees2Kernel::importOfModule()
   - a SME identifier is associated with the compiled tree.
   - a link between the compiled tree and the module is kept (setDeclarationAttribute())
   - for a non intrinsic library, addImportModule() is called.
*/
_VIRTUAL_ boolean SME_Translator::importOfModule(ntrees aImportedModule, ptrepfi * aModule)
{
  boolean b;
  ntrees t;

  b = Trees2Kernel::importOfModule(aImportedModule, aModule);
  if (b)
    {
      setSMEId(aImportedModule);
      t = kl_getTreeReference(*aModule);
      setDeclarationAttribute(aImportedModule, t, NILTREE);
      if (!intrinsicLoading)
	addImportModule(t);
    }
  return b;
}

/**
   Compiling a importation of modules (use M1, M2, ....Mn), specialisation of Trees2Kernel::importOfModules().
   Specialisation:
   - the classical compiling (Trees2Kernel::importOfModules()) is applied
   - a SME identifier is associated with the compiled tree.
*/
_VIRTUAL_ void SME_Translator::importOfModules(ntrees limports, ptlrepid * lvfi)
{
  Trees2Kernel::importOfModules(limports, lvfi);
  setSMEId(limports);
}

/**
   Compiling an external graph specification, specialisation of Trees2Kernel::cExternalGraph().
   Specialisation:
   - the classical compiling (Trees2Kernel::cExternalGraph()) is applied
   - a SME identifier is associated with the compiled tree.
*/
_VIRTUAL_ void SME_Translator::cExternalGraph(ntrees p_graph_extern, ptrepfi vfi, ntrees p_model, ptlrepid inList, ptlrepid outList, ptlrepid locList, ptlrepid llabels, boolean toplevel)
{
  ntrees p_expr;

  Trees2Kernel::cExternalGraph(p_graph_extern, vfi, p_model, inList, outList, locList, llabels, toplevel);
  if (!isUndefinedBodyTree(p_graph_extern))
    {
      p_expr = nchild(2, p_graph_extern);
      if (!isUndefinedBodyTree(p_expr))
	setSMEId(p_expr);
    }
}

/**
   Compiling  a tree (p) such that E1 :=: E2.
   @return "true" if the compiling succeeds, "false" otherwise.
   
   It produces:
   - vklExprProc the kernel representation of the expression 
   - inputs the inputs of the expression (signals, labels)

*/
_VIRTUAL_ boolean SME_Translator::cExpDefEgal(ntrees p, ptlrepid * inputs, ptklExprProc * vklExprProc)
{

  ptlrepid indexList, lparam;
  ntrees vc;
  ptklExprSig vklExprSig1, vklExprSig2;


  indexList = lparam = NILLREPID;
  if (!isChild(1, p))
    return pk_false;
  vklExprSig1 = cExprSig(operande, inputs, &lparam, &indexList);
  if (vklExprSig1 == NILEXPRSIG)
    return pk_false;
  if (!isChild(2, p))
    return pk_false;
  vklExprSig2 = cExprSig(operande, inputs, &lparam, &indexList);
  if (vklExprSig2 == NILEXPRSIG)
    return pk_false;
  *vklExprProc = kl_mkEquation(opdefegal, *inputs, NILLREPID, lparam, p, vklExprSig1, vklExprSig2);
  /* ---- */
  setSMEId(p);
  vc = addClockRelationConnection(p, 1);
  moveAttribute(p, nchild(1, p), opsrc_clockc, opdst_clockc);
  vc = addClockRelationConnection(p, 2);
  moveAttribute(p, nchild(2, p), opsrc_clockc, opdst_clockc);
  return pk_true;
}
