/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor.
   Such an object offers some functionalities (unparse(), print(), fileUnparse()) for printing a Signal abstract tree
   into a text.
   See initPrinter() for the initialisations.
*/
  SSME_Printer::SSME_Printer ()
{
  GeneralInit ();
}


/**
   Destructor.
*/
SSME_Printer::~SSME_Printer ()
{
}

/**
   @copydoc SignalPrinter::fileUnparse
   @param absFileName (out) the absolute name of the generated file: dir/name.sme where dir is the "current" output directory.
*/
_VIRTUAL_ boolean
SSME_Printer::fileUnparse (ntrees p, ptrepbuf cbuf, ptrepstr processName, ptrepstr * fileNamePath)
{

  if (!initPrinter ())
    return pk_false;
  if (!this_intePrinter->open_bufferMakePathName (cbuf, processName, "ssme", fileNamePath))
    return pk_false;
  ssmeTreeRoot = p;
  SSMEBegin ();
  unparse_tree (p, 999999);
  SSMEEnd ();
  intpriclose ();
  return pk_true;
}

/**
   @copydoc SignalPrinter::fileUnparse
   @param absFileName (out) the absolute name of the generated file: dir/name.sme where dir is the "current" output directory.
*/
_VIRTUAL_ boolean
SSME_Printer::fileUnparse (ntrees p, ptrepbuf cbuf, ptrepstr fileNamePath)
{

  if (fileNamePath == nilstr)
    return pk_false;
  if (!initPrinter ())
    return pk_false;
  if (!inteprint_open_buffer (cbuf, fileNamePath))
    return pk_false;
  ssmeTreeRoot = p;
  SSMEBegin ();
  unparse_tree (p, 999999);
  SSMEEnd ();
  intpriclose ();
  return pk_true;
}

/**
   It resets the SME printer: a same SSME_Printer can be used several times.
*/
_VIRTUAL_ void
SSME_Printer::reset ()
{
}
