/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
#include "pKio.h"
#include "common_type.h"
#include "stringmemStr_type.h"
#include "stringmemRegister_type.h"
#include "GHDC_type.h"
#include "kernel_type.h"
#include "polychronyExitStatus.h"
#include "opersignal.h"
#include "tksignal.h"
#include "sysfile_type.h"
#include "phylaSignal.h"
#include "buffer_type.h"
#include "atlist_type.h"
#include "messignalTables.h"
#include "messignal_0.h"
#include "SyndexTree_type.h"
#include "messignal_3.h"
#include "elemTableGenCode_type.h"
#include "AffineClockExportType.h"
#include "TraceabilityManagerExportType.h"
/* @End_import_type@ */
#include "DCG.h"
#include "DCG_basic.h"
#include "DCG_poly.h"
#include "DCG_endo.h"
#include "DCG_bool.h"
#include "DCG_sch.h"
#include "DCG_flat.h"
#include "DCG_Syndex.h"
#include "DCG_Lustre.h"
#include "markServer_g.h"
#include "atfg_g.h"
#include "atlinstances_g.h"
#include "ExprManager.h"
#include "NodeManager.h"
#include "ident_g.h"
#include "DCG2SignalTree.h"
#include "Pruner.h"
#include "ProfileBuilder.h"
#include "InterFormatsTranslator.h"
#include "DCGEndo2DCGBoolTranslator.h"
#include "DCGBool2DCGFlatTranslator.h"
#include "DCGBool2DCGSchTranslator.h"
#include "booleans2events.h"
#include "assertions2events.h"
#include "RunOnMark.h"
#include "separate_selected_others.h"
#include "typeSelection.h"
#include "mkBoolOther.h"
#include "mkBooleanBoundedIntegerOther.h"
#include "mkFiniteTypesOther.h"
#include "mkStatesOther.h"
#include "mkStatesOther.h"
#include "Closure.h"
#include "mkaggregationsAndMultiplexings.h"
#include "TrainClusters.h"
#include "statesVariables.h"
#include "exclusiveClocks.h"
#include "downstreamRetiming.h"
#include "mkDelayTree.h"
#include "StrScanner.h"
#include "AbstractHomomorphism.h"
#include "Homomorphism.h"
#include "Duplicator.h"
#include "SubGraphExtract.h"
#include "Inclusion.h"
#include "Colouring.h"
#include "ColouringGenCode.h"
#include "ColouringAbstraction.h"
#include "ColouringHierarchy.h"
#include "inteprintc.h"
#include "cGenerator.h"
#include "cGeneratorTasks.h"
#include "cGeneratorDist.h"
#include "inteprintcpp.h"
#include "cppGenerator.h"
#include "inteprintjava.h"
#include "javaGenerator.h"
#include "threadsJavaGenerator.h"
#include "DCG2KernelTranslator.h"
#include "atInodesList_g.h"
#include "atlist_g.h"
#include "atclocklist_g.h"
#include "Pragmas_g.h"
#include "PragmasTrees_g.h"
#include "eTypes.h"
#include "atmark_g.h"
#include "atListAttributes_g.h"
#include "atinterfaceGet_g.h"
#include "atinterfaceSet_g.h"
#include "atClosure_g.h"
#include "stringmemStr_g.h"
#include "decsig_g.h"
#include "z3zShaper.h"
#include "z3zAttributes.h"
#include "arrayClockUtl.h"
#include "HierarchyStatistics_g.h"
#include "tree_api.h"
#include "intetreesGarbage_g.h"
#include "signal_treeSet_g.h"
#include "signal_treeGet_g.h"
#include "splitting.h"
#include "intemkIGraph.h"
#include "UtilityClock.h"
#include "predefSignalNames_g.h"
#include "messenger.h"
#include "Signal2Lustre.h"
#include "DCG2LustreTree.h"
#include "LustrePrinter.h"
#include "DCG2dot.h"
#include "CommManager.h"
#include "GraphManager.h"
#include "TraceabilityManager.h"
#include "buffer.h"
#include "messenger.h"
#include "SyndexAttribute.h"
#include "SyndexPrinter.h"
#include "SyndexTreeInterface.h"
#include "DCG2Syndex.h"
#include "Service.h"
#include "SyndexService.h"
#include "XmlDecls.h"
#include "DCG2Geneauto.h"
#include "GeneautoService.h"
#include "SigaliService.h"
#include "FiacreShaper.h"
#include "FiacreService.h"
#include "LustreService.h"
#include "SMTService.h"
#include "DotService.h"
#include "CodeGenerator.h"
#include "CCodeGenerator.h"
#include "CPPCodeGenerator.h"
#include "JavaCodeGenerator.h"
#include "string.h"
#include "stringmemStr_g.h"
#include "stringmemSet_g.h"
#include "stringStd_g.h"
#include "atlist_g.h"
#include "sysfile_g.h"
#include "predefSignalNames_g.h"
#include "ident_g.h"
#include "tree_api.h"
#include "atInodesList_g.h"
#include "decsig_g.h"
#include "signal_tree.h"
#include "TreesAttributes.h"
#include "Pragmas_g.h"
#include "PragmasTrees_g.h"
#include "Trees2Kernel.h"
#include "SME_Translator.h"
#include "SME_SyntaxRestriction.h"
#include "atlrepid_g.h"
#include "TypeManager_g.h"
#include "TypeGarbage_g.h"
#include "IdentifierTable.h"
#include "atpileIni_g.h"
#include "atrepfiGet_g.h"
#include "sigparse_g.h"
#include "SignalPrinter.h"
#include "DCG2SignalTree.h"
#include "SME_Printer.h"
#include "SSME_Printer.h"
#include "pKServicesAPI.h"
#include "inteattr_g.h"
#include "morphismPredefinedNames.h"
#include "kernel2graph.h"
#include "vhdlCodeGenerator.h"
#include "VLCodeGenerator.h"

/* @Begin_prog@ */
#include "pKServicesAPI_m.c"
#include "pKServicesAPI_g.c"
#include "pKServicesAPIGraph_g.c"
