/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/bdd_util.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 02:50:47 $
 *
 */

#include "bdd_export_type.h"
#include <stddef.h>
#include "boolean.h"
#include "BDD_StdTypes.h"
#include "ansi.h"
#include "bdd_export.h"

#include "bdd_int.h"
#include "bdd_int_hashcache.h"
#include "bdd_int_safenodeframe.h"
#include "bdd_int_bddBlock.h"
#include "bdd_int_nodeBlock.h"
#include <stdio.h>
#include "bdd_inner.h"
#include "bdd_int_manager.h"

#include <string.h>
#include <stdlib.h>
#include "util.h"

#include "st.h"
#include "boolean_ops_g.h"
#include "print_stats_g.h"
#include "bdd_util_g.h"
#include "adhoc_cache_g.h"
#include "find_or_add_g.h"
#include "external_ptr_g.h"
#include "garb_collect_g.h"

/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/bdd_util.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 02:50:47 $
 * $Log: bdd_util.c,v $
 * Revision 1.2  1992/09/19  02:50:47  shiple
 * Version 2.4
 * Prefaced compile time debug switches with BDD_. Added typecast to void to some function calls.
 * Subtracted one from return value of bdd_size because it was wrong.  Added get_mem_stats function.
 * In bdd_get_branches, return NIL for then and else pointers if argument is the constant 0.
 * Added bdd_will_exceed_mem_limit function.
 *
 * Revision 1.1  1992/07/29  00:26:55  shiple
 * Initial revision
 *
 * Revision 1.2  1992/05/06  18:51:03  sis
 * SIS release 1.1
 *
 * Revision 1.1  92/01/08  17:34:34  sis
 * Initial revision
 * 
 * Revision 1.1  91/03/27  14:35:36  shiple
 * Initial revision
 * 
 *
 */

#define BDD_ONE_MEGABYTE 1048576

static void get_mem_stats (bdd_manager_pointer manager);

/*
 *    bdd_create_variable - create a new variable and add it to the bdd
 *
 *    A bdd formula of the form (v, 1, 0) is returned.
 *    The 'v' is a unique new variable.
 *
 *    return a bdd_t*
 */
bdd_t_pointer
bdd_create_variable (bdd_manager_pointer manager)
{
  bdd_t_pointer var;

  var = bdd_get_variable (manager, manager->bdd.nvariables++);

  return (var);
}

/*
 *    bdd_dup - return a copy of the given bdd
 *
 *    Actually what this does is allocates a new external
 *    pointer for the user.  It is the user's responsibility
 *    then to manage the allocation and deallocation of
 *    these external pointers.
 *
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_dup (bdd_t_pointer f)
{
  // bdd_manager_pointer manager;
  bdd_t_pointer f1;

  if (f == (NIL_bdd_t_pointer))
    return ((NIL_bdd_t_pointer));	/* robustness */

  //  manager = f->bdd;

  /* WATCHOUT - no safe frame is declared here (b/c its not needed) */

  f1 = bdd_make_external_pointer (f->bdd, f->node);


  return (f1);
}

/*
 *    bdd_free - free the given bdd_t
 *
 *    WATCHOUT - don't call with BDDs assigned to a node -- use bdd_f_noderee.
 *    WATCHOUT - there is no way to assertion-check the above condition either
 *               the only possible error message is core-dumped (later)
 *
 *    return nothing, just do it.
 */
void
bdd_free (bdd_t_pointer f)
{
  // bdd_manager_pointer manager;

  if (f == (NIL_bdd_t_pointer))
    return;			/* for compatibility with v2.2 */

  //  manager = f->bdd;

  (void) bdd_destroy_external_pointer (f);
}

/*
 *    bdd_else - return the else-side of a bdd
 *
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_else (bdd_t_pointer f)
{
  bdd_safeframe frame;
  bdd_safenode reg_f, ret;
  bdd_manager_pointer manager;
  bdd_t_pointer e;

  if (f == (NIL_bdd_t_pointer))
    fail ("bdd_else: invalid BDD");

  if (BDD_IS_CONSTANT (f->bdd, f->node))
    fail ("bdd_else: constant BDD");

  manager = f->bdd;

  /*
   *    After the input is checked for correctness, start the safe frame
   *    f is already an external pointer so it need not be protected
   */
  bdd_safeframe_start (manager, frame);
  bdd_safenode_declare (manager, reg_f);
  bdd_safenode_declare (manager, ret);

  reg_f.node = BDD_REGULAR (f->node);
  ret.node = BDD_IS_COMPLEMENT (f->node) ? BDD_NOT (reg_f.node->E) : reg_f.node->E;

  /*
   *    End the safe frame and return the result
   */
  bdd_safeframe_end (manager);
  e = bdd_make_external_pointer (manager, ret.node);


  return (e);
}

/* 
 *    bdd_then - return the then-side of a bdd
 *    
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_then (bdd_t_pointer f)
{
  bdd_safeframe frame;
  bdd_safenode reg_f, ret;
  bdd_manager_pointer manager;
  bdd_t_pointer t;

  if (f == (NIL_bdd_t_pointer))
    fail ("bdd_then: invalid BDD");

  if (BDD_IS_CONSTANT (f->bdd, f->node))
    fail ("bdd_then: constant BDD");

  manager = f->bdd;

  /*
   *    After the input is checked for correctness, start the safe frame
   *    f is already an external pointer so it need not be protected
   */
  bdd_safeframe_start (manager, frame);
  bdd_safenode_declare (manager, reg_f);
  bdd_safenode_declare (manager, ret);

  reg_f.node = BDD_REGULAR (f->node);
  ret.node = BDD_IS_COMPLEMENT (f->node) ? BDD_NOT (reg_f.node->T) : reg_f.node->T;

  /*
   *    End the safe frame and return the result
   */
  bdd_safeframe_end (manager);
  t = bdd_make_external_pointer (manager, ret.node);


  return (t);
}

/*
 *    bdd_get_branches - get the branches of a node
 *
 *    return nothing, just do it.
 */
void
bdd_get_branches (bdd_node_pointer f, bdd_node_pointer * f_T, bdd_node_pointer * f_E)
		/* return */
		/* return */
{
  bdd_node_pointer reg_f;
  Integer is_constant;

  reg_f = BDD_REGULAR (f);

  /* 
   * If f is the constant 0, we just want to return NIL for the then 
   * and else branches.  Since we can't access the manager from this
   * routine, we simply check the id of the node.
   */
  is_constant = (reg_f->id == BDD_ONE_ID) ? pk_true : pk_false;

  if (BDD_IS_COMPLEMENT (f) && !(is_constant))
    {
      *f_T = BDD_NOT (reg_f->T);
      *f_E = BDD_NOT (reg_f->E);
    }
  else
    {
      *f_T = reg_f->T;
      *f_E = reg_f->E;
    }
}

/* 
 *    bdd_top_var - return the top variable of the bdd
 *    
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_top_var (bdd_t_pointer f)
{
  bdd_safeframe frame;
  bdd_safenode reg_f, ret;
  bdd_t_pointer v;
  bdd_manager_pointer manager;

  if (f == (NIL_bdd_t_pointer))
    fail ("bdd_top_var: invalid BDD");

  if (f->node == BDD_ONE (f->bdd))
    return (bdd_one (f->bdd));

  if (f->node == BDD_ZERO (f->bdd))
    return (bdd_zero (f->bdd));

  manager = f->bdd;

  /*
   *    After the input is checked for correctness, start the safe frame
   *    f and g are already external pointers so they need not be protected
   */
  bdd_safeframe_start (manager, frame);
  bdd_safenode_declare (manager, reg_f);
  bdd_safenode_declare (manager, ret);

  reg_f.node = BDD_REGULAR (f->node);
  ret.node = bdd_find_or_add (manager, reg_f.node->id, BDD_ONE (manager), BDD_ZERO (manager));

  /*
   *    End the safe frame and return the result
   */
  bdd_safeframe_end (manager);
  v = bdd_make_external_pointer (manager, ret.node);

  return (v);
}

/*
 *    bdd_get_variable - get or create a new variable and add it to the bdd
 *
 *    The variable had better be in the range 0 .. nvariables else a failure occurs.
 *    A bdd of the form (v, 1, 0) is returned.   This is the only acceptable way of
 *    getting or creating variables in the manager.   It should be noted that the
 *    value of bdd.nvariables was established in bdd_start at manager initialization
 *    time and is not changed over the life of the manager.  This is for reasons of
 *    strong checking; ensure that the user doesn't wildly create variables.
 *
 *    return a bdd_t*
 */
bdd_t_pointer
bdd_get_variable (bdd_manager_pointer manager, bdd_variableId variable_i)
{
  bdd_safeframe frame;
  bdd_safenode v;
  bdd_t_pointer var;

  bdd_safeframe_start (manager, frame);
  bdd_safenode_declare (manager, v);

  v.node = bdd_find_or_add (manager, variable_i, BDD_ONE (manager), BDD_ZERO (manager));

  bdd_safeframe_end (manager);

  var = bdd_make_external_pointer (manager, v.node);

  return (var);
}

/*
 *    bdd_get_stats - get the statistics from the bdd package
 *
 *    return nothing, just do it.
 */
static void
bdd_get_stats (bdd_manager_pointer manager, bdd_stats * stats)
			/* return */
{
  (void) get_mem_stats (manager);
  *stats = manager->heap.stats;
}

/*
 *    bdd_size - return the size of the bdd (in nodes)
 *
 *    Compute the number of nodes in the bdd by traversal (e.g counting)
 *
 *    The convention is that the one constant counts as 1
 *    while the zero constant counts as 0 (b/c it is derived
 *    from the one constant).
 *
 *    return the count (under the above counting convention)
PLG REMOVE bdd_size(bdd_t_pointer f)
PLG REMOVE bdd_fail
 */

/*
 *    bdd_memfail - declare a memory allocation failure
 *
 *    PrInteger out the statistics, so the guy knows why he died!
 *
 *    don't return, just do it
 */
void
bdd_memfail (bdd_manager_pointer manager, char *function)
			/* may be (NIL_bdd_manager_pointer) in case of bdd_start() */
{
  bdd_stats stats;

  /*
   *    WATCHOUT - assume NOTHING about the manager except that you can
   *    go in and recover the statistics from it.  During garbage_collection
   *    the manager is in a state if semi-inconsistence, so it is not
   *    possible to do many operations on it.
   */

  /*
   *    If there was a bdd_manager, then get some stats before death
   */
  if (manager != (NIL_bdd_manager_pointer))
    {
      (void) fprintf (stderr, "\nBDD Manager Death Statistics\n\n");
      (void) bdd_get_stats (manager, &stats);
      (void) bdd_print_stats (stats, stderr);
    }

  (void) fprintf (stderr, "Fatal error: memory allocation failure in %s\n", function);
  (void) fflush (stdout);
  exit (1);
}

/*
 * get_mem_stats  
 *
 * Compute the memory used by the BDD manager and fill in the appropriate 
 * fields of the stats.memory data structure.
 */
static void
get_mem_stats (bdd_manager_pointer manager)
{
  bdd_stats *stats;

  /*
   * Get the current stats for the manager.
   */
  stats = &manager->heap.stats;

  /*
   * Update last_sbrk.
   */
  stats->memory.last_sbrk = (Integer) malloc (0);

  /*
   * Memory for the manager itself.
   */
  stats->memory.manager = sizeof (bdd_manager);

  /*
   * Memory used by bdd_nodes.  Multiply by 2 since there are two halfspaces.
   */
  stats->memory.nodes = (stats->blocks.total * 2) * sizeof (bdd_nodeBlock);

  /*
   * Memory used by the unique table.
   */
  stats->memory.hashtable = manager->heap.hashtable.nbuckets * sizeof (bdd_node_pointer *);

  /*
   * Memory used by external pointers.
   */
  stats->memory.ext_ptrs = stats->extptrs.blocks * sizeof (bdd_bddBlock);

  /*
   * Memory used by ITE table entries and ITE table buckets.
   */
  stats->memory.ITE_cache = (manager->heap.cache.itetable.nbuckets * sizeof (bdd_hashcache_entry *))
    + (manager->heap.cache.itetable.nentries * sizeof (bdd_hashcache_entry));

  /*
   * Memory used by consttable entries and consttable buckets.
   */
  stats->memory.ITE_const_cache = (manager->heap.cache.consttable.nbuckets * sizeof (bdd_constcache_entry *))
    + (manager->heap.cache.consttable.nentries * sizeof (bdd_constcache_entry));

  /*
   * Memory used by the adhoc table: table + buckets + entries.
   */
  if (manager->heap.cache.adhoc.table == (NIL_st_table))
    {
      stats->memory.adhoc_cache = 0;
    }
  else
    {
      stats->memory.adhoc_cache = (sizeof (st_table)
				   + (manager->heap.cache.adhoc.table->num_bins * sizeof (st_table_entry *))
				   + (st_count (manager->heap.cache.adhoc.table) * (sizeof (st_table_entry) + sizeof (bdd_adhoccache_key))));
    }

  /*
   * Total memory used.
   */
  stats->memory.total = stats->memory.manager + stats->memory.nodes
    + stats->memory.hashtable + stats->memory.ext_ptrs + stats->memory.ITE_cache + stats->memory.ITE_const_cache + stats->memory.adhoc_cache;

}


/*
 * bdd_will_exceed_mem_limit  
 *
 * Test to see if allocating alloc_size will cause the manager's memory limit to
 * be exceeded.  Return pk_true if limit will be exceeded, else pk_false. If memory limit
 * will be exceeded and if call_daemon is pk_true, then don't return; instead, call the daemon.
 */
boolean
bdd_will_exceed_mem_limit (bdd_manager_pointer manager, Integer alloc_size, Integer call_daemon)
{

  /*
   * If there is no memory limit, then return pk_false.
   */
  if (manager->memory.limit == BDD_NO_LIMIT)
    {
      return pk_false;
    }

  /* 
   * Compute the memory statistics for the manager. TODO: if get_mem_stats is taking too much time,
   * then we should update the memory stats incrementally each time we do an allocation.
   */
  (void) get_mem_stats (manager);

  /*
   * If allocating alloc_size bytes will not cause the limit to be exceeded, then return pk_false.
   */
  if ((manager->heap.stats.memory.total + alloc_size) < (manager->memory.limit * BDD_ONE_MEGABYTE))
    {
      return pk_false;
    }
  /*
   * Allocating alloc_size bytes will cause the memory limit to be exceeded.  If call_daemon is
   * pk_false, then just return pk_true.  Else, clean up, and then call the daemon.
   */
  if (call_daemon == pk_false)
    {
      return pk_true;
    }
  /*
   * Reset the safeframe pointer of the manager.  All of the data structures associated with the
   * safeframe mechanism are automatic (e.g. stack) variables which are on the stack somewhere.
   * Thus, we simply need to tell the manager that there are no safeframes.
   */
  manager->heap.internal_refs.frames = (NIL_bdd_safeframe);

  /*
   * Free the adhoc table if it is in use.
   */
  if (manager->heap.cache.adhoc.table != (NIL_st_table))
    {
      (void) bdd_adhoccache_uninit (manager);
    }

  /*
   * Call the garbage collector to get rid of the current computation, and to try
   * to free up additional space.
   */
  (void) bdd_garbage_collect (manager);

  /*
   * Call the daemon specified by the application.
   */
  if (manager->memory.daemon != ((void *) NULL))
    {
      (*(manager->memory.daemon)) (manager);
    }
  else
    {
      fail ("bdd_will_exceed_mem_limit: memory limit set, but no daemon registered");
    }
  return pk_true;		/* PLG daemon and fail should return here */
}

/* bdd_t_pointer is typedef struct __bdd_t* defined in bdd_export
 * (bdd)->node is bdd_node_pointer
 *  bdd_node_pointer is struct __bdd_node* defined in bdd_export_define
 */
extern bdd_variableId
topvar_id (bdd_t_pointer bdd)
{
  /* WARNING : access to the structure of a bdd.  */
  return BDD_REGULAR (bdd->node)->id;
}

extern bdd_variableId
lastvar_id (bdd_t_pointer bdd)
{
  /* WARNING : access to the structure of a bdd.  */
  return BDD_REGULAR (bdd->node)->last;
}
