/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/boolean_ops.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 01:47:17 $
 *
 */

#include <stddef.h>
#include "boolean.h"
#include "BDD_StdTypes.h"
#include "bdd_export_type.h"
#include "ansi.h"
#include "bdd_export.h"

#include "bdd_int_hashcache.h"

#include "bdd_int_safenodeframe.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "util.h"

#include "bdd_int.h"
#include "bdd_int_bddBlock.h"
#include "bdd_inner.h"
#include "bdd_int_manager.h"

#include "ite_g.h"
#include "external_ptr_g.h"
#include "ite_constant_g.h"

/* HERE */


/*
 * Revision Control Information
 *
 * $Source: /vol/hyper/hyper3/shiple/sis/bdd/RCS/boolean_ops.c,v $
 * $Author: shiple $
 * $Revision: 1.2 $
 * $Date: 1992/09/19 01:47:17 $
 * $Log: boolean_ops.c,v $
 * Revision 1.2  1992/09/19  01:47:17  shiple
 * Version 2.4
 * Prefaced compile time debug switches with BDD_.
 *
 * Revision 1.1  1992/07/29  00:26:56  shiple
 * Initial revision
 *
 * Revision 1.2  1992/05/06  18:51:03  sis
 * SIS release 1.1
 *
 * Revision 1.1  92/01/08  17:34:34  sis
 * Initial revision
 * 
 * Revision 1.1  91/04/11  20:59:01  shiple
 * Initial revision
 * 
 *
 */

/*
 *    bdd_and - returns the and of two bdd's
 *
 *    bdd_ITE Identity: F * G  =  ITE(F, G, 0)
 *
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_and (bdd_t_pointer f, bdd_t_pointer g, boolean f_phase, boolean g_phase)
				/* if ! f_phase then negate f */
				/* if ! g_phase then negate g */
{
  bdd_safeframe frame;
  bdd_safenode real_f, real_g, ret;
  bdd_manager_pointer manager;
  bdd_t_pointer h;

  if (f == (NIL_bdd_t_pointer) || g == (NIL_bdd_t_pointer))
    fail ("bdd_and: invalid BDD");

  if (f->bdd != g->bdd)
    fail ("bdd_and: different bdd managers");

  manager = f->bdd;		/* either this or g->bdd will do */

  /*
   *    After the input is checked for correctness, start the safe frame
   *    f and g are already external pointers so they need not be protected
   */
  bdd_safeframe_start (manager, frame);
  bdd_safenode_declare (manager, real_f);
  bdd_safenode_declare (manager, real_g);
  bdd_safenode_declare (manager, ret);

  real_f.node = f_phase ? f->node : BDD_NOT (f->node);
  real_g.node = g_phase ? g->node : BDD_NOT (g->node);

  ret.node = bdd__ITE_ (manager, real_f.node, real_g.node, BDD_ZERO (manager));

  /*
   *    End the safe frame and return the result
   */
  bdd_safeframe_end (manager);
  h = bdd_make_external_pointer (manager, ret.node);

  return (h);
}

/*
 *    bdd_not - returns the not of the given bdd.
 *
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_not (bdd_t_pointer f)
{
  bdd_t_pointer g;

  if (f == (NIL_bdd_t_pointer))
    fail ("bdd_not: invalid BDD");

  //  manager = f->bdd;

  /* WATCHOUT - no safe frame is declared here (b/c its not needed) */

  g = bdd_make_external_pointer (f->bdd, BDD_NOT (f->node));

  return (g);
}

/*
 *    bdd_or - returns the or of two bdd's
 *
 *    bdd_ITE Identity:  F + G  =  ITE(F, 1, G)
 *
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_or (bdd_t_pointer f, bdd_t_pointer g, boolean f_phase, boolean g_phase)
				/* if ! f_phase then negate f */
				/* if ! g_phase then negate g */
{
  bdd_safeframe frame;
  bdd_safenode real_f, real_g, ret;
  bdd_manager_pointer manager;
  bdd_t_pointer h;

  if (f == (NIL_bdd_t_pointer) || g == (NIL_bdd_t_pointer))
    fail ("bdd_or: invalid BDD");

  if (f->bdd != g->bdd)
    fail ("bdd_or: different bdd managers");

  manager = f->bdd;		/* either this or g->bdd will do */

  /*
   *    After the input is checked for correctness, start the safe frame
   *    f and g are already external pointers so they need not be protected
   */
  bdd_safeframe_start (manager, frame);
  bdd_safenode_declare (manager, real_f);
  bdd_safenode_declare (manager, real_g);
  bdd_safenode_declare (manager, ret);

  real_f.node = f_phase ? f->node : BDD_NOT (f->node);
  real_g.node = g_phase ? g->node : BDD_NOT (g->node);

  ret.node = bdd__ITE_ (manager, real_f.node, BDD_ONE (manager), real_g.node);

  /*
   *    End the safe frame and return the result
   */
  bdd_safeframe_end (manager);
  h = bdd_make_external_pointer (manager, ret.node);

  return (h);
}

/*
 *    bdd_one - return a new copy of the 'one' constant
 *
 *    return the new bdd (external pointer)
 */
bdd_t_pointer
bdd_one (bdd_manager_pointer manager)
{
  bdd_t_pointer one;

  if (manager == (NIL_bdd_manager_pointer))
    fail ("bdd_one: bad bdd manager");

  /* WATCHOUT - no safe frame is declared here (b/c its not needed) */

  one = bdd_make_external_pointer (manager, BDD_ONE (manager));

  return (one);
}

/*
 *    bdd_zero - return a new copy of the 'zero' constant
 *
 *    return the new manager (external pointer)
 */
bdd_t_pointer
bdd_zero (bdd_manager_pointer manager)
{
  bdd_t_pointer zero;

  if (manager == (NIL_bdd_manager_pointer))
    fail ("bdd_zero: bad bdd manager");

  /* WATCHOUT - no safe frame is declared here (b/c its not needed) */

  zero = bdd_make_external_pointer (manager, BDD_ZERO (manager));

  return (zero);
}

/* 
 *    bdd_equal - whether the two bdds are the same
 *
 *    return {pk_true, pk_false} on {is, is not}.
 */
boolean
bdd_equal (bdd_t_pointer f, bdd_t_pointer g)
{
  // bdd_manager_pointer manager;
  boolean ret;

  if (f == (NIL_bdd_t_pointer) || g == (NIL_bdd_t_pointer))
    fail ("bdd_equal: invalid BDD");

  if (f->bdd != g->bdd)
    fail ("bdd_equal: different bdd managers");

  //  manager = f->bdd;		/* or g->bdd */

  /* WATCHOUT - no safe frame is declared here (b/c its not needed) */

  ret = f->node == g->node ? pk_true : pk_false;

  return (ret);
}
