/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/****************************************************************************/
extern form_type
form_create (void)
{
  form_type result;
  result = MALLOC (form_type, sizeof (form_struct));
  ASSERT_MALLOC_OK (result);
  return result;
}

/****************************************************************************/
extern void
form_destroy (form_type form)
{
  switch (form->op)
    {
    case FORM_AND:
    case FORM_OR:
    case FORM_DIFF:
      {
	form_destroy (form->what.args.form1);
	form_destroy (form->what.args.form2);
	FREE (form);
	break;
      }
    default:
      /* we do not destroy literals or the null formula */
      break;
    }
}

/****************************************************************************/
extern void
form_set_op (form_type form, op_type op)
{
  form->op = op;
}

/****************************************************************************/
extern void
form_set_args (form_type form, form_type form1, form_type form2)
{
  form->what.args.form1 = form1;
  form->what.args.form2 = form2;
}

/****************************************************************************/
extern void
form_set_literal (form_type form, literal_type literal)
{
  form->what.literal = literal;
}

/****************************************************************************/
extern op_type
form_get_op (form_type form)
{
  return form->op;
}

/****************************************************************************/
extern void
form_get_args (form_type form, form_type * form1, form_type * form2)
{
  *form1 = form->what.args.form1;
  *form2 = form->what.args.form2;
}

/****************************************************************************/
extern literal_type
form_get_literal (form_type form)
{
  return form->what.literal;
}

/****************************************************************************/
extern form_type
form_clone (form_type form)
{
  form_type result, aux1, aux2;
  if (form->op == FORM_LIT)
    return form;
  result = MALLOC (form_type, sizeof (form_struct));
  ASSERT_MALLOC_OK (result);
  form_get_args (form, &aux1, &aux2);
  result->op = form->op;
  result->what.args.form1 = form_clone (aux1);
  result->what.args.form2 = form_clone (aux2);
  return result;
}

/****************************************************************************/
extern form_type
form_one (void)
{
  return &one;
}

/****************************************************************************/
extern form_type
form_zero (void)
{
  return &zero;
}

/****************************************************************************/
extern boolean
form_equal (form_type f1, form_type f2)
{
  if (f1->op != f2->op) return pk_false;
  if (f1->op == FORM_LIT) return (f1->what.literal == f2->what.literal);
  return ((f1->what.args.form1 == f2->what.args.form1) && (f1->what.args.form2 == f2->what.args.form2));
}
