_PUBLIC_
/* [NULL_char] is the character that terminates ANSI C char *
 * [Character] is an ANSI C char but [NULL_char]
 * [White] is an ANSI C white space but [eol] (' ' '\v' '\f' '\t')
 */
/* =============================================================================
 * elementary scanning functions
 */
/* 2004-11-03
   this remains unchanged
   success: string is open, not completed, cursor position is 0
   failure: string is not open or is completed or cursor is not at first position 
 */
  boolean LineAnalyzer::is_start ()
{
  return isSTARTPOS ();
}


/* 2004-11-03
   this remains unchanged
   success: BUFFER = [Character*] . [Character+][NULL_char]
   *            or [Character+] . [NULL_char]
            returns cursor position in BUFFER
   failure: this is not open or this is [NULL_char]
            returns -1
 */
int
LineAnalyzer::get_cursor ()
{
  if (isNOTEMPTY ())
    return thisLineCursor;
  return -1;
}

/* 2004-11-03
 * this remains unchanged
 * success: BUFFER = [Character*] . [Character+][NULL_char]
 * failure: this is closed or current character [NULL_char]
 */
boolean
LineAnalyzer::is_notCompleted ()
{
  return isNOTCOMPLETED ();
}

/* 2004-11-03
 * this remains unchanged
 * success: this is not open or BUFFER = [Character+] . [NULL_char]
 * failure: BUFFER = [Character*] . [Character+] [NULL_char]
 */
boolean
LineAnalyzer::is_lineCompleted ()
{
  return isCOMPLETED ();
}

/* 2004-11-03
   this remains unchanged
   success: BUFFER = [Character*] . <cc>[Character*][NULL_char]
   failure: string is not open or is completed or current character is not <cc>
 */
boolean
LineAnalyzer::is_ (char cc)
{
  return isCURRENT (cc);
}

/* 2004-10-14
   this remains unchanged
   success: BUFFER = . <cc>[Character*][NULL_char]
   failure: string is not open or is completed or cursor is not at first position or current character is not <cc>
 */
boolean
LineAnalyzer::is_start (char cc)
{
  return isSTART (cc);
}

/* 2004-10-14
 * this remains unchanged
 * success: BUFFER = [Character*] . <cc>[Character*][NULL_char], <cc> belongs to <charSet>
 * failure: this is not open or this is completed or 
 *          current character does not belong to <charSet> (or <charSet> is NIL_String)
 */
boolean
LineAnalyzer::is_in (const char *charSet)
{
  if ((isCOMPLETED ()) || (charSet == NIL_String))
    return pk_false;
  return ((strchr (charSet, unsafe_getCURRENTCHAR ())) != NIL_String);
}

/* 2004-11-03
 * this remains unchanged
 * success: BUFFER = [Character*] . [Character] <cc> [Character*][NULL_char] and <cc> is not NULL_char
 * failure: this is not open or <cc> is NULL_char
 *          or BUFFER = [Character*] . opt{[Character_but_<cc>][Character*]}[NULL_char]
 */
boolean
LineAnalyzer::is_forward (char cc)
{
  return isFORWARD (cc);
}

boolean
LineAnalyzer::is_ (const char *str)
{
  if (str == NIL_String)
    return pk_true;
  if (isCOMPLETED ())
    return pk_false;
  return isstrprefix (str, (thisLineBuffer + thisLineCursor));
}

/* 2004-11-03
 * this remains unchanged
 * success: BUFFER = [Character*] . [Character1] [Character2] [Character*][NULL_char]
 *          and [Character2] belongs to <charSet> (then <charSet> is not NIL_String)
 * failure: this is not open or <charSet> is NIL_String
 *          or BUFFER = [Character*] . [NULL_char]
 *          or BUFFER = [Character*] . [Character] [Character*][NULL_char] 
 *              and first character in [Character*][NULL_char] does not belong to <charSet>
 */
boolean
LineAnalyzer::is_forwardIn (const char *charSet)
{
  if ((charSet != NIL_String) && (isNOTLASTPOSITION ()))
    return (strchr (charSet, unsafe_getFORWARDCHAR ()) != NIL_String);
  return pk_false;
}

/* 2004-11-02
   this remains unchanged
   success: BUFFER = [Character*] . [Character] [Character] [Character*][NULL_char]
            returns second [Character] 
   failure: this is not open or BUFFER = [Character*] . {[Character]} [NULL_char]
            returns NULL_char
 */
char
LineAnalyzer::get_forward ()
{
  if (isNOTLASTPOSITION ())
    return unsafe_getFORWARDCHAR ();
  return NULL_char;
}

/* 2004-11-03
 * this remains unchanged
 * success: this is [Character+] [NULL_char]
 *          duplicates [Character+] in a new allocated string
 * failure: this is not open or this is [NULL_char]
 *          returns NIL_String
 */
char *
LineAnalyzer::strdupWholeString ()
{
  if (isNOTEMPTY ())
    return newCstr_String (thisLineBuffer);
  return NIL_String;
}

/* 2004-11-03
 * this remains unchanged
 * NOT STANDARD PKString manangement
 * success: BUFFER = [Character*] . [Character+] [NULL_char]
 *          duplicates [Character+] in a new allocated string
 * failure: this is not open or current character is [NULL_char]
 *          returns NIL_String
 */
char *
LineAnalyzer::strdupRemainingString ()
{
  if (isNOTCOMPLETED ())
    return newCstr_String ((thisLineBuffer + thisLineCursor));
  return NIL_String;
}

/* =============================================================================
 * stderr elementary printing
 */

/* 2004-12-08
 * this remains unchanged
 * output stderr buffer
 */
void
LineAnalyzer::stderr_ln ()
{
  fprintf (stderr, "\n");
}

/* 2004-12-08
 * this remains unchanged
 * appends <cc> to stderr buffer
 */
void
LineAnalyzer::stderr_ (char cc)
{
  fprintf (stderr, "%c", cc);
}

/* 2004-12-08
 * this remains unchanged
 * appends <cc> followed by space to stderr buffer
 */
void
LineAnalyzer::stderr_sp (char cc)
{
  fprintf (stderr, "%c ", cc);
}

/* 2004-12-08
 * this remains unchanged
 * appends <cc> to stderr buffer and output stderr
 */
void
LineAnalyzer::stderr_ln (char cc)
{
  fprintf (stderr, "%c\n", cc);
}

/* 2004-12-08
 * this remains unchanged
 * appends <str> to stderr buffer
 */
void
LineAnalyzer::stderr_ (const char *str)
{
  if (str != NIL_String)
    fprintf (stderr, "%s", str);
}

/* 2004-12-08
 * this remains unchanged
 * appends <str> followed by space to buffer 
 */
void
LineAnalyzer::stderr_sp (const char *str)
{
  if (str != NIL_String)
    fprintf (stderr, "%s ", str);
}

/* 2004-12-08
 * this remains unchanged
 * appends <str> stderr buffer and output stderr
 */
void
LineAnalyzer::stderr_ln (const char *str)
{
  if (str == NIL_String)
    stderr_ln ();
  else
    fprintf (stderr, "%s\n", str);
}

/* 2004-12-08
 * this remains unchanged
 * appends <n> <cc> occurrences to stderr buffer
 */
void
LineAnalyzer::stderr_NChars_ (int n, char cc)
{
  int i;
  for (i = 1; i <= n; i = i + 1)
    fprintf (stderr, "%c", cc);
}

/* =============================================================================
 * these attributes stderr printing
 */

/* 2004-12-08
 * this remains unchanged
 * appends this buffer to stderr buffer and output stderr
 */
void
LineAnalyzer::stderrWhole_ln ()
{
  stderr_ln (thisLineBuffer);
}

/* 2004-12-08
 * this remains unchanged
 * appends not consumed content of this buffer to stderr buffer and output stderr
 */
void
LineAnalyzer::stderrTail_ln ()
{
  stderr_ln (unsafe_getTAIL ());
}

/* 2004-12-08
 * this remains unchanged
 * appends this atom content to stderr buffer
 */
void
LineAnalyzer::stderrAtom_ ()
{
  stderr_ (thisAtomBuffer);
}

/* 2004-12-08
 * this remains unchanged
 * appends this content to stderr buffer and output stderr
 * appends this cursor state to stderr buffer and output stderr
 */
void
LineAnalyzer::stderr_CharIn_ln ()
{
  stderr_ ('>');
  if (isNOTEMPTY ())
    {
      stderrWhole_ln ();
      stderr_ ('>');
      if (isNOTCOMPLETED ())
	{
	  stderr_NChars_ (thisLineCursor, ' ');
	  stderr_ln (unsafe_getCURRENTCHAR ());
	}
      else
	stderr_ln (" line is completed");
    }
  else
    stderr_ln (" line is empty");
}

/* 2004-12-08
 * this remains unchanged
 * appends cursor position followed by space to stderr buffer 
 */
void
LineAnalyzer::stderr_Column_sp (const char *strMess)
{
  fprintf (stderr, "==> %ccolumn %i %s ", '/', thisLineCursor + 1, strMess);
}

/* 2004-12-08
 * this remains unchanged
 * appends cursor position followed by <strMess> to stderr buffer and output stderr
 * appends this content to stderr buffer and output stderr
 * appends this cursor state to stderr buffer and output stderr
 */
void
LineAnalyzer::stderr_Column_ln (const char *strMess)
{
  stderr_Column_sp (strMess);
  stderr_ln ();
  stderr_CharIn_ln ();
}

/* 2004-12-08
 * this remains unchanged
 * appends cursor position followed by <strMess1>, followed by <strMess2> to stderr buffer and output stderr
 * appends this content to stderr buffer and output stderr
 * appends this cursor state to stderr buffer and output stderr
 */
void
LineAnalyzer::stderr_Column_ln (const char *strMess1, const char *strMess2)
{
  stderr_Column_sp (strMess1);
  stderr_ln (strMess2);
  stderr_CharIn_ln ();
}

/* 2004-12-08
 * this remains unchanged
 * prints in stderr debugging informations and output stderr
 * prints in stderr "unexpected ... in", followed by space, followed by <mess> in stderr and output stderr
 */
void
LineAnalyzer::stderr_unexpected_ln (const char *mess)
{
  if (isNOTCOMPLETED ())
    stderr_Column_ln ("unexpected char in", mess);
  else
    {
      stderr_sp ("unexpected end of line in");
      stderr_ln (mess);
    }
}

/* =============================================================================
 * these attributes file printing
 */

/* 2004-12-08
 * this remains unchanged
 * target is <outBuffer> or by default to stderr buffer 
 * appends this buffer to target
 * output target
 */
void
LineAnalyzer::printWhole_ln (ref2Buffer outBuffer)
{
  if (outBuffer == NIL_Buffer)
    stderrWhole_ln ();
  else
    outBuffer->line (thisLineBuffer);
}

/* 2004-12-08
 * this remains unchanged
 * target is <outBuffer> or by default to stderr buffer 
 * appends this not consumed content to target
 * output target
 */
void
LineAnalyzer::printTail_ln (ref2Buffer outBuffer)
{
  if (outBuffer == NIL_Buffer)
    stderrTail_ln ();
  else
    outBuffer->line (unsafe_getTAIL ());
}

/* 2004-12-08
 * this remains unchanged
 * target is <outBuffer> or by default to stderr buffer 
 * appends this atom content to target
 * output target
 */
void
LineAnalyzer::printAtom_ (ref2Buffer outBuffer)
{
  if (outBuffer == NIL_Buffer)
    stderrAtom_ ();
  else
    outBuffer->print (thisAtomBuffer);
}

/* 2004-12-08
 * this remains unchanged
 * outBuffer is not NIL (not checked)
 * appends this content to <outBuffer>
 * output <outBuffer>
 * appends this cursor state to <outBuffer> buffer and output <outBuffer>
 */
void
LineAnalyzer::ln_traceBuffer_ln (ref2LogPrinter outBuffer)
{
  //  fprintf(stderr,"DEBUG LineAnalyzer::ln_traceBuffer_ln\n");
  outBuffer->freeze_indent (pk_true);
  outBuffer->line ();
  if (isNOTEMPTY ())
    {
      if (isNOTCOMPLETED ())
	{
	  // const char *currentMark =outBuffer->setLeftMark (">"); 
	  outBuffer->setLeftMark (">");
	  outBuffer->line (thisLineBuffer);
	  outBuffer->printNChars (' ', thisLineCursor);
	  outBuffer->print (unsafe_getCURRENTCHAR ());
	  outBuffer->line_removeLeftMark ();
	}
      else
	{
	  outBuffer->noIndent_ln ("> line is completed");
	}
    }
  else
    {
      outBuffer->noIndent_ln ("> line is empty");
    }
  outBuffer->release_indent (pk_false);
}

/* 2004-12-08
 * this remains unchanged
 * target is <outBuffer> or by default to stderr buffer 
 * appends this content to target
 * output target
 * appends this cursor state to target buffer and output target
 */
void
LineAnalyzer::ln_traceLine_ln (ref2LogPrinter outBuffer)
{
  if (outBuffer == NIL_LogPrinter)
    {
      stderr_ln ();
      stderr_CharIn_ln ();
    }
  else
    ln_traceBuffer_ln (outBuffer);
}

/* 2004-12-08
 * this remains unchanged
 * target is <outBuffer> or by default to stderr buffer 
 * prints in target debugging informations and output target
 * prints in target "unexpected ... in", followed by space, followed by <mess> in target and output target
 */
void
LineAnalyzer::logError_unexpected (ref2LogPrinter outBuffer, const char *mess)
{
  if (outBuffer == NIL_LogPrinter)
    {
      stderr_ln ();
      stderr_unexpected_ln (mess);
    }
  else
    {
      outBuffer->line ();
      ln_traceBuffer_ln (outBuffer);
      if (isNOTCOMPLETED ())
	outBuffer->errorColumn_sp (thisLineCursor + 1, "unexpected char in");
      else
	outBuffer->error__sp ("unexpected end of line in");
      outBuffer->line (mess);
    }
}

/* =============================================================================
 */
/* =============================================================================
 * standard scanning functions
 */

/* 2004-11-03
 * this remains unchanged
 * success:  <cc> is not NULL_char and BUFFER = [Character*] . [Character*] <cc> [Character*][NULL_char]
 *              where [Character*] does not contain <cc> occurrence,
 *          returns the length of [Character*]
 * failure: <cc> is NULL_char or BUFFER = [Character*] . [Character*][NULL_char] 
 *            where [Character*][NULL_char] does not contain <cc> occurrence
 *          returns -1
 */
int
LineAnalyzer::prefixLengthUntil (char cc)
{
  if ((cc == NULL_char) || (isCOMPLETED ()))
    return -1;
  return prefixLengthUntilChar (unsafe_getTAIL (), cc);
}

/* 2004-11-03
 * this remains unchanged
 * success:  <cc1> and <cc2> are not NULL_char and BUFFER = [Character*] . [Character*] <cc1> <cc2> [Character*][NULL_char]
 *              where [Character*] does not contain occurrence of <cc1> <cc2> sequence
 *          returns the length of [Character*]
 * failure: <cc1> or <cc2> is NULL_char or BUFFER = [Character*] . [Character*][NULL_char] 
 *            where [Character*][NULL_char] does not contain occurrence of <cc1> <cc2> sequence
 *          returns -1
 */
int
LineAnalyzer::prefixLengthUntil (char cc1, char cc2)
{
  if ((cc1 == NULL_char) || (cc2 == NULL_char) || (isCOMPLETED ()))
    return -1;
  return prefixLengthUntilCharChar (unsafe_getTAIL (), cc1, cc2);
}

/* =============================================================================
 * standard step forward functions; this may become completed after processing
 */

/* 2004-10-14
 * this atom remains unchanged
 * cursor is moved to current position + 1 if new position is an index in the string
 *              to strlen otherwise
 * cursor position <= thisLineActualLength
 * cursor position < 0 implies thisLineActualLength < 0
 */
void
LineAnalyzer::step_ ()
{
  if (thisLineCursor < thisLineActualLength)
    thisLineCursor = thisLineCursor + 1;
  else
    thisLineCursor = thisLineActualLength;
}

/* 2004-10-14
 * set line scanning terminated
 * this atom remains unchanged
 * cursor is moved to first [NULL_char] occurrence in this line
 * cursor position == thisLineActualLength
 */
void
LineAnalyzer::set_completed ()
{
  thisLineCursor = thisLineActualLength;
}

/* 2004-11-03
 * this atom remains unchanged
 * cursor position is scanned
 * success: BUFFER = [Character*] . [White*] # [Character_but_White][Character*][NULL_char]
 * failure: BUFFER = [Character*] . [White*] # [NULL_char]
 */
boolean
LineAnalyzer::step_whStar_nWh ()
{
  if (isNOTCOMPLETED ())
    {
      SKIP_WHITEsTAR ();
      return isNOTCOMPLETED ();
    }
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * cursor position is scanned
 * success: BUFFER = [Character*] . [White*] # [Character_but_White][Character*][NULL_char]
 * failure: BUFFER = [Character*] . [White*] # [NULL_char]
 */
boolean
LineAnalyzer::step_whStar_nWh (int *whiteNumber)
{
  if (isNOTCOMPLETED ())
    {
      (*whiteNumber) = strspn (unsafe_getTAIL (), WHITESPACE);
      thisLineCursor = thisLineCursor + (*whiteNumber);
      return isNOTCOMPLETED ();
    }
  return pk_false;
}

/* 2004-11-09
 * this atom remains unchanged
 * success: BUFFER = [Character*] . [White+] [Character_but_White][Character*]
 *             cursor is moved to [Character_but_White] starting [Character_but_White][Character*]
 *             <isEol> is set FALSE;
 *          BUFFER = [Character*] . [White] [WhiteString]
 *             cursor is moved to next NULL_char
 *             <isEol> is set TRUE;
 * failure: BUFFER = [Character*] . [Character_but_White][Character*]
 *             <isEol> is set FALSE;
 *          current character is NULL_char
 *             <isEol> is set TRUE;
 *          this remains unchanged
 */
boolean
LineAnalyzer::step_whPlus_ (boolean * isEol)
{
  if (isCOMPLETED ())
    (*isEol) = pk_true;
  else if (IS_WHITE ())
    {
      SKIP_WHITEsTAR ();
      (*isEol) = isCOMPLETED ();
      return pk_true;
    }
  else
    (*isEol) = pk_false;
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
   success: this is open, this is not completed, current character is not NULL_char <cc>
            cursor is moved one position forward
   failure: this is not open or this is completed or current character is not <cc> or <cc> is NULL_char
            this remains unchanged
 */
boolean
LineAnalyzer::stepIf_arg_ (char cc)
{
  if (isCURRENT (cc))
    {
      thisLineCursor = thisLineCursor + 1;
      return pk_true;
    }
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * success: <cc> is not NULL_char 
 *          BUFFER = [Character*] . <cc> [Character] [Character*][NULL_char]
 *               cursor is moved to [Character] following <cc>
 *          BUFFER = [Character*] . <cc> [NULL_char]
 *               cursor is moved to [NULL_char]
 * failure: BUFFER = [Character*] . [NULL_char]
 *          <cc> is not NULL_char, BUFFER = [Character*] . [Character] [Character*][NULL_char], [Character] is not <cc>
 *          this remains unchanged
 * <isEol> is set according to final cursor position
 */
boolean
LineAnalyzer::stepIf_arg_ (char cc, boolean * isEol)
{
  if (isCURRENT (cc))
    {
      thisLineCursor = thisLineCursor + 1;
      (*isEol) = isCOMPLETED ();
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-12-03
 * this atom remains unchanged
 * success: <cc1> and <cc2> are not NULL_char 
 *          BUFFER = [Character*] . <cc1> <cc2> { [NULL_char] + [Character] }
 *          cursor is moved after <cc2> occurrence
 * failure: <cc1> or <cc2> is not NULL_char or this is not open 
 *            or BUFFER = [Character*] . [NULL_char] + [Character][NULL_char]
 *            or BUFFER = [Character*] . [Character] [Character] 
 *                 and first [Character] is not <cc1> or second [Character] is not <cc2>
 *          this remains unchanged
 * <isEol> is set according to final cursor position
 */
boolean
LineAnalyzer::stepIf_arg_ (char cc1, char cc2, boolean * isEol)
{
  if ((isFORWARD (cc2)) && (unsafe_isCURRENTCHAR (cc1)))
    {
      thisLineCursor = thisLineCursor + 2;
      (*isEol) = isCOMPLETED ();
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-12-03
 * this atom remains unchanged
 * success: <cc1> and <cc2> are not NULL_char 
 *          BUFFER = [Character*] . <cc1> [Character2]{ [NULL_char] + [Character] }, [Character2] is not <cc2> 
 *          or BUFFER = [Character*] . <cc1> [NULL_char]
 *          cursor is moved after <cc1> occurrence
 * failure: <cc1> or <cc2> is not NULL_char or this is not open 
 *            or BUFFER = [Character*] . [NULL_char] + [Character][NULL_char]
 *            or BUFFER = [Character*] . [Character] [Character] 
 *                 and first [Character] is not <cc1> or second [Character] is not <cc2>
 *          this remains unchanged
 * <isEol> is set according to final cursor position
 */
boolean
LineAnalyzer::stepIf_char1_notChar2_ (char cc1, char cc2, boolean * isEol)
{
  if (unsafe_isCURRENTCHAR (cc1))
    {
      if (isFORWARD (cc2))
	{
	  (*isEol) = isCOMPLETED ();
	  return pk_false;
	}
      thisLineCursor = thisLineCursor + 1;
      (*isEol) = isCOMPLETED ();
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * cursor always starts NIL_String [Character*][NULL_char]
 * success: BUFFER = [Character*] . <str> [NULL_char]
 *               cursor is moved to [NULL_char]
 *               <isEol> becomes pk_true
 *          BUFFER = [Character*] . <str> [Character] 
 *               cursor is moved to [Character]
 *               <isEol> becomes false
 * failure: <str> is not NIL_String, current character is NULL_char
 *               <isEol> becomes pk_true
 *           <str> is not NIL_String, current character is not NULL_char, cursor does not start <str> [Character*]
 *               <isEol> becomes false
 *          this remains unchanged
 */
boolean
LineAnalyzer::stepIf_arg_ (const char *str, boolean * isEol)
{
  if (str == NIL_String)
    {
      (*isEol) = isCOMPLETED ();
      return pk_true;
    }
  if (isCOMPLETED ())
    {
      (*isEol) = pk_true;
      return pk_false;
    }
  if (isstrprefix (str, (thisLineBuffer + thisLineCursor)))
    {
      thisLineCursor = thisLineCursor + strlen (str);
      (*isEol) = isCOMPLETED ();
      return pk_true;
    }
  (*isEol) = pk_false;
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * success: BUFFER = [Character*] . [Character+] [NULL_char]
 *          first [Character+] character is returned in <cc> 
 *          cursor is moved to second position in [Character+] [NULL_char]
 * failure: this is not open or this is completed
 *          this remains unchanged
 *          <cc> remains unchanged
 */
boolean
LineAnalyzer::getThenStep_ (char *cc)
{
  if (isNOTCOMPLETED ())
    {
      *cc = unsafe_getCURRENTCHAR ();
      thisLineCursor = thisLineCursor + 1;
      return pk_true;
    }
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * success: BUFFER = [Character*] .[Character] <str> # opt{[not_IdentCharacter] [Character*]}[NULL_char]
 * failure: this is not open or BUFFER = [Character*] . {[Character]} [NULL_char]
 *          or BUFFER = [Character*] . [Character] <str> [identCharacter]
 *          or BUFFER = [Character*] . [Character] [Character*] and <str> is not a prefix of [Character*]
 *          this remains unchanged
 */
boolean
LineAnalyzer::stepIf_anyChar_thisFullIdentifier_ (const char *str)
{
  if (isNOTLASTPOSITION ())
    if (isstrprefix (str, (thisLineBuffer + thisLineCursor + 1)))
      {
	int nextIndex = thisLineCursor + 1 + strlen (str);
	if (isCIdChar (thisLineBuffer[nextIndex]))
	  return pk_false;
	thisLineCursor = nextIndex;
	return pk_true;
      }
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * success: <cc> is not NULL_char
 *          BUFFER = [Character*] . <cc>[White*] # [Character_but_White][Character*][NULL_char]
 *               <isEol> becomes false
 *          BUFFER = [Character*] . <cc> [White*] # [NULL_char]
 *               <isEol> becomes pk_true
 * failure: BUFFER = [Character*] .# [NULL_char].
 *               <isEol> becomes pk_true
 *          BUFFER = [Character*] .# [Character_but_<cc>][Character*][NULL_char]
 *               <isEol> becomes false
 *          this remains unchanged
 */
boolean
LineAnalyzer::stepIf_arg_whStar_ (char cc, boolean * isEol)
{
  if (isCURRENT (cc))
    {
      thisLineCursor = thisLineCursor + 1;
      if (isCOMPLETED ())
	(*isEol) = pk_true;
      else
	{
	  SKIP_WHITEsTAR ();
	  (*isEol) = isCOMPLETED ();
	}
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * success: <cc> is not NULL_char
 *          BUFFER = [Character*] . <cc>[White*] # [NULL_char], 
 *               <isSpaced> becomes pk_true
 *          BUFFER = [Character*] . <cc> [White+] # [Character_but_White][Character*][NULL_char]
 *               <isSpaced> becomes pk_true
 *          BUFFER = [Character*] . <cc> # [Character_but_White][Character*][NULL_char]
 *               <isSpaced> becomes false
 * failure: BUFFER = [Character*] .# [NULL_char]
 *               <isEol> becomes pk_true
 *          BUFFER = [Character*] .# [Character_but_<cc>][Character*][NULL_char]
 *               <isEol> becomes false
 *          this remains unchanged
 * <isEol> is set according to # position
 */
boolean
LineAnalyzer::stepIf_arg_whStar_ (char cc, boolean * isSpaced, boolean * isEol)
{
  if (isCURRENT (cc))
    {
      thisLineCursor = thisLineCursor + 1;
      if (isCOMPLETED ())
	{
	  (*isSpaced) = pk_true;
	  (*isEol) = pk_true;
	}
      else
	{
	  if (IS_WHITE ())
	    {
	      (*isSpaced) = pk_true;
	      SKIP_WHITEsTAR ();
	      (*isEol) = isCOMPLETED ();
	    }
	  else
	    {
	      (*isSpaced) = pk_false;
	      (*isEol) = pk_false;
	    }
	}
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-11-16
 * this atom remains unchanged
 * success: neither <cc1> nor <cc2> is NULL_char
 *          BUFFER = [Character*] . <cc1><cc2>[White*] # [Character_but_White][Character*][NULL_char]
 *          BUFFER = [Character*] . <cc1><cc2>[White*] # [NULL_char]
 * failure: BUFFER = [Character*] .# [NULL_char]
 *          BUFFER = [Character*] .# [Character_but_<cc1>][Character*][NULL_char]
 *              or   [Character*] .# <cc1> opt{[Character_but_<cc2>][Character*]}[NULL_char]
 *              or   [Character*] .# [Character*][NULL_char] if  <cc1> or <cc2> is [NULL_char]
 * <isEol> is set according to # position
 */
boolean
LineAnalyzer::stepIf_arg_whStar_ (char cc1, char cc2, boolean * isEol)
{
  if ((isFORWARD (cc2)) && (unsafe_isCURRENTCHAR (cc1)))
    {
      thisLineCursor = thisLineCursor + 2;
      if (isCOMPLETED ())
	(*isEol) = pk_true;
      else
	{
	  SKIP_WHITEsTAR ();
	  (*isEol) = isCOMPLETED ();
	}
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-11-16
 * this atom remains unchanged
 * success: neither <cc1> nor <cc2> is NULL_char
 *          BUFFER = [Character*] . <cc1>[White+] # opt{[Character_but_White][Character*]}[NULL_char]
 *          BUFFER = [Character*] . <cc1> # opt{[Character_but_White_but_<cc2>][Character*]}[NULL_char]
 * failure: BUFFER = [Character*] .# <cc1><cc2>[Character*][NULL_char]
 *          BUFFER = [Character*] .# opt{[Character_but_White_but_<cc1>][Character*]}[NULL_char]
 * <isEol> is set
 */
boolean
LineAnalyzer::stepIf_char1_notChar2_whStar_ (char cc1, char cc2, boolean * isEol)
{
  if (unsafe_isCURRENTCHAR (cc1))
    {
      if (isFORWARD (cc2))
	{
	  (*isEol) = isCOMPLETED ();
	  return pk_false;
	}
      thisLineCursor = thisLineCursor + 1;
      if (isCOMPLETED ())
	(*isEol) = pk_true;
      else
	{
	  SKIP_WHITEsTAR ();
	  (*isEol) = isCOMPLETED ();
	}
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * success: BUFFER = [Character*] . <str> [White*] # opt{[Character_but_White][Character*]}[NULL_char]
 *              <str> may be NIL_String
 * failure: BUFFER = [Character*] . [Character*][NULL_char]
 *              <str> is not NIL_String, second [Character*] does not start with <str>
 * <isEol> is set according to # position
 *          this remains unchanged
 */
boolean
LineAnalyzer::stepIf_arg_whStar_ (const char *str, boolean * isEol)
{
  if (isstrprefix (str, (thisLineBuffer + thisLineCursor)))
    {
      if (str != NIL_String)
	thisLineCursor = thisLineCursor + strlen (str);
      if (isCOMPLETED ())
	(*isEol) = pk_true;
      else
	{
	  SKIP_WHITEsTAR ();
	  (*isEol) = isCOMPLETED ();
	}
      return pk_true;
    }
  (*isEol) = isCOMPLETED ();
  return pk_false;
}

/* 2004-11-03
 * this atom remains unchanged
 * success: BUFFER = [Character*] . [Character] # [NULL_char], [Character] does not belong to <str>
 *          BUFFER = [Character*] . [Character] # [Character+] [NULL_char], first [Character] does not belong to <str>
 *          [Character] is returned in <cc>
 * failure: this is not open or BUFFER = [Character*] .# [NULL_char]
 *              NULL_char is returned in <cc>
 *          BUFFER = [Character*] .# [Character] [Character*][NULL_char], [Character] occurs in <str>
 *              [Character] is returned in <cc>
 *          this remains unchanged
 * <isEol> is set according to cursor final position
 */
boolean
LineAnalyzer::getThenStepIf_notIn_ (const char *str, char *cc, boolean * isEol)
{
  if (isCOMPLETED ())
    {
      *cc = NULL_char;
      (*isEol) = pk_true;
      return pk_false;
    }
  *cc = unsafe_getCURRENTCHAR ();
  if (STRCONTAINSCHR (str, *cc))
    {
      (*isEol) = pk_false;
      return pk_false;
    }
  thisLineCursor = thisLineCursor + 1;
  (*isEol) = isCOMPLETED ();
  return pk_true;
}

/* =============================================================================
 * this atom functions
 */

/* PLG 2004-10-25
 * this buffer remains unchanged provided that it does not share common space with this atom
 * WARNING may result in side effect on previously loaded string
 * this atom is emptied if its size is > 0
 * this atom length is set to 0
 */
void
LineAnalyzer::empty_atom ()
{
  if (thisAtomSize > 0)
    thisAtomBuffer[0] = NULL_char;
  thisAtomLength = 0;
}

/* 2004-10-14
 * this remains unchanged
 * returns pk_true if atom is empty
 */
boolean
LineAnalyzer::isAtomEmpty ()
{
  return (thisAtomLength <= 0);
}

/* 2004-10-14
 * this remains unchanged
 * returns pk_true if atom contents is equal to aString
 */
boolean
LineAnalyzer::isAtomEq (const char *aString)
{
  if (strcmp (aString, thisAtomBuffer) == 0)
    return pk_true;
  return pk_false;
}

/* 2004-10-14
 * this remains unchanged
 * if the atom contains a <cc1>, <cc2> sequence
 *     returns the length of the longest sub-string beginning at cursor position, terminated before next occurrence of sequence
 * if the atom does not contain a <cc1>, <cc2> sequence
 *     returns -1
 */
int
LineAnalyzer::atomPrefixLengthUntil (char cc1, char cc2)
{
  if ((cc1 == NULL_char) || isAtomEmpty ())
    return -1;
  return prefixLengthUntilCharChar (thisAtomBuffer, cc1, cc2);
}

/* 2004-10-14
 * this remains unchanged
 * returns atom adress
 */
char *
LineAnalyzer::unsafeGetAtom ()
{
  return thisAtomBuffer;
}

/* 2004-12-06, 2008_07-04
 * extracts a string XX as the longest not empty substring beginning at cursor position,
 *               and made of characters listed in atomAllowedChars
 * success: BUFFER = [Character*] . <atomAllowedChars>+ # [NULL_char], 
 *          BUFFER = [Character*] . <atomAllowedChars>+ # [Character_but_<atomAllowedChars>][Character*]
 *          <atomAllowedChars>+ is loaded in this atom
 *          <correct> remains unchanged
 * failure: BUFFER = [Character*] .# [NULL_char]
 *              <correct> remains unchanged
 *          BUFFER = [Character*] .# [Character_but_<atomAllowedChars>][Character*]
 *              <correct> remains unchanged
 *          internal error: atom becomes unsafe
 *              BUFFER = [Character*] .# <atomAllowedChars>[Character*]
 *              <correct> is set FALSE
 *          empties this atom
 * <isEol> is set 
 */
_PUBLIC_ boolean LineAnalyzer::stepAndLoad_charPlus_in_ (const char *atomAllowedChars, boolean * correct, boolean * isEol)
{
  if (isCOMPLETED ())
    (*isEol) = pk_true;
  else
    {
      int
	length = strspn (unsafe_getTAIL (), atomAllowedChars);
      if (length > 0)
	{
	  if (reset_atom (unsafe_getTAIL (), length))
	    {
	      thisLineCursor = thisLineCursor + length;
	      (*isEol) = isCOMPLETED ();
	      return pk_true;
	    }
	  (*correct) = pk_false;	/* internal error */
	  return pk_false;
	}
      (*isEol) = pk_false;
    }
  empty_atom ();
  return pk_false;
}

/* 2004-12-06
 * extracts a (possibly empty) string XX as the longest sub-string beginning at cursor position,
 *                      and terminated before next occurrence of <terminator>
 *          if <terminator> is NULL_char, extracts current remaining string if string is not completed
 * success: <terminator> is not NULL, it does not occur in [Character*] 
 *          BUFFER = [Character*] . [Character*] <terminator> #[Character*]
 *          [Character*] is loaded in this atom
 *          <correct> remains unchanged
 * failure: BUFFER = [Character*] .# [Character*], <terminator> is NULL or <terminator> does not occur in [Character*] 
 *             empties this atom
 *             <correct> remains unchanged
 *          BUFFER = [Character*] . [Character*] <terminator> # [Character*]
 *             Some memory error occurs
 *             <correct> is set FALSE
 * <isEol> is set according to final cursor position
 */
boolean
LineAnalyzer::stepAndLoad_charStar_before_ (char terminator, boolean * correct, boolean * isEol)
{
  if (isCOMPLETED ())
    (*isEol) = pk_true;
  else
    {
      int length = prefixLengthUntil (terminator);
      if (length >= 0)
	{
	  if (length > 0)
	    {
	      if (reset_atom (&thisLineBuffer[thisLineCursor], length))
		thisLineCursor = thisLineCursor + length + 1;
	      else
		{
		  thisLineCursor = thisLineCursor + length + 1;
		  (*isEol) = isCOMPLETED ();
		  (*correct) = pk_false;
		  return pk_false;
		}
	    }
	  else
	    {			/* length == 0 */
	      empty_atom ();
	      thisLineCursor = thisLineCursor + 1;
	    }
	  (*isEol) = isCOMPLETED ();
	  return pk_true;
	}
      (*isEol) = pk_false;
    }
  empty_atom ();
  return pk_false;
}

/* 2004-12-06, 2008_07-04
 * extracts a (possibly empty) string XX as the longest sub-string beginning at cursor position,
 *                      and terminated before next occurrence of <cc1><cc2>
 * success: <cc1> and <cc2> are not NULL, <cc1><cc2> does not occur in [Character*] 
 *          BUFFER = [Character*] . [Character*] <cc1><cc2> # [Character*]
 *          [Character*] is loaded in this atom
 *          <correct> remains unchanged
 * failure: BUFFER = [Character*] .#[Character*], <cc1> is NULL or <cc2> is NULL or <cc1><cc2> does not occur in [Character*] 
 *             empties this atom
 *             <correct> remains unchanged
 *          BUFFER = [Character*] . [Character*] <cc1><cc2> # [Character*]
 *             memory error occurs
 *             <correct> is set FALSE
 * <isEol> is set according to final position
 */
boolean
LineAnalyzer::stepAndLoad_charStar_before_ (char cc1, char cc2, boolean * correct, boolean * isEol)
{
  if (isCOMPLETED ())
    (*isEol) = pk_true;
  else
    {
      int length = prefixLengthUntil (cc1, cc2);
      if (length >= 0)
	{
	  if (length > 0)
	    {
	      if (reset_atom (unsafe_getTAIL (), length))
		thisLineCursor = thisLineCursor + length + 2;
	      else
		{
		  thisLineCursor = thisLineCursor + length + 2;
		  (*isEol) = isCOMPLETED ();
		  (*correct) = pk_false;
		  return pk_false;
		}
	    }
	  else
	    {			/* length == 0 */
	      empty_atom ();
	      thisLineCursor = thisLineCursor + 2;
	    }
	  (*isEol) = isCOMPLETED ();
	  return pk_true;
	}
      (*isEol) = pk_false;
    }
  empty_atom ();
  return pk_false;
}

/* 2004-10-13
 * cursor remains unchanged
 * if this atom is empty returns NIL_String
 * if this atom is not empty returns a new fresh copy of this atom
 * empties this atom
 */
char *
LineAnalyzer::strcatchAtom ()
{
  char *res = detachPKString (thisAtomRegister);	/* NIL_String if <thisAtomRegister> is empty or errored */
  reset_atom ();
  return res;
}

/** open this line analyzer with <aStr> as new content
    this is set equal to <aStr> (WARNING NOT A COPY)
    this line cursor is set at first position in this line
 * empties this atom
 */
void
LineAnalyzer::lineOpen (char *aStr)	/* cannot be const */
{
  set_line (aStr);		/* cannot be const */
  reset_atom ();
}

/** open this line analyzer with <aStr> as new content
    this is set equal to <aStr> (WARNING NOT A COPY)
         after removing last char if it is an end of line mark (WARNING SIDE EFFECT)
    this line cursor is set at first position in this line
 * empties this atom
 * returns this line length
 */
int
LineAnalyzer::lineOpenString (char *str)
{
  set_line (str);		/* cannot be const */
  remove_lineLastEol ();
  return thisLineActualLength;
}

/* prevents further reading not preceded by lineOpen:
 * this atom is emptied 
 * this line becomes NULL container
 * this string length and this line cursor position are set to -1
 */
void
LineAnalyzer::lineClose ()
{
  set_line ();
  empty_atom ();
}

/** Creates an empty LineAnalyzer
    sets this atom register
    sets this atom empty
    sets this line empty
    sets cursor position to -1
 */
LineAnalyzer::LineAnalyzer ()
{
  set_atomNull ();
  set_line ();
}

/** Creates an empty LineAnalyzer
    sets this atom register to contain <aSize> characters
    sets this atom empty
    sets this line empty
    sets cursor position to -1
 */
LineAnalyzer::LineAnalyzer (int aSize)
{
  set_atomNew (aSize);
  set_line ();
}

/** Creates a LineAnalyzer loaded with str
    sets this atom register
    this is set equal to <aStr> (WARNING NOT A COPY)
    this line cursor is set at first position in this line
 * empties this atom
 */
LineAnalyzer::LineAnalyzer (char *aStr)	/* cannot be const */
{
  set_atomNull ();
  lineOpen (aStr);		/* cannot be const */
}

LineAnalyzer::~LineAnalyzer ()
{
  if (thisAtomSize >= 0)
    thisAtomRegister = deletePKString (thisAtomRegister);
}
