/* 
   thors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/* ======> debug facilities 
 */
extern void
nDEBUG_treeRoot (int indentVal, int depth, const char *mess, ntrees aSubtree)
{
  DEBUG_treeRoot (indentVal, depth, mess, (trees) aSubtree);
}

extern void
nDEBUG_treeHyperRoot (int indentVal, int depth, const char *mess, ntrees aSubtree)
{
  DEBUG_treeHyperRoot (indentVal, depth, mess, (trees) aSubtree);
}

extern void
nDEBUG_subtree (int indentVal, int depth, const char *mess, ntrees aSubtree)
{
  DEBUG_subtree (indentVal, depth, mess, (trees) aSubtree);
}

extern void
nDEBUG_attribute (int indentVal, int depth, const char *mess, nattribute anAttribute)
{
  DEBUG_attribute (indentVal, depth, mess, (ptattribute) anAttribute);
}

extern void
nDEBUG_aSubreeAttributes (int indentVal, int depth, const char *mess, ntrees aSubtree)
{
  DEBUG_aSubreeAttributes (indentVal, depth, mess, (trees) aSubtree);
}

/* ======> Global property test  
 */

/** Testing tree. 
   @return "true" if operand tree aSubtree is the undefined empty tree. 
*/
extern boolean
nisniltree (ntrees aSubtree)
{
  return (boolean) (IS_NILTREE (aSubtree));
}

/** 
   @return true if aSubtree is an uvalued atom . 
*/
extern boolean
nisatomnil (ntrees aSubtree)
{
  return isTM_nilAtom ((trees) aSubtree);
}

/** 
   @return true if aSubtree is a string valued atom . 
*/
extern boolean
nisatomchar (ntrees aSubtree)
{
  return isTM_charAtom ((trees) aSubtree);
}

/** 
   @return true if aSubtree is a string valued atom . 
*/
extern boolean
nisatomstring (ntrees aSubtree)
{
  return isTM_TMStringAtom ((trees) aSubtree);
}

/** 
   @return true if aSubtree is a string valued atom . 
*/
extern boolean
nisatomint (ntrees aSubtree)
{
  return isTM_intAtom ((trees) aSubtree);
}

/** Getting Tree arity. 
   @return the arity (number of sons) of the operand tree aSubtree.
           0 if aSubtree is nil 
 */
extern int
ntreearity (ntrees aSubtree)
{
  return getTM_arity ((trees) aSubtree);
}

/** Testing tree. 
   @return true if the operand tree aSubtree is an empty list tree. 
*/
extern boolean
nisemptyList (ntrees aSubtree)
{
  return isTM_emptyList ((trees) aSubtree);
}

/** Testing tree. 
   @return "true" if operand tree aSubtree is a list tree. 
*/
extern boolean
nislist (ntrees aSubtree)
{
  return isTM_listSubtree ((trees) aSubtree);
}


/* ======> navigation  
 */

/** Getting nil tree. 
   @return the nil tree representation. 
*/
extern ntrees
ngetniltree ()
{
  return NILTREE;
}

/** @return the operator of NN if @param aSubtree is a correct node NN
 *  @return the LTA operator OPUNDEF (-8)
 *  may return a LTA operator (OPTROU -7, OPNIL -6, OPENTIER -5, OPCAR -4, OPCHAINE -3, OPSTAR -2, OPREPERE -1, OPARBRE 07 )
 * or a user defined operator as a positive integer
 */
extern noperators
noper (ntrees aSubtree)
{
  return getTM_operator ((trees) aSubtree);
}

/** Getting rank. 
   @return the rank of the operand tree aSubtree. 
          0 if aSubtree is nil
 */
extern int
nposit (ntrees aSubtree)
{
  return getTM_place ((trees) aSubtree);
}

/** Getting father. 
   @return the father of the operand tree aSubtree. 
        nil if aSubtree is nil
*/
extern ntrees
nfather (ntrees aSubtree)
{
  return (ntrees) getTM_father ((trees) aSubtree);
}

extern ntrees
nroot (ntrees aSubtree)
{
  return (ntrees) getTM_root ((trees) aSubtree);
}

extern ntrees
nhyperRoot (ntrees aSubtree)
{
  return (ntrees) getTM_hyperRoot ((trees) aSubtree);
}

/** Getting a son. 
   @return the i-th son of the operand tree aSubtree. 
        nil if aSubtree is nil or aSubtree has less than i children
 */
extern ntrees
nchild (int i, ntrees aSubtree)
{
  return (ntrees) getTM_nChild (i, (trees) aSubtree);
}

/** Getting a brother (right). 
    @param i is expected to be a strictly positive integer
    @return the i-th right brother of aSubtree
    @return niltree if i < 1, if aSubtree is NIL or does not have i elements on its right
*/
extern ntrees
ntrees_getINext (int i, ntrees aSubtree)
{
  return (ntrees) getTM_rightNeighbor (i, (trees) aSubtree);
}

/** Getting a brother (right). 
    @return the next right brother of aSubtree
    @return niltree if aSubtree is NIL or if aSubtree is a last child
*/
extern ntrees
ntrees_getNext (ntrees aSubtree)
{
  return (ntrees) getTM_rightNeighbor (1, (trees) aSubtree);
}

/** Getting a brother (left). 
    @param i is expected to be a strictly positive integer
    @return the Ith previous left brother of aSubtree
    @return niltree if aSubtree is NIL or if aSubtree has less than I elements on its left.
*/
extern ntrees
ntrees_getIPrevious (int i, ntrees aSubtree)
{
  return (ntrees) getTM_leftNeighbor (i, (trees) aSubtree);
}


/** Getting a brother (left). 
    @return the previous left brother of aSubtree
    @return niltree if aSubtree is NIL or if aSubtree is a first child.
*/
extern ntrees
ntrees_getPrevious (ntrees aSubtree)
{
  return (ntrees) getTM_leftNeighbor (1, (trees) aSubtree);
}

/** Getting atomic tree value(string). 
   @return the string denoted by an atomic tree aSubtree (assumed). 
   @return nilstr if aSubtree is not a non empty string atom
   \see{natomstring}
*/
extern ptrepstr
nstringof (ntrees aSubtree)
{
  return (ptrepstr) getTM_TMStringAtomValue ((trees) aSubtree);
}

/** Getting atomic tree value(int). 
   @return the integer value denoted by an atomic tree aSubtree (assumed). 
   @return 0 if aSubtree is not a non empty integer atom
   \see{natomint}
*/
extern long int
nintof (ntrees aSubtree)
{
  return getTM_intValue ((trees) aSubtree);
}

/** Getting atomic tree value(char). 
   @return the char value denoted by an atomic tree aSubtree (assumed). 
   @return \0 if aSubtree is not a non empty char atom
   \see{natomchar}
*/
extern char
ncharof (ntrees aSubtree)
{
  return getTM_charAtomValue ((trees) aSubtree);
}

/** Build an unvalued atomic tree. 
   @return a new atomic tree with pop as operator
*/
extern ntrees
natomnil (noperators pop)
{
  return (ntrees) makeTM_nilAtom (pop);
}

/** Build string atomic tree. 
   @return a new atomic tree with pop as operator and p as string value. 
*/
extern ntrees
natomstring (noperators pop, ptrepstr p)
{
  return (ntrees) makeTM_TMStringAtom (pop, (TMString) p);
}

/** Build int atomic tree. 
   @return a new atomic tree with pop as operator and p as int value. 
*/
extern ntrees
natomint (noperators pop, long int integerValue)
{
  return (ntrees) makeTM_intAtom (pop, integerValue);
}

/** Build char atomic tree.
   @return a new atomic tree with pop as operator and c as char value. 
*/
extern ntrees
natomchar (noperators pop, char c)
{
  return (ntrees) makeTM_charAtom (pop, c);
}

/** Build unary tree.
   @return a new (unary) tree with pop as operator and aSubtree as son. 
*/
extern ntrees
nmakeunary (noperators pop, ntrees aSubtree)
{
  return (ntrees) makeTM_unaryTree (pop, (trees) aSubtree);
}

/** Build binary tree.
   @return a new (binary) tree with pop as operator and aSubtree1, aSubtree2 as sons. 
*/
extern ntrees
nmakebinary (noperators pop, ntrees aSubtree1, ntrees aSubtree2)
{
  // nDEBUG_subtree(4,2,"entering intetreesGetSet_g::nmakebinary , son subtree 1 is", (trees)aSubtree1); 
  // nDEBUG_subtree(4,2,"entering intetreesGetSet_g::nmakebinary , son subtree 2 is", (trees)aSubtree2); 
  return (ntrees) makeTM_binaryTree (pop, (trees) aSubtree1, (trees) aSubtree2);
}

/** Build ternary tree.
    @return a new (ternary) tree with pop as operator and aSubtree1, aSubtree2, aSubtree3 as sons. 
*/
extern ntrees
nmaketernary (noperators pop, ntrees aSubtree1, ntrees aSubtree2, ntrees aSubtree3)
{
  return (ntrees) makeTM_ternaryTree (pop, (trees) aSubtree1, (trees) aSubtree2, (trees) aSubtree3);
}

/** Build quaternary tree.
   @return a new (quaternary) tree with pop as operator and aSubtree1, aSubtree2, aSubtree3, aSubtree4 as sons. 
*/
extern ntrees
nmakequaternary (noperators pop, ntrees aSubtree1, ntrees aSubtree2, ntrees aSubtree3, ntrees aSubtree4)
{
  return (ntrees) makeTM_quaternaryTree (pop, (trees) aSubtree1, (trees) aSubtree2, (trees) aSubtree3, (trees) aSubtree4);
}

/** Build empty list tree.
   @return a new (list) empty tree with pop as operator. 
*/
extern ntrees
nmakemptylist (noperators pop)
{
  return (ntrees) makeTM_emptyListTree (pop);
}

/** Build list tree.
   @return a new (list) tree with pop as operator and intilialized with aSubtree as single element. 
*/
extern ntrees
nmakelist (noperators pop, ntrees aSubtree)
{
  return (ntrees) makeTM_listTree (pop, (trees) aSubtree);
}

/** Build list tree.
   @return a new (list) tree the operator of which is pop, the lements of which are substrings of aString separated by separator 
*/
extern ntrees
CStringSplit_makelist (char *aString, noperators listOp, noperators elemOp, char separator)
{
  ptreg wreg;

  // printf("DEBUG entering CStringSplit_makelist aString is <<%s>>\n", aString);  fflush(stdout);
  if (new_reg_CString (&wreg, aString))
    {
      ntrees listRes;
      int stringNumber, i, lg;
      ptrepstr iRepstr;
      lg = strlen (aString);
      listRes = makeTM_emptyListTree (listOp);
      stringNumber = strregsplit (wreg, separator, lg);
      for (i = 0; i < stringNumber; i++)
	{
	  iRepstr = strreggetsubstring (wreg, i);
	  appendTM_lastInListTree (listRes, natomstring (elemOp, iRepstr));
	}
      strregfree (wreg);
      return listRes;
    }
  return NILTREE;
}

/** Adding in a list(end). 
    @return the list aListTree in which the element aSubtree is added at the end of the list. 
           nil if aListTree is nil (operator is not available)
*/
extern ntrees
ntreeList_appendLast (ntrees aListTree, ntrees aSubtree)
{
  if (IS_NILTREE (aListTree)) return aListTree;
  if (IS_NILTREE (aSubtree))
    {
      // printf ("DEBUG: (ntreeList_appendLast) Warning: Inserting a NILTREE\n");
      return aListTree;
    }
  return (ntrees) appendTM_lastInListTree ((trees) aListTree, (trees) aSubtree);
}

/** 
    Adding in a list(head). 
    @return the list aListTree in which the element aSubtree is added at the begin of the list. 
           nil if aListTree is nil (operator is not available)
*/
extern ntrees
ntreeList_insertFirst (ntrees insertedSubtree, ntrees aListTree)
{
  if (IS_NILTREE (aListTree)) return aListTree;
  if (IS_NILTREE (insertedSubtree))
    {
      // printf ("DEBUG: (ntreeList_insertFirst) Warning: Inserting a NILTREE\n");
      return aListTree;
    }
  return (ntrees) insertTM_firstInListTree ((trees) insertedSubtree, (trees) aListTree);
}

/**
   Adding in a list(after). 
   Inserts insertedSubtree at the right of the i-th son of the list nodeInList. 
   When i=0, it is added at the head of the list.
           nil if nodeInList is nil (operator is not available)
*/
extern void
ninsertright (int i, ntrees insertedSubtree, ntrees nodeInList)
{
  if (i == 0)
    ntreeList_insertFirst (insertedSubtree, nodeInList);
  else
    insertTM_afterIinListTree (i, (trees) insertedSubtree, (trees) nodeInList);	// remains unchanged if nodeInList is not a child in a list subtree or insertedSubtree is NIL
}

/**
   Add insertedSubtree after nodeInList.
   Do nothing if insertedSubtree is NIL or nodeInList is not a node in a list
*/
extern void
ninsertafter (ntrees insertedSubtree, ntrees nodeInList)
{
  (void) insertTM_afterPredInListTree ((trees) insertedSubtree, (trees) nodeInList);
}

/** 
    Adding in a list (before). 
    Inserts insertedSubtree at the left of the i-th son of the list aListTree. 
           nil if aListTree is nil (operator is not available)
*/
extern void
ninsertleft (int i, ntrees insertedSubtree, ntrees aListTree)
{
  if (IS_NILTREE (aListTree))
    return;
  insertTM_beforeIinListTree (i, (trees) insertedSubtree, (trees) aListTree);
}

/**
   Add insertedSubtree before nodeInList.
   Do nothing if insertedSubtree is NIL or nodeInList is not a node in a list
*/
extern void
ntreeList_insertBefore (ntrees insertedSubtree, ntrees nodeInList)
{
  (void) insertTM_beforeSuccInListTree ((trees) insertedSubtree, (trees) nodeInList);
}

/** Copying. 
   @return a copy TT of aSubtree
   @return  nil if aSubtree is nil
   if copyAttributes is pk_true TT contains a copy of attributes otherwise attributes are NIL in TT
*/
extern ntrees
ncopytree (ntrees aSubtree, boolean copyAttributes)
{
  return (ntrees) copyTM_subtree ((trees) aSubtree, copyAttributes);
}

/**
   Modifying. 
   It modifies the value assigned to an (assumed) atomic int tree anIntAtom. 
             Null if anIntAtom is nil
   @param val (in) the value assigned to anIntAtom. 
*/
extern void
nmodifvalint (ntrees anIntAtom, long int val)
{
  if (IS_NILTREE (anIntAtom))
    return;
  // printf("DEBUG entering nmodifvalint, tree adr is %li, integer value is %li\n", (long int)anIntAtom, val); fflush(stdout);
  setTM_intValue ((trees) anIntAtom, val);
}

/** Modifying. 
   It modifies the value assigned to an (assumed) atomic char tree aCharAtom. 
             Null if aCharAtom is nil
   @param val (in) the value assigned to aCharAtom. 
*/
extern void
nmodifvalcar (ntrees aCharAtom, char val)
{
  if (IS_NILTREE (aCharAtom))
    return;
  setTM_charAtomValue ((trees) aCharAtom, val);
}

/** Modifying. 
   It deletes the pos-th son of the tree aSubtree. 
             Null if aListTree is nil
   @return the deleted sub-tree. 
             nil if aListTree is nil
*/
extern ntrees
ndestroyListElem (ntrees aListTree, int pos)
{
  // DEBUG_treeRoot(4, 100, "entering ndestroyListElem", aListTree);
  aListTree = (ntrees) removeTM_inListSubtree ((trees) aListTree, pos);
  // DEBUG_treeRoot(4, 100, "exiting ndestroyListElem", aListTree);
  return aListTree;
}

/** Modifying. 
             Null if aSubtree is nil
   It replaces the i-th son of aSubtree by treeValue. 
*/
extern void
nchangchild (ntrees aSubtree, ntrees treeValue, int i)
{
  resetTM_nchild ((trees) aSubtree, (trees) treeValue, i);
}

/**
   Setting tree operator: op becomes the operator of the tree aSubtree. No controls are done on the arity of t. 
             Null if aSubtree is nil
  
*/
extern ntrees
nsetop (ntrees aSubtree, noperators op)
{
  setTM_operator ((trees) aSubtree, op);	/* NIL_trees is assumed to be tested */
  return aSubtree;
}

/**
   Removing a tree (p) from a list. 
   If p is an element of a tree, it is removed from it.
             Null if aSubtree is nil
*/
extern void
ntreesDestroy (ntrees aSubtree)
{
  (void) removeTM_inListSubtree (getTM_father ((trees) aSubtree), getTM_place ((trees) aSubtree));
}

/**
   Removing a tree (p) from a list. 
   If p is an element of a tree, it is removed from it.
             Null if aSubtree is nil
*/

/** @return true if aSubtree1 and aSubtree2 are equal up to their respective attributes (not tested)
 */
extern boolean
ntreesEq (ntrees aSubtree1, ntrees aSubtree2)
{
  // nDEBUG_subtree(0,2,"DEBUG entering ntreesEq aSubtree1", aSubtree1);
  // nDEBUG_subtree(0,2,"DEBUG entering ntreesEq aSubtree2", aSubtree2);
  if (aSubtree1 == aSubtree2)
    return pk_true;
  if (IS_NILTREE (aSubtree1))
    return IS_NILTREE (aSubtree2);
  if (IS_NILTREE (aSubtree2))
    return pk_false;
  if (isTM_hole (aSubtree1))
    return isTM_hole (aSubtree2);
  if (isTM_hole (aSubtree2))
    return pk_false;

  if (getTM_operator ((trees) aSubtree1) == getTM_operator ((trees) aSubtree2))
    {
      int arity1 = getTM_arity ((trees) aSubtree1);
      if (arity1 == getTM_arity ((trees) aSubtree2))
	{
	  if (isTM_emptyList ((trees) aSubtree1))
	    return pk_true;
	  if (isTM_nilAtom ((trees) aSubtree1))
	    return pk_true;

	  if (isTM_charAtom ((trees) aSubtree1))
	    return getTM_charAtomValue ((trees) aSubtree1) == getTM_charAtomValue ((trees) aSubtree2);
	  if (isTM_TMStringAtom ((trees) aSubtree1))
	    return getTM_TMStringAtomValue ((trees) aSubtree1) == getTM_TMStringAtomValue ((trees) aSubtree2);
	  if (isTM_intAtom ((trees) aSubtree1))
	    return getTM_intValue ((trees) aSubtree1) == getTM_intValue ((trees) aSubtree2);

	  if (arity1 > 0)
	    {
	      int i;
	      for (i = 1; i <= arity1; i++)
		{
		  if (!ntreesEq ((ntrees) getTM_nChild (i, (trees) aSubtree1), (ntrees) getTM_nChild (i, (trees) aSubtree2)))
		    return pk_false;
		}
	      // printf("DEBUG exiting ntreesEq 8 TRUE\n"); fflush(stdout);
	      return pk_true;
	    }
	}
    }
  // printf("DEBUG  exiting ntreesEq *\n"); fflush(stdout);
  return pk_false;
}


/**
   Misc.
   It prints the state of the tree memory.
 */
extern void
ntrees_printTreeMemoryState ()
{
  printTreeMemoryState ();
}
