/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


/** Loading Tree from a file. 
   @param filename (in) a name of a file.
   @return true if loading succeeds; the loaded object is assigned to res.
*/
extern boolean
ntrees_load (char *filename, ntrees * res)
{
  (*res) = (ntrees) loadTM_tree (filename);
  return !(IS_NILTREE (*res));
}

/** Saving a Tree into a file. 
   @param filename (in) a name of a file.
   @param e (in) a tree.
   @return true if the operation is performed successfully.
           the operation fails when aTree is nil or file cannot be open
*/
extern boolean
ntrees_save (char *filename, ntrees aTree)
{
  if (IS_NILTREE (aTree))
    return pk_false;
  return (saveTM_subTree (filename, (trees) aTree));
}


/** Saving a Tree into a file. 
   @param filename (in) a name of a file.
   @param e (in) a tree.
   @return true if the operation is performed successfully.
           the operation fails when aTree is nil or file cannot be open
*/
extern boolean
ntrees_formatSave (char *filename, ntrees aTree, char *formatLanguage, char *languageVersion)
{
  int totalNodeNumber;
  FILE *this_file;
  // printf("DEBUG ntrees_formatSave\n");
  if (IS_NILTREE (aTree))
    return pk_false;
  this_file = fopen (filename, "w");
  if (this_file == NULL)
    return pk_false;
  if (formatLanguage == (char *) 0)
    putString_formatSave (this_file, "**Unknown language**");
  else
    putString_formatSave (this_file, formatLanguage);
  fputc (' ', this_file);
  if (languageVersion == (char *) 0)
    putString_formatSave (this_file, "**Unknown version**");
  else
    putString_formatSave (this_file, languageVersion);
  fputc ('\n', this_file);
  totalNodeNumber = subtree_formatSave (this_file, aTree, 1, 0) - 1;
  fclose (this_file);
  return pk_true;
}

/** Loading a Tree in Intermediate Format. 
   @param filename (in) a name of a file.
   @param e (out) a tree.
   @return true if the operation is performed successfully.
*/
extern boolean
ntrees_formatLoad (char *filename, ntrees * aTree)
{
  FILE *this_file;
  char nextChar;
  ptrepstr loadedString;
  // printf("DEBUG ntrees_formatLoad\n");
  this_file = fopen (filename, "r");
  if (this_file == NULL)
    return pk_false;
  /* this_file=?!space* LANGUAGESTRING space+  VERSIONSTRING char* EOF */
  nextChar = skipSpaces (this_file);
  if (nextChar == EOF)
    {				/* this_file= ?space*! EOF */
      printf ("WARNING the file %s is empty\n", filename);
      (*aTree) = (ntrees) makeTM_hole ();
      return pk_false;
    }
  if (nextChar == '"')
    {				/* this_file=?space* !" LANGUAGESTRING space+  VERSIONSTRING char* EOF */
      nextChar = fgetc (this_file);	/* to get '"' character */
      if (getString_formatLoad (this_file, &loadedString))
	{
	  /* this_file=?space* " LANGUAGESTRING " ! space+  VERSIONSTRING char* EOF */
	  this_language = repstr2string (loadedString);
	  nextChar = skipSpaces (this_file);
	  if (nextChar == EOF)
	    {			/* this_file= ?space*! EOF */
	      printf ("WARNING the file %s has no tree occurrence\n", filename);
	      (*aTree) = (ntrees) makeTM_hole ();
	      return pk_true;
	    }
	  if (nextChar == '"')
	    {			/* this_file=?space* !" LANGUAGESTRING space+  VERSIONSTRING char* EOF */
	      nextChar = fgetc (this_file);	/* to get '"' character */
	      if (getString_formatLoad (this_file, &loadedString))
		this_version = repstr2string (loadedString);
	      else
		{
		  (*aTree) = (ntrees) makeTM_hole ();
		  set_errored ("language or version string");
		  return pk_false;
		}
	    }
	  else
	    {
	      this_version = this_language;
	      this_language = NIL_String;
	      printf ("WARNING the language or the version is missing in file %s, %s is assumed to be a version \n", filename, this_version);
	    }
	}
      else
	{
	  (*aTree) = (ntrees) makeTM_hole ();
	  set_errored ("language or version string");
	  return pk_false;
	}
    }
  else
    {
      this_version = NIL_String;
      this_language = NIL_String;
      printf ("WARNING the language and the version in file %s are missing\n", filename);
    }
  if (subtree_formatLoad (this_file, aTree, pk_false));
  fclose (this_file);
  //  return not is_errored; 
  return !is_errored;
}
