
;; chargement automatique du mode sigtra

(if (assoc "_TRA\\.SIG$" auto-mode-alist)
    t ; l'association y est deja
  (setq auto-mode-alist (append auto-mode-alist 
				'(("_TRA\\.SIG$" . sigtra-mode))))
  )

(defvar sigtra-mode-abbrev-table nil
  "Abbrev table in use in sigtra-mode buffers.")
(define-abbrev-table 'sigtra-mode-abbrev-table ())

(defvar sigtra-mode-map ()
  "Keymap used in Sigtra mode.")
(setq sigtra-mode-map (make-sparse-keymap))
(define-key sigtra-mode-map "\[229z" 'sigtra-extract-clocks-tree) ;; F6
;(define-key sigtra-mode-map "[229z" 'sigtra-hnot-search) ;; Shift F6
(define-key sigtra-mode-map "\[230z" 'sigtra-show-sig-clock-occ) ;; F7
(define-key sigtra-mode-map "\[193z" 'sigtra-show-sig-clock-in-clock-def) ;; F12
(define-key sigtra-mode-map "\[231z" 'sigtra-substitute-clock-by-def) ;; F8
(define-key sigtra-mode-map "\[232z" 'sigtra-show-sig-clock-def) ;; F9
(define-key sigtra-mode-map "\[-1z" 'sigtra-show-clock-of-sig) ;; F10
(define-key sigtra-mode-map "\[192z" 'sigtra-number-clocks) ;; F11
;(define-key sigtra-mode-map "\[193z" 'describe-mode) ;; F12

(defvar sigtra-mode-syntax-table nil
  "Syntax table in use in SIGTRA-mode buffers.")

(if sigtra-mode-syntax-table
    ()
  (setq sigtra-mode-syntax-table (make-syntax-table))
  ()
  )

(defconst sigtra-indent-level 2
  "*Indentation of sigtra statements with respect to containing block.")


(defconst sigtra-auto-newline nil
  "*Non-nil means automatically newline before and after braces,
and after colons and semicolons, inserted in OCCAM code.")

(defun sigtra-mode ()
  "     Quelques manipulations sur le TRA 

    - F6       : extraction de l'arborescence d'horloges;
    - F7       : visualisation des occurrences d'un signal ou d'une horloge
                 dans un process horloge;
    - F8       : substitution d'une horloge par sa definition;
    - F9       : visualisation de la definition d'une horloge ou d'un signal;
    - F10      : visualisation de l'horloge d'un signal;
    - F11      : numerotation des horloges en fonction de leur profondeur
                 dans l'arbre.
    - F12      : visualisation des occurrences d'un signal ou d'une horloge
                 dans les definitions d'horloges;
"


  (interactive)
  (kill-all-local-variables)
  (use-local-map sigtra-mode-map)
  (setq major-mode 'sigtra-mode)
  (setq mode-name "SIGTRA")
  (setq local-abbrev-table sigtra-mode-abbrev-table)
  (set-syntax-table sigtra-mode-syntax-table)
  (make-local-variable 'comment-start)
  (make-local-variable 'comment-end)
  (setq comment-start "%")
  (setq comment-end "%")
  (make-local-variable 'current-block)
  (setq current-block "")
  (make-local-variable 'sig-tree)
  (setq sig-tree nil)
  (make-local-variable 'clocks-tree-buf)
  (setq clocks-tree-buf nil)
  (make-local-variable 'num-clocks-tree-buf)
  (setq num-clocks-tree-buf nil)
  (make-local-variable 'pretty-decomp-pos)
  (setq pretty-decomp-pos 0)
  (modify-syntax-entry ?_ "w")
  (run-hooks 'sigtra-mode-hook)
  )

(defun last (l)
  (interactive)
  (cond
   ((null l)
    l)
   ((null (cdr l))
    l)
   (t
    (last (cdr l)))
   )
  )

(defun append1 (list elt)
  (cond
   ((null list) (list elt))
   ((rplacd (last list) (list elt)) list)
   )
  )

(defun sigtra-loc-extract-clocks-tree (signal-process-name)
  (save-excursion
    (let ((tree (list signal-process-name)) clock-name stop-point start-point (llochorl nil) (where-point 0) (end-point 0))
      (if (re-search-forward (concat "^[	 ]*process " signal-process-name) (point-max) t)
	  (progn
	    (setq start-point (point))
;	    (message "Process traite : %s" signal-process-name)
;	    (sit-for 3)
	    (message "Working")
	    (save-excursion
	      (search-forward "!")
	      (save-excursion
		(search-forward "}")
		(setq stop-point (point))
		)
	      (if (re-search-forward "\\bevent\\b" stop-point t)
		  (progn
		    (setq stop-point (min (save-excursion
					    (if (re-search-forward ";" stop-point t)
						(point)
					      stop-point
					      )
					    )
					  (save-excursion
					    (if (re-search-forward "}" stop-point t)
						(point)
					      stop-point
					      )
					    )
					  )
			  )
		    (forward-word 1)
		    (while (< (point) stop-point)
		      (backward-word 1)
		      (setq llochorl (append1 llochorl (buffer-substring (point)
									 (progn
									   (forward-word 1)
									   (point)
									   ))))
		      (forward-word 1)
		      )
		    )
		)
	      )
;	    (message "Apres traitement des sorties")
;	    (sit-for 3)
	    (setq stop-point (min (save-excursion
				    (if (re-search-forward "^[	 ]*where\\b" (point-max) t)
					(progn
					  (setq where-point (point))
					  (point)
					  )
				      (point-max)
				      )
				    )
				  (save-excursion
				    (if (re-search-forward "^[	 ]*end\\b" (point-max) t)
					(progn
					  (setq end-point (point))
					  (point)
					  )
				      (point-max)
				      )
				    )
				  )
		  )
;	    (message "where-point : %d ; stop-point : %d" where-point stop-point)
;	    (sit-for 3)
	    (if (= where-point stop-point)
		(save-excursion
		  (goto-char where-point)
		  (setq end-point (min (save-excursion
				    (if (re-search-forward "^[	 ]*process\\b" (point-max) t)
					(point)
				      (point-max)
				      )
				    )
				  (save-excursion
				    (if (re-search-forward "^[	 ]*end\\b" (point-max) t)
					(point)
				      (point-max)
				      )
				    )
				  )
			)
		  (if (re-search-forward "\\bevent\\b" end-point t)
		      (progn
			(setq end-point (save-excursion
					  (if (re-search-forward ";" end-point t)
					      (point)
					    end-point
					    )
					  ))
;			(message "Dans les event du where")
;			(sit-for 3)
			(forward-word 1)
			(while (< (point) end-point)
			  (backward-word 1)
			  (setq llochorl (append1 llochorl (buffer-substring (point)
									     (progn
									       (forward-word 1)
									       (point)
									       ))))
			  (forward-word 1)
			  )
			)
		    )
		  )
	      )
	    (message "           ") 
	    (while llochorl
	      (goto-char start-point)
	      (setq clock-name (car llochorl))
	      (setq llochorl (cdr llochorl))
;	      (message "Clock traitee : %s" clock-name)
;	      (sit-for 0)
	      (if (re-search-forward (concat clock-name "[	 ]*:=") stop-point t)
		  (progn
;		    (message "Definition de Clock trouve : %s" clock-name)
;		    (sit-for 3)
		    (setq tree (append1 tree (sigtra-loc-extract-clocks-tree clock-name)))
					;	      (message "Retour clock traitee : %s" clock-name)
					;	      (sit-for 7)
		    )
					;	    (message "Retour process traite : %s" signal-process-name)
					;	    (sit-for 7)
		)
	      )
	    )
	)
      tree
      )
    )
  )

(defun sigtra-extract-clocks-tree ()
  "Extraction et decompilation de l'arbre des horloges"
  (interactive)
  (let (sig-proc-name clocks-tree (fioritures-respons ""))
    (goto-char (point-min))
    (if (re-search-forward (concat "^[	 ]*process ") (point-max) t)
	(progn
	  (setq sig-proc-name (buffer-substring (point) (save-excursion
							  (search-forward "=")
							  (- (point) 1)
							  )))
	  (goto-char (point-min))
	  (setq clocks-tree (sigtra-loc-extract-clocks-tree sig-proc-name))
	  )
      (setq clocks-tree nil)
      )
    (setq sig-tree clocks-tree)
    (if (not clocks-tree-buf)
	(setq clocks-tree-buf (concat (substring (buffer-name) 0 (if (string-match "\\." (buffer-name))
								     (string-match "\\." (buffer-name))
								   (length (buffer-name))
								   ))
				      "_CT.sig"))
      )
    (save-excursion
      (set-buffer (get-buffer-create clocks-tree-buf))
      (make-local-variable 'ct-tree)
      (setq ct-tree clocks-tree)
      (make-local-variable 'last-clock)
      (setq last-clock " ")
      (make-local-variable 'ass-sub-tree)
      (setq ass-sub-tree nil)
      (make-local-variable 'max-prof)
      (setq max-prof 0)
      (kill-region (point-min) (point-max))
      (insert "
")
      (modify-syntax-entry ?_ "w")
      (setq fioritures-respons (read-from-minibuffer "Decompilation avec fioritures (O/N) [O] : "))
      (if (or (string-equal fioritures-respons "")
	      (string-equal fioritures-respons "O")
	      (string-equal fioritures-respons "o"))
	  (progn
	    (setq pretty-decomp-pos 0)
	    (if (cdr (cdr clocks-tree))
		(sigtra-pretty-decomp-clocks clocks-tree 0 1)
	      (sigtra-pretty-decomp-clocks (car (cdr clocks-tree)) 0 1)
	      )
	    )
	(if (cdr (cdr clocks-tree))
	    (sigtra-decomp-clocks clocks-tree 0)
	  (sigtra-decomp-clocks (car (cdr clocks-tree)) 0)
	  )
	)
      )
    (display-buffer (get-buffer-create clocks-tree-buf))
    )
  )


(defun sigtra-decomp-clocks (sub-tree ind-niv)
  "Fonction locale de decompilation de l'arbre des horloges"
  (let (laux clock cur-ind-niv num-clock (i 0))
    (if sub-tree
	(progn
	  (setq cur-ind-niv (+ 1 ind-niv))
	  (setq laux (cdr sub-tree))
	  (setq clock (car sub-tree))
	  (while (< i ind-niv)
	    (insert "	")
	    (setq i (+ 1 i))
	    )
	  (insert (concat clock "
"))
	  (while laux
;	    (message "in while cur-ind-niv : %d for %s " cur-ind-niv clock)
;	    (sit-for 1)
;	    (message "---------------------------------")
;	    (sit-for 1)
	    (sigtra-decomp-clocks (car laux) cur-ind-niv)
	    (setq laux (cdr laux))
	    )
	  )
      )
    )
  )


(defun sigtra-pretty-decomp-clocks (sub-tree ind-niv child-num)
  "Fonction locale de decompilation avec fioritures de l'arbre des horloges"
  (let (laux clock cur-ind-niv cur-pos cur-child-num op-line)
    (if sub-tree
	(progn
	  (goto-char (point-min))
	  (if (> ind-niv 0)
	      (progn
		(forward-line (* 3 ind-niv))
		(beginning-of-line)
		(setq op-line (- (* 3 ind-niv) (count-lines (point-min) (point))))
;	  (message "indentation courante : %d ; niveau courant : %d ; op-line : %d" (eval pos) ind-niv op-line)
;	  (sit-for 6)
		(end-of-line)
		(open-line op-line)
		(next-line op-line)
		(previous-line 2)
		(end-of-line)
;		(message "avant niveau 1")
;		(sit-for 5)
		(if (= child-num 1)
		    (progn
		      (while (< (current-column) pretty-decomp-pos)
			(insert " ")
			)
		      (insert "  |")
		      )
		  (while (< (current-column) pretty-decomp-pos)
		    (insert "_")
		    )
		  (insert "_")
		  )
		(next-line 1)
		(end-of-line)
;		(message "avant niveau 2")
;		(sit-for 5)
		(while (< (current-column) pretty-decomp-pos)
		  (insert " ")
		  )
		(if (= child-num 1)
		    (insert "  |")
		  (insert " \\")
		  )
		(next-line 1)
		)
	    )
	  (setq cur-ind-niv (+ 1 ind-niv))
	  (setq laux (cdr sub-tree))
	  (setq clock (car sub-tree))
	  (end-of-line)
;	  (message "avant niveau 3")
;	  (sit-for 5)
	  (while (< (current-column) pretty-decomp-pos)
	    (insert " ")
	    )
					;	  (if (= child-num 1)
					;	      (insert " ")
					;	    )
	  (insert clock)
	  (setq cur-pos (+ (current-column) 2))
	  (setq cur-child-num 1)
	  (while laux
	    (sigtra-pretty-decomp-clocks (car laux) cur-ind-niv cur-child-num)
	    (setq cur-child-num (+ 1 cur-child-num))
	    (setq laux (cdr laux))
	    )
	  (if (> cur-pos pretty-decomp-pos)
	      (progn
					;		(message "avant setq pretty-decomp-pos cur-pos")
					;		(sit-for 1)
		(setq pretty-decomp-pos cur-pos)
					;		(message "apres set pretty-decomp-pos cur-pos")
					;		(sit-for 1)
		)
	    )
	  )
      )
    )
  )


(defun sigtra-loc-renumber-clocks (sub-tree str-num)
  "Fonction locale de renumerotation des horloges en fonction de leur profondeur
dans l'arbre."
  (goto-char (point-min))
  (let (laux clock niv-ind num-clock)
    (if sub-tree
	(progn
	  (setq laux (cdr sub-tree))
	  (setq clock (car sub-tree))
	  (setq niv-ind 0)
	  (goto-char (point-min))
	  (setq num-clock (substring clock 2 (string-match "_H" clock)))
;	  (message "num-clock : %s" num-clock)
;	  (sit-for 4)
	  (replace-regexp (concat "\\bX_" num-clock "_X\\b") (concat "H_" str-num "_H"))
	   (while laux
;	     (message "in while niv-ind : %d " niv-ind)
;	     (sit-for 1)
	     (goto-char (point-min))
	     (setq niv-ind (+ niv-ind 1))
	     (sigtra-loc-renumber-clocks (car laux) (concat str-num "_" (int-to-string niv-ind)))
	     (setq laux (cdr laux))
	     )
	   )
      )
    )
  )

(defun sigtra-renumber-clocks ()
  "Renumerotation des horloges en fonction de leur profondeur dans l'arbre."
  (interactive)
  (goto-char (point-min))
  (if sig-tree
      (progn
	(replace-regexp "\\bH_\\([0-9_]*\\)_H\\b" "X_\\1_X")
	(sigtra-loc-renumber-clocks (car (cdr sig-tree)) "1")
	(goto-char (point-min))
	(if (re-search-forward "\\bX_[0-9_]_X\\b" (point-max) t)
	    (error "Il existe une ou plusieurs horloges qui n'existaient lors de la definition de l'arbre (X_num_X)")
	  )
	)
    (error "Arborescence non definie => F6")
    )
  )


(defun sigtra-loc-show-sig-clock-def (sig-name)
  "Fonction locale de visualisation de la definition d'une horloge ou d'un signal"
  (save-excursion
    (let (searched-str (point-def nil) sse)
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (goto-char (point-min))
      (if (re-search-forward (concat "\\b" sig-name "[	 ]*:=") (point-max) t)
	  (setq point-def (point))
	(goto-char (point-min))
	(if (re-search-forward (concat "{[^}]*\\b" sig-name "\\b[^}]*}[	 ]*:=") (point-max) t)
	    (setq point-def (point))
	  )
	)
      (if point-def
	  (sigtra-loc-show-pos)
	(modify-syntax-entry ?_ sse)
	(error "Definition non trouvee")
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )
			     
(defun sigtra-show-sig-clock-def ()
  "Visualisation de la definition d'une horloge ou d'un signal"
  (interactive)
  (save-excursion
    (let (sse (ind-h nil) sig-clock-name prompt sig-clock-name-respons)
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (if (string-equal "w" (char-to-string (char-syntax (char-after (point)))))
	  (forward-char 1)
	)
      (backward-word 1)
      (setq sig-clock-name (buffer-substring (point) (save-excursion
					   (forward-word 1)
					   (point)
					   )))
      (setq prompt (concat "Nom de l'horloge ou du signal (defaut : " sig-clock-name ") : "))
      (setq sig-clock-name-respons (read-from-minibuffer prompt))
      (if (string-equal sig-clock-name-respons "")
	  (sigtra-loc-show-sig-clock-def sig-clock-name)
	(sigtra-loc-show-sig-clock-def sig-clock-name-respons)
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )

(defun sigtra-loc-show-clock-of-sig (sig-name)
  "Fonction locale de visualisation de l'horloge d'un signal"
  (save-excursion
    (let (searched-str (point-def nil) sse clock-name)
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (goto-char (point-min))
      (if (re-search-forward (concat "\\bsynchro[	 ]*{[^}]*\\b" sig-name "\\b[^}]*}") (point-max) t)
	  (progn
	    (setq point-def (point))
	    (re-search-backward "\\bsynchro\\b")
	    (forward-word 2)
	    (backward-word 1)
	    (setq clock-name (buffer-substring (point) (save-excursion
							 (forward-word 1)
							 (point))))
	    )
	(goto-char (point-min))
	(if (re-search-forward (concat "_H[	 ]*:=[	 ]*\\bevent\\b[	 ]" sig-name) (point-max) t)
	    (progn
	      (setq point-def (point))
	      (search-backward ":=")
	      (backward-word 1)
	      (setq clock-name (buffer-substring (point) (save-excursion
							   (forward-word 1)
							   (point))))
	      )
	  )
	)
      (if point-def
	  (progn
	    (sigtra-loc-show-pos)
	    (message "Horloge de %s : %s" sig-name clock-name)
	    )
	(modify-syntax-entry ?_ sse)
	(error "Horloge non trouvee")
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )
			     
(defun sigtra-show-clock-of-sig ()
  "Visualisation de l'horloge d'un signal"
  (interactive)
  (save-excursion
    (let (sse (ind-h nil) sig-clock-name prompt sig-clock-name-respons)
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (if (string-equal "w" (char-to-string (char-syntax (char-after (point)))))
	  (forward-char 1)
	)
      (backward-word 1)
      (setq sig-clock-name (buffer-substring (point) (save-excursion
					   (forward-word 1)
					   (point)
					   )))
      (setq prompt (concat "Nom du signal (defaut : " sig-clock-name ") : "))
      (setq sig-clock-name-respons (read-from-minibuffer prompt))
      (if (string-equal sig-clock-name-respons "")
	  (sigtra-loc-show-clock-of-sig sig-clock-name)
	(sigtra-loc-show-clock-of-sig sig-clock-name-respons)
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )

(defun sigtra-loc-substitute-clock-by-def (sig-name)
  (save-excursion
    (let (searched-str sig-def (point-def nil) sse (beg-def-point nil) (end-def-point nil) (out-point nil)
		       make-repl (beg-expr-def-point nil) (beg-expr-point nil) (end-expr-point nil))
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (goto-char (point-min))
      (if (re-search-forward (concat "\\b" sig-name "[	 ]*:=") (point-max) t)
	  (progn 
	    (setq beg-expr-def-point (point))
	    (save-excursion
	      (backward-word 1)
	      (setq beg-def-point (point))
	      (backward-char 1)
	      (while (looking-at "[	 ]")
		(backward-char 1)
		)
	      (if (looking-at "[&;|]")
		  (progn
		    (setq beg-def-point (point))
		    (backward-char 1)
		    (if (looking-at "(")
			(setq beg-expr-point (point))
		      )
		    )
		)
	      )
	    (save-excursion
	      (if (re-search-forward "\\bwith\\b" (point-max) t)
		  (progn
		    (setq out-point (- (point) 5))
		    (setq end-def-point (- (point) 5))
		    )
		)
	      )
	    (save-excursion
	      (if (re-search-forward "[@]" (point-max) t)
		  (if end-def-point
		      (if (< (point) end-def-point)
			  (progn
			    (setq out-point (- (point) 2))
			    (setq end-def-point (- (point) 2))
			    )
			)
		    (progn
		      (setq out-point (- (point) 2))
		      (setq end-def-point (- (point) 2))
		      )
		    )
		)
	      )
	    (save-excursion
	      (if (re-search-forward "[&;|]" (point-max) t)
		  (if end-def-point
		      (if (< (point) end-def-point)
			  (setq end-def-point (- (point) 2))
			)
		    (setq end-def-point (- (point) 2))
		    )
		)
	      )
	    (save-excursion
	      (if (re-search-forward "\\bwhere\\b" (point-max) t)
		  (if end-def-point
		      (if (< (point) end-def-point)
			  (setq end-def-point (- (point) 6))
			)
		    (setq end-def-point (- (point) 6))
		    )
		)
	      )
	    (save-excursion
	      (if (re-search-forward "\\bend\\b" (point-max) t)
		  (if end-def-point
		      (if (< (point) end-def-point)
			  (setq end-def-point (- (point) 4))
			)
		    (setq end-def-point (- (point) 4))
		    )
		)
	      )
	    (if out-point
		(if (= out-point end-def-point)
		    (progn
		      (sigtra-loc-show-pos)
		      (modify-syntax-entry ?_ sse)
		      (error "Cette definition ne peut etre substituee")
		      )
		  )
	      )
	    (if end-def-point
		(progn
		  (goto-char end-def-point)
		  (save-excursion
		    (forward-char 1)
		    (if (looking-at "[&;|]")
			(progn
			  (forward-char 1)
			  (while (looking-at "[	 ]")
			    (forward-char 1)
			    )
			  (if (looking-at ")")
			      (setq end-expr-point (point))
			    )
			  )
		      )
		    )
		  (while (looking-at "[	 ]")
		    (backward-char 1)
		    )
		  (while (looking-at "[
]")
		    (backward-char 1)
		    )
		  (forward-char 1)
		  (setq sig-def (concat "(" (regexp-quote (buffer-substring beg-expr-def-point (point))) ")"))
		  (if end-expr-point
		      (setq end-def-point end-expr-point)
		    )
		  (while (and beg-expr-point end-expr-point)
		    (goto-char beg-expr-point)
		    (setq beg-expr-point nil)
		    (backward-char 1)
		    (while (looking-at "[
	 ]")
		      (backward-char 1)
		      )
		    (if (looking-at "[&;|]")
			(progn
			  (setq beg-def-point (point))
			  (backward-char 1)
			  (if (looking-at "(")
			      (setq beg-expr-point (point))
			    )
			  )
		      )
		    (goto-char end-expr-point)
		    (forward-char 1)
;		    (message "end-expr-point")
;		    (sit-for 3)
		    (while (looking-at "[
	 ]")
		      (forward-char 1)
		      )
		    (if (looking-at "[&;|]")
			(progn
			  (forward-char 1)
			  (if (looking-at ")")
			      (progn
				(setq end-expr-point (point))
				(setq end-def-point end-expr-point)
				)
			    )
			  )
		      )
		    )
		  (kill-region beg-def-point (+ end-def-point 1))
		  (goto-char (point-min))
		  (setq sig-name (concat "\\b" sig-name "\\b"))
		  (while (re-search-forward sig-name (point-max) t)
		    (setq make-repl nil)
		    (save-excursion
		      (backward-word 2)
		      (if (or (looking-at "\\bwhen\\b")
			      (looking-at "\\bdefault\\b")
			      (looking-at "\\bnot\\b")
			      (looking-at "\\bsynchro\\b")
			      )
			  (setq make-repl t)
			(forward-word 3)
			(backward-word 1)
			(if (or (looking-at "\\bdefault\\b")
				(looking-at "\\bwhen\\b")
				)
			    (setq make-repl t)
			  )
			)
		      )
		    (if make-repl
			(progn
;			  (sit-for-millisecs 500)
			  (backward-word 1)
			  (kill-word 1)
			  (insert sig-def)
			  )
		      )
		    )
		  )
	      (progn
		(sigtra-loc-show-pos)
		(modify-syntax-entry ?_ sse)
		(error "Cette definition ne peut etre substituee")
		)
	      )
	    )
	(goto-char (point-min))
	(if (re-search-forward (concat "{[^}]*\\b" sig-name "\\b[^}]*}[	 ]*:=") (point-max) t)
	    (progn
	      (sigtra-loc-show-pos)
	      (modify-syntax-entry ?_ sse)
	      (error "Cette definition ne peut etre substituee")
	      )
	  (modify-syntax-entry ?_ sse)
	  (error "Definition non trouvee")
	  )
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )

(defun sigtra-hnot-search ()
  "Recherche des horloges HNOT a partir de la position courante"
  (interactive)
  (if (re-search-forward ":=[	 ]*when[	 ]*([	 ]*([	 ]*not\\b")
      (backward-word 3)
    )
  )

			     
(defun sigtra-substitute-clock-by-def ()
  "Substitution d'une horloge par sa definition"
  (interactive)
  (save-excursion
    (let (sse (ind-h nil) sig-clock-name prompt sig-clock-name-respons)
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (if (string-equal "w" (char-to-string (char-syntax (char-after (point)))))
	  (forward-char 1)
	)
      (backward-word 1)
      (setq sig-clock-name (buffer-substring (point) (save-excursion
					   (forward-word 1)
					   (point)
					   )))
      (setq prompt (concat "Nom de l'horloge a substituer (defaut : " sig-clock-name ") : "))
      (setq sig-clock-name-respons (read-from-minibuffer prompt))
      (if (string-equal sig-clock-name-respons "")
	  (sigtra-loc-substitute-clock-by-def sig-clock-name)
	(sigtra-loc-substitute-clock-by-def sig-clock-name-respons)
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )

  
(defun sigtra-loc-show-sig-clock-occ (sig-name)
  (save-excursion
    (let ((proc-list nil) sse)
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (goto-char (point-min))
      (while (re-search-forward (concat "\\b" sig-name "\\b") (point-max) t)
	(save-excursion
	  (if (re-search-backward "\\bprocess\\b" (point-min) t)
	      (setq proc-list (append1 proc-list (buffer-substring (progn
								     (forward-word 2)
								     (backward-word 1)
								     (point)
								     )
								   (progn
								     (forward-word 1)
								     (point)
								     )
								   )))
	    )
	  )
	)
      (save-excursion
	(set-buffer (get-buffer-create clocks-tree-buf))
	(goto-char (point-min))
	(replace-string "*" " ")
	(while proc-list
	  (goto-char (point-min))
	  (if (re-search-forward (car proc-list) (point-max) t)
	      (progn
		(if (not (looking-at "$"))
		    (delete-char 1)
		  )
		(insert "*")
		)
	    )
	  (setq proc-list (cdr proc-list))
	  )
	)
      (display-buffer (get-buffer-create clocks-tree-buf))
      (modify-syntax-entry ?_ sse)
      )
    )
  )

(defun sigtra-show-sig-clock-occ ()
  "Visualisation des occurrences d'un signal ou d'une horloge dans les processus horloges"
  (interactive)
  (if sig-tree
      (save-excursion
	(let (sse (ind-h nil) sig-clock-name prompt sig-clock-name-respons)
	  (setq sse (char-to-string (char-syntax ?_)))
	  (modify-syntax-entry ?_ "w")
	  (if (string-equal "w" (char-to-string (char-syntax (char-after (point)))))
	      (forward-char 1)
	    )
	  (backward-word 1)
	  (setq sig-clock-name (buffer-substring (point) (save-excursion
							   (forward-word 1)
							   (point)
							   )))
	  (setq prompt (concat "Nom de l'horloge ou du signal a rechercher (defaut : " sig-clock-name ") : "))
	  (setq sig-clock-name-respons (read-from-minibuffer prompt))
	  (if (string-equal sig-clock-name-respons "")
	      (sigtra-loc-show-sig-clock-occ sig-clock-name)
	    (sigtra-loc-show-sig-clock-occ sig-clock-name-respons)
	    )
	  (modify-syntax-entry ?_ sse)
	  )
	)
    (error "Arborescence non definie => F6")
    )
  )

(defun sigtra-loc-show-sig-clock-in-clock-def (sig-name)
  (save-excursion
    (let ((proc-list nil) sse is-real-occ)
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (goto-char (point-min))
      (while (re-search-forward (concat "\\b" sig-name "\\b") (point-max) t)
	(setq is-real-occ nil)
;	(message "occurrence")
;	(sit-for 4)
	(if (not (looking-at "[	 ]*:="))
	    (save-excursion
	      (backward-word 2)
	      (if (or (looking-at "\\bwhen\\b")
		      (looking-at "\\bdefault\\b")
		      (looking-at "\\bnot\\b")
		      )
		  (setq is-real-occ t)
		(forward-word 3)
		(backward-word 1)
		(if (or (looking-at "\\bdefault\\b")
			(looking-at "\\bwhen\\b")
			)
		    (setq is-real-occ t)
		  )
		)
	      (if is-real-occ
		  (progn
		    (re-search-backward ":=" (point-min) t)
		    (backward-word 1)
;		    (message "clock")
;		    (sit-for 4)
		    (if (looking-at "H_[0-9_]*_H")
			(setq proc-list (append1 proc-list (buffer-substring (point)
									     (progn
									       (forward-word 1)
									       (point)
									       )
									     )))
		      )
		    )
		)
	      )
	  )
	)
      (save-excursion
	(set-buffer (get-buffer-create clocks-tree-buf))
	(goto-char (point-min))
	(replace-string "*" " ")
	(while proc-list
	  (goto-char (point-min))
	  (if (re-search-forward (car proc-list) (point-max) t)
	      (progn
		(if (not (looking-at "$"))
		    (delete-char 1)
		  )
		(insert "*")
		)
	    )
	  (setq proc-list (cdr proc-list))
	  )
	)
      (display-buffer (get-buffer-create clocks-tree-buf))
      (modify-syntax-entry ?_ sse)
      )
    )
  )


(defun sigtra-show-sig-clock-in-clock-def ()
  "Visualisation des occurrences d'une horloge ou d'un signal dans la definition d'une horloge"
  (interactive)
  (if sig-tree
      (save-excursion
	(let (sse (ind-h nil) sig-clock-name prompt sig-clock-name-respons)
	  (setq sse (char-to-string (char-syntax ?_)))
	  (modify-syntax-entry ?_ "w")
	  (if (string-equal "w" (char-to-string (char-syntax (char-after (point)))))
	      (forward-char 1)
	    )
	  (backward-word 1)
	  (setq sig-clock-name (buffer-substring (point) (save-excursion
							   (forward-word 1)
							   (point)
							   )))
	  (setq prompt (concat "Nom de l'horloge ou du signal a rechercher (defaut : " sig-clock-name ") : "))
	  (setq sig-clock-name-respons (read-from-minibuffer prompt))
	  (if (string-equal sig-clock-name-respons "")
	      (sigtra-loc-show-sig-clock-in-clock-def sig-clock-name)
	    (sigtra-loc-show-sig-clock-in-clock-def sig-clock-name-respons)
	    )
	  (modify-syntax-entry ?_ sse)
	  )
	)
    (error "Arborescence non definie => F6")
    )
  )


(defun sigtra-loc-show-pos ()
  "Positionnement d'un marqueur en debut de ligne courante.
Si la ligne n'est pas visible coupure en deux de la fenetre
pour la rendre visible."
  (interactive)
  (if (pos-visible-in-window-p)
      (progn
	(beginning-of-line 1)
	(sit-for 1)
	(setq overlay-arrow-string "=>")
	(setq overlay-arrow-position (make-marker))
	(set-marker overlay-arrow-position (point))
	)
    (progn
      (split-window-vertically 4)
      (beginning-of-line 1)
      (setq overlay-arrow-string "=>")
      (setq overlay-arrow-position (make-marker))
      (set-marker overlay-arrow-position (point))
      (other-window 1)
      )
    )
  )

(defun sigtra-number-clocks ()
  "Visualisation dans un buffer de la numerotation des horloges en fonction
de leurs positions dans l'arbre."
  (defun sigtra-loc-number-clocks (sub-tree str-num)
    (let (laux clock niv-ind num-clock)
      (if sub-tree
	  (progn
	    (setq laux (cdr sub-tree))
	    (setq clock (car sub-tree))
	    (setq niv-ind 0)
	    (insert (concat clock " (" str-num ")"))
	    (open-line 1)
	    (next-line 1)
	    (while laux
					;	     (message "in while niv-ind : %d " niv-ind)
					;	     (sit-for 1)
	      (setq niv-ind (+ niv-ind 1))
	      (sigtra-loc-number-clocks (car laux) (concat str-num "," (int-to-string niv-ind)))
	      (setq laux (cdr laux))
	      )
	    )
	)
      )
    )
  
  (interactive)
  (let (num-sig-tree)
    (goto-char (point-min))
    (if sig-tree
	(progn
	  (if (not num-clocks-tree-buf)
	      (setq num-clocks-tree-buf (concat (substring (buffer-name) 0 (if (string-match "\\." (buffer-name))
									       (string-match "\\." (buffer-name))
									     (length (buffer-name))
									     ))
						"_NUMCLOCK.sig"))
	    )
	  (save-excursion
	    (setq num-sig-tree sig-tree)
	    (set-buffer (get-buffer-create num-clocks-tree-buf))
	    (modify-syntax-entry ?_ "w")
	    (kill-region (point-min) (point-max))
	    (sigtra-loc-number-clocks (car (cdr num-sig-tree)) "1")
	    )
	  (display-buffer (get-buffer-create num-clocks-tree-buf))
	  )
      (error "Arborescence non definie => F6")
      )
    )
  )

(defun sigtra-found-clock (clock sub-tree)
  (let (laux found)
    (if sub-tree
	(progn
	  (setq laux (cdr sub-tree))
	  (if (string-equal clock (car sub-tree))
	      (progn
		(setq ass-sub-tree sub-tree)
		t)
	    (progn
	      (setq found nil)
	      (while (and laux (not found))
		(setq found (sigtra-found-clock clock (car laux)))
		(setq laux (cdr laux))
		)
	      found
	      )
	    )
	  )
      )
    )
  )
(defun sigtra-def-max-prof (sub-tree niv)
  (let (laux sub-niv)
    (if sub-tree
	(progn
	  (setq laux (cdr sub-tree))
	  (setq sub-niv (+ 1 niv))
	  (if (< max-prof sub-niv)
	      (setq max-prof sub-niv)
	    )
	  (while laux
					;		(message "Laux in while : %s" laux)
					;		(sit-for 5)
	    (sigtra-def-max-prof (car laux) sub-niv)
	    (setq laux (cdr laux))
	    )
	  )
      )
    )
  )

(defun sigtra-move-left-sub-tree ()
  (interactive)
  (let (sse (clock nil) (min-rect 0) (min-x 0) i)
    (save-excursion
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (if (looking-at "\\b[a-zA-Z]")
	  (setq clock (buffer-substring (point) (progn
						 (forward-word 1)
						 (point))))
	(if (looking-at "[a-zA-Z0-9_]+")
	    (setq clock (buffer-substring (progn
					    (backward-word 1)
					    (point))
					  (progn
					    (forward-word 1)
					    (point))))
	  (backward-char 1)
	  (if (looking-at "[a-zA-Z0-9_]")
	      (setq clock (buffer-substring (progn
					      (backward-word 1)
					      (point))
					    (progn
					      (forward-word 1)
					      (point))))
	    )
	  )
	)
      (if clock
	  (progn
	    (backward-word 1)
	    (previous-line 1)
	    (forward-char 1)
	    (if (looking-at "\\\\")
		(if (or (string-equal clock last-clock)
			(sigtra-found-clock clock ct-tree))
		    (progn
		      (goto-char (point-min))
		      (if (re-search-forward (concat "\\b" clock "\\b") (point-max) t)
			  (progn
			    (if (not (string-equal clock last-clock))
				(progn
				  (setq max-prof -1)
				  (sigtra-def-max-prof ass-sub-tree -1)
				  )
			      )
			    (backward-word 1)
			    (backward-char 2)
			    (setq min-x (current-column))
			    (setq min-rect (save-excursion
					     (previous-line 2)
					     (beginning-of-line)
					     (forward-char min-x)
					     (point)))
			    (setq ok (looking-at " "))
			    (setq i 0)
			    (while (and (< i (* 3 max-prof)) ok)
			      (next-line 1)
			      (beginning-of-line)
			      (forward-char min-x)
			      (setq ok (looking-at " "))
			      (setq i (+ i 1))
			      )
			    (if ok
				(progn
				  (setq last-clock clock)
				  (goto-char min-rect)
				  (delete-char 1)
				  (forward-char 2)
				  (if (looking-at " ")
				      (insert " ")
				    (if (looking-at "_")
					(insert "_")
				      )
				    )
				  (next-line 1)
				  (beginning-of-line)
				  (forward-char min-x)
				  (delete-char 1)
				  (forward-char 3)
				  (if (looking-at " ")
				      (insert " ")
				    )
				  (next-line 1)
				  (beginning-of-line)
				  (forward-char min-x)
				  (delete-char 1)
				  (forward-word 1)
				  (if (looking-at " ")
				      (insert " ")
				    )
				  (setq i 0)
				  (while (< i (* 3 max-prof))
				    (next-line 1)
				    (beginning-of-line)
				    (forward-char min-x)
				    (delete-char 1)
				    (forward-char 4)
				    (while (looking-at "_")
				      (forward-char 1)
				      )
				    (setq but (+ (current-column) 1))
				    (if (looking-at " ")
					(insert " ")
				      )
				    (next-line 1)
				    (beginning-of-line)
				    (forward-char min-x)
				    (delete-char 1)
				    (beginning-of-line)
				    (forward-char but)
;				    (message "Avant but %d" but)
;				    (sit-for 3)
				    (if (looking-at " ")
					(insert " ")
				      )
				    (next-line 1)
				    (beginning-of-line)
				    (forward-char min-x)
				    (delete-char 1)
				    (beginning-of-line)
				    (forward-char but)
;				    (message "but")
;				    (sit-for 3)
				    (forward-word 1)
				    (if (looking-at " ")
					(insert " ")
				      )
				    (setq i (+ i 3))
				    )
				  )
			      (modify-syntax-entry ?_ sse)
			      (error "Deplacement du sous-arbre impossible")
			      )
			    )
			(modify-syntax-entry ?_ sse)
			(error "Erreur interne : horloge non trouvee")
			)
		      )
		  (modify-syntax-entry ?_ sse)
		  (error "Erreur interne : horloge non trouvee dans l'arbre")
		  )
	      (modify-syntax-entry ?_ sse)
	      (error "Deplacement du sous-arbre impossibleeeeee")
	      )
	    )
	(modify-syntax-entry ?_ sse)
	(error "Positionnez-vous sur une horloge")
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )

(global-set-key "" 'sigtra-move-left-sub-tree) ;; Ctrl <

(defun sigtra-move-right-sub-tree ()
  (interactive)
  (let (sse (clock nil) (max-larg 0) (min-rect 0) (min-x 0) i)
    (save-excursion
      (setq sse (char-to-string (char-syntax ?_)))
      (modify-syntax-entry ?_ "w")
      (if (looking-at "\\b[a-zA-Z]")
	  (setq clock (buffer-substring (point) (progn
						 (forward-word 1)
						 (point))))
	(if (looking-at "[a-zA-Z0-9_]+")
	    (setq clock (buffer-substring (progn
					    (backward-word 1)
					    (point))
					  (progn
					    (forward-word 1)
					    (point))))
	  (forward-char 1)
	  (if (looking-at "[a-zA-Z0-9_]+")
	      (setq clock (buffer-substring (point)
					    (progn
					      (forward-word 1)
					      (point))))
	    )
	  )
	)
      (if clock
	  (progn
	    (backward-word 1)
	    (setq i (current-column))
	    (previous-line 1)
	    (beginning-of-line)
	    (forward-char (+ i 1))
	    (if (looking-at "\\\\")
		(if (or (string-equal clock last-clock)
			(sigtra-found-clock clock ct-tree))
		    (progn
		      (goto-char (point-min))
		      (if (re-search-forward (concat "\\b" clock "\\b") (point-max) t)
			  (progn
			    (setq max-larg (current-column))
			    (backward-word 1)
			    (setq min-x (current-column))
;			    (message "ass-sub-tree : %s" ass-sub-tree)
;			    (sit-for 3)
			    (if (not (string-equal clock last-clock))
				(progn
				  (setq max-prof -1)
				  (sigtra-def-max-prof ass-sub-tree -1)
				  )
			      )
			    (save-excursion
			      (forward-word 1)
			      (setq i 0)
			      (setq ok (save-excursion
					 (or (looking-at "$")
					     (progn
					       (forward-char 1)
					       (if (looking-at " ")
						   (progn
						     (forward-char 1)
						     (looking-at " ")
						     )
						 )
					       )
					     )
					 )
				    )
			      (while (and (< i (* 3 max-prof)) ok)
				(next-line 1)
				(beginning-of-line)
				(forward-char (+ 3 min-x))
				(while (looking-at "_")
				  (forward-char 1)
				  )
				(if (looking-at "$")
				    (setq i (* 3 max-prof))
				  (progn
				    (setq max-larg (current-column))
				    (next-line 2)
				    (beginning-of-line)
				    (forward-char max-larg)
				    (forward-word 1)
				    (if (not (looking-at "$"))
					(progn
					  (if (looking-at " ")
					      (progn
						(forward-char 1)
						(if (not (looking-at "$"))
						    (setq ok (looking-at " "))
						  )
						)
;					    (message "ca coince")
;					    (sit-for 4)
					    (setq ok nil)
					    )
					  )
				      )
				    (setq i (+ 3 i))
				    )
				  )
				)
			      )
			    (if ok
				(progn
					;				  (message "avant decalage")
					;				  (sit-for 3)
				  (setq last-clock clock)
				  (previous-line 2)
				  (beginning-of-line)
				  (forward-char (+ min-x 1))
				  (if (looking-at "$")
				      (insert "_")
				    (if (looking-at " ")
					(progn
					  (delete-char 1)
					  (insert "_")
					  )
				      )
				    )
				  (next-line 1)
				  (beginning-of-line)
				  (forward-char (+ min-x 2))
				  (if (not (looking-at "$"))
				      (delete-char 1)
				    )
				  (backward-char 2)
				  (insert " ")
				  (next-line 1)
				  (beginning-of-line)
				  (forward-char min-x)
				  (insert " ")
				  (forward-word 1)
				  (if (not (looking-at "$"))
				      (delete-char 1)
				    )
				  (setq i 0)
				  (while (and (< i (* 3 max-prof)) ok)
				    (next-line 1)
				    (beginning-of-line)
				    (forward-char (+ 3 min-x))
				    (while (looking-at "_")
				      (forward-char 1)
				      )
				    (setq max-larg (current-column))
				    (if (not (looking-at "$"))
					(delete-char 1)
				      )
				    (beginning-of-line)
				    (forward-char min-x)
				    (insert " ")
				    (next-line 1)
				    (beginning-of-line)
				    (forward-char max-larg)
				    (save-excursion
				      (if (not (looking-at "$"))
					  (progn
					    (if (looking-at "\\\\")
						(progn
						  (forward-char 1)
						  (if (not (looking-at "$"))
						      (delete-char 1)
						    )
						  )
					      (delete-char 1)
					      )
					    )
					)
				      )
				    (beginning-of-line)
				    (forward-char min-x)
				    (insert " ")
				    (next-line 1)
				    (beginning-of-line)
				    (forward-char max-larg)
				    (forward-word 1)
				    (if (not (looking-at "$"))
					(delete-char 1)
				      )
				    (beginning-of-line)
				    (forward-char min-x)
				    (insert " ")
				    (setq i (+ 3 i))
				    )
				  )
			      (modify-syntax-entry ?_ sse)
			      (error "Deplacement du sous-arbre impossible")
			      )
			    )
			(modify-syntax-entry ?_ sse)
			(error "Erreur interne : horloge non trouvee")
			)
		      )
		  (modify-syntax-entry ?_ sse)
		  (error "Erreur interne : horloge non trouvee dans l'arbre")
		  )
	      (modify-syntax-entry ?_ sse)
	      (error "Deplacement du sous-arbre impossibleeeeeeeeee")
	      )
	    )
	(modify-syntax-entry ?_ sse)
	(error "Positionnez-vous sur une horloge")
	)
      (modify-syntax-entry ?_ sse)
      )
    )
  )

(global-set-key "" 'sigtra-move-right-sub-tree) ;; Ctrl >



(global-set-key "{2500z" 'scroll-right) ;; Ctrl R7

(global-set-key "{2501z" 'scroll-left) ;; Ctrl R9



(defun sigtra-scroll-right ()
  (interactive)
  (scroll-right 1)
  )

(defun sigtra-scroll-left ()
  (interactive)
  (scroll-left 1)
  )

(global-set-key "{2503z" 'sigtra-scroll-right) ;; Ctrl R13

(global-set-key "{2502z" 'sigtra-scroll-left) ;; Ctrl R15
		

