_PUBLIC_ _VIRTUAL_ const char *
IdentifierList::myClassName ()
{
  return "IdentifierList";
}

/* returns this successor
 */
_PUBLIC_ ref2IdentifierList IdentifierList::next ()
{
  return cdr;
}

/* returns this successor
 */
_PUBLIC_ ref2IdentifierList IdentifierList::last ()
{
  if (cdr == NIL_IdentifierList)
    return this;
  return cdr->last ();
}

/* returns this identifier
 */
_PUBLIC_ ref2Identifier IdentifierList::content ()
{
  return car;
}

/* returns this identifier name
 */
_PUBLIC_ char *
IdentifierList::get_name ()
{
  return car->get_name ();
}

/* PLG 2005-12-08
 * assumes: <anArgument> is not NIL (not checked)
 * types are assumed to be correct
 * success: sets expansion of <anArgument> as this parameter value
 * failure: this is not safe
 *          error message is printed
 */
_PUBLIC_ boolean IdentifierList::linkThis_expand_ (ref2Argument anArgument)
{
  if (car == NIL_Identifier)
    return pk_false;
  return car->link_expand_ (anArgument);
}

/* returns in <res> the first identifier in this list which name (thisString field) is <str>
 * <str> is assumed to differ from NIL
 * failure: <str> is not an identifier name in this list
 *          <res> remains unchanged
 */
_PUBLIC_ boolean IdentifierList::find (char *str, ref2Identifier * res)
{
  ref2IdentifierList
    tailIdentifierList = this;
  ref2Identifier
    curIdentifier;
  do
    {
      curIdentifier = tailIdentifierList->content ();
      if (curIdentifier->is_isomorphic (str))
	{
	  (*res) = curIdentifier;
	  return pk_true;
	}
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (tailIdentifierList != NIL_IdentifierList);
  return pk_false;
}

_PUBLIC_ boolean IdentifierList::find (ref2MetaArgument genericIdent, ref2Identifier * res)
{
  ref2IdentifierList
    tailIdentifierList = this;
  ref2Identifier
    curIdentifier;
  do
    {
      curIdentifier = tailIdentifierList->content ();
      if (curIdentifier->is_isomorphic (genericIdent))
	{
	  (*res) = curIdentifier;
	  return pk_true;
	}
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (tailIdentifierList != NIL_IdentifierList);
  return pk_false;
}

_PUBLIC_ boolean IdentifierList::find (ref2CommandArgument genericIdent, ref2Identifier * res)
{
  ref2IdentifierList
    tailIdentifierList = this;
  ref2Identifier
    curIdentifier;
  do
    {
      curIdentifier = tailIdentifierList->content ();
      if (curIdentifier->is_isomorphic (genericIdent))
	{
	  (*res) = curIdentifier;
	  return pk_true;
	}
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (tailIdentifierList != NIL_IdentifierList);
  return pk_false;
}

/* 
 * <identName> is assumed to differ from NIL
 * <isAList> is pk_true iff identifier is to be created with list type
 * success: <identName> was not an identifier name in this list
 *          a new identifier named <identName> is appent to the end of this list and retruned in <created_identifier>
 * failure: <identName> is an identifier name in this list
 *          <created_identifier> remains unchanged
 */
_PUBLIC_ boolean IdentifierList::append_uniqueNewIdentifier (char *identName, boolean isAList, ref2Identifier * created_identifier)
{
  ref2IdentifierList
    tailIdentifierList = this;
  ref2Identifier
    curIdentifier;
  do
    {
      curIdentifier = tailIdentifierList->content ();
      if (curIdentifier->is_isomorphic (identName))
	return pk_false;
      if ((tailIdentifierList->cdr) == NIL_IdentifierList)
	{
	  (*created_identifier) = new Identifier (identName, isAList);
	  (tailIdentifierList->cdr) = new IdentifierList (*created_identifier);
	  return pk_true;
	}
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (pk_true);
}

/* 
 * <identName> is assumed to differ from NIL
 * <isAList> is pk_true iff identifier is to be created with list type
 * success: <identName> was not an identifier name in this list
 *          a new identifier named <identName> is appent to the end of this list and retruned in <created_identifier>
 * failure: <identName> is an identifier name in this list
 *          <created_identifier> remains unchanged
 */
_PUBLIC_
  boolean IdentifierList::append_uniqueNewIdentifier (char *identName, boolean isAList, ref2Argument formula, ref2Identifier * created_identifier)
{
  ref2IdentifierList
    tailIdentifierList = this;
  ref2Identifier
    curIdentifier;
  do
    {
      curIdentifier = tailIdentifierList->content ();
      if (curIdentifier->is_isomorphic (identName))
	return pk_false;
      if ((tailIdentifierList->cdr) == NIL_IdentifierList)
	{
	  (*created_identifier) = new Identifier (identName, isAList, formula);
	  (tailIdentifierList->cdr) = new IdentifierList (*created_identifier);
	  return pk_true;
	}
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (pk_true);
}

/* 
 * <metaIdent> is assumed to differ from NIL
 * <isAList> is pk_true iff identifier is to be created with list type
 * success: <metaIdent> was not an identifier name in this list
 *          a new identifier named <metaIdent> is appent to the end of this list and retruned in <created_identifier>
 * failure: <metaIdent> is an identifier name in this list
 *          <created_identifier> remains unchanged
 */
_PUBLIC_
  boolean IdentifierList::append_uniqueNewIdentifier (ref2MetaContext owner, ref2MetaArgument metaIdent, boolean isAList, ref2Argument formula,
						      ref2Identifier * created_identifier)
{
  // logDEBUG_in("append_uniqueNewIdentifier ref2MetaArgument");
  ref2IdentifierList
    tailIdentifierList = this;
  ref2Identifier
    curIdentifier;
  do
    {
      curIdentifier = tailIdentifierList->content ();
      if (curIdentifier->is_isomorphic (metaIdent))
	return pk_false;
      if ((tailIdentifierList->cdr) == NIL_IdentifierList)
	{
	  (*created_identifier) = new MetaIdentifier (owner, (ref2Argument) metaIdent, isAList, formula);
	  (tailIdentifierList->cdr) = new IdentifierList (*created_identifier);
	  return pk_true;
	}
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (pk_true);
}

/* 
 * <metaIdent> is assumed to differ from NIL
 * <isAList> is pk_true iff identifier is to be created with list type
 * success: <metaIdent> was not an identifier name in this list
 *          a new identifier named <metaIdent> is appent to the end of this list and retruned in <created_identifier>
 * failure: <metaIdent> is an identifier name in this list
 *          <created_identifier> remains unchanged
 */
_PUBLIC_
  boolean IdentifierList::append_uniqueNewIdentifier (ref2MetaContext owner, ref2CommandArgument commandIdent, boolean isAList, ref2Argument formula,
						      ref2Identifier * created_identifier)
{
  // logDEBUG_in("append_uniqueNewIdentifier ref2CommandArgument");
  ref2IdentifierList
    tailIdentifierList = this;
  ref2Identifier
    curIdentifier;
  do
    {
      curIdentifier = tailIdentifierList->content ();
      if (curIdentifier->is_isomorphic (commandIdent))
	return pk_false;
      if ((tailIdentifierList->cdr) == NIL_IdentifierList)
	{
	  (*created_identifier) = new MetaIdentifier (owner, (ref2Argument) commandIdent, isAList, formula);
	  (tailIdentifierList->cdr) = new IdentifierList (*created_identifier);
	  return pk_true;
	}
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (pk_true);
}

boolean
IdentifierList::append_uniqueNewIterator (char *identName, ref2Iterator iterator)
{
  ref2IdentifierList tailIteratorList = this;
  ref2Iterator curIterator;
  char *curName;
  do
    {
      curIterator = (ref2Iterator) tailIteratorList->content ();
      curName = curIterator->get_name ();
      if (curName != NIL_String)
	if (strcmp (curName, identName) == 0)
	  return pk_false;
      if ((tailIteratorList->cdr) == NIL_IdentifierList)
	{
	  (tailIteratorList->cdr) = new IdentifierList (iterator);
	  return pk_true;
	}
      tailIteratorList = tailIteratorList->cdr;
    }
  while (pk_true);
}

/* ========== Writer
 */

/* 
 */
void
IdentifierList::spaceThis_name (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (car != NIL_Identifier)
    car->print_nameValue_sp (printingContext, outBuffer);
}

/* 
 */
void
IdentifierList::printThis_typedName (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (car != NIL_Identifier)
    car->print_typedName_sp (printingContext, outBuffer);
}

/* prints list of names beginning with this, separated by spaces
 */
void
IdentifierList::print_sp (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2IdentifierList tailIdentifierList = this;
  while (tailIdentifierList != NIL_IdentifierList)
    {
      tailIdentifierList->spaceThis_name (printingContext, outBuffer);
      tailIdentifierList = tailIdentifierList->cdr;
    }
}

/* prints list of names beginning with this, separated by spaces
 * a list typed name is preceded by list mark
 */
void
IdentifierList::print_typedFormula (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2IdentifierList tailIdentifierList = this;
  while (tailIdentifierList != NIL_IdentifierList)
    {
      tailIdentifierList->printThis_typedName (printingContext, outBuffer);
      tailIdentifierList = tailIdentifierList->cdr;
    }
}

/* 
 */
void
IdentifierList::printThis_value (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  if (car != NIL_Identifier)
    {
      car->print_expandedValue (printingContext, outBuffer, pk_true);
      outBuffer->space ();
    }
}

/* prints list of _values beginning with this, separated by spaces
 */
void
IdentifierList::print_expandedValue (ref2Lexical printingContext, ref2IntePrinter outBuffer)
{
  ref2IdentifierList tailIdentifierList = this;
  while (tailIdentifierList != NIL_IdentifierList)
    {
      tailIdentifierList->printThis_value (printingContext, outBuffer);
      tailIdentifierList = tailIdentifierList->cdr;
    }
}

/* ========== Log
 */
/* 
 */
void
IdentifierList::logExtendedDebug_this ()
{
  if (car == NIL_Identifier)
    LOG_STR_LN ("== is NIL");
  else
    car->logExtendedDebug_formula_ln (1);
}

void
IdentifierList::logDebug_this (ref2Context aContext, int depth)
{
  if (car == NIL_Identifier)
    LOG_STR_SP ("??");
  else
    car->logDebug_formula_sp (depth);
}

void
IdentifierList::logThis_formula_ ()
{
  if (car == NIL_Identifier)
    LOG_STR_SP ("??");
  else
    car->log_formula_ ();
}

_PUBLIC_ _VIRTUAL_ void
IdentifierList::logExtendedDebug_formula_ln (int level)
{
  ref2IdentifierList tailIdentifierList = this;
  LOG_STR_SP_STR_LN (myClassName (), ", my identifiers are");
  while (tailIdentifierList != NIL_IdentifierList)
    {
      tailIdentifierList->logExtendedDebug_this ();
      tailIdentifierList = tailIdentifierList->cdr;
    }
}

_PUBLIC_ _VIRTUAL_ void
IdentifierList::logDEBUG_formula_ln (ref2Context aContext, int depth)
{
  ref2IdentifierList tailIdentifierList = this;
  LOG_STR_SP_STR_LN (myClassName (), ", my identifiers are");
  while (tailIdentifierList != NIL_IdentifierList)
    {
      tailIdentifierList->logDebug_this (aContext, depth);
      tailIdentifierList = tailIdentifierList->cdr;
    }
  LOG_LN ();
}

_PUBLIC_ _VIRTUAL_ void
IdentifierList::log_formula_ ()
{
  ref2IdentifierList tailIdentifierList = this;
  LOG_STR_SP_STR_LN (myClassName (), ", my identifiers are");
  while (tailIdentifierList != NIL_IdentifierList)
    {
      tailIdentifierList->logThis_formula_ ();
      tailIdentifierList = tailIdentifierList->cdr;
    }
  LOG_LN ();
}

/* ========== Generator
 */

/* PLG 2005-12-08
 * success: sets expansion of elements in <actualValues> as formal parameter values
 * failure: this is not safe or expansion failed or types do not fit
 *          error message is printed
 */
boolean
IdentifierList::listLink_listExpand_ (ref2ArgumentList actualValues)
{
  ref2IdentifierList tailIdentifierList = this;
  ref2ArgumentList tailArgumentList = actualValues;
  ref2Argument curActualArgument;

  // openDEBUG_in("listLink_listExpand_");
  do
    {
      if (tailArgumentList == NIL_ArgumentList)
	{
	  Soundness::OUPS ("too few actual parameters in IdentifierList::listLink_listExpand_");
	  // closeDEBUG_in("listLink_listExpand_ false");
	  return pk_false;
	}
      curActualArgument = tailArgumentList->content ();
      if (!(tailIdentifierList->linkThis_expand_ (curActualArgument)))
	{
	  // closeDEBUG_in("listLink_listExpand_ false");
	  return pk_false;
	}
      tailArgumentList = tailArgumentList->next ();
      tailIdentifierList = tailIdentifierList->cdr;
    }
  while (tailIdentifierList != NIL_IdentifierList);
  if (tailArgumentList == NIL_ArgumentList)
    {
      // closeDEBUG_in("listLink_listExpand_ pk_true");
      return pk_true;
    }
  Soundness::OUPS ("too much actual parameters in IdentifierList::listLink_listExpand_");
  // closeDEBUG_in("listLink_listExpand_ false");
  return pk_false;
}

/* PLG 2005-12-07
 * success: this referred to identifier is a parameter or a static variable
 *              this referred to identifier is a parameter 
 *                   sets this_argument_value equal to this_linkedIdentifier actual string list value
 *          <isEmptyList> is set to pk_true if this_argument_value is an empty list
 *          <isEmptyList> is set to false if this_argument_value is a not empty list
 * failure: this referred to identifier is  NIL, in error or does not have list type 
 *          system error message is printed
 */
_PUBLIC_ boolean IdentifierList::linkThis_expandedList_IdVal_ (boolean * isEmptyList)
{
  // logDEBUG_in("linkThis_expandedList_IdVal_");
  if (car == NIL_Identifier)
    {
      unsafe_use ("illegal call to Iterator::linkThis_expandedList_IdVal_, Iterator is NIL");
      return pk_false;
    }
  return car->link_expandedList_IdVal_ (isEmptyList);
}

/* PLG 2005-12-07
 * success: all referred to identifiers are parameters or static variables
 *              those referred to identifier list values are expanded
 *          <hasEmptyList> is set to pk_true if some identifier is linked to an empty list
 *          <hasEmptyList> is set to false if all identifiers are linked to not empty lists
 * failure: some identifier is  NIL, in error or does not have list type 
 */
_PUBLIC_ boolean IdentifierList::link_iteratorList_ (boolean * hasEmptyList)
{
  // logDEBUG_in("link_iteratorList_");
  ref2IdentifierList
    tailIteratorList = this;
  boolean
    isThisEmptyList = pk_false;

  (*hasEmptyList) = pk_false;
  while (tailIteratorList != NIL_IdentifierList)
    {
      if (tailIteratorList->linkThis_expandedList_IdVal_ (&isThisEmptyList))
	{
	  if (isThisEmptyList)
	    (*hasEmptyList) = pk_true;
	}
      else
	return pk_false;
      tailIteratorList = tailIteratorList->next ();
    }
  return pk_true;
}

/* PLG 2005-12-08
 * this_expandedArgument is assumed to be linked to a list argument
 * success: sets this current (sub-)list as this_expandedArgument 
 * failure: this is safe, this expanded list value is empty
 *               <correct> remains unchanged;
 *          this is not safe, or this identifier is NIL
 *               system error message is printed
 *               <correct> is set to pk_false;
 */
_PUBLIC_ boolean IdentifierList::downThisCurrentList (boolean * correct)
{
  // logDEBUG_in("downThisCurrentList");
  if (car != NIL_Identifier)
    return car->downCurrentList (correct);
  (*correct) = pk_false;
  Soundness::OUPS ("illegal call to Iterator::downThisCurrentList(), identifier is NIL");
  return pk_false;
}

/* PLG 2005-12-08
 * this_expandedArgument is assumed to be linked to a list argument
 * success: this current sublist is <e,L>, <L> is not empty
 *          sets this current sublist as <L>
 * failure: 
 *          this is safe, this current sublist has at most 1 element
 *          this is not safe, or this identifier is NIL
 *               system error message is printed
 */
_PUBLIC_ boolean IdentifierList::stepThisCurrentList ()
{
  if (car != NIL_Identifier)
    return car->stepCurrentList ();
  Soundness::OUPS ("illegal call to Iterator::stepThisCurrentList(), identifier is NIL");
  return pk_false;
}

/* PLG 2005-12-08
 * this_expandedArgument is assumed to be linked to a list argument
 * success: sets this current (sub-)list as this_expandedArgument 
 * failure: this is safe, this expanded list value is empty
 *               <correct> remains unchanged;
 *          this is not safe, or this identifier is NIL
 *               system error message is printed
 *               <correct> is set to pk_false;
 */
_PUBLIC_ boolean IdentifierList::coDownThisCurrentList (boolean * correct)
{
  // logDEBUG_in("coDownThisCurrentList");
  ref2IdentifierList
    curr_list = this;
  while (curr_list != NIL_IdentifierList)
    {
      if ((curr_list->car) == NIL_Identifier)
	{
	  (*correct) = pk_false;
	  Soundness::OUPS ("illegal call to Iterator::coDownThisCurrentList(), identifier is NIL");
	  return pk_false;
	}
      if (!((curr_list->car)->downCurrentList (correct)))
	return pk_false;
      curr_list = curr_list->cdr;
    }
  return pk_true;
}

/* PLG 2005-12-08
 * this_expandedArgument is assumed to be linked to a list argument
 * success: this current sublist is <e,L>, <L> is not empty
 *          sets this current sublist as <L>
 * failure: 
 *          this is safe, this current sublist has at most 1 element
 *          this is not safe, or this identifier is NIL
 *               system error message is printed
 */
_PUBLIC_ boolean IdentifierList::coStepThisCurrentList ()
{
  ref2IdentifierList
    curr_list = this;
  while (curr_list != NIL_IdentifierList)
    {
      if ((curr_list->car) == NIL_Identifier)
	{
	  Soundness::OUPS ("illegal call to Iterator::coStepThisCurrentList(), identifier is NIL");
	  return pk_false;
	}
      if (!((curr_list->car)->stepCurrentList ()))
	return pk_false;
      curr_list = curr_list->cdr;
    }
  return pk_true;
}

/* ========== Builder
 */

/* creates and returns a new string list cell named <name>
 * used in IdentifierList_static
 */
IdentifierList::IdentifierList (ref2Identifier anIdentifier)
{
  car = anIdentifier;
  cdr = NIL_IdentifierList;
}

/* creates and returns a new string list cell named <aString>
 * used in IdentifierList_static
 */
IdentifierList::~IdentifierList ()
{
  if (cdr != NIL_IdentifierList)
    delete cdr;
}
