_PUBLIC_ _VIRTUAL_ const char *
MetaContext::myClassName ()
{
  return "MetaContext";
}

/* ==========  ==========  ==========  ========== TITLE: these attributes
 */
_VIRTUAL_ ref2IdentifierList
MetaContext::get_parameterList ()
{
  return ((ref2DynamicContext) this_nestingContext)->get_parameterList ();
}


/* ==========  ==========  ==========  ========== TITLE: DEBUG
 */

_PUBLIC_ _VIRTUAL_ void
MetaContext::logDebug (int level)
{
  if (these_genericElements == NIL_GenericElementList)
    log_NILComponent_ln (level, "node generator");
  else
    {
      ref2GenericElementList cur_generic = these_genericElements;
      log_subTitle_ln (level, "this node generator is");
      while (cur_generic != NIL_GenericElementList)
	{
	  cur_generic->logExtendedDebug_formula_ln (level + 1);
	  LOG_LN ();
	  cur_generic = cur_generic->next ();
	}
    }
}

/* ==========  ==========  ==========  ========== TITLE: PRINTING
 */

_VIRTUAL_ void
MetaContext::print_funcProfile_ln (ref2Lexical printingContext, ref2IntePrinter outBuffer, const char *funcName)
{
  ((ref2DynamicContext) this_nestingContext)->print_funcProfile_ln (printingContext, outBuffer, funcName);
}

_PUBLIC_ _VIRTUAL_ void
MetaContext::print (ref2IntePrinter outBuffer)
{
  // logDEBUG_in("print");
  if (these_genericElements != NIL_GenericElementList)
    {
      outBuffer->line ();
      these_genericElements->print_formula (this_lexicalContext, outBuffer);
    }				/* PLG HERE vs AncestorFunctionElemen */
}

/* ==========  ==========  ==========  ========== TITLE: FATAL
 */

void
MetaContext::ASFatal_missingDecl (boolean printMess, const char *category, ref2Argument id_name)
{
  this_lexicalContext->ASFatal_missingDecl (printMess, "generic identifier", "");
}

void
MetaContext::ASFatal_doubleDecl (ref2Argument id_name)
{
  this_lexicalContext->ASFatal_doubleDecl ("generic identifier");
}

void
MetaContext::fatal_typeIdentifier_ (ref2Argument id_name)
{
  error_typeIdentifier_ ("generic identifier");
  set_fatal ();
}

/* ==========  ==========  ==========  ========== TITLE: SPECIFIC ACTIONS
 */


/* ========== Declaration linking
*/
_PUBLIC_ _VIRTUAL_ boolean MetaContext::parameterListLink_listExpand_ (ref2ArgumentList parameterValues)
{
  // logDEBUG_in("parameterListLink_listExpand_");
  ((ref2DynamicContext) this_nestingContext)->parameterListLink_listExpand_ (parameterValues);
  return pk_true;
}

/* ========== Declaration linking
*/
_PUBLIC_ _VIRTUAL_ boolean MetaContext::getIf_string_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_string_id_ 1, ident is", CString_name);
  return DynamicContext::getIf_string_id_ (CString_name, correct, argument_res);
}

/* 2005-01-17
 * this_nestingContext not NIL (not checked)
 * <id_name> is a (not empty) genric identifier (not checked)
 * success: <id_name> is a dynamic identifier which declaration is (recursively) found in a dynamic context
 *              <argument_res> is a (NamedArgument) reference to this identifier declaration
 *              <argument_res> scope is dynamic
 *          <id_name> is a static identifier which declaration,  not found in nested dynamic context, 
 *                 has been recursively found in a static context
 *              <argument_res> is the (static) StringArgument value of this static identifier; 
 *              <argument_res> scope is static
 *          <argument_res> type is string
 *          <correct> remains unchanged
 * failure: <id_name> declaration has not been (recursively) found
 *              <correct> remains unchanged
 *          <id_name> declaration has unexpected type
 *              <correct> is set to false
 *              error message is printed
 *          <argument_res> remains unchanged
 */
_PUBLIC_ _VIRTUAL_ boolean MetaContext::getIf_string_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_string_id_ 2");
  /* DEBUG id_name->log_formula_ln(); */
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  if (id_res->has_string_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_false);
	      return pk_true;
	    }
	  fatal_typeIdentifier_ ((ref2Argument) id_name);
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_string_id_ (id_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean MetaContext::getIf_string_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_string_id_");
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  if (id_res->has_string_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_false);
	      return pk_true;
	    }
	  fatal_typeIdentifier_ ((ref2Argument) id_name);
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_string_id_ (id_name, correct, argument_res);
}

_VIRTUAL_ boolean
MetaContext::getIf_list_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_list_id_, ident is", CString_name);
  return DynamicContext::getIf_list_id_ (CString_name, correct, argument_res);
}

_VIRTUAL_ boolean
MetaContext::getIf_list_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_list_id_");
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  if (id_res->has_list_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_true);
	      return pk_true;
	    }
	  fatal_typeIdentifier_ ((ref2Argument) id_name);
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_list_id_ (id_name, correct, argument_res);
}

_VIRTUAL_ boolean
MetaContext::getIf_list_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  // logDEBUG_in("getIf_list_id_");
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  if (id_res->has_list_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_true);
	      return pk_true;
	    }
	  fatal_typeIdentifier_ ((ref2Argument) id_name);
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_list_id_ (id_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean MetaContext::getIf_argument_id_ (char *CString_name, boolean * correct, ref2Argument * argument_res)
{
  return DynamicContext::getIf_argument_id_ (CString_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean MetaContext::getIf_argument_id_ (ref2MetaArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  if (id_res->has_list_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_true);
	      return pk_true;
	    }
	  if (id_res->has_string_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_false);
	      return pk_true;
	    }
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_argument_id_ (id_name, correct, argument_res);
}

_PUBLIC_ _VIRTUAL_ boolean MetaContext::getIf_argument_id_ (ref2CommandArgument id_name, boolean * correct, ref2Argument * argument_res)
{
  if (this_localIdentifierList != NIL_IdentifierList)
    {
      ref2Identifier
	id_res;
      if (this_localIdentifierList->find (id_name, &id_res))
	{
	  if (id_res->has_list_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_true);
	      return pk_true;
	    }
	  if (id_res->has_string_type ())
	    {
	      (*argument_res) = new NamedArgument (id_res, pk_false);
	      return pk_true;
	    }
	  (*correct) = pk_false;
	  return pk_false;
	}
    }
  return this_nestingContext->getIf_argument_id_ (id_name, correct, argument_res);
}

_PRIVATE_
  boolean MetaContext::append_uniqueNewMetaIdentifier (ref2MetaArgument metaName, boolean isAList, ref2Argument formula,
						       ref2Identifier * createdIdent)
{
  // logDEBUG_in("append_uniqueNewMetaIdentifier");
  if (this_localIdentifierList == NIL_IdentifierList)
    {
      // logDEBUG_in("append_uniqueNewMetaIdentifier, this_localIdentifierList is empty");
      (*createdIdent) = new MetaIdentifier (this, (ref2Argument) metaName, isAList, formula);
      this_localIdentifierList = new IdentifierList (*createdIdent);
      // logDebug_identifierList_ln("DEBUG MetaContext::append_uniqueNewMetaIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  // logDebug_identifierList_ln("DEBUG MetaContext::append_uniqueNewMetaIdentifier, this_localIdentifierList is ----------------");
  if (this_localIdentifierList->append_uniqueNewIdentifier (this, metaName, isAList, formula, createdIdent))
    {
      // logDebug_identifierList_ln("DEBUG MetaContext::append_uniqueNewMetaIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  ASFatal_doubleDecl ((ref2Argument) metaName);
  return pk_false;
}

_PRIVATE_
  boolean MetaContext::append_uniqueNewCommandIdentifier (ref2CommandArgument commandName, boolean isAList, ref2Argument formula,
							  ref2Identifier * createdIdent)
{
  // logDEBUG_in("append_uniqueNewCommandIdentifier");
  // logDEBUG_in("append_uniqueNewCommandIdentifier, formula  is");
  /* DEBUG formula->log_formula_ln();  */
  // log_ASDEBUG("MetaContext::append_uniqueNewCommandIdentifier");
  if (this_localIdentifierList == NIL_IdentifierList)
    {
      (*createdIdent) = new MetaIdentifier (this, (ref2Argument) commandName, isAList, formula);
      this_localIdentifierList = new IdentifierList (*createdIdent);
      // logDebug_identifierList_ln("DEBUG MetaContext::append_uniqueNewCommandIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  // logDebug_identifierList_ln("DEBUG MetaContext::append_uniqueNewCommandIdentifier, this_localIdentifierList is ----------------");
  if (this_localIdentifierList->append_uniqueNewIdentifier (this, commandName, isAList, formula, createdIdent))
    {
      //logDebug_identifierList_ln("DEBUG MetaContext::append_uniqueNewCommandIdentifier, this_localIdentifierList becomes ----------------");
      return pk_true;
    }
  ASFatal_doubleDecl ((ref2Argument) commandName);
  return pk_false;
}

/* ==========  ==========  ==========  ========== TITLE: READING
 */

/*
 * success: FILE = PREIFX.$(..
 */
_PRIVATE_ boolean MetaContext::readIf_varMeta_fSpStar_nSp (char *CString_namePrefix, boolean * correct, ref2MetaArgument * id_name)
{
  // log_ASDEBUG("MetaContext::readIf_varMeta_fSpStar_nSp 1");
  boolean
    lCorrect = pk_true;
  ref2ArgumentList
    tailList;
  if (stepIf_referenceHeadedIdentifier_fSpStar_nSp (&lCorrect, &tailList))
    {
      (*id_name) = new MetaArgument (new ArgumentList (CString_namePrefix, tailList));
      return pk_true;
    }
  if (lCorrect)
    return pk_false;
  set_fatal ();
  (*correct) = pk_false;
  return pk_false;
}

_PRIVATE_ boolean MetaContext::readIf_varMeta_fSpStar_nSp (boolean * correct, ref2MetaArgument * id_name)
{
  // logDEBUG_in("readIf_varMeta_fSpStar_nSp 2, CString_namePrefix is empty");
  boolean
    lCorrect = pk_true;
  ref2ArgumentList
    tailList;
  if (stepIf_referenceHeadedIdentifier_fSpStar_nSp (&lCorrect, &tailList))
    {
      (*id_name) = new MetaArgument (tailList);
      /* DEBUG (*id_name)->log_formula_ln(); */
      return pk_true;
    }
  if (lCorrect)
    return pk_false;
  set_fatal ();
  (*correct) = pk_false;
  return pk_false;
}

_PRIVATE_ boolean MetaContext::read_varMeta_fSpStar_nSp (ref2MetaArgument * id_name)
{
  // log_ASDEBUG("MetaContext::read_varMeta_fSpStar_nSp");
  // fprintf(stderr, "DEBUG Context::read_varMeta_fSpStar_nSp");
  boolean
    lCorrect = pk_true;
  if (readIf_varMeta_fSpStar_nSp (&lCorrect, id_name))
    return pk_true;
  if (lCorrect)
    {
      ASFatal_expected ("generic argument identification");
      set_fatal ();
    }
  return pk_false;
}

/* 2004-12-01
 * reads a new allocated named string argument 
 * success: cursor starts [Identifier] [fileSpace*_nSp] 
 *          cursor is moved to end of [fileSpace*_nSp]
 *          returns in <res> a new argument containing a new m'allocated copy of this identifier
 *          <correct> remains unchanged
 *          current atom is set to empty
 * failure: returns NIL_Argument in <res> 
 *          reference to missing identifier declaration is read
 *              error message is printed
 *              <correct> is set to false
 *          cursor starts [Identifier] [fileSpace*] [EOF]
 *              error message is printed
 *              <correct> is set to false
 *          next character does not begin an identifier reference
 *              current character,  <correct> remain unchanged
 */
_VIRTUAL_ boolean
MetaContext::idVal_readIf_stringIdentifier_fSpStar_nSp (boolean * correct, ref2Argument * argument_res)
{
  // log_ASDEBUG("MetaContext::idVal_readIf_stringIdentifier_fSpStar_nSp");
  char *CString_namePrefix;
  ref2MetaArgument id_name;
  boolean lCorrect = pk_true;
  boolean isSpaced;
  if (this_lexicalContext->stepIf_userIdentifier_fSpStar_nSp (&lCorrect, &CString_namePrefix, &isSpaced))
    {
      if (isSpaced)
	{
	  /* FILE = PREFIX . [Identifier][fileSpace+] # [Character_but_Space][Any*][EOF]
	   *        [Identifier] is not a reserved word, <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier] */
	  if (getIf_string_id_ (CString_namePrefix, &lCorrect, argument_res))
	    return pk_true;
	  Context::ASFatal_missingDecl (lCorrect, "string meta identifier", CString_namePrefix);
	}
      else
	{			/* FILE = PREFIX . [Identifier] # [Character_but_Space_but_IdentCharacter][Any*][EOF]
				 *        [Identifier] is not a reserved word, <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier]  */
	  if (readIf_varMeta_fSpStar_nSp (CString_namePrefix, &lCorrect, &id_name))
	    {
	      if (getIf_string_id_ (id_name, &lCorrect, argument_res))
		return pk_true;
	      ASFatal_missingDecl (lCorrect, "string meta identifier", id_name);
	    }
	  else if (lCorrect)
	    {
	      if (getIf_string_id_ (CString_namePrefix, &lCorrect, argument_res))
		return pk_true;
	      Context::ASFatal_missingDecl (lCorrect, "string meta identifier", CString_namePrefix);
	    }
	}
    }
  else if (lCorrect)
    {
      /*  FILE = PREFIX .# [Any_but_IdentCharacter][Any*][EOF] */
      if (readIf_varMeta_fSpStar_nSp (&lCorrect, &id_name))
	{
	  if (getIf_string_id_ (id_name, &lCorrect, argument_res))
	    return pk_true;
	  ASFatal_missingDecl (lCorrect, "string identifier", id_name);
	}
      else if (lCorrect)
	return pk_false;
      (*correct) = pk_false;
      return pk_false;
    }
  /* FILE = PREFIX . [Identifier][fileSpace*] # [EOF]
   *                <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier]
   *          FILE = PREFIX . [Identifier][fileSpace+] # [Any*][EOF]
   *                 [Identifier] is a reserved word
   *          FILE = PREFIX . ... # [Any*][EOF], reading failed
   *          error message is printed */
  (*correct) = pk_false;
  return pk_false;
}

/* 2004-12-01
 * reads a new allocated named list argument 
 * success: cursor starts [Identifier] [fileSpace*_nSp] 
 *          cursor is moved to end of [fileSpace*_nSp]
 *          returns in <res> a new argument containing a new m'allocated copy of this identifier
 *          <correct> remains unchanged
 *          current atom is set to empty
 * failure: returns NIL_Argument in <res> 
 *          reference to missing identifier declaration is read
 *              error message is printed
 *              <correct> is set to false
 *          cursor starts [Identifier] [fileSpace*] [EOF]
 *              error message is printed
 *              <correct> is set to false
 *          next character does not begin an identifier reference
 *              current character,  <correct> remain unchanged
 */
_PROTECTED_ _VIRTUAL_ boolean MetaContext::idVal_readIf_listIdentifier_fSpStar_ (boolean * correct, ref2Argument * argument_res, boolean * isEof)
{
  // log_ASDEBUG("MetaContext::idVal_readIf_listIdentifier_fSpStar_");
  char *
    CString_namePrefix;
  ref2MetaArgument
    id_name;
  boolean
    lCorrect = pk_true;
  boolean
    isSpaced;
  if (this_lexicalContext->stepIf_userIdentifier_fSpStar_nSp (&lCorrect, &CString_namePrefix, &isSpaced))
    {
      if (isSpaced)
	{
	  /* FILE = PREFIX . [Identifier][fileSpace+] # [Character_but_Space][Any*][EOF]
	   *        [Identifier] is not a reserved word, <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier] */
	  if (getIf_list_id_ (CString_namePrefix, &lCorrect, argument_res))
	    return pk_true;
	  Context::ASFatal_missingDecl (lCorrect, "list meta identifier", CString_namePrefix);
	}
      else
	{			/* FILE = PREFIX . [Identifier] # [Character_but_Space_but_IdentCharacter][Any*][EOF]
				 *        [Identifier] is not a reserved word, <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier] */
	  if (readIf_varMeta_fSpStar_nSp (CString_namePrefix, &lCorrect, &id_name))
	    {
	      if (getIf_list_id_ (id_name, &lCorrect, argument_res))
		return pk_true;
	      ASFatal_missingDecl (lCorrect, "list meta identifier", id_name);
	    }
	  else if (lCorrect)
	    {
	      if (getIf_list_id_ (CString_namePrefix, &lCorrect, argument_res))
		return pk_true;
	      Context::ASFatal_missingDecl (lCorrect, "list meta identifier", CString_namePrefix);
	    }
	}
    }
  else if (lCorrect)
    {
      /* FILE = PREFIX .# [Any_but_IdentCharacter][Any*][EOF] */
      if (readIf_varMeta_fSpStar_nSp (&lCorrect, &id_name))
	{
	  if (getIf_list_id_ (id_name, &lCorrect, argument_res))
	    return pk_true;
	  ASFatal_missingDecl (lCorrect, "list identifier", id_name);
	}
      else if (lCorrect)
	return pk_false;
      (*correct) = pk_false;
      return pk_false;
    }
  /* FILE = PREFIX . [Identifier][fileSpace*] # [EOF]
   *                <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier]
   *          FILE = PREFIX . [Identifier][fileSpace+] # [Any*][EOF]
   *                 [Identifier] is a reserved word
   *          FILE = PREFIX . ... # [Any*][EOF], reading failed
   *          error message is printed */
  (*correct) = pk_false;
  return pk_false;
}

/* 2004-12-01
 * reads a new allocated named argument 
 * success: cursor starts [Identifier] [fileSpace*_nSp] 
 *          cursor is moved to end of [fileSpace*_nSp]
 *          returns in <res> a new argument containing a new m'allocated copy of this identifier
 *          <correct> remains unchanged
 *          current atom is set to empty
 * failure: returns NIL_Argument in <res> 
 *          reference to missing identifier declaration is read
 *              error message is printed
 *              <correct> is set to false
 *          cursor starts [Identifier] [fileSpace*] [EOF]
 *              error message is printed
 *              <correct> is set to false
 *          next character does not begin an identifier reference
 *              current character,  <correct> remain unchanged
 */
_VIRTUAL_ boolean
MetaContext::idVal_readIf_argumentIdentifier_fSpStar_nSp (boolean * correct, ref2Argument * argument_res)
{
  // log_ASDEBUG("MetaContext::idVal_readIf_argumentIdentifier_fSpStar_nSp");
  char *CString_namePrefix;
  ref2MetaArgument id_name;
  boolean lCorrect = pk_true;
  boolean isSpaced;
  if (this_lexicalContext->stepIf_userIdentifier_fSpStar_nSp (&lCorrect, &CString_namePrefix, &isSpaced))
    {
      if (isSpaced)
	{
	  /*FILE = PREFIX . [Identifier][fileSpace+] # [Character_but_Space][Any*][EOF],
	   *        [Identifier] is not a reserved word, <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier]   */
	  if (getIf_argument_id_ (CString_namePrefix, &lCorrect, argument_res))
	    return pk_true;
	  Context::ASFatal_missingDecl (lCorrect, "argument meta identifier", CString_namePrefix);
	}
      else
	{			/* FILE = PREFIX . [Identifier] # [Character_but_Space_but_IdentCharacter][Any*][EOF]
				 *        [Identifier] is not a reserved word, <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier] */
	  if (readIf_varMeta_fSpStar_nSp (CString_namePrefix, &lCorrect, &id_name))
	    {
	      if (getIf_argument_id_ (id_name, &lCorrect, argument_res))
		return pk_true;
	      ASFatal_missingDecl (lCorrect, "argument meta identifier", id_name);
	    }
	  else if (lCorrect)
	    {
	      if (getIf_argument_id_ (CString_namePrefix, &lCorrect, argument_res))
		return pk_true;
	      Context::ASFatal_missingDecl (lCorrect, "argument meta identifier", CString_namePrefix);
	    }
	}
    }
  else if (lCorrect)
    {
      /* FILE = PREFIX .# [Any_but_IdentCharacter][Any*][EOF] */
      if (readIf_varMeta_fSpStar_nSp (&lCorrect, &id_name))
	{
	  if (getIf_argument_id_ (id_name, &lCorrect, argument_res))
	    return pk_true;
	  ASFatal_missingDecl (lCorrect, "argument identifier", id_name);
	}
      else if (lCorrect)
	return pk_false;
      (*correct) = pk_false;
      return pk_false;
    }
  /* FILE = PREFIX . [Identifier][fileSpace*] # [EOF]
   *                <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier]
   *          FILE = PREFIX . [Identifier][fileSpace+] # [Any*][EOF]
   *                 [Identifier] is a reserved word
   *          FILE = PREFIX . ... # [Any*][EOF], reading failed
   *          error message is printed */
  (*correct) = pk_false;
  return pk_false;
}

/* PLG 2008-06-06 functionName added */
_PRIVATE_ boolean MetaContext::read_varGeneric_fSpStar_nSp (int commandCode, char *functionName, ref2CommandArgument * id_name)
{
  ref2StringFunctionCall
    functionCall;
  boolean
    isEof = pk_false;
  ref2Argument
    ArgumentBid;
  // log_ASDEBUG("entering MetaContext::read_varGeneric_fSpStar_nSp");
  if (read_string_call_argList_fSpStar_ (commandCode, functionName, &functionCall, &isEof, &ArgumentBid))
    {
      (*id_name) = new CommandArgument (pk_false, functionCall, pk_false);
      return pk_true;
    }
  return pk_false;
}


_PRIVATE_
  boolean MetaContext::new_metaArgumentElement (ref2ArgumentList comments, boolean isGlobal, char *identName, ref2Argument formula,
						boolean isList, ref2GenericElementList * genericDecl_res)
{
  // logDEBUG_in("new_metaArgumentElement, ident is", identName);
  ref2Identifier
    _identifier;
  if (append_uniqueNewIdentifier (identName, isList, formula, &_identifier))
    {
      (*genericDecl_res) = new MetaArgumentElement (this, comments, !isList, isGlobal, new StringArgument (identName), formula);
      return pk_true;
    }
  set_fatal ();
  return pk_false;
}

_PRIVATE_
  boolean MetaContext::new_metaArgumentElement (ref2ArgumentList comments, boolean isGlobal, ref2MetaArgument id_name, ref2Argument formula,
						boolean isList, ref2GenericElementList * genericDecl_res)
{
  // logDEBUG_in("new_metaArgumentElement MetaArgument id_name");
  ref2Identifier
    _identifier;
  if (append_uniqueNewMetaIdentifier (id_name, isList, formula, &_identifier))
    {
      (*genericDecl_res) = new MetaArgumentElement (this, comments, !isList, isGlobal, id_name, formula);
      return pk_true;
    }
  set_fatal ();
  return pk_false;
}

_PRIVATE_
  boolean MetaContext::new_metaArgumentElement (ref2ArgumentList comments, boolean isGlobal, ref2CommandArgument id_name, ref2Argument formula,
						boolean isList, ref2GenericElementList * genericDecl_res)
{
  // logDEBUG_in("new_metaArgumentElement MetaArgument CommandArgument id_name");
  ref2Identifier
    _identifier;
  if (append_uniqueNewCommandIdentifier (id_name, isList, formula, &_identifier))
    {
      (*genericDecl_res) = new MetaArgumentElement (this, comments, !isList, isGlobal, id_name, formula);
      return pk_true;
    }
  set_fatal ();
  return pk_false;
}

_PRIVATE_ boolean MetaContext::new_metaVariableNode_nSp (ref2ArgumentList comments, ref2Argument id_name, ref2GenericElementList * genericDecl_res)
{
  // log_ASDEBUG("entering DEBUG MetaContext::new_metaVariableNode_nSp");
  ref2MetaVariableNode
    metaVariable_res;
  metaVariable_res = new MetaVariableNode (this, comments, id_name);
  if (metaVariable_res->read_metaBody_fSpStar_nSp ())
    {
      (*genericDecl_res) = metaVariable_res;
      return pk_true;
    }
  set_fatal ();
  return pk_false;
}

_PRIVATE_ boolean MetaContext::new_metaFileNode_nSp (ref2ArgumentList comments, ref2Argument id_name, ref2GenericElementList * genericDecl_res)
{
  // log_ASDEBUG("entering DEBUG MetaContext::new_metaFileNode_nSp");
  ref2MetaFileNode
    metaFile_res;
  metaFile_res = new MetaFileNode (this, comments, id_name);
  if (metaFile_res->read_metaBody_fSpStar_nSp ())
    {
      (*genericDecl_res) = metaFile_res;
      return pk_true;
    }
  set_fatal ();
  return pk_false;
}


_PRIVATE_
  boolean MetaContext::read_genericFunctionCall_argList_fSpStar_nSp (ref2ArgumentList comments, char *functionName,
								     ref2GenericElementList * genericDecl_res)
{
  // logDEBUG_in("read_genericFunctionCall_argList_fSpStar_nSp");
  // log_ASDEBUG("entering MetaContext::read_genericFunctionCall_argList_fSpStar_nSp");
  // fprintf(stderr, "          functionName is %s\n", functionName);
  ref2ElementFunctionCall
    elementFunctionCall;
  boolean
    isEof = pk_false;
  if (read_elementFunctionCall_argList_fSpStar_ (comments, functionName, &elementFunctionCall, &isEof))
    {
      printORlevelInfo_node_functionCallReading (functionName, (ref2FunctionCall) elementFunctionCall, comments);
      (*genericDecl_res) = new CommandElement (this, elementFunctionCall, comments);
      return pk_true;
    }
  return pk_false;
}

_PRIVATE_ boolean MetaContext::read_genericList_inNewIteratorContext_fSpStar_nSp (ref2ArgumentList comments, ref2GenericElementList * genericDecl_res)
{
  // logDEBUG_in("read_genericList_inNewIteratorContext_fSpStar_nSp");
  ref2ElementFunctionCall
    elementFunctionCall;
  if (read_elementList_inNewIteratorContext_fSpStar_nSp (comments, &elementFunctionCall))
    {
      (*genericDecl_res) = new CommandElement (this, elementFunctionCall, comments);
      return pk_true;
    }
  return pk_false;
}

_PRIVATE_ boolean MetaContext::read_argument_fSpStar_nSp (boolean isList, ref2Argument * formula)
{
  if (isList)
    return read_list_fSpStar_nSp (formula);
  return read_string_fSpStar_nSp (formula);
}

_PRIVATE_
  boolean MetaContext::read_varMeta_argumentEquation_fSpStar_nSp (ref2ArgumentList comments, boolean isGlobal, boolean isList,
								  ref2GenericElementList * genericDecl_res)
{
  // logDEBUG_in("readIf_varMeta_argumentEquation_fSpStar_nSp");
  // log_ASDEBUG("MetaContext::read_varMeta_argumentEquation_fSpStar_nSp");
  char *
    CString_namePrefix;
  ref2MetaArgument
    id_name;
  boolean
    lCorrect = pk_true;
  boolean
    isSpaced;
  ref2Argument
    formula;
  if (this_lexicalContext->stepIf_userIdentifier_fSpStar_nSp (&lCorrect, &CString_namePrefix, &isSpaced))
    {
      if (isSpaced)
	{
	  /* FILE = PREFIX . [Identifier][fileSpace+] # [Character_but_Space][Any*][EOF]
	   *        [Identifier] is not a reserved word, CString_namePrefix is loaded with a new m'allocated copy of [Identifier] */
	  if (read_argument_fSpStar_nSp (isList, &formula))
	    return new_metaArgumentElement (comments, isGlobal, CString_namePrefix, formula, isList, genericDecl_res);
	  return pk_false;
	}
      /* FILE = PREFIX . [Identifier] # [Character_but_Space_but_IdentCharacter][Any*][EOF],
       *        [Identifier] is not a reserved word, <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier]*/
      if (readIf_varMeta_fSpStar_nSp (CString_namePrefix, &lCorrect, &id_name))
	{
	  if (read_argument_fSpStar_nSp (isList, &formula))
	    return new_metaArgumentElement (comments, isGlobal, id_name, formula, isList, genericDecl_res);
	  return pk_false;
	}
      if (lCorrect)
	if (read_argument_fSpStar_nSp (isList, &formula))
	  return new_metaArgumentElement (comments, isGlobal, CString_namePrefix, formula, isList, genericDecl_res);
      return pk_false;
    }
  if (lCorrect)
    {
      /* FILE = PREFIX .# [Any_but_IdentCharacter][Any*][EOF] */
      if (read_varMeta_fSpStar_nSp (&id_name))
	if (read_argument_fSpStar_nSp (isList, &formula))
	  return new_metaArgumentElement (comments, isGlobal, id_name, formula, isList, genericDecl_res);
      return pk_false;
    }
  /* FILE = PREFIX . [Identifier][fileSpace*] # [EOF]
   *                <CString_namePrefix> is loaded with a new m'allocated copy of [Identifier]
   *          FILE = PREFIX . [Identifier][fileSpace+] # [Any*][EOF]
   *                 [Identifier] is a reserved word
   *          FILE = PREFIX . ... # [Any*][EOF], reading failed
   *          error message is printed */
  return pk_false;
}

/* PLG 2008-06-06 functionName added */
_PRIVATE_
  boolean MetaContext::read_metaArgumentEquation_fSpStar_nSp (ref2ArgumentList comments, boolean isList, ref2GenericElementList * genericDecl_res)
{
  // logDEBUG_in("read_metaArgumentEquation_fSpStar_nSp");
  boolean
    isGlobal;
  if (this_lexicalContext->stepOpt_scope_fSpStar_nSp (&isGlobal))
    {
      /* FILE = PREFIX . <commandChar>'static'[fileSpace*] # [Character_but_Space][Any*][EOF]
       *             first character in [fileSpace*][Character_but_Space] is not an [IdentCharacter]
       *             <isGlobal> is set TRUE
       *          FILE = PREFIX .# opt{[Any_but_<commandChar>][Any*]}[EOF]
       *             <isGlobal> is set FALSE */
      int
	commandCode;
      char *
	functionName;
      ref2Argument
	formula;
      if (this_lexicalContext->stepIf_varIdentCommandHead_fSpStar_nSp (&commandCode, &functionName))
	{
	  /* FILE = PREFIX . XXX '$:'[Identifier] [fileSpace*] # [Character_but_Space][Any*][EOF]
	   *             [Identifier] is in { 'string',  'stringval', 'hostval'}
	   *                <identifierCode> is set according to Identifier
	   *             [Identifier] is not a reserved word
	   *                <identifierCode> is set userCODE */
	  ref2CommandArgument
	    id_name;
	  if (read_varGeneric_fSpStar_nSp (commandCode, functionName, &id_name))
	    if (read_argument_fSpStar_nSp (isList, &formula))
	      return new_metaArgumentElement (comments, isGlobal, id_name, formula, isList, genericDecl_res);
	  return pk_false;
	}
      if (commandCode == notACommandCODE)
	/* FILE = PREFIX . XXX [Character_but_Space_but_'$:'][Any*][EOF]
	 *             <identifierCode> is set notACommandCODE */
	return read_varMeta_argumentEquation_fSpStar_nSp (comments, isGlobal, isList, genericDecl_res);
      /* FILE = PREFIX . XXX ... # [EOF], reading failed
       * FILE = PREFIX . XXX'$:'[Identifier] [fileSpace*] # [EOF]
       * FILE = PREFIX . XXX'$:' # opt{[Any_but_IdentHeaderCharacter][Any*]}[EOF]
       * FILE = PREFIX . XXX'$:'[Identifier] [fileSpace*] # [Character_but_Space][Any*][EOF]
       *                 [Identifier] is a reserved word not in { 'string',  'stringval', 'hostval'}
       * an error message is printed
       */
      return pk_false;
    }
  /* FILE = PREFIX .# <commandChar>[Identifier][fileSpace*] # opt{[Character_but_Space][Any*]}[EOF]
   *             [Identifier] is not  'static'
   * or FILE = PREFIX . <commandChar>'static'[fileSpace*] # [EOF]
   * or FILE = PREFIX . <commandChar> # [Any_but_IdentCharacter][Any*][EOF]
   * or FILE = PREFIX . ... [EOF], reading failed
   * or file is closed
   * error message is printed
   */
  return pk_false;
}

_PRIVATE_ boolean MetaContext::read_nodeIdentification_fSpStar_nSp (ref2Argument * id_name, boolean * isVar, boolean * isFile)
{
  // logDEBUG_in("read_nodeIdentification_fSpStar_nSp");
  // log_ASDEBUG("DEBUG MetaContext::read_metaNodeEquation_fSpStar_nSp");
  boolean
    isEol = pk_false;;
  if (read_nodeIdentification_spStar_ (id_name, isVar, isFile, &isEol))
    {
      if (isEol)
	return this_lexicalContext->req_nl_fSpStar_nSp ();	/* success: FILE = PREFIX . [eol][fileSpace*] #[Character_but_Space][Any*][EOF]
								 * failure: FILE = PREFIX.opt{[eol][fileSpace*]}#[EOF] */
      return pk_true;
    }
  return pk_false;
}

_PRIVATE_ boolean MetaContext::read_metaNodeEquation_fSpStar_nSp (ref2ArgumentList comments, ref2GenericElementList * genericDecl_res)
{
  // log_ASDEBUG("entering DEBUG MetaContext::read_metaNodeEquation_fSpStar_nSp");
  ref2Argument
    id_name;
  boolean
    isVar,
    isFile;
  if (read_nodeIdentification_fSpStar_nSp (&id_name, &isVar, &isFile))
    {
      if (isVar)
	return new_metaVariableNode_nSp (comments, id_name, genericDecl_res);
      if (isFile)
	return new_metaFileNode_nSp (comments, id_name, genericDecl_res);
      (get_lexicalContext ()->logERR_ln_LineColumn_ln__ln) ("an extern reference cannot be generic");
      set_error ();
    }
  return pk_false;
}

_PRIVATE_
  boolean MetaContext::read_genericElement_C_fSpStar_ (ref2ArgumentList comments, boolean inNewLexical, ref2GenericElementList * genericDecl_res,
						       boolean * isEof)
{
  // log_ASDEBUG("entering MetaContext::read_genericElement_C_fSpStar_");
  int
    commandCode;
  char *
    functionName;
  boolean
    lCorrect = pk_true;
  if (this_lexicalContext->stepOpt_elementCategory_fSpStar_nSp (&commandCode, &functionName))
    {
      switch (commandCode)
	{
	  /* commandCode belongs to userCODE, stringCODE, listCODE, iterateCODE, notACommandCODE */
	case userCODE:
	  lCorrect = read_genericFunctionCall_argList_fSpStar_nSp (comments, functionName, genericDecl_res);
	  break;
	case iterateCODE:
	  lCorrect = read_genericList_inNewIteratorContext_fSpStar_nSp (comments, genericDecl_res);
	  break;
	case stringCODE:	/* 3 [commandChar] 'string' [fileSpace*] has been read */
	  lCorrect = read_metaArgumentEquation_fSpStar_nSp (comments, pk_false, genericDecl_res);
	  break;
	case listCODE:		/* 3 [commandChar] 'list' [fileSpace*] has been read */
	  lCorrect = read_metaArgumentEquation_fSpStar_nSp (comments, pk_true, genericDecl_res);
	  break;
	case notACommandCODE:
	  lCorrect = read_metaNodeEquation_fSpStar_nSp (comments, genericDecl_res);	/* READ COMMAND 1 */
	  break;
	default:
	  ASFatal_expected ("generic element header");
	  lCorrect = pk_false;
	}
      if (inNewLexical)
	this_lexicalContext = this_lexicalContext->pop ();
      if (lCorrect)
	return this_lexicalContext->req_declClosing_fSpStar_ (isEof);
      /* success: FILE = PREFIX . '}'[fileSpace*] # opt{[Character_but_Space][Any*]}[EOF], '}' is not followed by ':'
       *             if [EOF] is reached, <isEof> is set TRUE and this is set COMPLETED */
      /* failure: FILE = PREFIX . [Any_but_'}'_but_'}:']+  # '}:'[Any*][EOF]
       *             if [EOF] is reached, <isEof> is set TRUE and this is set ERRORCOMPLETED
       *             if [EOF] is not reached, <isEof> remains unchanged, and this is set ERRORREADY
       *          FILE = PREFIX . [Any_but_'}'_but_'}:']+ '}'[fileSpace*] # opt{[Character_but_Space][Any*]}[EOF]
       *             if [EOF] is reached, <isEof> is set TRUE and this is set ERRORCOMPLETED
       *             if [EOF] is not reached, <isEof> remains unchanged, and this is set ERRORREADY
       *          FILE = PREFIX . [Any_but_'}'_but_'}:']*  # [EOF]
       *             <isEof> is set TRUE and this is set ERRORCOMPLETED */
      if (this_lexicalContext->recover_main_fSpStar_ (isEof));
      return pk_false;
    }				/* READ COMMAND 2 */
  if (inNewLexical)
    this_lexicalContext = this_lexicalContext->pop ();
  if (this_lexicalContext->recover_main_fSpStar_ (isEof));
  return pk_false;
}

/* 2005-04-04
 * reads a new node function body
 * success: FILE = PREFIX.[inner_NODE_GENERATOR_fSp+] # }: SUFFIX
 *          where
 *             [inner_NODE_GENERATOR] ::= '{' [fileSpace*] [_NODE_GENERATOR_fSp] '}' 
 * failure: a syntax or context error has occurred in [Character*]
 *             cursor starts [Character*] '}:', '{' '}' are well balanced in [Character*]
 *               cursor is moved to '}:'
 *             cursor starts [Character*] [EOF], '{' '}' are well balanced in [Character*]
 *               cursor is moved to [EOF]
 *             error message is printed
 */
_PUBLIC_ boolean MetaContext::starRead_genericElement_fSpStar_nSp ()
{
  // log_ASDEBUG("entering MetaContext::starRead_genericElement_fSpStar_nSp");
  ref2ArgumentList
    comments = NIL_ArgumentList;
  boolean
    isEof = pk_false;
  boolean
    lCorrect = pk_true, fullyCorrect = pk_true;
  boolean
    inNewLexical = pk_false;
  boolean
    isEmpty = pk_true;
  ref2GenericElementList
    nextGeneric = NIL_GenericElementList;
  ref2GenericElementList
    currentGeneric = NIL_GenericElementList;
  // PLG TODO CREATE AN INNER DYNAMIC CONTEXT AND APPEND A NEW CONTEXTTREE  LEVEL ?<
  while (!isEof)
    {				/* open, not completed */
      if (readIf_declOpening_fSpStar_nSp (&lCorrect, &comments))
	{			/* 2009-09-07 OK */
	  /* FILE = PREFIX_i .'{'[fileSpace*] # [Character_but_Space][Any*][EOF] */
	  if (readOpt_lexical_fSpStar_ (&inNewLexical, &isEof))
	    {
	      /* FILE = PREFIX_i .'{'[fileSpace*] opt{[SyntaxAdaptor_fSpPlus]} # opt{[Character_but_Space][Any*]} [EOF] */
	      if (isEof)
		{
		  this_lexicalContext->ASFatal_eof ();
		  return pk_false;
		}
	      if (read_genericElement_C_fSpStar_ (comments, inNewLexical, &nextGeneric, &isEof))
		{
		  if (isEof)
		    {
		      this_lexicalContext->ASFatal_eof ();
		      return pk_false;
		    }
		  if (isEmpty)
		    {
		      isEmpty = pk_false;
		      these_genericElements = nextGeneric;
		    }
		  else
		    currentGeneric->append (nextGeneric);
		  currentGeneric = nextGeneric;
		}
	      else
		{		/* should be [EOF] or recovered */
		  if (isEof)
		    return pk_false;
		  fullyCorrect = pk_false;
		}
	      inNewLexical = pk_false;
	    }
	  else
	    {
	      if (isEof)
		return pk_false;
	      if (!(this_lexicalContext->recover_main_fSpStar_ (&isEof)))
		return pk_false;
	      fullyCorrect = pk_false;
	    }
	}
      else
	{
	  if (lCorrect)
	    /* FILE = PREFIX_i .# [Character_but_'{'][Any*] [EOF] */
	    return fullyCorrect;
	  /* FILE = PREFIX . [ELEMENT*] '{' [fileSpace*] # [EOF]
	   * FILE = PREFIX . [ELEMENT*] '{' ... # [EOF], reading failed
	   * error message has been printed
	   */
	  return pk_false;
	}
    }
  return pk_false;
}

/* ==========  ==========  ==========  ========== TITLE: GENERATOR
 */

_PUBLIC_ _VIRTUAL_ boolean MetaContext::build_instance_in (ref2Context callingContext, ref2ArgumentList parameterValues)
{
  // logDEBUG_in("build_instance_in in MetaContext");
  if (((ref2DynamicContext) this_nestingContext)->build_instance_in (callingContext, parameterValues))
    {
      ref2DynamicContext
	innerContext = new DynamicContext (callingContext);
      return these_genericElements->build_elementDeclaration (innerContext);
    }
  return pk_false;
}

/* ==========  ==========  ==========  ========== TITLE: BUILDER
 */

/* creates the dynamic context for a function declaration embedded in <embedding_context> with same substitution marks
 * <embedding_context> is not NIL (not checked)
 *          this scope is set to DYNAMIC
 */
_PUBLIC_ MetaContext::MetaContext (ref2DynamicContext owner_context):DynamicContext ((ref2Context) owner_context)
{
  // log_ASDEBUG("MetaContext::MetaContext 2");
  these_genericElements = NIL_GenericElementList;
}
