/* =============== Diagnostics
*/
#define LOG_SPACES_STR_LN(_STR_) LOG_SPN_STR_LN(4, _STR_)
#define LOG_SEPARATORLINE() LOGFILE()->put_separatorLine(GET_VERBOSELEVEL())

#define LOG_TIME_LN(_CPUTIME_, _LAPS_) LOGFILE()->put_timeCounts_ln(_CPUTIME_, _LAPS_);
#define LOG_CLOSE() LOGFILE()->close()

const char *argErrMess[6] = { "root node cannot be defined twice",
  "root name is expected after -node option",
  "unknown option",
  "new_pKMaked file cannot be defined twice",
  "not a correct file root name" "not a correct variable definition"
};

const char *stepMess[4] = { "main", "File reading", "Dependence Checking", "File generation" };

/* ==========  ==========  ==========  ========== TITLE: these attributes
 */

_PUBLIC_ void
new_pKMakeReporting::infoCorrectArgs ()
{
  LOG_LN_INFO_STR_LN ("usage new_pKMake -h ");
  LOG_STR_LN ("   print this help, other arguments are ignored");
  LOG_STR_LN ("usage new_pKMake [file] [<OPTION> list]");
  LOG_STR_LN ("   <file> is present: new_pKMake data are read in <file>");
  LOG_STR_LN ("   <file> is omitted: new_pKMake data are read from standard input");
  LOG_STR_LN ("   MKF variable definition is given by");
  LOG_STR_LN ("         -=varName STRING to define static string variable <varName> and give it the value STRING");
  LOG_STR_LN ("   root generation options are");
  LOG_STR_LN ("         -node varName to use <varName> as the constructive root node");
  LOG_STR_LN ("         -all for collecting all constructive nodes in <file> (implicit action)");
  LOG_STR_LN ("         -ALL for recursively collecting all constructive nodes from root make file");
  LOG_STR_LN ("   mode generation options are");
  LOG_STR_LN ("         [-r] to force building up to inner files, even if files are up to date");
  LOG_STR_LN ("         [-R] to recursively force building even if files are up to date");
  LOG_STR_LN ("   log options are");
  LOG_STR_LN ("         [-v,-V,-d,-D] to collect more and more information about make execution");
  LOG_STR_LN ("   other options are");
  LOG_STR_LN ("         [-ignore] to continue to proceed despite errors");
  LOG_STR_LN ("         [-stat, -STAT] to display statistics information");
}

_PUBLIC_ _VIRTUAL_ const char *
new_pKMakeReporting::myClassName ()
{
  return "Soundness";
}

_PUBLIC_ void
new_pKMakeReporting::openStep (int step)
{
  NodeElement::incrStepIndex ();
  stepCpuTimeStarting = getcputime ();
  if (ISLVERBOSE ())
    {
      if (ISDEBUG ())
	LOG_STEPTITLE_STR_LN (stepMess[step]);
      else
	LOG_LN_INFO_STR_LN (stepMess[step]);
    }
}

_PUBLIC_ void
new_pKMakeReporting::closeStep (int step)
{
  int totalTime, cpuTime;
  stepCpuTimeEnding = getcputime ();	/* in  fileSystemTime_e.h */
  totalTime = getTotalTime ();	/* in  fileSystemTime_e.h */
  if (stepCpuTimeStarting <= stepCpuTimeEnding)
    {
      cpuTime = (int) (stepCpuTimeEnding - stepCpuTimeStarting);
      totalCpuTime = totalCpuTime + cpuTime;
    }
  else
    cpuTime = -1;
  if (ISLVERBOSE ())
    {
      LOG_LN ();
      if (ISDEBUG ())
	LOG_BACKTAB ();
    }
  if (ISVERBOSE ())
    {
      LOG_LN_INFO_STR_SP (stepMess[step]);
      LOG_STR_SP ("terminated after");
      LOG_TIME_LN (cpuTime, totalTime - totalFullTime);
    }
  totalFullTime = totalTime;
}

_PUBLIC_ void
new_pKMakeReporting::log_fileArgumentErrorClosing (const char *rootPath)
{
  LOG_LN_ERRO_STR_SP_STR_LN (rootPath, argErrMess[4]);
  LOG_CLOSE ();
}

_PUBLIC_ void
new_pKMakeReporting::log_errorClosing ()
{
  LOG_LN_INFO_STR_LN ("use -h as single argument to get correct argument list");
  LOG_CLOSE ();
}

_PUBLIC_ void
new_pKMakeReporting::log_argumentErrorClosing (const char *arg, int argErrNum)
{
  LOG_LN_ERRO_STR_SP_STR_LN (arg, argErrMess[argErrNum]);
  log_errorClosing ();
}

_PUBLIC_ void
new_pKMakeReporting::log_mainErrorContinue ()
{
  if (ISDEBUG ())
    {
      LOG_LN_ERRO_STR_LN ("continue despite errors; some built files may be corrupted");
    }
}

_PUBLIC_ void
new_pKMakeReporting::log_mainErrorClosing ()
{
  if (ISVERBOSE ())
    {
      LOG_SEPARATORLINE ();
      LOG_ERRO_STR_SP ("make file failed after");
      LOG_TIME_LN (totalCpuTime, getTotalTime ());
    }
  else
    LOG_LN_ERRO_STR_LN ("make file failed");
  LOG_CLOSE ();
}

_PUBLIC_ void
new_pKMakeReporting::log_mainErrorClosing (int step)
{
  if (ISVERBOSE ())
    {
      LOG_SEPARATORLINE ();
      LOG_ERRO_STR_SP ("make file failed after");
      LOG_TIME_LN (totalCpuTime, getTotalTime ());
      LOG_SPN_STR_SP (4, "during");
      LOG_STR_LN (stepMess[step]);
    }
  else
    LOG_LN_ERRO_STR_SP_STR_LN ("make file failed during", stepMess[step]);
  LOG_CLOSE ();
}

_PUBLIC_ void
new_pKMakeReporting::log_mainSuccessClosing (boolean somethingDone)
{
  if (ISVERBOSE ())
    {
      LOG_SEPARATORLINE ();
      LOG_INFO_STR_SP ("make file done after");
      LOG_TIME_LN (totalCpuTime, getTotalTime ());
      if (somethingDone)
	{
	  LOG_SPACES_STR_LN ("selected components have been updated");
	}
      else
	{
	  LOG_SPACES_STR_LN ("selected components are up to date");
	}
    }
  LOG_CLOSE ();
}

_PUBLIC_ void
new_pKMakeReporting::log_resetRecursivityArg ()
{
  LOG_STR_LN ("building mode is reset, highest level is kept");
}

_PUBLIC_ void
new_pKMakeReporting::log_resetHashTuningArg ()
{
  LOG_STR_LN ("hash tuning is reset, highest level is kept");
}

_PUBLIC_ void
new_pKMakeReporting::log_resetStatArg ()
{
  LOG_STR_LN ("statistics option is reset, highest level is kept");
}

_PUBLIC_ void
new_pKMakeReporting::log_ignoreArg ()
{
  LOG_STR_LN ("-ignore is reset");
}


_PUBLIC_ void
new_pKMakeReporting::reset_verbose (int level)
{
  if (ISVERBOSE ())
    LOG_STR_LN ("reporting mode is reset, highest level is kept");
  if ((GET_VERBOSELEVEL ()) < level)
    {
      SET_VERBOSELEVEL (level);
      return;
    }
}

/* PLG 2007-02-23
 */
_PUBLIC_ new_pKMakeReporting::new_pKMakeReporting ():Soundness ()
{
  totalFullTime = 0;
  totalCpuTime = 0;
  SET_SILENT ();
}
