/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

// Added L Besnard (July 2006)
#define TRACE(s)		/* s */

/* ===================================================================== 
 * Memory level actions
 *
 */
/* creates and returns a new action line list element valued with actionLine
 */
static ref2ActionLstType newActionLst(char *actionLine)
{
  ref2ActionLstType newRefAction = (ref2ActionLstType) malloc((unsigned)(sizeof(*newRefAction)));
  newRefAction->car = actionLine;
  newRefAction->cdr = NIL_ActionLstType;
  return (newRefAction);
}

/* appends and returns a new action line list element after the last element in the list prevAction
 */
static ref2ActionLstType appendActionLst(ref2ActionLstType prevAction, char *actionLine)
{
  ref2ActionLstType newRefAction = newActionLst(actionLine);
  if (prevAction == NIL_ActionLstType)
    return newRefAction;
  while (prevAction->cdr != NIL_ActionLstType)
    prevAction = prevAction->cdr;
  prevAction->cdr = newRefAction;
  return (newRefAction);
}

/*  creates and returns a new (predecessor object, associated actions) list element
 */
static ref2PredecessorLstType newPredecessorLst(ref2ObjectType predObject, boolean neverBuild)
{
  ref2PredecessorLstType newRefPredecessor = (ref2PredecessorLstType) malloc((unsigned)(sizeof(*newRefPredecessor)));
  newRefPredecessor->car = predObject;
  newRefPredecessor->neverBuild = neverBuild;
  newRefPredecessor->postConsAction = NIL_ActionLstType;
  newRefPredecessor->cdr = NIL_PredecessorLstType;
  return (newRefPredecessor);
}

/*  creates and returns a new action component to be referred to in object
 */
static ref2ActionType newAction(boolean handmade, boolean direct, boolean conditional)
{
  ref2ActionType newRefAction = (ref2ActionType) malloc((unsigned)(sizeof(*newRefAction)));
  newRefAction->handmade = handmade;
  newRefAction->direct = direct;
  newRefAction->conditional = conditional;
  newRefAction->prelConsAction = NIL_ActionLstType;
  newRefAction->innerLst = NIL_PredecessorLstType;
  newRefAction->noconsAction = NIL_ActionLstType;
  return newRefAction;
}

/* ===================================================================== 
 * Memory level object management
 *
 */
/* initialize the object hash table
 */
static void initObjectStorage()
{
  int i;
  for (i = 0; i < hashsize; i++)
    hashEntries[i] = NIL_LstRefObjType;
}

/* creates and returns a new (reference to object) list element valued with curObject
 */
static ref2LstRefObjType newLstRefObj(ref2ObjectType curObject)
{
  ref2LstRefObjType newRefLstRefObj = (ref2LstRefObjType) malloc((unsigned)(sizeof(*newRefLstRefObj)));
  newRefLstRefObj->car = curObject;
  newRefLstRefObj->cdr = NIL_LstRefObjType;
  return (newRefLstRefObj);
}

/* appends and returns a new (reference to object) list element after the last element in the list prevRefObject
 */
static ref2LstRefObjType appendLstRefObj(ref2LstRefObjType prevRefObject, ref2ObjectType curObject)
{
  ref2LstRefObjType newRefLstRefObj = newLstRefObj(curObject);
  if (prevRefObject == NIL_LstRefObjType)
    return newRefLstRefObj;
  while (prevRefObject->cdr != NIL_LstRefObjType)
    prevRefObject = prevRefObject->cdr;
  prevRefObject->cdr = newRefLstRefObj;
  return (newRefLstRefObj);
}

/* appends and returns a new (reference to object) list element after the last element in the list prevRefObject
 */
static ref2LstRefObjType insertLstRefObj(ref2LstRefObjType succRefObject, ref2ObjectType curObject)
{
  ref2LstRefObjType newRefLstRefObj = newLstRefObj(curObject);
  if (succRefObject == NIL_LstRefObjType)
    return newRefLstRefObj;
  newRefLstRefObj->cdr = succRefObject;
  return (newRefLstRefObj);
}

/* creates and returns a new object descriptor
 *        isvar pk_true, isFile false if the label denotes a variable
 *        isvar pk_false, isFile pk_true if the label denotes an application file designation
 *        isvar pk_false, isFile false if the label denotes a system file
 *        definition is pk_true if the object is created during in a definition term
 */
static ref2ObjectType newObject(char *name, boolean isvar, boolean isFile, boolean definition)
{
  int i;
  ref2ObjectType newRefObject = (ref2ObjectType) malloc((unsigned)(sizeof(*newRefObject)));

  newRefObject->inError = pk_false;
  newRefObject->internal = isvar;
  newRefObject->isFile = isFile;
  newRefObject->defined = definition;
  newRefObject->owner = NIL_ObjectType;
  newRefObject->include = pk_false;
  newRefObject->ordinal = objectNumber;
  objectNumber = objectNumber + 1;

  newRefObject->objectLabel = name;
  newRefObject->objectAlias = NIL_String;
  newRefObject->created = NIL_time;

  newRefObject->visited = pk_false;	/* pk_true when read for a mkf file */
  newRefObject->made = NIL_ObjectType;	/* file for which it has been made */
  newRefObject->stated = pk_false;

  newRefObject->actions = NIL_ActionType;
  newRefObject->firstPred = NIL_LstRefObjType;
  newRefObject->lastPred = NIL_LstRefObjType;

  return newRefObject;
}

/* ===================================================================== 
 * Low level object management
 *
 */

/* set owner as the owner of object and append object to owner child list
 */
static void setObjectOwner(ref2ObjectType object, ref2ObjectType owner)
{
  object->owner = owner;
  curFileObject->lastPred = appendLstRefObj(curFileObject->lastPred, object);
  if ((curFileObject->firstPred) == NIL_LstRefObjType)
    {
      curFileObject->firstPred = curFileObject->lastPred;
    }
}

/*
 * if forced is pk_true states its birthday to apocalypse
 */
static void registerDirectFile(ref2ObjectType newObject, boolean forced)
{
  ref2LstRefObjType newatom = appendLstRefObj(lastDirectFile, newObject);
  lastDirectFile = newatom;
  if (firstDirectFile == NIL_LstRefObjType)
    firstDirectFile = lastDirectFile;
  if (forced)
    newObject->created = apocalypse;
}

/*
 *
 */
static void registerObject(ref2ObjectType newObject, int hashi)
{
  ref2LstRefObjType newatom = insertLstRefObj(hashEntries[hashi - 1], newObject);
  hashEntries[hashi - 1] = newatom;
}

/* ===================================================================== 
 * Memory level MKF in/out file management
 *
 */

static ref2LstRefObjType initMKFReading(ref2ObjectType stdinObj)
{
  firstRefFileObject = newLstRefObj(stdinObj);
  lastRefFileObject = firstRefFileObject;
  return (firstRefFileObject);
}

static void appendMKFReading(ref2ObjectType newFile)
{
  lastRefFileObject = appendLstRefObj(lastRefFileObject, newFile);
}

static void appendFileConstruction(ref2ObjectType newFile)
{
  TRACE(fprintf(stderr, "lastFileConstruction newFile=%s\n", newFile->objectLabel);) lastFileConstruction = appendLstRefObj(lastFileConstruction, newFile);
  if (firstFileConstruction == NIL_LstRefObjType)
    firstFileConstruction = lastFileConstruction;
TRACE(fprintf(stderr, "lastFileConstruction firstFile=%s\n", firstFileConstruction->car->objectLabel);)}

static void initMakeGraph()
{
  lastFileConstruction = firstFileConstruction = NIL_LstRefObjType;
  lastDirectFile = firstDirectFile = NIL_LstRefObjType;
}

/* ===================================================================== 
 * Low level object management
 */
/* search the hash table for an object which label is the last read object
 *        isvar pk_true, isFile false if the label denotes a variable
 *        isvar pk_false, isFile pk_true if the label denotes a file designation
 *        isvar pk_false, isFile false if the label denotes a blind  designation
 * set hashi to hash code value for this label
 * success: return pk_true
 *          set refrefObject pointing to the object reference
 *          set in error the object in case of type conflict
 * failure: return false
 *          refrefObject is NIL_ObjectType
 */
static boolean findObject(boolean isvar, boolean isFile, int *hashi, ref2ObjectType * refrefObject)
{
  ref2LstRefObjType lenew;
  *hashi = getAtomHash();
  lenew = hashEntries[*hashi - 1];
  while (lenew != NIL_LstRefObjType)
    {
      *refrefObject = lenew->car;
      if (isAtomEq((*refrefObject)->objectLabel))
	{
	  if ((*refrefObject)->internal != isvar)
	    {
	      if (conflictError(conflictErrMess, *refrefObject))
		(*refrefObject)->inError = pk_true;
	    }
	  else if (!isvar)
	    if ((*refrefObject)->isFile != isFile)
	      {
		if (conflictError(conflictErrMess, *refrefObject))
		  (*refrefObject)->inError = pk_true;
	      }
	  return pk_true;
	}
      else
	lenew = lenew->cdr;
    }
  *refrefObject = NIL_ObjectType;
  return pk_false;
}

/* ===================================================================== 
 * Object management
 *
 */
static ref2ObjectType createStdinObject()
{
  ref2ObjectType curFObj = newObject((char *)"stdin", pk_false, pk_true, pk_false);
  curFObj->defined = pk_true;
  curFObj->include = pk_true;

  curFObj->created = NIL_time;

  curFObj->stated = pk_true;
  return curFObj;
}

/* ===================================================================== 
 * Printing
 *
 */
/*
 * Object printers
 */
static void printDeclHeadLine(ref2ObjectType object)
{
  if (object->include)
    logPrintIncludeDecl(object);
  else
    {
      logPrintOpenObjectMark(object);
      if (object->actions != NIL_ActionType)
	{
	  if (object->actions->handmade)
	    logPrintHandMadeMark();
	  else if ((object->actions->direct))
	    logPrintDirectMark();
	  if (object->actions->conditional)
	    logPrintDonotMark();
	}
      logPrintLine();
    }
}

static void printActionList(ref2ActionLstType actList)
{
  while (actList != NIL_ActionLstType)
    {
      logPrintString(actList->car);
      actList = actList->cdr;
    }
}

static void printPrelude(ref2ActionType allActions)
{
  if ((allActions->prelConsAction != NIL_ActionLstType))
    {
      logPrintPreludeOpening();
      printActionList(allActions->prelConsAction);
    }
}

static void printInnerList(ref2ActionType allActions)
{
  ref2PredecessorLstType predActList = allActions->innerLst;
  while (predActList != NIL_PredecessorLstType)
    {
      logPrintLine();
      logPrintPredecessorLine(predActList->car);
      printActionList(predActList->postConsAction);
      predActList = predActList->cdr;
    }
}

static void printDefault(ref2ActionType allActions)
{
  if ((allActions->noconsAction != NIL_ActionLstType))
    {
      logPrintDefaultOpening();
      printActionList(allActions->noconsAction);
    }
}

static void printObject(ref2ObjectType object)
{
  logPrintObjectHeading(object);
  printDeclHeadLine(object);
  if (object->actions != NIL_ActionType)
    {
      printPrelude(object->actions);
      printInnerList(object->actions);
      printDefault(object->actions);
    }
  logPrintObjectClosure();
}

static void printDirectFileList()
{
  ref2LstRefObjType curObject = firstDirectFile;

  logSkipLine();
  logPrintInfoTitle(verbose, directFileTitle);
  while (curObject != NIL_LstRefObjType)
    {
      logPrintObjectRefDateLine(curObject->car);
      curObject = curObject->cdr;
    }
}

static void printConstructionFileList()
{
  ref2LstRefObjType curObject = firstFileConstruction;

  logSkipLine();
  logPrintInfoTitle(verbose, constructionFileTitle);
  while (curObject != NIL_LstRefObjType)
    {
      logPrintObjectRefDateLine(curObject->car);
      curObject = curObject->cdr;
    }
}

static void printRegisteredObjects()
{
  int i;
  ref2LstRefObjType lat;
  for (i = 0; i < hashsize; i++)
    {
      lat = hashEntries[i];
      while (lat != NIL_LstRefObjType)
	{
	  logPrintLine();
	  printObject(lat->car);
	  lat = lat->cdr;
	}
    }
}

/* ===================================================================== 
 * Management of mkf files
 *
 */
static boolean getNextMKFReading()
{
  if (curRefFileObject == lastRefFileObject)
    return pk_false;
  curRefFileObject = curRefFileObject->cdr;
  curFileObject = curRefFileObject->car;
  return pk_true;
}

/* ===================================================================== 
 * 
 *
 */

/*
 * try to insert a file or variable name definition
 *        isvar pk_true if the label denotes a variable
 *        isvar pk_false, isFile pk_true if the label denotes a file designation
 *        isvar pk_false, isFile false if the label denotes a blind  designation
 * success: returns pk_true
 *          creates or gets an atom
 *          set curBuiltObject
 *          set isInclude = pk_true if the variable denotes an existing definition
 *               of a file to be included (overreferencing allowed)
 * failure: returns false on conflict
 *          emits error message
 */
static boolean getDefObject(boolean isvar, boolean isFile, boolean * isInclude)
{
  int hashi;

  if (findObject(isvar, isFile, &hashi, &curBuiltObject))
    {				/* curBuiltObject set */
      if (curBuiltObject->inError)
	return pk_false;
      if (curBuiltObject->defined)
	{
	  if (curBuiltObject->include)
	    {
	      *isInclude = pk_true;
	      return pk_true;
	    }
	  if (conflictError(doubleErrMess, curBuiltObject)) curBuiltObject->inError = pk_true;
	  return pk_false;
	}
      curBuiltObject->defined = pk_true;
      *isInclude = pk_false;
      setObjectOwner(curBuiltObject, curFileObject);
      if (curBuiltObject->internal)
	curBuiltObject->created = curFileObject->created;
      return pk_true;
    }				/* first character following an object closing */
  *isInclude = pk_false;
  curBuiltObject = newObject(strdupAtom(), isvar, isFile, pk_true);
  setObjectOwner(curBuiltObject, curFileObject);
  registerObject(curBuiltObject, hashi);
  return pk_true;
}

/*
 * try to read and insert a list of action lists associated to a file or variable 
 * beginning on next line
 * terminates on eof or at the beginning of a MKF tagged line
 * creates and links actions
 */
static ref2ActionLstType getActionLst()
{
  ref2ActionLstType prevActionLst, actionLstHead;

  if (!(fileNextLine()))
    return NIL_ActionLstType;
  actionLstHead = NIL_ActionLstType;
  while (!((isEof()) || (isMKFTag())))
    {
      if (actionLstHead == NIL_ActionLstType)
	actionLstHead = prevActionLst = newActionLst(appendLine());
      else
	prevActionLst = appendActionLst(prevActionLst, appendLine());
    }
  return actionLstHead;
}

/*
 * try to insert a file or variable name reference to
 *        isvar pk_true if the label denotes a variable
 *        isvar pk_false, isFile pk_true if the label denotes a file designation
 *        isvar pk_false, isFile false if the label denotes a blind  designation
 * success: return pk_true
 *          creates or gets an atom
 *          set curRefObject
 * failure: return false
 *          if found object is in error
 */
static boolean getRefObject(boolean isvar, boolean isFile)
{
  int hashi;

  if (findObject(isvar, isFile, &hashi, &curRefObject))
    {				/* curBuiltObject set */
      if (curRefObject->inError)
	return pk_false;
    }
  else
    {
      curRefObject = newObject(strdupAtom(), isvar, isFile, pk_false);
      registerObject(curRefObject, hashi);
    }
  return pk_true;
}

/*
 * get a new reference headin in the current line and find or creates an object describing it
 * success: set curRefObject pointing to this object
 *          set current character immediately after the atom
 *          set success to false if there is a type conflict
 * failure: return false if the current line is not a correct predecessor line
 *          set success to false if failure is due to syntax error
 */
static boolean getLineReference(boolean * success)
{
  boolean isvar, isFile;

  if (!(lineNextPredecessorTag()))
    return pk_false;
  if (readObjectName(&isvar, &isFile))
    {				/*current character immediately after the atom */
      if (!(getRefObject(isvar, isFile)))
	{
	  TRACE(fprintf(stderr, "getLineReference 1 curRefObject->objectLabel=%s, created=%i\n", curRefObject->objectLabel, curRefObject->created);) * success = pk_false;
	}
      return pk_true;
    }
  *success = pk_false;
  if (isEof())
    {
      if (syntaxError((char *)"Unexpected eof")) ;
    }
  else
    {
      if (syntaxError((char *)"Unwaited character")) ;
    }
  return pk_false;
}

/*
 * read prelude action if current line is a prelude opening action
 * success: currentchar is the first character of the line beginning inmmediately after the prelude, or Eof
 *          curBuiltObject->actions->prelConsAction  is set
 * failure:
 */
static void getPrelude()
{
  if (isOpenPrelude())
    curBuiltObject->actions->prelConsAction = getActionLst();
  else
    curBuiltObject->actions->prelConsAction = NIL_ActionLstType;
}

/*
 * read a new ref2PredecessorLstType predecessor beginning with the current line, and returns it in newPredActions
 * success: currentchar is the first character of the line beginning inmmediately after the predecessor, or Eof
 *          newPredActions contains the (predecessor,action) element list
 *          newPredObject contains the predecessor element list
 *          curBuiltObject->lastPred is set
 *          let success unchanged
 * failure: return pk_false;
 *          set success to false if failure is due to syntax error
 *          newPredActions, newPredObject remain unchanged on syntax error
 */
static boolean getPredecessor(ref2PredecessorLstType * newPredActions, ref2LstRefObjType * newPredObject, boolean * success)
{
  boolean neverBuild = pk_false;
  if (!(getLineReference(success)))
    return pk_false;
  *newPredObject = newLstRefObj(curRefObject);

  if (lineNextNonSpace())
    {
      if (isDonotMark())
	{
	  neverBuild = pk_true;
	  if (lineNextChar())
	    {
	      if (lineNextNonSpace()) ;
	    }
	  else
	    {
	      if (syntaxError((char *)"Unexpected eof")) ;
	      return pk_false;
	    }
	}
    }
  if (isCommentTag())
    {
      if (skipTopKEoln()) ;
    }
  else if (!isEol())
    {
      if (syntaxError((char *)"Unexpected character")) ;
      return pk_false;
    }

  *newPredActions = newPredecessorLst(curRefObject, neverBuild);
  (*newPredActions)->postConsAction = getActionLst();	// terminates on openTag or closeTag or Eof()

  return *success;
}

/*
 * read a new ref2PredecessorLstType list of predecessora beginning with the current line,
 * assumes that curBuiltObject, curBuiltObject->actions are properly set
 * success: currentchar is the first character of the line beginning inmmediately after the predecessor list
 *          curBuiltObject->actions->innerLst contains the (predecessor,action) list
 *          curBuiltObject->firstPred and curBuiltObject->lastPred are set
 * 
 */
static boolean getPredecessors()
{
  boolean success = pk_true;
  ref2PredecessorLstType prevPredActions = NIL_PredecessorLstType;
  ref2PredecessorLstType newPredActions = NIL_PredecessorLstType;
  ref2LstRefObjType newPredObject = NIL_LstRefObjType;
  ref2LstRefObjType prevPredObject = NIL_LstRefObjType;

  if (!(getPredecessor(&prevPredActions, &prevPredObject, &success)))
    return success;

  curBuiltObject->firstPred = prevPredObject;
  curBuiltObject->actions->innerLst = prevPredActions;
  while (getPredecessor(&newPredActions, &newPredObject, &success))
    {
      prevPredActions->cdr = newPredActions;
      prevPredObject->cdr = newPredObject;
      prevPredActions = newPredActions;
      prevPredObject = newPredObject;
    }
  prevPredActions->cdr = NIL_PredecessorLstType;
  prevPredObject = NIL_LstRefObjType;
  curBuiltObject->lastPred = newPredObject;
  return success;
}

static void getDefaultAction()
{
  if (isOpenDefault())
    curBuiltObject->actions->noconsAction = getActionLst();	// terminates on openTag or closeTag or Eof()
  else
    curBuiltObject->actions->noconsAction = NIL_ActionLstType;

}

/*
 * get a new definition head and find or creates an object describing it
 * current character is examined
 * set curBuiltObject pointing to this object
 * set current character immediately after the atom
 * success: returns pk_true
 *          set isInclude = pk_true if the variable denotes an existing definition
 *               of a file to be included (overreferencing allowed)
 *          set isvar pk_true if the label denotes a variable
 *          set isvar pk_false, otherwise
 *          
 */
static boolean skip2nextDef(boolean * isvar, boolean * isInclude)
{
  boolean isFile;
  while ((skipafterNewDefTag()))
    {
      if (readObjectName(isvar, &isFile))
	{			/*current character immediately after the atom */
	  if (getDefObject(*isvar, isFile, isInclude))
	    return pk_true;
	}
      else if (isEof())
	{
	  if (syntaxError((char *)"Unwaited eof")) ;
	  return pk_false;
	}
    }
  return pk_false;
}

/*
 * input
 *       fileObject: a file object
 *       isNewFAllowed pk_true if file is not required to exist
 * set fileObject->objectAlias to fileName expansion, 
 * success:
 *    fileObject->created is set to file date or apocalypse if file doesn't exist
 */
static boolean setFileProps(ref2ObjectType fileObject, boolean isNewFAllowed)
{
  char *fileName;
  boolean newf;

  if (fileObject->isFile)
    {
      fileName = (convertParemeterizedName(fileObject->objectLabel));
      fileObject->objectAlias = fileName;
      fileObject->created = getFDate(fileName, &newf, curErrMess);
      TRACE(fprintf(stderr, "%s birthday is %i \n", fileName, fileObject->created);) if (fileObject->created == NIL_time)
	{
	  logPrintErrObjMess(fileObject, *curErrMess);
	  fileObject->inError = pk_true;
	  return pk_false;
	}
      if (newf && !(isNewFAllowed))
	{
	  logPrintErrObjMess(fileObject, (char *)" missing");
	  fileObject->inError = pk_true;
	  return pk_false;
	}
      return pk_true;
    }
  return pk_true;
}

/*
 * set curFileObject->made = itself, curFileObject-> = pk_true
 */
static boolean openMakeFile()
{
  curFileObject->made = curFileObject;
  if (setFileProps(curFileObject, pk_false))
    return (openReadMKF(curFileObject->objectAlias));
  else
    return pk_false;
}

static boolean openwf(ref2ObjectType curObject)
{
  if (curObject->objectAlias == NULL)
    return pk_false;
  currentWriteFile = wFOpen(curObject->objectAlias, curErrMess);
  if (currentWriteFile == NIL_File)
    {
      logPrintErrObjMess(curObject, *curErrMess);
      return pk_false;
    }
  return pk_true;
}

/*
 * set undefined objects in error
 * set undefined missing files in error
 *
 */
static void checkUndefinedList(void)
{
  boolean newf, printTitle = pk_true, warToPrint = pk_false;
  register int i;
  ref2LstRefObjType lat;
  for (i = 1; i <= hashsize; i++)
    {
      lat = hashEntries[i - 1];
      while (lat != NIL_LstRefObjType)
	{
	  ref2ObjectType thisObject = lat->car;
	  if (!(thisObject->defined))
	    if (thisObject->internal)
	      {
		thisObject->inError = pk_true;
		logPrintErrObj(thisObject);
	      }
	    else if (thisObject->created == NIL_time)
	      {
		if (setFileProps(thisObject, pk_false))
		  warToPrint = pk_true;
	      }
	    else
	      warToPrint = pk_true;
	  if (verbose && warToPrint)
	    {
	      warToPrint = pk_false;
	      if (printTitle)
		{
		  logPrintWarTitle(nodefTitle);
		  printTitle = pk_false;
		}
	      logPrintObjectRefDateLine(thisObject);
	    }
	  lat = lat->cdr;
	}
    }
}

/*
 * checks that system file dependences may be properly stated
 * print object reference to
 * set cyclic to pk_true if a cyclic dependence is found
 * set obj->inError to pk_true if an error occurs
 * set obj->stated to pk_true
 * returns false iff the object is in error
 */
static boolean commonGraphChecking(ref2ObjectType obj, boolean * cyclic)
{
  if ((obj->created) == genesia)
    {
      /* cycle */
      obj->inError = pk_true;
      logPrintErrCycle(obj);
      *cyclic = pk_true;
      return pk_false;
    }
  if (obj->stated)
    {
      if (verbose)
	logPrintObjectRefDate(obj);
      return (!(obj->inError));
    }				/* building this object is yet decided */
  if (obj->inError)
    {
      logPrintErrObj(obj);
      return pk_false;
    }
  if (verbose)
    logPrintObjectRefDate(obj);
  return pk_true;

}

static void checkSystemFile(ref2ObjectType obj, boolean * cyclic);

/*
 * set obj->inError to pk_true if some error occurs
 */
static void chekSystemFilePred(ref2ObjectType obj)
{
  time_t birthday;
  boolean cyclic = pk_false, isCorrect = pk_true;
  ref2PredecessorLstType antelst, prev;

  TRACE(fprintf(stderr, "checkSystemFilePred 1 obj->objectLabel=%s, created=%i\n", obj->objectLabel, obj->created);) if (verbose)
    logPrintLine();
  imbric = imbric + 1;
  birthday = (obj->created);
  (obj->created) = genesia;	/* cycle detection */
  prev = NIL_PredecessorLstType;
  antelst = obj->actions->innerLst;
  while (antelst != NIL_PredecessorLstType)
    {
      checkSystemFile(antelst->car, &cyclic);
      if (cyclic)
	{			/*remove faulty dependence */
	  if (prev == NIL_PredecessorLstType)
	    obj->actions->innerLst = antelst->cdr;
	  else
	    prev->cdr = antelst->cdr;
	  TRACE(fprintf(stderr, "chekSystemFilePred cyclic antelst->car=%s\n", antelst->car->objectLabel);) obj->inError = pk_true;
	  cyclic = pk_false;
	}
      else if (antelst->car->inError)
	{
	  obj->inError = pk_true;
	  TRACE(fprintf(stderr, "chekSystemFilePred not correct antelst->car=%s\n", antelst->car->objectLabel);) isCorrect = pk_false;
	}
      if (verbose)
	logPrintLine();
      prev = antelst;
      antelst = antelst->cdr;
    }
  (obj->created) = birthday;
  TRACE(fprintf(stderr, "checkSystemFilePred 2 obj->objectLabel=%s, created=%i\n", obj->objectLabel, obj->created);) imbric = imbric - 1;
}

/*
 * checks that system file dependences are properly stated
 * set cyclic to pk_true if a cyclic dependence is found
 * set obj->inError to pk_true if an error occurs
 * set obj->stated to pk_true
 * actions are not taken into account
 */
static void checkSystemFile(ref2ObjectType obj, boolean * cyclic)
{
  ref2ActionType ptObjActions;

  TRACE(fprintf(stderr, "checkSystemFile 1 obj->objectLabel=%s, created=%i\n", obj->objectLabel, obj->created);) if (!(commonGraphChecking(obj, cyclic)))
    return;
  obj->stated = pk_true;
  /* from this point obj is neither stated nor inError */
  if (verbose)
    logPrintObjectRefDate(obj);
  if (obj->isFile)
    {
      logPrintErr(pk_false, (char *)"cannot precedes a system file");
      obj->inError = pk_true;
      return;
    }
  ptObjActions = obj->actions;
  if (ptObjActions == NIL_ActionType)
    return;
  if (ptObjActions->innerLst == NIL_PredecessorLstType)
    return;
  chekSystemFilePred(obj);
  obj->made = obj;
TRACE(fprintf(stderr, (char *)"checkSystemFile 2 obj->objectLabel=%s, created=%i\n", obj->objectLabel, obj->created);)}

static boolean makeit(time_t contextTime, ref2ObjectType obj, boolean * cyclic);

/*
 * assumes objects and missing files are either in error or defined 
 * returns pk_true if some internal object is younger or is to be built
 * context_time is the time of the file which building can be decided when looking at current object
 * returns false if all internal objects remain older or are inError
 * set obj->inError to pk_true if some error occurs
 */
static boolean makeactl(time_t contextTime, ref2ObjectType obj)
{
  time_t birthday, sContextTime;
  boolean cyclic = pk_false, isCorrect = pk_true, todo = pk_false;
  boolean isSonStd, sonToDo;
  ref2PredecessorLstType antelst, prev;

  TRACE(fprintf(stderr, "makeactl 0 obj=%s, created=%i\n", obj->objectLabel, obj->created);) if (verbose)
    logPrintLine();
  imbric = imbric + 1;

  TRACE(fprintf(stderr, "       makeactl, name=%s\n", obj->objectLabel);) if (obj->isFile)
    sContextTime = obj->created;
  else
    sContextTime = contextTime;
  birthday = (obj->created);
  (obj->created) = genesia;	/* cycle detection */
  prev = NIL_PredecessorLstType;
  antelst = obj->actions->innerLst;
  while (antelst != NIL_PredecessorLstType)
    {
      TRACE(fprintf(stderr, "     makeactl son 1 antelst->label=%s, antelst->created=%i\n", antelst->car->objectLabel, antelst->car->created);) isSonStd = antelst->car->isFile
	  || antelst->car->internal;
      if (isSonStd)
	{
	  sonToDo = makeit(sContextTime, antelst->car, &cyclic);
	  if (sonToDo)
	    if (!(todo))
	      {
		todo = pk_true;
		birthday = now;
	      }
	}
      else
	{
	  checkSystemFile(antelst->car, &cyclic);
	  sonToDo = pk_false;
	}
      if (!sonToDo)
	{
	  if (cyclic)
	    {			/*remove faulty dependence */
	      if (prev == NIL_PredecessorLstType)
		obj->actions->innerLst = antelst->cdr;
	      else
		prev->cdr = antelst->cdr;
	      TRACE(fprintf(stderr, "makeactl cyclic obj=%s\n", obj->objectLabel);) obj->inError = pk_true;
	      cyclic = pk_false;
	    }
	  else if (antelst->car->inError)
	    {
	      obj->inError = pk_true;
	      TRACE(fprintf(stderr, "makeactl not correct obj=%s\n", obj->objectLabel);) isCorrect = pk_false;
	    }
	  else if (obj->internal)
	    {
	      TRACE(fprintf(stderr, "makeactl son 1.1 antelst->created=%i, birthday =%i\n", antelst->car->created, birthday);) if (birthday < antelst->car->created)
		birthday = antelst->car->created;
	    }
	  else if (isSonStd && (birthday < antelst->car->created))
	    {
	      birthday = now;
	      todo = pk_true;
	    }
	}
      if (verbose)
	logPrintLine();
      TRACE(fprintf(stderr, "makeactl son 2 antelst->label=%s, antelst->created=%i\n", antelst->car->objectLabel, antelst->car->created);) prev = antelst;
      antelst = antelst->cdr;
    }
  TRACE(fprintf(stderr, "       makeactl 2, name=%s\n", obj->objectLabel);) obj->created = birthday;
  TRACE(fprintf(stderr, "makeactl 2 antelst->car=%s, created=%i\n", obj->objectLabel, obj->created);) imbric = imbric - 1;
  return (isCorrect && todo);
}

/*
 * obj is internal or is a file (not a system file)
 * assumes objects and missing files being either in error or defined 
 * context_time is the time of the file which building can be decided when looking at current object
 * return pk_true if this file or internal object has to be built
 * return false if object is up-to-date or inError, or if it cannot be built
 *              set cyclic to pk_true if a cyclic dependence is found
 *              set obj->inError to pk_true
 *              set obj->made to itself if obj is a file
 * set obj->stated to pk_true
 * complete the well ordered list of files to do
 */
static boolean makeit(time_t contextTime, ref2ObjectType obj, boolean * cyclic)
{
  ref2ActionType ptObjActions;
  TRACE(fprintf(stderr, "makeit 0, name=%s, date =%i \n", obj->objectLabel, obj->created);) if (!(commonGraphChecking(obj, cyclic)))
    return pk_false;
  /* obj is not in error */
  if (obj->stated)
    {
      TRACE(fprintf(stderr, "makeit stated, name=%s, date =%i \n", obj->objectLabel, obj->created);) if (contextTime < (obj->created))
	if (verbose)
	  logPrintString((char *)" +++ younger");
      return ((obj->created) == now);
    }				/* building this object is yet decided */
  /* obj is not stated */
  obj->stated = pk_true;
  if (obj->include)
    {
      if (contextTime < (obj->created))
	if (verbose)
	  logPrintString((char *)" +++ younger");
      return pk_false;
    }

  TRACE(fprintf(stderr, "makeit 1, name=%s, date=%i\n", obj->objectLabel, obj->created);) if (!(obj->defined))
    {
      /* obj is then an existing file */
      if (contextTime < (obj->created))
	if (verbose)
	  logPrintString((char *)" +++ younger");
      return pk_false;
    }

  /* obj has a constructive definition */
  ptObjActions = obj->actions;
  if (ptObjActions == NIL_ActionType)
    {
      if (contextTime <= (obj->created))
	if (verbose)
	  logPrintString((char *)" +++ younger");
      return pk_false;
    }
  TRACE(fprintf(stderr, "makeit 2, name=%s, date=%i\n", obj->objectLabel, obj->created);) if (ptObjActions->innerLst == NIL_PredecessorLstType)
    {
      if (contextTime <= (obj->created))
	if (verbose)
	  logPrintString((char *)" +++ younger");
      return pk_false;
    }
  /* obj has a non empty predecessor list */
  TRACE(fprintf(stderr, "makeit 1, name=%s\n", obj->objectLabel);) if (makeactl(contextTime, obj))
    {
      TRACE(fprintf(stderr, "makeit makeactl pk_true, name=%s, date=%i\n", obj->objectLabel, obj->created);)
	  /* obj has a younger predecessor */
	  if (verbose)
	logPrintBuildingObj(obj);
      if (obj->isFile)
	appendFileConstruction(obj);
      return pk_true;
    }
  /* obj predecessors are not younger */
  TRACE(fprintf(stderr, "makeit makeactl false 1, name=%s, date=%i\n", obj->objectLabel, obj->created);) if (obj->inError)
    {
      logPrintErrObj(obj);
      return pk_false;
    }
  TRACE(fprintf(stderr, "makeit makeactl false 2, name=%s, date=%i\n", obj->objectLabel, obj->created);) if (obj->isFile)
    {
      TRACE(fprintf(stderr, "makeit isfile false 2, name=%s, date=%i\n", obj->objectLabel, obj->created);) if (obj->created == apocalypse)
	{
	  TRACE(fprintf(stderr, "makeit has apocalypse, name=%s, date=%i\n", obj->objectLabel, obj->created);) obj->created = now;
	  appendFileConstruction(obj);
	  return pk_true;
	}
      TRACE(fprintf(stderr, "makeit doesnot have apocalypse, name=%s, date=%i\n", obj->objectLabel, obj->created);) if (obj->actions->conditional)
	{
	  appendFileConstruction(obj);
	  return pk_false;
	}
      obj->made = obj;
    }
  TRACE(fprintf(stderr, "makeit is false for obj->objectLabel=%s\n", obj->objectLabel);) return pk_false;
}

/*
 * builds the well ordered list of files to be built from firstDirectFile
 * firstDirectFile contains only located files (neither internal objects nor system files)
 * each of them is marked visited and not made
 * construction root is the direct file list
 * effective list of files to build is firstFileConstruction
 */
static boolean dependenceChecking()
{
  boolean correct;
  boolean cyclic;
  ref2LstRefObjType curRefObjectLst;
  ref2ObjectType curObject;

  correct = pk_true;
  curRefObjectLst = firstDirectFile;
  logPrintInfoTitle(verbose, graphTitle);
  while (curRefObjectLst != NIL_LstRefObjType)
    {
      curObject = curRefObjectLst->car;
      if (!(makeit(apocalypse, curObject, &cyclic)))
	{
	  if (curObject->inError)
	    {
	      TRACE(fprintf(stderr, "dependenceChecking not correct curObject=%s\n", curObject->objectLabel);) correct = pk_false;
	    }
	}
      if (verbose)
	logPrintLine();
      curRefObjectLst = curRefObjectLst->cdr;
    }
  return correct;
}				/* dependenceChecking */

static void printDirectFileLineList(ref2ActionLstType ll)
{
  int res;
  while (ll != NIL_ActionLstType)
    {
      res = fputs(ll->car, currentWriteFile);
      ll = ll->cdr;
    }
}

/*
 * Appends default action of obj to the embedding object
 *     
 */
static void consDefaultObject(ref2ObjectType obj)
{
  ref2ActionType ptObjActions;

  if (obj->inError)
    logPrintErr(pk_true, (char *)" Damned! current file is in Error ");
  ptObjActions = obj->actions;
  if (ptObjActions == NIL_ActionType)
    return;
  printDirectFileLineList(ptObjActions->noconsAction);
}

static void consObject(ref2ObjectType obj, ref2ObjectType curBuiltFile);

/*
 * Internal object
 *     an internal object can be built several times (one for each built file)
 *     obj is the object to build
 *     curBuiltFile is the nearest file that that depends on obj
 *     
 */
static void consActions(ref2ActionType ptObjActions, ref2ObjectType curBuiltFile)
{
  ref2PredecessorLstType innerLst;

  printDirectFileLineList(ptObjActions->prelConsAction);
  innerLst = ptObjActions->innerLst;
  if (innerLst == NIL_PredecessorLstType)
    return;
  imbric = imbric + 1;
  while (innerLst != NIL_PredecessorLstType)
    {
      if (innerLst->neverBuild)
	consDefaultObject(innerLst->car);
      else
	consObject(innerLst->car, curBuiltFile);
      printDirectFileLineList(innerLst->postConsAction);
      innerLst = innerLst->cdr;
    }
  imbric = imbric - 1;
}

/*
 * Internal object
 *     an internal object can be built several times (one for each built file)
 *     obj is the object to build
 *     curBuiltFile is the nearest file that that depends on obj
 *     
 */
static void consObject(ref2ObjectType obj, ref2ObjectType curBuiltFile)
{
  ref2ActionType ptObjActions;

  if (obj->inError)
    logPrintErr(pk_true, (char *)" Damned! current file is in Error ");
  ptObjActions = obj->actions;
  if (ptObjActions == NIL_ActionType)
    return;
  if (verbose)
    logPrintBuildingObj(obj);
  if (obj->internal)
    {
      if ((obj->made) == curBuiltFile)
	printDirectFileLineList(ptObjActions->noconsAction);
      else
	{
	  obj->made = curBuiltFile;
	  if ((ptObjActions->conditional) && (obj->created != now))
	    printDirectFileLineList(ptObjActions->noconsAction);
	  else
	    consActions(ptObjActions, curBuiltFile);
	}
    }
  else
    {
      if ((obj->made) == obj)
	printDirectFileLineList(ptObjActions->noconsAction);
      else
	{
	  obj->made = obj;
	  if ((ptObjActions->conditional) && (obj->created != now))
	    {
	      printDirectFileLineList(ptObjActions->noconsAction);
	    }
	  else
	    {
	      consActions(ptObjActions, obj);
	    }
	}
    }
}

/*
 * built the files registered in 
 */
static void buildfile()
{
}

/*
 * inserts the current filename as a make file to be read
 * skip after the next closeTag
 * success: returns pk_true
 * failure to skip: returns false
 */
static void linkNewMKF()
{
  curBuiltObject->defined = pk_true;
  curBuiltObject->include = pk_true;
  appendMKFReading(curBuiltObject);
}

/*
 * build actions of the curBuiltObject
 * try to skip after the end of this object
 * success: returns pk_true
 * failure to skip: returns false
 */
static boolean buildAllActions(boolean handmade, boolean direct, boolean conditional)
{				/* new object definition */

  boolean success;
  curBuiltObject->include = pk_false;
  curBuiltObject->actions = newAction(handmade, direct, conditional);
  getPrelude();
  success = getPredecessors();
  getDefaultAction();
  return success;
}

/*
 * sucess: set c to the non space character that immediately follows the current closeTag
 *          returns pk_true
 * failure: let the buffer unchanged or on Eof()
 *          returns false
 */
static boolean buildHandmade()
{
  if (curBuiltObject->internal)
    {
      if (conflictError(conflictErrMess, curBuiltObject))      return pk_false;
    }				/* first character of a new object */
  if (!(fileNextNonSpaceAfter()))
    {
      if (syntaxError((char *)"Unwaited eof")) ;
      return pk_false;
    }				/* first character of a new object */
  return buildAllActions(pk_true, pk_false, pk_true);
}

/*
 * input
 *   forced is pk_true if building must be done whatever the date is
 * sucess: set c to the non space character that immediately follows the current writeFileTag
 *          links curObject in RefObject list
 *          returns pk_true
 * failure: let the buffer unchanged or on Eof()
 *          returns false
 */
static boolean buildDirectFile()
{
  if (curBuiltObject->internal)
    {
      if (conflictError(conflictErrMess, curBuiltObject)) return pk_false;
    }				/* first character of a new object */
  if (!(fileNextNonSpaceAfter()))
    {
      if (syntaxError((char *)"Unwaited eof")) ;
      return pk_false;
    }				/* first character of a new object */
  if (isForcedMark())
    {
      if (!(fileNextNonSpaceAfter()))
	{
	  if (syntaxError((char *)"Unwaited eof")) ;
	  return pk_false;
	}			/* first character of a new object */
      registerDirectFile(curBuiltObject, pk_true);
      return buildAllActions(pk_false, pk_true, pk_false);
    }
  registerDirectFile(curBuiltObject, pk_false);
  if (isDonotMark())
    {
      if (!(fileNextNonSpaceAfter()))
	{
	  if (syntaxError((char *)"Unwaited eof")) ;
	  return pk_false;
	}
      return buildAllActions(pk_false, pk_true, pk_true);
    }
  return buildAllActions(pk_false, pk_true, pk_false);
}

/*
 * sucess: set c to the non space character that immediately follows the current writeFileTag
 *          links curObject in RefObject list
 *          returns pk_true
 * failure: let the buffer unchanged or on Eof()
 *          returns false
 */
static boolean buildStandardObject(boolean optional)
{
  if (!(fileNextNonSpace()))
    {
      if (syntaxError((char *)"Unwaited eof")) ;
      return pk_false;
    }				/* first character of a new object */
  return buildAllActions(pk_false, pk_false, optional);
}

static boolean readObject(boolean isvar, boolean isInclude)
{
  if (isvar)
    curBuiltObject->created = curFileObject->created;
  else if (curBuiltObject->isFile)
    if (setFileProps(curBuiltObject, pk_true)) ;	/* something to do if not curBuiltObject->isFile ? */
  if (lineNextNonSpace())
    {
      TRACE(fprintf(stderr, "true curBuiltObject->objectLabel=%s\n", curBuiltObject->objectLabel);) if (isCommentTag())
	{
	  if (skipTopKEoln()) ;
	  if (isInclude)
	    {
	      if (conflictError(doubleErrMess, curBuiltObject)) return pk_false;
	    }
	  return buildStandardObject(pk_false);
	}
      if (isIncludeMark())
	{			/* makefile to be loaded */
	  if (isInclude)
	    conflictWarning(doubleErrMess, curBuiltObject);
	  else
	    linkNewMKF();
	  return pk_true;
	}			/* c after the object label */
      if (isInclude)
	{
	  if (conflictError(doubleErrMess, curBuiltObject)) return pk_false;
	}
      if (isLoadMark())		/* File not built using this makefile */
	return buildHandmade();
      if (isDirectMark())
	return buildDirectFile();
      if (isDonotMark())
	{
	  if (!(fileNextNonSpaceAfter()))
	    {
	      if (syntaxError((char *)"Unexpected eof")) ;
	      return pk_false;
	    }
	  return buildAllActions(pk_false, pk_true, pk_true);
	}
      if (syntaxError((char *)"Unexpected character")) ;
      return pk_false;
    }
  /* eol or eof */
  return buildStandardObject(pk_false);
}

static boolean readMKF()
{
  boolean res = pk_true, isvar, isInclude;
  if (verbose)
    logPrintReadMKF(curFileObject);
  while (skip2nextDef(&isvar, &isInclude))
    {
      if (readObject(isvar, isInclude))
	{
	  /* current character set to first character after the current line */
	  if (!(lineNextCharAfterClosing()))
	    {
	      if (syntaxError((char *)"Close object tag missing")) ;
	    }
	}
      else
	{
	  TRACE(fprintf(stderr, "readMKF not correct curBuiltObject=%s\n", curBuiltObject->objectLabel);) curBuiltObject->inError = pk_true;
	  res = pk_false;
	}
      if (verbose)
	printObject(curBuiltObject);
    }
  if (!(res))
    {
      curFileObject->inError = pk_true;
    }
  return res;
}

static boolean readAllMkf()
{
  boolean res;
  objectNumber = 0;
  curFileObject = createStdinObject();
  curRefFileObject = initMKFReading(curFileObject);
  if (!(openStdin()))
    return pk_false;
  res = readMKF();
  closeStdin();
  while (getNextMKFReading())
    {
      if (openMakeFile())
	{
	  if (!(readMKF()))
	    res = pk_false;
	  closeMKF();
	}
      else
	res = pk_false;
    }
  return res;
}

static void printUnreferred()
{
  int i;
  ref2LstRefObjType lat;
  boolean titre;

  titre = pk_true;
  if (hashsize > 0)
    for (i = 1; i <= hashsize; i++)
      {
	lat = hashEntries[i - 1];
	while (lat != NIL_LstRefObjType)
	  {
	    if (!((lat->car->stated) || (lat->car->include)))
	      {
		if (titre)
		  {
		    titre = pk_false;
		    logPrintWarTitle(norefTitle);
		  }
		logPrintObjectRef(lat->car);
		logPrintLine();
	      }
	    lat = lat->cdr;
	  }
      }
}

static void printTableStatistics(boolean byLine)
{
  int i, count, total = 0;
  int smallest = objectNumber;
  int greatest = 0;
  ref2LstRefObjType lat;

  for (i = 0; i < hashsize; i++)
    {
      count = 0;
      lat = hashEntries[i];
      while (lat != NIL_LstRefObjType)
	{
	  count = count + 1;
	  lat = lat->cdr;
	}
      if (byLine)
	{
	  fprintf(stderr, "table entry %i has %i objects\n", i, count);
	  fflush(stderr);
	}
      if (count < smallest)
	smallest = count;
      if (count > greatest)
	greatest = count;
      total = total + count;
    }
}

/* 
 * Generates eligible files registered in curRefObjectLst (up to a nil pointer)
 * 
 */
static boolean generateAllFiles(ref2LstRefObjType curRefObjectLst)
{
  ref2ObjectType curObject;

  boolean correct = pk_true;
  while (curRefObjectLst != NIL_LstRefObjType)
    {
      curObject = curRefObjectLst->car;
      if (curObject->actions != NIL_ActionType)
	if (curObject->actions->direct)
	  {
	    logPrintInfoObjTitle(verbose, buildTitle, curObject);
	    if (openwf(curObject))
	      {
		consObject(curObject, curObject);
		rwFClose(currentWriteFile);
	      }
	    else
	      correct = pk_false;
	  }
      curRefObjectLst = curRefObjectLst->cdr;
    }
  return correct;
}

int main(int argc, char *argv[])
{
  boolean correct = pk_true;
  /* add exec tag and execution of exec tagged files */
  //  system("ls -l $utl_root");
  if (argc > 1)
    {
      if (strcmp("-v", argv[1]) == 0)
	verbose = pk_true;
      else
	verbose = pk_false;
    }
  curErrMess = (ref2String) malloc((unsigned)(sizeof(*curErrMess)));
  now = initDate(&apocalypse);
  objectNumber = 0;
  initObjectStorage();
  initFileSystemEnv();
  initGlobalReader();
  initPrinter();
  initMakeGraph();
  logPrintInfoTitle(verbose, readTitle);
  correct = readAllMkf();
  checkUndefinedList();
  if (verbose)
    printDirectFileList();
  if (verbose)
    printTableStatistics(pk_false);
  if (!correct)
    {
      logPrintInfoTitle(verbose, (char *)" Error during file reading: exit before graph building ");
      exit(-1);
    }
  else if (lastDirectFile == NIL_LstRefObjType)
    {
      logPrintInfoTitle(verbose, (char *)" No direct file to generate: exit before graph building ");
      exit(-1);
    }
  else
    {
      correct = dependenceChecking();
      printUnreferred();
      if (verbose)
	printConstructionFileList();
      TRACE(fprintf(stderr, "main 0 firstFile=%s\n", firstFileConstruction->car->objectLabel);) if (correct)
	{
	  TRACE(fprintf(stderr, "main 1 firstFile=%s\n", firstFileConstruction->car->objectLabel);) if (generateAllFiles(firstFileConstruction))
	    {
	      TRACE(fprintf(stderr, "main 1 firstFile=%s\n", firstFileConstruction->car->objectLabel);) logPrintInfoTitle(verbose, doneTitle);
	      exit(0);
	    }
	}
    }
  logPrintInfoTitle(verbose, failTitle);
  exit(-1);
}
